#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
// #define DEBUG

#include "FileSystemInterface.h"
#include "RecordStream.h"
#include "FixedRecordFile.h"
#include "FixedRecordStream.h"
#include "VariableRecordFile.h"
#include "VariableRecordStream.h"
#include "PersistentArray.h"
#include "SymbolicLink.h"
#include "SymbolicLinkStar.h"
#include "SystemInterface.h"
#include "PersistentObject.h"
#include "MemoryObjectDictionary.h"
#include "MemoryObjectDictionaryStar.h"
#include "Class.h"
#include "OutputStream.h"
#include "OutputStreamStar.h"
#include "UNIXFile.h"
#include "FileSystem.h"
#include "Convert.h"
#include "String.h"
#include "Tools.h"
#include "NameServer.h"

extern char HOME[];
extern FileSystemInterface * FSI;
extern ClassStar ClassClass;
extern ClassStar RecordStreamClass;
extern PersistentClassStar SymbolicLinkClass;
extern PersistentClassStar MemoryObjectDictionaryClass;

/*
 *	modefmt	- convert an inodes mode filed to a printable format 
 *
 *		mode - a UNIX inodes mode field 
 *		buf - an array of at least 11 characters
 */		
void
modefmt( int mode, char * buf )
{
	/*
	 *	Break the mode filed into its four subfields
	 */
	int type   = mode & UNIXTypeMask;
	int owner  = mode & UNIXOwnerMask;
	int group  = ( mode & UNIXGroupMask  ) << 3;
	int publc  = ( mode & UNIXPublicMask ) << 6;

	/*
	 *	Convert the file type subfield.
	 */
	if     ( UNIXTypeBlock        == type ) buf[0] = 'b';
	else if( UNIXTypeDirectory    == type ) buf[0] = 'd';
	else if( UNIXTypeCharacter    == type ) buf[0] = 'c';
	else if( UNIXTypeSymbolicLink == type ) buf[0] = 'l';
	else if( UNIXTypeSocket       == type ) buf[0] = 's';
	else if( UNIXTypeRegular      == type ) buf[0] = '-';
	else buf[0] = '?';

	/*
	 *	Convert the owner permissions subfield.
	 */
	if( UNIXReadMask    & owner ) buf[1] = 'r';	else buf[1] = '-';
	if( UNIXWriteMask   & owner ) buf[2] = 'w';	else buf[2] = '-';
	if( UNIXExecuteMask & owner ) buf[3] = 'x';	else buf[3] = '-';

	/*
	 *	Convert the group permissions subfield.
	 */
	if( UNIXReadMask    & group ) buf[4] = 'r';	else buf[4] = '-';
	if( UNIXWriteMask   & group ) buf[5] = 'w';	else buf[5] = '-';
	if( UNIXExecuteMask & group ) buf[6] = 'x';	else buf[6] = '-';

	/*
	 *	Convert the public permissions subfield.
	 */
	if( UNIXReadMask    & publc ) buf[7] = 'r';	else buf[7] = '-';
	if( UNIXWriteMask   & publc ) buf[8] = 'w';	else buf[8] = '-';
	if( UNIXExecuteMask & publc ) buf[9] = 'x';	else buf[9] = '-';

	/*
	 *	Handle the setuid, setgroupid, "sticky" bits.
	 */
	if( UNIXSetUserId & mode ) {
		if( buf[3] == 'x' ) buf[3] = 's';
		else		    buf[3] = 'S';
	}
	if( UNIXSetGroupId & mode ) {
		if( buf[6] == 'x' ) buf[6] = 's';
		else		    buf[6] = 'S';
	}
	if( UNIXSticky & mode ) {
		if( buf[9] == 'x' ) buf[9] = 't';
		else		    buf[9] = 'T';
	}

	/*
	 *	Add null terminator.
	 */
	buf[10] = '\0';
}

void
PrintEntry( UNIXFileInfo * infoBuffer, char * key )
{
	char	modeString[12];

	modefmt( infoBuffer->mode, modeString );
	stdout->formattedWrite( "%s %2d %4d",
		modeString, infoBuffer->links, infoBuffer->inumber );
	stdout->formattedWrite( "%6d %6d %8d %8d ",
		infoBuffer->uid, infoBuffer->gid, infoBuffer->size,
		infoBuffer->blocks );
	stdout->formattedWrite( "%s", key );
	if( ( infoBuffer->mode & UNIXTypeMask ) == UNIXTypeSymbolicLink ) {
		int status = 0;
		SymbolicLinkStar sl = FSI->open( key,
				    SymbolicLinkClass, status, O_NFOLLOW );
		if( sl != 0 ) {
			char *linkBuffer = new char[infoBuffer->size+1];
			sl->at( 0, linkBuffer, infoBuffer->size );
			linkBuffer[infoBuffer->size] = 0;
			stdout->formattedWrite( " -> %s", linkBuffer );
			delete linkBuffer;
		}
	}

	stdout->formattedWrite( "\n" );
}

int
append( int argc, char *argv[] )
{
	if( argc < 3 ) {
	    stdout->formattedWrite( "usage: append filename ... filename\n" );
	    return( NumberOfArguments );
	}

	int status = 0;

	--argc;
	RecordStreamStar OFile = FSI->open( argv[argc], RecordStreamClass,
					       status );
	if( OFile == 0 ) {
		stdout->formattedWrite( "cannot open %s\n", argv[2] );
		return( status );
	}

	OFile->setRecordNumber( 0, 2 );

	int bufSize = 512;
	char * buf = new char[bufSize];

	for( int i = 1; i < argc; i++ ) {
		RecordStreamStar IFile = FSI->open( argv[i],
						  RecordStreamClass, status );
		if( IFile == 0 ) {
			stdout->formattedWrite( "cannot open %s\n", argv[i] );
			break;
		}
		else {
			int amt = IFile->read( buf, bufSize, status );
			while( amt > 0 ) {
				OFile->write( buf, amt, status );
				amt = IFile->read( buf, bufSize );
			}

		}
	}

	delete buf;

	return( status );
}

int
cat( int argc, char *argv[] )
{
	if( argc < 2 ) {
		stdout->formattedWrite( "usage: cat filename [...]\n" );
		return( NumberOfArguments );
	}
	for( int i = 1; i < argc; i++) {
		int status = 0;
		RecordStreamStar RFile =
			FSI->open( argv[i], RecordStreamClass, status );

		if( RFile != 0 ) {
                        const int lineSize = 40;
                        char buf[lineSize+1];

                        int status = 0;
                        int amt = RFile->read( buf, lineSize, status );
                        while( status != EndOfFile ) {
                                Assert( amt >= 0 );
                                buf[amt] = '\0';
                                if( status == 0 ) {
                                        stdout->formattedWrite( "%s", buf );
                                        if( RFile->recordSize() > 1 ) {
                                                stdout->formattedWrite( "\n" );
                                        }
                                }
                                amt = RFile->read( buf, lineSize, status );
                        }
                }
                else {
                        stdout->formattedWrite( "cat: cant open %s\n", argv[i] )
;
                        return( status );
                }
        }
        return( 0 );
}

int
create( int argc, char *argv[] )
{
        if( argc < 3 ) {
                stdout->formattedWrite( "usage: create class [record-size] files
...\n" );
                return( NumberOfArguments );
        }

        ClassStar aClass = StandardNameServer->lookup( argv[1], "Class" );

        if( aClass == 0 ) {
                stdout->formattedWrite( "create: bad class name\n" );
                return( NumberOfArguments );
        }

        int status = 0;
        extern ClassStar IndexedRecordFileClass;
        RecordStreamStar OFile = 0;
        int start = 0;
        int recordSize = 0;
        if( aClass->isASubclassOf( IndexedRecordFileClass ) ) {
                start = argc - 1;
                recordSize = atoi( argv[2] );
                if( argc < 6 ) {
                        stdout->formattedWrite( "usage: create cls rsz s1 l1 fn\
n" );
                        return( NumberOfArguments );
                }
        }
        else if( aClass->isASubclassOf( FixedRecordFileClass ) ) {
                start = 3;
                recordSize = atoi( argv[2] );
                if( argc < 4 ) {
                        stdout->formattedWrite( "usage: create cls rsz fn\n" );
                        return( NumberOfArguments );
                }
        }
        else if( aClass->isASubclassOf( VariableRecordFileClass ) ) {
                start = 2;
        }
        else {
                stdout->formattedWrite( "usage: create bad class name\n" );
                return( NumberOfArguments );
        }
        for( int i = start; i < argc; i++ ) {
                int initArgc = 0;
                if( aClass->isASubclassOf( IndexedRecordFileClass ) ) {
                    OFile = FSI->open( argv[i], FixedRecordStreamClass,
                                       status, O_CREAT, 0777, aClass );
                    initArgc = argc - 3;
                }
                else if( aClass->isASubclassOf( FixedRecordFileClass ) ) {
                    OFile = FSI->open( argv[i], FixedRecordStreamClass,
                                       status, O_CREAT, 0777, aClass );
                    initArgc = 1;
                }
                else if( aClass->isASubclassOf( VariableRecordFileClass ) ) {
                    OFile = FSI->open( argv[i], VariableRecordStreamClass,
                                       status, O_CREAT, 0777, aClass );
                }
                if( OFile != 0 ) {
                    OFile->init( initArgc, &argv[2] );

                    char * buf = new char[256];

                    while( gets( buf ) ) {
                        if( ( buf[0] == '.' ) && ( buf[1] == '\0' ) ) break;
                        int amt = strlen( buf );
                        int status = 0;
                        char nbuf[32];
                        gets( nbuf );
                        OFile->setRecordNumber( atoi( nbuf ), 0 );
                        OFile->write( buf, recordSize ? recordSize : amt, status
 );
                    }

                    delete buf;
                }
        }
        return( 0 );
}


int
chgroup( int argc, char *argv[] )
{
	if( argc < 3 ) {
		stdout->formattedWrite( "usage: chgrp gid filename [...]\n" );
		return( NumberOfArguments );
	}
	int gid = atoi( argv[1] );
	for( int i = 2; i < argc; i++) {
		int status = 0;
		PersistentObjectStar so = FSI->open( argv[i], PersistentObjectClass,
						 status );
		if( so != 0 ) {
			status = so->setGroupId( gid );
		}
		if( status != 0 ) {
			stdout->formattedWrite( "chgrp: cannot chown %s\n", argv[i] );
			return( status );
		}
	}
	return( 0 );
}

int
chowner( int argc, char *argv[] )
{
	if( argc < 3 ) {
		stdout->formattedWrite( "usage: chown gid filename [...]\n" );
		return( NumberOfArguments );
	}
	int uid = atoi( argv[1] );
	for( int i = 2; i < argc; i++) {
		int status = 0;
		PersistentObjectStar so = FSI->open( argv[i], PersistentObjectClass,
						 status );
		if( so != 0 ) {
			status = so->setUserId( uid );
		}
		if( status != 0 ) {
			stdout->formattedWrite( "chown: cannot chown %s\n", argv[i] );
			return( status );
		}
	}
	return( 0 );
}

int
chmode( int argc, char *argv[] )
{
	if( argc < 3 ) {
		stdout->formattedWrite( "usage: chmod omode filename [...]\n" );
		return( NumberOfArguments );
	}
	int omode = atoi( argv[1], 8 );
	for( int i = 2; i < argc; i++) {
		int status = 0;
		PersistentObjectStar so = FSI->open( argv[i], PersistentObjectClass,
						 status );
		if( so != 0 ) {
			status = so->setMode( omode );
		}
		if( status != 0 ) {
			stdout->formattedWrite( "chmod: cannot chmod %s\n", argv[i] );
			return( status );
		}
	}
	return( 0 );
}

int
cd( int argc, char *argv[] )
{
	if( argc == 1 ) return( FSI->chdir( HOME ) );
	else if( argc == 2 ) return( FSI->chdir( argv[1] ) );
	else return( NumberOfArguments );
}

int
cdroot( int argc, char *argv[] )
{
	if( argc != 2 ) return( NumberOfArguments );
	return( FSI->chroot( argv[1] ) );
}

int
cp( int argc, char *argv[] )
{
	if( argc != 3 ) {
		stdout->formattedWrite( "usage: cp filename filename\n" );
		return( NumberOfArguments );
	}

// FIX try to push a dir
	int status = 0;
	if( argc != 3 ) {
		stdout->formattedWrite( "usage: cp filename filename\n" );
		return( NumberOfArguments );
	}
	PersistentObjectStar IFile = FSI->open( argv[1],
					PersistentObjectClass, status );
	if( IFile == 0 ) {
		stdout->formattedWrite( "cannot open %s\n", argv[1] );
		return( status );
	}
// FIX dont use 0666 for modes
        UNIXFileInfo statBuffer;
        IFile->info( &statBuffer );
        ClassStar aClass = IFile->classOf();
        PersistentArrayStar OFile = FSI->open( argv[2], aClass, status,
                                                O_CREAT|O_TRUNC,
                                               ( statBuffer.mode & 0777 ),
                                               aClass );
	if( OFile == 0 ) {
		stdout->formattedWrite( "cannot create %s\n", argv[2] );
		return( status );
	}

	status = IFile->copy( OFile );
	
	return( status );
}

int
dump( int argc, char *argv[] )
{
        if( argc < 2 ) {
                stdout->formattedWrite( "usage: dump filename ...\n" );
                return( NumberOfArguments );
        }
        for( int i = 1; i < argc; i++) {
                int status = 0;
                RecordStreamStar RFile =
                        FSI->open( argv[i], RecordStreamClass, status );

                stdout->formattedWrite( "\n%s: ", argv[i] );

                if( RFile != 0 ) {
                    int recSize = RFile->recordSize();
                    if( recSize <= 1 ) recSize = RFile->maximumRecordSize();
                    Assert( recSize > 0 );
                    char * buf = new char[recSize];

                    int status = 0;
                    int i = 0;
                    int amt = RFile->read( buf, recSize, status );
                    while( status != EndOfFile ) {
                        Assert( amt >= 0 );
                        if( status == 0 ) {
                            for( int j = 0; j < amt; i++ ) {
                                if( ( i % 8 ) == 0 ) {
                                    stdout->formattedWrite( "\n%7o ", i );
                                }
                                if( ( buf[j] >= ' ' ) && ( buf[j] <= '~' ) ) {
                                    stdout->formattedWrite( "   %c",buf[j] );
                                }
                                else if( buf[j] == '\0' ) {
                                        stdout->formattedWrite( "  \\0" );
                                }
                                else if( buf[j] == '\n' ) {
                                        stdout->formattedWrite( "  \\n" );
                                }
                                else if( buf[j] == '\t' ) {
                                        stdout->formattedWrite( "  \\t" );
                                }
                                else stdout->formattedWrite( "%4o", 255&buf[j]);
                                if( ++j < amt ) stdout->formattedWrite( " " );
                                else if( recSize > 1 ) {
                                        stdout->formattedWrite( "|" );
                                }
                            }
                        }
                        amt = RFile->read( buf, recSize, status );
                    }
                    delete buf;
                }
                else {
                        stdout->formattedWrite( "cat: cant open %s\n", argv[i] )
;
                        return( status );
                }
                stdout->formattedWrite( "\n" );
        }
        return( 0 );
}

int
inst( int argc, char *argv[] )
{
	StandardSystemInterface->controlSimpleInstruments( argc, argv );
	return( 0 );
}

int
ln( int argc, char *argv[] )
{
	if( argc == 3 ) {
		return( FSI->link( argv[1], argv[2] ) );
	}

	if( ( argc == 4 ) && ( strcmp( argv[1], "-s" ) == 0 ) ) {
		int status = 0;
		SymbolicLinkStar OFile = FSI->open( argv[3], SymbolicLinkClass,
						status, O_CREAT|O_EXCL,
						0644, SymbolicLinkClass );
		if( status != 0 ) {
			stdout->formattedWrite( "cannot create %s\n", argv[3] );
			return( status );
		}

		OFile->atPut( 0, argv[2], strlen( argv[2] ) );
		return( status );
	}

	else {
		stdout->formattedWrite( "usage: ln filename filename\n" );
		stdout->formattedWrite( "usage: ln -s filename filename\n" );
		return( NumberOfArguments );
	}
}

int
ls( int argc, char *argv[] )
{
	UNIXFileInfo * infoBuffer = new UNIXFileInfo;
	char * path  = new char [1024];
	char * keyBuffer  = new char [32000];
	int status;

	if ( argc == 1 ) {
		argc = 2;
		argv[1] = ".";
	}

Debug <<  "ls(" << argc << ", " << asAddress(argv) << ")\n" << eor;

	stdout->formattedWrite( "   mode    nlnk ino   uid    gid    size     blks name\n" );

	for( int a = 1; a < argc; a++) {
	    Debug <<  "getting status of " << argv[a] << "\n" << eor;
            status = FSI->info( argv[a], infoBuffer, 0 );
	    Debug <<  "got status of " << argv[a] << "\n" << eor;
	       if( status == 0 ) {
		  if( ( infoBuffer->mode & 0170000 ) == 040000 ) {
		     Debug << "opening " << argv[a] << "\n" << eor;
		     MemoryObjectDictionaryStar od = FSI->open( argv[a],
					MemoryObjectDictionaryClass, status );
		     Debug << "opened " << argv[a] << "\n" << eor;
		     if( od != 0 ) {
		       int count;
		     Debug << "getting keys " "\n" << eor;
		       status = od->keys( keyBuffer, 32000, count );
		     Debug << "got keys " "\n" << eor;
		       if( status == 0 ) {
			strcpy( path, argv[a] );
			int pathLength = strlen( path );
		        if( strcmp( path, ".") != 0 ) {
			    stdout->formattedWrite( "%s:\n", path );
		        }
			path[pathLength++] = '/';

			char * keyStart = keyBuffer;
			char * keyEnd = keyBuffer;

			while( *keyStart ) {
			    while( *keyEnd != '\n' ) keyEnd++;
			    *keyEnd = '\0';

			    strcpy( &path[pathLength], keyStart );

                            status = FSI->info( path, infoBuffer, 0 );
                            if( status == 0 ) PrintEntry( infoBuffer, keyStart );
			    else {
				stdout->formattedWrite( "Cannot stat %s\n", keyStart );
			    }

			    keyEnd++;
			    keyStart = keyEnd;
			}
		        stdout->formattedWrite( "\n" );
		       }
		     }
		  }
		  else {
			PrintEntry( infoBuffer, argv[a] );
		  }
	    }
	    else {
		stdout->formattedWrite( "Cannot stat %s\n", argv[a] );
	    }
	}
	delete keyBuffer;
	delete infoBuffer;
	delete path;
	return( 0 );
}

int
makedir( int argc, char *argv[] )
{
	int status = 0;
	if( argc == 2 ) {
		extern PersistentClassStar MemoryObjectDictionaryClass;
		PersistentObjectStar so = FSI->open( argv[1],
			MemoryObjectDictionaryClass, status,
			O_CREAT|O_EXCL, 0777, MemoryObjectDictionaryClass );
		so = 0;
	}
	else {
		stdout->formattedWrite( "usage: mkdir dirname\n" );
		status = NumberOfArguments; 
	}
	return( status ); 
}

int
makepo( int argc, char *argv[] )
{
	int status = 0;
	if( argc >= 3 ) {
		ClassStar aClass = StandardNameServer->lookup( argv[1],
							     ClassClass );
                if( aClass == 0 ) {
                        stdout->formattedWrite( "bad class name\n" );
                        status = NumberOfArguments;
                }
                else if( aClass->isASubclassOf( PersistentObjectClass ) ) {
                        PersistentObjectStar po = FSI->open( argv[2], aClass,
                                                         status, O_CREAT|O_EXCL,
                                                         0777, aClass );
                        if( po == 0 ) {
                                stdout->formattedWrite( "mkpo failed\n" );
                                status = NumberOfArguments;
                        }
                        else status = po->init( argc - 3, &argv[3], FSI );
                }
                else {
		        PersistentObjectStar po = FSI->open( argv[2], aClass,
						status, O_CREAT|O_EXCL,
						   0777, aClass );
                       if( po == 0 ) {
                                stdout->formattedWrite( "mkpo failed\n" );
                                status = NumberOfArguments;
                        }
                }
        }
	else {
		stdout->formattedWrite( "usage: mkpo classname filename\n" );
		status = NumberOfArguments; 
	}
	return( status ); 
}

int
mv( int argc, char *argv[] )
{
	if( argc != 3 ) {
		stdout->formattedWrite( "usage: mv filename filename\n" );
		return( NumberOfArguments );
	}
	int status = FSI->link( argv[1], argv[2] );
	if( status == 0 ) {
		return( FSI->unlink( argv[1] ) );
	}
	else {
		int status = 0;
		PersistentObjectStar IFile = FSI->open( argv[1],
						PersistentObjectClass, status );
		if( IFile == 0 ) {
			stdout->formattedWrite( "cannot open %s\n", argv[1] );
			return( status );
		}
                UNIXFileInfo statBuffer;
                IFile->info( &statBuffer );
                ClassStar aClass = IFile->classOf();
                PersistentArrayStar OFile = FSI->open( argv[2], aClass, status,
                                                O_CREAT|O_TRUNC,
                                               ( statBuffer.mode & 0777 ),
                                               aClass );
		if( OFile == 0 ) {
			stdout->formattedWrite( "cannot create %s\n", argv[2] );
			return( status );
		}
	
		status = IFile->copy( OFile );
		return( FSI->unlink( argv[1] ) );
	}
}

int
rm( int argc, char *argv[] )
{
	if( argc != 2 ) {
		stdout->formattedWrite( "usage: rm filename\n" );
		return( NumberOfArguments );
	}
	return( FSI->unlink( argv[1] ) );
}

int
remdir( int argc, char *argv[] )
{
	if( argc != 2 ) {
		stdout->formattedWrite( "usage: rmdir dirname\n" );
		return( NumberOfArguments );
	}
	return( FSI->unlink( argv[1] ) );
}

int
synch( int argc, char *argv[] )
{
	argc = argc;
	return( FSI->synchronize() );
}

int
mnt( int argc, char *argv[] )
{
	if( argc != 3 ) {
		stdout->formattedWrite( "usage: mount special pathname\n" );
		return( NumberOfArguments );
	}
	return( FSI->mount( argv[1], argv[2], 0 ) );
}

int
unmnt( int argc, char *argv[] )
{
	if( argc != 2 ) {
		stdout->formattedWrite( "usage: umount pathname\n" );
		return( NumberOfArguments );
	}
	return( FSI->unmount( argv[1] ) );
}
