/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Taed Nelson (nelson@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * iconvert.h: Converter for use with instrumentation.
 *
 */


#define INSTRUMENT
#include "InstrumentTypes.h"
#include "InstrumentFormat.h"


struct CpuSyncNode {
/* Keeps track of how many buffer dumps a specific CPU has done,
 *   so that we know when it is safe to write out a given buffer to disk.
 * This should be based on timestamps instead of number of dumps.
 */
	
	unsigned int cpuid;
	unsigned int sync;
	CpuSyncNode *next;

	/* Method CpuSyncNode:		The constructor.		*/
	CpuSyncNode (unsigned int _cpuid)
		{ cpuid = _cpuid;  sync = 1;  next = NULL; };
};


class CpuSyncClass {
/* This used the above structure to determine which buffers can be
 *   written out by determining their overlap (buffers are assumed to
 *   be of the same size, which is not quite true).
 * This should be based on timestamps instead of number of dumps.
 * In the current Multimax implementation, this class has no real purpose
 *   since all buffer dumps take place at the same time.
 */
	
private:
	// What is the maximum lag before it is save to write the data
	const unsigned int bufferOverlap = 1;
	
	CpuSyncNode *cpuSyncList;

protected:

public:
	/* Method CpuSyncClass:		The constructor.		*/
	CpuSyncClass ();

	/* Method CpuSyncClass:		The constructor.		*/
	~CpuSyncClass ();

	/* Method sync:		Called every time we hit a buffer sync
	 *			when reading in the data file.
	 *			Returns TRUE if we should write data.	*/
	Boolean sync (unsigned int cpuid);

	/* Method resync:	Called every time we hit a buffer sync
	 *			while writing out the buffer data.
	 *			Returns TRUE to continue writing data.	*/
	Boolean resync (unsigned int cpuid);
};


class TitleNodeClass {
/* Stores an instrument-id/title matching which is active between two
 *   timestamps.
 */

	friend class TitleListClass;

 private:
	Boolean  stopActivityKnown;
	timeType startActivity;
	timeType stopActivity;
	
	unsigned int ownerid;
	unsigned int selfid;

	char *title;

	TitleNodeClass *next;
	
 public:
	/* Method TitleNodeClass:	The constructor.		*/
	TitleNodeClass (timeType Start, unsigned int Ownerid,
			unsigned int Selfid, char *Title);
};


class TitleListClass {
/* Just keep a list of the titles and provide support routines.
 */

 private:
	TitleNodeClass *head;
	
 public:
	/* Method TitleListClass:	The constructor.		*/
	TitleListClass ();

	/* Method lookup:	Find the appropriate title.		*/
	char * lookup (timeType lookupTime, unsigned int ownerid, unsigned int selfid);

	/* Method update:	Register a change in title status.	*/
	Boolean update (timeType updateTime, unsigned int ownerid,
			unsigned int selfid, char *title);
};


class GeneralEventClass {
/* Stores all information needed for a single event.
 */

 public:
	/* Method GeneralEventClass:	The constructor.		*/
	GeneralEventClass ();

	GeneralEventClass *next;
	GeneralEventClass *prev;

#ifdef DEBUG
	// Where in the input file the event was found
	unsigned int fileIndex;
#endif

	// Are we a buffer sync event?
	Boolean syncevent;
	
	char *title;

	timeType timestamp;
	iIdentifierType header;

	countType count;
	timeType deltatime;

	unsigned int messagelength;
	char *message;

	unsigned int cpuid;

	/* Method setup:	Initialize the event.			*/
	void setup (unsigned int cpu, TitleListClass *list,
		    Boolean sync = FALSE);

	/* Method read:		Read in an event from a file.		*/
	Boolean read (FILE *fp);

	/* Method write:	Write out the text format to a file.	*/
	void write (FILE *fp, TitleListClass *titleList);
};


class EventListClass {
/* Just a list of events with associated helpers.
 */

 private:
	GeneralEventClass *head;
	GeneralEventClass *current;

	CpuSyncClass *cpuSync;

	TitleListClass *titleList;
	FILE *fp;

 public:
	/* Method EventListClass:	The constructor.	*/
	EventListClass (TitleListClass *_titleList);

	/* Method add:	Add a new event.	*/
	void add (GeneralEventClass * event);

	/* Method reset:	Reset the insert pointer for that cpu.	*/
	void reset (unsigned int cpuid);

	/* Method setOutputFile:	Set the file to output events.	*/
	void setOutputFile (FILE *_fp)
		{ fp = _fp; };
	
	/* Method write:	Write out everthing that is cpu synced.	*/
	void write ();

	/* Method forcedWrite:	Write out everything no matter what.	*/
	void forcedWrite ();
};
