#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MultimaxEMCDisk.cc: Multimax EMC Disk driver.
 */
#include "Assert.h"
#include "Raid.h"
#include "MultimaxEMCDisk.h"
#include "MemoryObjectPartition.h"
#include "String.h"
#include "Class.h"
#include "ClassStar.h"
#include "stdio.h"

extern ClassStar StoredObjectClass;

MultimaxEMCDisk::MultimaxEMCDisk( int, int, int )
: ( 1 )
{
	initDebug();
	char diskname[8];

	Raid( RaidConstructor )( "%N::MultimaxEMCDisk()\n", this );

	strcpy( diskname, "vdisk0" );

	_log2UnitSize = 9;
	int start = 0;
	for( _partitions = 0; _partitions < PARTITIONS; _partitions++ ) {
		diskname[5] = '0' + _partitions;
		_layout.l_partitions[_partitions].part_fd = ::open( diskname, 2 );
		if( _layout.l_partitions[_partitions].part_fd < 0 ) break;

		_layout.l_partitions[_partitions].part_off = start;
		_layout.l_partitions[_partitions].part_size =
			( ::lseek( _layout.l_partitions[_partitions].part_fd, 0, 2 ) +
					 unitSize() - 1 ) / unitSize();
		start += _layout.l_partitions[_partitions].part_size;
	
		Raid( RaidDetail )( "%N::MultimaxEMCDisk( %d, %d, %d )\n",
		    this, _layout.l_partitions[_partitions].part_fd,
		    _layout.l_partitions[_partitions].part_off, _layout.l_partitions[_partitions].part_size );
	}
	_numberOfUnits += start;
	Assert( _partitions > 0 );
	Assert( _partitions < PARTITIONS );
}

MultimaxEMCDisk::~MultimaxEMCDisk()
{
	for( _partitions--; _partitions >= 0 ; _partitions-- ) {
		::close( _layout.l_partitions[_partitions].part_fd );
	}
}

int
MultimaxEMCDisk::transient()
{
	return( 0 );
}

int
MultimaxEMCDisk::read( unsigned int pblk, int sectors, char * buffer, int )
{
	for( int p = 0; p < _partitions; p++ ) {
		if( pblk < ( _layout.l_partitions[p].part_off + _layout.l_partitions[p].part_size ) ) break;
	}
	Assert( p < _partitions );

	Raid( RaidMethod )( "%N::read( %d, %d, %x )\n",
		this, pblk, sectors, buffer );
	Raid( RaidDetail )( "%N::read() start %d length %d\n", this,
		_layout.l_partitions[p].part_off, _layout.l_partitions[p].part_size );

	pblk -= _layout.l_partitions[p].part_off;
	Assert( pblk >= 0 );
	Assert( ( pblk + sectors ) < _layout.l_partitions[p].part_size );
	int i = lseek( _layout.l_partitions[p].part_fd, pblk*unitSize(), 0 );
	Assert( i != - 1);
	for( int s = 0; s < sectors; s++ ) {
		i = ::read( _layout.l_partitions[p].part_fd, &buffer[s * unitSize()], unitSize() );
		Assert( i == unitSize() );
	}
	return( sectors );
}

int
MultimaxEMCDisk::write( unsigned int pblk, int sectors, char * buffer, int )
{
	for( int p = 0; p < _partitions; p++ ) {
		if( pblk < ( _layout.l_partitions[p].part_off + _layout.l_partitions[p].part_size ) ) break;
	}
	Assert( p < _partitions );

	Raid( RaidMethod )( "%N::write( %d, %d, %x )\n",
		   this, pblk, sectors, buffer );
	Raid( RaidDetail )( "%N::read() start %d length %d\n", this,
		_layout.l_partitions[p].part_off, _layout.l_partitions[p].part_size );

	pblk -= _layout.l_partitions[p].part_off;
	Assert( pblk >= 0 );
	Assert( ( pblk + sectors ) < _layout.l_partitions[p].part_size );
	int i = lseek( _layout.l_partitions[p].part_fd, pblk*unitSize(), 0 );
	Assert( i != - 1);
	for( int s = 0; s < sectors; s++ ) {
		i = ::write( _layout.l_partitions[p].part_fd, &buffer[s * unitSize()], unitSize() );
		Assert( i == unitSize() );
	}
	return( sectors );
}

void
MultimaxEMCDisk::initialize()
{
}

void
MultimaxEMCDisk::handleAttention()
{
}

void
MultimaxEMCDisk::handleInterrupt()
{
}

void
MultimaxEMCDisk::diskInterruptProcessEntry( MultimaxEMCDisk * )
{
}

int
MultimaxEMCDisk::doio( unsigned int, int, char *, int, int )
{
	Assert( NOTREACHED );
	return( 0 );
}

int
MultimaxEMCDisk::basicDoio( unsigned int, char *, int )
{
    Assert( NOTREACHED );
    return( DiskOperFailed );
}

int
MultimaxEMCDisk::reset()
{
    return( DiskOperSucceded );
}

int
MultimaxEMCDisk::partitions()
{
	Raid( RaidMethod )( "%N::partitions() %d\n", this, _partitions );
	return( _partitions );
}

int
MultimaxEMCDisk::locatePartition( int idNumber, int &start, int &units )
{
	Debug << "MultimaxEMCDisk::locatePartition( idNumber:" << 
		idNumber << " )\n" << eor;

	if( _layout.l_partitions[ idNumber ].part_size == 0 ) {
		Console().printf( "%N::locatePartition: empty partition %d\n",
				 this, idNumber );
		return( -1 );
	}

	start = _layout.l_partitions[ idNumber ].part_off;
	units = _layout.l_partitions[ idNumber ].part_size;

	return( 0 );
}



#include "ClassInstaller.h"
DEFINECLASS("MultimaxEMCDisk",MultimaxEMCDisk,Disk,0,0);
