#define DEBUG
#ifdef __GNUC__
#pragma implementation
#endif


#include "Assert.h"
#include "Byte.h"
#include "Class.h"
#include "Domain.h"
#include "Exception.h"
#include "Kernel.h"
#include "OutputStream.h"
#include "PROCESSOR.h"
#include "SoftClock.h"
#include "SoftInt.h"
#include "SourceOfZerosMemoryObject.h"
#include "Store.h"
#include "UnixTime.h"
#include "ThisCPU.h"
#include "ThisProcess.h"
#include "TimeOfDay.h"

#ifdef SS1

#include "NonReentrant.h"
#include "SPARCTranslation.h"
#include "SS1Configuration.h"
#define ASM
#include "SS1devices.h"
#include "SS1Timer.h"
#include "SS1util.h"
#define HZ ClockFrequency;

/* 

SS1 DMA Memory layout 
    all values offset from DMA_BASE 

This assumes mapToDMA is no longer in use.

0x200000-0x4fffff 		used by zalloc mo
	0x200000-0x2fffff	uncached, used for ring descriptors etc.
	0x300000-0x3fffff	cached used by transmit buffers
	0x400000-0x4fffff	128 or MSIZE mbufs, cached
0x500000-0x500000+(NMBCLUSTERS*CLBYTES) used by mbuf clusters

*/

const int ZALLOC_MOSIZE = 0x00300000; // 3 Mb
char * const BSD_DMA_BASE    = DMA_BASE + 0x00200000;
char * const ZALLOC_DMA_BASE = BSD_DMA_BASE;
char * const MBUF_DMA_BASE = ZALLOC_DMA_BASE + ZALLOC_MOSIZE;

#define INTREG_SOFT_INT1 0x02

#endif

#define NVNODE 256
#define NCALLOUT 64

struct link {
	struct link *forward;
	struct link *backward;
};

extern "C" {

#include "ioctl.h"
#include "time.h"
#include "param.h"
#include "map.h"
#include "proc.h"
#include "user.h"
#include "syslog.h"
#include "vnode.h"
#include "callout.h"
#include "domain.h"
#include "file.h"
#include "syslimits.h"
#include "malloc.h"
#include "mbuf.h"
#include "machine/mtpr.h"
/* #include "machine/pte.h" */
#include "socket.h"
#include "../netinet/in.h"
#include "../net/if.h"
#include "../net/netisr.h"
#include "../net/route.h"
/* #include "../net/if_arp.h" */
#include "../netinet/if_ether.h"
#include "../choices/debug.h"

// called by others

void _insque(struct link *after, struct link *element);
void _remque(struct link *element);
void bcopy(char * from, char *to, unsigned len);
void bzero(char * dest, unsigned len);
int copyin(char *udaddr, char *kaddr, unsigned int length);
int copyout(char *kaddr, char *udaddr, unsigned int length);
void * currentprocess();
void free(char * p, int type);
void free_mbuf(char * p, int type);
int gotsir();
void hardclock();
void leattach();
int log(int level, char *fmt, unsigned x1, unsigned x2, unsigned x3, unsigned x4);
void makeaddressable(char * vaddr, int npg);
long int *malloc(unsigned long size, int type);
long int *malloc_mbuf(unsigned long size, int type);
void microtime(struct timeval *mtv);
int memall(struct pte *pte, int size, struct proc *p, int type);
//char * memcpy(char *dest, char *src, unsigned n);
void munmapfd(int fd);
void nchinit();
void ovbcopy(char * from, char *to, unsigned len);
int oc_cksum(char * buffer, unsigned count, unsigned startval);
int panic(char *str);
struct proc * pfind(int pid);
void psignal(int pid, int signal);
void gsignal(int gid, int signal);
void printf(char *s, unsigned x1, unsigned x2, unsigned x3, unsigned x4);
void dprintetherheader( struct ether_header * header );
void tablefull(char *str);
int selwakeup(struct proc *p, int coll);
void setsoftnet();
int splx(int s);
int spl0();
int splimp();
int splnet();
int splhigh();
int splsoftcall();
int timeout(int (*fun)(), caddr_t arg, int t);
int untimeout(int (*fun)(), caddr_t arg);
/* int suser(struct ucred *cred, short *acflag); */
void vinvalbuf(struct vnode *vp, int save);

// i call these

int rminit(struct map *mp, long size, long addr, char *name, int mapsize);
void vfsinit();
void mbinit();
void loattach();
void ifinit();
void softclock();
void domaininit();
#ifdef INET
int ipintr();
#endif
#ifdef NS
int nsintr();
#endif
#ifdef ISO
int clnlintr();
#endif

// variables

int debugBSD = DEBUGBSD_ERROR | DEBUGBSD_WARNING;
int hz = HZ;
int tick = 1000000 / HZ;
int dmmax;
long desiredvnodes = NVNODE;
#ifdef NEVER
int nmbclusters = NMBCLUSTERS;
int ncallout = NCALLOUT;
int nfile;
struct map *mbmap;
struct callout *callfree;
struct callout *callout;
struct file * file;
struct file * fileNFILE;
#endif
unsigned ssir = 0;
char *mbutl;
long tickdelta = 0;
long timedelta = 0;
struct user u;
struct proc proc0;

extern struct domain inetdomain;
extern struct domain * domains;

};

int MaxMbuf = NMBCLUSTERS * CLBYTES;

void ifconfig();

int net_init()
{
	/* bcopy depends on these */

	Assert(sizeof(long) == 4);
	Assert(sizeof(short) == 2);

	/* makeaddressable depends on these */

	Assert(CLBYTES == MainStore->unitSize());
	Assert(CLSIZE == 1);
	Assert(NBPG == MainStore->unitSize());

	/* needed by microtime */

	Assert(theSystemClock != 0);

	/* be sure u is all zeros */

	bzero((char *) &u, sizeof(struct user));	

	/* initialize file */

	nfile = OPEN_MAX;
	int fileSize = sizeof(struct file) * nfile;
	file = (struct file *) new char[fileSize];
	fileNFILE = file + nfile;
	bzero((char *) file, fileSize);
	
	/* initialize callout */

	callout = (struct callout *)new char[sizeof(struct callout) * ncallout];
	struct callout * callout1 = (struct callout *) callout;
	callfree = (struct callout *) callout;
	for (int i = 1; i < ncallout; i++)
		callout1[i-1].c_next = &callout1[i];

	/* setup mbutl */

	extern ClassStar RootMemoryObjectCacheClass;
	MemoryObject * mbutl_mo = new SourceOfZerosMemoryObject;
	mbutl_mo->setNumberOfUnits(NMBCLUSTERS * CLBYTES);
	mbutl_mo->setName("mbutl");
	mbutl_mo->setCacheClass(RootMemoryObjectCacheClass);
#ifndef SS1
	mbutl = (char *) TheKernel->domain()->add(mbutl_mo, NoAccess);
#else 
	mbutl = (char *) TheKernel->domain()->add(mbutl_mo, MBUF_DMA_BASE, 
						  NoAccess);
	Assert(mbutl == MBUF_DMA_BASE);
	
	/* setup another mo for SS1 DMA */
	
	void ss1zallocinit();
	ss1zallocinit();
#endif

	/* setup mbmap */

	mbmap = new map[nmbclusters/4];
	rminit(mbmap, (long)(nmbclusters * MCLBYTES / NBPG), (long)CLSIZE,
	       "mbclusters", nmbclusters/4);
	
	/* fate a dummy u and proc[0] */

	// u.u_procp = &proc0;
	u.u_cmask = 0;
	u.u_lastfile = -1;
	
	/* duplicate init_main */

	vfsinit();

	/* setup dummy credentials */

	u.u_cred = crget();
	u.u_cred->cr_ngroups = 1;
	Assert(u.u_uid == 0);

	/* init networking code */

	mbinit(); 
	loattach();
	int s = splimp();
	domaininit();

	leattach();
	s = splimp();
	ifinit();
	splx(s);
	void ifconfig();
	ifconfig();
}

void _insque(struct link *element, struct link *after)
{
	IntStorage intState ; PROCESSOR::disableInterrupts(intState);
	element->forward = after->forward;
	element->backward = after;
	after->forward = element;
	element->forward->backward = element;
	PROCESSOR::restoreInterrupts(intState);
}

void _remque(struct link *element)
{
	IntStorage intState ; PROCESSOR::disableInterrupts(intState);
	element->forward->backward = element->backward;
	element->backward->forward = element->forward;
	PROCESSOR::restoreInterrupts(intState);
}

void * currentprocess()
{
	return thisProcess();
}

void free(char *p, int type)
{
	delete p;
}

#ifndef SS1
void free_mbuf(char *p, int type)
{
	delete p;
}
#endif

int log(int level, char *fmt, unsigned x1, unsigned x2, unsigned x3, unsigned x4) 
{
	IntStorage intState ; PROCESSOR::disableInterrupts(intState);
	Console().formattedWrite("syslog level %d:", level);
	Console().formattedWrite(fmt, x1, x2, x3, x4);
	Console().flush();
	PROCESSOR::restoreInterrupts(intState);
}

void makeaddressable(char * vaddr, int npg)
{
	TheKernel->domain()->makeAddressable(vaddr, npg*CLBYTES);
/*
#ifdef SS1
	SPARCTranslation::makeUncachable(vaddr, npg*CLBYTES);
#endif
 */
}

long int * malloc(unsigned long size, int type)
{
	return (long int *) (new char[size]);
}

#ifndef SS1
long int * malloc_mbuf(unsigned long size, int type)
{
	return (long int *) (new char[size]);
}
#endif

void microtime(struct timeval *mtv)
{
	UnixTime unixtime;
	theSystemClock->getUnixTime(unixtime);
	mtv->tv_sec = unixtime.seconds();
	mtv->tv_usec = unixtime.useconds();
}

#ifdef MYBCOPY

const unsigned long longalignmask = sizeof(long) - 1;
const unsigned long shortalignmask = sizeof(short) - 1;

inline int shortalign(char * from)
{
	return (((unsigned long) from) & shortalignmask); 
}

inline int longalign(char * from)
{
	return (((unsigned long) from) & longalignmask);
}

inline int shortalign(char *from, char *to)
{
	return shortalign(from) && shortalign(to);
}

inline int longalign(char *from, char *to)
{
	return longalign(from) && longalign(to);
}

inline 
void bcopyforward(char *from, char *to, unsigned len)
{
	if (longalign(from) == longalign(to)) {
		if (shortalign(from) != 0) {
			*to++ = *from++;
			len--;
		}
		if (longalign(from) == 2 && len >= 2) {
			*((short *) to)++ = *((short *) from)++;
			len-=2;
		}
		while(len >= 4) {
			*((long *) to)++ = *((long *) from)++;
			len-=4;
		}
		if (len >= 2) {
			*((short *) to)++ = *((short *) from)++;
			len-=2;
		}
		if (len > 0) {
			*to++ = *from++;
			len--;
		}
	} else if (shortalign(from) == shortalign(to)) {
		if (shortalign(from) != 0) {
			*to++ = *from++;
			len--;
		}
		while(len >= 2) {
			*((short *) to)++ = *((short *) from)++;
			len-=2;
		}
		if (len > 0) {
			*to++ = *from++;
			len--;
		}
	} else {
		while(len > 0) {
			*to++ = *from++;
			len--;
		}
	}
}

inline 
void bcopybackward(char *from, char *to, unsigned len)
{
	from += len;
	to += len;
	if (longalign(from) == longalign(to)) {
		if (shortalign(from) != 0) {
			--to;
			--from;
			*to = *from;
			len--;
		}
		if (longalign(from) == 2 && len >= 2) {
			to -= 2;
			from -= 2;
			*((short *) to) = *((short *) from);
			len-=2;
		}
		while(len >= 4) {
			to -= 4;
			from -= 4;
			*((long *) to) = *((long *) from);
			len-=4;
		}
		if (len >= 2) {
			to -= 2;
			from -= 2;
			*((short *) to) = *((short *) from);
			len-=2;
		}
		if (len > 0) {
			--to;
			--from;
			*to = *from;
			len--;
		}
	} else if (shortalign(from) == shortalign(to)) {
		if (shortalign(from) != 0) {
			--to;
			--from;
			*to = *from;
			len--;
		}
		while(len >= 2) {
			to -= 2;
			from -= 2;
			*((long *) to) = *((long *) from);
			len-=2;
		}
		if (len > 0) {
			--to;
			--from;
			*to = *from;
			len--;
		}
	} else {
		while(len > 0) {
			--to;
			--from;
			*to = *from;
			len--;
		}
	}
}

inline
void bcopyinline(char *from, char *to, unsigned len)
{
	if ( len == 0 )
		return;
	if ( to > from ) 
		bcopyforward(from, to, len);
	else
		bcopybackward(from, to, len);
}

void bcopy(char *from, char *to, unsigned len)
{
	bcopyinline(from, to, len);
}

void ovbcopy(char *from, char *to, unsigned len)
{
	bcopyinline(from, to, len);
}

char * memcpy(char *dest, char *src, unsigned len)
{
	bcopyforward(src, dest, len);
	return dest;
}

int copyin(char *udaddr, char *kaddr, unsigned int length)
{
	bcopyforward(udaddr, kaddr, length);
	return 0;
}

int copyout(char *kaddr, char *udaddr, unsigned int length)
{
	bcopyforward(kaddr, udaddr, length);
	return 0;
}

void bzero(char *dest, unsigned len)
{
	if (shortalign(dest) != 0) {
		*dest++ = 0;
		len--;
	}
	if (longalign(dest) == 2 && len >= 2) {
		*((short *) dest)++ = 0;
		len-=2;
	}
	while(len >= 4) {
		*((long *) dest)++ = 0;
		len-=4;
	}
	if (len >= 2) {
		*((short *) dest)++ = 0;
		len-=2;
	}
	if (len > 0) {
		*dest++ = 0;
		len--;
	}
}

#else /* not MYBCOPY */
void ovbcopy(char *from, char *to, unsigned len)
{
        bcopy(from, to, len);
}
 
char * memcpy(char *dest, char *src, unsigned len)
{
        bcopy(src, dest, len);
        return dest;
}
 
int copyin(char *udaddr, char *kaddr, unsigned int length)
{
        bcopy(udaddr, kaddr, length);
        return 0;
}
 
int copyout(char *kaddr, char *udaddr, unsigned int length)
{
        bcopy(kaddr, udaddr, length);
        return 0;
}
#endif /* MYBCOPY */

void munmapfd(int fd)
{
	Console() << "BSDNet munmapfd( " << fd << " )\n" << eor;
	Assert(NOTREACHED);
}

void nchinit()
{
	/* name cache init */
	return;
}

int panic(char * str)
{
	Console() << "BSDNet panic( \"" << str << "\" )\n" << eor;
	Assert(NOTREACHED);
}

struct proc * pfind(int pid)
{
	return (struct proc *) 4;
}

void psignal(int pid, int signal)
{
	Console() << "BSDNet psignal( " << signal << " )\n" << eor;
	Assert( NOTREACHED );
}

void gsignal(int pid, int signal)
{
	Console() << "BSDNet gsignal( " << signal << " )\n" << eor;
	Assert( NOTREACHED );
}

void printf(char *s, unsigned x1, unsigned x2, unsigned x3, unsigned x4)
{
	Console().formattedWrite(s, x1, x2, x3, x4);
	Console().flush();
}

void
printetheraddress( unsigned char addr[6] )
{
	static char sym[17] = "0123456789abcdef";
 	char buf[32];
	char * p = buf;
	for( int i = 0; i < 6; i++ ) {
		*p++ = sym[addr[i] >> 4];
		*p++ = sym[addr[i] & 0xf];
		*p++ = ':';
	}
	p--;
	*p = 0;	
	Console() << buf;
}

void dprintetherheader( struct ether_header * header )
{
	if ( !debugBSD )
		return;
        Console() << "etherheader dest ";
	printetheraddress( header->ether_dhost );
        Console() << " source ";
	printetheraddress( header->ether_shost );
	Console() << " type " << header->ether_type << "\n" << eor;
}

void tablefull(char *tab)
{

        log(LOG_ERR, "%s: table is full\n", (unsigned long) tab, 0, 0, 0);
}

int selwakeup(struct proc *p, int coll)
{
	Assert(NOTREACHED);
}

#ifdef NEVER
int suser(struct ucred *cred, short *acflag)
{
	/* always superuser */
	return 0; 
}
#endif

extern "C"
int timeout(int (*fun)(), caddr_t arg, int t)
{
	TheSoftwareClockManager->add( (SoftClockFunc) fun, arg, t );
	return 0;
} 

extern "C"
int untimeout(int (*fun)(), caddr_t arg) 
{
	return TheSoftwareClockManager->remove( (SoftClockFunc) fun, arg );
}

void vinvalbuf(struct vnode *vp, int save)
{
	Assert(vp->v_dirtyblkhd == 0 && vp->v_cleanblkhd == 0);
	return;
}

/*
 *
 *	MACHINES AND PROCESSOR DEPENDENT CODE
 *		May require modifications to native Choices code.
 *		
 */

#ifdef SS1

extern "C" {

void ether_copy(void *dest, void *src);
void installetherexception(int (*func)(void));
int localetheraddr(char * hint, char * result);
char * kernelmap_zalloc(unsigned int size, int duumy);
char * kernelmap_zalloc_uncachable(unsigned int size, int duumy);
//int strcmp(char *s1, char *s2);
//int strcpy(char *dest, char *src);
int soclose(struct socket *so);
int socreate(int dom, struct socket **aso, int type, int proto);
char * thisprocess();
int ifioctl(struct socket *so, int cmd, struct ifreq *ifr);
int rtinit(struct ifaddr *ifa, int cmd, int flags);
}

void ether_copy(void *a, void *b) 
{ 
	((short *)b)[0]=((short *)a)[0]; 
 	((short *)b)[1]=((short *)a)[1]; 
	((short *)b)[2]=((short *)a)[2]; 
}

int strcmp(char *s1, char *s2)
{
        while (*s1 == *s2++)
                if (*s1++=='\0')
                        return (0);
        return (*s1 - *--s2);
}

int netintr()
{
#ifdef INET
        if (netisr & (1 << NETISR_IP)) {
                netisr &= ~(1 << NETISR_IP);
                ipintr();
        }
#endif
#ifdef NS
        if (netisr & (1 << NETISR_NS)) {
                netisr &= ~(1 << NETISR_NS);
                nsintr();
        }
#endif
#ifdef ISO
        if (netisr & (1 << NETISR_ISO)) {
                netisr &= ~(1 << NETISR_ISO);
                clnlintr();
        }
#endif
}

void setsoftnet()
{
	TheSoftwareInterruptManager->add( (SoftCallFunc) &netintr, 0 );
}

int splx(int level)
{
	PROCESSOR::restoreInterrupts((IntStorage)level);
	return level;
}

inline int splraise(int desired_level)
{
	int current_level = interruptLevel();
	if (current_level < desired_level)
		PROCESSOR::restoreInterrupts((IntStorage)desired_level);
	return current_level;
}

int splsoftcall()
{
//	const int SPLSOFTCALL = 6;
	const int SPLSOFTCALL = INTERRUPT_HIGH;
	return splraise(SPLSOFTCALL);
}

int splimp()
{
//	const int SPLIMP = 6;
	const int SPLIMP = INTERRUPT_HIGH;
	return splraise(SPLIMP);
}

int splnet()
{
//	const int SPLNET = 1;
	const int SPLNET = INTERRUPT_HIGH;
	return splraise(SPLNET);
}

int splhigh()
{
	const unsigned SPLHIGH = INTERRUPT_LOCK;
	return splraise(SPLHIGH);
}

int spl0()
{
	const unsigned SPL0 = INTERRUPT_LOW;
	return splx(SPL0);
}

int localetheraddr(char * hint, char * result)
{
	if (result) {
		ethernetAddress(result);
	}
	return(1);
}

class EtherException : public Exception {
protected:
	int (*_func)();
        virtual void basicRaise( char * info );
public:  
        EtherException(int (*func)()) : Exception() { _func = func; }
};

void EtherException::basicRaise(char * info)
{
//	static NonReentrant nonreentrant("Ethernet");
//	int my_nesting = nonreentrant.enter();
	int s = splimp();
	static int inUse = 0;
	Assert( inUse == 0 );
	inUse++;	
	_func();
	inUse--;
	Assert( inUse == 0 );
	splx(s);
//	nonreentrant.exit( my_nesting );
}

extern "C"
void installetherexception(int (*func)(void))
{
	Exception * exception = new EtherException(func);
	thisCPU()->setException(21, exception);
	Console() << "installed exception for ethernet device le\n" << eor;
}

char * thisprocess()
{
	return (char *) thisProcess();
}

class Zalloc {
	char * _base;
	char * _end;
	char * _allocated;
	char * _resident;
	int _cacheit;
	static MemoryObject * _mo;
	static char * _nextZalloc;
	const int _moSize = ZALLOC_MOSIZE;
	char * const _baseZalloc = ZALLOC_DMA_BASE;
	char * const _lastZalloc = _baseZalloc + _moSize;
public:
	Zalloc( unsigned size, int cacheit );
	char * alloc( unsigned size );
	char * base() { return _base; }
	char * end() { return _end; }
	char * allocated() { return _allocated; }
};

MemoryObject * Zalloc::_mo = 0;
char * Zalloc::_nextZalloc = 0;

Zalloc::Zalloc( unsigned size, int cacheit )
{
	_cacheit = cacheit;
	if ( _mo == 0 ) {
		extern ClassStar RootMemoryObjectCacheClass;
		_mo = new SourceOfZerosMemoryObject;
		_mo->setNumberOfUnits(_moSize);
		_mo->setCacheClass(RootMemoryObjectCacheClass);
		char *addr = (char *) TheKernel->domain()->add
			(_mo, _baseZalloc, NoAccess);
		Assert(addr == _baseZalloc);
		_nextZalloc = _baseZalloc;
	}
	_base = _nextZalloc;
	_nextZalloc += size;
	_end = _nextZalloc;
	_allocated = _base;
	_resident = _base;
}

inline
char * Zalloc::alloc(unsigned int size)
{
	char * orig_allocated = _allocated;
	_allocated += size;
	if ( size & 0x7 ) {
		_allocated += ( 8 - ( size & 0x7 ) );
	}
	if (_allocated > _end)
		return 0;
	if (_allocated > _resident) {
		unsigned count = CLBYTES;
		char * new_resident = _resident + CLBYTES;
		while (new_resident <= _allocated) {
			new_resident += CLBYTES;
			count += CLBYTES;
		}
		TheKernel->domain()->makeAddressable(_resident, count);
		if ( !_cacheit )
			SPARCTranslation::makeUncachable(_resident, count);
		_resident = new_resident;
	}
	return orig_allocated;
}

Zalloc * uncached_zalloc = 0;
Zalloc * cached_zalloc = 0;
Zalloc * mbuf_zalloc = 0;

void ss1zallocinit()
{
	uncached_zalloc = new Zalloc( 1024 * 1024, 0 );
	cached_zalloc = new Zalloc( 1024 * 1024, 1 );
	mbuf_zalloc = new Zalloc( 1024 * 1024, 1 );
	Assert( uncached_zalloc != 0 && cached_zalloc != 0 && mbuf_zalloc != 0 );
}

char * kernelmap_zalloc(unsigned int size, int duumy)
{
	return cached_zalloc->alloc( size );
}

char * kernelmap_zalloc_uncachable(unsigned int size, int duumy)
{
	return uncached_zalloc->alloc( size );
}

static struct mbuf * FreeBuf = 0;

#ifdef MBUF_DEBUG

struct mbuf_debug_info {
	unsigned _code;
	struct mbuf * _addr;
};

class mbuf_debug {
public:
	const int _size = 512;
	mbuf_debug_info _log[_size];
	int _head;
	mbuf_debug();
	void add(unsigned code, struct mbuf * addr);	
	void dump();
};

mbuf_debug::mbuf_debug()
{
	for (int i = 0; i < _size; i++) {
		_log[i]._code = 0;
		_log[i]._addr = 0;
	}
	_head = 0;
}

void mbuf_debug::add(unsigned code, struct mbuf * addr)
{
	_log[_head]._code = code;
	_log[_head]._addr = addr;
	_head++;
	if ( _head == _size )
		_head = 0;
}

void mbuf_debug::dump()
{
	Console() << "head = " << _head << "\n" << eor;;
	for (int i = _head; i < _size; i++ ) {
		Console() << " " << i << " " << _log[i]._code << " " 
			<< asAddress(_log[i]._addr) << "\n" << eor; 
	}
	for (i = 0; i < _head; i++ ) {
		Console() << " " << i << " " << _log[i]._code << " " 
			<< asAddress(_log[i]._addr) << "\n" << eor; 
	}
}

mbuf_debug mbd;

#endif MBUF_DEBUG

long int * malloc_mbuf( unsigned long size, int type )
{
	Assert( size == MSIZE );
	int s = splhigh();
	struct mbuf * m;
	int dummy;
	if ( FreeBuf ) {
		m = FreeBuf;
		FreeBuf = FreeBuf->m_next;
	} else {
		m = (struct mbuf *) mbuf_zalloc->alloc( MSIZE );
	}
#ifdef MBUF_DEBUG
	mbd.add( 1, m );
#endif
	splx(s);
	return (long int *) m;
}

void free_mbuf( char * p, int type )
{
	struct mbuf * m = (struct mbuf *) p;
	Assert( p >= mbuf_zalloc->base() && p < mbuf_zalloc->allocated() );
	int s = splhigh();
	m->m_next = FreeBuf;
	FreeBuf = m;
#ifdef MBUF_DEBUG
	mbd.add( 2, m );
#endif
	splx(s);
}

typedef char Eaddress[6];
typedef char Iaddress[4];

struct EtoIAddress {
	Eaddress e;
	char pad[2];
	Iaddress i;
} EtoITable[] = {
	{ { 0x8, 0x0, 0x20, 0x0b, 0x26, 0x73 }, {0, 0}, { 128, 174, 238, 31 } },
	{ { 0x8, 0x0, 0x20, 0x13, 0x17, 0x5e }, {0, 0}, { 128, 174, 238, 32 } },
	{ { 0x8, 0x0, 0x20, 0x0b, 0x09, 0x64 }, {0, 0}, { 128, 174, 238, 33 } },
	{ { 0x8, 0x0, 0x20, 0x0b, 0x0f, 0x4a }, {0, 0}, { 128, 174, 238, 34 } },
	{ { 0x8, 0x0, 0x20, 0x0b, 0x30, 0x58 }, {0, 0}, { 128, 174, 238, 35 } },
	{ { 0x8, 0x0, 0x20, 0x0b, 0x26, 0xda }, {0, 0}, { 128, 174, 238, 36 } },
	{ { 0x8, 0x0, 0x20, 0x0b, 0x38, 0x72 }, {0, 0}, { 128, 174, 238, 37 } },
	{ { 0x8, 0x0, 0x20, 0x0f, 0xc5, 0x3f }, {0, 0}, { 128, 174, 238, 38 } },
	{ { 0x8, 0x0, 0x20, 0x0f, 0x97, 0xb0 }, {0, 0}, { 128, 174, 238, 39 } }
};

const int EtoITableSize = sizeof(EtoITable) / sizeof(EtoIAddress);

const unsigned long defaultBroadcastAddr = 0x80aeee00;
const unsigned long defaultDstAddr = 0;
const unsigned long defaultNetmask = 0xffffff00;
const unsigned long defaultGateway = 0x80aeeef3;

unsigned long MyIPAddress = 0;

void ifconfig()
{
	Console() << "ifconfig()\n" << eor;
	char myEAddress[6];
	ethernetAddress(myEAddress);
	char * myIPAddress = 0;
	EtoIAddress * ep = EtoITable;
	for( int i = 0; i < EtoITableSize; i++, ep++ ) {
		if (ep->e[0] == myEAddress[0] &&
		    ep->e[1] == myEAddress[1] &&
		    ep->e[2] == myEAddress[2] &&
		    ep->e[3] == myEAddress[3] &&
		    ep->e[4] == myEAddress[4] &&
		    ep->e[5] == myEAddress[5]) {
			myIPAddress = ep->i;
			break;
		}
	}
	Assert( myIPAddress != 0 );
	MyIPAddress = *((unsigned long *) myIPAddress);
	struct ifreq ifr;

	struct socket *inso;
	int result = socreate(AF_INET, &inso, SOCK_DGRAM, IPPROTO_UDP);
	Assert(result == 0);
	Assert(inso != 0);

	struct sockaddr_in * mySockAddr = 
		(struct sockaddr_in *) &(ifr.ifr_addr);

#define INADDR_LOOPBACK 0x7f000001

	strcpy( ifr.ifr_name, "lo0" );
	bzero((char *) mySockAddr, 16);
        mySockAddr->sin_len = 16;
        mySockAddr->sin_family = AF_INET;
        mySockAddr->sin_port = 0;
        mySockAddr->sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	Console() << "ifconfig() lo0 SIOCSIFADDR " 
                  << asHex(ntohl(mySockAddr->sin_addr.s_addr))
                  << "\n" << eor;
        result = ifioctl( inso, SIOCSIFADDR, &ifr );
	Debug 	<< "ifconfig() SIOCSIFADDR result " 
		<< result
	      	<< "\n" << eor;

	strcpy( ifr.ifr_name, "le0" );
	bzero((char *) mySockAddr, 16);
	mySockAddr->sin_len = 16;
	mySockAddr->sin_family = AF_INET;
	mySockAddr->sin_port = 0;
	mySockAddr->sin_addr.s_addr = MyIPAddress;
	Console() << "ifconfig() le0 SIOCSIFADDR " 
		  << asHex(ntohl(mySockAddr->sin_addr.s_addr))
		  << "\n" << eor;
	result = ifioctl( inso, SIOCSIFADDR, &ifr );
	Debug 	<< "ifconfig() SIOCSIFADDR result " 
		<< result
	      	<< "\n" << eor;
/*
	struct sockaddr * myEtherAddr = 
		(struct sockaddr *) &(ifr.ifr_addr);
	bzero((char *) myEtherAddr, 16);
	myEtherAddr->sa_family = AF_UNSPEC;
	myEtherAddr->sa_len = 16;
	bcopy( myEAddress, myEtherAddr->sa_data, 6 );
	Console() << "ifconfig() le0 SIOCSIFADDR ethernet address" 
		  << "\n" << eor;
	result = ifioctl( inso, SIOCSIFADDR, &ifr );
	Debug 	<< "ifconfig() SIOCSIFADDR result " 
		<< result
	      	<< "\n" << eor;
 */
	struct sockaddr_in * dstAddr = 
		(struct sockaddr_in *) &(ifr.ifr_dstaddr);

/* 
 * DOES NOT APPLY FOR ETHERNET

	bzero((char *) dstAddr, 16);
	dstAddr->sin_len = 16;
	dstAddr->sin_family = AF_INET;
	dstAddr->sin_port = 0;
	dstAddr->sin_addr.s_addr = defaultDstAddr;
	Console() << "ifconfig() le0 SIOCSIFDSTADDR " 
		  << asHex(dstAddr->sin_addr.s_addr)
	      	  << "\n" << eor;
	result = ifioctl( inso, SIOCSIFDSTADDR, &ifr );
	Debug 	<< "ifconfig() SIOCSIFDSTADDR result " 
		<< result
	      	<< "\n" << eor;
 */

	struct sockaddr_in * broadcastAddr = 
		(struct sockaddr_in *) &(ifr.ifr_broadaddr);
	bzero((char *) broadcastAddr, 16);
	broadcastAddr->sin_len = 16;
	broadcastAddr->sin_family = AF_INET;
	broadcastAddr->sin_port = 0;
	broadcastAddr->sin_addr.s_addr = defaultBroadcastAddr;
	Console() << "ifconfig() le0 SIOCSIFBRDADDR " 
		  << asHex(broadcastAddr->sin_addr.s_addr)
	      	  << "\n" << eor;
	result = ifioctl( inso, SIOCSIFBRDADDR, &ifr );
	Debug 	<< "ifconfig() SIOCSIFBRDADDR result " 
		<< result
	      	<< "\n" << eor;

	struct sockaddr_in * netmaskAddr = 
		(struct sockaddr_in *) &(ifr.ifr_addr);
	bzero((char *) netmaskAddr, 16);
	netmaskAddr->sin_len = 16;
	netmaskAddr->sin_family = AF_INET;
	netmaskAddr->sin_port = 0;
	netmaskAddr->sin_addr.s_addr = defaultNetmask;
	Console() << "ifconfig() le0 SIOCSIFNETMASK " 
		  << asHex(netmaskAddr->sin_addr.s_addr)
	      	  << "\n" << eor;
	result = ifioctl( inso, SIOCSIFNETMASK, &ifr );
	Debug 	<< "ifconfig() SIOCSIFNETMASK result " 
		<< result
	      	<< "\n" << eor;

	ifr.ifr_flags = IFF_DEBUG | IFF_UP | IFF_BROADCAST | IFF_NOTRAILERS |
			IFF_RUNNING;
/*
	ifr.ifr_flags = IFF_DEBUG | IFF_UP | IFF_BROADCAST | IFF_NOTRAILERS |
			IFF_RUNNING | IFF_PROMISC;
 */
	Console() << "ifconfig() SIOCIFFLAGS " 
		  << asHex(ifr.ifr_flags) 
	      	  << "\n" << eor;
	result = ifioctl( inso, SIOCSIFFLAGS, &ifr );
	Debug 	<< "ifconfig() SIOCIFFLAGS result " 
		<< result
	      	<< "\n" << eor;

	soclose(inso);

#ifdef ROUTE
	/* don't really need, ifioctl SIOCSIFADDR sets it up automatically */

	/* set up route table entries */

	Console() << "ifconfig() adding routing table entries\n" << eor;

	struct sockaddr_in ifa_addr;
	struct sockaddr_in ifa_dstaddr;
	struct sockaddr_in ifa_netmask;
	struct ifaddr ifa = 
		{ (struct sockaddr *) &ifa_addr, 
		  (struct sockaddr *) &ifa_dstaddr, 
		  (struct sockaddr *) &ifa_netmask, 
		  0, 0, 0, 0, 0, 0 };

	struct sockaddr_in * gatewayAddr = &ifa_addr;
	bzero((char *) gatewayAddr, 16);
	gatewayAddr->sin_len = 16;
	gatewayAddr->sin_family = AF_INET;
	gatewayAddr->sin_port = 0;
	gatewayAddr->sin_addr.s_addr = 0;

	dstAddr = &ifa_dstaddr;
	bzero((char *) dstAddr, 16);
	dstAddr->sin_len = 16;
	dstAddr->sin_family = AF_INET;
	dstAddr->sin_port = 0;
	dstAddr->sin_addr.s_addr = 0;

	netmaskAddr = &(ifa_netmask);
	bzero((char *) netmaskAddr, 16);
	netmaskAddr->sin_len = 16;
	netmaskAddr->sin_family = AF_INET;
	netmaskAddr->sin_port = 0;
	netmaskAddr->sin_addr.s_addr = 0;

	gatewayAddr->sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	dstAddr->sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	netmaskAddr->sin_addr.s_addr = 0xffffffff;
	ifa.ifa_flags = RTF_GATEWAY | RTF_UP;
	Debug	<< "ifconfig() adding loopback route\n" << eor;
	result = rtinit(&ifa, RTM_ADD, 0);
	Console() << "ifconfig route add lo0 dest "
		  << asHex(dstAddr->sin_addr.s_addr)
		  << " gateway "
		  << asHex(gatewayAddr->sin_addr.s_addr)
		  << " result " << result << "\n" << eor;

	gatewayAddr->sin_addr.s_addr = *((unsigned long *) myIPAddress);
	dstAddr->sin_addr.s_addr = defaultBroadcastAddr;
	netmaskAddr->sin_addr.s_addr = defaultNetmask;
	ifa.ifa_flags = RTF_GATEWAY | RTF_UP;
	Debug	<< "ifconfig() adding le0 route\n" << eor;
	result = rtinit(&ifa, RTM_ADD, 0);
	Console() << "ifconfig route add le0 dest "
		  << asHex(dstAddr->sin_addr.s_addr)
		  << " gateway "
		  << asHex(gatewayAddr->sin_addr.s_addr)
		  << " result " << result << "\n" << eor;

	gatewayAddr->sin_addr.s_addr = *((unsigned long *) myIPAddress);
	// gatewayAddr->sin_addr.s_addr = defaultGateway;
	dstAddr->sin_addr.s_addr = 0;
	netmaskAddr->sin_addr.s_addr = 0;
	ifa.ifa_flags = RTF_GATEWAY | RTF_UP;
	Debug 	<< "ifconfig() adding default gateway " 
		<< asHex(gatewayAddr->sin_addr.s_addr)
	      	<< "\n" << eor;
	result = rtinit(&ifa, RTM_ADD, 0);
	Console() << "ifconfig route add dest "
		  << asHex(dstAddr->sin_addr.s_addr)
		  << " (default) gateway "
		  << asHex(gatewayAddr->sin_addr.s_addr)
		  << " result " << result << "\n" << eor;
	
#endif // ROUTE
}
#endif /* SS1 */
