#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * Copyright (c) 1990 IBM Corporation. All Rights reserved
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Author: Panos Kougiouris ( kougiour@cs.uiuc.edu )
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * AIXContainer.c - 
 *
 */
#include "AIXAllocator.h"
#include "AIXContainer.h"
#include "AIXInode.h"
#include "Assert.h"
#include "Class.h"
#include "FileObject.h"
#include "FileSystem.h"
#include "Raid.h"
#include "UNIXFile.h"

#ifdef __GNUC__
#pragma implementation
#endif

extern ClassStar MemoryObjectDictionaryClass;

static const int SUPERBLOCK = 1;

static const int FsOKAY = 0x7c269d38;		/* s_state: clean */
static const int FsACTIVE = 0x5e72d81a;		/* s_state: active */
static const int FsBAD = 0xcb096f43;		/* s_state: bad root */

AIXContainer::AIXContainer( PersistentMemoryObject * source )
: ( source )
{
    initDebug();
    Raid( RaidConstructor )( "%N::AIXContainer( %N )\n", this, source );

    /*
     *	Calculate constants needed by UNIXContainer.
     */
    _log2FragmentToSector = 3;		
    _dinodeSize = sizeof( AIXdinode );
    _inodeBlockSize = _source->unitSize() << _log2FragmentToSector;
    _inodesPerBlock = _inodeBlockSize / _dinodeSize;

    /*
     *	Create super block buffer, chainize it, read it from the disk,
     *	and check its magic number.
     */
    _superBlock = (aixfilsys *) new char[ _inodeBlockSize ];
    _source->read( SUPERBLOCK << _log2FragmentToSector, 
		1 << _log2FragmentToSector, (char *) _superBlock, 4096 );
    if( ( _superBlock->s_magic != FaMAGIC ) ) {
	Raid( RaidError )( "_superBlock->s_magic %x\n", _superBlock->s_magic );
	Assert( ( _superBlock->s_magic == FaMAGIC ) );
    }


    /*
     *	Allocate and clear in memory and on disk inode pointers.
     *	Create one more object than there are inodes because object
     *	0 will be unused since inumbers start at 1.
     */
    int containerSize = ( ( _superBlock->s_isize - 2 ) * _inodesPerBlock ) + 1;
    UNIXContainer::initializeArrays( containerSize );
    _allocator = new AIXAllocator( _source, _superBlock,
				    _log2FragmentToSector );
}

AIXContainer::~AIXContainer()
{
    Raid( RaidConstructor )( "%N::~AIXContainer\n", this );

    synchronize();

    delete _superBlock;
}

/*
 * Find block number containing this inode.
 */
int
AIXContainer::mapInumber( int inumber )
{
    return( 2 + ( ( inumber - 1 ) / _inodesPerBlock ) );
}

/*
 * Find first inode in block containing this inode.
 */
int
AIXContainer::firstInumber( int inumber )
{
    return( ( ( ( inumber - 1 ) / _inodesPerBlock ) * _inodesPerBlock ) + 1 );
}

/*
 *	Find an unused inode.
 *	Returns the inumber of the inode found, or zero if none found.
 *	The inode is marked in use by this function.
 */
int
AIXContainer::getUnusedInode( int )
{
    Raid( RaidMethod )( "AIXContainer::getUnusedInode()\n" );

    int inumber = 0;
    if( _superBlock->s_tinode != 0 ) {
	
	// get a free inode from the i-list if there is one
	if( _superBlock->s_ninode > 0 ) {
	    inumber = _superBlock->s_inode[--_superBlock->s_ninode];
	    _superBlock->s_tinode--;
	}

	// otherwise go through all inodes on the disk, placing free
	// ones ( up to 100 ) into the s_inode array ( inodes 0 and 1
	// are exceptions ).
	else {

      	    // inodes begin at block 2  
      	    int chk_inode_blk = 2;

      	    // add free inodes to the free list until it is full or there
       	    // are no more 
      	    while( ( _superBlock->s_ninode < AIXNICINOD ) && 
                ( chk_inode_blk < ( _superBlock->s_isize ) ) ) {
       
  		// if this block of inodes is in memory then use those else 
	 	// read them from disk  
		int first_inode_in_blk = ( ( chk_inode_blk-2 ) * _inodesPerBlock )+1;
		if( _diskInodes[first_inode_in_blk] == 0 ) 
        		readDinode( first_inode_in_blk );

        	// check them and add free ones to list ( skip 1 and 2 ) 
        	int count = first_inode_in_blk;
        	if( chk_inode_blk == 2 ) count = count + 2;
        	while( count < first_inode_in_blk + _inodesPerBlock ) {
      
	  	    // point to the next AIXdinode in the buffer 
	  	    AIXdinode *nextDinode = (AIXdinode *) _diskInodes[count];
		    Assert( nextDinode != 0 );
 
          	    // add to free list if there is room and it is free 
          	    if( ( _superBlock->s_ninode<AIXNICINOD ) && ( nextDinode->di_nlink==0 ) )
	    		_superBlock->s_inode[_superBlock->s_ninode++] = count + 1;
          	    count++;
   		}
		chk_inode_blk++;
      	    }

      	    // now get a free inode from the i-list and update count 
      	    if( _superBlock->s_ninode > 0 ) {
        	inumber = _superBlock->s_inode[--_superBlock->s_ninode];
		_superBlock->s_tinode--;
      	    }
	}

	if ( inumber < _superBlock->s_lasti ) _superBlock->s_nbehind--;

	// The superblock is now modified.
	_modified = 1;
	_superBlock->s_time = 0;
    }

    Raid( RaidMethod )( "AIXContainer::getUnusedInode, return = %d\n", inumber );
    return( inumber );
}

/*
 *	Open returns a pointer to a sub-memory-object ( inode )
 *	with the given number.
 *	If the inumber is negative, an unused one is allocated an initialized.
 *	Also, the new inumber is passed back thru the reference parameter.
 *	If the inode is already in use,
 *	update its refernce count.
 *	Otherwise create a new in memory inode and return a pointer to it.
 *	Creating a new memory inode will also read in the disk inode if needed.
 */
FileObject *
AIXContainer::basicOpen( int inumber, int & status )
{
    Raid( RaidMethod )( "%N::basicOpen( inumber %d )\n",  this, inumber );

    if( inumber < basicRootId() ) {
	status = IdNumberOutOfRange;
	return( 0 );
    }

    /*
     *	Read this inode from the disk,
     *	Check to see if it exists,
     *	and Construct its in memory structure.
     */
    acquire();
    _fileObjects[ inumber ] = (FileObject *) ~0;
    AIXdinode * thisDinode = (AIXdinode *) readDinode( inumber );

    if( thisDinode->di_nlink == 0 ) {
	status = UnusedIdNumber;
	_fileObjects[ inumber ] = 0;
	writeDinode( inumber, 1 );
	release();
	return( 0 );
    }
    release();

    status = 0;
    return( new AIXInode( numFragments( thisDinode->di_size ),
			  _source->log2UnitSize() + _log2FragmentToSector,
			  inumber, this, _source, thisDinode ) );
}

/*
 *	create returns a pointer to a new sub-memory-object ( inode )
 *	The inumber passed in is just used as a hint to the file system.
 *	Also, the new inumber is passed back thru the same parameter.
 *	The mode parameter is primarily used to set the type of the
 *	file, it should not change during the life of the file.
 *	Uid, gid, and file permissions can be set via methods on an Inode.
 */
FileObject *
AIXContainer::basicCreate( int inumber, Class * aClass, int & status )
{
    Raid( RaidMethod )( "%N::basicCreate( inumber %d, %N, )\n",
			this, inumber, aClass );

    acquire();
    inumber = getUnusedInode( 0 );
    if( inumber <= 0 ) {
	status = ContainerFull;
	release();
	return( 0 );
    }

    /*
     *	Read the inode from disk, and reset all of its fields.
     */

    _fileObjects[ inumber ] = (FileObject *) ~0;
    AIXdinode *thisDinode = (AIXdinode *) readDinode( inumber );

    thisDinode->di_nlink = 0;
    thisDinode->di_mode = UNIXAccessMask;
    if( aClass->isASubclassOf( MemoryObjectDictionaryClass ) ) {
	thisDinode->di_mode |= UNIXTypeDirectory;
	thisDinode->di_dflag = DIALLOC | DILONGDIR | DISTORE;
    }
    else {
	thisDinode->di_mode |= UNIXTypeRegular;
	thisDinode->di_dflag = DIALLOC | DISTORE;
    }
    thisDinode->di_uid = 0;
    thisDinode->di_gid = 0;
    thisDinode->di_size = 0;
    // FIX: what time is it?
    thisDinode->di_atime = 
    thisDinode->di_mtime = 
    thisDinode->di_ctime = 0;
    for( int i = 0; i < 40; i++ ) thisDinode->di_addr[i] = 0;

    /*
     *	Create the new in memory inode.
     */
    status = 0;

    setModified( inumber );

    release();

    return( new AIXInode( numFragments( thisDinode->di_size ),
			  _source->log2UnitSize() + _log2FragmentToSector,
			  inumber, this, _source, thisDinode ) );
}

/*
 */
void
AIXContainer::basicClose( FileObject * thisInode )
{
    Raid( RaidMethod )( "%N::basicClose( %N )\n", this, thisInode );

    int inumber = thisInode->idNumber();

    Raid( RaidMethod )( "%N::basicClose() inumber %d \n", this, inumber );

    /*
     *	If their are no links to this inode, then this call will free its space.
     *	FIX: this isnt ready yet.
     */
    int links = thisInode->links( 0 );

    Raid( RaidMethod )( "%N::basicClose() links %d \n", this, links );

    MemoryObjectContainer::basicClose( thisInode );

    /*
     *	Write the disk inode back to disk.
     */

    acquire();
    writeDinode( inumber, 1 );

    if( links == 0 ) {
	// put inode in superblock free list if there is room
	if( _superBlock->s_ninode < AIXNICINOD )
	    _superBlock->s_inode[_superBlock->s_ninode++] = inumber;

	if ( inumber < _superBlock->s_lasti ) _superBlock->s_nbehind++;

	// update the free count
	_superBlock->s_tinode++;
	_modified = 1;
	_superBlock->s_time = 0;
    }
    release();
}

int
AIXContainer::basicRootId()
{
    return( 2 );
}

/*
 *	Synchronize the in-core data structures with the disk.
 */
void
AIXContainer::basicSynchronize()
{
    Raid( RaidMethod )
	    ( "%N->AIXContainer::basicSynchronize() %d\n", this, _modified );

    _superBlock->s_fmod = 0;
    _source->write( SUPERBLOCK << _log2FragmentToSector, 
		1 << _log2FragmentToSector, (char *) _superBlock, 4096 );

    basicSynchronizeInodes();
}

ProxiableObject * 
AIXContainerConstructor( ProxiableObject * source )
{
    if( source->isKindOf( PersistentMemoryObjectClass ) ) {
        return( new AIXContainer( (PersistentMemoryObject *)source ) );
    }
    else return( 0 );
}

int
AIXContainer::numFragments( int size )
{
    /* Other code crashes if this value is 0 */
    //if( size <= SMBLKSZ ) return 0;
    int fragmentSize = _source->unitSize() << _log2FragmentToSector;
    return( ( size + fragmentSize - 1 ) / fragmentSize );
}


#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("AIXContainer",AIXContainer,UNIXContainer,&AIXContainerConstructor,0);

