#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Contributing Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * AIXDirectory.c - 
 */
#include "AIXDirectory.h"
#include "AIXInode.h"
#include "Assert.h"
#include "Byte.h"
#include "FileSystem.h"
#include "PersistentMemoryObject.h"
#include "Raid.h"
#include "String.h"

#ifdef __GNUC__
#pragma implementation
#endif

class aixdirect {
	friend class AIXDirectory;

	int	d_ino;
	short	d_reclen;
	short	d_namlen;
	char	d_name[MaxEntrySize + 1];
};

/*
 * AIXDirectory allows directory entries to be located in,
 * added to, and removed from a AIX directory file.
 */
AIXDirectory::AIXDirectory( PersistentMemoryObject * source )
: ( source )
{
    initDebug();
    Raid( RaidConstructor )( "%N::AIXDirectory( %N )\n", this, source );
    Raid( RaidDetail )( "%N::AIXDirectory idNumber %d _size %d\n",
	     this, source->idNumber(), _size );

     _maxKeyLength = MaxEntrySize;

    if( _size > SMBLKSZ ){
    	/* make directory file size a multiple of the directory size */
    	int roundedSize = ( _size + DirectoryBlockSize - 1 ) / 
			  DirectoryBlockSize * DirectoryBlockSize;
    	if( _size != roundedSize ) {
	_size = roundedSize;
	source->setNumberOfBytes( _size );
    	}
    }
    if( _size == 0 ) {
	add( ".", source );
    }
}

AIXDirectory::~AIXDirectory()
{
    Raid( RaidConstructor )( "%N::~AIXDirectory\n size: %d", this, _size );
}

/*
 * search the entire directory for an occurrence of key
 * if found, return 1; otherwise 0
 * side effects: set _previousOffset & _offset dir entry ofs and
 * findKey will also try to find an _empty space big enough 
 * to hold a new directory entry of _size _size.
 */
int
AIXDirectory::findKey( char * key, int needSize ) 
{
    aixdirect * dp;

    Raid( RaidMethod )( "%N::findKey( key %s, needSize %d )\n",
		this, key, needSize );

    int StillLooking = ( needSize > 0);
    _emptyOffset = -1;

    _previousOffset = 0;
    for( _offset = 0; _offset < _size; _offset += dp->d_reclen ) {
        dp = ( aixdirect * ) &_buffer[_offset];

	Raid( RaidDetail )( "%N::findKey ofs %d ", this, _offset); 
	Raid( RaidDetail )( "ino %d rl %d nl %d key %s\n",
		   dp->d_ino, dp->d_reclen, dp->d_namlen, dp->d_name );

	Assert( dp->d_reclen > 0 );

	/* check to see if hole is big enough */
        if( StillLooking ) {
            int sp = dp->d_reclen;
            if( dp->d_ino != 0 ) sp -= entrySize( dp->d_namlen );
            if( sp >= needSize ) {
                _emptyOffset = _offset;
                StillLooking = 0;
        	Raid( RaidDetail )( "%N::findKey space @ %d\n", this, _offset );
            }
        }

	/* return if key is found */
        if( ( dp->d_ino != 0 ) && ( strcmp( key, dp->d_name ) == 0 ) ) {
	    _idNumberOfKey = dp->d_ino;
            Raid( RaidDetail )( "%N::findKey found @ %d\n", this, _offset );
            return( 1 );
	}
	_previousOffset = _offset;
    }
    return( 0 );
}

int
AIXDirectory::insertEntry( char * key, int keyLength, int idNumber  )
{
    aixdirect * edp;
    aixdirect dir;

    Raid( RaidMethod )( "%N::insertEntry( key %s, keyLength %d, idNumber %d )\n",
		this, key, keyLength, idNumber );

    /* set up the new entry ( for _size estimation ) */
    dir.d_ino = idNumber;
    dir.d_namlen = keyLength ;

    if( _emptyOffset < 0 ) {
        dir.d_reclen = DirectoryBlockSize ;
        _offset = _size;
        _size += DirectoryBlockSize ; 
        _source->setNumberOfBytes( _size );
        int newUnits = _source->numberOfUnits();
	if( _units != newUnits )  {
	    reMap();
	    _units = newUnits;
	}
    }
    else {
        edp = ( aixdirect * ) &_buffer[_emptyOffset];
        int bytesInUse = 0;
        if( edp->d_ino != 0 ) bytesInUse = entrySize( edp->d_namlen );
        dir.d_reclen = edp->d_reclen - bytesInUse;
        edp->d_reclen = bytesInUse;
	_offset = _emptyOffset + bytesInUse;
    }

    Raid( RaidDetail )( "%N::insertEntry _offset %d _emptyOffset %d\n",
		this, _offset, _emptyOffset );

    /* move new entry into the blk */
    edp = ( aixdirect * ) &_buffer[_offset];
    for( int i = 0 ; i < entrySize( keyLength ) ; i++ )
		_buffer[ _offset+i ] = 0;
    edp->d_ino    = dir.d_ino;
    edp->d_reclen = dir.d_reclen;
    edp->d_namlen = dir.d_namlen;
    strcpy( edp->d_name, key );

    int retVal = writeChanges();
    _source->setNumberOfBytes( this->_size );
    return( retVal );

}

/*
 */
int
AIXDirectory::clearEntry()
{
    Raid( RaidMethod )( "%N::clearEntry()\n", this );

    aixdirect * cdp = ( aixdirect * ) &_buffer[_offset];

    if( ( _offset & ( DirectoryBlockSize - 1 ) ) == 0 ) cdp->d_ino = 0;
    else {
	aixdirect * pdp = ( aixdirect* ) &_buffer[_previousOffset];
        pdp->d_reclen += cdp->d_reclen;
	
    }
    /* Ok now Null pad the entry */
    char * p = (char *) cdp;
    for( int i = cdp->d_reclen; i > 0 ; i-- ) 
    	*(p++) = 0;

    int retVal = writeChanges();
    _source->setNumberOfBytes( this->_size );
    return( retVal );
}

/*
 *	return positive count of keys if the buffer was big enough,
 *	negative count otherwise.
 *	optimize and improve later, perhaps via memory-mapping a precalculated
 *	list-o-keys.
 */
int
AIXDirectory::basicKeys( char * keyBuffer, int bufferSize, int & count ) 
{
    Raid( RaidMethod )( "%N::keys( buffer %x, bufferSize %d ) _size %d\n",
		this, keyBuffer, bufferSize, _size );

    count = 0;
    int limit = ( bufferSize >= _size ) ? _size : bufferSize;
    char * t = keyBuffer;
    aixdirect * dp = ( aixdirect * ) _buffer;
    for( _offset = 0; _offset < limit; _offset += dp->d_reclen ) {

        dp = ( aixdirect * ) &_buffer[_offset];

	Raid( RaidDetail )( "%N::keys ofs %d ", this, _offset); 
	Raid( RaidDetail )( "ino %d rl %d nl %d key %s\n",
		   dp->d_ino, dp->d_reclen, dp->d_namlen, dp->d_name );

	Assert( dp->d_reclen > dp->d_namlen );

	if( dp->d_ino != 0) {
		Assert( dp->d_namlen > 0 );

		if( t != 0 ) {
			for( char * s = dp->d_name; *s; ) *t++ = *s++; 
			*t++ = '\n';
		}
		count++;
	}
    }
    if( t != 0 ) *t = '\0';
    if( limit == _size ) return( 0 );
    else return( TooManyKeys );
}

int
AIXDirectory::entrySize( int keyLength )
{
    return( ( keyLength + 15 + sizeof( aixdirect ) - MaxEntrySize ) & ~15 );
}

int
 AIXDirectory::hint()
{
	return( _source->idNumber() );
}

ProxiableObject *
AIXDirectoryConstructor( ProxiableObject * source )
{
        if( source->isKindOf( PersistentMemoryObjectClass ) ) {
                return( new AIXDirectory( (PersistentMemoryObject *)source ) );
        }
        else return( 0 );
}


#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("AIXDirectory",AIXDirectory,MemoryObjectDictionary,&AIXDirectoryConstructor,0);
//DEFINECLASS("AIXDirectory",AIXDirectory,MemoryObjectDictionary,&AIXDirectoryConstructor,0);
