#ifdef __GNUC__
#pragma implementation
#endif

/* This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * Copyright (c) 1990 IBM Corporation. All Rights reserved
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Contr. Author: Panos Kougiouris ( kougiour@cs.uiuc.edu )
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * AIXInode.c - 
 */
#include "AIXInode.h"
#include "Assert.h"
#include "Class.h"
#include "FileSystem.h"
#include "MemoryObjectAsync.h"
#include "MemoryObjectContainer.h"
#include "Raid.h"
#include "UNIXFile.h"

#ifdef __GNUC__
#pragma implementation
#endif

extern ClassStar PersistentArrayClass;
extern ClassStar AIXDirectoryClass;

/* 
 * Initialize the in-memory inode
 */
AIXInode::AIXInode( unsigned int numberOfUnits,
		    int log2UnitSize,
		    int inumber,
		    MemoryObjectContainer * container,
		    PersistentMemoryObject * source,
		    AIXdinode * thisDinode )
: ( numberOfUnits, log2UnitSize, inumber, container, source, 0 )
{
    initDebug();
    Raid( RaidConstructor )( "%N::AIXInode()\n", this );

    /*
     *	Initialize variables from parameters
     */
    _diskInode = thisDinode;

    /*
     *	Set up indirect variables for mapUnit function.
     */
    singleIndirect = 10;
    doubleIndirect = singleIndirect + ( 1 << indexShift );
    tripleIndirect = doubleIndirect + ( 1 << ( indexShift + indexShift ) );
}

Class *
AIXInode::supports( Class * aClass )
{
    Raid( RaidProtected )( "%N::supports( %N )\n", this, aClass );

    Class * supportedClass = 0;

    if( ( _diskInode->di_mode & UNIXTypeMask ) == UNIXTypeDirectory ) {
	if( AIXDirectoryClass->isASubclassOf( aClass ) ) {
		supportedClass = AIXDirectoryClass;
	}
    }
    else if( PersistentArrayClass->isASubclassOf( aClass ) ) {
        supportedClass = PersistentArrayClass;
    }


    Raid( RaidProtected )( "%N::supports() returns %N\n",
                           this, supportedClass );

    return( supportedClass );
}

int
AIXInode::isDictionary()
{
    return ( ( _diskInode->di_mode & UNIXTypeMask ) == UNIXTypeDirectory );
}

/*
 * Clean up the in memory inode data
 */
AIXInode::~AIXInode()
{
    Raid( RaidConstructor )( "%N::~AIXInode()\n", this );

    if( _diskInode->di_nlink == 0 ) {
	_diskInode->di_mode = 0;
	_diskInode->di_dflag = DIDEL;
        setNumberOfBytes( 0 );
	setModified();
    }

    setNumberOfBytes( _diskInode->di_size );
    
    if( _diskInode->di_nlink == 0 ) {
	char * p =  (char *) _diskInode;
	for( int i = 0 ; i < 512 ; i++)
		*(p++) = 0;
	setModified();
    }
}


int
AIXInode::numberOfBytes()
{
    Raid( RaidMethod )( "%N::numberOfBytes() %d\n", this, _diskInode->di_size );
    return( _diskInode->di_size );
}

int
AIXInode::basicSetNumberOfBytes( int size )
{
    Raid( RaidProtected )( "%N::basicSetNumberOfBytes( %d ) %d\n",
		this, size, _diskInode->di_size );

    if( size < 0 ) size = 0;
    if( size != _diskInode->di_size ) setModified();
    _diskInode->di_size = size;

    /* AIX different here too */
    if( size <= SMBLKSZ )  _diskInode->di_sbflag = 1;
    else _diskInode->di_sbflag = 0;
	
    if ( _diskInode->di_sbflag  ){
	basicSetNumberOfUnits( 0 );
	MemoryObject::basicSetNumberOfUnits( 1 );
    }
    else{
    	short numberOfUnits = ( size + unitSize() - 1 ) >> log2UnitSize();
    	int result = basicSetNumberOfUnits( numberOfUnits );
	Assert( result == numberOfUnits );
    }
    return( size );
}

/*
 *	this function allows other objects to access and modify the
 *	link count ( or reference count ) of this inode.
 *	links returns the new number of links for this inode.
 */
int
AIXInode::links( int delta )
{
    Raid( RaidMethod )( "%N::links( delta %d )\n", this, delta );
    _mutex.P();
    _diskInode->di_nlink += delta;
    int result = _diskInode->di_nlink;
    _mutex.V();
    if ( delta ) setModified();
    Raid( RaidMethod )( "%N::links returns %d\n", this, result );
    return( result );
}

int
AIXInode::setMode( int mode )
{
    Raid( RaidMethod )( "%N::setMode( mode %o ) %o\n",
		this, mode, _diskInode->di_mode );
    _mutex.P();
    _diskInode->di_mode &= ~07777;
    _diskInode->di_mode |= mode & 07777;
    _mutex.V();
    setModified();
    return( 0 );
}

int
AIXInode::setUserId( int uid )
{
    Raid( RaidMethod )( "%N::setUserId( uid %d ) %d\n",
		this, uid, _diskInode->di_uid );
    _diskInode->di_uid = uid;
    setModified();
    return( 0 );
} 

int
AIXInode::setGroupId( int gid )
{
    Raid( RaidMethod )( "%N::setGroupId( gid %d ) %d\n",
		this, gid, _diskInode->di_gid );
    _diskInode->di_gid = gid;
    setModified();
    return( 0 );
} 

/* 
 * This does the UNIX Stat( 2 ) function.
 */
void
AIXInode::info( UNIXFileInfo * infoBuffer )
{
    Raid( RaidMethod )( "%N::info( %x ) %d\n", this, infoBuffer, _idNumber );
    infoBuffer->device	= 0;
    infoBuffer->inumber	= _idNumber;
    infoBuffer->mode	= _diskInode->di_mode;
    infoBuffer->links	= _diskInode->di_nlink;
    infoBuffer->uid	= _diskInode->di_uid;
    infoBuffer->gid	= _diskInode->di_gid;
    infoBuffer->size	= _diskInode->di_size;
    infoBuffer->atime	= _diskInode->di_atime;
    infoBuffer->mtime	= _diskInode->di_mtime;
    infoBuffer->ctime	= _diskInode->di_ctime;
    infoBuffer->blockSize = blockSize() << log2UnitSize(); 
    infoBuffer->blocks	= _diskInode->di_blocks; 
    infoBuffer->flags	= 0;
    infoBuffer->rdevice	= 0;
}

/* 
 * Allocate 1 block.
 */
int 
AIXInode::allocBlock()
{
	Raid( RaidMethod )( "%N::allocBlock() id %d\n", this, idNumber() );
	_diskInode->di_blocks++;
	setModified();
	return( _container->allocate( 0, 0, 1 ) );
}

/* 
 * Free 1 physical block.
 */
void
AIXInode::freeBlock( int physicalBlock )
{
	Raid( RaidMethod )( "%N::freeBlock( physicalBlock %d ) id %d\n",
		this, physicalBlock, idNumber() );
	_diskInode->di_blocks--;
	setModified();
	_container->free( 0, physicalBlock & 0x00FFFFFF, 1 );
}

/*
 * Get the physical block number of the indicated direct block pointer
 * from the disk inode. AIX is different from the SVID here.
 */
int 
AIXInode::getDirect( int index )
{
	Raid( RaidMethod )( "%N::getDirect( index %d )\n", this, index );
	Assert( index >= 0 );
	Assert( index < 10 );

	return( _diskInode->di_addr[index] );
}

/*
 * Set the physical block number of the indicated direct block pointer
 * in the disk inode.
 */
void
AIXInode::setDirect( int index, int value )
{
	Raid( RaidMethod )( "%N::setDirect( index %d, value %d )\n",
		   this, index, value );
	Assert( index >= 0 );
	Assert( index < 10 );

	_diskInode->di_addr[index] = value;
}

/*
 * Get the physical block number of the indicated indirect block pointer
 * in the disk inode.
 */
int
AIXInode::getIndirect( int level )
{
	Assert( level >= 1 );
	Assert( level <= 3 );
 	return(	_diskInode->di_addr[level+9] );
}

/*
 * Set the physical block number of the indicated indirect block pointer
 * in the disk inode.
 */
void
AIXInode::setIndirect( int level, int value )
{
	Assert( level >= 1 );
	Assert( level <= 3 );

	_diskInode->di_addr[level+9] = value;
}

int
AIXInode::accessMode()
{
    return( _diskInode->di_mode & UNIXAccessMask );
}

int
AIXInode::userId()
{
    return( _diskInode->di_uid );
}

int
AIXInode::groupId()
{
    return( _diskInode->di_gid );
}


extern "C" { void PhysCopyBlock( unsigned char *, unsigned char *, int );}
/* In order to improve performance AIX stores small files inside the Inode.
 * So we take care of it here.
 */

int 
AIXInode::rawWrite( unsigned startUnit, int count, char * addr, int nbytes,
		    int writeThrough, IOAsyncClosure * c )
{
	_mutex.P();

	if( _diskInode->di_sbflag ){
    	         /* Here we are!!! The problem is that the
                 * chain contains a physical address which we
                 * cannot access because the MMU is on. What we
                 * do is we translate the address of the inode
                 * data to physical address and then we call
                 * a physical memory copy function.
                 */

		 _diskInode->di_sbflag = 0;
		 int firstDirect = this->allocBlock();
		 if( firstDirect == 0 ) return 0;
		
		 this->setDirect( 0, firstDirect );

		 /* Now copy everything to the new block */
		 
		 char * p = new char[unitSize()];
		 for(int i = 0; i < SMBLKSZ; i++)
			p[i] = _diskInode->di_sbbuf[i];
		 for( i = SMBLKSZ ; i < unitSize(); i++)
			p[i] = 0;
		 
		 fileObjectWrite( firstDirect, 1, p, unitSize(), 
				  writeThrough, 0 );

		 delete p;
	}

	if( startUnit == 0 ){
		 for(int i = 0; i < SMBLKSZ; i++)
			_diskInode->di_sbbuf[i] = addr[i];
		
		 _modified = 1;
	}

	_mutex.V();
	return UNIXInode::rawWrite( startUnit, count, addr, nbytes, 
				    writeThrough, c );

}

/* In order to improve performance AIX stores small files inside the Inode.
 * So we take care of it here.
 */
int 
AIXInode::rawRead( unsigned startUnit, int count, char * addr, int nbytes,
		   IOAsyncClosure * c )
{
	_mutex.P();
	if( _diskInode->di_sbflag ){
		 for(int i = 0; i < SMBLKSZ; i++)
			addr[i] = _diskInode->di_sbbuf[i];
		 for( i = SMBLKSZ ; i < unitSize(); i++)
			addr[i] = 0;
		 
		 _mutex.V();
		 if ( c ) {
			c->activate( count );
		 }	
		 return( count );
	} else {
		 _mutex.V();
		 return UNIXInode::rawRead( startUnit, count, addr, nbytes, c );
	}
}


/* Here is another problem in the AIX PS/2 1.2 filesystems. The block addresses
 * inside the first indirect block have the first byte 1. For example the
 * block with address 0x0000008a appears to have address 0x0100008a. I do not
 * know why. The only thing I could do for the time being is to mask out this
 * 0x01
 */
int
AIXInode::basicMapUnit( int logicalBlock, int addMapping )
{
	return( UNIXInode::basicMapUnit( logicalBlock, addMapping ) & 0x00FFFFFF );
}

/* Copy has a small problems for files that reside completely inside the
 * inode. Essentially what we do here is the following:
 * We handle the small case difeerently, otherwiae we call the
 * superclass copy to take care of it.
 */
int
AIXInode::basicCopy( MemoryObject * dest )
{
	Raid( RaidProtected )( "%N::basicCopy( %N )\n", this, dest );

	int returnValue = 0;

	_mutex.P();

	/* If it is the small case */
	if( _diskInode->di_sbflag ){
		/* This assertion should hold for this implementation.
		 * If it does not hold we want a different implementation
		 * that breaks the SMBLKSZ bytes in dest->unitSize().
		 */
		Assert( dest->unitSize() > SMBLKSZ);

		dest->setNumberOfUnits( 0 );
		/* Copy the contents of the file to a buffer */
                 char * p = new char[dest->unitSize()];
		/* Copy the first SMBLKSZ bytes */
                 for(int i = 0; i < SMBLKSZ; i++)
                        p[i] = _diskInode->di_sbbuf[i];
		/* Clear the rest of the block */
                 for( i = SMBLKSZ ; i < dest->unitSize(); i++)
                        p[i] = 0;

                if( dest->write( 0, 1, p, dest->unitSize() ) != 1 )
			returnValue = WriteError;
                if( dest->isKindOf( PersistentMemoryObjectClass ) ) 
                      ((PersistentMemoryObject *)dest)->setNumberOfBytes(
							    numberOfBytes() );

		_mutex.V();
	}
	else {
		_mutex.V();
		returnValue = UNIXInode::copy( dest ) ;
	}

	Raid( RaidMethod )( "%NAIXInode::copy done ( %N )\n", this, dest );

	return( returnValue );
}

#include "ClassInstaller.h"
DEFINECLASS("AIXInode",AIXInode,UNIXInode,0,0);
