#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * BSDAllocator.c - 
 */
#include "Assert.h"
#include "Raid.h"
#include "BSDAllocator.h"
#include "BSDContainer.h"

#ifdef __GNUC__
#pragma implementation
#endif

BSDAllocator::BSDAllocator( MemoryObject * source, BSDContainer * container,
			    ag ** allocatorGroup, int size,
			    int * totalFreeBlocks, int * totalFreeFragments,
			    int log2FragmentToSector, int fragmentsPerBlock,
			    int fragmentsPerGroup, int sectorsPerCylinder,
			    int cylinderGroups, int sectors, 
			    int inodesPerGroup )
: BlockAllocator( source ), _container( container )
{
	initDebug();
	Raid( RaidConstructor )( "%N::BSDAllocator( %N, , %d, %d, ", this,
				 source, log2FragmentToSector,
				 fragmentsPerBlock );
	Raid( RaidConstructor )( "%d, %d, %d, %d )\n", fragmentsPerGroup,
				 sectorsPerCylinder, cylinderGroups, sectors );

	/*
	 * Assign instance variables from parameters.
	 */
	_size = size;
	_totalFreeBlocks = totalFreeBlocks;
	_totalFreeFragments = totalFreeFragments;
	_allocatorGroup = allocatorGroup;
	_log2FragmentToSector = log2FragmentToSector;
	_fragmentsPerBlock = fragmentsPerBlock;
	_fragmentsPerGroup = fragmentsPerGroup;
	_sectorsPerCylinder = sectorsPerCylinder;
	_cylinderGroups = cylinderGroups;
	_sectors = sectors;
	_inodesPerGroup = inodesPerGroup;

	/*
	 * Calculate _blockMask used to check for block boundaries.
	 * Ensure that fragmentsPerBlock is a power of 2.
	 */
	_blockMask = _fragmentsPerBlock - 1;
	Assert( ( _blockMask & _fragmentsPerBlock ) == 0 );
#ifdef RAID
	if ( ( _debugMask & RaidDetail ) && ( _debugMessages > 0 ) ) {
		for( unsigned i = 0; i < _cylinderGroups; i++ )
			Console() 
				<< "Group " << i << " " 
				<< *(_allocatorGroup[i])
				<< "\n" << eor;
	}
#endif // RAID
}

inline 
int
BSDAllocator::blockIsFree( unsigned int cg, unsigned int fragment )
{
	/*
	 * Ensure fragment is on a block boundary.
	 */
	Assert( ( fragment & _blockMask ) == 0 );

	/*
	 * This function is optimized for the lions share of cases,
	 * i.e., BSD file systems almost always have 8 fragments per block.
	 */
	if( _fragmentsPerBlock == 8 ) {
		return( _allocatorGroup[cg]->ag_free[fragment >> 3] == 0377 );
	}

	/*
	 * Just in case someone decides to use less fragments per block,
	 * this less efficient code is provided.
	 */
	else {
		unsigned char * map = _allocatorGroup[cg]->ag_free;
		for( int i = 0; i < _fragmentsPerBlock; i++ ) {
			if( bitIsClear( map, fragment++ ) ) return( 0 );
		}
		return( 1 );
	}
}

/*
 * Calculate and update disk information.
 * This information is not use by Choices,
 * and is just updated to avoid fsck error messages.
 */
inline
void
BSDAllocator::updateDiskInfo( unsigned int cg, unsigned int frag, int delta )
{
	unsigned int sector = frag << _log2FragmentToSector;
	unsigned int cylinder = sector / _sectorsPerCylinder;
	unsigned int rotationalPosition = sector % _sectors * NRPOS / _sectors;

	Assert( cylinder < MAXCPG );
	Assert( rotationalPosition < NRPOS );

	_allocatorGroup[cg]->ag_btot[cylinder] += delta;
	_allocatorGroup[cg]->ag_b[(cylinder*NRPOS)+rotationalPosition] += delta;
}

/*
 * Allocate a contiguous set of fragments.
 * Tries to put the new fragments near to the fragment number stored in "near".
 * FIX: currently creates a new fragmented block no matter what.
 */
int
BSDAllocator::basicAllocate( int & blocks, int near, int numberOfFragments )
{
	Raid( RaidProtected )( "%N::basicAllocate( %d, near %d, frags %d )\n",
				this, blocks, near, numberOfFragments );

	/*
	 * Pick the proper cynlinder group in which to start searching
	 * for the free block.
	 * Also pick the fragment number within the group and make sure its
	 * on a block boundary and is next to "near".
	 */

	unsigned int cg = near / _fragmentsPerGroup;
	unsigned int newBlock = near % _fragmentsPerGroup;
	newBlock &= ~_blockMask;
	newBlock += _fragmentsPerBlock;

	Assert( cg < _cylinderGroups );

	/*
	 * Find a group with a free block.
	 */
	unsigned groupsChecked = 0;
	_container->getCG( cg );
	_lock.acquire();
	while( *(_allocatorGroup[cg]->ag_nbfree) == 0 ) {
		_lock.release();
		_container->releaseCG( cg, 0 );
		if( ++cg == _cylinderGroups ) cg = 0;
// FIX: probably should not just step thru the groups if the first
//	one doesnt have the right sized block.
		if( groupsChecked++ == _cylinderGroups ) {
			Raids( RaidError )
			<< this 
		        << "::basicAllocate() all cylinder groups full\n"
			<< eor;
			return( 0 );
		}
	}

	Assert( cg < _cylinderGroups );

	/*
	 * Search for the free block within the chosen group.
	 */
	unsigned int fragmentsInThisGroup = *(_allocatorGroup[cg]->ag_ndblk);
	unsigned f;
	for( f = 0; f < fragmentsInThisGroup; f += _fragmentsPerBlock ) {
		if( newBlock >= fragmentsInThisGroup ) newBlock = 0;
		Assert( ( newBlock & _blockMask ) == 0 );
		if( blockIsFree( cg, newBlock ) ) break;
		newBlock += _fragmentsPerBlock;
	}

	/*
	 * Make sure a block was found.
	 */
	Assert( f < fragmentsInThisGroup );

	/*
	 * See if a new fragmented block was created
	 * due to the allocation of less than a full block.
	 * Update count for the appropriate fragment size if needed. 
	 */
	int newFreeFragments = _fragmentsPerBlock - numberOfFragments;
	if( newFreeFragments != 0 ) {
		_allocatorGroup[cg]->ag_frsum[newFreeFragments]++;
		(*_totalFreeFragments) += newFreeFragments;
	}

	/*
	 * Mark the fragments allocated as used.
	 * The first case is optimized for whole, 8-fragment blocks.
	 */
	if( numberOfFragments == 8 ) {
		Assert( ( newBlock & _blockMask ) == 0 );
		Assert( blockIsFree( cg, newBlock ) );
		_allocatorGroup[cg]->ag_free[newBlock >> 3] = 0;
	}
	else for( int i = 0; i < numberOfFragments; i++ ) {
		unsigned char * map = _allocatorGroup[cg]->ag_free;
		Assert( bitIsSet( map, newBlock + i ) );
		clearBit( map, newBlock + i );
	}

	updateDiskInfo( cg, newBlock, -1 );

	/*
	 * The cylinder group now has one less free data block.
	 */
	(*(_allocatorGroup[cg]->ag_nbfree))--;
	(*_totalFreeBlocks)--;
	Raids( RaidDetail ) 
		<< this << "::basicAllocate() cyl group " << cg << " "
		<< *(_allocatorGroup[cg]) << "\n" << eor;

	/*
	 * Charge the inode for the disk sectors that have just been allocated.
	 */
	blocks += numberOfFragments << _log2FragmentToSector;

	/*
	 * newBlock is relative to its cylinder group,
	 * caulculate the base for that group and add it to newBlock
	 * as the return value.
	 */
	int base = _fragmentsPerGroup * cg;

	_lock.release();
	_container->releaseCG( cg, 1 );

	Raid( RaidProtected )( "%N::basicAllocate = %d+%d\n",
				this, base, newBlock );
	return( base + newBlock );
}

void BSDAllocator::updateBlocks( int & blocks, int numberOfFragments )
{
	// Assert( blocks >= ( numberOfFragments << _log2FragmentToSector ) );
	if ( blocks < ( numberOfFragments << _log2FragmentToSector ) ) {
		Console().formattedWrite
		( "%N::basicFree( %d, #frags %d ) blocks set to %d\n", 
		  this, blocks, numberOfFragments,
		  numberOfFragments << _log2FragmentToSector );
		blocks = numberOfFragments << _log2FragmentToSector;
	}

	Assert( numberOfFragments > 0 );

	/*
	 * No longer charge the inode for these disk sectors.
	 */
	blocks -= numberOfFragments << _log2FragmentToSector;
}

/*
 *	Free the specified set of contiguous fragments.
 */
void
BSDAllocator::basicFree( int & blocks, int oldFrags, int numberOfFragments )
{
	Raid( RaidProtected )( "%N::basicFree( %d, oldFrags %d, #frags %d )\n",
				this, blocks, oldFrags, numberOfFragments );

	/*
	 * Calculate which cylinder owns the old fragments,
	 * and the start fragment within that group.
	 */
	unsigned int cg = oldFrags / _fragmentsPerGroup;
	unsigned int fragment = oldFrags % _fragmentsPerGroup;

	unsigned int endcg = ( oldFrags + numberOfFragments - 1 ) / 
			     _fragmentsPerGroup;

	Assert( endcg == cg || endcg == ( cg + 1 ) );
	Assert( endcg < _cylinderGroups );

  	/*
	 * Handle the most common case efficiently.
	 */

	Assert( _fragmentsPerBlock == 8 );
	Assert( numberOfFragments < 8 || ( fragment & _blockMask ) == 0 );

	if ( endcg == cg ) {
	    _container->getCG( cg );
	    _lock.acquire();
	    updateBlocks( blocks, numberOfFragments );
	    while( numberOfFragments >= 8 ) {
		/*
		 * Ensure blocks are really 8 fragments and that the
		 * fragments being freed are on a block boundary.
		 */

		/*
		 * Ensure fragments are really allocated.
		 */
		Assert( _allocatorGroup[cg]->ag_free[fragment>>3] == 0 );

		/*
		 * Free the fragments.
		 */
		_allocatorGroup[cg]->ag_free[fragment>>3] = 0377;

		updateDiskInfo( cg, fragment, 1 );

		/*
		 * The cylinder group now has one more free data block.
		 */
		(*(_allocatorGroup[cg]->ag_nbfree))++;
		(*_totalFreeBlocks)++;

		numberOfFragments -= 8;
		fragment += 8;
	    }
	} else {
	    _container->getCG( cg );
	    _lock.acquire();
	    updateBlocks( blocks, numberOfFragments );

	    unsigned fragmentsInThisGroup = _fragmentsPerGroup - fragment;
	    numberOfFragments -= fragmentsInThisGroup;

	    while( fragmentsInThisGroup >= 8 ) {
		/*
		 * Ensure blocks are really 8 fragments and that the
		 * fragments being freed are on a block boundary.
		 */

		/*
		 * Ensure fragments are really allocated.
		 */
		Assert( _allocatorGroup[cg]->ag_free[fragment>>3] == 0 );

		/*
		 * Free the fragments.
		 */
		_allocatorGroup[cg]->ag_free[fragment>>3] = 0377;

		updateDiskInfo( cg, fragment, 1 );

		/*
		 * The cylinder group now has one more free data block.
		 */
		(*(_allocatorGroup[cg]->ag_nbfree))++;
		(*_totalFreeBlocks)++;

		fragmentsInThisGroup -= 8;
		fragment += 8;
	    }
	    _lock.release();
	    _container->releaseCG( cg, 1 );
	    Assert( ( fragment % _fragmentsPerGroup ) == 0 );
	    fragment = 0;
	    cg++;
	    _container->getCG( cg );
	    _lock.acquire();
	    while( numberOfFragments >= 8 ) {
		/*
		 * Ensure blocks are really 8 fragments and that the
		 * fragments being freed are on a block boundary.
		 */

		/*
		 * Ensure fragments are really allocated.
		 */
		Assert( _allocatorGroup[cg]->ag_free[fragment>>3] == 0 );

		/*
		 * Free the fragments.
		 */
		_allocatorGroup[cg]->ag_free[fragment>>3] = 0377;

		updateDiskInfo( cg, fragment, 1 );

		/*
		 * The cylinder group now has one more free data block.
		 */
		(*(_allocatorGroup[cg]->ag_nbfree))++;
		(*_totalFreeBlocks)++;

		numberOfFragments -= 8;
		fragment += 8;
	    }
        }
	if ( numberOfFragments == 0 ) {
	    	_lock.release();
	    	_container->releaseCG( cg, 1 );
		return;
	}

	/*
	 * Ensure each fragment was in use, then mark it free.
	 */
	Assert( numberOfFragments > 0 && numberOfFragments < 8 );
	for( int i = 0; i < numberOfFragments; i++ ) {
		unsigned char * map =  _allocatorGroup[cg]->ag_free;
		Assert( bitIsClear( map, fragment + i ) );
		setBit( map, fragment + i );
	}

	/*
	 * Calculate beginning of block in which fragments belong.
	 * Calculate number of fragments on each side of the ones being freed.
	 */
	unsigned block = fragment & ~_blockMask;
	unsigned preFrags = fragment - block;
	unsigned postFrags = _fragmentsPerBlock - ( numberOfFragments + preFrags );

	/*
	 * If the whole block is free now, then adjust counters to show
	 * there is another free block, and one or two less free fragments.
	 */
	if( blockIsFree( cg, block ) ) {
		updateDiskInfo( cg, block, 1 );

		/*
		 * The cylinder group now has one more free data block.
		 */
		(*(_allocatorGroup[cg]->ag_nbfree))++;
		(*_totalFreeBlocks)++;

		/*
		 * The newly freed fragments must be coalesced with others
		 * in the same block.
		 */
		if( preFrags )  _allocatorGroup[cg]->ag_frsum[preFrags]--;
		if( postFrags ) _allocatorGroup[cg]->ag_frsum[postFrags]--;
		(*_totalFreeFragments) -= ( preFrags + postFrags );
	}

	/*
	 * The whole block was not free, coalesce newly freed fragments
	 * with others in the same block, if possible.
	 */
	else {
		unsigned char * map =  _allocatorGroup[cg]->ag_free;

		/*
		 * The newly freed fragments must be coalesced with others
		 * in the same block, if necessary. The coalseced fragments
		 * are free, any smaller groups of fragments on each side
		 * are no longer free.
		 */
		unsigned i;
		for( i = 0; i < preFrags; i++ ) {
			if( bitIsClear( map, ( fragment - 1 )  - i ) ) break;
			numberOfFragments++;
		}
		if( i > 0 ) {
			_allocatorGroup[cg]->ag_frsum[i]--;
			(*_totalFreeFragments) -= i;
		}

		fragment += (numberOfFragments - i);

		for( i = 0; i < postFrags; i++ ) {
			if( bitIsClear( map, fragment + i ) ) break;
			numberOfFragments++;
		}
		if( i > 0 ) {
			_allocatorGroup[cg]->ag_frsum[i]--;
			(*_totalFreeFragments) -= i;
		}

		_allocatorGroup[cg]->ag_frsum[numberOfFragments]++;
		(*_totalFreeFragments) += numberOfFragments;
	}
	_lock.release();
	_container->releaseCG( cg, 1 );
}

int
BSDAllocator::allocate( int & blocks, int nearPhysicalUnit, int units )
{
	Raid( RaidMethod )( "%N::basicAllocate( %d, %d, %d )\n",
			    this, blocks, nearPhysicalUnit, units );

	int block = basicAllocate( blocks, nearPhysicalUnit, units );

	if ( block == 0 ) {
		Console() 
			<< "=== WARNING ===\n"
			<< this << "::allocate() " << _source 
			<< " is full, allocate failed\n" << eor;
	}
	Raid( RaidMethod )( "%N::basicAllocate returns %d\n", this, block );
	return( block );
}

inline
void
BSDAllocator::markBlockInUse( int cg, int block )
{
	Raids( RaidMethod ) << this << "::markBlockInUse(cg " << cg 
		<< ", block " << block << ")\n" << eor;

	ag * allocGroup = _allocatorGroup[ cg ];
	if( _fragmentsPerBlock == 8 ) {
		Assert( ( block & _blockMask ) == 0 );
		allocGroup->ag_free[block >> 3] = 0;
	}
	else {
		unsigned char * map = allocGroup->ag_free;
		for( int i = 0; i < _fragmentsPerBlock; i++ ) {
			Assert( bitIsSet( map, block + i ) );
			clearBit( map, block + i );
		}
	}
	updateDiskInfo( cg, block, -1 );
	(*(allocGroup->ag_nbfree))--;
	(*_totalFreeBlocks)--;
}

int
BSDAllocator::allocateBlocksFromGroup( int & blocks, int cg, int * buf, 
				       int count, int pref )
{
	_container->getCG( cg );
	_lock.acquire();
	ag * allocGroup = _allocatorGroup[ cg ];
	int * nbfree = allocGroup->ag_nbfree; 
	int userotor;
	if ( *nbfree == 0 ) {
		_lock.release();
		_container->releaseCG( cg, 0 );
		return 0;
	}
	int base = _fragmentsPerGroup * cg;
	int allocated = 0;
	if ( pref == 0 ) {
		pref = *(allocGroup->ag_rotor);
		int mod = pref & (_fragmentsPerBlock - 1);
		if (mod) {
			pref += (_fragmentsPerBlock - mod );
		}
		userotor = 1;
	} else {
		pref = ( pref & ~( _fragmentsPerBlock - 1 ) ) % 
			_fragmentsPerGroup;
		if ( blockIsFree( cg, pref ) ) {
			userotor = 0;
			markBlockInUse( cg, pref );
			*buf = base + pref;
			buf++;
			count --;
			allocated++;
		} else {
			pref = *(allocGroup->ag_rotor);
			int mod = pref & (_fragmentsPerBlock - 1);
			if (mod) {
				pref += (_fragmentsPerBlock - mod );
			}
			userotor = 1;
		}
	}
#ifdef ASSERT
	int tries = 0;
	int maxtries = _fragmentsPerGroup / _fragmentsPerBlock;
#endif
	while ( count > 0 && *nbfree > 0 ) {
		pref += _fragmentsPerBlock;
		if ( pref >= _fragmentsPerGroup )
			pref = 0;
		if ( blockIsFree( cg, pref ) ) {
			markBlockInUse( cg, pref );
			*buf = base + pref;
			buf++;
			count --;
			allocated++;
		}
		Assert( tries++ <= maxtries );
	}	
	Assert( allocated > 0 );
	if ( userotor ) {
		int newrotor = *(buf - 1) - base;
		Assert( newrotor >= 0 && newrotor < _fragmentsPerGroup );
		*(allocGroup->ag_rotor) = newrotor;
	}

	blocks += ( allocated * _fragmentsPerBlock ) << _log2FragmentToSector;

	_lock.release();
	_container->releaseCG( cg, 1 );

	return allocated;
}

int
BSDAllocator::allocateBlocks( int & blocks, int inumber, 
			      int *buf, int start, int count, 
			      int pref )
{
	Raids( RaidMethod ) << this << "::allocateBlocks( "
		<< inumber << ", " << asAddress( buf ) << ", "
		<< start << ", " << count << ", " << pref << "\n" << eor;

	unsigned i;
	unsigned remaining = count;
	int * startBuf = buf + start;
	if ( pref >= _size )
		pref = 0;
	unsigned cg;
	if ( pref == 0 ) {
		cg = inumber / _inodesPerGroup;
	} else {
		cg = pref / _fragmentsPerGroup;
	}
	int icg = cg;
	int result = allocateBlocksFromGroup
			( blocks, cg, startBuf, remaining, pref );
	remaining -= result;
	if ( remaining == 0 ) {
		goto done;
	}
	startBuf += result;
	for ( i = 1; i < _cylinderGroups; i *= 2 ) {
		cg += i;
		if ( cg >= _cylinderGroups )
			cg -= _cylinderGroups;
		result = allocateBlocksFromGroup
			( blocks, cg, startBuf, remaining, 0 );
		if ( result ) {
			remaining -= result;
			if ( remaining == 0 ) {
				goto done;
			}
			startBuf += result;
		}
	}
	cg = ( icg + 2 ) % _cylinderGroups;
	for ( i = 2; i < _cylinderGroups; i++ ) {
		result = allocateBlocksFromGroup
			( blocks, cg, startBuf, remaining, 0 );
		if ( result ) {
			remaining -= result;
			if ( remaining == 0 ) {
				goto done;
			}
			startBuf += result;
		}
		cg++;
		if ( cg == _cylinderGroups )
			cg = 0;
	}
	Console() << this << "::allocateBlocks() out of space\n" << eor;
	count = count - remaining;
done:
	// Assert( remaining == 0 );
	Raids( RaidMethod ) << this << "::allocateBlocks() returns " << count
		<< "\n" << eor;
	return count;
}

void
BSDAllocator::free( int & blocks, int physicalUnit, int units )
{
	Raid( RaidMethod )( "%N::basicFree( %d, %d, %d )\n",
			    this, blocks, physicalUnit, units );

	basicFree( blocks, physicalUnit, units );
}

int
BSDAllocator::extendFragmentsToBlock( int & blocks, int physicalBlock, 
				      int noOfFragments )
{
	Raids(RaidProtected) << this << "::extendFragmentsToBlock( "
		<< physicalBlock << ", " << noOfFragments << " )\n" << eor;
	Assert( noOfFragments > 0 && noOfFragments < _fragmentsPerBlock );
	if ( ( physicalBlock & _blockMask ) != 0 )
		return 0;
	int cg = physicalBlock / _fragmentsPerGroup;
	int fragment = physicalBlock % _fragmentsPerGroup;
	int count = _fragmentsPerBlock - noOfFragments;
	int check = fragment + noOfFragments;
	unsigned char * map = _allocatorGroup[cg]->ag_free;
	_container->getCG( cg );
	_lock.acquire();
        for( int i = 0; i < count; i++ ) {
                if( bitIsClear( map, check++ ) ) {
			_lock.release();
			Raids(RaidProtected) 
				<< this << "::extendFragmentsToBlock() failed\n"
				<< eor;
			return 0;
		}
        }
	_allocatorGroup[cg]->ag_frsum[ count ]--;
	(*_totalFreeFragments) -= count;
	if ( _fragmentsPerBlock == 8 ) {
		map[fragment >> 3] = 0;
	} else {
		check = fragment + noOfFragments;
		for( int i = 0; i < count; i++ ) {
			clearBit( map, check++ );
		}
	}
	
	blocks += ( count << _log2FragmentToSector );
	_lock.release();
	_container->releaseCG( cg, 1 );
	Raids(RaidProtected) << this << "::extendFragmentsToBlock() extended "
		<< physicalBlock << " cg " << cg << " fragment " << fragment
		<< " org noOfFragments " << noOfFragments << "\n" << eor;
	return 1;
}

#include "ClassInstaller.h"
DEFINECLASS("BSDAllocator",BSDAllocator,BlockAllocator,0,0);
