#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Contributing Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * BSDDirectory.c - 
 */
#include "Assert.h"
#include "Raid.h"
#include "FileSystem.h"
#include "PersistentMemoryObject.h"
#include "BSDDirectory.h"
#include "String.h"
#include "Byte.h"

#ifdef __GNUC__
#pragma implementation
#endif

struct direct {
	int	d_ino;
	short	d_reclen;
	short	d_namlen;
	char	d_name[MaxEntrySize + 1];

	int	size();
};

inline 
int 
direct::size()
{
	int bytes = sizeof(d_ino) + sizeof(d_reclen) + sizeof(d_namlen)
		    + d_namlen + 1;
	int remainder = bytes & 0xf;
	if ( remainder ) {
		bytes += ( 0x10 - remainder );
	}
	return bytes;
}

/*
 * BSDDirectory allows directory entries to be located in,
 * added to, and removed from a BSD directory file.
 */
BSDDirectory::BSDDirectory( PersistentMemoryObject * source )
: MemoryObjectDictionary( source )
{
    initDebug();
    Raid( RaidConstructor )( "%N::BSDDirectory( %N )\n", this, source );
    Raid( RaidDetail )( "%N::BSDDirectory idNumber %d _size %d\n",
	     this, source->idNumber(), _size );

     _maxKeyLength = MaxEntrySize;

    /* make directory file size a multiple of the directory size */
    int roundedSize = ( _size + DirectoryBlockSize - 1 ) / DirectoryBlockSize
			* DirectoryBlockSize;
    if( _size != roundedSize ) {
	_size = roundedSize;
	source->setNumberOfBytes( _size );
    }
    if( _size == 0 ) {
	add( ".", source );
    } else {
	reMap();
    }
    _hint = source->idNumber() + 1;
}

BSDDirectory::~BSDDirectory()
{
    Raid( RaidConstructor )( "%N::~BSDDirectory\n", this );
}

/*
 * search the entire directory for an occurrence of key
 * if found, return 1; otherwise 0
 * side effects: set _previousOffset & _offset dir entry ofs and
 * findKey will also try to find an _empty space big enough 
 * to hold a new directory entry of _size _size.
 */
int
BSDDirectory::findKey( char * key, int needSize ) 
{
    direct * dp;

    Raid( RaidMethod )( "%N::findKey( key %s, needSize %d )\n",
		this, key, needSize );

    int StillLooking = ( needSize > 0 );
    _emptyOffset = -1;

    _previousOffset = 0;
    for( _offset = 0; _offset < _size; _offset += dp->d_reclen ) {
        dp = ( direct * ) &_buffer[_offset];

	/*	Raid( RaidDetail )( "%N::findKey ofs %d ", this, _offset); 
		Raid( RaidDetail )( "ino %d rl %d nl %d key %s\n", dp->d_ino, dp->d_reclen, dp->d_namlen, dp->d_name );*/

 Debug<<this<<"::findKey("<<key<<") "<<dp->d_reclen<<":"<<dp->d_name<<"\n"<<eor;

	Assert( dp->d_reclen > 0 );

	/* check to see if hole is big enough */
        if( StillLooking ) {
            int sp = dp->d_reclen;
            if( dp->d_ino != 0 ) sp -= entrySize( dp->d_namlen );
            if( sp >= needSize ) {
                _emptyOffset = _offset;
                StillLooking = 0;
        	Raid( RaidDetail )( "%N::findKey space @ %d\n", this, _offset );
            }
        }

	/* return if key is found */
        if( ( dp->d_ino != 0 ) && ( strcmp( key, dp->d_name ) == 0 ) ) {
	    _idNumberOfKey = dp->d_ino;
            Raid( RaidDetail )( "%N::findKey found @ %d\n", this, _offset );
            return( 1 );
	}
	_previousOffset = _offset;
    }
    return( 0 );
}

int
BSDDirectory::insertEntry( char * key, int keyLength, int idNumber  )
{
    direct * edp;
    direct dir;

    Raid( RaidMethod )( "%N::insertEntry( key %s, keyLength %d, idNumber %d )\n",
		this, key, keyLength, idNumber );

    /* set up the new entry ( for _size estimation ) */
    dir.d_ino = idNumber;
    dir.d_namlen = keyLength;

    if( _emptyOffset < 0 ) {
        dir.d_reclen = DirectoryBlockSize;
        _offset = _size;
        _size += DirectoryBlockSize;
        _source->setNumberOfBytes( _size );
        int newUnits = _source->numberOfUnits();
	if( _units != newUnits )  {
		reMap();
		_units = newUnits;
	}
    }
    else {
        edp = ( direct * ) &_buffer[_emptyOffset];
        int bytesInUse = 0;
        if( edp->d_ino != 0 ) bytesInUse = entrySize( edp->d_namlen );
        dir.d_reclen = edp->d_reclen - bytesInUse;
        edp->d_reclen = bytesInUse;
	_offset = _emptyOffset + bytesInUse;
    }

    Raid( RaidDetail )( "%N::insertEntry _offset %d _emptyOffset %d\n",
		this, _offset, _emptyOffset );

    /* move new entry into the blk */
    edp = ( direct * ) &_buffer[_offset];
    edp->d_ino    = dir.d_ino;
    edp->d_reclen = dir.d_reclen;
    edp->d_namlen = dir.d_namlen;
    strcpy( edp->d_name, key );

    /* new hint */
    _hint = idNumber + 1;

    return( writeChanges() );
}

/*
 */
int
BSDDirectory::clearEntry()
{
    Raid( RaidMethod )( "%N::clearEntry()\n", this );

    direct * cdp = ( direct * ) &_buffer[_offset];

    if( ( _offset & ( DirectoryBlockSize - 1 ) ) == 0 ) cdp->d_ino = 0;
    else {
	direct * pdp = ( direct* ) &_buffer[_previousOffset];
        pdp->d_reclen += cdp->d_reclen;
    }
    return( writeChanges() );
}

/*
 *	return positive count of keys if the buffer was big enough,
 *	negative count otherwise.
 *	optimize and improve later, perhaps via memory-mapping a precalculated
 *	list-o-keys.
 */
int
BSDDirectory::basicKeys( char * keyBuffer, int bufferSize, int & count ) 
{
    Raid( RaidMethod )( "%N::keys( buffer %x, bufferSize %d ) _size %d\n",
		this, keyBuffer, bufferSize, _size );

    count = 0;
    int limit = ( bufferSize >= _size ) ? _size : bufferSize;
    char * t = keyBuffer;
    direct * dp = ( direct * ) _buffer;
    for( _offset = 0; _offset < limit; _offset += dp->d_reclen ) {

        dp = ( direct * ) &_buffer[_offset];

	Raid( RaidDetail )( "%N::keys ofs %d ", this, _offset); 
	Raid( RaidDetail )( "ino %d rl %d nl %d key %s\n",
		   dp->d_ino, dp->d_reclen, dp->d_namlen, dp->d_name );

	Assert( dp->d_reclen > dp->d_namlen );

	if( dp->d_ino != 0) {
		Assert( dp->d_namlen > 0 );

		if( t != 0 ) {
			for( char * s = dp->d_name; *s; ) *t++ = *s++; 
			*t++ = '\n';
		}
		count++;
	}
    }
    if( t != 0 ) *t = '\0';
    if( limit == _size ) return( 0 );
    else return( TooManyKeys );
}

/*
 *	return positive count of keys if the buffer was big enough,
 *	negative count otherwise.
 *	optimize and improve later, perhaps via memory-mapping a precalculated
 *	list-o-keys.
 */
int
BSDDirectory::basicAssociations( char * keyBuffer, int bufferSize,
				 int & count )
{
   int offset = 0;
   return basicAssociations( keyBuffer, bufferSize, count, offset );
}

int
BSDDirectory::basicAssociations( char * keyBuffer, int bufferSize,
		 		 int & count, int & offset )
{
    Raid( RaidMethod )( "%N::basicAssociations( buf %x, bufSize %d )_size %d\n",
			this, keyBuffer, bufferSize, _size );

    Assert( keyBuffer != 0 );
    char * t = keyBuffer;
    char * tlimit = keyBuffer + bufferSize;
    direct * dp = ( direct * ) _buffer;

    for( _offset = offset ; _offset < _size; _offset += dp->d_reclen ) {

        dp = ( direct * ) &_buffer[_offset];

	Raid( RaidDetail )( "%N::associations ofs %d ", this, _offset); 
	Raid( RaidDetail )( "ino %d rl %d nl %d key %s\n",
		   dp->d_ino, dp->d_reclen, dp->d_namlen, dp->d_name );

	Assert( dp->d_reclen > dp->d_namlen );

	if( dp->d_ino != 0) {
	    Assert( dp->d_namlen > 0 );
	    if( t + dp->d_reclen > tlimit) {
		break;
	    }
	    int cpcount = dp->size();
	    ByteCopy( dp, t, cpcount );
	    ((direct *) t)->d_reclen = cpcount;
	    t += cpcount;
	}
    }
    offset = _offset;
    count = t - keyBuffer;
    Assert( count <= bufferSize );
    if( _offset >= _size ) return( 0 );
    else return( TooManyKeys );
}

int
BSDDirectory::entrySize( int keyLength )
{
    return( ( keyLength + 3 + sizeof( direct ) - MaxEntrySize ) & ~3 );
}

int
BSDDirectory::hint()
{
	return( _hint );
}

ProxiableObject * 
BSDDirectoryConstructor( ProxiableObject * source )
{
        if( source->isKindOf( PersistentMemoryObjectClass ) ) {
                return( new BSDDirectory( (PersistentMemoryObject *)source ) );
        }
        else return( 0 );
}



#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("BSDDirectory",BSDDirectory,MemoryObjectDictionary,&BSDDirectoryConstructor,0);
