#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * COFFContainer.cc - COFF Inode System class
 */

#include "COFF.h"
#include "COFFContainer.h"
#include "RootEntry.h"
#include "FileSystem.h"
#include "String.h"
#include "Raid.h"

#ifdef __GNUC__
#pragma implementation
#endif

COFFContainer::COFFContainer( PersistentMemoryObject * source )
: MemoryObjectContainer( source ), _byteSource( 0 )
{
	initDebug();

	Raid( RaidConstructor )( "%N::COFFContainer( %N )\n", this, _source );

	Console()<<this<<"::COFFContainer()\n";

	_valid = 0;

	int objects = 0;
	unsigned bytes = _source->numberOfBytes();

	if( bytes < sizeof( filehdr ) ) {
		Raid( RaidDetail )( "%N::COFFContainer() too small %d\n",
				    this, _source->numberOfBytes() );
	}
	else {
		_byteSource = new FileObject( bytes, 0, 0, 0, _source );

		_filehdr = new filehdr;
		_byteSource->read( 0, sizeof( filehdr ), (char *)_filehdr,
				sizeof( filehdr ) );

		if( ( _filehdr->f_magic != NS32GMAGIC )  &&
		    ( _filehdr->f_magic != NS32SMAGIC ) ) {
			Raid( RaidDetail )( "%N::COFFContainer()bad magic %o\n",
					    this, _filehdr->f_magic );
			Console()<<this<<" bad magic number: "<<_filehdr->f_magic<<"\n";
			delete _filehdr;
		}
		else {
			_scnhdrs = 0;
			_valid = 1;
		}
	}
	if( _valid ) {
		Assert( ( _filehdr->f_opthdr == sizeof( aouthdr ) ) ||
			( _filehdr->f_opthdr == 0 ) );

		int offset = sizeof( filehdr ) + _filehdr->f_opthdr;
		int length = sizeof( scnhdr ) * _filehdr->f_nscns;

		_scnhdrs = new scnhdr[_filehdr->f_nscns];
		_byteSource->read( offset, length, (char *)_scnhdrs, length );

		objects = SectionIds + ( 4 * _filehdr->f_nscns );
	}
	Console()<<"COFFContainer initializing file objects ("<<objects<<")\n";
	initializeArrays( objects );
}

COFFContainer::~COFFContainer()
{
	Raid( RaidConstructor )( "%N::~COFFContainer()\n", this );

	if( _valid ) {
		delete _scnhdrs;
		delete _filehdr;
	}

	Raid( RaidConstructor )( "%N::~COFFContainer done\n", this );
}

FileObject *
COFFContainer::basicOpen( int id, int & status, int flags )
{
	Raid( RaidMethod )( "%N::basicOpen( inumber %d )\n", this, id );

	Console()<<this<<"::basicOpen("<<id<<") RootId="<<RootId<<"\n";
	if( id < RootId ) {
		status = IdNumberOutOfRange;
		return( 0 );
	}
	else {
		status = 0;
		int offset = 0;
		int units = 0;
		if( id == RootId ) {
			extern ClassStar COFFDictionaryClass;
			return( new RootEntry(	sizeof( filehdr ) +
						_filehdr->f_opthdr +
						_filehdr->f_nscns *
						sizeof( scnhdr ), 0,
						id, this, _source,
						COFFDictionaryClass, 0 ) );
		}
		else if( id == FileHeaderId ) {
			offset = 0;
			units = sizeof( filehdr );
			return( new FileObject( units, 0,
						id, this, _source, offset ) );
		}
		else if( id == AOutHeaderId ) {
			if( _filehdr->f_opthdr > 0 ) {
				offset = sizeof( filehdr );
				units = sizeof( aouthdr );
			}
			return( new FileObject( units, 0,
						id, this, _source, offset ) );
		}
		else if( id == SymbolTableId ) {
			if( _filehdr->f_symptr > 0 ) {
				offset = _filehdr->f_symptr;
				units = _filehdr->f_nsyms * sizeof( syment );
			}
			return( new FileObject( units, 0,
						id, this, _source, offset ) );
		}
		else if( id == StringTableId ) {
			if( _filehdr->f_symptr > 0 ) {
				offset = _filehdr->f_symptr +
					 _filehdr->f_nsyms * sizeof( syment );
				units = _source->numberOfBytes() - offset;
			}
			return( new FileObject( units, 0,
						id, this, _source, offset ) );
		}
		else {
			int section = id - SectionIds;
			int type = section & 3;
			section >>= 2; 

			switch( type ) {
			case Header:
				offset = sizeof( filehdr ) +
					 _filehdr->f_opthdr +
					 section * sizeof( scnhdr );
				units = sizeof( scnhdr );
				break;
			case Section:
				offset = _scnhdrs[section].s_scnptr;
				units = _scnhdrs[section].s_size;
				break;
			case Reloc:
				offset = _scnhdrs[section].s_relptr;
				units = _scnhdrs[section].s_nreloc * RELOCSIZE;
				break;
			case LineNo:
				offset = _scnhdrs[section].s_lnnoptr;
				units = _scnhdrs[section].s_nlnno * LINENOSIZE;
				break;
			default:
				Assert( NOTREACHED );
			}
			return( new FileObject( units, 0,
						id, this, _source, offset ) );
		}
	}
}

int
COFFContainer::basicRootId()
{
	return( RootId );
}

ProxiableObject * 
COFFContainerConstructor( ProxiableObject * source )
{
	if( source->isKindOf( PersistentMemoryObjectClass ) ) {
		return( new COFFContainer( (PersistentMemoryObject *)source ) );
	}
	else return( 0 );
}



#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("COFFContainer",COFFContainer,MemoryObjectContainer,&COFFContainerConstructor,0);
