#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * ContainerIndex.cc
 */

#include "Assert.h"
#include "ContainerIndex.h"
#include "MemoryObjectContainer.h"
#include "OutputStream.h"

#ifdef __GNUC__
#pragma implementation
#endif

const int ContainerIndexMaxFreeCount = 64;
Lock ContainerIndex::_lock;


ContainerIndex theContainerIndex;
ContainerIndex * TheContainerIndex = &theContainerIndex;

int _activeFiles = 0;

ContainerIndex::ContainerIndex()
{
	for (int i = 0; i < CONTAINERINDEX_HASHTABLESIZE; i++) {
#ifdef DEBUG_CONTAINERINDEX
		_hashTableCount[i] = 0;
#endif
		_hashTable[i] = 0;
	}
	_free = 0;
	_freeCount = 0;
#ifdef PREALLOCATE_CONTAINERINDEX
	for ( i = 0; i < ContainerIndexMaxFreeCount; i++ ) {
		_IndexEntry * e = new _IndexEntry();
		e->_next = _free;
#ifdef DEBUG_CONTAINERINDEX
		e->_list = &_free;
#endif
		_free = e;
		_freeCount++;
	} 
	Assert( _freeCount == ContainerIndexMaxFreeCount );
#endif // PREALLOCATE_CONTAINERINDEX
}

void
ContainerIndex::errfull()
{
	Console() << "ContainerIndex::errfull() open file table full\n"
		  << "recompile with larger table\n" << eor;
	Assert( NOTREACHED );
}

FileObject * 
ContainerIndex::lookup(MemoryObjectContainer * c, int index)
{
	acquire();
	int slot = hash(c, index);
	_IndexEntry ** list = _hashTable + slot;
	_IndexEntry * e = *list;
#ifdef DEBUG_CONTAINERINDEX
	int count = 0;
	FileObject * result = 0;
#endif
	while(e) {
#ifdef DEBUG_CONTAINERINDEX
		Assert(e->_list == list);
#endif
		if (e->_index == index && e->_container == c) {
#ifdef DEBUG_CONTAINERINDEX
			result = e->_fileObject;	
#else
			release();
			return e->_fileObject;
#endif
		}
		e = e->_next;
#ifdef DEBUG_CONTAINERINDEX
		count++;
#endif
	}
#ifdef DEBUG_CONTAINERINDEX
	Assert(count == _hashTableCount[slot]);
#endif 
	release();
#ifdef DEBUG_CONTAINERINDEX
	return result;
#else
	return 0;
#endif
}

void 
ContainerIndex::bind(MemoryObjectContainer *c, int index, FileObject * f)
{
	acquire();
	int slot = hash(c, index);
	_IndexEntry ** list = _hashTable + slot;
	_IndexEntry ** prev = list;
	_IndexEntry * e;
#ifdef DEBUG_CONTAINERINDEX
	int count = 0;
	int found = 0;
#endif
	e = *prev;
	while(e) {
#ifdef DEBUG_CONTAINERINDEX
		Assert(e->_list == list);
#endif
		if (e->_index == index && e->_container == c) {
			Assert(e->_fileObject == (FileObject *) 0xffffffff);
			e->_fileObject = f;
#ifdef DEBUG_CONTAINERINDEX
			found++;
#else
			release();
			return;
#endif
		}
		prev = &(e->_next);
		e = *prev;
#ifdef DEBUG_CONTAINERINDEX
		count++;
#endif
	}
#ifdef DEBUG_CONTAINERINDEX
	Assert(count == _hashTableCount[slot]);
	if (found) {
		Assert( found == 1 );
		release();
		return;
	}
#endif
	if ( _free ) {
		_freeCount--;
		Assert( _freeCount >= 0 );
		e = _free;
		_free = _free->_next;
#ifdef DEBUG_CONTAINERINDEX
		Assert( e->_list == &_free );
#endif
	} else {
		e = new _IndexEntry;
	}
#ifdef DEBUG_CONTAINERINDEX
	_hashTableCount[slot]++;
#endif
	_activeFiles ++;
	e->_container = c;
	e->_index = index;
	e->_fileObject = f;
	e->_next = 0;
	*prev = e;
#ifdef DEBUG_CONTAINERINDEX
	e->_list = list;
#endif
	release();
}

void 
ContainerIndex::unbind(MemoryObjectContainer *c, int index)
{
	acquire();
	int slot = hash(c, index);
	_IndexEntry ** list = _hashTable + slot;
	_IndexEntry ** prev = list;
	_IndexEntry * e = *prev;
#ifdef DEBUG_CONTAINERINDEX
	int found = 0;
	int count = 0;
#endif
	while(e) {
#ifdef DEBUG_CONTAINERINDEX
		Assert(e->_list == list);
#endif
		if (e->_index == index && e->_container == c) {
			*prev = e->_next;
			if ( _freeCount < ContainerIndexMaxFreeCount ) {
#ifdef DEBUG_CONTAINERINDEX
				e->_list = &_free;
#endif
				e->_next = _free;
				_free = e;
				_freeCount++;
			} else {
				delete e;
			}
			_activeFiles--;
#ifdef DEBUG_CONTAINERINDEX
			found++;
			count++;
			e = *prev;
			continue;
#else
			release();
			return;
#endif
		}
		prev = &(e->_next);
		e = *prev;
#ifdef DEBUG_CONTAINERINDEX
		count++;
#endif
	}
#ifdef DEBUG_CONTAINERINDEX
	Assert(count == _hashTableCount[slot]);
	Assert(found == 1);
	_hashTableCount[slot]--;
#endif
	release();
}

FileObject ** 
ContainerIndex::getActiveFileObjects(MemoryObjectContainer * c)
{
	const int size = CONTAINERINDEX_HASHTABLESIZE;
	acquire();
	int count = 0;
	int i;
	for(i = 0; i < size; i++) {
		_IndexEntry ** list = _hashTable + i;
		_IndexEntry * e = *list;
		while (e) {
			if (e->_container == c) {
				count++;
			}
			e = e->_next;
		}
	}
	FileObject ** fo = new FileObject * [count + 1];
	FileObject ** p = fo;
	for(i = 0; i < size; i++) {
		_IndexEntry ** list = _hashTable + i;
		_IndexEntry * e = *list;
		while (e) {
#ifdef DEBUG_CONTAINERINDEX
			Assert(e->_container != 0 || e->_list == list);
#endif
			if (e->_container == c) {
				*p++ = e->_fileObject;
				count--;
			}
			e = e->_next;
		}
	}
	Assert( count == 0 );
	*p = 0;
	release();
	return fo;
}

void
ContainerIndex::freeActiveFileObjects(FileObject ** fo)
{
	delete fo;
}

