#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The Mobile Operating Systems Group
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 201 N. Goodwin Ave.
 *		 Urbana, IL	61801
 *
 * Copyright 2005 University of Illinois Board of Trustees. All Rights Reserved.
 *
 * Author: Jeffrey Carlyle (jcarlyle@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 */
/*
 * CramfsContainer.cc - Container for cramfs file system object.
 */

#include "Byte.h"
#include "CramfsContainer.h"
#include "CramfsFileObject.h"
#include "CramfsDirObject.h"
#include "Raid.h"
#include "FileSystem.h"

#ifdef __GNUC__
#pragma implementation
#endif

/* #define JEFF */

/**
 * Construct a cramfs container.
 * @param source Object that contains the file system data and metadata.
 */
CramfsContainer::CramfsContainer(PersistentMemoryObject *source) 
  : MemoryObjectCachingContainer(source)
{
    initDebug();
    
    Raid( RaidConstructor )( "%N::CramfsContainer( %N )\n", this, _source);

    /* create a byte addressable view of the disk */
    _byteSource = new FileObject(_source->numberOfBytes(), 0, 0, 0, _source);
    Assert(_byteSource != NULL);

    /* read the superblock */
    int count = _byteSource->read(0,
	sizeof(CramfsSuperblock),
	(char *)&_superBlock, sizeof(CramfsSuperblock));
    Assert( count == sizeof(CramfsSuperblock) );

    /* check the magic number */
    Assert( _superBlock.magic == CramfsMagicNumber );

    /* check for older version of the file system header */
    if( !(_superBlock.flags & CramfsFlagFsinfoVersion2) ) {
      _superBlock.size = 1<<28;
      _superBlock.fsinfo.blocks = 0;
      _superBlock.fsinfo.files = 0;
    }

#ifdef JEFF
    _superBlock.root.dumpTo(Console());

    int32 size = _superBlock.root.size;
    int32 offset = _superBlock.root.offset<<2;
    while(size>0)
    {
      CramfsInode inode;
      count = _byteSource->read(offset, sizeof(CramfsInode),
	  (char *)&inode, sizeof(CramfsInode));
      Assert(count == sizeof(CramfsInode));
      inode.dumpTo(Console());

      offset += count;
      size -= count;

      int nameLength = inode.namelen<<2;

      char *filename = new char[nameLength+1];
      count = _byteSource->read(offset,
	  nameLength, filename, nameLength);
      Assert(count == nameLength);
      filename[nameLength] = '\0';

      offset += count;
      size -= count;

      Console() << "count: " << asHex(count) << " filename: ["
	<< filename << "] offset:" << asHex(offset)
	<< " size: " << asHex(size) << "\n" << eor;
    }
#endif

    initializeArrays(_superBlock.fsinfo.files);
}

/**
 * Destroy the cramfs container.
 */
CramfsContainer::~CramfsContainer()
{
}

/**
 * Determine if an idNumber is in the range considered valid on the container.
 * @param idNumber ID number to validate
 * @returns Always returns true in cramfs.
 */
bool CramfsContainer::IsIdNumberInRange( int idNumber )
{
  return true;
}

/**
 * Open a file from the cramfs file system.
 * @param idNumber The inode indicating which file to open.
 * @param status [out] Value indicating the status of the operation.
 * @param flags File system specific contols.
 * @return A file object representing the newly opened file.
 */
FileObject *CramfsContainer::basicOpen(int idNumber, int &status, int flags)
{
  Raid( RaidMethod )( "%N::basicOpen( inumber %d )\n", this, idNumber );
#ifdef JEFF
  Console() << this << "::basicOpen( inumber " << idNumber << " )\n" << eor;
#endif

  FileObject *fo = NULL;

  if(idNumber < 1) {
    status = IdNumberOutOfRange;
  }
  else {
    CramfsInode *pInode = new CramfsInode();
    int bytesRead;

    if(idNumber == 1) {
      memcpy(pInode, &(_superBlock.root), sizeof(CramfsInode));
      idNumber = offsetof(struct CramfsSuperblock, root);
#ifdef JEFF
      Console() << "new idNumber is " << (pInode->offset<<2) << ".\n" << eor;
#endif
    }
    else {
      /* read the CramFS inode */
      bytesRead = _byteSource->read(idNumber, sizeof(CramfsInode),
	  (char *)pInode, sizeof(CramfsInode));
      Assert( bytesRead == sizeof(CramfsInode) );
    }

#ifdef JEFF
    pInode->dumpTo(Console());
#endif

    fo = (pInode->mode & CramfsIsDirMode) ? 
      new CramfsDirObject(pInode->offset<<2, this, _byteSource, pInode) :
      new CramfsFileObject(idNumber, this, _byteSource, pInode);
    status = 0;
  }

  Raid( RaidMethod )( "%N::basicOpen returns %N %d\n", this, fo, status );
#ifdef JEFF
  Console() << this << "::" << __FUNCTION__ << " returns fo " << fo <<
    ", status " << status << "\n" << eor;
#endif
  return fo;
}

/**
 * Attempt to create a Cramfs object using the source as a backing store.
 * @param source A PersistentMemoryObject containg file system data and metadata.
 * @return A new cramfs container.
 */
ProxiableObject *
CramfsContainerConstructor(ProxiableObject *source)
{
  if(source->isKindOf(PersistentMemoryObjectClass)) { 
    return new CramfsContainer((PersistentMemoryObject *)source);
  }
  else {
    return NULL;
  }
}

#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("CramfsContainer",CramfsContainer,MemoryObjectContainer,&CramfsContainerConstructor,0);
