#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The Mobile Operating Systems Group
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 201 N. Goodwin Ave.
 *		 Urbana, IL	61801
 *
 * Copyright 2005 University of Illinois Board of Trustees. All Rights Reserved.
 *
 * Author: Jeffrey Carlyle (jcarlyle@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 */
/*
 * CramfsDictionary.cc - Directory table for cramfs file system.
 */

#include "Raid.h"
#include "CramfsDictionary.h"
#include "Cramfs.h"
#include "FileSystem.h"
#include "Convert.h"
#include "String.h"
#include "PersistentMemoryObject.h"

#ifdef __GNUC__
#pragma implementation
#endif

/* #define JEFF */

/**
 * Construct a new directory table object.
 * @param source Backing store for directory table.
 */
CramfsDictionary::CramfsDictionary(PersistentMemoryObject * source)
  : MemoryObjectDictionary(source)
{
  initDebug();
  Raid( RaidConstructor ) ( "%N::CramfsDictionary( %N )\n", this, source );

  // reMap();

  _maxKeyLength = CramfsMaxPath;
}

/**
 * Destructor for directory table.
 */
CramfsDictionary::~CramfsDictionary()
{
  Raid( RaidConstructor ) ( "%N::~CramfsDictionary()\n", this );
}

/**
 * Find the inode matching the specified key.
 * 
 * Side-effects expected by MemoryObjectDictionary:
 *  - _offset should be set to location of found key.
 *  - _emptyOffset should be set to an empty slot. (Unused in Cramfs.)
 *  - _idNumberOfKey should be set to the inode of the found key.
 * 
 * @param key Name of entry to locate.
 * @param needSize Find directory hole of specified size. (Unused in Cramfs.)
 * @return 1 if key is found; 0 otherwise.
 */
int CramfsDictionary::findKey(char *key, int needSize)
{
  Raid( RaidMethod )( "%N::findKey( key %s, needSize %d )\n", this,
      key, needSize );
#ifdef JEFF
  Console() << this << "::" << __FUNCTION__ << "( key " << key << ", needSize "
    << needSize << " )\n" << eor;
#endif /* JEFF */

  _emptyOffset = -1;
  if( ( strcmp( key, "." ) == 0 ) ||
      ( strcmp( key, ".." ) == 0 ) )
  {
    _idNumberOfKey = 1;
    return 1;
  }
#ifdef JEFF
  Console() << this << "::" << __FUNCTION__ << "( _source " << _source
    << " )\n" << eor;
#endif /* JEFF */

  int offset = 0;
  while(offset < _size) {
    CramfsInode inode;
    char filename[CramfsMaxPath + 1];
    
    int bytesRead = this->readEntry(offset, &inode, filename,
	sizeof(filename));
    
#ifdef JEFF
    Console() << this << "::" << __FUNCTION__ << "( filename " << filename
      << " )\n" << eor;
#endif /* JEFF */

    if( strcmp( key, filename ) == 0 ) {
      _idNumberOfKey = _source->idNumber() + offset;
      _offset = offset;

#ifdef JEFF
      Console() << this << "::" << __FUNCTION__
	<< " returns 1, _idNumberOfKey " << _idNumberOfKey
	<< ", _offset " << _offset << "\n" << eor;
#endif /* JEFF */
      return 1;
    }

    offset += bytesRead;
  }
  
#ifdef JEFF
  Console() << this << "::" << __FUNCTION__ << " returns 0\n" << eor;
#endif /* JEFF */
  return 0;
}

/**
 * Read one directory entry.
 * @param offset [in] Offset in the directory table at which to read entry.
 * @param pInode [out] Target for inode.
 * @param filename [out] Target for the name of the file.
 * @param nbytes [in] The size of the filename buffer.
 * @returns The total number of bytes read.
 */
int CramfsDictionary::readEntry(int offset, CramfsInode *pInode,
    char *filename, int nbytes)
{
  Assert( nbytes >= (CramfsMaxPath + 1) );

  int count = _source->read(offset, sizeof(CramfsInode), (char *)pInode,
      sizeof(CramfsInode));
  Assert( count == sizeof(CramfsInode) );
#ifdef JEFF
  pInode->dumpTo(Console());
#endif

  offset += count;

  int nameLength = (pInode->namelen)<<2;

  count += _source->read(offset, nameLength, filename, nameLength);
  Assert( count == (nameLength + (int)sizeof(CramfsInode)) );

  filename[nameLength] = '\0';

  return count;
}

/**
 * Dump the contents of an inode to an output stream.
 * @param out Target stream.
 */
void CramfsInode::dumpTo(OutputStream &out)
{
  out << "CramfsInode:mode:" << asOctal(mode) << ":uid:" << uid << ":size:"
    << asHex(size) << ":gid:" << gid << ":namelen:" << asHex(namelen<<2)
    << ":offset:" << asHex(offset<<2) << "\n" << eor;
}

/**
 * Get the list of named objects.
 * @param keyBuffer [out] Target for list of keys.
 * @param bufferSize The size of keyBuffer.
 * @param count [out] The number of keys found.
 * @return Status of the search.
 */
int CramfsDictionary::basicKeys(char* keyBuffer, int bufferSize, int& count)
{
  Raid( RaidMethod )( "%N::basicKeys( buffer %x ) _size %d\n",
      this, keyBuffer, bufferSize, _size );
#ifdef JEFF
  Console() << this << "::" << __FUNCTION__ << "( keyBuffer " <<
    asHex(keyBuffer) << " ) _size " << _size << "\n" << eor;
#endif

  int bytes = (bufferSize >= _size ? _size : bufferSize) - 1;
  char *t = keyBuffer;

  if(bytes <= 6)
    return TooManyKeys;

  strcpy(t, ".\n..\n");
  count = 2;
  bytes -= 5;
  t += 5;

  char *next = t;

  int offset = 0;
  for(; (offset < _size) && (bytes > 0); bytes--) {
    CramfsInode inode;
    char filename[CramfsMaxPath + 1];
    
    offset += this->readEntry(offset, &inode, filename,
	sizeof(filename));
    
#ifdef JEFF
    Console() << this << "::" << __FUNCTION__ << "( filename " << filename
      << " )\n" << eor;
#endif

    char *s = filename;
    while( (*s != '\0') && (bytes > 0) ) {
#ifdef JEFF
      Console() << *s << eor;
#endif
      *t++ = *s++;
      bytes--;
    }

    if(bytes == 0) {
      break;
    }

    *t++ = '\n';
    bytes--;

    next = t;
    count++;
  }

  *next = '\0';
#ifdef JEFF
  Console() << "\n" << eor;
#endif

  /* check if we got the entire list of directories */
  if(offset == _size)
    return 0;

  return TooManyKeys;
}

/**
 * Construct a directory table from the specified source object, if possible.
 * @param The backing store for the directory table.
 */
ProxiableObject *
CramfsDictionaryConstructor( ProxiableObject * source )
{
  if( source->isKindOf( PersistentMemoryObjectClass ) ) {
    return( new CramfsDictionary( (PersistentMemoryObject *)source ) );
  }
  else
    return 0;
}

#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("CramfsDictionary",CramfsDictionary,MemoryObjectDictionary,&CramfsDictionaryConstructor,0);
