#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The Mobile Operating Systems Group
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 201 N. Goodwin Ave.
 *		 Urbana, IL	61801
 *
 * Copyright 2005 University of Illinois Board of Trustees. All Rights Reserved.
 *
 * Author: Jeffrey Carlyle (jcarlyle@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 */
/*
 * CramfsDirObject.cc - Cramfs directory entry (inode) code.
 */
#include "CramfsDirObject.h"
#include "Raid.h"
#include "UNIXFile.h"

#ifdef __GNUC__
#pragma implementation
#endif

/* #define JEFF */

extern ClassStar CramfsDictionaryClass;

/**
 * Construct an object representing a file in the Cram filesytem.
 * @param idNumber Offset of the inode for this file
 * @param container CramFS filesystem container.
 * @param source Disk containing the file system.
 * @param pInode CramFS inode for the file.
 */
CramfsDirObject::CramfsDirObject(int idNumber,
    MemoryObjectContainer *container, FileObject *source, CramfsInode *pInode)
  : FileObject( pInode->size, 0, idNumber, container, source,
      (pInode->offset<<2)), _pInode(pInode), _supports(CramfsDictionaryClass)
{
  initDebug();
  Raids( RaidConstructor ) << this << "::" << __FUNCTION__ << "(  idNumber "
    << idNumber << ", container " << container << ", source "
    << source << ", pInode " << pInode << ")\n" << eor;
#ifdef JEFF
  Console() << this << "::" << __FUNCTION__ << "(  idNumber "
    << idNumber << ", container " << container << ", source "
    << source << ", pInode " << pInode << ")\n" << eor;
#endif
}

/**
 * Destructor.
 */
CramfsDirObject::~CramfsDirObject()
{
  Raids( RaidConstructor ) << this << "::~CramfsDirObject()\n" << eor;
}

/**
 * Determine if this fileObject can export the requested interface.
 * @param aClass Requested interface.
 * @return Class supporting the requested interface or null.
 */
Class *CramfsDirObject::supports(Class * aClass)
{
  Raid( RaidProtected )( "%N::supports( %N )\n", this, aClass );

  Class * supportedClass = NULL;

  if((this->isDictionary()) && 
      (_supports->isASubclassOf(aClass))) {
    supportedClass = _supports;
  }

  Raid( RaidProtected )( "%N::supports() returns %N\n", this, supportedClass );

  return supportedClass;
}

/**
 * Determine if this fileObject is a directory file.
 * @return 1 if this is a directory; false otherwise.
 */
int CramfsDirObject::isDictionary(void)
{
  return _pInode->mode & CramfsIsDirMode;
}

/**
 * Get the UID of the current fileObject.
 * @return The UID of the owner of the current fileObject.
 */
int CramfsDirObject::userId(void)
{
  return _pInode->uid;
}

/**
 * Get the GID of the current fileObject.
 * @return The GID of the group that owns the current fileObject.
 */
int CramfsDirObject::groupId(void)
{
  return _pInode->gid;
}

/**
 * Get the access mode of the current file.
 * @return The UNIX-style access mode mask for the current file.
 */
int CramfsDirObject::accessMode(void)
{
  return _pInode->mode & UNIXAccessMask;
}

/**
 * Get a stat(2) compatible block of information about this fileObject.
 * @param infoBuffer Target for information.
 */
void CramfsDirObject::info(UNIXFileInfo *infoBuffer)
{
  Raid( RaidMethod )( "%N::info() %d\n", this, _idNumber );

  infoBuffer->device  = 0;
  infoBuffer->inumber = _idNumber;
  infoBuffer->mode    = accessMode();
  infoBuffer->mode   |= isDictionary() ? UNIXTypeDirectory : UNIXTypeRegular;
  infoBuffer->links   = 1;
  infoBuffer->uid     = userId();
  infoBuffer->gid     = groupId();
  infoBuffer->flags   = 0;    //FIX
  infoBuffer->rdevice = 0;    //FIX
  infoBuffer->size    = numberOfBytes();
  infoBuffer->atime   = 0;
  infoBuffer->mtime   = 0;
  infoBuffer->ctime   = 0;
  infoBuffer->blockSize = unitSize();
  infoBuffer->blocks  = numberOfUnits();
}

#include "ClassInstaller.h"
DEFINECLASS("CramfsDirObject",CramfsDirObject,FileObject,0,0);
