#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The Mobile Operating Systems Group
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 201 N. Goodwin Ave.
 *		 Urbana, IL	61801
 *
 * Copyright 2005-2006 University of Illinois Board of Trustees.
 * All Rights Reserved.
 *
 * Author: Jeffrey Carlyle (jcarlyle@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 */
/*
 * CramfsFileObject.cc - Cramfs directory entry (inode) code.
 */
#include "CramfsFileObject.h"
#include "Raid.h"
#include "UNIXFile.h"
#include "zlib.h"
#include "Byte.h"
#include "ELFContainer.h"
#include "PersistentClass.h"

#ifdef __GNUC__
#pragma implementation
#endif

/* #define JEFF */

/**
 * Construct an object representing a file in the Cram filesytem.
 * @param idNumber Offset of the inode for this file
 * @param container CramFS filesystem container.
 * @param source Disk containing the file system.
 * @param pInode CramFS inode for the file.
 */
CramfsFileObject::CramfsFileObject(int idNumber,
    MemoryObjectContainer *container, FileObject *source, CramfsInode *pInode)
  : CramfsDirObject( idNumber, container, source, pInode )
{
  initDebug();
  Raids( RaidConstructor ) << this << "::" << __FUNCTION__ << "(  idNumber "
    << idNumber << ", container " << container << ", source "
    << source << ", pInode " << pInode << ")\n" << eor;
#ifdef JEFF
  Console() << this << "::" << __FUNCTION__ << "(  idNumber "
    << idNumber << ", container " << container << ", source "
    << source << ", pInode " << pInode << ")\n" << eor;
#endif
}

/**
 * Destructor.
 */
CramfsFileObject::~CramfsFileObject()
{
  Raids( RaidConstructor ) << this << "::~CramfsFileObject()\n" << eor;
}

/**
 * Determine if this file contains another file system.
 *
 * @return  Non-zero for true; zero for false.
 */
int CramfsFileObject::isContainer()
{
    return (this->supports(ELFContainerClass) != NULL);
}

/**
 * Test if the file supports the given class.
 *
 * @param   aClass  Class to test support for.
 *
 * @return  Type of supported class or NULL if not supported.
 */
Class *CramfsFileObject::supports(Class *aClass)
{
#ifdef JEFF
    Console() << this << "::" << __FUNCTION__ << "()\n" << eor;
#endif

    /* test for ELF object */
    if(ELFContainerClass->isASubclassOf(aClass)) {
	/* we want the first four bytes of the file */
	unsigned numUnits = ((4-1) / this->unitSize())+1;
	unsigned bufSize = numUnits * this->unitSize();
	char *buffer = new char[bufSize];

	AssertOrThrow(buffer != NULL);

	this->read(0, numUnits, buffer, bufSize);

	const char magic_word[] = {'\x7f', 'E', 'L', 'F'};
	bool matched = true;

	for(int i = 0; i<4; i++) {
	    if(magic_word[i] != buffer[i]) {
		matched = false;
		break;
	    }
	}

	delete[] buffer;

	if(matched) {
#ifdef JEFF
	    Console() << this << "::" << __FUNCTION__ << " returning "
		<< ELFContainerClass << "\n" << eor;
#endif
	    return ELFContainerClass;
	}
    }
    
    return this->CramfsDirObject::supports(aClass);
}


/**
 * Read some number of blocks out of the file.
 * @param startUnit First block to read.
 * @param count Number of block to read.
 * @param dest Target for read blocks.
 * @param nbytes Size of target.
 * @param c Closure for the operation.
 * @return Number of bytes read.
 */
int CramfsFileObject::rawRead(unsigned startUnit, unsigned count, char * dest,
    unsigned nbytes, IOAsyncClosure * c)
{
  Raids( RaidProtected ) << this
    << "::" << __FUNCTION__ << "( startUnit " << startUnit
    << ", count " << count 
    << ", dest " << asHex( (int) dest )
    << ", nbytes " << nbytes 
    << " )\n" << eor;
#ifdef JEFF
  Console() << this
    << "::" << __FUNCTION__ << "( startUnit " << startUnit
    << ", count " << count 
    << ", dest " << asHex( (int) dest )
    << ", nbytes " << nbytes 
    << " )\n" << eor;
#endif /* JEFF */

  Assert((unsigned)nbytes >= (count * unitSize()));

  /* does not currently implement asynchronous support */

  int bytesRead, totalBytesRead = 0;
  unsigned stopUnit = startUnit + count;
  unsigned startBlock = startUnit / CramfsBlockSize;
  unsigned stopBlock = (stopUnit - 1) / CramfsBlockSize;
  unsigned blockCount = ((numberOfBytes() - 1) / CramfsBlockSize) + 1;
  unsigned blockStartOffset, nextBlockStartOffset, chunkSize;
  char *ptr = dest;

  if(startBlock==0) {
    blockStartOffset = ((_pInode->offset)<<2) + (blockCount * sizeof(int));
  }
  else {
    bytesRead = _source->read(((_pInode->offset)<<2) +
	((startBlock-1) * sizeof(unsigned)), sizeof(unsigned),
	(char *)&blockStartOffset, sizeof(unsigned));
    Assert( bytesRead == sizeof(unsigned) );
  }

  for(unsigned block = startBlock; block <= stopBlock; block++) {
    bytesRead = _source->read(((_pInode->offset)<<2) +
	(block * sizeof(unsigned)), sizeof(unsigned),
	(char *)&nextBlockStartOffset, sizeof(unsigned));
    Assert( bytesRead == sizeof(unsigned) );

    chunkSize = nextBlockStartOffset - blockStartOffset;
#ifdef JEFF
    Console() << "chunkSize " << chunkSize << " nextBlockStartOffset "
      << nextBlockStartOffset << " blockStartOffset " << blockStartOffset
      << " pInode->offset<<2 " << (_pInode->offset<<2) << " numberOfUnits "
      << numberOfUnits() << "\n" << eor;
#endif /* JEFF */
    char *chunk = new char[chunkSize];
    Assert( chunk != NULL );

    char *page = new char[CramfsBlockSize];
    Assert( page != NULL );

    bytesRead = 0;
    int tmpBytesRead = 0;
    while(bytesRead < (int)chunkSize) {
        tmpBytesRead = _source->read(blockStartOffset+tmpBytesRead,
                chunkSize-tmpBytesRead, chunk+tmpBytesRead,
                chunkSize-tmpBytesRead);
#ifdef JEFF
        Console() << "bytesRead: " << bytesRead << " tmpBytesRead: "
            << tmpBytesRead << " totalBytesRead: " << totalBytesRead
            << "\n" << eor;
#endif /* JEFF */
        Assert( tmpBytesRead > 0 );

        bytesRead += tmpBytesRead;
    }
    Assert( bytesRead == (int)chunkSize );

    unsigned blockSize = CramfsBlockSize;
#ifdef JEFF
    Console() << "blockSize " << blockSize << "\n" << eor;
#endif /* JEFF */
    int zret = uncompress((Bytef*)page, (uLongf*)&blockSize,
	(Bytef*)chunk, chunkSize);
#ifdef JEFF
    Console() << "blockSize " << blockSize << " zRet " << zret << "\n" << eor;
#endif /* JEFF */
    Assert( zret == Z_OK );
    Assert( blockSize <= CramfsBlockSize );

    delete[] chunk;

    unsigned byteStart = (block == startBlock)
      ? startUnit - (block * CramfsBlockSize)
      : 0;

    unsigned numCopy;

    if((block == startBlock)&&(block != stopBlock))
      numCopy = CramfsBlockSize - byteStart;
    else if((block == startBlock)&&(block == stopBlock))
      numCopy = stopUnit - startUnit;
    else if((block != startBlock)&&(block == stopBlock))
      numCopy = stopUnit - (block*CramfsBlockSize);
    else
      numCopy = CramfsBlockSize;

#ifdef JEFF
    Console() << "byteStart: " << byteStart << " numCopy: " << numCopy
      << " blockSize: " << blockSize << " ptr: " << asAddress(ptr)
      << "\n" << eor;
#endif /* JEFF */
    ByteCopy(page+byteStart, ptr, blockSize < numCopy ? blockSize : numCopy);
    delete[] page;

    totalBytesRead += blockSize;
    ptr += blockSize;

    blockStartOffset = nextBlockStartOffset;
  }

  if(c) {
    c->activate(totalBytesRead);
  }

#ifdef JEFF
  Console() << this
    << "::" << __FUNCTION__ << " done"
    << " return = "
    << totalBytesRead << "\n" << eor;
#endif /* JEFF */
  Raids( RaidProtected ) << this
    << "::" << __FUNCTION__ << " done"
    << " return = "
    << totalBytesRead << "\n" << eor;
  return totalBytesRead;
}

#include "ClassInstaller.h"
DEFINECLASS("CramfsFileObject",CramfsFileObject,CramfsDirObject,0,0);
