#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * ELFContainer.cc - ELF Inode System class : dthomas1@cs.uiuc.edu
 */

#include "NewAndDelete.h"
#include "elF.h"
#include "ELFContainer.h"
#include "RootEntry.h"
#include "FileSystem.h"
#include "String.h"
#include "Raid.h"

#ifdef __GNUC__
#pragma implementation
#endif


using namespace llvm::ELF;

ELFContainer::ELFContainer( PersistentMemoryObject * source ) 
    : MemoryObjectContainer( source ), _byteSource( 0 )
{
    initDebug();

    Raid( RaidConstructor )( "%N::ELFContainer( %N )\n", this, _source );

    _valid = 0;

    int objects = 0;
    unsigned bytes = _source->numberOfBytes();

    if( bytes < sizeof( filehdr ) ) {
        Raid( RaidDetail )( "%N::ELFContainer() too small %d\n",
                    this, _source->numberOfBytes() );
    }
    else {
        _byteSource = new FileObject( bytes, 0, 0, 0, _source );

        _filehdr = new filehdr;
        _byteSource->read( 0, sizeof( filehdr ), (char *)_filehdr,
                sizeof( filehdr ) );

        if (! _filehdr->checkMagic() ) {
            Console()<<"ELFContainer: Error! Bad magic number\n";
            delete _filehdr;
        }
        else {
            _scnhdrs = 0;
            _valid = 1;
        }
    }
    if( _valid ) {
	// Verify that the program header size is zero
	// We restrict ourself to relocatable files @DCT
        // Assert( ( _filehdr->e_phnum == 0 ) );

        int offset =  _filehdr->e_shoff;
        int length =_filehdr-> e_shentsize* _filehdr->e_shnum;

        _scnhdrs = new scnhdr[_filehdr->e_shnum];
        _byteSource->read( offset, length, (char *)_scnhdrs, length );


        if( _filehdr->e_shstrndx != SHN_UNDEF ) {
                _strTable = new char[ _scnhdrs[_filehdr->e_shstrndx].sh_size];
                offset = _scnhdrs[_filehdr->e_shstrndx].sh_offset;
                length = _scnhdrs[_filehdr->e_shstrndx].sh_size;
                _byteSource->read( offset, length, (char *)_strTable, length );
        }

	objects = SectionIds + ( 4 * _filehdr->e_shnum );
    }
    initializeArrays( objects );
    Console() << "ELFContainer " << this << " initialized OK\n";
}

ELFContainer::~ELFContainer()
{
    Raid( RaidConstructor )( "%N::~ELFContainer()\n", this );

    if( _valid ) {
        delete _scnhdrs;
        delete _filehdr;
    }

    Raid( RaidConstructor )( "%N::~ELFContainer done\n", this );
}

int ELFContainer::isValid(void) const
{
    return _valid;
}

FileObject *
ELFContainer::basicOpen( int id, int & status, int flags )
{
    Raid( RaidMethod )( "%N::basicOpen( inumber %d )\n", this, id );

Console()<<this<<"::basicOpen( "<<id<<" )\n";

    if( id < RootId ) {
        status = IdNumberOutOfRange;
        return( 0 );
    }
    else {
        status = 0;
        int offset = 0;
        int units = 0;
        if( id == RootId ) {
            Console()<<"Opening ELF file root..."<<eor;
            extern ClassStar ELFDictionaryClass;
            FileObject *re=new RootEntry(_filehdr->e_shoff +    
                        _filehdr->e_shnum * _filehdr->e_shentsize,
                        0,
                        id, this, _source,
                        ELFDictionaryClass, 0 ) ;
            Console()<<"done\n";
            return re;
        }
        else if( id == FileHeaderId ) {
            offset = 0;
            units = _filehdr->e_ehsize; //sizeof( filehdr );
            return( new FileObject( units, 0,
                        id, this, _source, offset ) );
        }
        else if( id == StringSymbolId ) {
                        for (int i=0; i<_filehdr->e_shnum; i++){
                                if (_scnhdrs[i].sh_type == SHT_SYMTAB){
                    offset = _scnhdrs[_scnhdrs[i].sh_link].sh_offset;
                    units = _scnhdrs[_scnhdrs[i].sh_link].sh_size;
                    Console() << "StringSymbolId" << offset << " ---- " << units << "\n";
                }
                        }

            return( new FileObject( units, 0,
                        id, this, _source, offset ) );
        }

        else if( id == SymbolTableId ) {

            for (int i=0; i<_filehdr->e_shnum; i++){
                if (_scnhdrs[i].sh_type == SHT_SYMTAB){
                    offset = _scnhdrs[i].sh_offset;
                  units  = _scnhdrs[i].sh_size;
                }
            }
            return( new FileObject( units, 0,
                        id, this, _source, offset ) );
        }
        else if( id == StringTableId ) {
            if( _filehdr->e_shstrndx != SHN_UNDEF ) {
                offset = _scnhdrs[_filehdr->e_shstrndx].sh_offset;
                units = _scnhdrs[_filehdr->e_shstrndx].sh_size;
            }
            return( new FileObject( units, 0,
                        id, this, _source, offset ) );
        }
        else {
            int section = id - SectionIds;
            int type = section & 3;
            section >>= 2; 

            int relocIndex;
            switch( type ) {
            case Header:
                offset = _filehdr->e_shoff +
                     section * _filehdr->e_shentsize;
                units = _filehdr->e_shentsize;
                break;
            case Section:
                offset = _scnhdrs[section].sh_offset;
                units = _scnhdrs[section].sh_size;
                break;
            case Reloc:
                //Search for the .rel{a}[name] in the section header
                relocIndex = getRelocID(_scnhdrs[section].sh_name);
                offset = _scnhdrs[relocIndex].sh_offset;
                units = _scnhdrs[relocIndex].sh_size;
                break;
#if 0
            case LineNo:
                offset = _scnhdrs[section].s_lnnoptr;
                units = _scnhdrs[section].s_nlnno * LINENOSIZE;
                break;
#endif
            default:
                Assert( NOTREACHED );
            }
            return( new FileObject( units, 0,
                        id, this, _source, offset ) );
        }
    }
}

// Create the source string first
// Prepend .rela to src string
// lookup section header again to see if any section matches it
// Return the SectionID of the matched string 
int
ELFContainer::getRelocID(int strIndex)
{
    char *origSection = &_strTable[strIndex];
    strcpy(_strName, ".rela");
    strcat(_strName, origSection);
    for (int i=0; i<_filehdr->e_shnum; i++){
        if (strcmp(&_strTable[_scnhdrs[i].sh_name],_strName) == 0){
            return (i);
        }
    }
  return (0); /* Not all sections have reloc sections */
}


int
ELFContainer::basicRootId()
{
    return( RootId );
}

ProxiableObject * 
ELFContainerConstructor( ProxiableObject * source )
{
    if( source->isKindOf( PersistentMemoryObjectClass ) ) {
        return( new ELFContainer( (PersistentMemoryObject *)source ) );
    }
    else return( 0 );
}

#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("ELFContainer",ELFContainer,MemoryObjectContainer,&ELFContainerConstructor,0);
