#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * Copyright (c) 1990 IBM Corporation. All Rights reserved
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Author: Panos Kougiouris ( kougiour@cs.uiuc.edu )
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Ext2Container.cc - 
 *
 */
#include "Ext2Allocator.h"
#include "Ext2Container.h"
#include "Ext2Inode.h"
#include "Assert.h"
#include "Class.h"
#include "FileObject.h"
#include "FileSystem.h"
#include "Raid.h"
#include "UNIXFile.h"
#include "UNIXContainer.h"
#include "UNIXInodeIndirect.h"

#ifdef __GNUC__
#pragma implementation
#endif

extern ClassStar MemoryObjectDictionaryClass;

/*
 * File system states in Linux
 */
static const int EXT2_VALID_FS = 0x0001;  /* Unmounted cleanly */
static const int EXT2_ERROR_FS = 0x0002;  /* Errors detected */

Ext2Container::Ext2Container( PersistentMemoryObject * source )
    :UNIXContainer( source )
{
    initDebug();
    Raid( RaidConstructor )( "%N::Ext2Container( %N )\n", this, source );

    /*
     *  Calculate constants needed by UNIXContainer.
     */
    _dinodeSize = sizeof( Ext2_dinode );
    //    swapl((void *) &_dinodeSize);

    /*
     *  Create super block buffer, chainize it, read it from the disk,
     *  and check its magic number.
     */
    _superBlock = new ext2filsys;
    _superBlock->u.ext2_sb.s_es = new ext2_super_block;

    // TODO: This assumes that source->unitSize() <= 1024 which means that ext2 fs can
    // only exist on disks with sector sizes <= 1024 bytes
    // Also assumes that sector sizes are a factor of 1024
    _source->read( 1024/_source->unitSize(), sizeof(ext2_super_block) >> _source->log2UnitSize(), 
                   (char *) (_superBlock->u.ext2_sb.s_es),
                   sizeof(ext2_super_block));

    if( ( _superBlock->u.ext2_sb.s_es->s_magic != EXT2_SUPER_MAGIC ) ) {
        Raid( RaidError )( "_superBlock->s_magic %x\n",
                           _superBlock->u.ext2_sb.s_es->s_magic );
    }
    Assert( _superBlock->u.ext2_sb.s_es->s_magic == EXT2_SUPER_MAGIC );

    _superBlock->s_blocksize = 1024 << _superBlock->u.ext2_sb.s_es->s_log_block_size;
    _superBlock->s_magic = _superBlock->u.ext2_sb.s_es->s_magic;

    _superBlock->u.ext2_sb.s_frag_size = 1 << _superBlock->u.ext2_sb.s_es->s_log_frag_size;
    _superBlock->u.ext2_sb.s_frags_per_block = _superBlock->s_blocksize/_superBlock->u.ext2_sb.s_frag_size;
    _superBlock->u.ext2_sb.s_inodes_per_block = _superBlock->s_blocksize/_superBlock->u.ext2_sb.s_es->s_inode_size;
    _superBlock->u.ext2_sb.s_frags_per_group = _superBlock->u.ext2_sb.s_es->s_frags_per_group;
    _superBlock->u.ext2_sb.s_blocks_per_group = _superBlock->u.ext2_sb.s_es->s_blocks_per_group;
    _superBlock->u.ext2_sb.s_inodes_per_group = _superBlock->u.ext2_sb.s_es->s_inodes_per_group;
    _superBlock->u.ext2_sb.s_itb_per_group = _superBlock->u.ext2_sb.s_inodes_per_group * _dinodeSize / _superBlock->s_blocksize;
    _superBlock->u.ext2_sb.s_desc_per_block = _superBlock->s_blocksize / sizeof(ext2_group_desc);
    _superBlock->u.ext2_sb.s_groups_count = _superBlock->u.ext2_sb.s_es->s_blocks_count/_superBlock->u.ext2_sb.s_es->s_blocks_per_group;
    if (_superBlock->u.ext2_sb.s_groups_count == 0) _superBlock->u.ext2_sb.s_groups_count = 1;
    _superBlock->u.ext2_sb.s_db_per_group = _superBlock->u.ext2_sb.s_groups_count/_superBlock->u.ext2_sb.s_desc_per_block;
    if (_superBlock->u.ext2_sb.s_db_per_group == 0) _superBlock->u.ext2_sb.s_db_per_group = 1;
    _superBlock->u.ext2_sb.s_sbh = 0;
    unsigned numGroupDescriptorBlocks = _superBlock->u.ext2_sb.s_groups_count/_superBlock->u.ext2_sb.s_desc_per_block;
    if (numGroupDescriptorBlocks == 0) numGroupDescriptorBlocks = 1;
    _superBlock->u.ext2_sb.s_group_desc = new buffer_head*[numGroupDescriptorBlocks];
    _superBlock->u.ext2_sb.s_inode_size = _superBlock->u.ext2_sb.s_es->s_inode_size;

    Assert(_dinodeSize == _superBlock->u.ext2_sb.s_es->s_inode_size);

    Assert(numGroupDescriptorBlocks == 1);
    // Load the group descriptor blocks
    for (unsigned i = 0; i < numGroupDescriptorBlocks; i++) {
        _superBlock->u.ext2_sb.s_group_desc[i] = new buffer_head;
        _superBlock->u.ext2_sb.s_group_desc[i]->b_data = new char[_superBlock->s_blocksize];
        _source->read( _superBlock->s_blocksize/_source->unitSize(), _superBlock->s_blocksize >> _source->log2UnitSize(),
                       _superBlock->u.ext2_sb.s_group_desc[i]->b_data,
                       _superBlock->s_blocksize);
    }
    _inodeBlockSize = _superBlock->s_blocksize;
    //    swapl((void *) &_inodeBlockSize);
    _inodesPerBlock = _superBlock->u.ext2_sb.s_inodes_per_block;
    //    swapl((void *) &_inodesPerBlock);

    _log2FragmentToSector = 10 + _superBlock->u.ext2_sb.s_es->s_log_frag_size - source->log2UnitSize();
    //    swapl((void *) &_log2FragmentToSector);
        
    _indirect = UNIXInodeIndirect::setup(10 + _superBlock->u.ext2_sb.s_es->s_log_frag_size);

    /*
     *  Allocate and clear in memory and on disk inode pointers.
     *  Create one more object than there are inodes because object
     *  0 will be unused since inumbers start at 1.
     */
    int tmp_s_inodes_count = _superBlock->u.ext2_sb.s_es->s_inodes_count;
    swapl((void *) &tmp_s_inodes_count);

    int containerSize = ( ( tmp_s_inodes_count - 2 ) * _inodesPerBlock ) + 1;
    UNIXContainer::initializeArrays( containerSize );
    _allocator = new Ext2Allocator( _source, _superBlock,
                                    _log2FragmentToSector );
}

Ext2Container::~Ext2Container()
{
    Raid( RaidConstructor )( "%N::~Ext2Container\n", this );

    synchronize();

    delete [] _superBlock->u.ext2_sb.s_group_desc;
    delete _superBlock;
}

/*
 * Find block number containing this inode.
 */
int
Ext2Container::mapInumber( int inumber )
{
    return( 4 + ( ( inumber - 1 ) / _inodesPerBlock ) );
    //    return( 10 + ( ( inumber - 1 ) / _inodesPerBlock ) );
}

/*
 * Find first inode in block containing this inode.
 */
int
Ext2Container::firstInumber( int inumber )
{
    return( ( ( ( inumber - 1 ) / _inodesPerBlock ) * _inodesPerBlock ) + 1 );
}

/*
 *  Find an unused inode.
 *  Returns the inumber of the inode found, or zero if none found.
 *  The inode is marked in use by this function.
 */
int
Ext2Container::getUnusedInode( int )
{
    Raid( RaidMethod )( "Ext2Container::getUnusedInode()\n" );

    unsigned inumber = 0;
    if( _superBlock->u.ext2_sb.s_es->s_free_inodes_count >= 0 ) {
        //        struct ext2_group_desc * group_desc = _allocator->blockPref( _idNumber, buf, start, startBlock );

        inumber = 3;

        while (inumber <= _superBlock->u.ext2_sb.s_es->s_inodes_count) {
            UNIXInodeIndirectInfo ** info = 0;            
            Ext2_dinode *nextDinode = (Ext2_dinode *) getDinode( inumber, info );
            Assert( nextDinode != 0 );
            
            // If it is free break out of loop
            // Hack for now
            if (nextDinode->di_size == 0) break;
            inumber++;
        }
    
        if (inumber <= _superBlock->u.ext2_sb.s_es->s_inodes_count) {
            _superBlock->u.ext2_sb.s_es->s_free_inodes_count--;
            // The superblock is now modified.
            _modified = 1;
            _superBlock->s_time = 0;
        } else {
            inumber = 0;
        }
    }
    Debug.printf( "Ext2Container::getUnusedInode, return = %d\n", inumber );
    Raid( RaidMethod )( "Ext2Container::getUnusedInode, return = %d\n", inumber );
    return( inumber );
}

/*
 *  Open returns a pointer to a sub-memory-object ( inode )
 *  with the given number.
 *  If the inumber is negative, an unused one is allocated and initialized.
 *  Also, the new inumber is passed back thru the reference parameter.
 *  If the inode is already in use, update its reference count.
 *  Otherwise create a new in memory inode and return a pointer to it.
 *  Creating a new memory inode will also read in the disk inode if needed.
 */
FileObject *
Ext2Container::basicOpen( int inumber, int & status ,int flag)
{

    if( inumber < basicRootId() ) {
        status = IdNumberOutOfRange;
        return( 0 );
    }

    /*
     *  Read this inode from the disk,
     *  Check to see if it exists,
     *  and Construct it in memory structure.
     */
    acquire();
    bind( inumber, (FileObject *) ~0 );
    UNIXInodeIndirectInfo ** info = 0;
 
    Ext2_dinode * thisDinode = (Ext2_dinode *) getDinode( inumber, info );

    if( thisDinode->di_nlink == 0 ) {
        status = UnusedIdNumber;
        unbind( inumber );
        writeDinode( inumber, 1 );
        release();
        return( 0 );
    }

    release();

    status = 0;
    //    UNIXInodeIndirectInfo *fake_indirect_inode = new UNIXInodeIndirectInfo();
    //    return( new Ext2Inode( numFragments( thisDinode->di_size ),
    //        _source->log2UnitSize() + _log2FragmentToSector,
    //             inumber, this, _source,
    //             _superBlock->u.ext2_sb.s_es->s_log_block_size,
    //             thisDinode, &fake_indirect_inode ) );
    unsigned int tmp_size = thisDinode->di_size;
    swapl((void *) &tmp_size);

    return( new Ext2Inode( numFragments( tmp_size ),
                           _source->log2UnitSize() + _log2FragmentToSector,
                           inumber, this, _source, 0,
                           // _superBlock->u.ext2_sb.s_es->s_log_block_size,
                           thisDinode, info ) );
}

/*
 *  create returns a pointer to a new sub-memory-object ( inode )
 *  The inumber passed in is just used as a hint to the file system.
 *  Also, the new inumber is passed back thru the same parameter.
 *  The mode parameter is primarily used to set the type of the
 *  file, it should not change during the life of the file.
 *  Uid, gid, and file permissions can be set via methods on an Inode.
 */
FileObject *
Ext2Container::basicCreate( int inumber, Class * aClass, int & status ,int flag)
{
    Raid( RaidMethod )( "basicCreate( inumber %d, Flag %d, )\n", inumber, flag);

    acquire();

    inumber = getUnusedInode( 0 );
    swapl((void *) &inumber);

    if( inumber <= 0 ) {
        status = ContainerFull;
        release();
        return( 0 );
    }

    /*
     *  Read the inode from disk, and reset all of its fields.
     */
    // TODO: Francis I don't think we need to pass _diskInodeCacheInfo
    Ext2_dinode *thisDinode = (Ext2_dinode *) getDinode( inumber, _diskInodeCacheInfo );
    UNIXInodeIndirectInfo ** info = _diskInodeCacheInfo;

    thisDinode->di_nlink = 0;
    swaps((void *) &(thisDinode->di_nlink));

    thisDinode->di_mode = UNIXAccessMask;

    if( aClass->isASubclassOf( MemoryObjectDictionaryClass ) ) {
        thisDinode->di_mode |= UNIXTypeDirectory;
        thisDinode->di_mode |= UNIXTypeDirectory;

        thisDinode->di_flags = EXT2_DIALLOC | EXT2_DILONGDIR | EXT2_DISTORE;
        swapl((void *) &thisDinode->di_flags);

    }
    else {
        thisDinode->di_mode |= UNIXTypeRegular;

        thisDinode->di_flags = EXT2_DIALLOC | EXT2_DISTORE;
        swapl((void *) &(thisDinode->di_flags));
    }

    swaps((void *) &(thisDinode->di_mode));

    thisDinode->di_uid = 0;
    swaps((void *) &thisDinode->di_uid);

    thisDinode->di_gid = 0;
    swaps((void *) &thisDinode->di_gid);

    thisDinode->di_size = 0;
    swapl((void *) &thisDinode->di_size);

    // FIX: what time is it?
    thisDinode->di_atime = 0;
    swapl((void *) &thisDinode->di_atime);

    thisDinode->di_mtime = 0;
    swapl((void *) &thisDinode->di_mtime);

    thisDinode->di_ctime = 0;
    swapl((void *) &thisDinode->di_ctime);

    /*
     *  Create the new in memory inode.
     */
    status = 0;

    //    setModified( inumber );
    *((*_diskInodeCacheInfo)->dirtyFlag()) = 1;

    release();

    return( new Ext2Inode( numFragments( thisDinode->di_size ),
                           _source->log2UnitSize() + _log2FragmentToSector,
                           inumber, this, _source,
                           0,
                           //_superBlock->u.ext2_sb.s_es->s_log_block_size,
                           thisDinode, info ) );
}

void
Ext2Container::basicClose( FileObject * thisInode )
{
    Raid( RaidMethod )( "%N::basicClose( %N )\n", this, thisInode );

    int inumber = thisInode->idNumber();

    Raid( RaidMethod )( "%N::basicClose() inumber %d \n", this, inumber );

    /*
     *  If there are no links to this inode, then this call will free its space.
     *  FIX: this isnt ready yet.
     */
    int links = thisInode->links( 0 );

    Raid( RaidMethod )( "%N::basicClose() links %d \n", this, links );

    MemoryObjectContainer::basicClose( thisInode );

    /*
     *  Write the disk inode back to disk.
     */

    acquire();
    writeDinode( inumber, 1 );

    if( links == 0 ) {
        // update the free count
        _superBlock->u.ext2_sb.s_es->s_free_inodes_count++;
        _modified = 1;
        _superBlock->u.ext2_sb.s_es->s_wtime = 0;
    }
    release();
}

int
Ext2Container::basicRootId()
{
    return( 2 );
}

/*
 *  Synchronize the in-core data structures with the disk.
 */
void
Ext2Container::basicSynchronize()
{
    Raid( RaidMethod )
    ( "%N->Ext2Container::basicSynchronize() %d\n", this, _modified );

    // TODO: Francis 
    // This is hard coded for specific block sizes
    _source->write( 2, 2, (char *) _superBlock, 1024 );

    basicSynchronizeInodes();
}

ProxiableObject * 
Ext2ContainerConstructor( ProxiableObject * source )
{
    if( source->isKindOf( PersistentMemoryObjectClass ) ) {
        return( new Ext2Container( (PersistentMemoryObject *)source ) );
    }
    else return( 0 );
}

unsigned
Ext2Container::numFragments( unsigned size )
{
    /* Other code crashes if this value is 0 */
    //if( size <= SMBLKSZ ) return 0;
    unsigned fragmentSize = _source->unitSize() << _log2FragmentToSector;
    return( ( size + fragmentSize - 1 ) / fragmentSize );
}

int
Ext2Container::blockPref( int inumber, int * buf, int start, int startBlock )
{
    if ( start % _superBlock->u.ext2_sb.s_es->s_blocks_per_group == 0 || buf[ start - 1 ] == 0 ) {
        if ( start < EXT2_NDIR_BLOCKS ) {
            int cg = inumber / _superBlock->u.ext2_sb.s_es->s_inodes_per_group;
            return ( _superBlock->u.ext2_sb.s_es->s_inodes_per_group * cg ); // + _superBlock->fs_frag ?
        }

        return 0;

    }

    return (buf[ start - 1 ] + 
            (_superBlock->u.ext2_sb.s_es->s_log_block_size / _superBlock->u.ext2_sb.s_es->s_log_frag_size));

}

#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("Ext2Container",Ext2Container,UNIXContainer,&Ext2ContainerConstructor,0);
