/*
 * This file is part of the Choices Operating System
 * Author: Yevgeniy Rabinovich <rabinvch@uiuc.edu>,
 *                             <yrabinov+uiuc@cs.uchicago.edu>
 * Using BSDDirectory.cc as base for development
 * Copyright (c) 2004
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 */
/*
 * Ext2Directory.cc
 */


#ifdef __GNUC__
#pragma implementation
#endif

#include "Assert.h"
#include "Raid.h"
#include "FileSystem.h"
#include "PersistentMemoryObject.h"
#include "Ext2Directory.h"
#include "String.h"
#include "Byte.h"

#ifdef __GNUC__
#pragma implementation
#endif

/*
 * The new version of the directory entry.  Since EXT2 structures are
 * stored in intel byte order, and the name_len field could never be
 * bigger than 255 chars, it's safe to reclaim the extra byte for the
 * file_type field.
 */
struct direct_2 {
    unsigned int    d_ino;                    /* Inode number */
    unsigned short  d_reclen;                 /* Directory entry length */
    char    d_namlen;                         /* Name length */
    char    d_file_type;                      /* File type */
    char    d_name[MaxEntrySize_e2 + 1];      /* File name */

    int size();
};

/*
 * Ext2 directory file types.  Only the low 3 bits are used.  The
 * other bits are reserved for now.
 */
enum {
    EXT2_FT_UNKNOWN,
    EXT2_FT_REG_FILE,
    EXT2_FT_DIR,
    EXT2_FT_CHRDEV,
    EXT2_FT_BLKDEV,
    EXT2_FT_FIFO,
    EXT2_FT_SOCK,
    EXT2_FT_SYMLINK,
    EXT2_FT_MAX
};

/*
 * EXT2_DIR_PAD defines the directory entries boundaries
 *
 * NOTE: It must be a multiple of 4
 */
#define EXT2_DIR_PAD            4
#define EXT2_DIR_ROUND          (EXT2_DIR_PAD - 1)

inline 
int 
direct_2::size()
{
    return (((d_namlen) + 8 + EXT2_DIR_ROUND) & ~EXT2_DIR_ROUND);
}


/*
 * Ext2Directory allows directory entries to be located in,
 * added to, and removed from a Ext2 directory file.
 */
Ext2Directory::Ext2Directory( PersistentMemoryObject * source )
    : MemoryObjectDictionary( source )
{
    initDebug();
    Raid( RaidConstructor )( "%N::Ext2Directory( %N )\n", this, source );
    Raid( RaidDetail )( "%N::Ext2Directory idNumber %d _size %d\n",
                        this, source->idNumber(), _size );

    _maxKeyLength = MaxEntrySize_e2;

    /* make directory file size a multiple of the directory size */
    int roundedSize = ( _size + DirectoryBlockSize_e2 - 1 ) / DirectoryBlockSize_e2
    * DirectoryBlockSize_e2;
    if( _size != roundedSize ) {
        _size = roundedSize;
        source->setNumberOfBytes( _size );
    }
    if( _size == 0 ) {
        add( ".", source );
    } else {
        reMap();
    }
    _hint = source->idNumber() + 1;
}


Ext2Directory::~Ext2Directory()
{
    Raid( RaidConstructor )( "%N::~Ext2Directory\n", this );
}


/*
 * search the entire directory for an occurrence of key
 * if found, return 1; otherwise 0
 * side effects: set _previousOffset & _offset dir entry ofs and
 * findKey will also try to find an _empty space big enough 
 * to hold a new directory entry of _size _size.
 */
int
Ext2Directory::findKey( char * key, int needSize ) 
{
    direct_2 * dp;

    Raid( RaidMethod )( "%N::findKey( key %s, needSize %d )\n",
                        this, key, needSize );

    int StillLooking = ( needSize > 0 );
    _emptyOffset = -1;

    _previousOffset = 0;
    unsigned short tmp_reclen;
    for( _offset = 0; _offset < _size; _offset += tmp_reclen ) {
        dp = ( direct_2 * ) &_buffer[_offset];

        /*  Raid( RaidDetail )( "%N::findKey ofs %d ", this, _offset); 
            Raid( RaidDetail )( "ino %d rl %d nl %d key %s\n", dp->d_ino, dp->d_reclen, dp->d_namlen, dp->d_name );*/

        tmp_reclen = dp->d_reclen;
        swaps((void *) &tmp_reclen);
    
        Debug<<this<<"::findKey("<<key<<") "<<tmp_reclen<<":"<<dp->d_name<<"\n"<<eor;

        Assert( tmp_reclen > 0 );

        /* check to see if hole is big enough */
        if( StillLooking ) {
            int sp = tmp_reclen;
            if( dp->d_ino != 0 ) sp -= entrySize( dp->d_namlen );
            if( sp >= needSize ) {
                _emptyOffset = _offset;
                StillLooking = 0;
                Raid( RaidDetail )( "%N::findKey space @ %d\n", this, _offset );
            }
        }

        /* return if key is found */
        /* This is different from BSD because BSD terminates directory entries with
           '\0' while Ext2 does not.
        */
        size_t key_len = strlen(key);
        if( ( dp->d_ino != 0 ) && 
            ( key_len == dp->d_namlen) &&
            ( strncmp( key, dp->d_name, key_len ) == 0 ) ) {
            _idNumberOfKey = dp->d_ino;
            swapl((void *) & _idNumberOfKey);
            Raid( RaidDetail )( "%N::findKey found @ %d\n", this, _offset );
            return( 1 );
        }
        _previousOffset = _offset;
    }
    return( 0 );
}


int
Ext2Directory::insertEntry( char * key, int keyLength, int idNumber  )
{
    direct_2 * edp;
    direct_2 dir;

    Raid( RaidMethod )( "%N::insertEntry( key %s, keyLength %d, idNumber %d )\n",
                        this, key, keyLength, idNumber );

    /* set up the new entry ( for _size estimation ) */
    dir.d_ino = idNumber;
    swapl((void *) & (dir.d_ino));

    dir.d_namlen = keyLength;

    if( _emptyOffset < 0 ) {
        dir.d_reclen = DirectoryBlockSize_e2;
        swaps((void *) & (dir.d_reclen));
        _offset = _size;
        _size += DirectoryBlockSize_e2;
        _source->setNumberOfBytes( _size );
        int newUnits = _source->numberOfUnits();
        if( _units != newUnits )  {
            reMap();
            _units = newUnits;
        }
    }
    else {
        edp = ( direct_2 * ) &_buffer[_emptyOffset];
        int bytesInUse = 0;
        if( edp->d_ino != 0 ) bytesInUse = entrySize( edp->d_namlen );
        unsigned short tmp_reclen = edp->d_reclen;
        swaps((void *) &tmp_reclen);
        dir.d_reclen = tmp_reclen - bytesInUse;
        swaps((void *) & (dir.d_reclen));
        edp->d_reclen = bytesInUse;
        swaps((void *) &(edp->d_reclen));
        _offset = _emptyOffset + bytesInUse;
    }

    Raid( RaidDetail )( "%N::insertEntry _offset %d _emptyOffset %d\n",
                        this, _offset, _emptyOffset );

    /* move new entry into the blk */
    /* No need to endian-swap here since both edp and dir are little-endian */
    edp = ( direct_2 * ) &_buffer[_offset];
    edp->d_ino    = dir.d_ino;
    edp->d_reclen = dir.d_reclen;
    edp->d_namlen = dir.d_namlen;
    /* This is different from BSD because BSD terminates directory entries with
       '\0' while Ext2 does not.
    */
    strncpy( edp->d_name, key, keyLength );

    /* new hint */
    _hint = idNumber + 1;

    return( writeChanges() );
}


/*
 */
int
Ext2Directory::clearEntry()
{
    Raid( RaidMethod )( "%N::clearEntry()\n", this );

    direct_2 * cdp = ( direct_2 * ) &_buffer[_offset];

    if( ( _offset & ( DirectoryBlockSize_e2 - 1 ) ) == 0 ) cdp->d_ino = 0;
    else {
        direct_2 * pdp = ( direct_2* ) &_buffer[_previousOffset];
        /* No need to byteswap here since copying between little-endian structs */
        pdp->d_reclen += cdp->d_reclen;
    }
    return( writeChanges() );
}


/*
 *  return positive count of keys if the buffer was big enough,
 *  negative count otherwise.
 *  optimize and improve later, perhaps via memory-mapping a precalculated
 *  list-o-keys.
 */
int
Ext2Directory::basicKeys( char * keyBuffer, int bufferSize, int & count ) 
{
    Raid( RaidMethod )( "%N::keys( buffer %x, bufferSize %d ) _size %d\n",
                        this, keyBuffer, bufferSize, _size );

    count = 0;
    int limit = ( bufferSize >= _size ) ? _size : bufferSize;
    char * t = keyBuffer;
    direct_2 * dp = ( direct_2 * ) _buffer;

    unsigned short tmp_reclen;
    for( _offset = 0; _offset < limit; _offset += tmp_reclen ) {

        dp = ( direct_2 * ) &_buffer[_offset];

        tmp_reclen = dp->d_reclen;
        swaps((void *) &tmp_reclen);
        int tmp_ino = dp->d_ino;
        swapl((void *) &tmp_ino);

        Raid( RaidDetail )( "%N::keys ofs %d ", this, _offset); 
        Raid( RaidDetail )( "ino %d rl %d nl %d key %s\n",
                            tmp_ino, tmp_reclen, dp->d_namlen, dp->d_name );

        Assert( tmp_reclen > dp->d_namlen );

        if( dp->d_ino != 0) {
            Assert( dp->d_namlen > 0 );

            if( t != 0 ) {
                for( char * s = dp->d_name; *s; ) *t++ = *s++; 
                *t++ = '\n';
            }
            count++;
        }
    }
    if( t != 0 ) *t = '\0';
    if( limit == _size ) return( 0 );
    else return( TooManyKeys );
}

/*
 *  return positive count of keys if the buffer was big enough,
 *  negative count otherwise.
 *  optimize and improve later, perhaps via memory-mapping a precalculated
 *  list-o-keys.
 */
int
Ext2Directory::basicAssociations( char * keyBuffer, int bufferSize,
                                  int & count )
{
    int offset = 0;
    return basicAssociations( keyBuffer, bufferSize, count, offset );
}

int
Ext2Directory::basicAssociations( char * keyBuffer, int bufferSize,
                                  int & count, int & offset )
{
    Raid( RaidMethod )( "%N::basicAssociations( buf %x, bufSize %d )_size %d\n",
                        this, keyBuffer, bufferSize, _size );

    Assert( keyBuffer != 0 );
    char * t = keyBuffer;
    char * tlimit = keyBuffer + bufferSize;
    direct_2 * dp = ( direct_2 * ) _buffer;

    unsigned short tmp_reclen;
    for( _offset = offset ; _offset < _size; _offset += tmp_reclen ) {

        dp = ( direct_2 * ) &_buffer[_offset];

        tmp_reclen = dp->d_reclen;
        swaps((void *) &tmp_reclen);
        int tmp_ino = dp->d_ino;
        swapl((void *) &tmp_ino);

        Raid( RaidDetail )( "%N::associations ofs %d ", this, _offset); 
        Raid( RaidDetail )( "ino %d rl %d nl %d key %s\n",
                            tmp_ino, tmp_reclen, dp->d_namlen, dp->d_name );

        Assert( tmp_reclen > dp->d_namlen );

        if( dp->d_ino != 0) {
            Assert( dp->d_namlen > 0 );
            if( t + tmp_reclen > tlimit) {
                break;
            }
            int cpcount = dp->size();
            ByteCopy( dp, t, cpcount );
            ((direct_2 *) t)->d_reclen = cpcount;
            swaps((void *) &(((direct_2 *) t)->d_reclen));
            t += cpcount;
        }
    }
    offset = _offset;
    count = t - keyBuffer;
    Assert( count <= bufferSize );
    if( _offset >= _size ) return( 0 );
    else return( TooManyKeys );
}

int
Ext2Directory::entrySize( int keyLength )
{
    return( ( keyLength + 3 + sizeof( direct_2 ) - MaxEntrySize_e2 ) & ~3 );
}

int
Ext2Directory::hint()
{
    return( _hint );
}

ProxiableObject * 
Ext2DirectoryConstructor( ProxiableObject * source )
{
    if( source->isKindOf( PersistentMemoryObjectClass ) ) {
        return( new Ext2Directory( (PersistentMemoryObject *)source ) );
    }
    else return( 0 );
}



#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("Ext2Directory",Ext2Directory,MemoryObjectDictionary,&Ext2DirectoryConstructor,0);
