/*
 * This file is part of the Choices Operating System
 * Author: Yevgeniy Rabinovich <rabinvch@uiuc.edu>,
 *                             <yrabinov+uiuc@cs.uchicago.edu>
 * Using BSDInode.cc as base for development
 * Copyright (c) 2004
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 */
/*
 * Ext2Inode.cc
 */

#ifdef __GNUC__
#pragma implementation
#endif

#ifndef COFFFile
#define COFFFile
#endif

#include "Assert.h"
#include "Byte.h"
#include "Ext2Allocator.h"
#include "Ext2Container.h"
#include "Ext2Inode.h"
#include "Class.h"
#include "FileSystem.h"
#include "FileSystemHints.h"
#include "MemoryObjectAsync.h"
#include "MemoryObjectContainer.h"
#include "MemoryObjectDictionary.h"
#include "Raid.h"
#include "RemoteFileSystem.h"
#include "SoftClock.h"
#include "Time.h"
#include "UNIXFile.h"
#include "UNIXInodeIndirect.h"

#include <string.h>

#ifdef __GNUC__
#pragma implementation
#endif

extern ClassStar PersistentArrayClass;
extern ClassStar SymbolicLinkClass;

int Ext2InodeDefaultCacheHints = (H_LCACHE << 16);

int Ext2debugHints = 0;

inline
void Ext2Inode::initDinode()
{
    /* Need to SWAP here. Do it UNIXInode */
    direct = _diskInode->di_db;
    singleIndex = _diskInode->di_db+EXT2_IND_BLOCK;
    doubleIndex = _diskInode->di_db+EXT2_DIND_BLOCK;
    tripleIndex = _diskInode->di_db+EXT2_TIND_BLOCK;
    _modifiedFlag = (*(_inodeData->_inodeIndirectInfo))->dirtyFlag();
//    Assert( _diskInode->di_nlink == _inodeData->_di_nlink );
}

void Ext2Inode::getDinode()
{
    *(_inodeData->_inodeIndirectInfo) = indirect->getIndirect
    ( _inodeData->_partition, 
      _inodeData->_inodeBlock, 
      _idNumber,
      *(_inodeData->_inodeIndirectInfo) );
    Ext2_dinode * newDiskInode = 
    ( (Ext2_dinode *) ( (*(_inodeData->_inodeIndirectInfo))->buf() ) ) + 
    _inodeData->_inodeOffset;
    if ( _diskInode != newDiskInode ) {
        _diskInode = newDiskInode;
        initDinode();
    }
}


void Ext2Inode::releaseDinode()
{
    indirect->releaseIndirect( *(_inodeData->_inodeIndirectInfo), 0 );
}


/* 
 * Initialize the in-memory inode
 */
Ext2Inode::Ext2Inode( unsigned int numOfUnits,
                      int log2UnitSize,
                      int inumber,
                      MemoryObjectContainer * container,
                      PersistentMemoryObject * source,
                      int log2blocksize,
                      Ext2_dinode * thisDinode, 
                      UNIXInodeIndirectInfo ** inodeIndirectInfo )
    : UNIXInode( numOfUnits, log2UnitSize, inumber, 
                 container, source, log2blocksize )
{
    initDebug();
 
    Raid( RaidConstructor )( "%N::Ext2Inode()\n", this );
    Debug.printf("Ext2Inode::Ext2Inode(numOfUnits = %d, log2UnitSize = %d, inumber = %d, container = 0x%x",numOfUnits, log2UnitSize, inumber, container);
    Debug.printf(", source = 0x%x, log2blocksize = %d)\n", source, log2blocksize);

    /* UNIXInode should know that Ext2 is little-endian */
    little_endian_fs = TRUE;
    /*
     *  Initialize variables from parameters.
     */
    _diskInode = thisDinode;
    _inodeData = new _InodeData;
    _version = _diskInode->di_generation; /*di_version in BSD ? */
    swapl((void *) &_version);
    /*    _hints = _diskInode->di_spare[2]; by Eugene */
    if (Ext2debugHints) {
        Console() 
        << this << "::ctor _hints " << asHex(_hints) << "\n" << eor;
    }
    _inodeData->_di_mode = _diskInode->di_mode;
    swaps((void *) &(_inodeData->_di_mode));
    _inodeData->_di_nlink = _diskInode->di_nlink;
    swaps((void *) &(_inodeData->_di_nlink));
    _inodeData->_di_flags = _diskInode->di_flags;
    swapl((void *) &(_inodeData->_di_flags));
    _inodeData->_di_size = _diskInode->di_size;
    swapl((void *) &(_inodeData->_di_size));
    
    _inodeData->_inodeIndirectInfo = inodeIndirectInfo;
    _inodeData->_partition = (*inodeIndirectInfo)->source();
    _inodeData->_inodeBlock = (*inodeIndirectInfo)->unit();
    /* Do I need to do any swapping here? Probably not */
    _inodeData->_inodeOffset = _diskInode - ( (Ext2_dinode *) ( (*inodeIndirectInfo)->buf() ) );

    initDinode();

    /*
     *  Set up indirect variables for mapUnit function.
     */
    /* Eugene: not really sure about these */
    singleIndirect = EXT2_NDIR_BLOCKS;
    doubleIndirect = singleIndirect + ( 1 << indexShift );
    tripleIndirect = doubleIndirect + ( 1 << ( indexShift + indexShift ) );

    /*
     * Used by allocBlock to allocate blocks near each other.
     */
    nearBlock = 0;

    /*
     * Calculate the number of fragments in the last logical block of the file.
     */
    /* Eugene: not really sure about this */
    unsigned int tmp_di_flags =  _diskInode->di_flags;
    swapl((void *) &tmp_di_flags);
    if( !( tmp_di_flags & IC_FASTLINK ) ) {
        fragmentsInLastBlock = numFragments();
        if( fragmentsInLastBlock > 0 ) {
            nearBlock = mapUnit( lastBlockNo(), 0 );
            if( nearBlock == 0 ) fragmentsInLastBlock = 0;
        }
    }

    /*
     * Set access time
     */
    _diskInode->di_atime = _theTime.sec();
    swapl((void *) &(_diskInode->di_atime));
    releaseDinode();

    MemoryObject::basicSetNumberOfUnits( _numberOfUnits );
}


Ext2Inode::~Ext2Inode()
{
    Raid( RaidConstructor )( "%N::~Ext2Inode() fragsInLastBlk %d\n",
                             this, fragmentsInLastBlock );


    if (Ext2debugHints) {
        Console() 
        << this << "::dtor _hints " << asHex(_hints) 
        << " zombieDelay " << zombieDelay() << "\n" << eor;
    }
    if( _inodeData->_di_nlink == 0 ) {
        getDinode();
        Assert( _inodeData->_di_nlink == _diskInode->di_nlink );
        Assert( zombieInfo()->type() == ZOMBIE_NOT );
        updateInode( 1 );
        _diskInode->di_mode = 0;
        _diskInode->di_atime = 0;
        _diskInode->di_ctime = 0;
        _diskInode->di_mtime = 0;

        /* Eugene: Ext2 also has dtime */
        _diskInode->di_dtime = _theTime.sec();
        swapl((void *) &(_diskInode->di_dtime));
        /* Eugene: size is a single integer in Ext2; version is generation */
        /*
          _diskInode->di_size[0] = 0;
          _diskInode->di_size[1] = 0;
          _diskInode->di_version++;
        */

        _diskInode->di_size = 0;
        swapl((void *) &(_diskInode->di_generation));
        _diskInode->di_generation++;
        _version = _diskInode->di_generation;
        swapl((void *) &(_diskInode->di_generation));

        /* Update the optional vars (OSD's) instead of spare */
        /*
          _diskInode->di_spare[0] = 0;
          _diskInode->di_spare[1] = 0;
          _diskInode->di_spare[2] = 0;
          _diskInode->di_spare[3] = 0;
        */
        memset((void *) &(_diskInode->osd1), 0, sizeof(_diskInode->osd1));
        memset((void *) &(_diskInode->osd2), 0, sizeof(_diskInode->osd2));
        setModified();
        releaseDinode(); 
    } else {
        if ( zombieInfo()->type() == ZOMBIE_NOT ||
             zombieInfo()->type() == ZOMBIE_NOTSYNC )
            synchronizeInode();
    }

    _diskInode = 0;
    delete _inodeData;
    _inodeData = 0;
}


Class *
Ext2Inode::supports( Class * aClass )
{
    Raid( RaidProtected )( "%N::supports( %N )\n", this, aClass );

    Class * supportedClass = 0;

    if( ( ( _inodeData->_di_mode & UNIXTypeMask ) == UNIXTypeRegular ) ) {
        if( PersistentArrayClass->isASubclassOf( aClass ) ) {
            supportedClass = PersistentArrayClass;
        }
    }
    else if( ( _inodeData->_di_mode & UNIXTypeMask ) == UNIXTypeDirectory ) {
        extern ClassStar Ext2DirectoryClass;
        /*  extern ClassStar HashedExt2DirectoryClass; */
        if( _inodeData->_di_flags & IC_HASHED_DIR ) {
            Console()<<"Ext2Inode::supports() - HashedExt2DirectoryClass is not implemented!!\n";
            /*
              if( HashedExt2DirectoryClass->isASubclassOf( aClass ) ) {
              supportedClass = HashedExt2DirectoryClass;
              }
            */
        }
        else {
            if( Ext2DirectoryClass->isASubclassOf( aClass ) ) {
                supportedClass = Ext2DirectoryClass;
            }
        }
    }
    else if( ( _inodeData->_di_mode & UNIXTypeMask ) == UNIXTypeSymbolicLink ) {
        if( SymbolicLinkClass->isASubclassOf( aClass ) ) {
            supportedClass = SymbolicLinkClass;
        }
    }

// TODO Everything below relies on some hint stored in the Inode. This needs to be reworked for Ext2
#ifdef FRANCIS
    else if( _inodeData->_di_spare3 == 2 ) {
        if( PersistentArrayClass->isASubclassOf( aClass ) ) {
            supportedClass = PersistentArrayClass;
        }
#ifdef ArFile
        else {
            extern ClassStar ArContainerClass;
            if( ArContainerClass->isASubclassOf( aClass ) ) {
                supportedClass = ArContainerClass;
            }
        }
#endif
    }
    else if( _inodeData->_di_spare3 == 3 ) {
        if( PersistentArrayClass->isASubclassOf( aClass ) ) {
            supportedClass = PersistentArrayClass;
        }
#ifdef MailFile
        else {
            extern ClassStar MailContainerClass;
            if( MailContainerClass->isASubclassOf( aClass ) ) {
                supportedClass = MailContainerClass;
            }
        }
#endif
    }
    else if( _inodeData->_di_spare3 == 4 ) {
        if( PersistentArrayClass->isASubclassOf( aClass ) ) {
            supportedClass = PersistentArrayClass;
        }
#ifdef TarFile
        else {
            extern ClassStar TarContainerClass;
            if( TarContainerClass->isASubclassOf( aClass ) ) {
                supportedClass = TarContainerClass;
            }
        }
#endif
    }
    else if( _inodeData->_di_spare3 == 5 ) {
        if( PersistentArrayClass->isASubclassOf( aClass ) ) {
            supportedClass = PersistentArrayClass;
        }
#ifdef COFFFile
        else {
            extern ClassStar COFFContainerClass;
            if( COFFContainerClass->isASubclassOf( aClass ) ) {
                supportedClass = COFFContainerClass;
            }
        }
#endif
    }    
    else if( _inodeData->_di_spare3 == 6 ) {
        if( PersistentArrayClass->isASubclassOf( aClass ) ) {
            supportedClass = PersistentArrayClass;
        }
        else {
            extern ClassStar ELFContainerClass;
            if( ELFContainerClass->isASubclassOf( aClass ) ) {
                supportedClass = ELFContainerClass;
            }
        }

    }
#endif // FRANCIS

    Raid( RaidProtected )( "%N::supports() returns %N %d\n", this, supportedClass );

    return( supportedClass );
}


void
Ext2Inode::hintsChanged( unsigned oldHints )
{
    /* Eugene: Ext2 does not have di_spare */
    /*
      getDinode();

      if ( _hints != (unsigned)_diskInode->di_spare[2] ) {
      _diskInode->di_spare[2] = _hints;
      setModified();
      }
      releaseDinode();
    */
}


int
Ext2Inode::basicRead( unsigned startUnit, int count, char * addr,
                      unsigned nbytes, IOAsyncClosure * c )
{
    getDinode();
    _diskInode->di_atime = _theTime.sec();
    swapl((void *) &(_diskInode->di_atime));
    releaseDinode();
    return UNIXInode::basicRead( startUnit, count, addr, nbytes, c );
}


int 
Ext2Inode::rawRead( unsigned startUnit, unsigned count, char * addr, unsigned nbytes,
                    IOAsyncClosure * c )
{
    Raids( RaidProtected ) 
    << this << "::rawRead( " << startUnit << ", " << count 
    << ", " << asAddress( addr ) << ", " << nbytes 
    << ", " << asAddress( c ) << " )\n" << eor;

    Debug<< this << "::rawRead( " << startUnit << ", " << count 
         << ", " << asAddress( addr ) << ", " << nbytes 
         << ", " << asAddress( c ) << " )\n" << eor;
   
    _mutex.P();

    if( ( ( _inodeData->_di_mode & UNIXTypeMask ) == UNIXTypeSymbolicLink ) 
        &&( _inodeData->_di_flags & IC_FASTLINK ) ) {

        getDinode();

        Assert( numberOfUnits() <= 1 );
        if( startUnit >= 1 ) return( 0 );

        unsigned bytes = numberOfBytes();
        Assert( bytes < unitSize() );
        Assert( bytes <= nbytes );

        unsigned int * tmp_di_db = _diskInode->di_db;
        swapl((void *) &tmp_di_db);
        ByteCopy( (char *) tmp_di_db, addr, bytes );
    
        releaseDinode();

        _mutex.V();

        if ( c ) {
            c->activate( 1 );
        }       
        return 1;
    }
    else {
        _mutex.V();
        return UNIXInode::rawRead( startUnit, count, addr, nbytes, c );
    }
}


int
Ext2Inode::basicWrite( unsigned startUnit, unsigned count, char * addr,
                       unsigned nbytes, IOAsyncClosure * c )
{
    getDinode();
    _diskInode->di_mtime = _theTime.sec();
    swapl((void *) &(_diskInode->di_mtime));
    releaseDinode();
    return UNIXInode::basicWrite( startUnit, count, addr, nbytes, c );
}


int
Ext2Inode::basicWriteThrough( unsigned startUnit, unsigned count, char * addr,
                              unsigned nbytes, IOAsyncClosure * c )
{
    getDinode();
    _diskInode->di_mtime = _theTime.sec();
    swapl((void *) &(_diskInode->di_mtime));
    releaseDinode();
    return UNIXInode::basicWriteThrough( startUnit, count, addr, nbytes, c );
}



int 
Ext2Inode::rawWrite( unsigned startUnit, unsigned count, char * src, unsigned nbytes,
                     int writeThrough, IOAsyncClosure * c )
{
    Raids( RaidProtected ) 
    << this << "::rawWrite( " << startUnit << ", " << count 
    << ", " << asAddress( src ) << ", " << nbytes 
    << ", writeThrough " << writeThrough 
    << ", " << asAddress( c ) << " )\n" << eor;

    /*
     * Dont perpetuate fast links.
     * I.e. writes to fast links should reset the fast link bit and
     * clear the direct and indirect block pointers,
     * since the fast link data were stored there. 
     * FIX: perpetuate fast links
     */
    _mutex.P();
    if( ( ( _inodeData->_di_mode & UNIXTypeMask ) == UNIXTypeSymbolicLink ) 
        && ( _inodeData->_di_flags & IC_FASTLINK ) ) {
    
        getDinode();

        unsigned int tmp_di_flags = _diskInode->di_flags;
        swapl((void *) &(tmp_di_flags));
        _inodeData->_di_flags = (tmp_di_flags &= ~IC_FASTLINK );
        swapl((void *) &(tmp_di_flags));
        _diskInode->di_flags = tmp_di_flags;

        // FIX: these assertions are restrictive
        Assert( startUnit == 0 );
        Assert( count == 1 );

        int i;
        /* Eugene */
        /*
          for( i = 0; i < NDADDR; i++ ) _diskInode->di_db[i] = 0;
          for(     i = 0; i < NIADDR; i++ ) _diskInode->di_ib[i] = 0;
        */
        for( i = 0; i < EXT2_N_BLOCKS; i++ ) _diskInode->di_db[i] = 0;

        setModified();

        releaseDinode();
    }
    _mutex.V();

    return( UNIXInode::rawWrite( startUnit, count, src, nbytes, writeThrough, c ) );
}


unsigned
Ext2Inode::numberOfBytes()
{
    return _inodeData->_di_size;
}


unsigned
Ext2Inode::basicSetNumberOfBytes(unsigned size )
{
    Raid( RaidProtected )( "%N::basicSetNumberOfBytes( %d ) %d\n",
                           this, size, numberOfBytes() );

    if( size < 0 ) size = 0;
    if( size == _inodeData->_di_size ) return size;
    getDinode();
    _diskInode->di_mtime = _theTime.sec();
    swapl((void *) &(_diskInode->di_mtime));

    /* Eugene: I don't think this is needed since in Ext2 di_size is an int
       instead of array of int's, as in BSDFS */
    /*
      union {
      int i;
      char    c[4];
      } littleEndian;
      littleEndian.i = 1;
      _diskInode->di_size[!littleEndian.c[0]] = size;
    */

    /* This is instead of the above */
    _diskInode->di_size = size;
    swapl((void *) &(_diskInode->di_size));

    _inodeData->_di_size = size;
    unsigned units;

    Debug<<"size = "<<size<<"\n"<<eor;

    if( size == 0 ) 
        units = 0;
    else
        units = ( size + unitSize() - 1 ) >> log2UnitSize();
    if( ( units >> log2BlockSize ) >= singleIndirect ) {
        int blocks = ( units + blockSize() - 1 ) >> log2BlockSize;
        units = blocks << log2BlockSize;
    }
    Debug<<"units = "<<units<<"\n"<<eor;
    if ( numberOfUnits() != units ) {
        unsigned  result = basicSetNumberOfUnits( units );
        Debug<<"result = "<<result<<"\n"<<eor;
        Assert( result == units );
    }
    setModified();
    releaseDinode();
    return( size );
}

unsigned
Ext2Inode::basicSetNumberOfUnits( unsigned units )
{
    Raid( RaidProtected )
    ( "%N::basicSetNumberOfUnits(Ext2)( units %d ) was %d\n",
      this, units, numberOfUnits() );
    unsigned numUnits = numberOfUnits();
    
    if( units == numUnits ) return( numUnits );

    getDinode();

#ifdef NEVER
    if( units == 0 ) {
        swapl((void *) &(_diskInode->di_generation));
        _diskInode->di_generation++;          /* Eugene: di_version in BSD ? */
        _version = _diskInode->di_generation;
        swapl((void *) &(_diskInode->di_generation));
        setModified();
    }
#endif

    unsigned int tmp_di_flags = _diskInode->di_flags;
    swapl((void *) &tmp_di_flags);
    if( tmp_di_flags & IC_FASTLINK ) {
        Assert( units == 0 );
        Assert( numUnits == 1 );

        int i;
        /* Eugene */
        /*
          for( i = 0; i < NDADDR; i++ ) _diskInode->di_db[i] = 0;
          for(     i = 0; i < NIADDR; i++ ) _diskInode->di_ib[i] = 0;
        */
        for( i = 0; i < EXT2_N_BLOCKS; i++ ) _diskInode->di_db[i] = 0;
        setModified();
    }

    // If the last block is fragmented, extend it out to a full block.
    // Unless it will soon be freed; if so, just free it now.
    // Eugene: We probably don't need this since Ext2 doesn't care
    // about cylinders
    /*
      if( fragmentsInLastBlock > 0 ) {

      int logicalBlock = lastBlockNo();
      int endBlock = ( units + blockSize() - 1 ) >> log2BlockSize;
      int physicalBlock = direct[ logicalBlock ];

      Assert( physicalBlock != 0 );

      if( endBlock <= logicalBlock ) {
      _container->free( _diskInode->di_blocks, physicalBlock, 
      fragmentsInLastBlock );
      direct[ logicalBlock ] = 0;
      numUnits -= fragmentsInLastBlock;
      MemoryObject::basicSetNumberOfUnits( numUnits );
      } else if ( ! ((Ext2Container *) _container)->allocator()->
      extendFragmentsToBlock( _diskInode->di_blocks,
      physicalBlock, 
      fragmentsInLastBlock ) ) {
      Raids( RaidProtected ) 
      << this << "::basicSetNumberOfUnits() extend failed\n"
      << eor;
      char * buffer = new char [unitSize() << log2BlockSize];

      ByteZero( buffer, unitSize() << log2BlockSize );
      fileObjectRead( physicalBlock, fragmentsInLastBlock,
      buffer, fragmentsInLastBlock << log2UnitSize(), 0 );
      _container->free( _diskInode->di_blocks, physicalBlock, 
      fragmentsInLastBlock );
      // FIX: check physicalBlock
      int newPhysicalBlock = 0;
      allocBlocks( &newPhysicalBlock, 0, 1, 
      physicalBlock & ~(blockMask()) );
      Assert( newPhysicalBlock != 0 );
      direct[ logicalBlock ] = newPhysicalBlock;
      fileObjectWrite( newPhysicalBlock, blockSize(),
      buffer, unitSize() << log2BlockSize, 0, 0 );
      delete buffer;
      }
      setModified();
      fragmentsInLastBlock = 0;
      }
    */
    releaseDinode();

    int result = UNIXInode::basicSetNumberOfUnits( units );

    return result;
}



/*
 *  this function allows other objects to access and modify the
 *  link count ( or reference count ) of this inode.
 *  links returns the new number of links for this inode.
 */
int
Ext2Inode::links( int delta )
{
    Raid( RaidMethod )( "%N::links( delta %d )\n", this, delta );
    int result;
    if ( delta == 0 ) {
        result = _inodeData->_di_nlink;
    } else {
        _mutex.P();
        getDinode();

        swaps((void *) &(_diskInode->di_nlink));
        Assert( _diskInode->di_nlink == _inodeData->_di_nlink );
        _inodeData->_di_nlink += delta;
        result = _diskInode->di_nlink = _inodeData->_di_nlink;
        swaps((void *) &(_diskInode->di_nlink));
    
        _diskInode->di_ctime = _theTime.sec();
        swapl((void *) &(_diskInode->di_ctime));
        setModified();
        releaseDinode();
        _mutex.V();
        if ( delta > 0 && _creatingDictionary ) {
            synchronizeCreatingDictionary();
        }
    }
    Raid( RaidMethod )( "%N::links = %d\n", this, result );
    return( result );
}


int
Ext2Inode::setMode( int mode )
{
    Raid( RaidMethod )( "%N::setMode( mode %o ) %o\n",
                        this, mode, _inodeData->_di_mode );

    short new_mode = ( _inodeData->_di_mode & ~07777 ) | ( mode & 07777 );
    if ( new_mode == _inodeData->_di_mode ) 
        return 0;
    _mutex.P();
    getDinode();
    _inodeData->_di_mode = new_mode;
    _diskInode->di_mode = new_mode;
    swaps((void *) &(_diskInode->di_mode));

    _diskInode->di_ctime = _theTime.sec();
    swapl((void *) &(_diskInode->di_ctime));

    setModified();
    releaseDinode();
    _mutex.V();
    return( 0 );
}


int
Ext2Inode::setUserId( int uid )
{
    getDinode();
    unsigned short tmp_di_uid = _diskInode->di_uid;
    swaps((void *) &tmp_di_uid);
    Raid( RaidMethod )( "%N::setUserId( uid %d ) %d\n",
                        this, uid, tmp_di_uid );
    _diskInode->di_uid = uid;
    swaps((void *) &(_diskInode->di_uid));
    _diskInode->di_ctime = _theTime.sec();
    swapl((void *) &(_diskInode->di_ctime));
    setModified();
    releaseDinode();
    return( 0 );
}


int
Ext2Inode::setGroupId( int gid )
{
    getDinode();
    unsigned short tmp_di_gid = _diskInode->di_gid;
    swaps((void *) &tmp_di_gid);
    Raid( RaidMethod )( "%N::setGroupId( gid %d ) %d\n",
                        this, gid, tmp_di_gid );
    _diskInode->di_gid = gid;
    swaps((void *) &(_diskInode->di_gid));
    _diskInode->di_ctime = _theTime.sec();
    swapl((void *) &(_diskInode->di_ctime));
    setModified();
    releaseDinode();
    return( 0 );
}



void
Ext2Inode::info( UNIXFileInfo * infoBuffer )
{
    Raid( RaidMethod )( "%N::info( %x ) %d\n", this, infoBuffer, _idNumber );
    getDinode();
    infoBuffer->device  = 0;
    infoBuffer->inumber = _idNumber;
    infoBuffer->mode    = _diskInode->di_mode;
    swaps((void *) &(infoBuffer->mode));
    infoBuffer->links   = _inodeData->_di_nlink;
    swaps((void *) &(infoBuffer->links));
    infoBuffer->uid = _diskInode->di_uid;
    swaps((void *) &(infoBuffer->uid));
    infoBuffer->gid = _diskInode->di_gid;
    swaps((void *) &(infoBuffer->gid));
    infoBuffer->size    = _inodeData->_di_size;
    swapl((void *) &(infoBuffer->size));
    infoBuffer->atime   = _diskInode->di_atime;
    swapl((void *) &(infoBuffer->atime));
    infoBuffer->mtime   = _diskInode->di_mtime;
    swapl((void *) &(infoBuffer->mtime));
    infoBuffer->ctime   = _diskInode->di_ctime;
    swapl((void *) &(infoBuffer->ctime));
    infoBuffer->blocks  = _diskInode->di_blocks;
    swapl((void *) &(infoBuffer->blocks));
    infoBuffer->blockSize = blockSize() << _log2UnitSize;
    infoBuffer->flags   = _inodeData->_di_flags;
    swapl((void *) &(infoBuffer->flags));
    releaseDinode();
}


void
Ext2Inode::getAttributes( RFSAttributes * attr )
{
    getDinode();
    attr->mode = _diskInode->di_mode;
    swaps((void *) &(attr->mode));
    attr->links = _inodeData->_di_nlink;
    swaps((void *) &(attr->links));
    attr->uid = _diskInode->di_uid;
    swaps((void *) &(attr->uid));
    attr->gid = _diskInode->di_gid;
    swaps((void *) &(attr->gid));
    attr->size = _inodeData->_di_size;
    swapl((void *) &(attr->size));
    attr->atime = _diskInode->di_atime;
    swapl((void *) &(attr->atime));
    attr->mtime = _diskInode->di_mtime;
    swapl((void *) &(attr->mtime));
    attr->ctime = _diskInode->di_ctime;
    swapl((void *) &(attr->ctime));
    attr->type.set( this );
    attr->log2BlockSize = log2BlockSize;
    attr->log2UnitSize = _log2UnitSize;
    attr->flags = RFSAttributes::AttrDirty;
    attr->cacheHints = _hints;
    releaseDinode();
}


void
Ext2Inode::setAttributes( RFSAttributes * attr )
{
    if ( !(attr->flags & RFSAttributes::AttrDirty) )
        return;
    getDinode();
    if ( !isDictionary() && _inodeData->_di_size != (unsigned)attr->size )
        setNumberOfBytes( attr->size );

    unsigned short tmp_short_id = _diskInode->di_uid;
    swaps((void *) &tmp_short_id);
    if ( tmp_short_id != attr->uid ) {
        _diskInode->di_uid = attr->uid;
        swaps((void *) &(_diskInode->di_uid));
        _diskInode->di_ctime = _theTime.sec();
        swapl((void *) &(_diskInode->di_ctime));
        setModified();
    }

    tmp_short_id = _diskInode->di_gid;
    swaps((void *) &tmp_short_id);
    if ( tmp_short_id != attr->gid ) {
        _diskInode->di_gid = attr->gid;
        swaps((void *) &(_diskInode->di_gid));
        _diskInode->di_ctime = _theTime.sec();
        swapl((void *) &(_diskInode->di_ctime));
        setModified();
    }

    tmp_short_id = _diskInode->di_mode;
    swaps((void *) &tmp_short_id);
    if ( (tmp_short_id & UNIXAccessMask) !=
         (attr->mode & UNIXAccessMask) ) {
        short new_mode = ( _inodeData->_di_mode & ~07777 ) | 
        ( attr->mode & 07777 );
        _inodeData->_di_mode = new_mode;
        swaps((void *) &(_diskInode->di_mode));
        _diskInode->di_mode = new_mode;
        _diskInode->di_ctime = _theTime.sec();
        swapl((void *) &(_diskInode->di_ctime));
        setModified();
    }
    releaseDinode();
    unsigned cacheHints = attr->cacheHints;
    if ( (_hints & 0xffff0000) != (cacheHints & 0xffff0000) ) {
        _hintLock.acquire();
        unsigned oldHints = _hints;
        _hints = (_hints & 0x0000ffff) | (cacheHints & 0xffff0000);
        _hintLock.release();
        hintsChanged( oldHints );
    }
}


int
Ext2Inode::allocBlock()
{
    Assert( NOTREACHED );
    return 0;
}


void
Ext2Inode::freeBlock( int physicalBlock )
{
    Raid( RaidDetail )( "%N::freeBlock( physicalBlock %d ) id %d\n",
                        this, physicalBlock, idNumber() );

    swapl((void *) & (_diskInode->di_blocks));
    _container->free((int &)_diskInode->di_blocks, physicalBlock, blockSize() );
    swapl((void *) & (_diskInode->di_blocks));

    setModified();
}

void
Ext2Inode::freeFragments( int physicalBlock, int count )
{
    Raid( RaidDetail )
    ( "%N::freeFragments( physicalBlock %d, count %d ) id %d\n",
      this, physicalBlock, count, idNumber() );

    swapl((void *) & (_diskInode->di_blocks));
    _container->free((int &) _diskInode->di_blocks, physicalBlock, count );
    swapl((void *) & (_diskInode->di_blocks));

    setModified();
}

int
Ext2Inode::allocBlocks( int * buf, int start, int count, int pref )
{
    // Eugene: replacing allocateBlocks() with allocate()
    // Calculate last logical and physical block in file.
    //int logicalBlock = lastBlockNo();
    //int physicalBlock = direct[ logicalBlock ];
    swapl((void *) &(_diskInode->di_blocks));
    // Eugene: replacing allocateBlocks() with allocate()
    // Eugene: I am not really sure about this part
    int result = _container->allocate((int &)_diskInode->di_blocks,
                                      buf[start], count);
    swapl((void *) &(_diskInode->di_blocks));

#ifdef FANCY
    /*    return ((Ext2Container *) _container)->allocator()->allocateBlocks
          ( _diskInode->di_blocks, _idNumber, buf, start, count, pref );
    */
#else
    /*
      int result = ((Ext2Container *) _container)->allocator()->allocateBlocks
      ( _diskInode->di_blocks, _idNumber, buf, start, count, pref );
    */
    if ( result > 0 )
        nearBlock = buf[ result - 1 ];
#endif
    for (int i=0; i < count; ++i) {
        swapl((void *) &(buf[start+i]));
    }
    return result;
}


int
Ext2Inode::blockPref( int * buf, int start, int startBlock )
{
#ifdef FANCY
    return ((Ext2Container *) _container)->blockPref
    ( _idNumber, buf, start, startBlock );
#else
    if( nearBlock == 0 ) {
        nearBlock = ((Ext2Container *) _container)->blockPref
        ( _idNumber, buf, start, startBlock );
        return nearBlock;
    }
    return nearBlock + blockSize();
#endif
}


void 
Ext2Inode::updateInode( int finalSync )
{
    Raids( RaidProtected ) 
    << this << "::updateInode( " << finalSync << " )\n" << eor;
    getDinode();
    unsigned units;
    unsigned short tmp_di_nlink = _diskInode->di_nlink;
    swaps((void *) &tmp_di_nlink);

    Assert( _inodeData->_di_nlink == tmp_di_nlink );
    if ( finalSync && _inodeData->_di_nlink == 0 ) {
        units = 0;
    } else {
        int bytes = numberOfBytes();
        if ( bytes == 0 ) {
            units = 0;
        } else {
            units = ( bytes + unitSize() - 1 ) >> log2UnitSize();
            if( ( units >> log2BlockSize ) >= singleIndirect ) {
                int blocks = ( units + blockSize() - 1 ) >> log2BlockSize;
                units = blocks << log2BlockSize;
            }
        }
    }
  
    if ( units != numberOfUnits() ) 
        basicSetNumberOfUnits( units );

    /*  If the file satisfies the following three conditions:
     *      1 ) Its last block is not yet fragmented
     *      2 ) It is not an abominable fast link
     *      3 ) It is not an empty file
     *  then the last block must be checked to see if it needs to be fragmented
     */
    unsigned int tmp_di_flags = _diskInode->di_flags;
    swapl((void *) &tmp_di_flags);

    if( ( fragmentsInLastBlock == 0 ) &&
        ( !( tmp_di_flags & IC_FASTLINK ) ) &&
        ( ( fragmentsInLastBlock = numFragments() ) > 0 ) ) {

        // Calculate last logical and physical block in file.
        int logicalBlock = lastBlockNo();
        int physicalBlock = direct[ logicalBlock ];

        swapl((void *) &physicalBlock);
        // Only existing physical blocks can ever need fragmentation
        if( physicalBlock != 0 ) {

            Raids( RaidDetail ) 
            << this << "::updateInode() fragmentsInLastBlock " 
            << fragmentsInLastBlock << " )\n" << eor;

#ifdef SLOW
            char * buffer = new char [unitSize() << log2BlockSize];

            // NOTE: these operations vvvv are assumed to be atomic 
            int status = fileObjectRead( physicalBlock, blockSize(), 
                                         buffer, unitSize() << log2BlockSize,
                                         0 );
            Assert( status == blockSize() );
            freeBlock( physicalBlock );

            swapl((void *) &(_diskInode->di_blocks));
            physicalBlock = _container->allocate( _diskInode->di_blocks,
                                                  physicalBlock, fragmentsInLastBlock );
            swapl((void *) &(_diskInode->di_blocks));
            Assert( physicalBlock != 0 );

            direct[ logicalBlock ] = physicalBlock;
            status = fileObjectWrite( physicalBlock, fragmentsInLastBlock, 
                                      buffer, 
                                      fragmentsInLastBlock << log2UnitSize(), 
                                      1, 0 );

            swapl((void *) &(direct[ logicalBlock ]));

            Assert( status == fragmentsInLastBlock );
            // NOTE: these operations ^^^^ are assumed to be atomic 

            delete buffer;
            setModified();
#else /* not SLOW */
            // this is a lot faster but may excessively fragment disk.
            swapl((void *) &(_diskInode->di_blocks));
            _container->free((int &) _diskInode->di_blocks,
                             physicalBlock + fragmentsInLastBlock,
                             blockSize() - fragmentsInLastBlock );
            swapl((void *) &(_diskInode->di_blocks));
            setModified();
#endif /* SLOW */
        }
    }
    releaseDinode();
}


void
Ext2Inode::synchronizeInode()
{
    if ( _diskInode != 0 ) {
        UNIXInode::synchronizeInode();
        updateInode( 0 );
    }
}


int
Ext2Inode::accessMode()
{
    return( _inodeData->_di_mode & UNIXAccessMask );
}

int
Ext2Inode::userId()
{
    getDinode();
    int result = _diskInode->di_uid;
    swapl((void *) &result);
    releaseDinode();
    return result;
}


int
Ext2Inode::groupId()
{
    getDinode();
    int result = _diskInode->di_gid;
    swapl((void *) &result);
    releaseDinode();
    return result;
}


int
Ext2Inode::isSymbolicLink()
{
    return( (_inodeData->_di_mode & UNIXTypeMask) == UNIXTypeSymbolicLink );
}

int
Ext2Inode::isDictionary()
{
    return( (_inodeData->_di_mode & UNIXTypeMask) == UNIXTypeDirectory );
}


int
Ext2Inode::isContainer()
{
// FIX: for now ok.
//IMPORTANT CODE CHANGE BY SCOTT KIRCHER
//Need to be able to identify files that are recursive filesystems
//This should be more selective somehow. It used to always return 0
//Now I made it always return 1, but neither is really correct
    return( 1 ); 
}

#include "ClassInstaller.h"
DEFINECLASS("Ext2Inode",Ext2Inode,UNIXInode,0,0);
