#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * FileIndex.c - 
 */
#include "Assert.h"
#include "Raid.h"
#include "FileIndex.h"
#include "PersistentMemoryObject.h"
#include "Convert.h"
#include "String.h"
#include "Byte.h"
#include "FileSystem.h"

#ifdef __GNUC__
#pragma implementation
#endif

struct KeyHeader {
	int	_number;
};

struct KeyStruct {
	KeyHeader _keyHeader;
	char	_key[1];
};

FileIndex::FileIndex( PersistentMemoryObject * source )
: PersistentArray( source )
{
	initDebug();

	Raid( RaidConstructor )( "%N::FileIndex( %N )\n", this, source );
}

FileIndex::~FileIndex()
{
	Raid( RaidConstructor )( "%N::~FileIndex\n", this );
}

int
FileIndex::init( int argc, char *argv[], FileSystemInterface * )
{
	if( argc > 0 ) {
		Raids( RaidMethod ) << this << "::init( " << argc << ", "
				    << argv[0] << " )\n" << eor;
		int keySize = atoi( argv[0] );
		Assert( keySize > 0 );
		_source->setRecordSize( keySize + sizeof( KeyHeader ) );
	}
	else {
		Assert( NOTREACHED );
	}
	return( 0 );
}

int
FileIndex::findKey( char * index, int & status )
{
	status = KeyNotFound;

	int records = _source->numberOfRecords();
	int recSize = _source->recordSize();
	int keySize = recSize - sizeof( KeyHeader );

	int key = 0;

	if( records < 1 ) {
		char * p = _arrayBuffer;
		for( key = 0; key < records; key++  ) {
			KeyStruct * k = (KeyStruct *)p;
			int cmp = strncmp( k->_key, index, keySize );
			p += recSize;
			if( cmp >= 0 ) {
				if( cmp == 0 ) status = KeyExists;
				break;
			}
		}
	}
	else {
		int lo = 0;
		int hi = records - 1;
		while( (lo <= hi ) && ( status == KeyNotFound ) ) {
			key = ( hi + lo ) >> 1;
			KeyStruct * k = (KeyStruct *)&_arrayBuffer[key*recSize];
			int cmp = strncmp( k->_key, index, keySize );
			if( cmp < 0 ) lo = key + 1;
			else if( cmp > 0 ) hi = key - 1;
			else status = KeyExists;
		}
		if( status == KeyNotFound ) key = lo;
	}

	Raids( RaidProtected ) << this << "::findKey( " << index
			       << ", ) return( " << key << " )\n" << eor;

	return( key );
}

int
FileIndex::at( char * index, int & status )
{
	int value = 0;
	int keyNumber = findKey( index, status );

	if( status == KeyExists ) {
		status = 0;
		int recSize = _source->recordSize();
		// Francis - CHECK - Does this really need a reinterpret cast?
		KeyStruct * k = reinterpret_cast<KeyStruct *>(_arrayBuffer[keyNumber*recSize]);
		value = k->_keyHeader._number;
	}

	Raids( RaidMethod ) << this << "::at( " << index << ", ) return( "
			    << value << " )\n" << eor;

	return( value );
}

int
FileIndex::atPut( char * index, int value, int & status )
{
	Raids( RaidMethod ) << this << "::atPut( " << index << ", "
			    << value << ", )\n" << eor;

	int keyNumber = findKey( index, status );

	if( status == KeyNotFound ) {
		status = 0;

		int records = _source->numberOfRecords();
		int recSize = _source->recordSize();
		int keySize = recSize - sizeof( KeyHeader );

		KeyStruct * newKey = (KeyStruct *)new char[recSize];
		newKey->_keyHeader._number = value;
		strncpy( newKey->_key, index, keySize );

		PersistentArray::atPut( records, (char *)newKey, 1 );

		if( keyNumber < records ) {
			char * dest = &_arrayBuffer[records*recSize];

			for( int j = records; j > keyNumber; j-- ) {
				char * source = dest - recSize;
				ByteCopy( source, dest, recSize );
				dest = source;
			}

			ByteCopy( newKey, dest, recSize );
		}

		delete newKey;
	}

	return( value );
}

int
FileIndex::atNext( char * index, int & status )
{
	int value = 0;
	int keyNumber = findKey( index, status );

	if( status == KeyExists ) {
		status = 0;
	
		int records = _source->numberOfRecords();
		int recSize = _source->recordSize();
		int keySize = recSize - sizeof( KeyHeader );

		// Francis - CHECK - Does this really need a reinterpret cast?
		KeyStruct * k = reinterpret_cast<KeyStruct *>(_arrayBuffer[keyNumber*recSize]);
		value = k->_keyHeader._number;

		if( ++keyNumber < records ) {
			// Francis - CHECK - Does this really need a reinterpret cast?
			k = reinterpret_cast<KeyStruct *>(_arrayBuffer[keyNumber*recSize]);
			strncpy( k->_key, index, keySize );
		}
		else index[0] = '\0';
	}

	Raids( RaidMethod ) << this << "::atNext( " << index << ", ) return( "
			    << value << " )\n" << eor;

	return( value );
}

ProxiableObject *
FileIndexConstructor( ProxiableObject * source )
{
        if( source->isKindOf( PersistentMemoryObjectClass ) ) {
		return( new FileIndex( (PersistentMemoryObject *)source ) );
        }
        else return( 0 );
}

/*
int
sizeOfFileIndex()
{
	return( sizeof( FileIndex ) );
}

Class *
FileIndex::classOf()
{
	return( FileIndexClass );
}
*/


#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("FileIndex",FileIndex,PersistentArray,&FileIndexConstructor,0);
