#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * FileObject.c - a view of another memory object.
 */

#include "Assert.h"
#include "Raid.h"
#include "Byte.h"
#include "FileObject.h"
#include "MemoryObjectAsync.h"
#include "MemoryObjectContainer.h"
#include "FileSystem.h"
#include "Class.h"

#ifdef __GNUC__
#pragma implementation
#endif

/* #define JEFF */

FileObject::FileObject( unsigned int numberOfUnits,
				    int log2USize,
				    int idNumber,
				    MemoryObjectContainer * container,
				    PersistentMemoryObject * source,
				    int offset )
: PersistentMemoryObject( numberOfUnits, log2USize, idNumber, container ), _source( source )
{

	_offset = offset;

	initDebug();
	Raid( RaidConstructor )( "%N::FileObject( %N, offset %d )\n",
				 this, source, offset );
#ifdef JEFF
	Console() << this << "::" << __FUNCTION__ << "(" << numberOfUnits
	  << " " << log2USize << ")\n" << eor;
#endif /* JEFF */

#ifdef FLATTEN
	_baseShift = -1;
#endif /* FLATTEN */

	if ( _source == 0 ) 
		return; //VChoices

	_segmentSize = _source->segmentSize();
	_unitShift = log2UnitSize() - _source->log2UnitSize();
	if( _unitShift < 0 ) {
		_buffered = 1;
		_unitShift = -_unitShift;
	}
	else {
		_buffered = 0;
	}
}

FileObject::~FileObject()
{
	Raid( RaidConstructor )( "%N::~FileObject()\n", this );
	_source = 0;
}

int
FileObject::transient()
{
	return( _source->transient() );
}

int
FileObject::rawRead( unsigned start, unsigned count, char * buf, unsigned nbytes,
		     IOAsyncClosure * c )
{
	Raids( RaidProtected ) 
		<< this << "::rawRead( start " << start
		<< ", count " << count << ", buf " << asAddress( buf )
		<< ", nbytes " << nbytes << ", closure " << asAddress( c )
		<< " )\n" << eor;

	Assert( ( start + count ) <= numberOfUnits() );

	int unitsRead = fileObjectRead( start, count, buf, nbytes, c );

	Raids( RaidProtected ) 
		<< this << "::rawRead return( " << unitsRead 
		<< " )\n" << eor;

	return unitsRead;
}

int
FileObject::rawWrite( unsigned start, unsigned count, char * buf, unsigned nbytes,
		      int writeThrough, IOAsyncClosure * c )
{
	Raids( RaidProtected ) 
		<< this << "::rawWrite( start " << start
		<< ", count " << count << ", buf " << asAddress( buf )
		<< ", nbytes " << nbytes 
		<< ", writeThrough " << writeThrough 
		<< ", closure " << asAddress( c ) << " )\n" << eor;

	Assert( ( start + count ) <= numberOfUnits() );

	int unitsWritten = 
		fileObjectWrite( start, count, buf, nbytes, writeThrough, c );
	Assert( unitsWritten <= (int)count );

	Raids( RaidProtected ) 
		<< this << "::rawWrite return( " << unitsWritten
		<< " )\n" << eor;

	return unitsWritten;
}

int
FileObject::fileObjectRead( unsigned start, int count, char * buf, int nbytes,
			    IOAsyncClosure * c )
{
	Raids( RaidProtected ) 
		<< this << "::fileObjectRead( start " << start
		<< ", count " << count << ", buf " << asAddress( buf )
		<< ", nbytes " << nbytes << ", closure " << asAddress( c )
		<< " )\n" << eor;
#ifdef JEFF
	Console() << "(" << __FILE__ << ":" << __LINE__ << ")"
		<< this << "::fileObjectRead( start " << start
		<< ", count " << count << ", buf " << asAddress( buf )
		<< ", nbytes " << nbytes << ", closure " << asAddress( c )
		<< " )\n" << eor;
#endif

	if( _buffered ) {

		/* do not support async operation for now */

		start += _offset;
		int offset = start & ( ( 1 << _unitShift ) - 1 );
		int endUnit = start + count - 1;
		start >>= _unitShift;
		endUnit >>= _unitShift;
		int units = endUnit - start + 1;

		char * buffer = new char[units * _source->unitSize()];

		int unitsRead = _source->read( start, units, buffer, 
					       units * _source->unitSize() );

		if( unitsRead == 0 )
			count = 0;
		else if( unitsRead < units ) {
			unitsRead = ( unitsRead << _unitShift ) - offset;
			if ( unitsRead < count )
				count = unitsRead;
		}

		Assert( count >= 0 );

		if( count > 0 ) {
			int maxBytes = count << log2UnitSize();
			ByteCopy( buffer + ( offset << log2UnitSize() ), 
				  buf, 
				  ( nbytes < maxBytes ? nbytes : maxBytes ) );
		}

		delete buffer;

		if( c ) {
			c->activate( count );
		}
	}
#ifdef FLATTEN
	else if ( _baseShift >= 0 ) {
		start <<= _baseShift;
		start += _baseOffset;
		count <<= _baseShift;
		if ( c ) {
			c->_rightShift += _baseShift;
		}
		count = _baseSource->rawRead( start, count, buf, nbytes, c )
			>> _baseShift;
	}
#endif /* FLATTEN */
	else {
		start <<= _unitShift;
		start += _offset;
		count <<= _unitShift;
		if ( c ) {
			c->_rightShift += _unitShift;
			_source->asyncRead( start, count, buf, nbytes, c );
		} else {
			count =  _source->read( start, count, buf, nbytes )
				 >> _unitShift;
		}
	}


	Raids( RaidProtected ) 
		<< this << "::fileObjectRead return( " << count
		<< " )\n" << eor;

	return count;
}

int
FileObject::fileObjectWrite( unsigned start, int count, char * buf, int nbytes,
			     int writeThrough, IOAsyncClosure * c )
{
	Raids( RaidMethod ) 
		<< this << "::fileObjectWrite( start " << start
		<< ", count " << count << ", buf " << asAddress( buf )
		<< ", nbytes " << nbytes 
		<< ", writeThrough " << writeThrough 
		<< ", closure " << asAddress( c ) << " )\n" << eor;

	if( _buffered ) {

		/* do not support async operation for now */

		start += _offset;
		int offset = start & ( ( 1 << _unitShift ) - 1 );
		int endUnit = start + count - 1;
		start >>= _unitShift;
		endUnit >>= _unitShift;
		int units = endUnit - start + 1;

		char * buffer = new char[units * _source->unitSize()];

		int unitsRead = _source->read( start, units, buffer,
					units * _source->unitSize() );
		Assert( unitsRead == units );

		ByteCopy( buffer + offset, buf, nbytes );

		int unitsWritten = _source->write( start, units, buffer,
					   units * _source->unitSize() );

		delete buffer;

		if( unitsWritten < units )
			count = ( unitsWritten << _unitShift ) - offset;

		if( c ) {
			c->activate( count );
		}
		
	}
#ifdef FLATTEN
	else if ( _baseShift >= 0 ) {
		start <<= _baseShift;
		start += _baseOffset;
		count <<= _baseShift;
		if ( c ) {
			c->_rightShift += _baseShift;
		}
		count = _baseSource->rawWrite( start, count, buf, nbytes, 
					       writeThrough, c ) 
		 	>> _baseShift;
	}
#endif /* FLATTEN */
	else {
		start <<= _unitShift;
		start += _offset;
		count <<= _unitShift;
		if ( writeThrough ) {
			if ( c ) {
				c->_rightShift += _unitShift;
				_source->asyncWriteThrough
					( start, count, buf, nbytes, c ); 
			} else {
				count = _source->writeThrough
					( start, count, buf, nbytes ) 
					>> _unitShift;
			}
		} else {
			if ( c ) {
				c->_rightShift += _unitShift;
				_source->asyncWrite
					( start, count, buf, nbytes, c );
			} else {
				count = _source->write
					( start, count, buf, nbytes ) 
					>> _unitShift;
			}
		}
	}

	Raids( RaidProtected ) 
		<< this << "::fileObjectWrite return( " << count
		<< " )\n" << eor;

	return count;
}

int
FileObject::basicCopy( MemoryObject * dest )
{
	Raids( RaidMethod )
		<< this << "::basicCopy( " << dest << " )\n" << eor;

	if( !_buffered ) 
		return( PersistentMemoryObject::basicCopy( dest ) );

	int returnValue = 0;
	int l2US = _source->log2UnitSize();
	int log2UnitDifference = l2US - dest->log2UnitSize();
	int readUnits = 1 << ( l2US - log2UnitSize() ) ;
	int writeUnits = 1;
	int bufferSize = _source->unitSize();

	if( log2UnitDifference > 0 ) {
		writeUnits <<= log2UnitDifference;
	}
	else if( log2UnitDifference < 0 ) {
		log2UnitDifference = -log2UnitDifference;
		readUnits <<= log2UnitDifference;
		bufferSize <<= log2UnitDifference;
	}

	Raid( RaidDetail )( "%N::basisCopy() ru %d wu %d bs %d\n",
			    this, readUnits, writeUnits, bufferSize );

	char * buffer = new char[bufferSize];

	unsigned w = 0;
	for( unsigned r = 0; r < numberOfUnits(); r += readUnits, w += writeUnits ){
	    int status = read( r, readUnits, buffer, bufferSize );
	    if( ( status < readUnits ) && ( r+status < numberOfUnits() ) ) {
		    returnValue = ReadError;
		    break;
	    }
	    if( dest->write( w, writeUnits, buffer, bufferSize )
							!= writeUnits ) {
		    returnValue = WriteError;
		    break;
	    }
	}
	delete buffer;

	if( dest->isKindOf( PersistentMemoryObjectClass ) ) {
		((PersistentMemoryObject *)dest)-> 
			setNumberOfBytes( numberOfBytes() );
	}

	Raids( RaidMethod )
		<< this << "::basicCopy( " << dest << " ) returns " 
		<< returnValue << "\n" << eor;
	return( returnValue );
}

void
FileObject::writeIdNumber( OutputStream * os )
{
	if (_source)
		_source->writeIdNumber( os );
	PersistentMemoryObject::writeIdNumber( os );
}

int
FileObject::userId()
{
	return( _source->userId() );
}

int
FileObject::groupId()
{
	return( _source->groupId() );
}

#include "ClassInstaller.h"
DEFINECLASS("FileObject",FileObject,PersistentMemoryObject,0,0);
