#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MadFileNode.c - 
 */
#include "Assert.h"
#include "MadFileNode.h"
#include "FileSystem.h"
#include "UNIXFile.h"
#include "MemoryObjectContainer.h"
#include "MadFileSystem.h"
#include "Class.h"
#include "Raid.h"
#include "PersistentArray.h"
#include "PersistentClass.h"

#ifdef __GNUC__
#pragma implementation
#endif


/* 
 * Initialize the in-memory inode
 */
MadFileNode::MadFileNode( unsigned int numOfUnits, int log2UnitSize,
			  int inumber, MemoryObjectContainer * container,
			  PersistentMemoryObject * source,
			  MadNode * madNode, Class * aClass )
: ( numOfUnits, log2UnitSize, inumber, container, source, 0 )
{
    initDebug();
    Raid( RaidConstructor )( "%N::MadFileNode()\n", this );

    /*
     *	Initialize variables from parameters
     */
    _madNode = madNode;
    _class = aClass;

    Raid( RaidDetail )( "%N::MadFileNode size %d units %d\n",
			this, _madNode->_size, numberOfUnits() );
}

Class *
MadFileNode::supports( Class * aClass )
{
    Raid( RaidProtected )( "%N::supports( %N ) %N\n", this, aClass, _class );

    Class * supportedClass = 0;

    if( _class->isASubclassOf( aClass ) ) {
	supportedClass = _class;
    }
    else if( PersistentArrayClass->isASubclassOf( aClass ) ) {
        supportedClass = PersistentArrayClass;
    }

    Raid( RaidProtected )( "%N::supports() returns %N\n",
                           this, supportedClass );

    return( supportedClass );
}

/*
 * Clean up the in memory inode data
 */
MadFileNode::~MadFileNode()
{
    Raid( RaidConstructor )( "%N::~MadFileNode()\n", this );
}

int
MadFileNode::mapUnit( int logicalBlock, int addMapping )
{
    Raid( RaidProtected )( "%N::mapUnit %d %d\n",
			   this, logicalBlock, addMapping);
    _mutex.P();
    Assert( logicalBlock < 16 );
    if( _madNode->_blocks[logicalBlock] == 0 ) {
	if( addMapping == 1 ) {
	    _madNode->_blocks[logicalBlock] = _container->allocate( 
					_madNode->_blocksAllocated, 0, 1 );
	} 
    } else if( addMapping == -1 ) {
	_container->free( _madNode->_blocksAllocated, 
			  _madNode->_blocks[logicalBlock], 1 );
	_madNode->_blocks[logicalBlock] = 0;
    }
    _mutex.V();
    Raid( RaidProtected )( "%N::mapUnit = %d\n",
			   this, _madNode->_blocks[logicalBlock] );
    return( _madNode->_blocks[logicalBlock] );
}

int
MadFileNode::numberOfBytes()
{
    Raid( RaidMethod )( "%N::numberOfBytes %d\n", this, _madNode->_size );
    return( _madNode->_size );
}

int
MadFileNode::basicSetNumberOfBytes( int size )
{
    Raid( RaidProtected )( "%N::basicSetNumberOfBytes( %d ) %d\n",
			   this, size, _madNode->_size );
    if( size < 0 ) size = 0;
    _madNode->_size = size;
    basicSetNumberOfUnits( ( size + unitSize() - 1 ) >> log2UnitSize() );
    return( size );
}

/*
 *	this function allows other objects to access and modify the
 *	link count ( or reference count ) of this inode.
 *	links returns the new number of links for this inode.
 */
int
MadFileNode::links( int delta )
{
    Raid( RaidMethod )( "%N::links( delta %d )\n", this, delta );
    _mutex.P();
    _madNode->_references += delta;
    int result = _madNode->_references;
    _mutex.V();
    Raid( RaidMethod )( "%N::links returns %d\n", this, result );
    return( result );
}

/* 
 * This does the UNIX Stat( 2 ) function.
 */
void
MadFileNode::info( UNIXFileInfo * infoBuffer )
{
    Raid( RaidMethod )( "%N::info( %x ) %d\n", this, infoBuffer, _idNumber );
    infoBuffer->device	= 0;
    infoBuffer->inumber	= _idNumber;

    extern ClassStar MemoryObjectDictionaryClass;
    extern ClassStar SymbolicLinkClass;

    if( _class->isASubclassOf( MemoryObjectDictionaryClass ) ) {
        infoBuffer->mode = 040000;
    }
    else if( _class->isASubclassOf( SymbolicLinkClass ) ) {
        infoBuffer->mode = 0120000;
    }
    else {
        infoBuffer->mode = 0100000;//FIX
    }
    infoBuffer->links	= _madNode->_references;
    infoBuffer->uid	= 0;//FIX
    infoBuffer->gid	= 0;//FIX
    infoBuffer->size	= _madNode->_size;
    infoBuffer->atime	= _madNode->_readTime;
    infoBuffer->mtime	= _madNode->_writeTime;
    infoBuffer->ctime	= _madNode->_createTime;
    infoBuffer->blocks	= _madNode->_blocksAllocated;
    infoBuffer->blockSize = 0; //FIX
    infoBuffer->flags	= 0;
    infoBuffer->rdevice	= 0;
}

int
MadFileNode::basicSetNumberOfUnits( unsigned int units )
{
    Raid( RaidProtected )( "%N::basicSetNumberOfUnits( units %d ) %d\n",
			   this, units, numberOfUnits() );

    if( units == numberOfUnits() ) return( units );

    _mutex.P();
    for( int i = units; i < 16; i++ ) {
	int physicalBlock = _madNode->_blocks[i];
	if( physicalBlock != 0 ) {
	    _container->free( _madNode->_blocksAllocated, physicalBlock, 1 );
	}
    }
    _mutex.V();

    setModified();

    return( MemoryObject::basicSetNumberOfUnits( units ) );
}


int
MadFileNode::recordSize()
{
        Assert( _madNode->_recordSize > 0 );
        return( _madNode->_recordSize );
}

int
MadFileNode::basicSetRecordSize( int size )
{
        if( size > 0 ) _madNode->_recordSize = size;
        return( _madNode->_recordSize );
}

int
MadFileNode::accessMode()
{
        return( UNIXAccessMask );
}


#include "ClassInstaller.h"
DEFINECLASS(MadFileNode",MadFileNode,UNIXInode,0,0);
