#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Author: Daniel E. Weber (weber@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#include "MailContainer.h"
#include "MailEntry.h"
#include "MailRootEntry.h"
#include "FileSystem.h"
#include "String.h"
#include "Raid.h"

#ifdef __GNUC__
#pragma implementation
#endif

/*
 * Either open the file as the Root entry (if the Id is 1) or open it as 
 * a plain MailEntry.
 */
FileObject *
MailContainer::basicOpen( int id, int & status, int flags )
{
   Raid( RaidMethod )( "%N::basicOpen( inumber %d )\n", this, id );

   if ( id < 1 ) {
	 status = IdNumberOutOfRange;
	 return ( 0 );
   }
   else {
	status = 0;
	if( id == 1 ) {
		return( new MailRootEntry( id, this, _source, _messages ) );
	}
   	else {
		return( new MailEntry( _offsets[id+1] - _offsets[id],
				       0, id, this, _source, _offsets[id] ) );
	}
   }
}


/*
 * Destroy all of the stored information.  First, each of the individual
 * attributes for each message.  Then the entire message structure itself.
 */
MailContainer::~MailContainer()
{
	Raid( RaidConstructor )( "%N::~MailContainer()\n", this );

	if ( _valid ) delete _offsets;
}

/* 
 * MailContainer parses the mail file.  Each time it encounters the string
 * "\nFrom " it will consider that a signal for the start of a new message and
 * add its header information to the data structure MailHeader.  It makes two
 * complete passes of the file (which is stored in one large buffer and then
 * later deleted); one to determine the number of messages, and the second to
 * store all of the header information.
 */
MailContainer::MailContainer( PersistentMemoryObject * source )
: ( source )
{
    initDebug();

    Raid( RaidConstructor )( "%N::MailContainer( %N )\n", this, _source );

/*
 * If the file is empty, set all information to be NULL.
 */
    unsigned int bytes = source->numberOfBytes();
    if( bytes == 0 ) {
	_valid = 0;
	initializeArrays( 0 );
    }
    else {

/*
 * Create a temporary buffer to read in all of the mail file for parsing.
 * Must use chains to read the file.
 */
        int units = _source->numberOfUnits();
        int bufferSize = units << _source->log2UnitSize();
        char * buffer = new char[bufferSize];
        _source->read( 0, units, (char *)buffer, bufferSize);

/*
 * Begin parsing the Mail File.  If the file is not a Mail File (it does not
 * begin with the keyword "From ") then set a flag to screen out this case.
 * Otherwise, make one quick pass of the file to determine the size of the
 * header block.
 */
	if( strncmp( buffer, "From ", 5 ) != 0 ) {
	    _valid = 0;
	    initializeArrays( 0 );
	}
	else {
	    _valid = 1;
	    _messages = 1;
	    for( int pos = 4; pos < bytes; pos++ ) {
	        if( strncmp( &buffer[pos], "\nFrom ", 6 ) == 0 ) _messages++;
	    }

	    _offsets = new int[_messages+3];
	    initializeArrays( _messages+2 );
	    int message = 2;

	    _offsets[message] = 0;
	    _offsets[_messages+2] = bytes;

	    for( pos = 4; pos < bytes; pos++ ) {
		if( strncmp( &buffer[pos], "\nFrom ", 6 ) == 0 ) {
		    _offsets[++message] = ++pos;
		}
	    }
	}
	delete buffer;
    }
}

/*
 * The source will be a new MailContainer.
 */
ProxiableObject * 
MailContainerConstructor( ProxiableObject * source )
{
    if ( source->isKindOf( PersistentMemoryObjectClass ) ) {
	return( new MailContainer( (PersistentMemoryObject *)source ) );
    }
    else return( 0 );
}


#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("MailContainer",MailContainer,MemoryObjectContainer,&MailContainerConstructor,0);
//DEFINECLASS("MailContainer",MailContainer,MemoryObjectContainer,&MailContainerConstructor,0);
