#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MemoryObjectCachingContainer.c - 
 */

#include "Assert.h"
#include "Async.h"
#include "BlockAllocator.h"
#include "FileObject.h"
#include "FileSystem.h"
#include "FileSystemHints.h"
#include "FileSystemSimpleInstruments.h"
#include "FreeRunningTimer.h"
#include "FreeRunningTimerStar.h"
#include "MemoryObjectCache.h"
#include "MemoryObjectCachingContainer.h"
#include "MemoryObjectDictionary.h"
#include "NotifyEvents.h"
#include "Process.h"
#include "ProcessStar.h"
#include "Raid.h"
#include "Semaphore.h"
#include "Store.h"
#include "ThisProcess.h"
#include "TimeoutTimer.h"
#include "ZombieInfo.h"

#ifdef __GNUC__
#pragma implementation
#endif

#define STANDARDZOMBIEDELAY 30000000 /* 30 secs */

unsigned MemoryObjectCachingContainerPeriod = STANDARDZOMBIEDELAY;
unsigned MemoryObjectCachingContainerTimeWrap = 0x70000000;
unsigned StandardZombieDelay = STANDARDZOMBIEDELAY;

ZombieInfoList StaticSyncZombieList( ZOMBIE_SYNC );
ZombieInfoList StaticNotSyncZombieList( ZOMBIE_NOTSYNC );
ZombieInfoList StaticExternZombieList( ZOMBIE_EXTERN );

unsigned MemoryObjectCachingContainer::MaxSyncZombieCount = 2000;
ZombieInfoList * MemoryObjectCachingContainer::SyncZombieList
	= &StaticSyncZombieList;

unsigned MemoryObjectCachingContainer::MaxNotSyncZombieCount = 300;
ZombieInfoList * MemoryObjectCachingContainer::NotSyncZombieList
	= &StaticNotSyncZombieList;

ZombieInfoList * MemoryObjectCachingContainer::ExternZombieList
	= &StaticExternZombieList;

Lock MemoryObjectCachingContainer::ActiveHelpLock;
unsigned MemoryObjectCachingContainer::ActiveHelpDaemonCount = 0;
unsigned MemoryObjectCachingContainer::MaxActiveHelpDaemon = 4;

int MemoryObjectCachingContainer::blocked;
Lock MemoryObjectCachingContainer::slock;

static SemaphoreStar daemonSem;
static ProcessStar daemonProcess;
static Semaphore sem(1);

extern ClassStar UNIXInodeCacheClass;

MemoryObjectCachingContainer::MemoryObjectCachingContainer
	( PersistentMemoryObject * src )
     : MemoryObjectContainer( src )
{
	initDebug();
	Raid( RaidConstructor )( "%N::MemoryObjectCachingContainer( %N )\n", 
				 this, _source );

	// check if need to do static initializtions and start daemon process

	sem.P();
	if ( daemonProcess != 0 ) {
		sem.V();
		return;
	}

	// start timeout deamon for to sync zombie file objects

	blocked = 1;
	daemonProcess = new SystemProcess
		( (PFV) &MemoryObjectCachingContainer::daemon, 
		  thisProcess()->domain(),
		  normal, 0 );
	Assert( daemonProcess != 0 );
	daemonProcess->setName( "MemoryObjectCachingContainerDaemon" );
	daemonProcess->ready();
	daemonSem = new Semaphore( 0 );

	// done with static initializations

	sem.V();
}						  

MemoryObjectContainerTimer
MemoryObjectCachingContainerOpen("MemoryObjectCachingContainer::open timer");

FileObjectRef
MemoryObjectCachingContainer::open( int idNumber, int & status, int flags )
{
	SInstr( int istart = MemoryObjectCachingContainerOpen.start() );

	Raid( RaidMethod )( "%N::open( idNumber %d )\n", this, idNumber );
#ifdef JEFF
	Console() << this << "::open( " << idNumber << " ) _numberOfFileObjects="
	  <<_numberOfFileObjects << "\n" << eor;
#endif /* JEFF */

	// check for valid id number
	if(!this->IsIdNumberInRange(idNumber)) {
	  status = IdNumberOutOfRange;
	  Debug << this << "::open( returns 0, status " << status
	    << " )\n" << eor;
	  SInstr( MemoryObjectContainerOpen.stop( istart ) );
	  return 0;
	}

	// make sure not in middle of container sync

	_syncMutex.acquireRead();

	// make sure someone else is not trying to close id number

	acquire( idNumber );
	FileObjectRef fo = lookup( idNumber );

	// check if it is already open or a zombie

	if( fo == 0 ) {

		// not open or zombie

		fo = basicOpen( idNumber, status, flags );
		if( fo != 0 ) {
			bind( idNumber, fo );
			acquire();
			if ( _openFileObjects == 0 ) 
				reference();
			_openFileObjects++;
			release();
		}
	} else {

		// open or zombie
		// check if zombie

		if ( ((PersistentMemoryObject *) fo)->referenceCount() == 1 ) {

			// it is a zombie, remove from appropriate zombie list

			ZombieInfo * zombieInfo = fo->zombieInfo();
			Assert( zombieInfo->type() != ZOMBIE_NOT );
#ifdef SAFESYNC
			if ( zombieInfo->type() == ZOMBIE_NOTSYNC ) {
				Assert( zombieInfo->expire() > 0 );
				zombieInfo->setExpire( 0 );
			} else {
				slock.acquire();
				zombieInfo->unlink(); 
				slock.release();
				Assert( zombieInfo->time() == 0 );
				Assert( zombieInfo->expire() == 0 );
			}
#else // not SAFESYNC
			if ( zombieInfo->type() == ZOMBIE_NOTSYNC ) {
				Assert( zombieInfo->time() == 
					zombieInfo->expire() );
				zombieInfo->setTime( 0 );
				zombieInfo->setExpire( 0 );
			} else {
				Assert( zombieInfo->time() == 0 );
				Assert( zombieInfo->expire() == 0 );
			}
			slock.acquire();
			zombieInfo->unlink();
			slock.release();
#endif // SAFESYNC
		}

		// indicate success

		status = 0;
	}

	if ( fo != 0) {
		parseFlags( fo, flags ); 
	}

	release( idNumber );

	_syncMutex.releaseRead();

	Raid( RaidMethod )( "%N::open returns %N %d\n", this, fo, status );
#ifdef JEFF
	Console() << this << "::" << __FUNCTION__ << " returns fo "
	  << fo << ", status " << status << "\n" << eor;
#endif /* JEFF */

	SInstr( MemoryObjectCachingContainerOpen.stop( istart ) );

	return( fo );
}

MemoryObjectContainerTimer
MemoryObjectCachingContainerClose("MemoryObjectCachingContainer::close timer");

void
MemoryObjectCachingContainer::close( FileObject * fo )
{
	SInstr( int istart = MemoryObjectCachingContainerClose.start() );

	Raid( RaidMethod )( "%N::close( %N )\n", this, fo );

	// don't need to acquire individual lock since it is
	// during unreference of file object
	// unreference needs to acquire lock to avoid race condition
	// between open/create and close

	// check if need to make it a zombie
	// don't make it a zombie if it is transient or
	// zombieDelay = 0;

	fo->notify( NotifyEvent_ContainerClose );

	if ( fo->transient() || fo->zombieDelay() == 0 ) {
		ZombieInfo * zombieInfo = fo->zombieInfo();
		ZombieInfoType type = zombieInfo->type();
#ifdef SAFESYNC
		Assert( type == ZOMBIE_NOT || type == ZOMBIE_NOTSYNC );
		if ( type == ZOMBIE_NOTSYNC ) {
			slock.acquire();
			zombieInfo->unlink();
			slock.release();
		}
#else // not SAFESYNC
		Assert( type == ZOMBIE_NOT );
#endif /*SAFESYNC*/
	        MemoryObjectContainer::close( fo );
		SInstr( MemoryObjectCachingContainerClose.stop( istart ) );
		return;
	}

	int idNumber = fo->idNumber();

	// make it a zombie which is not synchronize yet, 
	// it can potentially contain dirty blocks that needs to be sync'ed

	ZombieInfo * zombieInfo = fo->zombieInfo();
	Assert( zombieInfo->expire() == 0 );

	slock.acquire();					
	unsigned expire = SystemTimer->time() + fo->zombieDelay();
	if ( expire < 2 )
		expire = 2;
	zombieInfo->setFO( fo );
#ifdef SAFESYNC
	int doInsert;
	if ( zombieInfo->type() == ZOMBIE_NOT ) {
		Assert( zombieInfo->time() == 0 );
		doInsert = 1;
	} else {
		Assert( zombieInfo->type() == ZOMBIE_NOTSYNC );
		Assert( zombieInfo->time() != 0 );
		if ( zombieInfo->expire() == 1 ) {
			zombieInfo->unlink();
			doInsert = 1;
		} else {
			doInsert = 0;
		}
	}
	zombieInfo->setExpire( expire );
	if ( doInsert ) {
		Assert( zombieInfo->type() == ZOMBIE_NOT );
		zombieInfo->setTime( expire );
		NotSyncZombieList->insertByTime( zombieInfo );
	}
#else // not SAFESYNC
	zombieInfo->setExpire( expire );
	zombieInfo->setTime( expire );
	Assert( zombieInfo->type() == ZOMBIE_NOT );
	NotSyncZombieList->insertByTime( zombieInfo );
#endif // SAFESYNC
	
	// if the sync daemon is blocked because there is nothing
	// waiting to be synced, then we should unblock it

	if ( blocked ) { 
		daemonSem->V();
		blocked = 0;
	}

	slock.release();

	// notify the file object about its status change to not sync zombie

//	fo->notify( NotifyEvent_ZombieStatus );

	// release lock acquired by unreference

	release( idNumber );

	// all done

	SInstr( MemoryObjectCachingContainerClose.stop( istart ) );

	if ( NotSyncZombieList->count() > (int)MaxNotSyncZombieCount ) {
		spawnHelpDaemon();
	}
}

void
MemoryObjectCachingContainer::killZombie( int idNumber )
{
	Raids( RaidMethod ) 
		<< this << "::killZombie( " << idNumber << " )\n" << eor;

	_syncMutex.acquireRead();
	acquire( idNumber );
	FileObject * fo = lookup( idNumber );
	if ( fo == 0 ) {
		release( idNumber );
		_syncMutex.releaseRead();
		return;
	}
	Assert( fo->alive() );
	ZombieInfo * info = fo->zombieInfo();
	ZombieInfoType type = info->type();
	if ( type == ZOMBIE_NOT ) {
		Assert( info->time() == 0 );
		release( idNumber );
		_syncMutex.releaseRead();
		return;
	}
#ifdef SAFESYNC
	if ( type == ZOMBIE_NOTSYNC && info->expire() == 0 ) {
		Assert( info->time() == 0 );
		slock.acquire();
		info->unlink();
		slock.release();
		release( idNumber );
		_syncMutex.releaseRead();
		return;
	}
#else // not SAFESYNC
	Assert( type != ZOMBIE_NOTSYNC || info->expire() == info->time() );
#endif // SAFESYNC
	slock.acquire();
	info->unlink();
	info->setType( type );
	slock.release();
	basicClose( fo );
	unbind( idNumber );
	release( idNumber );
	acquire();
	_openFileObjects --;
	Assert( _openFileObjects >= 0 );
	int noOpenFileObjects = ( _openFileObjects == 0 );
	release();
	_syncMutex.releaseRead();
	if( noOpenFileObjects ) 
		unreference();
}

int
MemoryObjectCachingContainer::shutdown()
{
	int total = 0;
	sem.P();
	Console() 
		<< "MemoryObjectCachingContainer::shutdown " 
		<< NotSyncZombieList->count()
		<< " not sync'ed zombie...\n" << eor;
	total += NotSyncZombieList->count();
	shutdown( NotSyncZombieList );
	Console() 
		<< "MemoryObjectCachingContainer::shutdown " 
		<< SyncZombieList->count()
		<< " sync'ed zombie...\n" << eor;
	total += SyncZombieList->count();
	shutdown( SyncZombieList );
	Console() 
		<< "MemoryObjectCachingContainer::shutdown " 
		<< ExternZombieList->count()
		<< " extern'ed zombie...\n" << eor;
	total += ExternZombieList->count();
	shutdown( ExternZombieList );
	Console()
		<< "MemoryObjectCachingContainer::shutdown wait 2 secs\n" 
		<< eor;
	TimeoutTimer timer;
	timer.start( 2 * 1000000 );
	timer.await();
	Console()
		<< "MemoryObjectCachingContainer::shutdown complete\n";
	sem.V();
	return total;	
}

int debugShutdown = 0;

void
MemoryObjectCachingContainer::shutdown( ZombieInfoList * list )
{
	slock.acquire();
	while( list->count() > 0 ) {
		ZombieInfo * info = list->head();
		FileObject * fo = info->fo();
		MemoryObjectCachingContainer * container = 
			 ( MemoryObjectCachingContainer * ) fo->fastContainer();
	        slock.release();
		if (debugShutdown) Console() 
			<< "    kill zombie " << fo << "\n" << eor;
		container->killZombie( fo->idNumber() );
	 	slock.acquire();
	}
	slock.release();
	return;	
}

void
MemoryObjectCachingContainer::synchronize()
{
	// make sure no one else is opening/creating files when
	// doing a container sync, if open/create takes place it
	// may not be sync'ed causing inconsistency
	// this also avoids the need to acquire lock for individual files

	_syncMutex.acquireWrite();

	// find all of my file objects which are open/zombie

	FileObject ** active = TheContainerIndex->getActiveFileObjects(this);

	// acquire exclusive lock for each of these file objects

	FileObject ** fo;
	for ( fo = active; *fo != 0; fo++ ) {
		Assert( (*fo)->alive() );
		(*fo)->acquireAll();
	}

#ifdef SAFESYNC
	for ( fo = active; *fo != 0; fo++ ) {
		ZombieInfo * info = (*fo)->zombieInfo();
		if ( info->type() == ZOMBIE_NOTSYNC ) {
			unsigned expireTime = info->expire();
			if ( expireTime == 0 ) {
				slock.acquire();
				info->unlink();
				info->setTime( 0 );
				slock.release();
			} else if ( expireTime >= 2 ) {
				slock.acquire();
				info->unlink();
				info->setTime( expireTime );
				info->setExpire( 1 );
				NotSyncZombieList->insertByTime( info );
				slock.release();
			} else {
				// expireTime == 1
				// do nothing
			}
		}
	}
#endif // SAFESYNC

	// now call subclass method to synchronize container

	_activeFileObjects = active;
	basicSynchronize();
	_activeFileObjects = 0;


	// release exclusive lock on each of my file objects

	for ( fo = active; *fo != 0; fo++ ) {
		Assert( (*fo)->alive() );
		(*fo)->releaseAll();
	}

	// release the data structure used to get active file objects

	TheContainerIndex->freeActiveFileObjects(active);

	// all done can allow open and creates now

	_syncMutex.releaseWrite();

	if ( SyncZombieList->count() > (int)MaxSyncZombieCount ) {
		killZombies();
	}
}

int debugMemoryObjectCachingContainerDaemon = 0;

void 
MemoryObjectCachingContainer::daemon()
{
	int currentTime;
	int idNumber;
	FileObject * fo;
	MemoryObjectCachingContainer * container;
	TimeoutTimer timer;
	ZombieInfo * zombieInfo = 0;

	timer.setName( "MemoryObjectCachingContainerTimer" );
	int sleepTime = MemoryObjectCachingContainerPeriod;

	while (1) {
		
		if ( sleepTime != 0 ) {
			timer.start( sleepTime );
			if ( debugMemoryObjectCachingContainerDaemon ) 
			    Debug
			    << "MemoryObjectCachingContainer::daemon sleeping "
			    << "for " << sleepTime << " ms\n" << eor;
			timer.await();
		}

		slock.acquire();
		if ( NotSyncZombieList->isEmpty() ) {
			blocked = 1;
			slock.release();
			daemonSem->P();
			sleepTime = MemoryObjectCachingContainerPeriod;
			continue;
		}

		zombieInfo = NotSyncZombieList->head();
		Assert( zombieInfo != NotSyncZombieList );
		fo = zombieInfo->fo();
		container = fo->cachingContainer();
		idNumber = fo->idNumber();
		slock.release();

		/*
		 * to prevent deadlock always acquire individual locks first
		 */

		if ( debugMemoryObjectCachingContainerDaemon ) Debug
			<< "MemoryObjectCachingContainer::daemon() acquiring "
			<< "locks on " << container 
			<< "(" << idNumber << ")\n" << eor;

		container->_syncMutex.acquireRead();

		container->acquire( idNumber );
		slock.acquire();

		/*
		 * the problem here is that first link may have changed
		 * since it was examined, now that it is locked, first link
		 * is locked and can no longer change check that it is 
		 * still the same. if it has changed, then fo just locked
		 * is no longer a zombie
		 */

		if ( NotSyncZombieList->head() != zombieInfo ) {

			/*
			 * oops, the fo just locked is no longer a zombie
			 */

			if ( debugMemoryObjectCachingContainerDaemon ) 
				Debug
				<< "MemoryObjectCachingContainer::daemon() "
				<< "oops fo no longer zombie\n" << eor;

			slock.release();
			container->release( idNumber );
			container->_syncMutex.releaseRead();
			sleepTime = 0;
			continue;
		}

		/*
		 * check if expiration time is up yet, figure out the
		 * amount of time remaining to sleep and then loop again
		 */

		unsigned zombieTime = zombieInfo->time();
		if ( NotSyncZombieList->count() <= (int)MaxNotSyncZombieCount ) {
			currentTime = SystemTimer->time();
			sleepTime = zombieTime - currentTime;

			if ( debugMemoryObjectCachingContainerDaemon ) 
			Debug
			<< "MemoryObjectCachingContainer::daemon() "
		        << "currentTime " << currentTime << "\n" << eor;

			if ( (int)zombieTime > currentTime &&
			     sleepTime < (int)MemoryObjectCachingContainerTimeWrap ){

		        	if ( debugMemoryObjectCachingContainerDaemon ) 
				Debug
				<< "MemoryObjectCachingContainer::daemon() "
				<< "time not expired\n" << eor;

				slock.release();
				container->release( idNumber );
				container->_syncMutex.releaseRead();
				continue;
			}
		} else {
			if ( debugMemoryObjectCachingContainerDaemon )
			Debug
			<< "MemoryObjectCachingContainer::daemon() "
			<< " NotSyncZombieCount " << NotSyncZombieList->count()
			<< " > MaxNotSyncZombieCount " 
			<< MaxNotSyncZombieCount
			<< "\n" << eor;
		}

		/*
		 * time is up, or
		 * exceed MaxNotSyncZombieCount,
		 * remove fo from zombie list 
		 */
		
		zombieInfo->unlink();
		unsigned expireTime = zombieInfo->expire();
//		unsigned notifyFO;
#ifdef SAFESYNC
		if ( expireTime == zombieTime || expireTime == 1 ) {
			zombieInfo->insertBefore( SyncZombieList );
			zombieInfo->setTime( 0 );
			zombieInfo->setExpire( 0 );
//			notifyFO = 1;
		} else if ( expireTime == 0 ) {
			zombieInfo->setTime( 0 );
//			notifyFO = 1;
		} else {
			zombieInfo->setTime( expireTime );
			zombieInfo->setExpire( 1 );
			NotSyncZombieList->insertByTime( zombieInfo );
//			notifyFO = 0;
		}
#else // not SAFESYNC
		Assert( expireTime == zombieTime );
		zombieInfo->insertBefore( SyncZombieList );
		zombieInfo->setTime( 0 );
		zombieInfo->setExpire( 0 );
//		notify = 1;
#endif // SAFESYNC

		slock.release();

		Assert( container->lookup( idNumber )->identity() == 
		        fo->identity() );

		if ( debugMemoryObjectCachingContainerDaemon ) Debug
			<< "MemoryObjectCachingContainer::daemon() closing " 
			<< fo << " idNumber " << idNumber << "\n" << eor;

//		if ( notifyFO ) 
//			fo->notify( NotifyEvent_ZombieStatus );
		if ( expireTime != 1 ) 
			fo->synchronize();
		container->release( idNumber );
		container->_syncMutex.releaseRead();
		if ( SyncZombieList->count() > (int)MaxSyncZombieCount ) {
			killZombies();
		}

		sleepTime = 0;
	}
}

class SpawnHelpDaemonClosure  : public AsyncClosure {
public:
	static void func( SpawnHelpDaemonClosure * c );
  SpawnHelpDaemonClosure() /*: _urgency( 1 )*/ {Assert(NOTREACHED); //_func = (AyncFunc)&func;
  }

};

void
SpawnHelpDaemonClosure::func( SpawnHelpDaemonClosure * c )
{
	MemoryObjectCachingContainer::helpDaemon();
	MemoryObjectCachingContainer::ActiveHelpLock.acquire();
	Assert(MemoryObjectCachingContainer::ActiveHelpDaemonCount > 0);
	MemoryObjectCachingContainer::ActiveHelpDaemonCount--;
	MemoryObjectCachingContainer::ActiveHelpLock.release();
	delete c;
}

void
MemoryObjectCachingContainer::spawnHelpDaemon()
{
	int spawn = 0;
	ActiveHelpLock.acquire();
	if ( ActiveHelpDaemonCount < MaxActiveHelpDaemon ) {
		spawn = 1;
		ActiveHelpDaemonCount++;
	}
	ActiveHelpLock.release();
	if (spawn) {
		SpawnHelpDaemonClosure * c =
			new SpawnHelpDaemonClosure();
		TheAsyncSupport->schedule( c );
	}
}

void
MemoryObjectCachingContainer::helpDaemon()
{
	int idNumber;
	FileObject * fo;
	MemoryObjectCachingContainer * container;
	ZombieInfo * zombieInfo;
	while ( NotSyncZombieList->count() > (int)MaxNotSyncZombieCount ) {
		slock.acquire();
		if ( NotSyncZombieList->isEmpty() ) {
			slock.release();
			break;
		}

		zombieInfo = NotSyncZombieList->head();
		Assert( zombieInfo != NotSyncZombieList );
		Assert( zombieInfo->type() == ZOMBIE_NOTSYNC );
		fo = zombieInfo->fo();
		container = fo->cachingContainer();
		idNumber = fo->idNumber();
		slock.release();

		/*
		 * to prevent deadlock always acquire individual locks first
		 */

		container->_syncMutex.acquireRead();
		container->acquire( idNumber );
		slock.acquire();

		/*
		 * the problem here is that first link may have changed
		 * since it was examined, now that it is locked, first link
		 * is locked and can no longer change check that it is 
		 * still the same. if it has changed, then fo just locked
		 * is no longer a zombie
		 */

		if ( NotSyncZombieList->head() != zombieInfo ) {

			/*
			 * oops, the fo just locked is no longer a zombie
			 */

			slock.release();
			container->release( idNumber );
			container->_syncMutex.releaseRead();
			continue;
		}

		unsigned expireTime = zombieInfo->expire();
		unsigned zombieTime = zombieInfo->time();
		zombieInfo->unlink();
//		unsigned notifyFO;
#ifdef SAFESYNC
		if ( expireTime == zombieTime || expireTime == 1 ) {
			zombieInfo->insertBefore( SyncZombieList );
			zombieInfo->setTime( 0 );
			zombieInfo->setExpire( 0 );
//			notifyFO = 1;
		} else if ( expireTime == 0 ) {
			zombieInfo->setTime( 0 );
//			notifyFO = 1;
		} else {
			zombieInfo->setTime( expireTime );
			zombieInfo->setExpire( 1 );
			NotSyncZombieList->insertByTime( zombieInfo );
//			notifyFO = 0;
		}
#else // not SAFESYNC
		Assert( expireTime == zombieTime );
		zombieInfo->insertBefore( SyncZombieList );
		zombieInfo->setTime( 0 );
		zombieInfo->setExpire( 0 );
//		notify = 1;
#endif // SAFESYNC

		slock.release();

		Assert( container->lookup( idNumber )->identity() == 
		        fo->identity() );

//		if ( notifyFO )
//			fo->notify( NotifyEvent_ZombieStatus );
		if ( expireTime != 1 ) 
			fo->synchronize();
		container->release( idNumber );
		container->_syncMutex.releaseRead();
		if ( SyncZombieList->count() > (int)MaxSyncZombieCount ) {
			killZombies();
		}
	}
}

int
MemoryObjectCachingContainer::killZombies()
{
	// urgency not used for now, may be used to remove unsynced zombie too

	int idNumber;
	int count = 0;
	FileObject * fo;
	MemoryObjectCachingContainer * container;
	ZombieInfo * zombieInfo = 0;

	slock.acquire();
	while ( ! SyncZombieList->isEmpty() &&
		( ( SyncZombieList->count() > (int)MaxSyncZombieCount ) ||
		  ( MainStore->totalExpectedAllocPaus() > MainStore->low() ) 
		) 
	      ) {
		
		zombieInfo = SyncZombieList->head();
		Assert( zombieInfo != SyncZombieList );
		Assert( zombieInfo->type() == ZOMBIE_SYNC );
		fo = zombieInfo->fo();
		container = fo->cachingContainer();
		idNumber = fo->idNumber();
		slock.release();

		/*
		 * to prevent deadlock always acquire individual locks first
		 */

		container->_syncMutex.acquireRead();
		container->acquire( idNumber );
		slock.acquire();

		/*
		 * the problem here is that first link may have changed
		 * since it was examined, now that it is locked, first link
		 * is locked and can no longer change check that it is 
		 * still the same. if it has changed, then fo just locked
		 * is no longer a zombie
		 */

		if ( SyncZombieList->head() != zombieInfo ) {

			/*
			 * oops, the fo just locked is no longer a zombie
			 */

			slock.release();
			container->release( idNumber );
			container->_syncMutex.releaseRead();
			slock.acquire();
			continue;
		}

		count++; 
		ZombieInfoType type = zombieInfo->type();
		zombieInfo->unlink();

		Assert( container->lookup( idNumber )->identity() == 
		        fo->identity() );

		MemoryObjectCache * cache = fo->cache();
		if ( cache && cache->hasCache( UNIXInodeCacheClass ) ) {
			zombieInfo->insertBefore( ExternZombieList );
			slock.release();
			fo->flush();
			container->release( idNumber );
			container->_syncMutex.releaseRead();
		} else {
			zombieInfo->setType( type );
			slock.release();
			container->basicClose( fo );
			container->unbind( idNumber );
			container->release( idNumber );
			container->acquire();
			container->_openFileObjects --;
			int noOpenFileObjects = 
				( container->_openFileObjects == 0 );
			container->release();
			container->_syncMutex.releaseRead();
			if( noOpenFileObjects ) 
				container->unreference();
		}
		
		slock.acquire();
	}
	slock.release();
	return count;
}

int
MemoryObjectCachingContainer::unmount()
{
	synchronize();

	_syncMutex.acquireWrite();

	FileObject ** active = TheContainerIndex->getActiveFileObjects(this);
	FileObject ** fo;

	for ( fo = active; *fo != 0; fo++ ) {
		Assert( (*fo)->alive() );
		ZombieInfo * info = (*fo)->zombieInfo();
		if ( info->type() == ZOMBIE_SYNC || 
		     info->type() == ZOMBIE_EXTERN ) {
			int idNumber = (*fo)->idNumber();
			acquire( idNumber );
			slock.acquire();
			ZombieInfoType type = info->type();
			info->unlink();
			info->setType( type );
			slock.release();
			basicClose( *fo );
			unbind( idNumber );
			release( idNumber );
			acquire();
			_openFileObjects --;
			Assert( _openFileObjects >= 0 );
			release();
		}
	}

	int noOpenFileObjects = ( _openFileObjects == 0 );

	TheContainerIndex->freeActiveFileObjects(active);

	_syncMutex.releaseWrite();

	if ( noOpenFileObjects ) 
		unreference();

	return 0;
}

#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("MemoryObjectCachingContainer",MemoryObjectCachingContainer,MemoryObjectContainer,0,0);

