#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MemoryObjectContainer.c - 
 */

#include "Assert.h"
#include "BlockAllocator.h"
#include "BitArray.h"
#include "FileObject.h"
#include "FileSystem.h"
#include "FileSystemInterface.h"
#include "FileSystemLocks.h"
#include "FileSystemSimpleInstruments.h"
#include "MemoryObjectCache.h"
#include "MemoryObjectContainer.h"
#include "MemoryObjectDictionary.h"
#include "Raid.h"
#include "Semaphore.h"
#include "ThisProcess.h"

#ifdef __GNUC__
#pragma implementation
#endif

MemoryObjectContainer::MemoryObjectContainer( PersistentMemoryObject * src )
     : PersistentObject( src ), _allocator( 0 ), _mutex( 1 ), 
       _fileObjectLock( this )
{
	initDebug();
	Raid( RaidConstructor )( "%N::MemoryObjectContainer( %N )\n", 
				 this, _source );
	_openFileObjects = 0;
	_modified = 0;
}

MemoryObjectContainer::~MemoryObjectContainer()
{
	Raid( RaidConstructor )( "%N::~MemoryObjectContainer()\n", this );
#ifdef ASSERT
	for (int i = 0; i < _numberOfFileObjects; i++) {
		Assert( lookup( i ) == 0 );
	}
#endif
	Raid( RaidConstructor )( "%N::~MemoryObjectContainer done\n", this );
}

void
MemoryObjectContainer::noRemainingReferences()
{
	_source->close( this );
}

FileObject *
MemoryObjectContainer::basicOpen( int, int & status, int flags )
{
	status = OperationNotSupported;
	return( 0 );
}

int 
MemoryObjectContainer::size()
{
	return _numberOfFileObjects;
}

int
MemoryObjectContainer::rootIdNumber()
{
	return basicRootId();
}

void 
MemoryObjectContainer::acquire()
{
	Raids( 64 ) << this << "::acquire()" << thisProcess() << "\n" << eor;

	_mutex.P();

	Raids( 64 ) << this << "::acquire()" << thisProcess() << " got it\n" 
		<< eor;
}

void
MemoryObjectContainer::release()
{
	Raids( 64 ) << this << "::release()" << thisProcess() << "\n" << eor;
	_mutex.V();
}

void
MemoryObjectContainer::acquire( int idNumber )
{
	Raids( 64 ) << this << "::acquire( " << idNumber << " ) " 
		<< thisProcess() << "\n" << eor;

	_fileObjectLock.acquire( idNumber, idNumber );

	Raids( 64 ) << this << "::acquire( " << idNumber << " ) " 
		<< thisProcess() << " got it\n" << eor;

}

void
MemoryObjectContainer::release( int idNumber )
{
	Raids( 64 ) << this << "::release( " << idNumber << " ) "
		<< thisProcess() << "\n" << eor;

	_fileObjectLock.release( idNumber, idNumber );
}

void
MemoryObjectContainer::parseFlags( FileObject * fo, int flags )
{
	fo->parseFlags( flags );
}

MemoryObjectContainerTimer
MemoryObjectContainerOpen("MemoryObjectContainer::open() timer");

/**
 * Determine if an idNumber is in the range considered valid on the container.
 * @param idNumber ID number to validate
 * @returns true if the ID number is in range; false otherwise.
 */
bool MemoryObjectContainer::IsIdNumberInRange( int idNumber )
{
  if( ( 0 > idNumber) || ( idNumber >= _numberOfFileObjects ) ) {
    return false;
  }

  return true;
}

FileObjectRef
MemoryObjectContainer::open( int idNumber, int & status, int flags )
{
	SInstr( int istart = MemoryObjectContainerOpen.start() );

	Debug << this << "::open( " << idNumber << " )_numberOfFileObjects="
	  <<_numberOfFileObjects << "\n" << eor;

	Raid( RaidMethod )( "%N::open( idNumber %d )\n", this, idNumber );

	if(!this->IsIdNumberInRange(idNumber)) {
	  status = IdNumberOutOfRange;
	  Debug << this << "::open( returns 0, status " << status
	    << " )\n" << eor;
	  SInstr( MemoryObjectContainerOpen.stop( istart ) );
	  return 0;
	}
	
	_syncMutex.acquireRead();

	acquire( idNumber );
	FileObjectRef fo = lookup( idNumber );
	if( fo == 0 ) {
		fo = basicOpen( idNumber, status, flags );
		if( fo != (FileObject*)0 ) {
			bind( idNumber, fo );
			parseFlags( fo, flags );
			release( idNumber );	
			acquire();
			if ( _openFileObjects++ == 0 ) reference();
			release();
		} else {
			Console()<<"What the hell?\n" << eor;
			release( idNumber );	
		}
	} else {
		Assert( fo != (FileObject *) ~0 );
		parseFlags( fo, flags );
		release( idNumber );
		status = 0;
	}

	_syncMutex.releaseRead();

	Raid( RaidMethod )( "%N::open returns %N %d\n", this, fo, status );

	SInstr( MemoryObjectContainerOpen.stop( istart ) );

	return( fo );
}

FileObject *
MemoryObjectContainer::basicCreate( int, Class *, int & status, int flags )
{
	status = OperationNotSupported;
	return( 0 );
}

MemoryObjectContainerTimer
MemoryObjectContainerCreate("MemoryObjectContainer::create() timer");

FileObjectRef
MemoryObjectContainer::create( int hint, Class * aClass, int & status,
			       int flags )
{
	SInstr( int istart = MemoryObjectContainerCreate.start() );

	Raid( RaidMethod )( "%N::create( hint %d, %N )\n", this, hint, aClass);

	_syncMutex.acquireRead();

	FileObjectRef fo = 0;

	fo = basicCreate( hint, aClass, status, flags );
	_modified = 1;

	if( fo != 0 ) {
		int idNumber = fo->idNumber();

		Assert( idNumber < _numberOfFileObjects );

		acquire( idNumber );
#ifdef ASSERT
		FileObject *oldfo = lookup( idNumber );
		Assert( oldfo == 0 || oldfo == (FileObject *) ~0 );
#endif 
		bind( idNumber, fo );
		parseFlags( fo, flags );
		release( idNumber );

		acquire();
		if ( _openFileObjects++ == 0 ) reference();
		release();
	}

	_syncMutex.releaseRead();

	Raid( RaidMethod )( "%N::create returns %N\n", this, fo );

	SInstr( MemoryObjectContainerCreate.stop( istart ) );

	return( fo );
}

MemoryObjectContainerTimer
MemoryObjectContainerClose("MemoryObjectContainer::close() timer");

void
MemoryObjectContainer::close( FileObject * fo )
{
	SInstr( int istart = MemoryObjectContainerClose.start() );

	Raid( RaidMethod )( "%N::close( %N )\n", this, fo );

	_syncMutex.acquireRead();

	int idNumber = fo->idNumber();

	// Assert( ( 0 <= idNumber ) && ( idNumber < _numberOfFileObjects ) );
	// Assert( ((PersistentMemoryObject *) fo)->referenceCount() == 0 );
	// Assert( lookup( idNumber )->identity() == fo->identity() );	
	
	basicClose( fo );
	unbind( idNumber );
		
	release( idNumber );

	acquire();

	--_openFileObjects;
		
	int noOpenFileObjects = ( _openFileObjects == 0 );

	release();

	_syncMutex.releaseRead();

	if( noOpenFileObjects ) unreference();

	SInstr( MemoryObjectContainerClose.stop( istart ) );
}

int
MemoryObjectContainer::basicRootId()
{
	return( 1 );
}

void
MemoryObjectContainer::basicClose( FileObject * fo )
{
	fo->destroy();
}

int
MemoryObjectContainer::allocate( int & blocks, int near, int numberOfBlocks )
{
	Raid( RaidMethod )( "%N::allocate( %d, %d, %d )\n",
			    this, blocks, near, numberOfBlocks );

	if( _allocator != 0 ) {
		acquire();
		_modified = 1;
		int blk = _allocator->allocate( blocks, near, numberOfBlocks );
		release();
		return( blk );
	}
	else return( 0 );
}

void
MemoryObjectContainer::free( int & blocks, int block, int numberOfBlocks )
{
	Raid( RaidMethod )( "%N::free( %d, %d, %d )\n",
			    this, blocks, block, numberOfBlocks );

	if( _allocator != 0 ) {
		acquire();
		_modified = 1;
		_allocator->free( blocks, block, numberOfBlocks );
		release();
	}
}

void
MemoryObjectContainer::basicSynchronize()
{
}

void
MemoryObjectContainer::synchronize()
{
	_syncMutex.acquireWrite();

	FileObject ** active = TheContainerIndex->getActiveFileObjects(this);
	FileObject ** fo;
	for ( fo = active; *fo != 0; fo++ ) {
		Assert( (*fo)->alive() );
		(*fo)->acquireAll();
	}

	_activeFileObjects = active;
	basicSynchronize();
	_activeFileObjects = 0;

	for ( fo = active; *fo != 0; fo++ ) {
		Assert( (*fo)->alive() );
		(*fo)->releaseAll();
	}

	TheContainerIndex->freeActiveFileObjects(active);

	_syncMutex.releaseWrite();
}

void
MemoryObjectContainer::basicSynchronize( int idNumber )
{
}

void
MemoryObjectContainer::synchronize( int idNumber )
{
	basicSynchronize( idNumber );
}

MemoryObjectDictionaryRef
MemoryObjectContainer::rootDictionary()
{
	int status;
	Debug << this << "::rootDictionary( basicRootId "
	  << basicRootId() << " )\n" << eor;
	FileObjectStar fo = open( basicRootId(), status );

	Raid( RaidMethod )( "%N::rootDictionary id %d, %N, status %o\n",
			    this, basicRootId(), fo, status );

	if( fo == 0 ) return( 0 );

	fo->setName( "/" );
	MemoryObjectDictionaryRef modfo= fo->asA( MemoryObjectDictionaryClass, status ); 
	return( modfo );
}

void
MemoryObjectContainer::initializeArrays( int containerSize )
{
	Assert( containerSize >= 0 );

	Raid( RaidProtected )( "%N::initializeArrays( %d )\n",
				this, containerSize );
	_numberOfFileObjects	= containerSize;
	Raid( RaidProtected )( "%N::initializeArrays() done\n", this ); 
}

#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("MemoryObjectContainer",MemoryObjectContainer,PersistentObject,0,0);
