#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-2006
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Contributing Author: Jeffrey C. Carlyle (jcarlyle@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MemoryObjectDictionary.c - abstract base class for UNIX, MSDOS,
 *              Mail, Ar, COFF, and ELF directories.
 */
#include "Assert.h"
#include "Class.h"
#include "Domain.h"
#include "Kernel.h"
#include "FileSystem.h"
#include "FileSystemInterface.h"
#include "FileSystemSimpleInstruments.h"
#include "MemoryObjectContainer.h"
#include "MemoryObjectDictionary.h"
#include "MemoryObjectView.h"
#include "PersistentMemoryObject.h"
#include "Raid.h"
#include "RangeLocks.h"
#include "Semaphore.h"
#include "String.h"

#ifdef __GNUC__
#pragma implementation
#endif

const int AccessUpdate = AccessExecute|AccessWrite;

int MemoryObjectDictionarySynchronizeOff = 0;

/*
 *   MemoryObjectDictionary -   initialize instance variables,
 *          these variables are used to simulate 
 *          memory-mapped directories
 */
#ifdef MULTIMAX
extern unsigned int getKernelStackPointer();
#endif
MemoryObjectDictionaryTimer
MemoryObjectDictionaryCtor("MemoryObjectDictionary::MemoryObjectDictionary() timer");
MemoryObjectDictionary::MemoryObjectDictionary( PersistentMemoryObject * source  )
    : PersistentObject( source ), _container( 0 )
{
    SInstr( int istart = MemoryObjectDictionaryCtor.start() );

    if ( _source ) 
        _container = _source->container();

    initDebug();

    Raid( RaidConstructor )( "%N::MemoryObjectDictionary( %N )\n",
                             this, source );
#ifdef MULTIMAX
    Raids( RaidDetail ) << this << "::ctor(): stack pointer "
                        << asHex(getKernelStackPointer()) << "\n" << eor;
#endif

    _maxKeyLength = 0;
    _buffer = 0;
    _view = 0;
    _log2ViewSize = 0;
    _dirty = 0;

    //Assert( _container != 0 );

    if ( source != 0 ) {

        _unitSize = _source->unitSize();
        _unitMask = _unitSize - 1;
        _size = _source->numberOfBytes();
        _units = _source->numberOfUnits();

    }

    /*
     * This object can now be used.
     */
    _valid = 1;
    _mutex = new MutexLock(this);

    SInstr( MemoryObjectDictionaryCtor.stop( istart ) );
}

/*
 *  ~MemoryObjectDictionary -   since these objects are always kept
 *          synchronized with their _source,
 *          just free up any memory allocated in constructor
 */

MemoryObjectDictionaryTimer 
MemoryObjectDictionaryDtor("MemoryObjectDictionary::~MemoryObjectDictionary() timer");

MemoryObjectDictionary::~MemoryObjectDictionary()
{
    SInstr( int istart = MemoryObjectDictionaryDtor.start() );

    Raid( RaidConstructor )( "%N::~MemoryObjectDictionary()\n", this );
    _container = 0;
    if ( _buffer != 0 ) {
        Assert( _view != 0 );
        const char * where = 
        TheKernel->domain()->remove( _view, _buffer );
        Assert( where == _buffer );
        _buffer = 0;
        _view = 0;
    }
    delete _mutex;

    SInstr( MemoryObjectDictionaryDtor.stop( istart ) );
}

MemoryObjectDictionaryTimer 
MemoryObjectDictionaryReMap("MemoryObjectDictionary::reMap() timer");

void
MemoryObjectDictionary::reMap()
{
    SInstr( int istart = MemoryObjectDictionaryReMap.start() );

    Assert( _source != 0 );
    int sourceLog2UnitSize = _source->log2UnitSize();
    int size = _source->numberOfUnits() << sourceLog2UnitSize;
    Assert( size > 0 );
    int currentLog2Size = _log2ViewSize;
    if ( _log2ViewSize == 0 ) 
        _log2ViewSize = 
        ( sourceLog2UnitSize > 14 ?
          sourceLog2UnitSize : 14 ); /* 16kb */
    while (size > (1 << _log2ViewSize))
        _log2ViewSize++;
    if ( _log2ViewSize != currentLog2Size ) {
        if ( _buffer != 0 ) {
            Assert( _view != 0 );
            const char * where = 
            TheKernel->domain()->remove( _view, _buffer );
            Assert( where == _buffer );
        }
        _view = new MemoryObjectView
        ( _source, 
          1 << ( _log2ViewSize - sourceLog2UnitSize),
          sourceLog2UnitSize, 0 );
        Assert( _view != 0 );
        _buffer = (char *) TheKernel->domain()->add( _view, NoAccess );
        Assert( _buffer != 0 );
    }

    SInstr( MemoryObjectDictionaryReMap.stop( istart ) );
}

/*
 *   open - answer address of PersistentMemoryObject associated
 *      with the given key (0 if unsuccessful)
 *
 *      return status code via reference parameter
 */

MemoryObjectDictionaryTimer 
MemoryObjectDictionaryOpen("MemoryObjectDictionary::open() timer");

PersistentMemoryObjectRef
MemoryObjectDictionary::open( char * key, int & status, int flags )
{
    SInstr( int istart = MemoryObjectDictionaryOpen.start() );

    Raid( RaidMethod )( "%N::open( %s )\n", this, key );

    /*
     * Return value must be initialized.
     */
    PersistentMemoryObjectRef mo = 0;

    _mutex->P();

    if     ( !_valid )               status = DirectoryInvalid;
    else if( !_source->access( AccessExecute ) ) status = FileAccessViolation;
    else if( !findKey( key, 0 ) )        status = KeyNotFound;
    else {
        mo = _container->open( _idNumberOfKey, status, flags );
        if( ( mo != 0 ) &&
            ( strcmp( key, ".") != 0 ) &&
            ( strcmp( key, "..") != 0 ) ) mo->setName( key );
    }

    _mutex->V();

    SInstr( MemoryObjectDictionaryOpen.stop( istart ) );

    return( mo );
}

/*
 *    keys -fill the given buffer with as many of this dictionarys keys
 *      as possible, also answer the number of keys put in the buffer
 *
 *      return status code 
 *      ( >= 0 if successful, otherwise an error code)
 */

MemoryObjectDictionaryTimer 
MemoryObjectDictionaryKeys("MemoryObjectDictionary::keys() timer");

int
MemoryObjectDictionary::keys( char * keyBuffer, int bufferSize, int & count ) 
{
    SInstr( int istart = MemoryObjectDictionaryKeys.start() );

    Raid( RaidMethod )( "%N::keys( buffer %x, bufferSize %d ) _size %d\n",
                        this, keyBuffer, bufferSize, _size );

    int status = 0;

    _mutex->P();

    if     ( !_valid )                status = DirectoryInvalid;
    else if( !_source->access( AccessRead ) ) status = FileAccessViolation;
    else status = basicKeys( keyBuffer, bufferSize, count );

    _mutex->V();

    SInstr( MemoryObjectDictionaryKeys.stop( istart ) );

    return( status );
}

/*
 *    associations -fill the given buffer with as many of this dictionarys
 *          associations as possible; also answer the number of
 *              associations in the buffer
 *
 *      return status code 
 *      ( >= 0 if successful, otherwise an error code)
 */

MemoryObjectDictionaryTimer 
MemoryObjectDictionaryAssociations("MemoryObjectDictionary::associations() timer");

int
MemoryObjectDictionary::associations( char * keyBuffer, int bufferSize,
                                      int & count ) 
{
    SInstr( int istart = MemoryObjectDictionaryAssociations.start() );

    Raid( RaidMethod )( "%N::associations( buf %x, bufSize %d ) _size %d\n",
                        this, keyBuffer, bufferSize, _size );

    int status = 0;

    _mutex->P();

    if     ( !_valid )                status = DirectoryInvalid;
    else if( !_source->access( AccessRead ) ) status = FileAccessViolation;
    else status = basicAssociations( keyBuffer, bufferSize, count );

    _mutex->V();

    SInstr( MemoryObjectDictionaryAssociations.stop( istart ) );

    return( status );
}

MemoryObjectDictionaryTimer 
MemoryObjectDictionaryAssociationsOffset("MemoryObjectDictionary::associations(offset) timer");

int
MemoryObjectDictionary::associations( char * keyBuffer, int bufferSize,
                                      int & count, int & offset ) 
{
    SInstr( int istart = MemoryObjectDictionaryAssociationsOffset.start() );

    Raid( RaidMethod )( "%N::associations( buf %x, bufSize %d ) _size %d\n",
                        this, keyBuffer, bufferSize, _size );

    int status = 0;

    _mutex->P();

    if     ( !_valid )                status = DirectoryInvalid;
    else if( !_source->access( AccessRead ) ) status = FileAccessViolation;
    else status = basicAssociations( keyBuffer, bufferSize, count, offset );

    _mutex->V();

    SInstr( MemoryObjectDictionaryAssociationsOffset.stop( istart ) );

    return( status );
}

/*
 *     create - answer address of PersistentMemoryObejct
 *      associated with the given key ( 0 if unsuccessful)
 *
 *              If the key-PersistentMemoryObject pair was found,
 *              create a PersistentMemoryObject in this dictionarys container,
 *              and add the new key-PersistentMemoryObject pair.
 *
 *      return status code via reference parameter
 */

MemoryObjectDictionaryTimer 
MemoryObjectDictionaryCreate("MemoryObjectDictionary::create() timer");

PersistentMemoryObjectRef
MemoryObjectDictionary::create( char * key, Class * aClass, int & status,
                                int flags )
{
    SInstr( int istart = MemoryObjectDictionaryCreate.start() );

    Raid( RaidMethod )( "%N::create( %s, %N )\n", this, key, aClass );

    PersistentMemoryObjectRef mo = 0;

    int keyLength = strlen( key );

    _mutex->P();

    if( !_valid )                status = DirectoryInvalid;
    else if( !_source->access( AccessExecute ) ) status = FileAccessViolation;
    else if( keyLength <= 0 )            status = KeyTooSmall;
    else if( keyLength > _maxKeyLength )     status = KeyTooBig;
    else if( findKey( key, entrySize( keyLength ) ) ) {
        mo = _container->open( _idNumberOfKey, status, flags );
        if( mo != 0 ) {
            ProxiableObjectStar po = mo->asA( aClass, status );
            if( po != 0 ) {
                mo->setName( key );
                status = KeyExists;
            }
            else mo = 0;
        }
    }
    else if( !_source->access( AccessWrite ) )   status = FileAccessViolation;
    else {
        mo = _container->create( hint(), aClass, status, flags );
        if( mo != 0 ) {
            ProxiableObjectStar po = mo->asA( aClass, status );
            if( po != 0 ) {
                mo->setName( key );
                status = insertEntry( key, keyLength, mo->idNumber() );
                if( status == 0 ) {
                    mo->links( 1 );
                    if( aClass->isASubclassOf( MemoryObjectDictionaryClass ) ) {
                        MemoryObjectDictionaryStar od = 
                        mo->asA( MemoryObjectDictionaryClass, status );
                        if( od != 0 ) {
                            status = od->add( "..", _source );
                        }
                    }
                    mo->setCreatingDictionary( this );
                }
                else mo = 0;
            }
            else mo = 0;
        }
    }

    _mutex->V();

    Raid( RaidMethod )( "%N::create() returns %N\n", this, mo );

    SInstr( MemoryObjectDictionaryCreate.stop( istart ) );

    return( mo );
}

/*
 *    add - Add the given key-PersistentMemoryObject pair to the dictionary,
 *      Do this only if the key is not already in this dictionary.
 *
 *      return status code 
 *      ( >= 0 if successful, otherwise an error code)
 */

MemoryObjectDictionaryTimer 
MemoryObjectDictionaryAdd("MemoryObjectDictionary::add() timer");

int
MemoryObjectDictionary::add( char * key, PersistentMemoryObject * mo )
{
    SInstr( int istart = MemoryObjectDictionaryAdd.start() );

    Raid( RaidMethod )( "%N::add( key %s, %N )\n", this, key, mo );

    int status = 0;
    int keyLength = strlen( key );
    int entryLength = entrySize( keyLength );

    _mutex->P();

    if     ( !_valid )              status = DirectoryInvalid;
    else if( keyLength <= 0 )           status = KeyTooSmall;
    else if( keyLength > _maxKeyLength )    status = KeyTooBig;
    else if( findKey( key, entryLength ) )  status = KeyExists;
    else status = insertEntry( key, keyLength, mo->idNumber() );
    if( status == 0 ) mo->links( 1 );

    _mutex->V();

    SInstr( MemoryObjectDictionaryAdd.stop( istart ) );

    return( status );
}

/*
 *     remove - Remove the key-PersistentMemoryObject pair corresponding
 *      to the given key.
 *      Also decrement the link count for the PersistentMemoryObject.
 *
 *      return status code 
 *      ( >= 0 if successful, otherwise an error code)
 */

MemoryObjectDictionaryTimer 
MemoryObjectDictionaryRemove("MemoryObjectDictionary::remove() timer");

int
MemoryObjectDictionary::remove( char * key, int * links )
{
    SInstr( int istart = MemoryObjectDictionaryRemove.start() );

    Raid( RaidMethod )( "%N::remove( %s )\n", this, key );

    int status = 0;

    _mutex->P();

    if     ( !_valid )              status = DirectoryInvalid;
    else if( !_source->access( AccessUpdate ) ) status = FileAccessViolation;
    else if( !findKey( key, 0 ) )       status = KeyNotFound;
    else {
        PersistentMemoryObjectStar mo = 
        _container->open( _idNumberOfKey, status, O_NCACHE );

        if( mo != 0 ) {
            int rm = ( mo->links( 0 ) == 1 );
            MemoryObjectDictionaryStar od =
            mo->asA( MemoryObjectDictionaryClass, status );
            if( od != 0 ) {

                if( od->identity() == identity() ) status = DirectoryBusy;
                else {
                    od->_mutex->P();
                    int foundDot = od->findKey( ".", 0 );
                    od->_mutex->V();
                    if ( foundDot ) rm = ( mo->links( 0 ) == 2 );
                    status = od->empty();
                }

            }
            else status = 0;

            if( status == 0 ) {
                status = clearEntry();
                if( status == 0 ) {
                    mo->links( -1 );
                    if( od != 0 && rm ) {
                        mo->links( -1 );
                        _source->links( -1 );
                    }
                    if ( links )
                        *links = mo->links( 0 );
                    Assert( !rm || mo->links( 0 ) == 0 );
                }
            }
        }
    }

    _mutex->V();

    Raid( RaidMethod )( "%N::remove return %d\n", this, status );

    SInstr( MemoryObjectDictionaryRemove.stop( istart ) );

    return( status );
}

/*
 * empty: returns true if directory has no keys
 */

MemoryObjectDictionaryTimer 
MemoryObjectDictionaryEmpty("MemoryObjectDictionary::empty() timer");

int
MemoryObjectDictionary::empty()
{
    SInstr( int istart = MemoryObjectDictionaryEmpty.start() );

    Raid( RaidProtected )( "%N::empty()\n", this );

    int status = 0;

    _mutex->P();

    int count = 0;
    basicKeys( 0, _size, count );

    if( !_valid )       status = DirectoryInvalid;
    else if( count > 2 )    status = DirectoryHasKeys;
    else _valid = 0;

    _mutex->V();

    Raid( RaidProtected )( "%N::empty returns %d\n", this, status );

    SInstr( MemoryObjectDictionaryEmpty.stop( istart ) );

    return( status );
}

/*
 * noRemainingReferences - Decrement reference count
 *     that is maintained by the underlying PersistentMemoryObject.
 */
void
MemoryObjectDictionary::noRemainingReferences()
{
    _source->close( this );
}

/*
 *     findKey -    Search for given key
 *      (also look for empty space if second parameter is non-zero).
 *
 *      Return 1 if key is found, otherwise return 0.
 *
 *      Expected side-effects:
 *          _offset = location of found key
 *          _emptyOffset = location of space to add a key-value pair
 *                  (-1 means no room).
 *          _idNumberOfKey = value associated with key
 */
int
MemoryObjectDictionary::findKey( char *, int )
{
    _offset = 0;
    _emptyOffset = -1;
    _idNumberOfKey = -1;
    return( 0 );
}

/*
 *     insertEntry -    Update dictionary with the given key and value.
 *      (key length is supplied as the second parameter).
 *
 *      NOTE: requires findKey() to have just been called
 *      within same critical section!
 *
 *      Return status code indicating success or failure of writing
 *      changes back to the underlying PersistentMemoryObject(_source).
 */
int
MemoryObjectDictionary::insertEntry( char *, int, int )
{
    return( OperationNotSupported );
}

/*
 *   hint - generate a value to help a container decide which inumber
 *      to use for a newly created object.
 *      (For SVID this is useless, for BSD this increases efficiency,
 *       for MSDOS this is absolutely necessary.
 *       i.e., this smells, looks, and quacks like a hack to me.)
 *
 *      NOTE: requires findKey() to have just been called
 *      within same critical section!
 */
int
MemoryObjectDictionary::hint()
{
    return( 0 );
}

/*
 *  entrySize - return the size of a dictionary given a key length.
 */
int
MemoryObjectDictionary::entrySize( int )
{
    return( 0 );
}

/*
 * clearEntry - remove a key-value pair from a dictionary
 *
 *      NOTE: requires findKey() to have just been called
 *      within same critical section!)
 */
int
MemoryObjectDictionary::clearEntry()
{
    return( OperationNotSupported );
}

/*
 *  basicKeys - fill the given buffer with as many of this dictionarys keys
 *      as possible, also answer the number of keys put in the buffer
 *
 *      NOTE: This does all the work for the public keys() method.
 *      It does not check directory validity, permissions, and
 *      assumes a critical section has already been entered.
 *
 *      return status code 
 *      ( >= 0 if successful, otherwise an error code)
 */
int
MemoryObjectDictionary::basicKeys( char *, int, int & )
{
    Assert( NOTREACHED );
    return( 0 );
}

/*
 *  basicAssociations - fill the given buffer with as many of this dictionarys
 *          assocations as possible; also answer the number of keys put
 *      in the buffer
 *
 *      NOTE: This does all the work for the public associations()
 *      method. It does not check directory validity, permissions,
 *      and assumes a critical section has already been entered.
 *
 *      return status code 
 *      ( >= 0 if successful, otherwise an error code)
 */
int
MemoryObjectDictionary::basicAssociations( char *, int, int & )
{ 
    Assert( NOTREACHED );
    return( 0 );
}

int
MemoryObjectDictionary::basicAssociations( char *, int, int &, int & )
{ 
    Assert( NOTREACHED );
    return( 0 );
}

/*
 * writeChanges - record dirty info
 *
 */

int
MemoryObjectDictionary::writeChanges()
{
    _dirty = 1;
    return( 0 );
}

/*
 * synchronize - if dirty, then write out
 *
 */

int
MemoryObjectDictionary::synchronize()
{
    if ( !MemoryObjectDictionarySynchronizeOff && _dirty ) {
        _dirty = 0;
        _source->synchronize();
    }
    return( 0 );
}

int
MemoryObjectDictionary::keySize()
{
    return( _size );
}

int
MemoryObjectDictionary::sharesContainer( PersistentMemoryObject * mo )
{
    int result = 0;
    {
        MemoryObjectContainerStar oc = mo->container();
        result = ( _container->identity() == oc->identity()  );
    }
    return( result );
}

#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("MemoryObjectDictionary",MemoryObjectDictionary,PersistentObject,0,0);
