#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MountTable.cc - mount table
 */

#include "MountTable.h"
#include "MountItem.h"
#include "FileSystem.h"
#include "FileSystemSimpleInstruments.h"
#include "PersistentMemoryObject.h"
#include "MemoryObjectDictionary.h"
#include "MemoryObjectContainer.h"
#include "Raid.h"

#ifdef __GNUC__
#pragma implementation
#endif

MountTable::MountTable()
{
	initDebug();

	Raid( RaidConstructor )( "%N::MountTable( )\n", this );

	_mountItems = 0;
}

MountTable::MountTable( MountTable * table )
{
    initDebug();

    Raid( RaidConstructor )( "%N::MountTable( %N )\n", this, table );

    _mountItems = 0;
    MountItem * next = table->_mountItems;
    while( next != 0 ) {
	PersistentMemoryObjectRef mp = next->mountPoint();
	PersistentMemoryObjectRef m  = next->mounted();
	addMount( mp, m );
	next = next->next();
    }
}

MountTable::~MountTable()
{
    Raid( RaidConstructor )( "%N::~MountTable( )\n", this );

    while( _mountItems != 0 ) {
	MountItem * next = _mountItems->next();
	delete _mountItems;
	_mountItems = next;
    }
}

int
MountTable::synchronize()
{
    Raid( RaidConstructor )( "%N::synchronize()\n", this );

    MountItem * mi = _mountItems;
    while( mi != 0 ) {
	MemoryObjectContainerStar oc = mi->mounted()->container();
	if( oc != 0 ) oc->synchronize();
	mi = mi->next();
    }
    return( 0 );
}

int
MountTable::addMount( PersistentMemoryObjectRef mp,
		      PersistentMemoryObjectRef md )
{
	Raid( RaidProtected )( "%N::addMount( %N under %N )\n", this, mp, md );
	if( _mountItems != 0 ) {
		if( _mountItems->lookupMounted( mp ) != 0 ) {
			Raid( RaidDetail )( "%N::addMount busy %N\n", this, mp );
			return( MountPointBusy );
		}
	}
	_mountItems = new MountItem( _mountItems, mp, md );
	Raid( RaidProtected )( "%N::addMount done\n", this );
	return( 0 );
}

int
MountTable::deleteMount( PersistentMemoryObject * mo )
{
	Raid( RaidProtected )( "%N::deleteMount( %N )\n", this, mo );
	if( _mountItems != 0 ) {
		MountItem * mi = _mountItems->lookupMounted( mo );
		if( mi != 0 ) {
			int status = mo->unmount();
			if ( status != 0 ) 
				return( status );
			if( mi->identity() == _mountItems->identity() ) {
				_mountItems = _mountItems->next();
			}
			delete mi;
			return( 0 );
		}
	}
	return( ObjectNotMounted );
}

MountTableTimer MountTableSubstitute("MountTable::substitute() timer");

void
MountTable::substitute( PersistentMemoryObjectRef  mo )
{
        SInstr( int istart = MountTableSubstitute.start() );

	Raid( RaidProtected )( "%N::substitute( %N )\n", this, mo );

	if( ( _mountItems != 0 ) && ( mo != 0 ) ){
		MountItem * mi = _mountItems->lookupMountPoint( mo );
		if( mi != 0 ) {
			mo = mi->mounted();
		}
	}

	Raid( RaidProtected )( "%N::substitute = %N\n", this, mo );

        SInstr( MountTableSubstitute.stop( istart ) );
}

MountTableTimer MountTableSubstituteUp("MountTable::substituteUp() timer");

void
MountTable::substituteUp( MemoryObjectDictionaryRef od, char * key )
{
        SInstr( int istart = MountTableSubstituteUp.start() );

	Raid( RaidProtected )( "%N::substituteUp( %N )\n", this, od );

	if( ( key[0] == '.' ) &&
	    ( key[1] == '.' ) &&
	    ( ( key[2] == '/' ) || ( key[2] == '\0' ) ) &&
	    ( _mountItems != 0 ) && ( od != 0 ) ){
		int status = 0;
		PersistentMemoryObjectStar mo = od->asA(
				 	PersistentMemoryObjectClass, status );
		MountItem * mi = _mountItems->lookupMounted( mo );
		if( mi != 0 ) {
			int status = 0;
			mo = mi->mountPoint();
			od = mo->asA( MemoryObjectDictionaryClass, status );
			Assert( od );
		}
	}

	Raid( RaidProtected )( "%N::substituteUp = %N\n", this, od );

        SInstr( MountTableSubstituteUp.stop( istart ) );
}



#include "ClassInstaller.h"
DEFINECLASS("MountTable",MountTable,ProxiableObject,0,0);
