#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * POSHContainer.cc - 
 */
#include "Assert.h"
#include "Raid.h"
#include "POSHContainer.h"
#include "POSHFileNode.h"
#include "POSHBlockAllocator.h"
#include "FileSystem.h"
#include "POSHFileSystem.h"
#include "Class.h"
#include "NameServer.h"
#include "String.h"

#ifdef __GNUC__
#pragma implementation
#endif

inline int
POSHContainer::mapNodeNumber( int node )
{
        
        int block = node / _superBlock->_nodesPerBlock;
        return( _superBlock->_nodeStart + block );
}

inline int
POSHContainer::firstNodeNumber( int node )
{
        int mask = ~( _superBlock->_nodesPerBlock - 1 );
        return( node & mask );
}

/*
 *	Read an disk inode into memory, if it isnt there already.
 *	( Always read an enclosing block of disk inodes. )
 */
void
POSHContainer::readNode( int nodeNumber )
{
    Raid( RaidMethod )( "%N::readNode( %d )\n", this, nodeNumber );

    //  Check to make sure it isnt in memory already.
    if( _nodes[nodeNumber] == 0 ) {

        // Allocate a block-sized buffer for the inodes.
        char * buffer = new char [_blockSize];

        // Calculate the location of inodes and read them from the disk.
        int physicalBlock = mapNodeNumber( nodeNumber ) << _log2BlockToSector;
        _source->read( physicalBlock, 1 << _log2BlockToSector, buffer, 
					_blockSize );

        // Set up pointers for all inodes just read in.
        int firstNode = firstNodeNumber( nodeNumber );
        for( int i = 0; i < _superBlock->_nodesPerBlock; i++ ) {
            _nodes[firstNode++] = (POSHNode *)buffer;
            buffer += _superBlock->_nodeSize;
        }
    }

    Raid( RaidMethod )( "%N::readNode done %d\n", this, nodeNumber );
}

/*
 *	Write this inode to the disk.
 *	If there are no more current references to inodes
 *	in its enclosing block of memory, then free it.
 *	( Always write the enclosing block of disk inodes. )
 */
void
POSHContainer::writeNode( int nodeNumber )
{
    Raid( RaidMethod )( "%N::writeNode( %d ) node = %x\n",
		this, nodeNumber, _nodes[nodeNumber] );

    Assert( _nodes[nodeNumber] != 0 );

    int firstNode = firstNodeNumber( nodeNumber );
    int physicalBlock = mapNodeNumber( nodeNumber ) << _log2BlockToSector;
    char * buffer = (char *)_nodes[firstNode];
    {
       _source->write( physicalBlock, 1 << _log2BlockToSector, buffer, 
					_blockSize );
    }

    // Determine if any of the disk inodes in the block are still in use.
    for( int i = 0; i < _superBlock->_nodesPerBlock; i++ ) 
        if( _fileObjects[firstNode++] != 0 ) return;

    // Reset the pointers and free the block if none were in use.
    firstNode = firstNodeNumber( nodeNumber );
    for( i = 0; i < _superBlock->_nodesPerBlock; i++ ){
	_nodes[firstNode++] = 0;
    }

    Raid( RaidMethod )( "%N::writeNode() delete buffer %x\n", this, buffer );

    delete buffer;
}

POSHContainer::POSHContainer( PersistentMemoryObject * source )
: ( source )
{
    initDebug();
    Raid( RaidConstructor )( "%N::POSHContainer( %N )\n", this, source );

    int unitSize = _source->unitSize();
    Assert( unitSize >= sizeof( POSHSuperBlock ) );
    Assert( unitSize <= POSHSuperBlockOffset );

    /*
     *	Create super block buffer, chainize it, read it from the disk,
     *	and check its magic number.
     */
    _superBlock = (POSHSuperBlock *) new char[unitSize];

    _source->read( POSHSuperBlockOffset / unitSize, 1, (char *)_superBlock,
					unitSize );
    Assert( _superBlock->_magicNumber == POSHMagic );

    _blockSize = _superBlock->_blockSize;
    _log2BlockToSector = _superBlock->_log2BlockSize -
			 _source->log2UnitSize();

    /*
     *	Allocate and clear in memory and on disk inode pointers.
     *	Create one more object than there are inodes because object
     *	0 will be unused since nodeNumber start at 1.
     */
    int containerSize = _superBlock->_nodes;
    MemoryObjectContainer::initializeArrays( containerSize );

    _nodes = new POSHNode * [containerSize];
    for( int i = 0; i < containerSize; i++ ) _nodes[i] = 0;

    int startSector = _superBlock->_nodeMapStart << _log2BlockToSector; 
    int sectors = ( _superBlock->_blockMapStart - _superBlock->_nodeMapStart )
		  << _log2BlockToSector;

    _nodeMap  = new unsigned char[sectors * unitSize];

    _source->read( startSector, sectors, (char *)_nodeMap, sectors * unitSize );

    startSector = _superBlock->_blockMapStart << _log2BlockToSector; 
    sectors = ( _superBlock->_nodeStart - _superBlock->_blockMapStart )
	      << _log2BlockToSector;

    _blockMap = new unsigned char[sectors * unitSize];

    _source->read( startSector, sectors, (char *)_blockMap, sectors * unitSize );

    startSector = _superBlock->_classNameStart << _log2BlockToSector; 
    sectors = ( _superBlock->_nodeMapStart - _superBlock->_classNameStart )
	      << _log2BlockToSector;

    _classNames = new char[sectors * unitSize];

    _source->read( startSector, sectors, _classNames, sectors * unitSize );

    _classes = sectors * unitSize / 4;
    _classNameList = new char * [ _classes ];
    _classList = new ClassStar [ _classes ];
    for( i = 0; i < _classes; i++ ) {
	_classNameList[i] = 0;
	_classList[i] = 0;
    }
    _classes = 1;
    for( i = 0; ; ) {
	_classNameList[_classes] = &_classNames[i];
	_classList[_classes] = StandardNameServer->lookup( _classNameList[_classes], "Class" );
	_classes++;
    	while( _classNames[i++] != '\0' );
	if( _classNames[i] == '\0' ) break;
    }
    _classNameList[_classes] = &_classNames[i];

    _allocator = new POSHBlockAllocator( _source, _superBlock,
					_blockMap, _log2BlockToSector );

    Raid( RaidConstructor )( "%N::POSHContainer done\n", this );
}

POSHContainer::~POSHContainer()
{
    Raid( RaidConstructor )( "%N::~POSHContainer\n", this );

    synchronize();

    /*
     *	Delete everything allocated in the constructor.
     */
    for( int i = 0; i < _numberOfFileObjects;
	i += _superBlock->_nodesPerBlock ) {
        delete _nodes[i];
    }
    delete _superBlock;
    delete _nodeMap;
    delete _blockMap;
    for( i = 0; i < _classes; i++ ) _classList[i] = 0;
    delete _classList;
    delete _classNameList;
    delete _classNames;
}

/*
 *	Find an unused inode.
 *	Returns the nodeNumber of the inode found, or zero if none found.
 *	The inode is marked in use by this function.
 */
int
POSHContainer::getUnusedNode( int )
{
    Raid( RaidMethod )( "POSHContainer::getUnusedNode()\n" );

    int nodeNumber = 0;
    if( _superBlock->_freeNodes != 0 ) {

	for( int i = 0; i < _superBlock->_nodes; i++ ) {
		if( _nodeMap[i] != 255 ) break;
	}
	Assert( i < _superBlock->_nodes );

	for( int b = 0; b < 8; b++ ) {
		if( ( _nodeMap[i] & ( 1 << b ) ) == 0 ) {
			_nodeMap[i] |= ( 1 << b );
			break;
		}
	}
	Assert( b < 8 );

	nodeNumber = ( i << 3 ) + b;

	_superBlock->_freeNodes--;
	_superBlock->_modified = 1;
	_superBlock->_timeModified = 0;
    }

    Raid( RaidMethod )( "POSHContainer::getUnusedNode, return = %d\n",
			nodeNumber );
    return( nodeNumber );
}

/*
 *	Open returns a pointer to a sub-memory-object ( inode )
 *	with the given number.
 *	If the nodeNumber is negative, an unused one is allocated an initialized.
 *	Also, the new nodeNumber is passed back thru the reference parameter.
 *	If the inode is already in use,
 *	update its refernce count.
 *	Otherwise create a new in memory inode and return a pointer to it.
 *	Creating a new memory inode will also read in the disk inode if needed.
 */
FileObject *
POSHContainer::basicOpen( int nodeNumber, int & status, int flags )
{
    Raid( RaidMethod )( "%N::basicOpen( nodeNumber %d )\n", this, nodeNumber );

    if( nodeNumber < basicRootId() ) {
	status = IdNumberOutOfRange;
	return( 0 );
    }

    /*
     *	Read this inode from the disk,
     *	Check to see if it exists,
     *	and Construct its in memory structure.
     */
    acquire();
    readNode( nodeNumber );
    release();
    POSHNode * thisNode = (POSHNode *) _nodes[nodeNumber];

    if( thisNode->_persistentObjectClass == 0 ) {
	status = UnusedIdNumber;
	return( 0 );
    }

    status = 0;
    return( new POSHFileNode( roundup( thisNode->_size, _blockSize ) ,
			    _log2BlockToSector + _source->log2UnitSize(),
			    nodeNumber, this, _source, thisNode,
			    _classList[thisNode->_persistentObjectClass] ) );
}

/*
 *	create returns a pointer to a new sub-memory-object ( inode )
 *	The nodeNumber passed in is just used as a hint to the file system.
 *	Also, the new nodeNumber is passed back thru the same parameter.
 *	The mode parameter is primarily used to set the type of the
 *	file, it should not change during the life of the file.
 *	Uid, gid, and file permissions can be set via methods on an Node.
 */
FileObject *
POSHContainer::basicCreate( int nodeNumber, Class * aClass, int & status,
			    int flags )
{
    Raid( RaidMethod )( "%N::basicCreate( nodeNumber %d, %N, )\n",
			this, nodeNumber, aClass );

    if( !aClass->isASubclassOf( PersistentObjectClass ) ) {
	status = ProtocolMismatch;
	return( 0 );
    }
    acquire();
    nodeNumber = getUnusedNode( 0 );
    if( nodeNumber <= 0 ) {
	status = ContainerFull;
	release();
	return( 0 );
    }

    /*
     *	Read the inode from disk, and reset all of its fields.
     */
    readNode( nodeNumber );
    POSHNode * thisNode = (POSHNode *) _nodes[nodeNumber];

    for( int i = 1; i < _classes; i ++ ) {
	if( _classList[i]->identity() == aClass->identity() ) break;
	if( !aClass->concrete() &&
	    _classList[i]->isASubclassOf( aClass ) ) break;
    }
    if( i == _classes ) {
	strcpy( _classNameList[_classes], aClass->name() );
	_classList[_classes] = aClass;
	_classes++;
	_classNameList[_classes] = _classNameList[i] + strlen( aClass->name() )
				   + 1;
    }
    thisNode->_persistentObjectClass = i;
    thisNode->_size = 0;
    thisNode->_block = 0;

    release();
    /*
     *	Create the new in memory inode.
     */
    status = 0;
    return( new POSHFileNode( roundup( thisNode->_size, _blockSize ) ,
			    _log2BlockToSector + _source->log2UnitSize(),
			    nodeNumber, this, _source, thisNode,
			    _classList[thisNode->_persistentObjectClass] ) );
}

void
POSHContainer::basicClose( FileObject * thisNode )
{
    Raid( RaidMethod )( "%N::basicClose( %N )\n", this, thisNode );

    int nodeNumber = thisNode->idNumber();

    MemoryObjectContainer::basicClose( thisNode );

    /*
     *	Write the disk inode back to disk.
     */
    acquire();
    writeNode( nodeNumber );
    release();
}

/*
 *	Synchronize the in-core data structures with the disk.
 */
void
POSHContainer::basicSynchronize()
{
    Raid( RaidMethod )( "%N::basicSynchronize() %d\n", this, _superBlock->_modified );

    // exit function if superblock has NOT been modified.
    if( !_superBlock->_modified ) return;

    _superBlock->_modified = 0;

    int unitSize = _source->unitSize();


    _source->write( POSHSuperBlockOffset / unitSize, 1, (char *)_superBlock,
					unitSize );

    int startSector = _superBlock->_nodeMapStart << _log2BlockToSector; 
    int sectors = ( _superBlock->_blockMapStart - _superBlock->_nodeMapStart )
		  << _log2BlockToSector;


    _source->write( startSector, sectors, (char *)_nodeMap, sectors * unitSize );

    startSector = _superBlock->_blockMapStart << _log2BlockToSector; 
    sectors = ( _superBlock->_nodeStart - _superBlock->_blockMapStart )
	      << _log2BlockToSector;


    _source->write( startSector, sectors, (char *)_blockMap, sectors * unitSize );

    startSector = _superBlock->_classNameStart << _log2BlockToSector; 
    sectors = ( _superBlock->_nodeMapStart - _superBlock->_classNameStart )
	      << _log2BlockToSector;


    _source->write( startSector, sectors, _classNames, sectors * unitSize );

    // Write out any in memory inodes.
    for( int j = 0; j < _numberOfFileObjects; j++ ) {
	if( _fileObjects[j] != 0 ) {
	    MemoryObjectCache * cache = _fileObjects[j].cache();
	    if ( cache ) 
	        cache->synchronize();
	}
    }
    for( j = 0; j < _numberOfFileObjects; j++ ) {
	if( _fileObjects[j] != 0 ) {
	    writeDinode( j, 0 );
	}
    }
}

ProxiableObject *
POSHContainerConstructor( ProxiableObject * source )
{
        if( source->isKindOf( PersistentMemoryObjectClass ) ) {
                return( new POSHContainer( (PersistentMemoryObject *)source ) );
        }
        else return( 0 );
}

#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("POSHContainer",POSHContainer,MemoryObjectContainer,&POSHContainerConstructor,0);
