#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * POSHFileNode.c - 
 */
#include "Assert.h"
#include "POSHFileNode.h"
#include "FileSystem.h"
#include "UNIXFile.h"
#include "MemoryObjectContainer.h"
#include "POSHFileSystem.h"
#include "Class.h"
#include "Raid.h"

#ifdef __GNUC__
#pragma implementation
#endif

/* 
 * Initialize the in-memory inode
 */
POSHFileNode::POSHFileNode( unsigned int numOfUnits, int log2UnitSize,
			    int inumber, MemoryObjectContainer * container,
			    PersistentMemoryObject * source,
			    POSHNode * madNode, Class * aClass )
: ( numOfUnits, log2UnitSize, inumber, container, source, 0 )
{
    initDebug();
    Raid( RaidConstructor )( "%N::POSHFileNode()\n", this );

    /*
     *	Initialize variables from parameters
     */
    _madNode = madNode;
    _class = aClass;

    Raid( RaidDetail )( "%N::POSHFileNode size %d units %d\n",
			this, _madNode->_size, numberOfUnits() );
}

Class *
POSHFileNode::supports( Class * aClass )
{
    Raid( RaidProtected )( "%N::supports( %N ) %N\n", this, aClass, _class );

    Class * supportedClass = 0;

    if( _class->isASubclassOf( aClass ) ) supportedClass = _class;

    Raid( RaidProtected )( "%N::supports() returns %N\n",
                           this, supportedClass );

    return( supportedClass );
}

/*
 * Clean up the in memory inode data
 */
POSHFileNode::~POSHFileNode()
{
    Raid( RaidConstructor )( "%N::~POSHFileNode()\n", this );
}

int
POSHFileNode::mapUnit( int logicalBlock, int addMapping )
{
    Raid( RaidProtected )( "%N::mapUnit %d %d\n",
			   this, logicalBlock, addMapping);
    Assert( logicalBlock < 1 );
    Assert( addMapping == 0 || addMapping == 1 );

    _mutex.P();

    if( _madNode->_block == 0 ) {
	if( addMapping ) {
	    _madNode->_block = _container->allocate( 0, 0, 1 );
	}
    }

    _mutex.V();

    Raid( RaidProtected )( "%N::mapUnit = %d\n", this, _madNode->_block );
    return( _madNode->_block );
}

int
POSHFileNode::numberOfBytes()
{
    Raid( RaidMethod )( "%N::numberOfBytes %d\n", this, _madNode->_size );
    return( _madNode->_size );
}

int
POSHFileNode::basicSetNumberOfBytes( int size )
{
    Raid( RaidProtected )( "%N::basicSetNumberOfBytes( %d ) %d\n",
		     	   this, size, _madNode->_size );
    if( size < 0 ) size = 0;
    _madNode->_size = size;
    basicSetNumberOfUnits( ( size + unitSize() - 1 ) >> log2UnitSize() );
    return( size );
}

int
POSHFileNode::links( int delta )
{
    Raid( RaidMethod )( "%N::links( delta %d )\n", this, delta );
    return( 0 );
}

int
POSHFileNode::isContainer()
{
    return( _class->isASubclassOf( MemoryObjectContainerClass ) );
}

int 
POSHFileNode::isDictionary()
{
    return( _class->isASubclassOf( MemoryObjectDictionaryClass ) );
}

int 
POSHFileNode::isSymbolicLink()
{
    return( _class->isASubclassOf( MemoryObjectSymbolicClass ) );
}

/* 
 * This does the UNIX Stat( 2 ) function.
 */
void
POSHFileNode::info( UNIXFileInfo * infoBuffer )
{
    Raid( RaidMethod )( "%N::info( %x ) %d\n", this, infoBuffer, _idNumber );
    infoBuffer->device	= 0;
    infoBuffer->inumber	= _idNumber;

    extern ClassStar MemoryObjectDictionaryClass;
    extern ClassStar SymbolicLinkClass;

    if( _class->isASubclassOf( MemoryObjectDictionaryClass ) ) {
	infoBuffer->mode = 040000;
    }
    else if( _class->isASubclassOf( SymbolicLinkClass ) ) {
	infoBuffer->mode = 0120000;
    }
    else {
	infoBuffer->mode = 0100000;//FIX
    }
    infoBuffer->links	= 1;
    infoBuffer->uid	= 0;
    infoBuffer->gid	= 0;
    infoBuffer->size	= _madNode->_size;
    infoBuffer->atime	= 0;
    infoBuffer->mtime	= 0;
    infoBuffer->ctime	= 0;
    infoBuffer->blocks	= ( _madNode->_size > 0 ) ? 1 : 0;
    infoBuffer->blockSize = 0; //FIX
    infoBuffer->flags	= 0;
    infoBuffer->rdevice	= 0;
}

int
POSHFileNode::basicSetNumberOfUnits( unsigned units )
{
    Raid( RaidProtected )( "%N::basicSetNumberOfUnits( units %d ) %d\n",
			   this, units, numberOfUnits() );

    if( units == numberOfUnits() ) return( units );
    setModified();
    if( units < 1 ) {
	_mutex.P();
	if( _madNode->_block != 0 ) _container->free( 0, _madNode->_block, 1 );
        _mutex.V();
    }
    MemoryObject::basicSetNumberOfUnits( units );
    return( units );
}

int
POSHFileNode::accessMode()
{
    return( UNIXAccessMask );
}

#include "ClassInstaller.h"
DEFINECLASS("POSHFileNode",POSHFileNode,UNIXInode,0,0);
