#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * PersistentArray.c - 
 */

#include "Assert.h"
#include "Byte.h"
#include "Convert.h"
#include "Domain.h"
#include "FileSystem.h"
#include "Kernel.h"
#include "MemoryObjectContainer.h"
#include "MemoryObjectView.h"
#include "NotifyEvents.h"
#include "PersistentArray.h"
#include "PersistentMemoryObject.h"
#include "Raid.h"
#include "ReplacementPolicy.h"

#ifdef __GNUC__
#pragma implementation
#endif

class FileSystemInterface;

unsigned PersistentArrayMaxMMapSize = 1 << 20;

PersistentArray::PersistentArray( PersistentMemoryObject * source )
: PersistentObject( source )
{
	initDebug();

	Raid( RaidConstructor )( "%N::PersistentArray( %N )\n", this, source );
	_log2UnitSize = _source->log2UnitSize();
	_unitSize =  1 << _log2UnitSize;
    	_unitSizeMask = _unitSize - 1;
	_address = 0;
	_log2ViewSize = 0;
	_buffer = 0;
}

PersistentArray::~PersistentArray()
{
	Raid( RaidConstructor )( "%N::~PersistentArray\n", this );
	if ( _address ) 
		delete _address;
	if ( _buffer ) {
		Assert( _view != 0 );
		const char *where = TheKernel->domain()->remove(_view, _buffer);
		Assert( where == _buffer );
		_buffer = 0;
		_view = 0;
	}
}

void 
PersistentArray::reMap()
{
	Assert( _source != 0 );
	int sourceLog2UnitSize = _source->log2UnitSize();
	int size = _source->numberOfUnits() << sourceLog2UnitSize;
	// Assert( size > 0 );
	unsigned currentLog2Size = _log2ViewSize;
	if ( _log2ViewSize == 0 ) 
		_log2ViewSize = 
			( sourceLog2UnitSize > 16 ?
			  sourceLog2UnitSize : 16 ); /* 64kb */
	while (size > (1 << _log2ViewSize))
		_log2ViewSize++;
	if ( _log2ViewSize != currentLog2Size || _buffer == 0 ) {
		if ( _buffer != 0 ) {
			Assert( _view != 0 );
			const char * where = 
				TheKernel->domain()->remove( _view, _buffer );
			Assert( where == _buffer );
		}
		_view = new MemoryObjectView
			( _source, 
			  1 << ( _log2ViewSize - sourceLog2UnitSize),
			  sourceLog2UnitSize, 0 );
		Assert( _view != 0 );
		_buffer = (char *) TheKernel->domain()->add( _view, NoAccess );
		Assert( _buffer != 0 );
	}
}

int 
PersistentArray::useMap( unsigned length )
{
#ifdef USEMAP
	if ( length > (4 << _cacheLog2UnitSize) )
		return 0; 
	if ( _source->numberOfBytes() > PersistentArrayMaxMMapSize ) {
		if ( _buffer != 0 ) {
			Assert( _view != 0 );
			const char * where = 
				TheKernel->domain()->remove( _view, _buffer );
			Assert( where == _buffer );
			_buffer = 0;
			_view = 0;
			_log2ViewSize = 0;
		}
		return 0;
	}
	return 1;
#else
	return 0;
#endif
}

void
PersistentArray::noRemainingReferences()
{
	Raid( RaidReference )( "%N::noRemainingReferences\n", this );
	_source->close( this );
}

int
PersistentArray::init( int argc, char *argv[], FileSystemInterface * )
{
	if( argc > 0 ) {
		Raids( RaidMethod ) << this << "::init( " << argc << ", "
				    << argv[0] << " )\n" << eor;
	} else {
		Raids( RaidMethod ) << this << "::init()\n" << eor;
	}

	if( argc > 0 ) {
		_source->setRecordSize( atoi( argv[0] ) );
	}
	return( 0 );
}

int
PersistentArray::at( unsigned int index )
{
	Raids( RaidMethod ) << this << "::at( " << index << " )\n" << eor;

	Assert( NOTREACHED );

	return( 0 );
}

int
PersistentArray::at( unsigned int offset, char * userBuffer, 
		     unsigned int length )
{
	Raid( RaidMethod )( "%N::at( %d, %x, %d )\n", 
			     this, offset, ( unsigned ) userBuffer,
			     length );

	unsigned int sourceSize = _source->numberOfBytes();
	if( offset >= sourceSize || length == 0 ) 
		return( 0 );
	if( ( offset + length ) > sourceSize ) {
		length = sourceSize - offset;
		if( length == 0 ) 
			return( 0 );
	}

	MemoryObjectCache * cache = _source->cache();
	if( cache ) {
		_cacheLog2UnitSize = cache->log2UnitSize();
		unsigned end = offset + length - 1;
		if ( useMap( length ) ) {
			if ( _buffer == 0 )
				reMap();
			ReplacementPolicy * rp = cache->replacementPolicy();
			if ( rp ) {
				unsigned cacheFrom = 
					offset >> cache->log2UnitSize(); 
				unsigned cacheTo = 
					end >> cache->log2UnitSize(); 
				rp->preCacheRead( cacheFrom, cacheTo );
				ByteCopy( _buffer + offset, userBuffer, 
					  length );
				rp->postCacheRead( cacheFrom, cacheTo );
			} else {
				ByteCopy( _buffer + offset, userBuffer, 
					  length );
			}
			return length;
		}
		unsigned from = offset >> _log2UnitSize;
		unsigned to = end >> _log2UnitSize;
		_source->acquire( from, to );
                length = cache->cacheRead( offset, length, userBuffer, 0 );
		_source->release( from, to );
		return length;
	}
	
	/* no cache */

	int unitNumber = offset >> _log2UnitSize;
	int byteOffset = offset & _unitSizeMask;
	
	char * userPointer = userBuffer;
	
	int bytesRead = 0;
	int bytesToRead;
	
	if( _address == 0 ) {
		_address = new char[ _unitSize ];
		Assert( _address != 0 );
	}

	if( byteOffset != 0 ) {
		bytesToRead = ( ( byteOffset + length ) > _unitSize ) ?
			      ( _unitSize - byteOffset ) : length;
		Raid( RaidDetail )
			( "%N::read pre byteOffset %d bytesToRead %d\n",
			  this, byteOffset, bytesToRead );
		_source->read( unitNumber, 1, _address, _unitSize );
		ByteCopy( _address + byteOffset, userPointer, bytesToRead );
		
		userPointer += bytesToRead;
		bytesRead += bytesToRead;
		length -= bytesToRead;
		
		unitNumber++;
	}
	
	bytesToRead = length & ~_unitSizeMask;
	
	if ( bytesToRead > 0 ) {
		Raid( RaidDetail )( "%N::read in bytesToRead %d\n", 
				    this, bytesToRead );
		_source->read( unitNumber, 
			       bytesToRead >> _log2UnitSize,
			       userPointer, bytesToRead );
		
		userPointer += bytesToRead;
		bytesRead += bytesToRead;
		length -= bytesToRead;
		
		unitNumber += ( bytesToRead >> _log2UnitSize );
	}
	
	if ( length > 0 ) {
		Raid( RaidDetail )( "%N::read post bytesToRead %d\n", 
				    this, bytesToRead );
		_source->read( unitNumber, 1, _address, _unitSize );
		ByteCopy( _address, userPointer, length );
		
		bytesRead += length;
	}
	
	Raid( RaidMethod )( "%N::read returns %d\n", this, bytesRead );
	return( bytesRead );
}

int
PersistentArray::at( char * index, int & status )
{
	Raids( RaidMethod ) << this << "::at( " << index << ", )\n" << eor;

	Assert( NOTREACHED );

	status = NotImplemented;

	return( 0 );
}

int
PersistentArray::atPut( unsigned int index, int value )
{
	Raids( RaidMethod ) << this << "::atPut( " << index << ", "
			    << value << " )\n" << eor;

	Assert( NOTREACHED );

	return( 0 );
}

int
PersistentArray::atPut( unsigned int offset, char * userBuffer, 
		        unsigned int length )
{
	Raid( RaidMethod )( "%N::atPut( %d, %x, %d )\n", 
			     this, offset, ( unsigned ) userBuffer,
			     length );

	unsigned int noOfBytes = offset + length;
	unsigned int sourceBytes = _source->numberOfBytes();
	if( noOfBytes > sourceBytes ) {
		_source->setNumberOfBytes( noOfBytes );
	}
Debug<<"noOfBytes = "<<noOfBytes<<"\n"<<eor;
Debug<<"_source->numberOfBytes() = " 
<<_source->numberOfBytes()<<"\n"<<eor;
Debug<<_source<<"\n"<<eor;

//IMPORTANT CODE ADDED BY Scott Kircher:
 if ( _source->numberOfBytes() < noOfBytes )
		 return 0; //couldn't allocate space

	MemoryObjectCache * cache = _source->cache();
        if( cache ) {
                _cacheLog2UnitSize = cache->log2UnitSize();
		unsigned end = offset + length - 1;
                if ( useMap( length ) ) { 
			if ( ( _buffer == 0 ) || 
			     ( (int)noOfBytes > ( 1 << _log2ViewSize ) ) )
				reMap();
			ReplacementPolicy * rp = cache->replacementPolicy();
			if ( rp ) {
				unsigned cacheFrom = 
					offset >> cache->log2UnitSize();
				unsigned cacheTo = 
					end >> cache->log2UnitSize();
				rp->preCacheWrite( cacheFrom, cacheTo );
				ByteCopy( userBuffer, _buffer + offset, 
					  length );
				rp->postCacheWrite( cacheFrom, cacheTo );
			} else {
				ByteCopy( userBuffer, _buffer + offset, 
					  length );
			}
			return length;
                }
		unsigned from = offset >> _log2UnitSize;
		unsigned to = end >> _log2UnitSize;
		_source->acquire( from, to );
Debug<<"Alpha\n"<<eor;
                length = cache->cacheWrite( offset, length, userBuffer, 0 );
		_source->release( from, to );
Debug<<"Beta\n"<<eor;
		return length;
        }

	/* no cache */

	int unitNumber = offset >> _log2UnitSize;
	int byteOffset = offset & _unitSizeMask;

	char * userPointer = userBuffer;
	
	int bytesWritten = 0;
	int bytesToWrite;
	
	if( _address == 0 ) {
		_address = new char[ _unitSize ];
		Assert( _address != 0 );
	}

	if( byteOffset != 0 ) {
		bytesToWrite = ( ( byteOffset + length ) > _unitSize ) ?
			       ( _unitSize - byteOffset ) : length;
		_source->read( unitNumber, 1, _address, _unitSize );
		ByteCopy( userPointer, _address + byteOffset, bytesToWrite );
		_source->write( unitNumber, 1, _address, _unitSize );
		
		userPointer += bytesToWrite;
		bytesWritten += bytesToWrite;
		length -= bytesToWrite;
		
		unitNumber++;
	}
	
	bytesToWrite = length & ~_unitSizeMask;
	
	if ( bytesToWrite > 0 ) {
		_source->write( unitNumber, bytesToWrite >> _log2UnitSize, 
			        userPointer, bytesToWrite );
		
		userPointer += bytesToWrite;
		bytesWritten += bytesToWrite;
		length -= bytesToWrite;
		
		unitNumber += ( bytesToWrite >> _unitSize );
	}
	
	if ( length > 0 ) {
		_source->read( unitNumber, 1, _address, _unitSize );
		ByteCopy( userPointer, _address, length );
		_source->write( unitNumber, 1, _address, _unitSize );
		
		bytesWritten += length;
	}
	
	Raid( RaidMethod )( "%N::write returns %d\n", this, bytesWritten );
	return( bytesWritten );
}

int
PersistentArray::atPut( char * index, int value, int & status )
{
	Raids( RaidMethod ) << this << "::atPut( " << index << ", "
			    << value << ", )\n" << eor;

	status = NotImplemented;

	Assert( NOTREACHED );

	return( 0 );
}

unsigned
PersistentArray::size()
{
	Raids( RaidMethod ) << this << "::size() return( "
			    << _source->numberOfRecords() << " )\n" << eor;

	return( _source->numberOfRecords() );
}

int
PersistentArray::notify( int event )
{
#ifdef USEMAP
	switch (event) {
	    case NotifyEvent_ZombieStatus: 
		if ( _buffer &&
		     ( _source->zombieInfo()->type() == ZOMBIE_SYNC ||
		       _source->zombieInfo()->type() == ZOMBIE_EXTERN ) 
		   ) {
			Assert( _view != 0 );
			const char *where = 
				TheKernel->domain()->remove(_view, _buffer);
			Assert( where == _buffer );
			_buffer = 0;
			_view = 0;
		}
		break;
	    default:
		break;
	}
#endif
	return event;
}

ProxiableObject *
PersistentArrayConstructor( ProxiableObject * source )
{
        if( source->isKindOf( PersistentMemoryObjectClass ) ) {
		return new PersistentArray( (PersistentMemoryObject *)source );
        }
        else return( 0 );
}

#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("PersistentArray",PersistentArray,PersistentObject,&PersistentArrayConstructor,0);
