#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu) 
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * PersistentMemoryObject.c - persistent memory object classes.
 */

#include "Assert.h"
#include "Byte.h"
#include "CheckSum.h"
#include "Class.h"
#include "FileSystem.h"
#include "FileSystemHints.h"
#include "FileSystemInterface.h"
#include "FileSystemLocks.h"
#include "FileSystemSimpleInstruments.h"
#include "Log2.h"
#include "NotifyEvents.h"
#include "MemoryObjectContainer.h"
#include "MemoryObjectCache.h"
#include "MemoryObjectDictionary.h"
#include "OutputStream.h"
#include "PersistentClass.h"
#include "PersistentMemoryObject.h"
#include "PhysicalMemoryChain.h"
#include "PhysicallyAddressableUnit.h"
#include "Raid.h"
#include "RangeLocks.h"
#include "RemoteFileSystem.h"
#include "ReplacementPolicy.h"
#include "ThisProcess.h"
#include "UNIXFile.h"

#ifdef __GNUC__
#pragma implementation
#endif

PersistentMemoryObjectTimer
PersistentMemoryObjectCtor("PersistentMemoryObject::PersistentMemoryObject() timer");

Semaphore PersistentMemoryObject::_asAMutex(1);

PersistentMemoryObject::PersistentMemoryObject( unsigned int numberOfUnits,
						int log2UnitSize,
						int idNumber,
						MemoryObjectContainer * moc )
     : MemoryObject( numberOfUnits, log2UnitSize ), 
        _container( moc ), _version( 0 ), _creatingDictionary( 0 ),
	_mutex( this ), _openFlags( 0 )
{
	SInstr( int istart = PersistentMemoryObjectCtor.start() );
	Assert( this != 0 );

	initDebug();
	Raids( RaidConstructor )
		<< this << "::PersistentMemoryObject( id " << idNumber
		<< ", container " << moc << " )\n" << eor;

        _persistentObject = 0;
	_cachedPersistentObject = 0;
	_idNumber = idNumber;
	extern unsigned StandardZombieDelay;
	_zombieDelay = StandardZombieDelay;
        // _protection = ReadWrite;
	
	Assert( _moLock == 0 );
	_moLock = new NormalRangeLock( this );
	Assert( _moLock != 0 );
	SInstr( PersistentMemoryObjectCtor.stop( istart ) );
}

PersistentMemoryObject::~PersistentMemoryObject()
{
	Raids( RaidConstructor ) 
		<< this << "::~PersistentMemoryObject() " << _container 
		<< "\n" << eor;

	Assert( _referenceCount.value() == 0 );

	if ( _moLock )
		delete _moLock;
	if ( _creatingDictionary ) {
		MemoryObjectDictionary * dict = _creatingDictionary;
		_creatingDictionary = 0;
		dict->unreference();
	}
}

void
PersistentMemoryObject::unreference()
{
        Raids( RaidReference ) << this << "::unreference()\n" << eor;

        Assert( alive() );

	if ( _container != 0 )
		_container->acquire( _idNumber );
	_referenceCount.decrement();
	if ( _referenceCount.value() == 0 ) {
		noRemainingReferences();
	} else {
		if ( _container != 0 )
			_container->release( _idNumber );
	}
}

void
PersistentMemoryObject::noRemainingReferences()
{

	if( _container != 0 ) {
		_container->close( (FileObject *) this );
	} else {
		destroy();
	}
}

int
PersistentMemoryObject::accessMode()
{
	return( UNIXReadMask|UNIXExecuteMask );
}

Class *
PersistentMemoryObject::supports( Class * aClass )
{
	Raid( RaidProtected )( "%N::supports( %N )\n", this, aClass );

	Class * supportedClass = 0;

        extern PersistentClassStar PersistentArrayClass;
        if( PersistentArrayClass->isASubclassOf( aClass ) ) {
            supportedClass = PersistentArrayClass;
        }

        Raid( RaidProtected )( "%N::supports() returns %N\n",
                                this, supportedClass );

	return( supportedClass );
}

void
PersistentMemoryObject::basicFlush()
{
	Raids( RaidProtected ) 
		<< this << "::basicFlush()\n" << eor;

        MemoryObject::basicFlush();
	if ( _container )
		_container->synchronize( _idNumber );
}

void
PersistentMemoryObject::BasicSynchronizeAsyncClosure::func
	( PersistentMemoryObject::BasicSynchronizeAsyncClosure * c )
{
	c->_pmo->basicSynchronizeClosure( c );
}

void
PersistentMemoryObject::basicSynchronizeClosure
	( PersistentMemoryObject::BasicSynchronizeAsyncClosure * c )
{
	if ( c->_state == 0 ) {
		c->_state = 1;
		TheAsyncSupport->schedule( c );
		return;
	}
	if ( c->_state == 1 ) {
		c->_state = 2;
		_container->synchronize( _idNumber );
		c->_parent->activate();
		delete c;
		return;
	}
	Assert( NOTREACHED );
}

void
PersistentMemoryObject::basicSynchronize( IOAsyncClosure * c )
{
	Raids( RaidProtected )
		<< this << "::basicSynchronize()\n" << eor;

	if ( c != 0 && _container != 0 ) {
		c = new BasicSynchronizeAsyncClosure( c, this );
	}
	MemoryObject::basicSynchronize( c );
	if ( c == 0 && _container != 0 ) {
		_container->synchronize( _idNumber );
	}
}

void
PersistentMemoryObject::basicSynchronize( unsigned from, unsigned to,
					  IOAsyncClosure * c )
{
	Raids( RaidProtected )
		<< this << "::basicSynchronize( " << from << ", " << to 
		<< " )\n" << eor;

	if ( c != 0 && _container != 0 ) {
		c = new BasicSynchronizeAsyncClosure( c, this );
	}
	MemoryObject::basicSynchronize( from, to, c );
	if ( c == 0 && _container != 0 ) {
		_container->synchronize( _idNumber );
	}
}

unsigned
PersistentMemoryObject::basicSetNumberOfBytes( unsigned bytes )
{
    Assert(NOTREACHED);
    return( numberOfBytes() );
}

unsigned
PersistentMemoryObject::basicSetNumberOfRecords( unsigned records )
{
        return( basicSetNumberOfBytes( recordSize() * records ) );
}

unsigned
PersistentMemoryObject::basicSetRecordSize( unsigned )
{
    Assert(NOTREACHED);
        return( recordSize() );
}

int
PersistentMemoryObject::basicCopy( MemoryObject * dest )
{
	Raids( RaidMethod ) 
		<< this << "::basicCopy( " << dest << " )\n" << eor;

	int returnValue = MemoryObject::basicCopy( dest );
	if ( dest->isKindOf( PersistentMemoryObjectClass ) ) {
		( ( PersistentMemoryObject * )dest ) -> 
			setNumberOfBytes( numberOfBytes() );
	}

	Raids( RaidMethod ) 
		<< this << "::basicCopy( " << dest << " ) returns "
		<< returnValue << "\n" << eor;

	return( returnValue );
}

int
PersistentMemoryObject::referenceCount()
{
	return _referenceCount.value();
}

extern ClassStar UNIXInodeCacheClass;

PersistentMemoryObjectTimer
PersistentMemoryObjectParseFlags("PersistentMemoryObject::parseFlags() timer");

int 
PersistentMemoryObject::parseFlags( int flags )
{
	SInstr( int istart = PersistentMemoryObjectParseFlags.start() );

	if ( ! ( flags & O_NCACHE ) ) {
		
		/*
		 * some assertions to make sure we don't buildCache for
		 * pmo's we have not tested for cache compatibility yet
		 */
		
		extern ClassStar MemoryObjectPartitionClass;
		extern ClassStar DiskClass;
		Assert( !isKindOf( MemoryObjectPartitionClass ) );
		Assert( !isKindOf( DiskClass ) );
		
		/*
		 * figure out cache unit size 
		 */

		if ( flags & O_CMASK ) {

			int log2Size = 0;
			if (( flags & O_CMASK ) == O_CSIZE ) {
				log2Size = 12;
				int bytes = numberOfBytes();
				while ( bytes > ( 1 << log2Size ) )
					log2Size++;
			} else {
				log2Size = 
					( ( flags & O_CMASK ) >> 28 ) + 12;
			}
			Assert( log2Size != 0 );
			setCacheLog2UnitSize( log2Size );

		}

		if ( flags & O_HINTMASK ) {
			updateFlagsAndHints( flags );
		}

		MemoryObjectCache * moc = buildCache();

		if ( moc != 0 && h_lcache( _hints ) && isRemote() ) {
			addCache( UNIXInodeCacheClass );
		}
	}

	SInstr( PersistentMemoryObjectParseFlags.stop( istart ) );
	return flags;
}

void
PersistentMemoryObject::hintsChanged( unsigned oldHints )
{
}

void
PersistentMemoryObject::updateFlagsAndHints( int flags )
{
	_hintLock.acquire();
	unsigned oldHints = _hints;
	if ( flags & O_CLRHINTMASK ) {
		if ( flags & O_NLCACHE ) {
			_openFlags = ( _openFlags | O_NLCACHE ) & ~O_LCACHE;
			_hints &= ~H_LCACHE;
		} 
		if ( flags & O_MULTI ) {
			_openFlags = ( _openFlags | O_MULTI ) & ~O_SINGLE;
			_hints &= ~H_SINGLE;
		}
		if ( flags & O_SMALL ) {
			_openFlags = ( _openFlags | O_SMALL ) & ~O_LARGE;
			_hints &= ~H_LARGE;
		}
		if ( flags & O_SEQ ) {
			_openFlags = ( _openFlags | O_SEQ ) & ~O_RANDOM;
			_hints &= ~H_RANDOM;
		}
		if ( flags & O_PERM ) {
			_openFlags = ( _openFlags | O_PERM ) & ~O_TEMP;
			_hints &= ~H_TEMP;
		}
	}
	if ( flags & O_SETHINTMASK ) {
		if ( flags & O_LCACHE ) {
			_openFlags = ( _openFlags | O_LCACHE ) & ~O_NLCACHE;
			_hints |= H_LCACHE;
		} 
		if ( flags & O_SINGLE ) {
			_openFlags = ( _openFlags | O_SINGLE ) & ~O_MULTI;
			_hints |= H_SINGLE;
		}
		if ( flags & O_LARGE ) {
			_openFlags = ( _openFlags | O_LARGE ) & ~O_SMALL;
			_hints |= H_LARGE;
		}
		if ( flags & O_RANDOM ) {
			_openFlags = ( _openFlags | O_RANDOM ) & ~O_SEQ;
			_hints |= H_RANDOM;
		}
		if ( flags & O_TEMP ) {
			_openFlags = ( _openFlags | O_TEMP ) & ~O_PERM;
			_hints |= H_TEMP;
		}
	}
	unsigned newHints = _hints;
	_hintLock.release();
	if ( newHints != oldHints ) {
		hintsChanged( oldHints );
	}
}

void 
PersistentMemoryObject::setHints( unsigned mask )
{
	_hintLock.acquire();
	unsigned oldHints = _hints;
	if ( h_lcache(mask) && !(_openFlags & O_NLCACHE) ) {
		_hints |= H_LCACHE;
	}
	if ( h_single(mask) && !(_openFlags & O_MULTI) ) {
		_hints |= H_SINGLE;
	}
	if ( h_large(mask) && !(_openFlags & O_SMALL) ) {
		_hints |= H_LARGE;
	}
	if ( h_random(mask) && !(_openFlags & O_SEQ) ) {
		_hints |= H_RANDOM;
	}
	if ( h_temp(mask) && !(_openFlags & O_PERM) ) {
		_hints |= H_TEMP;
	}
	unsigned newHints = _hints;
	_hintLock.release();
	if ( newHints != oldHints ) {
		hintsChanged( oldHints );
	}
}

void
PersistentMemoryObject::clearHints( unsigned mask )
{
	_hintLock.acquire();
	unsigned oldHints = _hints;
	if ( h_lcache(mask) && !(_openFlags & O_LCACHE) ) {
		_hints &= ~H_LCACHE;
	}
	if ( h_single(mask) && !(_openFlags & O_SINGLE) ) {
		_hints &= ~H_SINGLE;
	}
	if ( h_large(mask) && !(_openFlags & O_LARGE) ) {
		_hints &= ~H_LARGE;
	}
	if ( h_random(mask) && !(_openFlags & O_RANDOM) ) {
		_hints &= ~H_RANDOM;
	}
	if ( h_temp(mask) && !(_openFlags & O_TEMP) ) {
		_hints &= ~H_TEMP;
	}
	unsigned newHints = _hints;
	_hintLock.release();
	if ( newHints != oldHints ) {
		hintsChanged( oldHints );
	}
}

void
PersistentMemoryObject::updateHints()
{
	MemoryObjectCache * moc = _cache;
	unsigned cacheLog2UnitSize = moc->log2UnitSize();
	unsigned bytesAccessed = moc->bytesAccessed();
	unsigned cacheUnitsAccessed =
		(bytesAccessed + ((1 << cacheLog2UnitSize) - 1))
		>> cacheLog2UnitSize;
	unsigned cacheNumberOfUnits = moc->numberOfUnits();
	extern int debugHints;
	if ( debugHints ) {
		Console()
			<< this << "::updateHints()"
			<< " H_SINGLE "
			<< ( cacheUnitsAccessed <= cacheNumberOfUnits )
			<< " H_LCACHE "
			<< !( moc->fromCacheUnits() <= cacheNumberOfUnits )
			<< "\n" << eor;
	}	
	if ( cacheUnitsAccessed <= cacheNumberOfUnits ) {
		setHints( H_SINGLE );
	} else {
		clearHints( H_SINGLE );
	}
	if ( moc->fromCacheUnits() <= cacheNumberOfUnits ) {
		clearHints( H_LCACHE );
	} else {
		setHints( H_LCACHE );
	}
}

void
PersistentMemoryObject::setCacheHints( unsigned userHints )
{
	_hintLock.acquire();
	unsigned oldHints = _hints;
	if ( h_lcache( userHints ) ) {
		_openFlags |= O_LCACHE;
		_openFlags &= ~ O_NLCACHE;
	} else {
		_openFlags |= O_NLCACHE;
		_openFlags &= ~ O_LCACHE;
	}
	if ( h_single( userHints ) ) {
		_openFlags |= O_SINGLE;
		_openFlags &= ~ O_MULTI;
	} else {
		_openFlags |= O_MULTI;
		_openFlags &= ~ O_SINGLE;
	}
	if ( h_large( userHints ) ) {
		_openFlags |= O_LARGE;
		_openFlags &= ~ O_SMALL;
	} else {
		_openFlags |= O_SMALL;
		_openFlags &= ~ O_LARGE;
	}
	if ( h_random( userHints ) ) {
		_openFlags |= O_RANDOM;
		_openFlags &= ~ O_SEQ;
	} else {
		_openFlags |= O_SEQ;
		_openFlags &= ~ O_RANDOM;
	}
	if ( h_temp( userHints ) ) {
		_openFlags |= O_TEMP;
		_openFlags &= ~ O_PERM;
	} else {
		_openFlags |= O_PERM;
		_openFlags &= ~ O_TEMP;
	}
	_hints = userHints;
	_hintLock.release();
	if ( userHints != oldHints ) {
		hintsChanged( oldHints );
	}
}

unsigned
PersistentMemoryObject::getCacheHints()
{
	return _hints;
}

PersistentMemoryObjectTimer
PersistentMemoryObjectDestroy("PersistentMemoryObject::destroy() timer");

void
PersistentMemoryObject::destroy()
{
	SInstr( int istart = PersistentMemoryObjectDestroy.start() );

	Raids( RaidProtected ) 
		<< this	<< "::destroy()\n" << eor;

	Assert( _referenceCount.value() == 0 );
	Assert( _persistentObject == 0 );

        if( _cachedPersistentObject != 0 ) {
                Raids( RaidDetail ) 
			<< this 
			<< "::destroy() cachedPersistentObject "
		        << _cachedPersistentObject << "\n" << eor;
		delete _cachedPersistentObject;
		_cachedPersistentObject = 0;
	}

	if( !transient() ) {
		if ( _cache  != 0) {
			updateHints();
		}
		flush();
#ifdef NEVER 
		ZombieInfoType type = zombieInfo()->type();
		if ( type == ZOMBIE_NOT || type == ZOMBIE_NOTSYNC ) {
			flush();
		} else {
			if ( _cache ) {
				_cache->free();
			}
		}
#endif
	}

	delete this;
	SInstr( PersistentMemoryObjectDestroy.stop( istart ) );
}

PersistentMemoryObjectTimer
PersistentMemoryObjectSetNumberOfUnits("PersistentMemoryObject::setNumberOfUnits() timer");

int
PersistentMemoryObject::setNumberOfUnits( unsigned units )
{
	SInstr( int istart = PersistentMemoryObjectSetNumberOfUnits.start() );

	if ( units != numberOfUnits() && ! fixedSize() ) {
		acquireAll();
		units = basicSetNumberOfUnits( units );
		releaseAll();
	}

	SInstr( PersistentMemoryObjectSetNumberOfUnits.stop( istart ) );
	return units;
}

unsigned
PersistentMemoryObject::numberOfRecords()
{
        unsigned recSize = recordSize();
        unsigned bytes = numberOfBytes();

        Assert( recSize >= 1 );
        Assert( ( bytes % recSize ) == 0 );

        return( bytes / recSize );
}

PersistentMemoryObjectTimer
PersistentMemoryObjectSetNumberOfRecords("PersistentMemoryObject::setNumberOfRecords() timer");

int
PersistentMemoryObject::setNumberOfRecords( unsigned records )
{
	SInstr( int istart = PersistentMemoryObjectSetNumberOfRecords.start() );
	int result;
	if ( fixedSize() )
		result =  numberOfRecords();
	else if ( records == numberOfRecords() )
		result = records;
	else {
		acquireAll();
        	result = basicSetNumberOfRecords( records );
		releaseAll();
	}
	SInstr( PersistentMemoryObjectSetNumberOfRecords.stop( istart ) );
	return result;
}

unsigned
PersistentMemoryObject::recordSize()
{
        return( 1 );
}

int
PersistentMemoryObject::setRecordSize( unsigned size )
{
	if ( recordSize() == size )
		return size;
	acquireAll();
        int result = basicSetRecordSize( size );
	releaseAll();
	return result;
}

unsigned
PersistentMemoryObject::numberOfBytes()
{
	return( numberOfUnits() << log2UnitSize() );
}

PersistentMemoryObjectTimer
PersistentMemoryObjectSetNumberOfBytes("PersistentMemoryObject::setNumberOfBytes() timer");

int
PersistentMemoryObject::setNumberOfBytes( unsigned size )
{
	SInstr( int istart = PersistentMemoryObjectSetNumberOfBytes.start() );
	int result;

	if ( fixedSize() )
		result =  numberOfBytes();
	else if ( numberOfBytes() == size )
		result = size;
	else {
		acquireAll();
        result = basicSetNumberOfBytes( size );
		releaseAll();
	}
	SInstr( PersistentMemoryObjectSetNumberOfBytes.stop( istart ) );
	return result;
}

void
PersistentMemoryObject::writeName( OutputStream * os )
{
	ProxiableObject::writeName( os );
	os->formattedWrite( "(" );
	writeIdNumber( os );
	os->formattedWrite( ")" );
}

void
PersistentMemoryObject::writeIdNumber( OutputStream * os )
{
	if( _idNumber >= 0 ) os->formattedWrite( ":%d", _idNumber );
}

int
PersistentMemoryObject::isSymbolicLink()
{
	return( 0 );
}

int
PersistentMemoryObject::isContainer()
{
	return( 0 );
}

int
PersistentMemoryObject::isDictionary()
{
	return( 0 );
}

int
PersistentMemoryObject::setMode( int )
{
	return( NotImplemented );
}
 
int
PersistentMemoryObject::setUserId( int )
{
	return( NotImplemented );
}
 
int
PersistentMemoryObject::setGroupId( int )
{
	return( NotImplemented );
}
 
void
PersistentMemoryObject::info( UNIXFileInfo * infoBuffer )
{
	Raid( RaidMethod )( "%N::info( ) %d\n", this, _idNumber );
	infoBuffer->device	= 0;
	infoBuffer->inumber	= _idNumber;
	infoBuffer->mode	= UNIXTypeRegular | accessMode();
	infoBuffer->links	= 1;
	infoBuffer->uid	        = userId();
	infoBuffer->gid	        = groupId();
	infoBuffer->flags	= 0;	//FIX
	infoBuffer->rdevice	= 0;	//FIX
	infoBuffer->size	= numberOfBytes();
	infoBuffer->atime	= 0;
	infoBuffer->mtime	= 0;
	infoBuffer->ctime	= 0;
	infoBuffer->blockSize   = unitSize();
	infoBuffer->blocks	= numberOfUnits();
}

MemoryObjectContainerRef 
PersistentMemoryObject::container()
{
	return( _container );
}

PersistentObjectRef
PersistentMemoryObject::persistentObject()
{
	return( _persistentObject );
}

PersistentMemoryObjectTimer
PersistentMemoryObjectAsA("PersistentMemoryObject::asA() timer");


ProxiableObjectRef
PersistentMemoryObject::asA( Class * aClass, int & status )
{
	SInstr( int istart = PersistentMemoryObjectAsA.start() );
	Raids( RaidMethod ) << this << "::asA( " << aClass << " )\n" << eor; 
	Debug << this << "::asA( " << aClass << " )\n" << eor; 

	int done = 0;
	Assert( &status != 0 );
	Assert( aClass != 0 );
	status = 0;
	PersistentObjectRef po;
	Class * supportedClass = 0;
	int supportedClassChecked = 0;

	if( isKindOf( aClass ) ) return( this );

#ifdef ASSERT
	static Process * asMutexAcquiredBy = 0;
	Process * me = thisProcess();
	Assert( asMutexAcquiredBy != me );
#endif
	_asAMutex.P();
#ifdef ASSERT
	asMutexAcquiredBy = me;
#endif

	if ( _persistentObject != 0 ) {
		if ( _persistentObject->isKindOf( aClass ) )
			po = _persistentObject;
		else
		  {
		    Console()<<"Badness1\n";
			status = ProtocolMismatch;
		  }
		done = 1;
		Raids( RaidDetail ) 
			<< this << "::asA() " << _persistentObject 
			<< " already created\n" << eor;
	} 
	else if ( _cachedPersistentObject != 0 ) {
		supportedClass = supports( aClass );
		supportedClassChecked = 1;
		if ( supportedClass == 0 ) {
			done = 1;
			Console()<<"Badness2\n";
			status = ProtocolMismatch;
			Raids( RaidDetail ) 
				<< this << "::asA() " 
				<< aClass 
				<< " not supported\n" << eor;
		}
		else if (_cachedPersistentObject->classOf()==supportedClass) {
			_persistentObject = _cachedPersistentObject;
			po = _persistentObject;
			this->reference();
			_cachedPersistentObject = 0;
			done = 1;
			Raids( RaidDetail ) 
				<< this << "::asA() " 
				<< _cachedPersistentObject 
				<< " already cached\n" << eor;
		} 
		else {
			Raids( RaidDetail ) 
				<< this << "::asA() wrong type "
				<< _cachedPersistentObject 
				<< " cached\n" << eor;
			delete _cachedPersistentObject;
			_cachedPersistentObject = 0;
			done = 0;
		}
	}

	if ( !done ) {
		if ( !supportedClassChecked )
		 	supportedClass = supports( aClass );
		if ( supportedClass != 0 ) {
			extern PersistentClassStar 
				AutoloadPersistentObjectClass;
			this->reference();
			Debug<<"supportedClass="<<supportedClass<<"\n"<<eor;
			po = supportedClass->constructor( this );
			if( ( po != 0 ) &&
			    ( numberOfUnits() > 0 ) &&
			    supportedClass->isASubclassOf
			    ( AutoloadPersistentObjectClass ) 
			   ) {
				int bufferSize = 
					numberOfUnits() << log2UnitSize();
				char * buffer = new char[bufferSize];
				read( 0, numberOfUnits(), buffer, bufferSize );
				int baseSize = PersistentObjectClass->sizeOf();
				int size = supportedClass->sizeOf() - baseSize;
				Assert( size >= 0 );
				char * dest = ((char *)(PersistentObject *)po) 
					+ baseSize;
				ByteCopy( buffer, dest, size );
				delete buffer;
			}
			if ( po != 0 ) {
				Assert( po->isKindOf( aClass ) );
				_persistentObject = po;
				if( ( objectName() != 0 ) && 
				    ( objectName()[0] != 0 ) ) 
					po->setName( objectName() );
			} else {
			  Console()<<"Badness3\n";
				status = ProtocolMismatch;
				this->unreference();
			}
		} else {
			Assert( po == 0 );
			Console()<<"Badness4\n";
			status = ProtocolMismatch;
		}
	}
	
#ifdef ASSERT
	asMutexAcquiredBy = 0;
#endif
	_asAMutex.V();

	Raids( RaidMethod ) << this << "::asA( " << aClass << " ) return( "
                        << po << " )\n" << eor; 
	Debug << this << "::asA( " << aClass << " ) return( "
          << po << " )\n" << eor; 

	SInstr( PersistentMemoryObjectAsA.stop( istart ) );
	return( po );
}

int
PersistentMemoryObject::userId()
{
	return( 0 );
}

int
PersistentMemoryObject::groupId()
{
	return( 0 );
}

extern int UID;
extern int GID;

int
PersistentMemoryObject::access( int mask )
{
	Raid( RaidMethod )( "%N::access(%o) UID %d GID %d\n", 
			    this, mask, UID, GID );

	int answer = 0;
	int mode = accessMode();
	
	if( UID != 0 ) {
		if     ( UID == userId() )  mode >>= UNIXOwnerShift;
		else if( GID == groupId() ) mode >>= UNIXGroupShift;
		else			    mode >>= UNIXPublicShift;
		
		answer = ( mask & mode ) == mask;
	}
	else {
		// FIX Execute for ODs should be handled separately
		if( mask & AccessExecute ) 
			answer = mode & UNIXExecuteMask;
		else 
			answer = 1;
	}
	
	Raid( RaidMethod )( "%N::access returns %o\n", this, answer );
	return( answer );
}

PersistentMemoryObjectTimer
PersistentMemoryObjectClose("PersistentMemoryObject::close() timer");

void
PersistentMemoryObject::close( PersistentObject * po )
{
	SInstr( int istart = PersistentMemoryObjectClose.start() );
	Raid( RaidMethod )( "%N::close( %N )\n", this, po );
	
        Assert( _persistentObject->identity() == po->identity() );
	Assert( _cachedPersistentObject == 0 );

        _persistentObject = 0;
	_cachedPersistentObject = po;
	this->unreference(); 	
	SInstr( PersistentMemoryObjectClose.stop( istart ) );
}

int
PersistentMemoryObject::idNumber()
{
	return( _idNumber );
}
 
int
PersistentMemoryObject::links( int )
{
	return( 1 );
}

int
PersistentMemoryObject::alloc( unsigned startUnit, int count )
{
	return count;
}

int
PersistentMemoryObject::free( unsigned startUnit, int count )
{
	return 0;
}

void
PersistentMemoryObject::synchronize()
{
        Raids( RaidMethod ) << this << "::synchronize()\n" << eor;
	if ( _creatingDictionary )
		synchronizeCreatingDictionary();
        acquireAll();
        basicSynchronize( 0 );
        releaseAll();
        Raids( RaidMethod ) << this << "::synchronize() done\n" << eor;
}

static Lock CreatingDictionaryLock;

void
PersistentMemoryObject::synchronizeCreatingDictionary()
{
	MemoryObjectDictionary * dict = 0;
	CreatingDictionaryLock.acquire();
	if ( _creatingDictionary ) {
		dict = _creatingDictionary;
		_creatingDictionary = 0;
	}
	CreatingDictionaryLock.release();
	if (dict) {
		dict->synchronize();
		dict->unreference();
	}
}

void 
PersistentMemoryObject::setCreatingDictionary( MemoryObjectDictionary * dict )
{
	CreatingDictionaryLock.acquire();
	MemoryObjectDictionary * old = _creatingDictionary;
	_creatingDictionary = dict;
	if (dict) {
		dict->reference();
	}
	CreatingDictionaryLock.release();
	if (old) {
		old->unreference();
	}
}

int
PersistentMemoryObject::unmount()
{
	int status = 0;
	MemoryObjectContainerStar oc = _container;
	setZombieDelay( 0 );
	// _asAMutex.P();
	if ( _persistentObject != 0 ) {
		Assert( _cachedPersistentObject == 0 );
		status = _persistentObject->unmount();
	} else if ( _cachedPersistentObject != 0 ) {
		Assert( _cachedPersistentObject->alive() );
		status = _cachedPersistentObject->unmount();
		delete _cachedPersistentObject;
		_cachedPersistentObject = 0;
	}
	// _asAMutex.V();
	if ( oc ) {
		oc->unmount();
		Console() << "Unmounting a directory in " << oc << "\n" << eor;
	}
	oc = 0;
	return status;
}

int
PersistentMemoryObject::notify( int event )
{
	if ( _cache != 0) {
#ifdef NEVER
		if ( event == NotifyEvent_ContainerClose ||
		     ( event == NotifyEvent_ZombieStatus && 
		       zombieInfo()->type() == ZOMBIE_EXTERN ) ) {
#endif
		if ( event == NotifyEvent_ContainerClose ) {
			updateHints();
			MemoryObjectCache * moc = _cache;
			if ( moc && moc->replacementPolicy() ) {
				moc->replacementPolicy()->notify( event );
			}
		}
	}
	PersistentObject * po = 0;
	if ( _persistentObject != 0 ) {
		po = _persistentObject;
	} else if ( _cachedPersistentObject != 0 ) {
		po = _cachedPersistentObject;
	}
	if ( po != 0 ) {
		return po->notify( event );
	}
	return event;
}

int 
PersistentMemoryObject::isRemote()
{
	return 0;
}

void
PersistentMemoryObject::getAttributes( RFSAttributes * attr )
{
        UNIXFileInfo uinfo;
        info( &uinfo );
        attr->mode = uinfo.mode;
        attr->links = uinfo.links;
        attr->uid = uinfo.uid;
        attr->gid = uinfo.gid;
        attr->size = uinfo.size;
        attr->atime = uinfo.atime;
        attr->mtime = uinfo.mtime;
        attr->ctime = uinfo.ctime;
        attr->type.set( this );
        attr->log2BlockSize = log2Low( uinfo.blockSize >> _log2UnitSize );
        attr->log2UnitSize = _log2UnitSize;
        attr->flags = 
		( fixedSize() ? RFSAttributes::FixedSize : 0 ) |
		RFSAttributes::AttrDirty;
}

void
PersistentMemoryObject::setAttributes( RFSAttributes * attr )
{
	if ( !(attr->flags & RFSAttributes::AttrDirty) )
		return;
        if ( !isDictionary() && numberOfBytes() != (unsigned)attr->size )
                setNumberOfBytes( (unsigned)attr->size );
        if ( userId() != attr->uid )
                setUserId( attr->uid );
        if ( groupId() != attr->gid )
                setGroupId( attr->gid );
        if ( (accessMode() & UNIXAccessMask) !=
             (attr->mode & UNIXAccessMask) )
                setMode( attr->mode & UNIXAccessMask );
}

#include "ClassInstaller.h"
DEFINECLASS("PersistentMemoryObject",PersistentMemoryObject,MemoryObject,0,0);
