#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * RemoteContainer.c - 
 */

#include "FileObject.h"
#include "FileSystem.h"
#include "MemoryObjectCache.h"
#include "Raid.h"
#include "RemoteContainer.h"
#include "RemoteFileObject.h"
#include "RemoteFileSystem.h"
#include "RemoteFileSystemRPC.h"
#include "RemoteTables.h"
#include "Semaphore.h"

#ifdef __GNUC__
#pragma implementation
#endif

RemoteContainer::RemoteContainer( RFSContainerHandle & handle )
        : MemoryObjectCachingContainer( 0 ), _handle( handle )
{
	initDebug();
	Raids( RaidConstructor ) << this << "::RemoteContainer()\n" << eor;
	initializeArrays( handle._size );
}

RemoteContainer::~RemoteContainer()
{
	Raids( RaidConstructor ) << this << "::~RemoteContainer()\n" << eor;
}

void 
RemoteContainer::unreference()
{
	Raids( RaidReference ) << this << "::unreference()\n" << eor;

	Assert( alive() );

	_referenceCount.decrement();
	if( _referenceCount.value() == 0 ) {
		noRemainingReferences();
	}
}

void 
RemoteContainer::noRemainingReferences()
{
	TheRemoteContainerRegistry->unbind( this );
	delete this;
}

FileObjectRef
RemoteContainer::fastOpen( RFSHandle & handle, RFSAttributes & attributes,
			   int & status, int flags, int create )
{
	Raids( RaidMethod ) 
		<< this << "::fastOpen( idNumber " << handle._idNumber 
		<< " )\n" << eor;

	int idNumber = handle._idNumber;

	if( ( 0 > idNumber ) || ( idNumber >= _numberOfFileObjects ) ) {
		status = IdNumberOutOfRange;
		return( 0 );
	}

	_syncMutex.acquireRead();

	acquire( idNumber );
	FileObject * fo = lookup( idNumber );
	if( fo == 0 ) {
		fo = new RemoteFileObject
			( this, handle, attributes, create );
		bind( idNumber, fo );
		acquire();
		if ( _openFileObjects++ == 0 ) reference();
		release();
	} else {
		RemoteFileObject * rfo = (RemoteFileObject *) fo;
		ZombieInfo * zombieInfo = fo->zombieInfo();
		ZombieInfoType zombieType = zombieInfo->type();
		if ( rfo->referenceCount() == 0 ) {
#ifdef SAFESYNC
			Assert( zombieType != ZOMBIE_NOT );
			if ( *(rfo->handle()) != handle ) {
				slock.acquire();
				zombieInfo->unlink();
				slock.release();
				zombieInfo->setType( zombieType );
				// stale handle
				// make it transient, no need to flush
				rfo->links(-1000000);	
				basicClose( rfo );
				fo = new RemoteFileObject
					( this, handle, attributes, create );
				bind( idNumber, fo );
			} else {
				if ( zombieType == ZOMBIE_NOTSYNC ) {
					Assert( zombieInfo->expire() > 0 );
					zombieInfo->setExpire( 0 );
				} else {
					slock.acquire();
					zombieInfo->unlink();
					slock.release();
					Assert( zombieInfo->time() == 0 );
					Assert( zombieInfo->expire() == 0 );
				}
			}
#else // not SAFESYNC
			Assert( zombieType != ZOMBIE_NOT );
			slock.acquire();
			zombieInfo->unlink();
			slock.release();
			if ( *(rfo->handle()) != handle ) {
				// stale handle
				// make it transient, no need to flush
				rfo->links(-1000000);	
				basicClose( rfo );
				fo = new RemoteFileObject
					( this, handle, attributes, create );
				bind( idNumber, fo );
			}
#endif // SAFESYNC
		} else {
#ifdef SAFESYNC
			Assert( zombieType == ZOMBIE_NOT ||
				zombieType == ZOMBIE_NOTSYNC );
			Assert( zombieInfo->expire() == 0 );
#else // not SAFESYNC
			Assert( zombieType == ZOMBIE_NOT );
#endif // SAFESYNC
			if ( *(rfo->handle()) != handle ) {
				Console() 
				<< this << " idNumber " << idNumber
				<< " is active but stale\n" << eor;
				// make it transient to close
				// as soon as possible when not active
				rfo->links(-1000000);	
			}
		}
		status = 0;
	}
	parseFlags( fo, flags ); 
	FileObjectRef ret = fo;
	release( idNumber );

	_syncMutex.releaseRead();

	Raids( RaidMethod )
		<< this << "::fastOpen returns " << ret << " status " << status
		<< "\n" << eor;

	return ret;
}

FileObject * 
RemoteContainer::basicOpen( int idNumber, int & status, int flags )
{
	Raids( RaidMethod ) 
		<< this << "::basicOpen( idNumber " << idNumber << " )\n" 
		<< eor;

	RFSHandle fileHandle;
	RFSAttributes attributes;

        status = RemoteFileSystemRPC::containerOpen
		( _handle, idNumber, flags, &fileHandle, &attributes );
	
	Raids( RaidMethod ) 
		<< this << "::basicOpen( idNumber " << idNumber 
		<< " ) RPC status " << status << "\n" << eor;

	if ( status != 0 )
		return 0;
	return new RemoteFileObject ( this, fileHandle, attributes, 0 );
}

FileObject * 
RemoteContainer::basicCreate( int hint, Class *aClass, int &status, int flags ) 
{
	Raids( RaidMethod )
		<< this << "::basicCreate( hint " << hint << ", " << aClass 
		<< " )\n" << eor;

	RFSObjectType type( aClass );
	RFSHandle fileHandle;
	RFSAttributes attributes;

	status = RemoteFileSystemRPC::containerCreate
		( _handle, hint, type, flags, &fileHandle, &attributes );

	Raids( RaidMethod ) 
		<< this << "::basicCreate() RPC status " << status 
		<< "\n" << eor;

	if ( status != 0 ) 
		return 0;
	return new RemoteFileObject( this, fileHandle, attributes, 1 );
}

void 
RemoteContainer::basicClose( FileObject * fo )
{
	Raid( RaidMethod )( "%N::basicClose( %N )\n", this, fo );

	Assert( fo->isKindOf( RemoteFileObjectClass ) );
	MemoryObjectContainer::basicClose( fo );
}

int 
RemoteContainer::basicRootId()
{
	return _handle._rootId;
}

void
RemoteContainer::basicSynchronize()
{
        // container and file objects have already been locked
        // Write out any in memory file objects.

	for( FileObject **fop = _activeFileObjects; *fop != 0; fop++ ) {
		RemoteFileObject * rfo = (RemoteFileObject *) (*fop);
		if( !rfo->transient() && 
		    ( rfo->zombieInfo()->type() == ZOMBIE_NOT ||
		      rfo->zombieInfo()->type() == ZOMBIE_NOTSYNC )
	 	  ) {
			MemoryObjectCache * cache = rfo->cache();
			if ( cache ) {
				cache->synchronize( 0 );
			}
		}
		rfo->syncAttributes(1, 0, 0);
	}
}

void
RemoteContainer::basicSynchronize( int idNumber )
{
	RemoteFileObject * rfo = (RemoteFileObject *) lookup( idNumber );
	Assert( rfo != 0 );
	rfo->syncAttributes(1, 0, 0);
}

#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("RemoteContainer",RemoteContainer,MemoryObjectCachingContainer,0,0);

