#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Contributing Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * RemoteFileObject.c - 
 */

#include "Assert.h"
#include "Byte.h"
#include "CacheAsync.h"
#include "FileSystemHints.h"
#include "FileSystemInterface.h"
#include "PersistentClass.h"
#include "Raid.h"
#include "RemoteFileObject.h"
#include "RemoteFileSystem.h"
#include "RemoteFileSystemRPC.h"
#include "RemoteTables.h"
#include "UNIXFile.h"

#ifdef __GNUC__
#pragma implementation
#endif

unsigned RemoteFileObjectDefaultHint = H_LCACHE;

RemoteFileObject::RemoteFileObject( MemoryObjectContainer * container,
				    RFSHandle & handle, 
				    RFSAttributes & attributes,
				    int create )
	: FileObject
		( attributes.size == 0 ? 
		  0 : ((attributes.size - 1) >> attributes.log2UnitSize) + 1,
		  attributes.log2UnitSize, handle._idNumber, container, 
		  0, 0 ),
	  _handle( handle ), _attributes( attributes )
{
	initDebug();

	Raids( RaidConstructor ) << this
		<< "::RemoteFileObject( "
		<< "idNumber " << _idNumber 
		<< " )\n" << eor;

	_unusedUnit = numberOfUnits();
	_modified = 0;
	_segmentSize = RemoteFileSystemMaxDataSize;
	if ( create ) {
		_hints = RemoteFileObjectDefaultHint;
		attributes.cacheHints = _hints << 16;
	} else {
		_hints = attributes.cacheHints;
	}
	MemoryObject::basicSetNumberOfUnits( _numberOfUnits );
}

RemoteFileObject::~RemoteFileObject()
{
	syncAttributes( 1, 1, theDummyAsyncClosure );
}

void
RemoteFileObject::setModified()
{
	_hintLock.acquire();
	_attributes.flags |= RFSAttributes::AttrDirty;
	_hintLock.release();
}

void
RemoteFileObject::clearModified()
{
	_hintLock.acquire();
	_attributes.flags &= ~RFSAttributes::AttrDirty;
	_hintLock.release();
}

void
RemoteFileObject::syncAttributes( int commit, int close, IOAsyncClosure * c )
{
#ifdef STATEFULL
	if ( close || (_attributes.flags & RFSAttributes::AttrDirty) ) {
#else
	if ( _attributes.flags & RFSAttributes::AttrDirty ) {
#endif
//		if ( isDictionary() ) {
//			commit = 0;
//		}
		commit = 0;
		int status = 0;
        	status = 
		RemoteFileSystemRPC::sync
			( _handle, _attributes, commit, close, c );
		clearModified();
		Raids( 64 ) 
			<< this << "::syncAttributes() clearModified\n" << eor;
	} else {
		if ( c != 0 )
			c->activate();
	}
}

typedef CacheIOState RemoteFileObjectAsyncState;
typedef CacheIOClosure RemoteFileObjectAsyncClosure;

class RemoteFileObjectRawReadAsyncClosure : public IOAsyncClosure {
protected:
	IOAsyncClosure * _parent;
	unsigned _done;
	static void func( RemoteFileObjectRawReadAsyncClosure * c );
public:
	RemoteFileObjectRawReadAsyncClosure
		( IOAsyncClosure * parent, unsigned done )
		: _parent( parent ), _done( done ), _func( &func ) {}
};

void
RemoteFileObjectRawReadAsyncClosure::func
	( RemoteFileObjectRawReadAsyncClosure * c )
{
	c->_parent->activate( c->_result + c->_done );
	delete c;
}

static int dummy;

int 
RemoteFileObject::rawRead( unsigned startUnit, int count, char * dest, 
			   int nbytes, IOAsyncClosure * c )
{
	Raids( RaidProtected ) 
		<< this << "::rawRead( " 
	        << startUnit << ", " << count << ", " 
		<< asHex((int) dest) << ", " 
		<< nbytes << " )\n" << eor;

	int total = 0;
	int noOfUnits = numberOfUnits();

	/* check if reading pass end of file object */

	Assert( startUnit + count <= noOfUnits );
	int orgCount = count;

	/* perform fill for parts no written to yet */

	int pastLast = startUnit + count;
	if ( pastLast > _unusedUnit ) {
		int fill = pastLast - 
			( _unusedUnit > startUnit ? _unusedUnit : startUnit );
		Assert( fill > 0 && fill <= count );
		total += fill;
		count -= fill;
		int skipBytes = ( count << _log2UnitSize );
		int fillBytes = nbytes - skipBytes;
		if ( fillBytes < 0 ) {
			fillBytes = 0;
		} else {
			ByteZero( dest + skipBytes, fillBytes );
			nbytes -= fillBytes;
		}	
	}

	/* check if done */

	if ( count <= 0 || nbytes <= 0 ) {
		if ( c ) 
			c->activate( total );
		return total;
	}
		
	/* if only single RPC required */

	if ( nbytes <= RemoteFileSystemMaxDataSize ) {
		if ( c != 0 && total != 0 ) {
			c = new RemoteFileObjectRawReadAsyncClosure( c, total );
		}
		int actualNoOfUnits = count;
		Raids( 64 ) 
			<< this << "::rawRead( " << startUnit << ","
			<< count << " ) do RPC\n" << eor;
		RemoteFileSystemRPC::read
			( this, _handle, startUnit, count, dest, nbytes,
			  &actualNoOfUnits, c );
		return actualNoOfUnits + total;
	}

	/* fragment request if more than one RPC */

	RemoteFileObjectAsyncState * state = 
		new RemoteFileObjectAsyncState( c );	

	int noOfBytes;

	int firstUnit = startUnit;
	int remainingBytes = nbytes;
	int remainingUnits = count;

	int maxSize = RemoteFileSystemMaxDataSize;
	Assert( unitSize() <= maxSize );

	while ( remainingBytes > 0 && remainingUnits > 0 ) {
		noOfBytes = ( remainingBytes > maxSize ? 
			     maxSize : remainingBytes);
		noOfUnits = ( ( noOfBytes - 1 ) >> log2UnitSize() ) + 1; 

		Raids( RaidDetail ) 
			<< this << "::rawRead() "
			<< firstUnit << ", " << noOfUnits << ", "
			<< noOfBytes << ", " 
			<< asHex( (unsigned) dest )
			<< " )\n" << eor;

		state->add( firstUnit, noOfUnits, dest, noOfBytes );

		dest += noOfBytes;
		remainingBytes -= noOfBytes;
		remainingUnits -= noOfUnits;
		firstUnit += noOfUnits;
	}

        state->_result = total;
        state->_expectedResult = orgCount;
	state->_spare = (unsigned) &(RemoteFileObject::rawRead);
     	RemoteFileObjectAsyncState::_Operation * op = state->_ops;
	int opCount = state->_count;
	Assert( opCount > 1 );
   	for( int i = 0; i < opCount; i++ ) {
		Raids( 64 ) 
			<< this << "::rawRead( " << op->_startUnit << ","
			<< op->_count << " ) do RPC n\n" << eor;
	    	RemoteFileSystemRPC::read
			( this, _handle, op->_startUnit, op->_count, op->_buf,
			  op->_spare, &dummy, &(op->_closure) );
	    	op++;
        }
        if ( c == 0 ) {
    	    	state->_barrier.wait();
	    	total = state->_result;
	    	delete state;
        }

	Raids( RaidProtected )
		<< this << "::rawRead return( " << total << " )\n" << eor;

	return total;
}

int 
RemoteFileObject::rawWrite( unsigned int startUnit, int count, char * src, 
			    int nbytes, int writeThrough, IOAsyncClosure * c )
{
	Raids( RaidProtected ) 
		<< this << "::rawWrite( " 
	        << startUnit << ", " << count << ", " 
		<< asHex((int) src) << ", " 
		<< nbytes << ", " << writeThrough << " )\n" << eor;

	/* check if reading pass end of file object */

	Assert( startUnit + count <= numberOfUnits() );

	/* check if done */

	if ( count <= 0 || nbytes <= 0 ) {
		if ( c ) 
			c->activate( 0 );
		return 0;
	}
		
	_modified = 1;
	if ( startUnit + count > _unusedUnit ) 
		_unusedUnit = startUnit + count;

	/* if only single RPC required */

	if ( nbytes <= RemoteFileSystemMaxDataSize ) {
		int actualNoOfUnits = count;
		Raids( 64 ) 
			<< this << "::rawWrite( " << startUnit << ","
			<< count << " ) do RPC\n" << eor;
		RemoteFileSystemRPC::write 
			( this, _handle, 
			  startUnit, count, src, nbytes, writeThrough,
			  &actualNoOfUnits, c );
		return actualNoOfUnits;
	}

	/* fragment request if more than one RPC */

	RemoteFileObjectAsyncState * state = 
		new RemoteFileObjectAsyncState( c );	

	int noOfBytes;
	int noOfUnits;

	int firstUnit = startUnit;
	int remainingBytes = nbytes;
	int remainingUnits = count;

	int maxSize = RemoteFileSystemMaxDataSize;
	Assert( unitSize() <= maxSize );

	while ( remainingBytes > 0 && remainingUnits > 0 ) {
		noOfBytes = ( remainingBytes > maxSize ? 
			     maxSize : remainingBytes);
		noOfUnits = ( ( noOfBytes - 1 ) >> log2UnitSize() ) + 1; 

		Raids( RaidDetail ) 
			<< this << "::rawWrite() "
			<< firstUnit << ", " << noOfUnits << ", "
			<< noOfBytes << ", " 
			<< asHex( (unsigned) src )
			<< " )\n" << eor;

		state->add( firstUnit, noOfUnits, src, noOfBytes );

		src += noOfBytes;
		remainingBytes -= noOfBytes;
		remainingUnits -= noOfUnits;
		firstUnit += noOfUnits;
	}

        state->_result = 0;
        state->_expectedResult = count;
	state->_spare = (unsigned) &(RemoteFileObject::rawWrite);
     	RemoteFileObjectAsyncState::_Operation * op = state->_ops;
	int opCount = state->_count;
	Assert( opCount > 1 );
   	for( int i = 0; i < opCount; i++ ) {
		Raids( 64 ) 
			<< this << "::rawWrite( " << op->_startUnit << ","
			<< op->_count << " ) do RPC n\n" << eor;
	    	RemoteFileSystemRPC::write
			( this, _handle, op->_startUnit, op->_count, op->_buf,
			  op->_spare, writeThrough,
			  &dummy, &(op->_closure) );
	    	op++;
        }
	int total = count;
        if ( c == 0 ) {
    	    	state->_barrier.wait();
	    	total = state->_result;
	    	delete state;
        }

	Raids( RaidProtected )
		<< this << "::rawWrite return( " << total << " )\n" << eor;

	return total;
}

void
RemoteFileObject::basicSynchronizeClosure
	( PersistentMemoryObject::BasicSynchronizeAsyncClosure * c )
{
	Assert( NOTREACHED );
}

void
RemoteFileObject::basicSynchronize( IOAsyncClosure * c )
{
//	if ( c ) {
		MemoryObject::basicSynchronize( c );
//	} else {
//		FileObject::basicSynchronize( c );
//	}
}

void
RemoteFileObject::basicSynchronize( unsigned from, unsigned to,
				    IOAsyncClosure * c )
{
//	if ( c ) {
		MemoryObject::basicSynchronize( from, to, c );
//	} else {
//		FileObject::basicSynchronize( from, to, c );
//	}
}

int
RemoteFileObject::links( int i ) 
{
	// Console() << "RemoteFileObject::links( " << i << " )\n" << eor;
	_attributes.links += i;
	if ( _attributes.links < 0 ) _attributes.links = 0;
	return _attributes.links;
}

Class *
RemoteFileObject::supports( Class * aClass )
{
	Raid( RaidProtected )( "%N::supports( %N )\n", this, aClass );

	Class * supportedClass = 0;

	extern ClassStar PersistentArrayClass;
	extern ClassStar SymbolicLinkClass;
	extern ClassStar RemoteDictionaryClass;

	if( PersistentArrayClass->isASubclassOf( aClass ) &&
	    ( _attributes.mode & UNIXTypeMask) == UNIXTypeRegular )
		supportedClass = PersistentArrayClass;
	else if( RemoteDictionaryClass->isASubclassOf( aClass ) && 
		 ( _attributes.mode & UNIXTypeMask ) == UNIXTypeDirectory )
		supportedClass = RemoteDictionaryClass;
        else if( SymbolicLinkClass->isASubclassOf( aClass ) &&
		 ( _attributes.mode & UNIXTypeMask ) == UNIXTypeSymbolicLink )
		supportedClass = SymbolicLinkClass;

	Raid( RaidProtected )( "%N::supports() returns %N %d\n",
			       this, supportedClass );
	
	return( supportedClass );
}

//int RemoteFileObjectCacheFlag = O_LCACHE;
//int RemoteFileObjectCacheFlag = 0;

int
RemoteFileObject::parseFlags( int flags )
{
    	Raid( RaidProtected )( "%N::parseFlags( %x )\n", this, flags );
    	if ( ( flags & O_CMASK ) == 0 ) {
       		int i = _attributes.log2BlockSize + _log2UnitSize - 12 /* 4K */;
		if ( i > 0 ) {
	    		flags |= ( i << 28 );
		}
      	}
    	if ( isDictionary() || isContainer() || !isSymbolicLink() ) {
        	flags = (flags | O_MULTI | O_PERM) & ~(O_SINGLE | O_TEMP);
#ifdef NEVER
	} else if ( (flags & O_NCACHE) == 0 ) {
			flags |= RemoteFileObjectCacheFlag;
		}
#endif
	}

    	return FileObject::parseFlags( flags );
}

MemoryObjectCache *
RemoteFileObject::buildCache()
{
	if ( ( _attributes.mode & UNIXTypeMask ) == UNIXTypeDirectory )
		return 0;
	return FileObject::buildCache();
}

int
RemoteFileObject::isSymbolicLink()
{
	return( ( _attributes.mode & UNIXTypeMask ) == UNIXTypeSymbolicLink );
}

int
RemoteFileObject::isDictionary()
{
	return( ( _attributes.mode & UNIXTypeMask ) == UNIXTypeDirectory );
}

int
RemoteFileObject::isContainer()
{
	return( _attributes.type.type() == Container );
}

int 
RemoteFileObject::isRemote()
{
	return 1;
}

int
RemoteFileObject::modified()
{
	return _modified;
}

int
RemoteFileObject::fixedSize()
{
	return _attributes.flags & RFSAttributes::FixedSize;
}

unsigned
RemoteFileObject::numberOfBytes()
{
	return (unsigned)_attributes.size;
}

unsigned
RemoteFileObject::basicSetNumberOfBytes( unsigned size )
{
	// Should only allow if open for write.

	Raid( RaidProtected )( "%N::basicSetNumberOfBytes( size %d ) was %d\n",
			        this, size, numberOfBytes() );

	if ( size == _attributes.size )
		return size;

	_attributes.size = size;

	int noOfUnits;
	if (size == 0) 
		noOfUnits = 0;
	else 
		noOfUnits = ((size - 1) >> log2UnitSize()) + 1;

	int orgNoOfUnits = numberOfUnits();
	if ( orgNoOfUnits != noOfUnits ) {
	        MemoryObject::basicSetNumberOfUnits( noOfUnits );
	}
	
	setModified();
	Raids( 64 ) 
		<< this << "::basicSetNumberOfUnits() setModified\n" << eor;

	return size;
}

unsigned
RemoteFileObject::basicSetNumberOfUnits( unsigned units )
{
	// Should only allows if open for write.

	Raid( RaidProtected )( "%N::basicSetNumberOfUnits( units %d ) was %d\n",
			       this, units, numberOfUnits() );

	if ( numberOfUnits() != units ) 
		basicSetNumberOfBytes( units << log2UnitSize() );

	return units;
}

int
RemoteFileObject::setMode( int accessMode )
{
	// Should only allows if open for write.

	if ( (_attributes.mode & UNIXAccessMask) == 
	     (accessMode & UNIXAccessMask) ) 
		return 0;
	
        _attributes.mode = ( _attributes.mode & ~UNIXAccessMask ) |
			   ( accessMode & UNIXAccessMask );
	setModified();
	Raids( 64 ) 
		<< this << "::setMode() setModified\n" << eor;
	return 0;
}

int
RemoteFileObject::setUserId( int uid )
{
	// Should only allow if open for write

	if ( _attributes.uid == uid )
		return 0;
	_attributes.uid = uid;
	setModified();
	Raids( 64 ) 
		<< this << "::setUserId() setModified\n" << eor;
	return 0;
}

int
RemoteFileObject::setGroupId( int gid )
{
	// Should only allow if open for write

	if ( _attributes.gid == gid )
		return 0;
	_attributes.gid = gid;
	setModified();
	Raids( 64 ) 
		<< this << "::setGroupId() setModified\n" << eor;
	return 0;
}

int
RemoteFileObject::basicCopy( MemoryObject * dest )
{
	return( PersistentMemoryObject::basicCopy( dest ) );
}

int
RemoteFileObject::accessMode()
{
    	return( _attributes.mode );
}

int
RemoteFileObject::userId()
{
	return( _attributes.uid );
}

int
RemoteFileObject::groupId()
{
	return( _attributes.gid );
}

int
RemoteFileObject::transient()
{
	return ( _attributes.links == 0 );
}

void
RemoteFileObject::writeIdNumber( OutputStream * os )
{
	PersistentMemoryObject::writeIdNumber( os );
}

void
RemoteFileObject::info( UNIXFileInfo * infoBuffer )
{
    	Raid( RaidMethod )( "%N::info( ) %d\n", this, _idNumber );
	infoBuffer->device	= 0;
	infoBuffer->inumber	= _idNumber;
	infoBuffer->mode	= _attributes.mode;
	infoBuffer->links	= _attributes.links;
	infoBuffer->uid	        = _attributes.uid;
	infoBuffer->gid	        = _attributes.gid;
	infoBuffer->flags	= 0;	//FIX
	infoBuffer->rdevice	= 0;	//FIX
	infoBuffer->size	= _attributes.size;
	infoBuffer->atime	= _attributes.atime;
	infoBuffer->mtime	= _attributes.mtime;
	infoBuffer->ctime	= _attributes.ctime;
	unsigned shift = _attributes.log2BlockSize + _attributes.log2UnitSize;
	infoBuffer->blockSize   = 1 << shift;
	infoBuffer->blocks	= 
		( infoBuffer->size == 0 ? 0 :
		  ( ( infoBuffer->size - 1 ) >> shift ) + 1);
}

void
RemoteFileObject::hintsChanged( unsigned oldHints )
{
	setModified();
}

void
RemoteFileObject::getAttributes( RFSAttributes * attr )
{
        if ( !(_attributes.flags & RFSAttributes::AttrDirty ) ) {
		attr->flags = 0;
		return;
	}
	// more eagerly update attributes to avoid additional message
	*attr = _attributes;
	attr->cacheHints = _hints << 16;
	clearModified();
	Raids( 64 ) 
		<< this << "::getAttributes() clearModified\n" << eor;
}

void
RemoteFileObject::setAttributes( RFSAttributes * attr )
{
	if ( !(attr->flags & RFSAttributes::AttrDirty) )
		return;
        if ( _attributes.size != attr->size ) {
		if ( isDictionary() ) {
			_attributes.size = attr->size;
		} else {
                	setNumberOfBytes( attr->size );
		}
	}
        if ( _attributes.uid != attr->uid )
		setUserId( attr->uid );
        if ( _attributes.gid != attr->gid )
                setGroupId( attr->gid );
        if ( (_attributes.mode & UNIXAccessMask) !=
             (attr->mode & UNIXAccessMask) )
                setMode( attr->mode & UNIXAccessMask );
	unsigned remoteHints = ( attr->cacheHints >> 16 ) & 0xffff; 
	if ( remoteHints != _hints ) {
		_hintLock.acquire();
		unsigned oldHints = _hints;
		if ( remoteHints & H_LCACHE ) {
			if ( !(_openFlags & O_NLCACHE) ) {
				_hints |= H_LCACHE;
			}
		} else {
			if ( !(_openFlags & O_LCACHE) ) {
				_hints &= ~H_LCACHE;
			}
		}
		if ( remoteHints & H_SINGLE ) {
			if ( !(_openFlags & O_MULTI) ) {
				_hints |= H_SINGLE;
			}
		} else {
			if ( !(_openFlags & O_SINGLE) ) {
				_hints &= ~H_SINGLE;
			}
		}
		if ( remoteHints & H_LARGE ) {
			if ( !(_openFlags & O_SMALL) ) {
				_hints |= H_LARGE;
			}
		} else {
			if ( !(_openFlags & O_LARGE) ) {
				_hints &= ~H_LARGE;
			}
		}
		if ( remoteHints & H_RANDOM ) {
			if ( !(_openFlags & O_SEQ) ) {
				_hints |= H_RANDOM;
			}
		} else {
			if ( !(_openFlags & O_RANDOM) ) {
				_hints &= ~H_RANDOM;
			}
		}
		if ( remoteHints & H_TEMP ) {
			if ( !(_openFlags & O_PERM) ) {
				_hints |= H_TEMP;
			}
		} else {
			if ( !(_openFlags & O_TEMP) ) {
				_hints &= ~H_TEMP;
			}
		}
		unsigned newHints = _hints;
		_hintLock.release();
		if ( newHints != oldHints ) {
			hintsChanged( oldHints );
		}
	}
}

char namebuf[128] = "UNIXInodeIndirect.cc";

void RemoteFileObject::setName( const char * str )
{
	FileObject::setName( str );
#ifdef NEVER
	if ( strcmp(str,namebuf) == 0 ) {
		setDebugMask(255);
		setDebugMessages(32767);
	}
#endif 
}

#include "ClassInstaller.h"
DEFINECLASS("RemoteFileObject",RemoteFileObject,FileObject,0,0);

