#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * RemoteFileSystemRPC.cc
 */

//#define DEBUG
#define RFSTIME

#include "Byte.h"
#include "Debug.h"
#include "FileSystem.h"
#ifdef RFSTIME
#include "FreeRunningTimer.h"
#endif
#include "Kernel.h"
#include "MemoryObjectAsync.h"
#include "Process.h"
#include "PROCESSOR.h"
#include "RemoteFileObject.h"
#include "RemoteFileSystem.h"
#include "RemoteFileSystemRPC.h"
#include "RemoteFileSystemMessages.h"
#include "RemoteTables.h"
#include "ResourceLock.h"
#include "Semaphore.h"
#include "SocketDef.h"
#include "SocketKernel.h"
#include "SoftClock.h"
#include "String.h"
#include "ThisProcess.h"
#include "TimeoutTimer.h"

#ifdef __GNUC__
#pragma implementation
#endif

#define LIMITVOLUME
// #define LIMITCOUNT
// #define MAXCOUNT 8
// #define DELAYSHIFT

const int defaultRetrySec = 2;
const int defaultMaxRetrySec = 60;

int RemoteFileSystemRPC::_debugMask = 0;
int RemoteFileSystemRPC::_debugMessages = 0;

char * RFSMessageTypeStrings[RFSLast] = {
	"Mount        ",
	"COpen        ",
	"CCreate      ",
	"Open         ",
	"Create       ",
	"Keys         ",
	"Remove       ",
        "Add          ",
        "Read         ",
        "Write        ",
        "Sync         ",
        "Assoc        ",
        "MountReply   ",
        "COpenReply   ",
        "CCreateReply ",
        "OpenReply    ",
        "CreateReply  ",
        "KeysReply    ",
        "RemoveReply  ",
        "AddReply     ",
        "ReadReply    ",
        "WriteReply   ",
        "SyncReply    ",
        "AssocReply   "
};

class RemoteFileSystemStatistics {
	const unsigned _size = RFSLast;
	struct statInfo {
		unsigned _count;
		unsigned _volume;
#ifdef RFSTIME
		unsigned _time;
#endif
	} _info[_size];
public:
	RemoteFileSystemStatistics();
	void reset();
	void info(OutputStream & s);
#ifdef RFSTIME
	void add(RFSMessage * msg, unsigned t = 0);
#else
	void add(RFSMessage * msg);
#endif
	void cmd(int argc, char *argv[]);
};

RemoteFileSystemStatistics::RemoteFileSystemStatistics()
{
	reset();
}

void
RemoteFileSystemStatistics::reset()
{
	for (int i = 0; i < _size; i++ ) {
		_info[i]._count =
		_info[i]._volume = 0;
#ifdef RFSTIME
		_info[i]._time = 0;
#endif
	}
}

void 
RemoteFileSystemStatistics::info(OutputStream & s)
{
	Console() << "RemoteFileSystem RPC statistics\n" << eor;
	unsigned sumCount = 0;
	unsigned sumVolume = 0;
#ifdef RFSTIME
	unsigned sumTime = 0;
#endif
	for (int i = 0; i < _size; i++ ) {
#ifdef RFSTIME
		Console().formattedWrite
			("%14s %12d %12d %12d\n",
 			 RFSMessageTypeStrings[i], 
			 _info[i]._count, _info[i]._volume,
			 _info[i]._time );
#else
		Console().formattedWrite
			("%14s %12d %12d\n",
 			 RFSMessageTypeStrings[i], 
			 _info[i]._count, _info[i]._volume );
#endif 
		Console().flush();
		sumCount += _info[i]._count;
		sumVolume += _info[i]._volume;
#ifdef RFSTIME
		sumTime += _info[i]._time;
#endif
	}
#ifdef RFSTIME
	Console().formattedWrite
		("%-14s %12d %12d %12d\n","Sum",sumCount,sumVolume,sumTime);
#else
	Console().formattedWrite("%-14s %12d %12d\n","Sum",sumCount,sumVolume);
#endif
	Console().flush();
}

inline
void
#ifdef RFSTIME
RemoteFileSystemStatistics::add(RFSMessage * msg, unsigned t)
#else
RemoteFileSystemStatistics::add(RFSMessage * msg)
#endif
{
	RFSMessageType type = msg->_messageType;
	unsigned volume = msg->_headerSize + msg->_dataSize;
	_info[(unsigned) type]._count++;
	_info[(unsigned) type]._volume += volume;
#ifdef RFSTIME
	_info[(unsigned) type]._time += t;
#endif
}

void
RemoteFileSystemStatistics::cmd(int argc, char *argv[])
{
	if ( argc <= 1 ) {
		info(Console());
		return;
	}
	char * opt = argv[1];
	while (*opt) {
		switch(*opt) {
		case 'r': reset(); break;
		case 'i': info(Console()); break;
		default : Console() << argv[0] << ": unknown option " << *opt
			  << "\n" << eor;
		}	
		opt++;
	}
}

RemoteFileSystemStatistics RFSStatistics;

void RemoteFileSystemStatistics_cmd(int argc, char *argv[])
{
	RFSStatistics.cmd(argc,argv);
}

struct OutstandingRPC {
	unsigned _sequenceNumber;
	int _socket;
	int _retries;
	int _maxRetries;
	unsigned _reserved;
#ifdef RFSTIME
	unsigned _time;
#endif
	msghdr * _msg;
	Process * _process;
	IOAsyncClosure * _closure;

	OutstandingRPC() : _sequenceNumber(0), _socket(-1) {}
};

class ToRedo {
	Semaphore _sem;
	Lock _lock;
	OutstandingRPC ** _head;
	OutstandingRPC ** _tail;
	OutstandingRPC ** _end;
	OutstandingRPC * _redoRPC[RemoteFileSystemMaxOutstandingRPC + 1];
public:
	ToRedo();
	void add( OutstandingRPC * rpc );
	OutstandingRPC * remove();
};

inline
ToRedo::ToRedo()
	: _sem( 0 )
{
	_head = _tail = _redoRPC;
	_end = _redoRPC + (RemoteFileSystemMaxOutstandingRPC + 1);
}

inline
void
ToRedo::add( OutstandingRPC * rpc )
{
	_lock.acquire();
	*_tail = rpc;
	_tail++;
	if ( _tail == _end )
		_tail = _redoRPC;	
	Assert( _tail != _head );
	_lock.release();
	_sem.V();
}

inline
OutstandingRPC *
ToRedo::remove()
{
	_sem.P();
	_lock.acquire();
	Assert( _head != _tail );
	OutstandingRPC * ret = *_head;
	_head++;
	if ( _head == _end )
		_head = _redoRPC;
	_lock.release();
	return ret;
}

static int _hz = 0;
static int _retryHz = 0;
static int _maxRetryHz = 0;
static int _retries = 0;
static int _redoInterval = 500000; /* msec, for now 0.5 sec */
static int _limit = 0;
static Process * _clientAsyncProcess = 0;
static Process * _clientRedoProcess = 0;
static int _clientSocket = -1;
static unsigned _sequenceNumber = 1;
static OutstandingRPC _outstandingRPC[RemoteFileSystemMaxOutstandingRPC];
static Lock _lock;
static ToRedo * _toRedo = 0;
#ifdef LIMITVOLUME
static ResourceLock * _limitLock = 0;
#endif
#ifdef LIMITCOUNT
static int _reserved = 0;
static Semaphore * _limitSem = 0;
#endif

class ServerAddressCache {
	const int _size = 31;
	struct Entry {
		Entry * _next;
		unsigned _host;
		sockaddr_in _inaddr;
	};
	Entry *_table[_size];
public:
	ServerAddressCache();
	sockaddr * get( unsigned host );
};

ServerAddressCache::ServerAddressCache()
{
	for( int i = 0; i < _size; i++ )
		_table[i] = 0;
}

sockaddr *
ServerAddressCache::get( unsigned host )
{
	int slot = host % _size;
	Entry ** p = _table + slot;
	Entry * e = *p;
	while ( e && e->_host != host ) {
		p = &(e->_next);
		e = *p;
	}
	if ( e == 0 ) {
		e = new Entry();
		*p = e;
		e->_host = host;
		e->_next = 0;
		sockaddr_in * inaddr = &(e->_inaddr);
		unsigned * p = (unsigned *) inaddr;
		p[2] = p[3] = 0;
		inaddr->sin_family = AF_INET;
		inaddr->sin_port = RemoteFileSystemServerPort;
		inaddr->sin_addr.s_addr = host;
	}
	return (sockaddr *) &(e->_inaddr);
}

ServerAddressCache _serverAddressCache;

int
buildSocket( unsigned long host, int port )
{
	int errno;
	extern int MaxMbuf;
	int reserve = MaxMbuf / 2;
	sockaddr_in insockaddr;
	ByteZero(&insockaddr, sizeof(insockaddr));
	insockaddr.sin_family = AF_INET;
	insockaddr.sin_port = port;
	insockaddr.sin_addr.s_addr = host;
	
	/* the following should be in subroutine but who cares for now */
	int s = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP, errno );
	if ( s < 0 ) {
		Console() 
			<< "RemoteFileSystemRPC::buildSocket( " 
			<< asHex( host ) << ", "
			<< port << " ) socket failed errno " 
			<< errno << "\n" << eor;
		return -1;
	}
	int result = setsockopt( s, SOL_SOCKET, SO_SNDBUF, 
			         (caddr_t) &reserve, sizeof(reserve), errno );
	if ( result < 0 ) {
		Console() 
			<< "RemoteFileSystemRPC::buildSocket( "
			<< asHex( host ) << ", "
			<< port << " ) setsockopt send buf failed errno "
			<< errno << "\n" << eor;
		return -1;
	}
	result = setsockopt( s, SOL_SOCKET, SO_RCVBUF, 
			         (caddr_t) &reserve, sizeof(reserve), errno );
	if ( result < 0 ) {
		Console() 
			<< "RemoteFileSystemRPC::buildSocket( "
			<< asHex( host ) << ", "
			<< port << " ) setsockopt recv buf failed errno "
			<< errno << "\n" << eor;
		return -1;
	}
	result = bind( s, (sockaddr *)&insockaddr, sizeof(insockaddr),
		       errno );
	if ( result < 0 ) {
		Console() 
			<< "RemoteFileSystemRPC::buildSocket( "
			<< asHex( host ) << ", "
			<< port << " ) bind failed errno " << errno 
			<< "\n" << eor;
		return -1;
	}
	return s;
}

int 
clientSocket( int port )
{
	unsigned long clienthostaddr = inet_myaddr();
	return buildSocket( clienthostaddr, port );	
}

int
RemoteFileSystemRPC::clientInit()
{
	static Lock lock;
	static int initialized = 0;
	lock.acquire();
	if ( initialized ) {
		lock.release();
		return 0;
	}
	initialized = 1;
	lock.release();

	extern int MaxMbuf;
#ifdef LIMITVOLUME
	_limit = ( MaxMbuf / 4 );
	_limitLock = new ResourceLock( _limit );
#endif // LIMITVOLUME
#ifdef LIMITCOUNT
#ifdef MAXCOUNT
        _limit = MAXCOUNT;
#else 
	_limit = ( MaxMbuf / 4 ) / RemoteFileSystemMaxDataSize;
	Assert( _limit >= 8 );
#endif // MAXCOUNT
	int rpclimit = ( RemoteFileSystemMaxOutstandingRPC * 2 ) / 3;
	if ( _limit > rpclimit )
		_limit = rpclimit;
	_limitSem = new Semaphore( _limit );
#endif // LIMITCOUNT
	Console() 
                << "RFS client max data size " << RemoteFileSystemMaxDataSize
		<< " max mbuf space " << MaxMbuf
		<< " rpc reserve limit " << _limit
                << "\n" << eor;

	Assert( _toRedo == 0 );
	_toRedo = new ToRedo();
	Assert( _toRedo != 0 );
	Assert( TheRemoteContainerRegistry == 0 );
	TheRemoteContainerRegistry = new RemoteContainerRegistry();
	Assert( TheRemoteContainerRegistry != 0 );
	Assert( TheRemotePathCache == 0 );
	TheRemotePathCache = new RemotePathCache();
	Assert( TheRemotePathCache != 0 );

	int port = RemoteFileSystemClientPort;
	_clientSocket = clientSocket( port );
	Assert( _clientSocket >= 0 );	
	port++;
	for( int i = 0; i < RemoteFileSystemMaxOutstandingRPC; i++ ) {
		_outstandingRPC[i]._socket = clientSocket( port );
		Assert( _outstandingRPC[i]._socket >= 0 );
		port++;
	}
	_hz = TheSoftwareClockManager->hz();
	_retryHz = ( _hz * defaultRetrySec ); 
	_maxRetryHz = _hz * defaultMaxRetrySec; 
        void clientAsyncRPCProcess();
        _clientAsyncProcess = new SystemProcess
                ( &clientAsyncRPCProcess, TheKernel->domain(), tiny, 0 );
        Assert( _clientAsyncProcess != 0 );
        _clientAsyncProcess->reference();
	_clientAsyncProcess->setName( "RFS Client Async" );
	_clientAsyncProcess->ready();
        Console()
                << "RFS client async process " 
		<< _clientAsyncProcess << "\n" << eor;
        void clientRedoRPCProcess();
        _clientRedoProcess = new SystemProcess
                ( &clientRedoRPCProcess, TheKernel->domain(), tiny, 0 );
        Assert( _clientRedoProcess != 0 );
        _clientRedoProcess->reference();
	_clientRedoProcess->setName( "RFS Client Redo" );
	_clientRedoProcess->ready();
        Console()
                << "RFS client redo process " 
		<< _clientRedoProcess << "\n" << eor;
	return 0;
}

void reserveBufSize( unsigned res )
{
#ifdef LIMITVOLUME
	_limitLock->acquire( res );
#endif // LIMITVOLUME
#ifdef LIMITCOUNT
	_limitSem->P();	
#endif // LIMITCOUNT
}

void freeBufSize( unsigned res )
{
#ifdef LIMITVOLUME
	_limitLock->release( res );
#endif // LIMITVOLUME
#ifdef LIMITCOUNT
	_reserved--;
	_limitSem->V();
#endif // LIMITCOUNT
}

int debugRPCRetry = 1;

void
redoRPCTimerExpired( OutstandingRPC * outrpc, int ticks )
{
	_toRedo->add( outrpc );
}

inline 
void 
redoRPC( OutstandingRPC * outrpc )
{
	if ( outrpc->_sequenceNumber == 0 )
		return;
	_lock.acquire();
	_retries++;
	outrpc->_retries++;
	int myretries = outrpc->_retries;
	if ( myretries > outrpc->_maxRetries ) {
		Console()
			<< "RFS client seq no " 
			<< outrpc->_sequenceNumber << " has "
			<< outrpc->_retries 
			<< " retries, exceeding max retries "
			<< outrpc->_maxRetries << "\n" << eor;
		outrpc->_retries = 0;
		outrpc->_sequenceNumber = 0;
		Process * p = outrpc->_process;
		IOAsyncClosure * c = outrpc->_closure;
		if ( outrpc->_reserved ) {
			freeBufSize( outrpc->_reserved );
		}
		_lock.release();
		if ( c == 0 ) {
			process_intr( p );
		} else {
			c->activate( -1 );
		}
		return;
	}
#ifdef DELAYSHIFT
#ifdef LIMITVOLUME
	int shift = ( _limitLock->waiting() > 0 ? 1 : 0 );
#endif // LIMITVOLUME
#ifdef LIMITCOUNT
	int shift = ( outrpc->_reserved > 0 && _reserved >= _limit ? 1 : 0 );
#endif // LIMITCOUNT
	if ( debugRPCRetry && shift ) Console()
		<< "RFS client retry shift " << shift << "\n" << eor;
	int delay = ( _retryHz * ( 1 << ( myretries - 1 ) ) ) << shift;
#else // no DELAYSHIFT
	int delay = ( _retryHz * ( 1 << ( myretries - 1 ) ) );
#endif // DELAYSHIFT
	if ( delay > _maxRetryHz )
		delay = _maxRetryHz;
	RFSMessage * reqmsg = (RFSMessage *) outrpc->_msg->msg_iov[0].iov_base;
	if ( debugRPCRetry >= myretries ) {
		int size = reqmsg->_headerSize + reqmsg->_dataSize;
		Console()
		<< "RFS client retrying RPC seq " 
		<< outrpc->_sequenceNumber 
		<< " retry " << myretries 
		<< " delay " << delay 
		<< " req size " << size 
		<< " type " 
		<< RFSMessageTypeStrings[(unsigned) reqmsg->_messageType]
		<< "\n" << eor;	
	}
	RFSStatistics.add( reqmsg );
	TheSoftwareClockManager->add
		( (SoftClockFunc) &redoRPCTimerExpired, (char *) outrpc, 
		  delay );
	int socket = 
		( outrpc->_closure == 0 ? outrpc->_socket : _clientSocket );
	msghdr * msg = outrpc->_msg;
	_lock.release();
	int errno;
	int result = sendmsg( socket, msg, 0, errno );
	if  ( result < 0 ) {
		Console() 
			<< "RFS client redo RPC sendmsg errno " 
			<< errno << "\n" << eor;
	}
}

void
clientRedoRPCProcess()
{
	TimeoutTimer timer;
	while( 1 ) {
		OutstandingRPC * outrpc = _toRedo->remove();
		redoRPC( outrpc );
		timer.start( _redoInterval );
		timer.await();
	}
}

void 
discardReply( int socket, RFSMessage * rfshdr )
{
	int errno;
	int result = recv( socket, (caddr_t) rfshdr, sizeof( RFSMessage ),
		       0, errno );
	if ( result < 0 ) {
		Console() 
			<< "RFS client discard duplicate reply errno " 
			<< errno << "\n" << eor;
	}
}

int
doRPC( unsigned server, msghdr * msg, IOAsyncClosure * c, unsigned reserve,
       unsigned maxRetries = 20 )
{
	if ( reserve )
		reserveBufSize( reserve );
	_lock.acquire();
#ifdef LIMITCOUNT
	if ( reserve )
		_reserved++;
#endif // LIMITCOUNT
	msg->msg_name = (caddr_t)
		_serverAddressCache.get( server );
	msg->msg_namelen = sizeof( sockaddr_in );
	Assert( msg->msg_iov != 0 && msg->msg_iov[0].iov_base != 0 );
	Assert( msg->msg_iov[0].iov_len >= sizeof( RFSMessage ) );
	msg->msg_control = 0;
	msg->msg_controllen = 0;
	msg->msg_flags = 0;
	
	RFSMessage * rfsmsg = (RFSMessage *) msg->msg_iov[0].iov_base;
	if (msg->msg_iovlen == 2 && rfsmsg->_messageType != RFSWrite) {
		unsigned size = rfsmsg->_headerSize + rfsmsg->_dataSize;
		if ( size >= 77 && size <= 80) {
			unsigned pad = 84 - size;
			static char dummy[8];
			msg->msg_iov[2].iov_len = pad;
			msg->msg_iov[2].iov_base = dummy;
			msg->msg_iovlen = 3;
			rfsmsg->_dataSize += pad;
		}
	}

	int count = 0;
	int slot = _sequenceNumber % RemoteFileSystemMaxOutstandingRPC;
	while ( _outstandingRPC[slot]._sequenceNumber != 0 && 
		count < RemoteFileSystemMaxOutstandingRPC ) {
		_sequenceNumber++;
		slot++;
		count++;
		if ( slot == RemoteFileSystemMaxOutstandingRPC )
			slot = 0;
	}
	if ( count >= RemoteFileSystemMaxOutstandingRPC ) {
		_lock.release();
		Console() << "RFS client too many outstanding RPCs\n" << eor; 
		return -1;
	}
	OutstandingRPC * outrpc = _outstandingRPC + slot;
	outrpc->_sequenceNumber = _sequenceNumber;
	rfsmsg->_sequenceNumber = _sequenceNumber;
	_sequenceNumber++;
	outrpc->_msg = msg;
	outrpc->_closure = c;
	outrpc->_process = ( c == 0 ? thisProcess() : 0 );
	outrpc->_retries = 0;
	outrpc->_maxRetries = maxRetries;
	outrpc->_reserved = reserve;
#ifdef RFSTIME
	outrpc->_time = SystemTimer->time();
#endif
	RFSStatistics.add(rfsmsg);
	_lock.release();

	rfsmsg->_checkSum = rfsmsg->checkSum();
	TheSoftwareClockManager->add
		( (SoftClockFunc) &redoRPCTimerExpired, 
		  (char *) outrpc, _retryHz );
	int socket = ( c == 0 ? outrpc->_socket : _clientSocket );
	int errno;
	int result = sendmsg( socket, msg, 0, errno );
	if  ( result < 0 ) {
		Console() 
			<< "RFS client do RPC sendmsg errno " << errno << "\n" 
			<< eor;
	}
	if ( c != 0 )
		return socket;
	RFSMessage rfshdr;
readmsg:
	/* peek at the message to figure out whether it is for me */
	result = recv( socket, (caddr_t) &rfshdr, sizeof( rfshdr ),
		       MSG_PEEK, errno );
	if ( result < 0 ) {
		if ( errno == EINTR ) {
			/* interrupted, probably too many retries */
			return -1;
		}
		Console() 
			<< "RFS client peek recv errno " 
			<< errno << "\n" << eor;
		return -1;
	}	
	/* make sure got enough info */
	if ( result != sizeof( rfshdr ) ||
	     rfshdr._checkSum != rfshdr.checkSum() ) {
		Console()
			<< "RFS client peek size result "
			<< result << " != expected " << sizeof( rfshdr )
			<< " || checksum incorrrect\n" << eor;
		discardReply( socket, &rfshdr );
		goto readmsg;
	}
	_lock.acquire();
	/* sanity check: seq no in msg must be < my next seq no */
	if ( rfshdr._sequenceNumber >= _sequenceNumber ) {
		Console()
			<< "RFS client insane sequence number " 
			<< rfshdr._sequenceNumber << " >= "
			<< " max sequence number " << _sequenceNumber 
			<< "\n" << eor;
		_lock.release();
		discardReply( socket, &rfshdr );
		goto readmsg;
	}
	/* check if seq no is what i expect */
	if ( outrpc->_sequenceNumber == 0 ||
	     outrpc->_sequenceNumber > rfshdr._sequenceNumber ) {
		Console() 
			<< "RFS client got duplicate reply seq "
			<< rfshdr._sequenceNumber << "\n" << eor;
		_lock.release();
		discardReply( socket, &rfshdr );
		goto readmsg;
	}
	if ( outrpc->_retries >= debugRPCRetry ) Console()
		<< "RFS client retried reply RPC seq " 
		<< outrpc->_sequenceNumber 
		<< " retry " << outrpc->_retries 
		<< " reply size " << rfshdr._headerSize + rfshdr._dataSize
		<< " type " 
		<< RFSMessageTypeStrings[(unsigned) rfshdr._messageType]
		<< "\n" << eor;
	outrpc->_sequenceNumber = 0;
	outrpc->_retries = 0;
	TheSoftwareClockManager->remove
		( (SoftClockFunc) &redoRPCTimerExpired, (char *) outrpc );
	if ( outrpc->_reserved ) {
		freeBufSize( outrpc->_reserved );
	}
#ifdef RFSTIME
	RFSStatistics.add(&rfshdr, SystemTimer->time() - outrpc->_time);
#else
	RFSStatistics.add(&rfshdr);
#endif
	_lock.release();
	return socket;
}
  
void
clientAsyncRPCProcess()
{
	while(1) {
		int errno;
		RFSMessage rfshdr;
		/* peek at the message to figure out how to dispatch it */
		int result = recv
			( _clientSocket, (caddr_t) &rfshdr, sizeof( rfshdr ),
			  MSG_PEEK, errno );
		if ( result < 0 ) {
			Console() 
				<< "RFS client peek recv errno " 
				<< errno << "\n" << eor;
			continue;
		}	
		/* make sure got enough info */
		if ( result != sizeof( rfshdr ) ||
		     rfshdr._checkSum != rfshdr.checkSum() ) {
			Console()
			<< "RFS client peek size result "
			<< result << " != expected " << sizeof( rfshdr )
			<< " || incorrect checksum\n" << eor;
			discardReply( _clientSocket, &rfshdr );
			continue;
		}
		_lock.acquire();
		/* sanity check: seq no in msg must be < my next seq no */
		if ( rfshdr._sequenceNumber >= _sequenceNumber ) {
			Console()
			<< "RFS client insane sequence number " 
			<< rfshdr._sequenceNumber << " >= "
			<< " max sequence number " << _sequenceNumber 
			<< "\n" << eor;
			_lock.release();
			discardReply( _clientSocket, &rfshdr );
			continue;
		}
		/* find outstanding rpc record */
		int slot = rfshdr._sequenceNumber % 
			   RemoteFileSystemMaxOutstandingRPC;
		OutstandingRPC * rpc = _outstandingRPC + slot;
		/* rpc record should be same sequence number returned in msg,
		   otherwise it is probably a repeat reply */
		if ( rpc->_sequenceNumber == 0 ||
		     rpc->_sequenceNumber > rfshdr._sequenceNumber ) {
			Console() 
				<< "RFS client got duplicate reply seq "
				<< rfshdr._sequenceNumber << "\n" << eor;
			_lock.release();
			discardReply( _clientSocket, &rfshdr );
			continue;
		}
		if ( rpc->_retries >= debugRPCRetry ) Console()
			<< "RFS client retried reply RPC seq " 
			<< rpc->_sequenceNumber 
			<< " retry " << rpc->_retries 
			<< " reply size " 
			<< rfshdr._headerSize + rfshdr._dataSize
			<< " type " 
			<< RFSMessageTypeStrings[(unsigned) rfshdr._messageType]
			<< "\n" << eor;
		/* now determine what to do */
		Assert( rpc->_closure != 0 );
		rpc->_sequenceNumber = 0;
		rpc->_retries = 0;
		TheSoftwareClockManager->remove
			( (SoftClockFunc) &redoRPCTimerExpired, (char *) rpc );
		if ( rpc->_reserved ) {
			freeBufSize( rpc->_reserved );
		}
#ifdef RFSTIME
		RFSStatistics.add(&rfshdr, SystemTimer->time() - rpc->_time);
#else
		RFSStatistics.add(&rfshdr);
#endif
		_lock.release();
		/* execute closure */
		rpc->_closure->activate( _clientSocket );
	}
}

int
RemoteFileSystemRPC::mount( unsigned server, char * path, 
			    RFSHandle * handle, RFSAttributes * attributes )
{
	RFSMessage reqmsg;
	reqmsg._headerSize = sizeof( RFSMountMessage );
	reqmsg._dataSize = strlen( path ) + 1;
	reqmsg._messageType = RFSMount;
	iovec iov[3] = { { (caddr_t) &reqmsg, reqmsg._headerSize },
			 { (caddr_t) path, reqmsg._dataSize } };
	msghdr msg = { 0, 0, iov, 2, 0, 0, 0 };
	int socket = doRPC( server, &msg, 0, 0, 0 );
	if ( socket == -1 )
		return CantConnectToSocket;
	RFSMountReplyMessage replymsg;
	int errno;
 	int result = recv( socket, (caddr_t) &replymsg, 
			   sizeof( replymsg ), 0, errno );
	Assert( result == sizeof( replymsg ) &&
		replymsg._headerSize == result &&
		replymsg._dataSize == 0 &&
		replymsg._messageType == RFSMountReply );
	*handle = replymsg._handle;
	*attributes = replymsg._attributes;
	return replymsg._status;
}

int
RemoteFileSystemRPC::containerOpen( RFSContainerHandle & handle,
				    int idNumber, int flags,
				    RFSHandle * fileHandle,
				    RFSAttributes * attributes )
{
	RFSContainerOpenMessage reqmsg;
	reqmsg._headerSize = sizeof( RFSContainerOpenMessage );
	reqmsg._dataSize = 0;
	reqmsg._handle._containerHandle = handle; 
	reqmsg._messageType = RFSContainerOpen;
	reqmsg._idNumber = idNumber;
	reqmsg._flags = flags;
	iovec iov[2] = { { (caddr_t) &reqmsg, reqmsg._headerSize } };
	msghdr msg = { 0, 0, iov, 1, 0, 0, 0 };
	int socket = doRPC( handle._host, &msg, 0, 0 );
	if ( socket == -1 )
		return CantConnectToSocket;
	RFSContainerOpenReplyMessage replymsg;
	int errno;
 	int result = recv( socket, (caddr_t) &replymsg, 
			   sizeof( replymsg ), 0, errno );
	Assert( result == sizeof( replymsg ) &&
		replymsg._headerSize == result &&
		replymsg._dataSize == 0 &&
		replymsg._messageType == RFSContainerOpenReply );
	*fileHandle = replymsg._handle;
	*attributes = replymsg._attributes;
	return replymsg._status;
}

int
RemoteFileSystemRPC::containerCreate( RFSContainerHandle & handle, int hint,
                                      RFSObjectType type,
                                      int flags, RFSHandle * fileHandle,
                                      RFSAttributes * attributes )
{
	RFSContainerCreateMessage reqmsg;
	reqmsg._headerSize = sizeof( RFSContainerCreateMessage );
	reqmsg._dataSize = 0;
	reqmsg._messageType = RFSContainerCreate;
	reqmsg._handle._containerHandle = handle; 
	reqmsg._hint = hint;
	reqmsg._objectType = type;
	reqmsg._flags = flags;
	iovec iov[2] = { { (caddr_t) &reqmsg, reqmsg._headerSize } };
	msghdr msg = { 0, 0, iov, 1, 0, 0, 0 };
	int socket = doRPC( handle._host, &msg, 0, 0 );
	if ( socket == -1 )
		return CantConnectToSocket;
	RFSContainerCreateReplyMessage replymsg;
	int errno;
 	int result = recv( socket, (caddr_t) &replymsg, 
			   sizeof( replymsg ), 0, errno );
	Assert( result == sizeof( replymsg ) &&
		replymsg._headerSize == result &&
		replymsg._dataSize == 0 &&
		replymsg._messageType == RFSContainerCreateReply );
	*fileHandle = replymsg._handle;
	*attributes = replymsg._attributes;
	return replymsg._status;
}

int 
RemoteFileSystemRPC::open( RFSHandle & handle, char * key, int flags, 
		  	   RFSHandle * fileHandle, RFSAttributes * attributes )
{
	RFSOpenMessage reqmsg;
	reqmsg._headerSize = sizeof( RFSOpenMessage );
	reqmsg._dataSize = strlen( key ) + 1;
	reqmsg._messageType = RFSOpen;
	reqmsg._handle = handle;
	reqmsg._flags = flags;
	iovec iov[3] = { { (caddr_t) &reqmsg, reqmsg._headerSize },
			 { (caddr_t) key, reqmsg._dataSize } };
	msghdr msg = { 0, 0, iov, 2, 0, 0, 0 };
	int socket = doRPC( handle._containerHandle._host, &msg, 0, 0 );
	if ( socket == -1 )
		return CantConnectToSocket;
	RFSOpenReplyMessage replymsg;
	int errno;
 	int result = recv( socket, (caddr_t) &replymsg, 
			   sizeof( replymsg ), 0, errno );
	Assert( result == sizeof( replymsg ) &&
		replymsg._headerSize == result &&
		replymsg._dataSize == 0 &&
		replymsg._messageType == RFSOpenReply );
	*fileHandle = replymsg._handle;
	*attributes = replymsg._attributes;
	return replymsg._status;
}

int
RemoteFileSystemRPC::create( RFSHandle & handle, RFSObjectType type,
		  	     char * key, int flags, 
			     RFSHandle * fileHandle, 
			     RFSAttributes * attributes )
{
	RFSCreateMessage reqmsg;
	reqmsg._headerSize = sizeof( RFSCreateMessage );
	reqmsg._dataSize = strlen( key ) + 1;
	reqmsg._messageType = RFSCreate;
	reqmsg._handle = handle;
	reqmsg._objectType = type;
	reqmsg._flags = flags;
	iovec iov[3] = { { (caddr_t) &reqmsg, reqmsg._headerSize },
			 { (caddr_t) key, reqmsg._dataSize } };
	msghdr msg = { 0, 0, iov, 2, 0, 0, 0 };
	int socket = doRPC( handle._containerHandle._host, &msg, 0, 0 );
	if ( socket == -1 )
		return CantConnectToSocket;
	RFSCreateReplyMessage replymsg;
	int errno;
 	int result = recv( socket, (caddr_t) &replymsg, 
			   sizeof( replymsg ), 0, errno );
	Assert( result == sizeof( replymsg ) &&
		replymsg._headerSize == result &&
		replymsg._dataSize == 0 &&
		replymsg._messageType == RFSCreateReply );
	*fileHandle = replymsg._handle;
	*attributes = replymsg._attributes;
	return replymsg._status;
}

int 
RemoteFileSystemRPC::keys( RFSHandle & handle,
		  	   char * buffer, int bufferSize,
			   int * count ) 
{
	RFSKeysMessage reqmsg;
	reqmsg._headerSize = sizeof( RFSKeysMessage );
	reqmsg._dataSize = 0;
	reqmsg._messageType = RFSKeys;
	reqmsg._handle = handle;
	reqmsg._bufferSize = bufferSize;
	iovec iov[2] = { { (caddr_t) &reqmsg, reqmsg._headerSize } };
	msghdr msg = { 0, 0, iov, 1, 0, 0, 0 };
	int socket = doRPC( handle._containerHandle._host, &msg, 0, 0 );
	if ( socket == -1 )
		return CantConnectToSocket;
	RFSKeysReplyMessage replymsg;
	iovec replyiov[2] = { 
		{ (caddr_t) &replymsg, sizeof(RFSKeysReplyMessage) },
		{ (caddr_t) buffer, bufferSize } 
	};
	msghdr replymsghdr = { 0, 0, replyiov, 2, 0, 0, 0 };
	int errno;
 	int result = recvmsg( socket, &replymsghdr, 0, errno );
	Assert( result >= sizeof( replymsg ) + replymsg._dataSize &&
		replymsg._headerSize == sizeof( replymsg ) &&
		replymsg._dataSize <= bufferSize &&
		replymsg._messageType == RFSKeysReply );
	if ( result > sizeof( replymsg ) + replymsg._dataSize ) 
		Console() 
			<< "RFSKeys client received " << result
			<< " > " << " expected " 
			<< sizeof( replymsg ) + replymsg._dataSize 
			<< "\n" << eor;
	*count = replymsg._count;
	return replymsg._status;
}
int 
RemoteFileSystemRPC::associations( RFSHandle & handle,
		  	           char * buffer, int bufferSize,
			   	   int startOffset,
				   int * count, int * resultOffset ) 
{
	RFSAssociationsMessage reqmsg;
	reqmsg._headerSize = sizeof( RFSAssociationsMessage );
	reqmsg._dataSize = 0;
	reqmsg._messageType = RFSAssociations;
	reqmsg._handle = handle;
	reqmsg._bufferSize = bufferSize;
	reqmsg._startOffset = startOffset;
	iovec iov[2] = { { (caddr_t) &reqmsg, reqmsg._headerSize } };
	msghdr msg = { 0, 0, iov, 1, 0, 0, 0 };
	int socket = doRPC( handle._containerHandle._host, &msg, 0, 0 );
	if ( socket == -1 )
		return CantConnectToSocket;
	RFSAssociationsReplyMessage replymsg;
	iovec replyiov[2] = { 
		{ (caddr_t) &replymsg, sizeof(RFSAssociationsReplyMessage) },
		{ (caddr_t) buffer, bufferSize } 
	};
	msghdr replymsghdr = { 0, 0, replyiov, 2, 0, 0, 0 };
	int errno;
 	int result = recvmsg( socket, &replymsghdr, 0, errno );
	Assert( result >= sizeof( replymsg ) + replymsg._dataSize &&
		replymsg._headerSize == sizeof( replymsg ) &&
		replymsg._dataSize <= bufferSize &&
		replymsg._messageType == RFSAssociationsReply );
	if ( result > sizeof( replymsg ) + replymsg._dataSize ) 
		Console() 
			<< "RFSAssociations client received " << result
			<< " > " << " expected " 
			<< sizeof( replymsg ) + replymsg._dataSize 
			<< "\n" << eor;
	*count = replymsg._count;
	*resultOffset = replymsg._resultOffset;
	return replymsg._status;
}

int 
RemoteFileSystemRPC::remove( RFSHandle & handle, char * key, int * links )
{
	RFSRemoveMessage reqmsg;
	reqmsg._headerSize = sizeof( RFSRemoveMessage );
	reqmsg._dataSize = strlen( key ) + 1;
	reqmsg._messageType = RFSRemove;
	reqmsg._handle = handle;
	iovec iov[3] = { { (caddr_t) &reqmsg, reqmsg._headerSize },
			 { (caddr_t) key, reqmsg._dataSize } };
	msghdr msg = { 0, 0, iov, 2, 0, 0, 0 };
	int socket = doRPC( handle._containerHandle._host, &msg, 0, 0 );
	if ( socket == -1 )
		return CantConnectToSocket;
	RFSRemoveReplyMessage replymsg;
	int errno;
 	int result = recv( socket, (caddr_t) &replymsg, 
			   sizeof( replymsg ), 0, errno );
	Assert( result == sizeof( replymsg ) &&
		replymsg._headerSize == result &&
		replymsg._dataSize == 0 &&
		replymsg._messageType == RFSRemoveReply );
	*links = replymsg._links;
	return replymsg._status;
}

int 
RemoteFileSystemRPC::add( RFSHandle & handle, char * key, int idNumber, 
		 	  int * links )
{
	RFSAddMessage reqmsg;
	reqmsg._headerSize = sizeof( RFSAddMessage );
	reqmsg._dataSize = strlen( key ) + 1;
	reqmsg._messageType = RFSAdd;
	reqmsg._handle = handle;
	reqmsg._idNumber = idNumber;
	iovec iov[3] = { { (caddr_t) &reqmsg, reqmsg._headerSize },
			 { (caddr_t) key, reqmsg._dataSize } };
	msghdr msg = { 0, 0, iov, 2, 0, 0, 0 };
	int socket = doRPC( handle._containerHandle._host, &msg, 0, 0 );
	if ( socket == -1 )
		return CantConnectToSocket;
	RFSAddReplyMessage replymsg;
	int errno;
 	int result = recv( socket, (caddr_t) &replymsg, 
			   sizeof( replymsg ), 0, errno );
	Assert( result == sizeof( replymsg ) &&
		replymsg._headerSize == result &&
		replymsg._dataSize == 0 &&
		replymsg._messageType == RFSAddReply );
	*links = replymsg._links;
	return replymsg._status;
}

class RemoteFileSystemRPCReadClosure : public IOAsyncClosure
{
public:
	IOAsyncClosure * _parent;
	RFSReadMessage _reqmsg;
	int _status;
	iovec _iov[2];
	char * _buffer;
	int * _actualNoOfUnits;
	msghdr _msghdr;
        RemoteFileObject * _rfo;

	static void func( RemoteFileSystemRPCReadClosure * c );

	RemoteFileSystemRPCReadClosure
		( IOAsyncClosure * parent, RemoteFileObject * rfo,
		  RFSHandle & handle,
		  int startUnit, int noOfUnits,
		  char * buffer, int noOfBytes, int * actualNoOfUnits ); 
};

inline
RemoteFileSystemRPCReadClosure::RemoteFileSystemRPCReadClosure
	( IOAsyncClosure * parent, RemoteFileObject * rfo,
	  RFSHandle & handle,
	  int startUnit, int noOfUnits, 
	  char * buffer, int noOfBytes, int * actualNoOfUnits ) 
	: _parent( parent ), _buffer( buffer ),
	  _actualNoOfUnits( actualNoOfUnits ), _rfo( rfo ),
	  _func( & func ) 
{
	_reqmsg._headerSize = sizeof( RFSReadMessage );
	_reqmsg._dataSize = 0;
	_reqmsg._messageType = RFSRead;
	_reqmsg._handle = handle;
	_reqmsg._startUnit = startUnit;
	_reqmsg._noOfUnits = noOfUnits;
	_reqmsg._noOfBytes = noOfBytes;
#ifdef RFS_UPDATE_ATTRIBUTES
	_rfo->getAttributes( &_reqmsg._attributes );
#endif
	_iov[0].iov_base = (caddr_t) &_reqmsg;
	_iov[0].iov_len = _reqmsg._headerSize;
	_msghdr.msg_name = 0;
	_msghdr.msg_namelen = 0;
	_msghdr.msg_iov = _iov;
	_msghdr.msg_iovlen = 1;
	_msghdr.msg_control = 0;
	_msghdr.msg_controllen = 0;
	_msghdr.msg_flags = 0;
}

// int debugReadClosure = 1;

void
RemoteFileSystemRPCReadClosure::func( RemoteFileSystemRPCReadClosure * c )
{
/*
	if ( debugReadClosure ) Console() 
		<< "RemoteFileSystemRPCReadClosure::func( "
		<< asAddress( c ) << " ) result " << c->_result << "\n" << eor;
 */
	if ( c->_result == -1 ) {
		c->_status = CantConnectToSocket;
		*(c->_actualNoOfUnits) = 0;
		if ( c->_parent ) {
			c->_parent->activate( 0 );
			delete c;
		}
		return;
	}
	RFSReadReplyMessage replymsg;
	iovec replyiov[2] = { 
		{ (caddr_t) &replymsg, sizeof(RFSReadReplyMessage) },
		{ (caddr_t) c->_buffer, c->_reqmsg._noOfBytes } 
	};
	msghdr replymsghdr = { 0, 0, replyiov, 2, 0, 0, 0 };
	int errno;
 	int result = recvmsg( c->_result, &replymsghdr, 0, errno );
	Assert( result >= sizeof( replymsg ) + replymsg._dataSize &&
		replymsg._headerSize == sizeof( replymsg ) &&
		replymsg._dataSize <= c->_reqmsg._noOfBytes &&
		replymsg._messageType == RFSReadReply );
	if ( result > sizeof( replymsg ) + replymsg._dataSize ) 
		Console() 
			<< "RFSRead client received " << result
			<< " > " << " expected " 
			<< sizeof( replymsg ) + replymsg._dataSize 
			<< "\n" << eor;
	c->_status = replymsg._status;
	*(c->_actualNoOfUnits) = replymsg._actualNoOfUnits;
#ifdef RFS_UPDATE_ATTRIBUTES
	if (c->_status == 0) {
		c->_rfo->setAttributes( &replymsg._attributes );
	}
#endif
/*
	if ( debugReadClosure && 
	     replymsg._actualNoOfUnits != c->_reqmsg._noOfUnits ) {
		Console() 
			<< "RFSRead client got " << replymsg._actualNoOfUnits
			<< " != expected " << c->_reqmsg._noOfUnits 
			<< "\n" << eor;
	}
 */
	if ( c->_parent ) {
		c->_parent->activate( replymsg._actualNoOfUnits );
		delete c;
	}
}

int
RemoteFileSystemRPC::read( RemoteFileObject * rfo,
			   RFSHandle & handle,
		  	   int startUnit, int noOfUnits, 
			   char * buffer, int noOfBytes, 
			   int * actualNoOfUnits, IOAsyncClosure * c )
{
	if ( c ) {
		RemoteFileSystemRPCReadClosure * rc = 
			new RemoteFileSystemRPCReadClosure
				( c, rfo, handle, startUnit, noOfUnits,
				  buffer, noOfBytes, actualNoOfUnits );
/*
		if (debugReadClosure) Console() 
			<< "RemoteFileSystemRPC::read( "
			<< startUnit << ", " << noOfUnits << ", "
			<< noOfBytes << ", " << asAddress( c ) 
			<< " ) new closure " << asAddress( rc )
			<< "\n" << eor;
 */
		int socket = doRPC( handle._containerHandle._host, 
				    &(rc->_msghdr), rc, noOfBytes );
		Assert( socket == _clientSocket );
		return 0;
	} else {
		RemoteFileSystemRPCReadClosure rc 
			( 0, rfo, handle, 
			  startUnit, noOfUnits, buffer, noOfBytes,
			  actualNoOfUnits );
		int socket = doRPC( handle._containerHandle._host, 
				    &(rc._msghdr), 0, noOfBytes );
		if ( socket == -1 )
			return CantConnectToSocket;
		Assert( socket != _clientSocket );
		rc.activate( socket );
		return rc._status;
	}
}

int debugWriteClosure = 0;

class RemoteFileSystemRPCWriteClosure : public IOAsyncClosure
{
public:
	IOAsyncClosure * _parent;
	RFSWriteMessage _reqmsg;
	int _status;	
	iovec _iov[3];
	msghdr _msghdr;
	int * _actualNoOfUnits;
	RemoteFileObject * _rfo;

	static void func( RemoteFileSystemRPCWriteClosure * c );
	RemoteFileSystemRPCWriteClosure
		( IOAsyncClosure * parent, RemoteFileObject * rfo,
		  RFSHandle & handle,
		  int startUnit, int noOfUnits,
		  char * buffer, int noOfBytes, 
		  int writeThrough, int * actualNoOfUnits );
};

inline
RemoteFileSystemRPCWriteClosure::RemoteFileSystemRPCWriteClosure
	( IOAsyncClosure * parent, RemoteFileObject * rfo,
	  RFSHandle & handle,
	  int startUnit, int noOfUnits,
	  char * buffer, int noOfBytes, 
	  int writeThrough, int * actualNoOfUnits ) 
	: _parent( parent ), _actualNoOfUnits( actualNoOfUnits ),
	  _func( & func ), _rfo( rfo )
{
	_reqmsg._headerSize = sizeof( RFSWriteMessage );
	_reqmsg._dataSize = noOfBytes;
	_reqmsg._messageType = RFSWrite;
	_reqmsg._handle = handle;
	_reqmsg._startUnit = startUnit;
	_reqmsg._noOfUnits = noOfUnits;
	_reqmsg._writeThrough = writeThrough;
#ifdef RFS_UPDATE_ATTRIBUTES
	_rfo->getAttributes( &_reqmsg._attributes );
#endif
	_iov[0].iov_base = (caddr_t) &_reqmsg;
	_iov[0].iov_len = _reqmsg._headerSize;
	_iov[1].iov_base = (caddr_t) buffer;
	_iov[1].iov_len = noOfBytes;
	_msghdr.msg_name = 0;
	_msghdr.msg_namelen = 0;
	_msghdr.msg_iov = _iov;
	_msghdr.msg_iovlen = 2;
	_msghdr.msg_control = 0;
	_msghdr.msg_controllen = 0;
	_msghdr.msg_flags = 0;
}

void
RemoteFileSystemRPCWriteClosure::func( RemoteFileSystemRPCWriteClosure * c )
{
	if ( debugWriteClosure ) Console() 
		<< "RemoteFileSystemRPCWriteClosure::func( "
		<< asAddress( c ) << " ) result " << c->_result << "\n" << eor;
	if ( c->_result == -1 ) {
		c->_status = CantConnectToSocket;
		*(c->_actualNoOfUnits) = 0;
		if ( c->_parent ) {
			c->_parent->activate( 0 );
			delete c;
		}
		return;
	}
	RFSWriteReplyMessage replymsg;
	int errno;
 	int result = recv( c->_result, (caddr_t) &replymsg, 
			   sizeof(replymsg), 0, errno );
	Assert( result == sizeof( replymsg ) &&
		replymsg._headerSize == sizeof( replymsg ) &&
		replymsg._dataSize == 0 &&
		replymsg._messageType == RFSWriteReply );
	c->_status = replymsg._status;
#ifdef RFS_UPDATE_ATTRIBUTES
	if (c->_status == 0) {
		c->_rfo->setAttributes( &replymsg._attributes );
	}
#endif
	*(c->_actualNoOfUnits) = replymsg._actualNoOfUnits;
	if ( c->_parent ) {
		c->_parent->activate( replymsg._actualNoOfUnits );
		delete c;
	}
}

int
RemoteFileSystemRPC::write( RemoteFileObject * rfo,
			    RFSHandle & handle, 
		  	    int startUnit, int noOfUnits, 
			    char * buffer, int noOfBytes, int writeThrough,
			    int * actualNoOfUnits, IOAsyncClosure * c )
{
	if ( c ) {
		RemoteFileSystemRPCWriteClosure * rc = 
			new RemoteFileSystemRPCWriteClosure
				( c, rfo, handle, startUnit, noOfUnits,
				  buffer, noOfBytes, writeThrough,
				  actualNoOfUnits );
		if (debugWriteClosure) Console() 
			<< "RemoteFileSystemRPC::write( "
			<< startUnit << ", " << noOfUnits << ", "
			<< noOfBytes << ", " << asAddress( c ) 
			<< " ) new closure " << asAddress( rc )
			<< "\n" << eor;
		int socket = doRPC( handle._containerHandle._host, 
				    &(rc->_msghdr), rc, noOfBytes );
		Assert( socket == _clientSocket );
		return 0;
	} else {
		RemoteFileSystemRPCWriteClosure rc 
			( 0, rfo, handle, 
			  startUnit, noOfUnits, buffer, noOfBytes,
			  writeThrough, actualNoOfUnits );
		int socket = doRPC( handle._containerHandle._host, 
				    &(rc._msghdr), 0, noOfBytes );
		if ( socket == -1 )
			return CantConnectToSocket;
		Assert( socket != _clientSocket );
		rc.activate( socket );
		return rc._status;
	}
}

class RemoteFileSystemRPCSyncClosure : public IOAsyncClosure
{
public:
	IOAsyncClosure * _parent;

	static void func( RemoteFileSystemRPCSyncClosure * c );
	RemoteFileSystemRPCSyncClosure( IOAsyncClosure * parent );
};

inline
RemoteFileSystemRPCSyncClosure::RemoteFileSystemRPCSyncClosure
	( IOAsyncClosure * parent )
	: _parent( parent ), _func( & func )
{
}

int debugSyncClosure = 0;

void
RemoteFileSystemRPCSyncClosure::func( RemoteFileSystemRPCSyncClosure * c )
{
	if ( debugSyncClosure ) Console() 
		<< "RemoteFileSystemRPCSyncClosure::func( "
		<< asAddress( c ) << " ) result " << c->_result << "\n" << eor;
	if ( c->_result == -1 ) {
		if ( c->_parent ) {
			c->_parent->activate( CantConnectToSocket );
			delete c;
		}
		return;
	}
	RFSSyncReplyMessage replymsg;
 	int result = recv( c->_result, (caddr_t) &replymsg, 
			   sizeof( replymsg ), 0, errno );
	Assert( result == sizeof( replymsg ) &&
		replymsg._headerSize == result &&
		replymsg._dataSize == 0 &&
		replymsg._messageType == RFSSyncReply );
	if ( c->_parent ) {
		c->_parent->activate( replymsg._status );
		delete c;
	}
}

int 
RemoteFileSystemRPC::sync( RFSHandle & handle, 
			   RFSAttributes & attributes, int commit, int close,
			   IOAsyncClosure * c )
{
	RFSSyncMessage reqmsg;
	reqmsg._headerSize = sizeof( RFSSyncMessage );
	reqmsg._dataSize = 0;
	reqmsg._messageType = RFSSync;
	reqmsg._handle = handle;
	reqmsg._attributes = attributes;
	reqmsg._commit = commit;
	reqmsg._close = close;
	iovec iov[2] = { { (caddr_t) &reqmsg, reqmsg._headerSize } };
	msghdr msg = { 0, 0, iov, 1, 0, 0, 0 };
	if ( c != 0 ) {
		c = new RemoteFileSystemRPCSyncClosure( c );
	}
	int socket = doRPC( handle._containerHandle._host, &msg, c, 0 );
	if ( c != 0 ) {
		Assert( socket == _clientSocket );
		return 0;
	}
	if ( socket == -1 )
		return CantConnectToSocket;
	RFSSyncReplyMessage replymsg;
	int errno;
 	int result = recv( socket, (caddr_t) &replymsg, 
			   sizeof( replymsg ), 0, errno );
	Assert( result == sizeof( replymsg ) &&
		replymsg._headerSize == result &&
		replymsg._dataSize == 0 &&
		replymsg._messageType == RFSSyncReply );
	return replymsg._status;
}
