#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * RemoteServer.cc
 */

//#define DEBUG

#include "Assert.h"
#include "Disk.h"
#include "DiskStar.h"
#include "Domain.h"
#include "FileObject.h"
#include "FileObjectStar.h"
#include "FileSystem.h"
#include "FileSystemInterface.h"
#include "FileSystemInterfaceStar.h"
#include "Kernel.h"
#include "Log2.h"
#include "MemoryObjectAsync.h"
#include "MemoryObjectContainer.h"
#include "MemoryObjectContainerStar.h"
#include "MemoryObjectDictionary.h"
#include "NameServer.h"
#include "Process.h"
#include "Raid.h"
#include "RemoteContainer.h"
#include "RemoteFileObject.h"
#include "RemoteFileSystem.h"
#include "RemoteFileSystemMessages.h"
#include "RemoteFileSystemRPC.h"
#include "RemoteServer.h"
#include "Semaphore.h"
#include "SocketDef.h"
#include "SocketKernel.h"
#include "ThisProcess.h"
#include "UNIXFile.h"

// #define SPEEDHACK

#ifdef __GNUC__
#pragma implementation
#endif


typedef sockaddr_in SockAddr;
const int SockAddrSize = sizeof( sockaddr_in );

DiskStar theRootDisk;
MemoryObjectContainerStar theRootContainer;

int RemoteServer::_debugMask = 0;
int RemoteServer::_debugMessages = 10000000;
int RemoteServer::_socket = -1;
int RemoteServer::_noOfProcess = 0;
Process ** RemoteServer::_processes = 0;
Semaphore RemoteServer::_mutex(1);
FileSystemInterfaceStar RemoteServer::_fsiStar;
FileSystemInterface * RemoteServer::_fsi = 0;

RemoteServer::DispatchFunc RemoteServer::_dispatchTable[32] = {
	&RemoteServer::mount,
	&RemoteServer::containerOpen,
	&RemoteServer::containerCreate,
	&RemoteServer::open,
	&RemoteServer::create,
	&RemoteServer::keys,
	&RemoteServer::remove,
	&RemoteServer::add,
	&RemoteServer::read,
	&RemoteServer::write,
	&RemoteServer::sync,
	&RemoteServer::associations,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid,
	&RemoteServer::invalid
};

void
RemoteServer::serverInit( int noOfProcess )
{
	Console() 
		<< "Initializing RFS server with " << noOfProcess 
		<< " processes\n" << eor;

	theRootDisk = TheKernel->rootDisk();
	Assert( theRootDisk != 0 );
	int status;
	theRootContainer = 
		theRootDisk->asA( MemoryObjectContainerClass, status );
	Assert( theRootContainer != 0 );

	Console() 
		<< "   root disk " << theRootDisk 
		<< ", root container " << theRootContainer
		<< "\n" << eor;

	Assert( noOfProcess > 0 );
	_noOfProcess = noOfProcess;
	_socket = buildSocket( inet_myaddr(), RemoteFileSystemServerPort );
	Assert( _socket >= 0 );
	
	_fsiStar = StandardNameServer->lookup
		( "fileSystemInterface", "FileSystemInterface" );
	_fsi = _fsiStar;
	Assert ( _fsi != 0 );
	_fsi->chdir( "/" );
	
	_processes = new Process * [ _noOfProcess ];
	Assert( _processes != 0 );
	for ( int i = 0; i < _noOfProcess; i ++ ) {
		_processes[ i ] = new SystemProcess
			( (PFV) &RemoteServer::daemon, TheKernel->domain(),
			  normal, i );
		Assert( _processes[ i ] != 0 );
		_processes[ i ]->reference();
		_processes[ i ]->setName( "RFS Server" );
		_processes[ i ]->ready();
	}
	Console() 
		<< "    RFS server ready.\n" << eor;
}

void
RemoteServer::daemon( int me )
{
	Process * process = thisProcess();
	Console() << process << "[" << me << "] is running\n" << eor;
     	while(1) {
		RFSMessage rfshdr;
		_mutex.P();
		int errno;
		Raids( RaidDetail ) 
			<< "RFS Server[" << me << "] is waiting for message\n"
			<< eor;
		int result = recv( _socket, (caddr_t) &rfshdr, sizeof( rfshdr ),
			 	   MSG_PEEK, errno );
		if ( result < 0 ) {
			Console() 
				<< "RFS server peek recv errno " 
				<< errno << "\n" << eor;
			_mutex.V();
			continue;
		}	
		/* do sanity check */
		if ( result != sizeof( rfshdr ) ||
		     rfshdr.checkSum() != rfshdr._checkSum ) {
			Console() 
				<< "RFS server insane message\n" << eor;
			recv( _socket, (caddr_t) &rfshdr, 
			      sizeof( rfshdr ), 0, errno );
			_mutex.V();
			continue;
		}
		Raids( RaidDetail ) 
			<< "RFS Server[" << me << "] got message "
			<< rfshdr;
		(*_dispatchTable[ rfshdr._messageType ])( &rfshdr );
	}
}	

void
RemoteServer::mount( RFSMessage * rfshdr )
{
	Assert( rfshdr->_messageType == RFSMount );
	char buf[sizeof(RFSMountMessage) + 256];
	RFSMountMessage * reqmsg = (RFSMountMessage *) buf;
	int size = rfshdr->_dataSize + rfshdr->_headerSize;
	Assert( size <= sizeof(RFSMountMessage) + 256 );
 	int errno;
	SockAddr addr;
	int addrlen = SockAddrSize;
	int result = recvfrom( _socket, (caddr_t) reqmsg, size, 0, 
			       (sockaddr *) &addr, &addrlen, errno );
	Assert( rfshdr->_sequenceNumber == reqmsg->_sequenceNumber );
	_mutex.V();
	if ( result < 0 ) {
		Console() 
			<< "RFS server recvfrom errno " << errno 
			<< "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == size );
	char * path = (char *)(reqmsg + 1);

	Raids( RaidMethod ) 
		<< "RemoteServer::mount( " << path << " )\n" << eor;

	int status = 0;
	ProxiableObjectStar po;

	PersistentMemoryObjectStar pmo;
	po = _fsi->open( path, MemoryObjectDictionaryClass, status );
	if ( status == 0 ) {
		extern ClassStar RemoteDictionaryClass;
		if ( po->isKindOf( RemoteDictionaryClass ) ) {
			pmo = 0;
			status = NotImplemented;
			Raids( RaidError ) 
				<< "RemoteServer::mount() " 
				<< path << " is remote object, "
				<< "cannot cross host boundary\n" << eor;
		} else {
			pmo = po->asA( PersistentMemoryObjectClass, status );
			Assert( status == 0 );
#ifdef STATEFULL
			pmo->reference();
#endif
		}
	} else {
		Raids( RaidError ) 
			<< "RemoteServer::mount() cannot open " 
			<< path << " )\n" << eor;
		Assert( po == 0 );
	}

	RFSMountReplyMessage replymsg;
	replymsg._sequenceNumber = reqmsg->_sequenceNumber;
	replymsg._headerSize = sizeof( replymsg );
	replymsg._dataSize = 0;
	replymsg._messageType = RFSMountReply;
	replymsg._status = status;
	if ( pmo ) {
		handle( pmo, replymsg._handle );
		pmo->getAttributes( &replymsg._attributes );
	}
	replymsg._checkSum = replymsg.checkSum();
	result = sendto( _socket, (caddr_t) &replymsg, sizeof( replymsg ), 
			 0, (sockaddr *) &addr, addrlen, errno );
	if ( result < 0 ) {
		Console() 
			<< "RFS server sendto errno " << errno << "\n" << eor;
		return;
	}	
	Assert( result == sizeof( replymsg ) );
}

void 
RemoteServer::containerOpen( RFSMessage * rfshdr )
{
	Assert( rfshdr->_messageType == RFSContainerOpen );
	RFSContainerOpenMessage reqmsg;
	Assert( rfshdr->_headerSize == sizeof( reqmsg ) &&
		rfshdr->_dataSize == 0 );
 	int errno;
	SockAddr addr;
	int addrlen = SockAddrSize;
	int result = recvfrom( _socket, (caddr_t) &reqmsg, sizeof( reqmsg ), 0, 
			       (sockaddr *) &addr, &addrlen, errno );
	Assert( rfshdr->_sequenceNumber == reqmsg._sequenceNumber );
	_mutex.V();
	if ( result < 0 ) {
		Console() 
			<< "RFS server recvfrom errno " << errno 
			<< "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == sizeof( reqmsg ) );

	int status = 0;

	FileObjectStar fo;
	MemoryObjectContainerStar moc = 
		container( reqmsg._handle._containerHandle, status );
	if ( status == 0 ) {
 		fo = moc->open( reqmsg._idNumber, status, reqmsg._flags );
		if ( status == 0 ) {
			Assert( fo );
#ifdef STATEFULL
			fo->reference();
#endif
		} else {
			Assert( fo == 0 );
		}
	} else {
		Assert( moc == 0 );
	}

	RFSContainerOpenReplyMessage replymsg;
	replymsg._sequenceNumber = reqmsg._sequenceNumber;
	replymsg._headerSize = sizeof( replymsg );
	replymsg._dataSize = 0;
	replymsg._messageType = RFSContainerOpenReply;
	replymsg._status = status;
	if ( fo ) {
		handle( fo, replymsg._handle );
		fo->getAttributes( &replymsg._attributes );
	}
	replymsg._checkSum = replymsg.checkSum();
	result = sendto( _socket, (caddr_t) &replymsg, sizeof( replymsg ), 0,
			 (sockaddr *) &addr, addrlen, errno );
	if ( result < 0 ) {
		Console() 
			<< "RFS server sendto errno " << errno << "\n" << eor;
		return;
	}	
	Assert( result == sizeof( replymsg ) );
}

void 
RemoteServer::containerCreate( RFSMessage * rfshdr )
{
	Assert( rfshdr->_messageType == RFSContainerCreate );
	RFSContainerCreateMessage reqmsg;
	Assert( rfshdr->_headerSize == sizeof( reqmsg ) &&
		rfshdr->_dataSize == 0 );
 	int errno;
	SockAddr addr;
	int addrlen = SockAddrSize;
	int result = recvfrom( _socket, (caddr_t) &reqmsg, sizeof( reqmsg ), 0, 
			       (sockaddr *) &addr, &addrlen, errno );
	Assert( rfshdr->_sequenceNumber == reqmsg._sequenceNumber );
	_mutex.V();
	if ( result < 0 ) {
		Console() 
			<< "RFS server recvfrom errno " << errno 
			<< "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == sizeof( reqmsg ) );

	int status = 0;

	FileObjectStar fo;
	MemoryObjectContainerStar moc =
		container( reqmsg._handle._containerHandle, status );
	if ( status == 0 ) {
 		fo = moc->create
			( reqmsg._hint, reqmsg._objectType.objectClass(),
			  status, reqmsg._flags );
		if ( status == 0 ) {
			Assert( fo );
#ifdef STATEFULL
			fo->reference();
#endif
		} else {
			Assert( fo == 0 );
		}
	} else {
		Assert( moc == 0 );
	}

	RFSContainerCreateReplyMessage replymsg;
	replymsg._sequenceNumber = reqmsg._sequenceNumber;
	replymsg._headerSize = sizeof( replymsg );
	replymsg._dataSize = 0;
	replymsg._messageType = RFSContainerOpenReply;
	replymsg._status = status;
	if ( fo ) {
		handle( fo, replymsg._handle );
		fo->getAttributes( &replymsg._attributes );
	}
	replymsg._checkSum = replymsg.checkSum();
	result = sendto( _socket, (caddr_t) &replymsg, sizeof( replymsg ), 0, 
			 (sockaddr *) &addr, addrlen, errno );
	if ( result < 0 ) {
		Console() 
			<< "RFS server sendto errno " << errno << "\n" << eor;
		return;
	}	
	Assert( result == sizeof( replymsg ) );
}

void
RemoteServer::open( RFSMessage * rfshdr )
{
	Assert( rfshdr->_messageType == RFSOpen );
	char buf[sizeof(RFSOpenMessage) + 256];
	RFSOpenMessage * reqmsg = (RFSOpenMessage *) buf;
	int size = rfshdr->_dataSize + rfshdr->_headerSize;
	Assert( size <= sizeof(RFSOpenMessage) + 256 );
 	int errno;
	SockAddr addr;
	int addrlen = SockAddrSize;
	int result = recvfrom( _socket, (caddr_t) reqmsg, size, 0, 
			       (sockaddr *) &addr, &addrlen, errno );
	Assert( rfshdr->_sequenceNumber == reqmsg->_sequenceNumber );
	_mutex.V();
	if ( result < 0 ) {
		Console() 
			<< "RFS server recvfrom errno " << errno 
			<< "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == size );
	char * key = (char *)(reqmsg + 1);

	int status = 0;
	PersistentMemoryObjectRef pmo;

	FileObjectStar fo = fileObject( reqmsg->_handle, status );
	if ( status == 0 ) {
		Assert( fo != 0 );
		MemoryObjectDictionaryRef dictRef = fo->asA
			( MemoryObjectDictionaryClass, status );
		// _fsi->substituteMountUp( dictRef, key );
		MemoryObjectDictionaryStar dict = dictRef;
		if ( status == 0 ) {
			pmo = dict->open( key, status, reqmsg->_flags );
			if ( status == 0 ) {
				Assert( pmo != 0 );
#ifdef STATEFULL
				pmo->reference();
#endif
			} else {
				Assert( pmo == 0 );
			}
		} else {
			Assert( dict == 0 );
		}
	}

	// _fsi->substituteMount( pmo );
	PersistentMemoryObjectStar pmoStar = pmo;
/*
	if ( pmoStar->isKindOf( RemoteFileObjectClass ) ) {
		status = NotImplemented;
		Raids( RaidError ) 
			<< "RemoteServer::open() " 
			<< key << " is remote object, "
			<< "cannot cross host boundary\n" << eor;
		pmoStar = 0;
	}
 */

	RFSOpenReplyMessage replymsg;
	replymsg._sequenceNumber = reqmsg->_sequenceNumber;
	replymsg._headerSize = sizeof( replymsg );
	replymsg._dataSize = 0;
	replymsg._messageType = RFSOpenReply;
	replymsg._status = status;
	if ( pmoStar ) {
		Assert( status == 0 );
		handle( pmoStar, replymsg._handle );
		pmoStar->getAttributes( &replymsg._attributes );
	}
	replymsg._checkSum = replymsg.checkSum();
	result = sendto( _socket, (caddr_t) &replymsg, sizeof( replymsg ), 0,
			 (sockaddr *) &addr, addrlen, errno );
	if ( result < 0 ) {
		Console() 
			<< "RFS server sendto errno " << errno << "\n" << eor;
		return;
	}	
	Assert( result == sizeof( replymsg ) );
}

void
RemoteServer::create( RFSMessage * rfshdr )
{
	Assert( rfshdr->_messageType == RFSCreate );
	char buf[sizeof(RFSCreateMessage) + 256];
	RFSCreateMessage * reqmsg = (RFSCreateMessage *) buf;
	int size = rfshdr->_dataSize + rfshdr->_headerSize;
	Assert( size <= sizeof(RFSCreateMessage) + 256 );
 	int errno;
	SockAddr addr;
	int addrlen = SockAddrSize;
	int result = recvfrom( _socket, (caddr_t) reqmsg, size, 0, 
			       (sockaddr *) &addr, &addrlen, errno );
	Assert( rfshdr->_sequenceNumber == reqmsg->_sequenceNumber );
	_mutex.V();
	if ( result < 0 ) {
		Console() 
			<< "RFS server recvfrom errno " << errno 
			<< "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == size );
	char * key = (char *)(reqmsg + 1);

	int status = 0;
	PersistentMemoryObjectRef pmo;

	FileObjectStar fo = fileObject( reqmsg->_handle, status );
	if ( status == 0 ) {
		Assert( fo != 0 );
		MemoryObjectDictionaryRef dictRef = fo->asA
			( MemoryObjectDictionaryClass, status );
		// _fsi->substituteMountUp( dictRef, key );
		MemoryObjectDictionaryStar dict = dictRef;
		if ( status == 0 ) {
			pmo = dict->create
				( key, reqmsg->_objectType.objectClass(),
				  status, reqmsg->_flags );
			if ( status == 0 || status == KeyExists) {
				Assert( pmo != 0 );
#ifdef STATEFULL
				pmo->reference();
#endif
			} else {
				Assert( pmo == 0 );
			}
		} else {
			Assert( dict == 0 );
		}
	}

	// _fsi->substituteMount( pmo );
	PersistentMemoryObjectStar pmoStar = pmo;
/*
	if ( pmoStar->isKindOf( RemoteFileObjectClass ) ) {
		status = NotImplemented;
		Raids( RaidError ) 
			<< "RemoteServer::open() " 
			<< key << " is remote object, "
			<< "cannot cross host boundary\n" << eor;
		pmoStar = 0;
	}
 */

	RFSCreateReplyMessage replymsg;
	replymsg._sequenceNumber = reqmsg->_sequenceNumber;
	replymsg._headerSize = sizeof( replymsg );
	replymsg._dataSize = 0;
	replymsg._messageType = RFSCreateReply;
	replymsg._status = status;
	if ( pmoStar ) {
		Assert( status == 0 || status == KeyExists );
		handle( pmoStar, replymsg._handle );
		pmoStar->getAttributes( &replymsg._attributes );
	}
/*
	Console() 
	<< "RemoteServer::create( " << key << " ) status " << status 
	<< "\n" << eor;
 */
	replymsg._checkSum = replymsg.checkSum();
	result = sendto( _socket, (caddr_t) &replymsg, sizeof( replymsg ), 0,
			 (sockaddr *) &addr, addrlen, errno );
	if ( result < 0 ) {
		Console() 
			<< "RFS server sendto errno " << errno << "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == sizeof( replymsg ) );
}

void
RemoteServer::keys( RFSMessage * rfshdr )
{
	Assert( rfshdr->_messageType == RFSKeys );
	RFSKeysMessage reqmsg;
	Assert( rfshdr->_headerSize == sizeof( reqmsg ) &&
		rfshdr->_dataSize == 0 );
 	int errno;
	SockAddr addr;
	int addrlen = SockAddrSize;
	int result = recvfrom( _socket, (caddr_t) &reqmsg, sizeof( reqmsg ), 0, 
			       (sockaddr *) &addr, &addrlen, errno );
	Assert( rfshdr->_sequenceNumber == reqmsg._sequenceNumber );
	_mutex.V();
	if ( result < 0 ) {
		Console() 
			<< "RFS server recvfrom errno " << errno 
			<< "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == sizeof( reqmsg ) );

	int status = 0;

	char * buffer = new char[ reqmsg._bufferSize + 
				  sizeof( RFSKeysReplyMessage) ];
	char * keysbuffer = buffer + sizeof( RFSKeysReplyMessage );
	RFSKeysReplyMessage * replymsg = (RFSKeysReplyMessage *) buffer;
	replymsg->_sequenceNumber = reqmsg._sequenceNumber;
	replymsg->_headerSize = sizeof( RFSKeysReplyMessage );
	replymsg->_dataSize = 0;
	replymsg->_messageType = RFSKeysReply;
	replymsg->_count = 0;	
	
	FileObjectStar fo = fileObject( reqmsg._handle, status );
	if ( status == 0 ) {
		Assert( fo != 0 );
		MemoryObjectDictionaryStar dict = fo->asA
			( MemoryObjectDictionaryClass, status );
		if ( status == 0 ) {
			Assert( dict != 0 );
			status = dict->keys( keysbuffer, reqmsg._bufferSize, 
					     replymsg->_count );
			if ( status == 0 ) {
				replymsg->_dataSize = strlen( keysbuffer ) + 1;
				Assert( replymsg->_dataSize <= 
					reqmsg._bufferSize );
			}
		} else {
			Assert( dict == 0 );
		}
	}

	replymsg->_status = status;
	replymsg->_checkSum = replymsg->checkSum();
	int size = replymsg->_dataSize + replymsg->_headerSize;
	result = sendto( _socket, (caddr_t) replymsg, size, 0,
			 (sockaddr *) &addr, addrlen, errno );
	delete buffer;
	if ( result < 0 ) {
		Console() 
			<< "RFS server sendto errno " << errno << "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == size );
}

void
RemoteServer::associations( RFSMessage * rfshdr )
{
	Assert( rfshdr->_messageType == RFSAssociations );
	RFSAssociationsMessage reqmsg;
	Assert( rfshdr->_headerSize == sizeof( reqmsg ) &&
		rfshdr->_dataSize == 0 );
 	int errno;
	SockAddr addr;
	int addrlen = SockAddrSize;
	int result = recvfrom( _socket, (caddr_t) &reqmsg, sizeof( reqmsg ), 0, 
			       (sockaddr *) &addr, &addrlen, errno );
	Assert( rfshdr->_sequenceNumber == reqmsg._sequenceNumber );
	_mutex.V();
	if ( result < 0 ) {
		Console() 
			<< "RFS server recvfrom errno " << errno 
			<< "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == sizeof( reqmsg ) );

	int status = 0;

	char * buffer = new char[ reqmsg._bufferSize + 
				  sizeof( RFSAssociationsReplyMessage) ];
	char * keysbuffer = buffer + sizeof( RFSAssociationsReplyMessage );
	RFSAssociationsReplyMessage * replymsg = 
		(RFSAssociationsReplyMessage *) buffer;
	replymsg->_sequenceNumber = reqmsg._sequenceNumber;
	replymsg->_headerSize = sizeof( RFSAssociationsReplyMessage );
	replymsg->_dataSize = 0;
	replymsg->_messageType = RFSAssociationsReply;
	replymsg->_count = 0;	
	replymsg->_resultOffset = 0;
	
	FileObjectStar fo = fileObject( reqmsg._handle, status );
	if ( status == 0 ) {
		Assert( fo != 0 );
		MemoryObjectDictionaryStar dict = fo->asA
			( MemoryObjectDictionaryClass, status );
		if ( status == 0 ) {
			Assert( dict != 0 );
			if (reqmsg._startOffset == -1) {
				status = dict->associations
					( keysbuffer, reqmsg._bufferSize, 
					  replymsg->_count );
			} else {
				replymsg->_resultOffset = reqmsg._startOffset;
				status = dict->associations
					( keysbuffer, reqmsg._bufferSize,
					  replymsg->_count, 
					  replymsg->_resultOffset );
			}
			replymsg->_dataSize = replymsg->_count;
			Assert( replymsg->_dataSize <= 
				reqmsg._bufferSize );
		} else {
			Assert( dict == 0 );
		}
	}

	replymsg->_status = status;
	replymsg->_checkSum = replymsg->checkSum();
	int size = replymsg->_dataSize + replymsg->_headerSize;
	result = sendto( _socket, (caddr_t) replymsg, size, 0,
			 (sockaddr *) &addr, addrlen, errno );
	delete buffer;
	if ( result < 0 ) {
		Console() 
			<< "RFS server sendto errno " << errno << "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == size );
}

void 
RemoteServer::remove( RFSMessage * rfshdr )
{
	Assert( rfshdr->_messageType == RFSRemove );
	char buf[sizeof(RFSRemoveMessage) + 256];
	RFSRemoveMessage * reqmsg = (RFSRemoveMessage *) buf;
	int size = rfshdr->_dataSize + rfshdr->_headerSize;
	Assert( size <= sizeof(RFSRemoveMessage) + 256 );
 	int errno;
	SockAddr addr;
	int addrlen = SockAddrSize;
	int result = recvfrom( _socket, (caddr_t) reqmsg, size, 0, 
			       (sockaddr *) &addr, &addrlen, errno );
	Assert( rfshdr->_sequenceNumber == reqmsg->_sequenceNumber );
	_mutex.V();
	if ( result < 0 ) {
		Console() 
			<< "RFS server recvfrom errno " << errno 
			<< "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == size );
	char * key = (char *)(reqmsg + 1);

	int status = 0;
	int links = 0;

	FileObjectStar fo = fileObject( reqmsg->_handle, status );
	if ( status == 0 ) {
		Assert( fo != 0 );
		MemoryObjectDictionaryStar dict = fo->asA
			( MemoryObjectDictionaryClass, status );
		if ( status == 0 ) {
			Assert( dict != 0 );
			status = dict->remove( key, &links );
		} else {
			Assert( dict == 0 );
		}
	} 

	RFSRemoveReplyMessage replymsg;
	replymsg._sequenceNumber = reqmsg->_sequenceNumber;
	replymsg._headerSize = sizeof( RFSRemoveReplyMessage );
	replymsg._dataSize = 0;
	replymsg._messageType = RFSRemoveReply;
	replymsg._links = links;	
	replymsg._status = status;
	replymsg._checkSum = replymsg.checkSum();
	result = sendto( _socket, (caddr_t) &replymsg, sizeof( replymsg ), 0,
			 (sockaddr *) &addr, addrlen, errno );
	if ( result < 0 ) {
		Console() 
			<< "RFS server sendto errno " << errno << "\n" << eor;
		return;
	}	
	Assert( result == sizeof( replymsg ) );
}

void
RemoteServer::add( RFSMessage * rfshdr )
{
	Assert( rfshdr->_messageType == RFSAdd );
	char buf[sizeof(RFSAddMessage) + 256];
	RFSAddMessage * reqmsg = (RFSAddMessage *) buf;
	int size = rfshdr->_dataSize + rfshdr->_headerSize;
	Assert( size <= sizeof(RFSAddMessage) + 256 );
 	int errno;
	SockAddr addr;
	int addrlen = SockAddrSize;
	int result = recvfrom( _socket, (caddr_t) reqmsg, size, 0, 
			       (sockaddr *) &addr, &addrlen, errno );
	Assert( rfshdr->_sequenceNumber == reqmsg->_sequenceNumber );
	_mutex.V();
	if ( result < 0 ) {
		Console() 
			<< "RFS server recvfrom errno " << errno 
			<< "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == size );
	char * key = (char *)(reqmsg + 1);

	int status = 0;
	int links = 0;

	FileObjectStar fo = fileObject( reqmsg->_handle, status );
	if ( status == 0 ) {
		Assert( fo != 0 );
		MemoryObjectDictionaryStar dict = fo->asA
			( MemoryObjectDictionaryClass, status );
		if ( status == 0 ) {
			Assert( dict != 0 );
			MemoryObjectContainerStar container = 
				fo->container();
			FileObjectStar toAdd = 
				container->open( reqmsg->_idNumber, status,
						 O_NCACHE );
			if ( status == 0 ) {
				Assert( toAdd != 0 );
				if ( dict->sharesContainer( toAdd ) ) {
					status = dict->add( key, toAdd );
					links = toAdd->links( 0 );
				} else {
					status = DifferentContainers;
				}
			} else {
				Assert( toAdd == 0 );
			}
		} else {
			Assert( dict == 0 );
		}
	}

	RFSAddReplyMessage replymsg;
	replymsg._sequenceNumber = reqmsg->_sequenceNumber;
	replymsg._headerSize = sizeof( RFSAddReplyMessage );
	replymsg._dataSize = 0;
	replymsg._messageType = RFSAddReply;
	replymsg._links = links;	
	replymsg._status = status;
	replymsg._checkSum = replymsg.checkSum();
	result = sendto( _socket, (caddr_t) &replymsg, sizeof( replymsg ), 0,
			 (sockaddr *) &addr, addrlen, errno );
	if ( result < 0 ) {
		Console() 
			<< "RFS server sendto errno " << errno << "\n" << eor;
		return;
	}	
	Assert( result == sizeof( replymsg ) );
}

class RemoteServerReadAsyncClosure : public IOAsyncClosure {
protected:
	static void func( RemoteServerReadAsyncClosure * c );
public:
	int _socket;
	int _noOfUnits;
	int _noOfBytes;
	int _log2UnitSize;
	FileObject * _fo;
	SockAddr _addr;
	int _addrlen;
	RFSReadReplyMessage _replymsg;
	RemoteServerReadAsyncClosure() : IOAsyncClosure(), _func( &func ) {}
};

void 
RemoteServerReadAsyncClosure::func( RemoteServerReadAsyncClosure * c )
{
#ifdef RFS_UPDATE_ATTRIBUTES
	c->_fo->getAttributes( &c->_replymsg._attributes );
#endif
	c->_replymsg._actualNoOfUnits = c->_result;
	Assert( c->_result <= c->_noOfUnits );
	int noOfBytes = 
		c->_replymsg._actualNoOfUnits << c->_log2UnitSize;
	if ( noOfBytes > c->_noOfBytes )
		noOfBytes = c->_noOfBytes;
	c->_replymsg._dataSize = noOfBytes;

	int size = c->_replymsg._dataSize + c->_replymsg._headerSize;
	int errno;
	int result;	
	if ( RemoteServer::_debugMask & 64 ) Console() 
		<< "RemoteServerReadAsyncClosure::func( "
		<< asAddress( c ) << " ) result " << c->_result << "\n" << eor;
	c->_replymsg._checkSum = c->_replymsg.checkSum();
	result = sendto( c->_socket, (caddr_t) &(c->_replymsg), size, 0,
			 (sockaddr *) &(c->_addr), c->_addrlen, errno );
	delete c;
	if ( result < 0 ) {
		Console() 
			<< "RFS server sendto errno " << errno << "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == size );
}

void
RemoteServer::read( RFSMessage * rfshdr )
{
	RFSReadMessage reqmsg;
	Assert( rfshdr->_messageType == RFSRead &&
		rfshdr->_headerSize == sizeof( reqmsg ) &&
		rfshdr->_dataSize == 0 );
 	int errno;
	SockAddr addr;
	int addrlen = SockAddrSize;
	int result = recvfrom( _socket, (caddr_t) &reqmsg, sizeof( reqmsg ), 0, 
			       (sockaddr *) &addr, &addrlen, errno );
	Assert( rfshdr->_sequenceNumber == reqmsg._sequenceNumber );
	_mutex.V();
	if ( result < 0 ) {
		Console() 
			<< "RFS server recvfrom errno " << errno 
			<< "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == sizeof( reqmsg ) );

	char * buffer = new char[ reqmsg._noOfBytes + 
				  sizeof( RemoteServerReadAsyncClosure ) ];
	RemoteServerReadAsyncClosure * c = 
		new ( buffer ) RemoteServerReadAsyncClosure();
	RFSReadReplyMessage * replymsg = &( c->_replymsg );
	char * dataBuffer = buffer + sizeof( RemoteServerReadAsyncClosure );

	c->_addr = addr;
	c->_addrlen = addrlen;
	c->_noOfUnits = reqmsg._noOfUnits;
	c->_noOfBytes = reqmsg._noOfBytes;
	c->_socket = _socket;

	replymsg->_sequenceNumber = reqmsg._sequenceNumber;
	replymsg->_headerSize = sizeof( RFSReadReplyMessage );
	replymsg->_dataSize = 0;
	replymsg->_messageType = RFSReadReply;
	replymsg->_actualNoOfUnits = 0;
	replymsg->_status = 0;
	
	Raids( RaidMethod ) 
		<< "RemoteServer::read() seqno " << reqmsg._sequenceNumber
		<< " start " << reqmsg._startUnit
		<< " count " << reqmsg._noOfUnits
		<< " noOfBytes " << reqmsg._noOfBytes
		<< " closure " << asAddress( c )
		<< "\n" << eor;

	FileObjectStar fo = fileObject( reqmsg._handle, replymsg->_status );
	if ( replymsg->_status == 0 ) {
		Assert( fo != 0 );
		c->_log2UnitSize = fo->log2UnitSize();
		c->_fo = fo;
#ifdef RFS_UPDATE_ATTRIBUTES
		fo->setAttributes( &reqmsg._attributes );
#endif
		fo->asyncRead
			( reqmsg._startUnit, reqmsg._noOfUnits,
			  dataBuffer, reqmsg._noOfBytes, c );
	} else {
		c->_log2UnitSize = 0;
		c->activate( 0 );	
	}
}

class RemoteServerWriteAsyncClosure : public IOAsyncClosure {
protected:
	static void func( RemoteServerWriteAsyncClosure * c );
public:
	int _socket;
	int _noOfUnits;
	SockAddr _addr;
	int _addrlen;
	FileObject * _fo;
	RFSWriteReplyMessage _replymsg;
	RFSWriteMessage _reqmsg;
	RemoteServerWriteAsyncClosure() : IOAsyncClosure(), _func( &func ) {}
};

void 
RemoteServerWriteAsyncClosure::func( RemoteServerWriteAsyncClosure * c )
{
#ifdef RFS_UPDATE_ATTRIBUTES
	c->_fo->getAttributes( &c->_replymsg._attributes );
#endif
	c->_replymsg._actualNoOfUnits = c->_result;
	Assert( c->_result <= c->_noOfUnits );
	if ( RemoteServer::_debugMask & 64 ) Console() 
		<< "RemoteServerWriteAsyncClosure::func( "
		<< asAddress( c ) << " ) result " << c->_result << "\n" << eor;
	int errno;
	int result;
	c->_replymsg._checkSum = c->_replymsg.checkSum();
	result = sendto( c->_socket, (caddr_t) &(c->_replymsg), 
		         sizeof( c->_replymsg ), 0,
			 (sockaddr *) &(c->_addr), c->_addrlen, errno );
	delete c;
	if ( result < 0 ) {
		Console() 
			<< "RFS server sendto errno " << errno << "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == sizeof( c->_replymsg ) );
}

void
RemoteServer::write( RFSMessage * rfshdr )
{
	Assert( rfshdr->_messageType == RFSWrite && 
		rfshdr->_headerSize == sizeof( RFSWriteMessage ) );
	int csize = rfshdr->_dataSize + sizeof( RemoteServerWriteAsyncClosure );
	int size = rfshdr->_dataSize + rfshdr->_headerSize;
	char * buffer = new char[csize];
	RemoteServerWriteAsyncClosure * c = 
		new ( buffer ) RemoteServerWriteAsyncClosure();
	c->_socket = _socket;
	c->_addrlen = SockAddrSize;
	RFSWriteMessage * reqmsg = &(c->_reqmsg);
	RFSWriteReplyMessage * replymsg = &(c->_replymsg);
	char * dataBuffer = buffer + sizeof( RemoteServerWriteAsyncClosure );
 	int errno;
	int result = recvfrom( _socket, (caddr_t) reqmsg, size, 0, 
			       (sockaddr *) &(c->_addr), &(c->_addrlen), 
			       errno );
	Assert( rfshdr->_sequenceNumber == reqmsg->_sequenceNumber );
	_mutex.V();
	if ( result < 0 ) {
		Console() 
			<< "RFS server recvfrom errno " << errno 
			<< "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == size );

	c->_noOfUnits = reqmsg->_noOfUnits;

	replymsg->_sequenceNumber = reqmsg->_sequenceNumber;
	replymsg->_headerSize = sizeof( RFSWriteReplyMessage );
	replymsg->_dataSize = 0;
	replymsg->_messageType = RFSWriteReply;
	replymsg->_status = 0;

	Raids( RaidMethod ) 
		<< "RemoteServer::write() seqno " << reqmsg->_sequenceNumber
		<< " start " << reqmsg->_startUnit
		<< " count " << reqmsg->_noOfUnits
		<< " noOfBytes " << reqmsg->_dataSize
		<< " closure " << asAddress( c )
		<< "\n" << eor;

	FileObjectStar fo = fileObject( reqmsg->_handle, replymsg->_status );
	if ( replymsg->_status == 0 ) {
		Assert( fo != 0 );
		c->_fo = fo;
#ifdef RFS_UPDATE_ATTRIBUTES
		fo->setAttributes( &reqmsg->_attributes );
#endif
		if ( reqmsg->_writeThrough ) {
			fo->asyncWriteThrough
			( reqmsg->_startUnit, reqmsg->_noOfUnits, dataBuffer,
			  reqmsg->_dataSize, c );
		} else {
			fo->asyncWrite
			( reqmsg->_startUnit, reqmsg->_noOfUnits, dataBuffer,
			  reqmsg->_dataSize, c );
		}
	} else {
		c->activate( 0 );
	}
}

void 
RemoteServer::sync( RFSMessage * rfshdr )
{
	Assert( rfshdr->_messageType == RFSSync );
	RFSSyncMessage reqmsg;
	Assert( rfshdr->_headerSize == sizeof( reqmsg ) &&
		rfshdr->_dataSize == 0 );
 	int errno;
	SockAddr addr;
	int addrlen = SockAddrSize;
	int result = recvfrom( _socket, (caddr_t) &reqmsg, sizeof( reqmsg ), 0, 
			       (sockaddr *) &addr, &addrlen, errno );
	Assert( rfshdr->_sequenceNumber == reqmsg._sequenceNumber );
	_mutex.V();
	if ( result < 0 ) {
		Console() 
			<< "RFS server recvfrom errno " << errno 
			<< "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == sizeof( reqmsg ) );

	int status = 0;

	FileObjectStar fo = fileObject( reqmsg._handle, status );
	if ( status == 0 ) {
		Assert( fo != 0 );
		fo->setAttributes( &reqmsg._attributes );
		if ( reqmsg._commit )
			fo->synchronize();
#ifdef STATEFULL
		if ( reqmsg._close ) {
			fo->unreference();
		}
#endif
#ifdef SPEEDHACK
		if ( reqmsg._close ) {
			fo->unreference();
			if ( fo->fastContainer() != 0 )
				fo->fastContainer()->unreference();
		}
#endif
	}

	RFSSyncReplyMessage replymsg;
	replymsg._sequenceNumber = reqmsg._sequenceNumber;
	replymsg._headerSize = sizeof( RFSSyncReplyMessage );
	replymsg._dataSize = 0;
	replymsg._messageType = RFSSyncReply;
	replymsg._status = status;
	replymsg._checkSum = replymsg.checkSum();
	result = sendto( _socket, (caddr_t) &replymsg, sizeof( replymsg ), 0,
			 (sockaddr *) &addr, addrlen, errno );
	if ( result < 0 ) {
		Console() 
			<< "RFS server sendto errno " << errno << "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == sizeof( replymsg ) );
}

void 
RemoteServer::invalid( RFSMessage * rfshdr )
{
	Console() 
		<< "RFS server got invalid message type "
		<< rfshdr->_messageType << " seq no " 
		<< rfshdr->_sequenceNumber << " header size " 
		<< rfshdr->_headerSize << " data size " 
		<< rfshdr->_dataSize << "\n" << eor;
	int size = rfshdr->_headerSize + rfshdr->_dataSize;
	char * buf = new char [ size ];
	RFSMessage * reqmsg = (RFSMessage *) buf;
 	int errno;
	char addr[16];
	int addrlen = sizeof(addr);
	int result = recvfrom( _socket, (caddr_t) buf, size, 0, 
			       (sockaddr *) addr, &addrlen, errno );
	Assert( rfshdr->_sequenceNumber == reqmsg->_sequenceNumber );
	_mutex.V();
	delete buf;
	if ( result < 0 ) {
		Console() 
			<< "RFS server recvfrom errno " << errno 
			<< "\n" << eor;
		return;
	}	
	/* make sure got enough info */
	Assert( result == size );

	/* do nothing */
	return;
}

void
RemoteServer::handle( PersistentMemoryObject * pmo, RFSHandle & result )
{
	if ( pmo->isKindOf( RemoteFileObjectClass ) ) {
		result = *((RemoteFileObject *) pmo)->handle();
		return;
	}
	result._containerHandle._host = inet_myaddr();
#ifdef SPEEDHACK
	result._containerHandle._containerId = 
		(unsigned) pmo->fastContainer();
	pmo->fastContainer()->reference();
#else
	Assert( pmo->fastContainer() != 0 );
	PersistentMemoryObject * containerSource =
		pmo->fastContainer()->source();
	extern ClassStar MemoryObjectPartitionClass;
	Assert( containerSource->isKindOf( MemoryObjectPartitionClass ) );
	Assert( ((FileObject *) containerSource)->source() == theRootDisk );
	result._containerHandle._containerId = containerSource->idNumber();
#endif
	result._containerHandle._size = pmo->fastContainer()->size();
	result._containerHandle._rootId = pmo->fastContainer()->rootIdNumber();
#ifdef SPEEDHACK
	result._idNumber = (unsigned) pmo;
	pmo->reference();
#else
	result._idNumber = pmo->idNumber();
#endif
	result._version = pmo->version();
}

MemoryObjectContainerRef
RemoteServer::container( RFSContainerHandle & handle, int & status )
{
	MemoryObjectContainerRef moc;
	Assert( handle._host = inet_myaddr() );
	status = 0;
#ifdef SPEEDHACK
	moc = (MemoryObjectContainer *) handle._containerId;
	Assert( moc->alive() );
	Assert( moc->isKindOf( MemoryObjectContainerClass ) );
#else
	FileObjectStar fo = 
		theRootContainer->open( handle._containerId, status, O_NCACHE );
	if ( status == 0 ) {
		Assert( fo != 0 );
		extern ClassStar MemoryObjectPartitionClass;
		Assert( fo->isKindOf( MemoryObjectPartitionClass ) );
		moc = fo->asA( MemoryObjectContainerClass, status );
	}
#endif
	Assert( moc != 0 || status != 0 );
	return moc;
}

FileObjectRef
RemoteServer::fileObject( RFSHandle & handle, int & status )
{
	status = 0;
	FileObjectRef fo;
#ifdef SPEEDHACK
	fo = (FileObject *) handle._idNumber;
	Assert( fo->alive() );
	Assert( fo->isKindOf( PersistentMemoryObjectClass ) );
#else
	MemoryObjectContainerStar moc = 
		container( handle._containerHandle, status );
	if ( status == 0 ) {
		fo = moc->open( handle._idNumber, status, O_NCACHE );
	}
#endif
	if ( fo && fo->version() != handle._version ) {
		status = NoSuchFile;
	}
	Assert( fo != 0 || status != 0 );
	return fo;
}


