#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * RemoteTables.c - 
 */

//#define DEBUG

#include "SoftClock.h"
#include "Debug.h"
#include "RemoteContainer.h"
#include "RemoteFileSystem.h"
#include "RemoteTables.h"

#ifdef __GNUC__
#pragma implementation
#endif

/*
 *  RemoteContainerRegistry
 */

RemoteContainerRegistry * TheRemoteContainerRegistry = 0;

RemoteContainerRegistry::RemoteContainerRegistry( unsigned size )
{
	_size = size;
	_info = new RemoteContainerRegistryInfo * [ size ];
	for( int i = 0; i < size; i++ ) {
		_info[ i ] = 0;
	}
}

RemoteContainer * 
RemoteContainerRegistry::lookup( RFSContainerHandle & handle )
{
	_lock.acquire();
	unsigned slot = handle.hash() % _size;
	RemoteContainerRegistryInfo * info = _info[ slot ];
	while( info && handle != *( info->_container->handle() ) ) {
		info = info->_next;
	}
	RemoteContainer * container;
	if ( info ) {
		container = info->_container;
		Debug 	<< "RemoteContainerRegistry::lookup() successful, "
			<< "return " << container << "\n" << eor;	       
	} else {
	        container = new RemoteContainer( handle );
		info = container->registryInfo();
		info->_container = container;
		info->_next = _info[ slot ];
		_info[ slot ] = info;
		Debug 	<< "RemoteContainerRegistry::lookup() "
	      		<< "created new container " 
		      	<< container << "\n" << eor;	
	}
	_lock.release();
	return container;
}

void 
RemoteContainerRegistry::unbind( RemoteContainer * remoteContainer )
{
	Debug	<< "RemoteContainerRegistry::unbind( " << remoteContainer 
	      	<< " )\n" << eor;
	_lock.acquire();
	unsigned slot = remoteContainer->handle()->hash() % _size;
	RemoteContainerRegistryInfo ** pInfo = _info + slot;
	RemoteContainerRegistryInfo * info = *pInfo;
	while( info && info->_container != remoteContainer ) {
		pInfo = & ( info->_next );
		info = *pInfo;
	}
	Assert( info != 0 );
	*pInfo = info->_next;
	_lock.release();
}

/*
 *  RemotePathCache
 */

RemotePathCache * TheRemotePathCache = 0;

RemotePathCache::RemotePathCache( unsigned size )
{
	_timeout = _defaultTimeout;
	_limit = ( size * 4 ) / 5;
	_noOfEntries = 0;
	_size = size;
	_info = new RemotePathCacheInfo * [ size ];
	for ( int i = 0; i < size; i++ )
		_info[ i ] = 0;
}

inline
unsigned
RemotePathCache::hash( RemoteDictionary * dictionary, char * key )
{
	const int max = 16;
	unsigned sum = (unsigned) dictionary;
	char * p = key;
	for ( int i = 0; *p && i < max; i++ ) 
		sum += *p++;
	return sum % _size;	
}

void 
RemotePathCache::garbageCollect()
{
	Debug	<< "RemotePathCache::garbageCollect()\n" << eor;
	for ( int i = 0; i < _size; i++ ) {
		RemotePathCacheInfo ** pInfo = _info + i;
		RemotePathCacheInfo * info = *pInfo;
		while( info ) {
			if ( info->_age < _time.sec() ) {
				*pInfo = info->_next;
				delete info;
				_noOfEntries--;
				info = *pInfo;
			} else {
				pInfo = & ( info->_next );
				info = *pInfo;
			}
		}
	}
}

void
RemotePathCache::bind( RemoteDictionary * dictionary, char * key,
		       RFSHandle & handle )	
{
	Debug	<< "RemotePathCache::bind( " << dictionary << ", " << key 
	      	<< " )\n" << eor;
	_lock.acquire();
	unsigned slot = hash( dictionary, key );
	RemotePathCacheInfo * info = _info[ slot ];
	while( info && 
	       ( dictionary != info->_dictionary ||
		 strcmp( key, info->key() ) != 0 ) ) {
		info = info->_next;
	}
	if ( info && handle != info->_handle ) {
		Debug 	<< "RemotePathCache::bind() binding changed\n" << eor;
		info->_handle = handle;
	} else {
		Debug 	<< "RemotePathCache::bind() new entry\n" << eor;
		info = ( RemotePathCacheInfo * ) new char
			[ strlen( key ) + 1 + sizeof( RemotePathCacheInfo  ) ];
		info->_dictionary = dictionary;
		strcpy( info->key(), key );
		info->_handle = handle;
		info->_next = _info[ slot ];
		_info[ slot ] = info;
		_noOfEntries++;
		if ( _noOfEntries > _limit )
			garbageCollect();
	}
	info->_age = _time.sec() + _timeout;
	_lock.release();
}

void
RemotePathCache::unbind( RemoteDictionary * dictionary, char * key )
{
	Debug	<< "RemotePathCache::unbind( " << dictionary << ", " << key 
	      	<< " )\n" << eor;
	_lock.acquire();
	unsigned slot = hash( dictionary, key );
	RemotePathCacheInfo ** pInfo = _info + slot;
	RemotePathCacheInfo * info = *pInfo;
	while( info && 
	       ( dictionary != info->_dictionary ||
	         strcmp( key, info->key() ) != 0 ) ) {
		pInfo = & ( info->_next );
		info = *pInfo;
	}
	Assert( info != 0 );
	*pInfo = info->_next;
	delete info;
	_noOfEntries--;
	_lock.release();
}

void
RemotePathCache::unbind( RemoteDictionary * dictionary )
{
	Debug	<< "RemotePathCache::unbind( " << dictionary << " )\n" << eor;
	_lock.acquire();
	for ( int i = 0; i < _size; i++ ) {
		RemotePathCacheInfo ** pInfo = _info + i;
		RemotePathCacheInfo * info = *pInfo;
		while( info ) {
			if ( dictionary == 0 ||
			     dictionary == info->_dictionary ) {
				*pInfo = info->_next;
				delete info;
				_noOfEntries--;
				info = *pInfo;
			} else {
				pInfo = & ( info->_next );
				info = *pInfo;
			}
		}
	}
	_lock.release();
}

int
RemotePathCache::lookup( RemoteDictionary * dictionary, char * key, 
			 RFSHandle * handle )
{
	Debug 	<< "RemotePathCache::lookup( " << dictionary << ", " << key 
	      	<< " )\n" << eor;
	_lock.acquire();
	unsigned slot = hash( dictionary, key );
	RemotePathCacheInfo * info = _info[ slot ];
	while( info && 
	       ( dictionary != info->_dictionary ||
	         strcmp( key, info->key() ) != 0 ) ) {
		info = info->_next;
	}
	int result;
	if ( info ) {
		if ( info->_age >= _time.sec() ) {
			Debug 	<< "RemoteContainerRegistry::lookup() found\n" 
		      		<< eor;
			*handle = info->_handle;
			result = 1;
		} else {
			Debug 	<< "RemoteContainerRegistry::lookup() found " 
		      		<< "but has timed out at " << info->_age
				<< "\n" << eor;
			result = 0;
		}
	} else {
		Debug << "RemoteContainerRegistry::lookup() not found\n" 
		      << eor;
		result = 0;
	}
	_lock.release();
	return result;
}

