#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * SVIDDirectory.c - 
 */
#include "Assert.h"
#include "Raid.h"
#include "FileSystem.h"
#include "PersistentMemoryObject.h"
#include "SVIDDirectory.h"
#include "Byte.h"
#include "String.h"

#ifdef __GNUC__
#pragma implementation
#endif

const int DirNameSize = 14;

class direct {
	friend class SVIDDirectory;

        short   d_ino;
        char    d_name[DirNameSize];
};

/*
 * SVIDDirectory allows directory entries to be located in,
 * added to, and removed from a SVID directory file.
 */
SVIDDirectory::SVIDDirectory( PersistentMemoryObject * source )
: MemoryObjectDictionary( source )
{
    initDebug();
    Raid( RaidConstructor )( "%N::SVIDDirectory( %N )\n", this, source );
    Raid( RaidDetail )( "%N::SVIDDirectory idNumber %d _size %d\n",
	     this, source->idNumber(), _size );

    _maxKeyLength = DirNameSize;

    if( _size == 0 ) {
	add( ".", source );
    }
}

SVIDDirectory::~SVIDDirectory()
{
    Raid( RaidConstructor )( "%N::~SVIDDirectory\n", this );
}

/*
 * search the entire directory for an occurrence of key
 * if found, return 1; otherwise 0
 * side effects: set _offset dir entry ofs and
 * findKey will also try to find an empty hole to hold a new directory entry,
 * if needSize is positive.
 */
int
SVIDDirectory::findKey( char * key, int needSize )
{
    direct * dp;

    Raid( RaidMethod )( "%N::findKey( key %s, needSize %d )\n",
                this, key, needSize );

    int StillLooking = ( needSize > 0 );
    _emptyOffset = -1;

    for( _offset = 0; _offset < _size; _offset += sizeof( direct ) ) {
        dp = ( direct * ) &_buffer[_offset];

        Raid( RaidDetail )( "%N::findKey ofs %d ", this, _offset);
        Raid( RaidDetail )( "ino %d key %s\n", dp->d_ino, dp->d_name );

        /* check to see if a hole */
        if( ( StillLooking ) && ( dp->d_ino == 0 ) ) {
            _emptyOffset = _offset;
            StillLooking = 0;
            Raid( RaidDetail )( "%N::findKey space @ %d\n", this, _offset );
        }

        /* return if key is found */
        if( ( dp->d_ino != 0 ) && ( strcmp( key, dp->d_name ) == 0 ) ) {
            _idNumberOfKey = dp->d_ino;
            Raid( RaidDetail )( "%N::findKey found @ %d\n", this, _offset );
            return( 1 );
        }
    }
    return( 0 );
}

/*
 *      return positive count of keys if the buffer was big enough,
 *      negative count otherwise.
 *      optimize and improve later, perhaps via memory-mapping a precalculated
 *      list-o-keys.
 */
int
SVIDDirectory::basicKeys( char * keyBuffer, int bufferSize, int & count )
{
    Raid( RaidMethod )( "%N::keys( buffer %x, bufferSize %d ) _size %d\n",
                this, keyBuffer, bufferSize, _size );

    count = 0;
    int limit = ( bufferSize >= _size ) ? _size : bufferSize;
    char * t = keyBuffer;
    direct * dp = ( direct * ) _buffer;
    for( _offset = 0; _offset < limit; _offset += sizeof( direct ) ) {

        dp = ( direct * ) &_buffer[_offset];

        Raid( RaidDetail )( "%N::keys ofs %d ", this, _offset);
        Raid( RaidDetail )( "ino %d key %s\n", dp->d_ino, dp->d_name );

        if( dp->d_ino != 0) {
                if( t != 0 ) {
                        for( char * s = dp->d_name; *s; ) *t++ = *s++;
                        *t++ = '\n';
                }
                count++;
        }
    }
    if( t != 0 ) *t = '\0';
    if( limit == _size ) return( 0 );
    else return( TooManyKeys );
}

int
SVIDDirectory::entrySize( int )
{
    return( sizeof( direct ) );
}

int
SVIDDirectory::insertEntry( char * key, int, int idNumber  )
{
    direct * edp;

    Raid( RaidMethod )( "%N::insertEntry( key %s, idNumber %d )\n",
		this, key, idNumber );

    if( _emptyOffset < 0 ) {
	/* FIX: COULD THIS BE INHERITED?? */
        _offset = _size;
        _size += sizeof( direct );
        _source->setNumberOfBytes( _size );
        int newUnits = _source->numberOfUnits();
	if( _units != newUnits )  {
		reMap();
		_units = newUnits;
	}
    }
    else {
	_offset = _emptyOffset;
    }

    Raid( RaidDetail )( "%N::insertEntry _offset %d _emptyOffset %d\n",
		this, _offset, _emptyOffset );

    /* move new entry into the blk */
    edp = ( direct * ) &_buffer[_offset];
    edp->d_ino    = idNumber;
    strcpy( edp->d_name, key );

    return( writeChanges() );
}

int
SVIDDirectory::clearEntry()
{
    ( (direct *) &_buffer[_offset] )->d_ino = 0;
    return( writeChanges() );
}

ProxiableObject * 
SVIDDirectoryConstructor( ProxiableObject * source )
{
        if( source->isKindOf( PersistentMemoryObjectClass ) ) {
                return( new SVIDDirectory( (PersistentMemoryObject *)source ) );
        }
        else return( 0 );
}


#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("SVIDDirectory",SVIDDirectory,MemoryObjectDictionary,&SVIDDirectoryConstructor,0);
//DEFINECLASS("SVIDDirectory",SVIDDirectory,MemoryObjectDictionary,&SVIDDirectoryConstructor,0);
