#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * SVIDInode.c - 
 */
#include "Assert.h"
#include "SVIDInode.h"
#include "FileSystem.h"
#include "UNIXFile.h"
#include "MemoryObjectContainer.h"
#include "Class.h"
#include "Raid.h"

#ifdef __GNUC__
#pragma implementation
#endif


extern ClassStar PersistentArrayClass;
extern ClassStar SVIDDirectoryClass;
extern ClassStar SVIDVersionedDirectoryClass;

/* 
 * Initialize the in-memory inode
 */
SVIDInode::SVIDInode( unsigned int numberOfUnits,
		      int log2UnitSize,
		      int inumber,
		      MemoryObjectContainer * container,
		      PersistentMemoryObject * source,
		      SVdinode * thisDinode )
: UNIXInode( numberOfUnits, log2UnitSize, inumber, container, source, 0 )
{
    initDebug();
    Raid( RaidConstructor )( "%N::SVIDInode()\n", this );

    /*
     *	Initialize variables from parameters
     */
    _diskInode = thisDinode;

    /*
     *	Set up indirect variables for mapUnit function.
     */
    singleIndirect = 10;
    doubleIndirect = singleIndirect + ( 1 << indexShift );
    tripleIndirect = doubleIndirect + ( 1 << ( indexShift + indexShift ) );
}

Class *
SVIDInode::supports( Class * aClass )
{
    Raid( RaidProtected )( "%N::supports( %N )\n", this, aClass );

    Class * supportedClass = 0;

    if( ( _diskInode->di_mode & UNIXTypeMask ) == UNIXTypeDirectory ) {
	if( SVIDDirectoryClass->isASubclassOf( aClass ) ) {
                supportedClass = SVIDDirectoryClass;
        }
        else if( SVIDVersionedDirectoryClass->isASubclassOf( aClass ) ) {
                supportedClass = SVIDVersionedDirectoryClass;
        }
    }
    else if( PersistentArrayClass->isASubclassOf( aClass ) ) {
        supportedClass = PersistentArrayClass;
    }

    Raid( RaidProtected )( "%N::supports() returns %N\n",
                           this, supportedClass );

    return( supportedClass );
}

int
SVIDInode::isDictionary()
{
    return( ( _diskInode->di_mode & UNIXTypeMask ) == UNIXTypeDirectory );
}

/*
 * Clean up the in memory inode data
 */
SVIDInode::~SVIDInode()
{
    Raid( RaidConstructor )( "%N::~SVIDInode()\n", this );

    int units = 0;
    if( _diskInode->di_nlink == 0 ) {
	_diskInode->di_mode = 0;
	_diskInode->di_size = 0;
    }
    else if( _diskInode->di_size > 0 ) {
        units = ( _diskInode->di_size + unitSize() - 1 ) >> log2UnitSize();
    }
  
    setNumberOfUnits( units );
    setModified();
}


int
SVIDInode::numberOfBytes()
{
    Raid( RaidMethod )( "%N::numberOfBytes() %d\n", this, _diskInode->di_size );
    return( _diskInode->di_size );
}

int
SVIDInode::basicSetNumberOfBytes( int size )
{
    Raid( RaidProtected )( "%N::basicSetNumberOfBytes( %d ) %d\n",
			   this, size, _diskInode->di_size );
    if( size < 0 ) size = 0;
    _diskInode->di_size = size;
    int units = ( size + unitSize() - 1 ) >> log2UnitSize();
    int result = basicSetNumberOfUnits( units );
    Assert( units == result );
    return( size );
}

/*
 *	this function allows other objects to access and modify the
 *	link count ( or reference count ) of this inode.
 *	links returns the new number of links for this inode.
 */
int
SVIDInode::links( int delta )
{
    Raid( RaidMethod )( "%N::links( delta %d )\n", this, delta );
    _mutex.P();
    _diskInode->di_nlink += delta;
    int result = _diskInode->di_nlink;
    _mutex.V();
    if ( delta ) setModified();
    Raid( RaidMethod )( "%N::links returns %d\n", this, result );
    return( result );
}

int
SVIDInode::setMode( int mode )
{
    Raid( RaidMethod )( "%N::setMode( mode %o ) %o\n",
		this, mode, _diskInode->di_mode );
    _mutex.P();
    _diskInode->di_mode &= ~07777;
    _diskInode->di_mode |= mode & 07777;
    _mutex.V();
    setModified();
    return( 0 );
}

int
SVIDInode::setUserId( int uid )
{
    Raid( RaidMethod )( "%N::setUserId( uid %d ) %d\n",
		this, uid, _diskInode->di_uid );
    _diskInode->di_uid = uid;
    setModified();
    return( 0 );
} 

int
SVIDInode::setGroupId( int gid )
{
    Raid( RaidMethod )( "%N::setGroupId( gid %d ) %d\n",
		this, gid, _diskInode->di_gid );
    _diskInode->di_gid = gid;
    setModified();
    return( 0 );
} 

/* 
 * This does the UNIX Stat( 2 ) function.
 */
void
SVIDInode::info( UNIXFileInfo * infoBuffer )
{
    Raid( RaidMethod )( "%N::info( %x ) %d\n", this, infoBuffer, _idNumber );
    infoBuffer->device	= 0;
    infoBuffer->inumber	= _idNumber;
    infoBuffer->mode	= _diskInode->di_mode;
    infoBuffer->links	= _diskInode->di_nlink;
    infoBuffer->uid	= _diskInode->di_uid;
    infoBuffer->gid	= _diskInode->di_gid;
    infoBuffer->size	= _diskInode->di_size;
    infoBuffer->atime	= _diskInode->di_atime;
    infoBuffer->mtime	= _diskInode->di_mtime;
    infoBuffer->ctime	= _diskInode->di_ctime;
    infoBuffer->blocks	= 0; //FIX
    infoBuffer->blockSize = 0; //FIX
    infoBuffer->flags	= 0;
    infoBuffer->rdevice	= 0;
}

/* 
 * Allocate 1 block.
 */
int 
SVIDInode::allocBlock()
{
    Raid( RaidMethod )( "%N::allocBlock() id %d\n", this, idNumber() );
    return( _container->allocate( 0, 0, 1 ) );
}

/* 
 * Free 1 physical block.
 */
void
SVIDInode::freeBlock( int physicalBlock )
{
    Raid( RaidMethod )( "%N::freeBlock( physicalBlock %d ) id %d\n",
	                this, physicalBlock, idNumber() );
    _container->free( 0, physicalBlock, 1 );
}

/*
 * Get the physical block number of the indicated direct block pointer
 * from the disk inode.
 */
int 
SVIDInode::getDirect( int index )
{
	Raid( RaidMethod )( "%N::getDirect( index %d )\n", this, index );
	Assert( index >= 0 );
	Assert( index < 10 );

	union {
		int	n;
		char	c[4];
	} convert;

	char * addr = &_diskInode->di_addr[index*3];
	convert.c[0] = *addr++;
	convert.c[1] = *addr++;
	convert.c[2] = *addr++;
	convert.c[3] = 0;

#ifdef MACINTOSH
        convert.n >>= 8;
#endif
        return( convert.n );
}

/*
 * Set the physical block number of the indicated direct block pointer
 * in the disk inode.
 */
void
SVIDInode::setDirect( int index, int value )
{
	Raid( RaidMethod )( "%N::setDirect( index %d, value %d )\n",
		   this, index, value );
	Assert( index >= 0 );
	Assert( index < 10 );

	union {
		int	n;
		char	c[4];
	} convert;

        convert.n = value;
#ifdef MACINTOSH
        convert.n <<= 8;
#endif

	char * addr = &_diskInode->di_addr[index*3];
	*addr++ = convert.c[0];
	*addr++ = convert.c[1];
	*addr   = convert.c[2];

	setModified();
}

/*
 * Get the physical block number of the indicated indirect block pointer
 * in the disk inode.
 */
int
SVIDInode::getIndirect( int level )
{
	Assert( level >= 1 );
	Assert( level <= 3 );
	union {
		int	n;
		char	c[4];
	} convert;

	char * addr = &_diskInode->di_addr[27 + (level*3)];
	convert.c[0] = *addr++;
	convert.c[1] = *addr++;
	convert.c[2] = *addr++;
	convert.c[3] = 0;

#ifdef MACINTOSH
        convert.n >>= 8;
#endif
        return( convert.n );
}

/*
 * Set the physical block number of the indicated indirect block pointer
 * in the disk inode.
 */
void
SVIDInode::setIndirect( int level, int value )
{
	Assert( level >= 1 );
	Assert( level <= 3 );

	union {
		int	n;
		char	c[4];
	} convert;

        convert.n = value;
#ifdef MACINTOSH
        convert.n <<= 8;
#endif

	char * addr = &_diskInode->di_addr[27 + (level*3)];
	*addr++ = convert.c[0];
	*addr++ = convert.c[1];
	*addr   = convert.c[2];
	
	setModified();
}

int
SVIDInode::accessMode()
{
    return( _diskInode->di_mode & UNIXAccessMask );
}

int
SVIDInode::userId()
{
    return( _diskInode->di_uid );
}

int
SVIDInode::groupId()
{
    return( _diskInode->di_gid );
}



#include "ClassInstaller.h"
DEFINECLASS("SVIDInode",SVIDInode,UNIXInode,0,0);
