#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * SVIDVersionedDirectory.c - SVID Directories with versioned entries
 */
#include "Assert.h"
#include "Raid.h"
#include "FileSystem.h"
#include "PersistentMemoryObject.h"
#include "SVIDVersionedDirectory.h"
#include "Byte.h"
#include "String.h"
#include "Convert.h"
#include "Ctype.h"

#ifdef __GNUC__
#pragma implementation
#endif


const int DirNameSize = 12;

class direct {
	friend class SVIDVersionedDirectory;

        short   d_ino;
        char    d_name[DirNameSize];
        short   d_version;
};

/*
 * SVIDVersionedDirectory allows directory entries to be located in,
 * added to, and iterated over a SVID versioned directory.
 */
SVIDVersionedDirectory::SVIDVersionedDirectory( PersistentMemoryObject *
						const source )
: SVIDDirectory( source )
{
    initDebug();
    Raid( RaidConstructor )( "%N::SVIDVersionedDirectory( %N )\n",
			     this, source );
    _maxKeyLength = DirNameSize;
}

SVIDVersionedDirectory::~SVIDVersionedDirectory()
{
    Raid( RaidConstructor )( "%N::~SVIDVersionedDirectory\n", this );
}

/*
 * search the entire directory for an occurrence of key
 * if found, return 1; otherwise 0
 *
 * side effects: set _offset dir entry ofs and
 *               try to find an empty hole to hold a new directory entry,
 *               if needSize is positive.
 *
 * NOTE: If needSize is positive, then always return 0, since that implies
 *       a create operations is being performed.
 *
 * Version format:
 *	filename	- most recent version of file
 *	filename;	- most recent version of file
 *	filename;0	- most recent version of file
 *	filename;-0	- most recent version of file
 *	filename;-n	- nth previous version of file
 *	filename;n	- nth version of file
 */
int
SVIDVersionedDirectory::findKey( char * key, int needSize )
{
    direct * dp;

    Raid( RaidProtected )( "%N::findKey( key %s, needSize %d )\n",
                	   this, key, needSize );

    _emptyOffset = -1;

    int stillLooking = ( needSize > 0 );
    int markPosition = 0;
    int versionGiven = getVersion( key, markPosition );

    char * oldKey = key;
    key = new char[markPosition + 1];
    strncpy( key, oldKey, markPosition );
    key[markPosition] = '\0';

    if( needSize > 0 ) {
	versionGiven = -1;
    }
    else if( versionGiven <= 0 ) {
	versionGiven = nthHighestVersion( key, markPosition, versionGiven );
    }

    for( _offset = 0; _offset < _size; _offset += sizeof( direct ) ) {
        dp = ( direct * ) &_buffer[_offset];

	int version = dp->d_version;
	if( version == 0 ) version = 1;

        Raid( RaidDetail )( "%N::findKey %s ofs %d ", this, key, _offset );
        Raid( RaidDetail )( "ino %d key %s version %d\n",
			    dp->d_ino, dp->d_name, version );

        /* check to see if a hole */
        if( ( stillLooking ) && ( dp->d_ino == 0 ) ) {
            _emptyOffset = _offset;
            stillLooking = 0;
            Raid( RaidDetail )( "%N::findKey space @ %d\n", this, _offset );
        }

        /* return if key is found */
        if( ( dp->d_ino != 0 ) &&
	    ( strcmp( key, dp->d_name ) == 0 ) &&
	    ( version == versionGiven ) ) {
            _idNumberOfKey = dp->d_ino;
            Raid( RaidDetail )( "%N::findKey found @ %d\n", this, _offset );
	    delete key;
            return( 1 );
        }
    }

    delete key;
    return( 0 );
}

int
SVIDVersionedDirectory::getVersion( char * key, int & markPosition )
{
    int keyLength = strlen( key );
    int sign = 0;
    for( markPosition = keyLength; markPosition > 0; ) {
	char c = key[--markPosition];

	if( c == ';' ) {
	    Raid( RaidProtected )( "%N::getVersion returns %d\n",
				   this, atoi( &key[markPosition + 1] ) );
	    return( atoi( &key[markPosition + 1] ) );
	}
	else if( c == '-' ) {
	    sign = 1;
	}
	else if( isdigit( c ) ) {
	    if( sign ) break;
	}
	else break;
    }
    Raid( RaidProtected )( "%N::getVersion returns 0\n", this );
    markPosition = keyLength;
    return( 0 );
}

/*
 *      return positive count of keys if the buffer was big enough,
 *      negative count otherwise.
 *      optimize and improve later, perhaps via memory-mapping a precalculated
 *      list-o-keys.
 */
int
SVIDVersionedDirectory::basicKeys( char * keyBuffer, int bufferSize, int & count )
{
    Raid( RaidMethod )( "%N::keys( buffer %x, bufferSize %d ) _size %d\n",
                this, keyBuffer, bufferSize, _size );

    count = 0;
    int limit = ( bufferSize >= _size ) ? _size : bufferSize;
    char * t = keyBuffer;
    direct * dp = ( direct * ) _buffer;
    for( _offset = 0; _offset < limit; _offset += sizeof( direct ) ) {

        dp = ( direct * ) &_buffer[_offset];

        Raid( RaidDetail )( "%N::keys ofs %d ", this, _offset );
        Raid( RaidDetail )( "ino %d key %s version %d\n",
			    dp->d_ino, dp->d_name, dp->d_version );

        if( dp->d_ino != 0) {
            if( t != 0 ) {
                for( char * s = dp->d_name; *s; ) *t++ = *s++;
		*t++ = ';';
		int version = dp->d_version;
		if( version == 0 ) version = 1;
		for( int i = 1, d = -1; i <= version; i *= 10, d++ );
		int digits = d + 1;
		for( ; d >= 0; d-- ) {
		    *(t + d) = "0123456789"[version % 10];
		    version /= 10;
		}
		t += digits;
                *t++ = '\n';
            }
            count++;
        }
    }
    if( t != 0 ) *t = '\0';
    if( limit == _size ) return( 0 );
    else return( TooManyKeys );
}

void
SVIDVersionedDirectory::insertVersion( int version, short * versions,
				       int & versionCount, int n )
{
    Raid( RaidProtected )( "%N::insertVersion( version %d, count %d, n %d )\n",
			   this, version, versionCount, n );

    Assert( versionCount <= ( n + 1) );
    if( versionCount <= n ) versions[versionCount++] = version;
    else if( versions[n] < version ) versions[n] = version;

    for( int i = 1; i < versionCount; i++ ) {
	for( int j = i; ( versions[j] > versions[j-1] ) && ( j > 0 ); j-- ) {
	    short temp = versions[j];
	    versions[j] = versions[j-1];
	    versions[j-1] = temp;
	}
    }

    for( i = 0; i <= n; i++ ) {
	Raid( RaidDetail )( "%N::insertVersion versions[%d] = %d\n",
			    this, i, versions[i] );
    }
}

int
SVIDVersionedDirectory::nthHighestVersion( char * key, int keyLength, int n )
{
    direct * dp;

    Raid( RaidProtected )( "%N::nthHighestVersion( key %s, keyLen %d, n %d )\n",
			   this, key, keyLength, n );
    Assert( n <= 0 );
    n = -n;

    short * versions = new short[n+1];
    for( int i = 0; i <= n; i++ ) versions[i] = 0;
    int versionCount = 0;

    for( int offset = 0; offset < _size; offset += sizeof( direct ) ) {
        dp = ( direct * ) &_buffer[offset];

        if( ( dp->d_ino != 0 ) &&
	    ( strncmp( key, dp->d_name, keyLength ) == 0 ) &&
	    ( ( keyLength == DirNameSize ) ||
	      ( dp->d_name[keyLength] == '\0' ) ) ) {
	    int version = dp->d_version;
	    if( version == 0 ) version = 1;
	    insertVersion( version, versions, versionCount, n );
        }
    }

    int nthHighest = versions[n];
    delete versions;

    Raid( RaidProtected )( "%N::nthHighestVersion returns %d\n",
			   this, nthHighest );

    return( nthHighest );
}

int
SVIDVersionedDirectory::nextHigherVersion( char * key, int givenVersion )
{
    direct * dp;
    int nextVersion = 0;
    int keyLength = strlen( key );

    Raid( RaidProtected )( "%N::nextHigherVersion( key %s, version %d )\n",
			   this, key, givenVersion );

    if( givenVersion <= 0 ) {
	for( int offset = 0; offset < _size; offset += sizeof( direct ) ) {
            dp = ( direct * ) &_buffer[offset];
	    int version = dp->d_version;
	    if( version == 0 ) version = 1;

            if( ( dp->d_ino != 0 ) &&
		( strncmp( key, dp->d_name, keyLength ) == 0 ) ) {
		nextVersion = ( version > nextVersion ) ? version : nextVersion;
            }
	}
	nextVersion++;
    }
    else {
	int found = 0;
	for( nextVersion = givenVersion; !found; nextVersion++ ) {
	    found = 1;
	    for( int offset = 0; offset < _size; offset += sizeof( direct ) ) {
        	dp = ( direct * ) &_buffer[offset];
		int version = dp->d_version;
		if( version == 0 ) version = 1;

        	if( ( dp->d_ino != 0 ) &&
		    ( strncmp( key, dp->d_name, keyLength ) == 0 ) &&
		    ( version == nextVersion ) ) {
			found = 0;
        	}
            }
	    if( found ) break;
	}
    }

    Raid( RaidProtected )( "%N::nextHigherVersion returns %d\n",
			   this, nextVersion );

    return( nextVersion );
}

int
SVIDVersionedDirectory::insertEntry( char * key, int, int idNumber  )
{
    direct * edp;

    Raid( RaidProtected )( "%N::insertEntry( key %s, idNumber %d )\n",
			   this, key, idNumber );

    int markPosition = 0;
    int versionGiven = getVersion( key, markPosition );

    char * oldKey = key;
    key = new char[markPosition + 1];
    strncpy( key, oldKey, markPosition );
    key[markPosition] = '\0';
    int version = nextHigherVersion( key, versionGiven );

    if( _emptyOffset < 0 ) {
	/* FIX: COULD THIS BE INHERITED?? */
        _offset = _size;
        _size += sizeof( direct );
        _source->setNumberOfBytes( _size );
        int newUnits = _source->numberOfUnits();
	if( _units != newUnits )  {
		reMap();
		_units = newUnits;
	}
    }
    else {
	_offset = _emptyOffset;
    }

    Raid( RaidDetail )( "%N::insertEntry %s _offset %d _emptyOffset %d\n",
			this, key, _offset, _emptyOffset );

    /* move new entry into the blk */
    edp = ( direct * ) &_buffer[_offset];
    edp->d_ino = idNumber;
    strcpy( edp->d_name, key );
    edp->d_version = version;

    delete key;
    return( writeChanges() );
}

ProxiableObject * 
SVIDVersionedDirectoryConstructor( ProxiableObject * source )
{
        if( source->isKindOf( PersistentMemoryObjectClass ) ) {
                return( new SVIDVersionedDirectory( (PersistentMemoryObject *)source ) );
        }
        else return( 0 );
}


#include "ClassInstaller.h"
DEFINEPERSISTENTCLASS("SVIDVersionedDirectory",SVIDVersionedDirectory,SVIDDirectory,&SVIDVersionedDirectoryConstructor,0);
//DEFINECLASS("SVIDVersionedDirectory",SVIDVersionedDirectory,SVIDDirectory,&SVIDVersionedDirectoryConstructor,0);
