#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Contributing Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * UNIXContainer.c - 
 */
#include "Assert.h"
#include "FileObject.h"
#include "MemoryObjectCache.h"
#include "PersistentMemoryObject.h"
#include "Raid.h"
#include "UNIXContainer.h"
#include "UNIXInode.h"
#include "UNIXInodeIndirect.h"

#ifdef __GNUC__
#pragma implementation
#endif

/*
 * Return the block number containing the disk inode.
 */
int
UNIXContainer::mapInumber( int )
{
    Assert( NOTREACHED );
    return( 0 );
}

/*
 * Return the first inode in the block containing "inumber" inode.
 */
int
UNIXContainer::firstInumber( int )
{
    Assert( NOTREACHED );
    return( 0 );
}

UNIXContainer::UNIXContainer( PersistentMemoryObject * source )
    : MemoryObjectCachingContainer( source )
{
    initDebug();

    Raid( RaidConstructor )( "%N::UNIXContainer( %N )\n", this, source );
}

UNIXContainer::~UNIXContainer()
{
    _indirect->free( _source );
    delete _diskInodeCacheInfo;
}

#ifdef NEVER
void
UNIXContainer::setModified( int inumber )
{
    int first = firstInumber( inumber );
    int diskInodeBufNo = first / _inodesPerBlock;
    UNIXInodeIndirectInfo *info = _diskInodeCacheInfo[ diskInodeBufNo ];
    Assert( info->lock() > 0 &&
            info->source() == _source &&
            info->unit() == mapInumber(inumber) << _log2FragmentToSector );
    Assert( info != 0 );
    *(info->dirtyFlag()) = 1;
}
#endif

/*
 *  Get an disk inode into memory 
 *  Also increment reference count on block on which inode resides
 *  ( Always read an enclosing block of disk inodes. )
 */
void *
UNIXContainer::getDinode( int inumber, UNIXInodeIndirectInfo ** & info )
{
    Raid( RaidMethod )( "%N::getDinode( %d )\n", this, inumber );

    int first = firstInumber( inumber );
    int diskInodeBufNo = first / _inodesPerBlock;
    Assert( diskInodeBufNo >= 0 && diskInodeBufNo < _noDiskInodeBufs );

    int physicalBlock = mapInumber( inumber ) << _log2FragmentToSector;
    _diskInodeCacheInfo[ diskInodeBufNo ] = 
    _indirect->getIndirect( _source, physicalBlock, inumber, _diskInodeCacheInfo[ diskInodeBufNo ] );
    info = &_diskInodeCacheInfo[ diskInodeBufNo ];
    /* inode buffer locked down */
    Raid( RaidMethod )( "%N::readDinode done inumber %d\n", this, inumber );
    return ((char *) (*info)->buf()) + (( inumber - first ) * _dinodeSize);
}

/*
 *  Decrement reference count on block with disk inode 
 */
void
UNIXContainer::releaseDinode( UNIXInodeIndirectInfo * info )
{
    Raid( RaidMethod )( "%N::releaseDinode( %x )\n", this, info );

    _indirect->releaseIndirect( info, 0 );
}

/*
 *  Write this inode to the disk.
 *  If there are no more current references to inodes
 */
void
UNIXContainer::writeDinode( int inumber, int release, unsigned delay )
{
    Raid( RaidMethod )( "%N::writeDinode( %d, %d, %d )\n",
                        this, inumber, release, delay );

    int first = firstInumber( inumber );
    int diskInodeBufNo = first / _inodesPerBlock;
    Assert( diskInodeBufNo >= 0 && diskInodeBufNo < _noDiskInodeBufs );

    UNIXInodeIndirectInfo * info = _diskInodeCacheInfo[ diskInodeBufNo ];
    int physicalBlock = mapInumber( inumber ) << _log2FragmentToSector;
    _indirect->synchronize( _source, (unsigned) physicalBlock, 
                            0xffffffff, info, delay, release );
}

void
UNIXContainer::basicSynchronizeInodes()
{
    // container and inodes have already been locked
    // Write out any in memory inodes.

    for( UNIXInode **inodep = (UNIXInode **) _activeFileObjects; 
         *inodep != 0; inodep++ ) {
        UNIXInode * inode = *inodep;
        if( !inode->transient() &&
            (inode->zombieInfo()->type() == ZOMBIE_NOT ||
             inode->zombieInfo()->type() == ZOMBIE_NOTSYNC) 
            ) {
            MemoryObjectCache * cache = inode->cache();
            if ( cache ) {
                setCreatingDictionary(inode, 0);
                cache->synchronize( 0 );
            }
            inode->synchronizeInode();
        }
    }

    acquire();

    _indirect->synchronize( _source );
    release();
}

void 
UNIXContainer::basicSynchronize( int idNumber )
{
    // acquire();
    writeDinode( idNumber, 0 );
    // release();
}

void
UNIXContainer::initializeArrays( int containerSize )
{
    MemoryObjectContainer::initializeArrays( containerSize );
    _noDiskInodeBufs = 
    ( containerSize + _inodesPerBlock - 1 ) / _inodesPerBlock;
    _diskInodeCacheInfo = new UNIXInodeIndirectInfo * [ _noDiskInodeBufs ];
    for( int i = 0; i < _noDiskInodeBufs; i++ ) 
        _diskInodeCacheInfo[i] = 0;
}


#include "ClassInstaller.h"
DEFINECLASS("UNIXContainer",UNIXContainer,MemoryObjectCachingContainer,0,0);
