#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Contributing Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * UNIXInode.c - 
 */

#include "Assert.h"
#include "Barrier.h"
#include "Byte.h"
#include "CacheAsync.h"
#include "FileSystem.h"
#include "FileSystemHints.h"
#include "FileSystemInterface.h"
#include "MemoryObjectCache.h"
#include "MemoryObjectContainer.h"
#include "MemoryObjectDictionary.h"
#include "Raid.h"
#include "UNIXContainer.h"
#include "UNIXFile.h"
#include "UNIXInode.h"
#include "UNIXInodeIndirect.h"

// #define UNIXINODECACHEWRITEALLOC

#ifdef __GNUC__
#pragma implementation
#endif

int
UNIXInode::fixedSize()
{
    return( 0 );
}

/*
 * Indicate whether the object is still "persistently" referenced or not.
 */
int
UNIXInode::transient()
{
    return( links( 0 ) == 0 );
}

/*
 * Allocate and return a block.
 */
int
UNIXInode::allocBlock()
{
    Assert( NOTREACHED );
    return( 0 ); 
}

/*
 * Free a block.
 */
void
UNIXInode::freeBlock( int )
{
    Assert( NOTREACHED );
}

/*
 * Free a fragments
 */
void
UNIXInode::freeFragments( int, int )
{
    Assert( NOTREACHED );
}

void
UNIXInode::getDinode()
{
    Assert( NOTREACHED );
}

void
UNIXInode::releaseDinode()
{
    Assert( NOTREACHED );
}

/*
 * Compute preferred block
 */
int 
UNIXInode::blockPref( int *, int, int )
{
    return 0;
}

/*
 * Allocate few blocks 
 */
int
UNIXInode::allocBlocks( int * buf, int start, int count, int pref )
{
    int * bp = buf + start;
    int i;
    for ( i = 0; i < count; i++ ) {
        Assert( *bp == 0 );
        *bp = allocBlock();
        if (little_endian_fs) {
            swapl((void *) bp);
        }

        if ( *bp == 0 )
            break;
        bp++;
    }
    return i;
}

int UNIXInodeLargestPossibleUnit = 0;
int UNIXInodeLog2CacheUnitSize = 0;
// int UNIXInodeCacheFlag = O_LCACHE;
int UNIXInodeCacheFlag = 0;

int
UNIXInode::parseFlags( int flags )
{
    Raid( RaidProtected )( "%N::parseFlags( %x )\n", this, flags );
    if ( ( flags & O_CMASK ) == 0 ) {
        if (UNIXInodeLargestPossibleUnit) { 
            if ( numberOfBytes() < 4096 ) {
                int i = log2BlockSize + _log2UnitSize - 12 /* 4K */;
                if ( i > 0 ) {
                    flags |= ( i << 28 );
                }
            } else {
                flags |= O_CSIZE;
            }
        } else {
            int i = log2BlockSize + _log2UnitSize - 12 /* 4K */;
            if ( UNIXInodeLog2CacheUnitSize ) {
                int j = UNIXInodeLog2CacheUnitSize - 12;
                if ( j > i )
                    i = j;
            }
            if ( i > 0 ) {
                flags |= ( i << 28 );
            }
        }
    }
    unsigned isDict = isDictionary();
    if ( !isDict && !isContainer() && !isSymbolicLink() &&
         (flags & O_NCACHE) == 0 ) {
        flags |= UNIXInodeCacheFlag;
    }
    if ( isDict ) {
        flags = (flags | O_MULTI | O_PERM) & ~(O_SINGLE | O_TEMP);
    }

    return FileObject::parseFlags( flags );
}

/*
 * Set this inode to have the indicated number of units.  This is used to
 * truncate a file ( or free all allocated blocks ).
 */
unsigned
UNIXInode::basicSetNumberOfUnits( unsigned units )
{
    Raid( RaidMethod )( "%N::basicSetNumberOfUnits( units %d ) %d\n",
                        this, units, numberOfUnits() );

    unsigned noOfUnits = numberOfUnits();
    if( units == noOfUnits ) 
        return( units );
    else if ( units > noOfUnits )
        return( MemoryObject::basicSetNumberOfUnits( units ) );

    getDinode();
    if ( units == 0 ) {
        basicMapUnitFree( 0, noOfUnits );
        fragmentsInLastBlock = 0;
    } else if ( log2BlockSize == 0 ) {
        basicMapUnitFree( units, noOfUnits - units );
    } else {
        // block align the last unit
        unsigned adjustedUnits;
        unsigned mask = ( units & blockMask() );
        if ( mask == 0 )
            adjustedUnits = units;
        else
            adjustedUnits = units + ( blockSize() - mask );
        Assert( (adjustedUnits & blockMask() ) == 0 );
        if ( noOfUnits > adjustedUnits ) {
            basicMapUnitFree( adjustedUnits, noOfUnits - adjustedUnits );
            fragmentsInLastBlock = 0;
        } else {
            if ( fragmentsInLastBlock != 0 ) {
                int fragments = noOfUnits - units;
                Assert( fragments < blockSize() );
                freeFragments( units, fragments );
                fragmentsInLastBlock -= fragments;
                Assert( fragmentsInLastBlock > 0 &&
                        fragmentsInLastBlock < blockSize() );
            }
        }
    }
    releaseDinode();

    unsigned retval = MemoryObject::basicSetNumberOfUnits( units );

    Assert( retval == units );
    return retval;
}

#ifdef UNIXINODECACHEWRITEALLOC

class UNIXInodeCacheWriteAsyncClosure : public IOAsyncClosure {
protected:
    IOAsyncClosure * _parent;
    unsigned _expected;
    unsigned _log2UnitSize;
    unsigned _startUnit;
    UNIXInode * _inode;
    static void func( UNIXInodeCacheWriteAsyncClosure * c );

public:
    UNIXInodeCacheWriteAsyncClosure( IOAsyncClosure * parent,
                                     unsigned log2UnitSize,
                                     unsigned expected,
                                     unsigned startUnit,
                                     UNIXInode * inode )
        : _parent( parent ), _expected( expected ), 
          _startUnit( startUnit ), _inode( inode ),
          _log2UnitSize( log2UnitSize ), _func( &func ) {}

};

void
UNIXInodeCacheWriteAsyncClosure::func( UNIXInodeCacheWriteAsyncClosure * c )
{
    unsigned bytesWritten = c->_result;
    int fragmentsWritten = 
    ( bytesWritten + ( ( 1 << c->_log2UnitSize ) - 1 ) ) >> 
    c->_log2UnitSize;
    Assert( c->_expected == fragmentsWritten );
    int allocated = c->_inode->alloc( c->_startUnit, fragmentsWritten );
    if ( allocated < fragmentsWritten ) {
        Console() 
        << "UNIXInodeCacheWriteAsyncClosure::func() out of space\n" << eor;
        fragmentsWritten = allocated;
    }
    if ( c->_parent ) {
        c->_parent->activate( fragmentsWritten );
    }
    delete c;
}

#endif // UNIXINODECACHWRITEALLOC

int 
UNIXInode::cacheWrite( unsigned int startUnit, unsigned int count,
                       char * addr, unsigned int nbytes, IOAsyncClosure * c )
{
    Raids( RaidProtected ) 
    << this << "::cacheWrite( " << startUnit << ", " 
    << count << ", " << asAddress( addr ) << ", " << nbytes << ", "
    << asAddress( c ) << " )\n" << eor;

    if ( count == 0 || nbytes == 0 ) {
        if ( c ) 
            c->activate( 0 );
        return 0;
    }

    // Assert( ( nbytes & ( unitSize() - 1 ) ) == 0 );
    Assert( nbytes <= ( count << log2UnitSize() ) );

    unsigned maxBytes = count << log2UnitSize();
    if ( nbytes > maxBytes ) 
        nbytes = maxBytes;

#ifdef UNIXINODECACHEWRITEALLOC
    unsigned fragmentsWritten;
    if ( c == 0 ) {
        unsigned bytesWritten = _cache->cacheWrite
        ( startUnit << log2UnitSize(), nbytes, addr, 0 );
        fragmentsWritten = 
        ( bytesWritten + ( unitSize() - 1 ) ) >> log2UnitSize();
        Assert( count == fragmentsWritten );
        int allocated = alloc( startUnit, fragmentsWritten );
        if ( allocated < fragmentsWritten ) {
            Console() << this << "::cacheWrite() out of space\n" << eor;
            fragmentsWritten = allocated;
        }
    } else {
        UNIXInodeCacheWriteAsyncClosure * myClosure =
        new UNIXInodeCacheWriteAsyncClosure( c, log2UnitSize(), count,
                                             startUnit, this );
        _cache->cacheWrite( startUnit << log2UnitSize(), nbytes, addr, 
                            myClosure );
        fragmentsWritten = count;
    }
#else // not UNIXINODECACHEWRITEALLOC
    unsigned fragmentsWritten = 
    FileObject::cacheWrite( startUnit, count, addr, nbytes, c );
#endif // UNIXINODECACHEWRITEALLOC

    Raids( RaidProtected ) 
    << this << "::cacheWrite return( "
    << fragmentsWritten << " )\n" << eor;

    return( fragmentsWritten );
}

typedef CacheIOState UNIXInodeAsyncState;
typedef CacheIOClosure UNIXInodeAsyncClosure;

#ifdef MULTIMAX
extern unsigned int getKernelStackPointer();
#endif
/*
 *  rawRead - retrieve data from an inode and store it in the buffer "dest"
 *  the data is found in the "count" logical fragments
 *  starting at "startUnit"
 *
 *  read optimizes reads of the file system by calling
 *  the file systems read function with block aligned
 *  requests if possible.
 */
int
UNIXInode::rawRead( unsigned startUnit, unsigned int count, char * dest, unsigned int nbytes,
                    IOAsyncClosure * c )
{
    Raids( RaidProtected ) 
    << this << "::rawRead( " << startUnit << ", "  
    << count << ", " << asHex( (int) dest ) << ", " << nbytes 
    << " )\n" << eor;

    if ( count == 0 || nbytes == 0 ) {
        if ( c )
            c->activate( 0 );
        return 0;
    }

    Assert( startUnit + count <= numberOfUnits() );
    // Assert( ( nbytes & ( unitSize() - 1 ) ) == 0 );
    Assert( nbytes == ( count << log2UnitSize() ) );

    int orgCount = count;
    int runLength = 0;
    int startRun = 0;
    int total = 0;
//    noOfReads++;

    UNIXInodeAsyncState * state = 0;
    unsigned firstStartUnit = 0;
    unsigned firstCount = 0;
    unsigned firstnBytes = 0;
    char * firstBuf = 0;

    _mutex.P();

    getDinode();

    while( count > 0 ) {
        MapUnitResult status = basicMapUnitRead( startUnit, count, startRun, runLength );
        count -= runLength;
        startUnit += runLength;
        unsigned bytes = runLength << log2UnitSize();
        Assert( status != MAP_FULL );
        if ( status == MAP_HOLE ) {
            Raids( RaidDetail ) 
            << this << "::rawRead() zeroing startRun " << startRun 
            << " runLength " << runLength << "\n" << eor;
            ByteZero( dest, bytes );
            dest += bytes;
            total += runLength;
            continue;
        }

        Raids( RaidDetail ) 
        << this << "::rawRead() startRun " << startRun 
        << " runLength " << runLength << "\n" << eor;

//        noOfReadsFromSource++;
        if ( firstBuf == 0 ) {
            firstStartUnit = startRun;
            firstCount = runLength;
            firstBuf = dest;
            firstnBytes = bytes;
        } else {
            if ( state == 0 ) {
                state = new UNIXInodeAsyncState( c );
                state->add( firstStartUnit, firstCount, firstBuf, firstnBytes );
            }
            state->add( startRun, runLength, dest, bytes );
        }
        dest += ( runLength << log2UnitSize() );
    }

    releaseDinode();

    _mutex.V();

    if ( state ) {
        state->_result = total;
        state->_expectedResult = orgCount;
        state->_spare = 63; //(unsigned) (&(UNIXInode::rawRead));
        UNIXInodeAsyncState::_Operation * op = state->_ops;
        int opCount = state->_count;
        for( int i = 0; i < opCount; i++ ) {
            fileObjectRead( op->_startUnit, op->_count, op->_buf, op->_spare,
                            &(op->_closure) );
            op++;
        }
        if ( c == 0 ) {
            state->_barrier.wait();
            total = state->_result;
            delete state;
        }
    } else if ( firstBuf ) {
        if ( c != 0 && total > 0 ) {
            c = new SumResultIOAsyncClosure( c, total );
        }
        total += fileObjectRead( firstStartUnit, firstCount, firstBuf,
                                 firstnBytes, c );
    } else if ( c ) {
        c->activate( total );
    }


    Raids( RaidProtected ) << this << "::rawRead return( "
                           << total << " )\n" << eor;
    return( total );
}

/*
 *  rawWrite - write the "count" logical fragments starting at
 *         "startUnit" from the buffer "src" to an inode.
 */
int 
UNIXInode::rawWrite( unsigned startUnit, unsigned int count, char * src, unsigned int nbytes,
                     int writeThrough, IOAsyncClosure * c )
{
    Raids( RaidProtected ) 
    << this << "::rawWrite( " << startUnit 
    << ", " << count << ", " << asHex( (int) src ) 
    << ", " << nbytes << ", " << writeThrough << " )\n" << eor;

    if ( count == 0 || nbytes == 0 ) {
        if ( c ) 
            c->activate( 0 );
        return 0;
    }

    Assert( startUnit + count <= numberOfUnits() );
    // Assert( ( nbytes & ( unitSize() - 1 ) ) == 0 );
    Assert( nbytes <= ( count << log2UnitSize() ) );

    if ( _creatingDictionary )
        synchronizeCreatingDictionary();

    _mutex.P();

    getDinode();

    // Update the write marks
    if( _highMark == -1 )   // First write, zero size file
        _lowMark = startUnit;   // Therefore the low is where we are going now
    if( _lowMark > (int)startUnit )
        _lowMark = startUnit;   // A new low
    if( (int)(startUnit + count - 1) > (int)_highMark )
        _highMark = startUnit + count -1;   // A new high
        
    int orgCount = count;
    int runLength = 0;
    int startRun;
    int total = 0;
//    noOfWrites++;

    UNIXInodeAsyncState * state = 0;
    unsigned firstStartUnit = 0;
    unsigned firstCount = 0;
    unsigned firstnBytes = 0;
    char * firstBuf = 0;

    while( count > 0 ) {

        MapUnitResult status = 
        basicMapUnitWrite( startUnit, count, startRun, runLength);
        count -= runLength;
        startUnit += runLength;
        Assert( status != MAP_HOLE );
        if ( status == MAP_FULL ) {
            Assert( runLength == 0 );
            Console() 
            << this << "::rawWrite() basicMapUnitWrite returns map full\n" 
            << eor;
            break;
        }
        Raids( RaidDetail ) 
        << this << "::rawWrite() startRun " << startRun 
        << " runLength " << runLength << "\n" << eor;
//        noOfWritesToSource++;
        int bytes = runLength << log2UnitSize();
        if ( firstBuf == 0 ) {
            firstStartUnit = startRun;
            firstCount = runLength;
            firstBuf = src;
            firstnBytes = bytes;
        } else {
            if ( state == 0 ) {
                state = new UNIXInodeAsyncState( c );
                state->add( firstStartUnit, firstCount, firstBuf, firstnBytes );
            }
            state->add( startRun, runLength, src, bytes );
        }    
        src += ( runLength << log2UnitSize() );
    }

    releaseDinode();

    _mutex.V();

    if ( state ) {
        state->_result = total;
        state->_expectedResult = orgCount;
        state->_spare = 83; //(unsigned) &(UNIXInode::rawWrite);
        UNIXInodeAsyncState::_Operation * op = state->_ops;
        int opCount = state->_count;
        for( int i = 0; i < opCount; i++ ) {
            fileObjectWrite( op->_startUnit, op->_count, op->_buf, op->_spare,
                             writeThrough, &(op->_closure) );
            op++;
        }
        if ( c == 0 ) {
            state->_barrier.wait();
            total = state->_result;
            delete state;
        }
    } else if ( firstBuf ) {
        if ( c != 0 && total > 0 ) {
            c = new SumResultIOAsyncClosure( c, total );
        }
        total += fileObjectWrite( firstStartUnit, firstCount, firstBuf,
                                  firstnBytes, writeThrough, c );
    } else if ( c ) {
        c->activate( total );
    }

    Raids( RaidProtected ) << this << "::rawWrite return( "
                           << total << " )\n" << eor;
    return( total ) ;
}

int
UNIXInode::basicCopy( MemoryObject * dest )
{
    Raid( RaidProtected )( "%N::basicCopy( %N )\n", this, dest );

    if ( _cache  != 0)
        _cache->pseudoSynchronize();

    int returnValue = 0;
    dest->setNumberOfUnits( 0 );
    int log2UnitDifference = log2UnitSize() + log2BlockSize - 
    dest->log2UnitSize();
    int readUnits = blockSize();
    int writeUnits = 1;
    int bufferSize = blockSize() << log2UnitSize();
    if( log2UnitDifference >= 0 ) {
    
        if( log2UnitDifference > 0 ) {
            writeUnits <<= log2UnitDifference;
        }
    
        char * buffer = new char[bufferSize];
        int w = _lowMark - (_lowMark%readUnits);
        for( int r = _lowMark - (_lowMark%readUnits); r <= _highMark;
             r += readUnits, w += writeUnits ) {
            int physicalUnit = mapUnit( r >> log2BlockSize, 0 );
            if( physicalUnit > 0 ) {
                int status = read( r, readUnits, buffer, bufferSize );
                if( ( status < readUnits ) && ( r+status < (int)numberOfUnits() ) ){
                    returnValue = ReadError;
                    break;
                }
                if( dest->write( w, writeUnits, buffer, bufferSize ) !=
                    writeUnits ) {
                    returnValue = WriteError;
                    break;
                }
            }
        }
        delete buffer;
    }
    else if( log2UnitDifference < 0 ) {
        log2UnitDifference = -log2UnitDifference;
        readUnits <<= log2UnitDifference;
        bufferSize <<= log2UnitDifference;
        char * buffer = new char[bufferSize];
        int unitDifference = 1 << log2UnitDifference;
        unsigned w = 0;
        for( unsigned r = 0; r < numberOfUnits(); r += readUnits, w += writeUnits ){
        
            int doReadWrite = 0;
            for( int u = 0; u < unitDifference; u++ ) {
                doReadWrite |= mapUnit( (r + u) >> log2BlockSize, 0 );
            }

            if( doReadWrite > 0 ) {
                int status = read( r, readUnits, buffer, bufferSize );
                if( ( status < readUnits ) && ( r+status < numberOfUnits() ) ){
                    returnValue = ReadError;
                    break;
                }
                if( dest->write( w, writeUnits, buffer, bufferSize ) !=
                    writeUnits ) {
                    returnValue = WriteError;
                    break;
                }
            }
        }
        delete buffer;
    }

    if( dest->isKindOf( PersistentMemoryObjectClass ) ) {
        ( ( PersistentMemoryObject * )dest ) ->
        setRecordSize( recordSize() );
        ( ( PersistentMemoryObject * ) dest ) ->
        setNumberOfRecords( numberOfRecords() ); 
        ( ( PersistentMemoryObject * ) dest ) ->
        setNumberOfBytes( numberOfBytes() );
    }

    Raid( RaidProtected )( "%N::basicCopy done ( %N )\n", this, dest );
    return( returnValue );
}

/*
 * Initialize the in-memory inode
 */
UNIXInode::UNIXInode( unsigned int numberOfUnits,
                      int log2USize,
                      int inumber,
                      MemoryObjectContainer * container,
                      PersistentMemoryObject * source,
                      int log2blocksize )
    : FileObject( numberOfUnits, log2USize, inumber, container, source, 0 )
{
    initDebug();

    little_endian_fs = false;

#ifdef FLATTEN
    _baseSource = ((FileObject *) source)->baseSource();
    _baseOffset = ((FileObject *) source)->baseOffset(); 
    _baseShift = ((FileObject *) source)->baseShift() + _unitShift;
/*
  Console() << this << " baseSource " << _baseSource << " baseOffset "
  << _baseOffset << " baseShift " << (int) _baseShift << "\n" 
  << eor;
*/
#endif /* FLATTEN */

    _lowMark = 0;
    _highMark = numberOfUnits - 1;
    
    Raid( RaidConstructor )( "%N::UNIXInode( inumber %d, log2blockSize %d )\n",
                             this, inumber, log2blocksize );

    Assert( _container != 0 );

    _modifiedFlag = 0;
    _myModifiedFlag = 0;

    // Initialize variables from parameters
    log2BlockSize = log2blocksize;

    indirect = UNIXInodeIndirect::indirect( log2BlockSize + log2UnitSize() );
    Assert(indirect != 0);

    // Clear index buffer cache variables
    direct = singleIndex = doubleIndex = tripleIndex = 0;
    singleBuffer = doubleBuffer = tripleBuffer= extra2Buffer = extra3Buffer = 0;

    // Set up indirect variables for mapUnit function.
    const unsigned log2IntegerSize = 2;
    Assert( sizeof(int) == ( 1 << log2IntegerSize ) );
    indexShift = log2BlockSize + log2UnitSize() - log2IntegerSize;
    indexMask = ( 1 << indexShift ) - 1;
    fragmentsInLastBlock = 0;

    // Zero stat counters
//    noOfWrites = noOfReads = noOfWritesToSource = noOfReadsFromSource = 0;
}

int UNIXInodeInfo = 0;

UNIXInode::~UNIXInode()
{
    Raid( RaidConstructor )( "%N::~UNIXInode()\n", this );

    synchronizeInode();

//    if ( UNIXInodeInfo ) {
//  Console() << this << "::~UNIXInode noOfReads " << noOfReads
//      << " noOfReadsFromSource " << noOfReadsFromSource
//      << " noOfWrites " << noOfWrites
//      << " noOfWritesToSource " << noOfWritesToSource
//      << "\n" << eor;
//    }
}

void
UNIXInode::setModified()
{
    * _modifiedFlag = 1;
    _myModifiedFlag = 1;
}

void UNIXInode::synchronizeInode()
{
    _mutex.P();
    indirect->synchronize( _source, _idNumber );
    _mutex.V();
}

void
UNIXInode::basicFlush()
{
    Raids( RaidProtected ) 
    << this << "::basicFlush()\n" << eor;

    if ( transient() )
        return;
    MemoryObject::basicFlush();
    synchronizeInode();
    if ( _myModifiedFlag && referenceCount() > 0 ) {
        _container->synchronize( _idNumber );
        _myModifiedFlag = 0;
    }
}

void
UNIXInode::basicSynchronizeClosure
( PersistentMemoryObject::BasicSynchronizeAsyncClosure * c )
{
    if ( c->_state == 0 ) {
        c->_state = 1;
        TheAsyncSupport->schedule( c );
        return;
    }
    if ( c->_state == 1 ) {
        synchronizeInode();
        if ( _myModifiedFlag ) {
            _container->synchronize( _idNumber );
            _myModifiedFlag = 0;
        }
        c->_parent->activate();
        delete c;
        return;
    }
}

void
UNIXInode::basicSynchronize( IOAsyncClosure * c )
{
    Raids( RaidProtected )
    << this << "::basicSynchronize()\n" << eor;
    
    if ( transient() ) {
        if ( c != 0 ) {
            c->activate();
        }
        return;
    }
    if ( c != 0 ) {
        c = new BasicSynchronizeAsyncClosure( c, this );
    }
    MemoryObject::basicSynchronize( c );
    if ( c == 0 ) {
        synchronizeInode();
        if ( _myModifiedFlag ) {
            _container->synchronize( _idNumber );
            _myModifiedFlag = 0;
        }
    }
}

void
UNIXInode::basicSynchronize( unsigned from, unsigned to, IOAsyncClosure * c )
{
    Raids( RaidProtected )
    << this << "::basicSynchronize( " << from << ", " << to 
    << " )\n" << eor;

    if ( transient() ) {
        if ( c != 0 ) {
            c->activate();
        }
        return;
    }
    if ( c != 0 ) {
        c = new BasicSynchronizeAsyncClosure( c, this );
    }
    MemoryObject::basicSynchronize( from, to, c );
    if ( c == 0 ) {
        synchronizeInode();
        if ( _myModifiedFlag ) {
            _container->synchronize( _idNumber );
            _myModifiedFlag = 0;
        }
    }
}

int
UNIXInode::alloc( unsigned startUnit, int count )
{
    _mutex.P();

    int total = 0;

    // Update the write marks

    if( _highMark == -1 )   // First write, zero size file
        _lowMark = startUnit;   // Therefore the low is where we are going now
    if( _lowMark > (int)startUnit )
        _lowMark = startUnit;   // A new low
    if( (int)(startUnit + count - 1) > (int)_highMark )
        _highMark = startUnit + count -1;   // A new high

    int startRun;
    int runLength;

    getDinode();

    while( count > 0 ) {

        MapUnitResult status = 
        basicMapUnitWrite( (int)startUnit, count, startRun, runLength );
        count -= runLength;
        total += runLength;
        startUnit += runLength;

        Assert( status != MAP_HOLE );
        if ( status == MAP_FULL ) {
            Assert( runLength == 0 );
            Console() 
            << this << "::alloc() basicMapUnitWrite returns map full\n" 
            << eor;
            break;
        }
    }

    releaseDinode();

    _mutex.V();

    return total;
}

int
UNIXInode::free( unsigned startUnit, int count )
{
    Raids( RaidMethod ) 
    << this << "::free( " << startUnit << ", " << count << " ) " 
    << _numberOfUnits << eor;

    Assert( count > 0 );
    Assert( log2BlockSize == 0 || 
            ( (startUnit & blockMask()) == 0 && (count & blockMask()) == 0 ));

    _mutex.P();
    getDinode();
 
    basicMapUnitFree( startUnit, count );    

    releaseDinode();
    _mutex.V();

    return count;
}

/*
 *  mapUnit - returns the file system block number corresponding
 *              to the given inode logical block number
 *
 */

int 
UNIXInode::mapUnit( int logicalBlock, int addMapping )
{
    _mutex.P();
    getDinode();
    int result = basicMapUnit( logicalBlock, addMapping );
    Assert( result >= 0 && result < (int)( _source->numberOfUnits() >> _unitShift ) );
    releaseDinode();
    _mutex.V();
    return result;
}

int
UNIXInode::basicMapUnit( int logicalBlock, int addMapping )
{
    Raid( RaidProtected )( "%N::basicMapUnit( %d, %d )\n",
                           this, logicalBlock, addMapping );
    Debug.printf( "%N::basicMapUnit( %d, %d )\n",
                           this, logicalBlock, addMapping );

    int physicalBlock = 0;
    unsigned int singleIndex_deref = *singleIndex;
    unsigned int doubleIndex_deref = *doubleIndex;
    unsigned int tripleIndex_deref = *tripleIndex;
    if (little_endian_fs) {
        swapl((void *) &singleIndex_deref);
        swapl((void *) &doubleIndex_deref);
        swapl((void *) &tripleIndex_deref);
    }

    Assert( logicalBlock >= 0 );
#ifdef ASSERT
    unsigned sourceNoOfUnits = _source->numberOfUnits() >> _unitShift;
#endif

    // This logical block is a direct block
    if( logicalBlock < (int)singleIndirect ) {

        physicalBlock = direct[ logicalBlock ];

        if (little_endian_fs) {
            swapl((void *) &physicalBlock);
        }

        // If there is no block and the flag is set, allocate a new block.
        if( ( physicalBlock == 0 ) && ( addMapping == 1 ) ) {
            physicalBlock = allocBlock();
            if ( physicalBlock == 0 ) {
                Raids( RaidError ) 
                << this << "::basicMapUnit() alloc direct block err\n"
                << eor;
            }
            direct[ logicalBlock ] = physicalBlock;

            if (little_endian_fs) {
                swapl((void *) &(direct[logicalBlock]));
            }

            setModified();
            Assert( physicalBlock > 0 && physicalBlock < (int)sourceNoOfUnits );
        } else if( ( physicalBlock != 0 ) && ( addMapping == -1 ) ) {
            freeBlock( physicalBlock );
            direct[ logicalBlock ] = 0;
            setModified();
        }
        Assert( physicalBlock >= 0 && physicalBlock < (int)sourceNoOfUnits );
    }

    // This logical block is a single indirect block
    else if( logicalBlock < (int)doubleIndirect ) {

        // Adjust logical block for easy arithmetic
        logicalBlock -= singleIndirect;

        if( *singleIndex == 0 ) {

            // Return 0 if this is not an addblock.
            if( addMapping == 0 || addMapping == -1 ) {
                return( 0 );
            }

            // Allocate the single indirect block.
            *singleIndex = allocBlock();
        
            if( *singleIndex == 0 ) {
                Raids( RaidError ) 
                << this 
                << "::basicMapUnit() alloc single indirect index err\n"
                << eor;
                return( 0 );
            }
            singleBuffer = indirect->newIndirect
            ( _source, *singleIndex << _unitShift, _idNumber );

            singleIndex_deref = *singleIndex;

            if (little_endian_fs) {
                swapl((void *) singleIndex);
            }

            setModified();

        } 
        // Read into the single indirect cache.
        else {

            singleBuffer = indirect->getIndirect
            ( _source, singleIndex_deref << _unitShift, 
              _idNumber, singleBuffer );
        }

        Assert( singleIndex_deref > 0 && singleIndex_deref < sourceNoOfUnits );

        // Get the physical block from the single indirect cache.
        physicalBlock = *( singleBuffer->buf( logicalBlock ) );
        if (little_endian_fs) {
            swapl((void *) &physicalBlock);
        }

        if( ( physicalBlock == 0 ) && ( addMapping == 1 ) ) {
            physicalBlock = allocBlock();
            if( physicalBlock == 0 ) {
                Raids( RaidError ) 
                << this << "::basicMapUnit() alloc single indirect block err\n"
                << eor;
                indirect->releaseIndirect( singleBuffer, 0 );
                return( 0 );
            }
            *( singleBuffer->buf( logicalBlock ) ) = physicalBlock;

            if (little_endian_fs) {
                swapl((void *) singleBuffer->buf( logicalBlock ));
            }

            indirect->releaseIndirect( singleBuffer, 1 );
        } else if( ( physicalBlock != 0 ) && ( addMapping == -1 ) ) {
            freeBlock( physicalBlock );
            *( singleBuffer->buf( logicalBlock ) ) = 0;
            indirect->releaseIndirect( singleBuffer, 1 );
        } else {
            indirect->releaseIndirect( singleBuffer, 0 );
        }
        Assert( physicalBlock >= 0 && physicalBlock < (int)sourceNoOfUnits );
    }

    // This logical block is a double indirect block
    else if( logicalBlock < (int)tripleIndirect ) {

        // Adjust logical block for easy arithmetic.
        logicalBlock -= doubleIndirect;

        // The double indirect cache is unused.
        if( *doubleIndex == 0 ) {

            // Return 0 if this is not an addblock.
            if( addMapping == 0 || addMapping == -1 ) {
                return( 0 );
            }

            // Allocate the double indirect block.
            *doubleIndex = allocBlock();
            if( *doubleIndex == 0 ) { 
                Raids( RaidError ) 
                << this 
                << "::basicMapUnit() alloc double indirect index err\n"
                << eor;
                return( 0 );
            }
            doubleBuffer = indirect->newIndirect
            ( _source, *doubleIndex << _unitShift, _idNumber );

            doubleIndex_deref = *doubleIndex;

            if (little_endian_fs) {
                swapl((void *) doubleIndex);
            }

            setModified();
        }
        else {

            doubleBuffer = indirect->getIndirect
            ( _source, doubleIndex_deref << _unitShift , _idNumber, 
              doubleBuffer );
        }

        Assert( doubleIndex_deref > 0 && doubleIndex_deref < sourceNoOfUnits );

        // Get the block number of the second-level indirect block.
        physicalBlock = 
        *( doubleBuffer->buf( calcDoubleIndex( logicalBlock ) ) );

        if (little_endian_fs) {
            swapl((void *) &physicalBlock);
        }

        // There is no second-level block.
        if( physicalBlock == 0 ) {

            // Return 0 if this is not an addblock.
            if( addMapping == 0 || addMapping == -1 ) {
                indirect->releaseIndirect( doubleBuffer, 0 );
                return( 0 );
            }

            physicalBlock = allocBlock();
            if( physicalBlock == 0 ) {
                Raids( RaidError ) 
                << this 
                << "::basicMapUnit() alloc double indirect 2 index err\n"
                << eor;
                indirect->releaseIndirect( doubleBuffer, 0 );
                return( 0 );
            }
            extra2Buffer = indirect->newIndirect
            ( _source, physicalBlock << _unitShift, _idNumber );
            *( doubleBuffer->buf( calcDoubleIndex( logicalBlock ) ) )
            = physicalBlock;

            if (little_endian_fs) {
                swapl((void *) doubleBuffer->buf(calcDoubleIndex(logicalBlock)));
            }

            indirect->releaseIndirect( doubleBuffer, 1 );
        }

        // Read in the single indirect block.
        else {
            extra2Buffer = indirect->getIndirect
            ( _source, physicalBlock << _unitShift , _idNumber,
              extra2Buffer );
            indirect->releaseIndirect( doubleBuffer, 0 );
        }
        Assert( physicalBlock > 0 && physicalBlock < (int)sourceNoOfUnits );

        // Get the physical block from the single indirect cache.
        physicalBlock = 
        *( extra2Buffer->buf( calcSingleIndex( logicalBlock ) ) );
        if (little_endian_fs) {
            swapl((void *) &physicalBlock);
        }

        if( ( physicalBlock == 0 ) && ( addMapping == 1 ) ) {
            physicalBlock = allocBlock();
            if( physicalBlock == 0 ) {
                Raids( RaidError ) 
                << this << "::basicMapUnit() alloc double indirect block err\n"
                << eor;
                indirect->releaseIndirect( extra2Buffer, 0 );
                return( 0 );
            }
            *( extra2Buffer->buf( calcSingleIndex( logicalBlock ) ) ) = 
            physicalBlock;
            if (little_endian_fs) {
                swapl((void *) extra2Buffer->buf(calcSingleIndex(logicalBlock)));
            }

            indirect->releaseIndirect( extra2Buffer, 1 );
        } else if( ( physicalBlock != 0 ) && ( addMapping == -1 ) ) {
            freeBlock( physicalBlock );
            *( extra2Buffer->buf( calcSingleIndex( logicalBlock ) ) ) = 0;
            indirect->releaseIndirect( extra2Buffer, 1 );
        } else {
            indirect->releaseIndirect( extra2Buffer, 0 );
        }
        Assert( physicalBlock >= 0 && physicalBlock < (int)sourceNoOfUnits );
    }

    // This logical block is a triple indirect block.
    else {

        // Adjust logical block for easy arithmetic
        logicalBlock -= tripleIndirect;

        // The triple indirect cache is unused.
        if( *tripleIndex == 0 ) {

            // Return 0 if this is not an addblock.
            if( addMapping == 0 || addMapping == -1 ) {
                return( 0 );
            }

            // Allocate the triple indirect block.
            *tripleIndex = allocBlock();
            if( *tripleIndex == 0 ) {
                Raids( RaidError ) 
                << this 
                << "::basicMapUnit() alloc triple indirect index err\n"
                << eor;
                return( 0 );
            }
            tripleBuffer = indirect->newIndirect
            ( _source, *tripleIndex << _unitShift, _idNumber );

            tripleIndex_deref = *tripleIndex;

            if (little_endian_fs) {
                swapl((void *) tripleIndex);
            }

            setModified();
        }
        else {

            tripleBuffer = indirect->getIndirect
            ( _source, tripleIndex_deref << _unitShift, _idNumber,
              tripleBuffer );
        }

        Assert( tripleIndex_deref > 0 && tripleIndex_deref < sourceNoOfUnits );

        // Get the block number of the second-level indirect block.
        physicalBlock = 
        *( tripleBuffer->buf( calcTripleIndex( logicalBlock ) ) );

        if (little_endian_fs) {
            swapl((void *) &physicalBlock);
        }

        // There is no second-level indirect block.
        if( physicalBlock == 0 ) {

            // Return 0 if this is not an addblock.
            if( addMapping == 0 || addMapping == -1 ) {
                indirect->releaseIndirect( tripleBuffer, 0 );
                return( 0 );
            }

            physicalBlock = allocBlock();
            if( physicalBlock == 0 ) {
                Raids( RaidError ) 
                << this << "::basicMapUnit() alloc triple indirect 2 idx err\n"
                << eor;
                indirect->releaseIndirect( tripleBuffer, 0 );
                return( 0 );
            }

            *( tripleBuffer->buf( calcTripleIndex( logicalBlock ) ) ) = 
            physicalBlock;

            if (little_endian_fs) {
                swapl((void *) tripleBuffer->buf(calcTripleIndex(logicalBlock)));
            }

            indirect->releaseIndirect( tripleBuffer, 1 );
            extra2Buffer = indirect->newIndirect
            ( _source, physicalBlock << _unitShift, _idNumber );
        }

        // Read in the second-level indirect block.
        else {
            indirect->releaseIndirect( tripleBuffer, 0 );
            extra2Buffer = indirect->getIndirect
            ( _source, physicalBlock << _unitShift, _idNumber,
              extra2Buffer );
        }

        Assert( physicalBlock > 0 && physicalBlock < (int)sourceNoOfUnits );

        // find the block number of the third-level indirect block.
        physicalBlock = 
        *( extra2Buffer->buf( calcDoubleIndex( logicalBlock ) ) );

        if (little_endian_fs) {
            swapl((void *) &physicalBlock);
        }

        // There is no third-level indirect block.
        if( physicalBlock == 0 ) {

            // Return 0 if this is not an addblock.
            if( addMapping == 0 || addMapping == -1 ) {
                indirect->releaseIndirect( extra2Buffer, 0 );
                return( 0 );
            }

            physicalBlock = allocBlock();
            if( physicalBlock == 0 ) {
                Raids( RaidError ) 
                << this 
                << "::basicMapUnit() alloc triple indirect 3 index err\n"
                << eor;
                indirect->releaseIndirect( extra2Buffer, 0 );
                return( 0 );
            }
            *( extra2Buffer->buf( calcDoubleIndex( logicalBlock ) ) ) = 
            physicalBlock;

            if (little_endian_fs) {
                swapl((void *) extra2Buffer->buf(calcDoubleIndex(logicalBlock)));
            }

            indirect->releaseIndirect( extra2Buffer, 1 );
            extra3Buffer = indirect->newIndirect
            ( _source, physicalBlock << _unitShift, _idNumber );
        }

        // Read in the single indirect block.
        else {
            indirect->releaseIndirect( extra2Buffer, 0 );
            extra3Buffer = indirect->getIndirect
            ( _source, physicalBlock << _unitShift, _idNumber,
              extra3Buffer );
        }

        Assert( physicalBlock > 0 && physicalBlock < (int)sourceNoOfUnits );

        // Get the physical block from the single indirect cache.
        physicalBlock = 
        *( extra3Buffer->buf( calcSingleIndex( logicalBlock ) ) );

        if (little_endian_fs) {
            swapl((void *) &physicalBlock);
        }

        if( ( physicalBlock == 0 ) && ( addMapping == 1 ) ) {
            physicalBlock = allocBlock();
            if( physicalBlock == 0 ) {
                Raids( RaidError ) 
                << this 
                << "::basicMapUnit() alloc triple indirect block err\n"
                << eor;
                indirect->releaseIndirect( extra3Buffer, 0 );
                return( 0 );
            }
            *( extra3Buffer->buf( calcSingleIndex( logicalBlock ) ) ) =
            physicalBlock;

            if (little_endian_fs) {
                swapl((void *) extra3Buffer->buf(calcSingleIndex(logicalBlock)));
            }

            indirect->releaseIndirect( extra3Buffer, 1 );
        } else if( ( physicalBlock != 0 ) && ( addMapping == -1 ) ) {
            freeBlock( physicalBlock );
            *( extra3Buffer->buf( calcSingleIndex( logicalBlock ) ) ) = 0;
            indirect->releaseIndirect( extra3Buffer, 1 );
        } else {
            indirect->releaseIndirect( extra3Buffer, 0 );
        }

        Assert( physicalBlock >= 0 && physicalBlock < (int)sourceNoOfUnits );
    }

    Raid( RaidProtected )( "%N::mapUnit returns %d\n", this, physicalBlock );

    return( physicalBlock );
}

int
UNIXInode::basicMapReadAssist( int * startCheck, int checkCount, 
                               int & remainingCount,
                               int & startRun,
                               int & expectedPhysicalBlock,
                               MapUnitResult & status )
{
    Assert( startCheck == 0 || 
            ( checkCount > 0 && checkCount <= ( indexMask + 1 ) ) );
    Assert( remainingCount > 0 );
    int blockSize = 1 << log2BlockSize;
    if ( checkCount > remainingCount )
        checkCount = remainingCount;
    int myCount = 0;
    if ( status == MAP_UNKNOWN ) {
        if ( startCheck == 0 ) {
            startRun = 0;
            status = MAP_HOLE;
            myCount = checkCount;
            expectedPhysicalBlock = 0;
        } else {
            int lastPhysicalBlock = *startCheck;
            if (little_endian_fs) {
                swapl((void *) &lastPhysicalBlock);
            }
            startRun = lastPhysicalBlock;
            if ( lastPhysicalBlock == 0 ) {
                status = MAP_HOLE;
                myCount++;
                startCheck++;
                while ( myCount < checkCount && *startCheck == 0 ) {
                    myCount++;
                    startCheck++;
                }
                expectedPhysicalBlock = 0;
            } else {
                status = MAP_REAL;
                myCount++;
                startCheck++;
                expectedPhysicalBlock = lastPhysicalBlock + blockSize;

                int startCheck_deref = *startCheck;
                if(little_endian_fs) {
                    swapl((void *) &startCheck_deref);
                }

                while ( myCount < checkCount && 
                        startCheck_deref == expectedPhysicalBlock ) {
                    myCount++;
                    startCheck++;
                    expectedPhysicalBlock += blockSize;
                    startCheck_deref = *startCheck;
                    if(little_endian_fs) {
                        swapl((void *) &startCheck_deref);
                    }
                }
            }
        }
    } else if ( status == MAP_HOLE ) {
        Assert( expectedPhysicalBlock == 0 );
        if ( startCheck == 0 ) {
            myCount = checkCount;
        } else {
            while ( myCount < checkCount && *startCheck == 0 ) {
                myCount++;
                startCheck++;
            }
        }
    } else {
        Assert( status == MAP_REAL ); 
        Assert( expectedPhysicalBlock != 0 );
        if ( startCheck != 0 ) {

            int startCheck_deref = *startCheck;
            if(little_endian_fs) {
                swapl((void *) &startCheck_deref);
            }

            while ( myCount < checkCount && 
                    startCheck_deref == expectedPhysicalBlock ) {
                myCount++;
                startCheck++;
                expectedPhysicalBlock += blockSize;
                startCheck_deref = *startCheck;
                if(little_endian_fs) {
                    swapl((void *) &startCheck_deref);
                }
            }
        }
    }
    remainingCount -= myCount;
    Raids( RaidProtected )
    << this << "::basicMapReadAssist()"
    << " checkCount " << checkCount 
    << " myCount " << myCount
    << " remaining " << remainingCount
    << " startRun " << startRun 
    << " expected " << expectedPhysicalBlock
    << " status " << status << "\n" << eor;
    return myCount;
}

MapUnitResult
UNIXInode::basicMapUnitRead( int start, int count, 
                             int & startRun, int & runLength )
{
    Raids( RaidMethod )
    << this << "::basicMapUnitRead( start " << start << ", count "
    << count << " )\n" << eor;

#ifdef ASSERT
    unsigned sourceNoOfUnits = _source->numberOfUnits() >> _unitShift;
#endif
    int startBlock = start >> log2BlockSize;
    int orgStartBlock = startBlock;
    int endBlock = ( start + count - 1 ) >> log2BlockSize;
    int fragmentOffset = start & blockMask();
    int remainingBlocks = endBlock - startBlock + 1;
    MapUnitResult status = MAP_UNKNOWN;
    int * startCheck;
    int checkCount;
    int expectedPhysicalBlock;
    int actual = 1;

    while ( startBlock <= endBlock && actual > 0 ) {
        // direct block
        if( startBlock < (int)singleIndirect ) {
            checkCount = singleIndirect - startBlock;

            startCheck = (int*)direct + startBlock;
            Raids( RaidDetail )
            << this << "::basicMapUnitRead() direct"
            << " startBlock " << startBlock << "\n" << eor;
            actual = basicMapReadAssist( startCheck, checkCount,
                                         remainingBlocks, 
                                         startRun,
                                         expectedPhysicalBlock, status );
        }
        // single indirect
        else if( startBlock < (int)doubleIndirect ) {
            checkCount = doubleIndirect - startBlock;
            // Adjust logical block for easy arithmetic
            unsigned logicalBlock = startBlock - singleIndirect;

            unsigned int singleIndex_deref = *singleIndex;
            if (little_endian_fs) {
                swapl((void *) &singleIndex_deref);
            }

            Raids( RaidDetail )
            << this << "::basicMapUnitRead() singleIndex "
            << singleIndex_deref << "\n" << eor;
            // There is no single indirect block.
            if( *singleIndex == 0 ) {
                Raids( RaidDetail )
                << this << "::basicMapUnitRead() no single indirect"
                << " startBlock " << startBlock 
                << " logicalBlock " << logicalBlock << "\n" << eor;
                actual = basicMapReadAssist( 0, checkCount, 
                                             remainingBlocks,
                                             startRun,
                                             expectedPhysicalBlock, 
                                             status );
            }
            // Read into the single indirect cache.
            else {

                unsigned int singleIndex_deref = *singleIndex;
                if (little_endian_fs) {
                    swapl((void *) &singleIndex_deref);
                }

                Assert( singleIndex_deref > 0 && singleIndex_deref < sourceNoOfUnits );
                singleBuffer = indirect->getIndirect
                ( _source, singleIndex_deref << _unitShift, 
                  _idNumber, singleBuffer );
                startCheck = singleBuffer->buf( logicalBlock );
                Raids( RaidDetail )
                << this << "::basicMapUnitRead() single indirect"
                << " startBlock " << startBlock
                << " logicalBlock " << logicalBlock << "\n" << eor;
                actual = basicMapReadAssist( startCheck, checkCount, 
                                             remainingBlocks,
                                             startRun,
                                             expectedPhysicalBlock, 
                                             status );
                indirect->releaseIndirect( singleBuffer, 0 );
            }
        }
        // double indirect
        else if( startBlock < (int)tripleIndirect ) {
            // Adjust logical block for easy arithmetic.
            unsigned logicalBlock = startBlock - doubleIndirect;

            unsigned int doubleIndex_deref = *doubleIndex;
            if (little_endian_fs) {
                swapl((void *) &doubleIndex_deref);
            }
        
            Raids( RaidDetail ) 
            << this << "::basicMapUnitRead() doubleIndex "
            << doubleIndex_deref << "\n" << eor;
            // There is no double indirect block.
            if( *doubleIndex == 0 ) {
                checkCount = tripleIndirect - startBlock;
                Raids( RaidDetail )
                << this << "::basicMapUnitRead() no double indirect"
                << " startBlock " << startBlock
                << " logicalBlock " << logicalBlock << "\n" << eor;
                actual = basicMapReadAssist( 0, checkCount, 
                                             remainingBlocks,
                                             startRun,
                                             expectedPhysicalBlock, 
                                             status );
            }
            // Read into the double indirect cache.
            else {
                Assert( doubleIndex_deref > 0 && doubleIndex_deref < sourceNoOfUnits );
                doubleBuffer = indirect->getIndirect
                ( _source, doubleIndex_deref << _unitShift, _idNumber,
                  doubleBuffer );
                // Get the block number of the second-level indirect block.
                unsigned physicalBlock =
                *( doubleBuffer->buf( calcDoubleIndex( logicalBlock ) ) );

                if (little_endian_fs) {
                    swapl((void *) &physicalBlock);
                }

                indirect->releaseIndirect( doubleBuffer, 0 );
                checkCount = 
                indexMask + 1 - calcSingleIndex( logicalBlock );
                Raids( RaidDetail ) 
                << this << "::basicMapUnitRead() doubleIndex2 "
                << physicalBlock << "\n" << eor;
                if ( physicalBlock == 0 ) {
                    Raids( RaidDetail )
                    << this << "::basicMapUnitRead() no double 2 indirect"
                    << " startBlock " << startBlock
                    << " logicalBlock " << logicalBlock 
                    << " offset " << calcSingleIndex( logicalBlock )
                    << "\n" << eor;
                    actual = basicMapReadAssist( 0, checkCount,
                                                 remainingBlocks,
                                                 startRun,
                                                 expectedPhysicalBlock,
                                                 status );
                } 
                // Read in the single indirect block.
                else {
                    Assert( physicalBlock > 0 && 
                            physicalBlock < sourceNoOfUnits );
                    extra2Buffer = indirect->getIndirect
                    ( _source, physicalBlock << _unitShift , _idNumber,
                      extra2Buffer );
                    startCheck = 
                    extra2Buffer->buf( calcSingleIndex( logicalBlock ) );

                    Raids( RaidDetail )
                    << this << "::basicMapUnitRead() double 2 indirect"
                    << " startBlock " << startBlock
                    << " logicalBlock " << logicalBlock 
                    << " offset " << calcSingleIndex( logicalBlock )
                    << "\n" << eor;
                    actual = basicMapReadAssist( startCheck, checkCount,
                                                 remainingBlocks,
                                                 startRun,
                                                 expectedPhysicalBlock,
                                                 status );
                    indirect->releaseIndirect( extra2Buffer, 0 );
                }
            }
        }
        // This logical block is a triple indirect block.
        else {
            // Adjust logical block for easy arithmetic
            unsigned logicalBlock = startBlock - tripleIndirect;

            unsigned int tripleIndex_deref = *tripleIndex;
            if (little_endian_fs) {
                swapl((void *) &tripleIndex_deref);
            }
        
            Raids( RaidDetail ) 
            << this << "::basicMapUnitRead() tripleIndex "
            << tripleIndex_deref << "\n" << eor;
            // There is no triple indirect block.   
            if( *tripleIndex == 0 ) {
                Raids( RaidDetail )
                << this << "::basicMapUnitRead() no triple indirect"
                << " startBlock " << startBlock
                << " logicalBlock " << logicalBlock << "\n" << eor;
                checkCount = ( 1 << ( 3 * indexShift ) ) - logicalBlock;
                actual = basicMapReadAssist( 0, checkCount, 
                                             remainingBlocks,
                                             startRun,
                                             expectedPhysicalBlock, 
                                             status );
            }
            // Read into the triple indirect cache.
            else {
                Assert( tripleIndex_deref > 0 && tripleIndex_deref < sourceNoOfUnits );
                tripleBuffer = indirect->getIndirect
                ( _source, tripleIndex_deref << _unitShift, _idNumber,
                  tripleBuffer );
                // Get the block number of the second-level indirect block.
                unsigned physicalBlock = 
                *( tripleBuffer->buf( calcTripleIndex( logicalBlock ) ) );

                if (little_endian_fs) {
                    swapl((void *) &physicalBlock);
                }

                indirect->releaseIndirect( tripleBuffer, 0 );
                Raids( RaidDetail ) 
                << this << "::basicMapUnitRead() tripleIndex2 "
                << physicalBlock << "\n" << eor;
                // There is no second-level indirect block.
                if( physicalBlock == 0 ) {
                    Raids( RaidDetail )
                    << this << "::basicMapUnitRead() no triple 2 indirect"
                    << " startBlock " << startBlock
                    << " logicalBlock " << logicalBlock << "\n" << eor;
                    unsigned maxcount = 1 << ( indexShift + indexShift );
                    checkCount = maxcount - ( logicalBlock & ( maxcount - 1 ) );
                    actual = basicMapReadAssist( 0, checkCount, 
                                                 remainingBlocks,
                                                 startRun,
                                                 expectedPhysicalBlock, 
                                                 status );
                } 
                // Read in the second-level indirect block.
                else {
                    Assert( physicalBlock > 0 && 
                            physicalBlock < sourceNoOfUnits );
                    extra2Buffer = indirect->getIndirect
                    ( _source, physicalBlock << _unitShift, _idNumber,
                      extra2Buffer );
                    // find the block number of the third-level indirect block.
                    physicalBlock = 
                    *(extra2Buffer->buf( calcDoubleIndex( logicalBlock ) ));
                    if (little_endian_fs) {
                        swapl((void *) &physicalBlock);
                    }

                    indirect->releaseIndirect( extra2Buffer, 0 );
                    Raids( RaidDetail ) 
                    << this << "::basicMapUnitRead() tripleIndex3 "
                    << physicalBlock << "\n" << eor;
                    checkCount = indexMask + 1 - ( logicalBlock & indexMask );
                    // There is no third-level indirect block.
                    if( physicalBlock == 0 ) {
                        Raids( RaidDetail )
                        << this << "::basicMapUnitRead() "
                        << "no triple 3 indirect"
                        << " startBlock " << startBlock
                        << " logicalBlock " << logicalBlock 
                        << " offset " << (logicalBlock & indexMask)
                        << "\n" << eor;
                        actual = basicMapReadAssist( 0, checkCount, 
                                                     remainingBlocks,
                                                     startRun,
                                                     expectedPhysicalBlock, 
                                                     status );
                    } 
                    // Read in the single indirect block.
                    else {
                        Assert( physicalBlock > 0 &&
                                physicalBlock < sourceNoOfUnits );
                        extra3Buffer = indirect->getIndirect
                        ( _source, physicalBlock << _unitShift, 
                          _idNumber, extra3Buffer );
                        startCheck = 
                        extra3Buffer->buf( calcSingleIndex( logicalBlock ));
                        Raids( RaidDetail )
                        << this << "::basicMapUnitRead() "
                        << "triple 3 indirect"
                        << " startBlock " << startBlock
                        << " logicalBlock " << logicalBlock 
                        << " offset " << ( logicalBlock & indexMask )
                        << "\n" << eor;
                        actual = basicMapReadAssist( startCheck, checkCount, 
                                                     remainingBlocks,
                                                     startRun,
                                                     expectedPhysicalBlock, 
                                                     status );
                        indirect->releaseIndirect( extra3Buffer, 0 );
                    }
                }
            }
        }
        startBlock += actual;
    }
    runLength = ( ( startBlock - orgStartBlock ) << log2BlockSize ) - 
    fragmentOffset;
    if ( runLength > count ) {
        runLength = count;
    }
    if ( startRun > 0 ) {
        startRun += fragmentOffset;
    }
    Assert( runLength > 0 );

    Assert( status != MAP_UNKNOWN );

    Raids( RaidMethod )
    << this << "::basicMapUnitRead() returns startRun " << startRun 
    << " runLength " << runLength << " status " << status
    << "\n" << eor;

    return status;
}

int
UNIXInode::basicMapWriteAssist( int * buf, int start,
                                int checkCount, 
                                int startBlock,
                                int & remainingCount,
                                int & startRun,
                                int & expectedPhysicalBlock,
                                MapUnitResult & status,
                                int & modified )
{
    Raids( RaidProtected )
    << this << "::basicMapWriteAssist()"
    << " start " << start
    << " checkCount " << checkCount
    << " startBlock " << startBlock
    << " remaining " << remainingCount
    << " startRun " << startRun
    << " expected " << expectedPhysicalBlock
    << " status " << status
    << "\n" << eor;
    modified = 0;
    Assert( checkCount > 0 && checkCount <= ( indexMask + 1 ) );
    Assert( remainingCount > 0 );
    int blockSize = 1 << log2BlockSize;
    if ( checkCount > remainingCount )
        checkCount = remainingCount;
    int myCount = 0;
    int pref;
    int count;
    int * check;
    Assert( status == MAP_UNKNOWN || status == MAP_REAL );
    int * startCheck = buf + start;
    if ( status == MAP_UNKNOWN ) {
        int lastPhysicalBlock = *startCheck;

        if (little_endian_fs) {
            swapl((void *) &lastPhysicalBlock);
        }

        if ( lastPhysicalBlock == 0 ) {
            // figure out how many blocks we need
            check = startCheck + 1;
            for ( count = 1; count < checkCount; count++ ) {
                if ( *check != 0 ) {
                    break;
                }
                check++;
            }

            pref = blockPref( buf, start, startBlock );
            count = allocBlocks( buf, start, count, pref );
            if ( count == 0 ) {
                status = MAP_FULL;
                goto done;
            }   
            lastPhysicalBlock = *startCheck;

            if (little_endian_fs) {
                swapl((void *) &lastPhysicalBlock);
            }

            modified = 1;
        }
        startRun = lastPhysicalBlock;
        myCount++;
        startCheck++;
        startBlock++;
        start++;
        expectedPhysicalBlock = lastPhysicalBlock + blockSize;
        status = MAP_REAL;
    }
    while ( myCount < checkCount ) {
        if ( *startCheck == 0 ) {
            // figure out how many blocks we need
            check = startCheck + 1;
            int limit = checkCount - myCount;
            for ( count = 1; count < limit; count++ ) {
                if ( *check != 0 ) {
                    break;
                }
                check++;
            }
            pref = blockPref( buf, start, startBlock );
            count = allocBlocks( buf, start, count, pref );
            modified = 1;
        }

        int startCheck_deref = *startCheck;
        if (little_endian_fs) {
            swapl((void *) &startCheck_deref);
        }

        if ( startCheck_deref == expectedPhysicalBlock ) {
            myCount++;
            startCheck++;
            startBlock++;
            expectedPhysicalBlock += blockSize;
            start++;
        } else {
            break;
        }
    }
 done:
    remainingCount -= myCount;
    if ( modified ) setModified();
    Raids( RaidProtected )
    << this << "::basicMapWriteAssist() return "
    << " remaining " << remainingCount
    << " startRun " << startRun
    << " expected " << expectedPhysicalBlock
    << " myCount " << myCount
    << " status " << status
    << " modified " << modified
    << "\n" << eor;
    return myCount;
}

MapUnitResult
UNIXInode::basicMapUnitWrite( int start, int count, 
                              int & startRun, int & runLength )
{
    Raids( RaidMethod )
    << this << "::basicMapUnitWrite( start " << start << ", count "
    << count << " )\n" << eor;

#ifdef ASSERT
    unsigned sourceNoOfUnits = _source->numberOfUnits() >> _unitShift;
#endif
    int startBlock = start >> log2BlockSize;
    int orgStartBlock = startBlock;
    int endBlock = ( start + count - 1 ) >> log2BlockSize;
    int remainingBlocks = endBlock - startBlock + 1;
    int fragmentOffset = start & blockMask();
    MapUnitResult status = MAP_UNKNOWN;
    int checkCount;
    int expectedPhysicalBlock;
    int actual = 1;
    int modified;

    while ( startBlock <= endBlock && actual > 0 ) {
        // direct block
        if( startBlock < (int)singleIndirect ) {
            checkCount = singleIndirect - startBlock;
            Raids( RaidDetail )
            << this << "::basicMapUnitWrite() direct\n" << eor;

            actual = basicMapWriteAssist( (int*)direct, startBlock, checkCount,
                                          startBlock,
                                          remainingBlocks, 
                                          startRun,
                                          expectedPhysicalBlock, status,
                                          modified );
            // if (modified) setModified();
        }
        // single indirect
        else if( startBlock < (int)doubleIndirect ) {
            checkCount = doubleIndirect - startBlock;
            // Adjust logical block for easy arithmetic
            unsigned logicalBlock = startBlock - singleIndirect;
            // There is no single indirect block.
            if( *singleIndex == 0 ) {
                Raids( RaidDetail )
                << this << "::basicMapUnitWrite() no single indirect\n"
                << eor;
                int pref = blockPref( 0, 0, startBlock );

                allocBlocks( (int*)singleIndex, 0, 1, pref );
                if ( *singleIndex == 0 ) {
                    Raids( RaidError )
                    << this << "::basicMapUnitWrite() "
                    << "alloc single indirect index err\n"
                    << eor;
                    if ( status == MAP_UNKNOWN )
                        status = MAP_FULL;
                    break;
                }
                setModified();
                unsigned int singleIndex_deref = *singleIndex;
                if (little_endian_fs) {
                    swapl((void *) &singleIndex_deref);
                }

                Raids( RaidDetail ) 
                << this << "::basicMapUnitWrite() singleIndex "
                << singleIndex_deref << "\n" << eor;

                singleBuffer = indirect->newIndirect
                ( _source, singleIndex_deref << _unitShift, _idNumber );
            } 
            // Read into the single indirect cache.
            else {
                unsigned int singleIndex_deref = *singleIndex;
                if (little_endian_fs) {
                    swapl((void *) &singleIndex_deref);
                }

                Raids( RaidDetail ) 
                << this << "::basicMapUnitWrite() singleIndex "
                << singleIndex_deref << "\n" << eor;
                Assert( singleIndex_deref > 0 && singleIndex_deref < sourceNoOfUnits );
                singleBuffer = indirect->getIndirect
                ( _source, singleIndex_deref << _unitShift, 
                  _idNumber, singleBuffer );
            }
            Raids( RaidDetail ) 
            << this << "::basicMapUnitWrite() single indirect\n" << eor;
            actual = basicMapWriteAssist( singleBuffer->buf( 0 ), logicalBlock,
                                          checkCount, 
                                          startBlock,
                                          remainingBlocks,
                                          startRun,
                                          expectedPhysicalBlock, 
                                          status, modified );
            indirect->releaseIndirect( singleBuffer, modified );
        }
        // double indirect
        else if( startBlock < (int)tripleIndirect ) {
            // Adjust logical block for easy arithmetic.
            unsigned logicalBlock = startBlock - doubleIndirect;
            // There is no double indirect block.
            if( *doubleIndex == 0 ) {
                Raids( RaidDetail )
                << this << "::basicMapUnitRead() no double indirect\n"
                << eor;
                // Allocate the double indirect block.
                int pref = blockPref( 0, 0, startBlock );
                allocBlocks( (int*)doubleIndex, 0, 1, pref );
                if( *doubleIndex == 0 ) {
                    Raids( RaidError )
                    << this
                    << "::basicMapUnitWrite() alloc double indirect index err\n"
                    << eor;
                    if ( status == MAP_UNKNOWN )
                        status = MAP_FULL;
                    break;
                }
                setModified();

                unsigned int doubleIndex_deref = *doubleIndex;
                if (little_endian_fs) {
                    swapl((void *) &doubleIndex_deref);
                }

                Raids( RaidDetail )
                << this << "::basicMapUnitWrite() doubleIndex "
                << doubleIndex_deref << "\n" << eor;
                doubleBuffer = indirect->newIndirect
                ( _source, doubleIndex_deref << _unitShift, _idNumber );
            }
            // Read into the double indirect cache.
            else {
                unsigned int doubleIndex_deref = *doubleIndex;
                if (little_endian_fs) {
                    swapl((void *) &doubleIndex_deref);
                }

                Raids( RaidDetail )
                << this << "::basicMapUnitWrite() doubleIndex "
                << doubleIndex_deref << "\n" << eor;
                Assert( doubleIndex_deref > 0 && doubleIndex_deref < sourceNoOfUnits );
                doubleBuffer = indirect->getIndirect
                ( _source, doubleIndex_deref << _unitShift, _idNumber,
                  doubleBuffer );
            }
            // Get the block number of the second-level indirect block.
            unsigned physicalBlock =
            *( doubleBuffer->buf( calcDoubleIndex( logicalBlock ) ) );

            if (little_endian_fs) {
                swapl((void *) &physicalBlock);
            }

            checkCount = indexMask + 1 - calcSingleIndex( logicalBlock );
            if ( physicalBlock == 0 ) {
                Raids( RaidDetail )
                << this << "::basicMapUnitRead() no double 2 indirect\n"
                << eor;
                int pref = blockPref( 0, 0, startBlock );

                allocBlocks( (int*)&physicalBlock, 0, 1, pref );
                if( physicalBlock == 0 ) {
                    Raids( RaidError )
                    << this << "::basicMapUnitWrite() "
                    << "alloc double indirect 2 index err\n"
                    << eor;
                    indirect->releaseIndirect( doubleBuffer, 0 );
                    if ( status == MAP_UNKNOWN )
                        status = MAP_FULL;
                    break;
                }
                setModified();
                *( doubleBuffer->buf( calcDoubleIndex( logicalBlock ) ) )
                = physicalBlock;
                /* swapl() is not needed here since allocBlocks
                   returns correct endiannes */
                /*
                  if (little_endian_fs) {
                  swapl((void *) doubleBuffer->buf(calcDoubleIndex(logicalBlock)));
                  }
                */
                unsigned tmp_physicalBlock = physicalBlock;
                if (little_endian_fs) {
                    swapl((void *) tmp_physicalBlock);
                }

                indirect->releaseIndirect( doubleBuffer, 1 );
                Raids( RaidDetail )
                << this << "::basicMapUnitWrite() doubleIndex2 "
                << tmp_physicalBlock << "\n" << eor;
                extra2Buffer = indirect->newIndirect
                ( _source, tmp_physicalBlock << _unitShift, _idNumber );
            } 
            // Read in the single indirect block.
            else {
                indirect->releaseIndirect( doubleBuffer, 0 );
                Raids( RaidDetail )
                << this << "::basicMapUnitWrite() doubleIndex2 "
                << physicalBlock << "\n" << eor;
                Assert( physicalBlock > 0 && 
                        physicalBlock < sourceNoOfUnits );
                extra2Buffer = indirect->getIndirect
                ( _source, physicalBlock << _unitShift , _idNumber,
                  extra2Buffer );
            }
            Raids( RaidDetail ) 
            << this << "::basicMapUnitWrite() double 2 indirect\n" << eor;
            actual = basicMapWriteAssist( extra2Buffer->buf( 0 ),
                                          calcSingleIndex( logicalBlock ),
                                          checkCount,
                                          startBlock,
                                          remainingBlocks,
                                          startRun,
                                          expectedPhysicalBlock,
                                          status, modified );
            indirect->releaseIndirect( extra2Buffer, modified );
        }
        // This logical block is a triple indirect block.
        else {
            // Adjust logical block for easy arithmetic
            unsigned logicalBlock = startBlock - tripleIndirect;
            // There is no triple indirect block.   
            if( *tripleIndex == 0 ) {
                Raids( RaidDetail )
                << this << "::basicMapUnitRead() no triple indirect\n"
                << eor;
                // Allocate the double indirect block.
                int pref = blockPref( 0, 0, startBlock );

                allocBlocks( (int*)tripleIndex, 0, 1, pref );
                if( *tripleIndex == 0 ) {
                    Raids( RaidError )
                    << this
                    << "::basicMapUnitWrite() alloc triple indirect index err\n"
                    << eor;
                    if ( status == MAP_UNKNOWN )
                        status = MAP_FULL;
                    break;
                }
                setModified();

                unsigned int tripleIndex_deref = *tripleIndex;
                if (little_endian_fs) {
                    swapl((void *) &tripleIndex_deref);
                }

                Raids( RaidDetail )
                << this << "::basicMapUnitWrite() tripleIndex "
                << tripleIndex_deref << "\n" << eor;
                tripleBuffer = indirect->newIndirect
                ( _source, tripleIndex_deref << _unitShift, _idNumber );
            }
            // Read into the triple indirect cache.
            else {
                unsigned int tripleIndex_deref = *tripleIndex;
                if (little_endian_fs) {
                    swapl((void *) &tripleIndex_deref);
                }

                Assert( tripleIndex_deref > 0 && tripleIndex_deref < sourceNoOfUnits );
                Raids( RaidDetail )
                << this << "::basicMapUnitWrite() tripleIndex "
                << tripleIndex_deref << "\n" << eor;
                tripleBuffer = indirect->getIndirect
                ( _source, tripleIndex_deref << _unitShift, _idNumber,
                  tripleBuffer );
                // Get the block number of the second-level indirect block.
            }
            unsigned physicalBlock = 
            *( tripleBuffer->buf( calcTripleIndex( logicalBlock ) ) );

            if (little_endian_fs) {
                swapl((void *) &physicalBlock);
            }

            // There is no second-level indirect block.
            if( physicalBlock == 0 ) {
                Raids( RaidDetail )
                << this << "::basicMapUnitRead() no triple 2 indirect\n"
                << eor;
                int pref = blockPref( 0, 0, startBlock );
                allocBlocks( (int*)&physicalBlock, 0, 1, pref );
                if( physicalBlock == 0 ) {
                    Raids( RaidError )
                    << this << "::basicMapUnitWrite() "
                    << "alloc triple indirect 2 index err\n"
                    << eor;
                    indirect->releaseIndirect( tripleBuffer, 0 );
                    if ( status == MAP_UNKNOWN )
                        status = MAP_FULL;
                    break;
                }
                setModified();
                *( tripleBuffer->buf( calcTripleIndex( logicalBlock ) ) )
                = physicalBlock;

                /* swapl() is not needed here since allocBlocks
                   returns correct endiannes */
                /*
                  if (little_endian_fs) {
                  swapl((void *) tripleBuffer->buf(calcTripleIndex(logicalBlock)));
                  }
                */
                unsigned tmp_physicalBlock = physicalBlock;
                if (little_endian_fs) {
                    swapl((void *) tmp_physicalBlock);
                }


                indirect->releaseIndirect( tripleBuffer, 1 );
                Raids( RaidDetail )
                << this << "::basicMapUnitWrite() tripleIndex2 "
                << tmp_physicalBlock << "\n" << eor;
                extra2Buffer = indirect->newIndirect
                ( _source, tmp_physicalBlock << _unitShift, _idNumber );
            }
            // Read in the second-level indirect block.
            else {
                indirect->releaseIndirect( tripleBuffer, 0 );
                Assert( physicalBlock > 0 && 
                        physicalBlock < sourceNoOfUnits );
                Raids( RaidDetail )
                << this << "::basicMapUnitWrite() tripleIndex2 "
                << physicalBlock << "\n" << eor;
                extra2Buffer = indirect->getIndirect
                ( _source, physicalBlock << _unitShift, _idNumber,
                  extra2Buffer );
            }
            // find the block number of the third-level indirect block.
            physicalBlock = 
            *( extra2Buffer->buf( calcDoubleIndex( logicalBlock ) ) );

            if (little_endian_fs) {
                swapl((void *) &physicalBlock);
            }

            // There is no third-level indirect block.
            if( physicalBlock == 0 ) {
                Raids( RaidDetail )
                << this << "::basicMapUnitRead() no triple 3 indirect\n"
                << eor;
                int pref = blockPref( 0, 0, startBlock );
                allocBlocks( (int*)&physicalBlock, 0, 1, pref );
                if( physicalBlock == 0 ) {
                    Raids( RaidError )
                    << this << "::basicMapUnitWrite() "
                    << "alloc triple indirect 3 index err\n"
                    << eor;
                    indirect->releaseIndirect( extra2Buffer, 0 );
                    if ( status == MAP_UNKNOWN )
                        status = MAP_FULL;
                    break;
                }
                setModified();
                *( extra2Buffer->buf( calcDoubleIndex( logicalBlock ) ) ) =
                physicalBlock;

                /* swapl() is not needed here since allocBlocks
                   returns correct endiannes */
                /*
                  if (little_endian_fs) {
                  swapl((void *) tripleBuffer->buf(calcTripleIndex(logicalBlock)));
                  }
                */
                unsigned tmp_physicalBlock = physicalBlock;
                if (little_endian_fs) {
                    swapl((void *) tmp_physicalBlock);
                }

                indirect->releaseIndirect( extra2Buffer, 1 );
                Raids( RaidDetail )
                << this << "::basicMapUnitWrite() tripleIndex3 "
                << tmp_physicalBlock << "\n" << eor;
                extra3Buffer = indirect->newIndirect
                ( _source, tmp_physicalBlock << _unitShift, _idNumber );
            }
            // Read in the single indirect block.
            else {
                indirect->releaseIndirect( extra2Buffer, 0 );
                Raids( RaidDetail )
                << this << "::basicMapUnitWrite() tripleIndex3 "
                << physicalBlock << "\n" << eor;
                Assert( physicalBlock > 0 &&
                        physicalBlock < sourceNoOfUnits );
                extra3Buffer = indirect->getIndirect
                ( _source, physicalBlock << _unitShift, _idNumber,
                  extra3Buffer );
            }
            Raids( RaidDetail ) 
            << this << "::basicMapUnitWrite() triple 3 indirect\n" << eor;
            checkCount = indexMask + 1 - calcSingleIndex( logicalBlock );
            actual = basicMapWriteAssist( extra3Buffer->buf( 0 ),
                                          calcSingleIndex( logicalBlock ),
                                          checkCount, 
                                          startBlock,
                                          remainingBlocks,
                                          startRun,
                                          expectedPhysicalBlock, 
                                          status, modified );
            indirect->releaseIndirect( extra3Buffer, modified );
        }
        startBlock += actual;
    }
    Assert( status != MAP_UNKNOWN );
    runLength = ( ( startBlock - orgStartBlock ) << log2BlockSize );
    Assert( runLength > 0 || status == MAP_FULL ); 
    if ( runLength > fragmentOffset ) {
        runLength -= fragmentOffset;
    }
    if ( runLength > count ) {
        runLength = count;
    }
    if ( startRun > 0 ) {
        startRun += fragmentOffset;
    }

    Raids( RaidMethod )
    << this << "::basicMapUnitWrite() returns startRun " << startRun 
    << " runLength " << runLength << " status " << status
    << "\n" << eor;

    return status;
}

inline 
int
UNIXInode::basicMapFreeIndirect( int * startBuf, int * startCheck, int count )
{
    int indexMax = 1 << indexShift;
    if ( count == indexMax ) 
        return 1;
    int * endCheck = startBuf + indexMax;
    while( startBuf < startCheck ) {
        if ( *startBuf != 0 )
            return 0;
        startBuf++;
    }
    startBuf = startCheck + count;
    while( startBuf < endCheck ) {
        if ( *startBuf != 0 )
            return 0;
        startBuf++;
    }
    return 1;
}

int
UNIXInode::basicMapFreeAssist( int * startCheck, int checkCount, 
                               int startBlock, int lastBlock,
                               int & remainingCount,
                               MapUnitResult & status, int & modified )
{
    modified = 0;
    Assert( checkCount > 0 && checkCount <= ( indexMask + 1 ) );
    Assert( remainingCount > 0 );
    int blockSize = 1 << log2BlockSize;
    if ( checkCount > remainingCount )
        checkCount = remainingCount;
    int myCount = 0;
    while( myCount < checkCount ) {
        unsigned lastPhysicalBlock;
        while ( ( lastPhysicalBlock = *startCheck ) == 0 ) {
            if (little_endian_fs) {
                swapl((void *) &lastPhysicalBlock);
            }
            startCheck++;
            startBlock++;
            myCount++;
            if ( myCount >= checkCount )
                goto done;
        }
        unsigned startRun = lastPhysicalBlock;
        unsigned runLength = blockSize;
        modified = 1;
        *startCheck = 0;
        startCheck++;
        startBlock++;
        myCount++;
        unsigned expectedPhysicalBlock = lastPhysicalBlock + blockSize;

        int startCheck_deref = *startCheck;
        if (little_endian_fs) {
            swapl((void *) &startCheck_deref);
        }

        while ( myCount < checkCount && startCheck_deref == (int)expectedPhysicalBlock ) {
            *startCheck = 0;
            startCheck++;
            startBlock++;
            myCount++;
            runLength += blockSize;
            expectedPhysicalBlock += blockSize;

            startCheck_deref = *startCheck;
            if (little_endian_fs) {
                swapl((void *) &startCheck_deref);
            }
        }
        if ( fragmentsInLastBlock != 0 && ( startBlock - 1 ) == lastBlock ) {
            runLength -= ( blockSize - fragmentsInLastBlock );
        }
        freeFragments( startRun, runLength );
    }
 done:
    status = MAP_HOLE;
    remainingCount -= myCount;
    if ( modified ) setModified();
    return myCount;
}

MapUnitResult
UNIXInode::basicMapUnitFree( int start, int count )
{
    Raid( RaidProtected )( "%N::basicMapUnitFree( %d, %d )\n",
                           this, start, count );
#ifdef ASSERT
    unsigned sourceNoOfUnits = _source->numberOfUnits() >> _unitShift;
#endif
    int startBlock = start >> log2BlockSize;
    int endBlock = ( start + count - 1 ) >> log2BlockSize;
    int remainingBlocks = endBlock - startBlock + 1;
    MapUnitResult status = MAP_UNKNOWN;
    int actual = 1;
    int lastBlock = lastBlockNo();
    int modified;

    if( startBlock < (int)singleIndirect ) {
        int checkCount = singleIndirect - startBlock;

        unsigned int * startCheck = direct + startBlock;
        actual = basicMapFreeAssist( (int*)startCheck, checkCount, 
                                     startBlock, lastBlock,
                                     remainingBlocks, 
                                     status, modified );
        // if ( modified ) setModified();
        startBlock += actual;
        Assert( remainingBlocks == 0 || startBlock == (int)singleIndirect );
        if ( remainingBlocks == 0 )
            goto done;
    }
    // single indirect
    if ( startBlock < (int)doubleIndirect ) {
        int checkCount = doubleIndirect - startBlock;
        // Adjust logical block for easy arithmetic
        unsigned logicalBlock = startBlock - singleIndirect;
        // There is no single indirect block.
        if( *singleIndex == 0 ) {
            actual = ( checkCount > remainingBlocks ? 
                       remainingBlocks : checkCount );
            remainingBlocks -= actual;
            startBlock += actual;
        } 
        // Read into the single indirect cache.
        else {
            unsigned int singleIndex_deref = *singleIndex;
            if (little_endian_fs) {
                swapl((void *) &singleIndex_deref);
            }

            Assert( singleIndex_deref > 0 && singleIndex_deref < sourceNoOfUnits );
            singleBuffer = indirect->getIndirect
            ( _source, singleIndex_deref << _unitShift, _idNumber, 
              singleBuffer );
            int * startCheck = singleBuffer->buf( logicalBlock );
            actual = basicMapFreeAssist( startCheck, checkCount, 
                                         startBlock, lastBlock,
                                         remainingBlocks,
                                         status, modified );
            startBlock += actual;
            if ( modified && 
                 basicMapFreeIndirect( singleBuffer->buf( 0 ), 
                                       startCheck, actual ) ) {
                indirect->freeIndirect( singleBuffer );
                freeBlock( singleIndex_deref );
                singleBuffer = 0;
                *singleIndex = 0;   
                // setModified();
            } else {
                indirect->releaseIndirect( singleBuffer, modified );
            }
        }
        Assert( remainingBlocks == 0 || startBlock == (int)doubleIndirect );
        if ( remainingBlocks == 0 )
            goto done;
    }
    // double indirect
    if( startBlock < (int)tripleIndirect ) {
        // There is no double indirect block.
        if( *doubleIndex == 0 ) {
            int checkCount = tripleIndirect - startBlock;
            actual = ( checkCount > remainingBlocks ? 
                       remainingBlocks : checkCount );
            remainingBlocks -= actual;
            startBlock += actual;
        }
        // Read into the double indirect cache.
        else {
            unsigned int doubleIndex_deref = *doubleIndex;
            if (little_endian_fs) {
                swapl((void *) &doubleIndex_deref);
            }

            Assert( doubleIndex_deref > 0 && doubleIndex_deref < sourceNoOfUnits );
            doubleBuffer = indirect->getIndirect
            ( _source, doubleIndex_deref << _unitShift, _idNumber,
              doubleBuffer );
            unsigned doubleMod = 0;
            int * initial = doubleBuffer->buf
            ( calcDoubleIndex( startBlock - doubleIndirect ) );
            int examined = 0;
            while( remainingBlocks > 0 && startBlock < (int)tripleIndirect ) {
                // Adjust logical block for easy arithmetic.
                unsigned logicalBlock = startBlock - doubleIndirect;
                // Get the block number of the second-level indirect block.
                unsigned physicalBlock =
                *( doubleBuffer->buf( calcDoubleIndex( logicalBlock ) ) );
                if (little_endian_fs) {
                    swapl((void *) &physicalBlock);
                }
                int checkCount = 
                indexMask + 1 - calcSingleIndex( logicalBlock );
                examined++;
                if ( physicalBlock == 0 ) {
                    actual = ( checkCount > remainingBlocks ? 
                               remainingBlocks : checkCount );
                    remainingBlocks -= actual;
                    startBlock += actual;
                    Assert( remainingBlocks == 0 ||
                            ( ( startBlock - doubleIndirect ) & indexMask )
                            == 0 ); 
                } 
                // Read in the single indirect block.
                else {
                    Assert( physicalBlock > 0 && 
                            physicalBlock < sourceNoOfUnits );
                    extra2Buffer = indirect->getIndirect
                    ( _source, physicalBlock << _unitShift , _idNumber,
                      extra2Buffer );
                    int * startCheck = 
                    extra2Buffer->buf( calcSingleIndex( logicalBlock ) );
                    actual = basicMapFreeAssist( startCheck, checkCount,
                                                 startBlock, lastBlock,
                                                 remainingBlocks,
                                                 status, modified );
                    startBlock += actual;
                    Assert( remainingBlocks == 0 ||
                            ( ( startBlock - doubleIndirect ) & indexMask )
                            == 0 ); 
                    if ( modified && 
                         basicMapFreeIndirect( extra2Buffer->buf( 0 ), 
                                               startCheck, actual ) ) {
                        indirect->freeIndirect( extra2Buffer );
                        freeBlock( physicalBlock );
                        *(doubleBuffer->buf(calcDoubleIndex( logicalBlock )))
                        = 0;
                        extra2Buffer = 0;
                        doubleMod = 1;
                    } else {
                        indirect->releaseIndirect( extra2Buffer, modified );
                    }
                }
            }
            // end of while loop for 2nd level indirect
            Assert( remainingBlocks == 0 || startBlock == (int)tripleIndirect );
            Assert( examined > 0 );
            if ( doubleMod && 
                 basicMapFreeIndirect( doubleBuffer->buf( 0 ), 
                                       initial, examined ) ) {
                indirect->freeIndirect( doubleBuffer );
                /* HERE */
                freeBlock( doubleIndex_deref );
                *doubleIndex = 0;
                doubleBuffer = 0;
                // setModified();
            } else {
                indirect->releaseIndirect( doubleBuffer, doubleMod );
            }
        }
        Assert( remainingBlocks == 0 || startBlock == (int)tripleIndirect );
        if ( remainingBlocks == 0 )
            goto done;
    }
    // triple indirect
    {
        // There is no triple indirect block.   
        if( *tripleIndex == 0 ) {
            // Adjust logical block for easy arithmetic
            unsigned logicalBlock = startBlock - tripleIndirect;
            int checkCount = ( 1 << ( 3 * indexShift ) ) - logicalBlock;
            actual = ( checkCount > remainingBlocks ? 
                       remainingBlocks : checkCount );
            remainingBlocks -= actual;
            startBlock += actual;
        }
        // Read into the triple indirect cache.
        else {
            unsigned int tripleIndex_deref = *tripleIndex;
            if (little_endian_fs) {
                swapl((void *) &tripleIndex_deref);
            }
            Assert( tripleIndex_deref > 0 && tripleIndex_deref < sourceNoOfUnits );
            tripleBuffer = indirect->getIndirect
            ( _source, tripleIndex_deref << _unitShift, _idNumber,
              tripleBuffer );
            unsigned tripleMod = 0;
            int * tripleInitial = 
            tripleBuffer->buf( calcTripleIndex( startBlock - tripleIndirect ) );
            int tripleExamined = 0;
            while( remainingBlocks > 0 ) {
                // Adjust logical block for easy arithmetic
                unsigned logicalBlock = startBlock - tripleIndirect;
                // Get the block number of the second-level indirect block.
                unsigned physicalBlock = 
                *( tripleBuffer->buf( calcTripleIndex( logicalBlock ) ) );
                if (little_endian_fs) {
                    swapl((void *) &physicalBlock);
                }
                // There is no second-level indirect block.
                unsigned maxcount = 1 << ( indexShift + indexShift );
                int checkCount = maxcount - ( logicalBlock & ( maxcount - 1 ) );
                tripleExamined++;
                if( physicalBlock == 0 ) {
                    actual = ( checkCount > remainingBlocks ? 
                               remainingBlocks : checkCount );
                    remainingBlocks -= actual;
                    startBlock += actual;
                    Assert( remainingBlocks == 0 ||
                            ( ( startBlock - tripleIndirect ) & 
                              ( maxcount - 1 ) ) == 0 );
                } 
                // Read in the second-level indirect block.
                else {
                    Assert( physicalBlock > 0 && 
                            physicalBlock < sourceNoOfUnits );
                    extra2Buffer = indirect->getIndirect
                    ( _source, physicalBlock << _unitShift, _idNumber,
                      extra2Buffer );
                    unsigned extra2Index = physicalBlock;
                    unsigned extra2Mod = 0;
                    unsigned endLoop = startBlock + checkCount;
                    Assert( ( ( endLoop - tripleIndirect ) & ( maxcount - 1 ) )
                            == 0 );
                    int * extra2Initial = 
                    extra2Buffer->buf(calcDoubleIndex(startBlock-tripleIndirect));
                    int extra2Examined = 0;
                    while( remainingBlocks > 0 && startBlock < (int)endLoop ) {
                        // Adjust logical block for easy arithmetic
                        unsigned logicalBlock2 = startBlock - tripleIndirect;
                        // find the block number of the third-level 
                        // indirect block.
                        physicalBlock = 
                        *(extra2Buffer->buf(calcDoubleIndex(logicalBlock2)));
                        if (little_endian_fs) {
                            swapl((void *) &physicalBlock);
                        }
                        // There is no third-level indirect block.
                        checkCount = 
                        indexMask + 1 - ( logicalBlock2 & indexMask );
                        extra2Examined++;
                        if( physicalBlock == 0 ) {
                            actual = ( checkCount > remainingBlocks ? 
                                       remainingBlocks : checkCount );
                            remainingBlocks -= actual;
                            startBlock += actual;
                            Assert( remainingBlocks == 0 ||
                                    ( ( startBlock - tripleIndirect ) & 
                                      indexMask ) == 0 );
                        } 
                        // Read in the single indirect block.
                        else {
                            Assert( physicalBlock > 0 &&
                                    physicalBlock < sourceNoOfUnits );
                            extra3Buffer = indirect->getIndirect
                            ( _source, physicalBlock << _unitShift, 
                              _idNumber, extra3Buffer );
                            int * startCheck = 
                            extra3Buffer->buf(calcSingleIndex(logicalBlock2));
                            actual = basicMapFreeAssist( startCheck, checkCount,
                                                         startBlock, lastBlock,
                                                         remainingBlocks,
                                                         status, modified );
                            startBlock += actual;
                            Assert( remainingBlocks == 0 ||
                                    ( ( startBlock - tripleIndirect ) & 
                                      indexMask ) == 0 );
                            if ( modified && 
                                 basicMapFreeIndirect( extra3Buffer->buf( 0 ),
                                                       startCheck, actual ) ) {
                                indirect->freeIndirect( extra3Buffer );
                                freeBlock( physicalBlock );
                                *(extra2Buffer->buf(calcDoubleIndex(logicalBlock2)))
                                = 0;
                                extra3Buffer = 0;
                                extra2Mod = 1;
                            } else {
                                indirect->releaseIndirect( extra3Buffer, 
                                                           modified );
                            }
                        }
                    }
                    // end of 3rd level indirect loop
                    if ( extra2Mod && 
                         basicMapFreeIndirect( extra2Buffer->buf( 0 ),
                                               extra2Initial, 
                                               extra2Examined ) ) {
                        indirect->freeIndirect( extra2Buffer );
                        freeBlock( extra2Index );
                        *(tripleBuffer->buf(calcTripleIndex(logicalBlock)))
                        = 0;
                        extra2Buffer = 0;
                        tripleMod = 1;
                    } else {
                        indirect->releaseIndirect( extra2Buffer, extra2Mod );
                    }
                }
                Assert( remainingBlocks == 0 ||
                        ( ( startBlock - tripleIndirect ) & 
                          ( maxcount - 1 ) ) == 0 );
            }
            // end of 2nd level indirect loop
            if ( tripleMod && 
                 basicMapFreeIndirect( tripleBuffer->buf( 0 ), 
                                       tripleInitial, tripleExamined ) ) {
                indirect->freeIndirect( tripleBuffer );
                unsigned int tripleIndex_deref = *tripleIndex;
                if (little_endian_fs) {
                    swapl((void *) &tripleIndex_deref);
                }
                freeBlock( tripleIndex_deref );
                tripleBuffer = 0;
                *tripleIndex = 0;
                // setModified();
            } else {
                indirect->releaseIndirect( tripleBuffer, tripleMod );
            }
        }
    }
 done:
    Assert( remainingBlocks == 0 );
    status = MAP_HOLE;

    Raid( RaidProtected )( "%N::mapUnitFree returns %d\n", this, (int) status );

    return status;
}
#include "ClassInstaller.h"
DEFINECLASS("UNIXInode",UNIXInode,FileObject,0,0);
