#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * UNIXInodeIndirect.cc - 
 */

#include "Assert.h"
#include "Async.h"
#include "Byte.h"
#include "MemoryObjectAsync.h"
#include "PersistentMemoryObject.h"
#include "Raid.h"
#include "RangeLocks.h"
#include "SoftClock.h"
#include "TimeoutTimer.h"
#include "UNIXInodeIndirect.h"

#define SAFE

NormalRangeLock _ioLock( 0 );

inline
unsigned 
acquireGetLock( PersistentMemoryObject * source, unsigned id )
{
    unsigned handle = ( ((unsigned) source) << 16 ) + id;
    _ioLock.acquire( handle, handle );
    return handle;
}

inline 
void
releaseGetLock( unsigned handle )
{
    _ioLock.release( handle, handle );
}

NormalRangeLock _blockLock( 0 );

inline
void
acquireBlockLock( PersistentMemoryObject * source, unsigned id )
{
    unsigned handle = ( ((unsigned) source) << 16 ) + id;
    _blockLock.acquire( handle, handle );
}

inline 
void
releaseBlockLock( PersistentMemoryObject * source, unsigned id )
{
    unsigned handle = ( ((unsigned) source) << 16 ) + id;
    _blockLock.release( handle, handle );
}


unsigned UNIXInodeIndirect::_age = 0;
static unsigned primeTable[] = 
{ 61, 127, 251, 383, 509, 761, 1021, 0 };
    
int UNIXInodeIndirectCacheSize = 256;

UNIXInodeIndirect * UNIXInodeIndirect::_indirect[32] = 
{ 0, 0, 0, 0, 0, 0, 0, 0, 
  0, 0, 0, 0, 0, 0, 0, 0, 
  0, 0, 0, 0, 0, 0, 0, 0, 
  0, 0, 0, 0, 0, 0, 0, 0 };

Lock UNIXInodeIndirect::_lock;

UNIXInodeIndirect *
UNIXInodeIndirect::setup( unsigned log2BytesPerBlock ) 
{
    UNIXInodeIndirect **uii;
    static Lock lock;
    lock.acquire();
    uii = _indirect + log2BytesPerBlock; 
    if ( *uii == 0 ) {
        *uii = new UNIXInodeIndirect( UNIXInodeIndirectCacheSize, 
                                      log2BytesPerBlock );
        Assert( *uii != 0 );
    }   
    lock.release();
    return *uii;
}

void
UNIXInodeIndirect_info()
{
    UNIXInodeIndirect::infoAll(Console());
}

void
UNIXInodeIndirect::infoAll( OutputStream & s )
{
    for (int i = 0; i < 32; i++) {
        if (_indirect[i] != 0) 
            _indirect[i]->info(s);
    }
    return;
}

UNIXInodeIndirect::UNIXInodeIndirect( unsigned size, unsigned log2BlockSize )
    : _size( size ), _log2BlockSize( log2BlockSize )
{
    Debug   
    << "UNIXInodeIndirect::UNIXInodeIndirect( size " << size 
    << ", log2BlockSize " << log2BlockSize << " )\n" << eor;

    _fastHit = 0;
    _slowHit = 0;
    _miss = 0;

    _blockSize = 1 << _log2BlockSize;
    
    _bufs = new char[ _size << _log2BlockSize ];
    _infos = new UNIXInodeIndirectInfo[ _size ];
    Assert( _bufs != 0 && _infos != 0 );

    // initialize free list

    char * p = _bufs;
    UNIXInodeIndirectInfo * info = _infos;
    unsigned i;
    for( i = 0; i < _size; i++ ) {
        info->_buf = (int*)p;
        p += _blockSize;
        info->_prev = info - 1;
        info->_next = info + 1;
        info->_neighbor = info + 1;
        info->_list = &_freeList;
        info++;
    }
    _infos[0]._prev = 0;
    _infos[_size - 1]._next = 0;
    _infos[_size - 1]._neighbor = 0;
    _freeList = _infos;
    _free = _size;
    _locked = 0;
    _unlocked = 0;
    _busy = 0;

    // construct hash table

    for( i = 0; primeTable[i] != 0; i++ ) {
        if ( primeTable[i] >= _size )
            break;
    }
    if ( primeTable[i] == 0 )
        i--;
    _hashSize = primeTable[i];
    _hashTable = new UNIXInodeIndirectInfo * [_hashSize];
    Assert( _hashTable != 0 );
    for( i = 0; i < _hashSize; i++ ) {
        _hashTable[i] = 0;
    }
}

void
UNIXInodeIndirect::info( OutputStream & s )
{
    s 
    << "UNIXInodeIndirect[" << _log2BlockSize << "]::info()\n"
    << "size " << _size << " free " << _free << " unlocked " << _unlocked   
    << " busy " << _busy << " locked " << _locked << "\n" << eor;

    unsigned total = _slowHit + _fastHit + _miss;
    s
    << "fast hit " << _fastHit << " (" << ( _fastHit * 100 ) / total 
    << "%) slow hit " << _slowHit << " (" << ( _slowHit * 100 ) / total
    << "%) miss " << _miss << " (" << ( _miss * 100 ) / total 
    << "%) total " << total << "\n" << eor;

    UNIXInodeIndirectInfo * info = _infos;
    int locks = 0;
    s.formattedWrite
    ("      info     source idnumber     unit      age d p lock\n");
    s.flush();
    while( info ) {
        s.formattedWrite("%10x %10x %8d %8d ",
                         (unsigned) info, (unsigned) info->_source,
                         info->_idNumber, info->_unit);
        s.formattedWrite("%8u %1d %1d %4d\n",
                         info->_age, info->_dirty,
                         info->_pending, info->_lock);
        s.flush();
        if (info->_lock > 0)
            locks += info->_lock;
        info = info->_neighbor;
    }
    extern int _activeFiles;
    s << "active files " << _activeFiles << " total locks " << locks 
      << "\n" << eor;
}

// inline 
void
UNIXInodeIndirect::writeDone( UNIXInodeIndirectInfo * info, unsigned result )
{
    acquire();
    info->_dirty = 0;
    info->_lock--;
    if (info->_lock == 0) {
        _locked--;
        _unlocked++;
    }
    Assert( info->_lock >= 0 );
    release();
    releaseBlockLock( info->_source, info->_unit );

    Debug   
    << "UNIXInodeIndirect[" << _log2BlockSize << "]::writeDone( "
    << asAddress( info ) << " ) source "
    << info->_source << " unit " << info->_unit 
    << ")\n" << eor;

#ifdef ASSERT
    int shift = _log2BlockSize - info->_source->log2UnitSize();
    Assert( shift >= 0 );
    unsigned sourceUnits = 1 << shift;
    Assert( result == sourceUnits );
#endif
}

unsigned
UNIXInodeIndirect::write( UNIXInodeIndirectInfo * info ) 
{
    PersistentMemoryObject * source = info->_source;
    unsigned unit = info->_unit;
    int * buf = info->_buf;

    Debug   
    << "UNIXInodeIndirect[" << _log2BlockSize << "]::write( "
    << asAddress( info ) << " ) source " << source 
    << " unit " << unit << ")\n" << eor;

    acquireBlockLock( source, unit );
    int shift = _log2BlockSize - source->log2UnitSize();
    Assert( shift >= 0 );
    int sourceUnits = 1 << shift;
    int result = source->writeThrough
    ( unit, sourceUnits, (char *) buf, _blockSize );
    releaseBlockLock( source, unit );

    Debug   
    << "UNIXInodeIndirect[" << _log2BlockSize << "]::write( "
    << asAddress( info ) << " ) returns\n" << eor;

    return ( result == sourceUnits );
}

void
UNIXInodeIndirect::write( UNIXInodeIndirectInfo * info, IOAsyncClosure * c ) 
{
    PersistentMemoryObject * source = info->_source;
    unsigned unit = info->_unit;
    int * buf = info->_buf;

    Debug   
    << "UNIXInodeIndirect[" << _log2BlockSize << "]::write( "
    << asAddress( info ) << ", " << asAddress( c ) 
    << " ) source " << source 
    << " unit " << unit << ")\n" << eor;

    acquireBlockLock( source, unit );
    int shift = _log2BlockSize - source->log2UnitSize();
    Assert( shift >= 0 );
    int sourceUnits = 1 << shift;
    source->asyncWriteThrough
    ( unit, sourceUnits, (char *) buf, _blockSize, c );
}

unsigned
UNIXInodeIndirect::read( PersistentMemoryObject * source, unsigned unit,
                         unsigned * buf )
{
    Debug   
    << "UNIXInodeIndirect[" << _log2BlockSize << "]::read( "
    << source << ", " << unit << ", " << asAddress( buf )
    << ")\n" << eor;

    acquireBlockLock( source, unit );
    int shift = _log2BlockSize - source->log2UnitSize();
    Assert( shift >= 0 );
    int sourceUnits = 1 << shift;
    int result = source->read( unit, sourceUnits, (char *) buf, 
                               _blockSize );
    releaseBlockLock( source, unit );

    Debug   
    << "UNIXInodeIndirect[" << _log2BlockSize << "]::read( "
    << source << ", " << unit << ", " << asAddress( buf )
    << ") done\n" << eor;

    return ( result == sourceUnits );
}

void
UNIXInodeIndirect::allocateMore()
{
    Assert( _freeList == 0 );
    for (int i = 0; i < 4; i++) {
        _size++;
        _free++;
        char * buf = new char[ _blockSize ];
        UNIXInodeIndirectInfo * info = new UNIXInodeIndirectInfo;
        Assert( buf != 0 && info != 0 );

        info->_buf = (int *) buf;
        info->_neighbor = _infos;
        _infos = info;
        addToList( &_freeList, info );
    }

    Debug 
    << "UNIXInodeIndirect[" << _log2BlockSize 
    << "]::allocateAnother() new size " << _size << "\n" << eor;
    if (( (int)_size > UNIXInodeIndirectCacheSize ) && ( _size % 32 == 0 )) {
        info( Console() );
    }

    return;
}

UNIXInodeIndirectInfo *
UNIXInodeIndirect::victimize()
{
    Debug   
    << "UNIXInodeIndirect[" << _log2BlockSize 
    << "]::victimize()\n" << eor;

    UNIXInodeIndirectInfo * found = _freeList;
    if ( found ) {
        Debug   
        << "UNIXInodeIndirect[" << _log2BlockSize 
        << "]::victimize() new " << asAddress( found )
        << " from free list\n" << eor;
        removeFromList( &_freeList, found );
        found->_lock = -1;
        _free--;
        _busy++;
        return found;
    }
    UNIXInodeIndirectInfo * info = _infos;
    unsigned foundAge = 0xffffffff;
    while( info ) {
        if ( info->_age < foundAge && info->_lock == 0 ) {
            foundAge = info->_age;
            found = info;
        }
        info = info->_neighbor;
    }
    if ( found == 0 ) {
        allocateMore();
        Debug   
        << "UNIXInodeIndirect[" << _log2BlockSize 
        << "]::victimize() new " << asAddress( found )
        << " from free list\n" << eor;
        found = _freeList;
        Assert( found != 0 );
        removeFromList( &_freeList, found );
        found->_lock = -1;
        _free--;
        _busy++;
        return found;
    }
    PersistentMemoryObject * source = found->_source;
    Assert( source != 0 );
    Debug   
    << "UNIXInodeIndirect[" << _log2BlockSize 
    << "]::victimize() found " << asAddress( found )
    << " source " << source << ", unit " << found->_unit 
    << "\n" << eor;
    unsigned foundSlot = hash( source, found->_unit );
    found->_lock = -1;
    _unlocked--;
    _busy++;
    removeFromList( _hashTable + foundSlot, found );
    if ( found->_dirty ) {
        release();
        unsigned ok = write( found );
        Assert( ok );
        acquire();
    }
        
    return found;
}

#ifdef SAFE

class UNIXInodeNewIndirectAsyncClosure : public IOAsyncClosure
{
    unsigned _expected;
    UNIXInodeIndirect * _indirect;
    UNIXInodeIndirectInfo * _info;
    static void func( UNIXInodeNewIndirectAsyncClosure * c );
public:
    UNIXInodeNewIndirectAsyncClosure( UNIXInodeIndirect * indirect,
                                      UNIXInodeIndirectInfo * info, 
                                      unsigned expected )
        : 
        _expected(expected),
        _indirect( indirect ),
        _info( info )
    {_func = (AsyncFunc) &func; }
};

void
UNIXInodeNewIndirectAsyncClosure::func( UNIXInodeNewIndirectAsyncClosure * c )
{
    if ( c->_result != c->_expected ) {
        Console() << "UNIXInodeIndirectAsyncClosure::func() "
                  << " expected " << c->_expected << " != " 
                  << " result " << c->_result << "\n" << eor;
    }
    c->_indirect->releaseIndirect( c->_info, 0 );
    releaseBlockLock( c->_info->source(), c->_info->unit() );
    delete c;
}

#endif // SAFE

UNIXInodeIndirectInfo *
UNIXInodeIndirect::newIndirect( PersistentMemoryObject * source, 
                                unsigned unit, unsigned idNumber )
{
    Debug   
    << "UNIXInodeIndirect[" << _log2BlockSize 
    << "]::newIndirect( "
    << source << ", " << unit << ", " << idNumber 
    << ")\n" << eor;

    acquire();
    unsigned slot = hash( source, unit );
    UNIXInodeIndirectInfo * found = victimize();
    Assert(found->_lock == -1);
    ByteZero( (char *) found->_buf, _blockSize );
    found->_source = source;
    found->_unit = unit;
    found->_idNumber = idNumber;
    found->_dirty = 1;
#ifdef SAFE 
    found->_lock = 2;
#else // SAFE
    found->_lock = 1;
#endif // SAFE
    _busy--;
    _locked++;
    found->_age = age();
    addToList( _hashTable + slot, found );
    Assert( found->_list ==
            _hashTable + hash(found->_source, found->_unit));
    release();
#ifdef SAFE
    int shift = _log2BlockSize - source->log2UnitSize();
    Assert( shift >= 0 );
    int sourceUnits = 1 << shift;
    UNIXInodeNewIndirectAsyncClosure * c = 
    new UNIXInodeNewIndirectAsyncClosure( this, found, sourceUnits );
    acquireBlockLock( source, unit );
    source->asyncWriteThrough( unit, sourceUnits, (char *) found->_buf, 
                               _blockSize, c );
#endif // SAFE
    return found;
}


UNIXInodeIndirectInfo * 
UNIXInodeIndirect::basicGetIndirect( PersistentMemoryObject * source, 
                                     unsigned unit, unsigned idNumber )
{
    Debug   
    << "UNIXInodeIndirect[" << _log2BlockSize 
    << "]::basicGetIndirect( "
    << source << ", " << unit << ", " << idNumber 
    << " )\n" << eor;

    release();
    unsigned lockHandle = acquireGetLock( source, unit );
    acquire();
    unsigned slot = hash( source, unit );
    UNIXInodeIndirectInfo * info = _hashTable[slot];
    while( info ) {
        if ( source == info->_source && unit == info->_unit &&
             info->_lock != -1 ) 
        {
            Assert( info->_lock >= 0 );
            if ( info->_lock == 0 ) {
                _unlocked--;
                _locked++;  
            }
            info->_lock ++;
            info->_age = age();
            _slowHit++;
            releaseGetLock( lockHandle );
            Debug   
            << "UNIXInodeIndirect[" << _log2BlockSize 
            << "]::basicGetIndirect() cached as "
            << asAddress( info ) << "\n" << eor;
            return info;
        }
        info = info->_next;
    }
    UNIXInodeIndirectInfo * found = victimize();
    Assert( found->_lock == -1 );
    release();
    int ok = read( source, unit, (unsigned int*)found->_buf );
    Assert( ok );
    acquire();
    _miss++;
    found->_source = source;
    found->_unit = unit;
    found->_idNumber = idNumber;
    found->_dirty = 0;
    found->_lock = 1;
    found->_age = age();
    _busy--;
    _locked++;
    addToList( _hashTable + slot, found );
    releaseGetLock( lockHandle );
    Assert( found->_list ==
            _hashTable + hash(found->_source, found->_unit));
    Debug   
    << "UNIXInodeIndirect[" << _log2BlockSize 
    << "]::basicGetIndirect( "
    << source << ", " << unit << ", " << idNumber 
    << " ) return " << asAddress( found ) << "\n" << eor;
    return found;
}

void
UNIXInodeIndirect::synchronize( UNIXInodeIndirectInfo * info )
{
    acquire();
    if ( info->_dirty && info->_lock != -1 && info->_source != 0 ) {
        if ( info->_lock == 0 ) {
            _unlocked--;
            _locked++;  
        }
        info->_lock++;
        Assert( info->_list ==
                _hashTable + hash(info->_source, info->_unit));
        release();
        unsigned ok = write( info );
        Assert( ok );
        acquire();
        info->_dirty = 0;
        info->_lock--;
        Assert( info->_lock >= 0 );
        if ( info->_lock == 0 ) {
            _unlocked++;
            _locked--;
        }
    }
    release();
}

void
UNIXInodeIndirectSynchronizeClosure::func( 
    UNIXInodeIndirectSynchronizeClosure * c )
{
    c->_func = (AsyncFunc)&funcWriteDone;
    c->_info->setPending( 0 );
    c->_indirect->write( c->_info, c );
}

void
UNIXInodeIndirectSynchronizeClosure::funcWriteDone( 
    UNIXInodeIndirectSynchronizeClosure * c )
{
    c->_indirect->writeDone( c->_info, c->_result );
    if ( c->_release ) {
        c->_indirect->releaseIndirect( c->_info, 0 );
    }
    // c->_info->_age = UNIXInodeIndirect::age();
    delete c;
}

void
UNIXInodeIndirect::timeout( UNIXInodeIndirectSynchronizeClosure * c )
{
    TheAsyncSupport->schedule( c );
}

void
UNIXInodeIndirect::synchronize( PersistentMemoryObject * source,
                                unsigned unit, unsigned idNumber, 
                                UNIXInodeIndirectInfo * hint, 
                                unsigned delay, int releaseWhenDone )
{
    acquire();
    UNIXInodeIndirectInfo * found = 0;
    if ( hint != 0 && hint->_source == source && hint->_unit == unit &&
         hint->_lock != -1 ) 
    {
        // Francis: TODO I don't think this assertion is correct
        // Assert( hint->_idNumber == idNumber );
        Assert( hint->_list ==
                _hashTable + hash( hint->_source, hint->_unit ) );
        found = hint;
    } else {
        unsigned slot = hash( source, unit );
        UNIXInodeIndirectInfo * info = _hashTable[slot];
        while( info ) {
            if ( source == info->_source && unit == info->_unit &&
                 info->_lock != -1 ) 
            {
                found = info;
                break;
            }
        }
        info = info->_next;
    }
    if ( found == 0 ) {
        Assert( ! releaseWhenDone );
        release();
        return;
    }
    UNIXInodeIndirectInfo * info = found;
    if ( info->_dirty && info->_pending == 0 ) {
        info->_pending++;
        if ( info->_lock == 0 ) {
            _unlocked--;
            _locked++;  
        }
        info->_lock++;
        release();

        UNIXInodeIndirectSynchronizeClosure * c =
        new UNIXInodeIndirectSynchronizeClosure
        ( this, info, releaseWhenDone );
#ifdef VCHOICES
        timeout(c);
#else
        if ( delay == 0 ) {
            timeout( c );
        } else {
            TheSoftwareClockManager->add
            ( ( SoftClockFunc ) &timeout, (char *) c, 
              delay * TheSoftwareClockManager->hz() );
        }
#endif
    } else {
        if ( releaseWhenDone ) {
            info->_lock--;
            Assert( info->_lock >= 0 );
            if ( info->_lock == 0 ) {
                _locked--;
                _unlocked++;
            }
        }
        // info->_age = age();
        release();
    } 
}

void
UNIXInodeIndirect::synchronize( PersistentMemoryObject * source, 
                                unsigned idNumber )
{
    acquire();
    UNIXInodeIndirectInfo * info = _infos;
    while ( info != 0 ) {
        if ( info->_dirty && info->_lock != -1 &&
             info->_idNumber == idNumber && info->_source == source ) {
            if ( info->_lock == 0 ) {
                _locked++;
                _unlocked--;
            }
            info->_lock ++;
            Assert( info->_list ==
                    _hashTable + hash(info->_source, info->_unit));
            release();
            unsigned ok = write( info );
            Assert( ok );
            acquire();
            info->_lock --;
            info->_dirty = 0;
            if ( info->_lock == 0 ) {
                _locked--;
                _unlocked++;
            }
        }
        info = info->_neighbor;
    }
    release();
}


void
UNIXInodeIndirect::synchronize( PersistentMemoryObject * source )
{
    acquire();
    UNIXInodeIndirectInfo * info = _infos;
    while ( info != 0 ) {
        if ( info->_dirty && info->_lock != -1 &&
             info->_source == source ) {
            Assert( info->_list ==
                    _hashTable + hash(info->_source, info->_unit));
            if ( info->_lock == 0 ) {
                _locked++;
                _unlocked--;
            }
            info->_lock ++;
            release();
            unsigned ok = write( info );
            Assert( ok );
            acquire();
            info->_lock --;
            info->_dirty = 0;
            if ( info->_lock == 0 ) {
                _locked--;
                _unlocked++;
            }
        }
        info = info->_neighbor;
    }
    release();
}

void
UNIXInodeIndirect::free( PersistentMemoryObject * source )
{
    acquire();
    UNIXInodeIndirectInfo * info = _infos;
    while ( info != 0 ) {
        if ( info->_source != source || info->_lock == -1 ) 
            continue;
        Assert( info->_lock == 0 && info->_dirty == 0 );
        removeFromList( _hashTable + hash( info->_source, info->_unit ),
                        info );
        info->_source = 0;
        info->_lock = 0;
        info->_unit = 0;
        info->_dirty = 0;
        info->_age = 0;
        addToList( &_freeList, info );
        _free++;
        _unlocked--;
        Debug
        << "UNIXInodeIndirect[" << _log2BlockSize 
        << "]::free( source " << source << " ) freeing "
        << asAddress(info) << "\n" << eor; 
        info = info->_neighbor;
    }
    release();
}

