#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Panos Kougiouris (kougiour@cs.uiuc.edu)
 * Contributing Author: See-mong Tan (stan@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * 1. Member methods of class Disk (file Disk.h)
 */

#include "Assert.h"
#include "Class.h"
#include "ClassStar.h"
#include "Convert.h"
#include "Disk.h"
#include "DiskStar.h"
#include "DiskScheduler.h"
#include "Domain.h"
#include "FileSystemSimpleInstruments.h"
#include "FreeRunningTimer.h"
#include "Kernel.h"
#include "MemoryObjectAsync.h"
#include "NameServer.h"
#include "OutputStream.h"
#ifndef SOMEDAY
#include "PhysicalMemoryChain.h"
#endif
#include "Raid.h"
#include "Semaphore.h"
#include "Store.h"
#include "ThisProcess.h"


#ifdef __GNUC__
#pragma implementation
#endif

#define DEFAULTDISKSCHEDULER CSCANDiskScheduler
// #define DEFAULTDISKSCHEDULER OneAtATimeDiskScheduler

/************************************************************************/
/*  PUBLIC Member methods for  Disk					*/
/************************************************************************/
extern  Store * MainStore;
#ifndef SOMEDAY
extern "C" { void PhysCopyBlock( unsigned char *, unsigned char *, int );}
#endif

Disk::Disk( short retries, DiskScheduler * diskScheduler )
	: PersistentMemoryObject( 0, 0, 0, 0 )
{
	initDebug();
	Assert( _moLock != 0 );
	delete _moLock;
	_moLock = 0;
	_retries = retries;
	if ( diskScheduler == 0 ) 
		diskScheduler = new DEFAULTDISKSCHEDULER( this );
	_diskScheduler = diskScheduler;
}

Disk::~Disk()
{
	delete _diskScheduler;
}

/**
 * Accessor for the disk scheduler for this disk.
 * @return The scheduler for this disk.
 */
DiskScheduler * Disk::diskScheduler()
{
  return _diskScheduler;
}

int
Disk::partitions()
{
        Raid( RaidMethod )( "%N::partitions() 1\n", this );
        return( 1 );
}

int
Disk::locatePartition( int idNumber, int &start, int &units )
{
	/*
	 * The default is to just have one partition, the whole disk.
	 */
	if ( idNumber != 0 )
		return( -1 );
	
	start = 0;
	units = numberOfUnits();
	
	return( 0 );
}

Class *
Disk::supports( Class * aClass)
{
        Raid( RaidProtected )( "%NDisk::supports( %N )\n", this, aClass );

        Class * supportedClass = 0;

	ClassStar diskContainerClass = thisProcess()->nameServer()->lookup(
			"DiskContainer", ClassClass );

        if( diskContainerClass->isASubclassOf( aClass ) ) {
            supportedClass = diskContainerClass;
        }

        Raid( RaidProtected )( "%NDIsk::supports() returns %N\n",
                                this, supportedClass);

        return( supportedClass );
}

int
Disk::isContainer()
{
	return 1;
}

/************************************************************************/
/*  PROTECTED 	DISK METHODS						*/
/************************************************************************/

int
Disk::rawWrite( unsigned int start, unsigned count, char * addr, unsigned nbytes,
	        int writeThrough, IOAsyncClosure * c )
{
	Raids( RaidProtected )
		<< this << "Disk::rawWrite( " << start << ", " << count
		<< ", " << asAddress( addr ) << ", " << nbytes 
		<< ", " << writeThrough 
		<< ", " << asAddress( c ) << " )\n" << eor;	
	unsigned realCount;
	Assert( nbytes == ( count << log2UnitSize() ) );
	if ( c ) {
		realCount = 0;
		DiskIORequest * ior = new DiskIORequest
				( DiskWriteOperation, start, count, addr,
				  nbytes, writeThrough, &(c->_result),
				  c, 0 );
		_diskScheduler->add( ior );
	} else {
		Semaphore sem(0);
		DiskIORequest ior( DiskWriteOperation, start, count, addr,
				   nbytes, writeThrough, &realCount,
				   0, &sem );
		_diskScheduler->add( &ior );
		sem.P();
	}
	return( realCount );
}

int
Disk::rawRead( unsigned int start, unsigned count, char * addr, unsigned nbytes,
	       IOAsyncClosure * c )
{
	Raids( RaidProtected )
		<< this << "Disk::rawRead( " << start << ", " << count
		<< ", " << asAddress( addr ) << ", " << nbytes 
		<< ", " << asAddress( c ) << " )\n" << eor;	
	unsigned realCount;
	Assert( nbytes == ( count << log2UnitSize() ) );
	if ( c ) {
		realCount = 0;
		DiskIORequest * ior = new DiskIORequest
				( DiskReadOperation, start, count, addr,
				  nbytes, 0, &(c->_result), c, 0 );
		_diskScheduler->add( ior );
	} else {
		Semaphore sem(0);
		DiskIORequest ior( DiskReadOperation, start, count, addr,
				   nbytes, 0, &realCount, 0, &sem );
		_diskScheduler->add( &ior );
		sem.P();
	}
	return( realCount );
}

int
Disk::doio( unsigned int start, int count, char * addr, int nbytes, int op )
{
    Debug << this << "Disk::doio( start=" << start << ", count=" <<
        count << ", addr=" << asHex( (int) addr ) << ", nbytes = " << nbytes <<
		", op = " << op << " )\n" << eor;

#ifndef SOMEDAY

    /* 
     * The code in this function is supposed to be portable but it 
     *   really isn't.  Panos wrote this but didn't try it on all
     *   machines.  At this point it is always overloaded by the child.
     */
    Assert( NOTREACHED );
    return( 0 );
#else
    PhysicalMemoryChain chain = thisProcess()->domain()->constructChain(
					addr, nbytes );
    thisProcess()->domain()->makeAddressable(addr, nbytes );

    if( start >= this->numberOfUnits() )
    {
        Console() << this << "Disk::doio() failed. "
                 <<  "Block:" << start << " out of range\n" << eor; 
         return 0 ;
    }


    if ( chain.lengthInBytes() < unitSize() * count ) {
        Console() << this << "Disk::doio: insufficient chain length\n" << eor;
        return 0;
    }

    /* iterate over the chain for count * unitSize() number of bytes */
    PhysicalMemoryChainIterator chainIterator( &chain );
    PhysicalMemoryChain * subChain = chainIterator.first();

    /* for count number of items, try to read from start */
    for(int numberOfSectorsRead = 0; numberOfSectorsRead < count;
        numberOfSectorsRead++, start++) 
    {

        Assert( subChain != 0 );
        /* want to iterate over the subChain for one sector size of memory */
        PhysicalMemoryChainContiguousBlockIterator
            contiguousBlockIterator( subChain, subChain->lengthInBytes() );

        unsigned int contiguousBlockSize = 0;
        char * transferAddress = (char *)
            contiguousBlockIterator.next( contiguousBlockSize );
        Debug << this << "Disk::doio transfer Address=" << 
		asHex((unsigned int) transferAddress) <<
            ", contiguousBlockSize=" << contiguousBlockSize << "\n" << eor;

	/* Here we solve the problem which arises because we get
	 * a memory chain as argument. The problem is that the
	 * number of bytes in the first PAU may not be enough to
	 * get a whole disk block. If this is the case we read 
	 * the block to a buffer and then we split it.
	 */

        switch( op ) { 
	case DiskReadOperation:
           {Debug << this << "Disk::doio: Reading sector " << start
		  << "\n" << eor;
	     
	    char * realTransferAddress;
	    PhysicallyAddressableUnit * unit;
            if ( contiguousBlockSize >= unitSize() ) {
                Debug << this << "Disk::doio: Read is contiguous\n" << eor;
		realTransferAddress = transferAddress;	
            }
            else {    // still a read, but discontiguous
                Debug << this << "Disk::doio: Read is not contiguous \n" << eor;

                /* get a free page from the main store, use it as a
                 * temporary space to copy to */
                 unit = MainStore->alloc();

                 realTransferAddress = (char *) unit->address();
	    }

	    if( doioWithRetries ( start,realTransferAddress,
 				        DiskReadOperation ) ==  DiskOperFailed )
	    {
        	Console() << this << "Disk::read()  "
                	 <<  "Block:" << start <<  " hard failed\n" << eor; 
	        if( transferAddress != realTransferAddress )
	  	       MainStore->dealloc( unit );    // return to main store
		 return( 0 );
	    }


	    /* If we did not transfer directly then ... */
	    if( transferAddress != realTransferAddress )
	    {

                /* now transfer from the temporary page to the real chain's
                 * first part, then the second part.
	 	 */
                int firstTransferSize = contiguousBlockSize;
                PhysCopyBlock( (unsigned char *) unit->address(),
                               (unsigned char *) transferAddress,
                              firstTransferSize );
                int nextTransferSize = unitSize() - firstTransferSize;
                transferAddress = (char *)
                    contiguousBlockIterator.next( contiguousBlockSize );
                /* In the next contiguous block it will have enough space
		 * unless if the disk unit size is bigger than the page size!
		 */
                Assert( contiguousBlockSize >= nextTransferSize );
                PhysCopyBlock( (unsigned char *) unit->address() +
                              firstTransferSize,
                               (unsigned char *) transferAddress,
                               nextTransferSize );
                MainStore->dealloc( unit );    // return to main store
           }
         } /* case */
	 break;
       case DiskWriteOperation:
         { Debug << this << "Disk::doio: Writing block " << start
		 << " to disk\n" << eor;

	    char * realTransferAddress;
	    PhysicallyAddressableUnit * unit;
            if ( contiguousBlockSize >= unitSize() ) {
                Debug << this << "Disk::doio(): write is contiguous\n" << eor;
		realTransferAddress = transferAddress;	
            }
            else 
            {    // still a write, but discontiguous
                Debug << this << "Disk::doio: write is not contiguous write\n"
		      << eor;

                /* get a free page from the main store, use it as a
                 * temporary space to copy to */
                 unit = MainStore->alloc();
                 realTransferAddress = (char *) unit->address();
                /* now transfer from the real chain's
                 * first part, to the temp page then the second part.
	 	 */
                int firstTransferSize = contiguousBlockSize;
                PhysCopyBlock( (unsigned char *) transferAddress,
                               (unsigned char *) realTransferAddress,
                              firstTransferSize );
                int nextTransferSize = unitSize() - firstTransferSize;
                transferAddress = (char *)
                    contiguousBlockIterator.next( contiguousBlockSize );
                /* In the next contiguous block it will have enough space
		 * unless if the disk unit size is bigger than the page size!
		 */
                Assert( contiguousBlockSize >= nextTransferSize );
                PhysCopyBlock( (unsigned char *) transferAddress,
			       (unsigned char *) realTransferAddress + 
							firstTransferSize,
                               nextTransferSize );

	    }

	    if( doioWithRetries( start,realTransferAddress, 
				DiskWriteOperation ) == DiskOperFailed )
	    {
        	Console() << this << "Disk::write() "
                	 <<  "Block:" << start  << " hard failed\n" << eor; 
	        if( transferAddress != realTransferAddress )
	  	       MainStore->dealloc( unit );    // return to main store
		return( 0 );
	    }


	    /* If we did not transfer directly then ... */
	    if( transferAddress != realTransferAddress )
                      MainStore->dealloc( unit );    // return to main store
      }
      break;
      default:
            Console() << this << "Disk::doio(): Operation " << op
		  << " is not read or write\n" << eor;
	    return 0;
      } /* switch */

        /* go onto next part of the chain */
        subChain = chainIterator.next( unitSize() );
  } /* for */
  Debug << this << "Disk::doio(): Successful\n" << eor;

  return numberOfSectorsRead;
#endif
}


/*********************************************************************
 *  PRIVATE DISK METHODS
 *********************************************************************/
	
int
Disk::doioWithRetries( unsigned int start, char * buf, short op ){

    int retries = _retries;            /* Retry so many times */
    int status = DiskOperFailed;       /* It will changed inside do..while */
    do{
         status = this->basicDoio( start, buf, op );
         if( status == DiskOperSucceded )  break; /* BREAK WHILE LOOP EXIT */
         /* We retry because we hope it was a transient problem.  */
	 Raids( RaidDetail ) << this << "Disk::doioWithRetries() failed. retries left: "
	       << retries << "\n" << eor; 
	/* Before retry initialize the controller */
	 while(( retries-- >= 0) && ( this->reset() != DiskOperSucceded ));
    }while ( retries-- >= 0 );

    return status;

}

void timeDisks( int argc, char * argv[] )
{
    if ( argc != 2 ) {
	Console() << "arguments: blockSize count\n" << eor;
	return;
    }
    DiskStar disk = TheKernel->rootDisk();
    if ( disk == 0 ) {
	Console() << "no root disk\n" << eor;
	return;
    }
    unsigned  blockSize = atoi(argv[0]);
    unsigned count = atoi(argv[1]);
    if ( blockSize < disk->unitSize() ||
	 ( blockSize % disk->unitSize() ) != 0 ) {
	Console() << "illegal block size " << blockSize
		  << " specified, disk unit size = " << disk->unitSize()
		  << "\n" << eor;
	return;
    }
    if (count == 0)
	return;
    char * buf = new char[blockSize];
    int unitsPerBlock = blockSize / disk->unitSize();
    int unitNo = 0;
    int sum = 0;
    unsigned start = SystemTimer->time();
    for( unsigned i = 0; i < count; i ++ ) {
	int result = disk->read( unitNo, unitsPerBlock, buf, blockSize );
	sum += result;
	unitNo += result;
	if ( result != unitsPerBlock ) {
	    Console() << "iteration " << i << ": result " << result << " != " 
		      << unitsPerBlock << "\n" << eor;
	    break;
	}
    }
    unsigned end = SystemTimer->time();
    unsigned total = end - start;
    delete buf;
    Assert( total != 0 );
    sum = sum << disk->log2UnitSize();
    int rate = ( sum * 1000 ) / total;
    Console() << "time taken " << total 
	      << " transferred bytes " << sum << "\n"
	      << "transfer rate " << rate << " Kbytes/sec\n" << eor;
    return;
}

#include "ClassInstaller.h"
DEFINECLASS("Disk",Disk,PersistentMemoryObject,0,0);
