#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#include "Async.h"
#include "Disk.h"
#include "DiskScheduler.h"
#include "Kernel.h"
#include "MemoryObjectAsync.h"
#include "OutputStream.h"
#include "Process.h"
#include "PROCESSOR.h"
#include "Semaphore.h"
#include "ThisProcess.h"

#ifdef __GNUC__
#pragma implementation
#endif

#define SEPARATEASYNCSUPPORT

// #define DEBUGCSCAN
// #define USESEM

OutputStream & operator<< ( OutputStream & stream, DiskIORequest & ior )
{
	stream << "op " << ior._op << ", start " << ior._start << ", count " 
	       << ior._count << ", through " << ior._through;
	return stream; 
}


DiskScheduler::DiskScheduler( Disk * disk )
	: _disk( disk )
{
	_queueLength = 0;
	_head = _tail = _current = 0;
	_process = new SystemProcess( (OPFV)&DiskScheduler::daemon, TheKernel->domain(),
				      small, (int) this );
	_process->reference();
#ifdef SEPARATEASYNCSUPPORT
	_asyncSupport = new AsyncSupport( 2 );
#else
	_asyncSupport = TheAsyncSupport;
#endif 
	//_process->ready();
	//_process = 0;
	//if ( thisProcess() != 0 ) startDaemon();
}

void
DiskScheduler::startDaemon()
{
	if ( _process != 0 )
		return;
	_process = new SystemProcess( (OPFV)&DiskScheduler::daemon, TheKernel->domain(),
				      small, (int) this );
	Assert( _process != 0 );
	_process->reference();
	_process->ready();
	//thisProcess()->relinquishProcessor();
	thisProcess()->relinquishProcessor();
}

void
DiskScheduler::daemon()
{
	IntStorage intState ; PROCESSOR::disableInterrupts(intState);
	_lock.acquire();

	while( 1 ) {
		if ( _queueLength == 0 ) {
			_lock.release();
			_process->block( this );
			_lock.acquire();
			Assert( _head != 0 );
		}
		Assert( _queueLength > 0 );
		DiskIORequest * ior = _current = _head;  
		_head = _head->_next;
		if ( _head == 0 )
			_tail = 0;
		else
			_head->_prev = 0;
		_lock.release();
		PROCESSOR::restoreInterrupts( intState );
		// Console() << "DiskScheduler::doio start\n" << eor;
		*(ior->_result) =
			_disk->doio( ior->_start, ior->_count, ior->_buf,
				     ior->_nbytes, ior->_op );
		// Console() << "DiskScheduler::doio done\n" << eor;
		if ( ior->_sem ) {
			ior->_sem->V();
		} else {
			Assert( ior->_closure != 0 );
			ior->_closure->_result >>= ior->_closure->_rightShift;
			ior->_closure->_urgency = 0;
			_asyncSupport->schedule( ior->_closure );
			delete ior;
		}
		PROCESSOR::disableInterrupts(intState);
		_lock.acquire();
		_queueLength--;
	}
}

void
FIFODiskScheduler::add( DiskIORequest * ior )
{
	IntStorage intState ; PROCESSOR::disableInterrupts(intState);
	_lock.acquire();
	_queueLength++;
	ior->_next = 0;
	ior->_prev = _tail;
	if ( _tail ) {
		_tail->_next = ior;
		_tail = ior;
	} else {
		_tail = _head = ior ;
	}
	int unblock = ( _queueLength == 1 );
	_lock.release();
	if ( unblock ) {
		Process * process = thisProcess();
		process->ready();
		process->giveProcessorTo( _process, ProcessReady );
	}
	PROCESSOR::restoreInterrupts( intState );
}

#ifdef DEBUGCSCAN
int debugCSCAN = 0;
#endif // DEBUGCSCAN

void
CSCANDiskScheduler::add( DiskIORequest * req )
{
	IntStorage intState ; PROCESSOR::disableInterrupts(intState);
	_lock.acquire();
	_queueLength++;
	if ( _head == 0 ) {
#ifdef DEBUGCSCAN
		if (debugCSCAN) Console() 
			<< "queue empty, new head " << *req << "\n" << eor;
#endif // DEBUGCSCAN
		Assert( _head == 0 && _tail == 0 );
		_head = _tail = req;
		req->_prev = req->_next = 0;
	} else {
		Assert( _current != 0 );
		unsigned last = _current->_start;
		DiskIORequest * nextReq = _head;
#ifdef DEBUGCSCAN
		if (debugCSCAN && nextReq) Console() 
			<< "examining " << *nextReq << "\n" << eor;
#endif // DEBUGCSCAN
		while ( nextReq != 0 && 
			!( req->_start >= last && 
			   req->_start < nextReq->_start ) &&
			nextReq->_start >= last ) {
			last = nextReq->_start;
			nextReq = nextReq->_next;
#ifdef DEBUGCSCAN
			if (debugCSCAN && nextReq) Console() 
				<< "examining " << *nextReq << "\n" << eor;
#endif // DEBUGCSCAN
		}
		last = 0;
#ifdef DEBUGCSCAN
		if (debugCSCAN && nextReq) Console() 
			<< "examining " << *nextReq << "\n" << eor;
#endif // DEBUGCSCAN
		while ( nextReq != 0 && 
		    	!( req->_start >= last && 
			   req->_start < nextReq->_start) ) {
			last = nextReq->_start;
			nextReq = nextReq->_next;
#ifdef DEBUGCSCAN
			if (debugCSCAN && nextReq) Console() 
				<< "examining " << *nextReq << "\n" << eor;
#endif // DEBUGCSCAN
		}
		DiskIORequest * prevReq;
		if ( nextReq == 0 )
			prevReq = _tail;
		else 
			prevReq = nextReq->_prev;
#ifdef DEBUGCSCAN
		if (debugCSCAN) {
			if ( prevReq == 0 )
				Console() << "insert at head";
			else
				Console() << "insert after " << *prevReq;
			if ( nextReq == 0 ) 
				Console() << " before tail\n" << eor;
			else
				Console() << " before " << *nextReq 
					  << "\n" << eor;
		}
#endif // DEBUGCSCAN
		req->_prev = prevReq;
		if ( prevReq ) {
			prevReq->_next = req;
		} else {
			_head = req;
		}
		req->_next = nextReq;
		if ( nextReq ) {
			nextReq->_prev = req;
		} else {
			_tail = req;
		}
	}

	int unblock = ( _queueLength == 1 );
	_lock.release();
	if ( unblock ) {
		Process * process = thisProcess();
		process->ready();
		process->giveProcessorTo( _process, ProcessReady );
	}
	PROCESSOR::restoreInterrupts( intState );
}

OneAtATimeDiskScheduler::OneAtATimeDiskScheduler( Disk * disk )
	: DiskScheduler( disk )
{
	_oneAtATimeMutex = new Semaphore( 1 );
	Assert( _oneAtATimeMutex != 0 );
}

class CheatAsyncClosure : public IOAsyncClosure
{
public:
	AsyncFunc getFunc() { return _func; }
};

void
OneAtATimeDiskScheduler::add( DiskIORequest * ior )
{
	_oneAtATimeMutex->P();
	*(ior->_result) =
		_disk->doio( ior->_start, ior->_count, ior->_buf,
			     ior->_nbytes, ior->_op );
	_oneAtATimeMutex->V();
	if ( ior->_sem ) {
		ior->_sem->V();
	} else {
		Assert( ior->_closure != 0 );
		Assert( &(ior->_closure->_result) == ior->_result );
		ior->_closure->_result >>= ior->_closure->_rightShift;
		ior->_closure->_urgency = 0;
		(*( ((CheatAsyncClosure *)(ior->_closure))->getFunc() ))( ior->_closure );	
		delete ior;
	}
}

