#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Panagiotis Kougiouris (kougiour@cs.uiuc.edu ) 
 * Contributing Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * RAMDisk.c: a RAM disk driver.
 *
 */

#include "Assert.h"
#include "OutputStream.h"
#include "PhysicalMemoryChain.h"
#include "RAMDisk.h"


#ifdef __GNUC__
#pragma implementation
#endif

/* This is the only machine dependent aspect of a RAMDisk. Each machine
 * should provide a function that copies a number of bytes (the third
 * argument) from a physical address (the first argument) to another
 * physical address (the second argument). Note that since the addresses
 * are physical the MMU must be off.
 */
extern "C" { void PhysCopyBlock( unsigned char *, unsigned char *, int );}



int 
RAMDisk::partitions()
{
    return( 1 );
}

RAMDisk::RAMDisk( char * Base, unsigned int numberOfUnits, int log2UnitSize )
    : Disk( 1 /* retries */ )
{
    RAMDiskBasePhysicalAddress = Base;
    _numberOfUnits = numberOfUnits;    /* defined in MemoryRange */
    _log2UnitSize = log2UnitSize;      /* defined in MemoryRange */
    _moLock = new NormalRangeLock( this );
}

/**
 * Read or write a single disk block.
 * @param start The number of the block on which to operate.
 * @param buf IO buffer; assumed to hae this->unitSize() bytes.
 * @param oper One of DiskReadOperation or DiskWriteOperation.
 * @return One of DiskOperFailed or DiskOperSucceded.
 */
int RAMDisk::basicDoio( unsigned int start, char * buf, int oper)
{
    return this->doio(start, 1, buf, this->unitSize(), oper);
}

/**
 * Read or write a series of disk blocks.
 * @param start The first block for the operation.
 * @param count The number of blocks on which to operate.
 * @param addr Buffer: holds data to be written or target for data read.
 * @param nbytes The size of the buffer at addr.
 * @param op One of DiskReadOperation or DiskWriteOperation.
 * @return number of blocks read
 */
int RAMDisk::doio(unsigned int start, int count, char *addr, int nbytes, int op)
{
    char * 
    diskAddress = RAMDiskBasePhysicalAddress + (start * unitSize());

    Debug << this << "RAMDisk::doio block number:" << start 
          << "(address: " << asHex((unsigned int)diskAddress) 
          << ") addr: " << asHex((unsigned int)addr)
          << " nbytes: " << nbytes
          << " operation: " << op << ".\n" << eor;

    Assert (start < numberOfUnits());
    Assert (start + count <= numberOfUnits());

    switch( op ){
        case DiskReadOperation:
            PhysCopyBlock(  (unsigned char *)diskAddress, 
                            (unsigned char *)addr, 
                            nbytes );
            break;
        case DiskWriteOperation:
            PhysCopyBlock(  (unsigned char *)addr,
                            (unsigned char *)diskAddress, 
                            nbytes );
            break;
        default:
            Console() << this << "RAMDisk::doio" << 
            "asked to perform unknown operation " << op
                      << "\n" << eor;
            return( 0 );
    } /* switch( op ) */

    Debug << this << "RAMDisk::doio: completed\n" << eor;
    
    return( count );
}

int 
RAMDisk::reset(){
    return( 1 );
}

#include "ClassInstaller.h"
DEFINECLASS("RAMDisk",RAMDisk,Disk,0,0);
