/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Author: Taed Nelson (nelson@cs.uiuc.edu)
 * Author: David Raila (raila@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * CPU.h: per-cpu private information class.
 */

#ifndef CPU_h
#define CPU_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Assert.h"		/* parent class */
#include "ProxiableObject.h"		/* parent class */
#include "ProcessContainer.h"
#include "TIMESLICE_TIMER.h"		/* to get the TIMESLICETIMER define */
#include "Types.h"
#include "InterruptManager.h"

/*****************************************************************************
 *	Typedef for mutex methods
 *****************************************************************************/
typedef volatile int * const MutexStateAddr;

class AddressFlushQueue;
class AddressTranslator;
class Interrupt;
class OutputStream;
class Process;

#ifdef INSTRUMENT
#include "InstrumentTypes.h"
#include "Instrument.h"

struct cpuSyncType {
	Boolean used;
	volatile Boolean done;
	unsigned long cpu;
};
class InstrumentBuffer;
#endif

/*****************************************************************************
 *	Things for thisCPU, and the global CPU table, etc.
 *****************************************************************************/
const unsigned MAX_PROCESSORS = 64;	// For now....
typedef unsigned CpuId;			// Sufficient for all current types

extern ClassStar CPUClass;
class CPU : public ProxiableObject {
    friend class ProcessorContext;
    friend class Process;
#ifdef INSTRUMENT
    friend class SystemInstrumentManager;
    friend class InstrumentBufferInterrupt;
    friend void InstrumentBufferDumpProcessEntry (int);
#endif



protected:
	/*
	 * The process to run when there is nothing else to do. Also, this
	 * is the first process dispatched when a processor is started.
	 * This and _currentProcess are first because they are used by
	 * assembler routines, so it's easier to figure out their structure
	 * offsets if they're together.
	 */
	Process * _idleProcess;

	/*
	 * A pointer to this processors AddressTranslator.
	 */
	AddressTranslator * _MMU;

	/*
	 * The current idleContainer the processor is using.
	 */
	ProcessContainer * _idleContainer;

	/*
	 * The timer this processor uses to do time-slicing.
	 */
	TIMESLICETIMER *	_timesliceTimer;

	/*
	 * A unique ID for this processor.
	 */
	CpuId	_uniqueIdentifier;

	/*
	 * Queue of pending address validation requests.
	 * It seems like this should be in AddressTranslator. --Vince
	 */
	AddressFlushQueue * _flushQueue;

	/*
	 * Interrupt Manager of the Machine
	 */
	static InterruptManager * _interruptManager;

	/*
	 * These methods are called to do CPU specific actions on
	 * the current processor.  If they are not overridden in a
	 * subclass they fail assertions.
	 * They are called using the PROCESSOR macro that selects
	 * the processor class for the machine, like 
	 * PROCESSOR::enableInterrupts(...)
	 *
	 * disableInterrupts() disables and returns int for use in
	 * restoreInterrupts(), which restores the passed state.
	 *
	 * enableInterrupts() is used to enable them without using
	 * disableInterrupts(), since machine independent code doesn't
	 * know what state to pass.  It can simply call restoreInterrupts
	 * with the proper arg.
	 *
	 * interruptsEnabled() returns 1 if all interrupts are not off.
	 * 
	 * initMutexState(MutexStateAddr) initializes the value of a spin lock
	 * state (integer).
	 *
	 * spinMutex(MutexStateAddr) does a spin mutex, using the state var.
	 *
	 * releaseMutex(MutexStateAddr) releases the mutex, 
	 * freeing other spinners
	 *
	 * cpuID() returns a unique integer for each CPU it runs on.
	 */
	//static void disableInterrupts(); 
	//static void disableInterrupts(IntStorage &); 
	//static void restoreInterrupts(int val);
	//static void enableInterrupts();
	//static int interruptsEnabled();
	//static void initMutexState(MutexStateAddr);
	//static void spinMutex(MutexStateAddr);
	//static void releaseMutex(MutexStateAddr);
	//static unsigned cpuID();

public:

	/*
	 * each processor sets up a boot stack and calls this bugger
	 * to start Choices
	 */
	static void processorStart();

	/*
	 * The piece of code all idle processor execute via their
	 * idle Process.
	 */
	static void idleLoop();

	/* 
	 * support for thisProcess() and thisCPU()
	 */
	static CPU * _cpus[MAX_PROCESSORS];
	static Process * volatile _runningProcesses[MAX_PROCESSORS];

	CPU( CpuId identifier,
		ProcessContainer * initialIdleContainer );
	~CPU();

	/*
	 * Overload this and it gets called in the idleLoop
	 * if there's nothing to dispatch.
	 */
	virtual void idleFunction() {};

	/*
	 * Used to bind interrupts to hardware vectors. (only used during
	 * bootup). This shouldn't be public!
	 */
	virtual void installInterrupts() = 0;

	/*
	 * To chip initializations. Mainly things like load processor
	 * registers. Subclasses are responsible for implementing this.
	 * A CPU::chipInitialize() function shouldn't allocate any new
	 * objects.  There is no heap to do so with.  This shouldn't
	 * be public either, but some Kernel::configure()s need to
	 * be able to handle interrupts.
	 */
	virtual void chipInitialize() = 0;


	/*
	 * Subclass implemented debugging assists.
	 * These should be moved to class process (or at least implemented
	 * there as well).
	 */
	// virtual int interruptsDisabled() = 0;
	virtual void debugBreakpoint();

	/*
	 * Access functions for most of the above stuff.
	 */
	TIMESLICETIMER * timesliceTimer();
	AddressTranslator * MMU();
	AddressFlushQueue * flushQueue();
	CpuId id();

	void setInterruptManager(InterruptManager *manager) { _interruptManager = manager; }
	InterruptManager *getInterruptManager() { return _interruptManager; }

	void setIdleContainer( ProcessContainer * container );

	/*
	 * Return the next ready to run process from this processors idle
	 * container. If there arent any, return this processors idle
	 * process.
	 */
	Process * getNextReadyProcess();

	/*
	 * Return the current process on this processor
	 * This should only be called with interrupts disabled.
	 */
	// Process * currentProcess();
	// Process * setCurrentProcess(Process *);
	Process * idleProcess(){return _idleProcess;}

	/*
	 * Subclasses define this to bind an interrupt to a hardware
	 * vector number. It should be implemented so that one processor
	 * calling it has the effect of activating the binding for all
	 * processors that potentially may receive the interrupt/trap.
	 * The easiest way to do this on a shared memory machine is for
	 * all processors to share the interrupt vector table.
	 * On a private memory machine the point is moot.
	 *
	 * Assist is the address of a function to be called when the
         * hardware vector is sent. If not specified, there should be a
         * default setting.
	 */
	virtual Interrupt * setInterrupt( int vector,
					Interrupt * interrupt,
					PFV assist = 0) = 0;
	virtual Interrupt * interrupt( int vector ) = 0;

	/*
	 * Flush a virtual address from the MMU TLB.
	 * I think this should be an AddressTranslator method. --Vince
	 */
	void flushAddressTranslationCache( const char * virtualAddress,
		unsigned int length );

	Class * classOf();

	/*
	 * This is a hack to make AddressFlushQueue portable until it is
	 * completly removed anyway. (AddressTranslation's will subsume the
	 * functionality).
	 */
	virtual void sendFlushRequestTo( unsigned long dest ) = 0;

	/*
	 * This is mainly here for objects which care whether or not
	 * they would get the idleProcess or not when they execute
	 * getNextReadyProcess.
	 */
	ProcessContainer * idleContainer();

#ifdef INSTRUMENT
	InstrumentBuffer *_logBuffer;

	Process *_dumpProcess, *_blockedByDumpProcess;
	ProcessContainer *_dumpContainer;
	int _didDumpFlag;

	unsigned int _icpuID;
	static cpuSyncType *_icpuSyncList;

	static unsigned int _highestInstrumentationID;
	
	unsigned int instrumentationID();

	void interruptAllCPUs (unsigned int vector);
	void syncWaitMyTurn ();
	void syncWaitForEveryoneElse ();
#endif

	/*
	 * Reset the CPU
	 */
	virtual void reset() {};
};

inline Process *
CPU::getNextReadyProcess()
{
	Process * p = _idleContainer->remove();
	if( p == 0 ) {
		p = _idleProcess;
	}
	Assert( p != 0 );
	return( p );
}

inline ProcessContainer *
CPU::idleContainer()
{
	return( _idleContainer );
}

inline TIMESLICETIMER *
CPU::timesliceTimer()
{
	return( _timesliceTimer );
}

inline AddressTranslator *
CPU::MMU()
{
	return( _MMU );
}

#ifdef NEVER
inline Process *
CPU::currentProcess()
{
	return( _currentProcess );
}

inline Process *
CPU::setCurrentProcess( Process *n )
{
	Process *p = _currentProcess;
	_currentProcess = n;
	return p;
}
#endif

inline AddressFlushQueue *
CPU::flushQueue()
{
	Assert (_flushQueue != 0);
	return( _flushQueue );
}

inline void
CPU::setIdleContainer( ProcessContainer * container )
{
	_idleContainer = container;
}

inline CpuId
CPU::id()
{
	return( _uniqueIdentifier );
}

inline void
CPU::debugBreakpoint()
{
}

#ifdef INSTRUMENT
inline unsigned int
CPU::instrumentationID()
{
	return ( _icpuID );
}
#endif

#endif /* CPU_h */
