/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Contributing Author: John L. Coolidge (coolidge@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Contributing Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Lee Lup Yuen (lup@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * CharacterQueue.h: character queueing Classes.
 */

#ifndef CharacterQueue_h
#define CharacterQueue_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Assert.h"
#include "Object.h"
#include "Lock.h"
#include "SemaphoreStar.h"


extern ClassStar CharacterQueueClass;
class CharacterQueue : public Object {
/*
 * This Class provides basic operations on a queue of characters.
 * It is an abstract base Class upon which more sophisticated queues are
 * built.  RawCharacterQueue and LineBufferedEchoingCharacterQueue are two
 * of the concrete subClasses.
 */

protected:
	Lock _lock;
	void lock(){_lock.acquire();}
	void unlock(){_lock.release();}

	int _size, _free;
	char * _queue;
	int _front;
	int _back;

	char head();

	char tail();
	int tailIsWS();


	int basicEnqueue( const char * buffer, int count = 1 );
	int basicDequeue( char * buffer, int count = 1 );

public:
	CharacterQueue( int size = 512 );
	~CharacterQueue();

	virtual int enqueue( const char * buffer, int count ) = 0;
	virtual int dequeue( char * buffer, int limit ) = 0;

	int full();
	int empty();

	virtual void flush();
	//  Set a special condition to indicate that there are no more
	//  characters to be enqueued.
	virtual void terminate () = 0;

	virtual void dump();

	Class * classOf();
};


inline int
CharacterQueue::full()
{
	return( _free == 0 );
}

inline int
CharacterQueue::empty()
{
	return( _free == _size );
}

inline char
CharacterQueue::head()
{
	Assert( !empty() );
	return( _queue[_front] );
}

inline char
CharacterQueue::tail()
{
	Assert( !empty() );
	return( _queue[_back] );
}

//  BlockCharacterQueues are for dequeueing blocks of characters.  Calls
//  to dequeue will block until the requested number of characters are
//  fully available, unless somebody has previously called terminate()
//  to indicate the end of the queue.

class BlockCharacterQueue : public CharacterQueue {
protected:
	SemaphoreStar _charBlockSemaphore;
	int _charsAvailable;
	int _charsRequested;
	int _terminated;
 public:
	BlockCharacterQueue( int size = 512 );
	
	int enqueue( const char * buffer, int count = 1 );
	int dequeue( char * buffer, int count = 1 );
	// void flush();
	void terminate();
};


class RawCharacterQueue : public CharacterQueue {
protected:
	Lock _lock;
public:
	RawCharacterQueue( int size = 512 );

	int enqueue( const char * buffer, int count = 1 );
	int dequeue( char * buffer, int limit = 1 );
	void terminate() {}
	void flush();
};


const int SpecialCharCount = 4;
const int Quote = 0;
const int Backspace = 1;
const int DeleteWord = 2;
const int DeleteLine = 3;

class OutputStream;

/*
 * This class now does NOT swallow CR/LF.  CR is converted to LF (which
 * may be questionable).  Attempts to dequeue from a queue block
 * until a full newline-terminated line is available.  The terminating
 * newline ('\n') is returned just as other characters are.
 *
 * Before a newline is enqueued, the enqueueing of an editing character
 * causes previously-enqueued characters to be discarded, for a primitive
 * line-editing system.  There is not yet any escape provision for
 * causing the editing characters themselves to be enqueued.
 */

class LineBufferedEchoingCharacterQueue : public CharacterQueue {
protected:
	Lock _lock;
	char _SpecialChar[ SpecialCharCount ];
	OutputStream * _echoStream;
	SemaphoreStar _lineSemaphore;
	int _lineLength;
	int _lineAvailable;
	int _requestingLine;
	int _terminated;

	virtual int unqueue( int n );
	int unqueueWord();
public:
	LineBufferedEchoingCharacterQueue( OutputStream * echoStream,
		int size = 256 );

	virtual int enqueue( const char * buffer, int count = 1 );
	virtual int dequeue( char * buffer, int limit = 1 );
	//void flush();
	void terminate();
	int setSpecialChar(int CharNum, char ch);
};


#endif /* CharacterQueue_h */
