/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Contributing Author: Lee Lup Yuen (lup@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * InputStream.h: definition of an input stream of characters.
 *
 */

#ifndef InputStream_h
#define InputStream_h


#ifdef __GNUC__
#pragma interface
#endif

#include "ProxiableObject.h"
#include "FormatAssists.h"

const unsigned inputStreamBufferSize = 200;  //  Read-back buffer size.

extern ClassStar InputStreamClass;
class InputStream : public ProxiableObject {
/*  An InputStream is a stream of characters, and operations are provided
 *  to read characters from that stream.  The stream of characters may be
 *  interpreted as textual representations for certain types - for example,
 *  formattedRead interprets the stream of characters into types like
 *  ints, floats, char*, etc. and returns the interpreted values to the
 *  user.  InputStreams may be subclassed to provide buffering (eg.
 *  CharacterInputStream) or to specify the source of characters (eg.
 *  ConsoleInputStream).
 */
 public:
	/* Method InputStream:
	 *  Initialize the input stream by setting the formattedRead
	 *  buffers and variables to null.
	 */
	InputStream();

	/*  Method read:
	 *  Read bufSize characters from the input stream into the buffer.
	 *  buffer must be big enough to hold bufSize characters.  Return
	 *  The actual number of characters read.
	 */
	proxiable virtual int read( char * buffer, int bufSize ) = 0;

	/*  Method formattedRead:
	 *  Works like scanf - read characters from the input stream,
	 *  interpret them according to the specified format, and write the
	 *  interpreted values into pointers a1 to a4.  (See the scanf
	 *  entry in an ANSI C manual for details about the format string.)
	 *  a1 through a4 must be pointers to variables, not actual variables.
	 *  a1 through a4 are optional.
	 */
	proxiable virtual int formattedRead( const char * format,
																			 const formatArgs & a1 = *(formatArgs *)0,
																			 const formatArgs & a2 = *(formatArgs *)0,
																			 const formatArgs & a3 = *(formatArgs *)0,
																			 const formatArgs & a4 = *(formatArgs *)0 );

	/*  Method multiFormattedRead:
	 *  Works like formattedRead, but supports more than 4 variables.
	 *  Read characters from the input stream, interpret them according
	 *  to the specified format, and write the interpreted values into
	 *  the variables pointed to by args[0], args[1], args[2], etc.
	 *  args must contain pointers to variables, and there must be as
	 *  many entries in args as the variables in the format string.
	 *  (See the scanf entry in an ANSI C manual for details about the
	 *  format string).
	 */
	virtual int multiFormattedRead( const char *format, int **args);

	/*  Method readBackChar:
	 *  Works like ungetc - push c back into a LIFO buffer so that it
	 *  will be interpreted later by formattedRead and multiFormattedRead.
	 *  Return -1 if c cannot be pushed.  Example: suppose the input stream
	 *  contains the characters "abcde" and formattedRead has already
	 *  consumed the first 2 characters, leaving "cde" in the stream.
	 *  If the user decides not to accept "b", he/she may call
	 *  readBackChar("b") to change the input stream buffer into "bcde".
	 *  Calling readBackChar("a") next will change the input stream buffer
	 *  buffer into "abcde".  Note that readBackChar may fail if the
	 *  buffer runs out of space.
	 */
	proxiable virtual int readBackChar( char c);
			
	/*  Method classOf: Return the class of this object.  */
	Class * classOf();
 private:
	//  These functions come straight from the BSD doscan implementation.
	//  see InputStream.cc for details.
	int _doscan( const char *, int **);
	int _innum( int **, int , int , int , int*);
	int _instr( char *,  int , int , int*);
	const char * _getccl( const char *);
	/*  Method getc: Get a character from the stream.  */
	char getc();
	/*  Method ungetc: Push back a character into the stream.  */
	char ungetc( char);
	
	/*  Buffer for characters to be interpreted by formattedRead.  */
	char readBuffer [inputStreamBufferSize + 1];
	int readBufferIndex;
	/*  Buffer for characters pushed back.  */
	char readBackBuffer [inputStreamBufferSize + 1];
	unsigned readBackBufferIndex;
};

/*
 * The 'StandardInput' variable is generally only used by applications.
 * It is usually connected to console input.
 */
extern InputStream * StandardInput;

#endif	/* InputStream_h */
