/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Contributing Author: Lee Lup Yuen (lup@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * OutputStream.h: definition of an output stream of characters.
 *
 */

#ifndef OutputStream_h
#define OutputStream_h


#ifdef __GNUC__
#pragma interface
#endif

#include "FormatAssists.h"
#include "ProxiableObject.h" /* parent */
#include "Debug.h"
/*
  #if defined (linux)
  extern "C" {
  #include <errno.h>
  }
  #endif
*/

/*
 * NOTES ON PROXIES: OutputStream.cc is copied into each Choices application.
 *   All non-virtual functions will be executed locally if called in user
 *   space.  This only works because there is no local data in this class
 *   so none of the member functions will try to operate on local data when
 *   it's really a Proxy.  Functions like 'formattedWrite' which take refs
 *   to formatArgs should NOT be virtual because they are a security hole:
 *   they actually pass a reference to a pointer which a user could specify to
 *   be protected memory and rely on the kernel to read it for them.
 */

enum OutputStreamPadChar { aZero, aSpace };
enum OutputStreamPadSide { padLeft, padRight };

//  Size of buffer for OutputStream formatting.
const unsigned outputStreamWriteBufferSize = 200;

extern ClassStar OutputStreamClass;
proxiable class OutputStream : public ProxiableObject {
/*  This is a stream of characters that may be written to.  OutputStream
 *  may be subclassed to specify where the stream of characters is sent
 *  to.  For example, you can make a subclass to send the characters to
 *  a terminal or a file.
 */
 public:
    /*  Method OutputStream:
     *  Initialize the output stream by setting the formattedWrite buffer
     *  to null.
     */
    OutputStream();

    /*  Method acquireExclusiveUse:
     *  Lock this output stream for exclusive use.  Block until we can
     *  use it exclusively.
     */
    proxiable virtual void acquireExclusiveUse();

    /*  Method releaseExclusiveUse:
     *  Unlock this output stream so that others may use it.
     */
    proxiable virtual void releaseExclusiveUse();

    /*  Method write:
     *  Write count characters from buffer to the output stream.  Return
     *  the actual number of characters written.
     */
    proxiable virtual int write( const char * buffer, int count ) = 0;

    /*  Method formattedWrite:
     *  Works like printf - translate variables a1 through a4 to their
     *  textual representations according to the specified format and
     *  write the text to the output stream.  formattedWrite is buffered -
     *  write() is not called until the entire format string has been
     *  interpreted.  Return the number of characters written.  (See
     *  an ANSI C manual for details on the format string.)
     */
    proxiable virtual int formattedWrite( const char * format,
                                          const formatArgs & a1 = *(formatArgs *)0,
                                          const formatArgs & a2 = *(formatArgs *)0,
                                          const formatArgs & a3 = *(formatArgs *)0,
                                          const formatArgs & a4 = *(formatArgs *)0);

    /*  Method flush:
     *  Flush all contents of the output stream to their destination.
     */
    proxiable virtual void flush();

    /*  Method classOf: Return the class of this object.  */
    virtual Class * classOf();

    /*  Method emitChar:
     *  Write c to the output stream.  Pad output with spaces to make
     *  "width" characters if width >1.  padSide is either padLeft or
     *  padRight to indicate whether the spaces should be padded on the
     *  left or the right of c.
     */
    void emitChar( char c, int width, OutputStreamPadSide padSide );

    /*  Method emitString:
     *  Write string to the output stream.  Pad output with spaces to make
     *  "width" characters if "width" > length of string.  padSide is
     *  either padLeft or padRight to indicate whether the spaces should
     *  be padded on the left or the right of c.
     */
    void emitString( const char * string, int width, OutputStreamPadSide );

    /*  Method emitSpaces: Write spaceCount spaces to the output stream. */
    void emitSpaces( int spaceCount );
    
    /*  Method emitZeros: Write zeroCount '0's to the output stream.  */
    void emitZeros( int zeroCount );
    
    /*  Method emitNumber:
     *  Convert arg into a text string using the specified base and write
     *  the text to the output stream.  base is either 2, 8, 10 or 16.
     *  If the resulting string has fewer than "width" characters, it is
     *  padded to make "width" characters.  padChar is set to aSpace
     *  or aZero to indicate whether spaces or '0's are used as the
     *  padding character.  padSize is either padLeft or padRight to
     *  indicate whether the padding should be done on the left or to
     *  the right of the string.  The string will contain a prefix to
     *  identify the base: "0b" for base 2, "0" for base 8, nothing or
     *  "-" for base 10, and "0x" for base 16.
     */
    void emitNumber( unsigned int arg, int base, int width,
                     OutputStreamPadSide padSide, OutputStreamPadChar padChar);

    /*  These operators write various types (eg. string, int) to the
     *  output stream.  They are used like this: "*StandardOutput << 123;".
     */
    OutputStream & operator <<( int );
    OutputStream & operator <<( unsigned int );
    OutputStream & operator <<( long );
    OutputStream & operator <<( unsigned long );
    OutputStream & operator <<( char );
    OutputStream & operator <<( const char * );
    OutputStream & operator <<( const void * );

    OutputStream & operator <<( Object * );
    OutputStream & operator <<( Object & );
    OutputStream & operator <<( ProxiableObjectRef & );
    OutputStream & operator <<( ProxiableObjectStar & );

    /*
     * The following methods are temporarily here for compatibility
     * with the old DiagnosticConsole class.  Do not use them any more!
     * They will be removed in future.
     */
    void printf( const char * f,
                 const formatArgs & a1 = *(formatArgs *)0,
                 const formatArgs & a2 = *(formatArgs *)0,
                 const formatArgs & a3 = *(formatArgs *)0,
                 const formatArgs & a4 = *(formatArgs *)0 );
    void cpuPrintf( const char * f,
                    const formatArgs & a1 = *(formatArgs *)0,
                    const formatArgs & a2 = *(formatArgs *)0,
                    const formatArgs & a3 = *(formatArgs *)0,
                    const formatArgs & a4 = *(formatArgs *)0 );
    void panic( const char * f,
                const formatArgs & a1 = *(formatArgs *)0,
                const formatArgs & a2 = *(formatArgs *)0,
                const formatArgs & a3 = *(formatArgs *)0,
                const formatArgs & a4 = *(formatArgs *)0 );
    void criticalPrintf( const char * f,
                         const formatArgs & a1 = *(formatArgs *)0,
                         const formatArgs & a2 = *(formatArgs *)0,
                         const formatArgs & a3 = *(formatArgs *)0,
                         const formatArgs & a4 = *(formatArgs *)0 );
    void ultimatePrintf( const char * f,
                         const formatArgs & a1 = *(formatArgs *)0,
                         const formatArgs & a2 = *(formatArgs *)0,
                         const formatArgs & a3 = *(formatArgs *)0,
                         const formatArgs & a4 = *(formatArgs *)0 );
 private:
    //  formattedWrite uses writeBuffer to buffer the output characters.
    char writeBuffer [outputStreamWriteBufferSize];
    unsigned writeBufferIndex;
    
    //  Note: some of these functions are supposed to accept double
    //  arguments, but due to a bug with g++ -mregparms on the
    //  Multimax, doubles cannot be passed to methods.  The kludge
    //  here uses the variable doubleParameter to pass doubles to
    //  OutputStream methods.
    
    //  These functions come straight from BSD doprnt.
    int doprnt(const char *format, formatArgs *args);
    void putChar(char ch);
    int convert(/* double,*/ int, int, char *, unsigned char, char *,
                char *);
    char *round(/* double, */ int *, char *, char *, char, char *);
    char *exponent(char *, int, unsigned char);
    char *memchr(char *, char, int);
    double modf(/* double, */ double *);
    double doubleParameter;
};


/*
 * A function to return an OutputStream to use as the system console.
 * The OutputStream returned can be used in any context since it
 * should never block.  This function is defined in OutputStream.cc.
 */
extern OutputStream &Console();

/*
 * The 'StandardOutput' variable is generally only used by applications.
 * The System should use the 'Console()' function; applications have that
 * function also defined to just return StandardOutput.  Applications use
 * this variable to write to the screen.
 */
extern OutputStream * StandardOutput;

inline 
OutputStream & operator << ( OutputStream & os, asHex hex )
{
    os.emitNumber( hex, 16, 0, padLeft, aZero );
    return os;
}

inline
OutputStream & operator << ( OutputStream & os, asAddress address)
{
    os << (char) '@';
    os.emitNumber( address, 16, 10, padLeft, aZero );
    return( os );
}

inline
OutputStream & operator << ( OutputStream & os, asOctal octal )
{
    os.emitNumber( octal, 8, 0, padLeft, aZero );
    return( os );
}

inline
OutputStream & operator << ( OutputStream & os, asBits bits)
{
    os.emitNumber( bits, 2, 0, padLeft, aZero );
    return( os );
}

inline 
void operator << ( OutputStream & os, Flush  fl) 
{ 
    os.flush();
}

#endif /* OutputStream_h */
