/*
 * This file is part of the Choices Operating System
 * Author: Yevgeniy Rabinovich <rabinvch@uiuc.edu>,
 *                             <yrabinov+uiuc@cs.uchicago.edu>
 * Using BSDInode.h as base for development
 * Copyright (c) 2004
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 */
/*
 * Ext2Inode.h - Ext2Inode class
 */

#ifndef Ext2Inode_h
#define Ext2Inode_h


#ifdef __GNUC__
#pragma interface
#endif

#include "endian_swap.h"
#include "Semaphore.h"
#include "UNIXInode.h"


const int EXT2_NDIR_BLOCKS = 12;
const int EXT2_IND_BLOCK = EXT2_NDIR_BLOCKS;
const int EXT2_DIND_BLOCK = (EXT2_IND_BLOCK + 1);
const int EXT2_TIND_BLOCK = (EXT2_DIND_BLOCK + 1);
const int EXT2_N_BLOCKS = (EXT2_TIND_BLOCK + 1);

// Should we use short, int, etc., or __u16, __u32?
// For now use short, int, etc., like Choices

class fs;
class cg;
class UNIXInodeIndirectInfo;

class Ext2_dinode {
    friend class Ext2Inode;
    friend class Ext2Container;

    unsigned short di_mode;   /*  0: File mode */
    unsigned short di_uid;    /*  2: Low 16 bits of Owner Uid */
    unsigned int   di_size;   /*  4: Size in bytes */
    unsigned int   di_atime;  /*  8: Access time */
    unsigned int   di_ctime;  /* 12: Creation time */
    unsigned int   di_mtime;  /* 16: Modification time */
    unsigned int   di_dtime;  /* 20: Deletion Time */
    unsigned short di_gid;    /* 24: Low 16 bits of Group Id */
    unsigned short di_nlink;  /* 26: Links count */
    unsigned int   di_blocks; /* 28: Blocks count */
    unsigned int   di_flags;  /* 32: File flags */

    // This is a actually a union of OS-dependant structures in Linux's ext2_fs.h
    union {
        struct {
            unsigned int l_i_reserved;
        } linux1;
    } osd1;

    unsigned int   di_db[EXT2_N_BLOCKS];/* Pointers to blocks: I think in ext2fs this is both direct and indirect blocks: The first 12 are direct? */
    unsigned int   di_generation; /* File version (for NFS) : di_version in BSD? */
    unsigned int   di_file_acl;   /* File ACL */
    unsigned int   di_dir_acl;    /* Directory ACL */
    unsigned int   di_faddr;      /* fragment address */

    union {
        struct {
            unsigned char l_i_frag;   /* Fragment number */
            unsigned char l_i_fsize;  /* Fragment size */
            unsigned short i_pad1;
            unsigned short l_i_uid_high;  /* these 2 fields    */
            unsigned short l_i_gid_high;  /* were reserved2[0] */
            unsigned int l_i_reserved2;
        } linux2;
    } osd2;
};

extern ClassStar Ext2InodeClass;

class Ext2Inode: public UNIXInode {
 protected:
    /*
     *    Structures and functions that almost could be inherited
     */
    struct _InodeData {
        UNIXInodeIndirectInfo ** _inodeIndirectInfo;
        PersistentMemoryObject * _partition;
        unsigned _inodeBlock;
        unsigned _inodeOffset;
        unsigned _di_spare3;
        unsigned _di_flags;
        unsigned _di_size;
        short    _di_mode;
        short    _di_nlink;
    };

    Ext2_dinode *_diskInode;
    _InodeData *_inodeData;

    void  initDinode();
    void  getDinode();
    void  releaseDinode();

    int   allocBlock();
    void  freeBlock( int physicalBlock );
    void  freeFragments( int physicalBlock, int count );
    int   blockPref( int * buf, int start, int startBlock );
    int   allocBlocks( int * buf, int start, int count, int pref );

    int   accessMode();
    int   userId();
    int   groupId();

    /*
     *    The following variables and functions handle fragmentation
     *    (Eugene: is this still needed?)
     */
    int   numFragments();

    /*
     *    nearBlock helps with BSDs block allocation optimizations
     */
    // int    nearBlock;
    ~Ext2Inode();

    Class * supports( Class * );

    unsigned   basicSetNumberOfBytes( unsigned size );
    unsigned   basicSetNumberOfUnits( unsigned numberOfUnits );

    int     basicRead( unsigned start, int count, 
                       char * addr, unsigned nbytes, IOAsyncClosure * c );
    int     basicWrite( unsigned start, unsigned count,
                        char * addr, unsigned nbytes, IOAsyncClosure * c );
    int     basicWriteThrough( unsigned start, unsigned count,
                               char * addr, unsigned nbytes, 
                               IOAsyncClosure * c );    
    int rawRead( unsigned startUnit, unsigned count, char * src, unsigned nbytes,
                 IOAsyncClosure * c );
    int rawWrite( unsigned startUnit, unsigned count, 
                  char * dest, unsigned n,
                  int writeThrough, IOAsyncClosure * c );
    void    updateInode( int finalSync );
    void    synchronizeInode();
    void    hintsChanged( unsigned oldHints );

 public:
    Class * classOf();
    Ext2Inode( unsigned int numberOfUnits,
               int log2UnitSize,
               int inumber,
               MemoryObjectContainer *,
               PersistentMemoryObject * source,
               int log2BlockSize,
               Ext2_dinode *, UNIXInodeIndirectInfo ** );

    unsigned numberOfBytes();
    int links( int delta );
    int setMode( int mode );
    int setUserId( int uid );
    int setGroupId( int gid );
    int isSymbolicLink();
    int isContainer();
    int isDictionary();
    void    info( UNIXFileInfo * );
    void    getAttributes( RFSAttributes * );
    void    setAttributes( RFSAttributes * );
};

inline int
Ext2Inode::numFragments()
{
    int ret = _inodeData->_di_size;
    if( ret > 0 ) {
        unsigned  units = ( ret + unitSize() - 1 ) >> log2UnitSize();
        if( ( units >> log2BlockSize ) >= singleIndirect ) ret = 0;
        else ret = units & blockMask();
    }
    return( ret );
}

#endif  /*Ext2Inode_h*/
