/*
** LogContainer.h
**
** header file for log based file system container routines
*/
#ifndef LogContainer_h
#define LogContainer_h


#ifdef __GNUC__
#pragma interface
#endif

#include "UNIXContainer.h"

#define FlMAGIC       0x10651065      /* sn_magic number                      */
const int FlMAGICOFFSET = 512; /* offset to magic number in info block */

#define LOGINFO  1   /* physical block containing file system information  */
#define SNFIRST  2   /* physical block containing first super node         */
#define SNFREQ  32   /* physical blocks between super nodes                */
#define SN_DIR   3   /* number of direct pointers in super node            */

/*
** class fsInfoNode
**
** this class holds file system parameters
** most important, it has field for the physical block of the current supernode
**
** upon mounting, the info node is read.  if the supernode block is set
** to 0, then that means the machine died ungracefully, and each super
** node on the system will have to be checked to see which one is most
** recent.  if it is not 0, and is a valid superNode block, then this is
** considered current, and the field is set to 0.
**
** upon umounting, the info node's field for current supernode is updated.
*/
class fsInfoNode {
   friend class LogContainer;

   int log_magic;      /* magic number                                */
   int log_fsize;      /* size in blocks of entire volume             */
   int log_snode;      /* physical block containing current supernode */
   char log_fill[500]; /* extra, to fill 512 byte block               */
};

/*
** class superNode:
**
** this is the structure of the super nodes as they are stored on disk,
**  similar to a regular inode.
*/
class superNode {
   friend class LogContainer;

   int  sn_addr[SN_DIR+3]; /* pointers to inodes, last three are indirect */
   char sn_fill[487];      /* extra, to fill 512 byte block               */
   char sn_counter;        /* which is most recent supernode              */
};

extern ProxiableObject * LogContainerConstructor( ProxiableObject * );
extern PersistentClassStar LogContainerClass;
class LogContainer : public UNIXContainer {
protected:
   fsInfoNode   *fs_infoblk;  /* information block                        */
   int          lg_last;      /* index to head of the log                 */
   int          writingSNTables;  /* SN is not to be written if flag true */
   char         *lg_free_cur; /* bitmap of current block usage            */
   char         *lg_free_old; /* bitmap of block usage at last super node */

   superNode    *SNode;       /* current super node                       */
   unsigned int *SNsingleBuffer;
   unsigned int *SNdoubleBuffer;
   unsigned int **SNdoubleBufferMem;
   unsigned int *SNtripleBuffer;
   unsigned int **SNtripleBufferMemDouble;
   unsigned int ***SNtripleBufferMemSingle;

   void resetBit ( int oldBlock );
   void setBit ( int newBlock );
   void copyBitmap ();
   char getBit ( int block, char * bitmap );
   void initBitmap ();
   void SNinitDirect(int physicalBlock, int inumber);
   void SNinitSingle(int physicalBlock, int inumber, unsigned int * &buffer);
   void SNinitDouble(int physicalBlock, int inumber, unsigned int * &buffer,
           unsigned int ** &mbuffer);
   void SNinitTriple(int physicalBlock, int inumber, unsigned int * &buffer,
           unsigned int ** &mbuffer, unsigned int *** &embuffer);
   void initPointer ( int physicalBlock, int level);
   void writeSNode ( int physicalBlock );
   void writeSNTables ( int inode );
   int  getSNode ( int inumber );
   void putSNode ( int inumber, int value );

   int  singleIndirect;
   int  doubleIndirect;
   int  tripleIndirect;
   int  ppblk;
   int  indexShift();
   int  calcSingleIndex( int block );
   int  calcDoubleIndex( int block );
   int  calcTripleIndex( int block );

   int  numFragments( int size );
   int  mapInumber( int inum );
   int  firstInumber( int inum );
   int  getUnusedInode( int notUsedInThisSubClass );
   void putUnusedInode( int inumber );
   int  nextUnusedInode;

   FileObject * basicOpen( int idNumber, int & status, int flags );
   FileObject * basicCreate( int idNumber, Class *, int & status, int flags );
   void basicClose( FileObject * mo );
   void basicSynchronize();

   ~LogContainer();
public:
   Class * classOf();

   LogContainer( PersistentMemoryObject * source );

   int  allocate( int & blocks, int nearPhysicalUnit, int units );
   void	free( int & blocks, int physicalUnit, int inits );
   void updateSNode (int inumber, int physicalBlock);
};

#include "FileObject.h"

inline int
LogContainer::numFragments( int size )
{
   int fragmentSize = _source->unitSize() << _log2FragmentToSector;
   return( ( size + fragmentSize - 1 ) / fragmentSize ); 	
}

#endif /*LogContainer_h*/
