/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Author: Panos Kougiouris ( kougiour@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MSDOSContainer.h - MSDOS Container Class
 */

#ifndef MSDOSContainer_h
#define MSDOSContainer_h


#ifdef __GNUC__
#pragma interface
#endif

#include "MemoryObjectContainer.h"

class MSDOSFile;
class MSDOSFAT;
class icentry;

extern ProxiableObject * MSDOSContainerConstructor( ProxiableObject * );
extern PersistentClassStar MSDOSContainerClass;

/* A class to represent an MSDOS file system. We support MSDOS file systems
 * with a small exceptions. The file system should not have files that
 * do not occupy disk space. The reason has to do with the fundamental
 * assumption of the framework that every stored object has a unique
 * number associated with it. 
 *    Since MSDOS does not have unique numbers
 * we assign to each file or subdirectory a number according to the
 * first cluster on the disk they occupy. 
 *     Cluster 0 and 1 do not exist however we use the idnumber 1
 * to refer to the root directory.
 *    Idnumber 2,3,4,... refer to the files or subdirectories starting
 * at cluster number 2,3,4 etc. Note that under this scheme if a file
 * occupies more than one clusters it makes unavailable more
 * than one inodes. Of course you understand that under this design
 * empty files would have been a big pain. So we decided to get rid of them.
 *   We remind that the cluster number 2 is the first cluster in the
 * Data Space of an MSDOS file system.  We also remind that a DOS filesystem 
 * consists of the four following  spaces: BOOT sector, File Allocator 
 * Tables, Root directory, Data space.
 */

class MSDOSContainer : public MemoryObjectContainer {
public:
	Class * classOf();

	/* DESCRIPTION: It builds an MSDOSConatiner from an MSDOS
	 *		file system supposed to be on the source.
	 *		Information from the boot block is get, a FATAllocator
	 *		is built, the inCore table is initialized and
	 *		the dentry table entries are filled by a recursive
	 *		procedure.
	 * INPUT:	1] source is where the MSDOS file system resides.
	 */
	MSDOSContainer( PersistentMemoryObject * source );

protected:
	/* DESCRIPTION: A pointer to the FAT.
	 * INITIALIZED: By the constructor.
	 */
	MSDOSFAT *	_fat;

	/* DESCRIPTION: A table of incore entries for each MSDOS entry.
	 *	        Entry 0 is not used.
	 * INITIALIZED: From the constructor.
	 */
	icentry * _inCoreEntries;

	/* DESCRIPTION: A table of offsets. Each is the offset of a file or
         * subdirectory from the beginning of the disk. Entry 0 is not used.
	 * INITIALIZED: From the constructor
	 */
	int *	_dentryOffsets;
	int	_log2ClusterToSector;
	int	_clusterSize;
	int	_unitSize;
	int	_inodesPerBlock;
	int	_dinodeSize;
	int	_inodeBlockSize;
	int	rootOffset;
	int	rootStartSector;
	int	rootSectors;
	int	dataStartSector;
	int	fatCopies;
	int	rootEntries;
	int	totalSectors;
	int	sectorsPerFat;
	int	dataSectors;
	int	numberOfClusters;
	int	fatMod;
	int	format12;

	/* DESCRIPTION: A method that returns the id number of the root
	 * 		Enrty.
	 */
	int 	basicRootId();

	/* DESCRIPTION: A method tro read the directroy entry of 
	 *		file or subdirectory in core.
	 *
	 * OUTPUT:	0 if everything is ok.
	 */
	int	readDinode( int inumber );
	int	writeDinode( int inumber );

	/* DESCRIPTION: Each file or subdirectory has an associated
	 *	directory entry. This directory entry resides in a 
	 *	unit of the source memory object. This method returns
	 *	the offset (in units from the start of the memory objects)
	 *	of this unit.
	 * INPUT:	1]inumber: The idnumber of the filr or subdirectory.
	 */
	int	mapInumber( int inumber );

	/* DESCRIPTION: Each file or subdirectory has an associated
	 *	directory entry. This directory entry resides in a 
	 *	unit of the source memory object. This method returns
	 *	the offset (in bytes from the start of the unit)
	 *	of this directory entry.
	 * INPUT:	1]inumber: The idnumber of the filr or subdirectory.
	 */
	int	firstInumber( int inumber );
	int	numClusters( int size );

	FileObject * basicOpen( int idNumber, int & status, int flags );
	FileObject * basicCreate( int idNumber, Class *, int &, int flags );
	void	basicClose( FileObject * );
 	void	basicSynchronize();
	~MSDOSContainer();

private:
	void	loadEntries( int cluster );
};

const int DentryOffsetUnknown = 0;
const int DentryOffsetFree = 1;
const int DentryOffsetInvalid = 2;

inline int
MSDOSContainer::numClusters( int size )
{
	int fragmentSize = _unitSize << _log2ClusterToSector;
	return( ( size + fragmentSize - 1 ) / fragmentSize ); 	
}

inline int
MSDOSContainer::basicRootId(){
	return(1);
}

#endif	/*MSDOSContainer_h*/
