/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Contributing Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MemoryObjectContainer.h 
 */

#ifndef MemoryObjectContainer_h
#define MemoryObjectContainer_h


#ifdef __GNUC__
#pragma interface
#endif

#include "BlockAllocatorStar.h"
#include "ContainerIndex.h"
#include "FileObject.h"
#include "FileObjectStar.h"
#include "FileSystemLocks.h"
#include "Lock.h"
#include "MemoryObjectDictionaryStar.h"
#include "PersistentMemoryObjectStar.h"
#include "PersistentObject.h"
#include "RangeLocks.h"
#include "Semaphore.h"
#include "SemaphoreStar.h"

// #define MAXACTIVEFILES

#ifndef NESTEDLOCK
#include "Semaphore.h"
#endif

extern PersistentClassStar MemoryObjectContainerClass;

class MemoryObjectContainer : public PersistentObject {
/*
 * A MemoryObjectContainer manages the contents of a PersistentMemoryObject as 
 * a collection of component FileObjects.
 * The MemoryObjectContainer class is an abstract superclass for all container
 * classes in the file system.  It implements public methods in terms of
 * protected methods, which subclasses must implement.
 * This class provides a table of all currently open (instantiated) FileObjects.
 */
public:
    /* Method classOf: return the Class of this object. */
    Class * classOf();

    /* Method MemoryObjectContainer: constructor. */
    proxiable MemoryObjectContainer( PersistentMemoryObject * source );

    /* Method open: return the FileObject corresponding to the give id. */
    proxiable virtual FileObjectRef open( int id, int & status, 
					  int flags = 0 );

    /* Method create: return a newly created FileObject
     *                that supports the given Class.
     */
    proxiable virtual FileObjectRef create( int id, Class *, int & sts, 
					    int flags = 0 );

    /* Method close: inform the container that the given FileObject
     *               is no longer being used.
     */
    proxiable virtual void close( FileObject * fo );

    /* Method allocate: allocate the requested number of units to the 
     *                  requesting FileObject from the source MemoryObject.
     */
    proxiable virtual int  allocate( int & blocks, int nearPhysicalUnit, int units );

    /* Method free: free the given number of units
     *              from the requesting FileObject.
     */
    proxiable virtual void free( int & blocks, int physicalUnit, int units );

    /* Method synhcronize: flush cached data structures
     *                     to source MemoryObject.
     */
    proxiable virtual void synchronize();

    /* Method synchronize: flush only the data structure associated with
     *	                   a particular file
     */
    proxiable virtual void synchronize( int idNumber );
  
    /* Method rootDictionary: return the root MemoryObjectDictionary
     *                        for this container.
     */
    proxiable virtual MemoryObjectDictionaryRef	rootDictionary();

    /* Method size: returns the max number of file objects
     */
    proxiable virtual int size();

    /* Method rootIdNumber: returns the idNumber of root object
     */
    proxiable virtual int rootIdNumber();

    /*
     * Method acquire and release: mutex for open/create/close
     */
    proxiable virtual void acquire();
    proxiable virtual void release();
    proxiable virtual void acquire( int idNumber );
    proxiable virtual void release( int idNumber );

protected:
    BlockAllocatorStar	_allocator;
    FileObject **	_activeFileObjects; // only valid for basicSynchronize
    int			_numberOfFileObjects;
    int			_modified;
    int		       	_openFileObjects;
#ifdef NESTEDLOCK
    NestedLock          _mutex;
#else
    Semaphore           _mutex;
#endif
    MRSWLock            _syncMutex;
    NormalRangeLock     _fileObjectLock;

#ifdef MAXACTIVEFILES
    static Semaphore	_maxActiveSem;
#endif
    
    virtual bool IsIdNumberInRange( int idNumber );

    /* Method parseFlags: parse flags passed in 
     */
    proxiable virtual void parseFlags( FileObject * fo, int flags );

    /* Method basicOpen: each subclass should return the FileObject
     *                   corresponding to the given idNumber. */
    proxiable virtual FileObject * basicOpen( int idNumber, int & status,
					      int flags );

    /* Method basicCreate: each subclass should create and return a FileObject
     *                     that supports the given class. */
    proxiable virtual FileObject * basicCreate( int idNumber, Class *, 
					        int & status, int flags );

    /* Method basicClose: each subclass should handle the returning of
     *                    FileObject control information to the source
     *                    MemoryObject.
     */
    proxiable virtual void basicClose( FileObject * mo );

    /* Method basicRootId: each subclass should return the idNumber that
     *                     corresponds to the contianer's root dictionary.
     */
    proxiable virtual int	 basicRootId();

    /* Method basicSynchronize: each subclass should flush specific 
     *                          cached information to the source MemoryObject.
     */
    proxiable virtual void basicSynchronize();

    /* Method basicSynchronize: each subclass should flush specific 
     *                          cached information to the source MemoryObject.
     */
    proxiable virtual void basicSynchronize( int idNumber );

    /* Method initializeArrays: initialize the table of open FileObjects
     *                          and set numberOfFileObjects variable. */
    proxiable virtual void initializeArrays( int containerSize );

    /* Method ~MemoryObjectContainer: destructor. */
    proxiable ~MemoryObjectContainer();

    /* Method noRemainingReferences: actions taken when no longer used. */
    proxiable virtual void noRemainingReferences();

    FileObject * lookup(int index) 
	{ return TheContainerIndex->lookup(this, index); }
    void bind(int index, FileObject * f) 
	{ TheContainerIndex->bind(this, index, f); }
    void unbind(int index)
	{ TheContainerIndex->unbind(this, index); }
    void setCreatingDictionary(FileObject * fo, MemoryObjectDictionary * d)
	{ fo->setCreatingDictionary(d); } 
};

#endif /*MemoryObjectContainer_h*/
