/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Contributing Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MemoryObjectDictionary.h - abstract superclass for directories
 */

#ifndef MemoryObjectDictionary_h
#define MemoryObjectDictionary_h


#ifdef __GNUC__
#pragma interface
#endif

#include "PersistentObject.h"
#include "PersistentMemoryObjectStar.h"
#include "MemoryObjectContainerStar.h"
#include "MemoryObjectViewStar.h"
#include "DomainStar.h"

class MutexLock;

extern PersistentClassStar MemoryObjectDictionaryClass;

class MemoryObjectDictionary : public PersistentObject {
/*
 * MemoryObjectDictionary is an abstract superclass for all dictionary
 * (or directory) classes.  It implements public methods in terms of
 * protected methods, which subclasses must implement.
 */

 protected:
    /*
     * The container that stores the objects that this dictionary references.
     */
    MemoryObjectContainerStar   _container;

    /*
     * All public member functions contain critical sections.
     * This semaphore provides the necessary mutual exclusions.
     */
    MutexLock *     _mutex;
    
    MemoryObjectViewStar _view;
    char *      _buffer;
    int         _log2ViewSize;
    int         _size;
    int         _units;
    int         _unitSize;
    int         _unitMask;
    int         _dirty;

    /*
     *  State instance variables:
     *      _idNumberOfKey  container index number associated
     *              with the key just found
     *      _offset     memory offset within dictionary buffer
     *              of a key that has been found
     *      _emptyOffset    memory offset within or at the end of
     *              the dictionary buffer that has space
     *              to add a key-value pair
     */
    int         _idNumberOfKey;
    int         _offset;
    int         _emptyOffset;

    /*
     * Flag to indicate whether this dictionary is still valid,
     * as opposed to having just been deleted. 
     * It is used to prevent a process that was
     * waiting on the _mutex semaphore when the dictionary got deleted
     * from using the dictionary after it has been deleted.
     */
    int         _valid;

    /*
     *  The maximum string length that any key in this dictionary could have.
     */
    int         _maxKeyLength;

    /*
     * Side-effects: _offset = location of found key
     *               _emptyOffset = location of space to add a key-value pair
     *                              (-1 means no room).
     *               _idNumberOfKey = value associated with key
     */

    /* Method basicKeys: fill the given buffer with as many of this dictionarys
     *                   keys as possible, also answer the number of keys put
     *                   in the buffer.  Return status code (>= 0 if successful,
     *                   otherwise an error code)
     */
    virtual int basicKeys( char * buffer, int bufferSize, int & count );
    /*
     *      NOTE: This does all the work for the public keys() method.
     *      It does not check directory validity, permissions, and
     *      assumes a critical section has already been entered.
     */

    /* Method basicAssociations: fill the given buffer with as many of this
     *                   dictionarys associations as possible; also answer
     *                   the number of keys put in the buffer. Return status
     *                   code (>= 0 if successful, otherwise an error code).
     */
    virtual int basicAssociations( char * buffer, int bufferSize,
                                   int & count );
    virtual int basicAssociations( char * buffer, int bufferSize,
                                   int & count, int & offset );

    /* Method insertEntry: Update dictionary with the given key and value.
     *                     (key length is supplied as the second parameter).
     *                     Return status code for success or failure of writing
     *                     changes back to the underlying MemoryObject(_source).
     */
    virtual int insertEntry( char * key, int keyLength, int idNumber );
    /*
     * NOTE: findKey() must have just been called within same critical section!
     */

    /* Method entrySize: return the size of an entry given a key length. */
    virtual int entrySize( int keyLength );

    /* Method hint: generate a value to help a container decide which inumber
     *              to use for a newly created object.
     */
    virtual int hint();
    /*
     * (For SVID this is useless, for BSD this increases efficiency,
     *  for MSDOS this is absolutely necessary, i.e., this smells, looks,
     *  and quacks like a hack to me.)
     * NOTE: findKey() must have just been called within same critical section!
     */


    /* Method clearEntry: remove a key-value pair from a dictionary */
    virtual int clearEntry();
    /*
     * NOTE: findKey() must have just been called within same critical section!
     */

    /* Method writeChanges: update underlying PersistentMemoryObject (_source).
     *                      Return status code from write operation.
     */
    virtual int writeChanges();

    /*
     *  REMOVE THIS FUNCTION WHEN MEMORY-MAPPED DICTIONARIES ARE IMPLEMENTED.
     */

    /* Method empty: returns true if directory has no keys. */
    virtual int empty();

    /* Method ~MemoryObjectDictionary: desctuctor. */
    ~MemoryObjectDictionary();

    /* Method noRemainingReferences: actions taken when no longer used. */
    virtual void noRemainingReferences();

    /* Method reMap: reMap after change in source size */
    virtual void reMap();

 public:
    /* Method classOf: returns the class of this object. */
    Class * classOf();

    /* Method MemoryObjectDictionary: contructor. */
    proxiable MemoryObjectDictionary( PersistentMemoryObject * source );

    /* Method open: returns the MemoryObject corresponding to the given key,
     *              if the key is in the dictionary. */
    proxiable virtual PersistentMemoryObjectRef open( char * key, 
                                                      int & status,
                                                      int flags = 0 );

    /* Method create: returns the MemoryObject corresponding to the given key
     *                if the key is in the dictionary.
     *                Otherwise a new MemoryObject is created
     *                and the key is stored in the dictionary. */
    proxiable virtual PersistentMemoryObjectRef create( char * key, Class *, 
                                                        int & sts,
                                                        int flags = 0 );

    /* Method add: inserts the given key into the dictionary only if it is not
     *             already in the dictionary. */
    proxiable virtual int add( char * key, PersistentMemoryObject * );

    /* Method remove: deletes the given key from the dictionary. */
    proxiable virtual int remove( char * key, int * links = 0 );

    /* Method keys: returns the list of keys in the dictionary.
     *              Also returns the number of keys. */
    proxiable virtual int keys( char * buffer, int bufferSize, int & count );

    /* Method associations: returns the list of associations in the
     *                      dictionary. Also returns the number of
     *                      associations. */
    proxiable virtual int associations( char * buffer, int bufferSize,
                                        int & count );
    proxiable virtual int associations( char * buffer, int bufferSize,
                                        int & count, int & offset );

    /* Method keySize: returns the sum of the length of the keys
     *                 in the dictionary.  */
    proxiable virtual int keySize();

    /* Method findKey: search for given key (also look for empty space
     *                 if second parameter is non-zero).
     *                 Return 1 if key is found, otherwise return 0.
     */
    virtual int findKey( char * key, int neededSize );

    /* Method keySize: returns true iff the given MemoryObject
     *                 has the same container as the dictionary.  */
    proxiable virtual int sharesContainer( PersistentMemoryObject * );

    /* Method synchronize: sync changes to dictionary */
    proxiable virtual int synchronize();

};

#endif /*MemoryObjectDictionary_h*/
