/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * PersistentMemoryObject.h: PersistentMemoryObject class descriptions.
 */

#ifndef PersistentMemoryObject_h
#define PersistentMemoryObject_h

#ifdef __GNUC__
#pragma interface
#endif

#include "ClassStar.h"
#include "MemoryObject.h"
#include "MemoryObjectAsync.h"
#include "MemoryObjectStar.h"
#include "MemoryObjectContainerStar.h"
#include "MemoryObjectDictionaryStar.h"
#include "PersistentObjectStar.h"
#include "RangeLocks.h"
#include "Semaphore.h"
#include "ZombieInfo.h"

class PersistentObject;
class UNIXFileInfo;
class RFSAttributes;

/*
 * NOTICE: this class is common to the kernel and the filesystem ring
 *    and the filesystem ring uses objects of this class in its own ring
 *    and it uses proxies to them in the kernel. 
 * Every virtual function must be proxiable in order to allow the filesystem
 *   ring to use the same interface for proxies and real objects.
 * It would be much better if the PersistentMemoryObject class did not have
 *   to be copied.  That would require either the Disk class being moved
 *   out of the kernel or it not being a sub-class of PersistentMemoryObject.
 * - Dave Dykstra, 9/6/91
 */

extern ClassStar PersistentMemoryObjectClass;
class PersistentMemoryObject : public MemoryObject {
/*
 * A PersistentMemoryObject supports read-block/write-block access
 * to typed persistent data.  It also supports the instantiation of
 * PersistentObjects that belong to appropriate types.
 * All PersistentMemoryObjects are either Disks or Files.
 */
 protected:

    class BasicSynchronizeAsyncClosure : public IOAsyncClosure {
    public:
        int _state;
        IOAsyncClosure * _parent;
        PersistentMemoryObject * _pmo;
        static void func( BasicSynchronizeAsyncClosure * c );
        BasicSynchronizeAsyncClosure
        ( IOAsyncClosure * parent, PersistentMemoryObject * pmo )
        :   _state( 0 ), _parent( parent ), _pmo( pmo ) {_func = (AsyncFunc)&func; }
              
    };
    friend class BasicSynchronizeAsyncClosure;
    virtual void basicSynchronizeClosure( BasicSynchronizeAsyncClosure *c );

    int         _idNumber;
    MemoryObjectContainer * _container;
    unsigned        _version;
    MemoryObjectDictionary * _creatingDictionary;
    MutexLock       _mutex;
    int         _openFlags;
    unsigned        _zombieDelay;
    ZombieInfo      _zombieInfo;
    PersistentObject *  _persistentObject;
    PersistentObject *  _cachedPersistentObject;
    static Semaphore        _asAMutex;

    /*
     * First check to see if aClass is legal for high-level access to this 
     * PersistentMemoryObject.
     * Second, set the requires variable to the level of permission needed
     * for access to this object using the given class.
     * Third, return a concrete (sub)class of aClass that can be 
     * instantiated
     */

    proxiable virtual       ~PersistentMemoryObject();

    virtual void            noRemainingReferences();
    virtual int             accessMode();

    virtual Class *         supports( Class * aClass );

    virtual void            basicFlush();
    virtual void            basicSynchronize( IOAsyncClosure * c );
    virtual void            basicSynchronize( unsigned from, unsigned to,
                                              IOAsyncClosure * c );

    virtual unsigned        basicSetNumberOfBytes( unsigned );
    virtual unsigned        basicSetNumberOfRecords( unsigned );
    virtual unsigned        basicSetRecordSize( unsigned );

    virtual int         basicCopy( MemoryObject * dest );
    
    virtual void        updateFlagsAndHints( int flags );
    virtual void        hintsChanged( unsigned oldHint );
    virtual void        updateHints();

    /* These functions should only be called by MemoryObjectContainer 
     * and its subclasses */

    friend class MemoryObjectContainer;
    friend class MemoryObjectCachingContainer;

    virtual int             referenceCount();
    virtual int             parseFlags( int flags );
    virtual void            destroy();
    MemoryObjectCachingContainer * cachingContainer()
    { return (MemoryObjectCachingContainer *) _container; }


    friend class MemoryObjectDictionary;
    
    void        setCreatingDictionary( MemoryObjectDictionary * dict );
    void        synchronizeCreatingDictionary();

 public:
    Class * classOf();

    proxiable PersistentMemoryObject( unsigned int numberOfUnits,
                                      int log2UnitSize,
                                      int idNumber,
                                      MemoryObjectContainer * container );
    
    proxiable virtual void  unreference();
    proxiable virtual void  synchronize();
//  proxiable virtual int   fixedSize();

    proxiable virtual int   setNumberOfUnits( unsigned numberOfUnits );

    proxiable virtual unsigned   numberOfRecords();
    proxiable virtual int   setNumberOfRecords( unsigned numberOfRecords );
    proxiable virtual unsigned   recordSize();
    proxiable virtual int   setRecordSize( unsigned recordSize );

    proxiable virtual unsigned   numberOfBytes();
    proxiable virtual int   setNumberOfBytes( unsigned numberOfBytes );

    proxiable virtual void  info( UNIXFileInfo * );
    proxiable virtual int   isSymbolicLink();
    proxiable virtual int   isContainer();
    proxiable virtual int   isDictionary();
    proxiable virtual int   setUserId( int uid );
    proxiable virtual int   setGroupId( int gid );
    proxiable virtual int   setMode( int mode );
    proxiable virtual int   idNumber();
    proxiable virtual int   links( int delta = 0 );
    proxiable virtual MemoryObjectContainerRef  container();

    proxiable virtual int   access( int mask );
    proxiable virtual ProxiableObjectRef asA( Class *, int & status );
    proxiable virtual PersistentObjectRef persistentObject();

    proxiable virtual void  close( PersistentObject * so );

    proxiable virtual void  writeName( OutputStream * );
    proxiable virtual void  writeIdNumber( OutputStream * );

    proxiable virtual int   userId();
    proxiable virtual int   groupId();

    proxiable virtual int   alloc( unsigned startUnit, int count );
    proxiable virtual int   free( unsigned startUnit, int count );

    proxiable virtual void     setCacheHints( unsigned userHints );
    proxiable virtual unsigned getCacheHints();

    virtual void        setHints( unsigned mask );
    virtual void        clearHints( unsigned mask );

    virtual int     unmount();
    virtual int     notify( int );
    virtual int     isRemote();
    virtual void    getAttributes( RFSAttributes * );
    virtual void    setAttributes( RFSAttributes * );
    
    unsigned        version() { return _version; }
    unsigned        zombieDelay() { return _zombieDelay; }
    void            setZombieDelay( unsigned delay ) { _zombieDelay = delay; }
    MemoryObjectContainer * fastContainer() { return _container; }
    ZombieInfo *        zombieInfo() { return &_zombieInfo; }
};

#endif /*PersistentMemoryObject_h*/
