/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Contributing Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * UNIXInode.h - unix inode abstract class
 */

#ifndef UNIXInode_h
#define UNIXInode_h


#ifdef __GNUC__
#pragma interface
#endif

#include "endian_swap.h"
#include "FileObject.h"

/* These are used in BSDInode and Ext2Inode */
const int IC_FASTLINK = 1;
const int IC_HASHED_DIR = 2;

extern ClassStar UNIXInodeClass;

class UNIXInodeIndirect;
class UNIXInodeIndirectInfo;

enum MapUnitResult {
    MAP_UNKNOWN,
    MAP_REAL,
    MAP_HOLE,
    MAP_FULL
};
    
class UNIXInode : public FileObject {
 protected:
    UNIXInodeIndirect * indirect;

    int                 indexMask;
    char                indexShift;
    unsigned char       log2BlockSize;
    char                _myModifiedFlag;
    unsigned            singleIndirect;
    unsigned int        doubleIndirect;
    unsigned int        tripleIndirect;
    UNIXInodeIndirectInfo * singleBuffer;
    UNIXInodeIndirectInfo * doubleBuffer;
    UNIXInodeIndirectInfo * tripleBuffer;
    UNIXInodeIndirectInfo * extra2Buffer;
    UNIXInodeIndirectInfo * extra3Buffer;
    unsigned int *  direct;
    unsigned int *  singleIndex;
    unsigned int *  doubleIndex;
    unsigned int *  tripleIndex;

    int     nearBlock;
    int     fragmentsInLastBlock;

    int     _lowMark;   // Lowest written block
    int     _highMark;  // Highest written block

    int *   _modifiedFlag;

    bool    little_endian_fs;

    virtual void basicSynchronizeClosure( BasicSynchronizeAsyncClosure *c );

    virtual int basicMapUnit( int logicalBlock, int addMapping );
    virtual int mapUnit( int logicalBlock, int addMapping );
    int     basicMapReadAssist( int * startCheck, int checkCount,
                                int & remainingCount, 
                                int & startRun,
                                int & expectedPhysicalBlock,
                                MapUnitResult & status );
    MapUnitResult   basicMapUnitRead( int unit, int count, 
                                      int &startRun, int &runLength );
    virtual int blockPref( int * buf, int start, int startBlock );
    virtual int allocBlocks( int * buf, int start, int count, 
                             int pref );
    int     basicMapWriteAssist( int * buf, int start,
                                 int checkCount,
                                 int startBlock,
                                 int & remainingCount, 
                                 int & startRun,
                                 int & expectedPhysicalBlock,
                                 MapUnitResult & status,
                                 int & modified );
    MapUnitResult   basicMapUnitWrite( int unit, int count, 
                                       int &startRun, int &runLength );
    int     basicMapFreeIndirect( int * startBuf, 
                                  int * startCheck, int count ); 
    int     basicMapFreeAssist( int * startCheck, int checkCount,
                                int startBlock, int lastBlock,
                                int & remainingCount, 
                                MapUnitResult & status,
                                int & modified );
    MapUnitResult   basicMapUnitFree( int unit, int count );
    int     lastBlockNo();
    int     blockSize();
    int     blockMask();
    int     calcSingleIndex( int block );
    int     calcDoubleIndex( int block );
    int     calcTripleIndex( int block );

    void    setModified();

    // Placeholder functions used by logicalToPhysical and addBlock
    // for access to disk inode object
    // and to the inode system block allocation
    // These functions must be redefined in subclasses 
    virtual int     allocBlock();
    virtual void    freeBlock( int physicalBlock );
    virtual void    freeFragments( int physicalBlock, int count );

    // Redefining superclass methods

    virtual int transient();

    virtual void    basicFlush();
    virtual void    basicSynchronize( IOAsyncClosure * c );
    virtual void    basicSynchronize( unsigned from, unsigned to,
                                      IOAsyncClosure * c );

    virtual int     basicCopy( MemoryObject * dest );
    virtual unsigned    basicSetNumberOfUnits( unsigned units );

    virtual int     cacheWrite( unsigned start, unsigned count,
                                char * addr, unsigned nbytes, 
                                IOAsyncClosure * c );

    friend class    UNIXInodeCache;

    virtual int     rawRead( unsigned start, unsigned count,
                             char * addr, unsigned nbytes,
                             IOAsyncClosure * c );

    virtual int     rawWrite( unsigned start, unsigned count,
                              char * addr, unsigned nbytes, int writeThrough,
                              IOAsyncClosure * c );

    virtual int     parseFlags( int flags );

    virtual void    getDinode();
    virtual void    releaseDinode();

    friend class    UNIXContainer;
    virtual void    synchronizeInode();

 public:
    Class * classOf();

    UNIXInode( unsigned int numberOfUnits,
               int log2UnitSize,
               int inumber,
               MemoryObjectContainer *,
               PersistentMemoryObject * source,
               int log2BlockSize );
    ~UNIXInode();

    virtual int     fixedSize();

    virtual int     alloc( unsigned startUnit, int count );
    virtual int     free( unsigned startUnit, int count );

    static  void    setup( unsigned log2BytesPerBlock );
};

inline int
UNIXInode::lastBlockNo()
{
/*
  int ret = numberOfBytes();
  if( ret > 0 ) 
  ret = ( ( ret - 1 ) >> ( log2BlockSize + log2UnitSize() ) );
*/
    int ret = _numberOfUnits;
    if ( ret > 0 )
        ret = ( ret - 1 ) >> log2BlockSize;
    return( ret );
}

inline int
UNIXInode::blockSize()
{
    return( 1 << log2BlockSize );
}

inline int
UNIXInode::blockMask()
{
    return( blockSize() - 1 );
}

inline int
UNIXInode::calcSingleIndex( int block )
{
    return( block & indexMask );
}

inline int
UNIXInode::calcDoubleIndex( int block )
{
    return( calcSingleIndex( block >> indexShift ) );
}

inline int
UNIXInode::calcTripleIndex( int block )
{
    return( calcDoubleIndex( block >> indexShift ) );
}

#endif  /*UNIXInode_h*/
