/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef Gdb_h
#define Gdb_h


#ifdef __GNUC__
#pragma interface
#endif

#include "ProcessorContext.h"

#define GDB

class Breakpoint {
 public:
	Breakpoint():valid(false) {}
	virtual ~Breakpoint() {}
	
	bool isValid() { return valid; }

	virtual void set(unsigned addr, unsigned insn) {
		this->addr = addr;
		save = *(unsigned *)addr;
		*(unsigned *)addr = insn;
		valid = true;
	}
	
	virtual void remove() {
		*(unsigned *)addr = save;
		valid = false;
	}
	
	unsigned getAddr() { return addr; }

 private:
	bool valid;
	unsigned addr;
	unsigned type;
	unsigned save;
};

// Francis TODO how to manipulate max breakpoints
const unsigned MAX_BREAKPOINTS = 20;

class Gdb {
 public:
	Gdb() {};
	virtual ~Gdb() {};

	static void handle(int vector, InterruptContext *context) {
		if (instance) instance->basicHandle(vector, context);
	}
	void setInstance(Gdb *instance) { this->instance = instance; }

	static unsigned causedInterrupt(int vector, InterruptContext *context) {
		if (instance) 
			return(instance->basicCausedInterrupt(vector, context));
		else
			return 0;
	}

 protected:
	
	void GdbHandler(int vector, InterruptContext *context);
	virtual void basicHandle(int vector, InterruptContext *context) = 0;
	virtual unsigned basicCausedInterrupt(int vector, InterruptContext *context) = 0;

	virtual unsigned getPC() = 0;
	virtual unsigned getNPC() { return getPC() + 4; }
	virtual void setPC(unsigned address) = 0;
	virtual void setBreakInsn(unsigned address) = 0;

/*
 * The structure that the trap routine fills in with register values
 * at trap time and restores (possibly changed) register values from
 * after the trap.
 */
	unsigned *registers ;


/* 
 * Interface definitions for the machine/processor dependent
 * portions of GDB.  Include this in your machine dependent
 * stuff.
 */
	virtual char GetDebugChar() = 0;
	virtual void PutDebugChar(char) = 0;
	virtual int ComputeSignal(int) = 0;
	virtual void GdbKill() = 0; 		// Kill & reboot.

	Breakpoint breakpoints[MAX_BREAKPOINTS];
	unsigned GdbBreakInsn;		// Opcode for break
	void setBreakpoint(unsigned addr);
	Breakpoint* BreakpointAt(unsigned addr);
	void removeBreakpoint(unsigned addr);
	
  unsigned NUM_REGISTERS;
	unsigned NUMREGBYTES;
/*                                                          
 * BUFMAX is  the maximum number of characters in inbound/outbound buffers
 * at least NUMREGBYTES*2 are needed for register I/O packets
 */
	unsigned BUFMAX;

/*
 * These are called at entry/exit of handler.  You can use them to set
 * up things like info to tell page fault handlers that faults should be
 * handled in user mode, etc.
 */
	virtual void GdbMachineHandlerStart(int, InterruptContext *context) = 0; 
	virtual void GdbMachineHandlerEnd(int, InterruptContext *context) = 0;

/*
 * Provided per machine:
 * Tell if Memory Access at addr -> addr+len is ok
 * for the type of access requested
 * return 0 for success, non-zero for failure
 */
	static const unsigned GDB_READ = 0;
	static const unsigned GDB_WRITE = 1;
	virtual int GdbMachineMem_ok( unsigned acc_type, char *addr, unsigned len) = 0;


/*
 * Stuff that is internal to the GDB handler, but you might want to
 * use it somewhere else, like to send a packet or something strange.
 */
	virtual void GdbGetPacket(char *);
	virtual void GdbPutPacket(char *);

 private:
	int scanhex(char *p, unsigned *var);
	int scan2hex(char *p, unsigned *var1, unsigned *var2);
	static const char hexchars[];
	char hex(char ch);
	char* mem2hex(char *mem, char *buf, int count);
	char* hex2mem(char *buf, char *mem, int count);
	
	static Gdb *instance;
};


/* 
 * Use C linkage so we can use _GDBhandler in the assembly 
 * code without name munging.
 */
extern "C" {
	inline static void GdbHandler(int vector, InterruptContext *context) {
		Gdb::handle(vector, context);
	}
}


#endif /*Gdb_h*/
