/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Panos Kougiouris (kougiour@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/* The file has two parts:
 * 1. Class Disk - Memory Object representing a disk. 
 * 2. Some useful constants for the disk operations
 */

#ifndef Disk_h
#define	Disk_h


#ifdef __GNUC__
#pragma interface
#endif

#include "PersistentMemoryObject.h"
#include "SemaphoreStar.h"

class DiskScheduler;
class PhysicalMemoryChain;


/**********************************************************************/
/*    CLASS DISK                                                      */
/**********************************************************************/
/*
 * An abstract class representing each disk drive as a persistent
 * MemoryObject. The
 * disk is supposed to consist of a number of partitions. The disk
 * contains a member variable called _retries. It is the number of times
 * we 'll retry to read or write a block in case of failure.
 *
 *   The first block of the disk is 0 and exceeds up to the size of the 
 * PersistentMemoryObject.
 * 
 *   The Disk are behave from an external point of view as a special
 * kind of Monitor. It allows only one thread of control executing inside the
 * Disk.  
 * Since C++ does not supports monitors we have to fake the Disk class
 * by using a semaphore inside the class. The semphore is initialized
 * during the construction to 1 and all the public methods have the
 * form:
 *          P()
 *          <action>
 *          V()
 *
 *   We have made the assumption that the units allocated from the
 * MainStore---that is the physicalMemory pages---have size greater or
 * equal than that of a disk block.
 */

/* The meatobject DiskClass represents the Disk Class. It's declaration
 * is built automatically. 
 */
extern ClassStar DiskClass;	

class Disk : public PersistentMemoryObject {
/*********************************************************************
 *           PUBLIC DISK METHODS				     *
 *********************************************************************/ 
public:
	/* DESCR: 	The class constructor. The constructor constucts
	 *		an "empty" persistent memory Object. It's
	 *		obligation of the the initialization routines
	 * 		to initialize the PersistentMemoryObject
	 *	  	inherited attributes.
         *		  WARNING: Before calling the constructor make
 	 *		sure that the new operator does work.
	 * INPUT:	The number of times we should retry to read
	 *		or write a block before reporting an error.
	 *	 	It is kept in the member variable _retries.
	 */
	Disk( short retries = 5,  DiskScheduler * diskScheduler = 0);

	~Disk();

	DiskScheduler * diskScheduler();

	Class * classOf();

	proxiable virtual int isContainer();

	/* DESCR:	A  virtual function which is supposed to return
	 *		the number of partitions on the disk. According to
	 * 		the model each disk is partitioned in a number of
	 *		contiguous collections of blocks. The default
         * 		behavior is that a disk has only one partition.
	 * OUTPUT:	The number of partitions on the disk.
	 */
	proxiable virtual int partitions();

	/* DESCR:	A  virtual function which is returns the start
	 *		and size of a partition on the disk.
	 * INPUT:	1] The partition id number, less than partitions()
	 *		2] reference for to return start of partition
	 *		3] reference to return number of units in partition
	 * OUTPUT:	-1 if bad partition id, 0 if successful.
	 */
	proxiable virtual int locatePartition( int idNumber, int &start,
					int &units );

/*********************************************************************
 *           PROTECTED DISK MEMBERS		 		     *
 *********************************************************************/ 
protected:
        /* DESCR: A function required from the PersistentMemoryObject
         *        superclass. As a persistent
         *        memory Object Disk should be able to answer if it supports
         *        a class, that is if it can be transformed to an object
         *        of this Class. The default behavior is to support
	 *	  a DiskContainer class.
         * OUTPUT: A subclass of the class superclass if any
         *         0 otherwise
         */
        proxiable virtual Class * 
	supports( Class * superclass);

	/* DESCR:	A method to write a number of blocks on the disk.
	 * INPUT:	1] The number of block to start writing.
 	 *		2] The number of blocks to write
	 *		3] The virtual address of the buffer.
	 *		4] The number of bytes in the buffer.
	 * OUTPUT:	The number of blocks actually written. The blocks
	 *		 written  may be less than the ones requested because
	 *		 of some hardware or sofware problem.
	 */
	proxiable virtual 
	int rawWrite( unsigned start, unsigned count, char * addr, unsigned nbytes, 
		      int writeThrough, IOAsyncClosure * c );

	/* DESCR:	A method to read a number of blocks from the disk.
	 * INPUT:	1] The number of block to start reading.
 	 *		2] The number of blocks to read
	 *		3] The virtual address of the buffer.
	 *		4] The number of bytes in the buffer.
	 * OUTPUT:	The number of blocks actually read. The blocks read
	 * 		may be less than the ones requested because of some
	 *		hardware or sofware problem.
	 */
	proxiable virtual
	int rawRead( unsigned start, unsigned count, char * addr, unsigned nbytes,
		     IOAsyncClosure * c );

	/* DESCRIPTION:	A virtual method that reads or writes a number
	 *		of blocks of data. The method iterates through
	 *		the PhysicalMemoryChain and uses the basicDoio.
	 *		The method also retries _retries number of
	 *		times before aborting when it tries to read a single
	 *		block. 
	 *		   Subclasses can redefine this method if
	 *		they want to gain in efficiency, if for example
	 *		they want to fully exploit the hardware.
	 *		   This method can be sure that only one thread of
	 *		control will ever be allowd inside this method.
	 * INPUT:	1] The number of block to start.
 	 *		2] The number of blocks to  read or write.
	 *		3] The virtual address of the buffer.
	 *		4] The number of bytes in the buffer.
	 *		5] DiskReadOperation mening read
			   DiskWriteOperation meaning write
	 * OUTPUT:	number of blocks read.
	 */
	friend class DiskScheduler;
	friend class OneAtATimeDiskScheduler;

	virtual int 
	doio( unsigned int start, int count, char * addr, int nbytes, int op );

	/* DESCR: A  pure  virtual function that does I/O of a single
	 *        block without retries. Subclasses are assumed to
	 *	  provide this method. The implementor of these 
	 *	  methods can be sure that only one thread of control
	 *        will ever allowed inside this method.
	 * INPUT:	1] The number of block to read or write.
	 *		2] A physical address representing where the 
	 *		   block should be transferred to. It is assumed
	 *		   that enough space to read or write a disk
	 *		   block exists there.
	 *		3] DiskReadOperation meaning read
	 *		   DiskWriteOperation meaning write
	 * OUTPUT:	DiskOperSucceded or DiskOperFailed
	 * 
	 */
	virtual int 
	basicDoio( unsigned int start, char * buf, int oper ) = 0;

	/* DESCR:	A function to initialize the Disk. Initialization
	 *		of all the variables of the class and the
	 *		superclass should have been done after the
	 *		call to this method. Here is the place where
	 *		a hardware reset should take place. 
	 * OUTPUT:	DiskOperSucceded or DiskOperFailed
	 */
	virtual 
	int reset() = 0;

/*********************************************************************
 *           PRIVATE DISK MEMBERS		 		     *
 *********************************************************************/ 
private:
	/* A member variable that knows how many times we should retry 
	 * to read or write a block. The variable is set during the
	 * object construction.
	 */
	short _retries;

	DiskScheduler * _diskScheduler;

	/* DESCR: A method that does I/O of a single block using
	 *	  basic I/O but retries _retries times before 
	 *        aborting. The method is called form doio.
	 * INPUT:	1] The number of block to read or write.
	 *		2] A physical address representing where the 
	 *		   block should be transferred to.
	 *		3] DiskReadOperation meaning read
	 *		   DiskWriteOperation meaning write
	 * OUTPUT:	DiskOperSucceded or DiskOperFailed
	 * 
	 */
	int 
	doioWithRetries( unsigned int start, char * buf, short oper);


};


/**********************************************************************/
/*    USEFUL CONSTANTS                                                */
/**********************************************************************/

/* Some useful constants now */
#define DiskReadOperation 	0
#define DiskWriteOperation 	1

/* Notice that methods that return a number of blocks when they
 * fail they usually return the value of 0 which matches the DiskOperFailed
 */
#define DiskOperSucceded	1
#define DiskOperFailed		!DiskOperSucceded

#endif /*Disk_h*/
