/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Roland G Ouellette
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/* This file has all sorts of stuff to do tty drivers.  Actually some
 * of it might even be general enough to be used for any sort of
 * device driver.  In doing I/O, there are generally two sorts of ways
 * that the user might want to program.
 * 
 * 1: The user issues a command & then just travels on her merry way.
 * Later when the I/O is complete, the OS changes some state in her
 * process to let her know that the I/O is done.  This state might be
 * an event flag, an IO completion block, or the running thread of
 * program code (an AST routine).  [the normal vms qio stuff...]
 *
 * 2: The user issues the command & waits for completion.  [this is
 * like a qio$w in vms... ]  This is the normal unix thing to do.
 *
 * For case 1, you can set up a conduit in the right sort of way:
 *
 * A) If you want AST style I/Os, the conduit has a pointer to the
 * procedure which gets called when the I/O is done; the OS will
 * magically swap the user's thread.  The user will REI back to the
 * right place when the AST is done.
 *
 * B) If you want event flag style completion, you just set the flag.
 * The user can use the flag to wait if necessary.
 *
 * C) I/O_blocks are kind of like event flags except they are cheeper
 * because they don't interlock.  They cannot however be waited upon.
 * Ususally each io gets its own io block, where with event flags one
 * gets shared in the whole process.  Event flags also offer the
 * possibility of putting time limits on your i/o's.
 *
 * For case 2:
 *
 * This whole mess could be emulated at the cost of the added over
 * head with any of the above schemes.  It is however a lot easier to
 * program and think about differently...
 *
 *  1. Incoming messages from the keyboard get inserted into the
 *     bottom of the driver.  They are simply queued up.
 *  2. The user at her leisure pulls whatever she likes out of the top
 *     of the driver.  Processing of input is delayed until she asks.
 *     This way she can change the mode of the terminal & the input
 *     will get the right kind of processing.
 *  3. Output is put into the top of the driver & it is forced out to
 *     the screen end of the I/O conduit we hook up to.
 *  4. As an optimization there might be a separate thread which
 *     actually does the output, but right now that's too hard for me
 *     to think about.  [This is kind of the inverse of the
 *     asynchronous case above...  You queue the IO & later you can
 *     check on its status.]
 */

#ifndef  TtyBuffer_h
#define  TtyBuffer_h


#ifdef __GNUC__
#pragma interface
#endif

class ConduitMessage;
class SemaphoreRegulatedMessageFIFO;
class DumbConduit;
#include "ProxiableObject.h"
#include "ConduitTypes.h"
#include "Lock.h"
#include "ClassStar.h"

typedef int status;		// for now...  Should be an enum or a class
				// let positive numbers be success
				// (num of chars sent or recieved).
				// let negative numbers be errors...
				// This may need a union typedef...

extern ClassStar UserTtyBufferClass;
extern ClassStar TtyBufferClass;
extern ClassStar TtyMessageBufferClass;


class UserTtyBuffer : public ProxiableObject {
/* This is the object which we give users proxy to...
 * The two methods will be implemented by the subclass TtyBuffer.
 */
public:
  /* Method classOf: return a pointer to the Class */
  Class * classOf();

  /* Method read: read a bunch of octets.  Status is the number of
   * chars read if positive; if negative it is an error code.
   */
  virtual status read(octet *s, int l);

  /* Method write: Write a bunch of octets.  Status will be positive
   * on success.
   */
  virtual status write(octet *s, int l);
};



class TtyBuffer : public UserTtyBuffer {
/* For now I'm just going to implement the second paradigm.
 *
 * I'm going to implement a low level buffer widget.  It talks to
 * JohnnyZ's conduits on the bottom & hands characters to the user on
 * the top end.  On top of this we will build the filters which belong
 * there.  This thing just gives RAW characters...  Nothing else...
 *
 * I don't want to even think about optimizing for multiple
 * character messages here.  Since that's such a weird case for
 * terminals, I think that the performance hit will be OK.  In any
 * case we shouldn't optimize code until we discover that we need to.
 * My thoughts are that if you need to do wicked fast I/O, you should
 * really be using the schemes from paradigm one...
 */
protected:
  SemaphoreRegulatedMessageFIFO *rawInput;
  DumbConduit *myConduit;
  Lock lock;		// This is for the next variable...
  int chars_left_input;		// This keeps track of the number of
				// unprocessed characters in incoming
				// messages so that we don't use up
				// all of the kernel heap space.
  // The previous two variables could be also implemented in ring
  // buffer fashion, but that's hard to think about.  By using the
  // semaphore the stuff below can also use the scheduler to avoid
  // busy waiting or dealing with the scheduler code ourself...
  ConduitMessage *current_message;
  octet *current_char;		// pointer into data...
  int chars_left_current_msg;
  friend class DumbConduit;
  /* Method next_message: dispose of the current message & start
   * unpacking the next one
   */
  void next_message();
public:
  /* Method classOf: return a pointer to the Class */
  Class * classOf();

  /* Method TtyBuffer: constructor */
  TtyBuffer(int max_msg, int max_chars);

  /* Method ~TtyBuffer: destructor */
  ~TtyBuffer();

  /* Method DumbConduit: give a pointer to our input port */
  DumbConduit *conduit();

  /* Method read: read a bunch of octets.  Status is the number of
   * chars read if positive; if negative it is an error code.
   */
  virtual status read(octet *s, int l);	// These two functions are what the

  /* Method write: Write a bunch of octets.  Status will be positive
   * on success.
   */
  status write(octet *s, int l); // user will call...

};

class TtyMessageBuffer: public TtyBuffer
{
  TtyMessageBuffer(int max_msg, int max_chars) : TtyBuffer(max_msg, max_chars)
  {}

  virtual status read( octet *s, int l);
public:
    Class * classOf();
};

inline DumbConduit *
TtyBuffer::conduit() {return(myConduit);}


inline octet *
move_bytes( octet * s1, octet * s2, int n )
{
  for ( ; n > 0 ; n-- )
    *s1++ = *s2++;
  return s1;
}
#endif  TtyBuffer_h
