/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Instrument.h: basic instruments for use anywhere
 *
 */

#ifndef Instrument_h
#define Instrument_h


#ifdef __GNUC__
#pragma interface
#endif

#ifdef INSTRUMENT

#include "ProxiableObject.h"

#include "InstrumentTypes.h"
#include "InstrumentFormat.h"

extern class InstrumentManager;

#include "InstrumentManager.h"

// ----------------------------------------------------------------------------

struct iInstrumentStateType {
	iInstrumentType type;
	iStateType currentState;
	Boolean active;
	Boolean verbal;
	Boolean loud;
	Boolean suspended;
	iStateType suspendedState;
};

extern ClassStar InstrumentClass;

class Instrument : public ProxiableObject {
/*
 * The base class for all instruments.
 * Note that Eventers are not the base class only due to the _header being
 * slightly different for each instrument.  The _header must be contiguous
 * memory, though.  I could just declare all the fields in the superclass
 * (which would then be an Eventer), but that isn't very clean or extendable.
 */

    friend class InstrumentManager;

	
private:
	
protected:
	char *_title;

	InstrumentManager *_owner;
	iInstrumentStateType _state;

	/* Method birthLog:	Log instrument creation.		*/
	virtual void birthLog () = 0;

	/* Method deathLog:	Log instrument destruction.		*/
	virtual void deathLog () = 0;
		
public:
	/* Method Instrument:	The constructor.			*/
	Instrument (InstrumentManager *iowner,
		    char *title = "Untitled",
		    iStateType state = DEAD);
	
	/* Method ~Instrument:	The destructor.				*/
	~Instrument ();

	/* Method changeState:	Change the instrument activity.		*/
	void changeState (iStateType newState);

	/* Method suspend:	Save old state and become DEAD.		*/
	virtual void suspend ();
	
	/* Method resume:	Restore the saved state, if any.	*/
	virtual void resume ();
	
	/* Method reset:	Set all instrument data to 0.		*/
	virtual void reset () = 0;

	/* Method getID:	Return the assigned instrument ID.	*/
	virtual unsigned int getID () = 0;

	Class * classOf ();

	/* Method dummy: 		required for proxify++ (bug)	*/
	proxiable virtual void dummy () {}
	
	virtual void begin ();
};

// ----------------------------------------------------------------------------

extern ClassStar EventerInstrumentClass;

class EventerInstrument : public Instrument {
/*
 * Just an event logger.
 */

private:

protected:
	struct {
		iIdentifierType id;
	} _header;

	/* Method birthLog:	Log instrument creation.		*/
	void birthLog ();

	/* Method deathLog:	Log instrument destruction.		*/
	virtual void deathLog ();

	/* Method log:		Log an event.				*/
	void log (unsigned int eventID,
		  char *message,
		  unsigned int messageLength);
	
public:

	/* Method EventerInstrument:		The constructor.	*/
	proxiable EventerInstrument (InstrumentManager *iowner,
				     char *title,
				     iStateType state = DEAD);
	
	/* Method ~EventerInstrument:	The destructor.			*/
	proxiable ~EventerInstrument ();

	/* Method reset:	Set all instrument data to 0.		*/
	proxiable virtual void reset ();

	/* Method getID:	Return the assigned instrument ID.	*/
	proxiable virtual unsigned int getID ();

	/* Method event:	Cause an event to occur.		*/
	proxiable virtual void event (unsigned int eventID,
				      char *message = NULL,
				      unsigned int messageLength = 0);
	Class * classOf ();
};

// ----------------------------------------------------------------------------

struct counterHeaderStruct {
	iIdentifierType id;
	countType count;
};

extern ClassStar CounterInstrumentClass;

class CounterInstrument : public Instrument {
/*
 * This is essentially the same as an eventer,
 * but it keeps track of the number of events.
*/

private:
	
protected:
	unsigned int _freq;

	counterHeaderStruct _header;

	/* Method birthLog:	Log instrument creation.		*/
	void birthLog ();

	/* Method deathLog:	Log instrument destruction.		*/
	virtual void deathLog ();

	/* Method log:		Log an event.				*/
	void log (unsigned int eventID,
		  char *message,
		  unsigned int messageLength);
	
public:

	/* Method CounterInstrument:		The constructor.	*/
	proxiable CounterInstrument (InstrumentManager *iowner,
				     char *title,
				     iStateType state = DEAD,
				     unsigned int freq = SUMMARY);
	
	/* Method ~CounterInstrument:	The destructor.			*/
	proxiable ~CounterInstrument ();

	/* Method reset:	Set all instrument data to 0.		*/
	proxiable virtual void reset ();

	/* Method getID:	Return the assigned instrument ID.	*/
	proxiable virtual unsigned int getID ();

	/* Method changeFreq:	Alter the event logging frequency.	*/
	proxiable virtual void changeFreq (unsigned int newFreq);
	
	/* Method event:	Cause an event to occur.		*/
	proxiable virtual void event (unsigned int eventID,
				      char *message = NULL,
				      unsigned int messageLength = 0);
	Class * classOf ();
};

struct timerHeaderStruct {
	iIdentifierType id;
	countType count;
	timeType time;
};

// ----------------------------------------------------------------------------

extern ClassStar TimerInstrumentClass;

class TimerInstrument : public Instrument {
/*
 * This is similar to a counter, but with a notion of change in time.
 */

private:
	timeType _startTime;
	
protected:
	iTimerStateType _timerState;
	iScopeType _timerScope;
	
	unsigned int _freq;

	timerHeaderStruct _header;
	
	/* Method birthLog:	Log instrument creation.		*/
	void birthLog ();

	/* Method deathLog:	Log instrument destruction.		*/
	virtual void deathLog ();

	/* Method log:		Log an event.				*/
	void log (unsigned int eventID,
		  char *message,
		  unsigned int messageLength);
		
public:

	/* Method TimerInstrument:	The constructor.		*/
	proxiable TimerInstrument (InstrumentManager *iowner,
				   char *title,
				   iStateType state = DEAD,
				   unsigned int freq = SUMMARY,
				   iScopeType scope = ALL);

	/* Method ~TimerInstrument:	The destructor.			*/
	proxiable ~TimerInstrument ();
	
	/* Method reset:	Set all instrument data to 0.		*/
	proxiable virtual void reset ();

	/* Method getID:	Return the assigned instrument ID.	*/
	proxiable virtual unsigned int getID ();

	/* Method suspend:	Save old state and become DEAD.		*/
	// Im not quite sure how to handle timer suspensions
	proxiable virtual void suspend ();

	/* Method resume:	Restore the saved state, if any.	*/
 	proxiable virtual void resume ();

	/* Method changeFreq:	Alter the event logging frequency.	*/
	proxiable virtual void changeFreq (unsigned int newFreq);

	/* Method changeScope:	Alter the timer scope.			*/
	proxiable virtual void changeScope (iScopeType newScope);
	
	/* Method start:	Start the timer.			*/
	proxiable virtual void start ();

	/* Method stop:		Stop the timer and log if appropriate.	*/
	proxiable virtual void stop (unsigned int eventID,
				     char *message = NULL,
				     unsigned int messageLength = 0);
	Class * classOf ();
};

// ----------------------------------------------------------------------------

unsigned int StringLength (const char *string);

void StringCopy (char *dest, const char *source);

#endif INSTRUMENT
#endif
