/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * InstrumentBuffer.h: buffer specifically for instrumentation use
 *
 * InstrumentBuffer is a *per* CPU buffer exclusively used by processes logging
 * events via the Choices instruments executing on the CPU.  This buffer is
 * *not* shared by any other processor.  The first log event is reserved for
 * recording the birth of the buffer during initialization and then used for
 * recording the last dumptime thereafter.  The rest of the buffer is available
 * for general event logging.
 *
 * Note: Unlike other instruments, the buffer is never used by the user
 *       directly and hence do not have an "event()" method for logging events.
 *	 However, the header is kept similar to other instruments with the
 *	 cpuLength and cpu are used instead of "messageLength" and "message".
 * 	 This allows post-processing conversion tools to treat the buffer
 *       creation and clear events to be processed wihtout introducing new
 * 	 event types. 
 *
 */

#ifndef InstrumentBuffer_h
#define InstrumentBuffer_h


#ifdef __GNUC__
#pragma interface
#endif

#ifdef INSTRUMENT

#include "InstrumentTypes.h"
#include "InstrumentFormat.h"
#include "InstrumentEvent.h"
#include "FreeRunningTimer.h"
#include "FreeRunningTimerStar.h"

struct bufferIdentifierType
{
	timeType dumpTime;		// timestamp of last dump
	iIdentifierType id;		// generic Instrument "id"
	unsigned int eventID;		// buffer specific event ID
	unsigned int cpuLength;		// sizeof(cpu)
	unsigned long cpu;		// thisCPU id
};

extern ClassStar InstrumentBufferClass;

class InstrumentBuffer {
	
    friend class SystemInstrumentManager;

private:

        unsigned int _mxSz;		// maximum buffer size
	unsigned int _size;		// useable buffer size
	char *_buffer;			// the actual buffer

public:

	/* Method InstrumentBuffer:		The constructor.	*/
	InstrumentBuffer (unsigned int bufferAlloc, unsigned long cpuNumber);

	/* Method ~InstrumentBuffer:		The destructor.		*/
	~InstrumentBuffer ();

	/* Method getBufferLocation:	Returns the buffer address.	*/
	char *getBufferLocation ();

	/* Method getBufferSize:	Returns the size of the buffer.	*/
	unsigned int getBufferSize ();

	/* Method getBufferSizeLeft:	Returns number of bytes left.	*/
	unsigned int getBufferSizeLeft ();

        /* Method getBufferMaxSize:     Returns the value of _mxSz   */
        /* Temporary function */
        unsigned int getBufferMaxSize ();
    
	/* Method clear:		Resets the buffer.		*/
	void clear ();

	/* Method timeStamp:		Logs the current time.		*/
	void timeStamp ();

	/* Method append:		Log data to the buffer.		*/
	void append (unsigned int dataLength, char *data);

	Class *classOf ();
};

#endif INSTRUMENT
#endif
