/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David K. Raila (raila@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
#ifndef SimpleInstruments_h
#define SimpleInstruments_h


#ifdef __GNUC__
#pragma interface
#endif

enum SimpleInstrumentState { ACTIVE, INACTIVE };

#ifdef SIMPLE_INSTRUMENTATION

#include "ProxiableObject.h"
#include "Lock.h"

extern ClassStar SimpleInstrumentClass;
extern ClassStar SimpleEventCounterClass;
extern ClassStar SimpleTimerClass;

extern void activateSimpleInstrument( ProxiableObject *);
extern void deactivateSimpleInstrument( ProxiableObject *);
extern void simpleInstrumentControl( int argc, char *argv[]); 

	/*
	 * Simple instrument classes 
	 */
class SimpleInstrument : public ProxiableObject {
	SimpleInstrumentState _state;	// active, inactive, etc.
	Lock	_lock;
public:
	SimpleInstrument(const char *const name, SimpleInstrumentState = INACTIVE);
	virtual ~SimpleInstrument(){}

	/*
	 * Turn it on/off
	 */
	void activate();
	void deactivate();
	int active(){return ( _state == ACTIVE );}

	/*
	 * Mutex used from subclasses
	 */
	virtual void lock();
	virtual void unlock();

	virtual void reset(){}
	virtual void displayOnConsole() = 0;

	Class * classOf();
};

inline 
SimpleInstrument::SimpleInstrument(
	const char *const name, SimpleInstrumentState initState)
{
	_state = initState;
	setName(name);
}

inline
void
SimpleInstrument::lock()
{
	_lock.acquire();
}

inline
void
SimpleInstrument::unlock()
{
	_lock.release();
}

inline
void
SimpleInstrument::activate()
{
	_state = ACTIVE;
}

inline
void
SimpleInstrument::deactivate()
{
	_state = INACTIVE;
}

class SimpleEventCounter : public SimpleInstrument {
protected:
	unsigned _count;

public:

	/*
	 * Pass name to ctor, INACTIVE if you want to
	 */
	SimpleEventCounter(const char *const name, SimpleInstrumentState = INACTIVE);

	/*
	 * returns the number of events.
	 */
	unsigned count();

	void log();

	void reset();
	Class * classOf();
};

inline 
SimpleEventCounter::SimpleEventCounter(const char *const name,
	SimpleInstrumentState iState) : SimpleInstrument(name, iState)
{
	_count = 0;
}

inline 
void
SimpleEventCounter::reset()
{
	_count = 0;
}

inline
unsigned
SimpleEventCounter::count()
{
	return _count;
}

inline 
void
SimpleEventCounter::log()
{
	lock();
	_count++;
	unlock();
}

class SimpleTimer : public SimpleInstrument {
private:
	unsigned _count;
	unsigned _totalTime;
	unsigned _min;
	unsigned _minSample;
	unsigned _max;
	unsigned _maxSample;

	unsigned basicStart();
	unsigned basicStop(const unsigned);
	unsigned basicSuspend(const unsigned);
	unsigned basicResume(const unsigned);
public:
	SimpleTimer(const char * const name, SimpleInstrumentState = INACTIVE);

	void reset();

	/*
	 * Start starts a timing, returning the start time to pass to stop
	 */
	unsigned start();

	/*
	 * Stop stops the current timing for the start time passed in.
	 * and returns the time since start.
	 * It also updates the total time and count of timings.
	 */
	unsigned stop(const unsigned);

	/*
	 * Suspend timing (like stop)
	 */
	unsigned suspend(const unsigned);

	/* 
         * Resume timing (like start)
	 */
	unsigned resume(const unsigned);

	/*
	 * Average returns the average time per call to date.
	 */
	unsigned avg();

	unsigned min();
	unsigned minSample();
	unsigned max();
	unsigned maxSample();
	unsigned count();

	/*
	 * Print stats to console
	 */
	void displayOnConsole();

	Class * classOf();
};

inline
unsigned
SimpleTimer::count()
{
	return _count;
}

inline
void
SimpleTimer::reset()
{
	_count = _totalTime = _max = 0;
	_min = 0xffffffff;
	SimpleInstrument::reset();
}

inline
SimpleTimer::SimpleTimer (const char * const name,
	SimpleInstrumentState initState) : SimpleInstrument(name, initState)
{
	reset();
}

inline
unsigned
SimpleTimer::min()
{
	return _min;
}

inline
unsigned
SimpleTimer::minSample()
{
	return _minSample;
}

inline
unsigned
SimpleTimer::max()
{
	return _max;
}

inline
unsigned
SimpleTimer::maxSample()
{
	return _maxSample;
}

inline
unsigned
SimpleTimer::avg()
{
	if (_count == 0)
		return 0;
	return _totalTime/_count;
}

inline
unsigned 
SimpleTimer::start()
{
	return(active() ? basicStart() : 0 );
}

inline
unsigned 
SimpleTimer::stop(const unsigned i)
{
	return(active() ? basicStop(i) : 0 );
}

inline
unsigned 
SimpleTimer::suspend(const unsigned i)
{
	return(active() ? basicSuspend(i) : 0 );
}

inline
unsigned 
SimpleTimer::resume(const unsigned i)
{
	return(active() ? basicResume(i) : 0 );
}

#else /* NOT SIMPLE_INSTRUMENT*/

class SimpleTimer {
public:
	SimpleTimer(){}
	SimpleTimer(const char * const){}
	void activate(){}
	void deactivate(){}
	unsigned start(){return 0;}
	unsigned stop(unsigned){ return 0;}
	unsigned suspend(unsigned){return 0;}
	unsigned resume(unsigned){return 0;}
	unsigned min(){return 0;}
	unsigned max(){return 0;}
	unsigned avg(){return 0;}
	unsigned minSample();
	unsigned maxSample();
	unsigned count();
};
#endif /*SIMPLE_INSTRUMENTATION*/


	/*
	 **************************************************************
	 * Simple instruments defined here.
	 * There is one class each for when SIMPLE_INSTRUMENTATION is defined
	 * and not.  Be sure to add/delete both!
	 **************************************************************
	 */
#ifdef SIMPLE_INSTRUMENTATION
extern ClassStar MemoryTimerClass;
class MemoryTimer : public SimpleTimer
{
public:
	MemoryTimer(const char * const name) : SimpleTimer(name) {}
	Class * classOf();
};

extern ClassStar ProcessTimerClass;
class ProcessTimer : public SimpleTimer
{
public:
	ProcessTimer(const char * const name) : SimpleTimer(name) {}
	Class * classOf();
};

extern ClassStar FileSystemTimerClass;
class FileSystemTimer : public SimpleTimer
{
public:
	FileSystemTimer(const char * const name) : SimpleTimer(name) {}
	Class * classOf();
};

#define SInstr(exp) exp

#else /* NOT SIMPLE_INSTRUMENTATION*/

/*
 * Swallow that semicolon and don't generate code
 */
#define SInstr(exp) while(0){} 

class MemoryTimer : public SimpleTimer {};
class ProcessTimer : public SimpleTimer {};
class FileSystemTimer : public SimpleTimer {};

#endif /*SIMPLE_INSTRUMENTATION*/


#endif /*SimpleInstruments_h*/










