/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Aamod Sane (russo@cs.uiuc.edu)
 * Author: David Raila (raila@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Kernel.h: Kernel description class.
 */

#ifndef Kernel_h
#define Kernel_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Object.h"

#include "Domain.h"
#include "DiskStar.h"
#include "DomainStar.h"
#include "FileSystemInterfaceStar.h"
#include "VMInterfaceStar.h"
#include "NameServerStar.h"
#include "PersistentMemoryObjectStar.h"
#include "MemoryObjectContainerStar.h"
#include "MemoryObjectDictionaryStar.h"
#include "PhysicalMemoryRangeStar.h"
#include "KernelCmdInstaller.h"

#include <map>
using namespace std;

class ProcessContainer;
class Allocator;
class Process;
class BSDProcess;
class MachineConfiguration;
#ifdef NETWORKING
class NetworkDriver;
class NetworkAdministrator;
#endif

class OutputStream;
extern OutputStream * ConsoleOutputStream;
class BootConsole;
extern BootConsole * bootConsole;

typedef void (*KCMD)(int, char**); // Kernel Command Pointer

extern ClassStar KernelClass;
class Kernel : public Object {

    friend void Main( int argc, char ** argv ); // to call initialize.
    friend void kernelStub( Kernel * );

 private:
    map<int, KCMD> _kernelCommands;
 protected:
    /*
     * The Domain to map all kernel memory objects into.
     */
    Domain          * _kernelDomain;

    /*
     * The kernel and 1-1 memory object references.
     * low, readonly, and high are the kernel MO's.
     * We make a variable number of 1-1 mapped MO's
     * based on the PhysicalMemoryMap.
     * These are used in the VM initialization code in
     * Kernel::initialize( Allocator * ).
     */
    MemoryObjectStar    _lowMemoryObject;
    MemoryObjectStar    _readOnlyMemoryObject;
    MemoryObjectStar    _highMemoryObject;
    MemoryObjectStar    _heapMemoryObject;
    MemoryObjectStar    * _oneToOneMemoryObjects;
    unsigned        _numOneToOneMemObjs;

    Allocator *     _kernelHeap;
    FileSystemInterfaceStar _fileSystemInterface;
    VMInterfaceStar         _VMInterface;

    /*
     * A source of empty backing store PersistentMemoryObjects. Eventually
     * we need to fix this to be a collection of MemoryObjectContainers
     * since, in reality, we will probably want to page simultaneously
     * to/from objects in multiple containers. For now we'll just use one.
     */
    MemoryObjectContainerStar _backingStoreSource;

    /*
     * This is temporary until Peter and I decide how to build a collection
     * of all the disks in all the machine on all the neworks in
     * all the countries on all the planets ...      --Vince
     */
    virtual void createFileSystemInterface();
    virtual void mountNonRootFileSystems();
    virtual void initializeBackingStore();

    /* Hack VMInterface for now */
    virtual void createVMInterface();

    /*  MachineConfiguration stuff.
     */
    MachineConfiguration *_configuration;

    /*
     * Used by the first processor to start up in order to initialize
     * physical memory and the initial virtual memory mappings.
     * Only Main() should ever call this. Usually this is not overloaded.
     */
    virtual void initialize( Allocator * bootHeap );

    /*
     * Defined by subclasses to do subclass specific initializations.
     */
    virtual void configureHardware() {}
    virtual void buildRunningCPU() = 0;

    /*
     * The subclass overloads this method if it needs to create other VM
     * mappings in memory objects before VM is turned on
     */
    virtual void basicAddMemoryObjects() {}

    /*
     * Do machine independent setup of devices and manager processes.
     */
    virtual void initializeDevicesAndManagers();

    /*
     * The subclass overloads this method to to machine specific
     * forms of the above.
     */
    virtual void basicInitializeDevicesAndManagers() = 0;

    /*
     * Put things to disable all further interrupts here.
     * Everything **should** be done by now
     */
    virtual void shutdown() {}

#ifdef NETWORKING
    /*
     *  Networking ...
     *
     */
    NetworkDriver *_ethernetDriver;
    NetworkAdministrator *_ethernetAdministrator;
    virtual void initializeNetworkProcesses();
    virtual void basicInitializeNetworkProcesses() = 0;
#endif

    /*
     * The code the top level process executes to dispatch applications.
     */
    virtual void dispatchApplications();

    /*
     * Assist static (class) functions.
     */
    static void wasteManagementProcessEntry( Kernel * k );

    /*
     * Subclasses overload this to do what the name says. Implementations
     * of this method should be aware that there is no allocator to
     * use to create new objects with when this method is invoked.
     */
    virtual void startOtherProcessors();

    /*
     * Called by the setupProcess to get things going.
     */
    void main();

 public:

// unsigned char rangeCheck( char * addr, Domain*, int strict = 0);

    //PhysicalMemoryRangeStar       _physicalMemoryMap;
    //char *                _kernelStart;
    virtual PhysicalMemoryRange *   physicalMemoryMap() = 0;
    virtual const unsigned      physicalMemorySize() = 0;
    virtual const char *        kernelStart() = 0;
    virtual const char *        applicationStart() = 0;
    virtual const unsigned      kernelHeapSize() = 0;
    virtual const unsigned      kernelSize() = 0;
    virtual const unsigned      applicationSize() = 0;
    virtual const unsigned      addressTranslationMemorySize() = 0;

    Kernel();
    virtual ~Kernel();

    /*
     * Where the setupProcess starts running.
     */
    static void setupProcessEntry( Kernel * );
    static void setupProcessTrampolineEntry( Kernel * );

    /*
     * Were process to be deleted are held.
     */
    // THIS WONT BE PUBLIC FOR LONG!!!
    ProcessContainer * _garbageCan;

    /*
     * The initial scheduler for all processors.
     */
    ProcessContainer * _initialScheduler;

    /*
     * Return the kernel's Domain.
     */
    Domain * domain();

    /*
     * Return a free PersistentMemoryObject available to use as a backing store.
     * This is a cleaner interface to the above backingStoreSource hack.
     */
    virtual PersistentMemoryObjectRef allocateMemoryObjectForBackingStore();

    /*
     * Support code to run *COFF* processes. A subclass could change the
     * object file format.
     * Actually, I'd like to see an ObjectFileParser class. The kernel
     * could have an instance of one set by the particular MD subclass.
     * Subclasses of ObjectFileParser could handle BSD, COFF, or
     * anything we come up with. These routines could the be replaced
     * by calls to this new object.  --Vince
     */
    virtual int isObjectFile( PersistentMemoryObject * objectFile ) = 0;
    virtual const char * loadObjectFile(
        PersistentMemoryObject * objectFile,
        Domain * newProcessesDomain ) = 0;
    virtual PersistentMemoryObjectRef getMemoryObjectOfInitialApplicationProcess();


    /* 
     * Accessors for configuration data.
     *
     * Overload buildConfiguration for special (diskless, VC, etc.)
     * machines.
     */
    MachineConfiguration *getConfiguration();
    virtual void buildConfiguration();

    /*
     * Called by getMemoryObjectOfInitialApplicationProcess() after
     * the input command is entered.  Kernels can define this and
     * intercept commands.  Return of 0 means the kernel then tries
     * to run the command, non-zero and the kernel goes back to the
     * prompt loop, assuming the subclass kernel ran the command.
     * Partnumber gets the numeric representation of argv[0].
     */
    static KernelCmd* installCmd(KernelCmd* cmd);
    virtual int tryCommand(int argc, char *argv[], int partnumber);
    virtual Process * buildProcessFromObjectFile( PersistentMemoryObject *,
                                                  int argc, char * argv[] );
    virtual Process * buildProcessFromObjectFile( PersistentMemoryObject *,
                                                  Domain *, int argc, char * argv[] );

    Class * classOf();
    FileSystemInterfaceRef fileSystemInterface();

    virtual DiskRef rootDisk() = 0;
    virtual MemoryObjectContainer * backingStoreSource();
};

extern Kernel * TheKernel;

inline Domain *
Kernel::domain()
{
    return( _kernelDomain );
}

inline MachineConfiguration *
Kernel::getConfiguration()
{
    return _configuration;
}

#endif /* Kernel_h */
