/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane sane@cs.uiuc.edu
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
#ifndef BitArray_h
#define BitArray_h

#ifdef __GNUC__
#pragma interface
#endif

#include "GPInclude.h"
#include "Assert.h"


extern const unsigned BitsInInt;
extern const unsigned LSB;
extern const unsigned MSB;
extern const unsigned ZERO;
extern const unsigned FFFF;

// Bit Array using the largest possible access unit in the machine.
// No checks are made on the args, since we want the functions to be
// simple inlines.
// Also, to keep it efficient, no virtual functions.

class IntBitArray {
    unsigned mask(const unsigned from, const unsigned to);
    unsigned mask(const unsigned pos);
  protected:
    unsigned* theInt;
  public:
    IntBitArray(unsigned*  r);
    ~IntBitArray();

    void set();
    void set(const unsigned pos);
    void set(const unsigned from, const unsigned to);

    int get();
    int get(const unsigned pos);
    int get(const unsigned from, const unsigned to);

    void clear();
    void clear(const unsigned pos);
    void clear(const unsigned from, const unsigned to);
    
    void invert();
    void invert(const unsigned pos);
    void invert(const unsigned from, const unsigned to);
    
    /*
     * Search for the next bit of type 1 or 0 in increasing (dir== 1)
     * or decreasing (dir==1) direction. Return -1 for not found, else
     * the position.
     */
    int bitSearch(unsigned pos, int type, int dir);
    
    /*
     * find the contiguos extent around pos of bits that have the same
     * value as bits. Record the extent in from and to - return the value of
     * the bit itself as the return value.
     */
    int getExtent(const unsigned pos, unsigned& from, unsigned& to);

    void inspect();
};

inline unsigned
IntBitArray::mask(const unsigned pos)
{
    return LSB << pos;
}

inline unsigned
IntBitArray::mask(const unsigned from, const unsigned to)
{
   return (((LSB << to) - 1) + (LSB << to)) & ~((LSB << from) - 1);
}

inline
IntBitArray::IntBitArray(unsigned* r)
: theInt(r)
{
}

inline
IntBitArray::~IntBitArray()
{
}

inline void
IntBitArray::set()
{
    (*theInt) = FFFF;
}

inline void
IntBitArray::set(const unsigned pos)
{
    (*theInt) |= mask(pos);
}

inline void
IntBitArray::set(const unsigned from, const unsigned to)
{
    (*theInt) |= mask(from,to);
}

inline int
IntBitArray::get()
{
    return ((*theInt) == FFFF) ? 1 : (((*theInt) == ZERO) ? 0 : -1) ;
}

inline int
IntBitArray::get(const unsigned pos)
{
    return ((*theInt) & mask(pos)) ? 1 : 0;
}

inline void
IntBitArray::clear()
{
    (*theInt) = ZERO;
}

inline void
IntBitArray::clear(const unsigned pos)
{
    (*theInt) &= ~mask(pos);
}

inline void
IntBitArray::clear(const unsigned from, const unsigned to)
{
    (*theInt) &= ~mask(from,to);
}

inline void
IntBitArray::invert()
{
    (*theInt) = ~(*theInt);
}

inline void
IntBitArray::invert(const unsigned pos)
{
    (*theInt) = ((*theInt) & mask(pos)) ? ((*theInt) & ~mask(pos)) : ((*theInt) | mask(pos));
}

inline void
IntBitArray::invert(const unsigned from, const unsigned to)
{
    (*theInt) = ((*theInt) & ~mask(from,to)) | ~((*theInt) & mask(from,to));
}

// Everything public, so that any users can be as efficient as
// possible. Check out the BitRanges class for an example

class BitArray {
  public:
    static unsigned inInt(const unsigned pos);
    static unsigned offsetInInt(const unsigned pos);
    static unsigned bitpos(const unsigned word,const unsigned offset);
    
    unsigned* array;
    unsigned  bits;
    unsigned  ints;
    
    BitArray(int n);
    virtual ~BitArray();
    
    void set();
    void set(const unsigned pos);
    void set(const unsigned from, const unsigned to);

    int get();
    int get(const unsigned pos);
    int get(const unsigned from, const unsigned to);

    void clear();
    void clear(const unsigned pos);
    void clear(const unsigned pos, const unsigned to);
    
    void invert();
    void invert(const unsigned pos);
    void invert(const unsigned pos, const unsigned to);
    
    /*
     * Search for the next bit of type 1 or 0 in increasing (dir== 1)
     * or decreasing (dir==1) direction. Return -1 for not found, else
     * the position.
     */
    int bitSearch(unsigned pos, int type, int dir);
    
    /*
     * find the contiguos extent around pos of bits that have the same
     * value as bits. Record the extent in from and to - return the value of
     * the bit itself as the return value.
     */
    int getExtent(const unsigned pos, unsigned& from, unsigned& to);

    virtual void inspect();
};

inline unsigned
BitArray::inInt(const unsigned pos)
{
    return pos/BitsInInt;
}

inline unsigned
BitArray::offsetInInt(const unsigned pos)
{
    return pos%BitsInInt;
}

inline unsigned
BitArray::bitpos(const unsigned word, const unsigned offset)
{
    return offset + word*BitsInInt;
}

inline
BitArray::~BitArray()
{
    delete array;
}

#endif /* BitArray_h */
