/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * List.h - List varieties
 *
 *	$Header: /projects/choices/old/cvsroot/choices/Includes/Libraries/GeneralPurpose/List.h,v 1.1 2004/06/03 01:58:43 jcarlyle Exp $
 */

#ifndef List_h
#define List_h

#ifdef __GNUC__
#pragma interface
#endif


#include "Assert.h"
#include "Link.h"

typedef void (*PFLV)(Link *);

// Testing List - Links are Lists get tested in Lists.cc
// Tests, solutions etc. are all in Applications/UNIXGP

// Lists are assumed to be circular in what follows.
// Entries of List will be Link's and the Link functions will get used
// in functions such as find();

class List {
protected:
    Link * _head;
    
    virtual Link * start();
public:
    List();
    List(Link * l);
    virtual ~List();
    
    // In some cases, access to actual list may be required.
    // But avoid using this unless absolutely essential
    Link * theLink();
    Link * first();
    Link * last();
    
    int isFirst(Link *l);
    int isLast(Link * l);
    
    // find uses Link::compare
    virtual Link * find( Link * e ); 
    // nth starts at 0
    virtual Link * find( Link * e, int nth ); 

    //
    // NOTE - oldEntry IS EXPECTED TO BE A POINTER TO AN EXISTING ENTRY
    // Use find to get the required entry if neccessary
    //

    virtual void add( Link * newEntry, Link * oldEntry, int before=0 );
    virtual void addFirst( Link * newEntry );
    virtual void addLast( Link * newEntry );
    virtual Link * remove( Link * entry, int nth=0 );
    virtual Link * removeFirst();
    virtual Link * removeLast();

    virtual void destroy();

    // uses Link::inspect()
    virtual void inspect();

    virtual void mapcar(PFLV f);
};
    
// Return a starting point for find
inline Link *
List::start()
{
    return _head;
}

inline
List::List(Link * l)
{
    _head = l;
}

inline
List::List()
{
    _head = 0;
}

inline
List::~List()
{
}

inline Link *
List::theLink()
{
    return _head;
}

inline Link *
List::first()
{
    return _head;
}

inline Link *
List::last()
{
    return _head->previous();
}

inline void
List::addFirst( Link * e )
{
    // add before _head
    add(e,_head,1);
}

inline void
List::addLast( Link * e )
{
    if (_head)
	// add after _head->previous()
	add(e,_head->previous(),0);
    else
	add(e,_head,0);
}

inline int
List::isFirst( Link * l )
{
    if (_head)
	return _head->compare(l);
    else
	return 0;
}

inline int
List::isLast( Link * l )
{
    if (_head)
	return _head->previous()->compare(l);
    else
	return 0;
}

// A list that counts how many entries are on it.

class CountingList : public List {
  protected:
    int _numItems;
  public:
    CountingList();
    CountingList(Link * l, int existingItems);
    ~CountingList();
    void add( Link* newEntry, Link* oldEntry, int before=0 );
    Link * remove( Link * oldEntry, int nth=0 );
    int numItems();

    virtual void destroy();
    virtual void inspect();
};

inline
CountingList::CountingList()
{
    _numItems = 0;
}

inline
CountingList::CountingList(Link * l, int existingItems)
: List(l)
{
    Assert( l != 0 && existingItems > 0 );
    _numItems = existingItems;
}

inline
CountingList::~CountingList()
{
}

inline int
CountingList::numItems()
{
    return _numItems;
}

inline void
CountingList::destroy()
{
    List::destroy();
    _numItems = 0;
}

// A list that maintains a hint for finding elements

class HintList : public CountingList {
  protected:
    Link * _hint;
    Link * start();
  public:
    HintList();
    HintList(Link * l, int existingItems);
    ~HintList();

    Link * find( Link * e ); 
    Link * find( Link * e, int nth ); 
    
    void add( Link* newEntry, Link* oldEntry, int before=0 );
    Link * remove( Link * oldEntry, int nth=0 );

    virtual void destroy();
};

inline
HintList::HintList()
{
    _hint = _head;
}

inline
HintList::HintList(Link * l, int existingItems)
: CountingList(l,existingItems)
{
    _hint = _head;	// ok since base class is constructed first
    Assert( _head == l );
}

inline
HintList::~HintList()
{
}

inline Link *
HintList::start()
{
    return _hint;
}

inline void
HintList::destroy()
{
    CountingList::destroy();
    _hint = _head;
}

#endif	/* List_h */
