/* 
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef FileDescriptor_h
#define FileDescriptor_h


#ifdef __GNUC__
#pragma interface
#endif

#include "PersistentMemoryObjectStar.h"
#include "ProxiableObjectStar.h"
#include "UnixConstants.h"

class OutputStream;
class InputStream;
class RecordStream;
class MemoryObjectDictionary;
class PersistentArray;
class PersistentMemoryObject;
#ifdef APPLICATIONCONDUIT
class ApplicationConduitStream;
#endif

class FileDescriptorTable;
class InputStreamFileDescriptor;
class OutputStreamFileDescriptor;
class RecordStreamFileDescriptor;
class DirectoryStreamFileDescriptor;
#if APPLICATIONCONDUIT
class ApplicationConduitStreamFileDescriptor;
#endif

struct UNIXFileInfo;

enum FileDescriptorClass {
	NullFileDescriptorClass = 0,
	InputStreamFileDescriptorClass = 1,
	OutputStreamFileDescriptorClass = 2,
	RecordStreamFileDescriptorClass = 3,
	ApplicationConduitStreamFileDescriptorClass = 4,
	DirectoryStreamFileDescriptorClass = 5,
};

enum FileDescriptorFlags {
	FDFree = 0,
	FDUsed = 0x0001,
	FDCloseOnExec = 0x0002
};

class FileDescriptor
{
	/*
	 * A FileDescriptor is a bridge between the Unix filedes interface
	 * and the Choices stream-object interface.  In other word, a
	 * FileDescriptor translates Unix filedes functions into calls to
	 * Choices InputStreams, OutputStreams, BufferedRecordStreams and
	 * ApplicationConduitStreams.
	 */

      protected:

	/*
	 * A stream adapter can represent one of the following.
	 * The must all be pointers to proxies in order for exec to work
	 * right.
	 * In addition, subclasses may not have instance variables.
	 */

	int _flags;
	int _fileFlags;
	long _offset;
	FileDescriptorClass _class;
	ProxiableObject * _stream;
	ProxiableObjectStar _streamStar;
	PersistentMemoryObject * _pmo;
	PersistentMemoryObjectStar _pmoStar;

	/* 
	 * Some lazy macros
	 */

	InputStreamFileDescriptor * inputStreamFileDescriptor() 
		{ return (InputStreamFileDescriptor *) this; }
	OutputStreamFileDescriptor * outputStreamFileDescriptor() 
		{ return (OutputStreamFileDescriptor *) this; }
	RecordStreamFileDescriptor * recordStreamFileDescriptor() 
		{ return (RecordStreamFileDescriptor *) this; }
#ifdef APPLICATIONCONDUIT
	ApplicationConduitStreamFileDescriptor * 
		applicationConduitStreamFileDescriptor() 
		{ return (ApplicationConduitStreamFileDescriptor *) this; }
#endif // APPLICATIONCONDUIT
	DirectoryStreamFileDescriptor * directoryStreamFileDescriptor() 
		{ return (DirectoryStreamFileDescriptor *) this; }

      public:
	FileDescriptor();

	void init(ProxiableObject * stream, PersistentMemoryObject * pmo,
		  FileDescriptorClass);
	void init(FileDescriptor * fd);

	int setFree();
	int used() { return _flags & FDUsed; }
	int setUsed() { _flags |= FDUsed; return _flags; }
	int flags() { return _flags; }
	int setFlags(int flags) { return (_flags = flags); }
	int flag(FileDescriptorFlags f) { return _flags & f; }
	int setFlag(FileDescriptorFlags f, int v) 
		{ return (v ? (_flags |= f) : (_flags &= ~f)); }
	int fileFlags() { return _fileFlags; };
	int setFileFlags(int f) { return _fileFlags = f; }
	long offset() { return _offset; }
	long setOffset(long offset) { return _offset = offset; }
	ProxiableObject * proxiableObject() { return _stream; }
	PersistentMemoryObject * persistentMemoryObject() { return _pmo; }
	void computePersistentMemoryObject();
	PersistentMemoryObject * getPersistentMemoryObject()
		{ if (_pmo == 0) computePersistentMemoryObject(); return _pmo; }
	FileDescriptorClass fileDescriptorClass() { return _class; }

	int close();

	int read(char * buf, int len);       
	int write(const char * buf, int len);
	int lseek(int offset, int origin);
	int tell();
	int stat(UNIXFileInfo * i);
	int sync();
	int ioctl(unsigned request, char * arg);
};

class InputStreamFileDescriptor : public FileDescriptor
{
	InputStream * stream() { return (InputStream *) _stream; }
		
      public:
	int close();

	int read(char *buf, int len);       
	int write(const char *buf, int len);
	int lseek(int offset, int origin);
	int tell();
	int stat(UNIXFileInfo *i);
	int sync();   	    	    	
	int ioctl(unsigned request, char * arg);
};

class OutputStreamFileDescriptor : public FileDescriptor
{	
	OutputStream * stream() { return (OutputStream *) _stream; };

      public:
	int close();

	int read(char * buf, int len);       
	int write(const char * buf, int len);
	int lseek(int offset, int origin);
	int tell();
	int stat(UNIXFileInfo * i);
	int sync();   	    	    	
	int ioctl(unsigned request, char * arg);
};

class RecordStreamFileDescriptor : public FileDescriptor
{
	PersistentArray * stream() { return (PersistentArray *) _stream; }
		
      public:
	int close();

	int read(char * buf, int len);       
	int write(const char * buf, int len);
	int lseek(int offset, int origin);
	int tell();
	int stat(UNIXFileInfo * i);
	int sync();   	    	    	
	int ioctl(unsigned request, char * arg);
};

#ifdef APPLICATIONCONDUIT

class ApplicationConduitStreamFileDescriptor : public FileDescriptor
{
	ApplicationConduitStream * stream()
		{ return (ApplicationConduitStream *) _stream; }

      public:
	int close();

	int read(char * buf, int len);       
	int write(const char * buf, int len);
	int lseek(int offset, int origin);
	int tell();
        int stat(UNIXFileInfo * i);
	int sync();   	    	    	
	int ioctl(unsigned request, char * arg);
};

#endif // APPLICATIONCONDUIT

class DirectoryStreamFileDescriptor : public FileDescriptor
{
	MemoryObjectDictionary * stream()
		{ return (MemoryObjectDictionary *) _stream; }

      public:
	int close();

	int read(char * buf, int len);       
	int write(const char * buf, int len);
	int lseek(int offset, int origin);
	int tell();
        int stat(UNIXFileInfo * i);
	int sync();   	    	    	
	int ioctl(unsigned request, char * arg);
};

class FileDescriptorTable {
	int used;
	FileDescriptor fd[UA_MAXFDS];
	int acquire();
	int release();
      public:
	FileDescriptorTable();
	int size() { return UA_MAXFDS; }
	int init();
	int closeAll();
	int closeOnExec();
	int allocFileDescriptor(int minfd = 0);
	FileDescriptor * fileDescriptor(int i);
};

#endif
