/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * IntegratorTimer.h - defines for the ARM Integrator timer
 */

#ifndef IntegratorTimer_h
#define IntegratorTimer_h

#ifdef __GNUC__
#pragma interface
#endif

#include "ARMCPU.h"
#include "Integratordevices.h"
#include "IntegratorInterrupts.h"

#define Integrator_MPU_TIMER1_BASE 0x13000000
#define Integrator_MPU_TIMER2_BASE 0x13000100
#define Integrator_MPU_TIMER3_BASE 0x13000200

/* Timer control register */
#define TIMER_EN        (1 << 7)

#define TIMER_MODE_FREE     (0 << 6)
#define TIMER_MODE_PERIODIC     (1 << 6)

#define TIMER_MODE_INTEN    (1 << 5)

#define TIMER_PRESCALE0     (0 << 2)
#define TIMER_PRESCALE16    (1 << 2)
#define TIMER_PRESCALE256   (2 << 2)

#define TIMER_SIZE16        (0 << 1)
#define TIMER_SIZE32        (1 << 1)

#define TIMER_WRAPPING      (0 << 0)
#define TIMER_ONESHOT       (1 << 0)

inline void loop(unsigned i) {
    while (i > 0) i--;
}

class IntegratorMPUTimer {
 private:
    // This class cannot have any other member variables. An object of this class
    // is designed to be directly placed on memory-mapped hardware. There should not
    // be any virtual functions because there cannot be any vptr.
    // The order of the member variables below has to match hardware.
    // This is a compiler dependent class. Do not modify this unless you are sure
    // you know what you are doing
    volatile unsigned LOAD;      // W
    volatile unsigned VALUE;     // R
    volatile unsigned CONTROL;   // R/W
    volatile unsigned INTCLR;
    volatile unsigned RIS;
    volatile unsigned MIS;
    volatile unsigned BGLOAD;
 public:
    inline void start(unsigned value) {
        CONTROL = TIMER_EN;
        LOAD = value;
        CONTROL = TIMER_EN | TIMER_MODE_FREE | TIMER_PRESCALE0 | TIMER_SIZE16 | TIMER_WRAPPING | TIMER_MODE_INTEN;
    }
    inline unsigned read() {
        return VALUE;
    }
    inline void clear() {
        INTCLR = 1;
    }
};

extern IntegratorMPUTimer *iTimer1;
extern IntegratorMPUTimer *iTimer2;
extern IntegratorMPUTimer *iTimer3;

/*
 * XXX 
 *
 * QEMU sets this system clock to 40,000,000 Hz (see hw/integratorcp.c). Timer1
 * uses the system clock. Timer2 and Timer2 on Integrator are set to
 * 1,000,000 Hz.
 *
 * We set the clock prescaler to 0 in the start function in IntegratorMPUTimer.
 *
 */
#define MPU_TIMER_TICKS_PER_SEC   (40000000)

/*
 * These are for the FreeRunningTimer, it 
 * interrupts about every 2 seconds.
 */
const unsigned FRTuSecPerInterrupt = 2000000;
const unsigned FRTTickPeriod = ((MPU_TIMER_TICKS_PER_SEC / 1000000 * FRTuSecPerInterrupt) - 1);

/*
 * These are for the timer manager, it 
 * interrupts every 100 microseconds.
 */
const unsigned TimerManageruSecPerInterrupt = 100;
const unsigned TimerManagerTickPeriod = ((MPU_TIMER_TICKS_PER_SEC / 1000000 * TimerManageruSecPerInterrupt) - 1);

#endif /* IntegratorTimer_h */
