/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MultimaxKernel.h: Machine description class for the Multimax Kernel.
 */

#ifndef MultimaxKernel_h
#define MultimaxKernel_h


#ifdef __GNUC__
#pragma interface
#endif

#include "CRQ.h"
#include "Kernel.h"
#include "Process.h"
#include "MultimaxEMCDiskStar.h"

class PhysicalMemoryRange;

static const int MAX_BOOT_NAME_SIZE = 64;

struct MEM_STATE {
	long mem_status;
	long mem_addr;		    /* filled in after configuration.	    */
	long mem_sn;		    /* serial number			    */
	long mem_rev;		    /* revision number			    */
	long mem_part;		    /* part number			    */
	unsigned short mem_flags;   /* miscellaneous flags.		    */
	unsigned short mem_size;    /* Size of memory board in MB's.	    */
			/* Bit 0 when set means do NOT configure the board. */
};

struct REQ_STATE {
	long req_status[NUM_SLOT_DEV * 3];  /* devices per board times 3 */
					    /* status of each dev on board */
	long req_board_id;		    /* From id prom, see idprom.h */
	long req_sn;			    /* From id prom, see HIT-0040 */
	unsigned short req_major_revision;
	unsigned short req_minor_revision;
	long req_flags;			    /* miscellaneous flags. */
			/* Bit 0 when set means do NOT configure the board. */
};

struct SYS_STATE {
	long sys_state_topmem;	    /* Top of memory.			    */
	long sys_state_flags;	    /* Boot flags.			    */
	long sys_state_console;	    /* Unit ID of console device.	    */
	long sys_state_bootid;	    /* ID of boot device.		    */
	short sys_state_boottype;   /* Type of boot device.		    */
	short sys_state_bootpun;    /* Physical unit number of boot device. */
				    /* Pathname of bootdevice.		    */
	char sys_state_bootname[MAX_BOOT_NAME_SIZE];
	long sys_state_cpu;	    /* ID of boot CPU.			    */
	long sys_state_config;	/* Bit map of permanently frozen requestors.*/

	MEM_STATE sys_state_mem[NUM_MEM];
	REQ_STATE sys_state_req[NUM_SLOT];
};

/*
 * Board id's
 */
enum {	UNDEF_BD = 0,
	EMC = 1,
	DPC = 2,
	SCC = 3,
	DLA = 4,
	EMCII = 5,
	DPCII = 6,
	SMC = 7,
	SMCII = 8,
	SCCII = 9,
	APC = 10,
	DLA_16_ASYNC = 11,
	DLA_8_SYNC = 12,
	ANNEX_UX_ASYNC = 13,	// Unix Only Annex
	MSC = 14,
	SMC_16 = 15,
	ANNEXII_MX_ASYNC = 16,	// UNIX only Annex II
	ANNEXII_UX_ASYNC = 17,
	ANNEXII_SYNC = 18,
	TIU_UX_ASYNC = 19,	// Proteon TIU
	MSC_TRACE = 20,		// Special for DPC address trace
	UIC = 21,
	UNKNOWN_BD = 22,	// Should be 1 more than last one.
};

/*
 * Return system configuration data.	(CRQOP_SCC_SYS_STATE)
 */
class CRQ_SYS_STATE_MSG : public CRQ_MSG {
public:
	SYS_STATE * sys_state_addr;
	short sys_state_status;

	CRQ_SYS_STATE_MSG( SYS_STATE * sys_state_addr = 0 );
};

inline
CRQ_SYS_STATE_MSG::CRQ_SYS_STATE_MSG( SYS_STATE * config )
	: CRQ_MSG( sizeof( CRQ_SYS_STATE_MSG ), CRQOP_SCC_SYS_STATE,
		UnitID( 0, SCC_SLOT, 0, 0 ), 0 )
{
	sys_state_addr = config;
}

class CRQ_EXEC_MSG : public CRQ_MSG {   //  START EXECUTION MESSAGE
public:
	PFV     exec_start_addr;        // Program start address
	long    exec_status;            // Result of exec

	CRQ_EXEC_MSG( int slotid = 0, int cpuid = 0, PFV addr = 0 );
};

inline
CRQ_EXEC_MSG::CRQ_EXEC_MSG( int slotid, int cpuid, PFV addr )
        : CRQ_MSG( sizeof( CRQ_EXEC_MSG ), CRQOP_EXEC_PROG,
                UnitID( 0, slotid, cpuid, 0 ), 0 )
{
	exec_start_addr = addr;
}

extern ClassStar MultimaxKernelClass;
class MultimaxKernel : public Kernel {
protected:
	SYS_STATE systemConfiguration;
	CRQ_SYS_STATE_MSG getConfigMessage;
	CRQ_EXEC_MSG execMsg;

	MultimaxEMCDiskStar theDisk;

	/*
	 * Multimax specific methods.
	 */
	void startProcessor( unsigned int slotid, unsigned int cpuid,
			     PFV addr );
	void determineConfiguration();
	PhysicalMemoryRangeRef configure();

	/*
	 * Overloads of parent class methods.
	 */
	virtual void buildRunningCPU();
	void configureHardware();
	PhysicalMemoryRange *   physicalMemoryMap() ;
	const char *            kernelStart() ;
	const char *            applicationStart() ;
	const unsigned          kernelHeapSize() ;
	const unsigned          kernelSize() ;
	const unsigned          applicationSize() ;
	const unsigned          addressTranslationMemorySize() ;


	void basicInitializeDevicesAndManagers();
	void basicInitializeNetworkProcesses();
	DiskRef rootDisk();
	virtual void mountNonRootFileSystems();
	virtual void initializeBackingStore();

	void startOtherProcessors();

	int isObjectFile( PersistentMemoryObject * );
	const char * loadObjectFile( PersistentMemoryObject *, Domain *);

public:
	MultimaxKernel();
	~MultimaxKernel();

	Class * classOf();
};

/*
 * Hardware address space that must be mapped into all domains.
 */
static const unsigned int HWADDR = 0xfffc0000;
static const unsigned int HWADDR2 = 0xfffd1000;
static const unsigned int HWSIZE = 0x00040000;
static const unsigned KERNELSTART = 0x1000;	// skip first page
static const int KERNELSIZE = 0x4000000-KERNELSTART;	// 64 Mb
static const int APPLICATIONSTART = KERNELSTART + KERNELSIZE;
static const unsigned int APPLICATIONSIZE = HWADDR2 - APPLICATIONSTART;
static const int KERNELHEAPSIZE = 0xa00000; 		// 10 Mb
static const int ADDRESSTRANSLATIONMEMORYSIZE = 4096 * 75;

inline
const char *
MultimaxKernel::kernelStart(){return (const char *)KERNELSTART;}

inline
const char *
MultimaxKernel::applicationStart(){return (const char *)APPLICATIONSTART;}

inline
unsigned
MultimaxKernel::kernelSize(){return KERNELSIZE;}

inline
unsigned
MultimaxKernel::applicationSize(){return APPLICATIONSIZE;}

inline
unsigned
MultimaxKernel::kernelHeapSize(){return KERNELHEAPSIZE;}

inline
unsigned
MultimaxKernel::addressTranslationMemorySize()
{
	return ADDRESSTRANSLATIONMEMORYSIZE;
}

#endif /* MultimaxKernel_h */
