/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *			   University of Illinois at Urbana-Champaign
 *			   Department of Computer Science
 *			   1304 W. Springfield Ave.
 *			   Urbana, IL	 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David W. Dykstra (dykstra@cs.uiuc.edu)
 * Contributing Author: See-Mong Tan (stan@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * DatakitChannel.h: Per-channel Datakit processing;
 *	  switch and urp protocol implementation
 */

#ifndef DatakitURPChannel_h
#define DatakitURPChannel_h
#ifdef CFRONT
#define CONST const
#define STATIC static
#else /* must be g++ */
#pragma interface
#if (__GNUG__ >= 2)
#if (!defined CONST)||(!defined STATIC)
#define CONST
#define STATIC
#endif
#else
#pragma once
#define CONST const
#define STATIC static
#endif
#endif

#include "Assert.h"
#include "ProxiableObject.h"
#include "Urp.h"
#include "Lock.h"
#include "SemaphoreStar.h"
#include "DatakitDriver.h"
#include "DatakitChannel.h"

/*
 * Everything from here on is for URP,
 * and is therefore probably wrong,
 * since reading from a network stream
 * should not presuppose any particular protocol.
 * for example, we may want just a raw channel,
 * or we may want to run msp or tcp/ip over the channel.
 * it should all look the same to a user.
 * probably the best way of doing this is to introduce
 * a streams model like dennis ritchie's, except with
 * real processes.
 * then we should break urp up into 2 pieces, one that
 * does the urp processesing and writes datakit packets
 * to another queue, and another that gives a queue
 * into a Read/Write interface.
 * it would also be nice to push and pop these modules
 * onto the channel.
 */
class DatakitPacketQueue {
protected:
	DatakitPacket	*head;
	DatakitPacket	*tail;
	int		size;
	int		maxSize;
	SemaphoreStar	queueCount;
	Lock		*queueLock;
	
public:
	DatakitPacketQueue(int max);
	int queueSize();
	int enqueue(DatakitPacket *);
	int append(DatakitPacketQueue *);
	DatakitPacket *dequeue();
	int undequeue(DatakitPacket *);         // push a packet back
	void trash();                       // throw everything away
};

class DatakitXmitBuf {
public:
	unsigned char	*buf;           // buffers
	short           len;            // buffer lengths
	char            avail;          // 1 if buffer is available for use
	char            eob;            // 1 if buffer is last in block
};

#define DATAKITSTATS
#ifdef DATAKITSTATS
#define STAT(x) ((x)++)
#else
#define STAT(x)
#endif

// URP transmitter side was extensively rewritten by Stan

// transmitter constants
const int URPXmitBlockSize = 60;
const int URPXmitWindowSize = 1;

extern ClassStar DatakitURPChannelClass;
class DatakitURPChannel : public DatakitChannel {
protected:

	// Transmitter variables
	int		service;		// quality of service provided
	int		xmitSeq;		// next number to send
	int		ackedSeq;		// last sequence number acknowledged
	Lock  		*xmitLock;
	int		writerWaiting;		// a writer is waiting to write
	SemaphoreStar	xmitWaitingSemaphore;	// sleep till window frees up
	DatakitXmitBuf	rexmitBuf[SEQ_NO_MASK + 1]; // for retransmission in gos 5
	int		xmitState;

	// Receiver variables
	DatakitPacketQueue	*readQ;		// packets ready to be read
	DatakitPacketQueue	*stage;		// packets waiting for SEQ
	Lock			*recvLock;
	unsigned char		recvSeq;	// last correctly received seq
	unsigned char		echo;		// last ECHO or REJ sent
	char			receiveState;	// state of the reciever
	unsigned short		receiveLen;	// bytes received
	int			receiveTrailers;// in INIT1 mode

#ifdef DATAKITSTATS
	int firststat;
	int statRecvError;
	int statInit0;
	int statInit1;
	int statInitreq;
	int statAinit;
	int xmitBuf;
	int saveXmitBuf;
	int wastedCopy;
	int statBot;
	int statSeq;
	int statEnq;
	int recvGoodBlock;
	int recvBadBlock;
	int xmitGoodBlock;
	int xmitBadBlock;
	int statEcho;
	int statAck;
	int statRej;
	int hog;
	int x_initreq;
	int x_init;
	int laststat;
#endif

	int getFreeChannel();
	void resetSeq();
	int canXmit();
	void receiveSeq(int);
	void accept(int);
	void reject(int);
	void ackTo(int);
	void freeWriters();
	void retransmit(char);
	void sendBlock(int seq, unsigned char *buf, int len, int endofblock);
	void sendControl(int);
	void clearBuf(DatakitXmitBuf *);
	void startTimer();
	void receiveEcho(int echoOrRej, int seq);

public:
	proxiable DatakitURPChannel(int channel = 0, int GradeOfService = 5);
	proxiable ~DatakitURPChannel();

	proxiable virtual int Write(unsigned char *, int);
	proxiable virtual int Read(unsigned char *, int);

	// the following are to be called only from the interrupt handler
	int ReceivePacket(DatakitPacket *);
	void ReceiveControl(int);

	// called only from timer
	static void timerHandler(DatakitURPChannel *);

	Class * classOf();
};

#endif  /* DatakitURPChannel_h */
