/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: See-Mong Tan (stan@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * PCi586Disk.h: PCi586 Disk Driver
 */

#ifndef PCi586Disk_h
#define PCi586Disk_h
#ifdef CFRONT
#define CONST const
#define STATIC static
#else /* must be g++ */
//#pragma interface
#if (__GNUG__ >= 2)
#if (!defined CONST)
#define CONST
#define STATIC
#endif
#else
#pragma once
#define CONST const
#define STATIC static
#endif
#endif

/* include stuff we need */
#include "Disk.h"
#include "Semaphore.h"

class PCi586DiskBackingStore;     // forward

/*
 * PC AT Hard disk controller definitions.
 */

#define	HD0	0x1f0	/* base for hard controller I/O port addresses */
#define	FDR	0x3f6	/* I/O port address for fixed disk register */

/*
 * Bit 3 of the fixed disk register must be set to 1 to access heads
 * 8 - 15 of a hard disk.
 */
#define	HD_EXTRAHDS	0x08	/* set into FDR to access high disk heads */
#define	HD_NOEXTRAHDS	0x00	/* set into FDR if no high disk heads */

/*
 * port offsets from base above.
 */
#define	HD_DATA		0x00	/* data register */
#define	HD_ERROR	0x01	/* error register/write precomp */
#define	HD_PRECOMP	0x01	/* error register/write precomp */
#define	HD_NSECT	0x02	/* sector count */
#define	HD_SECT		0x03	/* sector number */
#define	HD_LCYL		0x04	/* cylinder low byte */
#define	HD_HCYL		0x05	/* cylinder high byte */
#define	HD_DRV		0x06	/* drive/head register */
#define	HD_STATUS	0x07	/* status/command register */
#define	HD_CMD		0x07	/* status/command register */

/*
 * Status bits
 */
#define	HD_STATUS_BUSY		0x80	/* controller busy */
#define	HD_STATUS_READY		0x40	/* drive ready */
#define	HD_STATUS_WRFAULT	0x20	/* write fault */
#define	HD_STATUS_SEEKDONE	0x10	/* seek operation complete */
#define	HD_STATUS_DATARQ	0x08	/* data request */
#define	HD_STATUS_ECC		0x04	/* ECC correction applied */
#define	HD_STATUS_INDEX		0x02	/* disk revolution index */
#define HD_STATUS_ERROR		0x01	/* error flag */

/*
 * Drive selectors
 */
#define	HD_DHFIXED	0xa0	/* bits always set in drive/head reg. */
#define	HD_DRIVE0	0x00	/* or into HD_DHFIXED to select drive 0 */
#define	HD_DRIVE1	0x10	/* or into HD_DHFIXED to select drive 1 */

/*
 * Hard disk commands. 
 */
#define	HD_RESTORE	0x10	/* restore cmd, bottom 4 bits set step rate */
#define	HD_SEEK		0x70	/* seek cmd, bottom 4 bits set step rate */
#define	HD_RDSEC	0x20	/* read sector cmd, bottom 2 bits set ECC and
					retry modes */
#define	HD_WRSEC	0x30	/* write sector cmd, bottom 2 bits set ECC and
					retry modes */
#define	HD_FORMAT	0x50	/* format track command */
#define	HD_RDVER	0x40	/* read verify cmd, bot. bit sets retry mode */
#define	HD_DIAG		0x90	/* diagnose command */
#define	HD_SETPARAM	0x91	/* set parameters command */

#define	HDTIMOUT	25000	/* how many 10usecs in a 1/4 sec.*/

#define NUMDRV  2	/* maximum number of drives */
#define SECSIZE 512	/* default sector size */
#define SECSHIFT 9
#define SECMASK (SECSIZE-1)
#define cylin   av_back

/* Values of hd_state */
#define HD_OPEN		0x01	/* drive is open */
#define HD_OPENING	0x02	/* drive is being opened */
#define HD_DO_RST	0x04	/* hardware restore command should be issued */
#define HD_DO_FMT	0x08	/* track is being formatted */
#define HD_VTOC_OK	0x10	/* VTOC (pdinfo, vtoc, alts table) OK */
#define HD_FMT_RST	0x20	/* restore needs to happen before format */
#define HD_BADBLK	0x40	/* bad block is being remapped */
#define HD_BBH_VFY	0x0080	/* potential bad block is being verified */
#define HD_BBH_MAP	0x0100	/* bad block is being assigned an alt.  */
#define HD_DO_VFY	0x0200	/* Sector(s) being verified.            */
#define HD_BADTRK	0x0400	/* bad block is in alt trk area being remapped */

/*
 * Logical blocks to physical blocks
 */
#define lbtopb(lb) (((lb) << BSHIFT) >> SECSHFT)
#define pbtolb(pb) (((pb) << SECSHFT) >> BSHIFT)

#define	DAM_NOT_FOUND	0x01	/* Data Address Mark not found   */
#define	TR000_ERR	0x02	/* Track 0 not found             */
#define	ABORTED		0x04	/* Command Aborted               */
#define	ID_NOT_FOUND	0x10	/* Sector ID not found           */
#define	ECC_ERR		0x40	/* Uncorrectable data read error */
#define	BAD_BLK		0x80	/* Bad block flag detected       */

// Sector number on disk where Physical Description Area is kept
// relative to start of UNIX partition
const int PCi586DiskPhysicalDescriptionAreaOffset = 29;

/* We allow these operations */
const int PCi586DiskReadOperation = 1;
const int PCi586DiskWriteOperation = 2;

/*
 * NOTE: This code assumes the disk structure and especially the layout is in
 * one-to-one phsyical memory.
 */

/* structure and data about the master boot block */

#define BOOTSZ		446	/* size of boot code in master boot block */
#define FD_NUMPART	4	/* number of 'partitions' in fdisk table */
#define MBB_MAGIC	0xAA55	/* magic number for mboot.signature */
#define DEFAULT_INTLV	4	/* default interleave for testing tracks */
#define MINPSIZE	4	/* minimum number of cylinders in a partition */
#define TSTPAT		0xE5	/* test pattern for verifying disk */

/*
 * Class structure to hold the fdisk partition table
 * This is initial partition under DOS - the entire UNIX
 * partition is given in the volume table of contents later,
 * in the physical description area.
 */
class PCi586DiskInitialPartitionTable {
friend class PCi586Disk;
 protected:
    unsigned char bootid;	/* bootable or not */
    unsigned char beghead;	/* beginning head, sector, cylinder */
    unsigned char begsect;	/* begcyl is a 10-bit number. High 2 bits */
    unsigned char begcyl;	/*     are in begsect. */
    unsigned char systid;	/* OS type */
    unsigned char endhead;	/* ending head, sector, cylinder */
    unsigned char endsect;	/* endcyl is a 10-bit number.  High 2 bits */
    unsigned char endcyl;	/*     are in endsect. */
    long    relsect;		/* first sector relative to start of disk */
    long    numsect;		/* number of sectors in partition */
};
/*
 * Values for bootid.
 */
#define BOOTID_NOTACTIVE	0
#define BOOTID_ACTIVE		128
/*
 * Values for systid.
 */
#define DOSOS12		1	/* DOS partition, 12-bit FAT */
#define PCIXOS		2	/* PC/IX partition */
#define DOSDATA		86	/* DOS data partition */
#define DOSOS16		4	/* DOS partition, 16-bit FAT */
#define EXTDOS		5	/* EXT-DOS partition */
#define OTHEROS		98	/* part. type for appl. (DB?) needs raw partition */
				/* ID was 0 but conflicted with DOS 3.3 fdisk    */
#define UNIXOS		99	/* UNIX V.x partition */
#define UNUSED		100	/* unassigned partition */
#define MAXDOS		65535L	/* max size (sectors) for DOS partition */

/*
 * structure to hold master boot block in physical sector 0 of the disk.
 * Note that partitions stuff can't be directly included in the structure
 * because of lameo '586 compiler alignment design.
 */
class  PCi586DiskMasterBootBlock {
friend class PCi586Disk;
 protected:
    char             bootinst[BOOTSZ];
    char             parts[FD_NUMPART * 
			   sizeof(PCi586DiskInitialPartitionTable)];
    unsigned short   signature;
public:
    void operator=( char * buffer );
};

inline
void
PCi586DiskMasterBootBlock::operator=( char * buffer )
{
    unsigned int i = 0;
    char * s = (char *) this;

    while( i < sizeof(PCi586DiskMasterBootBlock) )
	  *s++ = buffer[i++];
}


#define V_NUMPAR        16              /* maximum number of partitions */

#define VTOC_SANE       0x600DDEEE      /* Indicates a sane VTOC */
#define V_VERSION       0x01            /* layout version number */

/* Partition identification tags */
#define V_BOOT		0x01		/* Boot partition */
#define V_ROOT		0x02		/* Root filesystem */
#define V_SWAP		0x03		/* Swap filesystem */
#define V_USR		0x04		/* Usr filesystem */
#define V_BACKUP	0x05		/* full disk */
#define V_ALTS          0x06            /* alternate sector space */
#define V_OTHER         0x07            /* non-unix space */
#define V_ALTTRK	0x08		/* alternate track space */

/* Partition permission flags */
#define V_UNMNT		0x01		/* Unmountable partition */
#define V_RONLY         0x10            /* Read only (except by IOCTL) */
#define V_OPEN          0x100           /* Partition open (for driver use) */
#define V_VALID         0x200           /* Partition is valid to use */
#define V_VOMASK        0x300           /* mask for open and valid */

/* Sanity word for the physical description area */
#define VALID_PD		0xCA5E600D

class PCi586DiskPartitionHeader {
friend class PCi586Disk;
friend class PCi586DiskContainer;
friend class PCi586Kernel;
 protected:
    unsigned short p_tag;	/*ID tag of partition*/
    unsigned short p_flag;	/*permision flags*/
    long p_start;         	/*physical start sector no of partition*/
    long p_size;		/*# of physical sectors in partition*/
};

class PCi586DiskVolumeTableOfContents {
friend class PCi586Disk;
friend class PCi586DiskContainer;
friend class PCi586Kernel;
 protected:
    unsigned long v_sanity;	/*to verify vtoc sanity*/
    unsigned long v_version;	/*layout version*/
    char v_volume[8];		/*volume name*/
    unsigned short v_nparts;		/*number of partitions*/
    unsigned short  v_pad;		/*pad for 286 compiler*/
    unsigned long v_reserved[10]; /*free space*/
    PCi586DiskPartitionHeader v_part[V_NUMPAR]; /*partition headers*/

public:
    void operator=( char * );
};

inline
void
PCi586DiskVolumeTableOfContents::operator=( char buffer[] )
{
    unsigned int i = 0;
    char * s = (char *) this;
    
    while( i < sizeof(PCi586DiskVolumeTableOfContents) )
	*s++ = buffer[i++];
}

class PCi586DiskPhysicalDescriptionArea {
friend class PCi586Disk;

 protected:
    unsigned long driveid;	/*identifies the device type*/
    unsigned long sanity;	/*verifies device sanity*/
    unsigned long version;	/*version number*/
    char serial[12];		/*serial number of the device*/
    unsigned long cyls;		/*number of cylinders per drive*/
    unsigned long tracks;	/*number tracks per cylinder*/
    unsigned long sectors;	/*number sectors per track*/
    unsigned long bytes;	/*number of bytes per sector*/
    unsigned long logicalst;	/*sector address of logical sector 0*/
    unsigned long errlogst;	/*sector address of error log area*/
    unsigned long errlogsz;	/*size in bytes of error log area*/
    unsigned long mfgst;	/*sector address of mfg. defect info*/
    unsigned long mfgsz;	/*size in bytes of mfg. defect info*/
    unsigned long defectst;	/*sector address of the defect map*/
    unsigned long defectsz;	/*size in bytes of defect map*/
    unsigned long relno;	/*number of relocation areas*/
    unsigned long relst;	/*sector address of relocation area*/
    unsigned long relsz;	/*size in sectors of relocation area*/
    unsigned long relnext;	/*address of next avail reloc sector*/

    /* added for the 80586 port */
    unsigned long vtoc_ptr;  /*byte offset of vtoc block*/
    unsigned short vtoc_len; /*byte length of vtoc block*/
    unsigned short vtoc_pad; /* pad for 16-bit machine alignment */
    unsigned long alt_ptr;   /*byte offset of alternates table*/
    unsigned short alt_len;  /*byte length of alternates table*/

public:
    void operator=( char * );
};

inline
void
PCi586DiskPhysicalDescriptionArea::operator=(char buffer[])
{
    unsigned int i=0;
    char * s = (char *) this;

    while( i < sizeof(PCi586DiskVolumeTableOfContents) )
	*s++ = buffer[i++];
}

/*
 * Defines the bad block table for the hard disk driver.
 * The same table structure is used for the bad track table.
 */

#define MAX_ALTENTS     253	/* Maximum # of slots for alts	*/
				/* allowed for in the table.	*/

#define ALT_SANITY      0xdeadbeef      /* magic # to validate alt table */
#define ALT_VERSION	0x02		/* version of table 		 */

class PCi586DiskAltTable {
public:
    unsigned short alt_used;	/* # of alternates already assigned	*/
    unsigned short alt_reserved;   /* # of alternates reserved on disk	*/
    long alt_base;	/* 1st sector (abs) of the alt area	*/
    long alt_bad[MAX_ALTENTS];	/* list of bad sectors/tracks	*/
};

class PCi586DiskAltInfo {	/* table length should be multiple of 512    */
public:
    long    alt_sanity;	/* to validate correctness		*/
    unsigned short  alt_version;     /* to corroborate vintage		*/
    unsigned short  alt_pad;	/* padding for alignment		*/
    PCi586DiskAltTable alt_trk;	/* bad track table	*/
    PCi586DiskAltTable alt_sec;	/* bad sector table	*/
};

//inline
//void
//bcopy( char *src, char * dest, int len )
//{
//    for(int i=0; i < len; i++)
//	*dest++ = *src++;
//}

/*
 * A class to represent each disk drive as a PersistentMemoryObject.
 */
extern ClassStar PCi586DiskClass;
class PCi586Disk : public Disk {
friend class PCi586Kernel;
friend class PCi586DiskContainer;
friend class PCi586DiskBackingStore;
    
    Lock mutex;

    // Drive specific information for this drive
    int myDrive;
    unsigned short   myCylinders;
    unsigned char    myHeads;
    unsigned char    mySectorsPerTrack;
    unsigned char    myPrecomp;

    // The structure in absolute sector 0 of the disk
    // Use it to find out which is the active UNIX partition
    PCi586DiskMasterBootBlock  mboot;
    
    // We need to read this in to find where the volume table of
    // contents is, where we can find the UNIX partition information
    PCi586DiskPhysicalDescriptionArea myPDA;

    // The Volume Table of Contents
    PCi586DiskVolumeTableOfContents myVTOC;

    // Alternate and bad sector/track information
    PCi586DiskAltInfo myAltInfo;

    // Bits are for tracks with at least one bad block.
    // Initialized with info obtained from the alternate 
    // sector table. bit map is later updated by BBH
    // each time a sector is detected as bad.
    char * bbh_trkmap;
    // Bits are set for all tracks recorded in the bad
    // track tabl" and those recorded in the bbh_trkmap.
    char * bth_trkmap;

    // The beginning and end of the whole UNIX partition
    // relative to the start of the disk
    int  startOfUnixPartition;       // in sectors
    int  sizeOfUnixPartition;        // in sectors
    int  beginTrackOfUnixPartition;  // in tracks

 protected:

#ifdef CFRONT
    int initDone;
#else
    volatile int initDone;
#endif

	      // do io on disk
    int doio ( unsigned int start, int count, char * addr, int nbytes, int op );
    void initialize();                         // initialize the disk driver
    unsigned char waitForControllerReady();    // return status of controller

    // for alternate sector/table information
    void bitSet( char * bitmap, long bitno );
    int bitTest( char * bitmap, long bitno );
    int tableScan( long * table, long value, unsigned short length );
    unsigned int badBlockMap( unsigned int block );

 public:

    Class * classOf();

    PCi586Disk( int );

    int read( unsigned int start, int count, char * addr, int nbytes );
    int write( unsigned int start, int count, char * addr, int nbytes );

    int basicDoio( unsigned int, char *, int );
    int reset();
    int partitions();
    int locatePartition( int idNumber, int &start, int &units );
};

// 1 meg in 512 byte blocks
#define sizeOfSingleSwapPartition 512

// a class for a backing store
extern ClassStar PCi586DiskBackingStoreClass;
class PCi586DiskBackingStore : public PersistentMemoryObject {
protected:
    static int initialized;
    static int wholePartitionStart;
    static int wholePartitionSize;
    static int availablePartitions;
    static int allocationArray[ 200 ];
    
    Lock mutex;
    int    startOfThisSwapPartition;
    int    sizeOfThisSwapPartition;
    PCi586Disk * theDisk;

    int rawRead( unsigned start, unsigned count,
		  char * addr, unsigned nbytes, IOAsyncClosure * c ) {}
    int rawWrite( unsigned start, unsigned count,
		  char * addr, unsigned nbytes,
		   int writeThrough, IOAsyncClosure * c ) {}
    
public:
    PCi586DiskBackingStore( PCi586Disk * source, int start, int size );
    ~PCi586DiskBackingStore();
    Class * classOf();
    void initialize( int partitionStart, int partitionSize );
    int read( unsigned int start, int count, char * addr, int nbytes );
    int write( unsigned int start, int count, char * addr, int nbytes );
    int setNumberOfUnits( unsigned int );
};


#endif  /* PCi586Disk_h */
