/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef Am7990_h
#define Am7990_h

#ifdef __GNUC__
#pragma interface
#endif


/*
 * Am7990 Lance ethernet driver superclass.
 */

#include "Am7990MachineConstants.h"
#include "Assert.h"
#include "SemaphoreStar.h"
#include "Types.h"
#include "EthernetDriver.h"
class Lock;
class NetworkBufferFreeList;

/***************************************************************************
 * Constants based on Am7990MachineContstants.h in the MI includes area
 ***************************************************************************/
const unsigned int ReceiveBuffers = 1 << Log2ReceiveBuffers;
const unsigned int TransmitBuffers = 1 << Log2TransmitBuffers;
const unsigned int MaxPacket = 1600;

typedef unsigned short Am7990Word;

/***************************************************************************
 * Am7990 register definitions
 ***************************************************************************/
class Am7990Reg {
private:
volatile Am7990Word *address;
public:
	Am7990Reg(Am7990Word *p) : address(p) {*p = 0;} 
	Am7990Word read();
	void write(Am7990Word);
};

inline
Am7990Word
Am7990Reg::read()
{
	unsigned short retval = *((unsigned short *)address);
	return retval;
}

inline
void
Am7990Reg::write(Am7990Word val)
{
	*address = val;
}

/***************************************************************************
 * Am7990 Buffer descriptor ring class
 ***************************************************************************/
class Am7990BufferDescriptor {
private:
	volatile Am7990Word rmd[4];
public:
	Am7990BufferDescriptor();
	Am7990BufferDescriptor(const Am7990BufferDescriptor &){} // shutup compiler
	~Am7990BufferDescriptor(){}
	void disown();			// Give up drivers rite of onwership
	int ownedByCPU();		// Test who owns it
	char *buffer();			// return address of buffer
	unsigned int length();		// length of data written by chip
	unsigned int status();		// return the status bits.

	void setBuffer(NetworkBuffer *); // Set address, length of buffer
	void clearError();		// Clear all error bits
	void clearTransmitError();	// Clear all error bits
	void setTransmitBits();		// Clear all error bits
	unsigned int error();		// return error bits
	unsigned int transmitError();	// return transmitter error bits
	int lastBuffer();			// Last buf for packet
	int firstBuffer();			// First buf for pckt.
	void clearMessage();		// Clear message length
};

inline
Am7990BufferDescriptor:: Am7990BufferDescriptor(){
	rmd[0] = (Am7990Word)0;
	rmd[1] = (Am7990Word)0;
	rmd[2] = (Am7990Word)0;
	rmd[3] = (Am7990Word)0;
}

inline
unsigned int
Am7990BufferDescriptor::status()
{
	return ( (rmd[1] >> 8) & 0xff);
}

inline
int
Am7990BufferDescriptor::lastBuffer()
{
	return (rmd[1] & 0x0100);
}

inline
int
Am7990BufferDescriptor::firstBuffer()
{
	return (rmd[1] & 0x0200);
}

inline
unsigned int
Am7990BufferDescriptor::error()
{
	return (rmd[1] & 0x4000);
}

inline
unsigned int
Am7990BufferDescriptor::transmitError()
{
	return (rmd[3] & 0x8c00);
}

inline
void
Am7990BufferDescriptor::clearError()
{
	rmd[1] &= ~0x7f00;
}

inline
void
Am7990BufferDescriptor::clearTransmitError()
{
	rmd[3] &= ~0x8c00;
}

inline
void
Am7990BufferDescriptor::disown()
{
	rmd[1] |= 0x8000;
}

inline
int 
Am7990BufferDescriptor::ownedByCPU()
{
	return ( (rmd[1] & 0x8000) != 0x8000 );
}

inline
char *
Am7990BufferDescriptor::buffer()
{
	char * retval = (char*)(((rmd[1] & 0xff) << 16) | rmd[0]);
	return retval;
}

inline
unsigned int
Am7990BufferDescriptor::length()
{
	unsigned int retval = ( rmd[3] & 0xfff);

	retval -= 4; // Take off for the CRC bytes at the end

	return retval;
}

inline
void
Am7990BufferDescriptor::setTransmitBits()
{
	rmd[1] |= 0x0300;
}

inline
void
Am7990BufferDescriptor::clearMessage()
{
	rmd[3] = 0;
}

/***************************************************************************
 * THe Am7990 LANCE Ethernet Controller
 * Initialization takes place thusly:
 * 	1. Subclass constructor creates ring and initBlock and
 *	initializes init block.
 *	2.  Subclass constructor calls init, return value
 *	specifies success.
 *	3.  Subclass constructor does anything necessary before
 *	chip is started.
 *	4.  Subclass constructor call startChip and returns.
 ***************************************************************************/
extern ClassStar Am7990Class;

class Am7990 : public EthernetDriver {
private:
	/*
	 * Instance vars
	 */
	Am7990Reg			_rap;
	Am7990Reg			_rdp;
	void 				transmitInterrupt();
	void 				receiveInterrupt();
	/* Transmit semaphores to queue the processes on until 
	 * the ring is open
	 */
	SemaphoreStar 			_transmitSemaphores[TransmitBuffers];
	/*
	 * Transmit completion semaphores and error return values
	 * since semaphores can't return values (yet).
	 * Use semaphores instead of spinning in case we get time sliced.
	 */
	SemaphoreStar 			_completionSemaphores[TransmitBuffers];
	unsigned char			_transmitErrors[TransmitBuffers];
	unsigned char			_transmitAsync[TransmitBuffers];
	NetworkBuffer			*_transmitBufs[TransmitBuffers];
	/*
	 * A lock to keep mutual exclusion between multiple
	 * processes updating _currentTransmitBuffer
	 */
	Lock * 				_transmitMutex;
	Lock * 				_interruptMutex;
	/*
	 * Mutual exclusion and reference counts for multicasts.
	 */
	Semaphore			*_multicastMutex;
	int				_multicastCount[MulticastAddresses];

protected:
	// for performance statistics
	unsigned startHandler;
	unsigned dmaFlushTime;
	unsigned fTimes;
// send
	unsigned sendMapDMA;
	unsigned sTimes;
	unsigned sendUnMapDMA;
	unsigned sUTimes;
// send interrupt

	unsigned sendITime;
	unsigned sITimes;
// recv interrupt

	unsigned rcvITime;
	unsigned rITimes;
// recv
	unsigned rcvUnMapDMA;
	unsigned rTimes;
	unsigned rcvBuf;
	unsigned rBufTimes;

	unsigned int 			_currentReceiveBuffer;
	/* _currentTransmitBuffer is the head of the driver ring */
	unsigned int 			_currentTransmitBuffer;
	/* _previousTransmitBuffer is where we expect the chip to be */
	unsigned int 			_previousTransmitBuffer;
	unsigned int			_missedPackets;
	unsigned int			_inputPackets;
	unsigned int			_inputErrors;
	unsigned int			_outputPackets;
	unsigned int			_outputErrors;
	unsigned int			_interrupts;

	/*
	 * NetworkBuffers given to us by the administrator and
	 * pre-fetched for use in the DMA buffer ring.
	 */
	NetworkBuffer **		_receiveNetworkBuffer;

private:
	unsigned int nextTransmitDescriptor(unsigned int current);
	unsigned int nextReceiveDescriptor(unsigned int current);
	int doTransmit(NetworkBuffer *buf, int async);

protected:
	/*
	 * These are set/used in subclass constructor
	 */
	Am7990Word *			_initBlock;	// set
	Am7990BufferDescriptor *	_receiveRing;	// set
	NetworkBuffer **		_recvBufs;	
	Am7990BufferDescriptor *	_transmitRing;	// set
	int 				_initialized;	// used
	void startChip();				// used
	void initChip();				// used
	void initBuffers();				// used

	/*
	 * These are what your subclass needs to define:
	 *
	 * The allocation routines allocate mem in DMA space and return
	 * the elements vaddrs.
	 * Your constructor allocates the init block, ring descriptors,
	 * initializes all the parts of the initblock, and sets _receiveRing,
	 * _transmitRing, and _initBlock.
	 */

	/* 
	 * getBuffer allocates a packet for recv ring from the administator.
	 * releaseBuffer releases a packet in the transmit ring.
	 */
	virtual NetworkBuffer*	getBuffer(unsigned int len);
	virtual void		releaseBuffer(NetworkBuffer *buf);

public:
	Am7990(volatile Am7990Word *rap,
	       volatile Am7990Word *rdp,
	       unsigned char debugMask = 0,NetworkBufferFreeList *list=0);
	virtual ~Am7990();
	static void disableChip( volatile Am7990Word *rap, volatile Am7990Word *rdp);
	virtual void interruptHandler();	// The interrupt handler
	virtual int transmitPacket(NetworkBuffer *);// To send a packet
	virtual int transmitPacketAsync(NetworkBuffer *);// To send a packet
	virtual void printInfo();		// Print stats to console
	virtual int  addMulticast(EthernetAddress *);
	virtual int  removeMulticast(EthernetAddress *);

	virtual void printStats();
	virtual void clearStats();
	virtual void flushInitBlock()=0;		
    	Class * classOf();
};

const int	CSR0		= 0;
const int	CSR1		= 1;
const int	CSR2		= 2;
const int	CSR3		= 3;

#define	LE_ERR		0x8000		/* RO BABL | CERR | MISS | MERR */
#define	LE_BABL		0x4000		/* RC transmitted too many bits */
#define	LE_CERR		0x2000		/* RC No Heartbeat */
#define	LE_MISS		0x1000		/* RC Missed an incoming packet */
#define	LE_MERR		0x0800		/* RC Memory Error; no acknowledge */
#define	LE_RINT		0x0400		/* RC Received packet Interrupt */
#define	LE_TINT		0x0200		/* RC Transmitted packet Interrupt */
#define	LE_IDON		0x0100		/* RC Initialization Done */
#define	LE_INTR		0x0080		/* RO BABL|MISS|MERR|RINT|TINT|IDON */
#define	LE_INEA		0x0040		/* RW Interrupt Enable */
#define	LE_RXON		0x0020		/* RO Receiver On */
#define	LE_TXON		0x0010		/* RO Transmitter On */
#define	LE_TDMD		0x0008		/* W1 Transmit Demand (send it now) */
#define	LE_STOP		0x0004		/* RW1 Stop */
#define	LE_STRT		0x0002		/* RW1 Start */
#define	LE_INIT		0x0001		/* RW1 Initialize */



/* Bits common to both rmds and tmds */
#define	LMD_OWN		0x80	/* Chip owns the descriptor */
#define	LMD_ERR		0x40	/* Error occurred */
#define	LMD_STP		0x02	/* Start of Packet */
#define	LMD_ENP		0x01	/* End of Packet */

/* Bits in rmd flags */
#define	RMD_FRAM	0x20	/* Framing error */
#define	RMD_OFLO	0x10	/* Internal Silo Overflowed. Valid if !ENP */
#define	RMD_CRC		0x08	/* CRC Error */
#define	RMD_BUFF	0x04	/* Didn't have a buffer for the packet */
/* bits in tmd flags */
#define	TMD_MORE	0x10	/* More than one retry was needed */
#define	TMD_ONE  	0x08	/* Exactly One Retry, valid only if !LCOL */
#define	TMD_DEF  	0x04	/* Deferred (net was initially busy) */

/* Bits for lmd_errflags */
#define	TMD_BUFF 0x8000		/* Buffer Error (imples underflow too) */
#define	TMD_UFLO 0x4000		/* Underflow Error */
#define	TMD_LCOL 0x1000		/* Late Collision */
#define	TMD_LCAR 0x0800		/* Loss of Carrier */
#define	TMD_RTRY 0x0400		/* More than 16 Retry's */
#define	TMD_TDR	 0x003f		/* Time Domain Reflectometry counter mask */

inline
unsigned int
Am7990::nextTransmitDescriptor(unsigned int current)
{
	unsigned int retval = (current + 1) % TransmitBuffers;
	return retval;
}

inline
unsigned int
Am7990::nextReceiveDescriptor(unsigned int current)
{
	unsigned int retval = (current + 1) % ReceiveBuffers;
	return retval;
}


#endif /* Am7990_h */
