/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
#ifndef ESP_h
#define ESP_h

#ifdef __GNUC__
#pragma interface
#endif


#include "SCSIDevice.h"

/*                ESP Register Offsets                           */
 unsigned char const LO_XFER_REG_OFFSET    = 0x0;
 unsigned char const HI_XFER_REG_OFFSET    = 0x1;
 unsigned char const FIFO_REG_OFFS         = 0x2;
 unsigned char const COMMAND_REG_OFFS      = 0x3;
 unsigned char const STATUS_REG_OFFS       = 0x4;
 unsigned char const BUS_ID_REG_OFFS       = 0x4;
 unsigned char const INTR_STATUS_REG_OFFS  = 0x5;
 unsigned char const TIMEOUT_REG_OFFS      = 0x5;
 unsigned char const SEQ_STEP_REG_OFFS     = 0x6;
 unsigned char const PERIOD_REG_OFFS       = 0x6;
 unsigned char const FIFO_FLAGS_REG_OFFS   = 0x7;
 unsigned char const OFFSET_REG_OFFS       = 0x7;
 unsigned char const CONFIG_REG_OFFS       = 0x8;
 unsigned char const RESERVA_REG_OFFSET    = 0x9;
 unsigned char const CLOCK_FCT_REG_OFFSET  = 0x9;
 unsigned char const RESERVB_REG_OFFSET    = 0xa;
 unsigned char const TEST_REG_OFFSET       = 0xa;

/*                ESP Commands                                   */

/* Command Register (RW) - Misc Commands */
 unsigned char const CMD_NOP             = 0x00;
 unsigned char const CMD_FLUSH_FIFO      = 0x01;
 unsigned char const CMD_RESET_CHIP      = 0x02;
 unsigned char const CMD_RESET_SCSI      = 0x03;
 unsigned char const DMA_NOP             = 0x80;
 unsigned char const DMA_FLUSH_FIFO      = 0x81;
 unsigned char const DMA_RESET_CHIP      = 0x82;
 unsigned char const DMA_RESET_SCSI      = 0x83;

/* Command Register (RW) - Disconnected Commands */
 unsigned char const CMD_RECONNECT       = 0x40;
 unsigned char const CMD_SEL_WO_ATN      = 0x41;
 unsigned char const CMD_SEL_W_ATN       = 0x42;
 unsigned char const CMD_SEL_W_ATN_STOP  = 0x43;
 unsigned char const CMD_ENABLE_SR       = 0x44;
 unsigned char const CMD_DISABLE_SR      = 0x45;
 unsigned char const DMA_RECONNECT       = 0xc0;
 unsigned char const DMA_SEL_WO_ATN      = 0xc1;
 unsigned char const DMA_SEL_W_ATN       = 0xc2;
 unsigned char const DMA_SEL_W_ATN_STOP  = 0xc3;
 unsigned char const DMA_ENABLE_SR       = 0xc4;
 unsigned char const DMA_DISABLE_SR      = 0xc5;

/* Target Commands */
 unsigned char const DMA_SEND_DATA 	 = 0xa2;
 unsigned char const DMA_RECEIVE_DATA 	 = 0xaa;

/* Command Register (RW) - Initiator Commands */
 unsigned char const CMD_TRANSFER_INFO   = 0x10;
 unsigned char const CMD_COMPL_SEQ       = 0x11;
 unsigned char const CMD_ACCEPT_MSG      = 0x12;
 unsigned char const CMD_TRANSFER_PAD    = 0x18;
 unsigned char const CMD_SET_ATN         = 0x1a;
 unsigned char const DMA_TRANSFER_INFO   = 0x90;
 unsigned char const DMA_COMPL_SEQ       = 0x91;
 unsigned char const DMA_ACCEPT_MSG      = 0x92;
 unsigned char const DMA_TRANSFER_PAD    = 0x98;
 unsigned char const DMA_SET_ATN         = 0x9a;

/* Configuration Register Bits (RW) */
 unsigned char const CF_BUS_ID           = 0x07;
 unsigned char const CF_PARITY_ENABLE    = 0x10;
 unsigned char const CF_PARITY_TEST_M    = 0x20;
 unsigned char const CF_SCSI_RES_INT_DIS = 0x40;
 unsigned char const CF_SLOW_CABLE       = 0X80;

/* Register Masks */
// unsigned char const SEQ_STEP_REG_MASK     = 0x07;
 unsigned char const SYNC_PERIOD_REG_MASK  = 0x17;
 unsigned char const FIFO_FLAGS_REG_MASK   = 0x17;
 unsigned char const SYNC_OFFSET_REG_MASK  = 0x17;
 unsigned char const CLOCK_FACTOR_REG_MASK = 0x07;

/*                ESP Register Classes                           */
class ESPRegister {
 protected:
  volatile unsigned char * const myAddress;
 public:
  ESPRegister(volatile unsigned char * const address) : myAddress(address){}
  unsigned char read() ;
  void write(unsigned char byte);
};

inline unsigned char
ESPRegister::read()
{
	return *myAddress;
}

inline void
ESPRegister::write(unsigned char byte)
{
	*myAddress = byte;
}


class ESPDoubleRegister {
 protected:
  ESPRegister low;
  ESPRegister high;
 public:
  ESPDoubleRegister(unsigned char *loAddress, unsigned char *hiAddress) : low(loAddress), high(hiAddress) {}
  unsigned int read() { return ((high.read() << 8) | low.read()); }
  void write(unsigned int val) { high.write((val >> 8) & 0xff); low.write(val & 0xff); }

};

class ESPReadRegister : public ESPRegister {
 public:
  ESPReadRegister(unsigned char *address) : ESPRegister(address) {}
  unsigned char read() {return *myAddress;}
  void write(unsigned char) {Assert(NOTREACHED);}
};

class ESPWriteRegister : public ESPRegister {
 public:
  ESPWriteRegister(unsigned char *address) : ESPRegister(address) {}
  unsigned char read() { Assert(NOTREACHED);return 0;}
  void write(unsigned char c) {*myAddress = c;}
};


/*******************************************************************
 *******************************************************************
 **                                                               **
 **                ESP  (NCR53C90) Class                          **
 **                                                               **
 *******************************************************************
 *******************************************************************/

enum ESPState {
  S_NEGOTIATE,	// Negotiating parameters
  S_INIT, S_INIT_DMA, S_RESEL, S_DISCON, S_SEND_DATA, S_RECEIVE_DATA,
  S_FINISH_CMD, S_GOT_MSG, S_CMD_OUT, S_MSG_OUT, S_MSG_IN,
  SYNC_OUT, SYNC_IN, SYNC_IN2, SYNC_IN3
  };


class SCSICommand; 
class DMAController;
class ESPStatus;


class ESP : public SCSIController {
  friend class ESPStatus;
protected:
  ESPState currentState;
  DMAController *myDMA;
  ESPStatus *myStatus;
  ESPDoubleRegister *transferCounter;
  ESPRegister *fifo;
  ESPRegister *commandRegister;
  ESPReadRegister *statusRegister;
  ESPWriteRegister *busIdRegister;
  ESPReadRegister *intrStatusRegister;
  ESPWriteRegister *timeoutRegister;
  ESPReadRegister *seqStepRegister;
  ESPWriteRegister *periodRegister;
  ESPReadRegister *fifoFlagsRegister;
  ESPWriteRegister *offsetRegister;
  ESPRegister *configRegister;
  ESPWriteRegister *clockFactorRegister;
  ESPWriteRegister *testRegister;

  /*
   * Message buffers, pointers
   */
  const unsigned MAX_MSG = 10;
  unsigned char inmsg[MAX_MSG];
  unsigned inmsgCount;
  unsigned char outmsg[MAX_MSG];
  unsigned outmsgCount;

  const unsigned MAX_CMD = 20;
  unsigned char outcmd[MAX_CMD];
  unsigned char outcmdLength;

  void enableRstIntr();
  void disableRstIntr();
  unsigned char remainingBytesInFIFO();
  void setSyncTransferPeriod(unsigned char period);
  void setSyncOffset(unsigned char offset);
  void setClockFactor(int clockFactor);
  void checkSCSIError();
  void checkSCSIPhase();
  void sendData();
  void receiveData();
  void initControllerAction(SCSICommand &, unsigned char deviceId, char *buf,
	unsigned buflen);
  void 		initTarget(unsigned char unit);
  void abortCommand();
  /*
   * Overrides parent to print regs, then use parent.
   */
  unsigned      waitForCompletion();

  int acceptMessage(unsigned char);
void dumpRegisters();

 public:

  ESP(unsigned char *baseAddress, int step, DMAController *dma);
  virtual ~ESP();
  void initialize();
  void configure(unsigned char id, int parity=1,
		 int resetIntrDisable=0, int slow=0);
  int interrupt();	// Called from the Interrupt handler, returns completed
			// status
  void completed() { complete(); } // Called by interrupt handler
};

inline void
 ESP::enableRstIntr()
{
  configRegister->write(configRegister->read() & ~CF_SCSI_RES_INT_DIS);
}

inline void
 ESP::disableRstIntr()
{
  configRegister->write(configRegister->read() | CF_SCSI_RES_INT_DIS);
}


inline unsigned char
 ESP::remainingBytesInFIFO()
{
  return fifoFlagsRegister->read() & FIFO_FLAGS_REG_MASK;
}


inline void
 ESP::setSyncTransferPeriod(unsigned char period)
{
  periodRegister->write(period & SYNC_PERIOD_REG_MASK);
}

inline void
 ESP::setSyncOffset(unsigned char offset)
{
  offsetRegister->write(offset & SYNC_OFFSET_REG_MASK);
}

inline void
 ESP::setClockFactor(int clockFactor)
{
  clockFactorRegister->write(clockFactor & CLOCK_FACTOR_REG_MASK);
}
 
#endif /*ESP_h*/

