/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
#ifndef ESPStatus_h
#define ESPStatus_h


#ifdef __GNUC__
#pragma interface
#endif


unsigned char * const SLOT_BASE_ADDR = (unsigned char *) 0xf8000000;

/* SCSI Register offsets */

/* Status Register Bits (RO) */

const unsigned char STAT_IO             = 0x01;
const unsigned char STAT_CD             = 0x02;
const unsigned char STAT_MSG            = 0x04;
const unsigned char STAT_XFER_COMPLETE  = 0x08;
const unsigned char STAT_XFER_COUNT_0   = 0x10;
const unsigned char STAT_PARITY_ERR     = 0x20;
const unsigned char STAT_GROSS_ERR      = 0x40;
const unsigned char STAT_INT		= 0x80;


/* Interrupt Status Register Bits  (RO) */

const unsigned char ISTAT_SEL           = 0x01;
const unsigned char ISTAT_SEL_W_ATN     = 0x02;
const unsigned char ISTAT_RESEL         = 0x04;
const unsigned char ISTAT_FUNC_COMPL    = 0x08;
const unsigned char ISTAT_BUS_SERVICE   = 0x10;
const unsigned char ISTAT_DISCONNECT    = 0x20;
const unsigned char ISTAT_ILLEGAL_CMD   = 0x40;
const unsigned char ISTAT_SCSI_RESET    = 0x80;


/* Sequential Step Register Bits (RO) */
 
const unsigned char SEQ_STEP_0          = 0x01;
const unsigned char SEQ_STEP_1          = 0x02;
const unsigned char SEQ_STEP_2          = 0x04;
const unsigned char SEQ_STEP_MASK       = 0x07;

/* SCSI Bus Phases */

unsigned char const DATA_OUT_PHASE = 0x0;
unsigned char const DATA_IN_PHASE  = 0x1;
unsigned char const COMMAND_PHASE  = 0x2;
unsigned char const STATUS_PHASE   = 0x3;
unsigned char const MSG_OUT_PHASE  = 0x6;
unsigned char const MSG_IN_PHASE   = 0x7;

const unsigned char STAT_PHASE_MASK = 0x7;
const unsigned char SEQ_STEP_REG_MASK= 0x07;

class ESPStatus {
  ESP *myESP;
  unsigned char _status;
  unsigned char _intStatus;
  unsigned char _sequenceStep;
public:
  ESPStatus(ESP *);
  unsigned char status() { return _status;}
  unsigned char intStatus() { return _intStatus;}
  unsigned char sequenceStep() { return _sequenceStep;}
  void update();
  void partialUpdate();
  int resetDetected() {return (_intStatus & ISTAT_SCSI_RESET);}
  int illegalCommand() {return (_intStatus & ISTAT_ILLEGAL_CMD);}
  int disconnected() {return (_intStatus & ISTAT_DISCONNECT);}
  int busService() {return (_intStatus & ISTAT_BUS_SERVICE);}
  int functionComplete() {return (_intStatus & ISTAT_FUNC_COMPL);}
  int reselected() {return (_intStatus & ISTAT_RESEL);}
  int selectedWAtn() {return (_intStatus & ISTAT_SEL_W_ATN);}
  int selected() {return (_intStatus & ISTAT_SEL);}
  unsigned char phase() {return (_status & STAT_PHASE_MASK);}
  int intr() {return (_status & STAT_INT);}
  int parityError() {return (_status & STAT_PARITY_ERR);}
  int grossError() {return (_status & STAT_GROSS_ERR);}
  int transferCountZero() {return (_status & STAT_XFER_COUNT_0);}
  unsigned char sequence() {return (_sequenceStep & SEQ_STEP_MASK);}
};

inline
 ESPStatus::ESPStatus(ESP *espChip)
{
  myESP = espChip;
}

inline void
 ESPStatus::update(){
   /* Read the interupt status register last */
   _status = myESP->statusRegister->read();
   _sequenceStep = myESP->seqStepRegister->read() & SEQ_STEP_REG_MASK;
   _intStatus = myESP->intrStatusRegister->read();
 }

inline void 
 ESPStatus::partialUpdate(){
   /* Do not save sequence step register */
   _status = myESP->statusRegister->read();
   _intStatus = myESP->intrStatusRegister->read();
 }

#endif
