/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
#ifndef SCSIData_h
#define SCSIData_h


#ifdef __GNUC__
#pragma interface
#endif

class SCSIData {
 protected:
  unsigned char myLength;
  unsigned char *byteArray;
 public:
  SCSIData(unsigned char *data);
  unsigned char byte(int i);
};

inline 
 SCSIData::SCSIData(unsigned char *data) {
   byteArray = data;
 }

inline unsigned char 
 SCSIData::byte(int i) {
   return byteArray[i];
 }


/*************************************************************************
 *                                                                       *
 *                   Inquiry  Data                                       *
 *                                                                       *
 *************************************************************************/


class InquiryData: public SCSIData {
 public:
  InquiryData(unsigned char *data, int allocLength);
  unsigned char deviceType();
  unsigned char deviceTypeQualifier();
  unsigned char removableMedium();
  unsigned char isoVersion();
  unsigned char ecmaVersion();
  unsigned char ansiVersion();
  unsigned char vendorUnique(int i);
  unsigned	additionalLength() { return byte(4);}
};

inline int MIN (int x, int y) { return (x<y)?x:y;}

inline
 InquiryData::InquiryData (unsigned char *data, int allocLength): SCSIData(data)
{
  if (allocLength < 5)
    myLength = allocLength;
  else
    myLength = MIN(allocLength,byte(4)+5);
}
  
inline unsigned char
 InquiryData::deviceType()
{
  return byte(0);
}
      
inline unsigned char
 InquiryData::deviceTypeQualifier()
{
  return byte(1) & 0x7f;
}

inline unsigned char
 InquiryData::removableMedium()
{
  return (byte(1) & 0x80) >> 7;
}

inline unsigned char
 InquiryData::isoVersion()
{
  return (byte(2) & 0x60) >> 6;
}

inline unsigned char
 InquiryData::ecmaVersion()
{
  return (byte(2) & 0x38) >> 3;
}

inline unsigned char
 InquiryData::ansiVersion()
{
  return (byte(2) & 0x03);
}

inline unsigned char
 InquiryData::vendorUnique(int i)
{
 return i+5<=myLength && byte(i+5);
}


/*************************************************************************
 *                                                                       *
 *                         Sense Data                                    *
 *                                                                       *
 *************************************************************************/

const unsigned char AD_VALID      = 0x80;
const unsigned char ERROR_CLASS   = 0x70; 
const unsigned char ERROR_CODE    = 0x0f;
const unsigned char VEND_UNIQUE   = 0xe0;
const unsigned char FILE_MARK     = 0x80;
const unsigned char END_OF_MED    = 0x40;
const unsigned char INC_LEN_IND   = 0x20;
const unsigned char SENSE_KEY     = 0x0f;

class SenseData : public SCSIData {
 public:
  SenseData(unsigned char *data, int allocLength);
  int isInExtendedFormat() {return errorCode() == 0x07;}
  unsigned char errorClass() {return (byte(0) & ERROR_CLASS) >> 4;}
  unsigned char errorCode() {return byte(0) & ERROR_CODE;}
  unsigned char vendorUnique() {return (byte(1) & VEND_UNIQUE) >> 5;}
  unsigned int logicalBlockAddr();
  unsigned char segmentNumber() {return byte(1);}
  int fileMark() {return (byte(2) & FILE_MARK)?1:0;}
  int incorrectLength() {return (byte(2) & INC_LEN_IND)?1:0;}
};

inline
 SenseData::SenseData (unsigned char *data, int allocLength): SCSIData(data)
{
  if (allocLength < 8)
    myLength = allocLength;
  else
    myLength = MIN(allocLength,byte(7)+8);
}

inline unsigned int
 SenseData::logicalBlockAddr() {
   if (! (byte(0) & AD_VALID))
     return 0;
   if (isInExtendedFormat())
     return byte(3)<<24 | byte(4)<<16 | byte(5)<<8 | byte(6);
   return (byte(1) & ~VEND_UNIQUE)<<16 | byte(2)<<8 | byte(3);
 }


/*************************************************************************
 *                                                                       *
 *                       Mode Sense Data                                 *
 *                                                                       *
 *************************************************************************/

const unsigned char WRITE_PROT  = 0x80;
const unsigned char BLK_DESCR_LEN = 0x08;

class BlockDescriptor:  public SCSIData {
 public:
  BlockDescriptor(unsigned char *data);
  unsigned char densityCode() {return byte(0);}
  unsigned int numOfBlocks() {return byte(1)<<16 | byte(2)<<8 | byte(3);}
  unsigned int blockLength() {return byte(5)<<16 | byte(6)<<8 | byte(7);}
};

class ModeSenseData : SCSIData {
 private:
  unsigned char presentDescriptors;
  BlockDescriptor *descriptor[8];  // There is usually just one.
  start(int i);
 public:
  ModeSenseData(unsigned char *data, int);
  ~ModeSenseData();
  unsigned char mediumType() {return byte(1);}
  int isWriteProtected() {return (byte(2) & WRITE_PROT)?0:1;}
  unsigned char totalBlockDescriptors() {return byte(3) >> 3;}
  unsigned char presentBlockDescriptors() {return presentDescriptors;}
  unsigned char densityCode(int i) {return descriptor[i]->densityCode();}
  unsigned int numOfBlocks(int i) {return descriptor[i]->numOfBlocks();}
  unsigned int blockLength(int i) {return descriptor[i]->blockLength();}
  unsigned char *vendor(){return byteArray + 4 + (8 * presentDescriptors);}
  unsigned vendorLen(){return myLength - 4 - (8 * presentDescriptors);}
  };

/*************************************************************
 *                                                           *
 *                Read Capacity Data                         *
 *                                                           *
 *************************************************************/
      
const unsigned char READ_CAP_DATA_LEN = 8;

class ReadCapacityData : public SCSIData {
 public:
  ReadCapacityData(unsigned char *data): SCSIData(data) 
	{myLength = READ_CAP_DATA_LEN;}
  unsigned int logicalBlockAddr();
  unsigned int blockLength();
};

inline unsigned int
 ReadCapacityData::logicalBlockAddr(){
   return byte(0)<<24 | byte(1)<<16 | byte(2)<<8 | byte(3);
 }

inline unsigned int
 ReadCapacityData::blockLength() {
   return byte(4)<<24 | byte(5)<<16 | byte(6)<<8 | byte(7);
 }

#endif
