/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef SS1DMA_h
#define SS1DMA_h


#ifdef __GNUC__
#pragma interface
#endif

class DMARegister {
  volatile unsigned int *const myAddress;
 public:
  DMARegister(volatile unsigned int * address) : myAddress( address){}
  unsigned int read();
  void write(unsigned int  data);
};

inline
unsigned 
int DMARegister::read() 
{
	return *myAddress;
}

inline
void 
DMARegister::write(unsigned int data) 
{
	*myAddress = data;
}




 unsigned int const DMA_CS_DEV_ID    = (unsigned int) 0xf0000000;
 unsigned int const DMA_CS_BYTE_ADDR = (unsigned int) 0x00003800;
 unsigned int const DMA_CS_PACK_CNT  = (unsigned int) 0x0000000c;
 unsigned int const DMA_CS_ILLAC     = (unsigned int) 1 << 15;
 unsigned int const DMA_CS_TC        = (unsigned int) 1 << 14;
 unsigned int const DMA_CS_EN_CNT    = (unsigned int) 1 << 13;
 unsigned int const DMA_CS_REQ_PEND  = (unsigned int) 1 << 10;
 unsigned int const DMA_CS_EN_DMA    = (unsigned int) 1 << 9;
 unsigned int const DMA_CS_WRITE     = (unsigned int) 1 << 8;
 unsigned int const DMA_CS_RESET     = (unsigned int) 1 << 7;
 unsigned int const DMA_CS_DRAIN     = (unsigned int) 1 << 6;
 unsigned int const DMA_CS_FLUSH     = (unsigned int) 1 << 5;
 unsigned int const DMA_CS_INT_EN    = (unsigned int) 1 << 4;
 unsigned int const DMA_CS_ERR_PEND  = (unsigned int) 1 << 1;
 unsigned int const DMA_CS_INT_PEND  = (unsigned int) 1 << 0;

/*
 * DMA+ extras
 */
 unsigned int const DMA_TURBO  = (unsigned int) 1 << 22;



class DMAController {
  DMARegister *myCSRegister;
  DMARegister *myAddressRegister;
  DMARegister *myByteCountRegister;
  DMARegister *myDiagnosticRegister; 
  char *	_dmaAddr;
  unsigned	_dmaLen;
 public:
  DMAController(volatile unsigned int *);
  unsigned int status() {return myCSRegister->read();}
  int byteCountExpired() {return myCSRegister->read() & DMA_CS_TC;}
  int requestPending() {return myCSRegister->read() & DMA_CS_REQ_PEND;}
  int errorPending() {return myCSRegister->read() & DMA_CS_ERR_PEND;}
  int intrPending() {return myCSRegister->read() & DMA_CS_INT_PEND;}
  void enable();
  void disable();
  void enableIntr();
  void disableIntr();
  void enableCount();
  void disableCount();
  void memoryToDeviceDirection();
  void deviceToMemoryDirection();
  void drain();
  void flush();
  void reset();
  //XXXXXXXXXX FIX ME: check MSByte
  void setByteCount(unsigned int nBytes) {myByteCountRegister->write(nBytes);}
  unsigned byteCount() {return myByteCountRegister->read();}
  void setAddress(unsigned char * addr, unsigned len);
  char* transferAddress(){ return _dmaAddr;}
  unsigned transferLen(){ return _dmaLen;}
  void release();
  unsigned address() {return myAddressRegister->read();}
};

class SS1DMAController : public DMAController
{
public:
	SS1DMAController(volatile unsigned int* base);
};

inline
SS1DMAController::SS1DMAController(volatile unsigned int *base) : DMAController(base)
{
}

  
inline 
 DMAController::DMAController(volatile unsigned int *baseAddr){
   myCSRegister = new DMARegister(baseAddr);
   myAddressRegister = new DMARegister(baseAddr+1);
   myByteCountRegister = new DMARegister(baseAddr+2);
   myDiagnosticRegister = new DMARegister(baseAddr+3);
   reset();
 }

inline void
 DMAController::disable(){
   myCSRegister->write(myCSRegister->read() & ~DMA_CS_EN_DMA);
 }

inline void
 DMAController::enable(){
   myCSRegister->write( myCSRegister->read() 
	| DMA_TURBO | DMA_CS_EN_DMA);
 }

inline void
 DMAController::enableIntr(){
   myCSRegister->write( myCSRegister->read() | DMA_CS_INT_EN);
 }

inline void
 DMAController::enableCount(){
   myCSRegister->write( myCSRegister->read() | DMA_CS_EN_CNT);
 }

inline void
 DMAController::memoryToDeviceDirection(){
   myCSRegister->write(myCSRegister->read() & ~DMA_CS_WRITE);
 }

inline void
 DMAController::deviceToMemoryDirection(){
   myCSRegister->write(myCSRegister->read() | DMA_CS_WRITE);
 }

inline void
 DMAController::disableIntr(){
   myCSRegister->write(myCSRegister->read() & ~DMA_CS_INT_EN);
 }

inline void
 DMAController::disableCount(){
   myCSRegister->write(myCSRegister->read() & ~DMA_CS_EN_CNT);
 }

inline void 
 DMAController::flush(){
   myCSRegister->write(myCSRegister->read() | DMA_CS_FLUSH);
 }

inline void 
 DMAController::reset(){
   /*XXXX FIX ME: reset must not happen if req_pend is 1 */
   myCSRegister->write(DMA_CS_RESET);
   myCSRegister->write(myCSRegister->read() & ~DMA_CS_RESET);
   setByteCount(0);
   _dmaAddr = 0;
   _dmaLen = 0;
 }
#endif SS1DMA_h
