/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * AddressTranslation.h: AddressTranslation class.
 */

#ifndef AddressTranslation_h
#define AddressTranslation_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Object.h" /* parent class */
#include "CPUSet.h"
#include "ProtectionLevel.h"

class PhysicalMemoryChain;  // forward reference...

enum AddressTranslationFailureCode {
    MMUCacheMiss,
    NonResidentMemory,
    ProtectionViolation,
    SomethingElse,
};

class AddressTranslationStatus {
 public:
    enum Values {
        Modified = 0x1,
        Referenced = 0x2
    };

    int value;
    AddressTranslationStatus(int v = 0) { value = v; }
};

extern ClassStar AddressTranslationClass;
class AddressTranslation : public Object {

    /*
     * This is a friend so it can get at activeOn() and noLongerActiveOn().
     */
    friend class AddressTranslator;

 protected:

    /*
     * This keeps track of which processors are using this translation.
     * This information is used by removeMapping to determine which
     * processor TLB caches need invalidated. The set is kept up
     * to date by AddressTranslation::activate().
     *
     * For a later optimizaton, move this into subclasses since it
     * isnt needed for uniprocessors at all. (I.e. the only TLB to
     * flush is the current one.)   --Vince
     */
    CPUSet  cpusActiveOn;

    void activeOn( CPU * );
    void noLongerActiveOn( CPU * );

    void flushFromAllActiveTranslationCaches( const char *, unsigned int );

 public:
    AddressTranslation();

    /*
     * Add a translation mapping from a virtual address into a chain of
     * physical memory frames.
     */
    virtual void addMapping( const char * vaddr,
                             PhysicalMemoryChain * chain, ProtectionLevel prot ) = 0;

    /*
     * Remove any mappings for a range of addresses.
     */
    void removeMapping( const char * vaddr, unsigned int nbytes );
    virtual void basicRemoveMapping( const char * vaddr,
                                     unsigned int nbytes ) = 0;

    /*
     * Alter the ProtectionLevel assigned to a range of addresses.
     */
    int changeProtection( const char * vaddr, unsigned int nbytes,
                          ProtectionLevel protection );
    virtual int basicChangeProtection( const char * vaddr,
                                       unsigned int nbytes, ProtectionLevel protection ) = 0;

    /*
     * Alter status of translation mappings
     */
    int changeStatus( const char * vaddr, unsigned int nbytes,
                      AddressTranslationStatus clear,
                      AddressTranslationStatus set );
    virtual int basicChangeStatus( const char * vaddr, unsigned int nbytes,
                                   AddressTranslationStatus clear,
                                   AddressTranslationStatus set ) = 0;

    /*
     * Get reference info
     */
    virtual int referenced( const char * vaddr, unsigned int nbytes, int clear );
    virtual int basicReferenced( const char * vaddr, unsigned int nbytes,
                                 int clear ) = 0;

    Class * classOf();
};

inline
AddressTranslation::AddressTranslation()
{
}

inline void
AddressTranslation::activeOn( CPU * cpu )
{
    cpusActiveOn.add( cpu );
}

inline void
AddressTranslation::noLongerActiveOn( CPU * cpu )
{
    cpusActiveOn.remove( cpu );
}

#endif /* AddressTranslation_h */
