/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Cache.h 
 */

#ifndef Cache_h
#define Cache_h
#ifdef __GNUC__
#pragma interface
#endif

#include "CacheUnitInfo.h"
#include "Lock.h"
#include "ProxiableObject.h"
#include "CacheAsync.h"

extern ClassStar CacheClass;

class CacheCounter {
 public:
    unsigned _count;
    unsigned _volume;
    CacheCounter() : _count( 0 ), _volume( 0 ) {}
    void add( unsigned volume ) 
    { _count++; _volume += volume; }
};

class CacheStatistics {
 public:
    CacheCounter _basicRead;
    CacheCounter _basicWrite;
    CacheCounter _moRead;
    CacheCounter _moWrite;
    CacheCounter _read;
    CacheCounter _write;
    CacheCounter _rawWrite; 
};

class Cache;
class CacheSynchronizeAsyncClosure : public IOAsyncClosure
{
 protected:
    friend class Cache;

    unsigned _startUnit;
    unsigned _endUnit;
    unsigned _unit;
    unsigned _maxUnits;
    unsigned _log2BufSize;
    unsigned _firstUnit;
    unsigned _noOfUnits;
    int _state;
    char * _realBuffer;
    char * & _buffer;
    Cache * _cache;
    IOAsyncClosure * _parent;

    static void func( CacheSynchronizeAsyncClosure * c );
 public:
    CacheSynchronizeAsyncClosure(
        int startUnit, int endUnit, int maxUnits, int log2BufSize,
        Cache * cache, IOAsyncClosure * parent );
};

class Cache : public ProxiableObject {
 protected:
    static Lock _lock;
    static CacheStatistics _stats;

    unsigned char _log2UnitSize;
    unsigned char _log2MoUnitSize;
    unsigned char _log2SegmentSize;


    MemoryObjectCache * _moc;
    Cache * _parent;
    Cache * _child;
    MemoryObject * _mo;
    CacheUnitInfoTable * _unitInfoTable;

    unsigned int _numberOfUnits;
    unsigned char _dirty;
    unsigned int _cachedUnits;

    virtual unsigned  basicSetNumberOfUnits( unsigned int unit );

    /*
     * methods to get and set status of a cache unit
     */

    CacheUnitInfo unitInfo( unsigned int unit );
    void setUnitInfo( unsigned int unit, CacheUnitInfo info );

    /*
     * advise allocation of addition units
     */

    virtual void basicAllocate( unsigned int count );

    /*
     * method used by invalidate to free space in cache
     */

    virtual void basicFreeUnits( unsigned int startUnit, unsigned int count );

    /*
     * methods to read and write into the store of this cache
     */

    virtual int basicRead ( unsigned int startUnit, unsigned int count, char * addr,
                            IOAsyncClosure * c ) = 0;
    virtual int basicWrite( unsigned int startUnit, unsigned int count, char * addr,
                            IOAsyncClosure * c ) = 0;

    /*
     * methods to read and write the real memory object
     */

    virtual int moRead ( unsigned int startUnit, unsigned int count, char * addr,
                         IOAsyncClosure * c );
    virtual int moWrite( unsigned int startUnit, unsigned int count, char * addr,
                         int writeThrough, IOAsyncClosure * c );

    /*
     * method to write modified units from MemoryObjectCache to mo
     */

    virtual int writeMOModified( unsigned int startUnit, unsigned int count, 
                                 CacheUnitInfo * info,
                                 char * addr, int writeThrough,
                                 IOAsyncClosure * c );

    /*
     * methods used by Cache to manipulate hierarchy
     */

    Cache * setParent( Cache * cache );
    Cache * setChild( Cache * cache );
    static void acquire();
    static void release();

 public:

    Class * classOf();

    Cache( MemoryObjectCache * moc );
    ~Cache();

    Cache * parent();
    Cache * child();

    /*
     * methods to manipulate cache hierarchy
     */

    virtual int insertAbove( Cache * cache );

    /*
     * methods to get and change number of units in the cache
     */

    unsigned int log2UnitSize();
    unsigned int unitSize();
    unsigned int numberOfUnits();
    virtual unsigned int setNumberOfUnits( unsigned int noOfUnits );

    /* 
     * methods to figure out if a unit exists in the cache or
     * its children
     */

    virtual int exists( unsigned unit );

    /*
     * methods to read from and write to the cache,
     * if addr == 0, the mo is mapped into the current domain, and
     *    the appropriate addr is computed from the baseAddr of the mo
     * write requires the status of the units being written
     * read also forwards requests to child if units desired are not 
     *    cached at this level
     */

    virtual int read ( unsigned int startUnit, unsigned int count, char * addr,
                       IOAsyncClosure * c );
    virtual int write( unsigned int startUnit, unsigned int count, CacheUnitInfo * info,
                       char * addr, IOAsyncClosure * c );

    /*
     * methods to write directly into the mo
     */

    virtual int rawWrite( unsigned int startUnit, unsigned int count, CacheUnitInfo * info,
                          char * addr, IOAsyncClosure * c );

    /*
     * methods to invalidate the cache
     */

    virtual void invalidate ( unsigned int startUnit, unsigned int count, int propagate );

    /* 
     * methods to synchronize and flush the cache
     * synchronize writes out all modified units
     * synchronize also forces child to synchronize
     */

    virtual void synchronize( unsigned int startUnit, unsigned int count, 
                              unsigned int & log2BufSize, char * & buffer );

    /*
     * methods to support asynchronize synchronize operation
     * returns 0 if closure will be executed later, else return 1
     */

    virtual void synchronize( unsigned int startUnit,unsigned  int count,
                              unsigned int log2BufSize, char * & buffer,
                              IOAsyncClosure * c );

    /*
     * method to return dirty status of cache
     */
    virtual int dirty();

    /*
     * return stats
     */
    CacheStatistics * stats();

 protected:

    friend class CacheSynchronizeAsyncClosure;
    void synchronizeClosure( CacheSynchronizeAsyncClosure * c );
};

inline
CacheStatistics * Cache::stats()
{
    return &_stats;
}

inline
Cache * Cache::parent() 
{
    return _parent;
}

inline
Cache * Cache::child()
{
    return _child;
}

inline
Cache * Cache::setParent( Cache * cache )
{
    _parent = cache;
    return _parent;
}

inline
Cache * Cache::setChild( Cache * cache )
{
    _child = cache;
    return _child;
}

inline
unsigned int Cache::log2UnitSize()
{
    return _log2UnitSize;
}

inline
unsigned int Cache::unitSize()
{
    return 1 << _log2UnitSize;
}

inline
unsigned int Cache::numberOfUnits()
{
    return _numberOfUnits;
}

inline
void Cache::acquire()
{
    _lock.acquire();
}

inline
void Cache::release()
{
    _lock.release();
}

inline
CacheUnitInfo Cache::unitInfo(unsigned  int unit )
{
    return _unitInfoTable->unitInfo( unit );
}

inline
void Cache::setUnitInfo( unsigned int unit, CacheUnitInfo info )
{
    _unitInfoTable->setUnitInfo( unit, info );
}

#endif /*Cache_h*/
