/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef DVMManager_h
#define DVMManager_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Lock.h"
#define LOCK Lock
#include "Semaphore.h"
#include "List.h"
#include "TimeoutTimer.h"
#include "MemoryObjectStar.h"
#include "State.h"
#include "DVMMessage.h"

class PhysicallyAddressableUnit;
class Process;
class NCDVMCache;

static const int MaxNumberOfDVMMachines = 32; // should fit in an int, otherwise
					      // the per page array in the
static const int MaxMOs = 32;

typedef void (*PFVPIV) (void *, int);

class GroupTimer {
  protected:
    static void GroupTimerEntry(int groupTimer);
    
    int _maxEvents;
    char * _eventTable;
    Semaphore _timeoutMutex;
    TimeoutTimer _timer;
    Process * _timeoutProcess;
    long _microSecondQuantum;

    void timeoutProcessEntry();

    PFVPIV _clientFunction;
    void * _clientObject;

  public:
    // clientFunction must be short and sweet to minimize effect on timer.
    GroupTimer(int max, long microSecondQuantum,
	       void * clientObject, PFVPIV clientFunction);
    ~GroupTimer();

    void enable(int eventNum);
    void disable(int eventNum);
};

class DVMServiceMachine : public StateMachine {
  public:
    DVMServiceMachine();
};

class DVMServiceState: public State {
  protected:
  public:
    DVMServiceState(StateMachine * machine) : State(machine) {};
    
    virtual State * joinLocal(DVMManager *, unsigned)
    { return error(); };
    virtual State * joinRemote(DVMManager *, unsigned)
    { return error(); };
    virtual State * remoteRequestJoin(DVMManager *, unsigned, DVMServiceMsg *)
    { return error(); };
    virtual State * remoteJoinDone(DVMManager *, unsigned, DVMServiceMsg *)
    { return error(); };
    
    virtual State * leaveLocal(DVMManager *, unsigned)
    { return error(); };
    virtual State * leaveRemote(DVMManager *, unsigned)
    { return error(); };
    virtual State * remoteRequestLeave(DVMManager *, unsigned, DVMServiceMsg *)
    { return error(); };
    virtual State * remoteLeaveDone(DVMManager *, unsigned,
				    DVMServiceMsg *)
    { return error(); };
    
    virtual State * timeout(DVMManager *, unsigned)
    { return error(); };
};
					      
class DVMManager {
/*
 * The DVMManager manages Transmit, Receive, Timeout and Action Queues.
 * Each queue has a server process. Transmit and Receive Processes are
 * awaken whenever there is work. The TimeoutProcess is awaken by a
 * timer. The action queue handles messages that invoke cache actions.
 *
 * The manager interacts with the cache. 
 * GetPage messages from a remote host will trigger
 * cache::remoteCache. The Cache itself uses the send and receive
 * methods to send and receive pages. The manager handles
 * fragmentation, timeouts, timestamps, old messages etc.
 *
 * Right now, we have one systemwide manager. Later, we might have
 * multiple managers, upto one per cache.
 */
  
  protected:
    int _thisMachine;		// bit position of this machine
    NCDVMCache * _cache;

    List _xmitQueue;
    LOCK _xmitMutex;
    Semaphore _awaitXmit;
    Process * _xmitProcess;
    DVMMsg * deqXmit();
    
    List _rcvQueue;
    LOCK _rcvMutex;
    Semaphore _awaitRcv;
    Process * _rcvProcess;
    DVMCacheMsg * deqRcv();
    
    List * _timeoutQueue;
    List * _backupQueue;
    LOCK _timeoutMutex;
    TimeoutTimer _timer;
    Process * _timeoutProcess;
    void acqTimeout();		// acquire timeout q
    void rlsTimeout();		// release timeout q
    void enqTimeout(DVMCacheMsg *);
    void enqBackup(DVMCacheMsg *);
    DVMCacheMsg * deqTimeout();
    DVMCacheMsg * removeTimeout(DVMCacheMsg *);
    void exchangeTOQ();
    
    List _actionQueue;
    LOCK _actionMutex;
    Semaphore _awaitAction;
    Process * _actionProcess;
    void enqAction(DVMCacheMsg *);
    DVMCacheMsg * deqAction();

    List _srvQueue;
    LOCK _srvMutex;
    Semaphore _awaitSrv;
    Process * _srvProcess;
    DVMServiceMsg * deqSrv();
    
    List _cacheQueue;
    LOCK _cacheMutex;
    Semaphore _awaitCache;
    Process * _cacheProcess;
    DVMCacheMsg * deqCache();

    unsigned _outgoing[10];
    unsigned _incoming[10];

    unsigned tsindex(unsigned);

  public:
    void xmitProcessEntry();
    void rcvProcessEntry();
    void actionProcessEntry();
    void timeoutProcessEntry();
    void srvProcessEntry();
    void cacheProcessEntry();
    
    DVMManager();
    ~DVMManager();

    void addCache(NCDVMCache * cache);

    unsigned id();
    
//     /*
//      * method send: send the contents of pau to destination
//      */
//     void send(int destination, unsigned unit, PhysicallyAddressableUnit * pau);
//     /*
//      * method rcv: receive the contents of unit into pau from source
//      */
//     void rcv(int source, unsigned unit, PhysicallyAddressableUnit * pau);

    void enqRcv(DVMCacheMsg *);
    void enqSrv(DVMServiceMsg *);
    void enqCache(DVMCacheMsg *);
    void enqXmit(DVMMsg *);

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Stuff for Setup/Teardown. Might separate this later into a separate class
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  protected:

friend class DVM_SRV_EMPTY;
friend class DVM_SRV_TOEMPTY;
friend class DVM_SRV_REMOTE;
friend class DVM_SRV_TOREMOTE;
friend class DVM_SRV_LOCAL;
friend class DVMServiceMsg;
friend class DVMJoinMsg;
friend class DVMAckJoinMsg;
friend class DVMLeaveMsg;
friend class DVMAckLeaveMsg;

    class _dvm_mo_entry {
      public:
	MemoryObjectStar mo;
	char * name;
	int machines;
	int localrefs;
	int remoteid;
	DVMServiceState * state;
    };

    _dvm_mo_entry _moTable[MaxMOs];

    GroupTimer _serviceTimer;

    Semaphore _moTableMutex;
    Semaphore _remoteMutex;
    Semaphore _remoteAwait;

    unsigned find(char * path);
    unsigned find(unsigned remote);
    unsigned installName(char * name);

    unsigned hostId(char * path);

    void createNCDVM(unsigned slot);
    void destroyDVM(unsigned slot);
    
    int refMemobj(unsigned slot) { return (_moTable[slot].localrefs += 1); };
    int unrefMemobj(unsigned slot) { return (_moTable[slot].localrefs -= 1); };
    int refs(unsigned slot) { return _moTable[slot].localrefs; };
    
    int setMachine(unsigned slot, int machine)
    { return _moTable[slot].machines |= machine; };
    int unsetMachine(unsigned slot, int machine )
    { return _moTable[slot].machines &= ~machine; };
    int machines(unsigned slot)
    { return _moTable[slot].machines; };

    int remoteId(unsigned slot) { return _moTable[slot].remoteid; };
    void setRemoteId(int remote, unsigned slot)
    { _moTable[slot].remoteid = remote; }

    
    char * name(unsigned slot) { return _moTable[slot].name; };
    
    unsigned timestamp(unsigned dst) { return _outgoing[tsindex(dst)]; };
    unsigned incoming(unsigned dst) { return _incoming[tsindex(dst)]; };
    unsigned setIncoming(unsigned dst, unsigned val)
    { return _incoming[tsindex(dst)] = val; };
    unsigned outgoing(unsigned dst) { return _outgoing[tsindex(dst)]; };
    unsigned setOutgoing(unsigned dst, unsigned val)
    { return _outgoing[tsindex(dst)] = val; };
    unsigned nextStamp(unsigned dst) { return _outgoing[tsindex(dst)]++; };

    void enableTimeout(unsigned slot);
    void disableTimeout(unsigned slot);

    // primitive for timeout
    void serviceTimeout(unsigned slot);
    static void DVMServiceTimeout(void *, int);
     
    // primitives invoked from messages
    void remoteSetupDone(char * path, DVMServiceMsg * msg );
    int setupDVM(char * path, DVMServiceMsg * msg);
    void teardownDVM(char * path, DVMServiceMsg * msg);
    void remoteTeardownDone(char * path, DVMServiceMsg * msg);
    
  public:

    int localId() { return _thisMachine; };
    
    // primitives for local dvm
    MemoryObjectRef setupDVM(char * path);
    void teardownDVM(char * path);
};

// inline void
// DVMManager::enqXmit(DVMMsg * msg)
// {
//     _xmitMutex.acquire();
//     _xmitQueue.addLast(msg);
//     _awaitXmit.V();
//     _xmitMutex.release();
// }

inline void
DVMManager::enqXmit(DVMMsg * msg)
{
    msg->transmit();
}

inline DVMMsg *
DVMManager::deqXmit()
{
    DVMMsg * msg;
    _xmitMutex.acquire();
    msg = (DVMMsg *)_xmitQueue.removeFirst();
    _xmitMutex.release();
    return msg;
}

// inline void
// DVMManager::enqRcv(DVMCacheMsg * msg)
// {
//     _rcvMutex.acquire();
//     _rcvQueue.addLast(msg);
//     _awaitRcv.V();
//     _rcvMutex.release();
// }


inline DVMCacheMsg *
DVMManager::deqRcv()
{
    DVMCacheMsg * msg;
    _rcvMutex.acquire();
    msg = (DVMCacheMsg *)_rcvQueue.removeFirst();
    _rcvMutex.release();
    return msg;
}

inline void
DVMManager::enqAction(DVMCacheMsg * msg)
{
    _actionMutex.acquire();
    _actionQueue.addLast(msg);
    _awaitAction.V();
    _actionMutex.release();
}

inline DVMCacheMsg *
DVMManager::deqAction()
{
    DVMCacheMsg * msg;
    _actionMutex.acquire();
    msg = (DVMCacheMsg *)_actionQueue.removeFirst();
    _actionMutex.release();
    return msg;
}

inline void
DVMManager::acqTimeout()
{
    _timeoutMutex.acquire();
}

inline void
DVMManager::rlsTimeout()
{
    _timeoutMutex.release();
}

inline void
DVMManager::enqTimeout(DVMCacheMsg * msg)
{
    acqTimeout();
    _timeoutQueue->addLast(msg);
    rlsTimeout();
}

inline void
DVMManager::enqBackup(DVMCacheMsg * msg)
{
    _backupQueue->addLast(msg);
}

inline DVMCacheMsg *
DVMManager::deqTimeout()
{
    DVMCacheMsg * msg;
    msg = (DVMCacheMsg *)_timeoutQueue->removeFirst();
    return msg;
}

inline DVMCacheMsg *
DVMManager::removeTimeout(DVMCacheMsg * target)
{
    DVMCacheMsg * msg;
    acqTimeout();
    msg = (DVMCacheMsg *)_timeoutQueue->remove(target);
    if (0 == msg)
	msg = (DVMCacheMsg *)_backupQueue->remove(target);
    rlsTimeout();
    return msg;
}

inline void
DVMManager::exchangeTOQ()
{
    List * l = _backupQueue;
    _backupQueue = _timeoutQueue;
    _timeoutQueue = l;
}

inline void
DVMManager::enqSrv(DVMServiceMsg * msg)
{
    _srvMutex.acquire();
    _srvQueue.addLast(msg);
    _awaitSrv.V();
    _srvMutex.release();
}

inline DVMServiceMsg *
DVMManager::deqSrv()
{
    DVMServiceMsg * msg;
    _srvMutex.acquire();
    msg = (DVMServiceMsg *)_srvQueue.removeFirst();
    _srvMutex.release();
    return msg;
}

inline void
DVMManager::enqCache(DVMCacheMsg * msg)
{
    _cacheMutex.acquire();
    _cacheQueue.addLast(msg);
    _awaitCache.V();
    _cacheMutex.release();
}

inline DVMCacheMsg *
DVMManager::deqCache()
{
    DVMCacheMsg * msg;
    _cacheMutex.acquire();
    msg = (DVMCacheMsg *)_cacheQueue.removeFirst();
    _cacheMutex.release();
    return msg;
}

inline void
DVMManager::addCache(NCDVMCache * cache)
{
    _cache = cache;
}

inline unsigned
DVMManager::id()
{
    return _thisMachine;
}


extern DVMManager * TheDVMManager;

// Stuff for DVMService

class DVM_SRV_EMPTY: public DVMServiceState {
    // JoinLocal to LOCAL
    // JoinRemote to REMOTE
  public:
    DVM_SRV_EMPTY(StateMachine * machine) : DVMServiceState(machine) {};
    virtual State * joinLocal(DVMManager *, unsigned);
    virtual State * joinRemote(DVMManager *, unsigned);
    virtual State * remoteRequestJoin(DVMManager *, unsigned, DVMServiceMsg *);
    virtual State * remoteRequestLeave(DVMManager *, unsigned, DVMServiceMsg *);
    virtual State * remoteLeaveDone(DVMManager *, unsigned,
				    DVMServiceMsg *);
};

class DVM_SRV_REMOTE: public DVMServiceState {
  public:
    DVM_SRV_REMOTE(StateMachine * machine) : DVMServiceState(machine) {};
    virtual State * joinRemote(DVMManager *, unsigned);
    virtual State * leaveRemote(DVMManager *, unsigned);
};

class DVM_SRV_LOCAL: public DVMServiceState {
  public:
    DVM_SRV_LOCAL(StateMachine * machine) : DVMServiceState(machine) {};
    virtual State * remoteRequestJoin(DVMManager *, unsigned, DVMServiceMsg *);
    virtual State * remoteRequestLeave(DVMManager *, unsigned, DVMServiceMsg *);
    virtual State * joinLocal(DVMManager *, unsigned);
    virtual State * leaveLocal(DVMManager *, unsigned);
};

class DVM_SRV_TOREMOTE: public DVMServiceState {
  public:
    DVM_SRV_TOREMOTE(StateMachine * machine) :
    DVMServiceState(machine) {};
    virtual State * remoteJoinDone(DVMManager *, unsigned, DVMServiceMsg *);
    virtual State * timeout(DVMManager *, unsigned);
};

class DVM_SRV_TOEMPTY: public DVMServiceState {
  public:
    DVM_SRV_TOEMPTY(StateMachine * machine) : DVMServiceState(machine) {};
    virtual State * remoteLeaveDone(DVMManager *, unsigned,
				    DVMServiceMsg *);
    virtual State * timeout(DVMManager *, unsigned);
};


extern DVMServiceMachine * TheDVMServiceMachine;
extern DVMServiceState * STATE_DVM_SRV_EMPTY;
extern DVMServiceState * STATE_DVM_SRV_LOCAL;
extern DVMServiceState * STATE_DVM_SRV_REMOTE;
extern DVMServiceState * STATE_DVM_SRV_TOEMPTY;
extern DVMServiceState * STATE_DVM_SRV_TOREMOTE;


#endif /* DVMManager_h */
