/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef DVMMessage_h
#define DVMMessage_h


#ifdef __GNUC__
#pragma interface
#endif

class Semaphore;



enum DVMMsgType {
    GetPage = 1,
    HerePage = 2,
    AckPage = 3,
    Join = 4,
    AckJoin = 5,
    Leave = 6,
    AckLeave = 7,
    OwnerHint = 8,
    GetRead = 9,
    HereRead = 10,
    GetUp = 11,
    HereUp = 12,
    AckUp = 13,
    Inv = 14,
    AckInv = 15
};

struct EthernetHeader {
    EthernetAddress dest;
    EthernetAddress source;
    unsigned char   type[2];
    unsigned char   dummy[2];       /* force word alignment */
    DVMMsgType _DVMType;
};


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// 
// DVMXmitMessages know how to transmit themselves, how to timeout and how
// to receive acknowledgements.
// They also support methods to make raw packed copies of themselves and
// to construct themselves from raw messages. rawCopyFrom is used only by
// those Xmit msgs that expect to receive data.
// Msgs that expects acks support the ack() and rcvType() methods
// 

class DVMMsg : public Link {
    friend OutputStream& operator<<(OutputStream&, DVMMsg *);
  protected:
    Semaphore * _sem;		// Sem on which sender waits
    DVMMsgType _type;
    unsigned _destination;
    unsigned _source;
    unsigned _timestamp;
    unsigned * _top;
    unsigned _rawLength;
    
    virtual int compare(Link *);
    
    virtual DVMMsgType rcvType();
    
    virtual void rawCopyTo(unsigned * packet) = 0;
    virtual unsigned length() = 0;
    
    virtual void inspect(OutputStream& os);

    virtual ~DVMMsg(){};
  public:

    // Ctor for transmission
    DVMMsg (Semaphore * sem, DVMMsgType type, unsigned dst,
	    unsigned src, unsigned ts);
    // Ctor for reception
    DVMMsg() {};
    
    virtual DVMMsgType type() { return _type; };
    virtual unsigned destination() { return _destination; };
    virtual unsigned source() { return _source; };
    virtual unsigned timestamp() { return _timestamp; };
    unsigned * top() { return _top; };
    unsigned rawLength() { return _rawLength; };
    
    virtual DVMMsg * transmit();
    virtual DVMMsg * timeout();
};

inline
DVMMsg::DVMMsg(Semaphore * sem, DVMMsgType type, unsigned dst,
		       unsigned src, unsigned ts)
{
    _sem = sem;
    _type = type;
    _destination = dst;
    _source = src;
    _timestamp = ts;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// 
// Messages that implement consistency protocols
//

class NCDVMCache;

class DVMCacheMsg : public DVMMsg {
  protected:
    unsigned _unit;
    unsigned * _address;
    unsigned _fragment;
    
    int compare(Link * l);
    
    virtual void rawCopyTo(unsigned *) = 0;
    virtual void cacheCopyFrom(DVMCacheMsg *) {};
    virtual DVMCacheMsg * response() = 0;
    
    virtual unsigned length() = 0;
    
    virtual void inspect(OutputStream& os);
  public:
    DVMCacheMsg (DVMMsgType type, unsigned dst,
		 unsigned src, unsigned ts, unsigned unit,
		 unsigned * address, unsigned frag);
    DVMCacheMsg( unsigned * top, unsigned length);

    unsigned unit() { return _unit; };
    unsigned fragment() { return _fragment; };
    unsigned * address() { return _address; };
    
    virtual DVMCacheMsg * receive(DVMCacheMsg *);

    virtual void action(NCDVMCache *) = 0;
};

inline
DVMCacheMsg::DVMCacheMsg(DVMMsgType type, unsigned dst,
			 unsigned src, unsigned ts,
			 unsigned unit, unsigned * address, unsigned frag)
: DVMMsg(0,type,dst,src,ts)
{
    _unit = unit;
    _address = address;
    _fragment = frag;
}

inline
DVMCacheMsg::DVMCacheMsg( unsigned * top, unsigned length)
: DVMMsg()
{
    _top = top;
    _rawLength = length;
    _address = 0;
}

class DVMGetPageMsg : public DVMCacheMsg {
  protected:
    virtual void cacheCopyFrom(DVMCacheMsg *);
    virtual DVMMsgType rcvType() { return HerePage; };
    virtual DVMCacheMsg * response();
    virtual unsigned length();
    virtual void rawCopyTo(unsigned * packet);

  public:
    DVMGetPageMsg(unsigned dst, unsigned src, unsigned ts,
		  unsigned unit, unsigned * address, unsigned frag)
	: DVMCacheMsg(GetPage,dst,src,ts,unit,address,frag) {};
    DVMGetPageMsg(unsigned * top, unsigned length)
	: DVMCacheMsg(top, length) {};
    
//     DVMMsg * transmit();
    DVMMsg * timeout();
    virtual void action(NCDVMCache *);
};

class DVMHerePageMsg : public DVMCacheMsg {
  protected:
    void rawCopyTo(unsigned * packet);
    virtual unsigned length();
    virtual DVMMsgType rcvType() { return AckPage; };
    virtual DVMCacheMsg * response();
  public:
    DVMHerePageMsg( unsigned dst, unsigned src, unsigned ts,
		   unsigned unit, unsigned * address, unsigned frag)
	: DVMCacheMsg(HerePage,dst,src,ts,unit,address,frag) {};
    DVMHerePageMsg(unsigned * top, unsigned length);
    
    DVMMsg * timeout();

    virtual DVMCacheMsg * selfack();
    virtual void action(NCDVMCache *);
};

inline
DVMHerePageMsg::DVMHerePageMsg( unsigned * top, unsigned length)
: DVMCacheMsg(top,length)
{
    _address = (unsigned *)this
	+ ((sizeof(DVMHerePageMsg) + sizeof(unsigned) - 1)/sizeof(unsigned));
}

class DVMAckPageMsg : public DVMCacheMsg {
  protected:
    virtual unsigned length();
    virtual void rawCopyTo(unsigned * packet);
    virtual DVMCacheMsg * response();
  public:
    DVMAckPageMsg(unsigned dst, unsigned src, unsigned ts,
		  unsigned unit, unsigned frag)
	: DVMCacheMsg(AckPage,dst,src,ts,unit,0,frag) {};
    DVMAckPageMsg(unsigned * top, unsigned length)
	: DVMCacheMsg(top, length) {};

    DVMCacheMsg * receive(DVMCacheMsg *);
    virtual void action(NCDVMCache *);
};

class DVMOwnerHintMsg : public DVMCacheMsg {
  protected:
    virtual unsigned length();
    virtual void rawCopyTo(unsigned * packet);
    virtual DVMCacheMsg * response();
  public:
    DVMOwnerHintMsg(unsigned dst, unsigned src, unsigned ts,
		    unsigned unit, unsigned owner)
	: DVMCacheMsg(OwnerHint,dst,src,ts,unit,0,owner) {};
    DVMOwnerHintMsg(unsigned * top, unsigned length)
	: DVMCacheMsg(top, length) {};

    DVMCacheMsg * receive(DVMCacheMsg *);
    virtual void action(NCDVMCache *);
};

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// RCDVM Messages

class DVMGetReadMsg : public DVMCacheMsg {
  protected:
    virtual void cacheCopyFrom(DVMCacheMsg *);
    virtual DVMMsgType rcvType() { return HereRead; };
    virtual DVMCacheMsg * response();
    virtual unsigned length();
    virtual void rawCopyTo(unsigned * packet);
  public:
    DVMGetReadMsg(unsigned dst, unsigned src, unsigned ts,
		  unsigned unit, unsigned * address, unsigned frag)
	: DVMCacheMsg(GetRead,dst,src,ts,unit,address,frag) {};
    DVMGetReadMsg(unsigned * top, unsigned length)
	: DVMCacheMsg(top, length) {};
    
//     DVMMsg * transmit();
    DVMMsg * timeout();
    virtual void action(NCDVMCache *);
};

class DVMHereReadMsg : public DVMCacheMsg {
  protected:
    void rawCopyTo(unsigned * packet);
    virtual unsigned length();
    virtual DVMMsgType rcvType() { return 0; };
    virtual DVMCacheMsg * response();
  public:
    DVMHereReadMsg( unsigned dst, unsigned src, unsigned ts,
		   unsigned unit, unsigned * address, unsigned frag)
	: DVMCacheMsg(HereRead,dst,src,ts,unit,address,frag) {};
    DVMHereReadMsg(unsigned * top, unsigned length);
    
    DVMMsg * timeout();

    virtual DVMCacheMsg * selfack();
    virtual void action(NCDVMCache *);
};

inline
DVMHereReadMsg::DVMHereReadMsg( unsigned * top, unsigned length)
: DVMCacheMsg(top,length)
{
    _address = (unsigned *)this
	+ ((sizeof(DVMHereReadMsg) + sizeof(unsigned) - 1)/sizeof(unsigned));
}

class DVMGetUpMsg : public DVMCacheMsg {
  protected:
    virtual unsigned length();
    virtual void rawCopyTo(unsigned * packet);
    virtual DVMCacheMsg * response();
  public:
    DVMGetUpMsg(unsigned dst, unsigned src, unsigned ts,
		    unsigned unit)
	: DVMCacheMsg(GetUp,dst,src,ts,unit,0,0) {};
    DVMGetUpMsg(unsigned * top, unsigned length)
	: DVMCacheMsg(top, length) {};

    DVMCacheMsg * receive(DVMCacheMsg *);
    virtual void action(NCDVMCache *);
};

class DVMHereUpMsg : public DVMCacheMsg {
  protected:
    virtual unsigned length();
    virtual void rawCopyTo(unsigned * packet);
    virtual DVMCacheMsg * response();
  public:
    DVMHereUpMsg(unsigned dst, unsigned src, unsigned ts,
		    unsigned unit)
	: DVMCacheMsg(HereUp,dst,src,ts,unit,0,0) {};
    DVMHereUpMsg(unsigned * top, unsigned length)
	: DVMCacheMsg(top, length) {};

    DVMCacheMsg * receive(DVMCacheMsg *);
    virtual void action(NCDVMCache *);
};


class DVMAckUpMsg : public DVMCacheMsg {
  protected:
    virtual unsigned length();
    virtual void rawCopyTo(unsigned * packet);
    virtual DVMCacheMsg * response();
  public:
    DVMAckUpMsg(unsigned dst, unsigned src, unsigned ts,
		    unsigned unit)
	: DVMCacheMsg(AckUp,dst,src,ts,unit,0,0) {};
    DVMAckUpMsg(unsigned * top, unsigned length)
	: DVMCacheMsg(top, length) {};

    DVMCacheMsg * receive(DVMCacheMsg *);
    virtual void action(NCDVMCache *);
};


class DVMInvMsg : public DVMCacheMsg {
  protected:
    virtual unsigned length();
    virtual void rawCopyTo(unsigned * packet);
    virtual DVMCacheMsg * response();
  public:
    DVMInvMsg(unsigned dst, unsigned src, unsigned ts,
		    unsigned unit)
	: DVMCacheMsg(Inv,dst,src,ts,unit,0,0) {};
    DVMInvMsg(unsigned * top, unsigned length)
	: DVMCacheMsg(top, length) {};

    DVMCacheMsg * receive(DVMCacheMsg *);
    virtual void action(NCDVMCache *);
};


class DVMAckInvMsg : public DVMCacheMsg {
  protected:
    virtual unsigned length();
    virtual void rawCopyTo(unsigned * packet);
    virtual DVMCacheMsg * response();
  public:
    DVMAckInvMsg(unsigned dst, unsigned src, unsigned ts,
		    unsigned unit)
	: DVMCacheMsg(AckInv,dst,src,ts,unit,0,0) {};
    DVMAckInvMsg(unsigned * top, unsigned length)
	: DVMCacheMsg(top, length) {};

    DVMCacheMsg * receive(DVMCacheMsg *);
    virtual void action(NCDVMCache *);
};

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// 
// Messages that implement DVM Admin stuff.
// 

class DVMManager;

class DVMServiceMsg: public DVMMsg {
  protected:
    unsigned _monumber;
    char * namecopy;
    
    virtual void rawCopyTo(unsigned *) = 0;
    virtual unsigned length() = 0;
    
    virtual void inspect(OutputStream& os);
  public:
    // ctor for transmit
    DVMServiceMsg (DVMMsgType type, DVMManager * mgr, unsigned slot,
		   DVMServiceMsg * msg = 0);
    // ctor for receive
    DVMServiceMsg(unsigned * top, unsigned length);
    ~DVMServiceMsg();
    
    char * moname() { return namecopy; };
    unsigned monumber() { return _monumber; };

    virtual void action(DVMManager *) {};
};

inline
DVMServiceMsg::DVMServiceMsg(unsigned * top, unsigned length)
: DVMMsg()
{
    _top = top;
    _rawLength = length;
    unsigned * data = (unsigned *)this
	+ ((sizeof(DVMServiceMsg) + sizeof(unsigned) - 1)/sizeof(unsigned));
    namecopy = (char *)data;
}

class DVMJoinMsg: public DVMServiceMsg {
    virtual void rawCopyTo(unsigned * packet);
    virtual unsigned length();
  public:
    DVMJoinMsg(DVMManager * mgr, unsigned slot);
    DVMJoinMsg(unsigned * top, unsigned length)
	: DVMServiceMsg(top,length) {};
    
    virtual void action(DVMManager *);
};

class DVMLeaveMsg: public DVMServiceMsg {
    virtual void rawCopyTo(unsigned * packet);
    unsigned length();
  public:
    DVMLeaveMsg(DVMManager * mgr, unsigned slot);
    DVMLeaveMsg(unsigned * top, unsigned length)
	: DVMServiceMsg(top,length) {};
    
    virtual void action(DVMManager *);
};

class DVMAckJoinMsg: public DVMServiceMsg {
    virtual void rawCopyTo(unsigned * packet);
    virtual unsigned length();
  public:
    DVMAckJoinMsg(DVMManager * mgr, unsigned slot,
		  DVMServiceMsg * msg);
    DVMAckJoinMsg(unsigned * top, unsigned length)
	: DVMServiceMsg(top,length) {};
    
    virtual void action(DVMManager *);
};

class DVMAckLeaveMsg: public DVMServiceMsg {
    virtual void rawCopyTo(unsigned * packet);
    unsigned length();
  public:
    DVMAckLeaveMsg(DVMManager * mgr, unsigned slot,
		   DVMServiceMsg * msg);
    DVMAckLeaveMsg(unsigned * top, unsigned length)
	: DVMServiceMsg(top,length) {};
    
    virtual void action(DVMManager *);
};

#endif /* DVMMessage_h */
