/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef DVMManager_h
#define DVMManager_h


#ifdef __GNUC__
#pragma interface
#endif

#include "LOCK.h"
#include "Semaphore.h"
#include "List.h"

class PhysicallyAddressableUnit;
class List;
class Process;
class DVMManagerAction;
class OneCopyDVMCache;

static const int MaxNumberOfDVMMachines = 32; // should fit in an int, otherwise
					      // the per page array in the
					      // caches should be altered

class DVMManager {
  protected:
    int * _machines;		// List of mapping machines
    int _numMachines;
    int _thisMachine;		// bit position of this machine

    List _actionQueue;
    LOCK _actionMutex;
    Semaphore _awaitAction;
    
    List _outstandingQueue;
    LOCK _outstandingMutex;

    Process * _commProcess;

    OneCopyDVMCache * _cache;

    void sendToRemoteMgr(DVMManagerAction *);

  public:
    void commLoop();
    
    DVMManager(int numMachines, int thisMachine, OneCopyDVMCache * cache);
    ~DVMManager();
    
    void add(DVMManager * machine);
    int id();
    
    void send(int machine, unsigned unit, PhysicallyAddressableUnit * pau);
    PhysicallyAddressableUnit * rcv(int machine, unsigned unit);

    void enqueueA(DVMManagerAction *); // Use the Action Queue
    DVMManagerAction * dequeueA();     // Use the Action Queue
    void enqueueO(DVMManagerAction *);               // Use the Outstanding Queue
    DVMManagerAction * dequeueO(DVMManagerAction *); // Use the Outstanding Queue
};


class DVMManagerAction : public Link {
friend class DVMManager;
  protected:
    Semaphore _sem;
    int _unit;
    PhysicallyAddressableUnit * _pau;
    int _target;
    int _sender;
  public:
    int unit() { return _unit; };
    PhysicallyAddressableUnit * pau() { return _pau; };
    int target() { return _target; };
    int sender() { return _sender; };
    
    DVMManagerAction (int target, int unit);
    DVMManagerAction (int target, int unit, PhysicallyAddressableUnit * pau);
    void sleep();
    void awake();
    virtual int compare(Link * l);
};

// Request are queued when a unit is required from the remote machine.
// After request is satisfied, the same packet is returned with a pau
// instead of a unit.
// Copies of the requests are queued into the remote machine. Remote
// machine comm process identifies them since the target is the remote
// machine. Its response is a response packet with the pau.

// A queue can have an incoming response and request, but no other
// combination. 

// An action with pau == 0 is a request for a pau
// A request is outgoing if target is another machine

inline
DVMManagerAction::DVMManagerAction(int target, int unit)
: _sem(0)
{
    _target = target;
    _unit = unit;
    _pau = 0;
}

// An action with pau != 0 is a response
// A response is incoming if target is this machine

inline
DVMManagerAction::DVMManagerAction(int target, int unit,
				   PhysicallyAddressableUnit * pau)
: _sem(0)
{
    _target = target;
    _unit = unit;
    _pau = pau;
}

inline void
DVMManagerAction::sleep()
{
    _sem.P();
}

inline void
DVMManagerAction::awake()
{
    _sem.V();
}

inline int
DVMManagerAction::compare(Link * l)
{
    return ((DVMManagerAction *)l)->_unit == _unit;
}

//////////////////////////////////////////////////////////////////////
//
// These functions use DVMManagerAction methods, so they come after
// that class definition is completed.
//
//////////////////////////////////////////////////////////////////////

inline void
DVMManager::enqueueA(DVMManagerAction * action)
{
    _actionMutex.acquire();
    _actionQueue.addFirst(action);
    _awaitAction.V();
    _actionMutex.release();
}

inline DVMManagerAction *
DVMManager::dequeueA()
{
    DVMManagerAction * action;
    _actionMutex.acquire();
    action = (DVMManagerAction *)_actionQueue.removeLast();
    _actionMutex.release();

    return action;
}

inline void
DVMManager::enqueueO(DVMManagerAction * outstanding)
{
    _outstandingMutex.acquire();
    _outstandingQueue.addFirst(outstanding);
    _outstandingMutex.release();
}

// Not really dequeue, but find an action with the corresponding unit
// and return it. The Action::compare() method is used by remove()
inline DVMManagerAction *
DVMManager::dequeueO(DVMManagerAction * action)
{
    DVMManagerAction * outstanding;
    _outstandingMutex.acquire();
    outstanding = (DVMManagerAction *)_outstandingQueue.remove(action);
    _outstandingMutex.release();
    Assert(outstanding != 0);
    
    return outstanding;
}

inline int
DVMManager::id()
{
    return _thisMachine;
}


#endif /* DVMManager_h */
