/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef NCDVM_h
#define NCDVM_h


#ifdef __GNUC__
#pragma interface
#endif

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// VM States
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "SimpleVM.h"

class NCDVMVMState: public SimpleVMState {
  public:
    // Ctor for new states
    NCDVMVMState(StateMachine * m)
	: SimpleVMState(m) {};
    // Ctor for derived states
    NCDVMVMState(StateMachine * m, State * d)
	: SimpleVMState(m,d) {}
    
    virtual State * localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
    virtual State * localWrite(MemoryObjectCache * cache, unsigned unit,
			       AddressTranslation * adt);
    virtual State * pageOut(MemoryObjectCache * cache, unsigned unit);
    
    virtual FaultErrorCode code();
    
    virtual State * remoteWrite(MemoryObjectCache * cache, unsigned
				unit) = 0;
};

class NCDVM_VM_TRANS: public NCDVMVMState {
    // VM in transition
    // all local events are errors
  public:
    NCDVM_VM_TRANS(StateMachine * m)
	: NCDVMVMState(m) {};
    
    virtual State * localRead(MemoryObjectCache * , unsigned ,
			      AddressTranslation * )
    { return error(); };
    
    virtual State * localWrite(MemoryObjectCache * , unsigned ,
			       AddressTranslation * )
    { return error(); };
    
    virtual State * pageOut(MemoryObjectCache * , unsigned )
    { return error(); };
    
    virtual State * remoteWrite(MemoryObjectCache * , unsigned )
    { return error(); };
};

class NCDVM_VM_NP: public NCDVMVMState {
    // No page - page has been paged out to disk
    // local functions are inherited,
    // remoteWrite first brings in page 
  public:
    NCDVM_VM_NP(StateMachine * m, SimpleVMState * d)
	: NCDVMVMState(m,d) {};
    
    virtual State * remoteWrite(MemoryObjectCache * cache, unsigned unit);
};


class NCDVM_VM_P: public NCDVMVMState {
    // No page - page has been paged out to disk
    // local functions are inherited,
    // remoteWrite first brings in page 
  public:
    NCDVM_VM_P(StateMachine * m, SimpleVMState * d)
	: NCDVMVMState(m,d) {};
    
    virtual State * remoteWrite(MemoryObjectCache * cache, unsigned unit);
};


class NCDVM_VM_PD: public NCDVMVMState {
    // Page distributed
    // localAccesses - bring in page to be filled
    // pageOut - should not happen
    // remoteAccesses - ignore
  public:
    NCDVM_VM_PD(StateMachine * m)
	: NCDVMVMState(m) {};
    
    virtual State * localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
    virtual State * localWrite(MemoryObjectCache * cache, unsigned unit,
			       AddressTranslation * adt);
    virtual State * pageOut(MemoryObjectCache * cache, unsigned unit);

    virtual FaultErrorCode code();
    
    virtual State * remoteWrite(MemoryObjectCache * cache, unsigned unit);
};

class NCDVMVMMachine: public SimpleVMMachine {
  public:
    NCDVMVMMachine();
};

extern NCDVMVMState * STATE_NCDVM_VM_TRANS;
extern NCDVMVMState * STATE_NCDVM_VM_NP;
extern NCDVMVMState * STATE_NCDVM_VM_P;
extern NCDVMVMState * STATE_NCDVM_VM_PD;

extern NCDVMVMMachine * TheNCDVMVMMachine;

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Networking states
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class NCDVMState;
class NCDVMCache;
class DVMCacheMsg;

class NCDVMNetState : public State {
  public:
    NCDVMNetState(StateMachine * m) : State(m) {};
    
    virtual State * hereRemoteWrite(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, NCDVMState *);
    virtual State * remoteWriteDone(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, NCDVMState *);
    virtual State * timeout(NCDVMCache * cache, unsigned unit);
};

class NCDVM_NET_Q: public NCDVMNetState {
    // Quiescent
    // hereRemoteWrite - return remoteWriteDone
    // remoteWriteDone - return error
    // timeout - return error
  public:
    NCDVM_NET_Q(StateMachine * m) : NCDVMNetState(m) {};
};

class NCDVM_NET_SP: public NCDVMNetState {
    // Sending Page
    // hereRemoteWrite - return remoteWriteDone - ok if old from destination
    // remoteWriteDone - return NET_Q if from destination else error
    // timeout - send hereRemoteWrite
  public:
    NCDVM_NET_SP(StateMachine * m) : NCDVMNetState(m) {};
    virtual State * remoteWriteDone(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, NCDVMState *);
    virtual State * timeout(NCDVMCache * cache, unsigned unit);
};

class NCDVM_NET_WP: public NCDVMNetState {
    // Waiting for page
    // hereRemoteWrite - to NET_Q if from source else return remoteWriteDone
    // remoteWriteDone - error?
    // timeout - send remoteWrite
  public:
    NCDVM_NET_WP(StateMachine * m) : NCDVMNetState(m) {};
    virtual State * hereRemoteWrite(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, NCDVMState *);
    virtual State * timeout(NCDVMCache * cache, unsigned unit);
};

class NCDVMNetMachine: public StateMachine {
  public:
    NCDVMNetMachine();
};

extern NCDVMNetState * STATE_NCDVM_NET_Q;
extern NCDVMNetState * STATE_NCDVM_NET_SP;
extern NCDVMNetState * STATE_NCDVM_NET_WP;

extern NCDVMNetMachine * TheNCDVMNetMachine;


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Ownership state
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class DVMCacheMsg;

class NCDVMOwnerState: public State {
  public:
    NCDVMOwnerState(StateMachine * m) : State(m) {};
    
    virtual int currentMsg(NCDVMCache * cache, unsigned unit,
		   DVMCacheMsg * msg) = 0;

    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
    virtual State * ownerHint(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};

class NCDVM_O_OWN: public NCDVMOwnerState {
    // I own page
    // remoteWrite - make page inaccessible and set owner
    // ownerhint - make sure message is old
  public:
    NCDVM_O_OWN(StateMachine * m) : NCDVMOwnerState(m) {};
    
    int currentMsg(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg);
		   
    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};
class NCDVM_O_NOT: public NCDVMOwnerState {
    // I do not own page
    // remoteWrite - return OwnerHint
    // ownerhint - make sure message is old
  public:
    NCDVM_O_NOT(StateMachine * m) : NCDVMOwnerState(m) {};
    
    int currentMsg(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg);
    
    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};
class NCDVM_O_TOOWN: public NCDVMOwnerState {
    // I will eventually own page
    // drop remoteR/W
    // ownerhint from the correct source - register owner else
    // make sure message is old
  public:
    NCDVM_O_TOOWN(StateMachine * m) : NCDVMOwnerState(m) {};
    
    int currentMsg(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg);
    
    virtual State * ownerHint(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};
class NCDVM_O_TONOT: public NCDVMOwnerState {
    // I am transferring ownership
    // remoteWrite - drop any requests
    // ownerhint  - ignore
  public:
    NCDVM_O_TONOT(StateMachine * m) : NCDVMOwnerState(m) {};
    
    int currentMsg(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg);
};

class NCDVMOwnerMachine: public StateMachine {
  public:
    NCDVMOwnerMachine();
};

extern NCDVMOwnerState * STATE_NCDVM_O_OWN;
extern NCDVMOwnerState * STATE_NCDVM_O_NOT;
extern NCDVMOwnerState * STATE_NCDVM_O_TOOWN;
extern NCDVMOwnerState * STATE_NCDVM_O_TONOT;

extern NCDVMOwnerMachine * TheNCDVMOwnerMachine;


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Combination - the actual DVM state
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class NCDVMState: public SimpleVMState {
  protected:
    NCDVMVMState *const _vm;
    NCDVMNetState *const _net;
    NCDVMOwnerState *const _own;
  public:
    NCDVMState(StateMachine * m, State * d, NCDVMVMState * vm,
	       NCDVMNetState * net, NCDVMOwnerState * owner);
    NCDVMState(StateMachine * m, NCDVMVMState * vm,
	       NCDVMNetState * net, NCDVMOwnerState * owner);
    
    virtual FaultErrorCode code()
    { return _vm->code(); };
    
    virtual int currentMsg(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg)
    { return _own->currentMsg(cache,unit,msg); };
    
    virtual State * localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt)
    { _vm->localRead(cache,unit,adt); return returnState(this); };
    
    virtual State * localWrite(MemoryObjectCache * cache, unsigned unit,
			       AddressTranslation * adt)
    { _vm->localWrite(cache,unit,adt);  return returnState(this);};
    
    virtual State * pageOut(MemoryObjectCache * cache, unsigned unit)
    { _vm->pageOut(cache,unit);  return returnState(this);};
    
    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
				DVMCacheMsg * msg)
    { _own->remoteWrite(cache,unit,msg);  return returnState(this);};
    
    virtual State * ownerHint(NCDVMCache * cache, unsigned unit,
			      DVMCacheMsg * msg)
    { _own->ownerHint(cache,unit,msg); return returnState(this); };
    
    virtual State * hereRemoteWrite(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg)
    { _net->hereRemoteWrite(cache,unit,msg,this);  return returnState(this);};
    
    virtual State * remoteWriteDone(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg)
    { _net->remoteWriteDone(cache,unit,msg,this);  return returnState(this);};
    
    virtual State * timeout(NCDVMCache * cache, unsigned unit)
    { _net->timeout(cache,unit);  return returnState(this);};
};

class NCDVM_P: public NCDVMState {
    // Page owned and present
    // to NCDVM_NP for pageOut
    // to NCDVM_SP for remoteR/W
  public:
    NCDVM_P(StateMachine * m, State * d)
	: NCDVMState(m,d,
		     STATE_NCDVM_VM_P,
		     STATE_NCDVM_NET_Q,
		     STATE_NCDVM_O_OWN) {};
    
    virtual State * pageOut(MemoryObjectCache * cache, unsigned unit);
    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
				DVMCacheMsg * msg);
};

class NCDVM_NP: public NCDVMState {
    // Page owned but paged out
    // to NCDVM_P for localR/W
    // to NCDVM_P to NCDVM_SP for remoteR/W
  public:
    NCDVM_NP(StateMachine * m, State * d)
	: NCDVMState(m,d,
		     STATE_NCDVM_VM_NP,
		     STATE_NCDVM_NET_Q,
		     STATE_NCDVM_O_OWN) {};
    
    virtual State * localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
    virtual State * localWrite(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
				DVMCacheMsg * msg);
};

class NCDVM_PD: public NCDVMState {
    // Page distributed, we have no copy
    // to DVM_WP for localR/W
  public:
    NCDVM_PD(StateMachine * m)
	: NCDVMState(m,
		     STATE_NCDVM_VM_PD,
		     STATE_NCDVM_NET_Q,
		     STATE_NCDVM_O_NOT) {};
    
    virtual State * localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
    virtual State * localWrite(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
};

class NCDVM_SP: public NCDVMState {
    // Sending page to another site
    // No local events should happen
    // to NCDVM_PD when remoteWriteDone
  public:
    NCDVM_SP(StateMachine * m)
	: NCDVMState(m,
		     STATE_NCDVM_VM_TRANS,
		     STATE_NCDVM_NET_SP,
		     STATE_NCDVM_O_TONOT) {};
    
    virtual State * remoteWriteDone(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg);
};

class NCDVM_WP: public NCDVMState {
    // Waiting for page
    // No local events should happen
    // to NCDVM_P when hereWrite
  public:
    NCDVM_WP(StateMachine * m)
	: NCDVMState(m,
		     STATE_NCDVM_VM_TRANS,
		     STATE_NCDVM_NET_WP,
		     STATE_NCDVM_O_TOOWN) {};
    
    virtual State * ownerHint(NCDVMCache * cache, unsigned unit,
			      DVMCacheMsg * msg);
    virtual State * hereRemoteWrite(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg);
};

typedef State * (NCDVMState::*PMFCUMS)(NCDVMCache *,unsigned,DVMCacheMsg *);

class NCDVMMachine: public SimpleVMMachine {
  public:
    NCDVMMachine();
};

extern NCDVMState * STATE_NCDVM_NP;
extern NCDVMState * STATE_NCDVM_P;
extern NCDVMState * STATE_NCDVM_PD;
extern NCDVMState * STATE_NCDVM_SP;
extern NCDVMState * STATE_NCDVM_WP;

extern NCDVMMachine * TheNCDVMMachine;


#endif /* NCDVM_h */
