/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef NCDVM_h
#define NCDVM_h


#ifdef __GNUC__
#pragma interface
#endif

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// VM States
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "SimpleVM.h"

class NCDVMVMState: public SimpleVMState {
  public:
    // Ctor for new states
    NCDVMVMState(StateMachine * m)
	: SimpleVMState(m) {};
    // Ctor for derived states
    NCDVMVMState(StateMachine * m, State * d)
	: SimpleVMState(m,d) {}
    
    virtual State * localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
    virtual State * localWrite(MemoryObjectCache * cache, unsigned unit,
			       AddressTranslation * adt);
    virtual State * pageOut(MemoryObjectCache * cache, unsigned unit);
    
    virtual FaultErrorCode code();
    
    virtual State * remoteWrite(MemoryObjectCache * cache, unsigned
				unit) = 0;
};

class NCDVM_VM_TRANS: public NCDVMVMState {
    // VM in transition
    // all local events are errors
  public:
    NCDVM_VM_TRANS(StateMachine * m)
	: NCDVMVMState(m) {};
    
    virtual State * localRead(MemoryObjectCache * , unsigned ,
			      AddressTranslation * )
    { return error(); };
    
    virtual State * localWrite(MemoryObjectCache * , unsigned ,
			       AddressTranslation * )
    { return error(); };
    
    virtual State * pageOut(MemoryObjectCache * , unsigned )
    { return error(); };
    
    virtual State * remoteWrite(MemoryObjectCache * , unsigned )
    { return error(); };
};

class NCDVM_VM_NP: public NCDVMVMState {
    // No page - page has been paged out to disk
    // local functions are inherited,
    // remoteWrite first brings in page 
  public:
    NCDVM_VM_NP(StateMachine * m, SimpleVMState * d)
	: NCDVMVMState(m,d) {};
    
    virtual State * remoteWrite(MemoryObjectCache * cache, unsigned unit);
};


class NCDVM_VM_P: public NCDVMVMState {
    // No page - page has been paged out to disk
    // local functions are inherited,
    // remoteWrite first brings in page 
  public:
    NCDVM_VM_P(StateMachine * m, SimpleVMState * d)
	: NCDVMVMState(m,d) {};
    
    virtual State * remoteWrite(MemoryObjectCache * cache, unsigned unit);
};


class NCDVM_VM_PD: public NCDVMVMState {
    // Page distributed
    // localAccesses - bring in page to be filled
    // pageOut - should not happen
    // remoteAccesses - ignore
  public:
    NCDVM_VM_PD(StateMachine * m)
	: NCDVMVMState(m) {};
    
    virtual State * localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
    virtual State * localWrite(MemoryObjectCache * cache, unsigned unit,
			       AddressTranslation * adt);
    virtual State * pageOut(MemoryObjectCache * cache, unsigned unit);

    virtual FaultErrorCode code();
    
    virtual State * remoteWrite(MemoryObjectCache * cache, unsigned unit);
};

class NCDVMVMMachine: public SimpleVMMachine {
  public:
    NCDVMVMMachine();
    NCDVMVMMachine(int num) : SimpleVMMachine(num) {};
};

extern NCDVMVMState * STATE_NCDVM_VM_TRANS;
extern NCDVMVMState * STATE_NCDVM_VM_NP;
extern NCDVMVMState * STATE_NCDVM_VM_P;
extern NCDVMVMState * STATE_NCDVM_VM_PD;

extern NCDVMVMMachine * TheNCDVMVMMachine;

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Networking states
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class NCDVMState;
class NCDVMCache;
class DVMCacheMsg;

class NCDVMNetState : public State {
  public:
    NCDVMNetState(StateMachine * m) : State(m) {};
    NCDVMNetState(StateMachine * m, State * d) : State(m,d) {};
    
    virtual State * hereRemoteWrite(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, NCDVMState *);
    virtual State * remoteWriteDone(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, NCDVMState *);
    virtual State * timeout(NCDVMCache * cache, unsigned unit);
};

class NCDVM_NET_Q: public NCDVMNetState {
    // Quiescent
    // hereRemoteWrite - return remoteWriteDone
    // remoteWriteDone - return error
    // timeout - return error
  public:
    NCDVM_NET_Q(StateMachine * m) : NCDVMNetState(m) {};
};

class NCDVM_NET_SP: public NCDVMNetState {
    // Sending Page
    // hereRemoteWrite - return remoteWriteDone - ok if old from destination
    // remoteWriteDone - return NET_Q if from destination else error
    // timeout - send hereRemoteWrite
  public:
    NCDVM_NET_SP(StateMachine * m) : NCDVMNetState(m) {};
    virtual State * remoteWriteDone(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, NCDVMState *);
    virtual State * timeout(NCDVMCache * cache, unsigned unit);
};

class NCDVM_NET_WP: public NCDVMNetState {
    // Waiting for page
    // hereRemoteWrite - to NET_Q if from source else return remoteWriteDone
    // remoteWriteDone - error?
    // timeout - send remoteWrite
  public:
    NCDVM_NET_WP(StateMachine * m) : NCDVMNetState(m) {};
    virtual State * hereRemoteWrite(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, NCDVMState *);
    virtual State * timeout(NCDVMCache * cache, unsigned unit);
};

class NCDVMNetMachine: public StateMachine {
  public:
    NCDVMNetMachine();
    NCDVMNetMachine(int num) : StateMachine(num) {};
};

extern NCDVMNetState * STATE_NCDVM_NET_Q;
extern NCDVMNetState * STATE_NCDVM_NET_SP;
extern NCDVMNetState * STATE_NCDVM_NET_WP;

extern NCDVMNetMachine * TheNCDVMNetMachine;


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Ownership state
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class DVMCacheMsg;

class NCDVMOwnerState: public State {
  public:
    NCDVMOwnerState(StateMachine * m) : State(m) {};
    NCDVMOwnerState(StateMachine * m, State * d) : State(m,d) {};
    
    virtual int currentMsg(NCDVMCache * cache, unsigned unit,
		   DVMCacheMsg * msg) = 0;

    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
    virtual State * ownerHint(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};

class NCDVM_O_OWN: public NCDVMOwnerState {
    // I own page
    // remoteWrite - make page inaccessible and set owner
    // ownerhint - make sure message is old
  public:
    NCDVM_O_OWN(StateMachine * m) : NCDVMOwnerState(m) {};
    
    int currentMsg(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg);
		   
    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};
class NCDVM_O_NOT: public NCDVMOwnerState {
    // I do not own page
    // remoteWrite - return OwnerHint
    // ownerhint - make sure message is old
  public:
    NCDVM_O_NOT(StateMachine * m) : NCDVMOwnerState(m) {};
    
    int currentMsg(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg);
    
    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};
class NCDVM_O_TOOWN: public NCDVMOwnerState {
    // I will eventually own page
    // drop remoteR/W
    // ownerhint from the correct source - register owner else
    // make sure message is old
  public:
    NCDVM_O_TOOWN(StateMachine * m) : NCDVMOwnerState(m) {};
    
    int currentMsg(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg);
    
    virtual State * ownerHint(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};
class NCDVM_O_TONOT: public NCDVMOwnerState {
    // I am transferring ownership
    // remoteWrite - drop any requests
    // ownerhint  - ignore
  public:
    NCDVM_O_TONOT(StateMachine * m) : NCDVMOwnerState(m) {};
    
    int currentMsg(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg);
};

class NCDVMOwnerMachine: public StateMachine {
  public:
    NCDVMOwnerMachine();
    NCDVMOwnerMachine(int num) : StateMachine(num) {};
};

extern NCDVMOwnerState * STATE_NCDVM_O_OWN;
extern NCDVMOwnerState * STATE_NCDVM_O_NOT;
extern NCDVMOwnerState * STATE_NCDVM_O_TOOWN;
extern NCDVMOwnerState * STATE_NCDVM_O_TONOT;

extern NCDVMOwnerMachine * TheNCDVMOwnerMachine;


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Combination - the actual DVM state
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class NCDVMState: public SimpleVMState {
  protected:
    NCDVMVMState *const _vm;
    NCDVMNetState *const _net;
    NCDVMOwnerState *const _own;
  public:
    NCDVMState(StateMachine * m, State * d, NCDVMVMState * vm,
	       NCDVMNetState * net, NCDVMOwnerState * owner);
    NCDVMState(StateMachine * m, NCDVMVMState * vm,
	       NCDVMNetState * net, NCDVMOwnerState * owner);
    
    virtual FaultErrorCode code()
    { return _vm->code(); };
    
    virtual int currentMsg(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg)
    { return _own->currentMsg(cache,unit,msg); };
    
    virtual State * localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt)
    { _vm->localRead(cache,unit,adt); return returnState(this); };
    
    virtual State * localWrite(MemoryObjectCache * cache, unsigned unit,
			       AddressTranslation * adt)
    { _vm->localWrite(cache,unit,adt);  return returnState(this);};
    
    virtual State * pageOut(MemoryObjectCache * cache, unsigned unit)
    { _vm->pageOut(cache,unit);  return returnState(this);};
    
    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
				DVMCacheMsg * msg)
    { _own->remoteWrite(cache,unit,msg);  return returnState(this);};
    
    virtual State * ownerHint(NCDVMCache * cache, unsigned unit,
			      DVMCacheMsg * msg)
    { _own->ownerHint(cache,unit,msg); return returnState(this); };
    
    virtual State * hereRemoteWrite(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg)
    { _net->hereRemoteWrite(cache,unit,msg,this);  return returnState(this);};
    
    virtual State * remoteWriteDone(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg)
    { _net->remoteWriteDone(cache,unit,msg,this);  return returnState(this);};
    
    virtual State * timeout(NCDVMCache * cache, unsigned unit)
    { _net->timeout(cache,unit);  return returnState(this);};
};

class NCDVM_P: public NCDVMState {
    // Page owned and present
    // to NCDVM_NP for pageOut
    // to NCDVM_SP for remoteR/W
  public:
    NCDVM_P(StateMachine * m, State * d)
	: NCDVMState(m,d,
		     STATE_NCDVM_VM_P,
		     STATE_NCDVM_NET_Q,
		     STATE_NCDVM_O_OWN) {};
    
    virtual State * pageOut(MemoryObjectCache * cache, unsigned unit);
    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
				DVMCacheMsg * msg);
};

class NCDVM_NP: public NCDVMState {
    // Page owned but paged out
    // to NCDVM_P for localR/W
    // to NCDVM_P to NCDVM_SP for remoteR/W
  public:
    NCDVM_NP(StateMachine * m, State * d)
	: NCDVMState(m,d,
		     STATE_NCDVM_VM_NP,
		     STATE_NCDVM_NET_Q,
		     STATE_NCDVM_O_OWN) {};
    
    virtual State * localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
    virtual State * localWrite(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
				DVMCacheMsg * msg);
};

class NCDVM_PD: public NCDVMState {
    // Page distributed, we have no copy
    // to DVM_WP for localR/W
  public:
    NCDVM_PD(StateMachine * m)
	: NCDVMState(m,
		     STATE_NCDVM_VM_PD,
		     STATE_NCDVM_NET_Q,
		     STATE_NCDVM_O_NOT) {};
    
    virtual State * localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
    virtual State * localWrite(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
};

class NCDVM_SP: public NCDVMState {
    // Sending page to another site
    // No local events should happen
    // to NCDVM_PD when remoteWriteDone
  public:
    NCDVM_SP(StateMachine * m)
	: NCDVMState(m,
		     STATE_NCDVM_VM_TRANS,
		     STATE_NCDVM_NET_SP,
		     STATE_NCDVM_O_TONOT) {};
    
    virtual State * remoteWriteDone(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg);
};

class NCDVM_WP: public NCDVMState {
    // Waiting for page
    // No local events should happen
    // to NCDVM_P when hereWrite
  public:
    NCDVM_WP(StateMachine * m)
	: NCDVMState(m,
		     STATE_NCDVM_VM_TRANS,
		     STATE_NCDVM_NET_WP,
		     STATE_NCDVM_O_TOOWN) {};
    
    virtual State * ownerHint(NCDVMCache * cache, unsigned unit,
			      DVMCacheMsg * msg);
    virtual State * hereRemoteWrite(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg);
};

typedef State * (NCDVMState::*PMFCUMS)(NCDVMCache *,unsigned,DVMCacheMsg *);

class NCDVMMachine: public SimpleVMMachine {
  public:
    NCDVMMachine();
    NCDVMMachine(int num) : SimpleVMMachine(num) {};
};

extern NCDVMState * STATE_NCDVM_NP;
extern NCDVMState * STATE_NCDVM_P;
extern NCDVMState * STATE_NCDVM_PD;
extern NCDVMState * STATE_NCDVM_SP;
extern NCDVMState * STATE_NCDVM_WP;

extern NCDVMMachine * TheNCDVMMachine;

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// RCDVM
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// VM States
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class RCDVMVMState: public NCDVMVMState {
  public:
    // Ctor for new states
    RCDVMVMState(StateMachine * m)
	: NCDVMVMState(m) {};
    // Ctor for derived states
    RCDVMVMState(StateMachine * m, NCDVMVMState * d)
	: NCDVMVMState(m,d) {}
    
    virtual State * remoteWrite(MemoryObjectCache * , unsigned );
    virtual State * remoteUpgrade(MemoryObjectCache * cache, unsigned
				  unit) = 0;
    virtual State * remoteRead(MemoryObjectCache * cache, unsigned
			       unit) = 0;
    virtual State * invalidate(MemoryObjectCache * cache, unsigned
			       unit) = 0;
};

class RCDVM_VM_TRANS: public RCDVMVMState {
    // VM in transition
    // all local events are errors
  public:
    RCDVM_VM_TRANS(StateMachine * m, NCDVMVMState * d)
	: RCDVMVMState(m,d) {};

    virtual State * remoteUpgrade(MemoryObjectCache *, unsigned)
    { return error(); };
    virtual State * remoteRead(MemoryObjectCache *, unsigned)
    { return error(); };
    virtual State * invalidate(MemoryObjectCache *, unsigned)
    { return error(); };
};

class RCDVM_VM_NP: public RCDVMVMState {
    // No page - page has been paged out to disk
    // local functions and remoteWrite are inherited,
    // remoteRead first brings in page (currently, remoteRead == remoteWrite)
  public:
    RCDVM_VM_NP(StateMachine * m, NCDVMVMState * d)
	: RCDVMVMState(m,d) {};
    
    virtual State * remoteUpgrade(MemoryObjectCache *, unsigned)
    { return error(); };
    virtual State * remoteRead(MemoryObjectCache * cache, unsigned unit);
    virtual State * invalidate(MemoryObjectCache *, unsigned)
    { return error(); };
};

class RCDVM_VM_P: public RCDVMVMState {
    // page present
    // local functions and remoteWrite are inherited,
    // remoteRead should change protection - currently because of a
    // problem with the format of protect, it makes page inaccessible
    // (currently, remoteRead == remoteWrite)
  public:
    RCDVM_VM_P(StateMachine * m, NCDVMVMState * d)
	: RCDVMVMState(m,d) {};
    
    virtual State * remoteUpgrade(MemoryObjectCache *, unsigned)
    { return error(); };
    virtual State * remoteRead(MemoryObjectCache * cache, unsigned unit);
    virtual State * invalidate(MemoryObjectCache *, unsigned)
    { return error(); };
};

class RCDVM_VM_PD: public RCDVMVMState {
    // Page distributed
    // local and remotWrite is inherited
  public:
    RCDVM_VM_PD(StateMachine * m, NCDVMVMState * s)
	: RCDVMVMState(m,s) {};
    
    virtual State * remoteUpgrade(MemoryObjectCache *, unsigned)
    { return null(); };
    virtual State * remoteRead(MemoryObjectCache * , unsigned )
    { return null(); };
    virtual State * invalidate(MemoryObjectCache * , unsigned )
    { return null(); };
};

class RCDVM_VM_PRC: public RCDVMVMState {
    // Page distributed
  public:
    RCDVM_VM_PRC(StateMachine * m)
	: RCDVMVMState(m) {};
    
    virtual State * localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
    virtual State * localWrite(MemoryObjectCache * , unsigned ,
			       AddressTranslation * )
    { return error(); };
    virtual State * pageOut(MemoryObjectCache * , unsigned )
    { return error(); };
    virtual FaultErrorCode code();
    virtual State * remoteWrite(MemoryObjectCache * , unsigned );
    virtual State * remoteUpgrade(MemoryObjectCache *, unsigned);
    virtual State * remoteRead(MemoryObjectCache * , unsigned )
    { return null(); };
    virtual State * invalidate(MemoryObjectCache * , unsigned );
};

class RCDVMVMMachine: public NCDVMVMMachine {
  public:
    RCDVMVMMachine();
};

extern RCDVMVMState * STATE_RCDVM_VM_TRANS;
extern RCDVMVMState * STATE_RCDVM_VM_NP;
extern RCDVMVMState * STATE_RCDVM_VM_P;
extern RCDVMVMState * STATE_RCDVM_VM_PD;
extern RCDVMVMState * STATE_RCDVM_VM_PRC;

extern RCDVMVMMachine * TheRCDVMVMMachine;

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Networking states
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class RCDVMState;
class RCDVMCache;
class DVMCacheMsg;

class RCDVMNetState : public NCDVMNetState {
  public:
    // Ctor for new states
    RCDVMNetState(StateMachine * m) : NCDVMNetState(m) {};
    // Ctor for derived states
    RCDVMNetState(StateMachine * m, NCDVMNetState * d) : NCDVMNetState(m,d) {}
    
    virtual State * hereRemoteWrite(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, NCDVMState *);
    virtual State * remoteWriteDone(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, NCDVMState *);
    virtual State * timeout(NCDVMCache * cache, unsigned unit);
    virtual State * hereRemoteUpgrade(RCDVMCache * cache, unsigned unit,
				      DVMCacheMsg * msg, RCDVMState *);
    virtual State * remoteUpgradeDone(RCDVMCache * cache, unsigned unit,
				      DVMCacheMsg * msg, RCDVMState *);
    virtual State * hereRemoteRead(RCDVMCache * cache, unsigned unit,
				   DVMCacheMsg * msg, RCDVMState *);
    virtual State * checkInvalidate(RCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, RCDVMState *);
};

class RCDVM_NET_Q: public RCDVMNetState {
    // Quiescent
    // hereRemoteRead and ackInv - ignore if old else error
    // checkInv - error
  public:
    RCDVM_NET_Q(StateMachine * m, NCDVMNetState * s) : RCDVMNetState(m,s) {};
};

class RCDVM_NET_SP: public RCDVMNetState {
    // Sending Page
    // hereRemoteRead and ackInv - ignore if old else error
    // checkInv - error
  public:
    RCDVM_NET_SP(StateMachine * m, NCDVMNetState * s) : RCDVMNetState(m,s) {};
    
    virtual State * remoteWriteDone(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, NCDVMState *);
    virtual State * timeout(NCDVMCache * cache, unsigned unit);
    virtual State * remoteUpgradeDone(RCDVMCache * cache, unsigned unit,
				      DVMCacheMsg * msg, RCDVMState *);
};

class RCDVM_NET_WP: public RCDVMNetState {
    // Waiting for page
    // hereRemoteWrite - to NET_Q if from source else return remoteWriteDone
    // remoteWriteDone - error?
    // timeout - send remoteWrite
  public:
    RCDVM_NET_WP(StateMachine * m, NCDVMNetState * s) : RCDVMNetState(m,s) {};
    
    virtual State * hereRemoteWrite(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, NCDVMState *);
    virtual State * checkInvalidate(RCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg, RCDVMState * state);
    virtual State * hereRemoteUpgrade(RCDVMCache * cache, unsigned unit,
				      DVMCacheMsg * msg, RCDVMState *);
    virtual State * timeout(NCDVMCache * cache, unsigned unit);
};

class RCDVM_NET_WR: public RCDVMNetState {
    // Waiting for read page
    // hereRemoteRead - to NET_Q if from source else ignore
    // timeout - send remoteRead
  public:
    RCDVM_NET_WR(StateMachine * m) : RCDVMNetState(m) {};
    virtual State * hereRemoteRead(RCDVMCache * cache, unsigned unit,
				   DVMCacheMsg * msg, RCDVMState * state);
    virtual State * timeout(NCDVMCache * cache, unsigned unit);
};

class RCDVM_NET_SI: public RCDVMNetState {
    // Waiting for read page
    // ackInvalidate - to NET_Q if from all done else record
    // timeout - send remoteRead
  public:
    RCDVM_NET_SI(StateMachine * m) : RCDVMNetState(m) {};
    virtual State * timeout(NCDVMCache * cache, unsigned unit);
};

class RCDVMNetMachine: public NCDVMNetMachine {
  public:
    RCDVMNetMachine();
};

extern RCDVMNetState * STATE_RCDVM_NET_Q;
extern RCDVMNetState * STATE_RCDVM_NET_SP;
extern RCDVMNetState * STATE_RCDVM_NET_WP;
extern RCDVMNetState * STATE_RCDVM_NET_WR;
extern RCDVMNetState * STATE_RCDVM_NET_SI;

extern RCDVMNetMachine * TheRCDVMNetMachine;

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Ownership state
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class RCDVMOwnerState: public NCDVMOwnerState {
  public:
    // Ctor for new states
    RCDVMOwnerState(StateMachine * m) : NCDVMOwnerState(m) {};
    // Ctor for derived states
    RCDVMOwnerState(StateMachine * m, NCDVMOwnerState * d) : NCDVMOwnerState(m,d) {}
    
    virtual int currentMsg(NCDVMCache * cache, unsigned unit,
		   DVMCacheMsg * msg);

    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
    virtual State * ownerHint(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
    
    virtual State * remoteRead(RCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
    virtual State * remoteUpgrade(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg);
    virtual State * invalidate(RCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
    virtual State * ackInvalidate(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg);
};

class RCDVM_O_OWN: public RCDVMOwnerState {
    // I own page
    // remote/Write/Upgrade - set owner or copyholder
    // remoteRead - send the page
    // ackInvalidate - if current, remove the copyholder
  public:
    RCDVM_O_OWN(StateMachine * m, NCDVMOwnerState * s)
	: RCDVMOwnerState(m,s) {};
     
    virtual State * remoteUpgrade(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg);
    virtual State * remoteRead(RCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
    virtual State * ackInvalidate(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg);
};
class RCDVM_O_NOT: public RCDVMOwnerState {
    // I do not own page
    // remoteWrite - return OwnerHint
    // ownerhint - make sure message is old
  public:
    RCDVM_O_NOT(StateMachine * m, NCDVMOwnerState * s)
	: RCDVMOwnerState(m,s) {};
    
    virtual State * ownerHint(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
    virtual State * remoteUpgrade(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg);
    virtual State * remoteRead(RCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
    virtual State * invalidate(RCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};
class RCDVM_O_TONOT: public RCDVMOwnerState {
    // I am transferring ownership
    // remoteUpgrade - drop older and equal requests (from copies)
    // ackInvalidate - this is something we are waiting for
  public:
    RCDVM_O_TONOT(StateMachine * m, NCDVMOwnerState * s)
	: RCDVMOwnerState(m,s) {};

    virtual State * remoteUpgrade(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg);
    virtual State * ackInvalidate(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg);
};
class RCDVM_O_TOOC: public RCDVMOwnerState {
    // I will eventually own or have a copy of page
    // ownerhint from the correct source - register owner else
    // make sure message is old
    // invalidate - acknowledge, but keep on timeouting
  public:
    RCDVM_O_TOOC(StateMachine * m, NCDVMOwnerState * s)
	: RCDVMOwnerState(m,s) {};
    
    virtual State * ownerHint(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
    virtual State * invalidate(RCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};
class RCDVM_O_COPY: public RCDVMOwnerState {
    // I own page
    // remote/Write/Upgrade - set owner or copyholder
    // remoteRead - send the page
    // ackInvalidate - if current, remove the copyholder
  public:
    RCDVM_O_COPY(StateMachine * m)
	: RCDVMOwnerState(m) {};
    
    virtual int currentMsg(NCDVMCache * cache, unsigned unit,
			   DVMCacheMsg * msg);
    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
				DVMCacheMsg * msg);
    virtual State * remoteUpgrade(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg);
    virtual State * remoteRead(RCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
    virtual State * invalidate(RCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};

class RCDVMOwnerMachine: public NCDVMOwnerMachine {
  public:
    RCDVMOwnerMachine();
};

extern RCDVMOwnerState * STATE_RCDVM_O_OWN;
extern RCDVMOwnerState * STATE_RCDVM_O_NOT;
extern RCDVMOwnerState * STATE_RCDVM_O_TONOT;
extern RCDVMOwnerState * STATE_RCDVM_O_TOOC;
extern RCDVMOwnerState * STATE_RCDVM_O_COPY;

extern RCDVMOwnerMachine * TheRCDVMOwnerMachine;

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Combination - the actual RCDVM state
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class RCDVMState: public NCDVMState {
  public:
    RCDVMState(StateMachine * m, NCDVMState * d, RCDVMVMState * vm,
	       RCDVMNetState * net, RCDVMOwnerState * own)
	: NCDVMState(m,d,vm,net,own) {};
    RCDVMState(StateMachine * m, RCDVMVMState * vm,
	       RCDVMNetState * net, RCDVMOwnerState * own)
	: NCDVMState(m,vm,net,own) {};
    
    virtual FaultErrorCode code();
    virtual int currentMsg(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg);
    virtual State * localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt);
    virtual State * localWrite(MemoryObjectCache * cache, unsigned unit,
			       AddressTranslation * adt);
    virtual State * pageOut(MemoryObjectCache * cache, unsigned unit);
    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
				DVMCacheMsg * msg);
    virtual State * ownerHint(NCDVMCache * cache, unsigned unit,
			      DVMCacheMsg * msg);
    virtual State * hereRemoteWrite(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg);
    virtual State * remoteWriteDone(NCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg);
    virtual State * timeout(NCDVMCache * cache, unsigned unit);
    
    
    virtual State * remoteUpgrade(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg)
    { ((RCDVMOwnerState *)_own)->remoteUpgrade(cache,unit,msg);
      return returnState(this);};
    
    virtual State * hereRemoteUpgrade(RCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg)
    { ((RCDVMNetState *)_net)->hereRemoteUpgrade(cache,unit,msg,this);
      return returnState(this);};
    
    virtual State * remoteUpgradeDone(RCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg)
    { ((RCDVMNetState *)_net)->remoteUpgradeDone(cache,unit,msg,this);
      return returnState(this);};
    
    virtual State * remoteRead(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg)
    { ((RCDVMOwnerState *)_own)->remoteRead(cache,unit,msg);
      return returnState(this);};
    
    virtual State * hereRemoteRead(RCDVMCache * cache, unsigned unit,
				    DVMCacheMsg * msg)
    { ((RCDVMNetState *)_net)->hereRemoteRead(cache,unit,msg,this);
      return returnState(this);};
    
    virtual State * invalidate(RCDVMCache * cache, unsigned unit,
			      DVMCacheMsg * msg)
    { ((RCDVMOwnerState *)_own)->invalidate(cache,unit,msg);
      return returnState(this); };

    virtual State * ackInvalidate(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg)
    { ((RCDVMOwnerState *)_own)->ackInvalidate(cache,unit,msg);
      return returnState(this); };
};

class RCDVM_P: public RCDVMState {
    // Page owned and present
    // to RCDVM_ORC for remoteR
  public:
    RCDVM_P(StateMachine * m, NCDVMState * d)
	: RCDVMState(m,d,
		     STATE_RCDVM_VM_P,
		     STATE_RCDVM_NET_Q,
		     STATE_RCDVM_O_OWN) {};
    
    virtual State * remoteRead(RCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};
class RCDVM_NP: public RCDVMState {
    // Page owned and present
    // to RCDVM_ORC for remoteR
  public:
    RCDVM_NP(StateMachine * m, NCDVMState * d)
	: RCDVMState(m,d,
		     STATE_RCDVM_VM_NP,
		     STATE_RCDVM_NET_Q,
		     STATE_RCDVM_O_OWN) {};
    
    virtual State * remoteRead(RCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};

class RCDVM_ORC: public RCDVMState {
    // Page owned and present
    // to RCDVM_ORC for remoteR
    // to RCDVM_SI for upgrade
    // to RCDVM_NC for localWrite
  public:
    RCDVM_ORC(StateMachine * m)
	: RCDVMState(m,
		     STATE_RCDVM_VM_PRC,
		     STATE_RCDVM_NET_Q,
		     STATE_RCDVM_O_OWN) {};
    
    virtual State * remoteWrite(NCDVMCache * cache, unsigned unit,
				DVMCacheMsg * msg);
    virtual State * remoteUpgrade(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg);
    virtual State * localWrite(MemoryObjectCache * cache, unsigned unit,
			       AddressTranslation * adt);
};

class RCDVM_PD: public RCDVMState {
    // Page owned and present
    // to RCDVM_WR for localRead
  public:
    RCDVM_PD(StateMachine * m, NCDVMState * d)
	: RCDVMState(m,d,
		     STATE_RCDVM_VM_PD,
		     STATE_RCDVM_NET_Q,
		     STATE_RCDVM_O_NOT) {};
    
    virtual State * localRead(MemoryObjectCache * cache, unsigned unit,
			       AddressTranslation * adt);
};

class RCDVM_RC: public RCDVMState {
    // Page owned and present
    // to RCDVM_PD for invalidate
    // to RCDVM_WO for localWrite
  public:
    RCDVM_RC(StateMachine * m)
	: RCDVMState(m,
		     STATE_RCDVM_VM_PRC,
		     STATE_RCDVM_NET_Q,
		     STATE_RCDVM_O_COPY) {};
    
    virtual State * invalidate(RCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
    virtual State * localWrite(MemoryObjectCache * cache, unsigned unit,
			       AddressTranslation * adt);
};

class RCDVM_SI: public RCDVMState {
    // Page owned and present
    // to RCDVM_SP for ackInvalidate
  public:
    RCDVM_SI(StateMachine * m)
	: RCDVMState(m,
		     STATE_RCDVM_VM_TRANS,
		     STATE_RCDVM_NET_SI,
		     STATE_RCDVM_O_TONOT) {};
    
    virtual State * ackInvalidate(RCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
};

class RCDVM_SP: public RCDVMState {
    // Page owned and present
    // to RCDVM_PD for remoteUpgradeDone
  public:
    RCDVM_SP(StateMachine * m, NCDVMState * d)
	: RCDVMState(m,d,
		     STATE_RCDVM_VM_TRANS,
		     STATE_RCDVM_NET_SP,
		     STATE_RCDVM_O_TONOT) {};
    
    virtual State * remoteUpgradeDone(RCDVMCache * cache, unsigned unit,
				      DVMCacheMsg * msg);
};

class RCDVM_NC: public RCDVMState {
    // Page owned and present
    // to RCDVM_P for ackInvalidate
    // Drop any remote requests
  public:
    RCDVM_NC(StateMachine * m)
	: RCDVMState(m,
		     STATE_RCDVM_VM_TRANS,
		     STATE_RCDVM_NET_SI,
		     STATE_RCDVM_O_OWN) {};
    
    virtual State * ackInvalidate(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg);
    
    virtual State * remoteWrite(NCDVMCache * , unsigned ,
				DVMCacheMsg * )
    { return null(); };
    virtual State * remoteUpgrade(RCDVMCache * , unsigned ,
				  DVMCacheMsg * )
    { return null(); };
    virtual State * remoteRead(RCDVMCache * , unsigned ,
				  DVMCacheMsg * )
    { return null(); };
};

class RCDVM_WP: public RCDVMState {
    // Waiting for page
    // No local events should happen
    // To RCDVM_P when done
  public:
    RCDVM_WP(StateMachine * m, NCDVMState * d)
	: RCDVMState(m,d,
		     STATE_RCDVM_VM_TRANS,
		     STATE_RCDVM_NET_WP,
		     STATE_RCDVM_O_TOOC) {};
    
    virtual State * invalidate(RCDVMCache * cache, unsigned unit,
				  DVMCacheMsg * msg);
};

class RCDVM_WR: public RCDVMState {
    // Waiting for page
    // No local events should happen
    // to RCDVM_RC when hereRead
  public:
    RCDVM_WR(StateMachine * m)
	: RCDVMState(m,
		     STATE_RCDVM_VM_PRC,
		     STATE_RCDVM_NET_WR,
		     STATE_RCDVM_O_TOOC) {};
    
    virtual State * hereRemoteRead(RCDVMCache * cache, unsigned unit,
				   DVMCacheMsg * msg);
};

class RCDVM_WO: public RCDVMState {
    // Waiting for page
    // No remote events should happen
    // to RCDVM_P when hereUpgrade
    // to RCDVM_PD when invalidate
  public:
    RCDVM_WO(StateMachine * m)
	: RCDVMState(m,
		     STATE_RCDVM_VM_PRC,
		     STATE_RCDVM_NET_WP,
		     STATE_RCDVM_O_COPY) {};
    
    virtual State * hereRemoteUpgrade(RCDVMCache * cache, unsigned unit,
				      DVMCacheMsg * msg);
    virtual State * invalidate(RCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg);
    
    virtual State * remoteWrite(NCDVMCache * , unsigned ,
				DVMCacheMsg * )
    { return null(); };
    virtual State * remoteUpgrade(RCDVMCache * , unsigned ,
				  DVMCacheMsg * )
    { return null(); };
    virtual State * remoteRead(RCDVMCache * , unsigned,
				  DVMCacheMsg * )
    { return null(); };
};

typedef State * (RCDVMState::*PMFCUMR)(RCDVMCache *,unsigned,DVMCacheMsg *);

class RCDVMMachine: public NCDVMMachine {
  public:
    RCDVMMachine();
};

extern RCDVMState * STATE_RCDVM_NP;
extern RCDVMState * STATE_RCDVM_P;
extern RCDVMState * STATE_RCDVM_PD;
extern RCDVMState * STATE_RCDVM_ORC;
extern RCDVMState * STATE_RCDVM_RC;
extern RCDVMState * STATE_RCDVM_NC;
extern RCDVMState * STATE_RCDVM_SP;
extern RCDVMState * STATE_RCDVM_SI;
extern RCDVMState * STATE_RCDVM_WP;
extern RCDVMState * STATE_RCDVM_WR;
extern RCDVMState * STATE_RCDVM_WO;

extern RCDVMMachine * TheRCDVMMachine;

#endif /* NCDVM_h */
