/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef NCDVMCache_h
#define NCDVMCache_h


#ifdef __GNUC__
#pragma interface
#endif

#include "SimpleVMCache.h"	/* parent class */
#include "DVMAccessType.h"
#include "DVMManager.h"
#include "OneCopyDVM.h"

class State;
class OneCopyDVMState;
class DVMManager;
class DVMCacheMsg;

/*
 * Define the size of units or "pages" in a NCDVMCache.
 * The size of the pages has to be a power of 2 multiple of the hardware
 * page size.
 * Should be consistent with RMOC..Size.
 */
static const unsigned int NCDVMCacheLog2UnitSize = 12;

extern ClassStar NCDVMCacheClass;
class NCDVMCache : public SimpleVMCache {
/*
 * A NCDVMCache is the normal cache for use by user
 * programs. It is implemented using a state machine, whereas
 * PagedMemoryObjectCache contains normal inlined code.
 */

  protected:
    
    unsigned * _machines;
    char * _fragments;
    unsigned * _timestamps;

    LOCK _localEventMutex;
    BitRanges * _localEventGate;
    LOCK _blockMutex;
    List * _blockedEvents;		// faulted Processes
    
    DVMManager * _manager;
    
    static void DVMCacheTimeout(void *, int);
    GroupTimer * _cacheTimer;
    void cacheTimeout(unsigned unit);
	
    TimeoutTimer _delayTimer;
    int _delayFlag;
    static void DVMActionProcess(int);
    List _actionQueue;
    LOCK _actionMutex;
    Semaphore * _awaitAction;
    Process * _actionProcess;
    DVMCacheMsg * deqAction();
    void actionProcessEntry();

    Semaphore _stateMutex;
    
    int areLocalEventsEnabled(unsigned from, unsigned to = 0);

  public:
  
    Class * classOf();

    void setManager( DVMManager * manager );
    
    NCDVMCache( MemoryObject * memobj, unsigned int length,
		     unsigned id, int copy );
    ~NCDVMCache();

//     virtual void pagesToNet(int requester, unsigned from, unsigned to = 0);
//     virtual void pagesFromNet(unsigned from, unsigned to = 0);
    
//     virtual void
// 	remoteCache(unsigned unit, DVMAccessType attemptedAccess, int requester);


    void sendMsg(DVMCacheMsg * msg) { _manager->enqXmit(msg); };
    
    void enableLocalEvents(unsigned from, unsigned to = 0);
    void disableLocalEvents(unsigned from, unsigned to = 0);

    unsigned timestamp(unsigned unit) { return _timestamps[unit]; };
    void nextStamp(unsigned unit, unsigned timestamp);

    unsigned fragment(unsigned unit) { return _fragments[unit]; };
    unsigned setFragment(unsigned unit, unsigned fragment)
    { return _fragments[unit] |= fragment; };
    void clearFragment(unsigned unit) { _fragments[unit] = 0; };

    void enableTimeout(unsigned unit) { _cacheTimer->enable(unit); };
    void disableTimeout(unsigned unit) { _cacheTimer->disable(unit); };

    unsigned owner(unsigned unit) { return _machines[unit]; };
    unsigned setOwner(unsigned unit, unsigned owner)
    { return _machines[unit] = owner; };
    unsigned clearOwner(unsigned unit, unsigned owner)
    { return _machines[unit] &= ~owner; };
    unsigned thisMachine() { return _manager->localId(); };

    unsigned * unitAddress(unsigned unit);
    
    void enqAction(DVMCacheMsg *);
    void delayActionProcess() { _delayFlag = 1; };

    void stateAction(PMFCUMS,DVMCacheMsg *);

    virtual PhysicalMemoryChain
	cache(unsigned int unit,
	      ProtectionLevel requiredProtection, FaultErrorCode & code,
	      AddressTranslation * adt, const char * vaddr, int addToAdt=0);
    virtual PhysicalMemoryChain
	mcache(unsigned int unit, int nunits,
	       ProtectionLevel requiredProtection, FaultErrorCode & code,
	       AddressTranslation * adt, const char * vaddr, int addToAdt=0);
    
};

inline void
NCDVMCache::enqAction(DVMCacheMsg * msg)
{
    _actionMutex.acquire();
    _actionQueue.addLast(msg);
    _awaitAction->V();
    _actionMutex.release();
}

inline DVMCacheMsg *
NCDVMCache::deqAction()
{
    DVMCacheMsg * msg;
    _actionMutex.acquire();
    msg = (DVMCacheMsg *)_actionQueue.removeFirst();
    _actionMutex.release();
    return msg;
}

inline void
NCDVMCache::setManager(DVMManager * manager)
{
    _manager = manager;
}

extern ClassStar RCDVMCacheClass;
class RCDVMCache : public NCDVMCache {
  protected:
    unsigned * _copyholders;
  public:
    Class * classOf();
    
    RCDVMCache( MemoryObject * memobj, unsigned int length,
		     unsigned id, int copy );
    ~RCDVMCache();
    
    unsigned copyholders(unsigned unit) { return _copyholders[unit]; };
    unsigned setCopyholders(unsigned unit, unsigned holder)
    { return _copyholders[unit] |= holder; };
    unsigned unsetCopyholders(unsigned unit, unsigned holder)
    { return _copyholders[unit] &= ~holder; };
    
    void stateAction(PMFCUMR,DVMCacheMsg *);
    void cacheTimeout(unsigned unit);
    
    void incrementStamp(unsigned unit) { ++_timestamps[unit]; };
    void setStamp(unsigned unit, unsigned stamp) { _timestamps[unit] = stamp; };
};

#endif /* NCDVMCache_h */
