/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef State_h
#define State_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Assert.h"

// The problem: Designing subclassable state machines.
// Example: Given a machine with states A,B and events p,q, we want to design a
// machine with state A,C::B with a new event r. Behavior of events
// for state C::B is the same as state B for event p, but different
// for event q.

// States are defined as objects, and events are member functions of
// the State object. Every event function returns the next state.
// A StateMachine class is also defined for every collection of
// states. Such a machine serves to instantiate the states and provide
// a virtual function table for returning states.

// In our example, (say) A::p() returns B. When we create a subclassed
// state machine that uses C::B that is derived from B, A::p() is
// expected to return C::B and not B. This is achieved by returning
// states through a virtual function call. 

// States are numbered. States derived from a single parent return the
// same number. Virtual functions that return states are defined to
// return numbered states. E.g function StateMachine::one() would
// return state number 1. The StateMachines keeps a table of states. A
// function such as one() indexes the table to return the proper states.

// When a state machine is subclassed, the derived machine
// instantiates that state table with subclasses of states or the
// original states as appropriate.
// In our example, the parent machine would have a table
// _states[0] = new A;
// _states[1] = new B;
// The derived machine would have a table
// _states[0] = new A;
// _states[1] = new C;
// The virtual function two() would return _states[1], thereby
// returning the correct state for each machine.

// In this case, we are using objects as jump tables (i.e. Code rather
// than data. Only one instance of each object exists in the system.

// For a detailed example of use, see the DVM system which is defined
// using state machines. Also see file State.cc for an example.

// A StateMachine creates an array of states

class State;

class StateMachine {
  private:
    int _newStatesStarted;	// to verify new states added after replaced
    int _numStates;
    State ** _states;
  protected:
    void newState(State *);
    void replaceState(State *);
  public:
    StateMachine(int numStates);
    ~StateMachine();
    
    State * returnState(State *);
};

// A State has a number which is used to place the state in the
// correct position in the array. It is also used to identify replace
// states when a subclass state is used in a subclassed State Machine.
// Every state in a given state machine should have a distinct number.

class State {
friend class StateMachine;
  private:
    StateMachine *const _machine;
    unsigned _number;
    void setNumber(unsigned n) { _number = n; };
  protected:
    State *const _delegatee;
    virtual State * error();
    virtual State * null();
    virtual State * returnState(State * s) { return _machine->returnState(s); };
  public:
    const unsigned number() { return _number; };
    
    State(StateMachine * m);
    State(StateMachine * m, State * d);
    virtual ~State();
};


// create a state and register with the state machine.
// initialize the global variable of the same name.
// must be called from within the StateMachine ctor;

#define NEWSTATE(state) { STATE_##state = new state(this); \
			      newState(STATE_##state); }

#define REPLACESTATE(state,ancestor) \
{ \
    Assert(STATE_##ancestor != 0); \
    Assert(STATE_##ancestor->number() > 0); \
    STATE_##state = new state(this,STATE_##ancestor); \
    if (STATE_##ancestor->number() != STATE_##state->number()) {\
        Console() << "STATE_" #state " ignores ancestor STATE_" #ancestor "\n" \
	    << eor; \
        Assert(NOTREACHED); \
    } \
    replaceState(STATE_##state); \
}

#endif /* State_h */
