/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef VMState_h
#define VMState_h


#ifdef __GNUC__
#pragma interface
#endif

// The virtual memory system is a state machine. Vanilla virtual
// memory is the simplest. Support for Copy-on-write, DVM can be added
// as subclasses of the state machine.

// When a state machine is subclassed, states and events can be added
// and redefined. If every state is an object, states in the refined
// machine are subclasses of state objects.

// An event is a member function of the state object. It executes some
// actions, and returns the next state. In a refined state machine, new
// transitions can be added as new member functions of the derived state.
// Existing events can have new actions or return different states.
// Every event takes a parameter representing the user of the machine,
// and state machine actions take effect on the user object.

// Events in a parent state machine return states from the parent machine.
// If events from the parent machine are reused in the child, and the
// returned state has been subclassed in the child machine, the
// subclassed state should be returned. This is achieved by returning
// the next state through member functions of the state machine.
// The member functions can be redefined when subclassing the entire
// state machine.


class VMStateMachine;

// VMState is the parent for all other state objects.

class VMState {
  protected:
    VMStateMachine * _machine;
  public:
    VMState(VMStateMachine * machine);
    
    virtual int number() = 0;
};

VMState::VMState(machine)
{
    Assert( machine != 0 );
    _machine = machine;
}

// VMStateMachine is parent state machine.

class VMStateMachine {
  protected:
    int _numStates;
    VMState ** _states;
  public:
    VMStateMachine(int numStates);
    virtual ~VMStateMachine();

    void newState( VMState * s );
};

inline
VMStateMachine::VMStateMachine(int numStates)
{
    _numStates = numStates;
    _states = new VMState * [ _numStates ];
    for (int i = 0; i < _numStates; i++)
	_states[i] = 0;
}

inline void
VMStateMachine::newState(VMState * s)
{
    Assert( s->number() < numStates );
    Assert( _states[s->number()] == 0 );
    _states[s->number()] = s;
}

#define VM_NEWSTATE(state) newState(new state())

#endif /* VMState_h */
