/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Domain.h: description of the Domain class. A Domain is a subclass of
 *	     SpaceMaps which adds reference counting and address translation.
 */

#ifndef Domain_h
#define Domain_h


#ifdef __GNUC__
#pragma interface
#endif

#include "AccessType.h"
//#include "AddressTranslation.h"
#include "DomainStar.h"
#include "FaultErrorCode.h"
#include "FileSystemInterfaceStar.h"
#include "Lock.h"
#include "MemoryObjectCache.h"
#include "MemoryObjectStar.h"
#include "MemoryObjectViewStar.h"
#include "NameServerStar.h"
#include "ObjectProxyManagerStar.h"
#include "ProtectionLevel.h"
#include "ProxiableObject.h"			/* parent class */
#include "Types.h"

class PhysicalMemoryChain;
class FileSystemInterface;
class LockHolder;
class MemoryObject;
class MemoryObjectView;
class ObjectProxy;
class NameServer;
class AddressTranslation;

static const char * const NotMapped = (const char *) 0xffffffff;

#ifdef MULTIMAX
class Module;
#endif

class Domain;
/*
 * The class 'DomainCleaner' is used to reference count the Domain
 *  by processes such that the domain can be 'cleaned' (that is, the
 *  name server and proxies deleted) when the last process dies and
 *  before the Domain itself is deleted.   The destructor of the
 *  DomainCleaner simply calls cleanup() on the Domain.
 */
class DomainCleaner : public ProxiableObject {
protected:
	Domain * _domain;
public:
	DomainCleaner( Domain * );
	~DomainCleaner();
};

extern ClassStar DomainClass;
class Domain : public ProxiableObject {
	friend class PhysicalMemoryChain;

protected:
	Lock lock;
	DomainCleaner * _cleaner;
	unsigned  maxMemoryObjectsInMap;
	unsigned  numberOfMemoryObjectsInMap;
	struct List {
		const char * baseAddr;
		const char * highAddr;
		int length;
		ProtectionLevel access;
		int shared;
		MemoryObjectView * memoryObjectView;
	} * list;

	char * _startAddress;
	unsigned _size;
	unsigned _last;

	AddressTranslation  * _translation;
	ObjectProxyManagerStar _proxyManager;
	ObjectProxy * _defaultProxy;
	PFV _returnFunct;
	DomainStar _parent;  /* for referencing direct parent Domain only */
	NameServerStar _nameServer;
	FileSystemInterfaceStar _fileSystemInterface;
#ifdef MULTIMAX
	Module * _moduleTableEntry;
#endif

	/*
	 * Look through our heritage returning the domain
	 * that manages the address or zero.
	 */
	Domain * managerOf( char *addr);

	/*
	 * These 'basic' methods really implement the corresponding
	 *  functionality as described below.  The methods without the
	 *  'basic' first determine whether to call this 'basic' method
	 *  or that of one of the parents.
	 */
	MemoryObjectViewRef basicLookup( const char * vaddr, 
		const char * & baseAddr, unsigned & length,
		ProtectionLevel & permissionDesired );

	FaultErrorCode basicRepairFault( const char * address,
				AccessType attemptedAccess );
	/*
	 * Force a unit to be cached in physical memory.
	 * Return the PhysicallyAddressableUnit it is in.
	 */
	virtual LockHolder *
		makeResidentInPhysicalMemory(MemoryObjectView * memobj, 
					     const char * baseAddr, 
					     unsigned size,
					     unsigned offset, 
					     unsigned length,
					     ProtectionLevel requestedProt,
					     ProtectionLevel & grantedProt,
					     FaultErrorCode & code,
					     AddressTranslation * adt,
					     const char * vaddr, 
					     CacheReason reason );

public:
	Domain( Domain * parentDomain , char *startaddress, unsigned size);
	~Domain();
	void cleanup();
	DomainCleaner * cleaner();

	/*
	 * See if this exact (not parent) domain manages the address
	 */
	int validAddress(const void * const address);

	/*
	 * Convert a virtual address into a (memory object,offset) pair.
	 */
	MemoryObjectViewRef lookup( const char * vaddr, 
		const char * & baseAddr, unsigned & length, 
		ProtectionLevel & permissionDesired );

	/*
	 * Called by the page fault handler to repair a page fault
	 */
	FaultErrorCode repairFault( const char * address,
				AccessType attemptedAccess );

	/*
	 * Force a range of addresses into physical memory and add the
	 * corresponding mappings to the address translation for this Domain.
	 */
	void makeAddressable( const char * address, unsigned int length );

	/*
	 * Force a range of addresses into physical memory and add the
	 * corresponding mappings to the address translation for this Domain.
	 */
	void makeUnAddressable( const char * address, unsigned int length );
       
	/*
	 * This method forces all the memory in a range to be cached and
	 * then returns a PMChain describing it.
	 */
	PhysicalMemoryChain constructChain( const char * baseVirtualAddress,
		unsigned int length, int makeAddr = 0 );

	/*
	 * Add a memory object at a partcular virtual address with a particular
	 * access permission.  If the memory object is marked as shared,
	 * it will not be copied to the new domain during a fork(); instead,
	 * it will be simply added to the new domain.  This lets us
	 * implement shared memory.
	 */
	const char * add( MemoryObject * memory,
		const char * vaddr, ProtectionLevel permission, int
		shared = 0 );

	/*
	 * Add a memory object wherever it will fit with a particular access
	 * permission.  If the memory object is marked as shared,
	 * it will not be copied to the new domain during a fork(); instead,
	 * it will be simply added to the new domain.  This lets us
	 * implement shared memory.
	 */
	const char * add( MemoryObject * memory, ProtectionLevel permission,
		int shared = 0 );


	/*
	 * Remove a memory object.
	 */
	const char * remove( MemoryObject * memory, const char * baseAddr );

	/*
	 * Get the object proxy manager for this domain.
	 */
	ObjectProxyManager * proxyManager();
	/*
	 * Get/Set the default return function for this domain.
	 */
	PFV returnFunct();
	void setReturnFunct( PFV funct );
	/*
	 * Get the defaultProxy for this domain.
	 */
	ObjectProxy * defaultProxy();

	/*
	 * Get the translation for this domain.
	 */
	AddressTranslation * translation();

	/*
	 * Return the NameServer for this Domain.
	 */
	NameServer * nameServer();

	/*
         * Set minor data elements of two domains equal. Assists copy.
         */
	void setDataEqual( Domain * oldDomain );

	/*
	 * Make one Domain's memory object mappings a copy of another.
	 */
	void copy( Domain * );

	/*
	 * Prepare for exec, clear all memory objects except the proxy 
	 * memory object
	 */
	void prepareForExec();

	/*
	 * return a parent of this domain.
	 */
	Domain * parent( );


#ifdef MULTIMAX
	/*
	 * set/get Module Table Entry for this domain
	 */
	void setModuleTableEntry( void * starttext, void * startdata );
	Module * moduleTableEntry();
#endif

	Class * classOf();
	
};

inline DomainCleaner *
Domain::cleaner()
{
	return ( _cleaner );
}

inline AddressTranslation *
Domain::translation()
{
	return( _translation );
}

inline ObjectProxyManager *
Domain::proxyManager()
{
	return( _proxyManager );
}

inline PFV
Domain::returnFunct()
{
	return( _returnFunct );
}

inline void
Domain::setReturnFunct( PFV funct )
{
	_returnFunct = funct;
}

inline Domain *
Domain::parent()
{
	return( _parent );
}

inline NameServer *
Domain::nameServer()
{
	return( _nameServer );
}

inline int
Domain::validAddress(const void * const addr)
{
        return( ((char*)addr >= _startAddress) &&
        	((char*)addr < _startAddress + _size) );
}

#ifdef MULTIMAX
inline Module *
Domain::moduleTableEntry( )
{
	return( _moduleTableEntry );
}
#endif

#endif /* Domain_h */
