/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1993
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * ReplacementPolicy.h
 */

#ifndef GlobalLRUReplacementPolicy_h
#define GlobalReplacementPolicy_h

#ifdef __GNUC__
#pragma interface
#endif

#include "PagedMemoryObjectCache.h"
#include "Process.h"
#include "RangeLocks.h"
#include "ReplacementPolicy.h"
#include "Semaphore.h"

#define NEW_GLRU

class MemoryObjectCache;

class GlobalLRUReplacementPolicy;

struct GlobalLRUInfo {
#ifdef NEW_GLRU
	int _prev;
	int _next;
#endif // NEW_GLRU
	GlobalLRUReplacementPolicy * _rp;
	unsigned _from;
	unsigned _to;
};

class GlobalLRUInfoTable {
	int _used;
	int _size;
	GlobalLRUInfo * _array;
	static Lock _lock;
	static void acquire() { _lock.acquire(); }
	static void release() { _lock.release(); }
	void grow();
	void garbageCollect();
      public:
	GlobalLRUInfoTable();
	int  size() { return _used; }
	void init();
	void check( GlobalLRUReplacementPolicy * rp, unsigned from, 
		    unsigned to, unsigned locked );
	void add( GlobalLRUReplacementPolicy * rp, unsigned from, 
		  unsigned to );
	void remove( GlobalLRUReplacementPolicy * rp, unsigned from, 
		     unsigned to, int hint );
	int next( int current, 
		  GlobalLRUReplacementPolicy * & rp,  
		  unsigned & from, unsigned & to );
};

class AsyncPageOutClosure;

extern ClassStar GlobalLRUReplacementPolicyClass;

class GlobalLRUReplacementPolicy : public ReplacementPolicy {
      protected:
	static GlobalLRUInfoTable _table;
	static int _paging;
	static Lock _pagingLock;
	static Semaphore _pagerWait;
	static Process * _pager;
	static int _firstHand;
	static int _distance;

	static void pageOutDaemon();
	virtual unsigned victim( unsigned count, 
				 unsigned & start, unsigned & end,
				 PhysicallyAddressableUnit * & pau );

#ifdef NEW_GLRU
	friend class GlobalLRUInfoTable;
	int _head;
#endif // NEW_GLRU

      public:
	static Process * pager() { return _pager; }
	static void initialize();
	Class * classOf();
	GlobalLRUReplacementPolicy( PagedMemoryObjectCache * cache );
	~GlobalLRUReplacementPolicy();
	virtual void unitsToStore( unsigned from, unsigned to );
	virtual void postPageIn( unsigned from, unsigned to );
	virtual unsigned pageOut( unsigned noOfUnits );
};

extern ClassStar SmarterGlobalLRUReplacementPolicyClass;

class SmarterGlobalLRUReplacementPolicy : public GlobalLRUReplacementPolicy {
      protected:
	static const unsigned _minLog2FreeBehindLag = 17;
	static const unsigned _slowLog2FreeBehindLag = 20;
	static const unsigned _log2ClusterSize = 16;

	static Lock _sglock;
	int _readSequential;
	int _writeSequential;
	int _doSyncBehind;
	int _doPrefetchFirst;
	unsigned _readPointer;
	unsigned _readAhead;
	unsigned _readFreeBehind;
	unsigned _writePointer;
	unsigned _writeFreeBehind;
	unsigned _nextSynchronize;
	unsigned _clusterSize;
	int _queued;
	unsigned _readAheadLag;
	unsigned _freeBehindLag;
	unsigned _slowFreeBehindLag;
	unsigned _maxFreeBehindLag;
	unsigned _minFreeBehindLag;
	unsigned _lastTime;

	friend class FreeBehindIOAsyncClosure;
	friend class ReadAheadIOAsyncClosure;
	friend class SynchronizeBehindAsyncClosure;

	void updateFreeBehindLag();
	void doFreeBehind( unsigned & from, unsigned current );
	void doSynchronizeBehind();
	void freeBehindDone( unsigned from, unsigned to );
	void readAheadDone( unsigned from, unsigned to );
	void synchronizeDone( unsigned from, unsigned to );

      public:
	Class * classOf();
	SmarterGlobalLRUReplacementPolicy( PagedMemoryObjectCache * cache );
	virtual void preCacheRead( unsigned from, unsigned to );
	virtual void postCacheRead( unsigned from, unsigned to );
	virtual void postCacheWrite( unsigned from, unsigned to );
	virtual unsigned pageOut( unsigned noOfUnits );
};

extern ClassStar HintedGlobalLRUReplacementPolicyClass;

class HintedGlobalLRUReplacementPolicy : 
      public SmarterGlobalLRUReplacementPolicy
{
      protected:
	int _hints;
	int _origMaxFreeBehindLag;
	void updateParams( int newHints );
      public:
	Class * classOf();
	HintedGlobalLRUReplacementPolicy( PagedMemoryObjectCache * cache );
	virtual void preCacheRead( unsigned from, unsigned to );
	virtual void preCacheWrite( unsigned from, unsigned to );
	virtual void notify( int );
};

#endif // GlobalLRUReplacementPolicy_h

