/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Map.h
 */

#ifndef Map_h
#define Map_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Lock.h"
#include "Object.h"
#include "ProtectionLevel.h"

class PhysicallyAddressableUnit;

class Map : public Object {
protected:
	Lock * _lock;

	unsigned _size;
	unsigned _residentSize;
	unsigned _log2UnitSize;
	unsigned _log2PauUnitSizeRatio;
	unsigned _pauUnitSizeRatio;
	unsigned _pauUnitSizeRatioMask;

	int	unitModified( PhysicallyAddressableUnit * );
	virtual PhysicallyAddressableUnit * basicPauList( unsigned unit, 
				PhysicallyAddressableUnit * & lastPau ) = 0;
	virtual PhysicallyAddressableUnit * basicRemoveUnit( unsigned unit,
				PhysicallyAddressableUnit * & lastPau ) = 0;

public:
	Map( int size, int log2UnitSize );
	~Map();
	void acquire();
	void release();
	unsigned log2UnitSize();
	unsigned residentSize();
	virtual void setSize( unsigned size ) = 0;
	virtual PhysicallyAddressableUnit * pauList( unsigned from, 
						     unsigned to );
	virtual void addPauList( PhysicallyAddressableUnit * plist,
			         unsigned from, unsigned to ) = 0;
	virtual PhysicallyAddressableUnit * findAndRemove( unsigned from, 
						           unsigned to,
							   unsigned &start,
							   unsigned &end );

	/* Methods to find next existing or non-existing range of units */

	virtual int findNextExist( unsigned from, unsigned to );
	virtual int findNextExist( unsigned from, unsigned to,
				   unsigned & start, unsigned & end,
				   PhysicallyAddressableUnit * & startPau );
	virtual int findNextNonExist( unsigned from, unsigned to,
				      unsigned & start, unsigned & end,
				      PhysicallyAddressableUnit * & fromPau );
	virtual int findNextModified( unsigned from, unsigned to,
				      unsigned & start, unsigned & end,
				      PhysicallyAddressableUnit * & fromPau );
	
	/* Methods to determine and modify unit state */

	int exists( unsigned unit );
	int modified( unsigned unit );
	int referenced( unsigned unit );
	ProtectionLevel protection( unsigned unit );
	void setModified( unsigned unit, int v );
	void setModified( unsigned from, unsigned to, int v );
	void setReferenced( unsigned unit, int v );
	void setReferenced( unsigned from, unsigned to, int v );
	void setProtection( unsigned unit, ProtectionLevel prot );
	void setProtection( unsigned from, unsigned to, ProtectionLevel );
	void setReferencedModified( unsigned unit,
				    int referenced, int modified );
	void setReferencedModified( unsigned from, unsigned to, 
				    int referenced, int modified );
};

inline void 
Map::acquire()
{
	_lock->acquire();
}

inline void
Map::release()
{
	_lock->release();
}

inline unsigned
Map::log2UnitSize()
{
	return _log2UnitSize;
}

inline unsigned
Map::residentSize()
{
	return _residentSize;
}

class TableMap : public Map {
protected:
	virtual PhysicallyAddressableUnit ** where( unsigned unit, int alloc )
		                             = 0;
	virtual PhysicallyAddressableUnit * basicPauList( unsigned unit, 
				PhysicallyAddressableUnit * & lastPau );
	virtual PhysicallyAddressableUnit * basicRemoveUnit( unsigned unit,
				PhysicallyAddressableUnit * & lastPau );
public:
        TableMap( int size, int log2UnitSize );
	virtual void addPauList( PhysicallyAddressableUnit * plist,
			         unsigned from, unsigned to ); 
};

inline 
TableMap::TableMap( int size, int log2PauUnitSizeRatio ) 
        : Map( size, log2PauUnitSizeRatio ) {}

class MultiLevelTable;

class MultiLevelTableMap : public TableMap {
protected:
	static Lock _staticLock;
	MultiLevelTable * _table;

	virtual PhysicallyAddressableUnit ** where( unsigned unit, int alloc );

public:
	MultiLevelTableMap( unsigned size, unsigned log2UnitSize );
	~MultiLevelTableMap();
	virtual void setSize( unsigned size );
};

class ConstantSizeTableMap: public TableMap {
protected:
	Lock _instanceLock;
	PhysicallyAddressableUnit ** _table;

	virtual PhysicallyAddressableUnit ** where( unsigned unit, int alloc );
public:
	ConstantSizeTableMap( int size, int log2UnitSize );
	~ConstantSizeTableMap();
	virtual void setSize( unsigned size );
};

class PremappedMap : public Map
{
	Lock _instanceLock;
	const char * _baseAddr;

protected:
	virtual PhysicallyAddressableUnit * basicPauList( unsigned unit,
				PhysicallyAddressableUnit * & lastPau );
	virtual PhysicallyAddressableUnit * basicRemoveUnit( unsigned unit,
				PhysicallyAddressableUnit * & lastPau );

public:
	PremappedMap( int size, int log2UnitSize, 
		      const char * baseAddr );
	virtual void setSize( unsigned size );
	virtual void addPauList( PhysicallyAddressableUnit * plist,
			         unsigned from, unsigned to ); 
};

#endif /*Map_h*/


