/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MemoryObject.h: MemoryObject class descriptions.
 */

#ifndef MemoryObject_h
#define MemoryObject_h

#ifdef __GNUC__
#pragma interface
#endif

#include "ClassStar.h"
#include "FaultErrorCode.h"
#include "MemoryRange.h"    /* parent class */
#include "MemoryObjectStar.h"
#include "MemoryObjectCacheStar.h"
#include "ProtectionLevel.h"
#include "RangeLocks.h"

#define INLINE_MOLOCKS

class AddressTranslation;
class IOAsyncClosure;
class RangeLock;
class PhysicalMemoryChain;
class Process;

extern ClassStar MemoryObjectClass;

class MemoryObject : public MemoryRange {
 protected:
    
    static Lock _lock;
    static Lock _hintLock;
    RangeLock * _moLock;

    /*
     * Used to set the maximum memory protection allowed.
     */
    int _protection : 3;

    /*
     * Used to figure out optimal caching policy.
     */
    unsigned char _cacheLog2UnitSize;
    unsigned char _cacheLog2UnitSizeRatio;
    Class * _cacheClass;
    Class * _replacementPolicyClass;
    MemoryObjectCacheStar _cache;

    /* 
     * Cache hints
     */
    unsigned _hints;
    unsigned _segmentSize;

    /*
     * Protected methods
     */
    ~MemoryObject();
    void noRemainingReferences();

    /*
     * This method is used to determine whether to flush the data in
     * the MemoryObjects cache back to the MemoryObject when it
     * has no remaining references.  Assume only called by
     * noRemainingReferences
     */
    virtual int transient();

    /*
     * Protected methods: 
     * These methods assume the proper locks has been acquired
     */

    friend class ReleaseIOAsyncClosure;
    friend class SynchronizeIOAsyncClosure;
    friend class FileObject;

    /*
     * assume entire mo locked
     */
    virtual void basicFlush();
    virtual void basicSynchronize( IOAsyncClosure * c );
    virtual void basicSynchronize( unsigned from, unsigned to, 
                                   IOAsyncClosure * c );
    virtual unsigned basicSetNumberOfUnits( unsigned unit );
    virtual int basicCopy( MemoryObject * dest );

    /*
     * actual read and write methods, range lock has already been
     * acquired
     */
    virtual int basicRead( unsigned start, unsigned count, 
                           char * addr, unsigned nbytes, IOAsyncClosure * c );
    virtual int basicWrite( unsigned start, unsigned count,
                            char * addr, unsigned nbytes, IOAsyncClosure * c );
    virtual int basicWriteThrough( unsigned start, unsigned count,
                                   char * addr, unsigned nbytes, 
                                   IOAsyncClosure * c );    
    virtual int cacheRead( unsigned start, unsigned count, 
                           char * addr, unsigned nbytes, IOAsyncClosure * c );
    virtual int cacheWrite( unsigned start, unsigned count,
                            char * addr, unsigned nbytes, IOAsyncClosure * c );
    friend class Cache;
    friend class DummyCache;
    virtual int rawRead( unsigned start, unsigned count,
                         char * addr, unsigned nbytes, IOAsyncClosure * c ) = 0;
    virtual int rawWrite( unsigned start, unsigned count,
                          char * addr, unsigned nbytes,
                          int writeThrough, IOAsyncClosure * c ) = 0;


 public:
    Class * classOf();

    proxiable MemoryObject( unsigned numberOfUnits, unsigned log2UnitSize );

    proxiable virtual int fixedSize();

    proxiable virtual int read( unsigned startUnit, int count,
                                char * addr, int nbytes );
    proxiable virtual int write( unsigned startUnit, int count,
                                 char * addr, int nbytes );
    proxiable virtual int writeThrough( unsigned startUnit, int count,
                                        char * addr, int nbytes );

    /*
     * Flush any cached data back to the backing store.
     */
    proxiable virtual void synchronize();
    proxiable virtual void flush();

    /*
     * Copy and change size of memory object
     */
    proxiable virtual int setNumberOfUnits( unsigned numberOfUnits );
    proxiable virtual int copy( MemoryObject * dest );

    /*
     * Stuff used by VM system when a MemoryObject is made directly
     * addressable (memory mapped).
     */
    RangeLock * moLock();
    virtual ProtectionLevel logicalProtectionLevel();
    virtual void setMemoryObjectCache( MemoryObjectCache * cache );
    virtual MemoryObjectCache * buildCache();
    virtual MemoryObjectCache * cache();
    proxiable virtual unsigned cacheLog2UnitSize();
    proxiable virtual unsigned setCacheLog2UnitSize( unsigned );
    proxiable virtual Class * cacheClass();
    proxiable virtual void setCacheClass( Class * );
    proxiable virtual Class * replacementPolicyClass();
    proxiable virtual void setReplacementPolicyClass( Class * );
    proxiable virtual unsigned maxResidentCacheUnits();
    proxiable virtual unsigned setMaxResidentCacheUnits( unsigned units );
    proxiable virtual void addCache( Class * cacheClass );

    /*
     * Asynchronous operations
     */

    virtual void asyncRead( unsigned int startUnit, int count, 
                            char * addr, int nbytes, IOAsyncClosure * c );
    virtual void asyncWrite( unsigned int startUnit, int count,
                             char * addr, int nbytes, IOAsyncClosure * c );
    virtual void asyncWriteThrough( unsigned int startUnit, int count,
                                    char * addr, int nbytes,
                                    IOAsyncClosure * c );

    /*
     * Methods for locking and unlocking
     * only to be used by subclasses, and 
     * subclasses of MemoryObjectContainer and MemoryObjectViews
     */

    Process * acquire( unsigned from, unsigned to, int block = 1, 
                       int locked = 0 );
    Process * release( unsigned from, unsigned to );
    void acquireAll();
    void releaseAll();
    void acquireSize();
    void releaseSize();
    void downgradeFromAll( unsigned from, unsigned to );

    /*
     * Cache hints
     */

    unsigned segmentSize()          { return _segmentSize; }
    unsigned setSegmentSize( unsigned s )   { _segmentSize = s; return s;}
    unsigned hints()            { return _hints; }
    virtual void setHints( unsigned mask );
    virtual void clearHints( unsigned mask );
};

#ifdef INLINE_MOLOCKS

inline
Process * 
MemoryObject::acquire( unsigned from, unsigned to, int block, int locked )
{
    return ( _moLock ? _moLock->acquire( from, to, block, locked ) : 0 );
}

inline
Process *
MemoryObject::release( unsigned from, unsigned to )
{
    return ( _moLock ? _moLock->release( from, to ) : (Process *) 0x4 );
}

inline
void 
MemoryObject::acquireAll()
{
    if ( _moLock )
        _moLock->acquireAll();
}

inline
void
MemoryObject::releaseAll()
{
    if ( _moLock )
        _moLock->releaseAll();
}

inline
void 
MemoryObject::acquireSize()
{
    if ( _moLock )
        _moLock->acquire( MaxUse, MaxUse );
}

inline
void
MemoryObject::releaseSize()
{
    if ( _moLock )
        _moLock->release( MaxUse, MaxUse );
}

inline
void 
MemoryObject::downgradeFromAll( unsigned from, unsigned to )
{
    if ( _moLock )
        _moLock->downgradeFromAll( from, to );
}

#endif /* INLINE_MOLOCKS */

#endif /*MemoryObject_h*/
