/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MemoryObjectCache.h: Cache Memory Object data in physical memory
 */

#ifndef MemoryObjectCache_h
#define MemoryObjectCache_h

#ifdef __GNUC__
#pragma interface
#endif

#include "AccessType.h"
#include "BlockedInfo.h"
#include "FaultErrorCode.h"
#include "Link.h"
#include "Lock.h"
#include "MemoryObjectViewStar.h"
#include "MemoryRange.h"
#include "ProtectionLevel.h"
#include "RangeLocks.h"
#include "Semaphore.h"
#include "StoreInfo.h"

class AddressTranslation;
class Cache;
class CacheUnitInfo;
class IOAsyncClosure;
class IOWindowInfo;
class LockHolder;
class Map;
class MemoryObject;         
class OutputStream;
class PhysicallyAddressableUnit;    
class PhysicalMemoryChain;
class Process;
class RangeLock;
class ReplacementPolicy;
class TranslationMap;

enum CacheReason {
    DomainRepairFault = 0x01,
    DomainMakeAddressable = 0x2,
    DomainConstructChain = 0x4
};

enum IOOperationCode {
    ToMemoryObject,
    ToCache,
    FromCache
};

#define UNITSTOSTORE_SKIPRP 0x1
#define UNITSTOSTORE_QUEUED     0x2

#define SAVEMODIFIED_FLUSH  0x1
#define SAVEMODIFIED_QUEUED 0x2

extern ClassStar MemoryObjectCacheClass;

class MemoryObjectCache : public MemoryRange {
/*
 * A MemoryObjectCache caches data from MemoryObjects into pages. It
 * also supports other vm operations such as protection manipulation,
 * pageout, etc. A MemoryObjectCache is the hub of the VM system.
 * MemoryObjectCache is a abstract class. If you define new methods
 * take care to see that SpecialPurposeCache is not affected.
 */

 protected:

    /*
     * Fundamental data structures
     */
    Map * _map;
    Cache * _cache;
    TranslationMap * _adtMap;
    ReplacementPolicy * _replacementPolicy;

    /*
     * For book keeping in Store
     */

    StoreInfo _storeInfo;

    /*
     * Relationship between memory object and cache unit
     */
    RangeLock * _moLock;
    MemoryObject * _memoryObject;
    int _moUnitSizeRatioMask;
    int _log2MoUnitSizeRatio;

    /*
     * Relationship between cache unit and pau unit
     */
    int _log2PauUnitSizeRatio;    

    /*
     * Relationship between kernel window size and unit size
     */
    unsigned char _log2WindowSize;  /* window size in log2 bytes */
    int _windowSizeRatio;
    int _windowSizeRatioMask;

    /*
     * Some info for hints
     */
    unsigned _bytesAccessed;
    void bytesAccessedAdd( unsigned bytes );
    unsigned _fromCacheUnits;
    void fromCacheUnitsAdd( unsigned units );

    /*
     * Kernel window data structures
     */
    static AddressTranslation * _translation;

    unsigned _noOfWindows;
    IOWindowInfo * _ioWindow;
    IOWindowInfo * _cacheWindow;

    friend class ReplacementPolicy;
    friend class IOWindowInfo;
    friend class IOWindowCache;

    PhysicallyAddressableUnit * forward( PhysicallyAddressableUnit * pau,
                                         unsigned noOfUnits );
    ProtectionLevel protection( PhysicallyAddressableUnit * pau,
                                unsigned noOfUnits );   
    void setProtection( PhysicallyAddressableUnit * pau,
                        unsigned noOfUnits, ProtectionLevel prot ); 
    void setReferenced( PhysicallyAddressableUnit * pau,
                        unsigned noOfUnits, int ref );
    void setModified( PhysicallyAddressableUnit * pau,
                      unsigned noOfUnits, int mod );
    void setReferencedModified( PhysicallyAddressableUnit * pau,
                                unsigned noOfUnits, int ref, int mod );
    void setCacheUnitInfo( PhysicallyAddressableUnit * pau, 
                           unsigned noOfUnits, CacheUnitInfo * cu );

    /* Methods for moLock acquisition and release */
    int log2WindowSize() { return _log2WindowSize; }
    int windowSizeRatio() { return _windowSizeRatio; }
    int windowSizeRatioMask() { return _windowSizeRatioMask; }
    unsigned noOfWindows() { return _noOfWindows; }
    void setNoOfWindows( int i ) { _noOfWindows = i; }
    int computeLog2WindowSize();
    Process * acquireMo( unsigned from, unsigned to, int block = 1,
                         int locked = 0 );
    Process * releaseMo( unsigned from, unsigned to, int locked = 0 );
    void acquireAllMo();
    void releaseAllMo();
    void acquireMoLock();
    void releaseMoLock();

    /* Method cleanup: release all pages held and delete window */
    virtual void cleanUp();

    /* Method preFault: grab required page range or block. */
    /* Method postFault: release page range and blocked processes. */
    void preFault( unsigned from, unsigned to );
    void postFault( unsigned from, unsigned to );

/*
 * Internal methods, mainly to facilitate subclassing
 */

    /* Method storeAlloc: Allocate pages from store. This method is
     * used to differentiate between store allocation calls for
     * special purpose kernel caches and normal caches. 
     * Used in unitsFromStore.
     */
    virtual PhysicallyAddressableUnit * storeAlloc( unsigned units ) = 0;
    
    /* Method storeDealloc: Deallocate pages to store. This method is
     * used to differentiate between store deallocation calls for
     * special purpose kernel caches and normal caches. 
     * Used in unitsToStore.
     */
    virtual void storeDealloc( PhysicallyAddressableUnit * pau, 
                               unsigned queued ) = 0;

    /* Method memoryObjectAlloc: used to allocate backing store
     */
    virtual unsigned memoryObjectAlloc( unsigned from, unsigned to ) = 0;

/*
 * Methods for page movement between MemoryObjectCaches and other
 * system objects
 */

    /* Method unitsFromStore: get pages from store, and handle null
     * returns when store is paging.
     */
    virtual PhysicallyAddressableUnit * unitsFromStore( unsigned from, 
                                                        unsigned to );
    
    /* Method unitsToStore: return pages to the Store. */
    virtual int unitsToStore(unsigned from, unsigned to, int code );
    
    /* Method to inform Store that units are in the process of being
     * freed 
     */
    virtual void unitsToStoreQueued( unsigned nunits ) = 0;

    /* Method saveModified: saveModified pages only to mo */
    virtual void saveModified( IOOperationCode operation,
                               unsigned from, unsigned to, int code,
                               IOAsyncClosure * c );
    /* Method load: bring non-exisintg units into cache */
    virtual unsigned load( unsigned from, unsigned to, IOAsyncClosure * c );
    
    /*
     * Cache and MemoryObject I/O
     */

    friend class MOCacheReadClosure;
    friend class MOCacheReadAggregateState;
    friend class MOCacheWriteClosure;
    friend class MOCacheWriteAggregateState;
    friend class IOOperationClosure;
    friend class IOOperationAggregateState;

    virtual unsigned ioOperation( IOOperationCode operation,
                                  unsigned from, unsigned to, int flush,
                                  PhysicallyAddressableUnit * pau,
                                  IOAsyncClosure * c );
    virtual unsigned toMemoryObject( unsigned from, unsigned to, 
                                     int flush, 
                                     PhysicallyAddressableUnit * pau,
                                     IOAsyncClosure * c );
    virtual unsigned toCache( unsigned from, unsigned to, int flush,
                              PhysicallyAddressableUnit * pau,
                              IOAsyncClosure * c );
    virtual unsigned fromCache( unsigned from, unsigned to,
                                PhysicallyAddressableUnit * pau,
                                IOAsyncClosure * c );

/*
 * Net I/O
 */

    /* Method toNet: Send pages to net */
    /* Method fromNet: Read pages from net */
    virtual void toNet(unsigned from, unsigned to );
    virtual void fromNet(unsigned from, unsigned to );
    
/*
 * Methods for interesting VM stuff. These methods have no
 * control, and are combined to get public methods such as ::cache.
 */
 
    /*
     * alter protection - flush adts, change pagetables
     */
    
    /* Method nprotect: change protection on given range. Will be
     * renamed to simply protect soon
     * Method whatProtection: find the protection level of the given
     * range. Currently unimplemented
     */
    virtual int nprotect( AddressTranslation * adt, const char * vaddr,
                          ProtectionLevel prot, 
                          unsigned from, unsigned to );
    virtual ProtectionLevel whatProtection( unsigned from, 
                                            unsigned to );
    
    /*
     * page table control
     */
    
    /* Method hackTranslat: For internal use, should be cleaned up.
     */

    virtual void hackTranslat( CacheReason reason, 
                               unsigned from, unsigned to,
                               ProtectionLevel prot, 
                               AddressTranslation * adt, 
                               const char * vaddr, 
                               MemoryObjectView * view,
                               const char * base, 
                               FaultErrorCode code,
                               PhysicallyAddressableUnit * pau );

/*
 * Methods to modify reference status, 
 */

    virtual int clearReferenced( unsigned from, unsigned to );
    virtual PhysicallyAddressableUnit * 
    checkReferenced( unsigned from, unsigned to, int mode );
    
 public:

    /* Method setDebugMask: set the bit-mask of messages types
     *                      that should be debugged.
     */
    proxiable virtual void setDebugMask( unsigned short mask );

    /* Method setDebugMessages: set the number of messages 
     *                          that should be debugged.
     */
    proxiable virtual void setDebugMessages( short messages );

/*
 * Public methods for using the cache
 */ 

    /* For Store */
    StoreInfo * storeInfo();
    
    /* Maintain relation with MemoryObject */
    
    /* Method memoryObject: return MemoryObject. */
    MemoryObject * memoryObject();

    /* Method to return bytesAccessed */
    unsigned bytesAccessed();
    void bytesAccessedReset();

    /* Method to return fromCacheUnits */
    unsigned fromCacheUnits();
    void fromCacheUnitsReset();

    /* Miscellaneous */
    
    /* Miscellaneous public methods: */

    /* Method writeName: Interface for Debug */
    void writeName( OutputStream * );
    
    /* Method classOf: Return MemoryObjectClass. */
    Class * classOf();
    
    /* Method MemoryObjectCache: Constructor */
    MemoryObjectCache( MemoryObject * memoryObject );

    /* Method ~MemoryObjectCache(): dtor, called by unref. */
    ~MemoryObjectCache();
    
/* 
 * These methods should be called by Domain only
 */

    /* Method cache: Cache the unit and update tlb if indicated. 
     * Method mcache: Cache the range and update tlb if indicated.   
     */
    virtual LockHolder *
    cache( unsigned unit,
           ProtectionLevel requiredProtection, 
           ProtectionLevel & grantedProtection, 
           FaultErrorCode & code,
           AddressTranslation * adt, const char * vaddr, 
           CacheReason reason, MemoryObjectView * view,
           const char * baseAddr );
    virtual LockHolder *
    mcache( unsigned unit, int nunits,
            ProtectionLevel requiredProtection, 
            ProtectionLevel & grantedProtection, 
            FaultErrorCode & code,
            AddressTranslation * adt, const char * vaddr, 
            CacheReason reason, MemoryObjectView * view,
            const char * baseAddr );

    /* Method translatable: put given range into page table with given
     * protection. If previous mappings exist, they are not replaced.
     * This means that old protection values can still exist. This
     * semantics is open for debate.
     */
    virtual void translatable( unsigned from, unsigned to, 
                               ProtectionLevel prot,
                               AddressTranslation * adt, 
                               const char * vaddr,
                               MemoryObjectView * view, 
                               const char * base,
                               PhysicallyAddressableUnit * pau );

    /* Method untranslatable: remove mappings from given adt or all
     * adts.
     */
    virtual void untranslatable(unsigned from, unsigned to, 
                                AddressTranslation * adt = 0,
                                const char * where = 0);
    virtual void untranslatable(AddressTranslation * adt,
                                MemoryObjectView * view,
                                const char * where);

/*
 * This method should be called by Store only
 */

    /* Method pageOut: Remove some page, write to MemoryObject if
     * dirty and return it.
     */
    virtual unsigned pageOut( unsigned noOfPaus ) = 0;  

/*
 * This method should be called LockHolder only
 */ 

    /* Method unlock: used by LockHolder to unlock and call postFault
     *                for the units it has locked
     */
    virtual void unlock( unsigned from, unsigned to );         

/*
 * Called by Allocator to release memory
 */
    virtual void free();
    virtual void free( unsigned from, unsigned to );

/* 
 * These methods should be called by MemoryObjects only
 * The memory object is expected do the locking the range
 */

    /* Method cacheRead:
     * Method cacheWrite:
     */
    virtual unsigned
    cacheRead( unsigned start, unsigned noOfBytes, char * buf,
               IOAsyncClosure * c );
    virtual unsigned
    cacheWrite( unsigned start, unsigned noOfBytes, char * buf,
                IOAsyncClosure * c );

    /* Method setNumberOfUnits: inform cache of changes in the memory 
     * object's size 
     */
    virtual int setNumberOfUnits( unsigned noOfUnits ) = 0;

    /* Method: to determine cache replacment policy
     */
    ReplacementPolicy * replacementPolicy();
    virtual void setReplacementPolicyClass( Class * policyClass ) = 0;

    /* Method: Flush all cached data back to the memory object, remove
     * all pages and translation mappings.
     */
    virtual void flush();
    virtual void flush( unsigned from, unsigned to );

    /* Method: Synchronize, write all modified data back to store
     */
    virtual void synchronize( IOAsyncClosure * c );
    virtual void synchronize( unsigned from, unsigned to, 
                              IOAsyncClosure * c );

    /* Method: PseudoSynchronize, write all allocation info back
     */
    virtual void pseudoSynchronize();

    /* Method: addCache, add a level to the memory hierarchy
     *         hasCache, check if cache type exists
     */
    virtual void addCache( Class * cacheClass ) = 0;
    virtual int hasCache( Class * cacheClass );

};

inline 
StoreInfo *
MemoryObjectCache::storeInfo()
{
    return & _storeInfo;
}

inline 
MemoryObject *
MemoryObjectCache::memoryObject()
{
    return _memoryObject;
}

inline
unsigned
MemoryObjectCache::bytesAccessed()
{
    return _bytesAccessed;
}

inline
unsigned
MemoryObjectCache::fromCacheUnits()
{
    return _fromCacheUnits;
}

inline 
ReplacementPolicy *
MemoryObjectCache::replacementPolicy()
{
    return _replacementPolicy;
}

inline
Process * 
MemoryObjectCache::acquireMo( unsigned from, unsigned to, int block, int locked )
{
    return _moLock->acquire
    ( from << _log2MoUnitSizeRatio, 
      ( to << _log2MoUnitSizeRatio ) + _moUnitSizeRatioMask,
      block, locked );
}

inline
Process *
MemoryObjectCache::releaseMo( unsigned from, unsigned to, int locked )
{
    return _moLock->release
    ( from << _log2MoUnitSizeRatio, 
      ( to << _log2MoUnitSizeRatio ) + _moUnitSizeRatioMask, 
      locked );
}

inline
void 
MemoryObjectCache::acquireAllMo()
{
    _moLock->acquireAll();
}

inline
void
MemoryObjectCache::releaseAllMo()
{
    _moLock->releaseAll();
}

inline
void 
MemoryObjectCache::acquireMoLock()
{
    _moLock->acquire();
}

inline
void
MemoryObjectCache::releaseMoLock()
{
    _moLock->release();
}

#endif /* MemoryObjectCache_h */
