/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MemoryObjectView.h - 

*/

#ifndef MemoryObjectView_h
#define MemoryObjectView_h


#ifdef __GNUC__
#pragma interface
#endif

#include "MemoryObject.h"
#include "MemoryObjectCache.h"

class PhysicallyAddressableUnit;

extern ClassStar MemoryObjectViewClass;

class MemoryObjectView : public MemoryObject {

 protected:
    unsigned _viewOffset;
    MemoryObject * _source;
    unsigned _startUnit;

    virtual int rawRead( unsigned startUnit, unsigned count, 
                         char * addr, unsigned nbytes, IOAsyncClosure * c );
    virtual int rawWrite( unsigned startUnit, unsigned count, 
                          char * addr, unsigned nbytes, int writeThrough,
                          IOAsyncClosure * c );
 public:
    Class * classOf();

    /* View specific methods */

    MemoryObject * source();
    unsigned startUnit();
    unsigned viewOffset();
    unsigned size();
    unsigned offsetToCacheUnit( unsigned offset ); 
    void setSource( MemoryObject * source );
    void setStartUnit( unsigned startUnit );
    void setLog2UnitSize( unsigned log2Size );

    MemoryObjectView( MemoryObject * source,
                      unsigned numberOfUnits,
                      int log2UnitSize,
                      unsigned startUnit );

    virtual MemoryObjectCache * buildCache();
    virtual MemoryObjectCache * cache();
    virtual ProtectionLevel logicalProtectionLevel();

    /* The following should not be reached */

    virtual int read( unsigned int startUnit, int count, 
                      char * addr, int nbytes );
    virtual int write( unsigned int startUnit, int count, 
                       char * addr, int nbytes );
    virtual int writeThrough( unsigned int startUnit, int count,
                              char * addr, int nbytes );
    virtual void asyncRead( unsigned int startUnit, int count, 
                            char * addr, int nbytes, IOAsyncClosure * c );
    virtual void asyncWrite( unsigned int startUnit, int count, 
                             char * addr, int nbytes, IOAsyncClosure * c );
    virtual void asyncWriteThrough( unsigned int startUnit, int count,
                                    char * addr, int nbytes, 
                                    IOAsyncClosure * c );
    virtual void synchronize();
    virtual void flush();
    virtual int copy( MemoryObject * dest );
    virtual void setMemoryObjectCache( MemoryObjectCache * cache );
    virtual unsigned cacheLog2UnitSize();
    virtual Class * cacheClass();
    virtual Class * cachePolicyClass();
    virtual unsigned setCacheLog2UnitSize( unsigned );
    virtual void setCacheClass( Class * );
    virtual void setCachePolicyClass( Class * );
    virtual RangeLock * moLock();
    virtual int setNumberOfUnits( unsigned log2Size );
};

inline MemoryObject * 
MemoryObjectView::source()
{
    return _source;
}

inline void
MemoryObjectView::setSource( MemoryObject * source )
{
    _source = source;
}

inline void
MemoryObjectView::setLog2UnitSize( unsigned log2Size )
{
    _log2UnitSize = log2Size;
}

inline void
MemoryObjectView::setStartUnit( unsigned startUnit )
{
    _startUnit = startUnit;
    _viewOffset = startUnit << _log2UnitSize;
}

inline unsigned 
MemoryObjectView::viewOffset()
{ 
    return _viewOffset;
}

inline unsigned 
MemoryObjectView::offsetToCacheUnit( unsigned offset )
{
    Assert( _source->cache() != 0 );
    return ( _viewOffset + offset ) >> 
    ( _source->cache()->log2UnitSize() );
}

inline unsigned
MemoryObjectView::startUnit()
{
    return _startUnit;
}

inline unsigned
MemoryObjectView::size()
{
    return _numberOfUnits << _log2UnitSize;
}

/*
 * NormalMemoryObjectView
 */

extern ClassStar NormalMemoryObjectViewClass;

class NormalMemoryObjectView: public MemoryObjectView
{
 public:
    Class * classOf();

    NormalMemoryObjectView( MemoryObject * source );
    NormalMemoryObjectView( MemoryObject * source,
                            unsigned numberOfUnits,
                            int log2UnitSize,
                            unsigned startUnit );

    ~NormalMemoryObjectView();

    virtual int read( unsigned int startUnit, int count, 
                      char * addr, int nbytes );
    virtual int write( unsigned int startUnit, int count, 
                       char * addr, int nbytes );
};

inline
NormalMemoryObjectView::NormalMemoryObjectView( MemoryObject * source )
    : MemoryObjectView( source, source->numberOfUnits(), source->log2UnitSize(), 0 )
{
    _source->reference();
}

inline
NormalMemoryObjectView::NormalMemoryObjectView( MemoryObject * source,
                                                unsigned numberOfUnits,
                                                int log2UnitSize,
                                                unsigned startUnit )
    : MemoryObjectView( source, numberOfUnits, log2UnitSize, startUnit )
{
    _source->reference();
    unsigned top = _viewOffset + ( _numberOfUnits << _log2UnitSize );
    unsigned sourceTop = _source->numberOfUnits() <<
    _source->log2UnitSize();
    Assert( top <= sourceTop );
}

inline
NormalMemoryObjectView::~NormalMemoryObjectView()
{
    _source->unreference();
}

/*
 * LockedMemoryObjectView: all pages must be in core
 */

extern ClassStar LockedMemoryObjectViewClass;

class LockedMemoryObjectView : public MemoryObjectView 
{
 public:
    Class * classOf();

    LockedMemoryObjectView( MemoryObject * source,
                            unsigned int numberOfUnits,
                            int log2UnitSize,
                            int startUnit );
};

inline 
LockedMemoryObjectView::LockedMemoryObjectView( MemoryObject * source,
                                                unsigned numberOfUnits,
                                                int log2UnitSize,
                                                int startUnit)
    : MemoryObjectView( source, numberOfUnits, log2UnitSize, startUnit )
{
    Assert( numberOfUnits > 0 );
    Assert( _source->cache() != 0 );
    Assert( _log2UnitSize == _source->cache()->log2UnitSize() );
}

#endif /*MemoryObjectView_h*/



