/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Contributing Author: Vince Russo (russo@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * PhysicalMemoryChain.h - a PhysicallyAddressableUnit chain with an initial
 *			   offset into the first one.
 */

#ifndef PhysicalMemoryChain_h
#define PhysicalMemoryChain_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Assert.h"
#include "NewAndDelete.h"
#include "Object.h"

class PhysicallyAddressableUnit;
class MemoryObjectCache;
class LockHolder;

class PhysicalMemoryChain : public Object {
protected:
	unsigned _offset;
	unsigned _length;		// in bytes...
	LockHolder * _lockHolder;
	PhysicallyAddressableUnit * _chain;
public:
	PhysicalMemoryChain();
	PhysicalMemoryChain( LockHolder * lockHolder, unsigned offset = 0 );
	PhysicalMemoryChain( PhysicallyAddressableUnit * start,
			     unsigned offset, unsigned length,
			     LockHolder * lockHolder );
	/*
	 * For building a PMC that is a (logically contiguous) subrange
	 * of another.
	 */
	PhysicalMemoryChain( PhysicalMemoryChain &, unsigned offsetFromStart,
			     unsigned int length );
	~PhysicalMemoryChain();

	/*
	 * For passing chains around.
	 */
	PhysicalMemoryChain( const PhysicalMemoryChain & );
	void operator = ( const PhysicalMemoryChain & );

	PhysicallyAddressableUnit * chain();
	unsigned offset();
	unsigned lengthInBytes();
	LockHolder * lockHolder();
};

inline PhysicallyAddressableUnit *
PhysicalMemoryChain::chain()
{
	return _chain;
}

inline unsigned
PhysicalMemoryChain::offset()
{
	return _offset;
}

inline unsigned
PhysicalMemoryChain::lengthInBytes()
{
	return _length;
}

inline LockHolder *
PhysicalMemoryChain::lockHolder()
{
	return _lockHolder;
}

/*
 * Iterate over a physical memory chain in chunks of contiguous physical
 * memory. Note: the memory is only valid while the PMChain stays in existance.
 */
class PhysicalMemoryChainContiguousBlockIterator : public Object {
protected:
	PhysicalMemoryChain * myChain;
	PhysicallyAddressableUnit * currentPau;
	int bias;
	unsigned int amountLeft;
public:
	PhysicalMemoryChainContiguousBlockIterator( PhysicalMemoryChain * c,
		unsigned int totalNumberOfBytes );

	const char * next( unsigned int & length );
};

inline
PhysicalMemoryChainContiguousBlockIterator::
PhysicalMemoryChainContiguousBlockIterator( PhysicalMemoryChain * chain,
	unsigned int totalNumberOfBytes )
{
	Assert( chain != 0 );
	myChain = chain;
	currentPau = chain->chain();
	Assert( currentPau != 0 );
	bias = chain->offset();
	amountLeft = totalNumberOfBytes;
}

/*
 * A class to iterate over a chain and return sub-chains of a given size.
 */
class PhysicalMemoryChainIterator {
	PhysicalMemoryChain * initialChain;
	PhysicalMemoryChain * currentChain;
	PhysicalMemoryChain current;
public:
	PhysicalMemoryChainIterator( PhysicalMemoryChain * chain );
	PhysicalMemoryChain * first();
	PhysicalMemoryChain * next( unsigned int increment );
};

inline
PhysicalMemoryChainIterator::
PhysicalMemoryChainIterator( PhysicalMemoryChain * chain )
{
	Assert( chain != 0 );
	initialChain = chain;
	currentChain = 0;
}

inline PhysicalMemoryChain *
PhysicalMemoryChainIterator::first()
{
	LockHolder * lockHolder = initialChain->lockHolder();
	PhysicalMemoryChain newChain( initialChain->chain(), 
				      initialChain->offset(), 
				      initialChain->lengthInBytes(),
				      lockHolder );
	current = newChain;
	currentChain = &current;
	return( currentChain );
}

#endif	/* PhysicalMemoryChain_h */
