/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * PhysicallyAddressableUnit.h - a physically addressable memory unit of
 *				 allocation.
 */

#ifndef PhysicallyAddressableUnit_h
#define PhysicallyAddressableUnit_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Assert.h"
#include "Process.h"
#include "ProtectionLevel.h"

class AddressTranslation;

/*
 * Current "bits" layout:
 *
 * Bit 0: Dirty (Unit has been written).
 * Bit 1: Referenced (Unit has been referenced).
 * Bits 2 & 3: Protection Level. (dependent on ProtectionLevel.h: meaning it
 *				  depends on ProtectionLevel fitting in 2 bits.)
 * Bits 4-31: Physical Address.  (units have to be > 16 bytes in size).
 */
static const unsigned DIRTYSHIFT = 0;
static const unsigned DIRTY = (1 << DIRTYSHIFT);
static const unsigned REFSHIFT = 1;
static const unsigned REFERENCED = (1 << REFSHIFT);
static const unsigned PROTSHIFT = 2;
static const unsigned PROTMASK = 0xc;
static const unsigned ADDRMASK = 0xfffffff0;

/*
 * Make this a tuneable parameter (should be system page size....)
 */
static const unsigned PAUUnitSize = 4096;
static const unsigned Log2PAUUnitSize = 12;

class Map;

class PhysicallyAddressableUnit {
friend class Store;
protected:
	unsigned int _bits;
	PhysicallyAddressableUnit * _next;
        Map * _owner;

	void reset();

public:
	PhysicallyAddressableUnit( const char * address );
	~PhysicallyAddressableUnit();

	void setNextLink( PhysicallyAddressableUnit * unit );
	PhysicallyAddressableUnit * nextLink();

	int referenced();
	int setReferenced( int value = 1 );
	int modified();
	int setModified( int value = 1 );
	ProtectionLevel protection();
	ProtectionLevel setProtection( ProtectionLevel );
        Map * owner();
        Map * setOwner( Map * map );
 
	const char * address();
	unsigned size();
	unsigned log2Size();
};

inline
PhysicallyAddressableUnit::PhysicallyAddressableUnit( const char * paddr )
{
	Assert( ( (unsigned int) paddr & ADDRMASK ) == (unsigned int) paddr );
	_next = 0;
	_bits = (unsigned int) paddr;
	_owner = 0;
}

inline void
PhysicallyAddressableUnit::reset()
{
}

inline const char *
PhysicallyAddressableUnit::address()
{
	return( (const char *) (_bits & ADDRMASK) );
}

inline unsigned
PhysicallyAddressableUnit::size()
{
	return( PAUUnitSize );
}

inline unsigned
PhysicallyAddressableUnit::log2Size()
{
	return( Log2PAUUnitSize );
}

inline ProtectionLevel
PhysicallyAddressableUnit::protection()
{
	return( (ProtectionLevel) (( _bits >> PROTSHIFT ) & 0x3) );
}

inline ProtectionLevel
PhysicallyAddressableUnit::setProtection( ProtectionLevel newLevel )
{
	ProtectionLevel oldLevel = protection();
	_bits &= ~PROTMASK;
	_bits |= (newLevel << PROTSHIFT);
	return( oldLevel );
}

inline int
PhysicallyAddressableUnit::referenced()
{
	return( ( _bits & REFERENCED ) == REFERENCED );
}

inline int
PhysicallyAddressableUnit::setReferenced( int value )
{
	int old = referenced();
	Assert( (value == 0) || (value == 1) );
	_bits &= ~REFERENCED;
	_bits |= (value << REFSHIFT);
	return( old );
}

inline int
PhysicallyAddressableUnit::modified()
{
	return( ( _bits & DIRTY ) == DIRTY );
}

inline int
PhysicallyAddressableUnit::setModified( int value )
{
	int oldval = modified();
	Assert( (value == 0) || (value == 1) );
	_bits &= ~DIRTY;
	_bits |= (value << DIRTYSHIFT);
	return( oldval );
}

inline
PhysicallyAddressableUnit::~PhysicallyAddressableUnit()
{
}

inline void
PhysicallyAddressableUnit::setNextLink( PhysicallyAddressableUnit * u )
{
	_next = u;
}

inline PhysicallyAddressableUnit *
PhysicallyAddressableUnit::nextLink()
{
	return( _next );
}

inline Map *
PhysicallyAddressableUnit::owner()
{
	return( _owner );
}

inline Map *
PhysicallyAddressableUnit::setOwner( Map * map )
{
	Map * oldOwner = owner();
	_owner = map;
	return( oldOwner );
}

#endif	/* PhysicallyAddressableUnit_h */




