/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1993
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * ReplacementPolicy.h
 */

#ifndef ReplacementPolicy_h
#define ReplacementPolicy_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Link.h"
#include "List.h"
#include "Lock.h"
#include "PagedMemoryObjectCache.h"
#include "PageOutAsync.h"
#include "ProxiableObject.h"
#include "Semaphore.h"

class PagedMemoryObjectCache;

extern ClassStar ReplacementPolicyClass;

class ReplacementPolicy : public ProxiableObject {
      protected:
	PagedMemoryObjectCache * _cache;
	Map * _map;
	unsigned _numberOfUnits;

	friend class PageOutAsyncClosure;

        Process * acquireMo( unsigned from, unsigned to, int block = 1,
                            int locked = 0 )
		{ return _cache->acquireMo( from, to, block, locked ); }
        Process * releaseMo( unsigned from, unsigned to, int locked = 0 )
		{ return _cache->releaseMo( from, to, locked ); }
        void acquireAllMo()
		{ _cache->acquireAllMo(); }
        void releaseAllMo()
		{ _cache->releaseAllMo(); }
        void acquireMoLock()
		{ _cache->acquireMoLock(); }
        void releaseMoLock()
		{ _cache->releaseMoLock(); }

	void cacheUntranslatable(unsigned from, unsigned to ) 
		{ _cache->untranslatable( from, to ); }
	int cacheUnitsToStore( unsigned from, unsigned to, unsigned code ) 
		{ return _cache->unitsToStore( from, to, code ); }
	void cacheUnitsToStoreQueued( unsigned nunits ) 
		{ _cache->unitsToStoreQueued( nunits ); }

	unsigned cacheCacheRead( unsigned start, unsigned noOfBytes, 
				 char * buf, IOAsyncClosure * c )
		{ return _cache->cacheRead( start, noOfBytes, buf, c ); }
	unsigned cacheCacheWrite( unsigned start, unsigned noOfBytes, 
				 char * buf, IOAsyncClosure * c )
		{ return _cache->cacheWrite( start, noOfBytes, buf, c ); }
	unsigned cacheFromCache( unsigned from, unsigned to, 
		                 PhysicallyAddressableUnit * pau, 
			         IOAsyncClosure * c ) 
		{ return _cache->fromCache( from, to, pau, c ); }	
	unsigned cacheToCache( unsigned from, unsigned to, int flush, 
		               PhysicallyAddressableUnit * pau, 
			       IOAsyncClosure * c ) 
		{ return _cache->toCache( from, to, flush, pau, c ); }	
	unsigned cacheToMemoryObject( unsigned from, unsigned to, int flush, 
		                      PhysicallyAddressableUnit * pau, 
			              IOAsyncClosure * c ) 
		{ return _cache->toMemoryObject( from, to, flush, pau, c ); }	
	unsigned cacheLoad( unsigned from, unsigned to, IOAsyncClosure * c ) 
		{ return _cache->load( from, to, c ); }
	void	 cacheSaveModified( IOOperationCode operation,
				    unsigned from, unsigned to, int code,
				    IOAsyncClosure * c ) 
		{ _cache->saveModified( operation, from, to, code, c ); }

	void cachePostFault( unsigned from, unsigned to )
		{ _cache->postFault( from, to ); }
	int cacheClearReferenced( unsigned from, unsigned to )
		{ return _cache->clearReferenced( from, to ); }
	PhysicallyAddressableUnit *
		 cacheCheckReferenced( unsigned from, unsigned to, int mode )
		{ return _cache->checkReferenced( from, to, mode ); }
	
	void doPageOutAsync( unsigned from, unsigned to, 
			     PhysicallyAddressableUnit * firstPau, 
			     PostFaultType postFaultType, 
			     int unitsToStoreCode );

	virtual unsigned victim(unsigned count, 
				unsigned &start, unsigned &end,
				PhysicallyAddressableUnit * &pau) = 0;

      public:
	Class * classOf();
	ReplacementPolicy( PagedMemoryObjectCache * cache );
	virtual unsigned maxResidentUnits();
	virtual unsigned setMaxResidentUnits( unsigned units );
	virtual void setNumberOfUnits( unsigned noOfUnits );
	virtual void unitsFromStore( unsigned from, unsigned to );
	virtual void unitsToStore( unsigned from, unsigned to );
	virtual void postPageIn( unsigned from, unsigned to );
	virtual void preCacheRead( unsigned from, unsigned to );
	virtual void preCacheWrite( unsigned from, unsigned to );
	virtual void postCacheRead( unsigned from, unsigned to );
	virtual void postCacheWrite( unsigned from, unsigned to );
	virtual unsigned pageOut( unsigned noOfUnits );
	virtual void notify( int );
};

extern ClassStar BoundedReplacementPolicyClass;

class BoundedReplacementPolicy : public ReplacementPolicy {
      protected:
	Lock _lock;
	Semaphore _pageOutMutex;
	unsigned _residentUnits;
	unsigned _maxResidentUnits;

      public:
	Class * classOf();
	BoundedReplacementPolicy( PagedMemoryObjectCache * cache );
	virtual unsigned maxResidentUnits();
	virtual unsigned setMaxResidentUnits( unsigned units );
	virtual void unitsFromStore( unsigned from, unsigned to );
	virtual void unitsToStore( unsigned from, unsigned to );
};

extern ClassStar SimpleReplacementPolicyClass;

class SimpleReplacementPolicy : public ReplacementPolicy {
      protected:
	unsigned _lastVictim;
	virtual unsigned victim(unsigned count, 
				unsigned &start, unsigned &end,
				PhysicallyAddressableUnit * &pau);
      public:
	Class * classOf();
	SimpleReplacementPolicy( PagedMemoryObjectCache * cache );
};

extern ClassStar FIFOReplacementPolicyClass;

class FIFOReplacementPolicyLink : public Link {
      protected:
	unsigned _unit;
      public:
	FIFOReplacementPolicyLink( unsigned unit ) { _unit = unit; }
	unsigned unit() { return _unit; }
};

class FIFOReplacementPolicy : public BoundedReplacementPolicy {
      protected:
	Lock _fifoLock;
	List _fifo;
	virtual unsigned victim(unsigned count, 
				unsigned &start, unsigned &end,
				PhysicallyAddressableUnit * &pau);
      public:
	Class * classOf();
	FIFOReplacementPolicy( PagedMemoryObjectCache * cache );
	~FIFOReplacementPolicy();
	virtual void postPageIn( unsigned from, unsigned to );
	virtual void unitsToStore( unsigned from, unsigned to );
};

#endif /*ReplacementPolicy_h*/
