/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Store.h - Physical memory allocation.
 */

#ifndef Store_h
#define Store_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Assert.h"
#include "BlockedInfo.h"
#include "Lock.h"
#include "OutputStream.h"
#include "ProxiableObject.h"    /* parent class */
#include "PhysicallyAddressableUnit.h"
#include "Semaphore.h"
#include "IteratingList.h"

class MemoryObjectCache;
class PhysicalMemoryRange;  // forward reference
class StoreInfo;

/*
 * A collection of PhysicallyAddressableUnits.
 */

extern ClassStar StoreClass;
class Store : public ProxiableObject {
    
 protected:

    class _AwaitInfo : public BlockedInfo {
    public:
        unsigned units() { return _buf[0]; }
        void setUnits( unsigned i ) { _buf[0] = i; }
    };

    unsigned _log2UnitSize;
    unsigned _numberOfUnits;
    unsigned _kernelUnits;
    unsigned _userUnits;
    unsigned _high;
    unsigned _low;
    unsigned _wantWaiting;           // no of alloc waiting for pager
    unsigned _wantUnits;             // no of units wanted by all pending allocs
    unsigned _wantAllocUnits;
    int _paging;        // Is pager operating
    int _queued;        // queued to be freed
    static Lock _lock;      // lock for free list.
    static void acquire();
    static void release();
    
    Process * _pageOutDaemon;
    Semaphore _pagerSem;    // Semaphore used as communication
    BlockedInfoList _blocked;
    
    // StoreInfo list
    unsigned _storeInfoCount;
    StoreInfo * _storeInfoHead;
    StoreInfo * _current;

    // the head of the free list.
    PhysicallyAddressableUnit * freeList;

    void freePages();
    
 public:
    static int _initDebugLevel;
    static int _initDebugMessages;
    static int _initDebugObjects;
    Class * classOf();
    
    Store( PhysicalMemoryRange * systemMemoryMap, char * name = "Store" );
    ~Store();
    
    void startPager();
    void info( OutputStream & );
    
    int isPaging();    
    unsigned unitSize();
    unsigned numberOfUnits();
    unsigned totalAvailPaus();
    unsigned totalAllocPaus();
    unsigned totalExpectedAllocPaus();
    unsigned userAvailPaus();
    unsigned userAllocPaus();
    unsigned low();
    unsigned high();
    unsigned wantWaiting();
    unsigned wantUnits();
    int queued();
    Process * pager();
    
    /*
     * No args means allocate one unit. It can be inlined to be faster
     * since its the most common case.
     */
    PhysicallyAddressableUnit * alloc();
    PhysicallyAddressableUnit * alloc( unsigned int nunits );
    PhysicallyAddressableUnit * alloc( const char * startAddr,
                                       unsigned int nunits );
    
    void dealloc( PhysicallyAddressableUnit * unit );
    
    PhysicallyAddressableUnit * userAlloc(MemoryObjectCache * c = 0 );
    PhysicallyAddressableUnit * userAlloc(unsigned int nunits,
                                          MemoryObjectCache * c = 0);
    void userDealloc(PhysicallyAddressableUnit * unit,
                     MemoryObjectCache * c, int queued );
    void userDeallocQueued(unsigned nunits, MemoryObjectCache * c );
};

inline void
Store::acquire()
{
    _lock.acquire();
}

inline void
Store::release()
{
    _lock.release();
}

inline unsigned
Store::unitSize()
{
    return( (unsigned) 1 << _log2UnitSize );
}

inline unsigned
Store::numberOfUnits()
{
    return( _numberOfUnits );
}

inline int
Store::isPaging()
{
    return _paging;
}

inline unsigned
Store::userAvailPaus()
{
    return _numberOfUnits - _kernelUnits;
}

inline unsigned
Store::userAllocPaus()
{
    return _userUnits;
}

inline unsigned
Store::totalAllocPaus()
{
    return _userUnits + _kernelUnits;
}

inline unsigned
Store::totalExpectedAllocPaus()
{
    return totalAllocPaus() + _wantUnits - _queued;
}

inline unsigned
Store::totalAvailPaus()
{
    return _numberOfUnits - totalAllocPaus();
}

inline int
Store::queued()
{
    return _queued;
}

inline unsigned
Store::low()
{
    return _low;
}

inline unsigned
Store::high()
{
    return _high;
}

inline unsigned
Store::wantWaiting()
{
    return _wantWaiting;
}

inline unsigned
Store::wantUnits()
{
    return _wantUnits;
}

inline Process *
Store::pager()
{
    return _pageOutDaemon;
}

inline void 
Store::userDeallocQueued( unsigned nunits, MemoryObjectCache * c )
{
    _lock.acquire();
    _queued += nunits;
    c->storeInfo()->_queued += nunits;
    _lock.release();
}

extern Store * MainStore;

#endif  /* Store_h */
