/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * TwoLevelPageTable.h:  Generic two-level page tables
 */

#ifndef TwoLevelPageTable_h
#define TwoLevelPageTable_h


#ifdef __GNUC__
#pragma interface
#endif

#include "AccessType.h"
#include "AddressTranslation.h" /* parent class */
#include "MemoryBlockAllocator.h"
#include "Lock.h"
#include "PAGE_TABLE_ENTRY.h"       /* #define PAGE_TABLE_ENTRY,
                                       POINTER_TABLE_ENTRY */


class PhysicalMemoryChain;  // forward reference...
class VirtualMemoryRange;   // forward reference...

extern ClassStar TwoLevelPageTableClass;
class TwoLevelPageTable : public AddressTranslation {
    friend void PhysicalMemorySetup();

 protected:
    static VirtualMemoryRange * kernelVirtualMemoryMap;
    static MemoryBlockAllocator * tableMemoryAllocator;

    /*
     * Table allocation routines.  These simply call the
     * tableMemoryAllocator and zero the memory returned.
     */
    virtual POINTER_TABLE_ENTRY * allocatePointerTable();
    virtual PAGE_TABLE_ENTRY * allocatePageTable();

    static int pointerTableIndex( const char * address );
    static int pageTableIndex( const char * address );

    /*
     * The instance data is a lock and a pointer to a first level
     * page table which is allocated from the classes allocator.
     */
    Lock lock;  // eventually refine this to be less coarse
    // i.e. maybe keep a lock/pte in the ptes themselves.
    POINTER_TABLE_ENTRY * thisPointerTable;

    /*
     * Also keep a pointer to the
     *   parent translation for purposes of sharing memory ranges.
     */
    TwoLevelPageTable * _parent;

    /*
     * setupPointerTable() is called by pointerTable() or addMapping() -
     * whichever is the first to discover that we
     * haven't yet allocated the first level page table for this
     * translation.
     */
    void setupPointerTable();

 public:
    /*
     * Used by MMU to get value to load into page table base register.
     * This used to be private, but then the MMU class has to be a
     * friend, which led to a problem with incomplete base types.
     * This way we don't even need to include MMU.h.
     */
    POINTER_TABLE_ENTRY * pointerTable();

    /*
     * Used by AddressTranslator to see what happened when
     * a page fault occurs.
     */
    AddressTranslationFailureCode determineError( const char * address,
                                                  AccessType attemptedAccess );

    static void init( unsigned int tableAreaSize,
                      VirtualMemoryRange * kernelVirtualMemoryMap );

    Class * classOf();

    //TwoLevelPageTable();
    TwoLevelPageTable( TwoLevelPageTable * parent );
    ~TwoLevelPageTable();

    void addMapping( const char * vaddr, PhysicalMemoryChain * chain,
                     ProtectionLevel prot );
    void basicRemoveMapping( const char * vaddr, unsigned int nbytes );
    int basicChangeProtection( const char * vaddr,
                               unsigned int nbytes, ProtectionLevel protection );
    
    virtual int basicReferenced ( const char * vaddr, unsigned int nbytes,
                                  int clear);
    
    virtual int basicChangeStatus ( const char * vaddr,
                                    unsigned int nbytes, AddressTranslationStatus clear,
                                    AddressTranslationStatus set);

    const char * mapping( const char * vaddr );

    void inhibitDataCaching(const char *vaddr, unsigned int nbytes);

    // Francis TODO: get rid of this eventually
    static VirtualMemoryRange * KernelVirtualMemoryMap() { return kernelVirtualMemoryMap; }
};

#endif /* TwoLevelPageTable_h */
