/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1990,1991 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 *
 * Author: Amit Parghi (parghi@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 *	NetworkBuffer.h
 *
 *	Buffers used for data coming in from the network.
 *
 * $Id: NetworkBuffer.h 7 2004-06-03 04:38:36Z jcarlyle $
 *
 */

#ifndef NetworkBuffer_h
#define NetworkBuffer_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Types.h"
#include "ConduitTypes.h"

#include "ProcessStar.h"
#include "ProxiableObject.h"
#include "Link.h"
#include "HexDump.h"

extern ClassStar NetworkBufferFreeListClass;

class NetworkWalker;
class NetworkBufferFreeList;
class NetworkBuffer;			// forward
class Lock;

class NetworkBuffer : public Link {
protected:
    unsigned int _time; // time off net
    unsigned int _length,			// Length of current frame
		 _maxLength;			// amount of space available
    octet *_packet; // dma address
    octet *_kpacket;// kernel non-dma address
    octet *_upacket;// user space address
    
    NetworkBufferFreeList *_freeList;		// point back to its list

	// Define our own, hyper-fast "bcopy" equivalent, at least for now:
    void copy(const octet *source, octet *dest, unsigned int length);

    void constructorCommon(octet *, unsigned int, NetworkBufferFreeList *);

    unsigned short _flags;			// assume 8 bits long

public:
	    		// allocate data area; for use by Administrator code
    NetworkBuffer(unsigned int length,
		  NetworkBufferFreeList *list = NULL);

// don't allocate data; used by protocol/user code
    NetworkBuffer(octet *buffer,
		  unsigned int length,
		  NetworkBufferFreeList *list = NULL);

    ~NetworkBuffer();

    void setTime(unsigned int);
    unsigned int getTime();

    void release();	// put self back on free list

    unsigned int length();
    void setLength(unsigned int len);
    unsigned int maxLength();

    octet *packet();				// get/set storage address
    octet *kpacket();				// get/set storage address
    octet *upacket();				// get/set storage address
    void setPacket(octet *p);
    void setKPacket(octet *p);
    void setUPacket(octet *p);

    void copyFrom(const octet *, unsigned int);	// copy from user data area
    unsigned int copyTo(octet *);		// copy to user data area

    NetworkBufferFreeList *freeList();

};

inline unsigned int
NetworkBuffer::getTime()
{
    return _time  ;
}

inline void
NetworkBuffer::setTime(unsigned int t)
{
    _time = t;
}

inline unsigned int
NetworkBuffer::length()				// accessor for _length
{
    return _length;
}

inline void
NetworkBuffer::setLength(unsigned int len)
{
    _length = len;
}

inline unsigned int
NetworkBuffer::maxLength()
{
    return _maxLength;
}

inline octet *
NetworkBuffer::packet()				// accessor for _packet
{
    return _packet;
}
inline octet *
NetworkBuffer::kpacket()				// accessor for _packet
{
    return _kpacket;
}
inline octet *
NetworkBuffer::upacket()				// accessor for _packet
{
    return _upacket;
}

inline void
NetworkBuffer::setUPacket(octet *p)
{
    _upacket = p;
}
inline void
NetworkBuffer::setKPacket(octet *p)
{
    _kpacket = p;
}
inline void
NetworkBuffer::setPacket(octet *p)
{
    _packet = p;
}

inline NetworkBufferFreeList *
NetworkBuffer::freeList()			// accessor for _freeList
{
    return _freeList;
}


inline OutputStream &
operator << (OutputStream &stream, NetworkBuffer &buff)
{
    HexDump* hd = new HexDump(buff.packet(), buff.length());
    stream << *hd;
    delete hd;
    //stream << HexDump(buff.packet(), buff.length());
    return stream;
}


	//	NetworkBufferFreeList -- linked list of NetworkBuffers.
	//	Various classes/processes may have one or more of these lists.
	//
class NetworkBufferFreeList :public ProxiableObject {
protected:
    NetworkBuffer *_list;
    Lock* _mutex;			// used to lock the free list

public:
    int _count;			// # of buffers currently free
    int _totalCount;		// total # of buffers 
    unsigned int _length;			// length of each buffer

    NetworkBufferFreeList( unsigned int bufferSize = 2048, // size per buffer
			  unsigned int listSize = 10, int init = 0);   // num of buffers

    ~NetworkBufferFreeList();

    NetworkBuffer *getBuffer();	// return first free buffer
    void releaseBuffer(NetworkBuffer *);// add buffer to free list
    void listElems();
    virtual void initialize(unsigned int,unsigned int);
    Class *classOf();
};

	//	Possible flags to be set in NetworkBuffer's 'flags' member
const unsigned int NetworkBufferFlags_DeletePacket = 1;

#endif /* ! NetworkBuffer_h */
