/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * Conduit.h
 *
 */

#ifndef Conduit_h
#define Conduit_h


#ifdef __GNUC__
#pragma interface
#endif

#include "ProxiableObject.h"
#include "Semaphore.h"
#include "ZootReturnCodes.h"
#include "ConduitTypes.h"
#include "Word.h"
#include "MessageFIFO.h"
#include "ConduitProtocolIdentifiers.h"
#include "ZootAddress.h"
#include "ConduitMessage.h"
#include "IPMessage.h"


//-----------------------------------------------------------------------------

extern ClassStar ConduitClass;

class Conduit : public ProxiableObject {
	/* The Conduit is an abstraction for communication, in the same sense
	 *  that Streams is.  Conduits are what Streams-modules wish they
	 *  could be.  Each Conduit has two parts, called Top and Bottom, which
	 *  represent, respectively, the application-program or user side of
	 *  communication and the system or network-interface side of the
	 *  communication.  Conduits are connected together (typically by a
	 *  controller), at which point they exchange protocol-information to
	 *  determine whether such a connection makes sense (TCP, for example,
	 *  must be connected to something that implements IP -- connecting a
	 *  TCP-Conduit to an Ethernet-Conduit directly would be an error).
	 *  Outgoing (downward) data is sent by inserting Message/Address
	 *  pairs into a Conduit.  Data coming the other direction is
	 *  delivered by a Conduit below inserting Messages into the receiving
	 *  Conduit.  In the latter case, the message is assumed to have come
	 *  over the network or from a device and must therefore contain
	 *  destination information.  The reason for including a separate
	 *  address in outgoing messages is that the message is ``filled in''
	 *  by the Conduits it travels through on the way down -- on the way
	 *  up, the message should have been filled in by whomever sent it,
	 *  so separate routing information is not required, or possibly not
	 *  available.  The idea is that a Conduit should not worry about the
	 *  details of what the Conduits above it expects.  Each Conduit
	 *  deals with its own `flavor' of messages coming and going from its
	 *  top.
	 */
	/* Method connectTo: Tells one conduit to connect to another. */
	/* Method connectFromAbove: Called by upper conduit to request conn. */
	/* Method connectFromBelow: Called by lower conduit to complete
		the connection handshake. */
	/* Method disconnectFrom: Self-explanatory. */
	/* Method disconnectFromAbove: Called by lower conduit to complete
		disconnection-handshake. */
	/* Method insertFromAbove: send message down the tube (i.e. over the net) */
	/* Method insertFromBelow: send message from net back up the tube */
	/* Method Reset: Self-explanatory. Evil magic. */
	/* Method classOf: return pointer to object's class. */

friend class ZootController;
protected:
	ProtocolID	my_proto;
	virtual void	Reset();
	    /* Self-explanatory. May do horrible, nasty things to some kinds
	     * of conduit.
	     */
public:
	Conduit();
	~Conduit();

	virtual Return_Code	connectTo( Conduit * c );
	    /* This is what the controller calls to link 2 Conduits. Basically,
	     *  it just calls connectFromAbove() to introduce itself.
	     * N.B.: connectTo() needs to be able to undo the effects of
	     * the connectFromBelow() call the connectFromAbove() call on
	     * the other conduit will generate, in case connectFromAbove()
	     * returns some sort of error-condition. Important!
	     */
	virtual Return_Code	connectFromAbove( Conduit * c,
				    ProtocolID proto ); //=0
	    /* Decides whether to allow connection to a conduit speaking
	     *  whatever protocol c speaks (proto), calls connectFromBelow()
	     *  to complete the connection-handshake.
	     */
	virtual Return_Code	connectFromBelow( Conduit * c,
					ProtocolID proto,
					int   his_handy,
					int & my_handy );
	    /* Decides whether to allow connection of this type.
	     * NOTE: connectFromAbove() and connectFromBelow() have to be
	     *  asymetric in their duties, or else there would be an
	     *  infinite-regress of the two conduits introducing themselves
	     *  to one-another.
	     * The Protocol-ID/Handy-Number pair can be used as the means
	     *  whereby a Conduit differentiates between the Conduits attached
	     *  to its top.  For example, IP needs to be able to look at each
	     *  incoming packet and decide which Conduit it is intended for on
	     *  the basis of the packet's PROTO field.  Ditto for Ethernet.
	     *  TCP, on the other hand, uses a different scheme entirely -- the
	     *  Conduits connected to the top pass in pointers to themselves
	     *  whenever they open connections, and the TCP Conduit use the
	     *  <source-port, dest-port, src-IP-address, dest-IP-addr> tuple
	     *  to do demultiplexing.  TCP ignores protocol-numbers and handy
	     *  numbers entirely.
	     * The in-parameter his_handy and the
	     *  out-parameter my_handy are just integers which might come
	     *  in handy when conduits are connecting together -- the
	     *  semantics are protocol-specific (see class TCPConduit for an
	     *  example).
	     */
	virtual Return_Code	disconnectFrom( Conduit * c );
	    /* This is what the controller calls to unlink 2 Conduits.
	     */
	virtual Return_Code	disconnectFromAbove( Conduit * c );
	    /* Forgets it ever met such a Conduit....
	     */

	virtual Return_Code   insertFromAbove( ConduitMessage * msg,
						NetworkAddress * addr );
	    /* This is how you send something -- specifying what to send
	     * (msg) and where to send it (addr).
	     */
	virtual Return_Code   insertFromBelow( ConduitMessage * msg );
	    /* This is how returning messages are handed up to the Conduit
	     * from the conduit(s) connected below.
	     */
	virtual Class * classOf();
	    /* Returns pointer to object's class.
	     */
        //  Proxify++ needs at least one proxiable method so that ConduitClass
        //  will be defined.
        proxiable virtual void dummyConduitMethod() {}
};

class EthernetConduit : public Conduit {
	/* Method connectTo: Tells one conduit to connect to another. */
	/* Method connectFromAbove: Called by upper conduit to request conn. */
	/* Method connectFromBelow: Called by lower conduit to complete
		the connection handshake. */
	/* Method disconnectFrom: Self-explanatory. */
	/* Method disconnectFromAbove: Called by lower conduit to complete
		disconnection-handshake. */
	/* Method insertFromAbove: send message down the tube (i.e. over the net) */
	/* Method insertFromBelow: send message from net back up the tube */
	/* Method Reset: Self-explanatory. Evil magic. */
	/* Method classOf: return pointer to object's class. */

protected:
	void	Reset();
public:
	EthernetConduit();
	~EthernetConduit();

	virtual Return_Code	connectTo( Conduit * c );
	virtual Return_Code	connectFromAbove( Conduit * c,
				    ProtocolID proto );
	virtual Return_Code	connectFromBelow( Conduit * c,
					ProtocolID proto,
					int   his_handy,
					int & my_handy );
	virtual Return_Code	disconnectFrom( Conduit * c );
	virtual Return_Code	disconnectFromAbove( Conduit * c );
	virtual Return_Code   insertFromAbove( ConduitMessage * msg,
						NetworkAddress * addr );
	virtual Return_Code   insertFromBelow( ConduitMessage * msg );
};
//-----------------------------------------------------------------------------
class IPConduit : public Conduit {
protected:
	Conduit *	_ethernet;
		// A pointer to the ethernet conduit I will use to send and
		//   receive IP messages (ether-type == 0x0800)
	Conduit * *	_top;
		// A table of pointers to the conduits connected to the top
		//   of this IP conduit. Since anyone connected to the top
		//   of an IP conduit expects to have IPMessages inserted into
		//   them from below (no enema/suppository jokes, please!),
		//   we don't need to distinguish them. They all need to
		//   implement the insertFromBelow(IPMessage*) method.
		// _top will have 256 elements, and IP just indexes directly
		//   into it using the PROTO field in incoming messages.
	   	// Protocol Numbers 0, 255 are reserved;
		//   6==TCP  17==UDP  18==MUX  1==ICMP
		//   2==IGMP   34-60,80-254==unassigned
		// *** see RFC1010 for more details ***
	int _size;
		// The size of the arrays top[] and protocols[].
	int _count;
		// The actual number of conduits currently connected.
public:
	IPConduit();
	~IPConduit();

	Return_Code	connectTo( Conduit * c );
	Return_Code	connectFromAbove( Conduit * c, ProtocolID proto );
	Return_Code	connectFromBelow( Conduit * c, ProtocolID proto,
						int   his_handy,
						int & my_handy );
				// IP uses the handy numbers to specify which
				//   IP-protocol field value c wishes to
				//   receive.  PORT_IN_USE and ILLEGAL_PORT
				//   are errors that cFB returns sometimes.
	Return_Code	disconnectFrom( Conduit * c );
	Return_Code	disconnectFromAbove( Conduit * c );

	Return_Code	insertFromAbove( ConduitMessage * msg,
					NetworkAddress * addr );
	Return_Code	insertFromBelow( ConduitMessage * msg );
	void		Reset();
};
//-----------------------------------------------------------------------------
class ICMPConduit : public Conduit {
protected:
	IPConduit * _IPConduit;
public:
	ICMPConduit();
	~ICMPConduit();

	Return_Code	connectTo( Conduit * c );
	Return_Code	connectFromAbove( Conduit * c, ProtocolID proto );
	Return_Code	connectFromBelow( Conduit * c, ProtocolID proto,
						int   his_handy,
						int & my_handy );
	Return_Code	disconnectFrom( Conduit * c );
	Return_Code	disconnectFromAbove( Conduit * c );

	Return_Code	insertFromAbove( ConduitMessage * msg,
					NetworkAddress * addr );
	Return_Code	insertFromBelow( ConduitMessage * msg );
};

extern volatile SemaphoreRegulatedMessageFIFO  * ICMPIncomingMessageFIFO;
			// Global Public data -- pretty evil. See ICMPConduit.cc
//-----------------------------------------------------------------------------
class TCPConnectionDescriptor;	// Defined in StateMachine.h

class TCPConduit : public Conduit {
protected:
	int _appsize;	// size of tables of attached application conduits
	int _appcount;	// count of currently-attached applications
	IPConduit * _IPConduit;
		// Just need one pointer to the IP conduit connected to
		//   the bottom -- "There can be only one." (-Highlander).
	TCPConnectionDescriptor * _connections;
		// Array of TCP Connection Descriptors, one per connection.
		//   Each keeps track of sequence- and acknowledgement numbers,
		//   contains IP-addresses and port-#s which decribe the two
		//   sockets on both ends of the connection, remembers windows,
		//   and holds message-descriptors giving the heads of the
		//   linked lists in the output-buffer and waiting-list of the
		//   outstanding msgs. This is what needs to be inspected to
		//   decide which messages in the re-xmit-buffer need to be de-
		//   leted as the result of an incoming ACK, etc.. Each also
		//   contains a pseudoheader for checksum-computation.
		//   Of course there's a pointer to the application conduit
		//   using this connection (so luser == 0 means no connection)
	MutexSemaphore * _mutexes;
		// Array of binary semaphores to provide mutual exclusion
		//   for the CD's when fiddling with things....
	MutexSemaphore _mutex;
		// Single semaphore to provide mutual exclusion when messing
		//   the entrire array of connection-descriptors.
	TCPMessage *	prepareIncomingMessage( IPMessage * ipmsg );
		/* This is just a utility that takes an IP-Message and turns
		 * it into a TCP-Message by twiddling some header/data
		 * pointers.
		 */
public:
	TCPConduit();
	~TCPConduit();

	virtual Return_Code	connectTo( Conduit * c );
	virtual Return_Code	connectFromAbove( Conduit * c,
					ProtocolID proto );
	virtual Return_Code	connectFromBelow( Conduit * c,
					ProtocolID proto,
					int   his_handy,
					int & my_handy );
	virtual Return_Code	disconnectFrom( Conduit * c );
	virtual Return_Code	disconnectFromAbove( Conduit * c );

	virtual Return_Code	insertFromAbove( ConduitMessage * msg,
					NetworkAddress * addr );
			/* This is how you send messages. La la la!
			 */
	virtual Return_Code	insertFromBelow( ConduitMessage * msg );
			/* This is how IP tells TCP about incoming messages .
			 */

	virtual Return_Code	open( Conduit * cond, const Word16 & srcport,
					int16 initialWindow,
					TCPAddress & conn );
			/* Opens a connection and places it in the LISTEN state
			 * (see RFC 793). Normally called a "passive" open...
			 * Newport specifies the port to be used for this end
			 * of the connection. Implicitly sets the destination
			 * socket to 0 -- i.e. allows any remote machine to
			 * connect from any remote port. InitialWindow
			 * specifies the amount of space in cond's receive
			 * buffer, being a count of the number of data-bytes
			 * (i.e. sequence-numbers) the receiver is able to
			 * accept -- see expandWindow() below.
			 */
	virtual Return_Code	open( Conduit * cond, const Word16 & srcport,
					const Word16 & destport,
					int16 initialWindow,
					TCPAddress & conn );
			/* Passive-open for connection that will only accept
			 * connection requests from the specified destination
			 * TCP port (at any machine).
			 */
	virtual Return_Code	open( Conduit * cond, const Word16 & srcport,
					const Word16 & destport,
					const IPAddress & destipaddr,
					int16 initialWindow,
					TCPAddress & conn );
			/* Opens a connection and places it in the SYN-SENT
			 * state as soon as the connection-request goes out.
			 * An "Active" open-request, with the sockets both
			 * completely specified for the connection.
			 * (I.e. remote port number 0 disallowed, and
			 * local port number could be restricted.)
			 */
	virtual Return_Code	close( TCPAddress & conn );
			/* Self-explanatory.
			 */
	virtual Return_Code	expandWindow( TCPAddress & conn,
					int16 additional );
			/* Increases the size of the TCP receive-window by
			 * the specified amount.  Any Conduit connected to
			 * TCP needs to let TCP know whenever its receive
			 * buffer has more space in it -- woe betide him who
			 * lies about the available capacity of his receive
			 * buffer! (Much bandwidth wastage and howls of
			 * derision at the very least....)
			 */
	virtual Return_Code	setTimeout( int32 newtimeout );
			/* For setting ARB retransmit-interval
			 */
//	Return_Code	query(	TCPAddress & connection,
//				TCB * controlblock, TCPState * &  currentstate,
//				int * outstanding, int * timeout  );
//  			Hmmm......
	virtual Return_Code	abort( TCPAddress & connection );
			/* Abort specified connection.
			 */
	virtual void	Reset();
			/* Resets the whole damn conduit. Evil magic!
			 */
};
extern volatile SemaphoreRegulatedMessageFIFO  * TCPIncomingMessageFIFO;
			// Global Public data -- pretty evil. See IPConduit.cc
			//   (_not_ TCPConduit.cc, like you'd expect) for
			//   the gory part....

//-----------------------------------------------------------------------------
extern ClassStar ApplicationConduitClass;

class ApplicationConduit : public Conduit {
protected:
	TCPAddress mysession;	// for keeping track of which session I am
				//   currently connected-to
	TCPConduit * myTCPConduit;
public:
	ApplicationConduit();
	~ApplicationConduit();

	Return_Code	connectTo( Conduit * c );
	Return_Code	connectFromAbove( Conduit * c, ProtocolID proto );
	Return_Code	connectFromBelow( Conduit * c, ProtocolID proto,
						int   his_handy,
						int & my_handy );
	Return_Code	disconnectFrom( Conduit * c );
	Return_Code	disconnectFromAbove( Conduit * c );

	Return_Code	insertFromBelow( ConduitMessage * msg );
	TCPAddress *	session();
	void		setSession( TCPAddress & newsession );
        Class *         classOf();
};

inline TCPAddress *
ApplicationConduit::session()
{
	return( & mysession );
}


class EchoingApplicationConduit : public ApplicationConduit {
	static void eacDaemon(EchoingApplicationConduit * eac);
		// Static member-fn is just global fn whose name is
		//   restricted in scope -- so when the EAC constructor
		//   calls it (with a pointer to the EAC being constructed),
		//   things should work sensibly.
	volatile Boolean	EAnotReady;
		// Used to signal the EAC when it has been connected to a
		//   TCP conduit (which ought not to happen until after
		//   that TCP-conduit has been fully initialized) and the
		//   FIFO is all ready.
protected:
	SemaphoreRegulatedMessageFIFO	_queue;
public:
	EchoingApplicationConduit();
	EchoingApplicationConduit(ProcessContainer * scheduler);
	~EchoingApplicationConduit();

	Return_Code			insertFromBelow( ConduitMessage * msg );
	SemaphoreRegulatedMessageFIFO *	queue();
	TCPConduit * 			tcpConduit();
};


inline SemaphoreRegulatedMessageFIFO *
EchoingApplicationConduit::queue()
{
	return( & _queue );
}

inline TCPConduit *
EchoingApplicationConduit::tcpConduit()
{
	return( myTCPConduit );
}

#endif
