/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * ConduitCharacterStream.h
 *
 */

#ifndef ConduitCharacterStream_h
#define ConduitCharacterStream_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Semaphore.h"

#include "ConduitTypes.h"
#include "Word.h"
#include "ZootReturnCodes.h"
#include "ZootAddress.h"
#include "ConduitMessage.h"
#include "ConduitProtocolIdentifiers.h"
#include "Conduit.h"


class ConduitCharacterStream {
	/* A CCS is basically a ring-buffer that is smart enough to interact
	 * with users either reading or writing chunks of characters on the
	 * one hand, and conduits sending and receiving messages on the
	 * other. The base class just worries about the fact that there is
	 * some storage -- Input- and Output-Streams define the methods for
	 * twiddling those octets.  Subclasses actually implement read() and
	 * write().
	 */
	/* Method empty: indicates whether there is anything in the buffer. */
protected:
	octet *	_buffer;	// pointer to actual array of octets
public:
	ConduitCharacterStream();
	virtual	~ConduitCharacterStream();
	virtual Boolean	empty();
			/* Returns TRUE if there's nothing in the buffer,
			 * FALSE otherwise.
			 */
};

class ConduitOutputStream : public ConduitCharacterStream {
	/* This is the character stream the user of the conduit writes
	 * to when sending data out over the net. Subclasses could do all
	 * sorts of hip stuff like knowing about the input-stream to do
	 * echoing and so forth.
	 */
protected:
public:
	ConduitOutputStream();
		/* The constructor might fail if there is not enough room
		 * in the Kernel for the buffer, for example.
		 */
	virtual	~ConduitOutputStream();

	virtual int	write( octet * buffer, int len );
			/* Asks that len octets be taken from the specified
			 * user buffer and put into the stream. If the COS
			 * decides it's time, it will do an insertTop() to
			 * send the data out on the conduit it's buffering.
			 * Returns the actual number of octets written (which
			 * may be less than len if the buffer is full and the
			 * implementation of write() is non-blocking).
			 */
	virtual void	flush();
			/* Tells the buffer to send a message out on the
			 * conduit regardless of how it normally decides
			 * when to send messages.
			 */
};

class ConduitInputStream : public ConduitCharacterStream {
protected:
public:
	ConduitInputStream();
	~ConduitInputStream();

	virtual int	read( octet * buffer, int maxlen);
			/* Asks that at most maxlen octets be copied from
			 * the stream into the specified user buffer. Returns
			 * the actual number of octets transferred.
			 */
};


class TCPOutputStream : public ConduitOutputStream {
	friend class TCPConnectionDescriptor;
protected:
	TCPConnectionDescriptor * cd;
public:
	TCPOutputStream();
	~TCPOutputStream();

	virtual int	write( octet * buffer, int len );
	virtual void	flush();
};

class TCPInputStream : public ConduitInputStream {
	friend class TCPConnectionDescriptor;
protected:
	TCPConnectionDescriptor * cd;
	virtual octet *	get( int offset );
			/* Returns a pointer to the octet that is offset
			 * from the beginning of the buffer (i.e. _buffer +
			 * offset unless offset is out of bounds).
			 */
public:
	TCPInputStream();
	~TCPInputStream();

	virtual int	read( octet * buffer, int maxlen);
	virtual octet *	urgent();
			/* Returns pointer to urgent data sitting in the
			 * buffer, 0 if there is none.
			 */
};

#endif /* ConduitCharacterStream_h */
