/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * ConduitMessage.h
 *
 */

#ifndef ConduitMessage_h
#define ConduitMessage_h


#ifdef __GNUC__
#pragma interface
#endif

#include "ConduitTypes.h" 
#include "Word.h"
#include "ConduitProtocolIdentifiers.h"
#include "ZootAddress.h"

class ConduitMessage {
	/* A ConduitMessage is conceptually a string of bytes. In order to
	 * provide a felxible mechanism for creating new messages and for
	 * embedding messages in other messages (as, for example, a TCP
	 * message is carried "inside" an IP message), the bytes are taken
	 * as being located in an array pointer to by the header-pointer,
	 * followed by bytes pointed to by the data-pointer, followed by
	 * bytes pointed to by the payload-pointer which is another
	 * ConduitMessage.  Further, "manual" memory-management  is
	 * facilitated by the ability to specify which pointers point to
	 * data that must be deleted when the message is deleted.  This is
	 * necessary since, for instance, the data pointer might point to
	 * the middle of a single array of bytes allocated beginning at
	 * the header -- it would be incorrect to delete the memory pointed
	 * to by data in this case!
	 */
	/* Method length: return total length of message. */
	/* Method protocol: return protocol field. */
	/* Method control: return control field. */
	/* Method header: retrun pointer to header octets. */
	/* Method data: return pointer to data octets. */
	/* Method headerLength: return length of header. */
	/* Method dataLength: return length of data. */
	/* Method link: return pointer to payload. */
	/* Method holder: return pointer to message whose payload I am. */
	/* Method status: return status-octet (which determines which of
		header/data/payload must be deleted when the message is
		deleted). */
	/* Method setProtocol: set protocol field. */
	/* Method setControl: set control field. */
	/* Method setHeader: set header pointer and length. */
	/* Method setData: set data pointer and length. */
	/* Method setLink: set pointer to payload. */
	/* Method pleaseDeleteHeader: set status bit to indicate that header
		should be deleted when message is deleted. */
	/* Method pleaseDeleteData: self-explanatory. */
	/* Method pleaseDeletePayload: self-explanatory. */
	/* Method pleaseDontDeleteHeader: self-explanatory. */
	/* Method pleaseDontDeleteData: self-explanatory. */
	/* Method pleaseDontDeletePayload: self-explanatory. */
	/* Method setStatus: set all status bits at once. */
	/* Method del: delete this message and all payload/carriers. */

protected:
	ProtocolID	_proto;
	int	_control;	/* Control information. */
	octet * _header;		// pointer to this msg's header
	int	_headerLength;
	octet * _data;		// pointer to this msg's data
	int	_dataLength;
	ConduitMessage * _link;	// pointer to message being carried.
			// Don't want to store its length, since it may have
			//  some screwy way of thinking about its own length
			//  and should be asked.
	ConduitMessage * _holder;	// pointer to message carrying me.
	int	_status;
			// Status is necessary to keep track of which pieces of
			//  memory ought to be deleted and which shouldn't.
			//  For example, if data points to the middle of the
			//  same array as header (as you might expect with an
			//  incoming message) doing a delete on it would be the
			//  Wrong Thing. On the other hand, if data space were
			//  nabbed dynamically (setData() needed more space,
			//  say), failing to do a delete on it would be bad.
			//      BIT:  Meaning:
			//       0     if 1, header needs to be deleted
			//       1     if 1, data needs to be deleted
			//       2     if 1, *_link needs to be deleted
public:
	ConduitMessage();
		/* default constructor allocates nothing.
		 */
	ConduitMessage( int hsz, int dsz );
		/* allocate hsz octets for head, dsz for data
		 */
	ConduitMessage( int hsz, int dsz, ConduitMessage * pld );
		/* same as above, set link to point to pld
		 */
	virtual	~ConduitMessage();
		/* Self-explanatory.
		 */
	// -------------------- Getters: --------------------
	// (Some are virtual, since subclasses may want to tweak/check things
	//  before letting any old fool look at/twiddle the fields.) 
	virtual int length();
		/* Returns length, measured in octets.
		 * It's virtual, since may be calculated strangely by some
		 * types of messages. Notice there is no setter for lying
		 * about a message's length.
		 */
	inline ProtocolID protocol();
		/* Self-explanatory.
		 */
	inline int control();
		/* Self-explanatory.
		 */
	inline octet * header();
		/* Pointer to header octets (if any -- otherwise returns 0).
		 */
	inline octet * data();
		/* Pointer to data octets (if any -- otherwise returns 0).
		 */
	inline int headerLength();
		/* Self-explanatory.
		 */
	inline int dataLength();
		/* Self-explanatory.
		 */
	inline ConduitMessage * link();
		/* Pointer to another message that is ``inside'' this message,
		 * being carried as payload. (Actually, it's best to think of
		 * it as a message which carries the octets following the
		 * header and data of this message.)
		 */
	inline ConduitMessage * holder();
		/* Pointer to another message that ``holds'' this message,
		 * as its payload.  It should have this message as its link.
		 */
	inline int status();
		/* status-bits keep track of which pointers (H/D/P) point to
		 * stuff that actually needs to be deleted when the message is
		 * deleted.
		 */
	// -------------------- Setters: --------------------
	inline void setProtocol(ProtocolID newproto);
		/* Self-explanatory.
		 */
	inline void setControl(int newcontrol);
		/* Self-explanatory.
		 */
	inline void setHeader( int newhlen, octet * newheader );
		/* Self-explanatory.
		 */
	inline void setData( int newdlen, octet * newdata );
		/* Self-explanatory.
		 */
	inline void setLink( ConduitMessage * newlink );
		/* Self-explanatory.
		 */
	inline void pleaseDeleteHeader();
		/* Sets a status bit
		 */
	inline void pleaseDeleteData();
		/* Sets a status bit
		 */
	inline void pleaseDeletePayload();
		/* Sets a status bit
		 */
	inline void pleaseDontDeleteHeader();
		/* Sets a status bit
		 */
	inline void pleaseDontDeleteData();
		/* Sets a status bit
		 */
	inline void pleaseDontDeletePayload();
		/* Sets a status bit
		 */
	inline void setStatus( int newstatus );
		/* Sets all status bits at one fell swoop.
		 */
	virtual void del();
	virtual ConduitMessage * clone();
		/* Creates a new instance of whatever class this
		 * message is, linked to nothing, with copies of the
		 * header and data this message has.
		 */
};


class NetworkBuffer; // see Administrator/NetworkBuffer.h

class EthernetMessage : public ConduitMessage {
protected:
	NetworkBuffer *	_buff;
		// Incoming ethernet messages arriving from the administrator
		//   live inside NetworkBuffers. If this pointer is nonzero,
		//   the actual octets of this EthernetMessage live in this
		//   buffer (which must be freed up when the EM is deleted)
public:
	EthernetMessage( int header_length,
			 int data_length,
			 ConduitMessage * link );
		// This constructor is for creating a new EM (which will
		//   always have a 14-octet ethernet header) that carries
		//   the message that link points to.
	EthernetMessage( NetworkBuffer * buff );
		// This is for incoming messages from the Administrator. It
		//   automagically sets up _header and _data to point into
		//   the buffer, and diddles the status bits appropriately.
	~EthernetMessage();

	ConduitMessage * clone();
	// All other methods are simply inherited from ConduitMessage.
};

inline
void
PhotoCopy( ConduitMessage * source, ConduitMessage * dest )
{
	// This function is used by clon(), and assumes that source and
	//   dest have the same header and data lengths, and that
	//   dest has not got a payload linked to it.
	if( 0 != source->headerLength() )
	    OctetCopy( source->header(), dest->header(),
		       source->headerLength() );
	if( 0 != source->dataLength() )
	    OctetCopy( source->data(), dest->data(), source->dataLength() );
	dest->setStatus( source->status() ); // Evil!
	dest->pleaseDontDeletePayload();
		// Just to be sure, since the dest hasn't got a payload
}


#endif
