/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * ConduitMessageBuffer.h
 *
 */

#ifndef ConduitMessageBuffer_h
#define ConduitMessageBuffer_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Assert.h"
#include "OutputStream.h"
#include "ProxiableObject.h"
#include "Semaphore.h"
#include "SemaphoreStar.h"

#include "ZootReturnCodes.h"
#include "ConduitTypes.h"
#include "Word.h"
#include "ZootClock.h"
#include "ZootAddress.h"
#include "ConduitMessage.h"
#include "Conduit.h"
#include "ZootStateMachine.h"

// As far as I can figure out, the only direction in which a buffer is
//   needed for message-flow is downward. TCP wants to throw IP messages into
//   a retransmit buffer on their way toward the IP conduit, but when messages
//   are received, they should get jammed right into their destination
//   conduits. A buffer inside a conduit (for example, the reassembly buffer
//   inside of IP for storing stuff until missing fragments show up) is a 
//   different sort of thing. Actually, subclasses could probably change the
//   semantics so that send() would do an insertFromBelow instead of an insertFromAbove
//   if desired....
// Anyway, the base-class Buffer is designed to sit between two conduits, and
//   the default insert(), send() and flush() operations do insertFromAbove()'s 
//   -- subclasses (such as the IP fragment reassembly buffer) may need to add 
//   ways to look at/scribble on the messages in the buffer, or change the
//   operations done on the conduit, etc....

class Conduit;

extern ClassStar ConduitMessageBufferClass;

class ConduitMessageBuffer : public ProxiableObject {
	/* Conceptually, a ConduitMessageBuffer is an object that sits inside
	 * one Conduit, buffering messages that will ultimately be sent to
	 * some other Conduit.
	 *
	 * For example, TCP needs to buffer outgoing
	 * messages until their receipt has been acknowledged by the TCP on
	 * the other end of the conversation.
	 *
	 * Some kinds of Conduit-Message Buffer can be connected to a
	 * Conduit for convenience so that messages can be inserted into
	 * the appropriate end by invoking the send() method. Depending on the
	 * particular subclass, this will do an insertFromAbove() or insertFromBelow()
	 * as desired. Typically, as buffer that knows how to connect to a
	 * Conduit will automatically send each message once as it is inserted.
	 */
    /* Method grab: locks out anyone from using the buffer. */
    /* Method release: unlocks the buffer. */
    /* Method insert: put a message into the buffer, sending it if the
		buffer knows how. Returns an ID. */
    /* Method reset: delete any messages in the buffer. */
    /* Method flush: send any messages in the buffer. */
    /* Method count: returns count of how many messages are in the buffer. */
    /* Method send: send specified message to connected Conduit. */
    /* Method connect: stash a pointer to a Conduit so buffer can send
		messages to it. */
    /* Method disconnect: Opposite of connect. */
    /* Method previous: Returns the ID of the message inserted immediately
		before the specified message. */
    /* Method del: Delete specified message. */
    /* Method lookAt: Retrieve pointer to specified message for monkeying with
		it while it's sitting in the buffer. */
    /* Method classOf: returns pointer to object's class. */
protected:
	SemaphoreStar	_mutex;
			// Semaphore to provide mutual exclusion when twiddling
			//   msgs[], addrs[], links, _count, etc.
	int	_size;	// maximum number of messages/addresses
	int	_count;	// number of messages/addresses in buffer
	int	_next;	// offset in array of next message to be inserted
	ConduitMessage	* * msgs; // Array of pointers to messages
	NetworkAddress	* * addrs; // Array of pointers to addresses
public:
	ConduitMessageBuffer();
			/* Default constructor makes a buffer with a
			 * capacity of 100 messages.
			 */
	ConduitMessageBuffer( int sz );
			/* sz specifies how many pointers to allocate.
			 */
	virtual ~ConduitMessageBuffer();
			/* Destructor also deletes any messages still around.
			 */

	void		grab();
			  /* Does a P() on the ARB's mutex-semaphore, so that
			   * the buffer's owner can monkey around and delete
			   * things and whatnot without anyone else (i.e. the
			   * daemon) interfering or getting confused.
			   */
	void		release();
			  /* Does a V() on the ARB's mutex-semaphore, undoing
			   * a previous grab().
			   */
	int			count();
				  /* Returns message-count so you can tell
				   * how many messages are lying around in
				   * the buffer.
				   */
	virtual void		reset();
				  /* Deletes all the messages, etc.
				   */
	virtual MessageID	insert( Return_Code & ret,
				    ConduitMessage * newmsg,
				    NetworkAddress * newaddr );
				  /* Copies pointer to the <message,address>
				   * pair into the buffer someplace, returning
				   * identifier that points to which slot it is
				   * sitting in. Will set ret to BUFFER_FULL if
				   * there is no more space. For a buffer that
				   * knows how to send() stuff, inserting a
				   * new message probably sends it as well.
				   *
				   * In the base class, the assumption is made
				   * that all the messages will be stored in a
				   * single block in the array.  Subclasses
				   * might want to allow spaces between them,
				   * probably requiring a linked list to be
				   * able to use the array of pointers
				   * efficiently.
				   */
	MessageID		previous( MessageID handle );
				  /* Returns ID of last message inserted before
				   * the specified message was (-1 indicates
				   * no such message). In the base class, this
				   * is essentially (handle - 1) -- subclasses
				   * might not assume all messages are stored
				   * sequentially in the array.
				   *
				   * The important thing is to be able to cdr
				   * down the list starting with the most
				   * recently inserted message and look at all
				   * the messages in the buffer in reverse
				   * chronological order by calling previous()
				   * until it returns -1.
				   *
				   * Previous(h) is undefined garbage if h is
				   * not the handle for a valid message.
				   */
	ConduitMessage * 	lookAt( MessageID handle );
				  /* Returns a pointer to msg specified by
				   * handle so you can look at it and tweak it
				   * if necessary.
				   */
	virtual Return_Code	del( MessageID handle );
				  /* Handle specifies which message to delete. 
				   */
	virtual Return_Code	connect( Conduit * target );
				  /* You connect a buffer to a Conduit so it
				   * can send stuff away for you automatically,
				   * or by calling send(), depending on what
				   * kind of buffer it is. It is the
				   * responsibility of the whoever's using the
				   * buffer to ensure that the right kind of
				   * Conduit is being specified -- this does
				   * not connect the buffer to the target in
				   * the same sense as two Conduits get
				   * connected -- just stores a pointer so the
				   * buffer can send off messages on the
				   * Conduit's behalf.
				   *
				   * Not all buffers know how
				   * to send stuff, so connect/send/flush are
				   * sometimes not allowed.
				   */
	virtual Return_Code	disconnect();
				  /* sets _dest = 0. Opposite of connect().
				   */
	virtual Return_Code	send( MessageID handle );
				  /* Non-automatic buffers need someone to tell
				   *  them when to push stuff down the conduit
				   *  and which stuff to push.
				   * Return_Code might be NO_SUCH_MESSAGE;
				   *  otherwise, returns what the Conduit
				   *  returns when you insertFromAbove the 
				   *  handle'th <msg,addr>.
				   */
	virtual void		flush();
				  /* Sends everything, regardless of time or
				   * whatever.
				   */
	virtual Class *		classOf();
};


extern ClassStar TCPWaitBufferClass;

class TCPWaitBuffer : public ConduitMessageBuffer {
	/* This is where a TCP-conduit stores all the messages that have been
	 *  inserted but weren't ready to go out on the wire (mostly ones
	 *  sent before a connection actually becomes established or when
	 *  the window was too small for them to be sent). It's very similar
	 *  to an AutoRetransmitBuffer, except there is no notion of sending
	 *  stuff/flushing, and it holds TCP-messages, not IP-messages. Also
	 *  note that del() just frees up a pointer -- it does {\em not}
	 *  actually delete the message.
	 */
    /* Method del: Resets pointer to message to 0 but does not delete it. */
    /* Method classOf: returns pointer to object's class. */
protected:
public:
	TCPWaitBuffer();
		/* Default size = 100 messages.
		 */
	TCPWaitBuffer(int sz);
		/* Sz specifies how many pointers to allocate.
		 */
	~TCPWaitBuffer();
		/* Destructor actually does delete any remaining messages.
		 */

	Return_Code		del( MessageID handle );
				/* Just frees up entry in table -- does _not_
				 *   actually delete the message!
				 */
	virtual Class *		classOf();
};

extern ClassStar RetransmitBufferClass;

class RetransmitBuffer : public ConduitMessageBuffer {
	/* A RetransmitBuffer is a ConduitMessageBuffer that knows how to
	 * connect to a Conduit, and which inserts messages into its top
	 * when they are inserted, and whenever send() or flush() make it
	 * necessary.
	 *
	 * It has no process associated with it -- its sole purpose in life is
	 * to store messages that occasionally get sent out ``by hand.''  To
	 * get automatic periodic retransmission, an AutoRetransmitBuffer is
	 * the right Thing.
	 */
    /* Method insert: put a message into the buffer, sending it if the
		buffer knows how. */
    /* Method reset: delete any messages in the buffer. */
    /* Method connect: stash a pointer to a Conduit so buffer can send
		messages to it. */
    /* Method send: send specified message to connected Conduit. */
    /* Method flush: send any messages in the buffer. */
    /* Method disconnect: Opposite of connect. */
    /* Method classOf: returns pointer to object's class. */
protected:
	Conduit *	_dest;
			// Conduit where msgs are headed
public:
	RetransmitBuffer();
			/* Default constructor makes a buffer with a
			 * capacity of 100 messages.
			 */
	RetransmitBuffer( int sz );
			/* sz specifies how many pointers to allocate.
			 */
	RetransmitBuffer( int sz, Conduit * initialDestination );
			/* Size as well as what it is connected to.
			 */
	~RetransmitBuffer();
			/* Destructor also deletes any messages still around.
			 */

	MessageID	insert( Return_Code & ret, ConduitMessage * newmsg,
						 NetworkAddress * newaddr );
			  /* See ConduitMessageBuffer. Inserting a message
			   * also sends it -- it is an error to insert a
			   * message if no Conduit is currently connect()ed.
			   */
	void		reset();
			  /* Zaps everything.
			   */
	Return_Code	connect( Conduit * target );
			  /* Sets destination to target -- see
			   * ConduitMessageBuffer.
			   */
	Return_Code	send( MessageID handle );
			  /* InsertTop the handle'th message in the connected
			   * Conduit.  It is an error to send a message if
			   * no Conduit is currently connected.
			   */
	void		flush();
			  /* Send everything right now.
			   */
	Return_Code	disconnect();
			  /* Set _dest = 0. Opposite of connect.
			   */
	virtual Class *	classOf();
};

//-----------------------------------------------------------------------------
//				inliners
//-----------------------------------------------------------------------------
inline
ConduitMessageBuffer::ConduitMessageBuffer()
{
	_mutex = new MutexSemaphore;
	_size = 100;
	_count = 0;
	_next = 0;
	msgs = new ConduitMessage *[100];
	addrs = new NetworkAddress *[100];
	for( int i = 0; i < _size; i++ ){
	   msgs[i] = 0;
	   addrs[i] = 0;
	}
}
inline
ConduitMessageBuffer::ConduitMessageBuffer( int sz )
{
	_mutex = new MutexSemaphore;
	_size = sz;
	_count = 0;
	_next = 0;
	msgs = new ConduitMessage *[ sz ];
	addrs = new NetworkAddress *[ sz ];
	for( int i = 0; i < _size; i++ ){
	   msgs[i] = 0;
	   addrs[i] = 0;
	}
}
inline int
ConduitMessageBuffer::count()
{
	return( _count );
}
inline MessageID
ConduitMessageBuffer::previous( MessageID handle )
{
// ASSUMES HANDLE IS VALID!
	if( handle == -1 ) return( -1 );
		// Sigh. Some things are easier to code if you ask for
		//   previous( m ) before checking if m is -1.
	// Trying to mutually exclude the operations deadlocks with
	//   trying to grab()/release() the whole buffer. Hmm...
	//_mutex->P();
	// Are there any messages in the buffer?
	if( _count == 0 ) {
	    //_mutex->V();
	    return( -1 );
	}
	// Is handle 0?
	if( handle == 0 ) {
	    if( msgs[ _size - 1 ] != 0 ) {
	        //_mutex->V();
		return( _size - 1 );
	    }
	    else {
	        //_mutex->V();
		return( -1 );
	    }
	}
	if( msgs[ handle - 1 ] != 0 ) {
	    //_mutex->V();
	    return( handle - 1 );
	}
	//_mutex->V();
	return( -1 );
}
inline ConduitMessage *
ConduitMessageBuffer::lookAt( MessageID handle )
{
	if( handle == -1 ) return( 0 );
	Assert( ( 0 <= handle ) && ( handle < _size ) );
	return( msgs[ handle ] );
}
inline void
ConduitMessageBuffer::grab()
{
	_mutex->P();
}

inline void
ConduitMessageBuffer::release()
{
	_mutex->V();
}
//----------------------------------------------------------------------------

extern ClassStar AutoRetransmitBufferClass;
class AutoRetransmitBuffer : public RetransmitBuffer {
	/* An AutoRetransmitBuffer has a daemon that looks at it every
	 * second or so to check what needs to be retransmitted. It adds
	 * time-stamping to the functionality of a RetransmitBuffer as
	 * well. Retransmit delay defaults to 30 seconds.
	 */
    /* Method check: Called by daemon to check if anything in the buffer
	is ready to be retransmitted. */
    /* Method insert: put a message into the buffer, sending it since this
	kind of buffer knows how. */
    /* Method setDelay: set retransmit-interval, which is measured in
	4 microsecond ticks as recommended in RFC 793. */
    /* Method kick: forces immediate retransmission of buffer contents. */
    /* Method classOf: return pointer to object's class. */

protected:
	int32		_delay;
			  /* Number of 4uS ticks to wait between re-sends
			   */
	int32 *		_lastsent;
			  /* Array of timestamps stating when each message
			   * was last sent. Measured in milliseconds.
			   */
	void		check();
			  /* Looks at the time since each thing was sent and
			   * re-sends anything if it is has been longer than
			   * _delay (measured in 4uS ticks) since it was
			   * last sent...
			   */
	static AutoRetransmitBuffer *	_ARBList;
				// class-variable (see below)
	AutoRetransmitBuffer *	_nextARB;
	static void	AutoRetransmitBufferDaemon();
			  /* Is a PreemptableSystemProcess that calls check()
			   * over and over, waiting about a second between
			   * calls. Runs down linked list of ARBs rooted in
			   * the class variable (static) _ARBlist.
			   */
public:
	AutoRetransmitBuffer();
			  /* Default constructor creates ARB of size 100
			   * messages, and adds the newly-created ARB to the
			   * global linked-list so the Daemon can check it.
			   */
	AutoRetransmitBuffer( int sz );
			  /* Specify a size.
			   */
	AutoRetransmitBuffer( int sz, int32 dlay );
			  /* Specify a size and a delay
			   * (measured in 4uS ticks).
			   */
	AutoRetransmitBuffer( int sz, int32 dlay,
				Conduit * initialDestination );
			  /* Specify all that as well as what to connect to.
			   */
	~AutoRetransmitBuffer();
			  /* Destroy any messages, unlink myself from global
			   * linked-list.
			   */

	virtual MessageID	insert( Return_Code & ret,
				    ConduitMessage * newmsg,
				    NetworkAddress * newaddr );
			  /* Need a version of insert() that sets the
			   * timestamp.
			   */
	void		setDelay( int32 newdelay );
			  /* Set the retransmit-interval to newdelay ticks.
			   */
	virtual void		kick();
			  /* retransmits everything right now (could be
			   * by TCP if it sees 3 identical ACKs to rexmit
			   * "by hand".
			   */
	virtual Class *	classOf();
			  /* Returns object's class.
			   */
};


//-----------------------------------------------------------------------------
inline void
AutoRetransmitBuffer::setDelay( int32 newdelay )
{
	_delay = newdelay;
}


#endif ConduitMessageBuffer_h
