/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funde\d by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * ConduitMessage.h
 *
 */

#ifndef IPMessage_h
#define IPMessage_h


#ifdef __GNUC__
#pragma interface
#endif

#include "ConduitMessage.h"

class IPMessage : public ConduitMessage {
protected:
public:
	IPMessage();
		// No data, no payload, header points to copy of default header
	IPMessage( int hsz, int dsz );
		// Header is copy of default header padded to hsz octets
		// data is dsz octets (all zero)
		// no payload
	//*** Note: to build an "empty" IPMessage -- i.e. no header or data at
	//	    all -- use this constructor with (0,0) for arguments; this
	//	    is the usual (idiomatic) way to create an IPMessage to be
	//	    used for an incoming message which already has storage.
	IPMessage( int hsz, int dsz, ConduitMessage * pld );
		// Same as above -- payload is specified
	// -------------------- Getters: --------------------
	nibble VERS();		// returns VERS field of IP header
	nibble HLEN();		// returns length of IP header
	octet TOS();		// returns Type Of Service field of IP header
	nibble Precedence();		// returns value of 3 Precedence bits
					// (0-7) in TOS field (bits 0-2)
	Boolean LowDelay();		// returns value of D bit in TOS field
					//  (bit 3)
	Boolean HighThroughput();	// returns value of T bit in TOS field
					//  (bit 4)
	Boolean HighReliability();	// returns value of R bit in TOS field
					//  (bit 5)
	Word16 TotalLength();	// returns Total Length field of IP header
	Word16 IDENT();		// returns IDENT field of IP header
	nibble FLAGS();		// returns FLAGS (only 2 relevant bits)
	Boolean DontFragment();	// returns status of second flag bit
	Boolean MoreFragments();// returns status of low order flag bit
	Word13 FragmentOffset();
				// returns Fragment Offset field of IP header
				// *** Remember *** FO is measured in 64-bit
				// chunks. FO lives in bits 0-12.
	octet TTL();		// returns Time To Live field of IP header
	octet PROTO();		// returns IP protocol identifier
				//     0, 255 are reserved;
				//     TCP==6  UDP==17  MUX==18  1==ICMP
				//     2==IGMP 34-60,80-254==unassigned
				//     *** see RFC1010 for more details ***
	Word16 HChecksum();	// returns IP header checksum
	Word32 SourceADDR();	// returns Source IP Address field of IP header
	Word32 DestADDR();	// returns Destination IP Address field
	octet * Options();	// returns pointer to Options segment of IP
			// header. For now, I'll be messy and manipulate
			// options by hand as an octet-stream. Fix later.
			// Notice: returns 0 if there are no options.
	// -------------------- Setters: --------------------
	void setHeader( int newhlen, octet * newheader );
	void setVERS( nibble newVERS );
	void setHLEN( nibble newHLEN );
	void setTOS( octet newTOS );	// for setting TOS field all at once
	void setPrecedence( nibble );	// sets value of 3 Precedence bits
					// (0-7) in TOS field
	void setLowDelay( Boolean );	// sets value of D bit in TOS field
	void setHighThroughput( Boolean );// sets value of T bit in TOS field
	void setHighReliability( Boolean );// sets value of R bit in TOS field
	void setTotalLength( Word16 & newTotalLength );
	void setIDENT( const Word16 & newIDENT );
	void setFLAGS( nibble newFLAGS ); // only contains 2 bits of useful data
			// in bits 1 and 2 (bit 0 must be 0; bit 4 ignored)
	void setDontFragment( Boolean );	// set bit 1 of Flags field
	void setMoreFragments( Boolean );	// set bit 2 of Flags field
	void setFragmentOffset( Word13 & newFragmentOffset );
	void setTTL( octet newTTL );
	octet decrementTTL();	// frequently want to decrement time to live;
			// it seems dumb to fetch and then reset -- this will
			// just decrement, and return the new value.
	void setPROTO( octet newPROTO );
	void setHChecksum( const Word16 & newHChecksum );
	void setSourceADDR( const Word32 & newSourceADDR );
	void setSourceADDR( const IPAddress & newSourceADDR );
	void setDestADDR( const Word32 & newDestADDR );
	void setDestADDR( const IPAddress & newDestADDR );
	void print();
	ConduitMessage * clone();
};

//-----------------------------------------------------------------------------
extern const octet DefaultIPHeader[];
extern const octet DefaultTCPHeader[];
//-----------------------------------------------------------------------------
class TCPMessage : public ConduitMessage {
protected:
	IPMessage * myIPmsg;	// pointer to IP message carrying this TCP msg
public:
	TCPMessage();
		// No data, no payload, header points to copy of default header
		// No IP message to carry this message in yet.
	TCPMessage( int hsz, int dsz );
		// Header is copy of default header padded to hsz octets
		// data is dsz octets (all zero)
		// no payload
		// No IP message to carry this message in yet.
	TCPMessage( int hsz, int dsz, ConduitMessage * pld );
		// Same as above -- payload is specified
	TCPMessage( int iphsz, int ipdsz, int hsz, int dsz, ConduitMessage * pld );
		// Everything specified (can be 0, of course) -- IP message is
		//   initialized to have a header of length iphsz, a data
		//   portion of length ipdsz, the TCP header is length hsz, the
		//   data portion is length dsz, and a (possibly 0) payload is
		//   specified.
	// -------------------- Getters: --------------------
	IPMessage * IPmsg();	// Pointer to IP message carrying this message
	// These 2 only work if myIPmsg != 0:
	IPAddress SourceAddress();
	IPAddress DestinationAddress();
	Word16 SourcePort();		// TCP port number of source
	Word16 DestinationPort();	// TCP port number of destination
	Word32 SequenceNumber();
	Word32 AckNumber();	// position of highest byte received by TCP
				//  on the other end.
	nibble Offset();	// offset from beginning of TCP header of user
				//  data (since Options variable length).
				//  (That is, header-length measured in 32-bit
				//   words....)
	octet CODE();		// Only 6 bits of useful stuff:
	Boolean URG();			// bit 5 of CODE field
	Boolean ACK();			// bit 4 of CODE field
	Boolean PSH();			// bit 3 of CODE field
	Boolean RST();			// bit 2 of CODE field
	Boolean SYN();			// bit 1 of CODE field
	Boolean FIN();			// bit 0 of CODE field
	Word16 Window();	// size of receive buffer on other (TCP) end
	Word16 Checksum();	// checksum of header and data as per RFC 793
	Word16 UrgentPointer();	// offset of urgent data in the byte stream
	octet * Options();	// same thing as for IP messages
			// Notice: returns 0 if there are no options.
	// -------------------- Setters: --------------------
	void setHeader( int newhlen, octet * newheader );
	void setIPmsg( IPMessage * newIPmsg );	// set myIPmsg pointer
	// These 4 only work if myIPmsg != 0:
	void setSourceAddress( const Word32 &newSourceAddr );
		// affects IP stuff
	void setSourceAddress( const IPAddress &newSourceAddr );
		// affects IP stuff
	void setDestinationAddress( const Word32 &newDestAddr );
		// affects IP stuff
	void setDestinationAddress( const IPAddress &newDestAddr );
		// affects IP stuff
	void setSourcePort( const Word16 & newSourcePort );
	void setDestinationPort( const Word16 & newDestPort );
	void setSequenceNumber( const Word32 &newSequenceNumber );
	void incrementSequenceNumber();
	void setAckNumber( const Word32 &newAckNumber );
	void setOffset( nibble newOffset );
	void setCODE( octet newCODE );	// Only 6 bits of useful stuff
					// (bits 6,7 ignored)
	void setURG( Boolean );			// bit 5 of CODE field
	void setACK( Boolean );			// bit 4 of CODE field
	void setPSH( Boolean );			// bit 3 of CODE field
	void setRST( Boolean );			// bit 2 of CODE field
	void setSYN( Boolean );			// bit 1 of CODE field
	void setFIN( Boolean );			// bit 0 of CODE field
	void setWindow( const Word16 & newWindow );
	void setChecksum( const Word16 & newChecksum );
	void setUrgentPointer( const Word16 & newUrgentPointer );
	// --------------- For debugging:
	void print();
	ConduitMessage * clone();
};

class TCPPseudoHeader {
protected:
	octet pseudoheader[12];
		// Conceptually, the pseudoheader is:
		//  +-------+-------+-------+-------+
		//  |        Source Address         | (4 octets)
		//  +-------+-------+-------+-------+
		//  |        Dest.  Address         | (4 octets)
		//  +-------+-------+-------+-------+
		//  |  0x00 | Proto |   TCPLength   | (4 octets)
		//  +-------+-------+-------+-------+
		// where Proto is one octet (0x06 for TCP over IP), and
		// TCPLength is the length of the TCP header + TCP data,
		// measured in octets (16 bits).
	Word16 chksum;
		// Since the only purpose of a pseudoheader is to calculate
		//   TCP segment checksums, a TCPPseudoHeader will always
		//   recalculate its checksum when it is changed.
	inline Word16 combineIPChecksums( Word16 &, Word16 & );
	inline int16 combineIPChecksums( int16, int16 );
public:
	TCPPseudoHeader();
	TCPPseudoHeader( IPAddress & src, IPAddress & dest, Word16 & len );
	TCPPseudoHeader( IPAddress & src, IPAddress & dest, int16 len );
	//----------Getters:----------
	IPAddress	sourceAddress();
	IPAddress	destinationAddress();
	Word16		length();
	Word16		checksum();
	//----------Setters:----------
	void		setSourceAddress( const IPAddress & newsrcaddr );
	void		setDestinationAddress( const IPAddress & newdestaddr );
	void		setLength( const Word16 & newlen );
	void		setLength( const int16 newlen );
			// Checksum updated automatically; no need to set
	void	doChecksum( TCPMessage * msg );
};
class UDPMessage : public ConduitMessage {
protected:
public:
	UDPMessage(){_proto = UDP;}
	UDPMessage( int hsz, int dsz ){_proto = UDP;}
	UDPMessage( int hsz, int dsz, ConduitMessage * pld ){_proto = UDP;}
	UDPMessage( int iphsz, int ipdsz, int hsz, int dsz, ConduitMessage * pld );
	Word16 SourcePort();		// UDP port number of source
	Word16 DestinationPort();	// UDP port number of destination
	Word16 Length();		// contents of UDP header LENGTH field
	Word16 Checksum();	// checksum of header and data as per RFC 768
	void setSourcePort( const Word16 & newSourcePort );
	void setDestinationPort( const Word16 & newDestPort );
	void setLength( const Word16 & newLength );
	void setChecksum( const Word16 & newChecksum );
	ConduitMessage * clone();
};

class ICMPMessage : public ConduitMessage {
protected:
	IPMessage * icmpIPmsg;	// pointer to IP message carrying this ICMP message
public:
	ICMPMessage();
		// No data, no payload, header points to copy of default header
		// No IP message to carry this message in yet.
	ICMPMessage( int hsz, int dsz );
		// Header is copy of default header padded to hsz octets
		// data is dsz octets (all zero)
		// no payload
		// No IP message to carry this message in yet.
	ICMPMessage( int hsz, int dsz, ConduitMessage * pld );
		// Same as above -- payload is specified
	ICMPMessage( int iphsz, int ipdsz, int hsz, int dsz, ConduitMessage * pld );
		// Everything specified (can be 0, of course) -- IP message is
		//   initialized to have a header of length iphsz, a data
		//   portion of length ipdsz, the TCP header is length hsz, the
		//   data portion is length dsz, and a (possibly 0) payload is
		//   specified.

	// -------------------- Getters: --------------------

	IPMessage * IPmsg();	// Pointer to IP message carrying this message
	octet MessageType(); // returns the ICMP message type
											// 0 : Echo Reply, 3: Destination Unreachable, 8: Echo Request etc.
											// See RFC 792.
	octet Code();				// Further qualifies the ICMP message
	Word16 HChecksum();	// returns ICMP header checksum


	// -------------------- Setters: --------------------
	void setIPmsg( IPMessage * newIPmsg );	// set myIPmsg pointer
	void setHeader( int newhlen, octet * newheader );
	void setMessageType(octet newMessageType);
	void setCode(octet newCode);
	void setHChecksum(const Word16 &newHChecksum);
};


//-----------------------------------------------------------------------------
inline
ConduitMessage::ConduitMessage()
{
	_proto = Undefined;
	_control = 0;
	_header = 0;
	_headerLength = 0;
	_data = 0;
	_dataLength = 0;
	_link = 0;
	_holder = 0;
	_status = 0;
}
inline ProtocolID
ConduitMessage::protocol()
{
	return( _proto );
}
inline int
ConduitMessage::control()
{
	return( _control );
}
inline octet *
ConduitMessage::header()
{
	return( _header );
}
inline octet *
ConduitMessage::data()
{
	return( _data );
}
inline int
ConduitMessage::headerLength()
{
	return( _headerLength );
}
inline int
ConduitMessage::dataLength()
{
	return( _dataLength );
}
inline ConduitMessage *
ConduitMessage::link()
{
	return( _link );
}
inline ConduitMessage *
ConduitMessage::holder()
{
	return( _holder );
}
inline int
ConduitMessage::status()
{
	return( _status );
}
inline void
ConduitMessage::setProtocol( ProtocolID newproto )
{
	_proto = newproto;
}
inline void
ConduitMessage::setControl( int newcontrol )
{
	_control = newcontrol;
}
inline void
ConduitMessage::setHeader( int newhlen, octet * newheader )
{
	_header = newheader;
	_headerLength = newhlen;
	_status |= 0x01;	// Sets bit 0 so destructor will delete
			//  header.
}
inline void
ConduitMessage::setData( int newdlen, octet * newdata )
{
	_data = newdata;
	_dataLength = newdlen;
	_status |= 0x02;	// Sets bit 1 so destructor will delete
			//  data.
}
inline void
ConduitMessage::setLink( ConduitMessage * newlink )
{
	_link = newlink;
	_status |= 0x04;	// Sets bit 2 so destructor will delete
			//  payload. The payload can set it to the
			//  Right Thing as necessary.
	newlink->_holder = this;
}
inline void
ConduitMessage::pleaseDeleteHeader()
{
	_status |= 0x01;	// Sets bit 0 so destructor will delete header
}
inline void
ConduitMessage::pleaseDeleteData()
{
	_status |= 0x02;	// Sets bit 1 so destructor will delete data
}
inline void
ConduitMessage::pleaseDeletePayload()
{
	_status |= 0x04;	// Sets bit 2 so destructor will delete payload
}
inline void
ConduitMessage::pleaseDontDeleteHeader()
{
	_status &= 0xfffe;	// Unset bit 0 so destructor won't delete header
}
inline void
ConduitMessage::pleaseDontDeleteData()
{
	_status &= 0xfffd;	// Unset bit 1 so destructor won't delete data
}
inline void
ConduitMessage::pleaseDontDeletePayload()
{
	_status &= 0xfffb;	// Unset bit 2 so destructor won't delete payload
}
inline void
ConduitMessage::setStatus( int newstatus )
{	// EVIL MAGIC! Should not need to use!
	// Keep in mind that subclasses may use the status bits 3, 4....
	_status = newstatus;
}

//------------------- class IPMessage inliners: -------------------------------
// The getters all do nasty bit-extraction type stuff. Very grungy.
//-----------------------------------------------------------------------------
inline nibble
IPMessage::VERS()
{
	return( (nibble)( (_header[0] & 0xf0 ) >> 4 ) );
}
inline nibble
IPMessage::HLEN()
{
	return( (nibble)( _header[0] & 0x0f ) );
}
inline octet
IPMessage::TOS()
{
	return( _header[1] );
}
inline nibble
IPMessage::Precedence()
{
	return( (nibble)( _header[1] & 0xe0 ) ); // prec is hi-order 3 bits!
}
inline Boolean
IPMessage::LowDelay()
{
	if( _header[1] & 0x10 ) {	// Extract bit 3 from TOS
		return( TRUE );
	}
	else {
		return( FALSE );
	}
}
inline Boolean
IPMessage::HighThroughput()
{
	if( _header[1] & 0x08 ) {	// Extract bit 4 from TOS
		return( TRUE );
	}
	else {
		return( FALSE );
	}
}
inline Boolean
IPMessage::HighReliability()
{
	if( _header[1] & 0x04 ) {	// Extract bit 5 from TOS
		return( TRUE );
	}
	else {
		return( FALSE );
	}
}
inline Word16
IPMessage::TotalLength()
{
	Word16	tmp;
	tmp.equals( _header + 2 ); // giving it an octet* should be right
	return( tmp );
}
inline Word16
IPMessage::IDENT()
{
	Word16	tmp;
	tmp.equals( _header + 4 ); // giving it an octet* should be right
	return( tmp );
}
inline nibble
IPMessage::FLAGS()
{
	return( (nibble)( ( _header[6] & 0xe0 ) >> 5 ) );
		// Set up so that the More Fragments bit will be the low
		//  order bit of the nibble. Seems sensible, though it costs
		//  a couple fo cycles.
}
inline Boolean
IPMessage::DontFragment()
{
	if( _header[6] & 0x40 ) {	// Bit 1 of FLAGS is DF
		return( TRUE );
	}
	else {
		return( FALSE );
	}
}
inline Boolean
IPMessage::MoreFragments()
{
	if( _header[6] & 0x20 ) {	// Bit 2 is MF
		return( TRUE );
	}
	else {
		return( FALSE );
	}
}
inline Word13
IPMessage::FragmentOffset()
{
	Word13 tmp;
	short unsigned int tmp1 = (short unsigned int)_header + 6;
	short unsigned int tmp2 = tmp.val() & 0x1fff;

	tmp.equals( tmp1 );	// Copy FLAGS+FragmentOffset into tmp
	tmp.equals( tmp2 );	//   and mask off FLAGS
	return( tmp );
}
inline octet
IPMessage::TTL()
{
	return( _header[8] );
}
inline octet
IPMessage::PROTO()
{
	return( _header[9] );
				// Assigned Protocol Numbers:
	   			//     0, 255 are reserved;
				//     TCP==6  UDP==17  MUX==18  1==ICMP
				//     2==IGMP 34-60,80-254==unassigned
				//     *** see RFC1010 for more details ***
}
inline Word16
IPMessage::HChecksum()
{
	Word16 tmp;
	tmp.equals( _header + 10 );
	return( tmp );
}
inline Word32
IPMessage::SourceADDR()
{
	Word32 tmp;
	tmp.equals( _header + 12 );
	return( tmp );
}
inline Word32
IPMessage::DestADDR()
{
	Word32 tmp;
	tmp.equals( _header + 16 );
	return( tmp );
}
inline octet *
IPMessage::Options()
{
	if( _header[0] & 0x0f == 0x05 ) { // LEN = 5 so no options
	   return( 0 );
	}
	else {
	   return( _header + 20 );
	}
}

// -------------------- Setters: --------------------
inline void
IPMessage::setHeader( int newhlen, octet * newheader )
{
	ConduitMessage::setHeader( newhlen, newheader );
	setHLEN( (nibble) newhlen / 4 );
}
inline void
IPMessage::setVERS( nibble newVERS )
{
	_header[0] &= 0x0f;
	_header[0] |= ( newVERS << 4 );	// relies on fact that nibble is really
					//   a char. I've never heard of a
					//   machine that really has a 4-bit
					//   type, so nibbles should always
					//   really be chars.
}
inline void
IPMessage::setHLEN( nibble newHLEN )
{
	_header[0] &= 0xf0;
	_header[0] |= newHLEN;
}
inline void
IPMessage::setTOS( octet newTOS )	// for setting TOS field all at once
{
	_header[1] = newTOS;
}
inline void
IPMessage::setPrecedence( nibble newPrec ) // sets value of 3 Precedence bits
{
	_header[1] &= 0x1f;
	_header[1] |= ( (octet)( newPrec & 0x07 ) << 5 );
}
inline void
IPMessage::setLowDelay( Boolean newD )	// sets value of D bit in TOS field
{
	if( newD ) {
	   _header[1] |= 0x10;
	}
	else {
	   _header[1] &= 0xef;
	}
}
inline void
IPMessage::setHighThroughput( Boolean newT )
					// sets value of T bit in TOS field
{
	if( newT ) {
	   _header[1] |= 0x08;
	}
	else {
	   _header[1] &= 0xf7;
	}
}
inline void
IPMessage::setHighReliability( Boolean newR )
					// sets value of R bit in TOS field
{
	if( newR ) {
	   _header[1] |= 0x04;
	}
	else {
	   _header[1] &= 0xfb;
	}
}
inline void
IPMessage::setTotalLength( Word16 & newTotalLength )
{
	newTotalLength.copyTo( _header + 2 );
}
inline void
IPMessage::setIDENT( const Word16 & newIDENT )
{
	newIDENT.copyTo( _header + 4 );
}
inline void
IPMessage::setFLAGS( nibble newFLAGS )
{
	_header[6] &= 0x1f;
	_header[6] |= ( ( (octet)newFLAGS & 0x07 ) << 5 );
}
inline void
IPMessage::setDontFragment( Boolean newDF )
{
	if( newDF ) {
	   _header[6] |= 0x40;
	}
	else {
	   _header[6] &= 0xbf;
	}
}
inline void
IPMessage::setMoreFragments( Boolean newMF )
{
	if( newMF ) {
	   _header[6] |= 0x20;
	}
	else {
	   _header[6] &= 0xdf;
	}
}
inline void
IPMessage::setFragmentOffset( Word13 & newFragmentOffset )
{
	octet tmp[2];
	newFragmentOffset.copyTo( tmp );
	_header[6] &= 0xe0;
	_header[6] |= tmp[0];
	_header[7] = tmp[1];
}
inline void
IPMessage::setTTL( octet newTTL )
{
	_header[8] = newTTL;
}
inline octet
IPMessage::decrementTTL()
{
	_header[8]--;
	return( _header[8] );
}
inline void
IPMessage::setPROTO( octet newPROTO )
{
	_header[9] = newPROTO;
}
inline void
IPMessage::setHChecksum( const Word16 & newHChecksum )
{
	newHChecksum.copyTo( _header + 10 );
}
inline void
IPMessage::setSourceADDR( const Word32 & newSourceADDR )
{
	newSourceADDR.copyTo( _header + 12 );
}
inline void
IPMessage::setSourceADDR( const IPAddress & newSourceADDR )
{
	newSourceADDR.copyTo( _header + 12 );
}
inline void
IPMessage::setDestADDR( const Word32 & newDestADDR )
{
	newDestADDR.copyTo( _header + 16 );
}
inline void
IPMessage::setDestADDR( const IPAddress & newDestADDR )
{
	newDestADDR.copyTo( _header + 16 );
}

// -------------------- Getters: --------------------
inline IPMessage *
TCPMessage::IPmsg()
{
	return ( myIPmsg );
}
#ifndef PROXIFY	/* this syntax is too convoluted for Proxify++ */
inline IPAddress
TCPMessage::SourceAddress()
{
       if( myIPmsg == 0 ){
	     return( IPAddress(NULL) );
        }

	return( IPAddress( myIPmsg->SourceADDR() ) );
}
inline IPAddress
TCPMessage::DestinationAddress()
{
        if( myIPmsg == 0 ){
	  return( IPAddress(NULL) );
	}

        return( IPAddress( myIPmsg->DestADDR() ) );
}
#endif
inline Word16
TCPMessage::SourcePort()
{
	Word16 tmp;
	tmp.equals( _header + 0 );
	return( tmp );
}
inline Word16
TCPMessage::DestinationPort()
{
	Word16 tmp;
	tmp.equals( _header + 2 );
	return( tmp );
}
inline Word32
TCPMessage::SequenceNumber()
{
	Word32 tmp;
	tmp.equals( _header + 4 );
	return( tmp );
}
inline Word32
TCPMessage::AckNumber()
{
	Word32 tmp;
	tmp.equals( _header + 8 );
	return( tmp );
}
inline nibble
TCPMessage::Offset()
{
	return( (nibble)( _header[12] >> 4 ) );
}
inline octet
TCPMessage::CODE()
{
	return( _header[13] );
}
inline Boolean
TCPMessage::URG()
{	// URG is bit 2
	if( _header[13] & 0x20 ) {
	   return( TRUE );
	}
	else {
	   return( FALSE );
	}
}
inline Boolean
TCPMessage::ACK()
{
	if( _header[13] & 0x10 ) {
	   return( TRUE );
	}
	else {
	   return( FALSE );
	}
}
inline Boolean
TCPMessage::PSH()
{
	if( _header[13] & 0x08 ) {
	   return( TRUE );
	}
	else {
	   return( FALSE );
	}
}
inline Boolean
TCPMessage::RST()
{
	if( _header[13] & 0x04 ) {
	   return( TRUE );
	}
	else {
	   return( FALSE );
	}
}
inline Boolean
TCPMessage::SYN()
{
	if( _header[13] & 0x02 ) {
	   return( TRUE );
	}
	else {
	   return( FALSE );
	}
}
inline Boolean
TCPMessage::FIN()
{
	if( _header[13] & 0x01 ) {
	   return( TRUE );
	}
	else {
	   return( FALSE );
	}
}
inline Word16
TCPMessage::Window()
{
	Word16 tmp;
	tmp.equals( _header + 14 );
	return( tmp );
}
inline Word16
TCPMessage::Checksum()
{
	Word16 tmp;
	tmp.equals( _header + 16 );
	return( tmp );
}
inline Word16
TCPMessage::UrgentPointer()
{
	Word16 tmp;
	tmp.equals( _header + 18 );
	return( tmp );
}
inline octet *
TCPMessage::Options()
{
	if( _header[12] & 0xf0 == 0x50 ) { // data offset of 5 means no options
	   return( 0 );
	}
	else {
	   return( _header + 20 );
	}
}
// -------------------- Setters: --------------------
inline void
TCPMessage::setHeader( int newhlen, octet * newheader )
{
	ConduitMessage::setHeader( newhlen, newheader );
	setOffset( (nibble)newhlen / 4 );
}
inline void
TCPMessage::setIPmsg( IPMessage * newIPmsg )
{
	myIPmsg = newIPmsg;
	myIPmsg->setPROTO( (octet) 6 );
}
inline void
TCPMessage::setSourceAddress( const IPAddress & newSourceAddr )
{
	if( myIPmsg != 0 ) myIPmsg->setSourceADDR( newSourceAddr.val() );
}
inline void
TCPMessage::setSourceAddress( const Word32 & newSourceAddr )
{
	if( myIPmsg != 0 ) myIPmsg->setSourceADDR( newSourceAddr );
}
inline void
TCPMessage::setDestinationAddress( const IPAddress & newDestAddr )
{
	if( myIPmsg != 0 ) myIPmsg->setDestADDR( newDestAddr.val() );
}
inline void
TCPMessage::setDestinationAddress( const Word32 & newDestAddr )
{
	if( myIPmsg != 0 ) myIPmsg->setDestADDR( newDestAddr );
}
inline void
TCPMessage::setSourcePort( const Word16 & newSourcePort )
{
	newSourcePort.copyTo( _header + 0 );
}
inline void
TCPMessage::setDestinationPort( const Word16 & newDestPort )
{
	newDestPort.copyTo( _header + 2 );
}
inline void
TCPMessage::setSequenceNumber( const Word32 &newSequenceNumber )
{
	newSequenceNumber.copyTo( _header + 4 );
}
inline void
TCPMessage::incrementSequenceNumber()		// Useless....
{
	Word32 tmp;	// Let tmp worry about byte-order for us
	tmp.equals( _header + 4 );
	tmp.equals( tmp.val() + 1 );
	tmp.copyTo( _header + 4 );
}
inline void
TCPMessage::setAckNumber( const Word32 & newAckNumber )
{
	newAckNumber.copyTo( _header + 8 );
}
inline void
TCPMessage::setOffset( nibble newOffset )
{
	_header[12] &= 0x0f;
	_header[12] |= ( newOffset << 4 );
}
inline void
TCPMessage::setCODE( octet newCODE )
{
	_header[13] = newCODE;
}
inline void
TCPMessage::setURG( Boolean newURG )
{
	if( newURG ) {
	   _header[13] |= 0x20;
	}
	else {
	   _header[13] &= 0xdf;
	}
}
inline void
TCPMessage::setACK( Boolean newURG )
{
	if( newURG ) {
	   _header[13] |= 0x10;
	}
	else {
	   _header[13] &= 0xef;
	}
}
inline void
TCPMessage::setPSH( Boolean newPSH )
{
	if( newPSH ) {
	   _header[13] |= 0x08;
	}
	else {
	   _header[13] &= 0xf7;
	}
}
inline void
TCPMessage::setRST( Boolean newRST )
{
	if( newRST ) {
	   _header[13] |= 0x04;
	}
	else {
	   _header[13] &= 0xfb;
	}
}
inline void
TCPMessage::setSYN( Boolean newSYN )
{
	if( newSYN ) {
	   _header[13] |= 0x02;
	}
	else {
	   _header[13] &= 0xfd;
	}
}
inline void
TCPMessage::setFIN( Boolean newFIN )
{
	if( newFIN ) {
	   _header[13] |= 0x01;
	}
	else {
	   _header[13] &= 0xfe;
	}
}
inline void
TCPMessage::setWindow( const Word16 & newWindow )
{
	newWindow.copyTo( _header + 14 );
}
inline void
TCPMessage::setChecksum( const Word16 & newChecksum )
{
	newChecksum.copyTo( _header + 16 );
}
inline void
TCPMessage::setUrgentPointer( const Word16 & newUrgentPointer )
{
	newUrgentPointer.copyTo( _header + 18 );
}

inline Word16
TCPPseudoHeader::combineIPChecksums( Word16 & chksum1, Word16 & chksum2 )
{
	int32 tmp;
	tmp = chksum1.val() + chksum2.val();
	while( (tmp & 0xffff0000)  != 0x00000000 ) {
	    // Need while-loop since may overflow upon adding in carry-bits
	    tmp = ( (tmp >> 16) & 0x0000ffff ) + ( tmp & 0x0000ffff );
	}
	return( Word16( tmp ) );
}

inline int16
TCPPseudoHeader::combineIPChecksums( int16 chksum1, int16 chksum2 )
{
	int32 tmp;
	int16 tmp2;
	tmp = chksum1 + chksum2;
	while( (tmp & 0xffff0000)  != 0x00000000 ) {
	    // Need while-loop since may overflow upon adding in carry-bits
	    tmp = ( (tmp >> 16) & 0x0000ffff ) + ( tmp & 0x0000ffff );
	}
	tmp2 = tmp;
	return( tmp2 );
}
//--------------------------- UDPMessage inliners -----------------------
inline Word16
UDPMessage::SourcePort()
{
	Word16 tmp;
	tmp.equals( _header + 0 );
	return( tmp );
}
inline Word16
UDPMessage::DestinationPort()
{
	Word16 tmp;
	tmp.equals( _header + 2 );
	return( tmp );
}
inline Word16
UDPMessage::Length()
{
	Word16 tmp;
	tmp.equals( _header + 4 );
	return( tmp );
}
inline Word16
UDPMessage::Checksum()
{
	Word16 tmp;
	tmp.equals( _header + 6 );
	return( tmp );
}
inline void
UDPMessage::setSourcePort( const Word16 & newSourcePort )
{
	newSourcePort.copyTo( _header + 0 );
}
inline void
UDPMessage::setDestinationPort( const Word16 & newDestPort )
{
	newDestPort.copyTo( _header + 2 );
}
inline void
UDPMessage::setLength( const Word16 & newLength )
{
	newLength.copyTo( _header + 4 );
}
inline void
UDPMessage::setChecksum( const Word16 & newChecksum )
{
	newChecksum.copyTo( _header + 6 );
}

//------------------- class ICMPMessage inliners: -------------------------------
inline IPMessage *
ICMPMessage::IPmsg()
{
	return ( icmpIPmsg );
}
inline octet
ICMPMessage::MessageType()
{
	return(_header[0]);
}
inline octet
ICMPMessage::Code()
{
	return(_header[1]);
}
inline Word16
ICMPMessage::HChecksum()
{
	Word16	tmp;
	tmp.equals( _header + 2 ); // giving it an octet* should be right
	return( tmp );
}

// -------------------- Setters: --------------------
inline void
ICMPMessage::setHeader( int newhlen, octet * newheader )
{
	ConduitMessage::setHeader( newhlen, newheader );
}
inline void
ICMPMessage::setIPmsg( IPMessage * newIPmsg )
{
	icmpIPmsg = newIPmsg;
	icmpIPmsg->setPROTO( (octet) 1 );
}
inline void
ICMPMessage::setMessageType( octet newMessageType)
{
	_header[0] = newMessageType;
}
inline void
ICMPMessage::setCode( octet newCode)
{
	_header[1] = newCode;
}
inline void
ICMPMessage::setHChecksum( const Word16 & newHChecksum )
{
	newHChecksum.copyTo( _header + 2 );
}


/**************** Broken since G++ is fucked up and won't pass structs! *******
class MessageID {
protected:
	int	Offset;	// usually, a msg-id is used to store an offset into
			//   an array (say, of connection-descriptors)
public:
	inline MessageID();
	inline MessageID( int newoffset );
	inline ~MessageID();
	inline int	offset();
	inline void	setOffset( int newoffset );
};
inline
MessageID::MessageID() { Offset = 0; }
inline
MessageID::MessageID( int newoffset ) { Offset = newoffset; }
inline
MessageID::~MessageID() { }
inline int
MessageID::offset() { return( Offset ); }
inline void
MessageID::setOffset( int newoffset ) { Offset = newoffset; }

// The idea is to be able to have arrays of message identifiers that just
//   provide quick access to messages in buffers, etc. For now, msg-ids will
//   just be the offset in the array holding relevant information; but later
//   it might be desirable to have msg-ids be pointers or memory-page numbers
//   or something. They should never be incremented or messed with in any way!
******************************************/
typedef	int	MessageID;

#endif /* IPMessage_h */
