/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 *
 *
 * FIle: Word.h
 *
 */

#ifndef Word_h
#define Word_h


#ifdef __GNUC__
#pragma interface
#endif

#include "ConduitTypes.h"
#include "MachineByteOrder.h"
#include "Assert.h"

/*****************************************************************************
 *  Now that we have numbers with known sizes, we have to worry about byte
 *  order.  We'll build up some types that do all the worrying about byte
 *  order for us by defining:
 *     WORDS that are smart enough to always be able to tell you their value in
 *       either network- or host byte order. Their value will always be in host
 *       byte order so arithmetic on values will work. You can assign to a Word
 *       from an octet* and the conversion is done automagically. Notice that
 *       I provide a function to do the copy, instead of providing a function
 *       that spits out a nbo-value since the copying may involve masking
 *       bits (for example, a Word13 has to mask off the 3 extra bits when
 *       it gets assigned into a datagram), or other nonsense.
 *****************************************************************************/

	// If the local machine's byte order is the same as network
	//   byte order, the constant HOFFSET should be set to 0
	//   since we can use the same copy of tha value for both
	//   orderings. Otherwise, HOFFSET=1 since the value array
	//   will need to store both a nbo and a hbo version.

//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
class Word8;
class Word16;
class Word32;

class	Word : public ByteSwapper {
	/* Words are used to provide byte-order transparent
	 * manipulation of data. The method val() returns the value stored
	 * in the Word in host byte order, and must therefore be defined
	 * in each subclass individually, since it returns a different
	 * type (number of bits) for each type of Word.  See Word.h for
	 * a lengthier discussion.
	 */
	/* Method equals: copy value (assumed to be in network byte order)
		from given pointer. */
	/* Method copyTo: copy value into memory pointed to by given
		pointer. */
protected:
	// Instance variables are subclass responsibility
	//   byte order is the same as net byte order, HOFFSET=0
	//   since we can use the same copy of the value for both
	//   orderings. Otherwise, HOFFSET=1 since the value array
	//   will need to store both a nbo and a hbo version.
public:
	Word();
	Word( const octet source ); // Facilitates foo = 3 if foo is a Word
	Word( const int16 source );
	Word( const int32 source );
	~Word();
	// The val() member function needs to return different types
	//   depending on the subclass, so it can't be defined here...
	void	copyFrom( const octet *source );
	void	equals( const octet *source );
			// use to grab value right out of a datagram
	void	equals( octet source );
	void	equals( const int16 source );		// int's always in hbo
	void	equals( const int32 source );
	void	equals( const Word8 & source );
	void	equals( const Word16 & source );
	void	equals( const Word32 & source );
	void	copyTo( octet *dest ) const;// copy value directly into datagram
					// (in nbo, obviously)
};


class	Word8 : public Word {
protected:
	octet	myval;	// Storage for value (no byte order worries!)
public:
	Word8();
	Word8( const octet source );
	Word8( const int16 source );
	Word8( const int32 source );
	~Word8();
	octet	val() const;
	void	copyFrom( const octet *source);
	void	equals( const octet *source);	// use to grab value right
						// out of a datagram
	void	equals( const octet source );
	void	equals( const int16 source );	// int's always in hbo
	void	equals( const int32 source );
	void	equals( const Word8 & source );
	void	equals( const Word16 & source );
	void	equals( const Word32 & source );
	void	copyTo( octet *dest ) const;// copy value directly into datagram
};


class	Word13 : public Word {
protected:
	int13	myvals[HOFFSET+1];	// Storage for (1 or 2 copies of) value
public:
	Word13();
	Word13( const octet source );
	Word13( const int16 source );
	Word13( const int32 source );
	~Word13();
	int13	val() const;
	void	copyFrom( const octet *source);
	void	equals( const octet *source);	// use to grab value right
						// out of a datagram
	void	equals( const octet source );
	void	equals( const int16 source );	// int's always in hbo
	void	equals( const int32 source );
	void	equals( const Word8 & source );
	void	equals( const Word16 & source );
	void	equals( const Word32 & source );
	void	copyTo( octet *dest) const;// copy value directly into datagram
};


class	Word16 : public Word {
protected:
	int16	myvals[HOFFSET+1];	// Storage for (1 or 2 copies of) value
public:
	Word16();
	Word16( const octet source );
	Word16( const int16 source );
	Word16( const int32 source );
	~Word16();
	int16	val() const;
	void	copyFrom( const octet *source);
	void	equals( const octet *source);
	void	equals( const octet source );
	void	equals( const int16 source );	// int's always in hbo
	void	equals( const int32 source );
	void	equals( const Word8 & source );
	void	equals( const Word16 & source );
	void	equals( const Word32 & source );
	void	copyTo( octet *dest) const;// copy value directly into datagram
};

class	Word32 : public Word {
protected:
	int32	myvals[HOFFSET+1];	// Storage for (1 or 2 copies of) value
public:
	Word32();
	Word32( const octet source );
	Word32( const int16 source );
	Word32( const int32 source );
	~Word32();
	int32	val() const;
	void	copyFrom( const octet *source);
	void	equals( const octet *source);
	void	equals( const octet source );
	void	equals( const int16 source );	// int's always in hbo
	void	equals( const int32 source );
	void	equals( const Word8 & source );
	void	equals( const Word16 & source );
	void	equals( const Word32 & source );
	void	copyTo( octet *dest ) const;// copy value directly into datagram
};

#ifndef PROXIFY

//-----------------------------------------------------------------------------
// These two helper functions copy 16- and 32-bit values from one place
//   to another regardless of their alignment.  I am using them to replace
//   code of the form:
//	int16 * p1;
//	char * p2;
//	*p1 = *((int16*)p2);
//   which works fine on the Multimax but causes a memory alignment interrupt
//   on the Sparc (sometimes).
//
inline void Copy2octets( const octet * from, octet * to )
{
	*to = *from;
	*(to + 1) = *(from + 1);
}

inline void Copy4octets( const octet * from, octet * to )
{
	*to = *from;
	*(to + 1) = *(from + 1);
	*(to + 2) = *(from + 2);
	*(to + 3) = *(from + 3);
}

inline int16 Align16( const octet * p )
{
	register union { int16 number; octet bytes[2]; } temp;
	temp.bytes[0] = *p;
	temp.bytes[1] = *(p + 1);
	return( temp.number );
}

inline int32 Align32( const octet * p )
{
	register union { int32 number; octet bytes[4]; } temp;
	temp.bytes[0] = *p;
	temp.bytes[1] = *(p + 1);
	temp.bytes[2] = *(p + 2);
	temp.bytes[3] = *(p + 3);
	return( temp.number );
}

//-------------------- class Word member functions: ---------------------------
inline
Word::Word()
{
}

inline
Word::Word( const octet source )
{}

inline
Word::Word( const int16 source )
{}

inline
Word::Word( const int32 source )
{}

inline
Word::~Word()
{}

inline
void
Word::copyFrom( const octet * source ) 
{	// NOTICE: this version of equals() will
				//   grab a whole word out of a datagram; the
				//   int16* version just grabs 2 octets, and
				//   the int32* version grabs 4.
	Assert ( NOTREACHED );	// Subclass responsibility
}


inline
void
Word::equals( const octet * source ) 
{	// NOTICE: this version of equals() will
				//   grab a whole word out of a datagram; the
				//   int16* version just grabs 2 octets, and
				//   the int32* version grabs 4.
	Assert ( NOTREACHED );	// Subclass responsibility
}

inline
void
Word::equals( const octet source ) 
{
	Assert ( NOTREACHED );	// Subclass responsibility
}

inline
void
Word::equals( const int16 source ) 
{
	Assert ( NOTREACHED );	// Subclass responsibility
}

inline
void
Word::equals( const int32 source ) 
{
	Assert ( NOTREACHED );	// Subclass responsibility
}

inline
void
Word::equals( const Word8 & source ) 
{
	Assert ( NOTREACHED );	// Subclass responsibility
}

inline
void
Word::equals( const Word16 & source ) 
{
	Assert ( NOTREACHED );	// Subclass responsibility
}

inline
void
Word::equals( const Word32 & source ) 
{
	Assert ( NOTREACHED );	// Subclass responsibility
}

inline
void
Word::copyTo( octet *dest ) const
{	// Copies directly into datagram, changing
				//   as many octets as needed to store value
	Assert ( NOTREACHED );	// Subclass responsibility
}



//-------------------- class Word8 member functions: --------------------------
inline
Word8::Word8() 
{
	myval = 0;
}

inline
Word8::Word8( const octet source )
{
	myval = source;
}

inline
Word8::Word8( const int16 source )
{
	myval = source;		// Will be truncated if necessary, of course
}

inline
Word8::Word8( const int32 source )
{
	myval = source;		// Will be truncated if necessary, of course
}

inline
Word8::~Word8() 
{}

inline octet
Word8::val() const
{
	return( myval );
}

inline void
Word8::copyFrom( const octet *source ) 
{
	myval = *source;
}

inline void
Word8::equals( const octet *source ) 
{
	myval = *source;
}

inline void
Word8::equals( const octet source ) 
{
	myval = source;
}

inline void
Word8::equals( const int16 source ) 
{
	myval = source;
}

inline void
Word8::equals( const int32 source ) 
{
	myval = source;
}

inline void
Word8::equals( const Word8 & source ) 
{
	myval = source.val();
}

inline void
Word8::equals( const Word16 & source ) 
{
	myval = source.val();
}

inline void
Word8::equals( const Word32 & source ) 
{
	myval = source.val();
}

inline void
Word8::copyTo( octet *dest )  const
{	// Copies directly into datagram.
	*dest = myval;
}

//-------------------- class Word13 member functions: -------------------------
inline
Word13::Word13() 
{
	myvals[NOFFSET] = 0;
	myvals[HOFFSET] = 0;
}

inline
Word13::Word13( const octet source ) 
{
	myvals[HOFFSET] = source;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline
Word13::Word13( const int16 source ) 
{
	myvals[HOFFSET] = source & 0x1fff;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline
Word13::Word13( const int32 source ) 
{
	myvals[HOFFSET] = source & 0x1fff;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline
Word13::~Word13()
{}

inline int13
Word13::val()  const
{
	return( myvals[HOFFSET] );
}

inline void
Word13::copyFrom( const octet *source ) 
{	// NOTE: copies 2 octets, not 1.
	myvals[NOFFSET] = Align16( source ) & hton( (int16) 0x1fff );
	myvals[HOFFSET] = ntoh( myvals[NOFFSET] );
}

inline void
Word13::equals( const octet *source ) 
{	// NOTE: copies 2 octets, not 1.
	myvals[NOFFSET] = Align16( source ) & hton( (int16) 0x1fff );
	myvals[HOFFSET] = ntoh( myvals[NOFFSET] );
}

inline void
Word13::equals( const octet source ) 
{
	myvals[HOFFSET] = source;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word13::equals( const int16 source ) 
{
	myvals[HOFFSET] = source & 0x1fff;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word13::equals( const int32 source ) 
{
	myvals[HOFFSET] = source & 0x1fff;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word13::equals( const Word8 & source ) 
{
	myvals[HOFFSET] = source.val();
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word13::equals( const Word16 & source ) 
{
	myvals[HOFFSET] = source.val() & 0x1fff;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word13::equals( const Word32 & source ) 
{
	myvals[HOFFSET] = source.val() & 0x1fff;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word13::copyTo( octet *dest ) const
{	// Copies directly into datagram, changing the contents of
	//   two octets to store the 13-bitvalue (that is, it will stomp
	//   the high-order 3 bits of the 16-bits dest points to)
	int16 temp = myvals[NOFFSET] & hton( (short unsigned int)0xe000 );
	Copy2octets( (octet*)( &temp ), dest );
}

//-------------------- class Word16 member functions: -------------------------
inline
Word16::Word16() 
{
	myvals[NOFFSET] = 0;
	myvals[HOFFSET] = 0;
}

inline
Word16::Word16( const octet source )
{
	myvals[HOFFSET] = source;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline
Word16::Word16( const int16 source )
{
	myvals[HOFFSET] = source;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline
Word16::Word16( const int32 source )
{
	myvals[HOFFSET] = source;		// Might truncate source
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline
Word16::~Word16()
{}

inline int16
Word16::val() const
{
	return( myvals[HOFFSET] );
}

inline void
Word16::copyFrom( const octet *source )
{	// NOTE: copies 2 octets, not 1.
	myvals[NOFFSET] = Align16( source );
	myvals[HOFFSET] = ntoh( myvals[NOFFSET] );
}

inline void
Word16::equals( const octet *source )
{	// NOTE: copies 2 octets, not 1.
	myvals[NOFFSET] = Align16( source );
	myvals[HOFFSET] = ntoh( myvals[NOFFSET] );
}

inline void
Word16::equals( const octet source ) 
{
	myvals[HOFFSET] = source;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word16::equals( const int16 source ) 
{
	myvals[HOFFSET] = source;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word16::equals( const int32 source ) 
{
	myvals[HOFFSET] = source;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word16::equals( const Word8 & source ) 
{
	myvals[HOFFSET] = source.val();
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word16::equals( const Word16 & source ) 
{
	myvals[HOFFSET] = source.val();
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word16::equals( const Word32 & source ) 
{
	myvals[HOFFSET] = source.val();
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word16::copyTo( octet *dest ) const
{	// Copies directly into datagram, changing
	//   as many octets as needed to store value
	Copy2octets( (octet*) (myvals + NOFFSET), dest );
}

//-------------------- class Word32 member functions: -------------------------
inline
Word32::Word32() 
{
	myvals[NOFFSET] = 0;
	myvals[HOFFSET] = 0;
}

inline
Word32::Word32( const octet source )
{
	myvals[HOFFSET] = source;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline
Word32::Word32( const int16 source )
{
	myvals[HOFFSET] = source;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline
Word32::Word32( const int32 source )
{
	myvals[HOFFSET] = source;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline
Word32::~Word32()
{}

inline void
Word32::copyFrom( const octet *source ) 
{	// NOTE: copies 4 octets, not 1.
	myvals[NOFFSET] = Align32( source );
	myvals[HOFFSET] = ntoh( myvals[NOFFSET] );
}

inline void
Word32::equals( const octet *source ) 
{	// NOTE: copies 4 octets, not 1.
	myvals[NOFFSET] = Align32( source );
	myvals[HOFFSET] = ntoh( myvals[NOFFSET] );
}

inline int32
Word32::val() const
{
	return( myvals[HOFFSET] );
}

inline void
Word32::equals( const octet source ) 
{
	myvals[HOFFSET] = source;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word32::equals( const int16 source ) 
{
	myvals[HOFFSET] = source;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word32::equals( const int32 source ) 
{
	myvals[HOFFSET] = source;
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word32::equals( const Word8 & source ) 
{
	myvals[HOFFSET] = source.val();
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word32::equals( const Word16 & source ) 
{
	myvals[HOFFSET] = source.val();
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word32::equals( const Word32 & source ) 
{
	myvals[HOFFSET] = source.val();
	myvals[NOFFSET] = hton( myvals[HOFFSET] );
}

inline void
Word32::copyTo( octet *dest ) const
{	// Copies directly into datagram, changing
	//   as many octets as needed to store value
	Copy4octets( (octet*) (myvals + NOFFSET), dest );
}

#endif // PROXIFY

#endif /* Word_h */
