/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Lee Lup Yuen (lup@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * ApplicationConduitStream.h
 */
 
#ifndef ApplicationConduitStream_h
#define ApplicationConduitStream_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Conduit.h"
#include "Process.h"
#include "Assert.h"
#include "CharacterInputStreamStar.h"
#include "OutputStreamStar.h"
#include "SemaphoreStar.h"

class List;
class Process;
class Semaphore;
class ClassStar;
class OutputStream;
class CharacterInputStream;
class UserSession;
extern ClassStar ApplicationConduitStreamClass;

class ApplicationConduitStream : public ApplicationConduit
{
    /* An ApplicationConduitStream provides a byte-based read/write interface
     * to an underlying message-based ApplicationConduit.  An Application-
     * ConduitStream creates a daemon to queue incoming messages to be read
     * later as bytes.  ApplicationConduitStreams may be created by
     * applications to offer the same functionality as sockets.  Application-
     * ConduitStream may be subclassed to provide networking services such
     * as FTP and Telnet.  Output is buffered, and byte reads may either be
     * line-buffered (by default), raw or character block.
     */
public:
    proxiable ApplicationConduitStream ();
    proxiable ~ApplicationConduitStream ();
    Class *classOf ();

    proxiable virtual Return_Code activeStart (long ipAddress, int localPort,
    	int remotePort);
    proxiable virtual Return_Code passiveStart (int port = -1);
    proxiable virtual void finish ();
    
    proxiable virtual int read (char *buffer, int count);
    proxiable virtual Return_Code write (const char *buffer, int count);
    virtual Return_Code rawWrite (const char *buffer, int count);
    virtual Return_Code rawWriteBlock (const char *buffer, int count);
    proxiable virtual Return_Code flush ();
    proxiable virtual void switchCharacterQueue (int type);
    
    void setUserSession (UserSession *session);
    UserSession *getUserSession ();
    Return_Code insertFromBelow (ConduitMessage *msg);
    TCPConduit *tcpConduit () { return myTCPConduit; }
protected:
    virtual int passivePort () { Assert (0); return 0; }
    static void applicationConduitStreamDaemon (ApplicationConduitStream
    	*client);
    virtual PFV *activeDaemon () {return (PFV*)applicationConduitStreamDaemon;}
    virtual PFV *passiveDaemon (){return (PFV*)applicationConduitStreamDaemon;}
    virtual char *processIncomingMessage (TCPMessage *msg, int &length);
    virtual char *processOutgoingMessage (const char *buffer, int &count);
    void startDaemon (PFV *conduitDaemon);
    void setEndOfStream ();
    void close ();

    SemaphoreRegulatedMessageFIFO incoming;  //  Queue for incoming messages.
    CharacterInputStreamStar characterQueue;    //  Queue for incoming bytes.
    OutputStreamStar echoStream;  //  Stream for echoing the bytes written.

    //  The quickBuffer is used for quickly buffering the first 8192 bytes
    //  written to this conduit.  This makes single-byte writes efficient
    //  (eg. interactive command shells, interactive screen editors).
    enum {quickBufferSize = 8192};
    char quickBuffer [quickBufferSize];
    char *quickBufferPos;
    int quickBufferCount, quickBufferFull;

    //  Queue outgoing is for outgoing byte blocks.  After quickBuffer has been
    //  filled, byte blocks to be written are put into this queue.  This
    //  procedure is slower than using quickBuffer.  outgoingCharacterCount
    //  keeps the total number of characters in quickBuffer and outgoing.
    List *outgoing;
    SemaphoreStar outgoingSemaphore;
    int outgoingCharacterCount;
    
    char *name;                        //  Name of this conduit, for debugging.
    Process *daemon;                   //  Daemon to process incoming messages.
    ApplicationConduitStream *spyStream;// Echo all characters to this stream.
    UserSession *userSession;          //  User session of this conduit.
    SemaphoreStar closed;              //  V'ed when the FIN message is recv'd.
public:
    //  Have to put these methods here because proxify++ complains about
    //  characterQueue and spyStream.
    CharacterInputStream *inputStream () { return characterQueue; }
    void setSpyStream (ApplicationConduitStream *stream) { spyStream=stream; }
    
    /* Method activeStart: Opens an active connection from the indicated
         local port number to the indicated IP address and remote port
	 number. */
    /* Method passiveStart: Waits for a passive connection on the indicated
         port number.  If no port number is specified, use the value returned
	 by function passivePort(). */
    /* Method finish: Close the connection. */
    /* Method read: Read count bytes from the conduit into buffer.  Return
         number of bytes read. */
    /* Method write: Write count bytes from buffer into the conduit. The
         writing is buffered.  Calls processOutgoingMessage before writing. */
    /* Method rawWrite: Write count bytes from buffer into the conduit.  The
         writing is unbuffered.  Called by write when buffer is full. */
    /* Method rawWriteBlock: Write count bytes from buffer into the conduit,
         retrying if unsuccessful.  The writing is unbuffered, and count is
	 assumed to be small enough to fit into a single IP packet.  Called by
	 rawWrite. */
    /* Method flush: Flush the contents of the buffer to the conduit.  Calls
         processOutgoingMessage before flushing. */
    /* Method switchCharacterQueue: Change the behaviour of read() - choose
         line-buffered, raw or character block reads.  See CharacterQueue.h. */
    /* Method setUserSession: Associate the user session with this conduit. */
    /* Method getUserSession: Return the user session associated with this
         conduit. */
    /* Method insertFromBelow: Derived from ApplicationConduit.  Processes
         incoming message and puts bytes into incoming buffer. */
    /* Method setSpyStream: All subsequent output to this object will also
         be sent to the indicated ApplicationConduitStream. */
    /* Method passivePort: Redefined by subclasses to indicate the default
         port number of this conduit for passive connections.  Eg. Telnet
	 uses port number 23.  Called by passiveStart. */
    /* Method applicationConduitStreamDaemon: Daemon to process incoming
         messages and put bytes into incoming byte buffer. */
    /* Method activeDaemon: Defined by subclass to indicate which daemon to
         use for active connections. */
    /* Method passiveDaemon: Defined by subclass to indicate which daemon to
         use for passive connections. */
    /* Method processIncomingMessage: Translate an incoming message into a
         byte array.  Called by daemon.  Used for input filters, such as
	 Telnet's IAC. */
    /* Method processOutgoingMessage: Translate a byte array into another
         byte array suitable for writing.  Used for output filters, such as
	 translating newlines into carriage-return+newlines for Telnet. */
    /* Method startDaemon: Run the daemon. */
    /* Method setEndOfStream: Called when the remote host closes this
         connection.  Indicate to client that there are no more bytes to
	 read. */
    /* Method close: Close up the connection. */
};

#endif
