/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * ZootStateMachine.h
 *
 */

#ifndef ZootStateMachine_h
#define ZootStateMachine_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Semaphore.h"
#include "SemaphoreStar.h"
#include "ConduitTypes.h"
#include "Word.h"
#include "ZootAddress.h"
#include "ConduitMessage.h"
#include "ZootReturnCodes.h"

// I am establishing a triumvirate between CONDUITS, CONNECTION-DESCRIPTORS and
//   STATES. A TCP conduit represents the functionality of transporting TCP
//   messages around. A TCP connection-descriptor represents everything there
//   is to know about a connection (except which TCP conduit is responsible for
//   managing it -- that's the TCP conduit's responsibility to keep track of).
//   A TCP state knows how to manipulate connection descriptors which live in
//   conduits. Telling a state about a connection, a conduit and (where appro-
//   priate) a message causes it to take appropriate action, which might in-
//   volve changing the state of the connection, moving messages around, 
//   checking things, etc....

// ------------------- TCPConnection-Descriptors: ------------------- 
//   (Based on TCBs (Transmission Control Blocks) from RFC 793....)

class TCPState;
class TCPConduit;
class AutoRetransmitBuffer;
class TCPWaitBuffer;
class Conduit;

class TCPConnectionDescriptor {
protected:
	TCPConduit *	_carrier;	// my TCP conduit
	Conduit *	_luser;	// local user's conduit
	IPAddress	_myIPAddress;	// copy of localHost.ipaddr()
	IPAddress	_destIPAddress;
	Word16		_srcPort;
	Word16		_destPort;
	TCPState *	_current_state;
	AutoRetransmitBuffer * _outgoingBuffer;
		// Messages will be dumped here on the way to my_IPConduit so
		//   that they will be automatically retransmitted until they
		//   are ack'ed.
	TCPWaitBuffer * _waitBuffer;
		// The WaitBuffer is for messages that have been insertTop()ed
		//   but can't go out yet, either because the connection isn't
		//   established, or the window is too small....
	MessageID	_outstandingList;
			// Handle for last message dropped into the output
			//   buffer for this connection (assumed to be a chain
			//   of msg-ids -- see Buffer.h) -- head of linked list
	MessageID	_waitingList;
			// Similarly, need a handle to point into the buffer
			//   of messages that aren't ready to be sent yet (z.B.
			//   sent before connection estblsh'd or when window
			//   is too small).
	Handle		_handle;
			// Need to be able to find out, from the CD, what the
			//   handle is, in order to communicate with the
			//   wait-buffer. 
	TCPPseudoHeader	_pseudoheader;
	Word32		SND_UNA;
			// Pointer to first byte of outstanding data that 
			//   hasn't been ack'd yet
	Word32		SND_NXT;
			// Next outgoing sequence number to be used
	Word16		SND_WND;
			// Allowable range of outstanding sequence numbers
	Word16		SND_UP;			// Send Urgent Pointer
	Word32		SND_WL1;
			// Segment sequence number used for last window update
	Word32		SND_WL2;
			// Segment acknowledgement number used for last window 
			//   update
	Word32		SND_ISN;		// ISN=Initial Sequence Number
	Word32		RCV_NXT;
			// Seqence number of next expected byte
	Word16		RCV_WND;
			// Allowable range of outstanding sequence numbers
	Word16		RCV_UP;
	Word32		RCV_ISN;
	Word32		_timestamp;
			// I implement TIME-WAIT state by timestamping the CD
			//   and re-using it only if it's been waiting long
			//   enough. So the transition from TIME-WAIT to
			//   CLOSED is not really automatic -- doesn't matter
	//SemaphoreStar	_mutex;
	Semaphore *	_mutex;
public:
	TCPConnectionDescriptor();
	virtual ~TCPConnectionDescriptor();

	// These are the methods the conduit invokes on a CD to service 
	//   requests from the conduits to which it is connected:
	virtual Return_Code	openConnection();
	virtual Return_Code	closeConnection();
	virtual Return_Code	abortConnection();
	virtual Return_Code	processOutgoingMessage(TCPMessage * msg);
	virtual Return_Code	processIncomingMessage(TCPMessage * msg);

	// These are the methods the state objects invoke on a CD to get
	//   the actual work done:
	TCPConduit *	carrier();
	Conduit *	user();
	virtual AutoRetransmitBuffer *	outgoingBuffer();
			/* Returns a pointer to the ARB -- used by all the
			 * state-machine stuff to send outgoing messages. 
			 */
	virtual TCPWaitBuffer  *	waitingBuffer();
			/* Also used by the state-machine to deal with messages
			 * that have been sent on a not-quite-ready
			 * TCP connection. 
			 */
	IPAddress	sourceIPAddress();
	IPAddress	destinationIPAddress();
	Word16		sourcePort();
	Word16		destinationPort();
	TCPState *	state();
	MessageID	outstanding();
				// Handle for retransmit-buffer to retrieve
				//   chain of msgs being retransmitted
	MessageID	waiting();
				// Same gazoo, but for the waiting-msg-buffer
	Handle		handle();
	TCPPseudoHeader	* pseudoheader();	// ptr to pseudoheader
	Word32		SendUnacknowledged();
	Word32		SendNext();
	Word16		SendWindow();
	Word16		SendUrgentPointer();
	Word32		SendWL1();
	Word32		SendWL2();
	Word32		SendInitialSequenceNumber();
	Word32		ReceiveNext();
	Word16		ReceiveWindow();
	Word16		ReceiveUrgentPointer();
	Word32		ReceiveInitialSequenceNumber();
	Word32		timestamp();
	Semaphore *	mutex();

	void	setCarrier( TCPConduit * newcarrier );
	void	setUser( Conduit * new_luser );
	void	connect( Conduit * target );
		  // Connects the wait-buffer and ARB to target
	void	setLocalIPAddress( const IPAddress & newlocalipaddr );
	void	setDestinationIPAddress( const IPAddress & newdestipaddr );
	void	setSourcePort( const Word16 & newscrport );
	void	setDestinationPort( const Word16 & newdestport );
	void	gotoState( TCPState * newstate );
	void	linkOutstanding( MessageID outstandingmsg );
	void	linkWaiting( MessageID waitingmsg );
	void	setHandle( Handle newhandle );
		// Play with pseudoheader by hand, using the pointer-getter
	void	setSendUnacknowledged( const Word32 & newSND_UNA );
	void	setSendNext( const Word32 & newSND_NXT );
	void	setSendWindow( const Word16 & newSND_WND );
	void	setSendUrgentPointer( const Word16 & newSND_UP );
	void	setSendWL1( const Word32 & newSND_WL1 );
	void	setSendWL2( const Word32 & newSND_WL2 );
	void	setSendInitialSequenceNumber( const Word32 & newSND_ISN );
	void	setReceiveNext( const Word32 & newRCV_NXT );
	void	setReceiveWindow( const Word16 & newRCV_WND );
	void	setReceiveUrgentPointer( const Word16 & newRCV_UP );
	void	setReceiveInitialSequenceNumber( const Word32 & newRCV_ISN );
	void	setTimestamp( const Word32 & newtimestamp );
};

// -------------------- TCPState: --------------------
// The classes TCPState, ClosedState, etc. are each only going to have one
//   (globally accessible) instance. By telling a state-object to process
//   a message, it will do all the bit-twiddling nonsense, put the appropriate
//   messages into the right buffers/conduits and return a pointer to the
//   resulting state (which it just put the connection-descriptor into).
// The idea is that each Connection will use its State-pointer to point to one
//   of the state objects, and re-assign to it to process each message. The
//   virtual functions basically replace gory switch-statements by v-table
//   dereferences.
// The returned state-pointer can come in handy, but will be 0 for illegal
//   operations-- so cd.gotoState( cd.state()-><operation>... ) can be
//   dangerous, as well as unnecessary.
// (Names taken right out of RFC793 -- pp.21, 22)...

class TCPState {
protected:
	//---------------------------------------------------------------------
	// The state objects use a couple of helper-functions to do mindless
	//   tasks that need to be done commonly (such as scribbling on a
	//   message's TCP header, copying stuff out of the CD). The sendFoo()
	//   methods embody the state-independent aspects of sending various
	//   sorts of messages, and do dumb stuff like set up the pseudoheader
	//   correctly, copy information from the CD into the message, figure
	//   the checksum, etc. The also link the messages they send into the
	//   CD's outstanding-message linked list. >>>> sendMessage() assumes
	//   that the TCP message has an IP message to live in! <<<< If there
	//   isn't one, trying to set the source- and destination-addresses
	//   will cause grief. prepareIncomingMessage() does the somewhat
	//   grungy translation from simple IP-message to TCP-message as 
	//   payload in that IP-message. checkIncomingMessage() does some
	//   processing common to most of the states (all but CLOSED, LISTEN,
	//   and SYN-SENT).  acknowledge() takes an ack-number and chases
	//   down the linked list of outstanding messages for the specified
	//   CD and deletes any messages that have just been acknowledged
	//   (it deletes all and only messages that have been totally ack'ed --
	//   if only part of a message is (somehow) ack'ed, it will remain
	//   in case the un-ack'ed part of its data needs to be resent).
	inline Word16		combineIPChecksums( const Word16 & chksum1,
					const Word16 & chksum2 );
	inline Boolean		isBetween( int32 segnum1, int32 segnum2,
					int32 segnum3 );
	Return_Code	sendConnectRequest( TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	Return_Code	sendAbort( TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	Return_Code	sendAbort( TCPConnectionDescriptor & cd,
				TCPConduit * carrier, const Word32 & seqnum );
	Return_Code	sendAbort( TCPConnectionDescriptor & cd,
				TCPConduit * carrier,
				const Word32 & seqnum, const Word32 & acknum );
	Return_Code	sendMessage( TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg);
	Return_Code	sendMessage( TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg,
				const Word32 & seqnum );
				// In a few cases, the SND_NXT field of the
				// cd is inappropriate as a sequence number,
				// so this is an ugly hack to allow the
				// Right Thing to be done then.
	Return_Code	sendCloseRequest( TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	Return_Code	checkIncomingMessage( TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg);
	void		acknowledge( TCPConnectionDescriptor & cd,
				TCPConduit * carrier, const Word32 & acknum );
public:
	TCPState();
	virtual ~TCPState();
	virtual TCPState * openConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	virtual TCPState * closeConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	virtual TCPState * abortConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	virtual	TCPState * processOutgoingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
	virtual	TCPState * processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
};

class ClosedState : public TCPState {
protected:
public:
	ClosedState();
	~ClosedState();
	TCPState * openConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
};

class ListenState : public TCPState {
protected:
public:
	ListenState();
	~ListenState();
	TCPState * openConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * closeConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * abortConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * processOutgoingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
	TCPState * processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
};

class Syn_SentState : public TCPState {
protected:
public:
	Syn_SentState();
	~Syn_SentState();
	TCPState * closeConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * abortConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * processOutgoingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
	TCPState * processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
};

class Syn_ReceivedState : public TCPState {
protected:
public:
	Syn_ReceivedState();
	~Syn_ReceivedState();
	TCPState * closeConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
		// Returns ERROR if there is outstanding data -- try again
		//   when the connection becomes established
	TCPState * abortConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * processOutgoingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
	TCPState * processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
};

class EstablishedState : public TCPState {
protected:
public:
	EstablishedState();
	~EstablishedState();
	TCPState * closeConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * abortConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * processOutgoingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
	TCPState * processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
};

class Fin_Wait_1State : public TCPState {
protected:
public:
	Fin_Wait_1State();
	~Fin_Wait_1State();
	TCPState * abortConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
};

class Fin_Wait_2State : public TCPState {
protected:
public:
	Fin_Wait_2State();
	~Fin_Wait_2State();
	TCPState * abortConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
};

class Close_WaitState : public TCPState {
protected:
public:
	Close_WaitState();
	~Close_WaitState();
	TCPState * closeConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
		// Returns ERROR if there is outstanding data -- try again
		//   after it has been acknowledged by the other end.
	TCPState * abortConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
};

class ClosingState : public TCPState {
protected:
public:
	ClosingState();
	~ClosingState();
	TCPState * abortConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
};

class Last_AckState : public TCPState {
protected:
public:
	Last_AckState();
	~Last_AckState();
	TCPState * abortConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier );
	TCPState * processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
};

// Right now, I'm nonplussed as to how to implement Time_Wait. The idea is to
//   leave a C-D around for awhile after the connection is closed so that any
//   incoming messages that need to be ack'ed still have a connection to refer
//   to.... I suppose the TCP conduit could check on waiting connections, or
//   have a daemon to kill things off later on. -JZ
class Time_WaitState : public TCPState {
protected:
public:
	Time_WaitState();
	~Time_WaitState();
	TCPState * processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg );
};

//-----------------------------------------------------------------------------
inline Word16
TCPState::combineIPChecksums( const Word16 & chksum1, const Word16 & chksum2 )
{
	int32 tmp;
	tmp = chksum1.val() + chksum2.val();
	while( (tmp & 0xffff0000)  != 0x00000000 ) {
	    // Need while-loop since may overflow upon adding in carry-bits
	    tmp = ( (tmp >> 16) & 0x0000ffff ) + ( tmp & 0x0000ffff );
	}
	return( Word16( tmp ) );
}
inline Boolean
TCPState::isBetween( int32 segnum1, int32 segnum2, int32 segnum3 )
{
	// I assume segment number 3 == segment number 1 plus a window size.
	//   Thus there are two cases: either segnum1 is so close to 2**32
	//   that segnum3 has wrapped around, or not. So just need to compare
	//   segnum1 and segnum3 to decide which inequality to check:
	if( segnum1 < segnum3 ) {
	    if( segnum1 <= segnum2 && segnum2 <= segnum3 ) {
		return( TRUE );
	    }
	    else {
		return( FALSE );
	    }
	}
	else {
	    if( segnum1 <= segnum2 || segnum2 <= segnum3 ) {
		return( TRUE );
	    }
	    else {
		return( FALSE );
	    }
	}
}
inline Semaphore *
TCPConnectionDescriptor::mutex()
{
	return( _mutex );
}
//-----------------------------------------------------------------------------
// Global pointers to the (single) instances of the class State objects:
extern ClosedState * CLOSED;
extern ListenState * LISTEN;
extern Syn_SentState * SYN_SENT;
extern Syn_ReceivedState * SYN_RECEIVED;
extern EstablishedState * ESTABLISHED;
extern Fin_Wait_1State * FIN_WAIT_1;
extern Fin_Wait_2State * FIN_WAIT_2;
extern Close_WaitState * CLOSE_WAIT;
extern ClosingState * CLOSING;
extern Last_AckState * LAST_ACK;
extern Time_WaitState * TIME_WAIT;

#endif
