#ifndef UDPConduit_h
#define UDPConduit_h


#ifdef __GNUC__
#pragma interface
#endif

#include "ZootReturnCodes.h"
#include "Conduit.h"
#include "ConduitProtocolIdentifiers.h"
#include "ZootAddress.h"
		// to get IPAddress


//----------------------------------------------------------------------------
//  A UDP address needs to specify both source- and destination port numbers
//  and destination IP address.  Since I plan on using implicit addressing
//  (i.e. filling in the fields in the UDP header), I don't know how often
//  these things will be used -- that's why I didn't add them to ZootAddress.h
//  yet. -JZ

class UDPAddress : public NetworkAddress {
	Word16	_sourcePort;
	Word16	_destinationPort;
	IPAddress	_destinationIPAddress;
protected:
public:
	UDPAddress();
	~UDPAddress();
	void		setSourcePort( int16 port );
	void		setSourcePort( Word16 port );
	void		setDestinationPort( int16 port );
	void		setDestinationPort( Word16 port );
	void		setDestinationIPAddress( IPAddress ipaddr );
	Word16		sourcePort();
	Word16		destinationPort();
	IPAddress	destinationIPAddress();
};
inline
UDPAddress::UDPAddress() : _destinationIPAddress()
{

  short unsigned int tmp1 = 0;

  _sourcePort = tmp1;
  _destinationPort = tmp1;

}
inline
UDPAddress::~UDPAddress()
{
}
inline void
UDPAddress::setSourcePort( int16 port )
{
	_sourcePort = port;
}
inline void
UDPAddress::setSourcePort( Word16 port )
{
	_sourcePort = port;
}
inline void
UDPAddress::setDestinationPort( int16 port )
{
	_destinationPort = port;
}
inline void
UDPAddress::setDestinationPort( Word16 port )
{
	_destinationPort = port;
}
inline void
UDPAddress::setDestinationIPAddress( IPAddress ipaddr )
{
	_destinationIPAddress = ipaddr;
}
inline Word16
UDPAddress::sourcePort()
{
	return( _sourcePort );
}
inline Word16
UDPAddress::destinationPort()
{
	return( _destinationPort );
}
inline IPAddress
UDPAddress::destinationIPAddress()
{
	return( _destinationIPAddress );
}

//----------------------------------------------------------------------------
//  A UDP Port Conduit is what the UDP Port Monster (manager) returns to you
//  when you connect to a UDP port.  Connecting a conduit to a UDPPC means it
//  is able to receive any messages coming in off the network whose
//  destination port number matches that for the port (demultiplexing further
//  amongst all the folks sending to that port is somebody else's problem).
//  Inserting a message into a UDPPC just sends it out through the UDPConduit
//  itself.  The existence of a UDPPC means that someone owns that port, so
//  to disconnect from a UDP port, the conduit needs to unreference (delete)
//  the UDPPC associated with it.  When I get around to writing ConduitStar.h,
//  using a ConduitStar for the UDPPortConduit should be the Right Thing. -JZ

//  Since my idea is to have a UDPPC be a totally generic Conduit (I would
//  prefer it if nobody but the UDPConduit/UDPPortMonster knew even of the
//  class' existence), my original idea of turning checksumming on/off by
//  diddling the UDPPC won't work.  My current plan is to do that by diddling
//  each UDP message -- setting the Checksum field in the header to some
//  nonzero value tells the UDPPC that it should calculate the checksum
//  before sending the thing. I might actually convince UDP messages to
//  checksum themselves, except that the pseudoheader involves knowing the
//  local IP address on top of everything else, and I am not sure that it
//  will ultimately be wise to force applications to totally mess around with
//  all that nonsense. -JZ

//  Notice that, since I am using implicit addressing (I don't see any
//  straightforward way of getting around it, nor a reason to want to),
//  the address passed in with insertFromAbove() should always be 0. I
//  suppose I will allow a UDPAddress to be passed in, and copy the stuff
//  out of it into the UDP- and IP headers, though I frown on this. -JZ

class UDPPortMonster;	// forward reference

class UDPPortConduit : public Conduit {
protected:
	int16	_port;
	UDPPortMonster * _monster;
	Conduit * _UDP;
	Conduit * _user;	// attached conduit using this port
public:
	UDPPortConduit( int16 port, UDPPortMonster * monster, Conduit * udp );
	~UDPPortConduit();
	virtual Return_Code	connectFromAbove( Conduit * c,
					ProtocolID proto );
	virtual Return_Code	insertFromAbove( ConduitMessage * msg,
					NetworkAddress * addr );
	virtual Return_Code	insertFromBelow( ConduitMessage * msg );

	// These methods should never be called on a UDPPC:
	virtual Return_Code	connectTo( Conduit * c );
	virtual Return_Code	connectFromBelow( Conduit * c,
					ProtocolID proto,
					int   his_handy,
					int & my_handy );
	virtual Return_Code	disconnectFrom( Conduit * c );
	virtual Return_Code	disconnectFromAbove( Conduit * c );
};

//----------------------------------------------------------------------------
//  A PortMonster is a manager that handles creating and destroying
//  connections, and interacts with a demultiplexer. It hands out PortConduits
//  (in this case, UDPPortConduits) that know what port they are responsible
//  for, in response to calls to the getPort() function. It also is the object
//  that creates the UDPConduit and connects it up to IP.
//  It makes sense for the demultiplexer itself to check, upon getPort-request,
//  whether someone is already attached to that port.
//  Deleting a PortConduit will ask the Monster to update the demux table and
//  delete the entry it created when that PortConduit was created.
//  The map() function is used by UDP to find the Conduit attached to a
//  particular port. Simple, no?

class Semaphore;
class UDPHashTable;

class UDPPortMonster {
protected:
	UDPHashTable *	_PortTable;
	Semaphore *	_mutex;
	Conduit * _UDP;
	friend class UDPConduit;	// allowed to call map()
	friend class UDPPortConduit;	// allowed to call releasePort()
	Conduit *	map( int16 port );
	void		releasePort( int16 port );
public:
	UDPPortMonster();
	virtual ~UDPPortMonster();
	Return_Code	getPort( Conduit *& port_conduit, int16 port );
};

//----------------------------------------------------------------------------
class ConduitMessage;

class UDPConduit : public Conduit {
protected:
	IPConduit *		_IPConduit;
	UDPPortMonster *	_monster;
public:
	UDPConduit( UDPPortMonster * monster );
	~UDPConduit();

	virtual Return_Code	insertFromAbove( ConduitMessage * msg,
					NetworkAddress * addr );
	virtual Return_Code	insertFromBelow( ConduitMessage * msg );

	virtual Return_Code	connectTo( Conduit * c );
	virtual Return_Code	connectFromBelow( Conduit * c,
					ProtocolID proto,
					int   his_handy,
					int & my_handy );
	virtual Return_Code	disconnectFrom( Conduit * c );

	// these functions are never called on a UDP conduit (use Monster):
	virtual Return_Code	connectFromAbove( Conduit * c,
					ProtocolID proto );
	virtual Return_Code	disconnectFromAbove( Conduit * c );
};

#endif
