/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *     University of Illinois at Urbana-Champaign
 *     Department of Computer Science
 *     1304 W. Springfield Ave.
 *     Urbana, IL 61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Process.h: Process class description
 */

#ifndef Process_h
#define Process_h


#ifdef __GNUC__
#pragma interface
#endif

#include "DomainStar.h"
#include "FileSystemInterfaceStar.h"
#include "FreeRunningTimer.h"
#include "FreeRunningTimerStar.h"
#include "InputStreamStar.h"
#include "Lock.h"
#include "NameServerStar.h"
#include "OutputStream.h"
#include "OutputStreamStar.h"
#include "PhysicalMemoryChain.h"
#include "ProcessContainer.h"
#include "ProcessorContext.h"
#include "ProcessStar.h"
#include "ProxiableObject.h"
#include "StackSize.h"
#include "Time.h"
#include "Types.h"
#include "ProcessorException.h"
#include "Interrupt.h"

class Domain;
class CPU;
class BootInterrupt;
class BlockedInfo;
class Process;
class Semaphore;
class NameServer;
class ObjectProxy;
class ProcessResourceUsage;
struct stackFrame;

typedef int (*ENTRYTYPE)(int, char **);

#ifdef INSTRUMENT_PROCESS
extern void ProcessConstructInstruments();
#endif

#define BUST_LOCKS
#define MAX_LOCKS_HELD 32

extern ClassStar SchedulerInfoClass;
class SchedulerInfo : public Object {
public:
  SchedulerInfo();
  ~SchedulerInfo();

  Class * classOf();
};

inline
SchedulerInfo::SchedulerInfo()
{
}

class ProcessResourceUsage {
public:
  TimeValue userTime;
  TimeValue systemTime;
  int maxResidentSetSize;
  int ixResidentSetSize;
  int idResidentSetSize;
  int isResidentSetSize; 
  int minorFault;
  int majorFault;
  int numSwap;
  int inBlock;
  int outBlock;
  int msgSend;
  int msgRecv;
  int sigRecv;
  int volContextSwitch;
  int involContextSwitch;
  ProcessResourceUsage() 
  {
    maxResidentSetSize = ixResidentSetSize =
    idResidentSetSize = isResidentSetSize =
    minorFault = majorFault  =
    numSwap = inBlock = outBlock =
    msgSend = msgRecv = sigRecv =
    volContextSwitch = involContextSwitch = 0;
  }
};

enum ProcessState {
  ProcessNew= 0,
  ProcessReady = 1,
  ProcessRunning = 2,
  ProcessBlocked = 3,
  ProcessDead = 4
};

extern int ProcessResourceUsageOn;
extern int ProcessResourceUsageReportOn;

class BlockedInfo {
protected:
    Process * _next;
    Process * _prev;
    unsigned _buf[10];
public:
    Process * next() { return _next; }
    Process * prev() { return _prev; }
    void setNext( Process * p ) { _next = p; }
    void setPrev( Process * p ) { _prev = p; }
};

extern ClassStar ProcessClass;
class Process : public ProxiableObject {
  friend class SVCTrap;
  friend class ProcessorContext;
  friend class CPU;
  friend class Kernel;
  friend class FIFOScheduler;
  friend class NonLockedFIFOScheduler;
  friend class BusyWait;
  friend class GraciousSemaphore;

protected:

  ~Process();

  ProcessorContext * _context;
  ProcessorContext * _newContext;
  ProcessorContext * _oldContext;
  ProcessContainer * _readyContainer;

  Object *  _blockedOn;
  BlockedInfo _blockedInfo;

  /* 
   * Accounting Information 
   */

  ProcessResourceUsage _usage;
  TimeValue _processorTime;
  ProcessState _state;
  unsigned int _processorStartTime;
  unsigned int _processorTimeInt;
  unsigned int _userStartTime;

  /*
   * The following few are for process containers to use.
   */
  int _quantum;
  int _residual;
  int _deadline;
  int _priority;
  int _event;
  int _completionStatus;
  unsigned int _serviceTime;
  Lock _completionLock;
  SchedulerInfo * _schedulerInfo;
  Semaphore * _notifySemaphore;
  int * _notifyStatus;
  Domain * _currentDomain;

  DomainStar _domain;
  ProxiableObjectStar _domainCleaner;

  CPU * runningCPU;

  ObjectProxy * _temporaryProxy;

  /*
   * Called once the a process's context is set up in order to
   * restore the state of the processor to that needed by the process.
   */
  friend class ApplicationProcess;
  friend class TimeSliceInterrupt;
  void restore( Process * previousProcess );

  /*
   * Subclass are responsible for implementing this to
   * "save any important state".
   */
  virtual void save(){}

  /*
   * Subclasses define this to do class specific stuff at restore
   * time.
   */
  virtual void basicRestore(){}

  /*
   * Build a console for this process the first time it is needed.
   */
  OutputStreamStar _console;

  BusyWait stackBusy;   /* Used to avoid race in Context Switch 
             * by locking stack while a processor is
             * running on it.                         
             * Nothing else should use this lock.    
             * Shouldnt really be public, but other
             * Processes must access it.
             */
  unsigned _stackBusy;

  /*
   * Used in the FIFOScheduler to avoid the scheduler having
   * to do memory allocation.
   */
  Process * _next;
  Process * _last;

  /*
   * Function to call when the process takes an SVC Trap. Is this
   * machine dependent, or can we assume that any Choices machine
   * will have something analogous to a supervisor/servicecall trap?
   */
  virtual void SVCHandler(stackFrame*);

  /*
   * These are for cooperation with Locking code, very helpful
   * in finding bugs
   */
  unsigned _numLocksHeld;   // Number of locks I'm holding

#ifdef BUST_LOCKS
  unsigned lockIndex;
  BusyWait *_held[MAX_LOCKS_HELD];
#endif

  int acquiredLock(BusyWait *b) { 
#ifdef BUST_LOCKS
      while(_held[lockIndex]) {
          lockIndex++;
          if (lockIndex == MAX_LOCKS_HELD) 
              lockIndex = 0;
      }
      _held[lockIndex] = b;
#endif
      return ++_numLocksHeld;
  }
  int releasedLock(BusyWait *b) {
#ifdef BUST_LOCKS
      while(_held[lockIndex] != b) {
          lockIndex--;
          if (lockIndex < 0) lockIndex = (MAX_LOCKS_HELD-1);
      }
      _held[lockIndex] = 0;
#endif
      return --_numLocksHeld;
  }

  bool _inGPT;
  ProcessorException *_exception;

  Process( Domain * domain, ProcessorContext * context );
public:

  int numLocksHeld() { return _numLocksHeld;}
  OutputStream & console();

  ProcessorContext * context();
  BlockedInfo * blockedInfo();

  /* 
   * accounting info 
   */
  void stopUserTime();
  void startUserTime();
  unsigned int processTime();
  TimeValue processorTime();
  ProcessResourceUsage * usage();

  /*
   * currentDomain() returns the domain in which the process is currently
   *  executing.  domain() returns the domain in which the process was
   *  created.
   */
  Domain * domain();
  proxiable Domain * currentDomain();
  void setCurrentDomain( Domain * );
  void setQuantum( int quantum );
  void setResidual( int residual );
  int getResidual();
  int getPriority();
  unsigned int getServiceTime();
  int getDeadline();
  int getProcessEvent();
  void setProcessEvent( int event );
  CPU *getMyCPU();
  void setSchedulerInfo( SchedulerInfo * schedulerInfo );
  int quantum();
  SchedulerInfo * schedulerInfo();

  ObjectProxy * temporaryProxy();

  void changeReadyContainer( ProcessContainer * );

  void inspect( OutputStream * );
  virtual int isPreemptable(){return 1;}

  proxiable virtual int waitForProcess( Process * other );
  proxiable virtual void notifyUponCompletion( Semaphore * sem, int * );
  proxiable virtual void wakeUpWaiter( int status = 0 );
  proxiable virtual int dead(){ return ( _state == ProcessDead );}
  proxiable virtual int completionStatus(){return _completionStatus;}

  /*
   * Give up the processor and resume "anotherProcess"
   */
  proxiable virtual void giveProcessorTo( Process * anotherProcess, 
            ProcessState newState );
  
  void forcedSwitchTo(Process *process);

  Class * classOf();

  void setException(ProcessorException *exception) { _exception = exception; }
  ProcessorException *getException() { return _exception; }
  

  /*
   * Called by a process when it wishes to terminate its own
   * execution.
   */
  proxiable virtual void die( int status = 0 );

  /*
   * Called by a process when it wishes to give up the processor
   * to the next available (ready to run) process.
   */
  proxiable virtual void relinquishProcessor();

  /*
   * Executing this method will place a Process in its readyContainer.
   * This will cause it to eventually start running when an idle
   * processor picks it out of the container. If the readyContainer
   * is ever changed to the garbageCan executing this will delete
   * the process.
   */
  proxiable virtual void ready();

  /*
   * Executing this method will place a Process in a fifoqueue.
   * The next available (ready to run) process will be given the 
   * processor. 
   */

  void block( Object * );

  void getStack();

  void releaseStack();

  /*
   * return name servers for this process.
   */
  NameServer * nameServer();

  /* return name server but for use by application process
   */
  proxiable virtual NameServerRef domainNameServer();

  /* Create and a new process with a copy of the current domain
   * and begin execution at address entry with the stack pointer set to
   * address stackPointer.  Only the writeable unshared memory objects
   * in the current domain are copied to the new domain.  The other
   * memory objects are directly inserted into the new domain.  Return
   * a pointer to the new process.  The new process will not run until
   * you ready() it.
   */
  proxiable virtual ProcessRef clone (const char *entry, const char
    *stackPointer);

  /*
   * exec
   */

  virtual void exec( OPFV startAddr, int arg1, char ** arg2 );
  virtual void exec(PFV startAddr, int arg1, char ** arg2 );

  /*
   * usage
   */

  proxiable virtual void getResourceUsage( ProcessResourceUsage * buf );

  //  Simple hack to get Telnet and pipes working.  These are per-process
  //  streams and file system interface.  - Lup
  OutputStreamStar standardOutput;
  InputStreamStar standardInput;
  FileSystemInterfaceStar fileSystemInterface;

};

//#ifdef notdef
/*
 * Cfront requires that inline functions be defined before they are used;
 * therefore, put all the basic access functions first.
 */

inline
void
Process::stopUserTime()
{
        if ( ProcessResourceUsageOn ) {
                Assert( _state == ProcessRunning );
    unsigned int time = SystemTimer->time();
    unsigned int elapsed = time - _userStartTime;
    Assert( elapsed < 0x80000000 );
                _usage.userTime += elapsed;
        }
}

inline 
void
Process::startUserTime()
{
        if ( ProcessResourceUsageOn ) {
                Assert( _state == ProcessRunning );
                _userStartTime = SystemTimer->time();
        }
}

inline
unsigned int
Process::processTime()
{
        if ( ProcessResourceUsageOn ) {
                if ( _state != ProcessRunning )
                        return _processorTimeInt;
    unsigned int time = SystemTimer->time();
    unsigned int elapsed = time - _processorStartTime;
    Assert( elapsed < 0x80000000 );
                return _processorTimeInt + elapsed;
        } else {
                return 0;
        }
}

inline
TimeValue 
Process::processorTime()
{
  if ( ProcessResourceUsageOn ) {
                if ( _state != ProcessRunning )
                        return _processorTime;
                unsigned int time = SystemTimer->time();
                unsigned int elapsed = time - _processorStartTime;
                Assert( elapsed < 0x80000000 );
                return _processorTime + elapsed;
        } else {
                TimeValue dummy;
    return dummy;
        }
}

inline
ProcessResourceUsage * 
Process::usage()
{
  return( & _usage );
}

inline
void
Process::getStack()
{
        stackBusy.acquire();    /* lock stack during context switch to avoid
                                race for stacks */
        _stackBusy = 1;
}

inline
void
Process::releaseStack()
{
        if (_stackBusy) {
                _stackBusy = 0;
                stackBusy.release(); /* unlock stack at end of context switch */
  }
}

inline int
Process::getPriority()
{
  return( _priority );
}

inline ProcessorContext *
Process::context()
{
  return( _context );
}

inline unsigned int
Process::getServiceTime()
{
  return( _serviceTime );
}

inline int
Process::getDeadline()
{
  return( _deadline );
}

inline int
Process::getProcessEvent()
{
  return( _event );
}

inline void
Process::setProcessEvent( int event )
{
  _event = event;
}

inline CPU *
Process::getMyCPU()
{
  return( runningCPU );
}

inline Domain *
Process::domain()
{
  return( _domain );
}

inline Domain *
Process::currentDomain()
{
  return( _currentDomain );
}

inline void
Process::setCurrentDomain( Domain * domain )
{
  _currentDomain = domain;
}

inline void
Process::setQuantum( int quantum )
{
  _quantum = quantum;
}

inline void
Process::setResidual( int residual )
{
  _residual = residual;
}

inline int
Process::getResidual()
{
  return _residual;
}

inline void
Process::setSchedulerInfo( SchedulerInfo * schedulerInfo )
{
  _schedulerInfo = schedulerInfo;
}

inline int
Process::quantum()
{
  return ( _quantum );
}

inline SchedulerInfo *
Process::schedulerInfo()
{
  return ( _schedulerInfo );
}

inline void
Process::notifyUponCompletion( Semaphore* sem, int * status )
{
  /* only one process can be notified */
  Assert( _notifySemaphore == 0);
  _notifySemaphore = sem;
  _notifyStatus = status;
  *status = 0;
}

inline 
OutputStream &
Process::console()
{
  Assert( _console != 0 );
  return( *(OutputStream *)_console );
}

inline 
BlockedInfo *
Process::blockedInfo()
{
  return ( &_blockedInfo );
}

/*
 * Normally, a process inherits its creators readyContainer. It (or it's
 * creator) may change it.
 */
inline void
Process::changeReadyContainer( ProcessContainer * newContainer )
{
  _readyContainer = newContainer;
}

/*
inline void
Process::ready()
{
  Assert( _readyContainer != 0 );
  Assert( _state != ProcessReady );
  _state = ProcessReady;
  _readyContainer->add( this );
}
 */
//#endif

extern ClassStar ApplicationProcessClass;
class ApplicationProcess : public Process {
protected:
  StackSize _userStackSize;

  ~ApplicationProcess();
  void save();
  void basicRestore();
public:
  /*
   * Process constructor for use by applications.
   */
  ApplicationProcess( ENTRYTYPE entry, int arg1 = 0,
    char ** arg2 = 0, StackSize userStackSize = normal );
  /*
   * Normal process constructor. Builds a stack of default size 16K
   * which can be overridden.
   */
  ApplicationProcess( OPFV, Domain *, int arg1, int arg2, StackSize = normal );
  proxiable ApplicationProcess( PFV, Domain *, int arg1, int arg2, StackSize = normal );
  /*
   * Process constructor for processes with already-present stack
   * memory.
   */
  ApplicationProcess( OPFV, Domain *, int arg1, int arg2, char * userStackPointer );
  ApplicationProcess( PFV, Domain *, int arg1, int arg2, char * userStackPointer );

  /*
   * Process constructor for cloned process
   */
  ApplicationProcess( Domain *, ApplicationProcess * process );

  /*
   * Can only clone application process
   */
  virtual ProcessRef clone ( const char *entry, const char
           *stackPointer);
  /*
         * Only application process can exec
   */
  virtual void exec( OPFV startAddr, int arg1, char ** arg2 );
  virtual void exec( PFV startAddr, int arg1, char ** arg2 );

  virtual StackSize userStackSize();

  Class * classOf();
};

//#ifndef PROXIFY /* Proxify doesnt like this and generates unresolved ctors */
extern ClassStar InterruptProcessClass;
class InterruptProcess : public Process {
protected:
  ~InterruptProcess();
public:
  InterruptProcess( OPFV, Domain *, StackSize, int arg );
  InterruptProcess( PFV, Domain *, StackSize, int arg );

  int isPreemptable(){return 0;}
  Class * classOf();
};

extern ClassStar SystemProcessClass;
class SystemProcess : public Process {
protected:
  ~SystemProcess();
public:
  SystemProcess( OPFV entry, Domain * domain, StackSize, int arg );
  SystemProcess( PFV entry, Domain * domain, StackSize, int arg );

  int isPreemptable(){return 0;}
  Class * classOf();
};

extern ClassStar PerCPUIdleProcessClass;
class PerCPUIdleProcess : public Process {
protected:
  ~PerCPUIdleProcess();
public:
  PerCPUIdleProcess( PFV entry, Domain * domain, StackSize, int arg );

  int isPreemptable(){return 0;}
  Class * classOf();
};

extern ClassStar PreemptableSystemProcessClass;
class PreemptableSystemProcess : public SystemProcess {
protected:
  ~PreemptableSystemProcess();
  void basicRestore();    // Insert timeslice timer
  void save();
public:
  PreemptableSystemProcess( OPFV, Domain *, StackSize, int arg );
  PreemptableSystemProcess( PFV, Domain *, StackSize, int arg );
  int isPreemptable(){return 1;}
  Class * classOf();
};
//#endif /* PROXIFY */

#define RESTART_LIMIT (1<<0)

#include "ThisProcess.h"
class RestartableProcess {
 public:
  RestartableProcess(PFV entry, unsigned policy):entry(entry),policy(policy) {};
  void run() {
    unsigned count = 0;
    while(1) {
      try {
        count++;
        entry();
      } catch (Exception *e) {
        if ((policy & RESTART_LIMIT)&&(count == 5)) {
          Console() << "Restart limit reached, killing process\n" << eor;
          thisProcess()->die();
          break;
        } else {
            //Console() << "Caught Exception and restarting\n" << eor;
        }
      }
    }
  }
 private:
  PFV entry;
  unsigned policy;
};

// Maybe eventually create a class to hold policies
#define MICROREBOOT_RETRY (1 << 0)
#define MICROREBOOT_SANITIZE (1 << 1)

#define MICROREBOOTABLE(OBJECT,METHOD,POLICY,POLICY_ARG)                \
  {                                                                     \
    unsigned numTries = 0;                                              \
    while(1) {                                                          \
      try {                                                             \
        numTries++;                                                     \
        OBJECT->METHOD;                                                 \
        break;                                                          \
      } catch (Exception *e) {                                          \
        Console() << "Retrying, attempt #"  << numTries << "\n" << eor; \
        if ((POLICY & MICROREBOOT_RETRY) && (numTries == POLICY_ARG)) { \
          Console() << "Retry limit reached\n" << eor;                  \
          break;                                                        \
        }                                                               \
        /*try {                                                           \
          if ((POLICY & MICROREBOOT_SANITIZE) && OBJECT->hasSanitizer()) \
            OBJECT->sanitize();                                         \
        } catch (Exception *e) {                                        \
          Console() << "Exception while sanitizing. Rethrowing\n" << eor; \
          throw e;                                                      \
        }*/                                                             \
      }                                                                 \
    }                                                                   \
  }


#endif /* Process_h */
