/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David Raila (raila@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef ARMCPU_h
#define ARMCPU_h


#ifdef __GNUC__
#pragma interface
#endif

#include "MachineTypes.h"
#include "CPU.h"
#include "ARMMMU.h"
#include "AsmBridge.h"
#include "ARMContext.h"

enum ARMCPUMode {
    ARM_ABT = 0x000017, 
    ARM_FIQ = 0x000011, 
    ARM_IRQ = 0x000012, 
    ARM_SVC = 0x000013, 
    ARM_SYS = 0x00001F, 
    ARM_UND = 0x00001B, 
    ARM_USR = 0x000010
};  

// Francis TODO Why is this necessary for compile to work? Pragma issues?
class ARMGenericContext;

class ARMCPU : public CPU {
 protected:

    /*
     * The interrupt table and interrupt vector table that all
     * CPU's share. This is class data so it will work for both a
     * private and shared memory machine.
     */
//  static Interrupt * interruptTable[ nInterrupts ];
    Interrupt *interrupt(int n) {return _interruptManager->getInterrupt(n);}

    void sendFlushRequestTo( unsigned long dest ){}

    ARMMMU myMMU;

    void chipInitialize() {};

    /*
     * Called by the kernel to install all the interrupts.
     */
    void installInterrupts();

 public:
    ARMCPU( unsigned long identifier,
            ProcessContainer * initialIdleContainer );
    ~ARMCPU();

    /*
     * This is the C++ part of the assembler asisst
     * vector is the trap vector
     */
    static void hardwareInterruptAssist(int vector, ARMInterruptContext *context);

    /*
     * The subclasses define these to bind and lookup the binding of
     * interrupt (hardware) vectors to Interrupts.
     */
    Interrupt * setInterrupt( int vector, Interrupt * interrupt, PFV assist = 0 );

    /*
     * Overridden from superclass
     */
    static void disableInterrupts(IntStorage &); 
    static void disableInterrupts() {IntStorage junk; disableInterrupts(junk);}
    static void restoreInterrupts(const IntStorage &val);
    static void enableInterrupts();
    static int interruptsEnabled();
    static void initMutexState(MutexStateAddr addr);
    static void releaseMutex(MutexStateAddr addr){initMutexState(addr);}
    static void spinMutex(MutexStateAddr addr);
    static unsigned cpuID(){return 0;}


    static unsigned getMode();
    static unsigned getMode(unsigned cpsr);
    static void setMode(unsigned mode);
    static void displayMode();
    static void getRegisters(unsigned regs[17]);

    /*
     * Debugger trap routine
     */
    void debugBreakpoint();

    void reset();

    enum regnames {
        R0=0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,
        FP,IP,SP,LR,PC 
    };
};

extern "C" {
    void reset_cpu();
}

inline void
ARMCPU::reset()
{
    reset_cpu();
}

#ifdef __GNUC__
#include "ARMInterrupts.h"
#endif // __GNUC__

inline void
ARMCPU::initMutexState(MutexStateAddr addr)
{
    Assert ( addr != 0);
    *addr = 0;
}

// Note: A real spin is not required for a uniprocessor as
// a Lock disables interrupts - BusyWait doesn't
inline void
ARMCPU::spinMutex( MutexStateAddr addr)
{
    Assert ( addr != 0);

    asm volatile ( "spin:\n"
                   "mov r2, #1\n"
                   "swp r3, r2, [%0]\n"
                   "cmp r3, #1\n"
                   "beq spin" :: "r" (addr): "r2","r3");
    Assert ( *addr == 1 );
}

inline void
ARMCPU::restoreInterrupts(const IntStorage &prevstate)
{
    unsigned cpsr = (prevstate & 0x3) << 6;
    asm( "mrs r1, cpsr; \
          bic r1, r1, #0x80; \
          orr r1, r1, %0; \
          msr cpsr_cxsf, r1" ::"r"(cpsr): "r1");
}

inline int
ARMCPU::interruptsEnabled()
{
    unsigned int prevState;
    asm volatile ("mrs %0, cpsr" : "=r"(prevState));
    return (~(prevState >> 7) & 1);
}


inline void
ARMCPU::disableInterrupts(IntStorage &level)
{
    unsigned cpsr;
    
    asm volatile ("mrs %0, cpsr" : "=r"(cpsr));

    //irq & fiq are offset by 6 in cpsr
    level = (cpsr >> 6) & 0x3;

    asm( "mrs r1, cpsr; \
          orr r1, r1, #0x80; \
          msr cpsr_cxsf, r1" :::"r1");
}

inline void
ARMCPU::enableInterrupts()
{
    asm( "mrs r1, cpsr; \
          bic r1, r1, #0x80; \
          msr cpsr_cxsf, r1" ::: "r1");
}

// Warning: Note that after a mode switch, banked
// registers are used. Users of this method should
// be careful to correctly set banked register values
// before switching modes
inline void
ARMCPU::setMode(unsigned mode) {
    register unsigned cpsr = 0;
    asm( "mrs %0, cpsr": "=r"(cpsr) );

    cpsr &= ~0x1f;
    cpsr |= mode;

    asm("msr cpsr_cxsf, %0" ::"r"(cpsr));
}

inline unsigned
ARMCPU::getMode() {
    register unsigned cpsr = 0;
    asm( "mrs %0, cpsr": "=r"(cpsr) );
    
    cpsr &= 0x1f;

    return cpsr;
}

inline unsigned
ARMCPU::getMode(unsigned cpsr) {
    cpsr &= 0x1f;

    return cpsr;
}

#endif /* ARMCPU_h */
