/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * ARMContext.h: Process context for an ARM.
 *
 */

#ifndef ARMContext_h
#define ARMContext_h

#ifdef __GNUC__
#pragma interface
#endif

#include "MemoryObjectStar.h"
#include "ProcessorContext.h"   /* parent class */
#include "StackSize.h"
#include "Types.h"
#include "AsmBridge.h"
#include "OutputStream.h"

class ObjectProxy;
class Domain;

#define STACK_DEPTH 20

class ARMStackTrace {
 public:
    ARMStackTrace() {index = 0;}
    void addFrame(unsigned frame, unsigned pc, unsigned lr, unsigned fp) {
        if (index < STACK_DEPTH) {
            _frame[index] = frame;
            _pc[index] = pc;
            _lr[index] = lr;
            _fp[index] = fp;
            index++;
        }
    }
    void display() {
        for (unsigned i = 0; i < index; i++) {
            Console() << "Frame: " << asHex(_frame[i]) << "\n" << eor;
            Console() << "pc: " << asHex(_pc[i])
                      << " lr: " << asHex(_lr[i])
                      << " fp: " << asHex(_fp[i]) << "\n" << eor;
        }
    }
 private:
    unsigned _frame[STACK_DEPTH];
    unsigned _pc[STACK_DEPTH];
    unsigned _lr[STACK_DEPTH];
    unsigned _fp[STACK_DEPTH];

    unsigned index;
};

class ARMInterruptContext: public InterruptContext {
 private:
    volatile unsigned volatileRegs[16];     // 16 volatile regs
    volatile unsigned cpsr;                // CPSR should be initialized by child classes
 public:
    volatile unsigned *getRegisters() { return static_cast<volatile unsigned *>(volatileRegs); }

    void setMode(unsigned mode) { cpsr &= ~0x1f; cpsr |= mode; }
    unsigned getMode() { return cpsr &= 0x1f; }
    void setCPSR(unsigned value) { cpsr = value; }
    volatile unsigned getCPSR() { return cpsr; }
    void dumpRegisters();
    void dumpStackTrace();
    ARMStackTrace getStackTrace();
};

extern ClassStar ARMGenericContextClass;
class ARMGenericContext : public ProcessorContext {
 protected:
    /*
     * Instance data
     */
    volatile unsigned volatileRegs[16];     // 16 volatile regs
    volatile unsigned cpsr;                // CPSR should be initialized by child classes
    volatile unsigned returnAddress;       // Used to return to correct address after context switch

    /*
     * Stack related things
     */
    char * _stackPointer;

    char *  _kernelStackTop;

    char * _userStackTop;
    unsigned _userStackSize;

    volatile Process * checkpoint();
    Process* basicRestore( Process * );

    static void defaultReturn();

    void restoreFromInitialContext( Process * );
    
    /* 
     * Allocate a stack of size in domain, set the MO star
     */
    void allocateStack(StackSize size, Domain *domain,
                       char * &, unsigned &, MemoryObjectStar &, char * &); 

    ARMGenericContext( Process * owner, PFV entry, StackSize stackSize, int arg0, int arg1 );

    ARMGenericContext( Process * owner, PFV entry, StackSize stackSize, OPFV arg0, int arg1 );

    // This constructor should only be used by ARMBootContext
    // Using 1 as an argument so that the Assert doesn't fail in ProcessorContext
    ARMGenericContext():ProcessorContext(reinterpret_cast<Process *>(1)) {};

 public:
    ~ARMGenericContext();
    Class * classOf();

    static void doExit(void);

    void loadFromInterruptStack(unsigned *stack);
    void restoreToInterruptStack(unsigned *stack);
    void loadFromInterruptContext(ARMInterruptContext *context);
    void restoreToInterruptContext(ARMInterruptContext *context);

    volatile unsigned *getRegisters() { return static_cast<volatile unsigned *>(volatileRegs); }

    void setMode(unsigned mode) { cpsr &= ~0x1f; cpsr |= mode; }
    void setCPSR(unsigned value) { cpsr = value; }
    volatile unsigned getCPSR() { return cpsr; }

    virtual unsigned getInterruptStack() { Assert(NOTREACHED); return 0; }
};


// Francis David
// ARMBootContext is used when interrupts/interrupts occur during the Boot process
// before entering the idle loop
class ARMBootContext : public ARMGenericContext {
 public:
    ARMBootContext():ARMGenericContext() { }
    virtual unsigned getInterruptStack() { unsigned stack; asm( "stmia %0, {sp}^"::"r"(&stack) ); return stack; }
};

extern ClassStar ARMSystemContextClass;
class ARMSystemContext : public ARMGenericContext {
 protected:
 public:
    ARMSystemContext( Process * owner, PFV entryPoint,
                      StackSize stackSize, int arg0, int arg1 );
    ARMSystemContext( Process * owner, PFV entryPoint,
                      StackSize stackSize, OPFV arg0, int arg1 );
    ARMSystemContext( Process * owner, OPFV entryPoint,
                      StackSize stackSize, int arg0, int arg1 );
    Class * classOf();
    virtual unsigned getInterruptStack() { unsigned stack; asm( "stmia %0, {sp}^"::"r"(&stack) ); return stack; }
};

extern ClassStar ARMUninterruptableSystemContextClass;
class ARMUninterruptableSystemContext : public ARMSystemContext {
 protected:
 public:
    ARMUninterruptableSystemContext( Process * owner, PFV entryPoint,
                                     StackSize stackSize, int arg0, int arg1 = 0);
    Class * classOf();
    virtual unsigned getInterruptStack() { unsigned stack; asm( "stmia %0, {sp}^"::"r"(&stack) ); return stack; }
};


extern ClassStar ARMApplicationContextClass;
class ARMApplicationContext : public ARMGenericContext {
 protected:

    MemoryObjectStar userStackMO;
    char * userStackBase;

 public:
    ~ARMApplicationContext(){}
    ARMApplicationContext( Process * owner, PFV entryPoint, 
                           StackSize stackSize, Domain *initialDomain, 
                           int arg0 = 0, int arg1 = 0, StackSize = application, 
                           char *userStackPointer = 0);
    Class * classOf();
    virtual unsigned getInterruptStack() {
        // If in interrupt context already, return current sp
        unsigned spsr;
        asm( "mrs %0, spsr": "=r"(spsr) );
        if (spsr & 0x1f != 0x000010) {
            unsigned stack; asm( "stmia %0, {sp}^"::"r"(&stack) ); return stack;
        } else
            return (unsigned)kernelStackBase()+kernelStackSize(); 
    }
};

#endif /* ARMContext_h */
