/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef ARMMMU_h
#define ARMMMU_h


#ifdef __GNUC__
#pragma interface
#endif

#include "AddressTranslator.h"  /* parent class */

#ifdef OMAP      // ARM926EJ-S core
#define CSIZE 13 // Cache size = 1 << CSIZE (8K)
#define CLINE 5  // Cache line size in bytes = 1 << CLINE
#define NWAY  2  // Set associativity = 1 << NWAY (4 way)
#define I7SET 5  // Index of CP15:C7 set
#define I7WAY 30 // Index of CP15:C7 way
#endif

#ifdef INTEGRATOR// ARM1026EJ-S core
#define CSIZE 13 // Cache size = 1 << CSIZE (8K)
#define CLINE 5  // Cache line size in bytes = 1 << CLINE
#define NWAY  2  // Set associativity = 1 << NWAY (4 way)
#define I7SET 5  // Index of CP15:C7 set
#define I7WAY 30 // Index of CP15:C7 way
#endif

#define NSET  CSIZE-NWAY-CLINE   // Cache lines per way = 1 << NSET

class ARMMMU : public AddressTranslator {
 protected:
    unsigned _enabled;
 public:
    ARMMMU();
    void basicActivate( AddressTranslation *);
    void enable();
    int enabled();
    void setDisabled() { _enabled = 0; _currentDomain = 0; }

    inline void flushITLB() { asm("mcr p15,0,%0,c8,c5,0" :: "r"(0)); }
    inline void flushICache() { asm("mcr p15,0,%0,c7,c5,0" :: "r"(0)); }

    inline void flushITLBandCache() { flushITLB(); flushICache(); }

    inline void flushDTLB() { asm("mcr p15,0,%0,c8,c6,0" :: "r"(0)); }
    inline void flushDCache() { asm("mcr p15,0,%0,c7,c6,0" :: "r"(0)); }

    // Cleans the data cache. The cache is not invalidated
    inline void cleanDCache() {
        register unsigned c7f = 0;
        for (register unsigned way = 0; way < (1 << NWAY); way++) {
            for (register unsigned set = 0; set < (1 << NSET); set++) {
                asm("mcr p15,0,%0,c7,c10,2" :: "r"(c7f)); // Clean D-Cache line
                c7f += (1 << I7SET);
            }
            c7f += (1 << I7WAY);
        }
    }

    // Cleans and flushes the data cache
    inline void cleanAndFlushDCache() {
        register unsigned c7f = 0;
        for (register unsigned way = 0; way < (1 << NWAY); way++) {
            for (register unsigned set = 0; set < (1 << NSET); set++) {
                asm("mcr p15,0,%0,c7,c14,2" :: "r"(c7f)); // Clean and flush D-Cache line
                c7f += (1 << I7SET);
            }
            c7f += (1 << I7WAY);
        }
    }

    inline void flushDTLBandCache() { 
        flushDTLB();
#ifndef RELIABILITY
        flushDCache();
#endif
    }

    inline void flushTLBandCaches() { 
        flushITLBandCache(); 
        flushDTLB();
#ifndef RELIABILITY
        cleanAndFlushDCache();
#endif
    }

    // The semantics of this function are to flush the address translation cache - ie the TLB, not the actual data/instr caches
    inline void flushCache( const char* address, unsigned int length) { flushITLB(); flushDTLB(); }
};

#endif /* ARMMMU_h */
