/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * ARMPageTableEntry.h:  Page table entries for a ARM MMU.
 */

#ifndef ARMPageTableEntry_h
#define ARMPageTableEntry_h


#ifdef __GNUC__
#pragma interface
#endif

#include "AccessType.h"
#include "PageTableConstants.h"
#include "ProtectionLevel.h"
#include "ThisCPU.h"
 
/*
 * The representation of these classes must match exactly the hardware
 * representation of page table entries.  This means NO VIRTUAL FUNCTIONS!
 */
 
/*
 * Here is the ARM pointer table entry format for fine page tables:
 * (We are using coarse page tables)
 * +------------------------+-----+---+---+------------+---+--+
 * |Page Table/Frame Address| SBZ |Domain|   | SBZ |  PT Type |
 * |                        |     |      |   |     |          |     
 * |  xxxxxxxxxxxxxxxxxxxx  | 000 | x    | 1 |  0  |   11     |
 * +------------------------+-----+---+--+---+-----+----------+
 *  31                    12      8     5    3    2          0
 * SBZ = Should be zero
 */

class ARMPageTableEntry;

class ARMPointerTableEntry {
    volatile unsigned entry;
 public:
    ARMPointerTableEntry();
    
    ARMPointerTableEntry & operator=( ARMPointerTableEntry & pte ) {
        return *this;
    }
    
    void mapToPageTable( ARMPageTableEntry * pageTable ) {
        //map a page table to this pointer table entry
        // Francis - TODO fix, we're allowed to overwrite old values
        //Assert( !valid() ); 
        // Francis - Page table does not have to be page aligned
        //Assert( PageAligned( (const char *) pageTable ) );        
        
        //Coarse (4k) pages
        entry = ((unsigned) (pageTable)) | L1_V | L1_DOMAIN;
    }
    
    ARMPageTableEntry * pageTable() {
        return( (ARMPageTableEntry *) ( entry & 0xfffffc00) );
    }
    
    unsigned int valid() { 
        return( (entry & L1_V) == L1_V );
    }

    operator unsigned() {return entry;}
 public:
    unsigned value() {return entry;}
    
    void setProtection( ProtectionLevel protection );
};

class ARMPageTableEntry {
    volatile unsigned entry;
    ARMPageTableEntry();
    
 public:
    ARMPageTableEntry( unsigned pte );

    unsigned operator=( unsigned pte );
    ARMPageTableEntry & operator=( ARMPageTableEntry & pte );

    void mapToAddress( const char * address, ProtectionLevel protection );
    char * addressMappedTo();

    operator unsigned() {return entry;}

    void unmap();
    unsigned valid();
    unsigned referenced();
    unsigned modified();
    unsigned value();

    void setProtection( ProtectionLevel protection );
    int okToPerform( AccessType attemptedAccess );

    void inhibitDataCaching();
};

inline
unsigned
ARMPageTableEntry::value()
{
    return entry ;
}

inline
ARMPageTableEntry::ARMPageTableEntry( unsigned pte )
{
    entry = pte;
    Assert(NOTREACHED);
}

inline unsigned int
ARMPageTableEntry::operator=( unsigned int pte )
{
    entry = pte;
    Assert(NOTREACHED);
    return( entry );
}

inline ARMPageTableEntry &
ARMPageTableEntry::operator= ( ARMPageTableEntry & pte )
{
    entry = pte.entry;
    Assert(NOTREACHED);
    return( *this );
}

inline unsigned int
ARMPageTableEntry::valid()
{
    return( (entry & L2_V) == L2_V);
}

inline void
ARMPageTableEntry::mapToAddress( const char * addr, ProtectionLevel protection )
{
    Assert( !valid() );
    Assert( PageAligned( (const char *) addr ) );
    entry = ((unsigned int)addr & 0xfffff000);
    setProtection( protection );
}

inline char *
ARMPageTableEntry::addressMappedTo()
{
    return( (char *) ( entry & 0xfffff000 ) );
}

inline void
ARMPageTableEntry::unmap()
{
    Assert( valid() );
    entry &= ~L2_V;
}

inline unsigned int
ARMPageTableEntry::referenced()
{
    // Not available on ARM
    return 1;
}

inline unsigned int
ARMPageTableEntry::modified()
{
    // Not available on ARM
    return 1;
}

inline void
ARMPageTableEntry::inhibitDataCaching()
{
    // Caching disabled, write buffer enabled
    entry &= ~L2_C;
}

inline void
ARMPageTableEntry::setProtection( ProtectionLevel protection )
{
    ARMProtection realProt;
    switch( protection ) {
        case ReadWrite:
            realProt = USERRW;
            break;
        case ReadOnly:
            realProt = USERRO;
            break;
        case NoAccess:
            realProt = SYSTEMRW;
            break;
        default:
            realProt = SYSTEMRO;    // shut up warning.
            Assert( NOTREACHED );
    }

    // small page
    entry = (entry & 0xfffff000) | (realProt << 4) | (realProt << 6) | (realProt << 8) | (realProt << 10) | L2_V;

    // enable write-back caching
    entry |= (L2_C | L2_B);
    ((ARMMMU *)thisCPU()->MMU())->cleanDCache();
}

inline int
ARMPageTableEntry::okToPerform( AccessType attemptedAccess )
{
    Assert(NOTREACHED);
    return 0;
}

#endif /* ARMPageTableEntry_h */

