/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MC68030Context.h: Process context for a MC68030.
 */

#ifndef MC68030Context_h
#define MC68030Context_h


#ifdef __GNUC__
#pragma interface
#endif

#include "ProcessorContext.h"	/* parent class */
#include "StackSize.h"

class ObjectProxy;
class Domain;


/*
 * Interrupt basicRaise() methods get a pointer to this as their
 * "info" argument.  The offsets of fields in this struct MUST match
 * those defined (for now) in MC68030ContextOffsets.h.
 */
struct interruptStackFrame {
	/*
	 * The following registers are saved by the interruptAssist().
	 * These are the registers that the compiler expects to have
	 * clobbered across a procedure call.
	 */
	int d0;
	int d1;
	int a0;
	int a1;

	/*
	 * The following fields are those common to all interrupt stack
	 * frames.  The hardware automatically pushes them on the
	 * supervisor stack when an interrupt occurs.
	 */
	unsigned short sr;
	unsigned int pc;
	unsigned short vector;

	/*
	 * The rest of these fields are valid only when the interrupt
	 * is a bus error.  They are also pushed on the supervisor
	 * stack automatically when the interrupt is taken.
	 */
	unsigned short internalRegister;
	unsigned short specialStatusRegister;
	unsigned short pipeStageC;
	unsigned short pipeStageB;
	char * faultAddress;
};


/*
 * The layout of this class MUST match exactly the offsets defined in
 * MC68030ContextOffsets.h.
 */
extern ClassStar MC68030GenericContextClass;
class MC68030GenericContext : public ProcessorContext {
friend ProcessorContext * thisProcessorContext();
protected:
	/*
	 * The supervisor stack base and bound are used to check for
	 * stack overflow in thisProcessorContext(); they are not used
	 * in any assembly code.
	 */
        char * supervisorStackPointer;
	char * supervisorStackBase;
	char * supervisorStackBound;

	/*
	 * Register save area.  The supervisor stack pointer is not
	 * included here; neither are compiler-volatile registers.
	 * For the MC68030, the save area includes d2-d7 and a2-a6.
	 */
	int d2;
	int d3;
	int d4;
	int d5;
	int d6;
	int d7;
	int a2;
	int a3;
	int a4;
	int a5;
	int a6;

	/*
	 * Finally, the program counter.
	 */
	int pc;

	/*
	 * Generic checkpoint/restore methods.
	 */
	virtual Process * checkpoint();
	virtual void basicRestore( Process * process );

	/*
	 * Interrupt management methods.
	 */
	virtual void becomeInterruptable();
	virtual int becomeUninterruptable();
	virtual int isInterruptable();
public:
	MC68030GenericContext( Process * owner, StackSize stackSize );
	~MC68030GenericContext();

        Class * classOf();
};


class InitialApplicationContextStackDescription {
	friend class MC68030ApplicationContext;
protected:
	/*
	 * The following are initialized by the MC68030ApplicationContext
	 * constructor and used by
	 * MC68030ApplicationContext::restoreFromInitialContext(). 
	 */
	int arg0;
	int arg1;
	const char * userStackPointer;

	/*
	 * These three form a stack frame suitable for a rte instruction
	 * and must remain in this order.
	 */
	unsigned short sr;
	PFV entryPoint;
	unsigned short format;
};

extern ClassStar MC68030ApplicationContextClass;
class MC68030ApplicationContext : public MC68030GenericContext {
friend unsigned int GateCross( interruptStackFrame * );	// TEMPORARY
protected:

	/*
	 * Additional registers saved by an application context.
	 * Floating point registers should also go here.
	 */
	char * userStackPointer;

	/*
	 * Overloaded to checkpoint/restore the additional registers above.
	 */
	virtual Process * checkpoint();
	virtual void basicRestore( Process * );

	void restoreFromInitialContext( Process * );
public:
	MC68030ApplicationContext( Process * owner, PFV entryPoint,
		StackSize stackSize, Domain *initialDomain,
		ObjectProxy *defaultProxy, int arg0, int arg1,
		StackSize userStackSize, char * userStackPointer = 0,
		int usesProxies = 1 );

	virtual Class * classOf();
};

extern ClassStar MC68030SystemContextClass;
class MC68030SystemContext : public MC68030GenericContext {
friend void idleProcessInitialRestoreAssist( ProcessorContext *, Process * );
protected:
	void restoreFromInitialContext( Process * );
	void initialRestoreAssist( Process * );
public:
	MC68030SystemContext( Process * owner, PFV entryPoint,
		StackSize stackSize, int arg0, int arg1 );

	virtual Class * classOf();
};

extern ClassStar MC68030UninterruptableSystemContextClass;
class MC68030UninterruptableSystemContext : public MC68030SystemContext {
public:
	MC68030UninterruptableSystemContext( Process * owner, PFV entryPoint,
		StackSize stackSize, int arg0, int arg1 );

	virtual Class * classOf();
};

#endif /* MC68030Context_h */
