/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MC68030Lock.h - low-level spin lock.
 */

#ifndef MC68030Lock_h
#define MC68030Lock_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Assert.h"
#include "PROCESSOR_ID.h"


class MC68030Lock {
protected:
	union LockUnion {
		struct {
			volatile unsigned char state;
			volatile unsigned int sr : 3;
			volatile unsigned int id : 5;
		} parts;
		short all;
	} data;

public:
	MC68030Lock();
	MC68030Lock( char * );
	void acquire();
	void release();

	/*
	 * The use of any of these will probably causes races...beware.
	 */
	int held();
	unsigned char heldBy();
	int heldByMe();
	int heldByAnother();
};

inline
MC68030Lock::MC68030Lock()
{
	data.parts.state = 0;
	data.parts.sr = 0;
	data.parts.id = -1;
}

inline
MC68030Lock::MC68030Lock( char * name )
{
	data.parts.state = 0;
	data.parts.sr = 0;
	data.parts.id = -1;
}

inline int
MC68030Lock::held()
{
	return( data.parts.state != 0 );
}

inline unsigned char
MC68030Lock::heldBy()
{
	return( data.parts.id );
}

inline int
MC68030Lock::heldByMe()
{
	return( data.parts.id == PerMemoryProcessorID() );
}

inline int
MC68030Lock::heldByAnother()
{
	return( held() && ( ! heldByMe() ) );
}

inline void
MC68030Lock::acquire()
{
	/*
	 * Save the current interrupt state, then disable interrupts.
	 */
	unsigned short SR;
	asm( "movew %%sr,%0;\
	      oriw  #0x0700,%%sr" : "=g" (SR) );
	
	Assert( ! heldByMe() );

	/*
	 * Now acquire the lock.  If we don't get the lock the first time,
	 * we spin only reading until it is released.  (Each instruction
	 * is 2 bytes long.)
	 */
	asm( "tas  %0@;\
	      beq  .+8;\
	      tstb %0@;\
	      bne  .-2;\
	      beq  .-8" : : "a" (this) );

	/*
	 * Now that we hold the lock, update its info with the holder
	 * and his previous interrupt state.
	 */
	data.parts.id = PerMemoryProcessorID();
	data.parts.sr = (SR >> 8);
}

inline void
MC68030Lock::release()
{
	unsigned short SR;

#ifdef ASSERT
	Assert( heldByMe() );

	/*
	 * Make sure interrupts are still disabled.
	 */
	asm( "movew %%sr,%0" : "=g" (SR) );
	Assert( (SR & 0x0700) == 0x0700 );
#endif

	/*
	 * Release the lock, making sure to save the sr beforehand.
	 * (Otherwise, someone else could clobber it before we
	 * restore it below.)
	 */
	unsigned short sr = data.parts.sr;
	data.parts.sr = 0;
	data.parts.id = -1;
	data.parts.state = 0;	// Poof!

	/*
	 * Restore interrupt mask to its value prior to acquiring the lock.
	 * Interrupts are still disabled until the last instruction is
	 * executed.
	 */
	asm( "movew %%sr,%0" : "=g" (SR) );
	SR &= ~0x0700;
	SR |= (sr << 8);
	asm( "movew %0,%%sr" : : "g" (SR) );
}



class MC68030BusyWait {
protected:
	volatile unsigned char state;
public:
	MC68030BusyWait();
	void wait();
	void releaseNextWaiter();
};

inline
MC68030BusyWait::MC68030BusyWait()
{
	state = 0;
}

inline void
MC68030BusyWait::wait()
{
	/*
	 * Do the spin part of a spin lock, but don't worry about interrupts.
	 */
	asm( "tas  %0@;\
	      beq  .+8;\
	      tstb %0@;\
	      bne  .-2;\
	      beq  .-8" : : "a" (this) );
}

inline void
MC68030BusyWait::releaseNextWaiter()
{
	state = 0;
}

#endif /* MC68030Lock_h */
