/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MC68030PageTableEntry.h:  Page table entries for a MC68030 MMU.
 */

#ifndef MC68030PageTableEntry_h
#define MC68030PageTableEntry_h


#ifdef __GNUC__
#pragma interface
#endif

#include "AccessType.h"
#include "PageTableConstants.h"
#include "ProtectionLevel.h"


/*
 * The representation of these classes must match exactly the hardware
 * representation of page table entries.  This means NO VIRTUAL FUNCTIONS!
 */

/*
 * We use only the following descriptors: 
 *
 * Short Format Table Descriptor:
 *
 * +--------------------------------------------------+-----------------+-----+
 * |               Page Table Address                 |Reference|Protect|Valid|
 * |                                                  |   (U)   | (WP)  |(DT) |
 * |           xxxxxxxxxxxxxxxxxxxxxxxxxxxx           |    x    |   x   | x x |
 * +--------------------------------------------------+---------+-------+-----+
 *  31                                               4     3        2     1 0
 *
 * Short Format Page Descriptor:
 *
 * +------------------------+-+------------+-+--------+---------+-------+-----+
 * |   Page Frame Address   | |CacheInhibit| |Modified|Reference|Protect|Valid|
 * |                        | |            | |        |         |       |     |
 * |xxxxxxxxxxxxxxxxxxxxxxxx|0|     x      |0|    x   |    x    |   x   | x x |
 * +------------------------+-+------------+-+--------+---------+-------+-----+
 *  31                     8        6             4        3        2     1 0
 *
 *
 * Here are the names of the fields and the values we're interested in:
 *
 *	DT   Descriptor type.
 *		00 - Invalid
 *		01 - Page Descriptor (Descriptor points to page)
 *		10 - Valid short format (Descriptor points to page table)
 *
 *	WP   Write Protect bit
 */

class MC68030PageTableEntry;

class MC68030PointerTableEntry {
	unsigned int entry;
public:
	MC68030PointerTableEntry();

	MC68030PointerTableEntry & operator=( MC68030PointerTableEntry & pte );

	void mapToPageTable( MC68030PageTableEntry * pageTable );
	MC68030PageTableEntry * pageTable();

	unsigned int valid();

	void setProtection( ProtectionLevel protection );
};


class MC68030PageTableEntry {
	unsigned int entry;
public:
	MC68030PageTableEntry();
	MC68030PageTableEntry( unsigned int pte );

	unsigned int operator=( unsigned int pte );
	MC68030PageTableEntry & operator=( MC68030PageTableEntry & pte );

	void mapToAddress( const char * address, ProtectionLevel protection );
	const char * addressMappedTo();
	void unmap();

	unsigned int valid();
	unsigned int referenced();
	unsigned int modified();

	void setProtection( ProtectionLevel protection );
	int okToPerform( AccessType attemptedAccess );

	void inhibitDataCaching();
};



inline
MC68030PointerTableEntry::MC68030PointerTableEntry()
{
	entry = 0;
}

inline MC68030PointerTableEntry &
MC68030PointerTableEntry::operator= ( MC68030PointerTableEntry & pte )
{
	entry = pte.entry;
	return( *this );
}

inline void
MC68030PointerTableEntry::mapToPageTable( MC68030PageTableEntry * pageTable )
{
	Assert( !valid() );
	Assert( PageAligned( (const char *) pageTable ) );

	entry = (unsigned int) pageTable | 0x2;
}

inline MC68030PageTableEntry *
MC68030PointerTableEntry::pageTable()
{
	return( (MC68030PageTableEntry *) (entry & 0xfffffff0) );
}

inline unsigned int
MC68030PointerTableEntry::valid()
{
	return( (entry & 0x3) == 0x2 );
}

inline void
MC68030PointerTableEntry::setProtection( ProtectionLevel protection )
{
	Assert( valid() );

	/*
	 * MC68030 translation table entries have only a single bit of
	 * protection information, so someone else has to worry about
	 * making NoAccess a meaningful protection level.
	 */
	MC68030Protection realProt = READONLY;
	switch( protection ) {
		case ReadWrite:
			realProt = READWRITE;
			break;
		case ReadOnly:
			realProt = READONLY;
			break;
		default:
			Assert( NOTREACHED );
	}

	entry &= ~0x4;
	entry |= realProt << 2;
}



inline
MC68030PageTableEntry::MC68030PageTableEntry()
{
	entry = 0;
}

inline
MC68030PageTableEntry::MC68030PageTableEntry( unsigned int pte )
{
	entry = pte;
}

inline unsigned int
MC68030PageTableEntry::operator=( unsigned int pte )
{
	entry = pte;
	return( entry );
}

inline MC68030PageTableEntry &
MC68030PageTableEntry::operator= ( MC68030PageTableEntry & pte )
{
	entry = pte.entry;
	return( *this );
}

inline void
MC68030PageTableEntry::mapToAddress( const char * addr,
	ProtectionLevel protection )
{
	Assert( !valid() );
	Assert( PageAligned( (const char *) addr ) );

	/*
	 * MC68030 translation table entries have only a single bit of
	 * protection information, so someone else has to worry about
	 * making NoAccess a meaningful protection level.
	 */
	MC68030Protection realProt = READONLY;
	switch( protection ) {
		case ReadWrite:
			realProt = READWRITE;
			break;
		case ReadOnly:
			realProt = READONLY;
			break;
		default:
			Assert( NOTREACHED );
	}

	entry = (unsigned int) addr | (realProt << 2) | 0x1;
}

inline const char *
MC68030PageTableEntry::addressMappedTo()
{
	return( (const char *) ( entry & 0xffffff00 ) );
}

inline void
MC68030PageTableEntry::unmap()
{
	Assert( valid() );
	entry &= 0xfffffffc;
}

inline unsigned int
MC68030PageTableEntry::valid()
{
	return( (entry & 0x3) == 0x1 );
}

inline unsigned int
MC68030PageTableEntry::referenced()
{
	return( (entry & 0x8) == 0x8 );
}

inline unsigned int
MC68030PageTableEntry::modified()
{
	return( (entry & 0x10) == 0x10 );
}

inline void
MC68030PageTableEntry::setProtection( ProtectionLevel protection )
{
	Assert( valid() );

	/*
	 * MC68030 translation table entries have only a single bit of
	 * protection information, so someone else has to worry about
	 * making NoAccess a meaningful protection level.
	 */
	MC68030Protection realProt = READONLY;
	switch( protection ) {
		case ReadWrite:
			realProt = READWRITE;
			break;
		case ReadOnly:
			realProt = READONLY;
			break;
		default:
			Assert( NOTREACHED );
	}

	entry &= ~0x4;
	entry |= realProt << 2;
}

inline int
MC68030PageTableEntry::okToPerform( AccessType attemptedAccess )
{
	int currentProtection = (entry & 0x4) >> 2;

	MC68030Protection neededProtection = READWRITE;
	switch( attemptedAccess ) {
		case ApplicationWrite:
		case SystemWrite:
			neededProtection = READWRITE;
			break;
		case ApplicationRead:
		case SystemRead:
			neededProtection = READONLY;
			break;
		default:
			Assert( NOTREACHED );
	}

	if( neededProtection == READWRITE && currentProtection == READONLY ) {
		return( 0 );
	} else {
		return( 1 );
	}
}

inline void
MC68030PageTableEntry::inhibitDataCaching()
{
	entry |= 0x40;
}

#endif /* MC68030PageTableEntry_h */
