/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * NS32332CPU.h: per-cpu private information class for the NS32332 processor.
 */

#ifndef NS32332CPU_h
#define NS32332CPU_h


#ifdef __GNUC__
#pragma interface
#endif

#include "CPU.h"
#include "NS32382MMU.h"
#include "Vectors.h"

/*
 * Hardware interrupt vector table entry.
 */
struct InterruptDescriptor {
	unsigned short module;
	unsigned short offset;
};

/*
 * 16 traps plus 256 vectored interrupts?
 * Actually I thought it is 256-16 vectored interrupts. We need to check this.
 */
static const int nInterrupts = 16 + 256;

class NS32332CPU : public CPU {
protected:
	/*
	 * The interrupt table and interrupt vector table that all
	 * CPU's share. This is class data so it will work for both a
	 * private and shared memory machine.
	 */
	static Interrupt * interruptTable[ nInterrupts ];
	static InterruptDescriptor interruptVectorTable[ nInterrupts ];

	/*
	 * These are assembler assists for interrupt/trap handling.
	 */
	static void proxyAssist();
	static void proxyReturnAssist();
	static void svcAssist();	
	static void trapAssist();
	static void interruptAssist();

	/*
	 * Per-Instance stuff.
	 */
	NS32382MMU myMMU;

	int bootSanityCheck();

	void chipInitialize();

	/*
	 * Called by the kernel to install all the interrupts.
	 */
	void installInterrupts();

	/*
	 * Helpers for trap/interrupt processing.
	 */
	static int getContextSaveHint( Interrupt *, char * );

public:
	NS32332CPU( unsigned long identifier,
		ProcessContainer * initialIdleContainer );
	~NS32332CPU();

	/*
	 * The subclasses define these to bind and lookup the binding of
	 * interrupt (hardware) vectors to Interrupts.
	 */
	Interrupt * setInterrupt( int vector, Interrupt * interrupt,
				  PFV assist = 0);
	Interrupt * interrupt( int vector );

	/*
	 * Preempt the processor.
	 */
	void preempt();

	
	/*
         * Overridden from superclass
         */
        static int disableInterrupts();
        static void restoreInterrupts(int val);
        static void enableInterrupts(){NS32332CPU::restoreInterrupts(1);}
        static void spinMutex(MutexStateAddr addr);
        static void initMutexState(MutexStateAddr addr);
        static void releaseMutex(MutexStateAddr addr);
        static unsigned cpuID();
        static int interruptsEnabled();

};

#ifdef NEVER
inline
void NS32332CPU::initMutexState(MutexStateAddr addr)
{
	Assert (addr != 0);
	*addr = 0;
}

inline
void NS32332CPU::releaseMutex(MutexStateAddr addr)
{
	Assert (addr != 0);
	*addr = 0;
}

inline
int
NS32332CPU::disableInterrupts()
{
	unsigned short psr = 0;
	asm volatile( "sprw psr,%0" : "=g" (psr) );	// save old state
	asm volatile( "bicpsrw $0x800" );		// disable
	return( (psr & 0x800) == 0x800 );
}

inline
void
NS32332CPU::restoreInterrupts(int val)
{
	Assert (val == 0 || val == 1);
	if (val) {
		asm volatile("bispsrw $0x800");
		}
}

inline
int
NS32332CPU::interruptsEnabled()
{
	unsigned short psr = 0;
	asm volatile( "sprw psr,%0" : "=g" (psr) );
	return( (psr & 0x800) == 0x800 );
}

inline 
void
NS32332CPU::spinMutex(MutexStateAddr addr)
{
	Assert (addr != 0);
	asm(   "sbitib  $0,0(%0); \
                bfc     .+10; \
                tbitb   $0,0(%0); \
                bfc     .-11; \
                br      .-6" : : "r" (addr) );
}

inline
unsigned
NS32332CPU::cpuID()
{
	unsigned result = *(volatile unsigned char *) 0xffffff54;
	return result;
}
#endif


#endif /* NS32332CPU_h */
