/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Contributing Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * NS32382PageTableEntry.h:  Page table entries for a NS32382 MMU.
 */

#ifndef NS32382PageTableEntry_h
#define NS32382PageTableEntry_h


#ifdef __GNUC__
#pragma interface
#endif

#include "AccessType.h"
#include "PageTableConstants.h"
#include "ProtectionLevel.h"


/*
 * The representation of these classes must match exactly the hardware
 * representation of page table entries.  This means NO VIRTUAL FUNCTIONS!
 */

/*
 * Here is the NS32382 descriptor format:
 *
 * +------------------------+---+---+---+------------+---+-------+-----+
 * |Page Table/Frame Address|   |Mod|Ref|CacheInhibit|   |Protect|Valid|
 * |                        |   |   |   |            |   |       |     |
 * |  xxxxxxxxxxxxxxxxxxxx  |000| x | x |      x     |000|  x  x |  x  |
 * +------------------------+---+---+---+------------+---+-------+-----+
 *  31                    12      8   7        6            2  1    0
 *
 */


class NS32382PageTableEntry;

class NS32382PointerTableEntry {
	unsigned int entry;
public:
	NS32382PointerTableEntry();

	NS32382PointerTableEntry & operator=( NS32382PointerTableEntry & pte );

	void mapToPageTable( NS32382PageTableEntry * pageTable );
	NS32382PageTableEntry * pageTable();

	unsigned int valid();

	void setProtection( ProtectionLevel protection );
};

inline 
unsigned int
NS32382PointerTableEntry::valid()
{
	return( (entry & 0x1) == 0x1 );
}


class NS32382PageTableEntry {
	unsigned int entry;
public:
	NS32382PageTableEntry();
	NS32382PageTableEntry( unsigned int pte );

	unsigned int operator=( unsigned int pte );
	NS32382PageTableEntry & operator=( NS32382PageTableEntry & pte );

	void mapToAddress( const char * address, ProtectionLevel protection );
	const char * addressMappedTo();
	void unmap();

	unsigned int valid();
	unsigned int referenced();
	unsigned int modified();

	void setProtection( ProtectionLevel protection );
	int okToPerform( AccessType attemptedAccess );

	void inhibitDataCaching();
};

inline 
unsigned int
NS32382PageTableEntry::valid()
{
	return( (entry & 0x1) == 0x1 );
}



inline
NS32382PointerTableEntry::NS32382PointerTableEntry()
{
	entry = 0;
}

inline NS32382PointerTableEntry &
NS32382PointerTableEntry::operator= ( NS32382PointerTableEntry & pte )
{
	entry = pte.entry;
	return( *this );
}

inline void
NS32382PointerTableEntry::mapToPageTable( NS32382PageTableEntry * pageTable )
{
	Assert( !valid() );
	Assert( PageAligned( (const char *) pageTable ) );

	entry = (unsigned int) pageTable | (USERRW << 1) | 0x1;
}

inline NS32382PageTableEntry *
NS32382PointerTableEntry::pageTable()
{
	return( (NS32382PageTableEntry *) (entry & 0xfffff000) );
}

inline void
NS32382PointerTableEntry::setProtection( ProtectionLevel protection )
{
	Assert( protection < Pte_PROTECT_LEVELS );
	Assert( valid() );

	NS32382Protection realProt;

	/*
	 * Convert protection to the corresponding NS32382 level.
	 */
	if( protection == ReadWrite ) {
		realProt = USERRW;
	}
	else if( protection == ReadOnly ) {
		realProt = USERRO;
	}
	else if( protection == NoAccess ) {
		realProt = SYSTEMRW;
	}
	else {
		realProt = SYSTEMRO;	// shut up compiler warning.
		Assert( NOTREACHED );
	}

	/*
	 * Set the protection.
	 */
	entry &= ~0x6;
	entry |= realProt << 1;
}



inline
NS32382PageTableEntry::NS32382PageTableEntry()
{
	entry = 0;
}

inline
NS32382PageTableEntry::NS32382PageTableEntry( unsigned int pte )
{
	entry = pte;
}

inline unsigned int
NS32382PageTableEntry::operator=( unsigned int pte )
{
	entry = pte;
	return( entry );
}

inline NS32382PageTableEntry &
NS32382PageTableEntry::operator= ( NS32382PageTableEntry & pte )
{
	entry = pte.entry;
	return( *this );
}

inline void
NS32382PageTableEntry::mapToAddress( const char * addr,
	ProtectionLevel protection )
{
	Assert( !valid() );
	Assert( PageAligned( (const char *) addr ) );

	entry = (unsigned int) addr | 0x1;

	NS32382Protection realProt;
	switch( protection ) {
		case ReadWrite:
			realProt = USERRW;
			break;
		case ReadOnly:
			realProt = USERRO;
			break;
		case NoAccess:
			realProt = SYSTEMRW;
			break;
		default:
			realProt = SYSTEMRO;	// shut up warning.
			Assert( NOTREACHED );
	}

	entry = (unsigned int) addr | ( realProt << 1 ) | 0x1;
}

inline const char *
NS32382PageTableEntry::addressMappedTo()
{
	return( (const char *) ( entry & 0xfffff000 ) );
}

inline void
NS32382PageTableEntry::unmap()
{
	Assert( valid() );
	entry &= ~0x1;
}

inline unsigned int
NS32382PageTableEntry::referenced()
{
	return( (entry & 0x80) == 0x80 );
}

inline unsigned int
NS32382PageTableEntry::modified()
{
	return( (entry & 0x100) == 0x100 );
}

inline void
NS32382PageTableEntry::setProtection( ProtectionLevel protection )
{
	Assert( protection < Pte_PROTECT_LEVELS );
	Assert( valid() );

	NS32382Protection realProt;

	/*
	 * Convert protection to the corresponding NS32382 level.
	 */
	if( protection == ReadWrite ) {
		realProt = USERRW;
	}
	else if( protection == ReadOnly ) {
		realProt = USERRO;
	}
	else if( protection == NoAccess ) {
		realProt = SYSTEMRW;
	}
	else {
		realProt = SYSTEMRO;	// shut up compiler warning.
		Assert( NOTREACHED );
	}

	/*
	 * Set the protection.
	 */
	entry &= ~0x6;
	entry |= realProt << 1;
}

inline int
NS32382PageTableEntry::okToPerform( AccessType attemptedAccess )
{
	NS32382Protection currentProtection;
	switch ((entry & 0x6) >> 1) {
		default:
			Assert(NOTREACHED)("Invalid page protection");
		case USERRO:
			currentProtection = USERRO;
			break;
		case USERRW:
			currentProtection = USERRW;
			break;
		case SYSTEMRW:
			currentProtection = SYSTEMRW;
			break;
		case SYSTEMRO:
			currentProtection = SYSTEMRO;
			break;
		};

	NS32382Protection neededProtection;
	switch( attemptedAccess ) {
		case ApplicationWrite:
			neededProtection = USERRW;
			break;
		case ApplicationRead:
			neededProtection = USERRO;
			break;
		case SystemWrite:
			neededProtection = SYSTEMRW;
			break;
		case SystemRead:
			neededProtection = SYSTEMRO;
			break;
		default:
			neededProtection = SYSTEMRO;	// shut up warning.
			Assert( NOTREACHED );
	}

	return( currentProtection >= neededProtection );
}

inline void
NS32382PageTableEntry::inhibitDataCaching()
{
	entry |= 0x40;
}

#endif /* NS32382PageTableEntry_h */
