/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David Raila (raila@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef SPARCCPU_h
#define SPARCCPU_h


#ifdef __GNUC__
#pragma interface
#endif

#include "MachineTypes.h"
#include "CPU.h"
#include "SPARCMMU.h"

typedef void (*PFV) ();

const int nInterrupts = 256;	// Number of interrupt in table


class SPARCCPU : public CPU {
protected:

	/*
	 * The interrupt table and interrupt vector table that all
	 * CPU's share. This is class data so it will work for both a
	 * private and shared memory machine.
	 */
	static Interrupt * interruptTable[ nInterrupts ];
	Interrupt *interrupt(int n) {return interruptTable[n];}

	void sendFlushRequestTo( unsigned long dest ){}

	/*
	 * This is the C++ part of the assembler asisst
	 * vector is the trap vector
	 */
	static void hardwareInterruptAssist(int vector, char *stackptr);

	SPARCMMU myMMU;

	void chipInitialize() {};

	/*
	 * Called by the kernel to install all the interrupts.
	 */
	void installInterrupts();

public:
	SPARCCPU( unsigned long identifier,
		ProcessContainer * initialIdleContainer );
	~SPARCCPU();

	/*
	 * The subclasses define these to bind and lookup the binding of
	 * interrupt (hardware) vectors to Interrupts.
	 */
	Interrupt * setInterrupt( int vector, Interrupt * interrupt,
				 PFV assist = 0 );

	/*
	 * Overridden from superclass
	 */
	static void disableInterrupts(IntStorage &); 
	static void disableInterrupts() {IntStorage junk; disableInterrupts(junk);}
	static void restoreInterrupts(const IntStorage &val);
	static void enableInterrupts(){IntStorage arg = 0; restoreInterrupts(arg);}
	static int interruptsEnabled();
	static void initMutexState(MutexStateAddr addr);
	static void releaseMutex(MutexStateAddr addr){initMutexState(addr);}
	static void spinMutex(MutexStateAddr addr);
	static unsigned cpuID(){return 0;}

	/*
	 * Debugger trap routine
	 */
	void debugBreakpoint();
};


#ifdef __GNUG__

#include "SPARCInterrupts.h"
#include <sys/psw.h>


const IntStorage HIGH_LEVEL = INTERRUPT_LOCK;

inline unsigned char
interruptLevel()
{
	register unsigned psr;
	asm volatile ("mov %%psr,%0" : "=r" (psr) );
	return ((psr & PSR_PIL) >> 8);
}

inline int
SPARCCPU::interruptsEnabled()
{
	return( interruptLevel() < INTERRUPT_HIGH );
}

inline void
SPARCCPU::restoreInterrupts(const IntStorage &prevstate)
{
	register unsigned psr;
	unsigned retval;
	Assert(prevstate < 16);

	asm volatile ("mov %%psr,%0" : "=r" (psr) );
	/*
	 * Disable traps, just in case this is interrupted since
	 * there is a bug in raising interrupts and enabling traps
	 * in the SPARC IU, where traps are enabled first and the
	 * interrupt is taken at the leve we're masking.
	 */
	psr &= ~PSR_ET;
	asm volatile ("mov %0,%%psr" : : "r" (psr));
	//asm volatile("nop;nop;nop");

	retval = (psr & PSR_PIL) >> 8;
	psr &= ~PSR_PIL;	/* mask out old level */
	psr |= ( (prevstate << 8) & PSR_PIL);
	asm volatile ("mov %0,%%psr" : : "r" (psr));
	asm volatile("nop;nop;nop");

	/*
	 * now that the level is set for sure we can re-enable
	 * traps
	 */
	psr |= PSR_ET;
	asm volatile ("mov %0,%%psr" : : "r" (psr));
	asm volatile("nop;nop;nop");
}

inline void
SPARCCPU::disableInterrupts(IntStorage &level)
{
	level = interruptLevel();
	if (level < HIGH_LEVEL)
		restoreInterrupts(HIGH_LEVEL);
}

#endif __GNUG__

inline void
SPARCCPU::initMutexState(MutexStateAddr addr)
{
	Assert (addr != 0);
	*addr = 0;
}

inline void
SPARCCPU::spinMutex( MutexStateAddr addr)
{
	Assert ( addr != 0);
	Assert ( *addr == 0)("Mutex already held!: value %d\n", *addr);
	*addr = 1;
}

#endif /* SPARCCPU_h */
