/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * SPARCPageTableEntry.h:  Page table entries for a SPARC MMU.
 */

#ifndef SPARCPageTableEntry_h
#define SPARCPageTableEntry_h


#ifdef __GNUC__
#pragma interface
#endif


#include "AccessType.h"
#include "PageTableConstants.h"
#include "ProtectionLevel.h"

/*
 * The representation of these classes must match exactly the hardware
 * representation of page table entries.  This means NO VIRTUAL FUNCTIONS!
 */

/*
 * Not really a problem for the SPARC, since there are not really any tables
 * in memory.  They reside in SRAM in hte MMU.
 */

class SPARCPageTableEntry {
	unsigned int entry;
public:
	SPARCPageTableEntry();
	SPARCPageTableEntry( unsigned int pte );

	unsigned int operator=( unsigned int pte );
	SPARCPageTableEntry & operator=( SPARCPageTableEntry & pte );

	void mapToAddress( const char * address, ProtectionLevel protection );
	char * addressMappedTo();

	operator unsigned() {return entry;}

	void unmap();
	unsigned int valid();
	unsigned int referenced();
	unsigned int modified();
	unsigned int value();

	void setProtection( ProtectionLevel protection );
	int okToPerform( AccessType attemptedAccess );

	void inhibitDataCaching();
};

inline
unsigned int
SPARCPageTableEntry::value()
{
	return entry ;
}

inline
SPARCPageTableEntry::SPARCPageTableEntry()
{
	entry = 0;
}

inline
SPARCPageTableEntry::SPARCPageTableEntry( unsigned int pte )
{
	entry = pte;
}

inline unsigned int
SPARCPageTableEntry::operator=( unsigned int pte )
{
	entry = pte;
	return( entry );
}

inline SPARCPageTableEntry &
SPARCPageTableEntry::operator= ( SPARCPageTableEntry & pte )
{
	entry = pte.entry;
	return( *this );
}

inline unsigned int
SPARCPageTableEntry::valid()
{
	return( (entry & PG_V) == PG_V);
}

inline void
SPARCPageTableEntry::mapToAddress( const char * addr,
	ProtectionLevel protection )
{
	Assert( !valid() );
	Assert( PageAligned( (const char *) addr ) );

	entry = (unsigned int) addr | 0x1;

	SPARCProtection realProt;
	switch( protection ) {
		case ReadWrite:
			realProt = USERRW;
			break;
		case ReadOnly:
			realProt = USERRO;
			break;
		case NoAccess:
			realProt = SYSTEMRW;
			break;
		default:
			realProt = SYSTEMRO;	// shut up warning.
			Assert( NOTREACHED );
	}

	entry = (unsigned int) addr | ( realProt << 1 ) | 0x1;
}

inline char *
SPARCPageTableEntry::addressMappedTo()
{
	unsigned fram = ( entry & (PG_PFNUM & ~PG_TYPE));
	return( (char *) (fram << Log2PhysicalPageSize));
}

inline void
SPARCPageTableEntry::unmap()
{
	Assert( valid() );
	entry &= ~PG_V;
}

inline unsigned int
SPARCPageTableEntry::referenced()
{
	return( (entry & PG_R) == PG_R );
}

inline unsigned int
SPARCPageTableEntry::modified()
{
	return( (entry & PG_M) == PG_M);
}

inline void
SPARCPageTableEntry::inhibitDataCaching()
{
	entry &= PG_NC;
}

#endif /* SPARCPageTableEntry_h */
