/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifndef SPARCTranslation_h
#define SPARCTranslation_h


#ifdef __GNUC__
#pragma interface
#endif

#include "AddressTranslation.h"	/* parent class */
#include "PAGE_TABLE_ENTRY.h"
#include "SPARCasmutil.h"
#include "Assert.h"
class DummyMemoryObject;
class Lock;

typedef unsigned char SEGMENT_TABLE_ENTRY;
typedef int CONTEXT;

class PhysicalMemoryChain;	// forward reference...
class VirtualMemoryRange;	// forward reference...
class SPARCTranslation;		// forward ref
class BitArray;
const unsigned nsegs = 4096;    //4k segments per ctxt.


/*
 * The SPARC ADT class does two things:
 * 1. Caches mappings in memory in case we use more ADTs than the number of
 * available contexts
 * 2. Manages the shared (static) PMEG table.
 */

class SPARCTranslation : public AddressTranslation {

friend void SS1MemoryInterrupt(char *stackPointer, int type);

	void localAcquire();
	void localRelease();
	void classAcquire();
	void classRelease();
protected:
	/*
	 * static data common to all translations
	 */
	static	int			_classInitialized;
	static	Lock *			_classLock;
	static	SPARCTranslation **	_contextMap;
	static  SPARCTranslation **	_pmegMap;
	static  char **			_pmegAddr;
	static  Lock *			_dmaLock;
	static	BitArray *		_dmaMap;
	static  const char **		_dmaMappings;
	static	unsigned *		_dmaLens;
	static	SPARCTranslation *	_kernelADT;

	/*
	 * Per ADT instance data
	 */
	CONTEXT 			_myContext;	// -1 for none & kernel
	Lock 				_lock;	

	/*
	 * Used when we are not in the MMU
	 */
	int			_stored;
	SEGMENT_TABLE_ENTRY 	_mySegmentPointers[nsegs];
	SEGMENT_TABLE_ENTRY	findPMEGForAddress(char *const);
	SEGMENT_TABLE_ENTRY	reclaimPMEG();

	/*
	 * Used to find a victim to flush from MMU
	 * Use round robin for now...... fix later dkr
	 */
	static CONTEXT			_nextVictim ;
	CONTEXT				findContextVictim();

	/*
	 * Flush and restore ourself from the MMU
	 */
	void				relinquishContext();
	static void			clearPMEG(unsigned which);

	/*
	 * Get in MMU, but leave the current context activated,
	 */
	void				inhabitMMU();
	void				restoreSegmentTable();
	void				initializeSegmentTable();

/*****************************************************************************/
	/*
	 *  The rest are for management of the mappings in the MMU
	 *  This stuff should probably be in a SPARCstationMMU class.
	 */
	static SEGMENT_TABLE_ENTRY segTableEntry( char *const  address);
	static PAGE_TABLE_ENTRY pageTableEntry( char *const  address);
	static void setSegTableEntry( char *const  address, SEGMENT_TABLE_ENTRY value);
	static void setPageTableEntry(char *, const char *, SPARCProtection);
	static void unsetPageTableEntry(char *);
	static void setContext(CONTEXT);
	static void nullPMEG(SEGMENT_TABLE_ENTRY);
	static CONTEXT getContext();
/*****************************************************************************/

	/*
	 * Cache flush thingies
	 */
	static void flushSegment(char *const addr);
	static void flushPage(char *const addr);
	static void flushContext();
	static void invalidateContext();
public:
	/*
	 * Used to fix kernel faults in the VM initialization,
	 * since the SPARC is already running with VM on.  Faults
	 * are in 1-1 memory
	 */
	static void fixKernelInitFault(char *addr);
	static void init(unsigned, VirtualMemoryRange *);
		/*
		 * initialize the MMU mappings
		 * 1. 	Initialize the PMEG map with kernel, HW, invlid,
		 *	and DMA entries
		 * 2.	Initialize the kernel mappings in context 0
		 *	and copy it to other contexts.
		 * THis method ignores the arg.
		 */

	void addMapping( const char *addr, PhysicalMemoryChain *,
	ProtectionLevel);

	void basicRemoveMapping( const char *vaddr, unsigned int nbytes );
	int			basicChangeProtection( const char *vaddr,
			unsigned int nbytes, ProtectionLevel protection );
	int  basicChangeStatus( const char *vaddr, unsigned int nbytes,
				AddressTranslationStatus clear,
				AddressTranslationStatus set );
	int  basicReferenced( const char *vaddr, unsigned int nbytes,
			      int clear );
	char *			mapping( char *vaddr );
	void			basicActivate();

	/*
	 * This is for DVMA device drivers and the vaddr cache.
	 * DMA happens in context 0 always, and in the DMA virtual
	 * area.  TO avoid cache coherency problems we have to map
	 * mod 64K (cache size).  To be efficient we just map mod 256K
	 * by setting the DVMA segptr to the user PTE's.  THe pages had
	 * better not go away whilst the DMA is going on!
	 *
	 * multiCtxt should be 1 where DMA and the CPU must communicate.
	 * It ensures that the DMA mappings are duplicated in all contexts
	 * since the CPU can be in any while the DMA is always in ctxt 0.
	 */
	static int		assertValidAddressRange(char *vaddr, unsigned len);
	static char *		mapToDMA(char *const vaddr, unsigned len, 
					int multiCtxt = 0);
	static char *		unMapDMA(char *const dmaAddr);
	static char *           prepareForDMA(char *const vaddr, unsigned len,
					      int read);
	static char *		postDMA(char *const vaddr, unsigned len,
					int read);
	static char *		makeUncachable(char *const vaddr, unsigned len);
 

	/*
	 * dmaFlush flushes the range passed in the current context
	 */
	static void		rangeFlush(char *addr, unsigned len);


	SPARCTranslation(SPARCTranslation *);
	~SPARCTranslation();
};

inline char *
segmentAlign(void *const  addr)
{
	return (char *)(
		(unsigned)addr & (~0 << Log2SegmentSize) );
}

inline char * 
pageAlign(void *  addr)
{
	return (char *) (
		(unsigned)addr & (~0 << Log2PhysicalPageSize) );
}

inline unsigned
segmentOffset(void * addr)
{
	return  (unsigned)addr % SegmentSize;
}

inline unsigned
pageOffset(void *const addr)
{
	return (unsigned)addr % PhysicalPageSize;
}

#endif /* SPARCTranslation_h */
