/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Contributing Author: See-Mong Tan (stan@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * i586CPU.h: per-cpu private information class for the i586 processor.
 */

#ifndef i586CPU_h
#define i586CPU_h
#ifdef CFRONT
#define CONST const
#define STATIC static
#else /* must be g++ */

#ifdef I586SYM
#pragma interface
#endif

#if (__GNUG__ >= 2)
#ifndef CONST
#define CONST
#define STATIC
#endif
#else
#pragma once
#define CONST const
#define STATIC static
#endif
#endif

#include "CPU.h"
#include "Vectors.h"
#include "i586MMU.h"
#include "i586Context.h"

extern void Halt();

extern "C" {
  unsigned int InterruptsEnabled ();
  unsigned int DisableInterrupts ();
  void EnableInterrupts ();
}

const int nInterrupts = 256;	// Number of Idt Gates; maximum is 256

/* 
 * Structure for the 80586 Gate Descriptors.  These are used in the
 * Interrupt Descriptor Table and in the Global Descriptor Table.
 * CallGates are only allowed in the Gdt.
 */
enum 	GateType { TaskGate=0x05, CallGate=0x0c, InterruptGate=0x0e,
							TrapGate=0x0f };
enum	GatePrivilegeLevel {LvlZero, LvlOne, LvlTwo, LvlThree};

class GateDescriptor {
public:
    unsigned int low_offset	:16;
    unsigned int selector	:16;
    unsigned int count		:5 ;	/* only in CallGates */
    unsigned int zeros		:3 ;
    unsigned int gate_type	:5 ;
    unsigned int dpl		:2 ;
    unsigned int present	:1 ;
    unsigned int high_offset	:16;
};

enum CodeDataType { CodeSegment = 0x1b, DataSegment = 0x13 };
enum Granularity { ByteGranularity = 0x4, PageGranularity = 0xc };

class CodeDataSegmentDescriptor {
public:
    unsigned int low_limit	:16;
    unsigned int low_base	:16;
    unsigned int mid_base	:8;
    unsigned int gate_type	:5;
    unsigned int dpl		:2;
    unsigned int present	:1;
    unsigned int high_limit	:4;
    unsigned int granularity	:4;
    unsigned int high_base	:8;
};

/*
 * Structure for the Interrupt Descriptor Table Register.
 */

struct InterruptDescriptorTableRegister {
    unsigned short pad;
    unsigned short length;
    unsigned int start;
};

class i586CPU : public CPU {
protected:

	/*
	 * The interrupt table and interrupt vector table that all
	 * CPU's share. This is class data so it will work for both a
	 * private and shared memory machine.
	 */
	static Interrupt * interruptTable[ nInterrupts ];

	//This static variable is designed to be set by installInterrupt once
	//the IDT is setup for this processor. Note that this static flag is
	//WRONG if we are using multiprocessors that each need their own IDT. 
	static bool IDTSetup; //MDB addition; 4/26/04

	/*
	 * These are assembler assists for interrupt and trap handling.
	 */
	static void interruptAssist();
	static void SVCAssist() {};
	static void ProxyAssist();
	static void ProxyReturnAssist();

	/*
	 * This is the C++ part of the assembler asisst
	 */
	static void hardwareInterruptAssist( struct stackFrame * );

	i586MMU myMMU;

	void setIdtGate( int gateNumber, GateType type, unsigned short sel,
			PFV addr, GatePrivilegeLevel priv );

	void setGdtGate( int gateNumber, GateType type, unsigned short sel,
		PFV addr, GatePrivilegeLevel priv, unsigned short count = 0);

	void setGdtSegment( int gateNumber, CodeDataType type,
		GatePrivilegeLevel priv, const char * base, unsigned int size );


public:
	i586CPU( unsigned long identifier,
		ProcessContainer * initialIdleContainer );
	~i586CPU();

	/*
	 * Called by kernel to initialize each ring
	 */
	void setRingRange( unsigned char, const char *, const char * );

	void chipInitialize();

	/*
	 * Called by the kernel to install all the interrupts.
	 */
	void installInterrupts();

	/*
	 * The subclasses define these to bind and lookup the binding of
	 * interrupt (hardware) vectors to Interrupts.
	 */
	Interrupt * setInterrupt( int vector, Interrupt * interrupt,
					PFV assist = 0);
	/*
	 * NOT needed on 80586
	 * Interrupt * interrupt( int vector );
         * void sendFlushRequestTo( unsigned long dest );
         * but must add since cannot inherit pure virtual.
	 */
	Interrupt * interrupt( int ) ;
        void sendFlushRequestTo( unsigned long );

	/*
	 * Preempt the processor.
	 */
	void preempt();
	
	static bool isIDTSetup() {return i586CPU::IDTSetup;} //MDB added; 4/26/04

	static int disableInterrupts() { return DisableInterrupts(); } 
	static int disableInterrupts(IntStorage &s) { disableInterrupts()?s=0x2:s=0; return s; }
	//Changed restoreInterrupts to only EnableInterrupts if we have appropriately setup the IDT.
	//MDB added; 4/26/04
	static void restoreInterrupts(int val) {if (isIDTSetup()&&val) EnableInterrupts();}
	static void enableInterrupts(){restoreInterrupts(1);}
	static int interruptsEnabled() {return InterruptsEnabled();}
	static void initMutexState(MutexStateAddr);
	static void spinMutex(MutexStateAddr);
	static void releaseMutex(MutexStateAddr addr){initMutexState(addr);}
	static int cpuID(){return 0;}
};

#endif /* i586CPU_h */
