/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David W. Dykstra (dykstra@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * i586Context.h: Process context for a i586.
 */

#ifndef i586Context_h
#define i586Context_h
#ifdef CFRONT
#define CONST const
#define STATIC static
#else /* must be g++ */

#ifdef I586SYM
#pragma interface
#endif

#if (__GNUC__ >= 2)
#if (!defined STATIC)||(!defined CONST)
#define CONST
#define STATIC
#endif
#else
#define CONST const
#define STATIC static
#endif
#endif

#include "ProcessorContext.h"	/* parent class */
#include "StackSize.h"
#include "EflagsRegister.h"
#include "Types.h"
//#include "CPU_CONFIGURATION.h"

class ObjectProxy;
class Domain;

/*
 * Interrupts get a pointer to this in their "info" argument.
 * Only registers which the compiler doesn't save need be saved here.
 */
class stackFrame {
public:
    unsigned int	ds;	/* first registers saved by interruptAssist */
    unsigned int	gs;
    unsigned int	fs;
    unsigned int	es;
    unsigned int	edx;
    unsigned int	ecx;
    unsigned int	eax;
    unsigned int	vector; /* this one saved by initial interrupt code */
    unsigned int	errno;  /* generated by processor or initial code */
    unsigned int	eip;	/* next three always generated by processor */
    unsigned int	cs;
    EflagsRegister	eflags;
    unsigned int	usp;	/* the next two are present only if trap */
    unsigned int	uss;	/*   was from user mode */
};

class LretFrame {
public:
    unsigned int	userDataSegment;

    /*
     * These next form a stack frame suitable for a lret instruction
     * and must remain in this order.
     */
    PFV		entryPoint;	
    unsigned int	userCodeSegment;
    unsigned int *	userStackPointer;
    unsigned int	userStackSegment;

};

class IretFrame {
public:
    unsigned int	userDataSegment;

    /*
     * These next form a stack frame suitable for a iret instruction
     * and must remain in this order.
     */
    PFV		entryPoint;	
    unsigned int	userCodeSegment;
    EflagsRegister	returnEflags;
    unsigned int *	userStackPointer;
    unsigned int	userStackSegment;

};

class UserProxyState {
public:
    char *  	oldKernelStackPointer;
    char *		oldUserStackPointer;
    unsigned char	oldUserRing;
    /*
     * These next are saved/restored by the proxy trap assembly code.
     * They are all the compiler non-volatile registers that the user
     *    can set: we can't trust that the user hasn't written his own
     *    assembly code to mess with them.
     */
    unsigned int	edi;
    unsigned int	esi;
    unsigned int	ebx;
    unsigned int	ebp;
};

class UserProxyStack {
    friend class i586GenericContext;
protected:
    LretFrame lretFrame;

    UserProxyState state;
};

extern ClassStar i586GenericContextClass;
class i586GenericContext : public ProcessorContext {
protected:
#if (defined CFRONT)||(defined i586)
    virtual void noOp();	/* hack to force vtable generation */
#endif
    /*
     * Stack pointers for every possible ring. 
     *   - stackPointer - current stack pointer.  The pointer for ring 0
     *	is updated on every context switch, the others are updated only
     *      when they make proxy calls.
     *   - stackBase - base of the stack for each ring.  Rings that have
     *      not been used yet for this process will have a stackBase of 0.
     *   - stackBound - bound of the stack for each ring.  Only used for
     *      checking stack bounds.
     * Note that the stackBase is above the stackBound since stacks
     * 	    grow down.
     */
    char * stackPointer[ 2 ];
    char * stackBase[ 2 ];
    char * stackBound[ 2 ];

    /*
     * Start of Kernel stack for use to write into the system
     *   TSS when doing a context switch 
     */
    char * startKernelStack;

    /*
     * Size of stack to allocate for shared rings.
     */
    int sharedStackSize;

    /*
     * Latest known value of the user stack pointer and the ring it
     *   belongs to.
     */
    char * userStackPointer;

    /*
     * Context Switch Register Save Area.
     *   Only registers which the compiler assumes need saving need
     *     to be saved here: register variables and the like.
     */
    unsigned int	edi;
    unsigned int	esi;
    unsigned int	ebx;
    unsigned int	ebp;
    unsigned int	pc;	/* that is, instruction pointer eip */

    /*
     * Generic checkpoint/restore methods.
     */

    Process * checkpoint();
    Process * basicRestore( Process * );

    /*
     * A subclass cannot inherit pure virtual, although this *is*
     * really a subclass.  Thus we have to add this method here.
     *                                 -- Stan
     */
    void restoreFromInitialContext( Process * );
public:
    char *saveUserStack (char *);
    i586GenericContext( Process * owner, StackSize stackSize );
    ~i586GenericContext();

    Class * classOf();
};

/*STATIC inline
  char * 
  i586GenericContext::saveUserStack( char * usp )
  {
  char * oldusp = userStackPointer;
  userStackPointer = usp;
  return ( oldusp );
  }*/


class InitialApplicationContextStackDescription {
    friend class i586ApplicationContext;
protected:
    unsigned int arg1;
    unsigned int arg0;
    PFV	returnPoint;
    ObjectProxy *defaultProxy;

    IretFrame iretFrame;
};

extern ClassStar i586ApplicationContextClass;
class i586ApplicationContext : public i586GenericContext {
protected:
#if (defined CFRONT)||(defined i586)
    virtual void noOp();	/* hack to force vtable generation */
#endif

    /*
     * Additional registers saved by an application context.
     * Floating point registers should also go here.
     */

    /*
     * Overloaded to checkpoint/restore the additional registers above.
     */
    //Process * checkpoint();
    //Process * basicRestore( Process * );

    void restoreFromInitialContext( Process * );
public:
    i586ApplicationContext( Process * owner, PFV entryPoint,
                            StackSize stackSize, Domain *initialDomain, int arg0, int arg1,
                            StackSize userStackSize, char * userStackPointer = 0 );

    void initialRestoreAssist( Process * );

    Class * classOf();
};

extern ClassStar i586SystemContextClass;
class i586SystemContext : public i586GenericContext {
protected:
#if (defined CFRONT)||(defined i586)
    virtual void noOp();	/* hack to force vtable generation */
#endif
    void restoreFromInitialContext( Process * );
    void initialRestoreAssist( Process * );
public:
    i586SystemContext( Process * owner, PFV entryPoint,
                       StackSize stackSize, int arg0, int arg1 );
    i586SystemContext( Process * owner, OPFV entryPoint,
                       StackSize stackSize, int arg0, int arg1 );
    i586SystemContext( Process * owner, PFV entryPoint,
                       StackSize stackSize, OPFV arg0, int arg1 );
    Class * classOf();
};

extern ClassStar i586UninterruptableSystemContextClass;
class i586UninterruptableSystemContext : public i586SystemContext {
protected:
public:
    i586UninterruptableSystemContext( Process * owner, PFV entryPoint,
                                      StackSize stackSize, int arg0, int arg1 );
    Class * classOf();
};

#endif /* i586Context_h */
