/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Contributing Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * i586PageTableEntry.h:  Page table entries for a i586 MMU.
 */

#ifndef i586PageTableEntry_h
#define i586PageTableEntry_h
#ifdef CFRONT
#define CONST const
#define STATIC static
#else /* must be g++ */
#pragma interface
#if (__GNUG__ >= 2)
#define CONST
#define STATIC
#else
#pragma once
#define CONST const
#define STATIC static
#endif
#endif

#include "AccessType.h"
#include "PageTableConstants.h"
#include "ProtectionLevel.h"


/*
 * The representation of these classes must match exactly the hardware
 * representation of page table entries.  This means NO VIRTUAL FUNCTIONS!
 */

/*
 * Here is the i586 descriptor format:
 *
 * +------------------------+-----+---+---+--+-------+-----+
 * |Page Table/Frame Address|     |Mod|Ref|  |Protect|Valid|
 * |                        |     |   |   |  |       |     |
 * |  xxxxxxxxxxxxxxxxxxxx  |00000| x | x |00|  x  x |  x  |
 * +------------------------+-----+---+---+--+-------+-----+
 *  31                    12        6   5       2  1    0
 */


class i586PageTableEntry;

class i586PointerTableEntry {
	unsigned int entry;
public:
	i586PointerTableEntry();

	i586PointerTableEntry & operator=( i586PointerTableEntry & pte );

	void mapToPageTable( i586PageTableEntry * pageTable );
	i586PageTableEntry * pageTable();

	unsigned int valid();

	void setProtection( ProtectionLevel protection );
};


class i586PageTableEntry {
	unsigned int entry;
public:
	i586PageTableEntry();
	i586PageTableEntry( unsigned int pte );

	unsigned int operator=( unsigned int pte );
	i586PageTableEntry & operator=( i586PageTableEntry & pte );

	void mapToAddress( const char * address, ProtectionLevel protection );
	const char * addressMappedTo();
	void unmap();

	unsigned int valid();
	unsigned int referenced();
	unsigned int modified();

	void setProtection( ProtectionLevel protection );
	int okToPerform( AccessType attemptedAccess );

	void inhibitDataCaching();
};



inline
i586PointerTableEntry::i586PointerTableEntry()
{
	entry = 0;
}

inline i586PointerTableEntry &
i586PointerTableEntry::operator= ( i586PointerTableEntry & pte )
{
	entry = pte.entry;
	return( *this );
}

inline unsigned int
i586PointerTableEntry::valid()
{
	return( (entry & 0x1) == 0x1 );
}

inline void
i586PointerTableEntry::mapToPageTable( i586PageTableEntry * pageTable )
{
	Assert( !valid() );
	Assert( PageAligned( (const char *) pageTable ) );

	entry = (unsigned int) pageTable | (USERRW << 1) | 0x1;
}

inline i586PageTableEntry *
i586PointerTableEntry::pageTable()
{
	return( (i586PageTableEntry *) (entry & 0xfffff000) );
}


inline
i586PageTableEntry::i586PageTableEntry()
{
	entry = 0;
}

inline
i586PageTableEntry::i586PageTableEntry( unsigned int pte )
{
	entry = pte;
}

inline unsigned int
i586PageTableEntry::operator=( unsigned int pte )
{
	entry = pte;
	return( entry );
}

inline i586PageTableEntry &
i586PageTableEntry::operator= ( i586PageTableEntry & pte )
{
	entry = pte.entry;
	return( *this );
}

inline unsigned int
i586PageTableEntry::valid()
{
	return( (entry & 0x1) == 0x1 );
}
#if 0
void
i586PageTableEntry::mapToAddress( const char * addr,
	ProtectionLevel protection )
{
	Assert( !valid() );
	Assert( PageAligned( (const char *) addr ) );

	entry = (unsigned int) addr | 0x1;

	i586Protection realProt;
	switch( protection ) {
		case ReadWrite:
			realProt = USERRW;
			break;
		case ReadOnly:
			realProt = USERRO;
			break;
		case NoAccess:
			realProt = SYSTEMRW;
			break;
		default:
			realProt = SYSTEMRO;	// shut up warning.
			Assert( NOTREACHED );
	}

	entry = (unsigned int) addr | ( realProt << 1 ) | 0x1;
}
#endif

inline const char *
i586PageTableEntry::addressMappedTo()
{
	return( (const char *) ( entry & 0xfffff000 ) );
}

inline void
i586PageTableEntry::unmap()
{
	Assert( valid() );
	entry &= ~0x1;
}

inline unsigned int
i586PageTableEntry::referenced()
{
	return( (entry & 0x20) == 0x20 );
}

inline unsigned int
i586PageTableEntry::modified()
{
	return( (entry & 0x40) == 0x40 );
}

inline void
i586PageTableEntry::inhibitDataCaching()
{
}

#endif /* i586PageTableEntry_h */
