/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * SchedulerUSS.h: a Universal Scheduler System process call.
 *
 * FIFOSchedulerUSS		- First-in First-out
 * RRSchedulerUSS		- Round robin
 * LIFOSchedulerUSS		- Last-in First-out
 * MLFSchedulerUSS		- Multi-level feedback
 * HRNSchedulerUSS		- Highest-Response-Ratio-Next
 * UNIXSchedulerUSS		- UNIX-like scheduling
 * VMSSchedulerUSS		- VMS-like scheduling
 * DeadlineSchedulerUSS		- Deadline scheduler
 * SRTSchedulerUSS		- Shortest remaining time
 */

#ifndef SchedulerUSS_h
#define SchedulerUSS_h


#ifdef __GNUC__
#pragma interface
#endif

#include "Process.h"
#include "ProcessContainer.h"
#include "ProcessEvents.h"
#include "FreeRunningTimer.h"
#include "SchedList.h"

class Process;	// forward reference.

/*
 * SchedulerUSS - base class for the universal scheduler system. 
 */
class SchedulerUSS : public ProcessContainer {
protected:
	SchedList * readyList;
	SchedList * runList;				// (Preemptive)
	int quantum;				// (Quantum-Oriented)
	virtual void priority( Process * process );
	virtual void decision( Process * process );
public:
	SchedulerUSS();
	~SchedulerUSS();
	void add( Process * process );
	Process * remove();
	int isEmpty();
};


/*
 * The schedulerInfo field of Process is defined to be a pointer to
 * scheduler specific info record that contains per-process data
 * needed by the scheduling algorithm.  This record is created for
 * a process on its first entry to the scheduler and is deleted when
 * the process is deleted.  It generally contains list pointers and
 * values needed for priority calculations.  It is used by the list
 * data structures.
 */
class USSSchedulerInfo : public SchedulerInfo {
protected:
	Process * myProcess;
	USSSchedulerInfo * prev;
	USSSchedulerInfo * next;
	int intPriority;		
	double doublePriority;
	int cpuPreempted;			// (Preemptive Schedulers)
public:
	USSSchedulerInfo( Process * process );
	~USSSchedulerInfo();
	Process * getProcess();
	USSSchedulerInfo * getNext();
	void setNext( USSSchedulerInfo * info );
	USSSchedulerInfo * getPrev();
	void setPrev( USSSchedulerInfo * info );
	int getIntPriority();
	void setIntPriority( int val );
	virtual double getDoublePriority( unsigned int );
	void setDoublePriority( double val );
	void setPreempt( int preempt );
	int getPreempt();
};

inline Process *
USSSchedulerInfo::getProcess()
{
	return( myProcess );
}

inline USSSchedulerInfo *
USSSchedulerInfo::getNext()
{
	return( next );
}

inline void
USSSchedulerInfo::setNext( USSSchedulerInfo * info )
{
	next = info;
}

inline USSSchedulerInfo *
USSSchedulerInfo::getPrev()
{
	return( prev );
}

inline void
USSSchedulerInfo::setPrev( USSSchedulerInfo * info )
{
	prev = info;
}

inline int
USSSchedulerInfo::getIntPriority()
{
	return( intPriority );
}

inline void
USSSchedulerInfo::setIntPriority( int val )
{
	intPriority = val;
}

inline double
USSSchedulerInfo::getDoublePriority( unsigned int )
{
	return( doublePriority );
}

inline void 
USSSchedulerInfo::setDoublePriority( double val )
{
	doublePriority = val;
}

inline void
USSSchedulerInfo::setPreempt( int preempt )
{
	cpuPreempted = preempt;
}

inline int
USSSchedulerInfo::getPreempt()
{
	return( cpuPreempted );
}

/*
 * FIFOSchedulerUSS - a non-preemptive FIFO run-to-completion scheduler
 */
class FIFOSchedulerUSS : public SchedulerUSS {
protected:
	void priority( Process * );
public:
	FIFOSchedulerUSS( char * name = "FIFOSchedulerUSS" );
};


/*
 * RRSchedulerUSS - a time-slicing round-robin scheduler.
 */
class RRSchedulerUSS : public FIFOSchedulerUSS {
public:
	RRSchedulerUSS( int thisQuantum, char * name = "RRSchedulerUSS" );
};


/*
 * LIFOSchedulerUSS - a last-in first-out scheduler
 */
class LIFOSchedulerUSS : public FIFOSchedulerUSS {
public:
	LIFOSchedulerUSS( char * name = "LIFOSchedulerUSS" );
};


/*
 * SchedulerInfo class for DeadlineSchedulerUSS.  
 */
class DeadlineSchedulerInfo : public USSSchedulerInfo {
protected:
	unsigned int myTimeLeft;
	unsigned int mustStartTime;
public:
	void setTimeLeft( unsigned int time );
	unsigned int getTimeLeft();
	void setStartTime();
	unsigned int getStartTime();
	DeadlineSchedulerInfo( Process * process );
};

inline void
DeadlineSchedulerInfo::setTimeLeft( unsigned int time )
{
	myTimeLeft = time;
}

inline unsigned int
DeadlineSchedulerInfo::getTimeLeft()
{
	return( myTimeLeft );
}

inline unsigned int
DeadlineSchedulerInfo::getStartTime()
{
	return( mustStartTime );
}

/*
 * DeadlineSchedulerUSS - a deadline scheduler
 */
class DeadlineSchedulerUSS : public SchedulerUSS {
protected:
	void priority( Process * process );
public:
	DeadlineSchedulerUSS( int thisQuantum, 
		char * name = "DeadlineSchedulerUSS" );
};


/*
 * SchedulerInfo class for UNIXSchedulerUSS.  
 */
class UNIXSchedulerInfo : public USSSchedulerInfo {
public:
	UNIXSchedulerInfo( Process * process );

	void setDoublePriority( double count );
	double getDoublePriority( unsigned int now );
};

inline void
UNIXSchedulerInfo::setDoublePriority( double count )
{
	doublePriority = count;
}

/*
 * UNIXSchedulerUSS - a UNIX-like scheduler
 */
class UNIXSchedulerUSS : public SchedulerUSS {
protected:
	void priority( Process *);
public:
	UNIXSchedulerUSS( int thisQuantum, char * name = "UNIXSchedulerUSS" );
};


/*
 * SchedulerInfo class for HRNSchedulerUSS.  
 */
class HRNSchedulerInfo : public USSSchedulerInfo {
	unsigned int startTime;
public:
	double getDoublePriority( unsigned int now );
	HRNSchedulerInfo( Process * process );
};

/*
 * HRNSchedulerUSS - a highest response-ratio next scheduler
 */
class HRNSchedulerUSS : public SchedulerUSS {
protected:
	void priority( Process * );
public:
	HRNSchedulerUSS( char * name = "HRNSchedulerUSS" );
};


/*
 * MLFSchedulerUSS - a multi-level feedback queue scheduler
 */
class MLFSchedulerUSS : public SchedulerUSS {
protected:
	int myQuantum;
	int levels;
	void priority( Process * );
	virtual int calcQuantum( int thisLevel );
public:
	MLFSchedulerUSS( int thisQuantum, int thisLevels, 
		char * name = "MLFSchedulerUSS" );
	Process * remove();
};


/*
 * SchedulerInfo class for VMSSchedulerUSS.  
 */
class VMSSchedulerInfo : public USSSchedulerInfo {
protected:
	SchedPriorityList * runList;
public:
	VMSSchedulerInfo( Process * process, SchedPriorityList * myList );
	~VMSSchedulerInfo();
};

/*
 * VMSSchedulerUSS - a VMS-like scheduler
 */
class VMSSchedulerUSS : public MLFSchedulerUSS {
protected:
	int myQuantum;
	int calcQuantum( int thisLevel );
	void priority( Process * );
public:
	VMSSchedulerUSS( int thisQuantum, char * name = "VMSSchedulerUSS" );
};
    	
/* 
 * SchedulerInfo class for SRTSchedulerUSS.  
 */
class SRTSchedulerUSS;
class SRTSchedulerInfo : public USSSchedulerInfo {
protected:
	SchedPriorityList * runList;
	unsigned int timeLeft;
	unsigned int startWaiting;
	SRTSchedulerUSS * myScheduler;
public:
	unsigned int getStart();
	void setStart();
	void setTimeLeft( unsigned int val );
	unsigned int getTimeLeft();
	double getDoublePriority( unsigned int now );
	SRTSchedulerInfo( Process * process, SchedPriorityList * myList ); 
	~SRTSchedulerInfo();
};

inline unsigned int
SRTSchedulerInfo::getStart()
{
	return( startWaiting );
}

inline void
SRTSchedulerInfo::setStart()
{
	startWaiting = SystemTimer->time();
}

inline void
SRTSchedulerInfo::setTimeLeft( unsigned int val )
{
	timeLeft = val;
}

inline unsigned int
SRTSchedulerInfo::getTimeLeft()
{
	return( timeLeft );
}

/* 
 * SRTSchedulerUSS - A preemptive shortest remaining time scheduler 
 */
class SRTSchedulerUSS : public SchedulerUSS {
protected:
	void priority( Process * );
public:
	SRTSchedulerUSS( char * name = "SRTSchedulerUSS" );
	Process * remove();
};

#endif /* SchedulerUSS_h */
