#ifdef __GNUC__
#pragma implementation
#endif

 /*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Instrument.cc: basic instruments for use anywhere
 */

#ifdef INSTRUMENT

#include "Assert.h"
#include "OutputStream.h"
#include "CPU.h"
#include "ThisCPU.h"
#include "Interrupts.h"
#include "Instrument.h"
#include "InstrumentManager.h"
#include "InstrumentEvent.h"


#ifdef __GNUC__
#pragma implementation
#endif

// ----------------------------------------------------------------------------

/*
Initialize all the variables.
Tell the instrumentManager about us and get a unique identifier.
*/

Instrument::Instrument (InstrumentManager *iowner,
			char *title,
			iStateType state)
{
	_owner = iowner;

	_title = new char[StringLength (title) + 1];
	StringCopy (_title, title);

	_state.suspended = FALSE;
	changeState (state);

	// Attempt to solve the suspend on boot problem on the system side
	if (iowner->suspendNewInstruments()) {
		suspend ();
	}
}

/*
Non-event instrument logging should occur in the subclasses due to reverse
  destruction of classes.
Tell the owner we're dying in the subclass since the owner needs to get the
  instrument ID which is in the subclass header field.
*/
Instrument::~Instrument ()
{
}

/*
Change the active/printing state.
*/
void 
Instrument :: changeState (iStateType newState)
{
	_state.currentState = newState;
	_state.active = (newState == AWAKE || newState == MUTE ||
			 newState == TALKING);
	_state.verbal = (newState == AWAKE || newState == ASLEEP ||
			 newState == TALKING || newState == SNORING);
	_state.loud = (newState == TALKING || newState == SNORING);
}

void 
Instrument :: suspend ()
{
	if (_state.suspended) return;

	_state.suspended = TRUE;
	_state.suspendedState = _state.currentState;

	changeState (DEAD);
}

void 
Instrument :: resume ()
{
	if ( ! _state.suspended) return;

	_state.suspended = FALSE;
	changeState (_state.suspendedState);
}

void 
Instrument :: begin ()
{
	if ( ! _state.suspended) return;

	_state.suspended = FALSE;
	changeState (_state.suspendedState);
}

//----------------------------------------------------------------------------

/*
Set the type fields to an eventer.
*/
EventerInstrument::EventerInstrument  (InstrumentManager *iowner,
				       char *title,
				       iStateType state)
	: (iowner, title, state)
{
	_state.type = EVENTER;
	_header.id.special = FALSE;
	_header.id.count = FALSE;
	_header.id.time = FALSE;
	_header.id.message = FALSE;
	_header.id.percpu = FALSE;
	_header.id.perprocess = TRUE;
	_header.id.ownerid = _owner->getID ();
	_header.id.mode = _owner->getMode ();
	_header.id.selfid = _owner->addInstrument (this);
	reset ();
	_header.id.special = TRUE;
	birthLog ();
	_header.id.special = FALSE;
}

/*
This is done in the subclass since the log definition will disappear...
*/
EventerInstrument::~EventerInstrument ()
{
	// Essentially a deathLog
	_header.id.special = TRUE;
	_header.id.message = FALSE;
	_owner->writeLogBuffer (sizeof(_header),
				(char *)&_header,
				InstrumentDeath);
	_header.id.special = FALSE;
	_owner->removeInstrument (this);
}

/*
The format for the event log is:
<time> <header> <eentId> {message length, message}
The message is ignored if the instrument is quiet.
*/
void 
EventerInstrument::birthLog ()
{
	_header.id.message = TRUE;
	_owner->writeLogBuffer (sizeof(_header),
				(char *)&_header,
				InstrumentBirth,
				StringLength (_title),
				_title);
}

void 
EventerInstrument::deathLog ()
{
	_header.id.message = FALSE;
	_owner->writeLogBuffer (sizeof(_header),
				(char *)&_header,
				InstrumentDeath);
}

void 
EventerInstrument::log (unsigned int eventID,
				    char *message,
				    unsigned int messageLength)
{
	if (_state.loud && message != NULL) {
		if (messageLength == 0) {
			messageLength = StringLength (message);
			Assert (messageLength < 256);
		}
		_header.id.message = TRUE;
		_owner->writeLogBuffer (sizeof(_header),
					(char *)&_header,
					eventID,
					messageLength, message);
	}
	else {
		_header.id.message = FALSE;
		_owner->writeLogBuffer (sizeof(_header),
					(char *)&_header,
					eventID);
	}
}

/*
Just logs an event.
*/
void 
EventerInstrument::event (unsigned int eventID,
			       char *message,
			       unsigned int messageLength)
{
	if (!_state.active || !_state.verbal)
		return;

	log (eventID, message, messageLength);
}

unsigned int 
EventerInstrument :: getID ()
{
	return (_header.id.selfid);
}

void 
EventerInstrument :: reset ()
{
        // eventer has nothing to reset
}

//----------------------------------------------------------------------------

/*
Set the type fields to a counter.
*/
CounterInstrument::CounterInstrument  (InstrumentManager *iowner,
				       char *title,
				       iStateType state,
				       unsigned int freq)
	: (iowner, title, state)
{
	_state.type = COUNTER;
	_header.id.special = FALSE;
	_header.id.count = TRUE;
	_header.id.time = FALSE;
	_header.id.message = FALSE;
	_header.id.percpu = FALSE;
	_header.id.perprocess = FALSE;
	_header.id.ownerid = _owner->getID ();
	_header.id.mode = _owner->getMode ();
	_header.id.selfid = _owner->addInstrument (this);
	_freq = freq;
	reset ();
	_header.id.special = TRUE;
	birthLog ();
	_header.id.special = FALSE;
}

/*
This is done in the subclass since the log definition will disappear...
*/
CounterInstrument::~CounterInstrument ()
{
	// Essentially a deathLog
	_header.id.special = TRUE;
	_header.id.message = FALSE;
	_owner->writeLogBuffer (sizeof(_header),
				(char *)&_header,
				InstrumentDeath);
	_header.id.special = FALSE;
	_owner->removeInstrument (this);
}

/*
The format for the event log is:
<time> <header> <count> <eventID> {message length, message}
*/
void 
CounterInstrument::birthLog ()
{
	_header.id.message = TRUE;
	_owner->writeLogBuffer (sizeof(_header),
				(char *)&_header,
				InstrumentBirth,
				StringLength (_title),
				_title);
}

void 
CounterInstrument::deathLog ()
{
	_header.id.message = FALSE;
	_owner->writeLogBuffer (sizeof(_header),
				(char *)&_header,
				InstrumentDeath);
}

void 
CounterInstrument::log (unsigned int eventID,
				    char *message,
				    unsigned int messageLength)
{
	if (_state.loud && message != NULL) {
		if (messageLength == 0) {
			messageLength = StringLength (message);
			Assert (messageLength < 256);
		}
		_header.id.message = TRUE;
		_owner->writeLogBuffer (sizeof(_header),
					(char *)&_header,
					eventID,
					messageLength,
					message);
	}
	else {
		_header.id.message = FALSE;
		_owner->writeLogBuffer (sizeof(_header),
					(char *)&_header,
					eventID);
	}
	
	reset ();
}

/*
Increments the counter by 1.
If the counter is equal to the counter bound, then it is placed in
the log.  Note that if counter bound = 1, this is identical to
an event logger; this is done so that if an event occurs too often,
its printing can be cut.
*/
void 
CounterInstrument::event (unsigned int eventID,
			       char *message,
			       unsigned int messageLength)
{
	if (!_state.active)
		return;

	if ((++_header.count >= _freq) && _state.verbal)
		log (eventID, message, messageLength);
}

unsigned int 
CounterInstrument :: getID ()
{
	return (_header.id.selfid);
}

void 
CounterInstrument :: reset ()
{
	_header.count = 0;
}

/*
Change the logging frequency to once every newFreq events.
*/
void 
CounterInstrument :: changeFreq (unsigned int newFreq)
{
	_freq = newFreq;
}

//----------------------------------------------------------------------------

/*
Set the type fields to a timer.
*/
TimerInstrument::TimerInstrument  (InstrumentManager *iowner,
				     char *title,
				     iStateType state,
				     unsigned int freq,
				     iScopeType scope)
	: (iowner, title, state)
{
	_state.type = TIMER;
	_header.id.special = FALSE;
	_header.id.count = TRUE;
	_header.id.time = TRUE;
	_header.id.message = FALSE;
	if (scope == SYSTEMPERCPU) {
		_header.id.percpu = TRUE;
	} else {
		_header.id.percpu = FALSE;
	}
	_header.id.perprocess = FALSE;
	_header.id.ownerid = _owner->getID ();
	_header.id.mode = _owner->getMode ();
	_header.id.selfid = _owner->addInstrument (this);
	_freq = freq;
	_timerScope = scope;
	reset ();
	_header.id.special = TRUE;
	birthLog ();
	_header.id.special = FALSE;
}

/*
Log the instrument if it is non-event.
This is done in the subclass since the log definition will disappear...
*/
TimerInstrument::~TimerInstrument ()
{
	// Essentially a deathLog
	_header.id.special = TRUE;
	_header.id.message = FALSE;
	_owner->writeLogBuffer (sizeof(_header),
				(char *)&_header,
				InstrumentDeath);
	_header.id.special = FALSE;
	_owner->removeInstrument (this);
}


/*
The format for the event log is:
<time> <header> <count> <delta time> <eventID> {message length, message}
Print out thyself.
*/

void 
TimerInstrument::birthLog ()
{
	_header.id.message = TRUE;
	_owner->writeLogBuffer (sizeof(_header),
				(char *)&_header,
				InstrumentBirth,
				StringLength (_title),
				_title);
}

void 
TimerInstrument::deathLog ()
{
	_header.id.message = FALSE;
	_owner->writeLogBuffer (sizeof(_header),
				(char *)&_header,
				InstrumentDeath);
}

void 
TimerInstrument::log (unsigned int eventID,
				  char *message,
				  unsigned int messageLength)
{
	if (_state.loud && message != NULL) {
		if (messageLength == 0) {
			messageLength = StringLength (message);
			Assert (messageLength < 256);
		}
		_header.id.message = TRUE;
		_owner->writeLogBuffer (sizeof(_header),
					(char *)&_header,
					eventID,
					messageLength, message);
	}
	else {
		_header.id.message = FALSE;
		_owner->writeLogBuffer (sizeof(_header),
					(char *)&_header,
					eventID);
	}
	
	reset ();
}

/*
Starts the timer running.  Places timer in RUNNING state.
*/
void 
TimerInstrument::start ()
{
	if (!_state.active) {
		return;
	}

	// Don't assert because of change state case
	Assert (_timerState != RUNNING);
	
	_timerState = RUNNING;
	_startTime = _owner->time ();
}

// ----------------------------------------------------------------------------

/*
Stops the current timer and adds it to the total timer.  Increment
the count by 1.  Places timer in STOPPED state.
*/
void 
TimerInstrument::stop (unsigned int eventID,
			    char *message,
			    unsigned int messageLength)
{
	if (!_state.active) {
		return;
	}

	// Don't assert because of change state case
	Assert (_timerState != STOPPED);

	if (_timerState != RUNNING) {
		_timerState = STOPPED;
		return;
	}
	
	timeType currentTime = _owner->time ();

	// Take care of system timer wrap-around
	if (currentTime >= _startTime)
		_header.time += currentTime - _startTime;
	else
		_header.time += ((timeType) -1) - _startTime + currentTime;

	_timerState = STOPPED;

	if ((++_header.count >= _freq) && _state.verbal)
		log (eventID, message, messageLength);
}

void 
TimerInstrument :: reset ()
{
	_header.count = 0;
	_header.time = 0;

	_timerState = STOPPED;
}

void 
TimerInstrument :: suspend ()
{
	if (_header.id.percpu == TRUE) {
		return;
	}
        if (_state.suspended) {
		return;
	}

        _state.suspended = TRUE;
        _state.suspendedState = _state.currentState;

        changeState (DEAD);
}

void 
TimerInstrument :: resume ()
{
	if (_header.id.percpu == TRUE) {
		return;
	}
        if ( ! _state.suspended) {
		return;
	}

        _state.suspended = FALSE;
        changeState (_state.suspendedState);
}

void 
TimerInstrument::changeScope (iScopeType newScope)
{
	if (newScope == SYSTEMPERCPU) {
		_header.id.percpu = TRUE;
	} else {
		_header.id.percpu = FALSE;
	}
        _timerScope = newScope;
}



unsigned int TimerInstrument :: getID ()
{
	return (_header.id.selfid);
}

void TimerInstrument :: changeFreq (unsigned int newFreq)
{
	_freq = newFreq;
}

/*
 * Inline string functions
 */
unsigned int StringLength (const char *string)
{
	for (register const char *temp = string; *temp; temp++)
		;
	return (temp - string);
}

void StringCopy (char *dest, const char *source)
{
	while (*(dest++) = *(source++))
		;
}

#include "ClassInstaller.h"

DEFINECLASS("Instrument",Instrument,ProxiableObject,0,0);
DEFINECLASS("CounterInstrument",CounterInstrument,Instrument,0,0);
DEFINECLASS("EventerInstrument",EventerInstrument,Instrument,0,0);
DEFINECLASS("TimerInstrument",TimerInstrument,Instrument,0,0);

#endif INSTRUMENT
