#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifdef INSTRUMENT

#include "Assert.h"
#include "OutputStream.h"
#include "PerCpuTimerInstrument.h"
#include "SystemInstrumentManager.h"
#include "InstrumentEvent.h"
#include "ThisCPU.h"


#ifdef __GNUC__
#pragma implementation
#endif

PerCpuTimerInstrument::PerCpuTimerInstrument  (SystemInstrumentManager *iowner,
					 char *title,
					 iStateType state,
					 unsigned int freq,
					 iScopeType scope)
	: ((InstrumentManager *) iowner, title, state, freq, scope)
{
	Assert(scope == ALL);

	_activeCPUs = iowner->numInstrumentedCPUs();
	_theInstruments = new TimerInstrument *[_activeCPUs];
	_theStates = new iInstrumentStateType [_activeCPUs];

	for (int i = 0; i < _activeCPUs; i++) {
	     _theInstruments[i] = new TimerInstrument(iowner,
			title, state, freq, scope);
		_theStates[i].suspended = FALSE;
		_theStates[i].currentState = state;
		// suspend per cpu instrument
	        _theInstruments[i]->changeScope(SYSTEMPERCPU);
		if (iowner->suspendNewInstruments()) {
			//suspend();  // suspend global instrument
			_theStates[i].suspended = TRUE;
			_theStates[i].suspendedState = 
				_theStates[i].currentState;
			_theInstruments[i]->changeState(DEAD);
		}
	}
}

PerCpuTimerInstrument::~PerCpuTimerInstrument ()
{
	deathLog ();
	
	for (int i = 0; i < _activeCPUs; i++) {
		delete _theInstruments[i];
		delete (&(_theStates[i]));
	}
	delete /*[]*/ _theInstruments;
}

void PerCpuTimerInstrument::deathLog ()
{
	// nothing?
}

void 
PerCpuTimerInstrument::birthLog ()
{
	// nothing?
	Assert(NOTREACHED);
}

void 
PerCpuTimerInstrument::log (unsigned int eventID,
				     char *message,
				     unsigned int messageLength)
{
	/* protected method--should not be invoked*/
	Assert(NOTREACHED);
}


void 
PerCpuTimerInstrument::start ()
{
iInstrumentStateType *s = &(_theStates[thisCPU()->instrumentationID()]);

	if (!s->active) {
		return;
	}
	TimerInstrument *t = _theInstruments[thisCPU()->instrumentationID()];
	t->start();
}

void 
PerCpuTimerInstrument::stop (unsigned int eventID,
			       char *message,
			       unsigned int messageLength)
{
iInstrumentStateType *s = &(_theStates[thisCPU()->instrumentationID()]);
	if (!s->active) {
		return;
	}
	TimerInstrument *t = _theInstruments[thisCPU()->instrumentationID()];
	t->stop(eventID, message, messageLength);
}

void 
PerCpuTimerInstrument::suspend ()
{
iInstrumentStateType *s = &(_theStates[thisCPU()->instrumentationID()]);
      if (s->suspended == TRUE) {
                return;
        }

        s->suspended = TRUE;
        s->suspendedState = s->currentState;

        changeState (DEAD);
}

void 
PerCpuTimerInstrument::resume ()
{
iInstrumentStateType *s = &(_theStates[thisCPU()->instrumentationID()]);
        if ( ! s->suspended) {
                return;
        }

        s->suspended = FALSE;
        changeState (s->suspendedState);
}

/*  resume them all */
void 
PerCpuTimerInstrument::begin ()
{
	for (int i = 0; i < _activeCPUs; i++) {
	     _theInstruments[i]->changeScope(ALL);
	     _theInstruments[i]->resume();
	     _theInstruments[i]->changeScope(SYSTEMPERCPU);
		iInstrumentStateType *s = &(_theStates[i]);
		if (s->suspended == TRUE) {
			s->suspended = FALSE;
			changeState (s->suspendedState);
		}
	}
}

// Call exactly once after enabling instruments--need interlock?
void 
PerCpuTimerInstrument::startAll ()
{
	for (int i = 0; i < _activeCPUs; i++) {
	     iInstrumentStateType *s = &(_theStates[i]);
		Assert( s->suspended == FALSE);
	     _theInstruments[i]->changeScope(ALL);
	     _theInstruments[i]->start();
	     _theInstruments[i]->changeScope(SYSTEMPERCPU);
	}
}

void 
PerCpuTimerInstrument::reset ()
{
	TimerInstrument *t = _theInstruments[thisCPU()->instrumentationID()];
	t->reset();
}

unsigned int
PerCpuTimerInstrument::getID ()
{
	TimerInstrument *t = _theInstruments[thisCPU()->instrumentationID()];
	return(t->getID());
}

void
PerCpuTimerInstrument::changeState(iStateType newState)
{
	iInstrumentStateType *s = &(_theStates[thisCPU()->instrumentationID()]);
	s->currentState = newState;
        s->active = (newState == AWAKE || newState == MUTE ||
                         newState == TALKING);
        s->verbal = (newState == AWAKE || newState == ASLEEP ||
                         newState == TALKING || newState == SNORING);
        s->loud = (newState == TALKING || newState == SNORING);
}

void
PerCpuTimerInstrument::changeFreq(unsigned int newFreq)
{
	TimerInstrument *t = _theInstruments[thisCPU()->instrumentationID()];
	t->changeFreq(newFreq);
}

void
PerCpuTimerInstrument::changeScope (iScopeType newScope)
{
	TimerInstrument *t = _theInstruments[thisCPU()->instrumentationID()];
	t->changeScope(newScope);
}

#include "ClassInstaller.h"
DEFINECLASS("PerCpuTimerInstrument",PerCpuTimerInstrument,TimerInstrument,0,0);
#endif INSTRUMENT
