#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * SystemInstrumentManager.cc: OS-only instrument manager
 *
 */

#ifdef INSTRUMENT

#include "Assert.h"
#include "CPU.h"
#include "ThisCPU.h"
#include "Kernel.h"
#include "Process.h"
#include "ThisProcess.h"
#include "MachineConfiguration.h"
#include "Vectors.h"
#include "OutputStream.h"
#include "FreeRunningTimer.h"
#ifndef iPSC2
#include "FileSystemInterface.h"
#include "PersistentArray.h"
#include "RecordStream.h"
#include "String.h"
#endif
#include "FIFOScheduler.h"
#include "PROCESSOR.h"
#include "SystemInstrumentManager.h"
#include "UserInstrumentManager.h"
#include "PerCpuTimerInstrument.h"


#ifdef __GNUC__
#pragma implementation
#endif

#define Instr(exp) exp

Boolean SystemInstrumentManager :: _loggingIsOn = FALSE;
char SystemInstrumentManager :: _lastLogFile[128];
#ifndef iPSC2
RecordStreamStar SystemInstrumentManager :: _logDumpFile = 0;
FileSystemInterfaceStar SystemInstrumentManager :: _FSI = 0;
#endif

extern SystemInstrumentManager *TheInstrumentManager;

extern PerCpuTimerInstrument *iProcessRestore;
#ifdef INSTRUMENT_IO
extern EventerInstrument *iIO;
#endif INSTRUMENT_IO


PerCpuTimerInstrument *iSystemInstrumentManagerDumpLogBuffer = NULL;
PerCpuTimerInstrument *iInstrumentBufferInterrupt = NULL;

/****  CONFIGURATION INFO--read from file? or from MachDep Include  *****/
#ifdef SPARC
	char *initialInstrumentationLogFileName = "/usr/Choices/silly";
#elif MULTIMAX
	char *initialInstrumentationLogFileName = "/home/srg/mcgrath/silly";
#elif VCHOICES
	char *initialInstrumentationLogFileName = "/instrument.out";
#else
	char *initialInstrumentationLogFileName = NULL;
#endif
/****  end CONFIGURATION INFO  *****/

void 
SystemInstrumentManagerConstructInstruments ()
{
        Debug << "SystemConstructingInstruments\n" << eor;
        iSystemInstrumentManagerDumpLogBuffer = new PerCpuTimerInstrument(
                TheInstrumentManager,
                "SystemInstrumentManagerDumpLogBuffer",
                TALKING,
                1,
                ALL);
        iInstrumentBufferInterrupt = new PerCpuTimerInstrument
                (TheInstrumentManager,
                "InstrumentBufferInterrupt",
                TALKING,
                1,
                ALL);
        Debug << "Created iIBE\n" << eor;
}

void 
SystemInstrumentManagerDeleteInstruments ()
{
        delete iSystemInstrumentManagerDumpLogBuffer;
        iSystemInstrumentManagerDumpLogBuffer = NULL;
        delete iInstrumentBufferInterrupt;
        iInstrumentBufferInterrupt = NULL;
}

unsigned int
mycpuID ()
{
	return (thisCPU()->_icpuID);
}

/*
The constructor.
*/
SystemInstrumentManager :: SystemInstrumentManager
(char *title, unsigned int instrumentAlloc, unsigned int managerAlloc)
	: (title, instrumentAlloc)
{
	Debug << this << "::SystemInstrumentManager()\n" << eor;

	_header.id.ownerid = 0;

	birthLog (title);

	_loggingIsOn = FALSE;
#ifndef iPSC2
	_FSI = NULL;
	_logDumpFile = NULL;
#endif

	_numInstrumentedCPUs = thisCPU()->_highestInstrumentationID + 1;

	_suspendNewInstruments = TRUE;

	_iManagerList = new InstrumentManager *[managerAlloc];
	
	_maxIManagers = managerAlloc;
	_numIManagers = 0;

	Debug << this << "End of IMS Cons\n" << eor;
}

/*
Log the managers.
Log the instruments.
Possibly kill all the instruments.
Close the log file.
*/
SystemInstrumentManager :: ~SystemInstrumentManager ()
{
	Debug << this << "::~SystemInstrumentManager()\n" << eor;

	deathLog ();

	_header.id.special = TRUE;
	_header.id.message = FALSE;
	writeLogBuffer (sizeof(_header), (char *)&_header, InstrumentDeath);
	_header.id.special = FALSE;

#ifndef iPSC2
	if (_loggingIsOn && _logDumpFile != NULL)
		closeLogDumpFile (); 
#endif

}

/*
Slightly different since we have the managers to take care of.
*/
void SystemInstrumentManager :: deathLog ()
{
	Debug << this << "::deathLog()\n" << eor;

	logAllInstrumentManagers ();

	logAllInstruments ();
}

/*
Assign an instrument manager ID to the user instrument manager.
*/
unsigned int 
SystemInstrumentManager :: addInstrumentManager(UserInstrumentManager * manager)
{
	Debug << this << "::addInstrumentManager( " << manager << " )\n" << eor;

	if (_numIManagers < _maxIManagers) {
		_iManagerList[_numIManagers++] = manager;
		return (_numIManagers);
	}
	else {
		register unsigned int i;

		// Find an unused entry
		for (i = 0;
		     ((i < _maxIManagers) &&
		      (_iManagerList[i] != NULL));
		     i++)
			;

		// If all entries are used
		if (i == _maxIManagers) {
			return (FALSE);
		}
		// We found an entry
		else {
			_iManagerList[i] = manager;

			return (i);
		}
	}
}


/*
Remove an active user instrument manager.
*/
void 
SystemInstrumentManager::removeInstrumentManager(
	UserInstrumentManager * manager)
{
	Debug << this << "::removeInstrumentManager( " << manager << " )\n" << eor;

	if (_iManagerList != NULL) {
		_iManagerList[((InstrumentManager *) manager)->getID() - 1] = NULL;
	}
}


/*
Command all managers to log themselves.
*/
void
SystemInstrumentManager :: logAllInstrumentManagers ()
{
	Debug << this << "::logAllInstrumentManagers()\n" << eor;

	register InstrumentManager **i = _iManagerList;
	
	for ( ; i != (_iManagerList + _numIManagers); i++) {
		if ((*i) != NULL)
			(*i)->deathLog ();
	}
}

Boolean
SystemInstrumentManager :: openInitialLogDumpFile()
{
#ifdef iPSC2
	MachineConfiguration *config = NULL;
#else
	MachineConfiguration *config = TheKernel->getConfiguration();
#endif
	int status = 1;
	char * fileToOpen = initialInstrumentationLogFileName;

	if (config != NULL) {
#ifdef iPSC2
		Assert(NOTREACHED);
#else
		Console() << "Lookup instrumentation log file.\n" << eor;
		status = config->lookup("InstrumentLog",fileToOpen);
		if (status == 0) {
			Console() << "Instrumentation will be logged to file " 
				<< fileToOpen << ".\n" << eor;
		} else {
			Console() << 
			"Instrumentation log file not found in configuration.\n"
			 << eor;
			fileToOpen = initialInstrumentationLogFileName;
		}
#endif
	}

	if (fileToOpen == NULL) {
		Console() << "No initial instrumentation log file specified, logging not enabled.\n" << eor;
		Console() << "Instrumentation data is not being saved.\n" << eor;
		return(NOFILESPECIFIED);
	} else {
		iLogOpenResult res = openLogDumpFile(fileToOpen);
		switch (res) {
		case OKNEW:
		case OKAPPEND:
			Console() << "Initial instrumentation log file " 
				<< fileToOpen 
				<< " open, logging enabled.\n" << eor;
			break;
		default:
			Console() << "Initial instrumentation log file " 
				<< fileToOpen 
			<< " could not be opened, logging not enabled.\n" 
			<< eor;
			Console() <<"Instrumentation data is not being saved.\n"
			<< eor;
			break;
		}
		return(res);
	}
}

/*
Specify the file to dump the log data to.
Log dumps prior to opening a file are lost.
Return FALSE if there is an error.
*/
iLogOpenResult
SystemInstrumentManager :: openLogDumpFile
(char * fileName, unsigned int mode = 0666)
{
#ifndef iPSC2
	Debug << this << "::openLogDumpFile()\n" << eor;

	if (fileName == NULL) {
		Console () << this << "::openLogDump: new log file is null.\nOld log still active.\n" << eor;
		Console().flush();
		return(NOFILESPECIFIED);
	}

	if (_logDumpFile != NULL) {
		int ret = closeLogDumpFile ();
		if (ret == FALSE) {
			Console () << this << "::openLogDump: can't close log file.\n" << eor;
			Console().flush();
			return(CANTCLOSE);
		}
	}

	// Somehow check if it is already open???

	// The way to do it in the new stable.
	_FSI = TheKernel->fileSystemInterface ();

	// REMOVE THIS WHEN BUG GETS FIXED
	//_FSI->unlink (fileName);

	// Make sure we don't get an error when opening the file.
	int fileStatus;
       _logDumpFile =  _FSI->open(fileName, RecordStreamClass, fileStatus, O_APPEND);
	if (fileStatus != 0) {
		//Console () << "Instrumentation log file could not be openned.\n" << eor;
	       _logDumpFile =  _FSI->open(fileName, RecordStreamClass, fileStatus, O_CREAT, mode, PersistentArrayClass);
		if (fileStatus != 0) {
			Console () << "Instrumentation log file " << fileName << "could not be found and could not be created.\n" << eor;
			_logDumpFile = NULL;
			return (CANTOPEN);
		}
		Console () << "New instrumentation log file " << fileName << ".\n" << eor;
		_loggingIsOn = TRUE;
		strncpy(_lastLogFile,fileName,sizeof(_lastLogFile));
		return (OKNEW);
	} else {
		Console () << "Instrumentation log appended to file " << fileName << ".\n" << eor;
		_loggingIsOn = TRUE;
		strncpy(_lastLogFile,fileName,sizeof(_lastLogFile));
		return (OKAPPEND);
	}
#else  /* iPSC2 */
	Console () << "Instrumentation logging not possible--no disks.\n" << eor;
	_loggingIsOn = FALSE;
	_lastLogFile[0] = '\0';
	return (CANTOPEN);  // Stub always fails
#endif
}

/*
Performs a logDump (all CPUs).
Closes the file.
Return FALSE if there is an error.
*/
Boolean 
SystemInstrumentManager :: closeLogDumpFile ()
{
#ifndef iPSC2
	Debug << this << "::closeLogDumpFile()\n" << eor;

	if (_logDumpFile == NULL) {
		Console () << "Log dump file doesn't exist or is not open!\n" << eor;
		Console().flush();
		return (FALSE);
	}
	
	dumpLogBuffer ();
	_loggingIsOn = FALSE;

	Assert ((_logDumpFile != NULL) && (_FSI != NULL));

  	_logDumpFile = NULL;
	
	_FSI->synchronize();

//  Bug with file system.	
//	_FSI = NULL;

	return (TRUE);
#else
	Debug << this << "::closeLogDumpFile()\n" << eor;
	Console () << "Log dump file doesn't exist or is not open!\n" << eor;
	return (FALSE);
#endif
}

/*
 *   copy name of file into buffer
 */
void
SystemInstrumentManager::lastLogDumpFile( char * name, int len )
{
int Len = len;
        if (Len > MaxInstrumentLogNameSize)  {
                Len = MaxInstrumentLogNameSize;
        }
        strncpy(name, _lastLogFile, Len);
        return;
}

Boolean
SystemInstrumentManager :: loggingIsOn()
{
		return(_loggingIsOn);
}

/*
Check if there is enough space left for the data and message.
If not, dump the log buffer.
Write the data and message to the buffer.
*/
void 
SystemInstrumentManager :: writeLogBuffer(unsigned int dataLength, char *data, 
	unsigned int eventID, unsigned int messageLength = 0, char *message = NULL)
{
	Debug << this << "::writeLogBuffer()\n" << eor;

	int intState = PROCESSOR::disableInterrupts();

	InstrumentBuffer *logBuffer = thisCPU()->_logBuffer;

	if ((logBuffer->getBufferSizeLeft()) <
	    (sizeof (timeType) + dataLength +
		sizeof (eventID) +
	     sizeof (messageLength) + messageLength))
		dumpLogBuffer ();
	
	logBuffer->timeStamp();

	logBuffer->append (dataLength, data);
	logBuffer->append (sizeof(eventID), (char *)&eventID);

	if (((iIdentifierType *) data)->message) {
		Assert (messageLength > 0);
		logBuffer->append (sizeof (messageLength),
				   (char *) &messageLength);
		logBuffer->append (messageLength, message);
	}

	PROCESSOR::restoreInterrupts(intState);
}

unsigned int 
SystemInstrumentManager :: numInstrumentedCPUs ()
{
	return (_numInstrumentedCPUs);
}

timeType 
SystemInstrumentManager :: time ()
{
	return (SystemTimer->time());
}

/*
Interrupts all processors and wait for them to acknowledge.
Everyone dumps their buffers to their appropriate file one at
a time after printing a header saying which CPU the following
data comes from -- this eliminates the need for putting the
CPU number anywhere in the logs.
All buffer variables are reset, even if there is no dump file.
Return FALSE if there is an error.
*/
Boolean
SystemInstrumentManager :: dumpLogBuffer ()
{
#ifndef iPSC2
	Debug << this << "::dumpLogBuffer()\n" << eor;

	if (_loggingIsOn == FALSE || _logDumpFile == NULL) {
		thisCPU()->_logBuffer->clear ();
		return (FALSE);
	}

	int intState = PROCESSOR::disableInterrupts();
	

	if ( iSystemInstrumentManagerDumpLogBuffer == NULL ) {
		Console() << this << ": instrument deleted\n" << eor;
		PROCESSOR::restoreInterrupts(intState);
		return(FALSE);
	}
	iSystemInstrumentManagerDumpLogBuffer->start();

	// Interrupt all instrumented processors, except itself
#ifdef MULTIMAX_TRASH
	thisCPU()->interruptAllCPUs (INSTRUMENT_Vector);
#endif

	Assert ( ! PROCESSOR::interruptsEnabled());

	// Remember that this procedure has ints disabled...
	extern InstrumentBufferInterrupt TheInstrumentBufferInterrupt;

	TheInstrumentBufferInterrupt.basicRaise (NULL);

	iSystemInstrumentManagerDumpLogBuffer->stop
		(SystemInstrumentManager_DumpLogBufferEvent);

	PROCESSOR::restoreInterrupts(intState);

	return (TRUE);
#else
	/*  iPSC2 */
	Debug << this << "::dumpLogBuffer()\n" << eor;
	Console () << "Log dump file is NULL!  Log Buffer lost.\n" << eor;
	thisCPU()->_logBuffer->clear ();
	return (FALSE);
#endif
}

void InstrumentBufferDumpProcessEntry (int arg)
{
	while (1) {
		Debug << "#InstrumentBufferDumpProcess::Entry()\n" << eor;

#ifndef iPSC2
		int bytesWritten;
		if ( SystemInstrumentManager::_loggingIsOn 
			&& SystemInstrumentManager::_logDumpFile != NULL) {
		bytesWritten =
			SystemInstrumentManager::_logDumpFile->write
				(thisCPU()->_logBuffer->getBufferLocation(),
				 thisCPU()->_logBuffer->getBufferSize());
		} else {
		bytesWritten = thisCPU()->_logBuffer->getBufferSize();
		}
#else
		int bytesWritten = thisCPU()->_logBuffer->getBufferSize();
#endif

		Debug << "#InstrumentBufferDumpProcess::LogDumpFileWrite("
			<< bytesWritten << ")\n" << eor;

		Assert (bytesWritten == thisCPU()->_logBuffer->getBufferSize());

		thisCPU()->_didDumpFlag = (int)TRUE;

		Assert (thisCPU()->idleContainer() == thisCPU()->_dumpContainer);

		Debug << "#InstrumentBufferDumpProcess::Exit()\n" << eor;

 		thisProcess()->ready();
		thisProcess()->giveProcessorTo
			(thisCPU()->_blockedByDumpProcess);
	}
}

InstrumentBufferInterrupt :: InstrumentBufferInterrupt ()
{
	_raised = NULL;
}

InstrumentBufferInterrupt :: ~InstrumentBufferInterrupt ()
{
	if (_raised != NULL) {
		delete /*[]*/ _raised;
	}
}

void 
InstrumentBufferInterrupt :: basicRaise (char * info)
{
	Debug << this << "::basicRaise()\n" << eor;
	Assert ( ! PROCESSOR::interruptsEnabled());

	if ( iInstrumentBufferInterrupt == NULL ) {
		Console() << this << ": instrument deleted\n" << eor;
		return;
	}
	// Record the time you got here (starts don't log)
	iInstrumentBufferInterrupt->start();

	// Wait for everyone else to get here
	// Syncronize appropriately (may not be MutEx on distributed machine)
	thisCPU()->syncWaitForEveryoneElse();

	// EVERYONE IS DOING THIS!!!
	// It is now safe to turn off instruments
	TheInstrumentManager->suspendAllInstruments ();

	// Wait for the lock (takes care of the cpu classes as well)
	thisCPU()->syncWaitMyTurn ();

	Debug << thisProcess() << eor;

	if (thisCPU()->_dumpContainer == NULL)
		thisCPU()->_dumpContainer = new FIFOScheduler();

	if (thisCPU()->_dumpProcess == NULL) {
		// I don't think the domain is right here, TheKernel->domain()?
		thisCPU()->_dumpProcess = new InterruptProcess (
			(PFV) InstrumentBufferDumpProcessEntry,
			TheKernel->domain(), normal, 0);
		thisCPU()->_dumpProcess->setName
			("#InstrumentManagerDumpProcess");
		thisCPU()->_dumpProcess->changeReadyContainer
			(thisCPU()->_dumpContainer);
	}
	else {
		// Get the dump process out of the container
		thisCPU()->_dumpContainer->remove();
		Assert (thisCPU()->_dumpContainer->isEmpty());
	}

	thisCPU()->_didDumpFlag = (int)FALSE;

	Debug << this << "::giveToDumpManager\n" << eor;

	// Give processor to dump manager
	ProcessContainer *normalCPUContainer = thisCPU()->idleContainer();
	thisCPU()->setIdleContainer (thisCPU()->_dumpContainer);
 	thisCPU()->_blockedByDumpProcess = thisProcess();
	thisProcess()->giveProcessorTo (thisCPU()->_dumpProcess);

	Debug << this << "::backFromDumpManager\n" << eor;

	// Take care of the funky if-idle-process stuff
	while ( ! thisCPU()->_didDumpFlag) {
		Assert (thisCPU()->idleContainer() == thisCPU()->_dumpContainer);
		Assert (thisProcess() == thisCPU()->_idleProcess);

		Debug << this << "::startFunkyInterruptable\n" << eor;


		// Here we go into the while loop
		PROCESSOR::enableInterrupts();

		while (thisCPU()->idleContainer()->isEmpty())
			;
		
		PROCESSOR::disableInterrupts();

		Debug << this << "::stopFunkyInterruptable\n" << eor;

		// Assert this again, just to make sure
		Assert (thisCPU()->idleContainer() == thisCPU()->_dumpContainer);
		
		if ( ! thisCPU()->_didDumpFlag) {
			Debug << this << "::tryItAgain\n" << eor;

			Process *readyProcess =
				thisCPU()->idleContainer()->remove();
			
			Assert (readyProcess == thisCPU()->_dumpProcess);
			

			// Give up the processor again
			thisProcess()->giveProcessorTo (readyProcess);
		}
	}

	// The dump process should now always be in that container
	Assert (!thisCPU()->_dumpContainer->isEmpty());

	thisCPU()->setIdleContainer (normalCPUContainer);
	
	thisCPU()->_logBuffer->clear ();

	// Release the lock (takes care of the halt stuff too)
	thisCPU()->syncWaitForEveryoneElse ();

	// It should now be safe to resume instrumenting
	TheInstrumentManager->resumeAllInstruments ();
	
	// Record the time you got here
	iInstrumentBufferInterrupt->stop(InstrumentBuffer_InterruptEvent);

	Debug << this << "::leavingBasicRaise\n" << eor;
}

void 
SystemInstrumentManager::suspendAllInstruments ()
{
        register Instrument **i = _instrumentList;

        for ( ; i != (_instrumentList + _numInstruments); i++) {
                if ((*i) != NULL)
                        (*i)->suspend ();
        }
}

void 
SystemInstrumentManager::resumeAllInstruments ()
{
        register Instrument **i = _instrumentList;

        for ( ; i != (_instrumentList + _numInstruments); i++) {
                if ((*i) != NULL)
                        (*i)->resume ();
        }
}

void 
SystemInstrumentManager::beginAllInstruments ()
{
        register Instrument **i = _instrumentList;

        for ( ; i != (_instrumentList + _numInstruments); i++) {
                if ((*i) != NULL)
                        (*i)->begin ();
        }
}

#include "ClassInstaller.h"

DEFINECLASS("SystemInstrumentManager",SystemInstrumentManager,InstrumentManager,
0,0);
#endif INSTRUMENT
