/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *     University of Illinois at Urbana-Champaign
 *     Department of Computer Science
 *     1304 W. Springfield Ave.
 *     Urbana, IL 61801
 *
 * Copyright (c) 1987-2005
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 */

#include "OutputStream.h"
#include "Process.h"
#include "CharacterInputStream.h"
#include "PeriodicTimer.h"
#include "PageTableConstants.h"

#ifdef __GNUC__
#pragma implementation
#endif
#include "AppInterface.h"
#include "ThisProcess.h"
#include "PersistentMemoryObject.h"
#include "SourceOfZerosMemoryObject.h"
#include "RootMemoryObjectCache.h"
#include "Kernel.h"
#include "Class.h"
#include <map>
#include <set>

class ObjLink;

std::multimap<Process*,ObjLink*> TheProcessObjMap;
AppInterface * TheAppInterface;

class ObjLink : public Link {
    void * obj;
    unsigned classID;
    unsigned numRef;
    std::set<Process *> processSet;
    public:
    ObjLink(void * o, unsigned id, Process* p, int ref=1) : obj(o), classID(id), numRef(ref)    {
        if (p == 0 )
            Assert(ref == 0);
        else
        {
            processSet.insert(p);
            Assert(numRef==processSet.size());
            TheProcessObjMap.insert(std::make_pair(p,this));
        }
    };
    ~ObjLink() {};
    void * getObj(){ return obj;}
    unsigned getClassID() { return classID; };
    unsigned refs(){ 
        Assert(numRef == processSet.size());
        return numRef; 
    }
    unsigned redRef(Process *p) {
        Assert(numRef == processSet.size());
        Assert(numRef>0);
        numRef--;
        processSet.erase(p);
        return numRef;
    }
    int compare ( Link * l ) { return obj == ((ObjLink*)l)->obj; };
    void inspect() { Console() << obj << eor; };
    int hash(int n) { return (unsigned)obj % n; }
};

//Dispatcher calls

#ifdef FRANCIS
//OutputStream
inline AppInterface::EXCODE OutputStreamDispatcher(HashTable * t, ObjLink * obj, unsigned methodID, void* inputParam, void* outputParam)
{
    OutputStream * p=0;
    if(methodID!=0){
        p=(OutputStream *)(obj->getObj());
        Assert(p->classOf() == OutputStreamClass);
    }
    switch(methodID){
        case 2://OutputStream & operator <<( int );
            (*p)<<*((int *)inputParam);
            return AppInterface::DONE;
        case 3://OutputStream & operator <<( unsigned int );
            (*p)<<*((unsigned int *)inputParam);
            return AppInterface::DONE;
        case 4://OutputStream & operator <<( long );
            (*p)<<*((long *)inputParam);
            return AppInterface::DONE;
        case 5://OutputStream & operator <<( unsigned long );
            (*p)<<*((unsigned long *)inputParam);
            return AppInterface::DONE;
        case 6://OutputStream & operator <<( char );
            (*p)<<*((char *)inputParam);
            return AppInterface::DONE;
        case 7://OutputStream & operator <<( const char * );
            (*p)<<*((char **)inputParam);
            return AppInterface::DONE;
        case 8://OutputStream & operator <<( const void * );
            (*p)<<*((void **)inputParam);
            return AppInterface::DONE;
    }
    return AppInterface::INVALIDMETHOD;
}

//InputStream
inline AppInterface::EXCODE InputStreamDispatcher(HashTable * t, ObjLink * obj, unsigned methodID, void* inputParam, void* outputParam)
{
    InputStream * p=0;
    if(methodID!=0){
        p=(InputStream *)(obj->getObj());
        //Assert(p->classOf() == InputStreamClass);
    }
    switch(methodID){
        case 2://InputStream & operator >>( int );
            p->formattedRead("%d", outputParam);
            return AppInterface::DONE;
    }
    return AppInterface::INVALIDMETHOD;
}
#endif

#include "AppInterfaceDispatcher.cc"

//AppInterface methods
AppInterface::AppInterface() :objTable(100) {}

AppInterface::EXCODE AppInterface::dispatch(unsigned classID, unsigned methodID, void * kernelObj, 
        void * inputParam, void *outputParam)
{
    Debug << "Calling Syscall dispatcher. ClassID: " << classID << " ,MethodID:  " << methodID 
        << " ,kernelObj: " << asHex((unsigned int)kernelObj)<< '\n' << eor;

    if(classID==0){ //AppInterface static calls
        ObjLink *f;
        PersistentMemoryObjectStar memObj;
        switch(methodID){
            case 0: //print
                Console() << (char*)inputParam << eor;
                break;
            case 1: //ConsoleOutput
            {
                OutputStream* out = &(Console());
                f = new ObjLink((void*)out,2,0, 0);
                if (objTable.find(f) == 0) 
                    objTable.insert(f);
                else
                    delete f;
                *((OutputStream **)outputParam) = out;
                break;
            }
            case 2: //ConsoleInput
            {
                InputStream* inp = ConsoleInputStream;
                f = new ObjLink((void*)inp,3,0,0);
                if (objTable.find(f) == 0) 
                    objTable.insert(f);
                else
                    delete f;
                *((InputStream **)outputParam) = inp;
                break;
            }
            case 3: //create heap
            {
                unsigned int nPages = *((unsigned int*)inputParam);

                MemoryObject *mo = new SourceOfZerosMemoryObject;
                mo->setName("HeapMo");
                mo->setNumberOfUnits(nPages*PhysicalPageSize);
                mo->setCacheClass( RootMemoryObjectCacheClass );
                MemoryObjectCache * cache = mo->buildCache();
                Assert (cache != 0);
                char *base = (char *)thisProcess()->domain()->add(mo, ReadWrite);
                Assert (base != 0);
                thisProcess()->domain()->makeAddressable(base, nPages*PhysicalPageSize);
                Assert( base != 0 );
                Debug << "Heap is at " << asHex(base) << "\n" << eor;
                *((char const **)outputParam) = base;
                break;
            }
            case 4: //return process
            {
                Process * p= thisProcess();
                f = new ObjLink((void*)p,5,0,0);
                if (objTable.find(f) == 0) 
                    objTable.insert(f);
                else
                    delete f;
                *((Process **)outputParam) = p;
                break;
            }
            case 5: //FileSystemInterface
            {
                FileSystemInterface * inter = TheKernel->fileSystemInterface();
                f = new ObjLink((void*)inter,19,0,0);
                if (objTable.find(f) == 0) 
                    objTable.insert(f);
                else
                    delete f;
                *((FileSystemInterface**)outputParam) = inter;
                break;
            }
            case 6: //theSystemClock
            {
                f = new ObjLink((void*)theSystemClock,22,0,0);
                if (objTable.find(f) == 0) 
                    objTable.insert(f);
                else
                    delete f;
                *((TimeOfDay**)outputParam) = theSystemClock;   
                break;  
            }
            case 7: //StandardSystemInterface
            {
                f = new ObjLink((void*)StandardSystemInterface,22,0,0);
                if (objTable.find(f) == 0) 
                    objTable.insert(f);
                else
                    delete f;
                *((SystemInterface**)outputParam) = StandardSystemInterface;    
                break;
            }   
            case 8: //StandardNameServer
            {
                f = new ObjLink((void*)StandardNameServer,22,0,0);
                if (objTable.find(f) == 0) 
                    objTable.insert(f);
                else
                    delete f;
                *((NameServer**)outputParam) = StandardNameServer;  
                break;  
            }
            default:
                return INVALIDMETHOD;
        }
    }
    else{  //dynamic objects
        ObjLink * obj=0;
        if(methodID != 0){
            ObjLink f(kernelObj,0, 0, 0);
            obj = (ObjLink *) (objTable.find(&f));
            if(obj == 0) {
                Debug <<"Invalid Object\n"<<eor;
                return INVALIDOBJECT;
            }
            /*if(obj->getClassID()!=classID) {
                Debug <<"Invalid Class\n"<<eor;
                return INVALIDCLASS;
            }*/
        }
#include "AppInterfaceSwitch.cc"
    }   
    return DONE;
};

//Cleanup
AppInterface::EXCODE AppInterface::cleanup(Process * p)
{
    std::multimap<Process*,ObjLink*>::iterator iter;
    Debug << "Entering process cleanup code " <<  TheProcessObjMap.size()  << "\n"<< eor;

    for (iter = TheProcessObjMap.begin(); iter !=  TheProcessObjMap.end();iter++)
    {
        Debug << "Process " << iter->first << " \n" << eor ;
        Debug << "ObjLink  " << iter->second->getClassID() << "\n" << eor ;
    }

    iter = TheProcessObjMap.find(p);
    if (iter == TheProcessObjMap.end() )
    {
        Debug << "No obj left for this process\n" << eor;
    }
    else {
        while( iter != TheProcessObjMap.end() )
        {
            Assert(iter->first == p);
            ObjLink * ol = iter->second;
            Assert(ol!=0);
            int ref = ol->redRef(p);
            if ( ref == 0)
            {
#include "AppInterfaceCleanup.cc"
            }
            else 
                Debug << "Remaining ref " << ref << " \n" << eor ;
            iter++;
        }
    }

    return DONE;

};
