#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * CPU.c: CPU class implementation.
 */

#include "AddressFlushQueue.h"
#include "AddressTranslator.h"
#include "Assert.h"
#include "CPU.h"
#include "ThisCPU.h"
#include "Process.h"
#include "ProcessContainer.h"
#include "Kernel.h"
#include "PROCESSOR.h"
#include "ThisProcess.h"

#ifdef INSTRUMENT
#include "SystemInstrumentManager.h"
#include "Vectors.h"
#include "PerCpuTimerInstrument.h"


#ifdef __GNUC__
#pragma implementation
#endif

extern SystemInstrumentManager *TheInstrumentManager;

cpuSyncType *CPU::_icpuSyncList;
unsigned int CPU::_highestInstrumentationID;
#endif

CPU * CPU::_cpus[MAX_PROCESSORS];
Process * volatile CPU::_runningProcesses[MAX_PROCESSORS];

InterruptManager *CPU::_interruptManager = NULL;

#ifdef INSTRUMENT_KERNEL
#define Instr(exp) exp
PerCpuTimerInstrument *iIdleProcessRunning = INULL;
static int idleProcessInstrumentReady = 0;
static int idleProcessInstrumentOK = 0;

void 
CPUConstructInstruments ()
{
	iIdleProcessRunning = new PerCpuTimerInstrument
		(TheInstrumentManager, "IdleProcessRunning", TALKING, 1, ALL);
	idleProcessInstrumentReady = 1;
}

void 
CPUDeleteInstruments ()
{
	delete iIdleProcessRunning;
	iIdleProcessRunning = NULL;
	idleProcessInstrumentReady = 0;
	idleProcessInstrumentOK = 0;
}

#else
#define Instr(exp)
#endif  //  INSTRUMENT_KERNEL

static NullProcessContainer idleProcessReadyContainer;

/*
 * Constructor for the CPU class. This is not always called by the processor
 * that the CPU object being constucted is to represent. The idleProcess() for
 * each processor will be do any "real" processor initialization. Make
 * sure CPU constructors don't rely on being able to address any per
 * processor registers.
 */
CPU::CPU( CpuId identifier, ProcessContainer * initialIdleContainer )
{
	Debug << this << "::CPU()\n" << eor;
	Assert( this != 0 );

	/*
	 * Tell the master CPU table about us
	 */
	_uniqueIdentifier = identifier;
	Assert (identifier < MAX_PROCESSORS)
			("CPU ID %d too large\n", identifier);
	Assert (_cpus[identifier] == 0);
	_cpus[identifier] = this;

	Assert( initialIdleContainer != 0 );
	_idleContainer = initialIdleContainer;
	_MMU = 0;	// subclass must set it!
	_timesliceTimer = 0; // ditto!

	_flushQueue = new AddressFlushQueue( _uniqueIdentifier );

#ifdef INSTRUMENT
	/*
	 * This is all kinda icky -- straighten it up sometime...
	 */
	static unsigned int currentInstrumentationID = 0;

	// Static member functions have to be initialized
	if (currentInstrumentationID == 0) {
		// Seems like a race, but all constructors are on the same CPU

		Console() << ">> This is an instrumented operating system! <<\n"			<< eor;

		_icpuSyncList = new cpuSyncType[MAX_PROCESSORS];

		register unsigned int i;
		for (i = 0; i < MAX_PROCESSORS; i++) {
			_icpuSyncList[i].used = FALSE;
		}
	}

	_icpuSyncList[currentInstrumentationID].used = TRUE;
	_icpuSyncList[currentInstrumentationID].done = FALSE;
	_icpuSyncList[currentInstrumentationID].cpu = _uniqueIdentifier;

	_icpuID = currentInstrumentationID;

	_logBuffer = new InstrumentBuffer(4000, _uniqueIdentifier);

	_dumpProcess = NULL;
	_blockedByDumpProcess = NULL;
	_dumpContainer = NULL;

	_highestInstrumentationID = currentInstrumentationID++;
#endif  //  INSTRUMENT

	/*
	 * Make the idle process now.
	 */
	// pass the initial domain in as an argument?
	Assert( TheKernel != 0 );
	Assert( TheKernel->domain() != 0 );

	// NOTE: TheKernel->domain() exists at this point but contains no
	// memory (MemoryObjects). This is ok because we wont use it until
	// it does.
	
	Debug << this << "::CPU() creating idle process\n" << eor;
	Assert (TheKernel != 0);
	_idleProcess = new PerCPUIdleProcess( &CPU::idleLoop,
				TheKernel->domain(), heapBootStack, 0 );
	_idleProcess->setName( "#IdleProcess" );

	/*
	 * The idleProcess is never stuck in the ready queue, only in a special
	 * process container. Be sure any processes built by the idleProcess
	 * are given a new ready container (there shouldnt be any anyway.)
	 */
	_idleProcess->changeReadyContainer( &idleProcessReadyContainer );

	/*
	 * The _currentProcess is not initialized until immediately
	 * before the first process (the idleProcess) is actually
	 * dispatched. 
	 *
	 * This is initialized by processorStart(), immediately
	 * before the idle process is started for the first time.
	 */
	Debug << this << "::CPU() done\n" << eor;
}

/*
 * This is the code which all idle processors run.
 */
void
CPU::idleLoop()
{
	Debug << "idleLoop started\n" << eor;
	PROCESSOR::disableInterrupts();

	Process * process = thisCPU()->idleProcess();
	Assert( process != 0 );

	/*
	 * Initialize the processor chip, timers, interrupt tables, etc.
	 * Implementations of this method should be careful to avoid
	 * allocating any new object as VM has yet to be enabled.
	 */
	CPU * cpu = thisCPU();
	Assert (cpu != 0);
	cpu->chipInitialize();

	/*
	 * Activate the proper domain on this processor.
	 */
	cpu->MMU()->activate( process->domain());

	/*
	 * Turn on the MMU()
	 */
	cpu->MMU()->enable();

	while( 1 ) {

		Debug << "idleLoop running\n" << eor;
		
		/*
		 * Keep looping (with interrupts on) waiting for something
		 * to do next.
		 * From now on we can potentially be interrupted and
		 * the process that interrupted us will
		 * put us on our readyContainer (the idleProcessReadyContainer).
		 * This is, in effect, a noop, since the idleProcess always
		 * needs explicitly restarted.
		 */
		Process * newProcess = 0;
		while( newProcess == 0 ) {
			PROCESSOR::enableInterrupts();
			Assert( PROCESSOR::interruptsEnabled() );
			
			/*
			 * If there is nothing else to do (an interrupts),
			 * keep looking at the cpu's idleContainter for jobs to
			 * run. If it looks like there is one there, try
			 * to grab it. (This reduces bus contention on the
			 * container.) Once there is something else to
			 * run, dispatch it.
			 */
			while( cpu->_idleContainer->isEmpty() ) {
				cpu->idleFunction();// Null or overloaded
				Assert( PROCESSOR::interruptsEnabled() );
			}
			
			/*
 			 * Make sure interrupts are off if the idleContainer()
			 * is ever removed from.
			 */
			PROCESSOR::disableInterrupts();
			newProcess = cpu->_idleContainer->remove();
		}
		Assert( newProcess != 0 );
		Assert( ! PROCESSOR::interruptsEnabled() );
		
		Debug << "idleLoop: resume " << newProcess << "\n" << eor;
		
#ifdef INSTRUMENT_KERNEL
		if (idleProcessInstrumentOK) {
			iIdleProcessRunning->stop (CPU_IdleLoopEvent,"IdleLoop");
		}
#endif

		/*
		 * This process will be destroyed by the process that it is
		 * dispatching in ::handleWhoDispatchedUs().
	 	 */
		/* Surprise:  The idle processes are never removed from
		 * the idleProcessReadyContainer so there is no need to do a 
		 * ready						    */    
		process->giveProcessorTo( newProcess, ProcessReady );
		Debug << "idleLoop resumed\n" << eor;

#ifdef INSTRUMENT_KERNEL
		if (idleProcessInstrumentOK) {
			iIdleProcessRunning->start ();
		} else {
			if (iIdleProcessRunning != NULL 
				&& idleProcessInstrumentReady) {
				iIdleProcessRunning->start ();
				idleProcessInstrumentOK = 1;
			}
		}
#endif

	}
}


#ifdef INSTRUMENT
	
void CPU :: interruptAllCPUs (unsigned int vector)
{
#ifdef MULTIMAX_TRASH
	unsigned int i;
	for (i = 0; i < MAX_PROCESSORS; i++) {
		if (_icpuSyncList[i].used && i != _icpuID) {
			TransmitDirectedVector (vector, _icpuSyncList[i].cpu);
		}
	}
#endif
}


void CPU :: syncWaitMyTurn ()
{
#ifdef MULTIMAX_TRASH
	Boolean doneFlag = (_icpuSyncList[_icpuID].done ? FALSE : TRUE);
	
	register unsigned int i;

	for (i = 0; i < _icpuID; i++) {
		if (i == 0) {
			SetCPUVectorClass (CPUCLASS_HALTED);
		}
		if (_icpuSyncList[i].used) {
			while (_icpuSyncList[i].done != doneFlag)
				;
		}
	}
	if (i != 0) {
		SetCPUVectorClass (CPUCLASS_RUNNING);
	}
#endif
}


void CPU :: syncWaitForEveryoneElse ()
{
#ifdef MULTIMAX_TRASH
	Boolean doneFlag = (_icpuSyncList[_icpuID].done ? FALSE : TRUE);

	_icpuSyncList[_icpuID].done = doneFlag;

	// Should really check if it's the last processor
	if (_icpuID != _highestInstrumentationID) {
		SetCPUVectorClass (CPUCLASS_HALTED);
	}

	unsigned int i;
	for (i = _icpuID + 1; i < MAX_PROCESSORS; i++) {
		if (_icpuSyncList[i].used) {
			while (_icpuSyncList[i].done != doneFlag)
				;
		}
	}

	// Should really check if it's the last processor
	SetCPUVectorClass (CPUCLASS_RUNNING);
#endif
}
#endif /*INSTRUMENT*/


CPU::~CPU()
{
	Assert( NOTREACHED );
}

void
CPU::flushAddressTranslationCache( const char * virtualAddress,
	unsigned int length )
{

	IntStorage state ; PROCESSOR::disableInterrupts(state);
	/*
	 * If "this" is us, then we can flush the entry ourself.
	 * Otherwise, we use the AddressFlushQueue.
	 */
	if( thisCPU() == this ) {
		Debug << this << "flush( " << asAddress( virtualAddress )
			<< " ) (local).\n" << eor;
		MMU()->flushCache( virtualAddress, length );
	}
	else {
		Debug << this << "flush( " << asAddress( virtualAddress )
			<< " ) (remote).\n" << eor;
		flushQueue()->add( virtualAddress, length );
	}
	PROCESSOR::restoreInterrupts(state);
}


void
CPU::processorStart()
{
	Assert (thisCPU() != 0)("thisCPU() for cpuID %d == 0!\n",
		PROCESSOR::cpuID());
	Assert (! PROCESSOR::interruptsEnabled() );
	CPU::_runningProcesses[PROCESSOR::cpuID()] = thisCPU()->idleProcess();
	thisCPU()->idleProcess()->context()->restoreFromInitialContext(0);
	Assert( NOTREACHED );
}


#include "ClassInstaller.h"
DEFINECLASS("CPU",CPU,ProxiableObject,0,0);
