#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Contributing Author: Lee Lup Yuen (lup@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * CharacterQueue.c: character queue implementations.
 */

#include "Assert.h"
#include "CharacterQueue.h"
#include "OutputStream.h"
#include "Semaphore.h"


#ifdef __GNUC__
#pragma implementation
#endif


CharacterQueue::CharacterQueue( int size )
{
	Assert( size > 0 );
	_size = size;
	_free = _size;
	_queue = new char [ size ];
	flush();
}

CharacterQueue::~CharacterQueue()
{
	flush();
	delete _queue;
}

/*
 * Note that CharacterQueue should not be instantiated.  A derived Class
 * must provide locking.
 */

LineBufferedEchoingCharacterQueue::LineBufferedEchoingCharacterQueue
	( OutputStream * echoStream, int size ) : CharacterQueue( size )
{
	Assert( echoStream != 0 );
	_echoStream = echoStream;

	_front = _back = 0;
	_SpecialChar[ Quote ] = '\\';
	_SpecialChar[ Backspace ] = '\177';	// Delete
	_SpecialChar[ DeleteWord ] = '\027';	// Ctrl-W
	_SpecialChar[ DeleteLine ] = '\025';	// Ctrl-U
	_lineAvailable = 0;
	_requestingLine = 0;
	_terminated = 0;

	//_lock = new Semaphore( 1 );
	//_lock->setName("#LineBufferedEchoingCharacterQueueLock");
	_lineSemaphore = new Semaphore( 0 );
	_lineSemaphore->setName("#LineBufferedEchoingCharacterQueueSemaphore");
}

void
CharacterQueue::flush()
{
	_front = _back = 0;
}

void
CharacterQueue::dump()
{
	Console() << "\n" << this << "::dump()\n"
		<< "\t_size: " << _size << ", _free: " << _free << "\n"
		<< "\t_queue: " << asAddress( _queue )
		<< ", _front: " << _front << ", _back: " << _back << "\n";

	Console() << "\tcontents: '";
	int count = _size - _free;
	int p = _front;
	while( count-- > 0 ) {
		Console() << _queue[p];
		p = (p + 1) % _size;
	}
	Console() << "'\n";
}

int
CharacterQueue::basicEnqueue( const char * buffer, int count )
{
	/*
	 * This could be done with one or two ByteCopy's (since this is a
	 * circular buffer).
	 */
	int charsEnqueued = 0;
	while( count-- && !full() ) {
		_queue[_back] = *buffer++;
		_back = (_back + 1) % _size;
		_free--;
		charsEnqueued++;
	}
	Assert( _free >= 0 );
	Assert( _free <= _size );
	Assert( _back >= 0 );
	Assert( _back < _size );
	return( charsEnqueued );
}

int
CharacterQueue::basicDequeue( char * buffer, int count )
{
	int charsDequeued = 0;
	while( count-- && !empty() ) {
		*buffer++ = _queue[_front];
		_front = (_front + 1) % _size;
		_free++;
		charsDequeued++;
	}
	Assert( _free >= 0 );
	Assert( _free <= _size );
	Assert( _back >= 0 );
	Assert( _back < _size );
	return( charsDequeued );
}

//  BlockCharacterQueues are for dequeueing blocks of characters.  Calls
//  to dequeue will block until the requested number of characters are
//  fully available, unless somebody has previously called terminate()
//  to indicate the end of the queue.

BlockCharacterQueue :: BlockCharacterQueue (int size) : CharacterQueue(size)
{
	//_lock = new Semaphore( 1 );
	//_lock->setName("#BlockCharacterQueueLock");
    	_charBlockSemaphore = new Semaphore (0);
	_charBlockSemaphore -> setName ("#BlockCharacterQueueSemaphore");
    	_charsAvailable = 0;
	_charsRequested = 0;
	_terminated = 0;
}

void
BlockCharacterQueue::terminate()
{
    	//  Indicate that there will be no more characters added to
	//  the queue.  If anybody makes a dequeue request, don't block
	//  them.
	lock(); //_lock -> P ();
    	_terminated = 1;
	if (_charsRequested > 0)
	{
	    unlock(); //_lock -> V ();
	    _charBlockSemaphore -> V ();
	}
	else unlock(); //_lock -> V ();
}

int
BlockCharacterQueue::enqueue( const char * buffer, int count )
{
    	int shouldDequeueChars = 0;
	lock(); //_lock -> P ();
	int charsEnqueued = CharacterQueue::basicEnqueue( buffer, count );
	_charsAvailable = _charsAvailable + charsEnqueued;
	//  Find out whether we can satisfy the pending request for a block
	//  of characters.
	if (_charsRequested != 0 && (_charsAvailable >= _charsRequested ||
	    _terminated))
	{
	    shouldDequeueChars = 1;
	    _charsRequested = 0;
	}
	unlock(); //_lock -> V ();

	//  If there is a request, and we can satisfy the request, then
	//  unblock it.
	if (shouldDequeueChars)
	    _charBlockSemaphore -> V ();
	Assert( charsEnqueued <= count );
	return( charsEnqueued );
}

int
BlockCharacterQueue::dequeue( char * buffer, int count )
{
    	if (count == 0) return 0;
	int madeRequest = 0;
	
	lock(); //_lock -> P ();
	if (!_terminated)
	{
	    //  If we need more characters than are available in the queue,
	    //  then make a request for the characters.
	    if (count > _charsAvailable)
	    {
	    	_charsRequested = count;
	    	madeRequest = 1;
    	    }
	}
	unlock(); //_lock -> V ();

	//  If we need more characters than what the queue contains, then
	//  block until our request can be satisfied.
	if (madeRequest)
	    _charBlockSemaphore -> P ();
	Assert (_charsRequested == 0 || _terminated);

	//  The queue has sufficient characters to satisfy our request.
	//  Just dequeue the required number of characters.
	lock(); //_lock -> P ();
	int charsDequeued = CharacterQueue::basicDequeue( buffer, count );
	_charsAvailable = _charsAvailable - charsDequeued;
	unlock(); //_lock -> V ();
	Assert( charsDequeued == count || _terminated );
	return( charsDequeued );
}

#ifdef NEVER
void
BlockCharacterQueue::flush()
{
	_lock -> P ();
	_lock -> V ();
}
#endif


/*
 * The RawCharacterQueue functions merely provide locking before calling
 * the CharacterQueue functions.
 */

RawCharacterQueue::RawCharacterQueue( int size ) : CharacterQueue( size )
{
	Debug << this << "::RawCharacterQueue()\n" << eor;
	///_lock = new Semaphore( 1 );
	///_lock->setName("#RawCharacterQueueLock");
}

int
RawCharacterQueue::enqueue( const char * buffer, int count )
{
	lock(); //_lock -> P();
	int charsEnqueued = CharacterQueue::basicEnqueue( buffer, count );
	unlock(); //_lock -> V();
	Assert( charsEnqueued <= count );
	return( charsEnqueued );
}

int
RawCharacterQueue::dequeue( char * buffer, int limit )
{
	lock(); //_lock -> P();
	int charsDequeued = CharacterQueue::basicDequeue( buffer, limit );
	unlock(); //_lock -> V();
	Assert( charsDequeued <= limit );
	return( charsDequeued );
}

void
RawCharacterQueue::flush()
{
	lock(); //_lock -> P();
	CharacterQueue::flush();
	unlock(); //_lock -> V();
}


int
LineBufferedEchoingCharacterQueue::setSpecialChar( int CharNum, char ch )
{
	if ( CharNum >= SpecialCharCount )
		return(-1);

	_SpecialChar[ CharNum ] = ch;
	return(0);
}

int
LineBufferedEchoingCharacterQueue::unqueue( int n )
{
	int count = 0;
	while( n-- > 0 && !empty() ) {
		if( tail() == '\n' ) {
			/* Can't unqueue past a newline! */
			break;
		} else {
			count++;
			_free++;
			_back--;
			if( _back < 0 ) {
				_back = _size - 1;
			}
		}
	}
	Assert( _free >= 0 );
	Assert( _free <= _size );
	Assert( _back >= 0 );
	Assert( _back < _size );
	return( count );
}

int
LineBufferedEchoingCharacterQueue::unqueueWord()
{
	static const int START = 0;	// Nothing or only white-space seen
	static const int CHAR = 1;	// A non white-space char has been seen

	int charsUnqueued = 0;
	int state = START;

	for( ;; ) {
		if( state == START ) {
			if( empty() || tail() == '\n' ) {
				break;
			} else if( tailIsWS() ) {
				charsUnqueued += unqueue( 1 );
			} else {
				state = CHAR;
				charsUnqueued += unqueue( 1 );
			}
		} else if( state == CHAR ) {
			if( empty() || tailIsWS() ) {
				break;
			} else {
				charsUnqueued +=  unqueue( 1 );
			}
		}
	}

	return( charsUnqueued );
}

void
LineBufferedEchoingCharacterQueue::terminate()
{
    	//  Indicate that there will be no more characters added to
	//  the queue.  If anybody makes a dequeue request, don't block
	//  them.
	lock(); //_lock -> P ();
    	_terminated = 1;
	if (_requestingLine)
	{
	    unlock(); //_lock -> V ();
	    _lineSemaphore -> V ();
	}
	else unlock(); //_lock -> V ();
}

/*
 * This should be rewritten sometime, though I guess it can never
 * enqueue the entire buffer at once because it might contain
 * editing characters.
 */
int
LineBufferedEchoingCharacterQueue::enqueue( const char * buffer, int count )
{
	int totalCharsEnqueued = 0;
	int charsToEnqueue = count;
	while( charsToEnqueue-- > 0 ) {
		char c = *buffer++;
		// Translate CR to LF
		if( c == '\r' ) {
			c = '\n';
		}

		/* Some characters exist only to control the output */

		if(( c == '\001' ) || ( c == '\023' )) {   
		     /* Ctrl-A     OR    CTRL-S  */
			_echoStream->write( &c, 1 );
			_echoStream->flush();
			return 0;
		}

		unsigned charsEnqueued = 0, charsToErase = 0;
		lock(); //_lock -> P ();
		if( c == _SpecialChar[Backspace] ) {
			charsToErase = unqueue( 1 );
		} else if( c == _SpecialChar[DeleteWord] ) {
			charsToErase = unqueueWord();
		} else if( c == _SpecialChar[DeleteLine] ) {
			charsToErase = unqueue( _size );
		} else {
			charsEnqueued = CharacterQueue::basicEnqueue( &c, 1 );
			totalCharsEnqueued += charsEnqueued;
		}
		unlock(); //_lock -> V ();

		if( charsToErase > 0 ) {
			for( unsigned int i = 0; i < charsToErase; i++ ) {
				_echoStream->write( "\b \b", 3 );
				_echoStream->flush();
			}
		} else if( c == '\n' ) {
			_echoStream->write( &c, 1 );
			_echoStream->flush();
			_lineSemaphore -> V();
		} else {
			_echoStream->write( &c, 1 );
			_echoStream->flush();
		}
	}
	Assert( totalCharsEnqueued <= count );
	return( totalCharsEnqueued );
}

int
LineBufferedEchoingCharacterQueue::dequeue( char * buffer, int limit )
{
	/*
	 * Dequeue at most one newline-terminated line, even if more
	 * data is available.  To do this, we have to dequeue a single
	 * character at a time.
	 */

	if (!_lineAvailable)
	{
	    //  No characters were left over from the last call to dequeue,
	    //  and we need more characters, so block until a full line is
	    //  available.
	    int shouldBlock = 0;
	    lock(); ///_lock -> P ();
	    if (!_terminated)
	    {
	    	_requestingLine = 1;
				shouldBlock = 1;
	    }
	    unlock(); //_lock -> V ();
	    if (shouldBlock) _lineSemaphore -> P();
	}

        //  Fetch characters from the queue, stopping if we see a newline,
	//  if we've enough characters, or if there are no more characters
	//  to read.
	lock(); //_lock -> P ();
	_requestingLine = 0;
	char c = '\0';
	int charsDequeued = 0;
	while( (charsDequeued < limit) && !empty() && (c != '\n') ) {
		CharacterQueue::basicDequeue( &c, 1 );
		*buffer++ = c;
		charsDequeued++;
	}
	unlock(); //_lock -> V ();
	if (limit > 0 && c == '\n')
	    //  If we have used up a newline, then the whole line has been
	    //  absorbed.
	    _lineAvailable = 0;
	else
	    //  There are some more characters left on this line.  Indicate
	    //  this fact so that we can use them during the next call to
	    //  dequeue.
	    _lineAvailable = 1;

	Assert( charsDequeued <= limit );
	return( charsDequeued );
}

#ifdef NEVER
void
LineBufferedEchoingCharacterQueue::flush()
{
	/*
	 * If we actually flush it, the semaphore will be out of
	 * sync with the actual queue, and we'll be *really* messed up.
	 */

	Console() << "Why are you flushing a LineBufferedCharacterQueue?\n"
		<< eor;
	_lock -> P();
	_lock -> V();
}
#endif

int
CharacterQueue::tailIsWS()
{
      if( !empty() && (tail() == ' ' || tail() == '\t' || tail() == '\n') ) {
              return( 1 );
      }
      return( 0 );
}

#include "ClassInstaller.h"
DEFINECLASS("CharacterQueue",CharacterQueue,Object,0,0);
