#ifdef __GNUC__
#pragma implementation
#endif

#include "Assert.h"
#include "DDriver.h"
#include "CPU.h"
#include "ThisProcess.h"
#include "ThisCPU.h"
#include "Process.h"

#include "Byte.h"
#include "Domain.h"
#include "Kernel.h"
#include "MemoryObject.h"
#include "MemoryObjectView.h"
#include "OutputStream.h"
#include "RootMemoryObjectCache.h"
#include "ARMContext.h"
#include "ARMInterrupts.h"
#include "ARMPageTableConstants.h"
#include "TwoLevelPageTable.h"
#include "ARMPageTableEntry.h"
#include "SourceOfZerosMemoryObject.h"
#include "ThisProcess.h"
#include "Process.h"
#include "ADDRESS_TRANSLATION.h"
#include "VirtualMemoryRange.h"
#include "MadAllocator.h"
#include "StupidAllocator.h"
#include "DummyMemoryObject.h"

DDriver * DDriver::currentDriver_;

extern Allocator * SetNewAllocator( Allocator * );

DDriver::DDriver () 
{   
    domain_ = NULL;
}
  
void DDriver::init() {

    if (domain_ != NULL) return;
  
    domain_ = new Domain(TheKernel->domain(), (char*)TheKernel->kernelStart(),TheKernel->kernelSize());
    domain_->setName("DDriver");

    allocateStack();
    allocateHeap();
  
    AddressTranslation* translation = domain_->translation();

    // Francis: TODO Fix Hopefully, all kernel stuff we need is marked ReadOnly
    VirtualMemoryRange* kernelVirtualMemoryMap = ADDRESS_TRANSLATION::KernelVirtualMemoryMap();

    for (VirtualMemoryRange * region = kernelVirtualMemoryMap; region != 0;
         region = region->next() ) {
        char * base = (char *) region->virtualAddress();
        int length = region->numberOfUnits() * region->unitSize();
        Debug << "Changing protection to RO base : " << asAddress(base) << " length: " << length << "\n" << eor;
        translation->changeProtection(base, length, ReadOnly);
    }

    basicInit();

    translation->changeProtection(thisProcess()->context()->kernelStackBase(), 
                                  thisProcess()->context()->kernelStackSize(), ReadWrite);
    translation->changeProtection(stackBase, stackPages * PhysicalPageSize, ReadWrite);
    translation->changeProtection(heapBase, heapPages * PhysicalPageSize, ReadWrite);
    translation->changeProtection(allocAddress_, allocPages_ * PhysicalPageSize , ReadWrite);
    translation->changeProtection((char*) heap_, PhysicalPageSize , ReadWrite);
#ifdef OMAP
    ((ARMTranslation *)translation)->inhibitDataCaching((char *)0xfff00000,0xf0000);
#endif
}

DDriver:: ~DDriver(){
    if(stackMemoryObject !=0){
        TheKernel->domain()->remove(stackMemoryObject, stackBase);
        //    TheKernel->domain()->remove(allocMemoryObject_, allocAddress_);
        //    TheKernel->domain()->remove(heapMemoryObject, heapBase);

        stackMemoryObject = (MemoryObject*)0;
        //heapMemoryObject = (MemoryObject*)0;
        //allocMemoryObject_ = (MemoryObject*)0;
    }
}


void DDriver::switchToDriverDomain () {

    ARMCPU * cpu = static_cast<ARMCPU*>(thisCPU());
  
    Assert(cpu->getMode() == ARM_SYS);

    Debug << "About to switch to driver\n" << eor;

    //this should rather be an assertion
  
    oldDomain_ = thisProcess()->domain();
    thisProcess()->setCurrentDomain(domain_);

    //JG: I commented this because it is not removed when switching back. 
    // I tried to remove it but cause some problems. 
    /*const char *vaddr = (const char*) thisProcess()->context()->kernelStackBase();
      const char* base;
      ProtectionLevel prot = ReadOnly;
      unsigned size = thisProcess()->context()->kernelStackSize();
      MemoryObjectViewRef stack = oldDomain_->lookup(vaddr,
      base,
      size,
      prot);
      thisProcessStack_ = stack->source();
      thisProcessStackStackBase= domain_->add(stack->source(), ReadOnly);
    */                        

    // TODO TOFIX This changes the allocator for everyone!!!
    oldHeap_ = SetNewAllocator(heap_);
  
    if( cpu->MMU()->currentlyActiveDomain() != domain_ ) {
        // Activate new domain
        cpu->MMU()->activate (domain_);
    }      

    //DDriver::currentDriver_ = this;
    cpu->setMode(ARM_USR);

    extern Allocator * GetNewAllocator();
    Assert(GetNewAllocator() == heap_);

#ifdef DEBUG
    GenericARMSerialPutS("Switched to user mode\n"); 
#endif
}

void DDriver::switchToKernelDomain () {
#ifdef DEBUG
    GenericARMSerialPutS("About to switch back to kernel mode\n");
#endif
    
    asm( "swi #0xfff" );  
  
    Assert (oldDomain_);

    // Restore domain
    thisProcess()->setCurrentDomain(oldDomain_);

    // Activate new domain
    CPU * cpu = thisCPU();
  
    //this should rather be an assertion!
    if( cpu->MMU()->currentlyActiveDomain() != oldDomain_) {
        cpu->MMU()->activate(oldDomain_);
    }    

    //DDriver::currentDriver_ = NULL;
  
    SetNewAllocator( oldHeap_ );
}


/*
 * Allocate a (non-paged and pre-allocated) stack of size into domain.
 * return the value to use as the stackpointer (stack grows down on ARM)
 */
void DDriver::allocateStack()
{
    stackPages = 4;
    Assert (TheKernel->domain() != 0);
    stackBase = 0;
    unsigned stackBytes = stackPages*PhysicalPageSize;

    ProtectionLevel prot =  NoAccess;
    stackMemoryObject = new SourceOfZerosMemoryObject;
    stackMemoryObject->setName("StackMo");
    stackMemoryObject->setNumberOfUnits(stackPages*PhysicalPageSize);
    stackMemoryObject->setCacheClass( RootMemoryObjectCacheClass );
    MemoryObjectCache * cache = stackMemoryObject->buildCache();
    Assert (cache != 0);
    stackBase = (char *) TheKernel->domain()->add(stackMemoryObject, prot);
    Assert (stackBase != 0);

    // The stack pointer should be word aligned
    newStackPointer_ = (char*) ((unsigned)stackBase + stackBytes);

    if ((unsigned)newStackPointer_ & 0x3) {
        unsigned aligned;
        aligned = (unsigned)newStackPointer_ & ~0x3;
        stackBytes -= ((unsigned)newStackPointer_ - aligned);
        newStackPointer_ = (char *)aligned;
    }

    TheKernel->domain()->makeAddressable(stackBase, stackBytes);

    Debug << " Stack Size " << stackBytes <<  " @ base " << asAddress(stackBase) << "\n" << eor;
}

/*
 * Allocate a (non-paged and pre-allocated) stack of size into domain.
 * return the value to use as the stackpointer (stack grows down on ARM)
 */
void DDriver::allocateHeap()
{
//allocPages_ = (int) ceil ((sizeof(MadAllocator) + allocatorStateSize) / PhysicalPageSize);
    allocPages_ = 2;
    allocMemoryObject_ = new RootMemoryObject(allocPages_, Log2PhysicalPageSize);
    allocAddress_ = (char *) TheKernel->domain()->add (allocMemoryObject_, NoAccess);  
    TheKernel->domain()->makeAddressable(allocAddress_, allocPages_ *PhysicalPageSize);
    Assert( allocMemoryObject_ != 0 );
  
    heapPages = 5;
  
    heapMemoryObject = new RootMemoryObject(heapPages, Log2PhysicalPageSize );
    heapBase = (char *)  TheKernel->domain()->add (heapMemoryObject, NoAccess);  
    TheKernel->domain()->makeAddressable(heapBase, heapPages * PhysicalPageSize); 
    Assert( heapMemoryObject != 0 );

#ifdef FRANCIS // MadAllocator isn't designed to work for these small components
    int allocatorStateSize = MadAllocator::stateSize(heapMemoryObject->numberOfUnits() );
  
    void * allocatorStateAddress = allocAddress_ + sizeof(MadAllocator);   /* GB: This is too high ~ 5832 bytes*/

    void *endBootHeap =  heapBase + heapPages * PhysicalPageSize;
    void * startBootHeap = heapBase ; 

    Console() << "heap stat: " << allocPages_ << " " << sizeof(MadAllocator) << " " << allocatorStateSize << "\n" << eor;
    Console() << "heap stat: " << asAddress(endBootHeap) << " " << asAddress(startBootHeap) << " \n" << eor;
    Console() << "heap stat: " <<  heapMemoryObject->numberOfUnits() << " " <<  heapMemoryObject->log2UnitSize()<< " \n" << eor;
  
    heap_ = new (allocAddress_) MadAllocator (allocatorStateAddress, heapBase,
                                              heapMemoryObject->numberOfUnits(),
                                              heapMemoryObject->log2UnitSize(), domain_,
                                              startBootHeap, endBootHeap, 1 );
#endif
    heap_ = new (allocAddress_) StupidAllocator(heapBase);

    Debug << "Heap Address: " << asAddress(heapBase) <<"\n" << eor;
}


DDriver * DDriver::curDriver()
{
    return currentDriver_;
}
