#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#include <ctype.h>
#include "Gdb.h"
#include "OutputStream.h"
#include "String.h"


#ifdef __GNUC__
#pragma implementation
#endif

const char Gdb::hexchars[] = "0123456789abcdef";
Gdb *Gdb::instance = 0;

/*
 * scanhex and scan2hex are replacements for sscanf'ing hex values
 * so libc.a is not needed.
 */
int Gdb::scanhex(char *p, unsigned *var)
{
    char upchar;
    int result = 0;
    int value = 0;
    while (isxdigit(*p)) {
        result++;
        if (isdigit(*p)) {
            value = 16 * value + (*p - '0');
        }
        else {
            if (islower(*p))
                upchar = (*p - 'a') + 'A';
            else
                upchar = *p;

            value = 16 * value + (upchar - 'A' + 10);
        }
        p++;
    }
    *var = value;
    return(result);
}

int Gdb::scan2hex(char *p, unsigned *var1, unsigned *var2)
{
    int result;

    if ((result = scanhex(p, var1)) == 0)
        return(0);
    if (*(p+result++) != ',')
        return(0);
    if (scanhex((p+result), var2) == 0)
        return(0);
    return(2);
}

char Gdb::hex(char ch)
{
    if ((ch >= 'a') && (ch <= 'f')) return (ch-'a'+10);
    if ((ch >= '0') && (ch <= '9')) return (ch-'0');
    return (0);
}

void Gdb::GdbGetPacket(char *buffer)
{
    unsigned char checksum;
    unsigned char xmitcsum;
    char *current;
    int  i;
    int  count;
    char ch;

    do{
        /* Wait for the start char */
        while ( (GetDebugChar()) != '$' ) ;
        checksum = 0;
        count = 0;
        current = buffer;

        while ( (ch = GetDebugChar()) != '#') {
            *current++ = ch;
            checksum += ch;
            count++;
        }
        *current = '\0';

        xmitcsum = hex(GetDebugChar()) << 4;
        xmitcsum |= hex(GetDebugChar());
        if (xmitcsum != checksum) {
            PutDebugChar('-');  /* failed checksum */
            Console().printf("GdbGetPacket: checksum error! 0x%x != 0x%x\n",
                             checksum, xmitcsum);
            Console() << "Packet: " << buffer << "\n" << eor;

            Debug << asHex(xmitcsum) << " != " << asHex(checksum) << "\n" << eor;
            Debug << "Packet: " << buffer << "\n" << eor;
        }   
        else {
            PutDebugChar('+');  /* successful transfer */
            /* if a sequence char is present, reply the sequence ID */
            if (buffer[2] == ':') {
                PutDebugChar( buffer[0] );
                PutDebugChar( buffer[1] );
                /* remove sequence chars from buffer */
                count = strlen(buffer);
                for (i=3; i <= count; i++) buffer[i-3] = buffer[i];
            }
        }
    } while (checksum != xmitcsum);

    //     Debug << "GdbGetPacket: got packet: " << buffer << "\n" <<eor;

}


/* send the packet in buffer.  The host get's one chance to read it.  
   This routine does not wait for a positive acknowledge.  */
void Gdb::GdbPutPacket(char *buffer)
{
    unsigned char checksum;
    int  count;
    char ch;
  
    /*  $<packet info>#<checksum>. */
    do {
        PutDebugChar('$');
        checksum = 0;
        count    = 0;
  
//  Debug << "GdbPutPacket(" << buffer << ")\n" << eor;

        while ( (ch=buffer[count]) != 0 ) {
            PutDebugChar(ch);
            checksum += ch;
            count += 1;
        }

        PutDebugChar('#');
        PutDebugChar(hexchars[checksum >> 4]);
        PutDebugChar(hexchars[checksum & 0xf]);

    } while (1 == 0);  /* (GetDebugChar() != '+'); */
  
}

/* convert the memory pointed to by mem into hex, placing result in buf */
/* return a pointer to the last char put in buf (null) */
char* Gdb::mem2hex(char *mem, char *buf, int count)
{
    int i;
    unsigned char ch;
    for (i=0;i<count;i++) {
        ch = *mem++;
        *buf++ = hexchars[ch >> 4];
        *buf++ = hexchars[ch & 0xf];
    }
    *buf = 0; 
    return(buf);
}

/* convert the hex array pointed to by buf into binary to be placed in mem */
/* return a pointer to the character AFTER the last byte written */
char* Gdb::hex2mem(char *buf, char *mem, int count)
{
    int i;
    unsigned char ch;
    for (i=0;i<count;i++) {
        ch = hex(*buf++) << 4;
        ch = ch + hex(*buf++);
        *mem++ = ch;
    }
    return(mem);
}


/*
 * This function does all command procesing for interfacing to gdb.
 */
void Gdb::GdbHandler(int interruptVector, InterruptContext *context)
{
    int    sigval;
    unsigned int    addr, length;
    char * ptr;
    static int stepping = 0;
    static unsigned int stepped_insn;
    static unsigned *steppedInsnPointer;
    static char *inbuffer = 0;
    static char *outbuffer = 0;

    if (!inbuffer) inbuffer = new char[BUFMAX];
    if (!outbuffer) outbuffer = new char[BUFMAX];

    int done=0;

//  Debug << "handle_interrupt(" << interruptVector << ")\n" << eor;

    /*
     * This can be used to set things in the VM system so we know that
     * we are running on behalf of a user process.  So that our memory
     * faults are translated into user faults.
     */
    GdbMachineHandlerStart(interruptVector, context);

    if (stepping) { 
        /*
         * We got here after a "s" message.
         * Replace the stepped insn and return
         */
        *steppedInsnPointer = stepped_insn;
        stepping = 0;
    }

    /* Notify host that an interrupt has occurred */
    sigval = ComputeSignal( interruptVector );
    outbuffer[0] = 'T';
    outbuffer[1] = hexchars[sigval/16];
    outbuffer[2] = hexchars[sigval];
    outbuffer[3] = '\0';
    GdbPutPacket(outbuffer); 

    while (!done) { 
        outbuffer[0] = 0;
        GdbGetPacket(inbuffer);
//    Debug << "handle_interrupt: got packet: " << inbuffer << "\n" << eor;

        switch (inbuffer[0]) {
            case '?' : 
                outbuffer[0] = 'S';
                outbuffer[1] = hexchars[sigval/16];
                outbuffer[2] = hexchars[sigval];
                outbuffer[3] = '\0';
                break; 

            case 'c' : /* cAA..AA    Continue at address AA..AA(optional) */
            {
                unsigned pc = 0;
                scanhex(&inbuffer[1],&pc);
                if (pc != 0)
                    setPC(pc);

                done = 1;
            }
            break;
            
            case 'g' : /* read the value of the CPU registers */
                mem2hex((char *) registers, outbuffer, NUMREGBYTES);
                break;

            case 'G' : /* set the value of the CPU registers - return OK */
                hex2mem(&inbuffer[1], (char *) registers, NUMREGBYTES);
                strcpy(outbuffer,"OK");
                break;
      
            case 'H':
                // Set Thread for subsequent operations
                // The subcommand is either a g or a c
                strcpy(outbuffer, "OK");
                break;

            case 'k' :  /* Kill Choices */
                //Console().printf("GdbGetPacket: K packet\n");
                GdbKill();
                break;

            case 'm' : /* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
                if (2 == scan2hex(&inbuffer[1],&addr,&length) &&
                    GdbMachineMem_ok(GDB_READ, (char *)addr, length) == 0) {
                    mem2hex((char*) addr, outbuffer, length);
                }
                else {
                    strcpy(outbuffer,"E01");
                    Debug << "malformed read memory command: " << inbuffer << eor;
                }     
                break;


            case 'M' : /* MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK */
                if (2 == scan2hex(&inbuffer[1],&addr,&length) &&    
                    GdbMachineMem_ok(GDB_WRITE, (char*)addr, length) == 0) {
                    ptr = index(inbuffer,':');
                    ptr += 1; /* point 1 past the colon */
                    hex2mem(ptr, (char*) addr, length);
                    strcpy(outbuffer,"OK");
                }
                else {
                    strcpy(outbuffer,"E02");
                    Debug << "malformed write memory command: " << inbuffer << eor;
                }     
                break;

            case 'p' : /* Read register */
            {
                char regnum[2];
                if (inbuffer[2] != '\0') {
                    regnum[0] = inbuffer[1];
                    regnum[1] = inbuffer[2];
                } else {
                    regnum[0] = '0';
                    regnum[1] = inbuffer[1];
                }
                char num;
                hex2mem(&regnum[0], &num, 1);
                mem2hex((char *)&registers[(int)num], outbuffer, 4);
            }
            break;

            case 'q': // Query interface
                switch(inbuffer[1]) {
                    case 'C':
                        // Current thread id
                        // Francis TODO currently returns 0
                        strcpy(outbuffer,"QC");
                        char id[8];
                        int value = 0;
                        int i = 4;
                        char *s = id;
                        while (i--)
                            *s++ = 0;
                        *s++ = (value >> 24) & 0xff;
                        *s++ = (value >> 16) & 0xff;
                        *s++ = (value >> 8) & 0xff;
                        *s++ = (value & 0xff);
                        mem2hex(id, outbuffer+2, 8);
                        break;
                }
                break;
                
            case 's' : /* sAA..AA   Step one instruction from AA..AA(optional) */
            {
                unsigned pc = 0;
                scanhex(&inbuffer[1],&pc);
                if (pc != 0)
                    setPC(pc);

                stepped_insn = *((unsigned int*)getNPC());
                stepping = 1;
                steppedInsnPointer = ((unsigned int *)getNPC());
                *steppedInsnPointer = GdbBreakInsn;
                done = 1;
            }
            break;

            case 'v' : /* Verbose packet prefix */
                // Unsupported
                break;

            case 'X': /* Write memory (binary) */
                // Unsupported
                break;

            case 'Z': /* Z[0,1],AA..AA,length   Set breakpoint at AA..AA 0->software 1->hardware */
                // Francis - TODO hardware breakpoints unsupported as of now
                unsigned addr;
                if (0 != scanhex(&inbuffer[3],&addr) ){
                    setBreakpoint(addr);
                    strcpy(outbuffer, "OK");
                } else {
                    strcpy(outbuffer, "E22");
                }
                break;

            case 'z': /* z[0,1],AA..AA,length   Remove breakpoint at AA..AA 0->software 1->hardware */
                if (0 != scanhex(&inbuffer[3],&addr)) {
                    removeBreakpoint(addr);
                    strcpy(outbuffer, "OK");
                } else {
                    strcpy(outbuffer, "E22");
                }
                break;

            default:
                Debug << "GdbHandler: Unsupported request: " <<  inbuffer << eor;
        } /* switch */ 
    
        /* reply to the request */
        if (!done) GdbPutPacket(outbuffer); 
    }

    GdbMachineHandlerEnd(interruptVector, context);
}


void Gdb::setBreakpoint(unsigned addr) {
    Breakpoint *free = 0;

    for (unsigned i = 0; i < MAX_BREAKPOINTS; i++) {
        if (breakpoints[i].isValid()) {
            if (breakpoints[i].getAddr() == addr) {
                return;
            }
        } else {
            if (free == 0) free = &breakpoints[i];
        }
    }
    
    free->set(addr, GdbBreakInsn);
}

Breakpoint* Gdb::BreakpointAt(unsigned addr) {
    for (unsigned i = 0; i < MAX_BREAKPOINTS; i++) {
        if (breakpoints[i].isValid() && breakpoints[i].getAddr() == addr) {
            return &breakpoints[i];
        }
    }
    return 0;
}

void Gdb::removeBreakpoint(unsigned addr) {
    for (unsigned i = 0; i < MAX_BREAKPOINTS; i++) {
        if (breakpoints[i].isValid() && breakpoints[i].getAddr() == addr) {
            breakpoints[i].remove();
        }
    }
}
