#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MadAllocator.cc: implementation of a Big Bag-o-Pages heap class.
 *
 */

/*
 * Debugging in this file is a bit of a problem now because calling
 * Console() (as both Debug << and Raid() do) may cause a
 * per-process console to be constructed, resulting in a recursive
 * call to allocate(), etc.
 */

#define PAGEOUT

#include "AddressTranslator.h"
#include "Assert.h"
#include "ClassStar.h"
#include "CPU.h"
#include "ThisCPU.h"
#include "Domain.h"
#include "MadAllocator.h"
#include "Raid.h"
#include "Lock.h"

#ifdef PAGEOUT
#include "MemoryObject.h"
#include "MemoryObjectCache.h"
#include "MemoryObjectCacheStar.h"
#include "MemoryObjectStar.h"
#include "MemoryObjectView.h"
#endif


#ifdef __GNUC__
#pragma implementation
#endif

/* 
   DEBUG_HEAP
   is useful for debugging access to deallocated data structures.

   It clears the content after allocation and after deallocation.
   The status of heap memory can be determined from the contents
   of the first 16 bytes.

   Warning: Use for debugging only, may be double heap space 
 	    requirements because of log2 bucket scheme.  Tagging
	    16 extra bytes can raise allocation to next bucket.
	    For 8K BSD file system, results in almost 50% memory
	    fragmentation.

 */

// #define DEBUG_HEAP

#define HEAP_INFO

#ifdef HEAP_INFO
unsigned heap_check_on = 0;
unsigned heap_check = 5;
unsigned heap_allocated[32];
unsigned heap_free[32];
unsigned heap_pages[32];

void heap_check_allocate(int nbytes, char * addr)
{
	Console() 
	<< asAddress(addr) << " heap_check_allocate " << nbytes << "\n" << eor;
}

void heap_check_free(char * addr)
{
	Console()
	<< asAddress(addr) << " heap_check_free\n" << eor;
}

#endif

void heap_info()
{
#ifdef HEAP_INFO
	unsigned total = 0;
	Console().formattedWrite("      Size  InUse   Free  Pages\n");
	Console().flush();
	for (unsigned i = 2; i < 24; i++) {
		Console().formattedWrite
		("%10d %6d %6d %6d\n", 
		 1 << i, heap_allocated[i], heap_free[i], heap_pages[i]);
		Console().flush();
		total += heap_pages[i];
	}
	Console().formattedWrite("Total pages %d\n", total);
	Console().flush();
#endif
}

#ifdef DEBUG_HEAP
#define HEAP_FREE 0xFEFEFEFE
#define HEAP_FREE_CHAR 0xFE
#define HEAP_ALLOC 0xACACACAC
#define HEAP_ALLOC_CHAR 0xAC
#endif

/*
 * Set the CatchBlock with a debugger and set a breakpoint at
 *   CaughtBlock() to catch when a particular block is allocated or freed.
 */
void * CatchBlock = 0;
void
CaughtBlock() { }

/*
 * The following class function computes the amount of space (in
 * bytes) required by the allocator to manage a heap of the
 * specified size.  An area of at least this size should be passed
 * to the constructor.
 */
int
MadAllocator::stateSize( int numberOfPages )
{
	return( numberOfPages * sizeof( char * ) );
}

MadAllocator::MadAllocator( void * stateAddress, void * baseAddress,
			    int numberOfPages, int log2PageSize,
			    Domain * domain, void * startOkRange,
			    void * endOkRange, int noMultiProcessing )
{
	initDebug();
	_canReclaim = 0;

	/*
	 *	FIX: _sizeToBucket should be static class data.
	 */
	_sizeToBuckets[0] = 11;
	int element = 1;
	int bit;
	for( bit = 0; bit < 10; bit++ ) {
		for( int i = 0; i < ( 1 << bit ); i++ ) {
			_sizeToBuckets[element++] = bit + 12;
		}
	}

	_sizeToBucket[0] = 2;
	_sizeToBucket[1] = 2;
	_sizeToBucket[2] = 2;
	element = 3;
	for( bit = 1; bit < 12; bit++ ) {
		for( int i = 0; i < ( 1 << bit ); i++ ) {
			_sizeToBucket[element++] = bit + 1;
		}
	}

	Assert( numberOfPages > 4 );
	Assert( log2PageSize >= 8 );
	Assert( log2PageSize <= 16 );

	/*
	 *
	 */
	Assert( domain->isKindOf( DomainClass ) );
	_domain = domain;

	/*
	 *	Set up instance variables based on constructor parameters.
	 *	Make sure the base address for allocations is page-aligned.
	 *	If re-alignment is needed, then we lose one page.
	 */
	_pageToBucket = (char *) stateAddress;
	_log2PageSize = log2PageSize;
	_pageSize = 1 << _log2PageSize;
	_pageMask = _pageSize - 1;
	if( (int) baseAddress & _pageMask ) {
		_numberOfPages = numberOfPages - 1;
		int alignedAddress = (int)baseAddress + _pageMask;
		_baseAddress = (char *)( alignedAddress & ~_pageMask );
	}
	else {
		_numberOfPages = numberOfPages;
		_baseAddress = (char *) baseAddress;
	}
	_upperBoundAddress = (char *) ((int) _baseAddress +
		(_numberOfPages * _pageSize));

	/*
	 *	Check to make sure aligment worked and
	 *	_baseAddress address is non-null.
	 */
	Assert( ( (int) _baseAddress & _pageMask ) == 0 );
	Assert( _baseAddress != 0 );
	Assert( _pageToBucket != 0 );

	/*
	 *	Initialize the buckets to page array,
	 *	set to zeros since no bucket points to a page with free memory.
	 */
	unsigned i;
	for( i = 0; i < NumberOBuckets; i++ ) {
		_bucketToPage[i] = 0;
		_bucketMutex[i] = 0;
		_bucketFreeCount[i] = 0;
		_bucketEnd[i] = 0;
#ifdef HEAP_INFO
		heap_free[i] = 0;
		heap_pages[i] = 0;
		heap_allocated[i] = 0;
#endif
	}

	/*
	 *	Initialize the page information array,
	 *	set to zeros since no pages are in use.
	 */
	for( i = 0; i < _numberOfPages; i++ ) _pageToBucket[i] = 0;

	_top = 1; // waste a page cause I am lazy and to be safe
	_topMutex = 0;

	_startOkRange = startOkRange;
	_endOkRange = endOkRange;

#ifdef PAGEOUT
	for( i = _log2PageSize + 1; i < NumberOBuckets; i++ ) {
		if ( i > 17 ) // 128kb
			_bucketMaxFreeCount[i] = 0;
		else
			_bucketMaxFreeCount[i] = 1 << ( 17 - i ) ;
	}

	const char * domainBaseAddr;
	ProtectionLevel prot;
	unsigned length;
	MemoryObjectViewStar view = _domain->lookup( _baseAddress, domainBaseAddr,
					         length, prot );

	Assert( view != 0 );
	Assert( view->isKindOf( NormalMemoryObjectViewClass ) );
	Assert( _baseAddress == domainBaseAddr );

	MemoryObject * mo = view->source();
	extern ClassStar RootMemoryObjectClass;
	Assert( mo->isKindOf( RootMemoryObjectClass ) );

	_cache = mo->cache();
	Assert( _cache != 0 );
	extern ClassStar RootMemoryObjectCacheClass;
	Assert( _cache->isKindOf( RootMemoryObjectCacheClass ) );
	Assert( (unsigned)_cache->log2UnitSize() == _log2PageSize );
#endif

	if ( !noMultiProcessing )
		enableMultiProcessing();

	Raid( RaidConstructor )( "%N::MadAllocator nOP %d l2PS %d ",
				 this, _numberOfPages, _log2PageSize );
	Raid( RaidConstructor )( "bA %A pI %A %N\n",
				 _baseAddress, _pageToBucket, _domain );
}

/*
 *	Nothing to do in the destructor.
 */
MadAllocator::~MadAllocator()
{
	Raid( RaidConstructor )( "%N::~MadAllocator\n", this );
	Assert( NOTREACHED );
}

/*
 * enableMultiProcessing has to be separated out of the constructor 
 *    because the 'new' that gets used may be the MadAllocator itself.
 */
void
MadAllocator::enableMultiProcessing()
{
	for( unsigned i = 0; i < NumberOBuckets; i++ ) {
		_bucketMutex[i] = new Lock;
	}
	_topMutex = new Lock;
}

/*
 *	findBucket quickly rounds nbytes up to the nearest power
 *	of 2, which is the bucket number.
 */
inline
unsigned int
MadAllocator::findBucket( unsigned int size )
{
	unsigned int bucket;
	if( size <= 0x1000 ) {
		bucket = _sizeToBucket[size];
	}
	else {
		if( size < 0x400000 )	bucket = _sizeToBuckets[size>>12];
		else			bucket = _sizeToBuckets[size>>22] + 10;

		/* Adjust if size is not a power of two */
		if( size & (size - 1) )	bucket++;
	}
	return( bucket );
}

/*
 *	allocate	given the size of a desired object,
 *			return a pointer to a block of memory whose
 *			size is the smallest power of 2 greater than or equal
 *			to nbytes and whose allignment is either on a page
 *			boundary or on a multiple of the region size.
 *
 *			return 0 if either the requestor asked for a 0 byte
 *			block-o-memory or we are out-o-memory.
 *
 *	This method absolutely MUST NOT be called in the interval between
 *	setting the allocator to this allocator and turning on VM!
 */
void *
MadAllocator::allocate( unsigned int nbytes )
{
	Raid( RaidMethod )( "%N::allocate( nbytes %d )\n", this, nbytes );
	//Assert( thisCPU()->MMU()->enabled() ); thisCPU can't be called
		// with interrupts enabled - raila 11-4-93

#ifdef DEBUG_HEAP
	nbytes += 16;
#endif

	char *	returnValue;

	int unsigned bucket = findBucket( nbytes );

	Assert( bucket > 1 );
	Assert( bucket < NumberOBuckets );

	/*
	 *	Get the appropriate bucket descriptor out of the
	 *	_bucket[i].toPage table.  A null descriptor means there is
	 *	currently no free space for this bucket.
	 */
	Lock *mutex = _bucketMutex[bucket];
	if ( mutex != 0 )
		mutex->acquire();

	if( ( returnValue = _bucketToPage[bucket] ) == 0 ) {
		returnValue = moreSpace( bucket );
	}
	else {
		_bucketToPage[bucket] = *(char **)returnValue;
		Assert( _bucketToPage[bucket] == 0 ||
			( _bucketToPage[bucket] >= _baseAddress && 
		          _bucketToPage[bucket] < _upperBoundAddress ) );
#ifdef PAGEOUT
		if( bucket > _log2PageSize ) {
			if ( --_bucketFreeCount[bucket] == 0 ) {
				Assert( _bucketToPage[bucket] == 0 &&
				  	_bucketEnd[bucket] == returnValue );
				_bucketEnd[bucket] = 0;
			} else {
				Assert( _bucketEnd[bucket] != 0 &&
					_bucketToPage[bucket] != 0 );
			}
			if ( *((char **) returnValue + 1) != 0 ) {
				Assert( *((char **) returnValue + 1) == 
					returnValue );
				unsigned pages = 
					1 << ( bucket - _log2PageSize );
				_domain->makeAddressable
					( returnValue + _pageSize, 
					  ( pages - 1 ) << _log2PageSize );
			} else {
				_canReclaim--;
			}
		}
#endif
	}

#ifdef HEAP_INFO
	heap_allocated[bucket]++;
	heap_free[bucket]--;
	if ( heap_check_on && heap_check == bucket ) {
		heap_check_allocate(nbytes, returnValue);
	}
#endif
#ifdef DEBUG_HEAP
	unsigned int * signature = (unsigned int *) returnValue;
	Assert( signature[1] == HEAP_FREE );
	signature[1] = HEAP_ALLOC;
	returnValue += 16;
	nbytes -= 16;
	unsigned char * cp = returnValue;
	for (int i = 0; i < (1 << bucket) - 16; i++)
		*cp++ = HEAP_ALLOC_CHAR;
#endif // DEBUG_HEAP

	if ( mutex != 0 )
		mutex->release();

	Raid( RaidMethod )( "%N::allocate returns %A next will be %A\n",
			    this, returnValue, _bucketToPage[bucket] );

#ifdef ASSERT
	if ( (void *)returnValue == CatchBlock )
		CaughtBlock();
#endif

	return( (void *)returnValue );
}

char *
MadAllocator::moreSpace( unsigned int bucket )
{
	Raid( RaidProtected )( "%N::moreSpace( bucket %d )\n", this, bucket );

	Assert( bucket > 1 );
	Assert( bucket < NumberOBuckets );

	unsigned pages = 1;
	if( bucket > _log2PageSize ) pages = 1 << ( bucket - _log2PageSize );

	Lock *mutex = _topMutex;
	if ( mutex != 0 )
		mutex->acquire();

	int oldTop = _top;

	Raid( RaidDetail )( "%N::moreSpace top %d pages %d total %d\n",
			    this, _top, pages, _numberOfPages );

	if( _top + pages > _numberOfPages ) {
		Console().formattedWrite
			( "%N::moreSpace out of free pages.\n", this );
		Console().flush();
		heap_info();
		Assert( NOTREACHED );
		return( 0 );
	}

	_top += pages;

	if ( mutex != 0 )
		mutex->release();

	_pageToBucket[oldTop] = bucket;

	unsigned units;
	if( bucket < _log2PageSize ) units = _pageSize >> bucket;
	else {
		units = 1;
		for( unsigned i = 1; i < pages; i++ ) _pageToBucket[oldTop + i] = 1;
	}

#ifdef HEAP_INFO
	heap_free[bucket] += units;
	heap_pages[bucket] += pages;
#endif 

	char * head = &_baseAddress[oldTop<<_log2PageSize];
	Assert( head >= _baseAddress );
	Assert( head < _upperBoundAddress );

	_domain->makeAddressable( head, pages << _log2PageSize );

#ifdef DEBUG_HEAP
	unsigned char * cp = (unsigned char *) head;
	for (int i = 0; i < _pageSize; i++)
		*cp++ = HEAP_FREE_CHAR;	
#endif // DEBUG_HEAP

	char * tail = head;
	int inc = 1 << bucket;
	for( ; units > 1; units-- ) {
		Assert( tail >= _baseAddress );
		Assert( tail < _upperBoundAddress );
		*(int *)tail = (int)tail + inc;
		tail += inc;
	}

	*(char **)tail = _bucketToPage[bucket];
	_bucketToPage[bucket] = *(char **)head;
	Assert( ( _bucketToPage[bucket] == 0 ) ||
		( _bucketToPage[bucket] >= _baseAddress ) );
	Assert( _bucketToPage[bucket] < _upperBoundAddress );
	
	return( head );
}

/*
 *	Given a pointer to a block of memory, return that memory to the
 *	heap.  Perform various checks to make sure the pointer is in
 *	the heap's range, has a valid alignment, and isn't returned 
 *	multiple times between allocations.
 *	(Not all checks performed all the time).
 *	Freeing address 0 is legal and a no-op.
 */
void
MadAllocator::free( void * vp )
{
#ifdef ASSERT
	if ( vp == CatchBlock )
		CaughtBlock();
#endif

	char * cp = (char *)vp;
	Raid( RaidMethod )( "%N::free( %A )\n", this, cp );

	if( ( cp != 0 ) && ( cp < _baseAddress ) ) {
		if ( ( vp >= _startOkRange ) && ( vp < _endOkRange ) )
			Debug.printf( "%N::free( %A ) Object in ok range: %A\n",
					this, cp, _baseAddress );
		else
			Console().printf(
				"%N::free( %A ) Object below heap base: %A\n",
					this, cp, _baseAddress );
		Debug.printf( ": %o %o %o %o\n",*cp,*(cp+1),*(cp+2),*(cp+3));
	}

	/*
	 *	Nothing to do if passed a null or an out-of-range pointer .
	 */
	if( cp < _baseAddress ) return;

#ifdef DEBUG_HEAP
	cp -= 16;
	unsigned int * signature = (unsigned int *) cp;
	Assert( signature[1] == HEAP_ALLOC );
	signature[1] = HEAP_FREE_CHAR;
#endif // DEBUG_HEAP

	/*
	 * 	Calculate the following:
	 *		1 ) pointer offset from base address
	 *		2 ) page number relative to address
	 *		3 ) offset from beginning of page
	 *	Make sure the pointer obeys the following: 
	 *		1 ) pointer is within managed region of memory
	 *		2 ) page number is within range
	 *		3 ) this page is in use
	 *		4 ) offset has the proper alignment for this page
	 */
	unsigned  offset = cp - _baseAddress;

	Assert( ( offset >> _log2PageSize ) < _numberOfPages );

	unsigned bucket = _pageToBucket[offset >> _log2PageSize];
	Assert( bucket > 1 );
	Assert( bucket < NumberOBuckets );

	int pageOffset = (int)cp & _pageMask;
	Assert( ! ( ( ( 1 << bucket ) - 1 ) & pageOffset ) );

	Raid( RaidDetail )( "%N::free bkt %d pg %d\n",
			    this, bucket, offset >> _log2PageSize );
	Lock *mutex = _bucketMutex[bucket];
	if ( mutex != 0 )
		mutex->acquire();

#ifdef HEAP_INFO
	heap_allocated[bucket]--;
	heap_free[bucket]++;
	if ( heap_check_on && heap_check == (unsigned)bucket ) {
		heap_check_free(cp);
	}
#endif

#ifdef DEBUG_HEAP
	unsigned char * p = cp;
	for( int i = 0; i < ( 1 << bucket ); i++ )
		*p++ = HEAP_FREE_CHAR;
#endif // DEBUG_HEAP

#ifdef PAGEOUT
	if( bucket > _log2PageSize ) {
	    	if ( _bucketFreeCount[bucket] >= _bucketMaxFreeCount[bucket] ) {
			if ( _bucketEnd[bucket] != 0 ) {
				Assert( *((char **) _bucketEnd[bucket]) == 0 &&
					_bucketFreeCount[bucket] > 0 );
				*((char **) _bucketEnd[bucket]) = cp;
			} else {
				Assert( _bucketFreeCount[bucket] == 0 &&
					_bucketToPage[bucket] == 0);
				_bucketToPage[bucket] = cp;
			}
			_bucketEnd[bucket] = cp;
			*(char **) cp = 0;
			unsigned pages = 1 << ( bucket - _log2PageSize );
			unsigned fromUnit = ( offset >> _log2PageSize ) + 1;
			unsigned toUnit = fromUnit + pages - 2;
			_cache->free( fromUnit, toUnit );
			*((char **) cp + 1) = cp;
		} else {
			_canReclaim++;
			*((char **) cp + 1) = 0;
			*(char **)cp = _bucketToPage[bucket];
			if ( _bucketToPage[bucket] == 0 ) {
				_bucketEnd[bucket] = cp;
			}
			_bucketToPage[bucket] = cp;
			Assert( _bucketToPage[bucket] >= _baseAddress &&
				_bucketToPage[bucket] < _upperBoundAddress );
		}
		_bucketFreeCount[bucket]++;
	} else {
		*(char **)cp = _bucketToPage[bucket];
		_bucketToPage[bucket] = cp;
		Assert( _bucketToPage[bucket] >= _baseAddress &&
			_bucketToPage[bucket] < _upperBoundAddress );
	}
#else
	*(char **)cp = _bucketToPage[bucket];
	_bucketToPage[bucket] = cp;
	Assert( _bucketToPage[bucket] >= _baseAddress &&
		_bucketToPage[bucket] < _upperBoundAddress );
#endif

	if ( mutex != 0 )
		mutex->release();

	Raid( RaidMethod )( "%N::free returns\n", this );
} 

int
MadAllocator::pagesUsed()
{
	return( _top );
}

int
MadAllocator::reclaimSpace()
{
#ifdef PAGEOUT
	if ( _canReclaim <= 0 )
		return 0;

	int result = 0;
	unsigned bucket;
	for ( bucket = _log2PageSize + 1; bucket < NumberOBuckets; bucket++ ) {
		if ( _bucketFreeCount[bucket] == 0 ) 
			continue;
		Lock *mutex = _bucketMutex[bucket];
Console() << "reclaim bucket " << bucket << "\n" << eor;
		if ( mutex != 0 )
			mutex->acquire();
		char ** cpp = (char **) _bucketToPage[bucket];
		while( cpp ) {
			if ( *(cpp + 1) == 0 ) {
				int offset = (char *) cpp - _baseAddress;
				unsigned pages = 
					1 << ( bucket - _log2PageSize );
				unsigned fromUnit = 
					( offset >> _log2PageSize ) + 1;
				unsigned toUnit = fromUnit + pages - 2;
				_cache->free( fromUnit, toUnit );
				*(cpp + 1) = (char *) cpp;
				result += ( pages - 1 );
				_canReclaim--;
			} else {
				Assert( *(cpp + 1) == (char *) cpp );
			}
			cpp = (char **) *cpp;
		}
		if ( mutex != 0 )
			mutex->release();
Console() << "reclaim bucket " << bucket << " done\n" << eor;
	}
	return result;
#else
	return 0;
#endif
}

#include "ClassInstaller.h"
DEFINECLASS("MadAllocator",MadAllocator,Allocator,0,0);
