#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MemoryBlockAllocator.cc: implementation of a Big Bag-o-Pages heap class.
 *
 */

#include "Assert.h"
#include "MemoryBlockAllocator.h"
#include "OutputStream.h"


#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This allocator is intended to manage relatively large, fixed-size blocks.
 * Therefore, we keep a bit vector _freeList to indicate which blocks are
 * currently in use.
 *
 * The _freeList consists of an array of chars, so if the _blockCount
 * managed by the allocator is not a multiple of 8, the last byte in the
 * _freeList is only partially used.  Since a '1' bit in _freeList denotes
 * an available block, the last byte of _freeList is initialized as though
 * the unused bits were already allocated.  The following table contains
 * the precomputed values needed, indexed by the number of bits used in
 * the last byte of _freeList.
 *
 * 0000 0000    0 bits used ... this case should never be used
 * 0000 0001    1 bit used, i.e., managing 8n + 1 blocks
 * 0000 0011    2 bits used, i.e., managing 8n + 2 blocks
 * 0000 0111
 * 0000 1111
 * 0001 1111
 * 0011 1111
 * 0111 1111
 */
static unsigned char initialFreeList[] = { 0xff, 0x01, 0x03, 0x07,
                                           0x0f, 0x1f, 0x3f, 0x7f };

MemoryBlockAllocator::MemoryBlockAllocator( int size, int blockSize )
{
    initDebug();

//  Debug << this << "::MemoryBlockAllocator( " << size << " bytes, "
//      << blockSize << "-byte blocks )\n" << eor;

    _blockSize = blockSize;
    int blockMask = blockSize - 1;

    /*
     * Allocate the number of bytes requested.  Since we align blocks
     * on _blockSize boundaries, there will be wasted space both before
     * and after the allocatable blocks unless the space we allocate
     * here happens to be _blockSize-aligned.
     */
    char * base = new char [ size ];

    /*
     * Compute the bases of the first and last _blockSize-aligned
     * blocks within the region just allocated.
     */
    _firstBlock = (char *)
    (((int) &base[_blockSize - 1]) & (~blockMask));
    _lastBlock = (char *)
    ((((int) &base[size - 1]) & (~blockMask)) - _blockSize);

    Assert( ((unsigned int) _firstBlock % _blockSize) == 0 );
    Assert( ((unsigned int) _lastBlock % _blockSize) == 0 );

    _blockCount = ((_lastBlock - _firstBlock) / _blockSize) + 1;
    int usableSpace = _blockCount * _blockSize;
    Assert( (usableSpace % _blockSize) == 0 );
    _totalBytesFree = usableSpace;

    unsigned wastedSpaceBefore = _firstBlock - base;
    unsigned wastedSpaceAfter = size - (wastedSpaceBefore + usableSpace);

    Debug << this << "::MemoryBlockAllocator: " << size
          << " bytes allocated, " << _blockCount << " blocks fit\n"
          << eor;
    Debug << this << "::MemoryBlockAllocator: wasted " << wastedSpaceBefore
          << " bytes before, " << wastedSpaceAfter
          << " after\n" << eor;
    Debug << this << "::MemoryBlockAllocator: _firstBlock: "
          << asAddress( _firstBlock ) << ", _lastBlock: "
          << asAddress( _lastBlock ) << "\n" << eor;

    /*
     * The _freeList requires one bit per block, so compute the number
     * of bytes needed for the list, allocate it, and initialize it.
     */
    _freeListSize = _blockCount / 8;
    int bitsLeftOver = _blockCount % 8;
    if( bitsLeftOver > 0 ) {
        _freeListSize += 1;
    }

    /*
     * If the _freeList will fit in the wasted space before or after
     * the _blockSize-aligned area, we might as well use that.  Otherwise
     * we have to allocate more space for it.
     */
    if( _freeListSize <= wastedSpaceBefore ) {
        _freeList = (unsigned char *) base;
    } else if( _freeListSize <= wastedSpaceAfter ) {
        _freeList = (unsigned char *) (_lastBlock + _blockSize);
    } else {
        _freeList = new unsigned char [ _freeListSize ];
    }

    Debug << this << "::MemoryBlockAllocator: " << _freeListSize
          << "-byte _freeList at " << asAddress( _freeList )
          << "\n" << eor;

    /*
     * Initialize the _freeList to show that no blocks have yet
     * been allocated.  If the last byte of the _freeList is only
     * partially used (i.e., if the _blockCount is not an even
     * multiple of 8), mark the unused bits as if the corresponding
     * blocks were already allocated.
     */
    unsigned i;
    for( i = 0; i < _freeListSize; i++ ) {
        _freeList[i] = 0xff;
    }
    if ( bitsLeftOver > 0 ) {
        _freeList[ _freeListSize - 1 ] = initialFreeList[ bitsLeftOver ];
    }
}

/*
 *  Nothing to do in the destructor.
 */
MemoryBlockAllocator::~MemoryBlockAllocator()
{
    Console() << this << "::~MemoryBlockAllocator()\n" << eor;
    Assert( NOTREACHED );
}

static unsigned char freeBlockMask[] = { 0x01, 0x02, 0x04, 0x08,
                                         0x10, 0x20, 0x40, 0x80 };

/*
 * Allocate the requested number of bytes.  For now, we require that the
 * size be a multiple of the block size, though I suppose we could just
 * allocate a bit more than requested if it's not.
 */
void *
MemoryBlockAllocator::allocateAligned( unsigned nbytes, unsigned align )
{
    if (align == 0)
        align = _blockSize;

    Debug << this << "::allocate( " << nbytes << " ) aligned at " << asHex(align) <<"\n" << eor;
    unsigned freeBlocksReqd = nbytes / _blockSize;
                        
    // For now...
    Assert( (nbytes % _blockSize) == 0 );

    unsigned i, j;
    int blockOffset = -1;
    // unsigned char oldFreeList, newFreeList;
    // oldFreeList = newFreeList = '\0';

    _mutex.acquire();
    for( i = 0; i < _freeListSize; i++ ) {
        if( _freeList[i] > 0x00 ) {
            for( j = 0; j < 8; j++ ) {
                if( (_freeList[i] & freeBlockMask[j]) != 0 ) {
                    // Ok. We found a free block
                    //oldFreeList = _freeList[i];
                    //_freeList[i] &= ~freeBlockMask[j];
                    //newFreeList = _freeList[i];
                    blockOffset = (i * 8) + j;
                    // Check to see if the free block we found is aligned the way we want it
                    if ( ((unsigned)(_firstBlock + (blockOffset * _blockSize)) % align) == 0 )
                    {
                        // Check to see if we have contiguous space for nbytes
                        int k;
                        for (k = blockOffset; k < (blockOffset + (int)freeBlocksReqd); k++)
                        {
                            if (( _freeList[ k / 8 ] & freeBlockMask[ k % 8 ] ) == 0 ) {
                                blockOffset = -1;
                                break;
                            }
                        }
                        if ( blockOffset != -1 )
                            break;
                    } else {
                        blockOffset = -1;
                    }
                }
            }
        }
        if ( blockOffset != -1 )
            break;
    }
    _mutex.release();

    char * block;
    if( blockOffset == -1 ) {
        Console() << this << "::allocate: out of space!\n" << eor;
        /*
         * Return an unlikely value to indicate failure.
         * -1 can never be a valid block address, unless we
         * are allocating 1-byte blocks.  If the caller
         * attempts to use the address anyway, he'll
         * probably get a page fault.
         */
        block = (char *) -1;
    } else {
        // Mark blocks as been used
        for (int k = blockOffset; k < (blockOffset + (int)freeBlocksReqd); k++)
        {
            _freeList[ k / 8 ] &= ~freeBlockMask[ k % 8 ];
        }
        
        // Debug << this << "::allocate: blockOffset: " << blockOffset
        //  << ", oldFreeList: " << asHex( oldFreeList )
        //  << ", newFreeList: " << asHex( newFreeList )
        //          << "\n" << eor;
        block = _firstBlock + (blockOffset * _blockSize);
        // Debug << this << "::allocate: " << nbytes << " at " << asAddress( block )
        //          << " align: " << asHex(align)
        //          << " free: " << _totalBytesFree - nbytes << "\n" << eor;
    }

    _totalBytesFree -= nbytes;

    return( (void *) block );
}

/*
 * Francis - TOFIX
 * Free does not free allocatations bigger than one blockSize correctly
 * Needs to be matched to updated allocateAligned implementation above
 */
void
MemoryBlockAllocator::free( void * vp )
{
    Debug << this << "::free( " << asAddress( vp ) << " )\n" << eor;

    int blockIndex = ((char *) vp - _firstBlock) / _blockSize;
    int freeListIndex = blockIndex / 8;
    int bitIndex = blockIndex % 8;

    Debug << this << "::free: blockIndex: " << blockIndex
          << ", freeListIndex: " << asHex( freeListIndex )
          << ", bitIndex: " << asHex( bitIndex )
          << "\n" << eor;

    _totalBytesFree += _blockSize;

    unsigned char oldFreeList, newFreeList;
    _mutex.acquire();
    oldFreeList = _freeList[ freeListIndex ];
    _freeList[ freeListIndex ] |= freeBlockMask[ bitIndex ];
    newFreeList = _freeList[ freeListIndex ];
    _mutex.release();

    Debug << this << "::free: oldFreeList: " << asHex( oldFreeList )
          << ", newFreeList: " << asHex( newFreeList )
          << "\n" << eor;
} 

#include "ClassInstaller.h"
DEFINECLASS("MemoryBlockAllocator",MemoryBlockAllocator,Allocator,0,0);
