#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * NameServer.cc: implementation of the NameServer class.
 */

#include "Assert.h"
#include "Class.h"
#include "AssocTable.h"
#include "NameServer.h"
#include "ProxiableObject.h"
#include "Raid.h"


#ifdef __GNUC__
#pragma implementation
#endif

NameServer *StandardNameServer;

NameServer::NameServer( NameServer *_parent )
{
	initDebug();
	hashTable = 0;
	parent = _parent;
}

NameServer::~NameServer()
{
	Raid( RaidConstructor )( "%N::~NameServer()\n" );

	if ( hashTable != 0 )
		delete hashTable;
}

static void
unbindNameServerEntry( int type, void * name, unsigned int entry )
{
	( (ProxiableObject *) entry )->unreference();
}

void
NameServer::bind( char * name, ProxiableObject * o )
{
	Raid( RaidMethod )( "%N::bind( %s, %N )\n", this, name, o );

	if ( o == 0 )
		return;

	/*
	 * Explicitly call reference() on the object since we can't use
	 *    Stars inside the hash table.
	 */
	o->reference();

	//lock.acquire();
	mutex.P();

	if ( hashTable == 0 ) {
		hashTable = new AssocTable( ( parent == 0 ) ?
			KERNELNAMESERVERHASHSIZE : OTHERNAMESERVERHASHSIZE,
				&unbindNameServerEntry );
		
		Assert( hashTable != 0 );
	}

// 	if ( !hashTable->bind( name, (unsigned int) o ) ) {
// 		/*
// 		 * The bind failed, try unbinding it first.
// 		 */
// 		int status;
// 		(void) hashTable->unbind( name, status );
// 		Assert( status != -1 );
// 
// 		status = hashTable->bind( name, (unsigned int) o );
// 		Assert( status != 0 );
// 	}
	unsigned int value;
	if (hashTable->unbind( name, value ))
		((ProxiableObject *) value) -> unreference ();
	hashTable->bind( name, (unsigned int) o );
	Assert( hashTable->lookup( name, value ) &&
	       value == (unsigned int) o );

	//lock.release();
	mutex.V();
	Raids( RaidMethod ) << eor;
}

ProxiableObjectRef
NameServer::lookup( char * name, Class * aClass )
{
	Raid( RaidMethod )( "%N::lookup( (char*)%s, (Class*)%N )\n", this, name, aClass );
	if( name == 0 ) {
		Raid( RaidMethod )( "%N::lookup: bad name (name == 0)\n", this);
		return( 0 );
		}

	/*
	 * aClass == 0 defaults to ClassClass to allow a user program
	 * to lookup a class when it doesn't have anything else.
	 */
	if( aClass == 0 ) {
		aClass = ClassClass;
	}

	ProxiableObject * obj = 0;
	//lock.acquire();
	mutex.P();
	if ( hashTable != 0 ) {
		unsigned int value;
		if (hashTable->lookup( name , value ))
		    obj = (ProxiableObject *)value;
// 		obj = (ProxiableObject *) hashTable->lookup( name, status );
// 		if ( status == -1 )
// 			obj = 0;
	}
	//lock.release();
	mutex.V();
	Raid( RaidMethod )( "%N::lookup: obj = %N from hash table\n", 
		this, obj );

	if ( obj != 0 ) {
		/*
		 * Getting classOf can cause an out-call if obj is a proxy.
		 */
		Class * objClass = obj->classOf();

		if( !objClass->isASubclassOf( aClass ) ) {
			Raid( RaidMethod )( "%N::lookup: bad class\n", this );
			return( 0 );
		}
	}
	else if ( parent != 0 ) {
		Raid( RaidMethod )( "%N::lookup: trying %s in parent(%N)\n", this, name, parent );

		return( parent->lookup( name, aClass ) );
	}
	
	Raid( RaidMethod )( "%N::lookup: returning %N\n", this, obj );

	/*
	 * The next statement can cause an outcall if obj is a proxy;
	 *   it calls reference() on obj.
	 */
	ProxiableObjectRef ret = obj;

	Raids( RaidMethod ) << eor;
	return( ret );
}

ProxiableObjectRef
NameServer::lookup( char * name, char * className )
{
	Raid( RaidMethod )( "%N::lookup( (char*)%s, (char*)%s )\n",
		this, name, className );
	if( className == 0 ) className = "Class";

	ClassStar aClass = lookup( className, ClassClass );

	ProxiableObjectRef ret;
	if( aClass != 0 ) ret = lookup( name, aClass );
	else ret = 0;
	Raid( RaidMethod )( "%N::lookup( %s, %s ) returns %N\n", this, name, 
		className, ret );
	Raids( RaidMethod ) << eor;
	return ret;
}

int
NameServer::unbind( char * name )
{

	Raid( RaidMethod )( "%N::unbind(%s)\n", this, name );

	int ret = 0;

	//lock.acquire();
	mutex.P();

	if ( hashTable != 0 ) {
// 		int status;
// 		(void) hashTable->unbind( name, status );
// 		if ( status != -1 )
// 			ret = 1;
		unsigned int value;
		if (hashTable->unbind( name, value )){
			((ProxiableObject *) value) -> unreference ();
			ret = 1;
			}
	}

	//lock.release();
	mutex.V();

	Raids( RaidMethod ) << eor;
	return( ret );
}

static Semaphore copySem( 1 );
static NameServer * copyToNameServer = 0;

static void copyNameServerEntry( int type, void * name, unsigned int entry )
{
	char * str = ( char * ) name;
	ProxiableObject * po = (ProxiableObject *) entry;
	copyToNameServer->bind( str, po );
}

void NameServer::clone( NameServer * nameServer )
{
	Raid( RaidMethod )( "%N::clone( %N )\n", this, nameServer );

	copySem.P();
	mutex.P();
	Assert( copyToNameServer == 0 );
	copyToNameServer = nameServer;
	hashTable->mapcar( (PFLV) &copyNameServerEntry );
	copyToNameServer = 0;
	mutex.V();
	copySem.V();
}

#include "ClassInstaller.h"
DEFINECLASS("NameServer",NameServer,ProxiableObject,0,0);
