#ifdef __GNUC__
#pragma implementation
#endif

#define NEW
/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David W. Dykstra (dykstra@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * ObjectProxyManager.cc: manager to allocate and free ObjectProxies
 */

#include "AddressTranslator.h"
#include "Allocator.h"
#include "Assert.h"
#include "Byte.h"
#include "Class.h"
#include "Domain.h"
#include "HashTable.h"
#include "Kernel.h"
#include "OutputStream.h"
#include "ObjectProxy.h"
#include "ObjectProxyManager.h"
#include "ProtectionLevel.h"
#include "Raid.h"


#ifdef __GNUC__
#pragma implementation
#endif

class OPMEntry : public Link {
  public:
    ObjectProxyKey k;
    ObjectProxy *  p;
    virtual int compare(Link * l);	// compare pointers to self
    virtual int hash(int);		// hash pointer to self.
};

int
OPMEntry::compare( Link * l )
{

return (
	k._proxiedObject == ((OPMEntry *)l)->k._proxiedObject &&
	k._proxiedDomain == ((OPMEntry *)l)->k._proxiedDomain &&
	k._table == ((OPMEntry *)l)->k._table
	);
	
}

int
OPMEntry::hash(int n)
{
    return (((((unsigned)k._proxiedObject) << 1 +
	      ((unsigned)k._proxiedDomain)) << 1 +
	     ((unsigned)k._table)) << 1) % n;
}
    
ObjectProxyManager::ObjectProxyManager( int maxProxies,
	ObjectProxy * _domainProxyMemory, MemoryObject * proxyMO, 
        ProxiableObject * freeFirst )
{
	initDebug();

	Raids( RaidConstructor ) << this << "::ObjectProxyManager( " 
		<< "freeFirst " << freeFirst << " )\n" << eor;

	domainProxyMemory = _domainProxyMemory;
	domainProxyMemoryLast = domainProxyMemory + (maxProxies - 1);
	if ( proxyMO == 0 ) 
		kernelProxyMemory = domainProxyMemory;
  	else
	        kernelProxyMemory = (ObjectProxy *) TheKernel->domain()->add
		    		    ( proxyMO, NoAccess);
	Assert( kernelProxyMemory != 0 );
	kernelProxyMemoryLast = kernelProxyMemory + (maxProxies - 1);
	kernelToDomainProxyMemoryOffset =
		(unsigned int) domainProxyMemory -
		(unsigned int) kernelProxyMemory;

	firstObjectProxy = kernelProxyMemory;
	nextObjectProxy = firstObjectProxy;
	lastObjectProxy = firstObjectProxy + maxProxies;
	sizeOfAllocatedProxies = 0;
	proxyMemoryObject = proxyMO;
	freeList = 0;

	firstToFree = freeFirst;

	/*
	 * Allocate space for a pseudo-ObjectProxy so we can use 
	 *   the next and prev pointers to easily maintain the doubly-
	 *   linked list.
	 */
	usedList = (ObjectProxy *) new char[ sizeof(ObjectProxy) ];
	Assert( usedList != 0 );
	usedList->next = usedList;
	usedList->prev = usedList;
	proxiesAllocated = 0;
#ifndef NOHASH
	hash = new HashTable( maxProxies + (maxProxies/2) );
#endif
}

ObjectProxyManager::~ObjectProxyManager()
{
	freeAll();
	delete (char *) usedList;
	if ( proxyMemoryObject != 0 ) {
	    	ObjectProxy * where = (ObjectProxy *) 
			TheKernel->domain()->remove( proxyMemoryObject, 
				             (const char *) kernelProxyMemory );
		Assert(where == kernelProxyMemory);
	}
#ifndef NOHASH
	delete hash;
#endif
}

ObjectProxy * 
ObjectProxyManager::domainToKernel( ObjectProxy * object )
{
	Assert( object >= domainProxyMemory &&
		object <= domainProxyMemoryLast );
	return (ObjectProxy *) 
	       ((char *) object - kernelToDomainProxyMemoryOffset);
}

ObjectProxy *    
ObjectProxyManager::kernelToDomain( ObjectProxy * object )
{ 
        Assert( object >= kernelProxyMemory && 
                object <= kernelProxyMemoryLast ); 
        return (ObjectProxy *) 
               ((char *) object + kernelToDomainProxyMemoryOffset);
} 

int
ObjectProxyManager::isValidProxy( ObjectProxy * object )
{
        /*
         * See if the object is a valid ObjectProxy for this manager
         */
         
        unsigned int offset;
        if (object >= domainProxyMemory &&
            object <= domainProxyMemoryLast) {
                offset = (unsigned int) object - 
                         (unsigned int) domainProxyMemory; 
        } else if (object >= kernelProxyMemory &&  
                 object <= kernelProxyMemoryLast) {
                offset = (unsigned int) object -  
                         (unsigned int) kernelProxyMemory;
        } else {  
                return 0; 
        }
        return ( ( ( offset % sizeof(ObjectProxy) ) == 0 ) &&
                   ( offset < sizeOfAllocatedProxies )  && 
                   ( object->prev != 0 ) );
}

void
ObjectProxyManager::freeAll()
{
	Raids( RaidMethod ) << this << "::freeAll() "
		<< "proxiesAllocated " << proxiesAllocated << "\n" << eor;

	if ( proxiesAllocated == 0 )
		return;

	/*
	 * Assume that while the ObjectProxyManager is being cleaned up no
	 *   other processes will be still allocating and freeing ObjectProxies.
	 * Since this is called by a kernel process and may not be in a
	 *   Domain where the proxies are can be accessed, need to add
	 *   the proxyMemoryObject to the kernel domain and offset the
	 *   pointers to where it gets mapped in the kernel domain.
	 */
	Domain * thisDom = TheKernel->domain();

	int offset = 0;
	ObjectProxy *obj = usedList;
	if ( firstToFree != 0 ) {
		/*
		 * free this proxy to this object first if it's there
		 */
		ObjectProxy * firstObj = 0;
#ifndef NOHASH
		OPMEntry key;
		key.k._proxiedObject = firstToFree;
		key.k._proxiedDomain = thisDom;
		key.k._table = firstToFree->classOf()->proxyTable();
		
		OPMEntry * r;
		if ((r = (OPMEntry *)hash->find(&key)))
		    firstObj = (ObjectProxy *)((char*)r->p - offset);
#else
		for ( firstObj = usedList->next; firstObj != usedList;
						firstObj = firstObj->next )
			if ( firstObj->proxiedObject() == firstToFree )
				break;
		
		if ( firstObj == usedList )
			firstObj = 0;
#endif
		
		if ( firstObj != 0 ) {
			Assert( firstObj->proxiedDomain() == thisDom );
			firstToFree->unreference();
			firstObjectProxy = (ObjectProxy *)
				( ( char *) firstObjectProxy - offset );
			free( firstObj );
			firstObjectProxy = (ObjectProxy *)
				( ( char *) firstObjectProxy + offset );
		}
	}

	while ( ( obj = usedList->next ) != usedList ) {
		/*
		 * free all the used ObjectProxies.  Need to unreference
		 *    each of the proxied objects first, in the Domain
		 *    of the proxied object.
		 */
#ifdef NEW
Assert (obj->proxiedDomain() == thisDom);
#endif
		if ( obj->proxiedDomain() == thisDom )
		 	obj->proxiedObject()->unreference();
		else {
			ObjectProxy *proxy = thisDom->proxyManager()
			    ->allocate( obj->proxiedObject(),
				ProxiableObjectClass, obj->proxiedDomain(), 1 );
			proxy->unreference();
		}
		firstObjectProxy = (ObjectProxy *)
			( ( char *) firstObjectProxy - offset );
		free( obj );
		firstObjectProxy = (ObjectProxy *)
			( ( char *) firstObjectProxy + offset );
	}

	if ( offset != 0 ) {
		char * whereItWas = (char *)
			thisDom->remove( proxyMemoryObject, (const char *) 
					 ( firstObjectProxy - offset ) );
		
		Assert( whereItWas == (char *) firstObjectProxy - offset );
	}

	Assert( proxiesAllocated == 0 );
}

ObjectProxy *
ObjectProxyManager::allocate( ProxiableObject * who, Class * whoClass,
					Domain * dom, int returnsRef )
{
	Raids( RaidMethod ) << this << "::allocate( "
		<< who << ", " << whoClass << ", " << dom << ", " 
		<< returnsRef << " )\n" << eor;

	ObjectProxy *obj = 0;
#ifndef NOHASH
	OPMEntry * key = new OPMEntry(); // To be inserted later.
	int KeyInserted = 0;	// Tell if key has been inserted.
#endif
	mutex.P();

#ifndef NOHASH
	key->k._proxiedObject = who;
	key->k._proxiedDomain = dom;
	if ( whoClass != 0 )
		key->k._table = whoClass->proxyTable();
	else
		key->k._table = 0;

	OPMEntry * r;

	if (who == 0 ||
	    !(r = (OPMEntry *)hash->find(key), (r ? obj = r->p : 0), r)) {
#endif
		if ( freeList == 0 ) {
			/*
			 * get a new one from above the high-water mark
			 */
			if ( nextObjectProxy == lastObjectProxy )
				obj = 0;
			else {
				obj = nextObjectProxy++;
				sizeOfAllocatedProxies += sizeof( ObjectProxy );
			}
		}
		else {
			/*
			 * take one off the free list
			 * Note: only the 'next' pointers are maintained on the
			 *   free list, not the 'prev' pointers.
			 */
			obj = freeList;
			freeList = obj->next;
		}

		Assert( obj != 0 ) 
			("ObjectProxyManager::allocate: out of proxies\n");

		Object *constructedObj = new ( (void *) obj ) ObjectProxy();
		Assert( constructedObj == obj );


#ifndef NOHASH
		if ( who != 0 ) {
		    Assert( !hash->find(key) );
		    key->p = obj;
		    hash->insert(key);
		    KeyInserted = 1;
		}
#endif
		proxiesAllocated++;
		obj->next = usedList->next;
		obj->prev = usedList;
		obj->next->prev = obj;
		usedList->next = obj;

		obj->notReferenced = 0;

#ifndef NOHASH
	}
	if (!KeyInserted)
	    delete key;
#endif

	obj->setProxiedObject( who, whoClass, dom, returnsRef );

	Raids( RaidMethod ) << this << "::allocate() returns " 
		<< asHex( (unsigned) obj ) << "\n" << eor;

	mutex.V();

	Debug.printf("ObjectProxyManager::allocate(%N) returns %A\n", who, obj);

	return( obj );
}

void
ObjectProxyManager::free( ObjectProxy * obj )
{
	Assert( isValidProxy( obj ) );
	Assert( proxiesAllocated > 0 );

	Raids( RaidMethod ) << this << "::free( " << asHex( (unsigned ) obj )
		<< " )\n" << eor;

	mutex.P();

	Raids( RaidDetail ) << this << "::free() "
		<< "proxied object " << obj->proxiedObject() << ", "
		<< "domain " << obj->proxiedDomain() << "\n" << eor;

	obj->prev->next = obj->next;
	obj->next->prev = obj->prev;
	proxiesAllocated--;

#ifndef NOHASH
	if ( obj->key._proxiedObject != 0 ) {
	    OPMEntry r;
	    r.k = obj->key;
	    Link * l = hash->remove(&r);
	    Assert(l);
	    delete l;
	}
#endif

	obj->next = freeList;
	obj->prev = 0;
	freeList = obj;
	obj->notReferenced = 0;

	mutex.V();
}

void
ObjectProxyManager::clone( ObjectProxyManager * opm )
{

	Raids( RaidMethod ) << this << "::clone( " << opm << " )\n" << eor;

	opm->mutex.P();
	mutex.P();

	Raids( RaidDetail ) << this << "::clone() acquired locks\n" << eor;

	/* offset values */

	int offset = (unsigned int) kernelProxyMemory -
		     (unsigned int) opm->kernelProxyMemory;

	/* copy values */

	freeList = 
		(ObjectProxy *) ((char *) opm->freeList + offset);
	firstObjectProxy = 
		(ObjectProxy *) ((char *) opm->firstObjectProxy + offset);
	nextObjectProxy = 
		(ObjectProxy *) ((char *) opm->nextObjectProxy + offset);
	lastObjectProxy = 
		(ObjectProxy *) ((char *) opm->lastObjectProxy + offset);
	sizeOfAllocatedProxies = opm->sizeOfAllocatedProxies;
	proxiesAllocated = opm->proxiesAllocated;
	firstToFree = opm->firstToFree;

	ByteCopy( opm->firstObjectProxy, firstObjectProxy,
		  (unsigned int) nextObjectProxy - 
		  (unsigned int) firstObjectProxy );

	/* update pointers within proxy memory object */

	ObjectProxy * obj = firstObjectProxy;
	while(obj < nextObjectProxy) {
		obj->next = (ObjectProxy *) ((char *) obj->next + offset);
		if (obj->prev) 
			obj->prev = (ObjectProxy *) 
				    ((char *) obj->prev + offset);
		obj++;
	}

	/* update pointers to and from used list */

	if ( opm->usedList->prev == opm->usedList ) {
		Assert( opm->usedList->prev == opm->usedList->next );
		usedList->next = usedList;
		usedList->prev = usedList;
	} else {
		Assert( opm->usedList->next != opm->usedList );
		ObjectProxy * usedListNext = ( ObjectProxy * ) 
			( (char *) firstObjectProxy + 
			  ( ( unsigned ) opm->usedList->next - 
			    ( unsigned ) opm->firstObjectProxy ) );
		usedListNext->prev = usedList;
		usedList->next = usedListNext;

		Assert( opm->usedList->prev != opm->usedList );
		ObjectProxy * usedListPrev = ( ObjectProxy * ) 
			( (char *) firstObjectProxy + 
			  ( ( unsigned ) opm->usedList->prev - 
			    ( unsigned ) opm->firstObjectProxy ) );
		usedListPrev->next = usedList;
		usedList->prev = usedListPrev;
	}

	/* reference proxiable objects in usedList */

	obj = usedList;
	while ( ( obj = obj->next ) != usedList ) {
		obj->proxiedObject()->reference();
#ifndef NOHASH
		OPMEntry * key = new OPMEntry();
		key->k._proxiedObject = obj->proxiedObject(); 
		key->k._proxiedDomain = obj->proxiedDomain();
		key->k._table = obj->table();
		key->p = obj;
		hash->insert( key );
#endif
	}

	mutex.V();
	opm->mutex.V();

	Raids( RaidMethod ) << this << "::clone() done\n" << eor;
}

void
ObjectProxyManager::incrementAllReferenceCounts()
{
	//  Double the reference counts of all the ObjectProxies.
	ObjectProxy *obj = usedList -> next;
	while (obj != usedList) {
		int v = obj -> _referenceCount.value ();
		for (int i = 1; i <= v; i++)
			obj -> _referenceCount.increment ();
		obj = obj -> next;
	}
}

#include "ClassInstaller.h"
DEFINECLASS("ObjectProxyManager",ObjectProxyManager,Object,0,0);

