#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David W. Dykstra (dykstra@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * PeriodicTimer.c - machine independent portion of periodic timers
 */

#include "Assert.h"
#include "Lock.h"
#include "TIMER_MANAGER.h"
#include "OutputStream.h"
#include "PeriodicTimer.h"
#include "Semaphore.h"


#ifdef __GNUC__
#pragma implementation
#endif

TimerManager *PeriodicManager;
int periodicTimerInitialized = 0;

Lock *PeriodicTimer::lock = 0;
PeriodicTimer *PeriodicTimer::first = 0;

void
PeriodicTimerEntry(int arg)
{
	/* This is the periodic timer process.  It calls the machine dependent */
	/*  PeriodicManager->processHandler() which calls back here to 	   */
	/*  perTickBookKeeping and expiryProcessing when appropriate.	   */

	Debug << "PeriodicTimerEntry( " << asHex( arg ) << " )\n" << eor;

	PeriodicTimer::initialize();

	PeriodicManager = new TIMER_MANAGER(&PeriodicTimer::perTickBookKeeping, &PeriodicTimer::expiryProcessing);

	Assert(PeriodicManager != 0);

	periodicTimerInitialized = 1;
	PeriodicManager->processHandler();

	Assert(NOTREACHED);
}

/*
 * Now for the regular, per-instance functions
 */

PeriodicTimer::PeriodicTimer()
{
	Debug << this << "::PeriodicTimer\n" << eor;

	PeriodicManager->initialize();	/* in case not started yet */

	periodicSemaphore = new Semaphore(0);
	Assert(periodicSemaphore != 0);
	maxCount = 0;
	running = 0;
	nextDifferent = (PeriodicTimer *) 0;
	nextSame = (PeriodicTimer *) 0;
}

PeriodicTimer::~PeriodicTimer()
{
	Debug << this << "::~PeriodicTimer\n" << eor;

	if (running)
		(void) stop();
}

void
PeriodicTimer::await()
{
	Debug << this << "::await()\n" << eor;
	Assert(periodicSemaphore != 0);
	periodicSemaphore->P();
}

unsigned int
PeriodicTimer::start( unsigned int microseconds )
{
	PeriodicTimer *ptr,*prevptr;
	int setTicks;

	Debug << this << "::start( " << microseconds << ")\n" << eor;

	Assert(microseconds != 0);

	if (running)
	{
		/* It's assumed that no other process will try to */
		/*  start this same timer after this point.       */
		(void) stop();
	}

	/* microsecondsToTicks() will round up to the nearest number of ticks */
	setTicks = PeriodicManager->microsecondsToTicks(microseconds);

	Debug << this << "::start: " << setTicks << " ticks\n" << eor;

	Assert(running == 0);
	Assert(maxCount == 0);
	Assert(nextDifferent == (PeriodicTimer *) 0);
	Assert(nextSame == (PeriodicTimer *) 0);

	prevptr = (PeriodicTimer *) 0;
	maxCount = setTicks;

	lock->acquire();

	for (ptr = first; ptr != (PeriodicTimer *) 0; ptr = ptr->nextDifferent) {
		if (ptr->maxCount == setTicks)
	    break;
		prevptr = ptr;
	}

	/* put this timer at the head in the current position */
	if (prevptr == (PeriodicTimer *) 0) {
		Assert(first == ptr);
		first = this;
	}
	else
		prevptr->nextDifferent = this;

	if (ptr == (PeriodicTimer *) 0) {
		/* No other timer with this period, at the tail of the Different list */
		count = setTicks;
	}
	else {
		/* found a timer with the same period, take its place in Same list */
		nextDifferent = ptr->nextDifferent;
		ptr->nextDifferent = (PeriodicTimer *) 0;
		nextSame = ptr;
		count = ptr->count;
		ptr->count = 0;
	}

	running = 1;

	lock->release();

	return(PeriodicManager->ticksToMicroseconds(setTicks));
}

unsigned int
PeriodicTimer::stop()
{
	unsigned int resid;
	PeriodicTimer *ptr, *prevptr;

	Debug << this << "::stop()\n" << eor;

	lock->acquire();
	if (running)
	{
		prevptr = (PeriodicTimer *) 0;
		for (ptr = first; ptr->maxCount != maxCount; ptr = ptr->nextDifferent) {
	    Assert(ptr != (PeriodicTimer *) 0);

	    prevptr = ptr;
		}
		resid = PeriodicManager->ticksToMicroseconds(ptr->count);
		if (ptr == this) {
	    /* it was at the head of its list */
	    if (nextSame == (PeriodicTimer *) 0) {
				/* it was the only one in its list */
				if (prevptr == (PeriodicTimer *) 0) {
					Assert(first == ptr);
					first = nextDifferent;
				}
				else
					prevptr->nextDifferent = nextDifferent;
	    }
	    else {
				/* there are more in this list, delete head */
				if (prevptr == (PeriodicTimer *) 0) {
					Assert(first == ptr);
					first = nextSame;
				}
				else
					prevptr->nextDifferent = nextSame;
				nextSame->nextDifferent = nextDifferent;
				Assert(nextSame->count == 0);
				Assert(nextSame->maxCount == maxCount);
				nextSame->count = count;
				nextSame = (PeriodicTimer *) 0;
	    }
	    count = 0;
	    nextDifferent = (PeriodicTimer *) 0;
		}
		else {
	    /* it was not at the head, find out which one */
	    prevptr = ptr;
	    for (ptr = prevptr->nextSame; ptr != this; ptr = ptr->nextSame) {
				Assert(ptr != (PeriodicTimer *) 0);

				prevptr = ptr;
	    }
	    prevptr->nextSame = nextSame;
	    nextSame = 0;
	    Assert(nextDifferent == (PeriodicTimer *) 0);
	    Assert(count == 0);
		}

		maxCount = 0;
		running = 0;
	}
	else
		resid = 0;

	lock->release();

	return(resid);
}

unsigned int
PeriodicTimer::residual()
{
	unsigned int resid;
	PeriodicTimer *ptr;

	Debug << this << "::residual()\n" << eor;

	lock->acquire();

	if (running) {
		for (ptr = first; ptr->maxCount != maxCount; ptr = ptr->nextDifferent) {
	    Assert(ptr != (PeriodicTimer *) 0);
		}
		resid = PeriodicManager->ticksToMicroseconds(ptr->count);
	}
	else
		resid = 0;

	lock->release();

	return(resid);
}

/*
 * Now for the class functions
 */

void
PeriodicTimer::initialize()
{
	lock = new Lock();
	Assert(lock != 0);

	first = 0;

}

int
PeriodicTimer::perTickBookKeeping()
{
	/* this is called directly from the interrupt handler */

	// Debug << "PeriodicTimer::perTickBookKeeping\n" << eor;

	int expiring = 0;
	PeriodicTimer *ptr;

	lock->acquire();

	for (ptr = first; ptr != (PeriodicTimer *) 0; ptr = ptr->nextDifferent) {
		Assert(ptr->count > 0);
		if (--ptr->count == 0)
	    expiring = 1;
	}
	lock->release();

	return(expiring);
}

void
PeriodicTimer::expiryProcessing()
{
	PeriodicTimer *ptr, *ptr2;

	/* At this point we are protected against any other ticks from going off */

	lock->acquire();

	for (ptr = first; ptr != (PeriodicTimer *) 0; ptr = ptr->nextDifferent) {
		if (ptr->count == 0) {
	    for (ptr2 = ptr; ptr2 != (PeriodicTimer *) 0; ptr2 = ptr2->nextSame) {
				ptr2->periodicSemaphore->V();
	    }
	    ptr->count = ptr->maxCount;
		}
	}

	lock->release();
}

#include "ClassInstaller.h"
DEFINECLASS("PeriodicTimer",PeriodicTimer,Timer,0,0);
