#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Pipe.cc
 */

#include "Assert.h"
#include "Byte.h"
#include "Pipe.h"
#include "Process.h"
#include "PROCESSOR.h"
#include "Raid.h"
#include "ThisProcess.h"


#ifdef __GNUC__
#pragma implementation
#endif

Pipe::Pipe( int size )
     : _size( size ), _copyOutSem( 0 ), _copyInSem( 0 ), 
       _readQueue( 1 ), _writeQueue( 1 )
{
	initDebug();
	Raids( RaidConstructor ) << this << "::Pipe( " << size 
		<< " )\n" << eor;
	Assert( _size != 0 );
	_buffer = new char [ _size ];
	Assert( _buffer != 0 );
	_count = 0;
	_head = 0;
	_tail = 0;
	_readBlocked = 0;
	_writeBlocked = 0;
	_readPending = 0;
	_writePending = 0;
	_readQueueLength = 0;
	_writeQueueLength = 0;
}

Pipe::~Pipe()
{
	Raids( RaidConstructor ) << this << "::~Pipe()\n" << eor;
	_lock.acquire();
	delete _buffer;	
	_buffer = 0;
	if ( _readQueueLength > 0 ) {
		Assert( _readBlocked );
		_copyOutSem.V();
		_readQueueLength--;
		while ( _readQueueLength-- > 0 ) 
			_readQueue.V();
	}
	if ( _writeQueueLength > 0 ) {
		Assert( _writeBlocked );
		_copyInSem.V();
		_writeQueueLength--;
		while ( _writeQueueLength-- > 0 )
			_writeQueue.V();
	}
	_lock.release();
}

int 
Pipe::read( char * buf, int count, int block)
{
	Raids( RaidMethod ) << this << "::read( "
		<< asAddress( buf ) << ", " << count << ", " << block
		<< " )\n" << eor;

	int completed = 0;

	_lock.acquire();
	if ( _readPending > ( _writePending + used() ) && !block ) {
		_lock.release();
		return 0;
	}

	_readQueueLength++;
	_readPending += count;
	_lock.release();
	_readQueue.P();
	if ( _buffer == 0 )
		return 0;
	_lock.acquire();

	do {
		if ( used() == 0 ) {
			_readBlocked = 1;
			_lock.release();
			_copyOutSem.P();
			if ( _buffer == 0 )
				return 0;
			_lock.acquire();
		}
		int nbytes = count > used() ? used() : count;
		removeFromBuffer( buf, nbytes );
		buf += nbytes;
		count -= nbytes;
		completed += nbytes;
		_readPending -= nbytes;
		if ( _writeBlocked ) {
			Assert( _writeQueueLength > 0 );
			Assert( _writePending > 0 );
			_writeBlocked = 0;
			_copyInSem.V();
		}
	} while ( count > 0 && ( _writePending + used() ) > 0 );

	_readPending -= count;
	_readQueueLength--;
	_readQueue.V();
	Assert( completed > 0 );
	Assert( _readQueueLength > 0 || _readPending == 0 );
 
	_lock.release();

	return completed;
}

int 
Pipe::write( char * buf, int count, int block)
{
	Raids( RaidMethod ) << this << "::write( "
		<< asAddress( buf ) << ", " << count << ", " << block
		<< " )\n" << eor;

	int completed = 0;

	_lock.acquire();
	if ( free() < ( _writePending + used() + count  - _readPending ) 
	     && !block ) {
		_lock.release();
		return 0;
	}

	_writeQueueLength++;
	_writePending += count;
	_lock.release();
	_writeQueue.P();
	if ( _buffer == 0 )
		return 0;
	_lock.acquire();

	while ( count > 0 ) {
		if ( free() == 0 ) {
			_writeBlocked = 1;
			_lock.release();
			_copyInSem.P();
			if ( _buffer == 0 )
				return 0;
			_lock.acquire();
		}
		int nbytes = count > free() ? free() : count;
		insertIntoBuffer( buf, nbytes );
		buf += nbytes;
		count -= nbytes;
		completed += nbytes;
		_writePending -= nbytes;
		if ( _readBlocked ) {
			Assert( _readQueueLength > 0 );
			Assert( _readPending > 0 );
			_readBlocked = 0;
			_copyOutSem.V();
		}
	}

	_writeQueueLength--;
	_writeQueue.V();
	Assert( _writeQueueLength > 0 || _writePending == 0 );

	_lock.release();

	return completed;
}	

void 
Pipe::insertIntoBuffer( char * buf, int count )
{
	Assert( count <= free() );
	int total = count;
	int nbytes = _size - _tail;
	if ( nbytes > total ) 
		nbytes = total;
	ByteCopy( buf, _buffer + _tail, nbytes );
	total -= nbytes;
	if ( total > 0 )
		ByteCopy( buf + nbytes, _buffer, total );
	_count += count;
	Assert( _count >= 0 && _count <= _size );
	_tail += count;
	if ( _tail >= _size )
		_tail -= _size;
}

void 
Pipe::removeFromBuffer( char * buf, int count )
{
	Assert( count <= used() );
	int total = count;
	int nbytes = _size - _head;
	if ( nbytes > total ) 
		nbytes = total;
	ByteCopy( _buffer + _head, buf, nbytes );
	total -= nbytes;
	if ( total > 0 )
		ByteCopy( _buffer, buf + nbytes, total );
	_count -= count;
	Assert( _count >= 0 && _count <= _size );
	_head += count;
	if ( _head >= _size )
		_head -= _size;
}

#include "ClassInstaller.h"
DEFINECLASS("Pipe",Pipe,ProxiableObject,0,0);

