#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * RaidOutputStream.cc
 *
 */

#include "Assert.h"
#include "Async.h"
#include "Byte.h"
#include "FileSystemInterface.h"
#include "FileSystemInterfaceStar.h"
#include "Kernel.h"
#include "Lock.h"
#include "PersistentMemoryObject.h"
#include "PersistentMemoryObjectStar.h"
#include "Process.h"
#include "PROCESSOR.h"
#include "RaidOutputStream.h"
#include "TimeoutTimer.h"


#ifdef RAIDOUTPUTSTREAM

RaidOutputStream StandardRaidStream;

const int RaidBufferSize = ( 1 << 18 );
const int RaidBufferHalfSize = RaidBufferSize / 2;
char RaidBuffer[RaidBufferSize+1];
Lock RaidLock;
int RaidConsole = 0;			// echo output to console
int RaidBufferDisable = 0;		// turn on RaidBuffer
int RaidLastWrittenHalf = 1;
int RaidWrap;
char * RaidHead;
char * RaidBufferLimit;
char * RaidBufferHalf;
char * RaidFileName = "/mnt/local/RaidLog";
PersistentMemoryObjectStar RaidFileObject = 0;

RaidAsyncSupport::RaidAsyncSupport() {
	_func = &func;
	_inUse = 0;
	_whichHalf = 0;
	_urgency = 0;
}

RaidAsyncSupport RaidAsync;

static int pointer = 0;

int RaidOpenFile()
{
	FileSystemInterfaceStar fsi = 0;
	if (TheKernel) { 
		fsi = TheKernel->fileSystemInterface();
	}
	if (fsi == 0) {
		Console()
 		<< "\nRaidLog not written, kernel or FSI not ready\n"
		<< eor;
		return -1;
	}
	int status = 0;
	RaidFileObject = 
		fsi->open(RaidFileName, 
			  PersistentMemoryObjectClass, 
			  status,
			  O_NCACHE | O_CREAT | O_TRUNC | O_WRONLY,
			  0777);
	if (RaidFileObject == 0) {
		Console()
		<< "\nCannot open " << RaidFileName 
		<< " as RaidLog, status " << status << "\n" << eor;
		return -1;
	}
	return 0;
}

void RaidAsyncSupport::func( AsyncClosure * _c )
{
	RaidAsyncSupport *c = static_cast<RaidAsyncSupport *>(_c);
	if (RaidFileObject == 0) {
		int status = RaidOpenFile();
		if (status != 0 ) {
			c->_inUse = 0;
			return;
		}
	}
	Console() 
		<< "\nRaidLog write starting at " 
		<< ( pointer << RaidFileObject->log2UnitSize() )
		<< "\n" << eor;
	int count = RaidBufferHalfSize; 
	int units = count >> RaidFileObject->log2UnitSize();
	char * addr = c->_whichHalf == 0 ? RaidBuffer : RaidBufferHalf;
	int written = RaidFileObject->writeThrough(pointer, units, addr, count);
	RaidFileObject->setNumberOfBytes
		( ( pointer + units ) << RaidFileObject->log2UnitSize() );
	if (written != units) {
		Console()
		<< "\nRaidLog write error only " << written << " units wrote\n"
		<< eor;
	} else { 
		Console()
		<< "\nRaidLog write successful\n" << eor;
	}
	RaidLastWrittenHalf = c->_whichHalf;
	c->_inUse = 0;
	pointer += units;
	return;
}

inline 
int whichHalf(char * p)
{
	return ( p < RaidBufferHalf ? 0 : 1 );
}

RaidOutputStream::RaidOutputStream()
{
	RaidConsole = 0;
	RaidWrap = 0;
	ByteZero(RaidBuffer, RaidBufferSize);
	RaidBufferLimit = RaidBuffer + RaidBufferSize;
	RaidBufferHalf = RaidBuffer + RaidBufferHalfSize;
	RaidHead = RaidBuffer;
	*RaidHead = 0;
	*RaidBufferLimit = 0;
}

int
RaidOutputStream::write(const char * buffer, int count)
{
	if (RaidConsole) {
		Console().write(buffer, count);
		Console().flush();
	}
	if (RaidBufferDisable) 
		return count;
	char * orgHead = RaidHead;
	int orgCount = count;
	RaidLock.acquire();
	const char * cp = buffer;
	while( count > 0 ) {
		if (RaidHead == RaidBufferLimit) {
			RaidHead = RaidBuffer;
			RaidWrap++;
		}
		*RaidHead++ = *cp++;
		count--;
	}
	*RaidHead = 0;
	int doWrite;
	if (whichHalf(orgHead) != whichHalf(RaidHead)) {
		if ( RaidAsync._inUse ) {
			Console() 
			<< "\nRaidBuffer too small, "
			<< "cannot stagger writes\n" << eor;
			doWrite = 0;
		} else {
			doWrite = 1;
			RaidAsync._whichHalf = whichHalf(orgHead);
			RaidAsync._inUse = 1;
		}
	} else {
		doWrite = 0;
	}
	RaidLock.release();
	if ( doWrite ) {
		if (TheAsyncSupport == 0) {
			Console()
			<< "\nRaidBuffer too small, "
			<< "cannot write before AsyncSupport is ready\n" << eor;
			RaidAsync._inUse = 0;
		} else {
			TheAsyncSupport->schedule(&RaidAsync);
		}
	}
	return orgCount;
}

void
RaidOutputStream::dump()
{
	Console() 
		<< "\nRaidOutputStream::dump\n\n"
		<< "Buffer size:           " << RaidBufferSize << "\n"
		<< "No of times wrapped:   " << RaidWrap << "\n"
		<< "\n\n" << eor;
	
	TimeoutTimer timer;
		
	int waitCount = 0;
	while (RaidAsync._inUse && waitCount < 5) {
		waitCount++;
		Console() << "RaidLog busy, wait...\n" << eor;
		timer.start(1000000);
		timer.await();
	}

	IntStorage state ; PROCESSOR::disableInterrupts(state);
	if (RaidAsync._inUse) {
		Console() 
		<< "\nRaidLog still in use, dumping duplicate just in case\n" << eor;
		char * addr = 
		    RaidAsync._whichHalf == 0 ? RaidBuffer : RaidBufferHalf;
		char save = addr[RaidBufferHalfSize]; 
		addr[RaidBufferHalfSize] = 0;
		Console() << addr << eor;
		addr[RaidBufferHalfSize] = save;
	}
	char * addr = 
	        whichHalf(RaidHead) == 0 ? RaidBuffer : RaidBufferHalf;
	if (whichHalf(RaidHead) == RaidLastWrittenHalf) {
		Console()
		<< "\nRaidBuffer weird, currentHalf == RaidLastHalfWritten\n"
		<< eor;
	}
	if (RaidConsole)
		Console() << addr << eor;
/*
	if (RaidWrap) {
		Console() << RaidHead + 1 << eor;
	}
	Console() << RaidBuffer << eor;
 */
	int log2UnitSize = RaidFileObject->log2UnitSize();
	int nbytes = RaidHead - addr;
	int units = ( nbytes + ( (1 << log2UnitSize) - 1 ) ) >> log2UnitSize;
	int count = units << log2UnitSize;
	PROCESSOR::restoreInterrupts(state);
	if (RaidFileObject == 0) {
		int status = RaidOpenFile();
		if (status != 0 ) 
			return;
	}
	int written = RaidFileObject->writeThrough(pointer, units, addr, count);
	RaidFileObject->setNumberOfBytes
		( ( pointer << log2UnitSize) + nbytes );
	if (written != units) {
		Console()
		<< "\nRaidBuffer write error only " 
		<< written << " units wrote, should be " << units
		<< eor;
	} else { 
		Console()
		<< "\nRaidBuffer write successful\n" << eor;
	}
}

void
RaidDump()
{
	StandardRaidStream.dump();
}

#else // no RAIDOUTPUTSTREAM

RaidOutputStream::RaidOutputStream()
{
}

int
RaidOutputStream::write(const char * buffer, int count)
{
	return count;
}

void 
RaidOutputStream::dump()
{
	return;
}

void
RaidDump()
{
	return;
}

#endif // RAIDOUTPUTSTREAM

#include "ClassInstaller.h"
DEFINECLASS("RaidOutputStream",RaidOutputStream,OutputStream,0,0);

