#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Author: Roland Ouellette (ouellette@tarkin.enet.dec.com)
 * Author: Roy H Campbell (roy@cs.uiuc.edu)
 * Author: See-Mong Tan (stan@cs.uiuc.edu)
 * Author: David Raila (raila@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Semaphore.c - Semaphore class implementation.
 */


#include "OutputStream.h"
#include "Semaphore.h"
#include "Process.h"
#include "ThisProcess.h"
#include "Raid.h"
#include "PROCESSOR.h"


#ifdef __GNUC__
#pragma implementation
#endif

Semaphore::Semaphore( int initialCount, int max )
{
	initDebug();
	Raid( RaidConstructor )( "%N::Semaphore( %d, %d )\n",
				 this, initialCount, max );
	Assert( this != 0 );
	count = initialCount;
	maxCount = max;
	_holder = 0;
	Assert( initialCount <= maxCount );
}

Semaphore::~Semaphore()
{
	Raid( RaidConstructor )( "%N::~Semaphore()\n", this );
	Assert( queue.isEmpty() );
}

void
Semaphore::P()
{
	Raid( RaidMethod ) ( "%N::P(), count = %d\n", this, count );

	Assert( this != 0 );

	IntStorage intState ; PROCESSOR::disableInterrupts(intState);
	lock.acquire();

	Process * process = thisProcess();
	Raid( RaidMethod ) ( "%N::P(), process = %N\n", this, process );
	// Assert( process == 0 || process->numLocksHeld() == 1);//("%N[%A]: Potential block while holding %d locks!\n", process, process, process->numLocksHeld());

	if( --count < 0 ) {
		/*
		 * We have to block so get another process to run next.
		 * If there is'nt one panic for now.
		 */
		queue.add( process );
		lock.release();
		Raid( RaidDetail ) ( "%N::P(), block process %N\n",
			this, process );
		process->block(this);
		_holder = process;
		Raid( RaidDetail ) ( "%N::P(), process %N unblocked\n", 
			this, process);
	}
	else {
		/*
		 * We definitely got the semaphore.
		 */
		Assert( count >= 0 );
		_holder = process;
		lock.release();
	}
	PROCESSOR::restoreInterrupts(intState);
}

void
Semaphore::V()
{
	Raid( RaidMethod ) ( "%N::V(), count = %d\n", this, count );

	Assert( this != 0 );

	IntStorage intState ; PROCESSOR::disableInterrupts(intState);
	lock.acquire();
	count++;
	Assert( count <= maxCount );
	if( count <= 0 ) {
		Process * waiter = queue.remove();
		_holder = waiter;
		lock.release();
		Assert( waiter != 0 );
		Raid( RaidDetail )( "%N::V(): %N resuming %N\n", 
			this, thisProcess(), waiter );
		waiter->ready();
	}
	else {
		_holder = 0;
		lock.release();
	}
	PROCESSOR::restoreInterrupts(intState);
}




BinarySemaphore::BinarySemaphore( int initialCount )
	: Semaphore( initialCount, 1 )
{
	initDebug();
	Raid( RaidConstructor )( "%N::BinarySemaphore( %d )\n",
				 this, initialCount );

	Assert( this != 0 );
	Assert( (initialCount == 0) || (initialCount == 1) );
}

BinarySemaphore::~BinarySemaphore()
{
	Raid( RaidConstructor )( "%N::~BinarySemaphore()\n", this );
}



GraciousSemaphore::GraciousSemaphore( int initialCount, int maxCount )
	: Semaphore( initialCount, maxCount )
{
	initDebug();

	Raid( RaidConstructor )( "%N::GraciousSemaphore( %d, %d )\n",
				 this, initialCount, maxCount );
    givingProcess = 0;
}


GraciousSemaphore::~GraciousSemaphore()
{
	Raid( RaidConstructor )( "%N::~GraciousSemaphore()\n", this );
}

void
GraciousSemaphore::P()
{
	Raid( RaidMethod ) ( "%N::P(), count = %d\n", this, count );

	Assert( this != 0 );

	IntStorage intState ; PROCESSOR::disableInterrupts(intState);
	lock.acquire();

	Process * process = thisProcess();
	Raid( RaidMethod ) ( "%N::P(), process = %N\n", this, process );
	// Assert( process == 0 || process->numLocksHeld() == 1);//("%N[%A]: Potential block while holding %d locks!\n", process, process, process->numLocksHeld());

	if( --count < 0 ) {
		/*
		 * We have to block so get another process to run next.
		 * If there is'nt one panic for now.
		 */
		queue.add( process );
		lock.release();
		Raid( RaidDetail ) ( "%N::P(), block process %N\n",
			this, process );
        // Give back control to the process that switched to us
        if (!givingProcess) {
            process->block(this);
        } else {
            if (process->_oldContext) {
                process->_oldContext = 0;
                delete process->_oldContext;
            }
            process->_blockedOn = this;
            process->giveProcessorTo( givingProcess, ProcessBlocked );
            process->_blockedOn = 0;
        }
		_holder = process;
		Raid( RaidDetail ) ( "%N::P(), process %N unblocked\n", 
			this, process);
	}
	else {
		/*
		 * We definitely got the semaphore.
		 */
		Assert( count >= 0 );
		_holder = process;
		lock.release();
	}
	PROCESSOR::restoreInterrupts(intState);
}

void
GraciousSemaphore::V()
{
	Raid( RaidMethod ) ( "%N::V(), count = %d\n", this, count );

	Assert( this != 0 );

	IntStorage intState ; PROCESSOR::disableInterrupts(intState);
	lock.acquire();
	count++;
	Assert( count <= maxCount );

	/*
	 * If there is a process waiting for the semaphore, run it
	 * instead of ourselves by handing the CPU to it. It will put
	 * us back in our readyContainer so we can run later or now
	 * on another processor if there is an idle one.
	 */
	if( count <= 0 ) {
		Process * waiter = queue.remove();
		_holder = waiter;
		lock.release();
		Assert( waiter != 0 );
		Raids( RaidMethod ) << this << "::V(): " << thisProcess()
			<< "::giveProcessorTo " << waiter << "\n" << eor;
        /*
         * ThIS MIGHT VIOLATE THE SCHEDULER RULES
         */
		givingProcess = thisProcess();
		// givingProcess->ready();  /* Put me on back of readyqueue*/
		givingProcess->giveProcessorTo( waiter, ProcessReady );
	}

	/*
	 * If the semaphore is not being waited for, just unlock and return.
	 */
	else {
		Assert (_holder != 0);
		_holder = 0;
		lock.release();
	}

	PROCESSOR::restoreInterrupts(intState);
}


LazySemaphore::LazySemaphore( int initialCount )
    : BinarySemaphore( initialCount )
{
	Raid( RaidConstructor )( "%N::LazySemaphore( %d )\n",
				 this, initialCount );
}

LazySemaphore::~LazySemaphore()
{
	Raid( RaidConstructor )( "%N::~LazySemaphore()\n", this );
}

void
LazySemaphore::V()
{
	Raid( RaidDetail ) ( "%N::V(), count = %d\n", this, count );

	Assert( this != 0 );

	IntStorage intState ; PROCESSOR::disableInterrupts(intState);
	lock.acquire();
	count++;
	if( count > maxCount )
	        count = maxCount;
	if( count <= 0 ) {
		Process * waiter = queue.remove();
		lock.release();
		Assert( waiter != 0 );
		Debug.printf( "LazySemaphore::V(): resuming %A\n", waiter );
		waiter->ready();
	}
	else {
		lock.release();
	}
	PROCESSOR::restoreInterrupts(intState);
}

MultiVSemaphore::MultiVSemaphore( int initialCount, int maxCount ) 
    : Semaphore( initialCount, maxCount )
{
	Raid( RaidConstructor )( "%N::MultiVSemaphore( %d, %d )\n",
				 this, initialCount, maxCount );
}

MultiVSemaphore::~MultiVSemaphore()
{
	Raid( RaidConstructor )( "%N::~MultiVSemaphore()\n", this );
}

void
MultiVSemaphore::V( int how_many )
{
	Raid( RaidDetail ) ( "%N::V( %d ), count = %d\n", this, how_many, count );

	Assert( this != 0 );

	IntStorage intState ; PROCESSOR::disableInterrupts(intState);
	lock.acquire();
	for ( ; how_many < 0; how_many-- ) {
		count++;
		Assert( count <= maxCount );
		if( count <= 0 ) {
			Process * waiter = queue.remove();
			Assert( waiter != 0 );
			Debug.printf( "Semaphore::V(): resuming %A\n",
					waiter );
			waiter->ready();
		}
		else {
			count += how_many;
			break;
		}
	}
	lock.release();
	PROCESSOR::restoreInterrupts(intState);
}

const char *
MultiVSemaphore::className()
{
	return( "MultiVSemaphore" );
}




#include "ClassInstaller.h"
DEFINECLASS("Semaphore",Semaphore,ProxiableObject,0,0);
DEFINECLASS("BinarySemaphore",BinarySemaphore,Semaphore,0,0);
DEFINECLASS("GraciousSemaphore",GraciousSemaphore,Semaphore,0,0);
