#ifdef __GNUC__
#pragma implementation
#pragma implementation "SystemInterfaceStar.h"
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * SystemInterface.cc - class for miscellaneous system calls
 */


#include "Assert.h"
#include "Byte.h"
#include "Domain.h"
#include "ExecInfo.h"
#include "FileSystem.h"
#include "Kernel.h"
#include "NameServer.h"
#include "NameServerStar.h"
#include "OutputStream.h"
#include "PageTableConstants.h"
#include "PersistentMemoryObject.h"
#include "PersistentMemoryObjectStar.h"
#include "Pipe.h"
#include "PipeInputStream.h"
#include "PipeOutputStream.h"
#include "Process.h"
#include "SimpleInstruments.h"
#include "String.h"
#include "SystemInterface.h"
#include "ThisProcess.h"
#include "VersionObject.h"


#ifdef __GNUC__
#pragma implementation
#endif

SystemInterfaceStar TheSystemInterface;
SystemInterface * StandardSystemInterface;

extern "C" void ConstructProxy();


void *
SystemInterface::initialCall( void * returnFunct )
{
	thisProcess()->domain()->setReturnFunct( (PFV) returnFunct );

	return( (int *) &ConstructProxy );
}

void
SystemInterface::versions( ProvidedVersionInstaller * provided,
					RequiredVersionInstaller *required )
{
    VersionObject::install( provided );
    VersionObject::check( required );
}

Process *
SystemInterface::getThisProcess()
{
	return( thisProcess() );
}

void
SystemInterface::killThisProcess()
{
	thisProcess()->die();
}

Process *
SystemInterface::buildProcessFromObjectFile( PersistentMemoryObject * obj,
					     int argc, char ** argv )
{
	return( TheKernel->buildProcessFromObjectFile( obj, argc, argv ) );
}

int 
SystemInterface::execFromObjectFile( PersistentMemoryObject * objFile, 
				     int argc, char ** argv,
				     ExecInfo * execInfo )
{
	Debug << this << "::execFromObjectFile( " << objFile << ", "
		<< argc << " )\n" << eor;

	/*
	 * Test objfile to see if it is a loadable object file, if
	 * not return error;
	 */

	if ( ! TheKernel->isObjectFile( objFile ) )
		return NotAProgram;

	/*
	 * Handle execInfo
	 */

	Assert( strcmp( execInfo->signature, EXEC_SIGNATURE ) == 0 );
	int argsize = execInfo->length;

	/*
	 * Compute size of argument memory object
	 */

	int i;
	for(i = 0; i < argc; i++) {
		argsize += ( strlen( argv[i] ) + 1 );
	}
	argsize += 1023; /* for spare */

	/*
	 * Size of argv vector, allocate some for spare
	 */

	int allocArgvSize = argc + 32;
	argsize += sizeof( char * ) * allocArgvSize;

	/*
	 * Allocate memory object for arguments
	 */

	PersistentMemoryObjectStar memObj = 
		TheKernel->allocateMemoryObjectForBackingStore();
	Assert( memObj != 0 );

	int unitSize = memObj->unitSize();
	memObj->setName( "ArgumentMO" );
	memObj->setNumberOfUnits( ( argsize + unitSize - 1 ) / unitSize );
	Debug << "Argument swap mo = " << memObj << "\n" << eor;

	/* 
	 * Copy arguments 
	 */

	const char * argloc = TheKernel->domain()->add( memObj, NoAccess );
	char * p = ( char * ) argloc;

	/*
         * Save space for argv first 
	 */

	char ** newArgv = (char **) p;
	p += sizeof( char * ) * allocArgvSize;
	
	/*
	 * argv[0] is execInfo
	 * argv[1] is null
	 * argv[2...] is the actual unix main argv[0...]
	 * the actual address which the args will be mapped into is
	 * unknown yet, place relative addresses into the argv
	 */

	newArgv[0] = (char *) ( p - argloc );
	ByteCopy( (char *) execInfo, p, execInfo->length );
	p += execInfo->length;
	newArgv[1] = 0;

	for( i = 0; i < argc; i++ ) {
		newArgv[i+2] = (char *) ( p - argloc );
		strcpy( p, argv[i] );
		p += strlen( argv[i] ) + 1;
	}
	newArgv[ argc+2 ] = 0;

	thisProcess()->domain()->remove( memObj, argloc );
	Debug << "memobj removed from Kernel domain\n" << eor;

	/*
	 * Prepare the current domain for exec
	 */

	thisProcess()->domain()->prepareForExec();

	/*
	 * Add to the domain a mapped cache for each section in the file.
	 * We dont reference the ObjectFile before passing it one because
	 * were done with it here.
	 */

	PFV entryPoint = (PFV)
		TheKernel->loadObjectFile( objFile, thisProcess()->domain() );
	Debug << this << "::buildProcessFromObjectFile: entry = "
		<< asAddress( (void*)entryPoint ) << "\n" << eor;

	const char * childArgloc = 
		thisProcess()->domain()->add( memObj, ReadWrite );
	Debug << this << "::BPFCF child argslog at "
		<< asAddress( childArgloc ) << "\n" << eor;

	/*
	 * Perform relocation of arguments 
	 */

	newArgv = (char **) childArgloc;
	p = (char *) childArgloc;
	for( i = 0; i < ( argc + 2 ); i++ ) {
		newArgv[i] = (char *) ( p + (int) newArgv[i] );
	}
	newArgv[1] = 0;

	memObj = 0; /* end of method will not be executed */

	/*
	 * Create the process.
	 */

	thisProcess()->setName(objFile->objectName());
	thisProcess()->exec( entryPoint, argc + 2, newArgv );

	Assert( NOTREACHED );
	return 0;
}

int
SystemInterface::isObjectFile( PersistentMemoryObject * file )
{
	return TheKernel->isObjectFile( file );
}

OutputStreamRef
SystemInterface::outputStream( ProxiableObject * obj )
{
	OutputStreamRef s = 0;
	if ( obj->isKindOf( PipeClass ) ) {
		s = new PipeOutputStream( (Pipe *) obj );
	}
	return s;	
}

InputStreamRef 
SystemInterface::inputStream( ProxiableObject * obj )
{
	InputStreamRef s = 0;
	if ( obj->isKindOf( PipeClass ) ) {
		s = new PipeInputStream( (Pipe *) obj );
	}
	return s;
}

DiskRef
SystemInterface::rootDisk()
{
	return( TheKernel->rootDisk() );
}

int 
SystemInterface::pageSize()
{
	return PhysicalPageSize;
}

void
SystemInterface::controlSimpleInstruments( int argc, char **argv )
{
	SInstr(simpleInstrumentControl( argc, argv ));
}

#include "ClassInstaller.h"
DEFINECLASS("SystemInterface",SystemInterface,ProxiableObject,0,0);
