#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David W. Dykstra (dykstra@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * TimeoutTimer.c - machine independent portion of timeout timers
 */

#include "Assert.h"
#include "Lock.h"
#include "TIMER_MANAGER.h"
#include "OutputStream.h"
#include "Semaphore.h"
#include "TimeoutTimer.h"


#ifdef __GNUC__
#pragma implementation
#endif


TimerManager *TimeoutManager;

unsigned int TimeoutTimer::tickTime;
TimeoutTimer *TimeoutTimer::table[WHEELHASHSIZE];
#ifdef ONELock
Lock *TimeoutTimer::lock;
#else
static Lock *TimeoutTimer::lock[WHEELHASHSIZE+1];
#endif

#ifdef ONELock
inline void
TimeoutTimer::acquireLock(int)
{
	lock->acquire();
}
inline void
TimeoutTimer::releaseLock(int)
{
	lock->release();
}
#else
inline void
TimeoutTimer::acquireLock(int indx) {
	Assert(indx >= 0);
	Assert(indx < WHEELHASHSIZE);
	lock[indx+1]->acquire();
}
inline void
TimeoutTimer::releaseLock(int indx) {
	Assert(indx >= 0);
	Assert(indx < WHEELHASHSIZE);
	lock[indx+1]->release();
}
#endif

inline int
TimeoutTimer::hashIndex( unsigned int index )
{
	return( index & WHEELHASHMASK );
}

inline int
TimeoutTimer::index()
{
	return( hashIndex( tickTime ) );
}

inline int
TimeoutTimer::hashRemainder( unsigned int index )
{
	return( index >> WHEELHASHPOWER );
}

inline int
TimeoutTimer::hashValue( unsigned int index )
{
	return( index << WHEELHASHPOWER );
}

inline unsigned int
TimeoutTimer::timeResidual( TimeoutTimer *ptr )
{
	/* Locks are assumed to be already acquired	*/
	/* and timer is assumed to be running		*/
	return( TimeoutManager->ticksToMicroseconds( ptr->expireTick - tickTime ) );
}

inline TimeoutTimer **
TimeoutTimer::bucket( int indx )
{
	Assert( indx >= 0 );
	Assert( indx < WHEELHASHSIZE );
	return( &table[indx] );
}

/*
 * End of inline functions
 */

int timeoutTimerInitialized = 0;
   
void
TimeoutTimerEntry(int arg)
{
	/* This is the timeout timer process.  It calls the machine dependent */
	/*  TimeoutManager->processHandler() which calls back here to 	  */
	/*  perTickBookKeeping and expiryProcessing when appropriate.	  */

	Debug << "TimeoutTimerEntry( " << asHex( arg ) << " )\n" << eor;

	TimeoutTimer::initialize();

	TimeoutManager = new TIMER_MANAGER(&TimeoutTimer::perTickBookKeeping, &TimeoutTimer::expiryProcessing);

	Assert(TimeoutManager != 0);

	timeoutTimerInitialized = 1;
	TimeoutManager->processHandler();

	Assert(NOTREACHED);
}

/*
 * Now for the regular, per-instance functions
 */

TimeoutTimer::TimeoutTimer( )
{
	Debug << this << "::TimeoutTimer\n" << eor;

	TimeoutManager->initialize();	/* in case not started yet */

	raceSemaphore = new Semaphore(0);
	Assert(raceSemaphore != 0);
	timeoutSemaphore = new Semaphore(0);
	Assert(timeoutSemaphore != 0);
	running = 0;
	expiring = 0;
	waiting = 0;
	_tableIndex = 0;
	tickCount = 0;
	expireTick = 0;
	prev = (TimeoutTimer *) 0;
	next = (TimeoutTimer *) 0;
    
}

TimeoutTimer::~TimeoutTimer()
{
	Debug << this << "::~TimeoutTimer\n" << eor;

	(void) stop(); /* just in case in the process of expiring */
}

void
TimeoutTimer::await()
{
	Debug << this << "::await()\n" << eor;
	Assert(timeoutSemaphore != 0);
	timeoutSemaphore->P();
}

unsigned int
TimeoutTimer::start( unsigned int microseconds )
{
	unsigned ticks;
	TimeoutTimer *ptr,*prevptr;
	int setticks;
	unsigned int saveTime;

	Debug << this << "::start( " << microseconds << " )\n" << eor;

	Assert(microseconds != 0);

	if (running)
	{
		/* It's assumed that no other process will try to */
		/*  start this same timer after this point.       */
		(void) stop();
	}

	/* microsecondsToTicks() will round up to the nearest number of ticks */
	setticks = ticks = TimeoutManager->microsecondsToTicks(microseconds);

	Debug << "::start: " << ticks << " ticks\n" << eor;

	Assert(running == 0);
	/* All the following Asserts must be true when timer is stopped, but   */
	/*  but they each could also potentially be true when timer is running */
	Assert(expireTick == 0);
	Assert(prev == (TimeoutTimer *) 0);
	Assert(next == (TimeoutTimer *) 0);
	Assert(tickCount == 0);

	prevptr = (TimeoutTimer *) 0;

#ifdef ONELock
	acquireLock(0);
	saveTime = tickTime;
	expireTick = saveTime + ticks;
	_tableIndex = hashIndex(expireTick);
#else
	while(1)
	{
		/* may need to try more than once if tick increments before get	*/
		/*  chance to acquire the lock.					*/
		saveTime = tickTime;
		expireTick = saveTime + ticks;
		_tableIndex = hashIndex(expireTick);
		acquireLock(_tableIndex);
		if (tickTime - saveTime < ticks)
		{
	    ticks -= tickTime - saveTime;
	    break;
		}
		releaseLock(_tableIndex);
	}
#endif
	running = 1;

	ptr = *bucket(_tableIndex);
	ticks = hashRemainder(ticks+WHEELHASHSIZE-1);
	Assert( ticks != 0 );

	/* find appropriate place in sorted list for this bucket */
	while (ptr != (TimeoutTimer *) 0)
	{
		if (ticks < ptr->tickCount)
	    break;
		ticks -= ptr->tickCount;
		prevptr = ptr;
		ptr = ptr->next;
	}

	/* insert into the list */
	tickCount = ticks;
	if (ptr != (TimeoutTimer *) 0)
	{
		ptr->tickCount -= ticks;
		ptr->prev = this;
	}
	if (prevptr == (TimeoutTimer *) 0)
		*bucket(_tableIndex) = this;
	else
		prevptr->next = this;
	next = ptr;
	prev = prevptr;

	releaseLock(_tableIndex);

	return(TimeoutManager->ticksToMicroseconds(setticks));
}

inline void
TimeoutTimer::waitExpire()
{
	/* Save the current index -- the only way it could change is if	*/
	/*   somebody tries to start this timer while it is being stopped	*/
	/*   and that is not allowed.					*/
	int indx = _tableIndex;

	acquireLock(_tableIndex);

	while (expiring)
	{
		/* if so, indicate it and wait for completion */
		Assert( waiting == 0 );
		waiting = 1;
		releaseLock(_tableIndex);
		raceSemaphore->P();

		acquireLock(_tableIndex);
		waiting = 0;
	}

	Assert(indx == _tableIndex);

	/* _tableIndex lock is left acquired */
}

unsigned int
TimeoutTimer::stop()
{
	unsigned int resid;

	Debug << this << "::stop()\n" << eor;

	/* Wait if this timer is in the process of being expired. */
	waitExpire();
	/* the _tableIndex lock is now acquired */

	if (running)
	{
		resid = timeResidual(this);

		if (prev != (TimeoutTimer *) 0)
	    prev->next = next;
		else
	    *bucket(_tableIndex) = next;
	
		if (next != (TimeoutTimer *) 0)
		{
	    next->prev = prev;
	    next->tickCount += tickCount;
		}

		next = prev = (TimeoutTimer *) 0;
		expireTick = 0;
		tickCount = 0;
		running = 0;
		timeoutSemaphore->V();
	}
	else
		resid = 0;

	releaseLock(_tableIndex);

	return(resid);
}

unsigned int
TimeoutTimer::residual()
{
	Debug << this << "::residual()\n" << eor;

	unsigned int resid;

	/* Wait if this timer is in the process of being expired. */
	waitExpire();
	/* the _tableIndex lock is now acquired */

	if (running)
		resid = timeResidual(this);
	else
		resid = 0;

	releaseLock(_tableIndex);

	return(resid);
}

/*
 * Now for the class functions
 */

void
TimeoutTimer::initialize()
{
#ifdef ONELock
	lock = new Lock();
	Assert(lock != 0);
#else
	for (i = 0; i <= WHEELHASHSIZE; i++) {
		lock[i] = new Lock();
		Assert(lock[i] != 0);
	}
#endif
	for(int i = 0; i < WHEELHASHSIZE; i++)
	{
		table[i] = (TimeoutTimer *) 0;
	}
	tickTime = 0;

}

int
TimeoutTimer::perTickBookKeeping()
{
	/* this is called directly from the interrupt handler */

	TimeoutTimer *ptr;

	acquireLock(hashIndex(tickTime+1));
	++tickTime;

	if ((ptr = *bucket(index())) != (TimeoutTimer *) 0)
	{
		if (--ptr->tickCount == 0)
		{
	    /* there is something expiring */
	    releaseLock(index());
	    return(1);
		}
		Debug << "perTickBookKeeping: tickTime: " << tickTime
					<< ", expireTick: " << ptr->expireTick
					<< ", count: " << ptr->tickCount << "\n" << eor;
	}
	/* nothing expiring this tick */
	releaseLock(index());
	return(0);
}

void
TimeoutTimer::expiryProcessing()
{
	TimeoutTimer *ptr,**tableptr;

	/* At this point we are protected against the tickTime from		*/
	/*  changing because ticks are disallowed through the busyFlag	*/
	tableptr = bucket(index());

	acquireLock(index());

	for (ptr = *tableptr; (ptr != (TimeoutTimer *) 0) && (ptr->tickCount == 0);
			 ptr = *tableptr)
	{
		Debug << "expiryProcessing: expireTick: " << ptr->expireTick
					<< ", tickTime: " << tickTime << "\n" << eor;
		Assert(ptr->expireTick == tickTime);

		Assert(ptr->prev == (TimeoutTimer *) 0);

		if ((*tableptr = ptr->next) != (TimeoutTimer *) 0)
		{
	    /* remove this timer from the list */
	    ptr->next->prev = (TimeoutTimer *) 0;
	    ptr->next = (TimeoutTimer *) 0;
		}

		/* indicate that the timer is expiring */
		Assert( ptr->expiring == 0 );
		ptr->expiring = 1;
		ptr->expireTick = 0;

		Assert( ptr->running == 1 );
		ptr->running = 0;

		releaseLock(index());

		Assert(ptr->timeoutSemaphore != 0);

		ptr->timeoutSemaphore->V();

		/* re-aquire the lock, may need to wait for somebody else */
		acquireLock(index());

		ptr->expiring = 0;
		if (ptr->waiting)
		{
	    releaseLock(index());
	    ptr->raceSemaphore->V();
	    acquireLock(index());
		}

	}
	releaseLock(index());
}

#include "ClassInstaller.h"
DEFINECLASS("TimeoutTimer",TimeoutTimer,Timer,0,0);
