/***********************************
*EthernetDriverLKM                 *
*Ethernet driver "implementation   *
*Actual implementation is provided *
*by the loaded Linux Kernel Module *
***********************************/
//Scott Kircher
//dthomas1@ VCEthernet Rx(interrupt)/Tx code + emulation for netlink API
//CS424
//LKM Adaptor for Choices Project
//a.k.a. "TooManyChoices"
//a.k.a. "Frankenstein"
extern "C" {
#include        <sys/filio.h>
#include        <sys/dlpi.h>
#include        <sys/pfmod.h>
#include        <memory.h>
#include        <signal.h>
#include        <stdio.h>
#include        <fcntl.h>
#include        <malloc.h>
#include        <stdlib.h>
#include        <unistd.h>
#include        <sys/param.h>
#include        <sys/file.h>
#include        <sys/ioctl.h>
#include        <sys/time.h>
#include        <stropts.h>
#include        <sys/socket.h>
#include        <net/if.h>
#include        <netinet/in.h>
#include        <netinet/if_ether.h>
#include        <netdb.h>
#include        "dlwrap.h"

void    bcopy(char *, char *, int);
int     dlattachreq(int, u_long);
int     dlokack(int , char*);
int     dlbindreq(int , u_long , u_long , u_long , u_long , u_long);
int     dlbindack(int , char*);
int     dlpromisconreq(int, u_long);
int     dstrioctl(int, int, int, int, char*);
int     dlphysaddrreq(int, u_long);
int     dlphysaddrack(int, char*);

}

#ifdef __GNUC__
#pragma implementation
#endif

#include	"ThisCPU.h"
#include	"VCInterrupts.h"
#include	"ThisCPU.h"
#include	"ThisProcess.h"
#include	"OutputStream.h"
#include	"Process.h"
#include	"AwaitedInterruptInterrupt.h"
#include	"NetworkBuffer.h"
#include	"NetworkAdministrator.h"
#include	"FreeRunningTimer.h"
#include	"FreeRunningTimerStar.h"
#include	"NameServer.h"
#include  "String.h"

#include "EthernetDriverLKM.h"
#include "LinuxNetStructures.h"

using namespace LinuxNetStructures;

//network-specific adaptor functions
//these provide the expected Linux style API for the loaded module
net_device *EthernetDriverLKM::_device_info=0; //currently we allow just one, but this normally would be a linked list

/* Bleach, we need a global variable to keep track of the driver */
static EthernetDriverLKM	*ethernet;

int EthernetDriverLKM::adaptor_register_netdev(net_device *device)
{
	_device_info=device;
	Console()<<"Net device registered\n";
	return 0;
}

void EthernetDriverLKM::adaptor_unregister_netdev(net_device *device)
{
	_device_info=0;
}

void EthernetDriverLKM::adaptor_ether_setup(net_device *device)
{
	//normally this would set up some internal junk in the net_device structure,
	//but we're not using most of it, so just a minimal amount is done here

	//which currently is absolutely nothing
}

int EthernetDriverLKM::adaptor_dev_get(const char *name)
{
	if(_device_info)
	{
		if(!strcmp(_device_info->name,name)) return 1;
	}
	return 0;
}

//Constructor
//Have to load the LKM, and initialize the ethernet
EthernetDriverLKM::EthernetDriverLKM(const char *LKMmodulename,unsigned char debugMask=0,NetworkBufferFreeList* nf =0)
    : EthernetDriver(debugMask,nf), LKMAdaptor()
{
	_ready=false;

	//register all the symbols we locally export
	//(these are LKMAdaptor base class calls)
	addExportedSymbol("register_netdev*",(void*)adaptor_register_netdev);
	addExportedSymbol("unregister_netdev*",(void*)adaptor_unregister_netdev);
	addExportedSymbol("ether_setup*",(void*)adaptor_ether_setup);
	addExportedSymbol("dev_get*",(void*)adaptor_dev_get);
	
	//now we load the module
	int status=loadELF(LKMmodulename);
	Assert(status==0);

	//Now that we have loaded our "implementation" (the LKM)
	//we can begin using it to set-up the ethernet device
	
	//first we call it's init_module function
	initModule();

	Console()<<"Loaded module initialized.\n";
	//if the device is valid, it should have registered its init function with us
	//using the register_netdev function we provided
	//so, try to use it
	if(_device_info)
	{
		Console()<<"Checking for ethernet init function\n";
		if(_device_info->init)
		{
			Console()<<"Calling ethernet init function\n";
			if(_device_info->init(_device_info)!=0)
			{
				Console()<<"Ethernet initialization provided by "<<LKMmodulename<<" FAILED!\n"<<eor;
				return;
			}
		}else
		{
			Console()<<"No ethernet initialization function was registered by "<<LKMmodulename<<"\n"<<eor;
			return;
		}
	}else
	{
		Console()<<"No network device was registered by "<<LKMmodulename<<"\n"<<eor;
		return;
	}

	//at this point, the ethernet device should be ready for use, and our _device_info
	//structure should have been filled with all the nice functions we need to use
	//Verify that this is so:
	if(!_device_info->open)
	{
		Console()<<"No device open() function was registered by "<<LKMmodulename<<"\n"<<eor;
		return;
	}

	if(!_device_info->hard_start_xmit)
	{
		Console()<<"No device hard_start_xmit() function was registered by "<<LKMmodulename<<"\n"<<eor;
		return;
	}

	if(!_device_info->stop)
	{
		Console()<<"No device stop() function was registered by "<<LKMmodulename<<"\n"<<eor;
		return;
	}

	//OK! So, now we're ready to rock'n'roll We just need to open the device, and GO
	if(_device_info->open(_device_info)!=0)
	{
		Console()<<"Ethernet open() provided by "<<LKMmodulename<<" FAILED!\n"<<eor;
		return;
	}

	_ready=true;
}

//destructor
//close down the device
EthernetDriverLKM::~EthernetDriverLKM()
{
	if(_device_info)
	{
		if(_device_info->stop)
		{
			_device_info->stop(_device_info);
		}
	}
}

//print info
void EthernetDriverLKM::printInfo()
{
	Console()<<"Hi there. This is the EthernetDriverLKM speaking\n";
	Console()<<"Normally, I could print some device stats here...\n";
	Console()<<"but we don't have time to deal with Linux's device\n";
	Console()<<"stats structure, so we'll ignore it.\n";
	Console()<<"All I can tell you right now is:\n";
	if(_ready) Console()<<this<<" is ready to rock and roll.\n";
	else Console()<<this<<" is not ready.\n";	
}


void
EthernetDriverLKM::interruptHandler()
{
  _interrupts++;

  for ( ; ; ) {
    NetworkBuffer *buf = NetworkDriver::getBuffer();

    Assert(buf->packet() != 0);
    Assert(buf->maxLength() > 60);

	/* Read the packet from the interface */
     _data.buf = (char *) buf->packet();
     _data.maxlen = buf->maxLength();
     _data.len  = 0;
     getmsg(_fd, NULL , &_data, &_flags);	
     
     int size = _data.len;
     //fprintf(stderr, "ETH driv interrupt pkt size: %d\n", size);
     EthernetAddress	to(buf->packet());
     EthernetAddress	from(buf->packet() + EthernetAddressLength);

    	/* If nothing was read, start reading from the pseudo-interface. */
    if (size <= 0 || (to.isBroadcast() && from == *_ethernetAddress))
    {
      int	bufferSize;

      ioctl(_serverSocket, FIONREAD, &size);
      if (size <= 0)
      {
				buf->release();
				return;
      }

      	/* First we read the size of the next buffer */
      read(_serverSocket, &bufferSize, sizeof(int));

      	/* Then we actually read the buffer */
      bufferSize = ntohs(bufferSize);
      if (bufferSize > buf->maxLength())
				bufferSize = buf->maxLength();

      size = read(_serverSocket, buf->packet(), bufferSize);
      Assert(size == bufferSize);
    }

	/* Filter out multicast packets we do not want */
    //EthernetAddress	to(buf->packet());
    if (!isValidAddress(&to))
      buf->release();
    else
    {
		  // We call the LKM Rx interrupt handler here
			_skbRx = convertBuffer(buf);
			struct sock	*sk  = new struct sock;
			sock -> protocol = 0x01;
			ethertap_rx(sk,0); // DO IT
			delete sk;

 
      buf->setTime(SystemTimer->time());
      buf->setLength(size);
			
			// Replace this function with skb specific function
			memcpy(buf->packet(), _skbRx->data, _skbRx -> len);
      NetworkDriver::receivePacket(buf);

			delete _skbRx;
    }
  }
}


void
VCEthernetInterruptProcessEntry()
{
  int		fd, server;
  EthernetDriverLKM	*ethernet;

  ethernet = (EthernetDriverLKM *) (ProxiableObject *)StandardNameServer->lookup("ethernetDriver", "EthernetDriverLKM");

	/* Get the file descriptors associated with the interfaces */
  ethernet->getDescriptors(&fd, &server);

  AwaitedInterruptInterrupt *interrupt = new AwaitedInterruptInterrupt();

	/* Add this descriptor to list that is checked for pending
	 * input when a SIGIO signal is received.
	 */
  VCIOInterrupt::setIOInterrupt(fd, interrupt);

  	/* Add the descriptor for the pseudo-interface. */
  VCIOInterrupt::setIOInterrupt(server, interrupt);

  for ( ; ; ) {
    interrupt->await();
//    fprintf(stderr, "An ethernet terrupt occured\n");
    ethernet->interruptHandler();
  }
}


void
startLocalServer(void)
{
  int			pid;
  int			ret;

  if ((pid = vfork()) < 0){
	perror("vfork");
	Assert (NOTREACHED) ("VCVirtualEthernetDaemon vfork failed!");
	}

  if (pid == 0)  {
	ret = execlp("VCVirtualEthernetDaemon", "VCVirtualEthernetDaemon",NULL);
	Assert (NOTREACHED) ("VCVirtualEthernetDaemon execve failed!");
	}

}

bool
EthernetDriverLKM::connectToLocalServer(void)
{
  struct sockaddr_in	address;
  struct hostent	*host;
  char			hostname[MAXHOSTNAMELEN];

  	/* Create a socket to connect to the server */
  _serverSocket = socket(AF_INET, SOCK_STREAM, 0);

  	/* Ick.  To connect to ourself, we have to look up
	   our address in the host table */
  gethostname(hostname, sizeof(hostname));
  host = gethostbyname(hostname);
  bcopy(host->h_addr_list[0], (char *) &address.sin_addr.s_addr, 
	sizeof(struct in_addr));
  address.sin_family = AF_INET;
  address.sin_port = htons(networkPort);

  	/* Try to connect */
  if ((connect(_serverSocket, (sockaddr *) &address, sizeof(address)) < 0))
  {

   fprintf(stderr, "We failed in the first attempt to connect\n");
    	/* If we don\'t succeed, start the server */
    //startLocalServer();

    int connected = 0;
    while (connected == 0) {
    	/* You can not connect twice to a socket, even if
	   the first attempt failed.  So create a new socket */
	    close(_serverSocket);
	    _serverSocket = socket(AF_INET, SOCK_STREAM, 0);
	    bcopy(host->h_addr_list[0], (char *) &address.sin_addr.s_addr, 
		sizeof(struct in_addr));
	    address.sin_family = AF_INET;
	    address.sin_port = htons(networkPort);

    	/* If we can't connect this time, something is wrong */
	    connected = (connect(_serverSocket, (sockaddr *) &address, sizeof(address)) == 0) ;
		if(!connected)
		{
		    fprintf(stderr, "That was our second attempt at a connection, which also failed\n");
			return false;
		}
	}
  }

  	/* Set socket so we receive SIGIO signals */
  fcntl(_serverSocket, F_SETFL, FASYNC);
  fcntl(_serverSocket, F_SETOWN, getpid());
  fprintf(stderr, "Connected to the local daemon\n");
  return true;
}

//transmission functions
//Note: reception is done via an interrupt handler which is provided by the loaded LKM
//and registered with the (adapted) netlink API
int EthernetDriverLKM::transmitPacket(NetworkBuffer *buf)
{
  int		retVal;

  retVal = doTransmit(buf);
  thisProcess()->relinquishProcessor();	// simulate a synchronous send

  return(retVal);
}

int EthernetDriverLKM::transmitPacketAsync(NetworkBuffer *buf)
{
  int		retVal;

  retVal = doTransmit(buf);
  buf->release();

  return(retVal);
}

int EthernetDriverLKM::doTransmit(NetworkBuffer *buf)
{
  
	EthernetAddress	dest(buf->packet());
  octet			*packet;
  int			packetSize;
	// We create 2 structures 
	// 1. net_device -- I stole this from loopback.c of linux
	// 2. sk_buff     -- We create it out of a NetworkBuffer
	struct sk_buff* skb      = convertBuffer(buf);
	struct net_device * dev = initnetDevice();

  packet = buf->packet();

  // Here we make a call to the loaded LKM, to do the actual transmission
	// Scott since I do not know the /name/ of the linux Tx routine
	// I used the name from ethertap.c

  /* If we're sending to someone not on this machine, use Ethernet */
  if (dest != *_ethernetAddress)
  {
		ethertap_start_xmit(skb, dev); /* We call the Linux xmit to actually send the pkt */
 	}

		/* I decided to send the rest by the daemon.
			 Maybe we won't need this anymore
		 */

  	/* If it is the same machine or a multicast address, use the 
	   pseudo-interface.  Note that multicast packets are sent
	   on both interfaces */
  if (dest == *_ethernetAddress || dest.isMulticast() || dest.isBroadcast())
  {
    	/* Now transmit the packet on the pseudo-interface */
    packetSize = htons(buf->length());
    write(_serverSocket, &packetSize, sizeof(int));
    write(_serverSocket, packet, buf->length());
  }

  Console() << "Ethernet transmit \n" ;

	//Cleanup memory before we leave
	delete skb;
	delete dev;
  
  return(0);
}

struct net_device* EthernetDriverLKM::initnetDevice()
{
	struct net_device* dev = new (struct net_device);

 	dev->mtu    = (16 * 1024) + 20 + 20 + 12;

#if 0 
  dev->hard_start_xmit  = loopback_xmit;
  dev->hard_header  = eth_header;
  dev->hard_header_cache  = eth_header_cache;
  dev->header_cache_update= eth_header_cache_update;
#endif 

  dev->hard_header_len  = EthernetAddressLength;   /* 14     */
  dev->addr_len   = 6;   /* 6      */
  dev->tx_queue_len = 0;
  dev->type   = 0x0001; /*ARPHRD_LOOPBACK; */ /* 0x0001   */
//  dev->rebuild_header = eth_rebuild_header;
  //dev->flags    = IFF_LOOPBACK;
  //dev->features   = NETIF_F_SG|NETIF_F_FRAGLIST|NETIF_F_NO_CSUM|NETIF_F_HIGHDMA;
  dev->priv = kmalloc(sizeof(struct net_device_stats), GFP_KERNEL);
  if (dev->priv == NULL)
      return NULL;
  memset(dev->priv, 0, sizeof(struct net_device_stats));
//  dev->get_stats = get_stats;

} 

/* 
	 Converts a Choices NetworkBuffer to Linux sk_buff
	 This is a good place to do optimizations
	 since this code is called for each packet Tx and Rx
 */

struct sk_buff* EthernetDriverLKM::convertBuffer(NetworkworkBuffer *buf)
{

	// Allocate memory for it first
	struct sk_buff* skb = dev_alloc_skb(buf -> length() + 2);

	if (!skb) {
		fprintf(stderr, "Unable to alloc memory to Tx pkt\n");
		return (NULL);
	} 

	// Now fill in the data fields
	memcpy(skb_put(skb, buf -> length()) , buf -> packet(), buf -> length());
 
	return (skb);
}

/* Data Link Packet Interface (DLPI) initialization routine */
int EthernetDriverLKM::initDLPI()
{
  struct strioctl       strioctl;
  struct ifreq          ifreq;
  struct packetfilt     filter;
  struct sockaddr       *sockaddr;
  u_short               *fwp;
  char                  hostname[64];
  struct hostent        *hostaddr;
  short                 *localAddr;
  long                  buf[MAXDLBUF];
  union   DL_primitives   *dlp;

  _interrupts = 0;
  _inputPackets = 0;
  _outputPackets = 0;

  _ready=false;

        /* Start up a local process to act as a pseudo-broadcast
           network.  This is necessary since the NIT interface does
           not send packets addressed to the local machine */
  if(!connectToLocalServer())
  {
          return;
  }

       /* Open the DLPI enabled Fast Ethernet driver directly */
  if ((_fd = open("/dev/eri", O_RDWR)) < 0)
    perror("/dev/eri");

        /*
         * Attach.
         */
   if (dlattachreq(_fd, 0) < 0)
      perror("dlattach:");

   if (dlokack(_fd, (char *)buf) < 0)
		perror("dlokack:");
	

	/*
	 * Bind for receiving only IP packets.
	 */
   if (dlbindreq(_fd, 0x0800, 0, DL_CLDLS, 0, 0) < 0)
		perror("dlbindreq:");
   if (dlbindack(_fd, (char*)buf) < 0)
		perror("dlbindack:"); 

	/* Enable promiscuous mode @DCT verify whether it's /really/ needed? */
    if (dlpromisconreq(_fd, DL_PROMISC_SAP) < 0)
		perror("dlpromisconreq:");
    if (dlokack(_fd, (char *)buf) < 0)
		perror("dlokack");

	/*
	 * Issue DLIOCRAW ioctl.
	 */
    if (dstrioctl(_fd, DLIOCRAW, -1, 0, NULL) < 0)
			perror("DLIOCRAW:");

	/* Set mode to message no-discard mode */
  	if (ioctl(_fd, I_SRDOPT, (char *) RMSGN) < 0)
    	perror("ioctl: discard");


  gethostname(hostname, sizeof(hostname));
  hostaddr = gethostbyname(hostname);
  localAddr = (short *) hostaddr->h_addr_list[0];

	/* Set up the packet filter.  We discard all IP packets
	 * that are destined for the UNIX machine we're running on.
	 */
  fwp = filter.Pf_Filter;

  *fwp++ = ENF_PUSHWORD + 6;		/* Push packet type */
  *fwp++ = ENF_PUSHLIT | ENF_CNAND;	/* Return true if not IP */
  *fwp++ = ETHERTYPE_IP;

  *fwp++ = ENF_PUSHWORD + 16;		/* Push second word of IP address */
  *fwp++ = ENF_PUSHLIT | ENF_CNOR;	/* Return false if for local host */
  *fwp++ = localAddr[1];

  filter.Pf_FilterLen = 6;

  strioctl.ic_cmd = PFIOCSETF;
  strioctl.ic_timout = 3;
  strioctl.ic_len = sizeof(filter);
  strioctl.ic_dp = (char *) &filter;

  if (ioctl(_fd, I_PUSH, "pfmod") < 0) 
    perror("push protocol filter");

  if (ioctl(_fd, I_STR, (char *) &strioctl) < 0)
    perror("set protocol filter");

	/* Get the ethernet address of the interface */

  /*
   * Get our physical address.
   */   
  dlphysaddrreq(_fd, DL_CURR_PHYS_ADDR);
  dlphysaddrack(_fd, (char *)buf);
  dlp = (union DL_primitives*) buf;
  _ethernetAddress = new EthernetAddress((octet *)
	OFFADDR(dlp, dlp->physaddr_ack.dl_addr_offset));

  	/* We really shouldn\'t do this */
  ethernet = this;

  StandardNameServer->bind("ethernetDriver", this);

  Process * p = new InterruptProcess ( VCEthernetInterruptProcessEntry,
				      thisProcess()->domain(),
				      normal, 0);
  Assert (p != 0);
  p->setName("#VCEthernetInterruptProcess");
  p->ready();
  //thisProcess()->relinquishProcessor(); // get it running

	/* Flush the stream, ask to receive SIGIO signals,
	 * and set descriptor to non-blocking mode.
	 */
  ioctl(_fd, I_FLUSH, (char *) FLUSHR);
  fcntl(_fd, F_SETFL, O_NDELAY);
  ioctl(_fd, I_SETSIG, (char *) S_INPUT);
  fprintf(stderr, "Finished constructor of VCEthernet\n");
  _ready=true;

  return (_fd); /* This is done purely for esoteric purposes */

}

/* The netlink API emulation routines */
struct sock * 
EthernetDriverLKM::netlink_kernel_create(int unit, void(*input)(struct sock *sk, int len))
{

	struct sock* sk = new (struct sock);

	//Call the DLPI init function
	sk->fd = initDLPI(); 
	// Store the function pointer to the member variable of the device class
	_lkmRecvFunc = input;
	return (sk);

}

int 
EthernetDriverLKM::netlink_broadcast(struct sock *ssk, struct sk_buff *skb, __u32 pid, int nonblock)
{
    	/* Transmit the packet on the interface */
    if (write(_fd, skb -> data,  skb -> len) < 0)
			perror ("Ethernet write:") ;

}

/* sk_buff emulation code */

struct sk_buff* EthernetDriverLKM::skb_dequeue(struct sk_buff_head *list)
{
	
	/* At this point we have a valid sk_buff in the 
	
	return (_skbRx);
}


void EthernetDriverLKM::netif_rx(struct sk_buff* skb)
{
	_skbRx = skb;
} 

#include "ClassInstaller.h"
DEFINECLASS("EthernetDriverLKM",EthernetDriverLKM,EthernetDriver,0,0);
