/*****************************
*LKMAdaptor                  *
*Uses ELFLoader to load an   *
*ELF binary version of a     *
*Linux Kernel Module, and    *
*provides the (static) syms  *
*that the LKM expects to     *
*link with(i.e. kernel calls)*
*Also provides the "member"  *
*interface for Choices.      *
*There should probably be one*
*Of these classes for each   *
*kind of driver base class,  *
*that provides the interface *
*that Choices expects for the*
*particular driver.          *
******************************
*Linux adaptor for Choices   *
*Project, CS423, Spring 2004 *
*****************************/
//SCOTT KIRCHER

#ifdef __GNUC__
#pragma implementation
#endif

#include "Raid.h"
#include "LKMAdaptor.h"
#include "String.h"
#include "FileSystem.h"
#include "OutputStream.h"

/********************ADAPTOR FUNCTIONS************/
//adaptor for printk
int LKMAdaptor::adaptor_printk(char *s, unsigned x1, unsigned x2, unsigned x3, unsigned x4)
{
	Console().formattedWrite(s, x1, x2, x3, x4);
	Console().flush();
	return 0;
}

//adaptor for memcpy
void *LKMAdaptor::adaptor_memcpy(void *s1, const void *s2, int n)
{	
  char *dest=(char*)s1;
  char *source=(char*)s2;
  Console() << "memcpy called: copying from " 
    << asAddress((void *)source) << "\n" << eor;
  //not the most efficient, but oh well
  for(int c=0;c<n;c++)
  {
    *dest++ = *source++;
  }
  Console()<<"memcpy done\n"<<eor;
  return s1;
}

//adaptor for bzero
void LKMAdaptor::adaptor_bzero(char *s,int n)
{
	adaptor_memset((void*)s,0,n);
}

//adaptor for kmalloc (ignores flags, so may not work for every kind of kmalloc call)
void *LKMAdaptor::adaptor_kmalloc(int size, int flags)
{
	return (void*)(new char[size]);
}

//adaptor for kfree
void LKMAdaptor::adaptor_kfree (const void *objp)
{
	delete [] ((char*)objp);
}

//adaptor for memset
void *LKMAdaptor::adaptor_memset(void *s, int c, int n)
{
	//not the most efficient, but oh well
	char *dest=(char*)s;
	for(int i=0;i<n;i++)
	{
		dest[i]=(unsigned char)c;
	}
	return s;
}

//helper class for sprintf adaptor
class StringOutputStream : public OutputStream
{
public:
	StringOutputStream(char *where) : OutputStream()
	{
		string_buffer=where;
		string_location=0;
	}
	proxiable virtual int write( const char * buffer, int count )
	{
		for(int c=0;c<count;c++)
		{
			string_buffer[string_location++]=buffer[c];
		}
		return count;		
	}
	void finish()
	{
		string_buffer[string_location]=0;
	}
private:
	char *string_buffer;
	int string_location;
};

//adaptor for sprintf
int LKMAdaptor::adaptor_sprintf(char * buffer, const char *format, unsigned x1, unsigned x2, unsigned x3, unsigned x4)
{
	StringOutputStream *outputter=new StringOutputStream(buffer);
	
	int rv=outputter->formattedWrite(format,x1,x2,x3,x4);
	outputter->finish();
	
	delete outputter;
	
	return rv;
}

//adaptor for set_bit
void LKMAdaptor::adaptor_set_bit (int nr, volatile void * addr)
{
	((unsigned char*)addr)[nr/8] |= 1<<(nr%8);
}

//adaptor for clear_bit
void LKMAdaptor::adaptor_clear_bit (int nr, volatile void * addr)
{
	Console()<<"Clearing bit "<<nr<<" of "<<asAddress((void *)addr)<<"\n";
	((unsigned char*)addr)[nr/8] &= ~(1<<(nr%8));
}


/********************END ADAPTOR FUNCTIONS************/
//List of adapted functions:
const int LKMAdaptor::number_of_exported_symbols=9;
ELFLoader::ExportTuple LKMAdaptor::exported_symbols[number_of_exported_symbols] =
{  
   {"printk*",(void*)adaptor_printk},
   {"*memcpy",(void*)adaptor_memcpy},
   {"*bzero",(void*)adaptor_bzero},
   {"kmalloc*",(void*)adaptor_kmalloc},
   {"kfree*",(void*)adaptor_kfree},
   {"*memset*",(void*)adaptor_memset},
   {"sprintf*",(void*)adaptor_sprintf},
   {"*set_bit",(void*)adaptor_set_bit},
   {"*clear_bit",(void*)adaptor_clear_bit}
};

/*****************LKMAdaptor IMPLEMENTATION***********/
//constructor
LKMAdaptor::LKMAdaptor()
{
	number_of_required_symbols=0;
	required_symbols_storage_size=0;
	required_symbols=0;
	number_of_special_exported_symbols=0;
	special_exported_symbols_storage_size=0;
	special_exported_symbols=0;
	//Every LKMAdaptor requires init_module and cleanup_module
	addRequiredSymbol("*init_module*",(void**)&init_module);
	addRequiredSymbol("*cleanup_module*",(void**)&cleanup_module);

	//add all the global exported symbols
	for(int c=0;c<number_of_exported_symbols;c++)
	{
		addExportedSymbol(exported_symbols[c].name,exported_symbols[c].address);
	}
}

//load constructor
LKMAdaptor::LKMAdaptor(const char *moduleName)
{
	number_of_required_symbols=0;
	required_symbols_storage_size=0;
	required_symbols=0;
	number_of_special_exported_symbols=0;
	special_exported_symbols_storage_size=0;
	special_exported_symbols=0;
	//Every LKMAdaptor requires init_module and cleanup_module
	addRequiredSymbol("*init_module*",(void**)&init_module);
	addRequiredSymbol("*cleanup_module*",(void**)&cleanup_module);

	//add all the global exported symbols
	for(int c=0;c<number_of_exported_symbols;c++)
	{
		addExportedSymbol(exported_symbols[c].name,exported_symbols[c].address);
	}

	if(loadELF(moduleName)==Success)
	{
		initModule();
	}
}

//destructor
LKMAdaptor::~LKMAdaptor()
{
	if(cleanup_module) cleanupModule();
	
	if(required_symbols)
	{
		for(int c=0;c<number_of_required_symbols;c++)
		{
			if(required_symbols[c].name) delete [] required_symbols[c].name;
		}
		delete [] required_symbols;
	}
	required_symbols=0;

	if(special_exported_symbols)
	{
		for(int c=0;c<number_of_special_exported_symbols;c++)
		{
			if(special_exported_symbols[c].name) delete [] special_exported_symbols[c].name;
		}
		delete [] special_exported_symbols;
	}
	special_exported_symbols=0;

}

//load function
int LKMAdaptor::loadELF(const char *moduleName)
{
	//use an ELFLoader to load and link the module
	ELFLoader *loader=new ELFLoader(moduleName,number_of_special_exported_symbols,special_exported_symbols,number_of_required_symbols,required_symbols);
	int rv=loader->getStatus();
	delete loader;
	return rv;
}

//add required symbol function
//child classes can use this to register new symbols that must be defined in the loaded LKM, in order for the load to succeed
void LKMAdaptor::addRequiredSymbol(char *name, void **who_wants_it)
{
	//first check to see if we need to expand the size of the required symbol table
	if(required_symbols_storage_size<=number_of_required_symbols)
	{
		//allocate a new symbol table
		if(required_symbols_storage_size<8) required_symbols_storage_size=8;
		else required_symbols_storage_size*=2;
		ELFLoader::ImportTuple *newtable=new ELFLoader::ImportTuple[required_symbols_storage_size];
		//copy over the entries from the old table (pointer copy is OK)
		for(int c=0;c<number_of_required_symbols;c++)
		{
			newtable[c]=required_symbols[c];
		}
		//delete the old table
		if(required_symbols) delete [] required_symbols;
		required_symbols=newtable; //update pointer
	}
	//initialize the address to zero (as a courtesy)
	*who_wants_it=0;
	//add a new required symbol entry
	required_symbols[number_of_required_symbols].name=new char[strlen(name)+1];
	strcpy(required_symbols[number_of_required_symbols].name,name);
	required_symbols[number_of_required_symbols].address=who_wants_it;
	number_of_required_symbols++;
}

//add special exported symbol functions
//exported symbols added in this way are only accessible to the class who
//added them (as opposed to the global exported symbol table defined at the top
//of this file)
//child classes can use this to register new symbols that may be undefined in the loaded LKM
void LKMAdaptor::addExportedSymbol(char *name, void *where_is_it)
{
	//first check to see if we need to expand the size of the required symbol table
	if(special_exported_symbols_storage_size<=number_of_special_exported_symbols)
	{
		//allocate a new symbol table
		if(special_exported_symbols_storage_size<8) special_exported_symbols_storage_size=8;
		else special_exported_symbols_storage_size*=2;
		ELFLoader::ExportTuple *newtable=new ELFLoader::ExportTuple[special_exported_symbols_storage_size];
		//copy over the entries from the old table (pointer copy is OK)
		for(int c=0;c<number_of_special_exported_symbols;c++)
		{
			newtable[c]=special_exported_symbols[c];
		}
		//delete the old table
		if(special_exported_symbols) delete [] special_exported_symbols;
		special_exported_symbols=newtable; //update pointer
	}
	//add a new exported symbol entry
	special_exported_symbols[number_of_special_exported_symbols].name=new char[strlen(name)+1];
	strcpy(special_exported_symbols[number_of_special_exported_symbols].name,name);
	special_exported_symbols[number_of_special_exported_symbols].address=where_is_it;
	number_of_special_exported_symbols++;
}

void LKMAdaptor::initModule()
{
	Console()<<"LKMAdaptor::initModule() called\n";
	Assert(init_module!=0);
	init_module();
}

void LKMAdaptor::cleanupModule()
{
	Console()<<"LKMAdaptor::cleanupModule() called\n";
	Assert(cleanup_module!=0);
	cleanup_module();
}
