#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
#include "GPInclude.h"
#include "BitRanges.h"

#ifdef __GNUC__
#pragma implementation
#endif


int
BitRanges::borderSearch(const unsigned pos, int dir)
{
    Assert( pos < _range.bits );
    
    unsigned stype = _screen.get(pos);
    
    // Test pattern opposite of bit to search
    unsigned rtest = FFFF;
    
    // Test for a screen of opposite type so pattern is opposite of
    // opposite i.e the same type as screen.
    unsigned stest = stype ? FFFF : ZERO ;
    
    int add, lim, ipos;
    
    if (dir) { 	// increasing bit index
	add = 1;
	lim = _range.ints;
	ipos = 0;
    } else { 	// decreasing bit index
	add = -1;
	lim = -1;
	ipos = BitsInInt-1;
    }
    
    int i = BitArray::inInt(pos);
    
    unsigned theInt = _range.array[i] &
	(stype ? _screen.array[i] : ~_screen.array[i]);
    
    int ret = IntBitArray(&theInt).
	bitSearch(BitArray::offsetInInt(pos), 0, dir);
    
    if ( ret == -1 )
    {
	while ( (i+=add) != lim && _range.array[i] == rtest &&
	       _screen.array[i] == stest )
	    ;
	if (i != lim) {
	    theInt = _range.array[i] &
		(stype ? _screen.array[i] : ~_screen.array[i]);
	    ret = IntBitArray(&theInt).bitSearch(ipos, 0, dir);
	}
    } 
    
    if (ret >= 0) {
	ret = BitArray::bitpos(i,ret);
	if ((unsigned)ret > _range.bits-1)
	    ret = _range.bits-1;
	else
	    ret = dir ? ret-1 : ret+1;
    } else {
	ret = dir ? _range.bits-1 : 0;
    }
    
    Assert( (-1 < ret) && ((unsigned)ret < _range.bits) );
    return ret;
}

int
BitRanges::inRange( const unsigned pos, unsigned& from, unsigned& to )
{
    Assert( pos < _range.bits );

    if (_range.get(pos)) {
	from = borderSearch(pos,0);
	to = borderSearch(pos,1);
	return 1;
    }
    else
	return 0;
}

/*
 * alloc - allocated bit range
 * _screen is 0 if any adjacent range has _screen 1
 * else _screen is 1
 * Return 0 if range cannot be allocated
 */
int
BitRanges::alloc(const unsigned pos)
{
    Assert( pos < _range.bits );

    // verify that range is unallocated
    if (_range.get(pos)) return 0;

    _range.set(pos);
    
    // if neither screens of adjacent ranges are set, set screen else
    // leave it alone.
    if (!(pos ? _screen.get(pos-1) : 0) &&
	!((pos < _screen.bits-1) ? _screen.get(pos+1) : 0))
	_screen.set(pos);
    
    return 1;
}

/*
 * alloc - allocated bit range
 * _screen is 0 if any adjacent range has _screen 1
 * else _screen is 1
 * Return 0 if range cannot be allocated
 */
int
BitRanges::alloc(const unsigned from, const unsigned to)
{
    Assert( to < _range.bits && to >= from );

    // verify that range is unallocated
    if (_range.get(from,to)) return 0;

    _range.set(from,to);
    
    // if neither screens of adjacent ranges are set, set screen else
    // leave it alone.
    if (!(from ? _screen.get(from-1) : 0) &&
	!((to < _screen.bits-1) ? _screen.get(to+1) : 0))
	_screen.set(from,to);
    
    return 1;
}

/*
 * free - free bit range
 * Zero out _range and _screen(if reqd.).
 * If an adjacent range has 0 _screen, it may be necessary to set it
 * Freeing across ranges is not allowed. This may prevent bugs of
 * locking, for which partitions are intended.
 */
int
BitRanges::free(const unsigned pos)
{
    Assert( pos < _range.bits );

    // verify that range is allocated
    if (_range.get(pos) != 1) return 0;

    _range.clear(pos);
    _screen.clear(pos);
    
    unsigned afrom, ato;
    
    // Set screens for adjacent ranges if they are zero.
    
    if (pos && _range.get(pos-1) && !_screen.get(pos-1)) {
	afrom = borderSearch(pos-1,0);
	if (!(afrom ? _screen.get(afrom-1) : 0))
	    _screen.set(afrom,pos-1);
    }

    if (pos < _range.bits-1 && _range.get(pos+1) && !_screen.get(pos+1)) {
	ato = borderSearch(pos+1,1);
	if (!((ato < _screen.bits-1) ? _screen.get(ato+1) : 0))
	    _screen.set(pos+1,ato);
    }

    return 1;
}


/*
 * free - free bit range
 * Zero out _range and _screen(if reqd.).
 * If an adjacent range has 0 _screen, it may be necessary to set it
 * Freeing across ranges is not allowed. This may prevent bugs of
 * locking, for which partitions are intended.
 */
int
BitRanges::free(const unsigned from, const unsigned to)
{
    Assert( to < _range.bits && to >= from );

    // verify that range is allocated
    if (_range.get(from,to) != 1) return 0;

    _range.clear(from,to);
    _screen.clear(from,to);
    
    unsigned afrom, ato;
    
    // Set screens for adjacent ranges if they are zero.
    
    if (from && _range.get(from-1) && !_screen.get(from-1)) {
	afrom = borderSearch(from-1,0);
	if (!(afrom ? _screen.get(afrom-1) : 0))
	    _screen.set(afrom,from-1);
    }

    if (to < _range.bits-1 && _range.get(to+1) && !_screen.get(to+1)) {
	ato = borderSearch(to+1,1);
	if (!((ato < _screen.bits-1) ? _screen.get(ato+1) : 0))
	    _screen.set(to+1,ato);
    }

    return 1;
}

void
BitRanges::inspect()
{
    Console() << "Range\n" << eor;
    _range.inspect();
    Console() << "\nScreen\n" << eor;
    _screen.inspect();
}
