#ifdef __GNUC__
#pragma implementation
#endif

/*
 * Portable combined random number generator (32 and 16 bit versions).
 *
 * Ported by: Gary Johnston.
 * From: L'Ecuyer, Pierre.  "Efficient and Portable Combined Random Number
 *		Generators" Communications of the ACM, vol. 31, no. 6,
 *		June 1988, pp. 742-749 and 774.
 *
 * The 32-bit generator is from figure 3, page 747.
 * The 16-bit generator is from figure 4, page 748.
 */

#include "Random.h"


#ifdef __GNUC__
#pragma implementation
#endif

#ifdef RANDOM32

/*
 * 32-bit version (assumes 32-bit integers).
 * With these numbers the period is about 2.30584E18.
 */

static int s1 = 0;
static const int m1 = (int)2147483563;
static const int a1 = 40014;
static const int q1 = 53668;
static const int r1 = 12211;

static int s2 = 0;
static const int m2 = (int)2147483399;
static const int a2 = 40692;
static const int q2 = 52774;
static const int r2 = 3791;

void
Seed( int a, int b )
{
	s1 = a;
	s2 = b;
}

double
Random()
{
	int k = s1 / q1;
	s1 = a1 * (s1 - k * q1) - k * r1;
	if ( s1 < 0 )
		s1 += m1;

	k = s2 / q2;
	s2 = a2 * (s2 - k * q2) - k * r2;
	if ( s2 < 0 )
		s2 += m2;

	int Z = s1 - s2;
	if ( Z < 1 )
		Z += m1 - 1;

	return (double) Z / (double) m1;
}

#else

/*
 * 16-bit version (assumes 16-bit integers).
 * With these numbers the period is about 8.12544E12.
 */

static int s1 = 0;
static const int m1 = 32363;
static const int a1 = 157;
static const int q1 = 206;
static const int r1 = 21;

static int s2 = 0;
static const int m2 = 31727;
static const int a2 = 146;
static const int q2 = 217;
static const int r2 = 45;

static int s3 = 0;
static const int m3 = 31657;
static const int a3 = 142;
static const int q3 = 222;
static const int r3 = 133;

void
Seed( int a, int b, int c )
{
	s1 = a;
	s2 = b;
	s3 = c;
}

double
Random()
{
	int k = s1 / q1;
	s1 = a1 * (s1 - k * q1) - k * r1;
	if ( s1 < 0 )
		s1 += m1;

	k = s2 / q2;
	s2 = a2 * (s2 - k * q2) - k * r2;
	if ( s2 < 0 )
		s2 += m2;

	k = s3 / q3;
	s2 = a1 * (s2 - k * q2) - k * r2;
	if ( s2 < 0 )
		s2 += m2;

	int Z = s1 - s2;
	if ( Z > (m1 - m3))
		Z -= m1 - 1;
	Z += s3;
	if ( Z < 1 )
		Z += m1 - 1;

	return (double)
		Z += m1 - 1;

	return (double) Z / (double) m1;
}

#endif
