#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#ifdef APPLICATIONCONDUIT
#include "ApplicationConduitInputStream.h"
#include "ApplicationConduitOutputStream.h"
#include "ApplicationConduitStream.h"
#endif // APPLICATIONCONDUIT
#include "Assert.h"
#include "CharacterInputStream.h"
#include "ClassStar.h"
#include "FileDescriptor.h"
#include "FileSystem.h"
#include "InputStream.h"
#include "MemoryObjectDictionary.h"
#include "MemoryObjectDictionaryStar.h"
#include "OutputStream.h"
#include "PersistentArray.h"
#include "PersistentMemoryObject.h"
#include "PersistentMemoryObjectStar.h"
#include "RecordStream.h"
#include "FileDescriptor.h"
#include "UnixApplication.h"
#include "UNIXFile.h"
extern "C" {
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/termios.h>
}

#ifdef __GNUC__
#pragma implementation
#endif

/*
 *
 *  FileDescriptorClass
 * 
 */


FileDescriptor::FileDescriptor()
     : _flags(0), _fileFlags(0), _class(NullFileDescriptorClass), 
       _stream(0), _pmo(0)
{
}

void FileDescriptor::init(ProxiableObject * stream, 
		          PersistentMemoryObject * pmo, 
			  FileDescriptorClass cl)
{
	Assert(cl != NullFileDescriptorClass);
	Assert(stream != 0);
	Assert(used());
	Assert(_stream == 0);

	_class = cl;
#ifdef STAR
	_streamStar = _stream;
	_pmoStar = _pmo;
#endif // STAR
	_stream = stream;
	_pmo = pmo;
	_offset = 0;
}

void FileDescriptor::init(FileDescriptor * fd)
{
	Assert(fd->fileDescriptorClass() != NullFileDescriptorClass);
	Assert(fd->proxiableObject() != 0);
	Assert(fd->used());
	Assert(used());
	Assert(_stream == 0);

	_class = fd->fileDescriptorClass();
#ifdef STAR
	_streamStar = _stream;
	_pmoStar = _pmo;
#endif // STAR
	_stream = fd->proxiableObject();
	_pmo = fd->persistentMemoryObject();
	_flags = fd->flags();
	_fileFlags = fd->fileFlags();
	_offset = fd->offset();
}

int
FileDescriptor::setFree()
{
#ifdef STAR
	_streamStar = 0;
	_pmoStar = 0;
#endif // STAR
	_stream = 0;
	_pmo = 0;
	_class = NullFileDescriptorClass;
	_flags = 0;
	_fileFlags = 0;
	_offset = 0;
	return 0;
}

void FileDescriptor::computePersistentMemoryObject()
{
        int err;
#ifdef STAR
        _pmoStar = _stream->asA(PersistentMemoryObjectClass, err);
        _pmo = _pmoStar;
#else
        _pmo = (PersistentMemoryObject*)_stream->asA(PersistentMemoryObjectClass, err);
#endif // STAR
	Assert( _pmo != 0 || err );
}

int FileDescriptor::close()
{
	int result;
	switch( _class ) {
	      case InputStreamFileDescriptorClass:
		result = inputStreamFileDescriptor()->close();
		break;
	      case OutputStreamFileDescriptorClass:
		result = outputStreamFileDescriptor()->close();
		break;
	      case RecordStreamFileDescriptorClass:
		result = recordStreamFileDescriptor()->close();
		break;
#ifdef APPLICATIONCONDUIT
	      case ApplicationConduitStreamFileDescriptorClass:
		result = applicationConduitStreamFileDescriptor()->close();
		break;
#endif // APPLICATIONCONDUIT
	      case DirectoryStreamFileDescriptorClass:
		result = directoryStreamFileDescriptor()->close();
		break;
	      default:
		Assert(NOTREACHED);
		result = -1;
		break;
	}

#ifdef STAR
	setFree();
#else // STAR
	if (_stream) _stream->unreference();
	if (_pmo) _pmo->unreference();
	setFree();
#endif // STAR

	return result;
}

int FileDescriptor::read(char * buf, int len)
{
	int result;
	switch( _class ) {
	      case InputStreamFileDescriptorClass:
		result = inputStreamFileDescriptor()->read(buf, len);
		break;
	      case OutputStreamFileDescriptorClass:
		result = outputStreamFileDescriptor()->read(buf, len);
		break;
	      case RecordStreamFileDescriptorClass:
		result = recordStreamFileDescriptor()->read(buf, len);
		break;
#ifdef APPLICATIONCONDUIT
	      case ApplicationConduitStreamFileDescriptorClass:
		result = applicationConduitStreamFileDescriptor()->
			read(buf, len);
		break;
#endif // APPLICATIONCONDUIT
	      case DirectoryStreamFileDescriptorClass:
		result = directoryStreamFileDescriptor()->read(buf, len);
		break;
	      default:
		Assert(NOTREACHED);
		result = -1;
		break;
	}
	return result;
}

int FileDescriptor::write(const char * buf, int len)
{
	int result;
	switch( _class ) {
	      case InputStreamFileDescriptorClass:
		result = inputStreamFileDescriptor()->write(buf, len);
		break;
	      case OutputStreamFileDescriptorClass:
		result = outputStreamFileDescriptor()->write(buf, len);
		break;
	      case RecordStreamFileDescriptorClass:
		result = recordStreamFileDescriptor()->write(buf, len);
		break;
#ifdef APPLICATIONCONDUIT
	      case ApplicationConduitStreamFileDescriptorClass:
		result = applicationConduitStreamFileDescriptor()->
			write(buf, len);
		break;
#endif // APPLICATIONCONDUIT
	      case DirectoryStreamFileDescriptorClass:
		result = directoryStreamFileDescriptor()->write(buf, len);
		break;
	      default:
		Assert(NOTREACHED);
		result = -1;
		break;
	}
	return result;
}

int FileDescriptor::lseek(int offset, int origin)
{
	int result;
	switch( _class ) {
	      case InputStreamFileDescriptorClass:
		result = inputStreamFileDescriptor()->lseek(offset, origin);
		break;
	      case OutputStreamFileDescriptorClass:
		result = outputStreamFileDescriptor()->lseek(offset, origin);
		break;
	      case RecordStreamFileDescriptorClass:
		result = recordStreamFileDescriptor()->lseek(offset, origin);
		break;
#ifdef APPLICATIONCONDUIT
	      case ApplicationConduitStreamFileDescriptorClass:
		result = applicationConduitStreamFileDescriptor()->
			lseek(offset, origin);
		break;
#endif // APPLICATIONCONDUIT
	      case DirectoryStreamFileDescriptorClass:
		result = directoryStreamFileDescriptor()->lseek(offset,origin);
		break;
	      default:
		Assert(NOTREACHED);
		result = -1;
		break;
	}
	return result;
}

int FileDescriptor::tell()
{
	int result;
	switch( _class ) {
	      case InputStreamFileDescriptorClass:
		result = inputStreamFileDescriptor()->tell();
		break;
	      case OutputStreamFileDescriptorClass:
		result = outputStreamFileDescriptor()->tell();
		break;
	      case RecordStreamFileDescriptorClass:
		result = recordStreamFileDescriptor()->tell();
		break;
#ifdef APPLICATIONCONDUIT
	      case ApplicationConduitStreamFileDescriptorClass:
		result = applicationConduitStreamFileDescriptor()->tell();
		break;
#endif // APPLICATIONCONDUIT
	      case DirectoryStreamFileDescriptorClass:
		result = directoryStreamFileDescriptor()->tell();
		break;
	      default:
		Assert(NOTREACHED);
		result = -1;
		break;
	}
	return result;
}

int FileDescriptor::stat(UNIXFileInfo * info)
{
	int result;
	switch( _class ) {
	      case InputStreamFileDescriptorClass:
		result = inputStreamFileDescriptor()->stat(info);
		break;
	      case OutputStreamFileDescriptorClass:
		result = outputStreamFileDescriptor()->stat(info);
		break;
	      case RecordStreamFileDescriptorClass:
		result = recordStreamFileDescriptor()->stat(info);
		break;
#ifdef APPLICATIONCONDUIT
	      case ApplicationConduitStreamFileDescriptorClass:
		result = applicationConduitStreamFileDescriptor()->stat(info);
		break;
#endif // APPLICATIONCONDUIT
	      case DirectoryStreamFileDescriptorClass:
		result = directoryStreamFileDescriptor()->stat(info);
		break;
	      default:
		Assert(NOTREACHED);
		result = -1;
		break;
	}
	return result;
}

int FileDescriptor::sync()
{
	int result;
	switch( _class ) {
	      case InputStreamFileDescriptorClass:
		result = inputStreamFileDescriptor()->sync();
		break;
	      case OutputStreamFileDescriptorClass:
		result = outputStreamFileDescriptor()->sync();
		break;
	      case RecordStreamFileDescriptorClass:
		result = recordStreamFileDescriptor()->sync();
		break;
#ifdef APPLICATIONCONDUIT
	      case ApplicationConduitStreamFileDescriptorClass:
		result = applicationConduitStreamFileDescriptor()->sync();
		break;
#endif // APPLICATIONCONDUIT
	      case DirectoryStreamFileDescriptorClass:
		result = directoryStreamFileDescriptor()->sync();
		break;
	      default:
		Assert(NOTREACHED);
		result = -1;
		break;
	}
	return result;
}

int FileDescriptor::ioctl(unsigned request, char * arg)
{
	int result;
	switch( _class ) {
	      case InputStreamFileDescriptorClass:
		result = inputStreamFileDescriptor()->ioctl(request, arg);
		break;
	      case OutputStreamFileDescriptorClass:
		result = outputStreamFileDescriptor()->ioctl(request, arg);
		break;
	      case RecordStreamFileDescriptorClass:
		result = recordStreamFileDescriptor()->ioctl(request, arg);
		break;
#ifdef APPLICATIONCONDUIT
	      case ApplicationConduitStreamFileDescriptorClass:
		result = applicationConduitStreamFileDescriptor()->
			ioctl(request, arg);
		break;
#endif // APPLICATIONCONDUIT
	      case DirectoryStreamFileDescriptorClass:
		result = directoryStreamFileDescriptor()->ioctl(request, arg);
		break;
	      default:
		Assert(NOTREACHED);
		result = -1;
		break;
	}
	return result;
}

/*
 *
 *  InputStreamFileDescriptorClass
 * 
 */

int InputStreamFileDescriptor::close()
{ 
	return 0;
}

int InputStreamFileDescriptor::read(char *buf, int len)
{ 
	return stream()->read(buf, len);
}

int InputStreamFileDescriptor::write(const char *buf, int len)
{       
	UA_errno = FileAccessViolation;
	errno = EPERM;
	return -1; 
}
    
int InputStreamFileDescriptor::lseek(int offset, int origin)
{
	return 0; 
}
    
int InputStreamFileDescriptor::tell()
{
	return 0;
}

int InputStreamFileDescriptor::stat(UNIXFileInfo *i)
{
	i->mode = 020620;
	i->blockSize = 8192;
	return 0;
}

int InputStreamFileDescriptor::sync()
{ 
	return 0; 
}

int InputStreamFileDescriptor::ioctl(unsigned request, char * arg)
{
	errno = EINVAL;
	return -1; 
}

/*
 *
 *  OutputStreamFileDescriptorClass
 * 
 */

int OutputStreamFileDescriptor::close()
{ 
	return 0;
}

int OutputStreamFileDescriptor::read(char *buf, int len)
{ 
	UA_errno = FileAccessViolation;
	errno = EPERM;
	return -1;
}

int OutputStreamFileDescriptor::write(const char *buf, int len)
{       
	return stream()->write(buf, len); 
}
    
int OutputStreamFileDescriptor::lseek(int offset, int origin)
{
	return 0; 
}
    
int OutputStreamFileDescriptor::tell()
{
	return 0;
}

int OutputStreamFileDescriptor::stat(UNIXFileInfo *i)
{
	i->mode = 020620;
	i->blockSize = 8192;
	return 0;
}

int OutputStreamFileDescriptor::sync()
{ 
	stream()->flush();
	return 0; 
}

int OutputStreamFileDescriptor::ioctl(unsigned request, char * arg)
{ 
	static struct termios dummy = { 
		0 /* c_iflag */, 
		0 /* c_oflag */, 
		0 /* c_cflag */,
		0 /* c_lflag */,
		{ /* c_cc */
			4    /*  0 VEOF */,
			0    /*  1 VEOL */,
			0    /*  2 VEOL2 */,
			127  /*  3 VERASE */,
			23   /*  4 VWERASE */,
			21   /*  5 VKILL */,
			18   /*  6 VREPRINT */,
			0    /*  7 spare */,
			3    /*  8 VINTR */,
		        28   /*  9 VQUIT */,
			25   /* 10 VSUSP */,
			0    /* 11 VDSUSP */,
			17   /* 12 VSTART */,
			19   /* 13 VSTOP */,
			22   /* 14 VLNEXT */,
			15   /* 15 VDISCARD */,
			0    /* 16 VIN */,
			0    /* 17 VTIME */,
			0    /* 18 VSTATUS */,
			0    /* 19 spare 2 */
		},
		38400 /* c_ispeed */,
		38400 /* c_ospeed */
	};
			  
	switch(request) {
        case TIOCGETA:
        {
            struct termios * t = (struct termios *) arg;
            *t = dummy;
            return 0;
        }
        case TIOCGWINSZ:
        {
            struct winsize * win = (struct winsize *) arg;
            win->ws_row = 24;
            win->ws_col = 80;
            win->ws_xpixel = 0;
            win->ws_ypixel = 0;
            return 0;
        }
	}
	errno = EINVAL;
	return -1; 
}
    
/*
 *
 *  RecordStreamFileDescriptorClass
 * 
 */

int RecordStreamFileDescriptor::close()
{ 
	return 0;
}

int RecordStreamFileDescriptor::read(char *buf, int len)
{ 
	int result = stream()->at(_offset, buf, len);
	if (result > 0) _offset += result;
	return result;
}

int RecordStreamFileDescriptor::write(const char *buf, int len)
{       
	int result = stream()->atPut(_offset, (char *) buf, len);
	if (result > 0) _offset += result;
	return result;
}
    
int RecordStreamFileDescriptor::lseek(int offset, int origin)
{
	if (origin == 0) _offset = offset;
	else if (origin == 1) _offset += offset;
	else if (origin == 2) _offset = stream()->size() + offset;
	return _offset;
}
    
int RecordStreamFileDescriptor::tell()
{
	return _offset;
}

int RecordStreamFileDescriptor::stat(UNIXFileInfo *i)
{
        getPersistentMemoryObject();
	_pmo->info(i);
        return 0;
}

int RecordStreamFileDescriptor::sync()
{ 
        getPersistentMemoryObject();
	_pmo->synchronize();
        return 0;
}

int RecordStreamFileDescriptor::ioctl(unsigned request, char * arg)
{ 
	if ( request == FIOSETHINT ) {
        	getPersistentMemoryObject();
		_pmo->setCacheHints( (unsigned) arg );
		errno = 0;
		return 0;
	} else if ( request == FIOGETHINT ) {
        	getPersistentMemoryObject();
		unsigned int * hintp = (unsigned int *) arg;
		*hintp = _pmo->getCacheHints();
		errno = 0;
		return 0;
	}
	errno = EINVAL;
	return -1; 
}

#ifdef APPLICATIONCONDUIT

/*
 *
 *  ApplicationConduitStreamFileDescriptorClass
 * 
 */

int ApplicationConduitStreamFileDescriptor::close()
{
	stream()->flush();
	stream()->finish();
	return 0;
}

int ApplicationConduitStreamFileDescriptor::read(char *buf, int len)
{ 
	return stream()->read(buf, len);
}

int ApplicationConduitStreamFileDescriptor::write(const char *buf, int len)
{       
	return stream()->write(buf, len); 
}
    
int ApplicationConduitStreamFileDescriptor::lseek(int offset, int origin)
{
	UA_errno = FileAccessViolation;
	errno = EPERM;
	Assert(NOTREACHED);
	return -1;
}
    
int ApplicationConduitStreamFileDescriptor::tell()
{
	UA_errno = FileAccessViolation;
	errno = EPERM;
	Assert(NOTREACHED);
	return -1;
}

int ApplicationConduitStreamFileDescriptor::stat(UNIXFileInfo *i)
{
	i->mode = 020620;
	i->blockSize = 8192;
	return 0;
}

int ApplicationConduitStreamFileDescriptor::sync()
{ 
	stream()->flush();
	return 0; 
}

int ApplicationConduitStreamFileDescriptor::ioctl(unsigned request, char * arg)
{ 
	errno = EINVAL;
	return -1; 
}

#endif // APPLICATIONCONDUIT

/*
 *
 *  DirectoryStreamFileDescriptorClass
 * 
 */

int DirectoryStreamFileDescriptor::close()
{
	return 0;
}

int DirectoryStreamFileDescriptor::read(char *buf, int len)
{ 
	UA_errno = UnknownNFSOp;
	errno = EISDIR;
	return -1;
}

int DirectoryStreamFileDescriptor::write(const char *buf, int len)
{       
	UA_errno = FileAccessViolation;
	errno = EISDIR;
	return -1;
}
    
int DirectoryStreamFileDescriptor::lseek(int offset, int origin)
{
    // Francis - TODO fix
	//Assert(origin == SEEK_SET);
	_offset = offset;
	return _offset;
}
    
int DirectoryStreamFileDescriptor::tell()
{
	UA_errno = FileAccessViolation;
	errno = EINVAL;
	Assert(NOTREACHED);
	return -1;
}

int DirectoryStreamFileDescriptor::stat(UNIXFileInfo *i)
{
        getPersistentMemoryObject();
	_pmo->info(i);
	return 0;
}

int DirectoryStreamFileDescriptor::sync()
{ 
        getPersistentMemoryObject();
	_pmo->synchronize();
	return 0; 
}

int DirectoryStreamFileDescriptor::ioctl(unsigned request, char * arg)
{ 
	if ( request == FIOSETHINT ) {
        	getPersistentMemoryObject();
		_pmo->setCacheHints( (unsigned) arg );
		errno = 0;
		return 0;
	} else if ( request == FIOGETHINT ) {
        	getPersistentMemoryObject();
		unsigned int * hintp = (unsigned int *) arg;
		*hintp = _pmo->getCacheHints();
		errno = 0;
		return 0;
	}
	errno = EINVAL;
	return -1; 
}

/*
 *
 *  FileDescriptorTable
 * 
 */

FileDescriptorTable::FileDescriptorTable()
{
	init();
}

int FileDescriptorTable::init()
{
	used = 0;
	return 0;
}

int FileDescriptorTable::acquire()
{
	return 0;
}

int FileDescriptorTable::release()
{
	return 0;
}

int FileDescriptorTable::closeAll()
{
	acquire();
	for(int i = 0; i < used; i++) {
		if (fd[i].used())
			fd[i].close();
	}
	release();
	return 0;
}

int FileDescriptorTable::closeOnExec()
{
	acquire();
	for(int i = 0; i < used; i++) {
		if (fd[i].flag(FDCloseOnExec)) {
			fd[i].close();
		}
	}
	release();
	return 0;
}
	
int FileDescriptorTable::allocFileDescriptor(int minfd)
{
	acquire();
    int f;
	for(f = (minfd < used ? minfd : used); 
	    f < used && fd[f].used(); 
	    f++);
	Assert(f < UA_MAXFDS); /* too many fd's inuse */
	if (f == used) {
		used++;
	}
	fd[f].setUsed();
	release();
	return f;
}

FileDescriptor * FileDescriptorTable::fileDescriptor(int i)
{
	if (i < 0 || i >= UA_MAXFDS)
		return 0;
	return &(fd[i]);
}

