#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

// #define DEBUG_REGISTERS

#include "Gdb.h"
#include "OutputStream.h"
#include "GenericARMConsole.h"
#include "GenericARMGdb.h"
#include "ARMPageTableConstants.h"
#include "ADDRESS_TRANSLATION.h"
#include "AsmBridge.h"
#include "ARMCPU.h"

#ifdef __GNUC__
#pragma implementation
#endif

/*******************************************************************
 *   FROM THE GDB SOURCE
 *   #define NUM_REGS 26
 *
 *   enum regnames {
 *    R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,
 *    FP,IP,SP,LR,PC,
 *    B0,B1,B2,B3,B4,B5,B6,B7,B8,
 *    CPSR
 *   };
 ********************************************************************/

GenericARMGdb::GenericARMGdb() {
    NUM_REGISTERS = 26;
    NUMREGBYTES = NUM_REGISTERS * 4;
    BUFMAX = NUMREGBYTES * 3;
    //ARM bkpt instruction
    GdbBreakInsn = 0xe1200070;
    setInstance(this);
}


void GenericARMGdb::GdbKill()
{
    extern void Halt();
    Halt();
}

unsigned GenericARMGdb::basicCausedInterrupt(int vector, InterruptContext *context) {

    volatile unsigned *savedRegisters = ((ARMInterruptContext *)context)->getRegisters();

    if (*(unsigned *)savedRegisters[PC] == GdbBreakInsn)
        return 1;
    else
        return 0;
}

void GenericARMGdb::basicHandle(int vector, InterruptContext *context)
{
    // We get here after a prefetch abort (BKPT instruction)
    // Switch modes to SVC
    ARMCPU::displayMode();

    unsigned * lastRegisters = registers;   // Save for later restore
    /*
     * 1.  Allocate and fill in the registers structure.
     * 2.  Point the global registers at it.
     * 3.  Call the gdb handler
     * 4.  Restore the registers
     */
    unsigned gdbRegisters[NUM_REGISTERS];
    volatile unsigned *savedRegisters;

/******************************************************************************
 * Store away the registers (from the stack) into gdbRegisters
 ******************************************************************************/
    /*
     * Get the saved globals
     */
    savedRegisters = ((ARMInterruptContext *)context)->getRegisters();

    gdbRegisters[CPSR] = ((ARMInterruptContext *)context)->getCPSR();

    // Saving R0-R15
    for (int i = 0; i < 16; i++)
        gdbRegisters[i] = savedRegisters[i];

/******************************************************************************
 * Point the registers structure there and go to the handler
 ******************************************************************************/
    registers = gdbRegisters;

#ifdef DEBUG_REGISTERS
    unsigned pre_registers[NUM_REGISTERS];  // Copy before call 
    unsigned *pre_reg_ptr = registers;
    for (i = 0; i < NUM_REGISTERS; i++)
        pre_registers[i] = registers[i];
#endif

    GdbHandler(vector, context);

#ifdef DEBUG_REGISTERS
    if (registers != pre_reg_ptr)
        Console() << "registers ptr changed from " 
                  << asAddress(pre_reg_ptr) << " to " 
                  << asAddress(registers) << "\n" << eor;
    for (i = 0; i < NUM_REGISTERS; i++)
        if (registers[i] != pre_registers[i]) {
            Console() << "registers[" << i << "] changed from "
                      << asHex(pre_registers[i]) << " to "
                      << asHex(registers[i]) << "\n" << eor;
        }
    Console() << "returning from handler\n" << eor;
#endif


/******************************************************************************
 * Restore the (possibly changed) registers structures to their savedRegisterss
 ******************************************************************************/
    /*
     * Restore the saved globals
     */
    // Restore CPSR
    ((ARMInterruptContext *)context)->setCPSR(gdbRegisters[CPSR]);

    if (*((unsigned *)gdbRegisters[PC]) == GdbBreakInsn) {
        gdbRegisters[PC] += 4;
    }

    // Restore R0-R15
    for (int i = 0; i < 16; i++)
        savedRegisters[i] = gdbRegisters[i];

/******************************************************************************
 * Restore the old registers pointer and get outta here
 ******************************************************************************/
    registers = lastRegisters;      // Restore old value
}

unsigned debuggingUser = 0;

void GenericARMGdb::GdbMachineHandlerStart(int vector, InterruptContext *context)
{
    // Francis TODO: Check if we are debugging user or kernel mode
    debuggingUser = 0;
    return;
}

void GenericARMGdb::GdbMachineHandlerEnd(int vector, InterruptContext *context)
{
    debuggingUser = 0;
}

// Francis TODO - Checking to see if ok to access memory
int GenericARMGdb::GdbMachineMem_ok(unsigned type, char *addr, unsigned len)
{
    return 0;
}

char GenericARMGdb::GetDebugChar()
{
    return GenericARMSerialGetChar();
}

void GenericARMGdb::PutDebugChar(char outch)
{
    GenericARMSerialPutChar(outch);
}

// TODO Fix this function
/* this function takes the interrupt number and attempts to
   translate this number into a unix compatible signal value */
int GenericARMGdb::ComputeSignal( int interruptVector )
{
    int sigval;

//  Debug << "ComputeSignal(" << interruptVector <<")\n" << eor;
    if (interruptVector == 2)       /* software interrupt */
        sigval = 5;
    else
        switch (interruptVector - 128) {
            case 1 : sigval = 11; break; /* seg. violation      */
            case 2 : sigval = 2;  break; /* illegal instruction */
            case 3 : sigval = 3; break; /* privilege violation  */
            case 4 : sigval = 8; break; /* FP interrupt         */
            case 7 : sigval = 10;break; /* alignment            */
            case 8 : sigval = 8; break; /* FP interrupt         */
            case 9 : sigval = 11;break; /* data access          */
            case 10 : sigval = 11;break; /* data access         */
            default:
                sigval = 5;         /* breakpoint */
        }
    return (sigval);
}


void GenericARMGdb::setPC(unsigned address) {
    registers[PC] = address;
}

void GenericARMGdb::setBreakInsn(unsigned address) {
    *((unsigned *)address) = GdbBreakInsn;
}
