#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-2006
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David Raila (raila@cs.uiuc.edu)
 * Author: Jeffrey Carlyle (jcarlyle@cs.uiuc.edu)
 * Author: Boris Capitanu (capitanu@cs.uiuc.edu)
 * Author: Ellick Chan (emchan@cs.uiuc.edu)
 * Author: Francis David (fdavid@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * GenericARMKernel.cc: GenericARM kernel class implementation.
 */

#include "defNETWORKING.h"
#include "defDISTRIBUTED.h"
#include "ADDRESS_TRANSLATION.h"
#include "Assert.h"
#include "Byte.h"
#include "DummyMemoryObject.h"
#include "Exception.h"
#include "Interrupt.h"
#include "MemoryObjectView.h"
#include "FileSystemInterface.h"
#include "InputStream.h"
#include "CharacterInputStream.h"
#include "MACHINE_CONFIGURATION.h"
#include "MemoryObjectDictionary.h"
#include "MemoryObjectContainer.h"
#include "NameServer.h"
#include "OutputStream.h"
#include "PAGE_TABLE_CONSTANTS.h"
#include "PROM.h"
#include "PhysicalMemoryRange.h"
#include "Process.h"
#include "ARMCPU.h"
#include "ARMInterrupts.h"
#include "GenericARMConsole.h"
#include "RAMDisk.h"
#include "GenericARMKernel.h"
#include "GenericARMTimeOfDay.h"
#include "GenericARMSoftClock.h"
#include "GenericARMSoftInt.h"
#include "PROCESSOR.h"
#include "AsmBridge.h"
#include "Store.h"
#include "GenericARMBootConsole.h"
#include "elF.h"

extern Store * MainStore;       // Main physical memory manager.

#include "RAMDiskInfo.h"

#define NONETWORKING 

#ifdef LINUXNET
#ifdef NONETWORKING
#undef NONETWORKING
#endif
#endif
#ifdef BSDNET
#ifndef NONETWORKING
#define NONETWORKING
#endif
#endif

#ifdef __GNUC__
#pragma implementation
#endif

#include "GenericARMGdb.h"

static const char *proc_arch[] = {
    "v3",
    "v4",
    "v4T",
    "v5",
    "v5T",
    "v5TE",
    "v5TEJ",
    "v6TEJ",
    "?(9)",
    "?(10)",
    "?(11)",
    "?(12)",
    "?(13)",
    "?(14)",
    "?(15)",
    "?(16)",
    "?(17)",
};

extern CharacterInputStreamStar ConsoleInputStream;

extern void showPTChoices(ARMPointerTableEntry* pointerTable);

GenericARMKernel::GenericARMKernel()
{
    _availableMemory = 0;
    _physicalMemorySize = 0;
}

/**
 * Dump information about the CPU to the console.
 */
void GenericARMKernel::DisplayCpuInfo(GenericARMPROM *prom)
{
    unsigned int cpuinfo;
    asm ("mrc p15, 0, %0, c0, c0" : "=r" (cpuinfo));

    unsigned int revision = cpuinfo & 0xf;
    unsigned int partNumber = (cpuinfo >> 4) & 0xfff;
    unsigned int architecture = (cpuinfo >> 16) & 0xff;
    unsigned int implementer = (cpuinfo >> 24) & 0xff;

    Console() << " (id: "  << asHex(prom->omap32_id) << eor;
    Console() << ", pid: " << asHex(prom->prod_id) << eor;
    Console() << ", rev: " << asHex(prom->dev_rev) << ")\n" << eor;

    Console() << "CPU: ";

    switch ((char)implementer) {
        case 'A':
            Console() << "ARM";
            break;

        case 'D':
            Console() << "Digital";
            break;

        default:
            Console() << "Unknown";
            break;
    };

    switch (partNumber) {
        case 0x926:
            Console() << "926";
            break;

        case 0xa26:
            Console() << "1026";
            break;

        default:
            Console() << asHex(partNumber);
            break;
    };

    switch (architecture) {
        case 0x4:
            Console() << "S";
            break;

        case 0x5:
            Console() << "E-S";
            break;

        case 0x6:
            Console() << "EJ-S";
            break;

        default:
            break;
    };

    Console() << " (" << proc_arch[architecture] << "), rev: "
            << asHex(revision) << "\n\n" << eor;
}

void
GenericARMKernel::buildRunningCPU()
{
    /*
     * Build a CPU object for this processor and install interrupts.
     */
    CPU * current = new ARMCPU( PROCESSOR::cpuID(), _initialScheduler );
    current->setName( "#CPU" );
    current->setInterruptManager(new MACHINEINTERRUPTMANAGER());
    current->setIdleContainer(_initialScheduler);
    current->installInterrupts();
}

PhysicalMemoryRange *
GenericARMKernel::physicalMemoryMap()
{
    Debug << "GenericARMKernel::physicalMemoryMap()\n" << eor;

    /*
     * Get the available physical memory
     */
    unsigned physMemory = 0;
    struct memlist *phys_available = 0;
    struct memlist *phys_total = 0;

    MACHINEPROM::setmemlists(phys_available, phys_total);

    /*
     * Find out how much memory we have on board to print
     * out a nice message.
     */
    struct memlist *chunk = phys_total;
    while (chunk)   {
        physMemory += chunk->size;
        chunk = chunk->next;
    }

    /*
     * Configure the Physical memory map, also find the amount
     * of memory available.
     *
     * The physicalMemoryMap needs to be laid out with the following
     * separate pieces on the ARM:
     *  1. Kernel
     *  2. Any other physical memory chunks not taken by the PROM
     */
    chunk = phys_available;
    
    Debug << "Starting physical memory map: "
          << asAddress((char*)chunk->address)
          << " size: " << asHex(chunk->size) << "\n" << eor;

    PhysicalMemoryRange * _physicalMemoryMap =
    new PhysicalMemoryRange( (char *)chunk->address,
                             chunk->size/PhysicalPageSize, Log2PhysicalPageSize,
                             "#Memory (RAM)" );
    _availableMemory += chunk->size;

    chunk = chunk->next;
    while (chunk) {
        Debug << "Adding to physical memory map: "
              << asAddress((char*)chunk->address) 
              << " size: " << asHex(chunk->size) << "\n" << eor;

        _physicalMemoryMap->add(
            new PhysicalMemoryRange ((char*)chunk->address,
                                     chunk->size/PhysicalPageSize, Log2PhysicalPageSize,
                                     "#Memory (RAM)" ));

        _availableMemory += chunk->size;
        chunk = chunk->next;
    }

    _physicalMemorySize = physMemory;

    Console() << "Total memory : " << physMemory << " ("
              << asHex(physMemory) << ")\n" << eor;
    Console() << "Available memory : " << _availableMemory
              << " (" << asHex(_availableMemory) << ")\n" << eor;

    Debug << "Exiting GenericARMKernel::physicalMemoryMap()\n" << eor;
    return _physicalMemoryMap;
}

const unsigned
GenericARMKernel::physicalMemorySize()
{
    Console() << "GenericARMKernel::physicalMemorySize() " << _physicalMemorySize << "\n" << eor;
    return (const unsigned) _physicalMemorySize;
}

void
GenericARMKernel::mountNonRootFileSystems(){
    _fileSystemInterface->mount( "!6/", "usr", 0 );
    _fileSystemInterface->mount( "!7/", "mnt/local", 0 );
}


void
GenericARMKernel::initialize( Allocator * bootHeap )
{
    Debug << this << "::initialize()\n" << eor;
    bootConsole = new GenericARMBootConsole();
    bootConsole->run();
    Kernel::initialize(bootHeap);
}

/**
 * Initialize the backing store. We don't have a backing store on GenericARM.
 */
void GenericARMKernel::initializeBackingStore()
{
#if 0
    MemoryObjectStar disk = rootDisk();
    Assert( disk != 0 );
    int status = 0;
    MemoryObjectContainerStar rootContainer =
    disk->asA( MemoryObjectContainerClass, status );
    Assert( status == 0 );
    Assert( rootContainer != 0 );
    MemoryObjectStar mo = rootContainer->open( 6, status, O_NCACHE );
    Assert( status == 0 );
    Assert( mo != 0 );
    _backingStoreSource = mo->asA( MemoryObjectContainerClass, status );
    Debug << "backingStoreSource: " << _backingStoreSource << "\n" << eor;
    Assert( _backingStoreSource != 0 );
    _backingStoreSource->setName( "backingStoreSource" );
#endif
}


/**
 * Load an ELF header. Used to implement isObjectFile and loadObjectFile.
 *
 * @param   file    [in] File to load ELF header from.
 * @param   pHeader [out] ELF header structure.
 *
 * @return  Zero on failure; non-zero on success.
 */
int GenericARMKernel::loadObjectHeader(PersistentMemoryObject * file,
        llvm::ELF::Elf32_Ehdr *header)
{
    FileObjectStar byteFile = new FileObject( file->numberOfBytes(), 0, 0, 0,
            file );
    Assert(byteFile != 0);

    /* read the elf header */
    int num = byteFile->read(0, sizeof(*header), (char*)header,
            sizeof(*header));
    Debug << "num: " << num << " header->e_machine: " << header->e_machine
        << "\n" << eor;
    if(num != sizeof(*header)) {
        Debug << this << "::" << __FUNCTION__ 
                << "(): can't read elf header!\n" << eor;
        return 0;
    }

    if(!header->checkMagic()) {
        Debug << this << "::" << __FUNCTION__
                << "(): file doesn't match ELF magic number!\n" << eor;
        return 0;
    }

    if(header->getFileClass() != llvm::ELF::ELFCLASS32) {
        Debug << this << "::" << __FUNCTION__
                << "(): file class is not ELF32!\n" << eor;
        return 0;
    }

    if(header->e_version != 1) {
        Debug << this << "::" << __FUNCTION__
                << "(): ELF header is not version 1!\n" << eor;
        return 0;
    }

    if(header->e_machine != llvm::ELF::EM_ARM) {
        Debug << this << "::" << __FUNCTION__
                << "(): not an ARM executable!\n" << eor;
        return 0;
    }

    if(header->e_type == llvm::ELF::ET_DYN) {
        Debug << this << "::" << __FUNCTION__
                << "(): Dynamic file!\n" << eor;
        return 0;
    }

#if 0
    /* left over from the a.out days; appears to be related to relocation */
    if ( (hdr->a_trsize + hdr->a_drsize) != 0 ) {
        Debug << this << "::loadObjectFile(): Bad trsize + drsize\n" << eor;
        delete hdr;
        return 0;
    }
#endif /* if 0 */

    return 1;
}


/**
 * Determine if a file is an ELF executable.
 *
 * @param   file    File to test.
 *
 * @return  Zero on failure; non-zero on success.
 */
int GenericARMKernel::isObjectFile( PersistentMemoryObject * file )
{
    Debug << this << "::" << __FUNCTION__ << "( file: " << file
        << ")\n" << eor;
    Debug << this << "::" << __FUNCTION__ << "( file: " << file 
        << ")\n" << eor;

    Assert ( file != 0 );

    llvm::ELF::Elf32_Ehdr header;
    if( loadObjectHeader(file, &header) == 0 ) {
        /* failure reading header */
        return 0;
    }

    return 1;
}


/**
 * Load an ARM ELF executable file into memory.
 *
 * @param   file        The executable file to load.
 * @param   newDomain   The virtual memory domain for the new application.
 *
 * @return  Virtual memory address withini newDomain of the application entry point; or zero on failure.
 */
const char *
GenericARMKernel::loadObjectFile( PersistentMemoryObject * file,
                            Domain * newDomain )
{
    Debug << this << "::loadObjectFile( file: " << file 
          << ", domain: " << newDomain << ")\n" << eor;
    
    Assert ( file != 0 );
    Assert ( newDomain != 0 );

    /* We should make sure all sections are loaded between applicationStart
     * and applicationEnd. */
    const char *applicationEnd = this->applicationStart() +
        this->applicationSize();

    llvm::ELF::Elf32_Ehdr header;
    if( loadObjectHeader(file, &header) == 0 ) {
        /* failure reading header */
        return 0;
    }

    /* sanity check */
    Assert(header.checkMagic());
    Assert(header.getFileClass() == llvm::ELF::ELFCLASS32);
    Assert(header.e_version == 1);
    Assert(header.e_machine == llvm::ELF::EM_ARM);
    Assert(header.e_type != llvm::ELF::ET_DYN);

    if( (header.e_entry < (unsigned int)this->applicationStart()) ||
            (header.e_entry >= (unsigned int)applicationEnd) ) {
        Debug << "entry point is outside of application space: "
            << asHex(header.e_entry) << "\n" << eor;
        return 0;
    }

    FileObjectStar byteFile = new FileObject( file->numberOfBytes(), 0, 0, 0,
            file );
    Assert(byteFile != 0);

    Debug << "application start: "
        << asHex((char *)this->applicationStart())
        << " end: " << asHex((char*)applicationEnd) << " size: "
        << asHex((int)this->applicationSize()) << "\n" << eor;
    Debug << "pht entries: " << header.e_phnum << " pht entry size: "
        << header.e_phentsize << "\n" << eor;
    Debug << "e_phoff: " << header.e_phoff << " fileObject l2us: "
        << byteFile->log2UnitSize() << "\n" << eor;

    for(int index = 0; index < header.e_phnum; index++) {
        llvm::ELF::Elf32_Phdr pht;
        byteFile->read(
            header.e_phoff + (index * header.e_phentsize),
            sizeof(pht), (char*)&pht, sizeof(pht));

        Debug << "ELF size: " << pht.p_memsz << " p_type: " << asHex(pht.p_type) 
              << " p_offset: " << asHex(pht.p_offset) << " vaddr " << asHex(pht.p_vaddr) << "\n" << eor;

        if( (pht.p_type == llvm::ELF::PT_LOAD) &&
                (pht.p_memsz > 0) ) {
            if( (pht.p_vaddr >= (unsigned int)this->applicationStart()) &&
                    ((pht.p_vaddr+pht.p_memsz)<(unsigned int)applicationEnd)) {
                MemoryObjectViewStar view =
                    new NormalMemoryObjectView(byteFile, pht.p_memsz, 0,
                            pht.p_offset);
                Assert( view != NULL );

                char *mem = new char[pht.p_memsz];
                Assert( mem != NULL );

                RAMDisk *disk = new RAMDisk(mem, pht.p_memsz, 0);
                Assert( disk != NULL );

                PersistentMemoryObjectRef mo = PersistentMemoryObjectRef(disk);
                Assert( mo != NULL );

                mo->setName(".text");
                mo->buildCache();
                mo->cache()->setName(".text");

                const char * source = TheKernel->domain()->add( view,
                        NoAccess );
                const char * dest = TheKernel->domain()->add( mo, NoAccess );

                Assert( pht.p_filesz <= pht.p_memsz );

                /* 
                 * TODO: anything between p_memsz and p_filesz is likely
                 * to be in the BSS segment; so eventually we need to make sure
                 * its filled with zeros.
                 */

                /* TODO: This isn't exactly copy-on-write. */
                ByteCopy( ( char * ) source, ( char * ) dest, pht.p_filesz );
                        
                TheKernel->domain()->remove( view, source );
                TheKernel->domain()->remove( mo, dest );

                const char *where = newDomain->add(mo,
                        (char*)pht.p_vaddr, 
                        pht.p_flags & llvm::ELF::SHF_WRITE ? 
                            ReadWrite : ReadOnly);
                Assert( where == (char*)pht.p_vaddr );
            }
            else
            {
                Debug << "ELF does not fit into application vm-area.\n"
                    << eor;
                return 0;
            }
        }
    }

    Debug << "application entrypoint: " << asHex(header.e_entry) << "\n"
        << eor;
    return (const char *)header.e_entry;
}

/*
 * Setup machine dependent devices and device managers.
 */
void GenericARMKernel::basicInitializeDevicesAndManagers()
{
    Debug << "GenericARMKernel::basicInitializeDevicesAndManagers()\n" << eor;
    Assert( _initialScheduler != 0 );

    // Add MemoryObject for interrupts
    PhysicallyResidentMemoryObject *mo = new PhysicallyResidentMemoryObject( 1,
            Log2PhysicalPageSize, (char *)KERNEL_START_VIRTUAL, 0);
    domain()->add( mo, (char *)0xffff0000, NoAccess );
    domain()->makeAddressable((char *)0xffff0000, PhysicalPageSize);
}

/**
 * Enable Console interrupts.
 */
void GenericARMKernel::EnableConsoleInterrupts(void)
{
    /*
     * Now that interrupts are available, set up the Console
     * for interrupts.
     */
    Debug << "Enabling Console interrupts\n" << eor;
    ((GenericARMConsole*)ConsoleOutputStream)->setInterruptMode();
    Debug << "Enabled Console interrupts\n" << eor;
}

/**
 * Build the time of day clock
 */
void GenericARMKernel::InstallTimeOfDayClock(void)
{
    theSystemClock = new GenericARMTimeOfDay();
    StandardNameServer->bind( "timeOfDay", theSystemClock );
}

/**
 * Now install software interrupt manager
 */
void GenericARMKernel::InstallSoftwareInterruptManager(void)
{
    TheSoftwareInterruptManager = new GenericARMSoftwareInterruptManager();
}

/**
 * Now install software clock manager
 */
void GenericARMKernel::InstallSoftwareClockManager(void)
{
    TheSoftwareClockManager = new GenericARMSoftwareClockManager();
}

/**
 * Start the Timeout timer & Periodic timer helper processes.
 */
void GenericARMKernel::InstallTimerProcesses(void)
{
    Debug << "Build Timeout Manager process\n" << eor;
    extern void TimeoutTimerEntry( int );
    Process *process = new SystemProcess( (PFV) TimeoutTimerEntry,
                                          domain(), normal, 0 );
    Debug << "Timeout manager process " << process << "\n" << eor;
    Assert( process != 0 );
    process->setName( "#TimeoutTimerManager" );
    process->ready();
    thisProcess()->relinquishProcessor();

    Debug << "Build Periodic Manager process\n" << eor;
    extern void PeriodicTimerEntry( int );
    process = new SystemProcess( (PFV) PeriodicTimerEntry,
                                 domain(), normal, 0 );
    Debug << "Periodic manager process " << process << "\n" << eor;
    Assert( process != 0 );
    process->setName( "#PeriodicTimerManager" );
    process->ready();
    thisProcess()->relinquishProcessor();
}


/**
 * Start up the RAM disk.
 *
 * @param   baseAddress     Base address of the RAM disk data.
 */
void GenericARMKernel::InstallTheRamDisk(char *baseAddress)
{
    theDisk = new RAMDisk(
            baseAddress,
            __CHOICES_VDISK_NUMBLOCKS__,
            __CHOICES_VDISK_LOG2BLKSIZE__);
    // theDisk->diskScheduler()->startDaemon();
}

void
GenericARMKernel::shutdown()
{
    Debug << "GenericARMKernel::shutdown()\n" << eor;
    PROCESSOR::disableInterrupts();
}

unsigned debugApplications = 0;

void hello_world_nodie(void *arg) {
    ARMCPU::displayMode();
    Console() << "Hello World!\n" << eor;
}

int
GenericARMKernel::tryCommand( int argc, char *argv[], int partnumber)
{
    int result = 1;
    unsigned debugMask;
    char* testtext = "Test for print functionality... OK\n";
    switch (partnumber) {
        case 1027: 
        {
            int status;
            PersistentMemoryObjectStar mo = _fileSystemInterface->open("/id",
                                                                       PersistentMemoryObjectClass, status);
            if(mo==NULL) {
                Console() << this << "::" << __FUNCTION__
                          << " open returned null: " << "status=" << status
                          << "\n" << eor;
                break;
            }
            else {
                Console() << this << "::" << __FUNCTION__ << "(" << __FILE__
                          << ":" << __LINE__ << ") mo " << mo << ", status "
                          << status << "\n" << eor;
            }

            const char *base = thisProcess()->domain()->add( mo, ReadWrite );
            char *pastEnd = ( char * ) base + mo->numberOfBytes();
            char *current = ( char * ) base;

            Console() << "number of bytes: " << mo->numberOfBytes() << " ["
                      << eor;

            try
            {
                while(current != pastEnd) {
                    Console() << *current++ << eor;
                }
    
                thisProcess()->domain()->remove( mo, base );
            }
            catch(Exception *e)
            {
                Console() << "caught an exception: " << asAddress(e) << "\n"
                          << eor;
                TrapToDebugger();
            }

            Console() << "]\n" << eor;
          
            break;
        }


        case 914:
            unsigned retval;
            unsigned pc;
            asm("mov %0, pc" : "=r"(pc));
            Console() << "Current domain is: " << this->domain() << " pc:" << asHex(pc) << "\r\n";
                
            ARMCPU::displayMode();
            Console() << "Changing mode to USR, current mode is: " << asHex(ARMCPU::getMode()) << "\r\n" << eor;
            ARMCPU::setMode(ARM_USR);
            GenericARMSerialPutS("Changed to USR mode\r\n");
            Console() << "Changing mode to USR, new mode is: " << asHex(ARMCPU::getMode()) << "\r\n" << eor;
            Console() << "Doing test syscall\r\n" << eor;
            asm("swi #0x7b; mov %0, r0" : "=r"(retval):: "r0");
            Console() << "Return from syscall, retval=" << asHex(retval) << "\r\n" << eor;
            ARMCPU::displayMode();

            //testing print
            Console() << "Testing PRINT syscall 0xffe\n" << eor;
            asm("mov r0, %0; swi #0xffe;" :: "r"(testtext) : "r0");

            //testing function calls
            Console() << "Testing FUNCTION syscall 0xffd\n" << eor;
            asm("mov r0, %0; swi #0xffd;" :: "r"(hello_world_nodie) : "r0");

            // return to system mode
            asm("swi #0xfff"::: "r0");
            ARMCPU::displayMode();
            break;
        case 915:
        {
            // Print page tables (via Choices)
            ARMTranslation* trans = (ARMTranslation*) TheKernel->domain()->translation();
            ARMPointerTableEntry* pointerTable = trans->pointerTable();
            showPTChoices(pointerTable);
            break;
        }
        case 444:
            debugMask = 0x1;
            if (debugApplications & debugMask)
                debugApplications &= ~debugMask;
            else
                debugApplications |= debugMask;
            Console() << this << ": debugApplications = " <<
            asHex(debugApplications) << "\n" << eor;
            Console() << this << ": Stop Application in _start\n" 
                      << eor;
            break;
        case 445:
            debugMask = 0x2;
            if (debugApplications & debugMask)
                debugApplications &= ~debugMask;
            else
                debugApplications |= debugMask;
            Console() << this << ": debugApplications = " <<
            asHex(debugApplications) << "\n" << eor;
            Console() << this << ": Stop Application before _main\n" 
                      << eor;
            break;
        case 446:
            debugMask = 0x4;
            if (debugApplications & debugMask)
                debugApplications &= ~debugMask;
            else
                debugApplications |= debugMask;
            Console() << this << ": debugApplications = " <<
            asHex(debugApplications) << "\n" << eor;
            Console() << this << ": Stop Application after _main\n" 
                      << eor;
            break;
        case 500:
            Assert (NOTREACHED);
            break;
        case 222:
#ifndef NONETWORKING
#ifdef NETWORKING
            extern GenericARMAm7990 * le0;
            if (le0 == 0){
                Console() << "le0 == 0!\n" << eor;
            }
            else    {
                le0->printInfo();
            }
#else
            Console() << "Networking is disabled\n" << eor;
#endif
#endif
            break;
        case 7000:
#ifdef DISTRIBUTED
            extern void startServer();
            startServer();
#endif // DISTRIBUTED
            break;
        case 7001:
#ifdef DISTRIBUTED
            extern void startClient();
            startClient();
#endif // DISTRIBUTED
            break;
        case 8000:
            // GenericARMBenchmarks(argc, argv);
            break;
        default:
            try {
                result = Kernel::tryCommand(argc, argv, partnumber);
            } catch (Interrupt *e) {
                Console() << "Caught Interrupt " << *e << " in tryCommand!\n" << eor;
            }
            break;
    }
    return result;
}

DiskRef
GenericARMKernel::rootDisk()
{
    return(theDisk);
}

PersistentMemoryObjectRef
GenericARMKernel::allocateMemoryObjectForBackingStore()
{
    char* mem = new char[4096 * 32];
    RAMDisk* disk = new RAMDisk(mem, 32, 12);
    return PersistentMemoryObjectRef(disk);
}

#include "ClassInstaller.h"
DEFINECLASS("GenericARMKernel",GenericARMKernel,Kernel,0,0);
