#ifdef __GNUC__
#pragma implementation
#endif

#include "IntegratorInterrupts.h"
#include "OutputStream.h"
#include "IntegratorInterruptManager.h"
#include "IntegratorSerial.h"
#include "IntegratorTimer.h"
#include "Vectors.h"
#include "ARMCPU.h"

unsigned IntegratorInterruptManager::initialized = 0;

IntegratorInterruptManager::IntegratorInterruptManager() : InterruptManager(nInterrupts) {
    // Need some locking if we are initializing for multiple CPUs
    if (!initialized) {
        setupInterrupts();
        enableInterrupts();
    }
}

IntegratorInterruptManager::~IntegratorInterruptManager() {
    Debug << "IntegratorInteruptManager destructor\n" << eor;
}

unsigned IntegratorInterruptManager::getInterruptNumber() {
    // Read the registers in PIC and SIC (if necessary) to determine the interrupt number

    unsigned status = *reinterpret_cast<unsigned *>(PIC_IRQ_STATUS);
    unsigned source1 = 0;

    // find first set bit
    while(!((status >> source1) & 1)) source1++;
 
    if (source1 == ARM_PIC_CPPLDINT) {
        unsigned source2 = *reinterpret_cast<unsigned *>(SIC_BASE + SIC_INT_STATUS);
        Debug << "Got an SIC interrupt: " << source2 << "\n\n" << eor;
        return source2 + ARM_PIC_VEC_CNT;
    } else {
        return source1 + ARM_PIC_VEC_BASE;
    }
}

void IntegratorInterruptManager::ackInterrupt(unsigned number) {
    // Acknowledge the interrupt to PIC and SIC
    if (number >= ARM_PIC_VEC_BASE) {
        //PIC interrupt
        unsigned* picexclr = (unsigned *)(PIC_IRQ_ENABLECLR);

        if (number == 8) {
            // Loopback soft interrupt
            Debug << "PIC Interrupt " << (number - ARM_PIC_VEC_BASE) << "\n" << eor;
            Console() << "Loopback soft interrupt " << number << " masked\n" << eor;
            *picexclr = ( 1 << (number - ARM_PIC_VEC_BASE));
        }
    }
}


void IntegratorInterruptManager::setupInterrupts() {
    initialized = 1;
}

void IntegratorInterruptManager::enableInterrupts() {
    unsigned* pic_irq_enable_set = (unsigned*) PIC_IRQ_ENABLESET;
    unsigned* sic_irq_enable_set = (unsigned*) SIC_INT_ENABLESET;

    //Disable PIC/SIC
    disableInterrupts();

    // Enable SIC
    Debug << "Enabling SIC\n" << eor;

    //unmask individual SIC interrupts
    *sic_irq_enable_set |= SIC_SOFTINT;

    Debug << "Enabled SIC\n" << eor;

    Debug << "Enabling PIC\n" << eor;
    // Individual interrupt mask register
//  *pic_irq_enable_set |= (1 << ARM_PIC_UARTINT0);

    // set softint
//  *pic_irq_enable_set |= (1 << ARM_PIC_SOFTINT);
//  *pic_irq_enable_set |= (1 << ARM_PIC_UARTINT1);
    *pic_irq_enable_set |= (1 << ARM_PIC_TIMERINT0);
    *pic_irq_enable_set |= (1 << ARM_PIC_TIMERINT1);
//  *pic_irq_enable_set |= (1 << ARM_PIC_TIMERINT2);
//  *pic_irq_enable_set |= (1 << ARM_PIC_CPPLDINT);
    Debug << "Enabled PIC\n" << eor;
}

void IntegratorInterruptManager::disableInterrupts() {
    // Mask all interrupts in PIC and SIC
    unsigned* pic_irq_enable_clr = (unsigned*) PIC_IRQ_ENABLECLR;
    unsigned* sic_irq_enable_clr = (unsigned*) SIC_INT_ENABLECLR;
    *pic_irq_enable_clr = 0xffffffff;
    *sic_irq_enable_clr = 0xffffffff;
}

void IntegratorInterruptManager::enableInterrupt(unsigned number) {
    unsigned* pic_irq_enable_set = (unsigned*) PIC_IRQ_ENABLESET;
    //unsigned* sic_irq_enable_set = (unsigned*) SIC_INT_ENABLESET;

    Debug << "IntegratorInterruptManager::enableInterrupt(" << number << ");\n" << eor;
    if (number >= 8) {
        //PIC or SIC
        if( number - ARM_PIC_VEC_BASE < ARM_PIC_VEC_CNT ) {
            //PIC
            Debug << "Enabling PIC int: " << (number - ARM_PIC_VEC_BASE) << "\n" << eor;
            *pic_irq_enable_set |=  1 << (number - ARM_PIC_VEC_BASE);
        } else if ( number - ARM_PIC_VEC_BASE - ARM_PIC_VEC_CNT < SIC_MAXCNT) {
            //SIC
            Debug << "Enabling SIC int: " << (number - ARM_PIC_VEC_BASE - ARM_PIC_VEC_CNT);
            //  *sic_irq_enable_set |=  1 << (number - ARM_PIC_VEC_BASE - ARM_PIC_VEC_CNT);
            Debug << "SIC interrupts not supported yet\n" << eor;
            while(1);
            Assert(NOTREACHED);
        } 
    }
}

void IntegratorInterruptManager::disableInterrupt(unsigned number) {
    Debug << "IntegratorInterruptManager::disableInterrupt() not supported yet\n" << eor;
    while(1);
    Assert(NOTREACHED);
}
