#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-2006
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Ellick Chan (emchan@cs.uiuc.edu)
 * Author: David Raila (raila@cs.uiuc.edu)
 * Contributing Authors: Francis David (fdavid@cs.uiuc.edu)
 *                       Jeffrey Carlyle (jcarlyle@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * IntegratorKernel.cc: Integrator kernel class implementation.
 */

#include "DummyMemoryObject.h"
#include "MachPROM.h"
#include "MemoryObject.h"
#include "IntegratorKernel.h"
#include "IntegratorSerial.h"
#include "PhysicalMemoryRange.h"
#include "RAMDisk.h"

#include "RAMDiskInfo.h"

IntegratorKernel::IntegratorKernel()
{
    IntegratorPROM *prom = new IntegratorPROM();

    Console() << "QEMU ARM Integrator CP board ";

    DisplayCpuInfo(prom);
}

PhysicalMemoryRange *
IntegratorKernel::physicalMemoryMap()
{
    Debug << "IntegratorKernel::physicalMemoryMap()\n" << eor;

    PhysicalMemoryRange *_physicalMemoryMap =
        GenericARMKernel::physicalMemoryMap();

    // NOTE: This function only causes PTE to be allocated. It doesn't
    // do any memory mapping. Use basicAddMemoryObjects() or
    // basicInitializeDevicesAndManagers() below for the actual mapping.

    // Add mapping for Memory Mapped I/O region
    _physicalMemoryMap->add(new PhysicalMemoryRange( (char *)UART_BASE,
                                                     0x1000/PhysicalPageSize, Log2PhysicalPageSize,
                                                     "#MemoryMapped UART I/O", 1 ));
    _physicalMemoryMap->add(new PhysicalMemoryRange( (char *)0xfff00000,
                                                     0x1000/PhysicalPageSize, Log2PhysicalPageSize,
                                                     "#MemoryMapped I/O", 1 ));
    _physicalMemoryMap->add(new PhysicalMemoryRange( (char *)PIC_BASE,
                                                     0x1000/PhysicalPageSize, Log2PhysicalPageSize,
                                                     "#Programmable interrupt controller", 1 ));
    _physicalMemoryMap->add(new PhysicalMemoryRange( (char *)SIC_BASE,
                                                     0x1000/PhysicalPageSize, Log2PhysicalPageSize,
                                                     "#Second programmable interrupt controller", 1 ));
    _physicalMemoryMap->add(new PhysicalMemoryRange( (char *)Integrator_MPU_TIMER1_BASE,
                                                     0x1000/PhysicalPageSize, Log2PhysicalPageSize,
                                                     "#Timer 0", 1 ));
    _physicalMemoryMap->add(new PhysicalMemoryRange( (char *)INITRD,
                                                     __CHOICES_VDISK_SIZE__/PhysicalPageSize, Log2PhysicalPageSize,
                                                     "#Initrd disk image", 1 ));

    Debug << "Exiting IntegratorKernel::physicalMemoryMap()\n" << eor;
    return _physicalMemoryMap;
}

const unsigned 
IntegratorKernel::kernelHeapSize()
{   
    return (const unsigned) 0x200000;
}

/**
 * Map virtual memory for upper interrupt region and memory mapped devices.
 * Consider using basicInitializeDevicesAndManagers().
 */
void
IntegratorKernel::basicAddMemoryObjects()
{
    // Add MemoryObject for Serial I/O; other peripherals are added later
    // in basicInitializeDevicesAndManagers
    PhysicallyResidentMemoryObject *mo = new PhysicallyResidentMemoryObject( 1,
            Log2PhysicalPageSize, (char *)UART_BASE, 1);
    domain()->add(mo, (char *)UART_BASE, NoAccess);
    domain()->makeAddressable((char *)UART_BASE, PhysicalPageSize);
}



/*
 * Setup machine dependent devices and device managers.
 */
void
IntegratorKernel::basicInitializeDevicesAndManagers()
{
    Debug << "IntegratorKernel::basicInitializeDevicesAndManagers()\n" << eor;

    GenericARMKernel::basicInitializeDevicesAndManagers();

    // Map in PIC/SIC
    PhysicallyResidentMemoryObject *mo =
        new PhysicallyResidentMemoryObject( 1, Log2PhysicalPageSize,
                (char *)PIC_BASE, 1);
    domain()->add(mo, (char *)PIC_BASE, NoAccess);
    domain()->makeAddressable((char *)PIC_BASE, PhysicalPageSize);

/* Enable secondary interrupt controller later
   mo = new PhysicallyResidentMemoryObject( 0x10000/PhysicalPageSize, Log2PhysicalPageSize, (char *)SIC_BASE, 1);
   domain()->add(mo, (char *)SIC_BASE, NoAccess);
   domain()->makeAddressable((char *)SIC_BASE, 0x10000);
*/

    /* Test PIC/SIC */
#ifdef DEBUG
    Debug << "Testing PIC SOFTINT\n" << eor;
    Assert(ARMCPU::interruptsEnabled());
    
    unsigned* pic_irq_softset = (unsigned*) PIC_INT_SOFTSET;
    *pic_irq_softset |= 1;

    Debug << "Tested PIC SOFTINT\n" << eor;
#endif //DEBUG

    /*
     * Now that interrupts are available, set up the Console
     * for interrupts.
     */
    EnableConsoleInterrupts();

    /* Test console interrupts */
#ifdef DEBUGINTSERIAL
    unsigned oldstat = 0, cnt=10;
    Debug << "Old UART raw int stat was: " << asHex(oldstat) << "\n" << eor;
    while(1) {
        unsigned rawintstat = *((unsigned*)(UARTTRIS));
        unsigned maskintstat = *((unsigned*)(UARTMIS));
        
        if( oldstat != rawintstat && cnt ) {
            Debug << "New UART raw int stat is: " << asHex(rawintstat) << "\n" << eor;
            Debug << "New UART masked int stat is: " << asHex(maskintstat) << "\n" << eor;
            Debug << "Old UART raw int stat was: " << asHex(oldstat) << "\n" << eor;
            oldstat = rawintstat;
            cnt--;
        }
    }
#endif

    /* Map in Timer0 */
    mo = new PhysicallyResidentMemoryObject( 1, Log2PhysicalPageSize, (char *)Integrator_MPU_TIMER1_BASE, 1);
    domain()->add(mo, (char *)Integrator_MPU_TIMER1_BASE, NoAccess);
    domain()->makeAddressable((char *)Integrator_MPU_TIMER1_BASE, PhysicalPageSize);

    /*
     * Start the Timeout timer & Periodic timer helper processes.
     *
     */
    InstallTimerProcesses();

    /*
     * Build the time of day clock
     */
    InstallTimeOfDayClock();

    /*
     * Now install software interrupt manager
     */
    InstallSoftwareInterruptManager();

    /*
     * Now install software clock manager
     */
    InstallSoftwareClockManager();

    /* Map in initial ramdisk */
    mo = new PhysicallyResidentMemoryObject(
            __CHOICES_VDISK_SIZE__/PhysicalPageSize,
            Log2PhysicalPageSize, (char *)INITRD, 1);
    domain()->add(mo, (char *)INITRD, NoAccess);
    domain()->makeAddressable((char *)INITRD, __CHOICES_VDISK_SIZE__);

    /*
     * Start up the disk.
     */
    InstallTheRamDisk((char *)INITRD);
}

#include "ClassInstaller.h"
DEFINECLASS("IntegratorKernel",IntegratorKernel,GenericARMKernel,0,0);
