#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * CRQ.c
 */

#include "Assert.h"
#include "OutputStream.h"
#include "Vectors.h"
#include "CRQ.h"
#include "PhysicallyAddressableUnit.h"
#include "PhysicalMemoryChain.h"


#ifdef __GNUC__
#pragma implementation
#endif

/*
 * Dump the contents of a CRQ to the console (for debugging).
 */
void
CRQ::dump()
{
	Debug.printf( "CRQ%T::dump\n", this );
	Assert( this != 0);

	Debug.printf( "CRQ: _mplock.lock_state:%x, .lock_level:%x\n",
		crq_mplock.lock_state, crq_mplock.lock_level );
	Debug.printf( "CRQ: _mplock.lock_owner:%x, .lock_type:%x, .lock_stats:%x\n",
		crq_mplock.lock_owner, crq_mplock.lock_type,
		crq_mplock.lock_stats );
	Debug.printf( "CRQ: _mode:%x, _opt:%x, _unitid:%x\n",
		crq_mode, crq_opt, crq_unitid );
	Debug.printf( "CRQ: _slave_vect:%x, _master_vect:%x\n",
		crq_slave_vect, crq_master_vect );
	Debug.printf( "CRQ: _cmd empty = %d\n", crq_cmd.isEmpty() );
	Debug.printf( "CRQ: _immedcmd empty = %d\n", crq_immedcmd.isEmpty() );
	Debug.printf( "CRQ: _rsp empty = %d\n", crq_rsp.isEmpty() );
	Debug.printf( "CRQ: _attn empty = %d\n", crq_attn.isEmpty() );
	Debug.printf( "CRQ: _free empty = %d\n", crq_free.isEmpty() );
	Debug.printf( "CRQ: _totcmds:%x, _totattns:%x, _totvects:%x\n",
		crq_totcmds, crq_totattns, crq_totvects );
	Debug.printf( "CRQ: _stats:%x\n", crq_stats );
}

CRQ::CRQ( unsigned int unitID, unsigned int master_vector, int mode )
	: SYSHDR( sizeof( CRQ ), TYPE_CRQ, 0 )
{
	crq_mplock.lock_state = LOCK_FREE;	// state is free;
	crq_mplock.lock_level = ChoicesIOLockLevel;
	crq_mplock.lock_owner = 0;		// no owner yet
	crq_mplock.lock_type = LOCK_NORMAL;	// uninstrumented.
	crq_mplock.lock_stats = 0;		// no stats since uninstrumented

	crq_mode = mode;
	crq_opt = CRQOPT_NULL;	// No extended statistics.
	crq_unitid = unitID;
	crq_master_vect = master_vector;

	crq_totcmds = 0;
	crq_totrsps = 0;
	crq_totattns = 0;
	crq_totvects = 0;

	crq_stats = 0;
}

/*
 * Send a command message on a CRQ
 */
void
CRQ::sendCommand( CRQ_MSG * command )
{
	Debug.printf( "CRQ%T::sendCommand( command:%x )\n", this, command );
	Assert( this != 0 );
	Assert( command != 0 );

	/*
	 * Fill in additional message fields.
	 * We should eventually add code to set the timestamp in the
	 * command to the time of day.
	 */
	command->crq_msg_crq = this;
	command->crq_msg_unitid = crq_unitid;	// experimental
	command->crq_msg_timestamp[0] = 0; 
	command->crq_msg_timestamp[1] = 0;

	lock();

	int commandQueueEmpty = crq_cmd.isEmpty();
	int isInterruptDriven = ( crq_mode == CRQMODE_INTR );
	int intFlag = commandQueueEmpty && isInterruptDriven;
	if ( intFlag ) crq_totvects++;
	crq_cmd.enqueue( command );

	command->setStatus( STS_QUEUED );
	crq_totcmds++;

	if ( intFlag ) {
		TransmitVector( (unsigned int) crq_slave_vect );
		Debug.printf( "CRQ::sendCommand vector %x\n", crq_slave_vect );
	}
	else {
		Debug.printf( "CRQ::sendCommand: NO vector\n" );
		Debug.printf( "CRQ::sendCommand empty/intr = %d/%d\n",
			commandQueueEmpty, isInterruptDriven );
	}
	unlock();
}

/*
 * Like above but P the semaphore of the message after it is sent.
 */
void
CRQ::sendCommandAndAwaitResponse( CRQ_MSG * command )
{
	sendCommand( command );
	Assert( command->semaphore() != 0 );
	command->semaphore()->P();
}

/*
 * Change the Master Vector of a CRQ.
 */
unsigned int
CRQ::changeMasterVector( unsigned int vector )
{
	Debug.printf( "CRQ%T::changeMasterVector( vector:%x )\n",
		this, vector );

	lock();
	unsigned int oldVector = (unsigned int) crq_master_vect;
	crq_master_vect = (long) vector;
	unlock();
	Debug.printf( "CRQ::changeMasterVector: return( %x )\n", oldVector );
	return( oldVector );
}

void
CRQ::handleStandardAttention( CRQ_MSG * attn )
{
	switch( attn->code() ) {
	   case CRQATN_LOCK_TMO:
		Console().cpuPrintf( "CRQ%T lock timed out\n", this );
		break;
	   case CRQATN_VECT_NOT_XMIT:
		Console().cpuPrintf( "CRQ%T vector not sent\n", this );
		break;
	   case CRQATN_VECT_NOT_RCV:
		Console().cpuPrintf( "CRQ%T vector not received\n", this );
		break;
	   case CRQATN_NO_FREE_MSG:
		Console().cpuPrintf( "CRQ%T no free messages left\n", this );
		break;
	   case CRQATN_INVLD_CRQ:
		Console().cpuPrintf( "CRQ%T invalid CRQ\n", this );
		break;
	   default:
		Console().cpuPrintf( "CRQ%T unknown attention message opcode (%d)\n",
			this, attn->code() );
		break;
	}
}

/*
 * Dequeue the next attention message.
 * Returns 0 if there are no attention messages to dequeue.
 */
CRQ_MSG *
CRQ::dequeueAttentionMessage()
{
	CRQ_MSG * attention;

	lock();

	if( crq_attn.isEmpty() ) {
		// the queues empty.
		attention = 0;
	}
	else {
		attention = crq_attn.dequeue();
		attention->setStatus( STS_PENDING );
		crq_totattns++;
		if( crq_opt == CRQOPT_EXTSTATS ) {
			crq_stats->crs_curattn--;
		}
	}

	unlock();
	return( attention );
}

/*
 * Dequeue the next response from the response queue.
 * Returns 0 if there are no responses to dequeue.
 */
CRQ_MSG *
CRQ::dequeueResponse()
{
	CRQ_MSG * response;

	lock();

	if( crq_rsp.isEmpty() ) {
		// the queues empty.
		response = 0;
	}
	else {
		response = crq_rsp.dequeue();
		Assert( response != 0 );
		crq_totrsps++;
		if( crq_opt == CRQOPT_EXTSTATS ) {
			crq_stats->crs_currsp--;
		}
	}
	unlock();
	return( response );
}

void
CRQ::enqueueFreeMessage( CRQ_MSG * msg )
{
	lock();
	msg->setStatus( STS_FREE );
	msg->crq_msg_crq = this;
	crq_free.enqueue( msg );
	if( crq_opt == CRQOPT_EXTSTATS ) {
		if( ++crq_stats->crs_curfree > crq_stats->crs_maxfree ) {
			crq_stats->crs_maxfree++;
		}
	}
	unlock();
}

void
BufferDescriptor::mapTo( PhysicalMemoryChain * chain )
{
	// length should come as an arg since chain might be longer than needed.

	unsigned int length = chain->lengthInBytes();
	Debug.printf( "BufferDescriptor::convertChain( chain:%A length:%d )\n",
		chain, length );

	Assert( this != 0 );
	if( length > 8*1024 ) length = 8192;
	Assert( length <= 8*1024 );	  // at present we can reference 8K.
					  // more can be done but I'm not
					  // worrying about it now.
	buf_desc * desc = descriptors();

	/*
	 * Were currently assuming all PhysicallyAddressableUnits are the
	 * same size and that there are only (at most) 3 in the chain.
	 */
	PhysicallyAddressableUnit * pau = chain->chain();
	Assert( pau != 0 );

	/*
	 * Convert the PAU chain elements to buf_desc's.
	 */
	const char * base = pau->address() + chain->offset();
	int bytes = pau->size() - chain->offset();
	if( bytes >= length ) {
		bytes = length;
	}
	Debug.printf( "BufferDesc::convertChain: bytes:%d length:%d\n",
		bytes, length );

	desc->map( base, bytes );
	desc++;
	length -= bytes;

	if( length == 0 ) {
		/*
		 * Terminate the list.
		 */
		desc->invalidate();
		return;
	}

	pau = pau->nextLink();
	Assert( pau != 0 );

	/*
	 * Do another buf_desc.
	 */
	base = pau->address();
	if( length > pau->size() ) {
		bytes = pau->size();
	}
	else {
		bytes = length;
	}

	desc->map( base, bytes );
	desc++;
	length -= bytes;

	if( length == 0 ) {
		/*
		 * Terminate the list.
		 */
		desc->invalidate();
		return;
	}

	/*
	 * One more time...
	 */
	pau = pau->nextLink();
	Assert( pau != 0 );

	/*
	 * Do another buf_desc.
	 */
	base = pau->address();
	Assert( length < pau->size() );
	bytes = length;

	desc->map( base, bytes );
	desc++;

	/*
	 * Terminate the list.
	 */
	desc->invalidate();
}

void
BufferDescriptor::mapTo( const char * data_paddr, int length )
{
	Assert( this != 0 );
	Assert( data_paddr != 0 );
	Assert( length <= 32*1024*1024 ); // at present we can reference 32M.
					  // more can be done but I'm not
					  // worrying about it now.
	buf_desc * desc = descriptors();

	/*
	 * Were hoping the physical memory isnt going anywhere here!
	 * (I.e. not being swapped out, although that doesnt make any
	 * sense anyway).
	 */

	desc->map( data_paddr, length );
	desc++;

	/*
	 * Terminate the list.
	 */
	desc->invalidate();
}

/*
 * Start the semaphore off at 0 since there are initially no messages
 * in the pool.
 */
CRQ_MSGPool::CRQ_MSGPool()
{
    resourceCountSemaphore = new Semaphore( 0 );
}

CRQ_MSGPool::~CRQ_MSGPool()
{
}

/*
 * This just sets everything to a default. The semaphore and the address
 * of the first buffer descriptor are the only things really known at
 * the time of the constructors call.
 */
MassStoreTransferRequest::MassStoreTransferRequest()
	: EM_MSG( sizeof( MassStoreTransferRequest ), CRQOP_MS_READ, 0,
		new BinarySemaphore( 0 ) )
{
	firstDescriptor = bd.descriptors();

	// we dont use this field for anything.
	sortkey = 0;
}


/*
 * This all assumes the locking is done at a higher level.
 */
DBL_LINK::DBL_LINK()
{
        Assert( this != 0 );
        dbl_fwd = this;
        dbl_bwd = this;
}

SYSHDR::SYSHDR( int size, int type, int version ) : DBL_LINK()
{
        Assert( this != 0 );
        Assert( size > 0 );
        Assert( ( type == TYPE_CRQ ) || ( type == TYPE_CRQMSG ) );
        Assert( version >= 0 );
        Assert( version < 10 );           // Arbitrary.

        hdr_size = size;
        hdr_type = type;
        hdr_version = version;
        hdr_alloc = 0;
}


CRQ_MSG::CRQ_MSG( int size, int opCode, int unitID, Semaphore * sem )
        : SYSHDR( size, TYPE_CRQMSG, CRQMSG_VERSION )
{
        Assert( this != 0 );
        Assert( size >= sizeof( CRQ_MSG ) );

        crq_msg_code = opCode;
        crq_msg_unitid = unitID;
        crq_msg_semaphore = sem;

        // we should set the timestamp someday.
        crq_msg_timestamp[0] = 0L;
        crq_msg_timestamp[1] = 0L;
}

CRQ_MSG *
CRQ_MSGPool::removeMessage()
{
        /*
         * Wait for a message to become available.
         */
        resourceCountSemaphore->P();

        Assert( ! links.isEmpty() );

        /*
         * Grab the message.
         * Lock the links to avoid corruption.
         */
        mutex.acquire();
        CRQ_MSG * msg = links.dequeue();
        mutex.release();

        return( msg );
}

