#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Contributing Author: Paul Petersen
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 *	EMCEthernet.c: ethernet driver code.
 */
#include "Assert.h"
#include "OutputStream.h"
#include "Vectors.h"
#include "CRQ.h"
#include "Semaphore.h"
#include "SemaphoreStar.h"
#include "PhysicalMemoryChain.h"
#include "Lock.h"
#include "Process.h"
#include "Process.h"
#include "ThisProcess.h"
#include "Domain.h"


#ifdef __GNUC__
#pragma implementation
#endif

static const int EMC_SLOT = 4;		// MAKE THIS A PARAMETER TO INIT!!
static const maxEthernetPacketSize = 1514;

/* 
 * EMC logical units for the Ethernet
 */
static const int EMCLUN_EN_CMD  = 61;
static const int EMCLUN_EN_RECVS = 62;
static const int EMCLUN_EN_RECVL = 63;

/*
 * Definition of largest small receive Ethernet packet. EMC controller defines
 * this to decide when to put received frame on the small receive CRQ and
 * when to put it on the large receive CRQ.
 */
static const int LargestSmallPacketSize = 112;

/*
 * Its a subclass of DBL_LINK so we can keep a list of them much like
 * a list of CRQ_MSG's
 */
class EthernetPacketBuffer : public DBL_LINK {
public:
	long size;
	char databuf[ maxEthernetPacketSize ];

	EthernetPacketBuffer();
};

inline
EthernetPacketBuffer::EthernetPacketBuffer()
{
}

static const int num_buffers = 20;
static EthernetPacketBuffer packetBuffers[ num_buffers ];
static Lock freeBufferPoolLock;
static Lock fullBufferPoolLock;
static SemaphoreStar fullBufferCount;
static DBL_LINK freeBufferPool;
static DBL_LINK fullBufferPool;

static CRQ_GENERIC_ATTN_MSG EtherAttentions[ FreeAttentions*3 ]; // per CRQ...

/*
 * EMC Ethernet device states.
 */
enum {
	EN_OFFLINE = 0,		// Offline -- receiving no packets
	EN_ONLINE = 1,		// Online -- receiving packets from network
	EN_INTLPBK = 2,		// Internal loopback
	EN_EXTLPBK = 3,		// External loopback
};

/*
 * Transmit side CRQ_MSG classes.
 */
class CRQ_EN_XMIT_MSG : public EM_MSG { /* TRANSMIT AN ETHERNET PACKET */
public:
	buf_desc * en_xmit_bdesc;
	int     en_xmit_bytes;

	CRQ_EN_XMIT_MSG( int unitID, Semaphore * sem, buf_desc * desc,
		int count );
};

inline
CRQ_EN_XMIT_MSG::CRQ_EN_XMIT_MSG( int unitID, Semaphore * sem,
	buf_desc * desc, int size )
	: EM_MSG( sizeof( CRQ_EN_XMIT_MSG ), CRQOP_EN_XMIT_FRAME, unitID, sem )
{
	en_xmit_bdesc = desc;
	en_xmit_bytes = size;
}

class ENET_XMIT_PKT : public CRQ_EN_XMIT_MSG {
public:
	BufferDescriptor bd;

	ENET_XMIT_PKT();
	ENET_XMIT_PKT( int unitID, buf_desc * desc, int count );
};

ENET_XMIT_PKT::ENET_XMIT_PKT()
	: CRQ_EN_XMIT_MSG( 0, new BinarySemaphore( 0 ), 0, 0 )
{
}

inline
ENET_XMIT_PKT::ENET_XMIT_PKT( int unitID, buf_desc * desc, int count )
	: CRQ_EN_XMIT_MSG( unitID, new BinarySemaphore( 0 ), desc, count )
{
}

/*
 * Recieve side CRQ_MSG classes.
 */
class CRQ_EN_RCV_MSG : public EM_MSG {  /* RECIEVE AN ETHERNET PACKET */
public:
	buf_desc * en_rcv_bdesc;

	CRQ_EN_RCV_MSG( int unitID, buf_desc * desc, Semaphore * sem );
};

inline
CRQ_EN_RCV_MSG::CRQ_EN_RCV_MSG( int unitID, buf_desc * desc, Semaphore * sem )
	: EM_MSG( sizeof( CRQ_EN_XMIT_MSG ), CRQOP_EN_RCV_FRAME, unitID, sem )
{
	en_rcv_bdesc = desc;

	em_status_code = 0;
	em_dev_errcode = 0;
	em_compltn_cnt = 0;
	em_xtnd_status = 0;
}

class ENET_RECV_PKT : public CRQ_EN_RCV_MSG {
public:
	BufferDescriptor bd;
	EthernetPacketBuffer * currentBuffer;

	ENET_RECV_PKT();
	ENET_RECV_PKT( int unitID, buf_desc * desc );
};

ENET_RECV_PKT::ENET_RECV_PKT()
	: CRQ_EN_RCV_MSG( 0, 0, new BinarySemaphore( 0 ) )
{
}

ENET_RECV_PKT::ENET_RECV_PKT( int unitID, buf_desc * desc )
	: CRQ_EN_RCV_MSG( unitID, desc, new BinarySemaphore( 0 ) )
{
}

/*
 * Subcodes for the Ethernet control command (CRQOP_EN_CONTROL).
 */
enum { 
	ENSUB_SET_ONLINE = 1,
	ENSUB_SET_OFFLINE = 2,
	ENSUB_RESET = 3,
	ENSUB_DSB_MC_FILTER = 4,
	ENSUB_SET_INT_LPBACK = 5,
	ENSUB_SET_EXT_LPBACK = 6,
	ENSUB_ENB_RCV_ON_ERR = 7,
	ENSUB_DSB_RCV_ON_ERR = 8,
	ENSUB_ENB_MC_FILTER = 9,
};

/*
 * Control side CRQ_MSG classes
 */
class CRQ_EN_CNTL_MSG : public EM_MSG {
public:
	short   en_cntl_subcode;

	CRQ_EN_CNTL_MSG( int unitid, Semaphore * sem, short subcode );
};

class ENET_CNTL_PKT : public CRQ_EN_CNTL_MSG {
public:

	ENET_CNTL_PKT( int unitid = 0, short subcode = ENSUB_SET_ONLINE );
};

inline
CRQ_EN_CNTL_MSG::CRQ_EN_CNTL_MSG( int unitid, Semaphore * sem, short subcode )
	: EM_MSG( sizeof( CRQ_EN_CNTL_MSG ), CRQOP_EN_CONTROL, unitid, sem )
{
	en_cntl_subcode = subcode;
	em_status_code = 0;
	em_dev_errcode = 0;
	em_compltn_cnt = 0;
	em_xtnd_status = 0;
}

inline
ENET_CNTL_PKT::ENET_CNTL_PKT( int unitid, short subcode )
	: CRQ_EN_CNTL_MSG( unitid, new BinarySemaphore( 0 ), subcode )
{
}

/*
 * Definition of the size of the largest Ethernet transmit/command CRQ
 * message.  This is used to allocate a pool of command messages for the
 * transmit/command Ethernet CRQ which can handle all messages.
 * Be sure to modify this if a longer message is ever added.
 * file!
 */
static const int MAX_EN_MSGSIZE = sizeof( CRQ_EN_XMIT_MSG );
	
/*
 * The transmit packet pool.
 */
static const int num_xmit_pkts = 5;			// # messages in pool
static ENET_XMIT_PKT XmitPoolMsg[ num_xmit_pkts ];	// transmit pool objs.
static Lock XmitPoolLock;				// lock for the pool.
static DBL_LINK XmitPool;				// transmit pool queue
static SemaphoreStar XmitPoolCount;			// alloc semaphore.

/*
 * The receive packet pool.
 */
static const int num_recv_pkts = 4;	// # messages/CRQ in pool
					// (small and large CRQ)
static ENET_RECV_PKT RecvPoolMsg[ num_recv_pkts * 2 ];	// receive pool objs.

/*
 * Use a single control packet since there isn't much call for it.
 */
static ENET_CNTL_PKT ControlMessage;	// the message packet
static SemaphoreStar ControlMutex;	// semaphore to regulate its use

/*
 * Definition of the ETH_STATS structure. This defines various statistics
 * the EMC can provide about the ethernet.
 */
class ETH_STATS {
public:
	short   st_firm_ver;		// Firmware version
	short   st_firm_rev;		// Firmware revision
	unsigned char  st_ethr_addr[6];	// Ethernet address
	short   st_dev_state;		// Current device state
	long    st_xmit_frames;		// Frames transmitted
	long    st_rcv_frames;		// Frames received
	long    st_xmit_bytes;		// Bytes transmitted
	long    st_rcv_bytes;		// Bytes received
	long    st_xmit_nocoll;		// Frames transmitted -- no collisions
	long    st_xmit_onecoll;	// Frames transmitted -- one collision
	long    st_xmit_mulcoll;	// Frames transmitted -- mult collisions
	long    st_xmit_exccoll;	// Xmits failed -- too many collisions
	long    st_xmit_undrnerr;	// Xmits failed -- DMA underrun
	long    st_rcv_crcerr;		// Received frames with CRC error
	long    st_rcv_alignerr;	// Rcvd frames w/ alignment errors
	long    st_rcv_rscerr;		// Rcvd frames lost-- no controller resp
	long    st_rcv_ovrnerr;		// Rcvd frames lost -- ctlr bus overrun
	long    st_rcv_hostbufs;	// Rcvd frames lost -- no host buffers
};

/*
 * The class of CRQ_MSG used to get the above statistics.
 */
class CRQ_EN_RDSTATS_MSG : public EM_MSG {
public:
	ETH_STATS * en_rdstats_bufptr;

	CRQ_EN_RDSTATS_MSG();
	CRQ_EN_RDSTATS_MSG( int unitID, ETH_STATS * stats );
};

inline
CRQ_EN_RDSTATS_MSG::CRQ_EN_RDSTATS_MSG()
{
}

inline
CRQ_EN_RDSTATS_MSG::CRQ_EN_RDSTATS_MSG( int unitID, ETH_STATS * stats )
        : EM_MSG( sizeof( CRQ_EN_RDSTATS_MSG ), CRQOP_EN_READ_STAT, unitID )
{
	en_rdstats_bufptr = stats;

	em_status_code = 0;
	em_dev_errcode = 0;
	em_compltn_cnt = 0;
	em_xtnd_status = 0;
}

// I dont know about this one. It was in the Encore definitions.
// class CRQ_EN_RRSTATS_MSG : public EM_MSG {
// public:
//	ETH_STATS * en_rrstats_bufptr;
// };

class MultimaxEMCEthernet {
protected:
	int slot;
public:
	CRQ controlCRQ;			// A CRQ for controls and Xmits
	CRQ smallCRQ;			// A CRQ for small receives
	CRQ largeCRQ;			// A CRQ for large receives
	ETH_STATS stats;		// A Copy of the unit's statistics

	MultimaxEMCEthernet( int slot );

	void init();
	void createChannels();
};

static MultimaxEMCEthernet theEthernet( EMC_SLOT );


MultimaxEMCEthernet::MultimaxEMCEthernet( int slot )
	: largeCRQ( UnitID( 0, slot, 0, EMCLUN_EN_RECVL ), ClassVector(
		ETHERL_Vector, CPUCLASS_RUNNING ), CRQMODE_INTR ),
	  smallCRQ( UnitID( 0, slot, 0, EMCLUN_EN_RECVS ), ClassVector(
		ETHERS_Vector, CPUCLASS_RUNNING ), CRQMODE_INTR ),
	  controlCRQ( UnitID( 0, slot, 0, EMCLUN_EN_CMD ), ClassVector(
		ETHER_Vector, CPUCLASS_RUNNING ), CRQMODE_INTR )

{
	this->slot = slot;

	XmitPoolCount = new Semaphore( num_xmit_pkts );
	ControlMutex = new Semaphore( 1 );
	fullBufferCount = new Semaphore( 0 );
}

extern unsigned int ClassVector( unsigned int, unsigned int );

static CreateChannelRequest CreateChannelMessage;
static CRQ_EN_RDSTATS_MSG ReadStatsMessage;	// Read Statistics Message


/*
 * User interface functions.
 */
void InitEther();
void EtherAddr( char * addr );
int  EtherXmit( char * data, unsigned int size );
int  EtherRecv( char * & data );
void EtherControl( unsigned int subcode );

/*
 * User interface routine to read the Ethernet Address from the
 * return information of the STATISTICS message.  Could also be
 * read directly at some magic address
 */
void
ethernetAddress( unsigned char * addr )
{
EtherAddr((char *) addr);
}

void
EtherAddr( char * addr )
{
	Debug.printf( "EtherAddr( addr:%A )\n", addr );
	Assert( addr != 0 );

	for( int i = 0; i < 6; i++ ) {
		addr[i] = theEthernet.stats.st_ethr_addr[i];
	}
}

/*
 * Return a received Ethernet packet buffer.
 */
int
EtherRecv( char * & usrbuf )
{
	Debug.printf( "Entering EtherRecv: %x\n", usrbuf );
	Assert( usrbuf == 0 );

	/*
	 * Get a received packet from the fullBufferPool.
	 * We loop because the Semaphore might be Ved more than necessary.
	 */
	for ( ; ; ) {
		fullBufferCount->P();
		Assert( ! fullBufferPoolLock.heldByMe() );
		fullBufferPoolLock.acquire();
		if ( ! fullBufferPool.isEmpty() )
			break;
		Assert( fullBufferPoolLock.heldByMe() );
		fullBufferPoolLock.release();
	}
	Assert( fullBufferPoolLock.heldByMe() );
	EthernetPacketBuffer * bufr =
		(EthernetPacketBuffer *) fullBufferPool.dequeue();
	Assert( bufr != 0 );
	Assert( bufr != &fullBufferPool );
	fullBufferPoolLock.release();

	/*
	 * Return the size and buffer.
	 * The receiver "promises" to do an EtherRelease( buffer ) ASAP.
	 */
	usrbuf = bufr->databuf;
	int size = (int) bufr->size;
	Assert( size > 0 );
	return size;
}

void
EtherRelease( char * & usrbuf )
{
	Assert( usrbuf != 0 );

	/*
	 * Determine the EthernetPacketBuffer.
	 * We are trusting the caller to only give us back something
	 * we gave to it via EtherRecv().
	 * The bufr->size checks are a weak attempt to verify this.
	 */
	EthernetPacketBuffer * bufr = // FIX: GAG ME WITH A PAITCH FORK
		(EthernetPacketBuffer *) (usrbuf-sizeof(long)-sizeof(DBL_LINK));
	usrbuf = 0;
	if( ( bufr->size <= 0 ) || ( bufr->size > maxEthernetPacketSize ) ) {
		Console() << "bufr " << asHex( (int)bufr ) << " size "
				<< bufr->size << "\n" << eor;
	}
	Assert( bufr->size > 0 );
	Assert( bufr->size <= maxEthernetPacketSize );

	/*
	 * Return the buffer to the free pool.
	 */
	freeBufferPoolLock.acquire();
	freeBufferPool.enqueue( (CRQ_MSG *) bufr );
	freeBufferPoolLock.release();
}

/*
 * User interface routine to write data onto the Ethernet
 * The size parameter determines the number of bytes to write
 *
 *  NOTE: this function always return the number of bytes written
 *        as soon as the data is moved to EMC memory.
 */

int
EtherXmit( char * data, unsigned int size )
{
	CRQ * crq = &theEthernet.controlCRQ;

	Debug.printf( "Entering EtherXmit: %x, (%d)\n", data, size );
	Assert( data != 0 );

	if( size == 0 ) return( 0 );	// Check for the NULL transfer
	Assert( size <= maxEthernetPacketSize );

	/*
	 * Get a message from the transmit message pool.
	 */
	XmitPoolCount->P();
	XmitPoolLock.acquire();
	ENET_XMIT_PKT * msg = (ENET_XMIT_PKT *) XmitPool.dequeue();
	XmitPoolLock.release();
	Assert( msg != 0 );

	Debug.printf("EtherXmit: about to build chain...\n");
	PhysicalMemoryChain chain = thisProcess()->domain()->
		constructChain( data, size );
	Debug.printf("EtherXmit: just built chain...\n");

	msg->bd.mapTo( &chain );

	msg = new ( msg ) ENET_XMIT_PKT( crq->crq_unitid,
		msg->bd.descriptors(), size );

	crq->sendCommand( msg );
	msg->semaphore()->P();
	Debug.printf( "EtherXmit: packet sent?\n" );

	XmitPoolLock.acquire();
	XmitPool.enqueue( msg );
	XmitPoolLock.release();
	XmitPoolCount->V();
	Debug.printf( "Exiting EtherXmit: %x (%d)\n", data, size );
	return( size );
}

/*
 * Control the state of the Ethernet.  The possible calls are:
 *
 * EtherControl( ENSUB_SET_ONLINE );
 * EtherControl( ENSUB_SET_OFFLINE);
 * EtherControl( ENSUB_RESET );
 * EtherControl( ENSUB_SET_INT_LPBACK );
 * EtherControl( ENSUB_SET_EXT_LPBACK );
 * EtherControl( ENSUB_ENB_RCV_ON_ERR );
 * EtherControl( ENSUB_DSB_RCV_ON_ERR );
 * EtherControl( ENSUB_ENB_MC_FILTER );
 * EtherControl( ENSUB_DSB_MC_FILTER );
 */
void
EtherControl( unsigned int subcode )
{
	CRQ * crq = &theEthernet.controlCRQ;

	/*
	 * Claim exclusive use of the control message
	 */
	ControlMutex->P();

	ENET_CNTL_PKT * msg = new ( &ControlMessage ) ENET_CNTL_PKT(
		crq->crq_unitid, subcode );

	Debug.printf( "EtherControl: msg %x (%d)\n", msg, subcode );

	/*
	 * Send the message and wait for a reply.
	 */
	crq->sendCommand( msg );
	msg->semaphore()->P();
	Assert( msg->status() == STS_SUCCESS );

	/*
	 * Release the control message.
	 */
	ControlMutex->V();
}

void
InitEther()
{
	theEthernet.init();
}

/*
 *  Main entry point when the Ethernet device is created.  This routine
 *  queues the data onto the free pools, and created the CRQ's for transmit
 *  small receive and large receive channels
 */
void
MultimaxEMCEthernet::init()
{
	Debug.printf( "MultimaxEMCEthernet::init()\n" );

	/*
 	 * Initilize and create the channels
	 */
	Debug.printf( "InitEther: creating channels to devices\n" );
	createChannels();
	Debug.printf( "InitEther: channels created.\n" );

	/*
	 * Install "FreeAttentions" worth of free attention messages
	 * on each CRQ.
	 */
	CRQ_GENERIC_ATTN_MSG * freeAttn = &EtherAttentions[0];
	for( int i = 0; i < FreeAttentions; i++ ) {
		controlCRQ.enqueueFreeMessage( freeAttn );
		freeAttn++;
	}
	for( i = 0; i < FreeAttentions; i++ ) {
		smallCRQ.enqueueFreeMessage( freeAttn );
		freeAttn++;
	}
	for( i = 0; i < FreeAttentions; i++ ) {
		largeCRQ.enqueueFreeMessage( freeAttn );
		freeAttn++;
	}
	Debug.printf( "InitEther: free attns enqueued\n" );

	/*
	 * Stick all the free Xmit messages in the XmitPool.
	 */
	for( i = 0; i < num_xmit_pkts; i++ ) {
		XmitPool.enqueue( &XmitPoolMsg[i] );
	}
	Debug.printf( "InitEther: free trnasmit packets enqueued\n" );

	/*
	 * Stick all the receive buffers in the free buffer pool.
	 */
	for( i = 0; i < num_buffers; i++ ) {
		freeBufferPool.enqueue( (CRQ_MSG *) &packetBuffers[i] );
	}
	Debug.printf( "Free receive buffers enqueued\n" );

	/*
	 * Now install some Receive commands onto the Large & Small Channels.
	 */
	freeBufferPoolLock.acquire();
	ENET_RECV_PKT * recv = &RecvPoolMsg[ 0 ];
	CRQ * crq = &smallCRQ;
	for ( i = 0; i < num_recv_pkts * 2; i++, recv++ ) {
		if( i == num_recv_pkts ) crq = &largeCRQ;
		int unitid = crq->crq_unitid;

		/*
		 * Dequeue a buffer for this message.
		 */
		EthernetPacketBuffer * buf = (EthernetPacketBuffer *)
			freeBufferPool.dequeue();
		Assert( buf != &freeBufferPool );

		/*
		 * Initialize the message.
		 */
		recv->currentBuffer = buf;

		// if kernel memory is every paged there could be a problem
		// here since when the chain goes out of scope the buffer
		// memory is a candidate for paging out.
		PhysicalMemoryChain chain = thisProcess()->domain()->
			constructChain( &buf->databuf[0],
			maxEthernetPacketSize );
		
		recv->bd.mapTo( &chain );

		recv = new ( recv ) ENET_RECV_PKT( unitid,
			recv->bd.descriptors() );

		/*
		 * Send the command to the CRQ.
		 */
		crq->sendCommand( recv );
	}
	freeBufferPoolLock.release();

	Debug.printf( "InitEther: recv requests sent\n" );

	CRQ_EN_RDSTATS_MSG * msg = &ReadStatsMessage;
	Debug.printf( "sendReadStatsMessage: msg %x\n", msg );
	crq = &theEthernet.controlCRQ;
	Assert( crq != 0 );

	msg = new ( msg ) CRQ_EN_RDSTATS_MSG( crq->crq_unitid,
		&theEthernet.stats );

	crq->sendCommand( msg );

// REENABLE IT AFTER I FIX IT
	EtherControl( ENSUB_SET_ONLINE );
	Debug.printf( "InitEther: done.\n");
}

/*
 *  Create a CRQ out of the given CRQ structure.
 */
void
MultimaxEMCEthernet::createChannels()
{
	CRQ * EMCCRQ = RequestorCRQ( slot, REQ_LUN );
	CreateChannelRequest * msg = &CreateChannelMessage;

	Debug.printf( "MultimaxEMCEthernet::createChannels()\n" );
	Assert( this != 0 );

	/*
	 * Send a Create channel message on the EMC CRQ for each ethernet
	 * CRQ (control/small/large) and wait for a repsonse.
	 */
	CRQ * crq;
	for( int i = 0; i < 3; i++ ) {
		switch( i ) {
		   case 0:
			crq = &controlCRQ;
			break;
		   case 1:
			crq = &smallCRQ;
			break;
		   case 2:
			crq = &largeCRQ;
			break;
		   default:
			crq = 0;
			Assert( NOTREACHED );
		}

		msg->setup( crq );
		EMCCRQ->sendCommand( msg );
		msg->semaphore()->P();

		/*
		 * Verify succesful channel creation
		 */
		Debug.printf( "InitEther: message/channel status %x/%x\n",
			msg->status(), msg->resultStatus() );

		Assert( msg->status() == STS_SUCCESS );
	}
}

/*
 *  Interrupt Service Routine used to handle interrupts from
 *  the Control/Transmit channel
 */
void
EtherControlInterruptHandler()
{
	CRQ_MSG * rsp;
	CRQ_MSG * attn;
	CRQ * crq = &theEthernet.controlCRQ;

	int didWork = 0;

	Debug.printf( "Ether: Control START\n" );
	do {
		/*
		 * If there is an attention message waiting, handle it.
		 */
		if( ( attn = crq->dequeueAttentionMessage() ) != 0 ) {
			switch( attn->code() ) {
			   default:
				crq->handleStandardAttention( attn );
				break;
			}
			crq->enqueueFreeMessage( attn );
			didWork = 1;
		}

		/*
		 * If there is a response message waiting, handle it.
		 */
		if( ( rsp = crq->dequeueResponse() ) != 0 ) {
			switch( rsp->code() ) {

			   case CRQOP_EN_READ_STAT:
				Debug.printf( "Ether: Transmit/Control Statistics\n" );
#ifdef OLDDEBUG
				ETH_STATS = ((CRQ_EN_RDSTATS_MSG *) rsp )->
					en_rdstats_bufptr;
				Printf( "Ethernet Statistics\n" );
				Printf( "Firmware version  > %d\n",
					stats->st_firm_ver );
				Printf( "Firmware revision > %d\n",
					stats->st_firm_rev );
				Printf( "Ethernet Address  > %02x",
					stats->st_ethr_addr[0] );
				for (int i = 1; i < 6; i++) {
				   Printf(":%02x", stats->st_ethr_addr[i] );
				}
				Printf("\n" );
				Printf( "Device State      > %d\n",
					stats->st_dev_state   );
				Printf( "Frames Transmitted> %d\n",
					stats->st_xmit_frames );
				Printf( "Frames Received   > %d\n",
					stats->st_rcv_frames );
				Printf( "Bytes  Transmitted> %d\n",
					stats->st_xmit_bytes  );
				Printf( "Bytes  Received   > %d\n",
					stats->st_rcv_bytes  );
				Printf( "Frames Xmit 0 Col > %d\n",
					stats->st_xmit_nocoll );
				Printf( "Frames Xmit 1 Col > %d\n",
					stats->st_xmit_onecoll );
				Printf( "Frames Xmit N Col > %d\n",
					stats->st_xmit_mulcoll );
				Printf( "Frames Xmit X Col > %d\n",
					stats->st_xmit_exccoll );
				Printf( "Frames Xmit DMA   > %d\n",
					stats->st_xmit_undrnerr);
				Printf( "Frames Recv CRC   > %d\n",
					stats->st_rcv_crcerr  );
				Printf( "Frames Recv ALIGN > %d\n",
					stats->st_rcv_alignerr);
				Printf( "Frames Recv RESP  > %d\n",
					stats->st_rcv_rscerr  );
				Printf( "Frames Recv OVR   > %d\n",
					stats->st_rcv_ovrnerr );
				Printf( "Frames Recv BUFF  > %d\n",
					stats->st_rcv_hostbufs);
				Printf( "\n\n" );
#endif
				break;

			   case CRQOP_EN_CONTROL:
			   case CRQOP_EN_XMIT_FRAME:
				Debug.printf( "EtherControl: Transmit or Control\n" );
				Assert( rsp->semaphore() != 0 );
				rsp->semaphore()->V();
				break;

			   case CRQOP_EN_RCV_FRAME:
				Console().cpuPrintf( "Receive frame on CONTROL channel\n" );
				break;

			   default:
				Console().cpuPrintf( "EtherControl: UNKNOWN code (%d)\n",
					rsp->code() );
				break;
			}
			didWork = 1;
		}
	} while( ( rsp != 0 ) || ( attn != 0 ) );

	if( ! didWork )
		Debug.printf( "Ether: Transmit/Control NO WORK?\n" );
	Debug.printf( "Ether: Transmit/Control done.\n" );
}

/*
 *  Interrupt Service Routine used to handle interrupts from
 *  the Receive channels
 */
void
EtherRecvInterruptHandler( CRQ * crq )
{
	CRQ_MSG * rsp;
	CRQ_MSG * attn;

	int didWork = 0;

	Debug.printf( "EtherRecv: START\n" );
	do {
		/*
		 * If there is an attention message waiting, handle it.
		 */
		if( ( attn = crq->dequeueAttentionMessage() ) != 0 ) {
			switch( attn->code() ) {
			   default:
				crq->handleStandardAttention( attn );
				break;
			}
			crq->enqueueFreeMessage( attn );
			didWork = 1;
		}

		/*
		 * If there is a response message waiting, handle it.
		 */
		if( ( rsp = crq->dequeueResponse() ) != 0 ) {
			switch( rsp->code() ) {

			   case CRQOP_EN_READ_STAT:
				Console().cpuPrintf( "CRQOP_EN_READ_STAT on receive channel???\n" );
				break;

			   case CRQOP_EN_XMIT_FRAME:
				Console().cpuPrintf( "CRQOP_EN_XMIT_FRAME on receive channel???\n" );
				break;

			   case CRQOP_EN_RCV_FRAME:
			   {
				ENET_RECV_PKT * recv = (ENET_RECV_PKT *) rsp;

				/*
				 * If there was an error, report it.
				 * If the received packet ends up on the
				 * fullBufferPool, its size is zero to indicate
				 * the error to the receiver.
				 * As you might have noticed, we don't deal with
				 * device errors very well yet.
				 */
				if ( recv->status() == STS_ERROR ) {
					Console().printf( "EtherRecv: STS_ERROR status_code = %d, dev_errcode = %d, compltn_cnt = %d, xtnd_status = %d\n",
					recv->em_status_code,
					recv->em_dev_errcode,
					recv->em_compltn_cnt,
					recv->em_xtnd_status );
					Assert( recv->em_compltn_cnt == 0 );
				}
				recv->currentBuffer->size =
					recv->em_compltn_cnt;

				/*
				 * Separate the buffer from the message.
				 */
				EthernetPacketBuffer * oldbuf =
					recv->currentBuffer;
				Assert( oldbuf != 0 );
				recv->currentBuffer = 0;

				/*
				 * Get a buffer from the freeBufferPool,
				 * if there is one.
				 */
				EthernetPacketBuffer * newbuf = 0;
				int newFull = 1;
				freeBufferPoolLock.acquire();
				if ( ! freeBufferPool.isEmpty() ) {
					newbuf = (EthernetPacketBuffer *)
						 freeBufferPool.dequeue();
					Assert( newbuf != 0 );
					Assert( newbuf != &freeBufferPool );
				}
				freeBufferPoolLock.release();

				/*
				 * If the freeBufferPool was empty,
				 * recycle a buffer from the fullBufferPool,
				 * if there is one.
				 */
				if ( newbuf == 0 ) {
					fullBufferPoolLock.acquire();
					if ( ! fullBufferPool.isEmpty() ) {
					    newbuf = (EthernetPacketBuffer *)
						     fullBufferPool.dequeue();
					    Assert( newbuf != 0 );
					    Assert( newbuf != &fullBufferPool );
					    newFull = 0;
					}
					fullBufferPoolLock.release();
				}

				/*
				 * If we got a new buffer,
				 * enqueue the received buffer on the
				 * fullBufferPool.
				 * Otherwise, we'll reuse the received buffer.
				 */
				if ( newbuf != 0 ) {
					fullBufferPoolLock.acquire();
					fullBufferPool.enqueue( (CRQ_MSG *)
								oldbuf );
					fullBufferPoolLock.release();
					if ( newFull )
						fullBufferCount->V();
				} else {
					newbuf = oldbuf;
				}

				// SEE COMMENT ABOVE ABOUT THE CHAIN
				Assert( newbuf != 0 );
				PhysicalMemoryChain chain = thisProcess()->
					domain()->constructChain(
					&newbuf->databuf[0],
					maxEthernetPacketSize );

				recv->bd.mapTo( &chain );
				recv->currentBuffer = newbuf;
				recv->em_status_code = 0;
				recv->em_dev_errcode = 0;
				recv->em_compltn_cnt = 0;
				recv->em_xtnd_status = 0;
				recv->en_rcv_bdesc = recv->bd.descriptors();
	
				crq->sendCommand( recv );
			        break;
			   }
			   default:
				Console().cpuPrintf( "EtherRecv: UNKNOWN code (%d).\n",
					rsp->code() );
				break;
			}
			didWork = 1;
		}
	} while( ( rsp != 0 ) || ( attn != 0 ) );

	if ( ! didWork )
		Debug.printf( "EtherRecv: NO WORK?\n" );
	Debug.printf( "EtherRecv: return.\n" );
}

void
EtherLargeInterruptHandler()
{
	EtherRecvInterruptHandler( &theEthernet.largeCRQ );
}

void
EtherSmallInterruptHandler()
{
	EtherRecvInterruptHandler( &theEthernet.smallCRQ );
}
