#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * MultimaxAPCCPU.cc: Multimax NS32332 Processors
 */

#include "Assert.h"
#include "OutputStream.h"
#include "MultimaxAPCCPU.h"
#include "NS32332Interrupts.h"
#include "NS32202.h"
#include "TimeSliceInterrupt.h"
#include "ThisCPU.h"
#include "MultimaxInterrupts.h"
#include "TimerManager.h"
#include "Vectors.h"
#include "PROCESSOR.h"
#ifdef INSTRUMENT
#include "SystemInstrumentManager.h"
#endif


#ifdef __GNUC__
#pragma implementation
#endif

/*
 * All the initial Interrupt handlers.
 */
StrayInterrupt PowerFailure( "Power Failure" );
StrayInterrupt SystemNMI( "System NMI" );
NanoBusInterrupt HardNanoBusError0( "Hard NanoBus error 0" );
NanoBusInterrupt HardNanoBusError1( "Hard NanoBus error 1" );
NanoBusInterrupt HardNanoBusError2( "Hard NanoBus error 2" );
NanoBusInterrupt HardNanoBusError3( "Hard NanoBus error 3" );
StrayInterrupt DestselParityError( "Destsel parity error" );
StrayInterrupt MasterOPEN( "Master ICU IR7" );
NanoBusInterrupt SoftNanoBusError0( "Soft NanoBus error 0" );
VectorBusInterrupt VectorBusInterruptInterrupt;
StrayInterrupt MasterFree1( "Master ICU IR10" );
StrayInterrupt MasterFree3( "Master ICU IR13" );
StrayInterrupt MasterFree4( "Master ICU IR14" );
StrayInterrupt MasterFree5( "Master ICU IR15" );
StrayInterrupt SlaveICUInterrupt( "slave ICU" );
AddressTranslatorFlushInterrupt MMUFlushInterrupt;
TimerManagerInterrupt TimerManagerInterruptInterrupt;
TimeSliceInterrupt TSliceInterrupt;
#ifdef INSTRUMENT
InstrumentBufferInterrupt TheInstrumentBufferInterrupt;
#endif

MultimaxAPCCPU::MultimaxAPCCPU( unsigned long identifier,
	ProcessContainer * initialIdleContainer )
        : NS32332CPU( identifier, initialIdleContainer ),
	  myTimer( TIMESLICE_Vector )
{
        Debug.printf( "MultimaxAPCCPU::ctor()\n" );

        _timesliceTimer = &myTimer;

	/*
	 * Set the name to the <slot,Number> format for printing.
	 */
	int slot = _uniqueIdentifier >> 2;
	int cpuid = _uniqueIdentifier & 0x3;
	char cpuidName = "ABCD"[ cpuid ];
	if( slot < 10 ) {
		name[0] = "0123456789"[slot];
		name[1] = cpuidName;
		name[2] = 0;
	} else {
		Assert( slot < 20 );
		slot -= 10;
		name[0] = '1';
		name[1] = "0123456789"[slot];
		name[2] = cpuidName;
		name[3] = 0;
	}
}

MultimaxAPCCPU::~MultimaxAPCCPU()
{
	Assert( NOTREACHED );
}

/*
 * APC Control Register bits/values.
 */
static unsigned volatile int * APCControlAddress = (unsigned int *) 0xffffff54;

static const unsigned int APCControlNormal  = 0x00000000;
static const unsigned int APCControlCacheOff = 0x00000100;
static const unsigned int APCControl = APCControlNormal;


void
MultimaxAPCCPU::chipInitialize()
{
	/*
	 * Make sure were who we think we are.
	 */
	Assert( _uniqueIdentifier == PROCESSOR::cpuID() );
	Assert( this == thisCPU() );

	/*
	 * Set various bits in the APC control registers.
	 */
	Debug << this << "::chipInitialize(): writing APCControl with " <<
		asHex( APCControl ) << "\n" << eor;
	*APCControlAddress = APCControl;
	Debug << this << "::chipInitialize(): wrote APCControl\n" << eor;

	/*
	 * Read the error status register to clear it.
	 */
	readESR();

	Debug << this << "::chipInitialize(): initializing ICUs\n" << eor;
	initICUs();

	/*
	 * Set our CPU vector class here because when we dispatch later on,
	 * interrupts could get turned on, and we must set our vector
	 * class before that happens.
	 */

	SetCPUVectorClass( CPUCLASS_RUNNING );

	/*
	 * Do the superclass stuff last.
	 */
	NS32332CPU::chipInitialize();
}

unsigned int
MultimaxAPCCPU::readESR()
{
	static const unsigned volatile int * const ErrorStatusRegister =
		(unsigned int *) 0xffffff44;
	return( *ErrorStatusRegister );
}

void
MultimaxAPCCPU::initICUs()
{
	Debug.printf( "MultimaxAPCCPU::initICUs()\n" );

	/*
	 * Read the ICU's and see what the register contents are.
	Debug.printf( "InitICUs: SVCT: Master %x Slave %x\n",
		*MasterSVCT, *SlaveSVCT );
	Debug.printf( "InitICUs: ELTG: Master %x/%x Slave %x/%x\n",
		*MasterELTGL, *MasterELTGH, *SlaveELTGL, *SlaveELTGH );
	Debug.printf( "InitICUs: TPL: Master %x/%x Slave %x/%x\n",
		*MasterTPLL, *MasterTPLH, *SlaveTPLL, *SlaveTPLH );
	Debug.printf( "InitICUs: IPND: Master %x/%x Slave %x/%x\n",
		*MasterIPNDL, *MasterIPNDH, *SlaveIPNDL, *SlaveIPNDH );
	Debug.printf( "InitICUs: ISRV: Master %x/%x Slave %x/%x\n",
		*MasterISRVL, *MasterISRVH, *SlaveISRVL, *SlaveISRVH );
	Debug.printf( "InitICUs: IMSK: Master %x/%x Slave %x/%x\n",
		*MasterIMSKL, *MasterIMSKH, *SlaveIMSKL, *SlaveIMSKH );
	Debug.printf( "InitICUs: CSRC: Master %x/%x Slave %x/%x\n",
		*MasterCSRCL, *MasterCSRCH, *SlaveCSRCL, *SlaveCSRCH );
	Debug.printf( "InitICUs: FPRT: Master %x/%x Slave %x/%x\n",
		*MasterFPRTL, *MasterFPRTH, *SlaveFPRTL, *SlaveFPRTH );
	Debug.printf( "InitICUs: MCTL: Master %x Slave %x\n",
		*MasterMCTL, *SlaveMCTL );
	Debug.printf( "InitICUs: OCASN: Master %x Slave %x\n",
		*MasterOCASN, *SlaveOCASN );
	Debug.printf( "InitICUs: CIPTR: Master %x Slave %x\n",
		*MasterCIPTR, *SlaveCIPTR );
	Debug.printf( "InitICUs: PDAT: Master %x Slave %x\n",
		*MasterPDAT, *SlavePDAT );
	Debug.printf( "InitICUs: IPS: Master %x Slave %x\n",
		*MasterIPS, *SlaveIPS );
	Debug.printf( "InitICUs: PDIR: Master %x Slave %x\n",
		*MasterPDIR, *SlavePDIR );
	Debug.printf( "InitICUs: CCTL: Master %x Slave %x\n",
		*MasterCCTL, *SlaveCCTL );
	Debug.printf( "InitICUs: CICTL: Master %x Slave %x\n",
		*MasterCICTL, *SlaveCICTL );
	Debug.printf( "InitICUs: LCSV: Master %x/%x Slave %x/%x\n",
		*MasterLCSVL, *MasterLCSVH, *SlaveLCSVL, *SlaveLCSVH );
	Debug.printf( "InitICUs: HCSV: Master %x/%x Slave %x/%x\n",
		*MasterHCSVL, *MasterHCSVH, *SlaveHCSVL, *SlaveHCSVH );
	Debug.printf( "InitICUs: LCCV: Master %x/%x Slave %x/%x\n",
		*MasterLCCVL, *MasterLCCVH, *SlaveLCCVL, *SlaveLCCVH );
	Debug.printf( "InitICUs: HCCV: Master %x/%x Slave %x/%x\n",
		*MasterHCCVL, *MasterHCCVH, *SlaveHCCVL, *SlaveHCCVH );
	 */

	/*
	 * Initialize master.
	 */
	*MasterMCTL = 0xca;	// CFRZ | COUTD | FRZ | NTAR.
	*MasterCCTL = 0x80;	// CCON.
	*MasterCIPTR = 0xbb;	// Counter is vector 11 on master.
	*MasterSVCT = 0x10;	// Master if vectors 0x10-0x1f.

	*MasterHCSVH = 0x00;	// Init. starting values.
	*MasterHCSVL = 0x7f;
	*MasterLCSVH = 0x00;
	*MasterLCSVL = 0x00;

	*MasterHCCVH = 0x00;	// Init. current values.
	*MasterHCCVL = 0x7f;
	*MasterLCCVH = 0x00;
	*MasterLCCVL = 0x00;

	*MasterCICTL = 0x11;	// WENH | WENL (clear).
	*MasterCICTL = 0x30;	// WENH | CIEH (enable H+L).

	*MasterIPS = 0xff;	// All are interrupt sources.
	*MasterPDIR = 0xff;	// All are inputs.
	*MasterOCASN = 0;	// No counter interrupt output.
	*MasterPDAT = 0;	// No port data.

	*MasterCSRCH = 0x01;	// Cascade to slave is IR8 on master.
	*MasterCSRCL = 0x00;

	*MasterELTGL = 0xff;	// Clear pending/in-service interrupts.
	*MasterELTGH = 0xff;
	*MasterIPNDL = 0x40;
	*MasterIPNDH = 0x40;
	*MasterISRVL = 0x00;
	*MasterISRVH = 0x00;

	*MasterTPLH = 0x10;	// Triggering polarity is 0x1040.
	*MasterTPLL = 0x40;
	*MasterELTGH = 0x19;	// Edge/level triggering is 0x1900.
	*MasterELTGL = 0x00;

	*MasterFPRTH = 0x00;	// First priority 0.

	*MasterIMSKH = 0xe4;	// Set mask.
	*MasterIMSKL = 0x80;

	*MasterMCTL = 0x02;	// NTAR.
	
	/*
	 * Initialize slave.
	 */
	*SlaveMCTL = 0xca;	// CFRZ | COUTD | FRZ | NTAR.
	*SlaveCCTL = 0x80;	// CCON.
	*SlaveCIPTR = 0xee;	// Counter is vector 14 on slave.
	*SlaveSVCT = 0x20;	// Slave is vectors 0x20-0x2f.

	*SlaveHCSVH = 0x00;	// Init. starting values.
	*SlaveHCSVL = 0x7f;
	*SlaveLCSVH = 0x00;
	*SlaveLCSVL = 0x00;

	*SlaveHCCVH = 0x00;	// Init. current values.
	*SlaveHCCVL = 0x7f;
	*SlaveLCCVH = 0x00;
	*SlaveLCCVL = 0x00;

	*SlaveCICTL = 0x11;	// WENH | WENL (clear).
	*SlaveCICTL = 0x30;	// WENH | CIEH (enable H+L).

	*SlaveIPS = 0xff;	// All are interrupt sources.
	*SlavePDIR = 0xff;	// All are inputs.
	*SlaveOCASN = 0;	// No counter interrupt output.
	*SlavePDAT = 0;	// No port data.

	*SlaveCSRCH = 0x00;	// No cascades into slave.
	*SlaveCSRCL = 0x00;

	*SlaveELTGL = 0xff;	// Clear pending/in-service interrupts.
	*SlaveELTGH = 0xff;
	*SlaveIPNDL = 0x40;
	*SlaveIPNDH = 0x40;
	*SlaveISRVL = 0x00;
	*SlaveISRVH = 0x00;

	*SlaveTPLH = 0x00;	// Triggering polarity is 0x0000.
	*SlaveTPLL = 0x00;
	*SlaveELTGH = 0x60;	// Edge/level triggering is 0x6000.
	*SlaveELTGL = 0x00;

	*SlaveFPRTH = 0x00;	// First priority 0.

	*SlaveIMSKH = 0xa0;	// Set mask.
	*SlaveIMSKL = 0x00;

	*SlaveMCTL = 0x02;	// NTAR.

	Debug.printf( "MultimaxAPCCPU::initICUs: initialization complete\n" );
}

void
MultimaxAPCCPU::installInterrupts()
{
	Assert( this != 0 );

	/*
	 * First start with the stuff thats processor (not machine)
	 * dependent.
	 */
	NS32332CPU::installInterrupts();

	/*
	 * Install the global (common amoung all CPUs executing this kernel)
	 * interrupt handlers. The ABT handler is especially important since
	 * it will allow page faults to occur and be handled properly.
	 */
	setInterrupt( POWERFAIL_Vector, &PowerFailure );	
	setInterrupt( SYSTEMNMI_Vector, &SystemNMI );
	setInterrupt( HARDNANOBUSERROR0_Vector, &HardNanoBusError0 );
	setInterrupt( HARDNANOBUSERROR1_Vector, &HardNanoBusError1 );
	setInterrupt( HARDNANOBUSERROR2_Vector, &HardNanoBusError2 );
	setInterrupt( HARDNANOBUSERROR3_Vector, &HardNanoBusError3 );
	setInterrupt( DESTSELPARITYERROR_Vector, &DestselParityError );
	setInterrupt( OPEN_Vector, &MasterOPEN );
	setInterrupt( SOFTNANOBUSERROR0_Vector, &SoftNanoBusError0 );
	setInterrupt( MASTERFREE1_Vector, &MasterFree1 );
	setInterrupt( VECTORBUSINTERRUPT_Vector, &VectorBusInterruptInterrupt );
	setInterrupt( MASTERFREE3_Vector, &MasterFree3 );
	setInterrupt( MASTERFREE4_Vector, &MasterFree4 );
	setInterrupt( MASTERFREE5_Vector, &MasterFree5 );
	setInterrupt( MMU_Vector, &MMUFlushInterrupt );
	setInterrupt( INTERVAL_Vector, &TimerManagerInterruptInterrupt );
	setInterrupt( TIMESLICE_Vector, &TSliceInterrupt );
#ifdef INSTRUMENT
	setInterrupt (INSTRUMENT_Vector, &TheInstrumentBufferInterrupt);
#endif

	for( int vector = 0x20; vector < 0x30; vector++ ) {
		setInterrupt( vector, &SlaveICUInterrupt );
	}
	Assert( ! PROCESSOR::interruptsEnabled() );
}

void
MultimaxAPCCPU::sendFlushRequestTo( unsigned long cpuid )
{
	TransmitDirectedVector( MMU_Vector, cpuid );
}

