#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * EMCDisk.c: Multimax EMC Disk driver.
 */

#include "Assert.h"
#include "OutputStream.h"
#include "Raid.h"
#include "Class.h"
#include "MultimaxEMCDisk.h"
#include "CRQ.h"
#include "Vectors.h"
#include "Semaphore.h"
#include "FileSystem.h"
#include "PhysicallyAddressableUnit.h"
#include "PhysicalMemoryChain.h"
#include "MultimaxInterrupts.h"
#include "CPU.h"
#include "ThisCPU.h"
#include "Process.h"
#include "ThisProcess.h"
#include "Domain.h"
#ifdef INSTRUMENT
#include "SystemInstrumentManager.h"
extern SystemInstrumentManager *TheInstrumentManager;
#endif


#ifdef __GNUC__
#pragma implementation
#endif

#ifdef INSTRUMENT_IO
#define Instr(exp) exp
#else
#define Instr(exp)
#endif  //  INSTRUMENT_IO


/*
 * NOTE: This code assumes the disk structure (CRQ's and Messages ) and the
 * _layout is in one-to-one phsyical memory.
 */

/*
 * A message to make new disk CRQ's with.
 */
static CreateChannelRequest CreateChannelMessage;


/*
 * The transfer message pool. This should be class data once G++ can handle
 * class data with constructors.
 */
static const int poolMessageCount = 10;		// # xfer messages in pool
static CRQ_MSGPool DiskTransferRequestMessagePool;
static MassStoreTransferRequest PoolMsg[ poolMessageCount ];// disk pool objects

/*
 * All the constructor really does is stick the free messages on the
 * CRQ's free list. Most of the real work is done in ::initialize()
 * The vector number should be dynamically determined.
 */
MultimaxEMCDisk::MultimaxEMCDisk( int whichSlot, int unit, int vector )
	: Disk( 1 ),
	  crq( UnitID( 0, whichSlot, 0, unit ),
	       ClassVector( vector, CPUCLASS_RUNNING ), CRQMODE_INTR ),
	  		myInterrupt( this )
{
	// size and amount are set in init...
	initDebug();

	/*
	 * Were not initialized until all our processes are run.
	 */
	alertSemaphore = new BinarySemaphore( 0 );
	initDone = 0;
	slot = whichSlot;
	logicalUnit = unit;
	responseVector = vector;
}

MultimaxEMCDisk::~MultimaxEMCDisk()
{
}


int
MultimaxEMCDisk::transient()
{
	return( 0 );
}

void
MultimaxEMCDisk::diskInterruptProcessEntry( MultimaxEMCDisk * disk )
{
	Assert( disk != 0 );
	Debug << "MultimaxEMCDisk::diskInterrupProcessEntry(" << disk
		<< " )\n" << eor;

	/*
	 * Loop forever waiting on the alert semaphore for critical messages
	 * and process them if they ever occur. The interrupt only V's this
	 * semaphore when something bad happening is detected in the interrupt
	 * handler. With normal I/O completions, the interrupt just V's the
	 * semaphore of the request to resume any waiter.
	 */
	while( 1 ) {
		disk->alertSemaphore->P();

		/*
		 * Process the alert condition.
		 */
		Console() << "MultimaxEMCDisk::diskInterruptProcess: ALERT\n" << eor;
		disk->handleAttention();
	}
}

void
MultimaxEMCDisk::initialize()
{
	Assert( this != 0 );
	initDebug();
	Raids( RaidConstructor ) << this << "::initialize()\n" << eor;

	/*
	 * Stick all the free Xfer messages in the Message pool.
	 */
	for( int i = 0; i < poolMessageCount; i++ ) {
		DiskTransferRequestMessagePool.addMessage( &PoolMsg[i] );
	}

	/*
	 * Stick our free attention messages on our CRQ.
	 */
	for( i = 0; i < FreeAttentions; i++ ) {
		crq.enqueueFreeMessage( &attention[i] );
	}

	/*
	 * Install the MultimaxEMCDiskInterrupt to handle
	 * interrupts from this disk.
	 */
	thisCPU()->setInterrupt( responseVector, &myInterrupt );

	/*
	 * Before trying to initialize the disk start up a process to wait
	 * for the disk interrupts. This is since the initialization code
	 * will try to read the disk to get the header and someone needs
	 * to be waiting for the request to finish.
	 */
	Process * process = new InterruptProcess( 
		(PFV) diskInterruptProcessEntry, thisProcess()->domain(),
		normal, (int) this );
	Assert( process != 0 );
	process->setName( "#DiskInterrrupt" );
	Raids( RaidDetail ) << this << "::initialize(): interrupt process is "
		<< process << "\n" << eor;

	/*
	 * Mark the interrupt process as ready to run and
	 * relinquish the processor. Since it's
	 * an interrupt process it should have high enough priority to
	 * run before this process is rescheduled. (I hope). If there are
	 * other processors, one of them might grab it and run it before
	 * this processor checks the scheduler for it. In that case, we
	 * will just get back here anyway.
	 */
	process->ready();
	thisProcess()->relinquishProcessor();

	/*
	 * Call the disk initialization code.
	 */
	Raids( RaidDetail ) << this << "initialize(): interrupt process started\n" << eor;

	/*
	 * Next, initialize the CRQ and read the geometry/partition info
	 * off the disk.
	 */

 	/*
	 * Get a pointer to the EMC's requestor CRQ.
	 */
	CRQ * EMCCRQ = RequestorCRQ( slot, REQ_LUN );
	Raids( RaidDetail ) << this << "::initialize(): Disk CRQ " << asAddress( &crq )
		<< " EMCCRQ " << asAddress( EMCCRQ ) << "\n" << eor;

	/*
	 * Initialize the CRQ as a channel for channel for disk requests
	 */
	CreateChannelMessage.setup( &crq );

	Raids( RaidDetail ) << this << "::initialize(): sending create chan and blocking\n"
		<< eor;
	EMCCRQ->sendCommandAndAwaitResponse( &CreateChannelMessage );

	/*
	 * Verify the reply.
	 */
	Raids( RaidDetail ) << this << "::initialize(): message/channel status: "
		<< asHex( CreateChannelMessage.status() ) << "/"
		<< asHex( CreateChannelMessage.resultStatus() ) << eor;
	Assert( CreateChannelMessage.status() == STS_SUCCESS );

	/*
	 * At this point we have a CRQ to the disk and can start asking it
	 * questions.
	 */
	Raids( RaidDetail ) << this << "::initialize(): Disk CRQ constructed\n" << eor;

	/*
	 * Read Disk Layout/Geometry blocks from unit
	 */
	MassStoreTransferRequest * msg = (MassStoreTransferRequest *)
		DiskTransferRequestMessagePool.removeMessage();
	Assert( msg != 0 );

	Raids( RaidDetail ) << this << "::initialize(): Enquiring Disk Layout ("
		<< sizeof( _layout) << " bytes at start of disk)\n" << eor;

	int pblk = 0;	// _layout is the first thing on the disk...
	msg->map( CRQOP_MS_READ, (const char *) &_layout, sizeof( _layout ),
		pblk );
	crq.sendCommandAndAwaitResponse( msg );
	Raids( RaidDetail ) << this << "::initialize(); Disk _layout request response status: "
		<< msg->status() << "\n" << eor;
	Assert( msg->status() == STS_SUCCESS );

	Raids( RaidDetail ) << this << "::initialize():  expected/actual magic_number: "
		<< SCC_MAGIC_NUMBER << "/" << _layout.l_SCC_magic_number <<
		"\n" << eor;
	Assert( SCC_MAGIC_NUMBER == _layout.l_SCC_magic_number );
	Assert( _layout.l_current_part <= MAXPARTITIONS );
	Assert( _layout.l_maximum_part <= MAXPARTITIONS );

	Raids( RaidDetail ) << this << "::initialize(): Disk Geometry: " <<
		_layout.lay_geom.total_blocks << " total blocks, " <<
 		_layout.lay_geom.byte_sector << " bytes/sector, " <<
		_layout.lay_geom.sector_track << " sectors/track, " <<
		_layout.lay_geom.track_cylinder << " tracks/cylinder, " <<
		_layout.lay_geom.cylinders << " total cylinders, " <<
		_layout.lay_geom.total_avail_blocks << " addressable blocks, " <<
		_layout.lay_geom.avail_sect_cylinder  << "sectors/cylinder\n"
		<< eor;
	Assert( _layout.lay_geom.byte_sector == 512 );

	int totalPartitions = 0;

	for( int part = 0; part < _layout.l_maximum_part; part++ ) {
		if( _layout.l_partitions[part].part_type != 0 ) {
			Raids( RaidDetail ) << this << "::initialize(): partition:"
				<< part << "(" <<
				(char *) &_layout.l_part_namesp +
				_layout.l_partitions[part].part_name_off
				<< ") size:" <<
				_layout.l_partitions[part].part_size <<
				" offset:" <<
				_layout.l_partitions[part].part_off <<
				"\n" << eor;
			totalPartitions++;
		}
		else {
			Assert( _layout.l_partitions[part].part_size == 0 );
			Assert( _layout.l_partitions[part].part_off == 0 );
		}
	}
	Assert( totalPartitions == _layout.l_current_part );
	_partitions = _layout.l_maximum_part;

	/*
	 * Return the message to the free list.
	 */
	DiskTransferRequestMessagePool.addMessage( msg );

	/*
	 * Set the MemoryObject/MemoryRange instance variables now.
	 */
	Assert( _layout.lay_geom.byte_sector == 512 );
	_log2UnitSize = 9;
	_numberOfUnits = _layout.lay_geom.total_avail_blocks;

	initDone = 1;

	Raids( RaidDetail ) << this << "::initialize(): disk is ready\n" << eor;
}

/*
 * Disk Attention handler.
 */
void
MultimaxEMCDisk::handleAttention()
{
	Assert( NOTREACHED );

	CRQ_MSG * attn = 0;
	int didWork = 0;

	do {
		if( ( attn = crq.dequeueAttentionMessage() ) != 0 ) {
			switch( attn->code() ) {
			   default:
				crq.handleStandardAttention( attn );
				break;
			}
			crq.enqueueFreeMessage( attn );
			didWork = 1;
		}
	} while( attn != 0 );

	if ( ! didWork ) {
		Console() << this << "::handleAlert: NO WORK?\n" << eor;
	}
}

void
MultimaxEMCDiskInterrupt::basicRaise( char * )
{
	client->handleInterrupt();
}

/*
 * DiskInterruptHandler. If an attention arrived let the attention handling
 * process deal with it. Otherwise, if its a normal message reponse just V
 * the semaphore in the message.
 */
void
MultimaxEMCDisk::handleInterrupt()
{
	CRQ_MSG * rsp = 0;
	int didWork = 0;
	int wasAlert;

	Raids( RaidProtected ) << thisProcess() << this << "::handleInterrupt()\n" << eor;

	/*
	 * The loop will handle multiple events for efficiency. Unfortunatly,
	 * we still may get multiple interrupts due to the way Choices
	 * does the interrupt acknowledgement.
	 */
	do {
		/*
		 * If there is an attention message waiting, wake up the
		 * interrupt manager process by V'ing the alertSemaphore.
		 */
		if( ! crq.crq_attn.isEmpty() ) {
			alertSemaphore->V();
			wasAlert = 1;
		}
		else {
			wasAlert = 0;
		}

		/*
		 * If there is a response message waiting, handle it.
		 */
		if( ( rsp = crq.dequeueResponse() ) != 0 ) {
			switch( rsp->code() ) {
			   case CRQOP_MS_READ:
			   case CRQOP_MS_WRITE:
				Assert( rsp->semaphore() != 0 );
				rsp->semaphore()->V();
				break;
			   default:
				Console() << this <<
					"::handleInterrupt() BAD RESPONSE\n"
					<< eor;
				break;
			}
			didWork = 1;
		}
	} while( ( rsp != 0 ) || ( wasAlert ) );

	if ( ! didWork ) {
		Console() << this << "::handleInterrupt(): NO WORK?\n"
			<< eor;
	}
	Raids( RaidProtected ) << this << "::handleInterrupt done\n" << eor;
}

/*
 * The other side (initiator) of the above. This gets a transfer message,
 * initializes it and then sends it to the disk channel.
 */
int
MultimaxEMCDisk::doio( unsigned int start, int count, char * addr, int nbytes,
							int op )
{
	Assert( nbytes == (count << log2UnitSize()) );
	PhysicalMemoryChain aChain = thisProcess()->domain()->constructChain(
					addr, nbytes, 1 );
	PhysicalMemoryChain * chain = &aChain;

	Raids( RaidMethod ) << thisProcess() << this << "::doio( start:" << start << " count:" << count <<
		" pauc:" << asAddress( chain ) << " op:" << op << ")\n" << eor;

	switch (op) {
		case DiskReadOperation:
			op = CRQOP_MS_READ;
			break;
		case DiskWriteOperation:
			op = CRQOP_MS_WRITE;
			break;
		default:
			Assert (NOTREACHED) 	
				("MultimaxEMCDisk::doio: invalid operation %d\n", op);
			break;
	};

	/*
	 * Get a MassStoreTransferRequest message from the pool to use for
	 * the transfer.
	 */
	MassStoreTransferRequest * msg = (MassStoreTransferRequest *)
		DiskTransferRequestMessagePool.removeMessage();
	Assert( msg != 0 );

	/*
	 * Set up the transfer command. (512 == bytes in a block)
	 */

	/* limit to 8192, imposed by CRQ code */
	if (nbytes > 8192) {
		nbytes = 8192;
		count = nbytes >> 9;
	}
	msg->map( op, chain, nbytes, start );
	Raids( RaidDetail ) << this << "::doio: sending message "
			    << asAddress( msg ) << "\n" << eor;
	crq.sendCommandAndAwaitResponse( msg );
	Raids( RaidDetail ) << this << "::doio: xferred " << count << " blocks\n" << eor;

	/*
	 * DiskInterruptHandler has already removed msg from the
	 * response queue so just check status. In the future,
	 * fix this to return the number transferred in a partially
	 * successful transfer if at all possible.
	 */
	int nxferred;
	if( msg->status() != STS_SUCCESS ) nxferred = 0;
	else nxferred = count;

	/*
	 * Return the message to the free list.
	 */
	DiskTransferRequestMessagePool.addMessage( msg );

	return( nxferred );
}

int
MultimaxEMCDisk::rawRead( unsigned int start, int count, char * addr, int nbytes )
{
	Raids( RaidMethod ) << this << "::rawRead( " << start << ", " << count
		<< " ) " << _numberOfUnits << "\n" << eor;
	return( doio( start, count, addr, nbytes, DiskReadOperation ) );
}

int
MultimaxEMCDisk::rawWrite( unsigned int start, int count, char * addr, int nbytes )
{
	Raids( RaidMethod ) << this << "::rawWrite( " << start << ", " << count
		<< " ) " << _numberOfUnits << "\n" << eor;
	return( doio( start, count, addr, nbytes, DiskWriteOperation ) );
}

int
MultimaxEMCDisk::basicDoio( unsigned int, char *, int )
{
    /*
     * NOTICE:  Someday somebody should take the essential part out of
     *   doio and implement basicDoio instead.  For now we are a subclass
     *   of Disk but overload the doio, most of which is already implemented
     *   in Disk::doio.  At that time should also be able to stop overloading
     *   the read() and write() methods from Disk.
     */
    Assert( NOTREACHED );
    return( DiskOperFailed );
}

int
MultimaxEMCDisk::reset()
{
    /* nothing for now */
    return( DiskOperSucceded );
}

int
MultimaxEMCDisk::locatePartition( int idNumber, int &start, int &units )
{
	Debug << "MultimaxEMCDisk::locatePartition( idNumber:" << 
		idNumber << " )\n" << eor;

	if( _layout.l_partitions[ idNumber ].part_size == 0 ) {
		Console().printf( "%N::locatePartition: empty partition %d\n",
				 this, idNumber );
		return( -1 );
	}

	start = _layout.l_partitions[ idNumber ].part_off;
	units = _layout.l_partitions[ idNumber ].part_size;

	return( 0 );
}

#include "ClassInstaller.h"
DEFINECLASS("MultimaxEMCDisk",MultimaxEMCDisk,Disk,0,0);
