#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Contributing Author: Nayeem Islam (nislam@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MultimaxKernel.cc: Multimax kernel class implementation.
 */

//#define DEBUG
//#define NONETWORKING

#include "Assert.h"
#include "CRQ.h"
#include "CPU.h"
#include "ThisCPU.h"
#include "PROCESSOR.h"
#include "OutputStream.h"
#include "Interrupt.h"
#include "MemoryObjectStar.h"
#include "MemoryObjectContainerStar.h"
#include "MemoryObjectContainer.h"
#include "MemoryObjectView.h"
#include "MultimaxAPCCPU.h"
#include "MultimaxEMCDisk.h"
#include "MultimaxEMCDiskStar.h"
#include "MultimaxKernel.h"
#include "NameServer.h"
#include "NS32202.h"
#include "PageTableConstants.h"
#include "PersistentMemoryObjectStar.h"
#include "PhysicalMemoryRange.h"
#include "RealTimeClock.h"
#include "Store.h"
#include "ThisProcess.h"
#include "Vectors.h"
#include "VirtualMemoryRange.h"
#include "MultimaxEMCDiskStar.h"
#include "FileSystemInterface.h"
#include "defNETWORKING.h"
#ifdef NETWORKING
#include "EthernetAdministrator.h"
#include "MultimaxEthernetDriver.h"
#endif
#include "MAGIC.h"
#include "FileObject.h"
#include "a.out.h"      // This assumes COFF is available, but so do
			//      other things in this file; will have to
			//      move it someday



#ifdef __GNUC__
#pragma implementation
#endif

MultimaxKernel::MultimaxKernel()
	: getConfigMessage( &systemConfiguration )
{
	Console() << "(ENCORE Multimax Version)\n" << eor;
}

MultimaxKernel::~MultimaxKernel()
{
	Assert( NOTREACHED );
}

/*
 * We could get the from the configuration information...
 */
DiskRef
MultimaxKernel::rootDisk()
{
	/*
	 * ::rootDisk() should never be called until after the disk
	 * is initialized.
	 */
	Assert( theDisk != 0 );
	Assert( theDisk->initDone );
	return( theDisk );
}

extern void TransmitVector( unsigned int );


/*
 * Start processor "cpuid" in slot "slotid".
 */
void
MultimaxKernel::startProcessor( unsigned int slotid, unsigned int cpuid,
	PFV addr )
{
	Assert( slotid < NUM_SLOT );
	Assert( cpuid < NUM_SLOT_DEV );
	Debug << "StartProcessor( slotid:" << slotid << " cpuid:"
		<< cpuid << " addr:" << asAddress( addr ) << " )\n" << eor;

	/*
	 * Get a pointer to the proper CRQ for the processor to be started.
	 */
	CRQ * crq = RequestorCRQ( slotid, cpuid );
	Assert( crq != 0 );
	Debug << "StartProcessor: crq: " << asAddress( crq ) << "\n" << eor;
#ifdef DEBUG
	crq->dump();
#endif
	crq->crq_mode = CRQMODE_POLL;

	/*
	 * Initialize command.
	 */
	CRQ_EXEC_MSG * cmd = new ( &execMsg ) CRQ_EXEC_MSG( slotid,
		cpuid, addr );
	Debug << "StartProcessor: cmd: " << asAddress( cmd ) << "\n" << eor;

	/*
	 * Enqueue the message and send an interrupt to the slave to wake
	 * it up.
	 */
	Debug << "StartProcessor: enqueing command " << asAddress( cmd )
		<< " on " << asAddress( crq ) << "\n" << eor;
	crq->sendCommand( cmd );

	TransmitVector( (unsigned int) crq->crq_slave_vect );

	Debug << "StartProcessor: request enqued\n" << eor;

	crq->crq_mode = CRQMODE_INTR;
}

/*
 * Go through each possible slot on a Multimax and see whats in it.
 * Try to start the A and B processor if its a processor card.
 * We know that any object creations done here will have their memory
 * mapped 1-to-1 once VM is enabled. This is useful for devices to know
 * since they can allocate DMA buffers, etc... right from physical memory.
 */
void
MultimaxKernel::configureHardware()
{
	determineConfiguration();
	Assert( _initialScheduler != 0 );

	/*
	 * Print the configuration and simultaneous do any board specific
	 * configurations. (Currently just start the other processors.)
	 */
	Console() << "Encore Multimax System Configuration:\n" << eor;
	for( int slot = 0; slot < NUM_SLOT; slot++ ) {
		long boardType =
			systemConfiguration.sys_state_req[slot].req_board_id;
		if( boardType != 0 ) {
			Console() << "\tRequestor Slot " << slot << " ";
			switch( boardType ) {
			   case EMC:
			   case EMCII:
				Console() << "Ethernet/Mass Storage"
					<< " Controller\n";

				/*
				 * For now assume there is only one EMC
				 * in the machine and it has only 1 disk
				 * LUN = 16, which uses vector DISK_Vector
				 * for it's interrupts.
				 */
				Assert( theDisk == 0 );
				theDisk = new MultimaxEMCDisk( slot, 16,
					DISK_Vector );
				Assert( theDisk != 0 );
				break;

			   case APC:
				Console() << "Advanced ";
			   case DPC:
			   case DPCII:
				Console() << "Dual Processor Card\n"
					<< eor;
				int myslot = PROCESSOR::cpuID() >> 2;
				int myid = PROCESSOR::cpuID() & 0x3;

				unsigned int pid0 = (slot << 2);
				unsigned int pid1 = (slot << 2) | 1;

				/*
				 * Build CPU object for each processor and
				 * start them.
				 */
				if( ( myslot != slot ) || ( myid != 0 ) ) {
					CPU * newCPU =
						new MultimaxAPCCPU( pid0,
						_initialScheduler );
					newCPU->setName( "#CPU" );
					// OnlineCPUs.add( pid0, newCPU );
					Debug << "built CPU for " << pid0
						<< " (I'm "
						<< PROCESSOR::cpuID()
						<< ")\n" << eor;
				}
				if( ( myslot != slot ) || ( myid != 1 ) ) {
					CPU * newCPU =
						new MultimaxAPCCPU( pid1,
						_initialScheduler );
					newCPU->setName( "#CPU" );
					// OnlineCPUs.add( pid1, newCPU );
					Debug << "built CPU for " << pid1
						<< " (I'm "
						<< PROCESSOR::cpuID()
						<< ")\n" << eor;
				}

				break;
			   case SCC:
			   case SCCII:
				Console() << "System Control Card\n";
				break;
			   case DLA:
				Console() << "DLA (whatever that is)\n";
				break;
			   case SMC:
				Console() << "Shared Memory Card (4Mb?)\n";
				break;
			   case DLA_16_ASYNC:
				Console() << "DLA_16_ASYNC (whatever that is)\n";
				break;
			   case DLA_8_SYNC:
				Console() << "DLA_8_SYNC (whatever that is)\n";
				break;
			   case ANNEX_UX_ASYNC:
				Console() << "ANNEX_UX_ASYNC (whatever that is)\n";
				break;
			   case MSC_TRACE:
				Console() << "Special ";
			   case MSC:
				Console() << "Mass Store Controller\n";
				break;
			   case SMC_16:
				Console() << "A 16Mb Shared Memory Card\n";
				break;
			   case ANNEXII_MX_ASYNC:
				Console() << "ANNEXII_MX_ASYNC (whatever that is)\n";
				break;
			   case ANNEXII_UX_ASYNC:
				Console() << "ANNEXII_UX_ASYNC (whatever that is)\n";
				break;
			   case ANNEXII_SYNC:
				Console() << "ANNEXII_SYNC (whatever that is)\n";
				break;
			   case TIU_UX_ASYNC:
				Console() << "TIU_UX_ASYNC (whatever that is)\n";
				break;
			   case UIC:
				Console() << "UIC (whatever that is)\n";
				break;
			   default:
				Console() << "Unknown board with type "
					<< boardType << "\n";
				break;
			}
			Console() << eor;
		}
	}

	/*
	 * Since the started processors didnt reply to the start command
	 * (see the TACKY ALERT above) there will be a VECTOR_BUS_NAK
	 * (Slave ICU pin 9) interrupt pending at this time.
	 * Clear all the interrupts in the HIGH IPND.
	 *
	 * This is truly gross! --Vince
	 */
	*MasterMCTL = 0xca;	// CFRZ | COUTD | FRZ | NTAR.
	*SlaveMCTL = 0xca;	// CFRZ | COUTD | FRZ | NTAR.

	*MasterELTGL = 0xff;
	*MasterELTGH = 0xff;

	*SlaveELTGL = 0xff;
	*SlaveELTGH = 0xff;
	
	*MasterIPNDL = 0x40;
	*MasterIPNDH = 0x40;
	*SlaveIPNDL = 0x40;
	*SlaveIPNDH = 0x40;

	*MasterISRVL = 0x00;
	*MasterISRVH = 0x00;
	*SlaveISRVL = 0x00;
	*SlaveISRVH = 0x00;

	*MasterMCTL = 0x02;	// NTAR
	*SlaveMCTL = 0x02;	// NTAR

	for( int i = 0; i < NUM_MEM; i++ ) {
		MEM_STATE * ms = &systemConfiguration.sys_state_mem[i];
		// for some reason the 310s report empty slots as 1Mb of memory.
		if( ms->mem_size > 1 ) {
			Console() << "\tMemory Slot " << i << ": "
				<< ms->mem_size << "Mb Shared Memory Card\n"
				<< eor;
		}
	}
	Console() << "\tTop of Memory at "
		<< asAddress( (void *) systemConfiguration.sys_state_topmem )
		<< "\n\n\n" << eor;
	
}

PhysicalMemoryRange *
MultimaxKernel::physicalMemoryMap()
{
	/*
	 * The Multimax has a very simple physical memory map.
	 */
	int unitsOfPhysicalMemory =
		systemConfiguration.sys_state_topmem / PhysicalPageSize;
	PhysicalMemoryRange * memoryMap = new PhysicalMemoryRange(
		0, unitsOfPhysicalMemory, Log2PhysicalPageSize,
						"#Shared Memory (RAM)" );

	int unitsOfIOSpace = HWSIZE / PhysicalPageSize;
	memoryMap->add( new PhysicalMemoryRange( (const char * ) HWADDR,
		unitsOfIOSpace, Log2PhysicalPageSize, "#I/O Space", 1 ) );

	return( memoryMap );
}


void
MultimaxKernel::buildRunningCPU()
{
	extern void InitModuleTable();
	InitModuleTable();

	/*
	 * Build a CPU object for THIS processor. configure() for
	 * even a uniprocessor would do this much. Multiprocessors
	 * continue and make a CPU object for each processor.
	 */
	CPU * current = new MultimaxAPCCPU( PROCESSOR::cpuID(),
		_initialScheduler );
	current->setName( "#FirstCPU" );
	// OnlineCPUs.add( PerMemoryProcessorID(), current );

	/*
	 * At least for the Multimax, all processors share the same interrupt
	 * table. So only 1 processor needs to call installInterrupts().
	 */
	current->installInterrupts();
	Assert( ! PROCESSOR::interruptsEnabled() );
}

void
MultimaxKernel::determineConfiguration()
{
	Debug << this << "::determineSystemConfiguration()\n" << eor;
	CRQ * SCCCRQ = RequestorCRQ( SCC_SLOT, 0 );

	SCCCRQ->crq_mode = CRQMODE_POLL;

	/*
	 * Enqueue the message.
	 */
	Debug << this << "::determineConfiguration: enqueing command "
		<< asAddress( &getConfigMessage ) << " on CRQ "
		<< asAddress( SCCCRQ ) << "\n" << eor;
	SCCCRQ->sendCommand( &getConfigMessage );

	TransmitVector( (unsigned int) SCCCRQ->crq_slave_vect );

	Debug << "DetermineConfig: request enqued\n" << eor;

	static const int maxLoops = 1000;
	static const int maxSpins = 100000;

	for( int loops = 0; loops < maxLoops; loops++ ) {
		/*
		 * Check for a response.
		 */
		SCCCRQ->lock();
		if( ! SCCCRQ->crq_rsp.isEmpty() ) {
			break;
		}
		SCCCRQ->unlock();
		Debug << "DetermineConfig: waiting\n" << eor;
		for( int i = 0; i < maxSpins; i++ ) ;
	}

	if( loops == maxLoops ) {
		Console().panic( "No response from SCC CRQ!\n" );
	}
	CRQ_MSG * response = SCCCRQ->crq_rsp.dequeue();
	SCCCRQ->unlock();

	Assert( response == &getConfigMessage );
	Assert( getConfigMessage.status() == STS_SUCCESS );

	SCCCRQ->crq_mode = CRQMODE_INTR;
}

/*
 * Setup machine dependent devices and device managers.
 */
void
MultimaxKernel::basicInitializeDevicesAndManagers()
{
	extern volatile int slot4initialized;
	extern volatile int slot12initialized;
	extern volatile int consoleInputInitialized;
	extern volatile int tapeInitialized;
#ifdef NETWORKING
	extern volatile int ethernetInitialized;
#endif NETWORKING
	extern volatile int timeoutTimerInitialized;
	extern volatile int periodicTimerInitialized;

	Assert( _initialScheduler != 0 );

	/*
	 * Set up machine dependent (Multimax) manager processes and
	 * devices.
	 */
	Process * process;

	/*
	 * The first interrupt processes that need constructed and
	 * executed are the Requestor interrupt processes. This is since the
	 * other device processes need them to create channels to the devices.
	 *
	 * Right now do only slots 12 and 4 (our SCC and EMC).
	 */
	Debug << "Building SCC interrupt process\n" << eor;
	extern void RequestorInterruptProcessEntry( int );
	process = new InterruptProcess( (PFV) RequestorInterruptProcessEntry,
		domain(), normal, 12 );
	process->setName( "#SCC process" );
	Debug << process << "\n" << eor;
	Assert( process != 0 );
	process->ready();
	while( slot12initialized == 0 ) {
		Debug << thisProcess() << " still waiting\n" << eor;
		thisProcess()->relinquishProcessor();
		Debug << thisProcess() << " running again\n" << eor;
	}

	Debug << "SCC requestor process taken care of\n" << eor;

	process = new InterruptProcess( (PFV) RequestorInterruptProcessEntry,
		domain(), normal, 4 );
	process->setName( "#EMC process" );
	Debug << process << "\n" << eor;
	Assert( process != 0 );
	process->ready();
	/* We _may_ have the only CPU. Give it up for the moment. */
	while( slot4initialized == 0 )
		thisProcess()->relinquishProcessor();

	Debug << "EMC requestor process taken care of\n" << eor;

	/*
	 * Build a stack for and construct the console manager process,
	 */
	Debug << "Building console manager process\n" << eor;
	extern void ConsoleManagerProcessEntry( int );
	process = new SystemProcess( (PFV) ConsoleManagerProcessEntry,
		domain(), normal, 0 );
	Debug << "console process " << process << "\n" << eor;
	Assert( process != 0 );
	process->setName( "#ConsoleManagerProcess" );
	process->ready();
	/* We _may_ have the only CPU. Give it up for the moment. */
	while( consoleInputInitialized == 0 )
		thisProcess()->relinquishProcessor();

	Debug << "Console taken care of\n" << eor;

	/*
	 * Next start up the disk.
	 */
	theDisk->initialize();

        /*
         * Next start up the tape management process.
         */
        extern void TapeManagerProcessEntry( int );
        process = new SystemProcess( (PFV) TapeManagerProcessEntry,
		domain(), normal, 0 );
        Debug << "tape process " << process << "\n" << eor;
        Assert( process != 0 );
	process->setName( "#TapeManager" );
	process->ready();
	/* We _may_ have the only CPU. Give it up for the moment. */
	while( tapeInitialized == 0 )
		thisProcess()->relinquishProcessor();

#ifndef NONETWORKING
#ifdef NETWORKING
	/*
	 * Now do the Ethernet manager process.
	 */
	Debug << "Building Ethernet manager stack/process\n" << eor;
	extern void EthernetManagerProcessEntry( int );
	process = new SystemProcess( (PFV) EthernetManagerProcessEntry,
		domain(), normal, 0 );
	Debug << "Ethernet manager process " << process << "\n" << eor;
	Assert( process != 0 );
	process->setName( "#EthernetManager" );
	process->ready();
	/* We _may_ have the only CPU. Give it up for the moment. */
	while( ethernetInitialized == 0 )
		thisProcess()->relinquishProcessor();

	Debug << "Ethernet manager process done\n" << eor;
#endif NETWORKING
#endif

	/*
	 * Now start the Timeout timer & Periodic timer helper processes.
	 *
	 */
	Debug << "Build Timeout Manager process\n" << eor;
	extern void TimeoutTimerEntry( int );
	process = new SystemProcess( (PFV) TimeoutTimerEntry,
		domain(), normal, 0 );
	Debug << "Timeout manager process " << process << "\n" << eor;
	Assert( process != 0 );
	process->setName( "#TimeoutTimerManager" );
	process->ready();
	/* We _may_ have the only CPU. Give it up for the moment. */
	while( timeoutTimerInitialized == 0 )
		thisProcess()->relinquishProcessor();

	Debug << "Build Periodic Manager process\n" << eor;
	extern void PeriodicTimerEntry( int );
	process = new SystemProcess( (PFV) PeriodicTimerEntry,
		domain(), normal, 0 );
	Debug << "Periodic manager process " << process << "\n" << eor;
	Assert( process != 0 );
	process->setName( "#PeriodicTimerManager" );
	process->ready();
	/* We _may_ have the only CPU. Give it up for the moment. */
	while( periodicTimerInitialized == 0 )
		thisProcess()->relinquishProcessor();

	// Start the system time of day clock
	theSystemClock = new MultimaxTimeOfDay();
	StandardNameServer->bind( "timeOfDay", theSystemClock );
}

void
MultimaxKernel::basicInitializeNetworkProcesses()
{
#ifdef NETWORKING
	//   Start up the EthernetDriver and EthernetAdministrator.
	Debug << "MultimaxKernel:\t\tstarting EthernetDriver\n" << eor;

	//   Start up the Administrator.
	Debug << "MultimaxKernel:\t\tstarting EthernetAdministrator\n" << eor;
	_ethernetAdministrator = new EthernetAdministrator(
						"ethernetAdministrator");
	Assert(_ethernetAdministrator != 0);

#	if defined(NETWORK_TEST)
		_ethernetDriver = new StubEthernetDriver();
		Debug << "\t\t\t\tStarted *STUB* EthernetDriver\n" << eor;
#	else
		_ethernetDriver = new MultimaxEthernetDriver();
#	endif /* NETWORK_TEST */

	Assert(_ethernetDriver != 0);
	_ethernetDriver->setName("ethernetDriver");
	StandardNameServer->bind("ethernetDriver", _ethernetDriver);

	_ethernetDriver->ready(_ethernetAdministrator);
	_ethernetAdministrator->ready(_ethernetDriver);

#endif NETWORKING
}


void
MultimaxKernel::startOtherProcessors()
{
	extern void processorEntry();


	Assert( thisCPU() != 0 );

	for( int i = 0; i < MAX_PROCESSORS; i++ ) {
		CPU * cpu = CPU::_cpus[i];
		if( cpu == thisCPU() ) {
			Debug << this << "::startOtherProcessors: skipping self\n"
				<< eor;
			continue;
		}
		if( cpu == 0 ) {
			continue;
		}

		unsigned long id = cpu->id();

		int slot = id >> 2;
		int cpuid = id & 0x3;

		Debug << this << "::startOtherProcessors: start ("
			<< slot << ", " << cpuid << ")\n" << eor;

		startProcessor( slot, cpuid, processorEntry );
	}
}

void
MultimaxKernel::mountNonRootFileSystems()
{
	_fileSystemInterface->mount( "!4/", "usr", 0 );
	_fileSystemInterface->mount( "!7/", "mnt", 0 );
}

void
MultimaxKernel::initializeBackingStore()
{
	MemoryObjectStar disk = rootDisk();
	Assert( disk != 0 );
	int status = 0;
	MemoryObjectContainerStar rootContainer =
		disk->asA( MemoryObjectContainerClass, status );
	Assert( status == 0 );
	Assert( rootContainer != 0 );
	MemoryObjectStar mo = rootContainer->open( 4, status, O_NCACHE );
	Assert( status == 0 );
	Assert( mo != 0 );
	_backingStoreSource = mo->asA( MemoryObjectContainerClass, status );
	Debug << "backingStoreSource: " << _backingStoreSource << "\n" << eor;
	Assert( _backingStoreSource != 0 );
	_backingStoreSource->setName( "backingStoreSource" );
}


/*
 * Interpret a memory object a a COFF file and fill a domain with a new
 * MemoryObjectCache mapping each section in the COFF file. The assumption
 * about COFF can be fixed my comment in the Kernel.h file. --Vince
 * It is assumed all the sections in the file are "unit" aligned both in
 * virtual memory and in the actual memory object of the file itself so they
 * can be paged directly from the file.
 */
/*
 * The magic numbers for COFF are defined in MACHINE_MAGIC.h
 * to make this file machine-independent.
 */
//#define Debug Console()
const char *
MultimaxKernel::loadObjectFile( PersistentMemoryObject * coffFile, Domain * domain )
{
	Debug << this << "::loadObjectFile( objectFile: " << coffFile
		<< " domain: " << domain << " )\n" << eor;

	int unitSize = coffFile->unitSize();

	char * HeaderBuffer = new char[ unitSize ];
	Assert( HeaderBuffer != 0 );
	char * TextBuffer = new char[ unitSize ];
	Assert( TextBuffer != 0 );

	Assert( domain != 0 );

	/*
	 * Read in the file header and a.out header. The current assumption is
	 * that all the COFF section information fits one unit of the file.
	 * (including all the Section headers (at least .text, .data & .bss).)
	 */
	Assert(( sizeof(filehdr)+sizeof(aouthdr)+3*sizeof(scnhdr)) <= unitSize);
	Debug << this << "::loadObjectFile read " << unitSize
		<< " bytes of header\n" << eor;
	int i = coffFile->read( 0, 1, HeaderBuffer, unitSize );
	if( i != 1 )
		thisProcess()->console() << this <<
			 "::loadObjectFile: read " << coffFile <<
			 " returns " << i <<
			 "\n" << eor;
	Assert( i == 1 );
	Debug << this << "::loadObjectFile: header read\n" << eor;

	filehdr * fileHeader = (filehdr *) HeaderBuffer;

	/*
	 * See if the fileheader has the right magic number in it.
	 */
	if( ( fileHeader->f_magic != MACHINE_MAGIC_1 ) &&
	    ( fileHeader->f_magic != MACHINE_MAGIC_2 ) &&
	    ( fileHeader->f_magic != MACHINE_MAGIC_3 ) &&
	    ( fileHeader->f_magic != MACHINE_MAGIC_4 ) &&
	    ( fileHeader->f_magic != MACHINE_MAGIC_5 ) &&
	    ( fileHeader->f_magic != MACHINE_MAGIC_6 ) ) {
		Console() << this << "::loadObjectFile: Bad magic number "
			<< asOctal( fileHeader->f_magic ) << "\n" << eor;
		delete HeaderBuffer;
		delete TextBuffer;
		return( 0 );
	}
	Debug << this << "::loadObjectFile uSize=" << unitSize
		<< " Bytes=" << coffFile->numberOfBytes()
		<< " Units=" << coffFile->numberOfUnits()
		<< " id=" << coffFile->idNumber()
		<< " link=" << coffFile->links( 0 ) << "\n" << eor;

	/*
	 * Checking the optional header seems to be the only way to get
	 * the entrypoint. I'd like to ignore it all together if I
	 * could or even better replace it with more useful information.
	 */
	aouthdr * aoutHeader = (aouthdr *) ( HeaderBuffer + sizeof( filehdr ) );

	const char * entryPoint = (const char *) aoutHeader->entry;
	Debug << this << "::loadObjectFile: " << fileHeader->f_nscns
		<< " sections, opt header size: "
		<< fileHeader->f_opthdr << " bytes, start: "
		<< asAddress( entryPoint ) << "\n" << eor;

	/*
	 * Ignore the rest of the optional header and interpret the sections
	 * one at a time.
	 *
	 * Here and everywhere else I need to put code to make sure the
	 * file wasn't short. --Vince
	 */
	scnhdr * sectionHeader = (scnhdr *) ( HeaderBuffer + sizeof( filehdr )
		+ fileHeader->f_opthdr );
	void * startText = 0;
	void * staticBaseOffset = 0;
	for( int section = 0; section < fileHeader->f_nscns; section++ ) {
		Assert( (char *) sectionHeader < &HeaderBuffer[ unitSize ] );

		Debug << this << "::loadObjectFile: " << sectionHeader->s_name
			<< "(#" << section << ") paddr: "
			<< asHex( (unsigned int) sectionHeader->s_paddr )
			<< " vaddr: "
			<< asHex( (unsigned int) sectionHeader->s_vaddr )
			<< " size: "
			<< (unsigned int) sectionHeader->s_size << " flags: "
			<< asHex( (unsigned int) sectionHeader->s_flags )
			<< "\n" << eor;

		/*
		 * Skip zero length sections.
		 */
		if( sectionHeader->s_size == 0 ) {
			Debug << this << "::loadObjectFile: Skipping zero length section\n" << eor;
			/*
			 * Point to the next section header.
			 */
			sectionHeader++;
			continue;
		}

		/*
		 * Build a PersistentMemoryObject to manage each section in the file.
		 */
		PersistentMemoryObjectStar memObj = 0;

/*
 * Fix this to use the protection in the COFF file not the implied ones.
 */

		ProtectionLevel protection;

		switch( sectionHeader->s_flags ) {

		   case STYP_TEXT:
			Debug << "COFF: text section\n" << eor;
			protection = ReadOnly;

			int startingUnitNumber = coffFile->offsetToUnit(
				(unsigned int) sectionHeader->s_scnptr );
			int numberOfUnits = coffFile->offsetToUnit(
				(Offset) ( sectionHeader->s_size +
				unitSize - 1 ) );

			Debug << "COFF: startingUnit:" << startingUnitNumber
				<< " numberOfUnits:" << numberOfUnits
				<< "\n" << eor;

/* old style
			memObj = new FileObject( numberOfUnits,
						 coffFile->log2UnitSize(),
						 0, 0,
						 coffFile, 
						 startingUnitNumber );
*/
			/* 
			 * Assume starting offset is always page aligned
			 */

			memObj = new NormalMemoryObjectView
				( coffFile, numberOfUnits,
				  coffFile->log2UnitSize(),
				  startingUnitNumber );

			memObj->setName( ".text" );
			Debug << "COFF: text view " << memObj << "\n" << eor;
			startText = (void *) sectionHeader->s_vaddr;

			/*
			 * Mirror the UMAX semantics for finding out what
			 * the static base register should be. The value,
			 * relative to the start of .bss, is always stored
			 * as an int 32 bytes (8 ints) into .text.
			 */
    			memObj->read( 0, 1, TextBuffer, unitSize );

			staticBaseOffset = ((void **)TextBuffer)[8];
			Debug << "COFF: staticBaseOffset: " <<
				asAddress(staticBaseOffset) <<
				"\n" << eor;
			break;

		   case STYP_DATA:
			{
			Debug << "COFF: data section\n" << eor;
			protection = ReadWrite;

			startingUnitNumber = coffFile->offsetToUnit(
				(unsigned int) sectionHeader->s_scnptr );
			numberOfUnits = coffFile->offsetToUnit(
				(Offset) ( sectionHeader->s_size +
				unitSize - 1 ) );

			Debug << "COFF: startingUnit:" << startingUnitNumber
				<< " numberOfUnits:" << numberOfUnits
				<< "\n" << eor;

			PersistentMemoryObjectStar view =
				new FileObject( numberOfUnits,
					coffFile->log2UnitSize(),
					0, 0,
					coffFile, startingUnitNumber );
			Debug << "COFF: data view " << view << 
				" nunits = " << view->numberOfUnits() << 
				" unitSize = " << view->unitSize() << "\n"
				<< eor;

			/*
			 * Allocate a new memory object to swap to.
			 */
			memObj = allocateMemoryObjectForBackingStore();
			Assert( memObj != 0 );
			memObj->setName( ".data" );

			Debug << "COFF: copy data view\n" << eor;
			view->copy( memObj );
			Debug << "COFF: data view copied\n" << eor;
			Debug << "COFF: data swap inode "
				<< memObj << " nunits = "
				<< memObj->numberOfUnits() << 
				" unitSize = " << memObj->unitSize() <<
				"\n" << eor;
			};
			break;

		   case STYP_BSS:
			/* fall through */
			
		   case STYP_NOLOAD:
			Debug << "COFF: noload or bss\n" << eor;
			protection = ReadWrite;

			/*
			 * Allocate a new memory object to swap to.
			 */
			memObj = allocateMemoryObjectForBackingStore();
			Assert( memObj != 0 );
			if( sectionHeader->s_flags == STYP_BSS ) {
				memObj->setName( ".bss" );
			}
			else memObj->setName( ".noload" );

			/*
			 * Since the actual memory object has zero length at
			 * this time, grow it out to be large enough to hold
			 * the maximum size it will grow to and then add it
			 * to the domain. I hope memobjs can handle this.
			 */
			numberOfUnits = memObj->offsetToUnit(
				(Offset) ( sectionHeader->s_size +
				memObj->unitSize() - 1 ) );

			memObj->setNumberOfUnits( numberOfUnits );

			Debug << "COFF: swap area in "
				<< memObj << " (" << numberOfUnits <<
				" units)\n" << eor;

			break;

		   default:
			protection = NoAccess;
			memObj = 0;
			Console().panic( "COFF: Unexpected flags for %s: %d\n",
				sectionHeader->s_name, sectionHeader->s_flags );
		}

		Debug << "COFF section "
			<< sectionHeader->s_name
			<< " is in " << memObj
			<< "\n" << eor;
		Assert( memObj != 0 );

		const char * where = domain->add( memObj,
			(const char *) sectionHeader->s_vaddr, protection );
		Assert( where == (const char *) sectionHeader->s_vaddr );

		/*
		 * Point to the next section header.
		 */
		sectionHeader++;
	}
	Debug << "COFF: staticBase:" <<	
		asAddress(staticBaseOffset) << "\n" << eor;

	domain->setModuleTableEntry( startText, staticBaseOffset );
	Debug << this << "::loadObjectFile: done!\n" << eor;

	delete HeaderBuffer;
	delete TextBuffer;
	return( entryPoint );
}


int
MultimaxKernel::isObjectFile( PersistentMemoryObject * coffFile )
{
	Debug << this << "::isObjectFile( objectFile: " << coffFile
		<< " )\n" << eor;

	int unitSize = coffFile->unitSize();

	char * HeaderBuffer = new char[ unitSize ];
	Assert( HeaderBuffer != 0 );

	/*
	 * Read in the file header and a.out header. The current assumption is
	 * that all the COFF section information fits one unit of the file.
	 * (including all the Section headers (at least .text, .data & .bss).)
	 */
	Assert(( sizeof(filehdr)+sizeof(aouthdr)+3*sizeof(scnhdr)) <= unitSize);
	Debug << this << "::isObjectFile read " << unitSize
		<< " bytes of header\n" << eor;
	int i = coffFile->read( 0, 1, HeaderBuffer, unitSize );
	if( i != 1 )
		thisProcess()->console() << this <<
			 "::loadObjectFile: read " << coffFile <<
			 " returns " << i <<
			 "\n" << eor;
	Assert( i == 1 );
	Debug << this << "::loadObjectFile: header read\n" << eor;

	filehdr * fileHeader = (filehdr *) HeaderBuffer;

	/*
	 * See if the fileheader has the right magic number in it.
	 */
	if( ( fileHeader->f_magic != MACHINE_MAGIC_1 ) &&
	    ( fileHeader->f_magic != MACHINE_MAGIC_2 ) &&
	    ( fileHeader->f_magic != MACHINE_MAGIC_3 ) &&
	    ( fileHeader->f_magic != MACHINE_MAGIC_4 ) &&
	    ( fileHeader->f_magic != MACHINE_MAGIC_5 ) &&
	    ( fileHeader->f_magic != MACHINE_MAGIC_6 ) ) {
		Console() << this << "::loadObjectFile: Bad magic number "
			<< asOctal( fileHeader->f_magic ) << "\n" << eor;
		delete HeaderBuffer;
		return( 0 );
	}
	delete HeaderBuffer;
	return( 1 );
}


#include "ClassInstaller.h"
DEFINECLASS("MultimaxKernel",MultimaxKernel,Kernel,0,0);
