#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David W. Dykstra (dykstra@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MultimaxTimerManager.c - machine dependent portion of timer managing
 */

#include "OutputStream.h"
#include "Assert.h"
#include "TIMER_MANAGER.h"
#include "CPU.h"
#include "ThisCPU.h"
#include "Lock.h"
#include "FreeRunningTimer.h"
#include "FreeRunningTimerStar.h"
#include "Vectors.h"
#include "PROCESSOR.h"


#ifdef __GNUC__
#pragma implementation
#endif

/* the ClockFrequency should be an even divisor of 1000000 */
static const unsigned int ClockFrequency = 100;   /* ticks/second */

extern TimerManager *TimeoutManager;
extern TimerManager *PeriodicManager;

void
TimerManagerInterrupt::basicRaise( char * )
{
    /*
    Debug.printf( "TimerManagerInterrupt%T::basicRaise()\n", this );
    */
    Assert( this != 0 );
    Assert( ! PROCESSOR::interruptsEnabled() );

    TimeoutManager->interruptHandler();

    PeriodicManager->interruptHandler();
}

MultimaxTimerManager::MultimaxTimerManager(PFI ptb, PFV ep)
	: TimerManager(ptb, ep)
{
    Debug.printf("MultimaxTimerManager::ctor()\n");

    /*
     * Don't actually initialize the hardware timer until necessary;
     *   that is, when the 'initialize' function is called.
     */

    /* QUESTION: should this be rounded up, or always be evenly divisible?*/
    unsigned int microTicks = 1000000/ClockFrequency;
    Assert((microTicks * ClockFrequency) == 1000000);
    Assert(microTicks > 0);

    /*
     * The microsecondsPerTick could conceivably be different for
     *   timeout timers and periodic timers, but they're the same here;
     *   in fact, there's no way to tell the difference at this point.
     */
    microsecondsPerTick = microTicks;

    setFreeRunningTime(SystemTimer->time());
}

static volatile unsigned int SCCREG_BASE = (unsigned int) 0xfffd0000;

/* interval timer control registers */
static unsigned volatile char * SCCREG_PITCNT0 =
	(unsigned char *) SCCREG_BASE;
static unsigned volatile char * SCCREG_PITCNT1 =
	(unsigned char *) (SCCREG_BASE + 0x4);
static unsigned volatile char * SCCREG_PITCNT2 =
	(unsigned char *) (SCCREG_BASE + 0x8);
static unsigned volatile char * SCCREG_PITCTL =
	(unsigned char *) (SCCREG_BASE + 0xc);

/* interval timer vector register */
static volatile long * SCCREG_PITVEC = (long *) (SCCREG_BASE + 0x100);

static const int SC0 =	0x00;	// select counter 0
static const int SC1 =	0x40;	// select counter 1

static const int TIM_MODE = 0x34;     // Load both counter bytes, rate generator

static Lock initTimerLock;	/* these two could be within the next routine */
static int intervalTimerStarted = 0;  /* but it gets "sorry, not implemented" */

void
MultimaxTimerManager::initialize()
{
    initTimerLock.acquire();
    /*
     * This is to avoid starting the hardware clock
     *  if no timers are created; initialize is called
     *  whenever a timer is created and only the first
     *  one successeds in initializing.
     */
    if (intervalTimerStarted)
    {
	initTimerLock.release();
	return;
    }
    intervalTimerStarted = 1;

    if (ClockFrequency == 0)
    {
	/* Clock turned off */
	initTimerLock.release();
	return;
    }

    Debug.printf("beginning of MultimaxTimerManager initialize\n");

    /*
     * I hate to have to do the following but if I don't and there's
     *   a long period of time between the time the TimerManagers are
     *   created and the time they are initialized, a lot of unnecessary
     *   work will be done because they will think they've missed a
     *   lot of ticks and try to catch up.
     */
    extern TimerManager *TimeoutManager;
    if (TimeoutManager != 0)
	TimeoutManager->setFreeRunningTime(SystemTimer->time());

    extern TimerManager *PeriodicManager;
    if (PeriodicManager != 0)
	PeriodicManager->setFreeRunningTime(SystemTimer->time());

    /* Set the vector register */
    *SCCREG_PITVEC = ClassVector(INTERVAL_Vector, CPUCLASS_RUNNING);

    /* Load counter 0 pre-scaler to divide 1Mhz down to 100Khz. */
    /* Requires minimum clock frequency of 2 Hertz in order to  */
    /*    fit in the 16 bits of counter 1.			    */
    Assert(ClockFrequency >= 2);
    Assert(ClockFrequency <= 100000);

    *SCCREG_PITCTL = (char)(SC0 | TIM_MODE);
    *SCCREG_PITCNT0 = (char)10;
    *SCCREG_PITCNT0 = (char)0;

    /* Load counter 1 */
    *SCCREG_PITCTL = (char)(SC1 | TIM_MODE); /* this stops it */
    *SCCREG_PITCNT1 = (char) ((100000/ClockFrequency) & 0xff);
    *SCCREG_PITCNT1 = (char) (((100000/ClockFrequency) >> 8) & 0xff);

    initTimerLock.release();

    Debug.printf("end of MultimaxTimerManager initialize\n");
}
