#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Gary Johnston (johnston@cs.uiuc.edu)
 * Contributing Author: Dave Dykstra (dykstra@cs.uiuc.edu)
 * Contributing Author: Doug Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MultimaxTimesliceTimer.cc
 */

#include "Assert.h"
#include "OutputStream.h"
#include "MultimaxTimesliceTimer.h"
#include "NS32202.h"
#include "Vectors.h"


#ifdef __GNUC__
#pragma implementation
#endif

static const unsigned int kHz10 = 9216;	// Input clock frequency (kHz) * 10.

static const unsigned int MaxTimerMicroseconds = 466033777; // ~466 second max.
				    // Avoids overflow in MicrosecondsToCount

static inline unsigned int
MicrosecondsToCount( unsigned int microseconds )
{
	return( ((microseconds / 1000) * kHz10) / 10 );
	//return( (microseconds * kHz10) / 10000 );
}

static inline unsigned int
CountToMicroseconds( unsigned int count )
{
	return (((count * 100) / kHz10) * 100);
}

static const unsigned char CFRZ = (unsigned char) 0x80;		// MCTL.CFRZ
static const unsigned char NTAR = (unsigned char) 0x02;		// MCTL.NTAR
static const unsigned char MCTL_RESET = NTAR;

static const unsigned char CCON = (unsigned char) 0x80;		// CCTL.CCON
static const unsigned char CRUNH = (unsigned char) 0x08;	// CCTL.CRUNH
static const unsigned char CCTL_RESET = CCON;

static inline void
stopTimer()
{
	*MasterCCTL = CCTL_RESET;
}

static inline void
startTimer()
{
	*MasterCCTL = CCTL_RESET | CRUNH;
}

static inline void
freezeTimer()
{
	*MasterMCTL = MCTL_RESET | CFRZ;
}

static inline void
unfreezeTimer()
{
	*MasterMCTL = MCTL_RESET;
}

MultimaxTimesliceTimer::MultimaxTimesliceTimer( unsigned int vector )
{
	Assert( this != 0 );
	Assert( vector >= 0x10 );
	Assert( vector <= 0x1f );

	stopTimer();					// Stop.
	vector &= 0x0f;
	vec = vector | (vector << 4);
	Debug.printf("Timer::Timer: CIPTR %A --> %x\n", MasterCIPTR, vec);
	*MasterCIPTR = vec;				// Set vector.
}

MultimaxTimesliceTimer::~MultimaxTimesliceTimer()
{
	Debug.printf( "MultimaxTimesliceTimer::~MultimaxTimesliceTimer()\n" );

	stopTimer();					// Stop.
}

unsigned int
MultimaxTimesliceTimer::start( unsigned int microseconds )
{
	Debug << "MultimaxTimesliceTimer::start( "
		<< microseconds << " )\n";
	Assert( microseconds <= MaxTimerMicroseconds );

	/*
	 * Stop the timer (actually it should already be stopped).
	 * Load the current value registers with the count.
	 * Restart the counter.
	 */
	stopTimer();						// Stop.

	/*
	 * microseconds == 0 means dont start it after all.
	 * BUT: set the current count to zero so we know it isn't running.
	 */
	if( microseconds == 0 ) {
		Debug.printf( "TimersliceTimer::start: Not starting\n" );
		*MasterCCTL |= 0x04;
		return(0);
	}

	unsigned int count = MicrosecondsToCount( microseconds ); // Get count.
	Debug.printf("MultimaxTimesliceTimer::start: count %d\n", count);

	freezeTimer();
	/* set the counter start value */
	*MasterLCCVL = (unsigned char) (count         & 0xff);	// Load count.
	*MasterLCCVH = (unsigned char) ((count >> 8)  & 0xff);
	*MasterHCCVL = (unsigned char) ((count >> 16) & 0xff);
	*MasterHCCVH = (unsigned char) ((count >> 24) & 0xff);

	/* set the automatic reload value */
	*MasterLCSVL = (unsigned char) (count         & 0xff);	// Load count.
	*MasterLCSVH = (unsigned char) ((count >> 8)  & 0xff);
	*MasterHCSVL = (unsigned char) ((count >> 16) & 0xff);
	*MasterHCSVH = (unsigned char) ((count >> 24) & 0xff);
	unfreezeTimer();

	//startTimer();

	Debug.printf( "MultimaxTimesliceTimer::start() exiting \n" );
	return(CountToMicroseconds(count));
}

void
MultimaxTimesliceTimer::set( unsigned int microseconds )
{
	Assert(microseconds <= MaxTimerMicroseconds);
	Debug.printf( "MultimaxTimesliceTimer::set() entering \n" );

	/*
	 * Load the current value registers with the count.
	 */

	unsigned int count = MicrosecondsToCount( microseconds ); // Get count.
	Debug.printf("Timer::start: count %d\n", count);

	freezeTimer();
	*MasterLCCVL = (unsigned char) (count         & 0xff);	// Load count.
	*MasterLCCVH = (unsigned char) ((count >> 8)  & 0xff);
	*MasterHCCVL = (unsigned char) ((count >> 16) & 0xff);
	*MasterHCCVH = (unsigned char) ((count >> 24) & 0xff);
	unfreezeTimer();
	Debug.printf( "MultimaxTimesliceTimer::set() exiting \n" );
}

unsigned int
MultimaxTimesliceTimer::stop()
{
	Debug << "MultimaxTimesliceTimer::stop()\n" << eor;

	*MasterCCTL &= 0xf7;		// set not running.

	/*
	 * If there is a pending timer interrupt then clear it and
	 * return 0. This happens if we disable interrupts before we
	 * get a chance to stop the timer. The process running is
	 * credited with using the time anyway.
	 */
	if( *MasterCICTL & 0x40 ) {
		*MasterCICTL = 0;
		return( 0 );
	}

	freezeTimer();
	unsigned int count = *MasterLCCVL;		// Get count.
	count |= (*MasterLCCVH << 8);
	count |= (*MasterHCCVL << 16);
	count |= (*MasterHCCVH << 24);
	unfreezeTimer();
	return ( CountToMicroseconds(count) );	// Return microseconds.
}

/*
 * residual - Returns the number of microseconds left before interrupt.
 */

unsigned int
MultimaxTimesliceTimer::residual()
{
	Debug.printf( "MultimaxTimesliceTimer::residual() entering \n" );

	/*
	 * If there is a pending timer interrupt then clear it and
	 * return 0.
	 */
	if( *MasterCICTL & 0x40 ) {
		*MasterCICTL = 0;
		return( 0 );
	}

	freezeTimer();
	unsigned int count = *MasterLCCVL;		// Get count.
	count |= (*MasterLCCVH << 8);
	count |= (*MasterHCCVL << 16);
	count |= (*MasterHCCVH << 24);
	unfreezeTimer();
	Debug.printf( "MultimaxTimesliceTimer::residual() exiting \n" );
	return ( CountToMicroseconds(count) );	// Return microseconds.
}

/*
 * idle - returns non-zero iff the timer is not running.
 */

int
MultimaxTimesliceTimer::idle()
{
	Debug.printf( "MultimaxTimesliceTimer::idle()\n" );
	return ( (*MasterCCTL & CRUNH) == 0 );	// Return !running.
}
