#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */



/*
 * Vectors.c: Interrupt and trap vector table and access routines.
 */

#include "Assert.h"
#include "OutputStream.h"
#include "Vectors.h"


#ifdef __GNUC__
#pragma implementation
#endif

/*
 * Vector Bus Stuff.
 */
static unsigned volatile int * const CSR =  (unsigned int *) 0xffffff54;
static const unsigned int VectorBusFIFONotEmpty = 0x1 << 13;
static const unsigned int VectorBusIdle = 0x1 << 12;
static const unsigned int VectorBusClassShift = 10;
static const unsigned int VectorBusClassMask = 0x3 << VectorBusClassShift;
static unsigned volatile int * const VBTR = (unsigned int *) 0xffffff4c;

union VectorBusTransmitRegister {
	unsigned int all;
	struct {
		unsigned int CPUClass : 2; // Interrupt class.
		unsigned int deviceID : 2; // CPU ID.
		unsigned int slot     : 4; // Slot number/class.
		unsigned int LAMPID0  : 1; // LAMP ID 0.
		unsigned int LAMPID1  : 1; // LAMP ID 1.
		unsigned int LAMPREQ0 : 1; // LAMP request.
		unsigned int          : 5; // RESERVED by HW.
		unsigned int number   : 8; // Interrupt vector number.
		unsigned int LAMPID2  : 1; // LAMP ID 2.
		unsigned int LAMPID3  : 1; // LAMP ID 3.
		unsigned int LAMPREQ1 : 1; // Same as LAMPREQ0.
		unsigned int          : 5; // RESERVED by HW.
	} field;
};

unsigned int
SetCPUVectorClass( unsigned int newClass )
{
	unsigned int oldCSR = *CSR;
	unsigned int oldClass = ( oldCSR & VectorBusClassMask ) >>
			    VectorBusClassShift;
	Debug.printf( "SetCPUVectorClass: class: %d -> %d\n",
		oldClass, newClass );
	Assert( newClass <= CPUCLASS_MAX );
	Assert( oldClass <= CPUCLASS_MAX );
	unsigned int newCSR = ( oldCSR & ~VectorBusClassMask ) |
			  ( newClass << VectorBusClassShift) ;
	Debug.printf( "SetCPUVectorClass: csr: %x -> %x\n", oldCSR, newCSR );
	*CSR = newCSR;
	return ( oldClass );
}

unsigned int
CPUVectorClass()
{
	unsigned int csr = *CSR;
	unsigned int cpuClass =
		( csr & VectorBusClassMask ) >> VectorBusClassShift;
	Debug.printf( "CPUVectorClass(): %d\n", cpuClass );
	Assert( cpuClass <= CPUCLASS_MAX );
	return ( cpuClass );
}

void
TransmitVector( unsigned int vector )
{
        // Debug.printf( "TransmitVector(%x)\n", vector );

	/*
	 * Wait for the vector go to be idle if it isn't already.
	 */
	if ( ( *CSR & VectorBusIdle ) == 0 ) {
		//Debug.printf( "Vector bus busy\n" );
		unsigned int loopCount = 0;
		static const unsigned int maxLoopCount = 100000;
		while ( ( *CSR & VectorBusIdle ) == 0 ) {
			if ( loopCount == maxLoopCount ) {
				Console().panic( "Vector bus hung\n" );
			}
			loopCount++;
		}
		// Debug.printf( "Vector bus idle\n" );
	}

	/*
	 * Transmit the vector.
	 */
	/*
	 *Debug.printf( "TransmitVector: writing vector %x to %x\n",
	 *		  vector, VBTR );
	 */
	*VBTR = vector;
}

unsigned int
DirectedVector( unsigned int number, unsigned int CPUID )
{
	Assert( number >= 16 );
	Assert( number < 256 );
	Assert( CPUID <= 0x3f );
	VectorBusTransmitRegister vbtxr;
	vbtxr.all = 0;
	vbtxr.field.CPUClass = 0;
	vbtxr.field.deviceID = CPUID & 0x3;
	vbtxr.field.slot = CPUID >> 2;
	vbtxr.field.number = number;
	Debug.printf( "DirectedVector(%x,%x): %x\n", number, CPUID,
		vbtxr.all );
	return( vbtxr.all );
}

unsigned int
ClassVector( unsigned int number, unsigned int CPUClass )
{
	Assert( number >= 16 );
	Assert( number < 256 );
	Assert( CPUClass < 0x4 );
	Assert( CPUClass != 0 );
	VectorBusTransmitRegister vbtxr;
	vbtxr.all = 0;
	vbtxr.field.CPUClass = CPUClass;
	vbtxr.field.number = number;
	Debug.printf( "ClassVector(%x,%x): %x\n", number, CPUClass,
		vbtxr.all );
	return( vbtxr.all );
}

unsigned int
VectorNumber( unsigned int vector )
{
	VectorBusTransmitRegister vbtxr;
	vbtxr.all = vector;
	return( vbtxr.field.number );
}
