#ifdef __GNUC__
#pragma implementation
#endif

#include "OMAPDevices.h"
#include "OutputStream.h"
#include "OMAPInterruptManager.h"

unsigned OMAPInterruptManager::initialized = 0;

OMAPInterruptManager::OMAPInterruptManager() : InterruptManager(nInterrupts) {
    // Need some locking if we are initializing for multiple CPUs
    if (!initialized) {
        setupInterrupts();
        enableInterrupts();
    }
}

OMAPInterruptManager::~OMAPInterruptManager() {
    Debug << "OMAPInteruptManager destructor\n" << eor;
}

unsigned OMAPInterruptManager::getInterruptNumber() {
    // Read the registers in IH1 and IH2 (if necessary) to determine the interrupt number
    // Find source of interrupt
    // Look in IH1
    unsigned source1 = *reinterpret_cast<unsigned *>(OMAP_IH1_BASE+IH1_SIR_IRQ);
//  Debug << "Got an IH1 interrupt :" << source1 << "\n" << eor;

    if (source1 == 0) {
        unsigned source2 = *reinterpret_cast<unsigned *>(OMAP_IH2_BASE+IH2_SIR_IRQ);
//      Debug << "Got an IH2 interrupt :" << source2 << "\n\n" << eor;
        return source2 + 40;
    } else {
        return source1 + 8;
    }
}

void OMAPInterruptManager::ackInterrupt(unsigned number) {
    // Acknowledge the interrupt to IH1 and IH2
    if (number >= 8) {
        if (number >= 40) {
            *((unsigned *)(OMAP_IH2_BASE+IH2_CONTROL)) |= 1;
        }
        *((unsigned *)(OMAP_IH1_BASE+IH1_CONTROL)) |= 1;
    }
}


void OMAPInterruptManager::setupInterrupts() {
    interruptBanks[0].registers = OMAP_IH1_BASE;
    interruptBanks[0].triggers = 0xb3fefe8f;
    interruptBanks[1].registers = OMAP_IH2_BASE;
    interruptBanks[1].triggers = 0xfdb7c1fd;
    interruptBanks[2].registers = OMAP_IH2_BASE + 0x100;
    interruptBanks[2].triggers = 0xffffb7ff;
    interruptBanks[3].registers = OMAP_IH2_BASE + 0x200;
    interruptBanks[3].triggers = 0xffffffff;

    for (int i = 0; i < numInterruptBanks; i++) {
        for (int j = 0; j < 32; j++) {
            (reinterpret_cast<unsigned *>(interruptBanks[i].registers+IH1_ILR0))[j] &= ~0x7F;
            (reinterpret_cast<unsigned *>(interruptBanks[i].registers+IH1_ILR0))[j] |= ((interruptBanks[i].triggers >> j) & 0x1) << 1;
        }
    }

    initialized = 1;
}

void OMAPInterruptManager::enableInterrupts() {
    // Enable IH2
    Debug << "Enabling IH2\n" << eor;
    // Individual interrupt mask register
    // Masking all interrupts now
    *reinterpret_cast<unsigned *>(OMAP_IH2_BASE+IH2_MIR) = 0xFFFFFFFF;
    *reinterpret_cast<unsigned *>(OMAP_IH1_BASE+IH2_ITR) = 0x0;
    // Enable IRQ and set the Global Mask bit to 0
    *reinterpret_cast<unsigned *>(OMAP_IH2_BASE+IH2_CONTROL) &= ~0x6;
    *reinterpret_cast<unsigned *>(OMAP_IH2_BASE+IH2_CONTROL) |= 0x1;
    Debug << "Enabled IH2\n" << eor;

    // Set interrupt level to falling edge sensitive
    // clearing lowest bits sets highest priority, makes it falling edge sensitve and makes it IRQ instead of FIQ
    //*reinterpret_cast<unsigned *>(OMAP_IH2_BASE+IH2_ILR0) &= ~0x7F;

    Debug << "Enabling IH1\n" << eor;
    // Individual interrupt mask register
    // Masking all interrupts now
    *reinterpret_cast<unsigned *>(OMAP_IH1_BASE+IH1_MIR) = 0xFFFFFFFE;
    *reinterpret_cast<unsigned *>(OMAP_IH1_BASE+IH1_ITR) = 0x0;
    // Reset IRQ output and enables new IRQ generation - reset on interrupt
    *reinterpret_cast<unsigned *>(OMAP_IH1_BASE+IH1_CONTROL) |= 1;
    Debug << "Enabled IH1\n" << eor;

    // Set interrupt level to falling edge sensitive
    // clearing lowest bits sets highest priority, makes it falling edge sensitve and makes it IRQ instead of FIQ
    // Update - this is now part of setupInterrupts
    //*reinterpret_cast<unsigned *>(OMAP_IH1_BASE+IH1_ILR0) &= ~0x7F;
    //*reinterpret_cast<unsigned *>(OMAP_IH1_BASE+IH1_ILR0) |= 0x2;

    // Making last bit 0 - global interrupts enabled
    *reinterpret_cast<unsigned *>(OMAP_IH1_BASE+IH1_GMR) &= ~0x1;
}

void OMAPInterruptManager::disableInterrupts() {
    // Mask all interrupts in IH1 and IH2
    *reinterpret_cast<unsigned *>(OMAP_IH1_BASE+IH1_MIR) = 0xffffffff;
    *reinterpret_cast<unsigned *>(OMAP_IH2_BASE+IH2_MIR) = 0xffffffff;
    // Making last bit 1 - global interrupts disabled
    *reinterpret_cast<unsigned *>(OMAP_IH1_BASE+IH1_GMR) |= 0x1;
}

void OMAPInterruptManager::enableInterrupt(unsigned number) {
    if (number >= 8) {
        number -= 8;
        if (number < 32) {
            *reinterpret_cast<unsigned *>(OMAP_IH1_BASE+IH1_MIR) &= ~(1 << number);
        } else {
            number -= 32;
            *reinterpret_cast<unsigned *>(OMAP_IH2_BASE+IH2_MIR) &= ~(1 << number);
            enableInterrupt(8);
        }
    }
}

void OMAPInterruptManager::disableInterrupt(unsigned number) {
    if (number >= 8) {
        number -= 8;
        if (number < 32) {
            *reinterpret_cast<unsigned *>(OMAP_IH1_BASE+IH1_MIR) |= (1 << number);
        } else {
            number -= 32;
            *reinterpret_cast<unsigned *>(OMAP_IH2_BASE+IH2_MIR) |= (1 << number);
        }
    }
}

int OMAPInterruptManager::isPending(unsigned number) {
    if (number >= 8) {
        number -= 8;
        if (number < 32) {
            return (*reinterpret_cast<unsigned *>(OMAP_IH1_BASE+IH1_ITR) >> number) & 0x1;
        } else {
            number -= 32;
            return (*reinterpret_cast<unsigned *>(OMAP_IH2_BASE+IH2_ITR) >> number) & 0x1;
        }
    }
    return 0; // Never reached - keep compiler happy
}
