#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-2006
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David Raila (raila@cs.uiuc.edu)
 * Author: Boris Capitanu (capitanu@cs.uiuc.edu)
 * Author: Ellick Chan (emchan@cs.uiuc.edu)
 * Author: Francis David (fdavid@cs.uiuc.edu)
 * Author: Jeffrey Carlyle (jcarlyle@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * OMAPKernel.cc: OMAP kernel class implementation.
 */

#include "DummyMemoryObject.h"
#include "MachPROM.h"
#include "MemoryObject.h"
#include "OMAPKernel.h"
#include "PhysicalMemoryRange.h"
#include "RAMDisk.h"
#include "ARMTranslation.h"
#include "RAMDiskInfo.h"
#include "Vectors.h"

#ifdef RELIABILITY
#include "OMAPWatchdogTimerDD.h"
OMAPWatchdogTimerDD watchdog;
#else
#include "OMAPWatchdogTimer.h"
OMAPWatchdogTimer watchdog;
#endif

// If this is defined, a watchdog reset is translated into an exception object and
// this is thrown in the context of the process that was running/locked at the time
// of the reset
//#define WATCHDOG_EXCEPTION

// Same value in OMAP_ULPD_RESET_STATUS_REGISTER
enum OMAPResetReason {
    UNAVAILABLE = 0,
    POWER_ON_RESET = 1,
    SECURE_WD_TIMEOUT = 2,
    SECURITY_VIOLATION = 4,
    WD_TIMEOUT = 8,
    INVALID = 16
};

OMAPResetReason resetReason = INVALID;
extern void OMAPSerialPutS(const char *s);
extern void StartTimer(void);
extern GenericARMSerial* GenericARMSerialPort;

// Whatever context is saved by bootloader when CPU is reset
ARMInterruptContext watchdogContext; 

extern "C" void OMAPHandleReset() {
    // Read and log reset reason
    switch(*(unsigned *)OMAP_ULPD_RESET_STATUS_REGISTER & 0xf) {
        case POWER_ON_RESET:
            resetReason = POWER_ON_RESET;
            break;
        case SECURE_WD_TIMEOUT:
            resetReason = SECURE_WD_TIMEOUT;
            break;
        case SECURITY_VIOLATION:
            resetReason = SECURITY_VIOLATION;
            break;
        case WD_TIMEOUT:
            resetReason = WD_TIMEOUT;
            break;
        default:
            resetReason = UNAVAILABLE;
    }

    // Clear reset status
    *(unsigned *)OMAP_ULPD_RESET_STATUS_REGISTER = 0x0;

#ifndef RELIABILITY
    return;
#endif
    
    // We are on the boot Sys/Usr stack thats allocated and initialized in boot.s
    // We shouldn't need to clean up the activation on the stack (this func never returns)
    // because it isn't used anywhere

    // This needs a modified bootloader that will just jump to choices
    // when resetReason == WD_TIMEOUT
    if (resetReason == WD_TIMEOUT) {
        Process * process = thisCPU()->idleProcess();
        Assert( process != 0 );

        CPU * cpu = thisCPU();
        Assert (cpu != 0);
        cpu->chipInitialize();
        GenericARMSerialPort->Initialize();

        OMAPSerialPutS("Recovered after watchdog timeout\n");

        ((ARMMMU *)cpu->MMU())->setDisabled();

        Process *crashingProcess = CPU::_runningProcesses[PROCESSOR::cpuID()];
        CPU::_runningProcesses[PROCESSOR::cpuID()] = process;
        /*
         * Activate the proper domain on this processor.
         */
        cpu->MMU()->activate( process->domain());

        /*
         * Turn on the MMU()
         */
        cpu->MMU()->enable();

        // Turn on interrupts
        OMAPInterruptManager *interruptManager = (OMAPInterruptManager *)thisCPU()->getInterruptManager();
        interruptManager->reinit();
        StartTimer();
        interruptManager->enableInterrupt(VEC_TIMER1);
        interruptManager->enableInterrupt(VEC_TIMER2);

        GenericARMSerialPort->enableInterrupts();
        interruptManager->enableInterrupt(VEC_UART);

        volatile unsigned *dest = watchdogContext.getRegisters();
        extern unsigned regsave;
        unsigned *src = &regsave;

        // r0 and r1 are technically available after reset in the bootloader
        // But their contents are clobbered in order to save the rest of the register set

        for( unsigned i = 0; i < 13; i++)
            dest[2+i] = src[i];

        // dest[15] is PC - This is irrecoverably lost
        // We approximate this with the first instruction in the current function
        dest[15] = *((unsigned *)dest[11])+0x4;

        // Set CPSR to ARM_SYS mode with Interrupts and FIQs enabled
        watchdogContext.setCPSR(ARM_SYS);
        
        watchdog.start(FRTuSecPerInterrupt/1000000 + 1);

        CPU::_runningProcesses[PROCESSOR::cpuID()] = crashingProcess;
#ifdef WATCHDOG_EXCEPTION
        // Should never return
        ARMWatchdogTimeoutInterrupt(&watchdogContext); 
#else
        // Kill the process that was running
        // Should never return
        crashingProcess->die();
#endif

        Assert(NOTREACHED);
    }
}

extern unsigned __CTOR_LIST__;
unsigned reloadKernelFromFlash(unsigned address) {
    unsigned flashBase = 0x00100000;
    unsigned ramBase = 0x10000000;

    // Map in using section table entries
    unsigned* ttb = (unsigned *)((TwoLevelPageTable *)thisProcess()->domain()->translation())->pointerTable();

    unsigned ttb1 = ttb[1];
    unsigned ttb2 = ttb[2];
    ttb[1] = 0x00100000 | 0xC12;
    ttb[2] = 0x00200000 | 0xC12;

    if (address) {
        if (((address - ramBase) > 0x200000)||(address < ramBase)) {
            ttb[1] = ttb1;
            ttb[2] = ttb2;
            return 0;
        }
        *(unsigned *)address = ((unsigned *)flashBase)[(address-ramBase)/4];
    } else {
        unsigned kernelReadOnlySize = (__CTOR_LIST__ - ramBase)/4;
        
        // We need to load kernelReadOnlySize words 
        // from flashBase to ramBase
        for (unsigned i = 0; i < kernelReadOnlySize; i++)
            ((unsigned *)ramBase)[i] = ((unsigned *)flashBase)[i];
    }
    ttb[1] = ttb1;
    ttb[2] = ttb2;

    ((ARMMMU *)thisCPU()->MMU())->cleanDCache();
    ((ARMMMU *)thisCPU()->MMU())->flushICache();

    return 1;
}

#include "CRC.h"
unsigned long getCRC32(unsigned address, unsigned numBytes) {
    // address to address+size should be in Virtual Memory
    Crc::Type const key = 0x04c11db7;
    FastCrc fastCrc (key);

    while(numBytes--)
    {
        fastCrc.PutByte ( (unsigned)((char *)address)[numBytes] );
    }

    return (unsigned long)fastCrc.Done ();
}

OMAPKernel::OMAPKernel()
{
    OMAPPROM *prom = new OMAPPROM();

    Console() << "TI OMAP ";

    switch (prom->omap32_id >> 16) {
        case 0x332:
            Console() << "1610";

            switch (prom->omap32_id & 0xffff) {
                case 0x100:
                    switch (prom->dev_rev) {
                        case 0x2:
                            Console() << "A";
                            break;

                        case 0x3:
                            Console() << "C";
                            break;

                        default:
                            Console() << "?";
                            break;
                    };
                    break;

                case 0x200:
                    Console() << "D";
                    break;

                case 0x300:
                    Console() << "E";
                    break;

                default:
                    Console() << "?";
                    break;
            };
            break;

        default:
            Console() << asHex(prom->omap32_id);
            break;
    };

    DisplayCpuInfo(prom);

    Console() << "Reset Reason: ";
    switch(resetReason) {
        case POWER_ON_RESET:
            Console() << "Power On\n";
            break;
        case SECURE_WD_TIMEOUT:
            Console() << "Secure Watchdog Timeout\n";
            break;
        case SECURITY_VIOLATION:
            Console() << "Security Violation\n";
            break;
        case WD_TIMEOUT:
            Console() << "Watchdog Timeout\n";
            break;
        case UNAVAILABLE:
            Console() << "Not Available\n";
            break;
        default:
            Console() << "OMAPResetHandler was not run!\n" << eor;
            while(1);
    }
}

PhysicalMemoryRange *
OMAPKernel::physicalMemoryMap()
{
    Debug << "OMAPKernel::physicalMemoryMap()\n" << eor;

    PhysicalMemoryRange *_physicalMemoryMap =
        GenericARMKernel::physicalMemoryMap();

    // NOTE: This function only causes PTE to be allocated. It doesn't
    // do any memory mapping. Use basicAddMemoryObjects() or
    // basicInitializeDevicesAndManagers() below for the actual mapping.

    // Add entries for Memory Mapped I/O region
    _physicalMemoryMap->add(
        new PhysicalMemoryRange( (char *)0xfff00000,
                                 0x100000/PhysicalPageSize, Log2PhysicalPageSize,
                                 "#MemoryMapped I/O", 1 ));

    // For virtual disk
    _physicalMemoryMap->add(
        new PhysicalMemoryRange( (char *)__CHOICES_VDISK_START__,
                                 __CHOICES_VDISK_SIZE__/PhysicalPageSize, Log2PhysicalPageSize,
                                 "#disk image", 1 ));

    Debug << "Exiting OMAPKernel::physicalMemoryMap()\n" << eor;
    return _physicalMemoryMap;
}

const unsigned 
OMAPKernel::kernelHeapSize()
{   
    return (const unsigned) ((_physicalMemorySize * 8) / 10);
}

/**
 * Map virtual memory for upper interrupt region and memory mapped devices.
 * Consider using basicInitializeDevicesAndManagers().
 */
void
OMAPKernel::basicAddMemoryObjects()
{
    Debug << this << "::" << __FUNCTION__ << "()\n" << eor;

    // Add MemoryObject for Serial I/O and other peripherals
    PhysicallyResidentMemoryObject *mo = new PhysicallyResidentMemoryObject(
        0xf0000/PhysicalPageSize, Log2PhysicalPageSize,
        (char *)0xfff00000, 1);
    domain()->add(mo, (char *)0xfff00000, NoAccess);
    domain()->makeAddressable((char *)0xfff00000, 0xf0000);
    ((ARMTranslation *)domain()->translation())->inhibitDataCaching((char *)0xfff00000,0xf0000);

    // For virtual disk
    mo = new PhysicallyResidentMemoryObject(
        0x100000/PhysicalPageSize, Log2PhysicalPageSize,
        (char *)0x11f00000, 1);
    domain()->add(mo, (char *)0x11f00000, NoAccess);
    domain()->makeAddressable((char *)0x11f00000, 0x100000);

    // For 2MB of NOR Flash starting at 0x100000
    // Only for read access
    // This doesn't work for some reason
    /*mo = new PhysicallyResidentMemoryObject(
        0x200000/PhysicalPageSize, Log2PhysicalPageSize,
        (char *)0x100000, 1);
    domain()->add(mo, (char *)0x100000, NoAccess);
    domain()->makeAddressable((char *)0x100000, 0x200000);*/
}

/*
 * Setup machine dependent devices and device managers.
 */
void
OMAPKernel::basicInitializeDevicesAndManagers()
{
    Debug << "OMAPKernel::basicInitializeDevicesAndManagers()\n" << eor;

    GenericARMKernel::basicInitializeDevicesAndManagers();

    /*
     * Now that interrupts are available, set up the Console
     * for interrupts.
     */
    EnableConsoleInterrupts();

    /*
     * Start the Timeout timer & Periodic timer helper processes.
     *
     */
    InstallTimerProcesses();

    /*
     * Build the time of day clock
     */
    InstallTimeOfDayClock();

    /*
     * Now install software interrupt manager
     */
    InstallSoftwareInterruptManager();

    /*
     * Now install software clock manager
     */
    InstallSoftwareClockManager();

    /*
     * Start up the disk.
     */
    InstallTheRamDisk((char *)__CHOICES_VDISK_START__);

    /*
     * Start the watchdog timer
     * Will bite if FRT doesn't interrupt normal period + 1 second
     */
    watchdog.start(FRTuSecPerInterrupt/1000000 + 1);
}

void infloop() {
    Console() << "Hello!\n" << eor;
    while(1);
}

void crashKernel() {
    Console() << "Crashing kernel\n" << eor;
    PROCESSOR::disableInterrupts();
    try {
        infloop();
    } catch ( ProcessorException *e ) {
        Console() << "Caught Exception " << *e << ":"
                  << e->getMessage() << "\n" << eor;
        e->displayContext();
        //e->displayStackTrace();
        delete e;
    }
    thisProcess()->die();
}

void codeReloadTest() {
    Console() << "Reload successful!\n" << eor;
}

int
OMAPKernel::tryCommand( int argc, char *argv[], int partnumber)
{
    switch(partnumber) {
        case 1243:
        {
            Process *process = new SystemProcess(crashKernel, TheKernel->domain(), normal, (int) 0 );
            Assert( process != 0 );
            process->setName( "#BadProcess" );
            process->ready();
            break;
        }
        case 1244:
        {
            Console() << "Corrupting instruction in boot.s..." << eor;
            *(unsigned *)0x1000021c = 0xcafebabe;
            break;
        }
        case 1245:
        {
            Console() << "Performing code reload test...\n" << eor;
            ((unsigned *)&codeReloadTest)[0] = 0xdeadbeef;
            codeReloadTest();
            break;
        }
        default:
            return GenericARMKernel::tryCommand(argc, argv, partnumber);
    }
    return 1;
}


#include "ClassInstaller.h"
DEFINECLASS("OMAPKernel",OMAPKernel,GenericARMKernel,0,0);
