/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *			   University of Illinois at Urbana-Champaign
 *			   Department of Computer Science
 *			   1304 W. Springfield Ave.
 *			   Urbana, IL	 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David W. Dykstra (dykstra@cs.uiuc.edu)
 * Contributing Author: See-Mong Tan (stan@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * DatakitDriver.c: AT&T 6586 CPM-HS Datakit driver
 */

#include "OutputStream.h"
#include "PCi586InterruptController.h"
#include "Vectors.h"
#include "Util.h"
#include "CPU.h"
#include "Interrupt.h"
#include "ThisProcess.h"
#include "DatakitDriver.h"
#include "DatakitChannel.h"
#include "Semaphore.h"
#include "PeriodicTimer.h"
//#include "DatakitUnixHost.h"
//#include "DatakitCon.h"

static const int DK_CSR_ADD =   0x202;
static const int DK_DATA_ADD =  0x204;
static const int DK_REARM_ADD = 0x206;

// bits for writing the CSR
static const int FULLRESET = 0x00;	// full reset of of hardware
static const int ALIVE_C = 0x01;	// alive set mask
static const int HLOOPR_C = 0x02;	// host loop request set mask
static const int INTENB_C = 0x04;	// Interrupt enable set mask
static const int FORCE_C = 0x08;	// force set mask
static const int INTCLR	= 0x80;		// clear interrupt request
static const int RESETBLK = 0xff00;	// used to block resets in csr
static const int XMTRES	= 0x0100;	// transmit fifo reset, active low
static const int BRDRES	= 0x0200;	// Board reset, active low
static const int RCVRES	= 0x0400;	// Receive fifo reset, active low

// bits for reading the CSR
static const int ALIVE_S = 0x01;	// alive test mask
static const int HLOOPR_S = 0x02;	// Host Loop Request test mask
static const int INTENB_S = 0x04;	// Interrupt enable	test mask*/
static const int EXOFLOW = 0x08;	// Xmit fifo overflow error
static const int IRQIN = 0x10;		// Interrupt request is pending, active HIGH
static const int MUTE = 0x20;		// Mute indicator, no activity from cpm-hs
static const int ERPARITY = 0x40;	// parity error
static const int ERFE = 0x80;		// frame error
static const int EROFLOW = 0x0100;	// Receive fifo overflow error
static const int FORCE_S = 0x0200;	// force test mask
static const int RHF = 0x0400;		// Receive fifo half-full flag, active low
static const int REF = 0x0800;		// Receive fifo empty flag, active low
static const int RFF = 0x1000;		// Receive fifo full flag, active low
static const int XEF = 0x2000;		// Transmit fifo empty flag, active low
static const int XFF = 0x4000;		// Transmit fifo full flag, active low
static const int XHF = 0x8000;		// Transmit fifo half-full flag, active low

// data register bits
static const int CHANMARK = 0x0200;	// Datakit Channel Mark
static const int FLUSH = 0x0400;	// flush previous output
static const int WCNTLMK = 0x0100;	// Write Control
static const int RDATAMK = 0x0100;	// Read Data
static const int EMPTY = 0x4000;	// fifo empty
static const int NND = 0x8000;		// Not Normal Data Mark

static const int XDEPTH = 4096;		// transmit fifo max depth
static const int RDEPTH = 4096;		// Receive fifo max depth

static const int DKENABS = (INTENB_C | RESETBLK| ALIVE_C);

static const int TIMERSEC = 3;		// period of "unstick" timer

DatakitChannel * DatakitChan[MAXCHAN];

class InterfaceStats {
public:
	int parity;
	int frame;
	int recvfifo;
	int xmitfifo;
	int stray;
	InterfaceStats();
};

static inline
InterfaceStats::InterfaceStats()
{
	parity = 0;
	frame = 0;
	recvfifo = 0;
	xmitfifo = 0;
	stray = 0;
}

InterfaceStats DatakitInterfaceStats;

// keeps a running count of how many times the driver got stuck -
// ie something in the queue, but no interrupt got caught
static int datakitStuckCount = 0;

static const int MAXERRORS = 15;

static inline void 
checkError(unsigned short csr)
{
	if (csr & (ERPARITY | ERFE | EROFLOW | EXOFLOW)) {
		if (csr & ERPARITY && !(++DatakitInterfaceStats.parity % MAXERRORS))
			Console().printf("DatakitDriver: Parity errors\n");
		if (csr & ERFE && !(++DatakitInterfaceStats.frame % MAXERRORS))
			Console().printf("DatakitDriver: Frame errors\n");
		if (csr & EROFLOW && !(++DatakitInterfaceStats.recvfifo % MAXERRORS))
			Console().printf("DatakitDriver: Recv FIFO overflow errors\n");
		if (csr & EXOFLOW && !(++DatakitInterfaceStats.xmitfifo % MAXERRORS))
			Console().printf("DatakitDriver: Xmit FIFO overflow errors\n");
	}
}

static inline void
DKOUT(int addr, int com)
{
#ifdef TRACEDKIO
	Debug.printf("DKOUT: %x, %x\n",addr,com);
#endif
	Outw(addr, com);
}

static inline unsigned short
DKIN(int addr)
{
	register unsigned short ret = Inw(addr);
#ifdef TRACEDKIO
	Debug.printf("DKIN:  %x, %x\n",addr,ret);
#endif
	return(ret);
}

// true if the recive
static inline int
DKReceiveFIFOEmpty()
{
    short c = DKIN(DK_CSR_ADD);
    Debug << "DKReceiveFIFOEmpty( csr=" << asHex((int) c) << " )\n" << eor;
    return ((c & REF) == 0);
}

static LOCK * datakitControllerLock;

class DatakitInterrupt : public Interrupt {
protected:
	void basicRaise( char * );
public:
	DatakitInterrupt();
};

static inline
DatakitInterrupt::DatakitInterrupt() :
		Interrupt()
{
	Debug << "DatakitInterrupt::DatakitInterrupt()\n" << eor;
}

void
DatakitInterrupt::basicRaise( char * )
{
    Debug.printf( "DatakitInterrupt::basicRaise()\n" );

    datakitStuckCount = 0;
    checkError(DKIN(DK_CSR_ADD));

    datakitControllerLock->acquire();

    do {
	DatakitInterruptHandler();
	DKOUT(DK_REARM_ADD, 0);              // re-arm interrupt

	// If a byte comes in to the 586HS when interrupts are being
	// re-armed, no interrupt will be generated.  Therefore we
	// must check the CSR to make sure that the FIFO isn't empty.
	// Otherwise we'll miss all further packets... bad deal.
    } while( ! DKReceiveFIFOEmpty() );

    Out(PIC2_CTRL_REG, PIC_EOI);
    Out(PIC_CTRL_REG, PIC_EOI);

    datakitControllerLock->release();
}

void
DatakitDriverTimerEntry( int )
{
    Debug << "DatakitDriverTimerEntry()\n" << eor;
    PeriodicTimer * driverTimer = new PeriodicTimer();
    Assert( driverTimer != 0 );

    driverTimer->start( TIMERSEC * 1000000 );
    for(;;) {
	driverTimer->await();
	Debug << "DatakitDriverTimerEntry: waking up\n" << eor;

	datakitControllerLock->acquire(); // grab the controller
	
	int csr = DKIN(DK_CSR_ADD);
	
	if ((csr & MUTE) && (++datakitStuckCount % (600/TIMERSEC)) == 2) {
	    /* repeat message every 10 minutes */
	    Console().printf("DatakitDriver: Unable to communicate with controller! (Receiver down)\n");
	}
	else {
	    checkError(csr);
	    /* is there something in the queue but no interrupt request? */
	    if ((csr & REF) && (!csr & IRQIN) && (++datakitStuckCount > 1)) {
		Debug << "DatakitDriver: missed interrupt, unsticking\n"
		    << eor;
		datakitStuckCount = 0;
		DatakitInterfaceStats.stray++;

		DatakitInterruptHandler();
	    }
	}   
	// release access to the controller
	datakitControllerLock->release();
    }
}	   

void
DatakitDriverProcessEntry( int )
{
    Debug << "DatakitDriverProcessEntry()\n" << eor;

    // clear the global array of channels
    for( int i = 0; i < MAXCHAN; i++ )
	DatakitChan[i] = 0;

    for (int j = 0; j < 2; j++) {
	/* two resets */
	DKOUT(DK_CSR_ADD, FULLRESET);

	for (int i = 0; i < 3500; i++)
	    /* little delay for reset to go to work */
	    ;
    }

    (void) DKIN(DK_CSR_ADD);		/* ignore first time */

    // DKOUT(DK_CSR_ADD, DKENABS);

    unsigned short csr;
    
    if ( (csr = DKIN(DK_CSR_ADD)) & ALIVE_S ) {
	Console().printf("DatakitDriver: CPM is dead\n");
    }
    if (csr & MUTE) {
	Console().printf("DatakitDriver: No optical connectivity\n");
    }

    /* set CPM to known state */
    DKOUT(DK_CSR_ADD, DKENABS | FORCE_C);

    // this Lock makes sure of exclusive access to the controller
    datakitControllerLock = new LOCK;

//	Debug << "DatatkitDriverProcessEntry: Spawning Control Channel entry\n"
//	  << eor;
//	extern void DatakitControlChannelEntry(int);
//	Process * controlChannelProcess = 
//	  new SystemProcess( (PFV) &DatakitControlChannelEntry,
//					  thisProcess()->domain(), normal, 0 );
//	Assert( controlChannelProcess != 0 );
//	controlChannelProcess->ready();

    // create a process to wake us up once in a while
//    Debug << "DatakitDriverProcessEntry: Spawning Timer Process\n" << eor;
//    extern void DatakitDriverTimerEntry( int );
//    Process * datakitTimerProcess =
//	new SystemProcess( (PFV) &DatakitDriverTimerEntry,
//			  thisProcess()->domain(), normal, 0 );
//    Assert( datakitTimerProcess != 0 );
//    datakitTimerProcess->ready();
    
    // The interrupt coming from the datakit 586HS board
    Interrupt * interrupt = new DatakitInterrupt();
    thisCPU()->setInterrupt(DK_Vector, interrupt);
    
    /* enable DK interrupts */
    Out(PIC2_MASK_REG, In(PIC2_MASK_REG) & ~PIC2_DK_MASK);
}

// chewdata -- return the next chanmark or no data character
static unsigned short
chewdata()
{
	unsigned short c;

	do
		c = DKIN(DK_DATA_ADD);
	while (c != 0xffff && !(c & CHANMARK));
	return c;
}

void
DatakitInterruptHandler()
{
    register unsigned char *p;
    register DatakitPacket *pack;
    register unsigned short c;
    DatakitChannel *channel;
    short chan = MAXCHAN;
    
    Debug << "DatakitInterruptHandler()\n" << eor;

    pack = 0;
    c = DKIN(DK_DATA_ADD);		// this should be the channel number
    while (c != 0xffff) {		// 0xffff is garbage, null stuff
	if (!(c & CHANMARK)) {
	    Debug << "DatakitDriver: expected chanmark, got " << c << "\n" 
		<< eor;
	    
	    c = chewdata();
	    continue;
	}
	
	if ((c &= 0x1ff) != chan) {
	    chan = c;
	    Debug << "DatakitDriver: new channel=" << chan << "\n" << eor;
	    if (chan < 0 || chan >= MAXCHAN || 
		(channel = DatakitChan[chan]) == 0) {
		c = chewdata();
		continue;
	    }
	    
	}
	
	int ctlcalls = channel->ctlCalls();
	for (;;) {
	    if (pack == 0) {
		if ((pack = new DatakitPacket()) == 0) {
		    Console().printf("DatakitDriver: can't allocate packet\n");
		    c = chewdata();
		    break;
		}
		p = pack->buf;
	    }
	    Assert( p != 0 );
	    c = DKIN(DK_DATA_ADD);
	    while( !(c & NND) ) {            // fill packet while normal data
		*p++ = (unsigned char) (c &~ NND);
		c = DKIN(DK_DATA_ADD);
	    }
	    if ((c & CHANMARK) || ctlcalls) {
		if ((pack->len = p - pack->buf) > 0) {
		    if (channel->ReceivePacket(pack))
			pack = 0;
		    else {	// recycle the packet
			pack->hibits = 0;
			p = pack->buf;
		    }
		}
	    }
	    else if ( ! ctlcalls ) {
		// to fix reading VCS_READ and VCS_WRITE in answer from 
		// the switch
		// just stuff it in the packet anyway
		*p++ = (c & ~NND);
		continue;
	    }
	    
	    if (c & CHANMARK)
		break;
	    if (ctlcalls)
		channel->ReceiveControl(c &~ NND);
	    else
		// set the 9th bit
		pack->hibits |= 0x8000 >> (p - pack->buf);
	}
	
    }
    
    if (pack != 0)
	delete pack;

    Debug << "Leaving DatakitInterruptHandler\n" << eor;
}

static LOCK xmitLock;

void
DatakitSendControl(int chan, int val)
{
	xmitLock.acquire();

	DKOUT(DK_DATA_ADD, chan | CHANMARK);
	DKOUT(DK_DATA_ADD, val | WCNTLMK);
	DKOUT(DK_DATA_ADD, FLUSH);

	xmitLock.release();
}

void
DatakitSendRawPacket( DatakitPacket * p )
{
    xmitLock.acquire();

    unsigned char * buf = &p->buf[0];
    unsigned char * endbuf = &p->buf[p->len];
    short hibits = (short) p->hibits;
    while ( buf < endbuf ) {			  // dump data packet to device
	if ( hibits < 0 )			 // make sure hibits go first
	    DKOUT( DK_DATA_ADD, *buf++ | WCNTLMK );
	else
	    DKOUT( DK_DATA_ADD, *buf++ );
	hibits <<= 1;
    }
    DKOUT( DK_DATA_ADD, FLUSH );

    xmitLock.release();
}

void
DatakitSendPacket( int chan, DatakitPacket * p, int /* waitcount */ )
{
    xmitLock.acquire();

    DKOUT(DK_DATA_ADD, chan | CHANMARK);

    unsigned char * buf = &p->buf[0];
    unsigned char * endbuf = &p->buf[p->len];
    short hibits = (short) p->hibits;
    while ( buf < endbuf ) {			  // dump data packet to device
	if ( hibits < 0 )			 // make sure hibits go first
	    DKOUT( DK_DATA_ADD, *buf++ | WCNTLMK );
	else
	    DKOUT( DK_DATA_ADD, *buf++ );
	hibits <<= 1;
    }
    DKOUT( DK_DATA_ADD, FLUSH );

    xmitLock.release();
}

// Transmit functions -- allow higher level packets to be created and flushed when done

int
DatakitTransmitStart(int chan)
{
	xmitLock.acquire();
	DKOUT(DK_DATA_ADD, chan | CHANMARK);
	return 0;
}

int
DatakitTransmitEnd()
{
	DKOUT(DK_DATA_ADD, FLUSH);
	checkError(DKIN(DK_CSR_ADD));
	xmitLock.release();
	return 0;
}

// xmitLock should already be acquired
int
DatakitTransmitData(unsigned char *buf, int len)
{
	unsigned char *endbuf;
	for (endbuf = buf + len; buf < endbuf; buf++)
		DKOUT(DK_DATA_ADD, *buf);
	return len;
}

int
DatakitTransmitControl(int val)
{
	DKOUT(DK_DATA_ADD, val | WCNTLMK);
	return 1;
}

/*******************
 *  the rest is machine independent and should go in another file
 */

static const int CMCSIZ = 12;
static unsigned char init_mesg1[CMCSIZ] = { 4, 1, 0, 0, '@', 0, 4, 0, 0, 0, 0, 0 };
static unsigned char init_mesg2[CMCSIZ] = { 3, 1, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
static unsigned char keep_alive[CMCSIZ] = { 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

//void
//DatakitControlChannelEntry( int )
//{
//#ifdef HOST
//	DatakitURPChannel * ControlChannel = new DatakitURPChannel(1);
//
//	Assert(ControlChannel != 0);
//	Assert(ControlChannel->channelOpened());
//
//	Debug << "DatakitControlChannelEntry: writing init_mesgs\n" << eor;
//	ControlChannel->Write(init_mesg1, CMCSIZ);
//	ControlChannel->Write(init_mesg2, CMCSIZ);
//
//	Debug << "DatakitControlChannelEntry: reading first message\n" << eor;
//	unsigned char buf[DKPSIZ];
//	int n = ControlChannel->Read(buf, DKPSIZ);
//	Debug << "DatakitControlChannelEntry: read " << n << " bytes\n" << eor;
//
//#else
//	DatakitSwitch * ControlChannel = new DatakitSwitch;
//
//	Assert( ControlChannel != 0 );
//	Assert( ControlChannel->channelOpened() );
//
//	Debug << "DatakitControlChannelEntry: reading location 9\n" << eor;
//	int val = ControlChannel->ReadLocation(9, 0);
//	Debug << "DatakitControlChannelEntry: returned " << val << "\n" << eor;
//
//	Debug<< "DatakitControlChannelEntry: reading location 0\n" << eor;
//	val = ControlChannel->ReadLocation(0, 0);
//	Debug << "DatakitControlChannelEntry: returned " << val << "\n" << eor;
//
//	int mod = 14;
//	int ccdhn = 1;
//	int iscommand = 0;
//	int ctrldata = 0;
//
//	Debug.printf("DatakitControlChannelEntry: GetChannel(%d,%d,%d,%d)\n",
//						mod,chn,iscommand,ctrldata);
//	val = ControlChannel->GetChannel(mod, chn, iscommand, ctrldata);
//	Debug.printf("DatakitControlChannelEntry: returned %d, (%d,%d,%d)\n",
//				val, mod, chn, iscommand);
//
//	mod = 2;
//	chn = 0;
//	Debug.printf("DatakitControlChannelEntry: GetCount(%d,%d,%d)\n",
//				mod, chn, ctrldata);
//	val = ControlChannel->GetCount(mod, chn, ctrldata);
//	Debug.printf("DatakitControlChannelEntry: returned %d\n", val);
//
//	mod = 14;
//	chn = 3;
//	int count = 5;
//	Debug.printf("DatakitControlChannelEntry: SetCount(%d,%d,%d,%d)\n",
//				mod, chn, count, ctrldata);
//	val = ControlChannel->SetCount(mod, chn, count, ctrldata);
//	Debug.printf("DatakitControlChannelEntry: returned %d\n", val);
//	chn = 2;
//	int mod2 = 2;
//	int chn2 = 461;
//	iscommand = 0;
//	Debug.printf("DatakitControlChannelEntry: SetChannel(%d,%d,%d,%d)",
//														mod,chn,mod2,chn2);
//	Debug.printf("%d,%d)\n", iscommand,ctrldata);
//	val = ControlChannel->SetChannel(mod, chn, mod2, chn2, iscommand, ctrldata);
//
//#endif
//	// it don't do too much yet!
//	thisProcess()->die();
//}
//
