/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David W. Dykstra (dykstra@cs.uiuc.edu)
 * Contributing Authors: See-Mong Tan (stan@cs.uiuc.edu)
 *                       Sean M. Dorward (smd@research.att.com)
 *                       Anand Iyengar (anand@research.att.com)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * DatakitSwitch.c: Per-channel Datakit processing; switch implementation
 */

#include "Assert.h"
#include "DatakitDriver.h"
#include "DatakitSwitch.h"
#include "OutputStream.h"
#include "Urp.h"
#include "LOCK.h"
#include "Semaphore.h"
#include "PeriodicTimer.h"
#include "Util.h"
#include "ThisProcess.h"

/*
 * Concentrator modules are not supported yet; in Radian they logically
 *  get modules 129-512 even though the switch thinks they're all in the
 *  the module of the concentrator and they get a range of channels there.
 */
static const int MAXMODS = 128;

class DatakitSwitchAddress {
public:
    int dataaddr;
    int controladdr;
    int channels;
};

DatakitSwitchAddress DatakitSwitchAddr[MAXMODS];

DatakitSwitch::DatakitSwitch()
{
    Debug << "DatakitSwitch::DatakitSwitch()\n" << eor;
    chan = 1;
    if (DatakitChan[chan] != 0) {
	chan = MAXCHAN;
	return;
    }
    Debug << this << ": Switch Channel on " << chan << "\n" << eor;
    ctlcalls = 0;	// no URP protocol
    waiting = 0;
    returnedPacket = 0;
    sendPacket.hibits = 0x8000;
    mutex = new Semaphore(1);

    waitSemaphore = new Semaphore( 0 );    // P on this waits for reply
    timerWakeup = new Semaphore( 0 );      // timer waits till called

    // timeout and retry process
    Process * proc = new SystemProcess( (PFV) DatakitSwitch::timerHandlerEntry,
				       thisProcess()->domain(),
				       normal, (int) this );
    Assert( proc != 0 );
    proc->ready();
     
    for (int i = 0; i < MAXMODS; i++)
	DatakitSwitchAddr[i].channels = -1;
    DatakitChan[chan] = this;
}

DatakitSwitch::~DatakitSwitch()
{
    if (chan != MAXCHAN)
	DatakitChan[chan] = 0;
}

void
DatakitSwitch::timerHandlerEntry()
{
    PeriodicTimer * timer = new PeriodicTimer;
    Debug << this << ": timer=" << asHex((int) timer) << ", waiting=" <<
	waiting << ", chan=" << chan << ", sendpacket=" << 
		asHex((int) &sendPacket) << ", waitSemaphore=" << 
			waitSemaphore << "\n" << eor;
    timer->start( 100000 );
    for(;;) {
	timerWakeup->P();       // wait till I'm needed
	for( int i=0; i < 6 ; i++ ) {
	    timer->await();
	    if ( waiting ) 
		DatakitSendPacket( chan, &sendPacket, 20 );
	    else
		break;
	}
	if ( i == 6 )
	    waitSemaphore->V();
    }
}

// ReceivePacket -- get a packet; return true if packet consumed
int
DatakitSwitch::ReceivePacket( DatakitPacket *packet )
{
    Debug << this << "::ReceivePacket()\n" << eor;

    if (waiting == 0) {
	Debug << this << ": unexpected packet received\n" << eor;
	return 0;
    }
    if (returnedPacket != 0) {
	Debug << this << ":DatakitSwitch: packet overrun\n" << eor;
	delete returnedPacket;
    }
    returnedPacket = packet;
    waitSemaphore->V();

    return 1;
}

void
DatakitSwitch::sendAndWait()
{
    Debug << this << "::sendAndWait()\n" << eor;

    Assert(waiting == 1);

    DatakitSendPacket(chan, &sendPacket, 20);

    timerWakeup->V();      // wakeup the retry and timeout process
    waitSemaphore->P();    // go to sleep and wait for a reply
}

int
DatakitSwitch::ReadLocation(int addr, int evenodd)
{
    Debug << this << "::ReadLocation( addr=" << addr << ", evenodd="
	<< evenodd << " )\n" << eor;
    mutex->P();

    Assert( waiting == 0 );

    sendPacket.buf[0] = 0x76;
    if (evenodd)
	sendPacket.buf[1] = (addr >> 8) | 0x80;
    else
	sendPacket.buf[1] = addr >> 8;
    sendPacket.buf[2] = addr;
    sendPacket.len = 3;

    waiting = 1;

    sendAndWait();

    int val;

    if (returnedPacket == 0) {
	Console().printf("DatakitSwitch: Can't read switch location %d\n", addr);
	val = -1;
    }
    else {
	// Verify structure of returned packet, channel numbers should
	// be same as bottom part
	unsigned char * p = &returnedPacket->buf[3];
	if ( *(p-3) != sendPacket.buf[0] || *(p-2) != sendPacket.buf[1] 
	     || *(p-1) != sendPacket.buf[2] ) { 
	    Console().printf("DatakitSwitch::ReadLocation: bad return structure\n");
	    Debug <<
		asHex((int) *(p-3)) << " " <<
		    asHex((int) *(p-2)) << " " <<
			asHex((int) *(p-1)) << "\n" <<
			    asHex((int) sendPacket.buf[0]) << " " <<
				asHex((int) sendPacket.buf[1]) << " " << 
				    asHex((int) sendPacket.buf[2]) << 
					"\n" << eor; 
	    val = -1;
	}
	else {
	    val = (*p++ & 0x7f) << 16;	/* strip parity bit */
	    val |= (*p++) << 8;
	    val |= *p;
	}
	delete returnedPacket;
	returnedPacket = 0;
    }
    waiting = 0;

    mutex->V();

    Debug.printf("DatakitSwitch::ReadLocation returns %d\n",val);

    return val;
}

int
DatakitSwitch::WriteLocation(int addr, int val, int evenodd)
{
    Debug << this << "::WriteLocation( addr=" << addr << ", val=" << val
	<< ", evenodd=" << evenodd << " )\n" << eor;

    mutex->P();

    Assert(waiting == 0);

    sendPacket.buf[0] = 0x77;
    if (evenodd)
	sendPacket.buf[1] = (addr >> 8) | 0x80;
    else
	sendPacket.buf[1] = addr >> 8;
    sendPacket.buf[2] = addr;
    sendPacket.buf[5] = val;
    sendPacket.buf[4] = (val >>= 8);
    sendPacket.buf[3] = (val >> 8);
    sendPacket.len = 6;

    waiting = 1;

    sendAndWait();

    if (returnedPacket == 0) {
	Console().printf("DatakitSwitch: Can't write switch location %d\n", addr);
	val = -1;
    }
    else {
	if (*((unsigned short *) &returnedPacket->buf[6]) == 0xffff) {
	    Console().printf("DatakitSwitch: Switch write of location %d was nacked\n", addr);
	    val = -1;
	}
	else
	    /* ok */
	    val = 0;

	delete returnedPacket;
	returnedPacket = 0;
    }

    waiting = 0;

    mutex->V();

    Debug.printf("DatakitSwitch::WriteLocation returns %d\n",val);

    return val;
}

int
DatakitSwitch::getSwitchAddress( int mod )
{
    Debug << this << "::getSwitchAddress( mod=" << mod << " )\n" << eor;

    Assert((mod >=0) && (mod < MAXMODS));

    /*
     * In a VCS node, the first 512 double-words of switch memory are
     *   a table for allocating switch memory.
     * The address of those 512 is 9 bits long: the lower 4 are for
     *   slot number within a shelf, and the upper 3 are for shelf number.
     *   The middle two bits are for slots which have more than 512
     *   channels.  For example, slot 2 corresponds to addresses 2, 18,
     *   34, and 60.
     * The even word is for data and the odd word is for control.  Each word
     *   contains the (length - 1) in the lower 9 bits and the address in the
     *   remaining bits.
     */
    DatakitSwitchAddress *dp = &DatakitSwitchAddr[mod];

    int addr = (mod & 0xf) + ((mod & 0x30) << 2);

    for (int offset = 0; offset < 0x40; offset += 0x10, addr += 0x10) {
	int val;

	if ((val = ReadLocation(addr, 0)) == -1) {
	    dp->channels = -1;
	    return(-1);
	}

	int dataaddr = val >> 9;
	int length = (val & 0x1ff) + 1;

	if ((val = ReadLocation(addr, 1)) == -1) {
	    dp->channels = -1;
	    return(-1);
	}

	int controladdr = val >> 9;
	if (((val & 0x1ff) + 1) != length)
	    Console().printf("DatakitSwitch: switch location %d has data length != control length\n",addr);

	/*
	 * A value of 0 in the length bits means 0 or 1; assume it's 1 if
	 *  this is the first iteration and 0 otherwise.
	 */
	if (offset == 0) {
	    dp->dataaddr = dataaddr;
	    dp->controladdr = controladdr;
	    dp->channels = length;
	}
	else if (length != 1) {
	    /*
	     * Assume dataaddr = dp->dataadr + dp->channels and
	     *   controladdr = dp->controladdr + dp->channels;
	     *   they are contiguous.
	     */
	    dp->channels += length;
	}

	if (length != 512)
	    break;
    }
}

int
DatakitSwitch::GetCount(int module, int channel, int ctrldata)
{
    Debug.printf("DatakitSwitch::GetCount(%d, %d, %d)\n",
						module,channel,ctrldata);

    DatakitSwitchAddress *dp = &DatakitSwitchAddr[module];

    if (dp->channels == -1) {
	if (getSwitchAddress(module) == -1)
	    return(-1);
    }

    if (channel >= dp->channels)
	return(-1);

    return (ReadLocation(((ctrldata ? dp->controladdr : dp->dataaddr) +
								channel), 1));
}

int
DatakitSwitch::GetChannel(int &module, int &channel, int &iscommand,
			  int ctrldata)
{
    Debug.printf("DatakitSwitch::GetChannel(%d, %d, %d, %d)\n",
					    module,channel,iscommand,ctrldata);

    DatakitSwitchAddress *dp = &DatakitSwitchAddr[module];

    if (dp->channels == -1) {
	if (getSwitchAddress(module) == -1)
	    return(-1);
    }

    if (channel >= dp->channels)
	return(-1);

    int val = ReadLocation(((ctrldata ? dp->controladdr : dp->dataaddr) +
								channel), 0);
    if (val == -1)
	return(-1);

    /* see comment in 'getSwitchAddress' for bits of module/channel */
    int mod = (val >> 1) & 0x1ff;
    module = (mod & 0xf) + ((mod & 0x1c0) >> 2);
    int chn = (val >> 11) & 0x1ff;
    channel = chn + ((mod & 0x30) << 1);

    if (val & 0x200000)
	iscommand = 1;
    else
	iscommand = 0;

    Debug.printf("DatakitSwitch::GetChannel returning (%d, %d, %d)\n",
			    module, channel, iscommand);

    return(0);
}

int
DatakitSwitch::SetCount(int module, int channel, int count, int ctrldata)
{
    Debug.printf("DatakitSwitch::SetCount(%d, %d, %d, %d)\n",
					module, channel, count, ctrldata);

    DatakitSwitchAddress *dp = &DatakitSwitchAddr[module];

    if (dp->channels == -1) {
	if (getSwitchAddress(module) == -1)
	    return(-1);
    }

    if (channel >= dp->channels)
	return(-1);

    return (WriteLocation(((ctrldata ? dp->controladdr : dp->dataaddr) +
					channel), count, 1));
}

static char parityTable[] = {
    0, 1, 1, 0, 1, 0, 0, 1,
    1, 0, 0, 1, 0, 1, 1, 0,
    1, 0, 0, 1, 0, 1, 1, 0,
    0, 1, 1, 0, 1, 0, 0, 1
};

static inline int
DatakitSwitch::switchParity(int val)
{
    return((parityTable[(val & 0x1f) ^ ((val >> 5) & 0xf)]) | (val << 1));
}

int
DatakitSwitch::SetChannel(int sourcemodule, int sourcechannel,
		int destmodule, int destchannel, int iscommand, int ctrldata)
{
    DatakitSwitchAddress *dp = &DatakitSwitchAddr[destmodule];

    if (dp->channels == -1) {
	/* we're only looking this up to do a range check on the destchannel */
	if (getSwitchAddress(destmodule) == -1)
	    return(-1);
    }

    if (destchannel >= dp->channels)
	return(-1);

    dp = &DatakitSwitchAddr[sourcemodule];

    if (dp->channels == -1) {
	if (getSwitchAddress(sourcemodule) == -1)
	    return(-1);
    }

    if (sourcechannel >= dp->channels)
	return(-1);

    /* see comment in 'getSwitchAddress' for bits of module/channel */
    int chn = destchannel & 0x1ff;
    int mod = (destmodule & 0xf) + ((destmodule << 2) & 0x1c0) +
		((destchannel & 0x600) >> 5);

    int val = switchParity(mod) + (switchParity(chn) << 10);

    if (iscommand)
	val |= 0x200000;

    if (WriteLocation(((ctrldata ? dp->controladdr : dp->dataaddr) +
		       sourcechannel), val, 0) == -1)
	return(-1);

    return(SetCount(sourcemodule, sourcechannel, 0, ctrldata));
}

int
DatakitSwitch::setDevice(int mod, int eaddr, int oaddr, int nrte)
{
    int	val;
    int	a, n, d, c, x;
    DatakitSwitchAddress * dp = &DatakitSwitchAddr[mod];

    a = eaddr; n = nrte; c = 0;
    while (n>0) {
	val = a;
	val <<= 9;
	val |= 511;
	getphys(mod, c, d, x);
	if (WriteLocation( d, val, 0 ) < 0 )
	    return -1;
	a += 512; c += 512; n -= 512;
    }
    while (c < 2048) {
	getphys(mod, c, d, x);
	if (WriteLocation( d, 0, 0 ) < 0)
	    return (-1);
	c += 512;
    }

    dp->dataaddr = eaddr;

    a = oaddr; n = nrte; c = 0;
    while (n>0) {
	val = a;
	val <<= 9;
	val |= 511;

	getphys(mod, c, d, x);
	if(WriteLocation( d, val, 1 ) < 0) 
	    return(-1);
	a += 512; c += 512; n -= 512;
    }
    while (c < 2048) {
	getphys(mod, c, d, x);
	if (WriteLocation( d, 0, 1) < 0)
	    return (-1);
	c += 512;
    }
    dp->controladdr = oaddr;
    dp->channels = nrte;

    return(0);
}

void
DatakitSwitch::getphys(int d, int c, int & dp, int & cp)
{
    d = ((d<<2)&0700) | (d&017) | ((c>>5)&060);
    c &= 0777 ;

    dp = d ;
    cp = c ;
}


int
DatakitSwitch::Rammify()
{
    int val=0; 

    Debug << this << "::Rammify( )\n" << eor;

    mutex->P();

    Assert(waiting == 0);

    sendPacket.buf[0] = 0171;
    sendPacket.len = 1;

    waiting = 1;

    sendAndWait();

    if (returnedPacket == 0) {
	Console().printf("Ramify: Can't RAM my baby location \n");
	val = -1;
    }
    else {
	if (*((unsigned short *) &returnedPacket->buf[6]) == 0xffff) {
	    Console().printf("Ramify: nacked\n");
	    val = -1;
	}
	else
	    /* ok */
	    val = 0;

	delete returnedPacket;
	returnedPacket = 0;
    }

    waiting = 0;

    mutex->V();

    Debug.printf("Ramify returns %d\n",val);

    return val;
}


int
DatakitSwitch::Rommify()
{
    int val=0; 

    Debug << this << "::Rammify( )\n" << eor;

    mutex->P();

    Assert(waiting == 0);

    sendPacket.buf[0] = 0170;
    sendPacket.len = 1;

    waiting = 1;

    sendAndWait();

    if (returnedPacket == 0) {
	Console().printf("Ramify: Can't RAM my baby location \n");
	val = -1;
    }
    else {
	if (*((unsigned short *) &returnedPacket->buf[6]) == 0xffff) {
	    Console().printf("Ramify: nacked\n");
	    val = -1;
	}
	else
	    /* ok */
	    val = 0;

	delete returnedPacket;
	returnedPacket = 0;
    }

    waiting = 0;

    mutex->V();

    Debug.printf("Ramify returns %d\n",val);

    return val;
}


#include "ClassInstaller.h"
DEFINECLASS("DatakitSwitch",DatakitSwitch,DatakitChannel,0,0);
