/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *	   University of Illinois at Urbana-Champaign
 *	   Department of Computer Science
 *	   1304 W. Springfield Ave.
 *	   Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David W. Dykstra (dykstra@cs.uiuc.edu)
 * Contributing Authors: See-Mong Tan (stan@cs.uiuc.edu)
 *					   Sean M. Dorward (smd@research.att.com)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * DatakitURPChannel.c: Per-channel Datakit processing; urp protocol implementation
 */

#include "Assert.h"
#include "DatakitURPChannel.h"
#include "OutputStream.h"
#include "Urp.h"
#include "LOCK.h"
#include "Semaphore.h"
#include "PeriodicTimer.h"
#include "Util.h"
#include "ThisProcess.h"
     
static const int DKPACKETHOG = 200;	// max packets/channel: 1024 bytes + trailer
     
// receive state
static const char GETBLOCK	= 0x00;		// normal state, receive data or BOT
// retransmission error state
static const char MLEN		= 0x01;
static const char MSEQ		= 0x04;
static const char MERROR	= 0x08;
static const char MSTATES	= 0x0f;
// Special messages -- increment seqs even if message is bad
static const char SLEN		= 0x10;
static const char SSEQ		= 0x40;
static const char SERROR	= 0x80;
static const char SSTATES	= 0x0f;

// xmitStates
static const char X_WAIT =	0x02;	// waiting for AINIT
static const char X_INITREQ =	0x80;	// send INITREQ next timeout
     
// NOT CHECKED
static const char X_NOINIT =	0x04;	// just sent INIT or INTREQ; flag for timer
static const char X_INIT =	0x08;	// transmitter initialized
static const char X_ACK =	0x10;	// ENQ sent, waiting for ACK
static const char X_REJ =	0x20;	// sent REJ; this is only cleared

// NOT SET OR CHECKED
static const char X_ENQ =	0x40;	// send ENQ next timeout
static const char X_OFF =	0x01;	// transmitter off

enum {		// services that can be provided on the channel
    RAW	=	0x0,	// no extra service
	 
    // flow control
    FLOW		=	0x01,
	 
    // error control
    DETECT	=	0x02,
    RETRANS	=	0x04,
};
	 
static char services[] = {		// services provided for each gos
    0,
    
    // character modes
    RAW,
    FLOW,
    
    // block modes
    DETECT,
    FLOW | DETECT,
    FLOW | RETRANS,
};


//   Packet Queues for the URP channel
DatakitPacketQueue::DatakitPacketQueue(int max)
{
    head = tail = 0;
    size = 0;
    maxSize = max;
    queueLock = new LOCK;
    Assert(queueLock != 0);
    queueCount = new Semaphore(0);
    Assert(queueCount != 0);
}

int
DatakitPacketQueue::queueSize()
{
    return size;
}

// append -- append q to this; empty q
int
DatakitPacketQueue::append(DatakitPacketQueue *q) 
{
    queueLock->acquire();
    q->queueLock->acquire();
    
    int qsize = q->queueSize();
    
    Debug << "DatakitPacketQueue::append() of " << qsize
	<< " packets to a queue of " << size << " packets\n" << eor;
    
    if (size + qsize > maxSize) {
	Console() << "DatakitPacketQueue: queue size exceeded\n" << eor;
	q->queueLock->release();
	queueLock->release();
	return 0;
    }
    
    Assert(q != 0);
    if (head == 0) {
	Assert(tail == 0);
	head = q->head;
	tail = q->tail;
    } else {
	Assert(tail != 0);
	tail->next = q->head;
	tail = q->tail;
    }
    q->head = q->tail = 0;
    q->size = 0;
    size += qsize;
    
    while (qsize--) {
	queueCount->V();
	q->queueCount->P();
    }
    
    queueLock->release();
    q->queueLock->release();
    
    Debug << "Appended\n" << eor;
    
    return 1;
}

int
DatakitPacketQueue::enqueue(DatakitPacket *packet)
{
    Assert(packet != 0);
    queueLock->acquire();
    
    if (size >= maxSize) {
	Debug << "DatakitPacketQueue: queue size exceeded\n" << eor;
	queueLock->release();
	return 0;
    }
    
    Assert(packet->next == 0);
    if (head == 0) {
	Assert(tail == 0);
	head = tail = packet;
    } else {
	Assert(tail != 0);
	tail = tail->next = packet;
    }
    
    size += 1;
    queueCount->V();
    queueLock->release();
    
    return 1;
}

int
DatakitPacketQueue::undequeue(DatakitPacket *Packet)
{
    Assert( Packet != 0 );
    queueLock->acquire();
    
    if (size >= maxSize) {
	Console() << "DatakitPacketQueue: queue size exceeded\n" << eor;
	queueLock->release();
	return 0;
    }
    
    Assert(Packet->next == 0);
    if (head == 0) {
	Assert(tail == 0);
	head = tail = Packet;
    } else {
	Assert(tail != 0);
    Packet->next = head;
	head = Packet;
    }
    size += 1;
    queueCount->V();
    queueLock->release();
    return 1;
}

DatakitPacket *
DatakitPacketQueue::dequeue()
{
    queueCount->P();
    queueLock->acquire();
    
    Assert(size > 0);
    DatakitPacket * p = head;
    
    if (head == tail) {
	Assert(p->next == 0);
	head = tail = 0;
    } else {
	head = head->next;
	Assert(head != 0);
	p->next = 0;
    }
    size -= 1;
    queueLock->release();
    
    return p;
}

void
DatakitPacketQueue::trash()
{
    while(queueSize() > 0)
	delete dequeue();
}

// getFreeChannel -- get a datakit channel that is not currently used
int
DatakitURPChannel::getFreeChannel()
{
    static int next = 0;
    int stop = next;
    
    do {
	if (++next == MAXCHAN)
	    next = 1;	   // don't use 0
    } while (DatakitChan[next] != 0 && next != stop);
    
    return next == stop ? 0 : next;
}

// clearBuf -- clear a retransmission buffer
void
DatakitURPChannel::clearBuf(DatakitXmitBuf *buf) {
    buf->buf = 0;
    buf->len = 0;
    buf->avail = 1;
    buf->eob = 0;
}

// resetSeq -- reset transmit sequence numbers and retransmit buffers
void
DatakitURPChannel::resetSeq()
{
    // move retransmission buffers to new places
    if (service & RETRANS) {
	if (xmitSeq != (ackedSeq + 1 & SEQ_NO_MASK) && ackedSeq != 0) {
	    int i, j;
	    for (i = 1, j = ackedSeq + 1 & SEQ_NO_MASK; j != xmitSeq;
		 i++, j = j + 1 & SEQ_NO_MASK) {
		     rexmitBuf[i] = rexmitBuf[j];
		     clearBuf(&rexmitBuf[j]);
		 }
	}
	xmitSeq = xmitSeq - ackedSeq & SEQ_NO_MASK;
	ackedSeq = 0;
	retransmit(1);
    }
    else {
	xmitSeq = 1;
	ackedSeq = 0;
    }
}

DatakitURPChannel::DatakitURPChannel(int channel, int gos)
{
    Debug << "DatakitURPChannel::DatakitURPChannel( channel=" << channel
	<< ", gos=" << gos << " )\n" << eor;
    
    chan = channel;
    
    if (chan == 0) {
	chan = getFreeChannel();
	
	Debug.printf("DatakitURPChannel: ctor selected channel=%d\n", chan);
    }
    
    if (chan <= 0 || chan >= MAXCHAN ||
	DatakitChan[chan] != 0 || gos < 1 || gos > 5) {
	    chan = MAXCHAN;
	    return;
	}
    
    static int initialized = 0;
    if (!initialized) {
	static LOCK initlock;
	
	initlock.acquire();
	if (initialized)
	    initlock.release();
	else {
	    initialized = 1;
	    initlock.release();
	    
	    startTimer();
	}
    }
    
    // Create receiver side stuff
    recvLock = new LOCK;
    Assert(recvLock != 0);
    readQ = new DatakitPacketQueue(DKPACKETHOG);
    Assert(readQ != 0);
    stage = new DatakitPacketQueue(DKPACKETHOG);
    Assert(stage != 0);
    recvSeq = 0;
    receiveLen = 0;
    receiveState = GETBLOCK;
    receiveTrailers = 0;
    
    // Create transmitter side stuff
    xmitLock = new LOCK;
    Assert(xmitLock != 0);
    xmitWaitingSemaphore = new Semaphore( 0 );
    Assert(xmitWaitingSemaphore != 0);
    writerWaiting = 0;
    
    xmitSeq = 1;
    ackedSeq = 0;
    
    ctlcalls = 1;
    service = services[gos];
    
#ifdef DATAKITSTATS
    ZeroBlock((char *)&firststat, (char *)&laststat - (char *)&firststat);
#endif
    
    DatakitXmitBuf * p, * endbuf = rexmitBuf + (SEQ_NO_MASK + 1);
    for (p = rexmitBuf; p < endbuf; p++)
	clearBuf(p);
    
    // initialize myself
    sendControl(service & (DETECT | RETRANS) ? INIT1 : INIT0);
    
    // Enter channel into big table
    DatakitChan[chan] = this;
}

DatakitURPChannel::~DatakitURPChannel()
{
    Debug.printf("DatakitURPChannel::dtor on channel %d\n", chan);
    
    if (xmitLock != 0)
	delete xmitLock;
    if (recvLock != 0)
	delete recvLock;
    if (chan != MAXCHAN)
	DatakitChan[chan] = 0;
}


// canXmit -- true if another packet can be transmitted
int
DatakitURPChannel::canXmit()
{
    return (!(service & FLOW) || (xmitSeq - ackedSeq & SEQ_NO_MASK) <= URPXmitWindowSize) &&
	!(xmitState & X_WAIT);
}

// Write -- write an URP packet; save the data for possible retransmission if needed
// much rewritten by Stan and Sean
int
DatakitURPChannel::Write(unsigned char *buf, int writelen)
{
    Debug.printf("DatakitURPChannel::Write(%x, %d) on chan %d\n",
		   buf, writelen, chan);
    
    if (!channelOpened())
	return -1;
    
    if (writelen == 0)
	return 0;
    int initialLen = writelen;
    
    // break the buffer into sendable pieces
    int packets = (writelen + URPXmitBlockSize - 1) / URPXmitBlockSize;
    DatakitXmitBuf *packet = new DatakitXmitBuf[packets];
    DatakitXmitBuf *p;
    DatakitXmitBuf *endpack = packet + packets;
    for (p = packet; p < endpack; p++) {
	int len = p->len = writelen > URPXmitBlockSize ? URPXmitBlockSize : writelen;
	p->avail = 0;
	p->eob = writelen <= URPXmitBlockSize;
	if (service & RETRANS) {
	    p->buf = new unsigned char[len];
	    CopyBlock(buf, p->buf, len);
	} else {
	    p->buf = buf;
	}
	writelen -= len;
	buf += len;
    }
    
    // send each packet
    xmitLock->acquire();
    for (p = packet; p < endpack; p++) {
	if (!canXmit()) {
	    do {
		writerWaiting++;
		Debug << this << "::Write(): waiting on ready\n" << eor;
		xmitLock->release();
		xmitWaitingSemaphore->P();
		xmitLock->acquire();
	    } while (!canXmit());
	}
	
	Debug << "Transmitting with xmitSeq=" << xmitSeq << "\n" << eor;
	Assert(canXmit());
	
	// Send the data through the datakit 586HS device driver call
	sendBlock(xmitSeq, p->buf, p->len, p->eob);
	
	// save for possible retransmission later
	if (service & RETRANS) {
	    DatakitXmitBuf *pack = &rexmitBuf[xmitSeq];
	    
	    Assert(pack->avail == 1 && pack->buf == 0);
	    *pack = *p;
	}
	
	xmitSeq = (xmitSeq + 1) & SEQ_NO_MASK;
    }
    xmitLock->release();
    
    delete packet;
    return initialLen;
}

// Read -- return the next readLen bytes or fewer if a null message is reached
int
DatakitURPChannel::Read(unsigned char *buf, int readLen)
{
    Debug.printf("DatakitURPChannel::Read(%x, %d) on chan %d\n",
		   buf, readLen, chan);
    
    if (!channelOpened())
	return -1;
    
    if (readLen < 0)
	return -1;
    
    int bytesRead = 0;
    while( readLen > 0 ) {
	DatakitPacket *packet = readQ->dequeue();
	
	if (packet->len == 0) {
	    // This is the end of message marker
	    delete packet;
	    break;
	}
	
	int copylen = ( readLen > packet->len ? packet->len : readLen );
	CopyBlock( (unsigned char *) packet->buf, buf, copylen );

	buf += copylen;
	readLen -= copylen;
	bytesRead += copylen;

	// Packet's length is decreased since we consumed it's data
	packet->len -= copylen;
	
	// put part of the packet back if it is not all consumed
	if (packet->len != 0) {
	    CopyBlock( packet->buf + copylen, packet->buf, 
		      packet->len );   // Copy to front
	    readQ->undequeue(packet);
	} 
	else
	    delete packet;
    }
    return bytesRead;
}

// The rest are only to be accessed by the interrupt handler or used internally.

// ReceivePacket -- 
int
DatakitURPChannel::ReceivePacket(DatakitPacket *packet)
{
    recvLock->acquire();
    int retval = 0;		// packet not consumed
    Debug << "DatakitURPChannel::ReceivePacket( chan=" << chan << 
	", len=" << packet->len << " )\n" << eor;
    switch (receiveState) {
    case GETBLOCK:
	Debug << "GETEBLOCK\n" << eor;
	receiveLen += packet->len;
	
	if (receiveTrailers)
	    retval = stage->enqueue(packet);
	else
	    retval = readQ->enqueue(packet);
	break;
    case SLEN:
    case MLEN:
	Debug << "MLEN\n" << eor;
	if (packet->len == 2 &&
	    packet->buf[0] == (receiveLen & 0x00ff) &&
	    packet->buf[1] == (receiveLen & 0xff00))
	    receiveState = receiveState == SLEN ? SSEQ : MSEQ;
	else
	    receiveState = receiveState & MSTATES ? MERROR : SERROR;
	break;
    default:
	Debug << "ERROR\n" << eor;
	receiveState = receiveState & MSTATES ? MERROR : SERROR;
	break;
    }
    Debug << "packet processed with retval=" << retval << "\n" << eor;
    recvLock->release();
    return retval;
}

// ReceiveControl -- receive a control message
//	receiver ignores any control messages that are not for it's mode
void
DatakitURPChannel::ReceiveControl(int c)
{
    Debug << "ReceiveControl(chan = " << chan << ", ctl = " << c << ")\n" << eor;
    switch (c) {
	// *******   Receiver side   ********
    case INIT0:
    case INIT1:
	Debug << "Receiver on chan=" << chan << " received INIT\n"
	    << eor;
	recvLock->acquire();
	xmitState &= ~X_INITREQ;
	if (c == INIT0) {
	    STAT(statInit0);
	    receiveTrailers = 0;
	} else {
	    STAT(statInit1);
	    receiveTrailers = 1;
	}
	sendControl(AINIT);
	echo = ECHO + 0;
	recvSeq = 0;
	receiveLen = 0;
	receiveState = GETBLOCK;
	stage->trash();
	recvLock->release();
	return;
	
    case BOT:
    case BOTM:
    case BOTS:
	recvLock->acquire();
	Debug << "Receiver on chan=" << chan << " received BOT\n" 
	    << eor;
	STAT(statBot);
	if (receiveTrailers)
	    if (receiveState == GETBLOCK)
		receiveState = c == BOTS ? SLEN : MLEN;
	    else {
		stage->trash();
		receiveState = c == BOTS ? SERROR : MERROR;
	    }
	recvLock->release();
	return;
	
    case SEQ + 0: case SEQ + 1: case SEQ + 2: case SEQ + 3:
    case SEQ + 4: case SEQ + 5: case SEQ + 6: case SEQ + 7:
	recvLock->acquire();
	STAT(statSeq);
	receiveSeq(c - SEQ);
	recvLock->release();
	return;
	
    case ENQ:
	recvLock->acquire();
	Debug << "Receiver on chan=" << chan << " received ENQ\n"
	    << eor;
	STAT(statEnq);
	if (receiveTrailers)
	    stage->trash();
	receiveLen = 0;
	receiveState = GETBLOCK;
	sendControl(echo);
	sendControl(ACK + recvSeq);
	recvLock->release();
	return;
	
	// *******   Transmitter side   ********<
    case INITREQ:
	xmitLock->acquire();
	STAT(statInitreq);
	// we can only be freed with an AINIT
	// sent from the other side.
	Debug << "Transmitter on " << chan << " received INITREQ\n" << eor;
	
	sendControl(service & (DETECT | RETRANS) ? INIT1 : INIT0);
	xmitLock->release();
	return;
	
    case AINIT:
	xmitLock->acquire();
	STAT(statAinit);
	if (xmitState & X_WAIT) {
	    Debug << "Transmitter on " << chan << " received AINIT\n" << eor;
	    // Reset variables so we next start with sequence number of 0
	    resetSeq();
	    // reset transmitter state
	    xmitState &= ~(X_NOINIT | X_WAIT);
	    // then we free up the writers who are waiting to write
	    freeWriters();
	} else {
	    Debug << "Transmitter on " << chan << " received unexpected AINIT\n" << eor;
	    resetSeq();
	    xmitState &= ~(X_NOINIT | X_WAIT);
	}
	xmitLock->release();
	return;
	
    case ECHO + 0: case ECHO + 1: case ECHO + 2: case ECHO + 3:
    case ECHO + 4: case ECHO + 5: case ECHO + 6: case ECHO + 7:
	xmitLock->acquire();
	STAT(statEcho);
	receiveEcho( ECHO, c - ECHO );
	xmitLock->release();
	return;
	
    case REJ + 0: case REJ + 1: case REJ + 2: case REJ + 3:
    case REJ + 4: case REJ + 5: case REJ + 6: case REJ + 7:
	xmitLock->acquire();
	STAT(statRej);
	Debug << "Transmitter on chan=" << chan << " received REJ"
	    << c - REJ << "\n" << eor;
	receiveEcho( REJ, c - REJ );
	xmitLock->release();
	return;
	
    case ACK + 0: case ACK + 1: case ACK + 2: case ACK + 3:
    case ACK + 4: case ACK + 5: case ACK + 6: case ACK + 7:
	xmitLock->acquire();
	STAT(statAck);
	Debug << "Transmitter on " << chan << " received ACK" << c - ACK
	    << "\n" << eor;
	xmitLock->release();
	return;
	
    default:
	// EOI, SOI, and CHECK are not supported,
	// as per the XUNET programmer's manual
	STAT(statRecvError);
	Debug.printf("DatakitURPChannel(%d): unrecognized control: %x\n", chan, c);
	return;
    }
}

// receiveSeq -- process a block
void
DatakitURPChannel::receiveSeq(int seq)
{
    Debug << "Receiver on " << chan << " got SEQ" << seq << "\n" << eor;
    
    if (!receiveTrailers) {
	accept(seq);
	return;
    }
    
    // check for proper sequence
    if (seq != (recvSeq + 1 & SEQ_NO_MASK)) {
	reject(seq);
	return;
    }
    
    // the length check has already bee proformed & is reflected in the state
    switch (receiveState) {
    case SSEQ:
	accept(seq);
	break;
    case MSEQ:
	// append staging queue to the main queue
	if (!readQ->append(stage))
	    reject(seq);
	else
	    accept(seq);
	break;
    default:	// error state
	reject(seq);
    }
}

// accept -- accept a packet with sequence number seq
void
DatakitURPChannel::accept(int seq) 
{
    STAT(recvGoodBlock);
    recvSeq = seq;
    
    // reply to sender
    sendControl(ECHO + seq);
    receiveLen = 0;
    
    // put terminator packet on the readQ to tell everyone it's good
    DatakitPacket * endofblock = new DatakitPacket;
    readQ->enqueue(endofblock);
    
    receiveState = GETBLOCK;
}

// reject -- reject a packet with sequence number seq
void
DatakitURPChannel::reject(int seq) 
{
    STAT(recvBadBlock);
    if (receiveState & SSTATES)
	recvSeq = seq;
    sendControl(REJ + recvSeq);
    if (service & RETRANS)
	stage->trash();
    receiveLen = 0;
    receiveState = GETBLOCK;
}

// receiveEcho -- update ackedSeq; reset xmitSeq if resetXmitSeq
void
DatakitURPChannel::receiveEcho(int echo, int seq)
{
    if ((xmitSeq - seq & SEQ_NO_MASK) - 1 > URPXmitWindowSize) {
	Debug << this << "::ReceiveControl: bad " <<
	    (echo == REJ ? "REJ " : "ECHO ") << seq <<
		" on chan " << chan << ", ackedSeq=" <<
		    ackedSeq << ", xmitSeq=" << xmitSeq << "\n" << eor;
	return;
    }
    Debug << "Transmitter on " << chan << " received " <<
	(echo == REJ ? "REJ " : "ECHO ") << seq << "\n" << eor;
    
    if (service & RETRANS) {
	ackTo(seq);
	if (echo == REJ)
	    retransmit(seq + 1 & SEQ_NO_MASK);
    } else if (service & DETECT) {
	ackedSeq = seq;
	if (echo == REJ)
	    xmitSeq = seq + 1 & SEQ_NO_MASK;
    } else if (service & FLOW) {
	// no retransmission -- just set acked seq no.
	ackedSeq = seq;
    }
    freeWriters();
}

// ackTo -- ack up to and including endseq
void
DatakitURPChannel::ackTo(int endSeq)
{
    if (ackedSeq != endSeq) {
	do {
	    ackedSeq = ackedSeq + 1 & SEQ_NO_MASK;
	    DatakitXmitBuf *p = &rexmitBuf[ackedSeq];
	    
	    // buf == 0 when ackedSeq == 0 on startup
	    if (p->buf != 0)
		delete p->buf;
	    p->buf = 0;
	    p->avail = 1;
	} while (ackedSeq != endSeq);
    }
}

// freeWriters -- free any writers who were waiting for window space
void
DatakitURPChannel::freeWriters()
{
    int freeable = URPXmitWindowSize - (xmitSeq - ackedSeq & SEQ_NO_MASK) + 1;

    freeable = freeable > writerWaiting ? writerWaiting : freeable;
    writerWaiting -= freeable;
    while (freeable--)
	xmitWaitingSemaphore->V();
}

// retransmit -- retransmit all packets from seq up to (but not including)
// xmitSeq.  We assume that the xmit lock is already acquired
void
DatakitURPChannel::retransmit(char seq)
{
    Debug << this << "::retransmit( seq=" << (int) seq
	<< ", xmitSeq=" << (int) xmitSeq << " )\n" << eor;

    Assert(service & RETRANS);

    DatakitXmitBuf *xbp;
    for (; seq != xmitSeq; seq = (seq + 1) & SEQ_NO_MASK) {
	xbp = &rexmitBuf[seq];
	Assert(xbp->buf != 0);

	sendBlock(seq, xbp->buf, xbp->len, xbp->eob);
    }
}

// sendBlock -- send a block along with any trailer needed
// assumes the mutex lock is already held
// NOTE -- error checking should be done here
void
DatakitURPChannel::sendBlock(int seq, unsigned char *buf, int len,
			     int endofblock)
{
    DatakitTransmitStart(chan);
    DatakitTransmitData(buf, len);
    if (service & RETRANS || service & DETECT) {
	DatakitTransmitControl(endofblock ? BOT : BOTM);
	unsigned char lenbuf[2];
	lenbuf[0] = len & 0xff;
	lenbuf[1] = len >> 8 & 0xff;
	DatakitTransmitData(lenbuf, 2);
    }
    if (service != RAW)
	DatakitTransmitControl(SEQ + seq);
    DatakitTransmitEnd();
}

// sendControl -- send a control message and set any state
// assumes the mutex lock is held already
void
DatakitURPChannel::sendControl(int ctl) 
{
    // Assume that the lock is already held by someone else
    Debug << this << "::sendControl(" << ctl << " === " << eor;
    switch (ctl) {
    case INITREQ:
	STAT(x_initreq);
	Debug << "INITREQ";
	xmitState |= X_NOINIT|X_INITREQ;
	break;
    case INIT0:
    case INIT1:
	Debug << "INIT";
	STAT(x_init);
	// gain exclusive access, must wait for AINIT from other side
	// to V() us, so we can continue
	xmitState |= X_NOINIT | X_WAIT;
	break;
    case ENQ:
	Debug << "ENQ";
	STAT(statEnq);
	xmitState &= ~X_REJ;
	xmitState |= X_ACK;
	break;
    case ECHO + 0: case ECHO + 1: case ECHO + 2: case ECHO + 3:
    case ECHO + 4: case ECHO + 5: case ECHO + 6: case ECHO + 7:
    case REJ + 0: case REJ + 1: case REJ + 2: case REJ + 3:
    case REJ + 4: case REJ + 5: case REJ + 6: case REJ + 7:
	Debug << ((ctl &~ SEQ_NO_MASK) == ECHO ? "ECHO" : "REJ");
	echo = ctl;
	break;
    default:
	break;
    }
    Debug << ")\n" << eor;
    
    DatakitSendControl(chan, ctl);
}

void
DatakitURPChannelTimerEntry(int)
{
    Debug << "DatakitURPChannelTimerEntry()\n" << eor;
    
    PeriodicTimer *timerRetry = new PeriodicTimer();
    Assert(timerRetry != 0);
    
    timerRetry->start(4 * 1000000);
    for (;;) {
	timerRetry->await();	// wait for expiry
	Debug << "DatakitURPChannelTimerEntry: waking up\n" << eor;
	DatakitURPChannel **p, **stop;
	
	stop = (DatakitURPChannel **)DatakitChan + MAXCHAN;
	for (p = (DatakitURPChannel **)DatakitChan; p != stop; p++)
	    if (*p != 0)
		DatakitURPChannel::timerHandler(*p);
    }
}

// startTimer -- start a timer process to wake up channels on timeout
void
DatakitURPChannel::startTimer()
{
    Debug << "DatakitURPChannelTimerEntry: Spawning Timer Process\n" << eor;
    
    Process *timerRetryProcess =
	new SystemProcess ((PFV)DatakitURPChannelTimerEntry,
			   thisProcess()->domain(), normal, 0);
    Assert(timerRetryProcess != 0);
    timerRetryProcess->ready();
}

static int writers[64]; 

void
DatakitURPChannel::timerHandler(DatakitURPChannel *chan)
{
    if (chan->xmitState & X_WAIT) {
	Debug << "Resending INIT1 on " << chan->chan << "\n" << eor;
	chan->sendControl(chan->service & (DETECT | RETRANS) ? INIT1 : INIT0);
    }
    // sweep the writers array
    if (writers[chan->chan] > 2) {
	if (chan->writerWaiting) {
	    // got to resynchronize
	    chan->sendControl( ENQ );
	}
    }
    if (chan->writerWaiting)
	writers[chan->chan] += 1;
    else
	writers[chan->chan] = 0;

    //	if (chan->xmitState & X_WINDOW) {
    //	  Debug << "Resending INIT1 on " << chan->chan << "\n" << eor;
    //	  chan->sendControl(chan->service & (DETECT | RETRANS) ? INIT1 : INIT0);
    //	}
    //	  MAJORLY CHANGE THIS IF IT IS ENABLED!!!!
    //  if ( chan->writerWaiting ) {
    //	  Debug << "Reinitializing protocol on " << chan->chan << "\n" << eor;
    //	  chan->xmitLock->acquire();
    //	  chan->xmitSeq = 0;
    //	  chan->ackedSeq = 0;
    //	  chan->recvSeq = 0;
    //	  chan->sendControl(chan->service & (DETECT | RETRANS) ? INIT1 : INIT0);
    //	  chan->xmitLock->release();
    //  }
}


#include "ClassInstaller.h"
DEFINECLASS("DatakitURPChannel",DatakitURPChannel,DatakitChannel,0,0);
