#define DEBUG
/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: See-Mong Tan (stan@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * DatakitUnixHost.cc: AT&T 6586 Datakit Unix Host
 */

#include "Assert.h"
#include "Class.h"
#include "OutputStream.h"
#include "PeriodicTimer.h"
#include "LOCK.h"
#include "ThisProcess.h"
#include "DatakitURPChannel.h"
#include "DatakitUnixHost.h"

// The Unix Host Manager known system wide
// should be bound to a name in the Nameserver
DatakitUnixHost * unixHostManager;

// The format of messages sent over the wire between the host and
// the unixcscp process on the datakit switch control computer
class Inpack {
public:
    char type;
    char srv;
    short chan;
    short p1;
    short p2;
    short p3;
    short uid;
};

// Fill in the blanks with the proper fields
inline void
FillInpack( char type, char srv, short chan, 
			short p1, short p2, short p3, short uid, 
			Inpack & inpack )
{
    inpack.type = type;
    inpack.srv = srv;
    inpack.chan = chan;
    inpack.p1 = p1;
    inpack.p2 = p2;
    inpack.p3 = p3;
    inpack.uid = uid;
}

// The Channel for common supervision
static const int UNIXCSCPCHANNEL = 1;

// Allowable types in the type field
static const char T_SRVREQ =   (char) 1;
static const char T_SRV2 =     (char) 107;
static const char D_SERV =     (char) 1;
static const char D_DIAL =     (char) 2;
static const char D_XINIT =    (char) 3;

static const char T_REPLY =    (char) 2;
static const char D_OPEN =     (char) 2;
static const char D_FAIL =     (char) 3;

static const char T_CHG =      (char) 3;
static const char D_CLOSE =    (char) 1;
static const char D_ISCLOSED = (char) 2;
static const char D_CLOSEALL = (char) 3;
static const char D_SPLICE =   (char) 4;
static const char D_REDIAL =   (char) 6;

static const char T_ALIVE =    (char) 4;
static const char D_STILL =    (char) 0;
static const char D_RESTART =  (char) 1;
static const char D_MAXCHAN =  (char) 2;

static const char T_RESTART =  (char) 8;


/*
 * datakit error numbers
 */

#define EdBUSY	        1
#define EdDEADNODE	2	/* or trunk inoperable */
#define EdNOANSWER	3
#define EdNONASS	4	/* non-assigned number */
#define EdREORDER	5
#define EdEXIST		6	/* server already exists */
#define	EdNOACC		7	/* access denied */
#define EdDIRASS	8
				/* 9 is unused */
#define EdTOOLONG	10
#define EdDIALER	11	/* obsolete -- replaced by 22-29 */

/* expanded REORDER reason code */
#define EdROUTERR	12	/* routing error */
#define EdTIMOUT	13	/* time out */
#define EdCONBRK	14	/* connection broken */
#define EdSMGERR	15	/* sendmsg failure */
#define EdCHALER	16	/* controller channel allocate failure */
#define EdCONERR	17	/* connect call failure */
#define EdFWDCER	18	/* forward call error, sendmsg failure in trk */
#define EdCOLTRK	19	/* collision in trunk */
#define EdCFGTER	20	/* trunk configuration error */
#define EdCFGHER	21	/* host configuration error */

/* do not use high order byte for dialer reason code */
#define EdDIALNEEDNUM	22	/* need valid number */
#define EdDIALREORT	23	/* reorder tone */
#define EdDIALNOIDT	24	/* no initial dial tone */
#define EdDIALNOSDT	25	/* no secondary dial tone */
#define EdDIALBUSY	26	/* number busy */
#define EdDIALNOANS	27	/* no answer */
#define EdDIALNOCT	28	/* no carrier tone */
#define EdDIALBADBAUD	29	/* bad baud rate */

/* error codes used by hosts */
#define EdDEVBAD	130	/* can't open the /dev file for this channel */
#define EdMODPUSH	131	/* can't push user's stream module */
#define EdBADPROTO	132	/* invalid protocol requested */
#define EdNOSRVTAB	133	/* missing dksrvtab file */
#define EdBADCHROOT	134	/* can't chroot to subrequest directory */

// Messages sent to CSCP from host
inline void
E_KPALIVE( Inpack & inp )
{
    FillInpack( T_ALIVE, D_STILL, 0, 0, 0, 0, 0, inp );
}
inline void
E_RESTART( Inpack & inp )
{
    FillInpack( T_ALIVE, D_RESTART, 0, 64, 4, 0, 0, inp );
}
inline void
E_DIAL( short chan, short windows, int uid, Inpack & inp )
{
    FillInpack( T_SRV2, D_DIAL, chan, windows, 0, 0, uid, inp );
}
inline void
E_SERV( short chan, Inpack & inp )
{
    FillInpack( T_SRVREQ, D_SERV, chan, 0, 0, 0, 0, inp );
}
inline void
E_CLOSE( short chan, Inpack & inp )
{
    FillInpack( T_CHG, D_CLOSE, chan, 0, 0, 0, 0, inp );
}
inline void
E_ISCLOSED( short chan, Inpack & inp )
{
    FillInpack( T_CHG, D_ISCLOSED, chan, 0, 0, 0, 0, inp );
}

// This is the control channel that talks to the common supervision
// process on the switch control computer side
static DatakitURPChannel * ControlChannel;

// Need this semaphore, since the control channel can be simultaneously
// written to (DatakitUnixHost::open and DatakitUnixHostTimer)
// However, only the HostReceiveProcessEntry reads it
static MutexSemaphore * ControlChannelWriteMutex;

static Semaphore * OpenAwaitSemaphores[ HIGHCHANMARK ];

// Sends keep alive messages every so often
void DatakitUnixHostTimerEntry( int arg )
{
    DatakitURPChannel * control = (DatakitURPChannel *) arg;
    Assert( control != 0 );
    Assert( control->classOf() == DatakitURPChannelClass );
    Assert( control->channel() == UNIXCSCPCHANNEL );

    Inpack inpack;
    E_KPALIVE( inpack );
    PeriodicTimer * timer = new PeriodicTimer();
    Assert( timer != 0 );

    timer->start(5 * 1000000 );
    for(;;) {
	timer->await();
	ControlChannelWriteMutex->P();
	control->Write( (unsigned char *) &inpack, sizeof(Inpack) );
	ControlChannelWriteMutex->V();
    }
}

// Entry point for the receive and listen process...
// we wait around waiting for the messages the CSCP sends to us
// from the switch control computer side.
void
HostReceiveProcessEntry( int arg )
{
    Debug << "HostReceiveProcessEntry()\n" << eor;

    DatakitURPChannel * control = (DatakitURPChannel *) arg;
    Assert( control != 0 );
    Assert( control->classOf() == DatakitURPChannelClass );
    Assert( control->channel() == UNIXCSCPCHANNEL );
    Debug << "HostReceiveProcessEntry(): ctrl chan=" << control << "\n" << eor;

    Inpack inp;
    for(;;) {
	control->Read( (unsigned char *) &inp, sizeof(Inpack) );
	
	Debug << "HostReceiveProcessEntry: type=" << (int) inp.type
	    << ", srv=" << (int) inp.srv << ", chan=" << inp.chan
		<< ", p1=" << inp.p1 << "\n" << eor;

	// let's go for the simple case and ignore the rest first
	if ( inp.type == T_REPLY && inp.srv == D_OPEN ) {
	    Debug << "HostReceiveProcessEntry: ACK of open of chan=" <<
		inp.chan << ", window=" << inp.p1 << "\n" << eor;
	    // wakey wakey
	    Assert( OpenAwaitSemaphores[inp.chan] != 0 );
	    OpenAwaitSemaphores[inp.chan]->V();
	}
	else if ( inp.type == T_REPLY && inp.srv == D_FAIL ) {
	    Debug << "HostReceiveProcessEntry: E_OPEN on chan=" << inp.chan
		<< "\n" << eor;
	}
	else if ( inp.type == T_CHG && inp.srv == D_CLOSE ) {
	    Debug << "HostReceiveProcessEntry: E_CLOSE on chan=" << inp.chan
		<< "\n" << eor;
	    if ( DatakitChan[inp.chan] != 0 )
		DatakitChan[inp.chan]->closeChannel();

	    // send a closing message to the CSCP process
	    Inpack akclose;
	    E_ISCLOSED( inp.chan, akclose );
	    ControlChannelWriteMutex->P();
	    control->Write((unsigned char *) &akclose, sizeof(Inpack));
	    ControlChannelWriteMutex->V();
	}
    }
}

DatakitUnixHost::DatakitUnixHost()
{
    Debug << "DatakitUnixHost::DatakitUnixHost()\n" << eor;

    Assert( initialized == 0 );
    initialized = 1;

    // To ensure mutually exclusive access to the Control Channel
    ControlChannelWriteMutex = new MutexSemaphore;

    // Create a control channel to talk to CSCP process on other side
    ControlChannel = new DatakitURPChannel( UNIXCSCPCHANNEL, 5 );
    Assert( ControlChannel != 0 );
    Assert( ControlChannel->classOf() == DatakitURPChannelClass );
    Assert( ControlChannel->channel() == UNIXCSCPCHANNEL );

    Debug << "DatakitUnixHost(): got " << ControlChannel << 
	"for a Controller Channel\n" << eor;

    // Create a process to receive messages from the control channel
    Debug << "DatakitUnixHost(): creating host receive\n" << eor;
    Process * process = new SystemProcess( (PFV) HostReceiveProcessEntry,
					  thisProcess()->domain(), normal,
					  (int) ControlChannel );
    Assert( process != 0 );
    process->ready();

    // Timer process
    Debug << "DatakitUnixHost(): creating keepalive timer\n" << eor;
    process = new SystemProcess( (PFV) DatakitUnixHostTimerEntry,
				thisProcess()->domain(), normal, 
				(int) ControlChannel );
    Assert( process != 0 );
    process->ready();

    // Send a reinitialize to the CSCP, to tell him we got rebooted
    // and just came back up
    Debug << this << ":DatakitUnixHost(): RESTART to CSCP\n" << eor;
    Inpack inp;
    E_RESTART( inp );
    ControlChannelWriteMutex->P();
    ControlChannel->Write( (unsigned char *) &inp, sizeof( Inpack ) );
    ControlChannelWriteMutex->V();
}

DatakitURPChannel *
DatakitUnixHost::open( char * dialstring )
{
    // Search for a free channel, when found mark it for dialing
    mutex.acquire();
    for( int chan = LOWCHANMARK; chan <= HIGHCHANMARK; chan-- ) {
	if ( channelStatus[chan] == 0 ) {
	    channelStatus[chan] = CHANNELDIALING;
	    break;
	}
	else if ( chan == LOWCHANMARK )
	    return 0;
    }
    mutex.release();

    Debug << this << "::open( dialstring=" << dialstring << ", chan="
	<< chan << " )\n" << eor;

    // write an E_DIAL message over it to request the CSCP to open
    // the channel, uid of 6000 is for stan@beren as a test
    Inpack inp;
    E_DIAL( (short) chan, (short) 10, 6000, inp );
    ControlChannelWriteMutex->P();
    ControlChannel->Write( (unsigned char *) &inp, sizeof( Inpack ) );
    ControlChannelWriteMutex->V();

    // Create an URP Channel, and write a dialstring over it.
    DatakitURPChannel * newChannel = new DatakitURPChannel( chan, 5 );
    for( int dstrlen = 0; ; dstrlen++ )
	if ( *( dialstring + dstrlen ) == '\0' ) break;
    newChannel->Write( (unsigned char *) dialstring, dstrlen );

    return (DatakitURPChannel *) DatakitChan[chan];
}



void
DatakitUnixHost::close( DatakitURPChannel * chan )
{
    Inpack inp;
    E_CLOSE( chan->channel(), inp );
    ControlChannelWriteMutex->P();
    ControlChannel->Write( (unsigned char *) &inp, sizeof( Inpack ) );
    ControlChannelWriteMutex->V();

    // and more stuff please...
}

#include "ClassInstaller.h"
DEFINECLASS DatakitUnixHost
