/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: See-Mong Tan (stan@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * PCi586Console.c: PCi586 Diagnostic console output routines.
 */

#include "Assert.h"
#include "OutputStream.h"
#include "PCi586Console.h"
#include "PCi586InterruptController.h"
#include "CharacterInputStream.h"
#include "CharacterQueue.h"
#include "Semaphore.h"
#include "Util.h"
#include "Process.h"

extern "C" {
  extern unsigned int InterruptsEnabled();
}

int Com0Busy = 0; /* this is set to 1 by the debugger if it's using Com 0 */

static int const ComBase = 0x3f8;	/* base address of Com 0 Port */
static int const ComDAT = ComBase + 0;	/* receive/transmit data */
static int const ComICR = ComBase + 1;	/* interrupt control register */
static int const ComISR = ComBase + 2;	/* interrupt status register */
static int const ComLCR = ComBase + 3;	/* line control register */
static int const ComMCR = ComBase + 4;	/* modem control register */
static int const ComLSR = ComBase + 5;	/* line status register */
static int const ComMSR = ComBase + 6;	/* modem status register */

/* ComICR register */
static int const ComRIEN = 0x01;	/* enable receiver interrupt */
static int const ComTIEN = 0x02;	/* enable transmitter interrupt */
static int const ComSIEN = 0x04;	/* enable receiver status interrupt */
static int const ComMIEN = 0x08;	/* enable modem status interrupt */

/* ComISR register */

static int const ComRSTATUS = 0x06;	/* change in receiver line status */
static int const ComRxRDY = 0x04;	/* receiver data available */
static int const ComTxRDY = 0x02;	/* transmitter holding register empty */
static int const ComMSTATUS = 0x00;	/* change in modem status */

/* ComLSR register */

static int const ComRCA = 0x01;		/* receive char available */
static int const ComOVRRUN = 0x02;	/* receive overrun */
static int const ComPARERR = 0x04;	/* parity error */
static int const ComFRMERR = 0x08;	/* framing/CRC error */
static int const ComBRKDET = 0x10;	/* break detected (null ch + frmerr) */
static int const ComXHRE = 0x20;	/* xmit holding register empty */
static int const ComXSRE = 0x40;	/* xmit shift register empty */


/* ComMSR register */

static int const ComDCTS = 0x01;	/* CTS has changed state */
static int const ComDDSR = 0x02;	/* DSR has changed state */
static int const ComDRI = 0x04;		/* RI has changed state */
static int const ComDDCD = 0x08;	/* DCD has changed state */
static int const ComCTS = 0x10;		/* state of CTS */
static int const ComDSR = 0x20;		/* state of DSR */
static int const ComRI = 0x40;		/* state of RI */
static int const ComDCD = 0x80;		/* state of DCD */



static unsigned char * const ScreenAddr = (unsigned char *) 0xb8000;
static unsigned char Spaces[80*2] = {
' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, 
' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, 
' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, 
' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, 
' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, 
' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, 
' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, 
' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, ' ', 7, 
};
unsigned char ComInitData[] = {
1, 0x00,	/* reset port */
3, 0x80,	/* point to modem speed regs */
0, 0x0c,	/* set speed to 9600 baud */
1, 0x00,
4, 0x0b,	/* set RTS and DTR */
3, 0x03,	/* point to data registers */
1, 0x01,	/* enable rcv intrs (1, 0x00 to disable all intrs) */
0xff
};


PCi586Console::PCi586Console()
{
	unsigned char *p;
	int delay;

	clearScreen();

	if (!Com0Busy)
	{
		/* initialize the serial port */
		for (p = &ComInitData[0]; *p != 0xff; p += 2)
		{
			Out(ComBase + *p, *(p+1));
			for (delay = 0; delay < 500; delay++)
				;
		}
		for (delay = 0; delay < 100000; delay++)
		    /* time for the serial port to reset */
		    ; 
	}

#ifdef DEBUG
	write("PCi586:: Console initialized\n",30);
#endif
}

/*
 * These violate one of my personal pet peeves of acquiring and releasing
 * a lock in two different function!! I really can see any better way.
 * We have to pray the parent class always calls them in pairs.
 */
void
PCi586Console::acquireExclusiveUse()
{
	_exclusiveUseLock.acquire();
}

void
PCi586Console::releaseExclusiveUse()
{
	_exclusiveUseLock.release();
}

int
PCi586Console::write( const char * buf, int count )
{
	for( int i = 0; i < count; i++ ) {
		char c = *buf++;
		putScreen( c );
		putSerial( c );
		if( c == '\n' ) {
		    putScreen( '\r' );
		    putSerial( '\r' );
		}
	}
	return(count);
}

static char ScreenOutputStopped = 0;
static int LeftShift, RightShift, Ctrl;

void
PCi586Console::putScreen( char c )
{
	int i, loop;

	switch(c)
	{
	case '\007': /*beep!*/
		c = In(0x61);
		for(loop = 0; loop < 200; ++loop) {
			Out(0x61, c & 0xfc);
			for(i = 0; i < 300; ++i)
				;
			Out(0x61, (c & 0xfc) | 2);
			for(i = 0; i < 200; ++i)
				;
		}
		Out(0x61, c);
		return;
	case '\t':
		do {
			putScreen(' ');
		} while(currentColumn % 8);
		goto finish;
	case '\b':
		if(currentColumn)
		{
			--currentColumn;
			screenPtr -= 2;
		}
		goto finish;
	case '\r':
		screenPtr -= currentColumn*2;
		currentColumn = 0;
		goto finish;
	case '\n':
		screenPtr += (80 - currentColumn)*2;
		currentColumn = 79;
		break;
	default:
		*screenPtr = c;
		screenPtr += 2;
		//JMT HACK TBD
		if (!InterruptsEnabled()) 
		  {
			/* stop if a 'control-s' is pressed */
			switch(In(0x60))
			{
			case 0x1f:
				if (Ctrl)
					ScreenOutputStopped = 1;
				break;
			case 29:
				Ctrl = 1;
				break;
			case 0x80|29:
				Ctrl = 0;
				break;
			}
		}

		while (ScreenOutputStopped)
		{
		  //JMT HACK TDB
		  if (!InterruptsEnabled()) 
		  {
				/* restart if 'control-q' is pressed */
				switch(In(0x60))
				{
				case 0x10:
					if (Ctrl)
						ScreenOutputStopped = 0;
					break;
				case 29:
					Ctrl = 1;
					break;
				case 0x80|29:
					Ctrl = 0;
					break;
				}
			}
		}
		break;
	}

	if(++currentColumn == 80)
	{
		currentColumn = 0;
		if(currentRow == 23)
		{
			/* scroll! */
			for(i = 0; i < 23; ++i) {
				CopyBlock(ScreenAddr+(i+1)*80*2, ScreenAddr+i*80*2, 80*2);
			}
			screenPtr = ScreenAddr+23*80*2;
			CopyBlock(Spaces, screenPtr, 80*2);
		} else
			++currentRow;
	}

finish:
	updateCursor();
}


static char SerialOutputStopped = 0;

void
PCi586Console::putSerial( char c )
{
	char rc;

	if (!Com0Busy)
	{
		if (In(ComLSR) & ComRCA)
		{
receive_char:
			/*
			 * Receiver ready with data
			 * This code should only be executed when interrupts
			 *   are disabled; throw away anything not control-s or
			 *   control-q.
			 */
			rc = In(ComDAT) & 0x7f;

			if (rc == 0x13)
			{
				/* control - s; wait for control - q */
				while (1)
				{
					if (In(ComLSR) & ComRCA)
					{
						if ((In(ComDAT) & 0x7f) == 0x11)
							break;
					}
				}
			}
			SerialOutputStopped = 0;
		}

		while (SerialOutputStopped || ((In(ComLSR) & ComXHRE) == 0)) {
			/* wait until everything's ready */
			if (In(ComLSR) & ComRCA)
				goto receive_char;
		}

		Out(ComDAT, c);

		if (c == '\n')
			putSerial('\r');
	}
}

void
PCi586Console::clearScreen( )
{
	int i;

	currentRow = 0;
	currentColumn = 0;
	screenPtr = ScreenAddr;
	for(i = 0; i < 24; ++i)
		CopyBlock(Spaces, screenPtr + i * 80 * 2, 80 * 2);
	updateCursor( );
}

void
PCi586Console::updateCursor( )
{
	int pos;

	pos = (screenPtr - ScreenAddr)/2;
	Out(0x3d4, 0xf);
	Out(0x3d5, pos);
	Out(0x3d4, 0xe);
	Out(0x3d5, pos >> 8);
}

// 1 if initialized
int consoleInputInitialized;

void
InitConsole()
{
    /* clear keyboard interrupt mask in the interrupt controller */
    Out(PIC_MASK_REG, In(PIC_MASK_REG) & ~PIC_KEYBOARD_MASK);

    if (!Com0Busy)
	Out(PIC_MASK_REG, In(PIC_MASK_REG) & ~PIC_COM0_MASK);

//  Debug << "InitConsole(): Initializing input queues\n" << eor;
//
//  These moved to PCi586Main(), because of Bjorn's changes to streams.
//         -- Stan
//    consoleInputQueue = new LineBufferedEchoingCharacterQueue( &Console() );
//	   Assert( consoleInputQueue != 0 );
//
//    ((LineBufferedEchoingCharacterQueue *)consoleInputQueue)->
//	setSpecialChar( Backspace, '\b' );

    consoleInputInitialized = 1;    // we're in business!
}

static unsigned char Char_table[] = {
0,
/*1-10*/ '\033', '1', '2', '3', '4', '5', '6', '7', '8', '9',
/*11-20*/ '0', '-', '=', '\b', '\t', 'q', 'w', 'e', 'r', 't',
/*21-30*/ 'y', 'u', 'i', 'o', 'p', '[', ']', '\r', 0, 'a',
/*31-40*/ 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';', '\'',
/*41-50*/ '`', 0, '\\', 'z', 'x', 'c', 'v', 'b', 'n', 'm',
/*51-..*/ ',', '.', '/', 0, 0, 0, ' ' };

static unsigned char Shift_table[] = {
0,
/*1-10*/ '\033', '!', '@', '#', '$', '%', '^', '&', '*', '(',
/*11-20*/ ')', '_', '+', '\b', '\t', 'Q', 'W', 'E', 'R', 'T',
/*21-30*/ 'Y', 'U', 'I', 'O', 'P', '{', '}', '\r', 0, 'A',
/*31-40*/ 'S', 'D', 'F', 'G', 'H', 'H', 'K', 'L', ':', '"',
/*41-50*/ '~', 0, '|', 'Z', 'X', 'C', 'V', 'B', 'N', 'M',
/*51-..*/ '<', '>', '?', 0, 0, 0, ' ' };

# define C(val) val-'a'+1
static unsigned char Ctrl_table[] = {
0,
/*1-10*/ '\033', 0, 0, 0, 0, 0, 0, 0, 0, 0,
/*11-20*/ 0, 0, 0, '\b', '\t', C('q'), C('w'), C('e'), C('r'), C('t'),
/*21-30*/ C('y'), C('u'), C('i'), C('o'), C('p'), 0, 0, '\r', 0, C('a'),
/*31-40*/ C('s'), C('d'), C('f'), C('g'), C('h'), C('j'), C('k'), C('l'), 0, 0,
/*41-50*/ 0, 0, 0, C('z'), C('x'), C('c'), C('v'), C('b'), C('n'), C('m'),
/*51-..*/ 0, 0, 0, 0, 0, 0, ' ' };
# undef C

int
getKeyboardCharacter()
{
    unsigned char c;

    c = In(0x60);
    switch(c)
    {
    case 42:
	LeftShift = 1;
	break;
    case 54:
	RightShift = 1;
	break;
    case 0x80|42:
	LeftShift = 0;
	break;
    case 0x80|54:
	RightShift = 0;
	break;
    case 29:
	Ctrl = 1;
	break;
    case 0x80|29:
	Ctrl = 0;
	break;
    default:
	if(c >= sizeof(Char_table))
	    break;
	if(Ctrl)
	    c = Ctrl_table[c];
	else if(LeftShift || RightShift)
	    c = Shift_table[c];
	else
	    c = Char_table[c];
	switch (c)
	{
	case 0x13:
	    /* control - s */
	    ScreenOutputStopped = 1;
	    break;
	case 0x11:
	    ScreenOutputStopped = 0;
	    break;
	default:
	    return(c);
	}
	break;
    }
    return(-1);
}

void
ConsoleInputInterruptHandler()
{
    int c;
    /* reset the keyboard */
    c = In(0x61);
    Out(0x61, c|0x80);
    Out(0x61, c);

    if ((c = getKeyboardCharacter()) != -1)
    {
	Debug.printf("ConsoleInputInterruptHandler: received %c (%x)\n",
			    c, c);
	// stick the stuff in the In queue for CharacterInputStream::read 
	// to pick up
	char ch;
	ch = c;
	ConsoleInputStream->enqueue( (const char *) &ch, 1 );
    }
    Out(PIC_CTRL_REG, PIC_EOI);
}

int
getSerialCharacter()
{
    if (In(ComLSR) & ComRCA)
    {
	/* receiver ready with data */
	char c = In(ComDAT) & 0x7f;
	switch (c)
	{
	case 0x13:
	    /* control - s */
	    SerialOutputStopped = 1;
	    break;
	case 0x11:
	    SerialOutputStopped = 0;
	    break;
	default:
	    return(c);
	}
    }
    return(-1);
}

void
ConsoleOutputInterruptHandler()
{
    int c;
    if ((c = getSerialCharacter()) != -1)
    {
	Debug.printf("ConsoleOutputInterruptHandler: received %c (%x)\n",
			    c, c);
	// stick the stuff in the In queue for CharacterInputStream::read 
	// to pick up
	char ch;
	ch = c;
	ConsoleInputStream->enqueue( (const char *) &ch, 1 );
    }
    Out(PIC_CTRL_REG, PIC_EOI);
}

int
PCi586ConsoleForUserOutput::write( const char * buf, int count )
{
	return( Console().write( (char *) buf, count ) );
}

