/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: See-Mong Tan (stan@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * PCi586Disk.c: PCi586 Disk Driver
 */

#include "Assert.h"
#include "OutputStream.h"
#include "Raid.h"
#include "Class.h"
#include "Vectors.h"
#include "Lock.h"
#include "MemoryObjectPartition.h"
#include "PhysicalMemoryChain.h"
#include "Util.h"
#include "ThisProcess.h"
#include "Domain.h"
#include "Store.h"
#include "PCi586Disk.h"
#include "PCi586Multiboot.h"

extern Store * MainStore;

extern "C" {
    void PhysCopyBlock( unsigned char *, unsigned char *, int );
    void PhysCopyInWordBlock( int, char *, int ); 
    void CopyInWordBlock( int, char *, int ); 
    void PhysCopyOutWordBlock( int, char *, int );
}

int PCi586DiskBackingStore::initialized;
int PCi586DiskBackingStore::wholePartitionStart;
int PCi586DiskBackingStore::wholePartitionSize;
int PCi586DiskBackingStore::availablePartitions;
int PCi586DiskBackingStore::allocationArray[ 200 ];

PCi586Disk::PCi586Disk( int drive=0 )
     : Disk( 1 )
{
    initDebug();
    Assert( drive == 0 || drive == 1 );

    Debug << "PCi586Disk::PCi586Disk()\n" << eor;

    myDrive = drive;
    
    /*
     *  From multiboot info, fill in drive specific information
     */
    myCylinders = drive_info[myDrive].drive_cylinders;

    Assert( myCylinders != 0 );       // No cylinders mean no drive!
    myHeads = drive_info[myDrive].drive_heads;
    mySectorsPerTrack = drive_info[myDrive].drive_sectors;
    myPrecomp = 0xFF;

    Debug << "# of heads=" << myHeads << 
      ", sectors per track=" << mySectorsPerTrack <<
      ", precomp=" << myPrecomp << "\n" << eor;
	
    /*
     * We're not initialized until all our processes are run.
     */
    initDone = 0;
}


unsigned char
PCi586Disk::waitForControllerReady()
{
    unsigned char status;

    while( (status = In( HD0 + HD_STATUS )) & HD_STATUS_BUSY );

    return status;
}

void
PCi586Disk::initialize()
{
    initDebug();
    Debug << "PCi586Disk::initialize()\n"  << eor;

    /*
     * Set the MemoryObject/MemoryRange instance variables now.
     */
    _log2UnitSize = 9;                       // 512 byte sectors
    _numberOfUnits = 280000;		     // calculate this sometime
					     //   ~ 140 meg

    /*
     *  Set drive controller characteristics for this drive
     */
    Debug << "initialize: Setting Drive Controller Characteristics for drive\n"
	<< eor;

    unsigned char nheads = HD_DHFIXED | (myHeads-1);
    nheads |= myDrive ? HD_DRIVE1 : HD_DRIVE0;
    mutex.acquire();
    Out( HD0 + HD_DRV, nheads );               // head select
    Out( HD0 + HD_NSECT, mySectorsPerTrack );
    Out( HD0 + HD_CMD, HD_SETPARAM );          // fire it up
    waitForControllerReady();
    mutex.release();

    Debug << "initialize: Controller characteristics set\n" << eor;

    Debug << "initialize: Reading the master boot block\n" << eor;

    char * readBuffer= new char[ unitSize() ];    // used for reading stuff in

    doio( (unsigned) 0, 1, readBuffer, unitSize(), PCi586DiskReadOperation );
    mboot = (char *) readBuffer;
    Assert( mboot.signature == MBB_MAGIC ) ;

    // Get the initial partition table under DOS
    // and scan it for an active UNIX partition
    Debug << "initialize: Looking for an active UNIX partition\n" << eor;
    PCi586DiskInitialPartitionTable * ipart = 
	(PCi586DiskInitialPartitionTable *) &(mboot.parts);

    int i;
    for(i = 0; i < FD_NUMPART; i++) {
	Debug << " Partition #" << i << "has bootid=" << ipart->bootid <<
	    ", systid =" << ipart->systid << "\n" << eor;
	if (ipart->bootid == BOOTID_ACTIVE)
	    break;
	else
	    ipart++;     // else examine the next partition
    }
    Assert( i != FD_NUMPART );

    // We've found an active UNIX partition here, so good
    // save some data about it
    startOfUnixPartition = (int) ipart->relsect;
    sizeOfUnixPartition = (int) ipart->numsect;
    beginTrackOfUnixPartition = (int) ipart->relsect / mySectorsPerTrack;
    Debug << "initialize: start=" << startOfUnixPartition << 
	", size=" << sizeOfUnixPartition << 
	    ", beginning track=" << beginTrackOfUnixPartition << 
		"\n" << eor;

    // Find the physical description area
    int pdaSector = startOfUnixPartition + 
	PCi586DiskPhysicalDescriptionAreaOffset;

    Debug << "initialize: Reading physical description area\n" << eor;

    doio( pdaSector, 1, readBuffer, unitSize(), PCi586DiskReadOperation );
    myPDA = readBuffer;
    Assert( myPDA.sanity == VALID_PD );

    // Now read the Volume Table of Contents
    int vtocSector = (int) startOfUnixPartition + 
	((int) myPDA.vtoc_ptr >> SECSHIFT);
    if (vtocSector != pdaSector) {
	Debug << "initialize: VTOC is in sector=" << vtocSector 
	    << "\n" << eor;
	// This means the Volume Table of Contents is in some other sector
	doio( vtocSector, 1, readBuffer, unitSize(), PCi586DiskReadOperation );
    }

    int vtocOffset = (int) myPDA.vtoc_ptr & SECMASK;
    myVTOC = readBuffer + vtocOffset;

    Assert( myVTOC.v_sanity == VTOC_SANE );
    Assert( myVTOC.v_part[0].p_start == startOfUnixPartition );

    Debug << "initialize: Correct Whole UNIX partition start, GOOD\n"
	<< eor;

    // Read the alternate sector/track information
    Debug << "initialize: attempting to find alternates\n" << eor;

    long block = startOfUnixPartition + ( myPDA.alt_ptr >> SECSHIFT );
    long offset = myPDA.alt_ptr & SECMASK;
    // n <= # of sectors needed for alt table.
    long n = (offset + myPDA.alt_len + SECMASK) >> SECSHIFT;
    if ( block != pdaSector || n > 1 ) {
	// Alternates table is in a different sector
	for( int i = 0; i < n; i++ ) {
	doio( (unsigned int) block+i, 1, readBuffer, unitSize(),
						PCi586DiskReadOperation );
	CopyBlock( (unsigned char *) readBuffer, 
		  ((unsigned char *) &myAltInfo) + ( i * unitSize() ),
		  unitSize() );
	}
    }
    else {
	// else copy it from this buffer into the structure directly
	CopyBlock( (unsigned char *) readBuffer + offset, 
		  (unsigned char *) &myAltInfo,
		  sizeof( PCi586DiskAltInfo ) );
    }

    Assert( myAltInfo.alt_sanity == ALT_SANITY );
    Assert( myAltInfo.alt_version == ALT_VERSION );
    Assert( myAltInfo.alt_trk.alt_used <= myAltInfo.alt_trk.alt_reserved );
    Assert( myAltInfo.alt_sec.alt_used <= myAltInfo.alt_sec.alt_reserved );

    // Initialize the bad track/sector bitmaps
    long mapsize = (( myVTOC.v_part[0].p_size + myPDA.sectors - 1 ) / myPDA.sectors 
		   + 7) / 8;
    Debug << "PCi586Disk::initialize: creating bit maps of size " << (int) mapsize
	<< "\n" << eor;
    bth_trkmap = new char[mapsize];
    Assert( bth_trkmap != 0 );
    bbh_trkmap = new char[mapsize];
    Assert( bbh_trkmap != 0 );
    for( i = 0; i < mapsize; i++ ) {
	bth_trkmap[i] = (char) 0;
	bbh_trkmap[i] = (char) 0;
    }
    long lsttrk = (myVTOC.v_part[0].p_start + myVTOC.v_part[0].p_size -1) / 
	myPDA.sectors;
    long begtrk = startOfUnixPartition / myPDA.sectors;
    for( n = 0; n < myAltInfo.alt_trk.alt_used; n++ ) {
	if ( myAltInfo.alt_trk.alt_bad[n] != -1 ) {
	    if ( myAltInfo.alt_trk.alt_bad[n] > lsttrk || 
		 myAltInfo.alt_trk.alt_bad[n] < -1 )
		break;
	    bitSet( bth_trkmap, myAltInfo.alt_trk.alt_bad[n] - begtrk );
	}
    }
    for( n = 0; n < myAltInfo.alt_sec.alt_used; n++ ) {
	if ( myAltInfo.alt_sec.alt_bad[n] != -1 ) {
	    long trk = myAltInfo.alt_sec.alt_bad[n] / myPDA.sectors;
	    if ( trk > lsttrk || trk >= -1 )
		break;
	    bitSet( bth_trkmap, trk - begtrk );
	}
    }

    delete readBuffer;
    initDone = 1;
}

void
PCi586Disk::bitSet( char * bitmap, long bitno )
{
    bitmap[bitno >> 3] |= (char) (1 << (bitno & 7));
}

int
PCi586Disk::bitTest( char * bitmap, long bitno )
{
    return bitmap[bitno >> 3] & (1 << (bitno & 7));
}

int
PCi586Disk::tableScan( long * table, long value, unsigned short length )
{
    for (int count = (int) length; count-- > 0;) {
	if (*table++ == value)
	    return length - count - 1;
    }
    return -1;
}
     
int
PCi586Disk::basicDoio( unsigned int, char *, int )
{
    /*
     * NOTICE:  Someday somebody should take the essential part out of
     *   doio and implement basicDoio instead.  For now we are a subclass
     *   of Disk but overload the doio, most of which is already implemented
     *   in Disk::doio.  At that time should also be able to stop overloading
     *   the read() and write() methods from Disk.
     */
    Assert( NOTREACHED );
    return( DiskOperFailed );
}

int
PCi586Disk::reset()
{
    /* nothing for now */
    return( DiskOperSucceded );
}

int
PCi586Disk::read( unsigned int start, int count, char * addr, int nbytes )
{
    Debug <<  "PCi586Disk::read( start=" << start << ", count=" <<  count 
	<< ", addr=" << asHex( (int) addr ) << " )\n" << eor;

    // Lock, so no one else bothers us
    mutex.acquire();
    int retval = doio( start, count, addr, nbytes, PCi586DiskReadOperation );
    mutex.release();

    return retval;
}

int
PCi586Disk::write( unsigned int start, int count, char * addr, int nbytes )
{
    Debug << "PCi586Disk::write( start=" << start << ", count=" << count <<
     ", addr=" << asHex( (int) addr ) << ", nbytes=" << nbytes << " )\n" << eor;
    
    mutex.acquire();
    int retval = doio( start, count, addr, nbytes, PCi586DiskWriteOperation );
    mutex.release();

    return retval;
}

int
PCi586Disk::doio( unsigned start, int count, char * addr, int nbytes, int op )
{
    PhysicalMemoryChain chain = thisProcess()->domain()->constructChain(
					addr, nbytes );
    thisProcess()->domain()->makeAddressable(addr, nbytes);

    Debug << "PCi586Disk::doio( start=" << start << ", count=" <<
	count << ", chain=" << &chain << " )\n" << eor;

    if ( chain.lengthInBytes() < unitSize() * count ) {
	Console() << "PCi586Disk::doio: insufficient chain length\n" << eor;
	return 0;
    }

    // iterate over the chain for count * unitSize() number of bytes
    PhysicalMemoryChainIterator chainIterator( &chain );
    PhysicalMemoryChain * subChain = chainIterator.first();

    int numberOfSectorsRead;
    // for count # of items, try to read from start
    for(numberOfSectorsRead = 0; numberOfSectorsRead < count;
	numberOfSectorsRead++, start++) {

        // do bad block mapping only if initialized
        int mappedStart;
        if ( initDone )
	    mappedStart = badBlockMap( start );
	else
	    mappedStart = start;

        Assert( subChain != 0 );
	// want to iterate over the subChain for one sector size of memory
	PhysicalMemoryChainContiguousBlockIterator 
	    contiguousBlockIterator( subChain, subChain->lengthInBytes() );

	unsigned int contiguousBlockSize = 0;
	char * transferAddress = (char *) 
	    contiguousBlockIterator.next( contiguousBlockSize );
	Debug << "transfer Address=" << (int) transferAddress << 
	    ", contiguousBlockSize=" << contiguousBlockSize << "\n" << eor;

        Debug << "Setting drive controller registers for operation\n" << eor;
	unsigned char precomp = myPrecomp >> 2;
	unsigned int track = mappedStart / mySectorsPerTrack;
	unsigned char drv = track % myHeads;
	unsigned int cyl = track / myHeads; // which cylinder
	unsigned char sect = mappedStart % mySectorsPerTrack;	// sector in track
	unsigned char nsect = 1;			// better be 1
	unsigned char cmd = (op == PCi586DiskReadOperation ? 
			     HD_RDSEC : HD_WRSEC);
	
	drv |= (HD_DHFIXED | (myDrive << 4));
	
	unsigned char status = waitForControllerReady();
	if ((status & HD_STATUS_WRFAULT) || ! (status & HD_STATUS_READY)) {
	    Console() << "doio: cannot do IO, HD fault!\n" << eor;
	    break;
	}
	
	//  Decide if we need the extra disk heads parameter, and set 
	//  the register accordingly
	Out( FDR, 
	    (myHeads > 8) ? 
	    (unsigned char ) HD_EXTRAHDS :
	    (unsigned char) HD_NOEXTRAHDS );

	Out( HD0 + HD_PRECOMP, precomp ); // write precomp
	Out( HD0 + HD_NSECT, nsect );	  // sector count

	// The controller sector numbering starts from 1, and we 
	// have been using base sector of 0, thus we add 1 to it.
	Out( HD0 + HD_SECT, sect + 1 );	// sector number
	Out( HD0 + HD_LCYL, cyl & 0xff ); // cylinder low byte
	Out( HD0 + HD_HCYL, (cyl >> 8) & 0xff ); // cylinder high byte
	Out( HD0 + HD_DRV, drv );		 // drive/head register
	Out( HD0 + HD_CMD, cmd );		 // status/command register
	
	//  Wait for controller ready, and check for data request.
	//  Then block transfer bytes either to or from data sector buffer.

	status = waitForControllerReady();

	if (status & HD_STATUS_WRFAULT) {
	    Console() << "PCi586Disk::doio: Hard disk controller write fault\n"
		<< eor;
	    break;
	}
	if (status & HD_STATUS_ERROR) {
	    Console() << "PCi586Disk::doio: Hard disk hard error\n" << eor;
	    break;
	}

	if (op == PCi586DiskReadOperation) { // 'twas a read
	    status = waitForControllerReady();
	    if ( (status & HD_STATUS_DATARQ) == 0 ) {
		Console() << "PCi586Disk::doio: - Data Request error on read\n"
		    << eor;
		break;
	    }
	    Debug << "PCi586Disk::doio: Reading sector " << mappedStart << 
		" from sector buffer\n" << eor;

 	    if ( contiguousBlockSize >= unitSize() ) {
		Debug << "PCi586Disk::doio: Copy is contiguous\n" << eor;

		PhysCopyInWordBlock( HD0 + HD_DATA, transferAddress, 
				     unitSize() / 2 );

		Debug << "PCi586Disk::doio: Safe Copy\n" << eor;
	    }
	    else {    // still a read, but< discontiguous
		int firstTransferSize = contiguousBlockSize;
		Debug << "PCi586Disk::doio: Discontiguous copy\n" << eor;
		// get a free page from the main store, use it as a
		// temporary space to copy to
		PhysicallyAddressableUnit * unit = MainStore->alloc();
		PhysCopyInWordBlock( HD0 + HD_DATA, (char *) unit->address(),
				     unitSize() / 2 );		
		// now transfer from the temporary page to the real chain's
		// first part, then the second part.
		PhysCopyBlock( (unsigned char *) unit->address(),
			       (unsigned char *) transferAddress,
			      firstTransferSize );
		int nextTransferSize = unitSize() - firstTransferSize;
		transferAddress = (char *)
		    contiguousBlockIterator.next( contiguousBlockSize );
		// better have enough space in the next contiguous block!
		Assert( contiguousBlockSize >= nextTransferSize );
		PhysCopyBlock( (unsigned char *) unit->address() +
			      firstTransferSize,
			       (unsigned char *) transferAddress,
			       nextTransferSize );
		MainStore->dealloc( unit );    // return to main store
	    }
	}
        else if ( op == PCi586DiskWriteOperation ) { // 'twas a write instead
	    Debug << "doio: Writing block to sector buffer " <<
		mappedStart << "\n" << eor;
	    
	    Assert( contiguousBlockSize >= unitSize() );
	    // Assume contiguous
	    PhysCopyOutWordBlock( HD0 + HD_DATA, transferAddress,
				 unitSize() / 2 );
	}
	else {
	    Debug << "PCi586Disk::doio: Bad Op\n" << eor;
	    Assert( NOTREACHED );
	}

	// go onto next part of the chain
	subChain = chainIterator.next( unitSize() );
    }
    Debug << "doio: Successful\n" << eor;

    return numberOfSectorsRead;
}

unsigned int
PCi586Disk::badBlockMap( unsigned int block )
{
    long firstsec = (long) block;
    long sec = firstsec;
    long trk = firstsec / myPDA.sectors;
    PCi586DiskAltTable * secptr = &(myAltInfo.alt_sec);
    PCi586DiskAltTable * trkptr = &(myAltInfo.alt_trk);
    long lastsec = firstsec;
    long begtrk = startOfUnixPartition / myPDA.sectors;
    // determine if sector is potentially bad
    if ( bitTest( bth_trkmap, trk - begtrk ) ) {
	// sector is potentially bad, see if it's in the alternate
	// sector table.
	if ( bitTest( bbh_trkmap, trk -begtrk ) ) {
	    int ndx = tableScan( &(secptr->alt_bad[0]), sec, secptr->alt_used );
	    if ( ndx != -1 ) {
		Debug << "bad block mapped in alt sector table\n" << eor;
		return (unsigned int) ( ndx + secptr->alt_base );
	    }
	}
	else {
	    // bit not set in bbh_trkmap, block must be
	    // a track recorded in the alternate track table
	    int ndx = tableScan( &(trkptr->alt_bad[0]), trk, trkptr->alt_used );
	    Debug << "bad block mapped in alt track table\n" << eor;
	    Assert( ndx != -1 );
	    return (unsigned int) ( trkptr->alt_base + (ndx * myPDA.sectors) 
				   + (sec % myPDA.sectors) );
	}
    }
    else
	return block;
}

int
PCi586Disk::partitions()
{
    return( 128 );
}

int
PCi586Disk::locatePartition( int idNumber, int &start, int &units )
{
    Debug << "PCi586Disk::locatePartition( idNumber:" << 
		idNumber << " )\n" << eor;

    int tag;
    switch( idNumber )
    {
    case 0:
	// for id number 0 we use the root partition
    	tag = V_ROOT;
	break;
    case 1:
	// this isn't used yet but may come in handy
	tag = V_USR;
	break;
    default:
	Assert( NOTREACHED );
    }

    // Find the partition bounds
    const int NumberOfPartitions = myVTOC.v_nparts;

    // partition information part of volume table of contents
    const PCi586DiskPartitionHeader * partitionHdr = myVTOC.v_part;

    int i;
    for(i = 0; i < NumberOfPartitions; i++) {
	if (partitionHdr->p_tag == tag)
	    break;                     // found it
	else
	    partitionHdr++;            // try next partition
    }
    if (i == NumberOfPartitions) {     // couldn't find it
	Debug << "basicOpen: can't find object with proper tag\n" << eor;
	return -1;
    }

    Debug << "basicOpen: gives partition=" << idNumber <<
	", start=" << (int) partitionHdr->p_start << ", size=" << 
	    (int) partitionHdr->p_size <<  "\n" << eor;
    
    start = (int) partitionHdr->p_start;
    units = (int) partitionHdr->p_size;

    return( 0 );
}

PCi586DiskBackingStore::PCi586DiskBackingStore( PCi586Disk * source,
						 int start, int size )
     : PersistentMemoryObject( sizeOfSingleSwapPartition, 9, 0, 0 )
{
    if ( ! initialized ) {
	initialized = 1;
	initialize( start, size );
    }
    Debug << "PCi586DiskBackingStore::PCi586DiskBackingStore()\n" << eor;
    theDisk = source;
    int i;
    for( i = 0; i < availablePartitions; i++ ) {
	if ( allocationArray[i] == 0 ) {
	    allocationArray[i] = 1;
	    break;
	}
    }
    Assert( i != availablePartitions );
    
    startOfThisSwapPartition = wholePartitionStart + 
	sizeOfSingleSwapPartition * i;
    sizeOfThisSwapPartition = sizeOfSingleSwapPartition;
}

void
PCi586DiskBackingStore::initialize( int partitionStart, int partitionSize )
{
    wholePartitionStart = partitionStart;
    wholePartitionSize = partitionSize;
    availablePartitions = wholePartitionSize / sizeOfSingleSwapPartition;
    for( int i = 0; i < availablePartitions; i++ )
	allocationArray[i] = 0;
}

PCi586DiskBackingStore::~PCi586DiskBackingStore()
{
    int whichPartition = startOfThisSwapPartition / sizeOfSingleSwapPartition;
    allocationArray[ whichPartition ] = 0;
}

int
PCi586DiskBackingStore::setNumberOfUnits( unsigned int units )
{
    if ( units > sizeOfThisSwapPartition )
	units = sizeOfThisSwapPartition;

    return( MemoryObject::setNumberOfUnits( units ) );
}

int
PCi586DiskBackingStore::read( unsigned int start, int count,
			      char * addr, int nbytes )
{
    Assert( start + count <= sizeOfThisSwapPartition );
    mutex.acquire();
    int retval = theDisk->doio( start + startOfThisSwapPartition,
			count, addr, nbytes, PCi586DiskReadOperation );
    mutex.release();
    return( retval );
}

int
PCi586DiskBackingStore::write( unsigned int start, int count,
			      char * addr, int nbytes )
{
    Assert( start + count <= sizeOfThisSwapPartition );  // don't overrun
    if ( start + count > numberOfUnits() )
	setNumberOfUnits( start + count );

    mutex.acquire();
    int retval = theDisk->doio( start + startOfThisSwapPartition,
		count, addr, nbytes, PCi586DiskWriteOperation );
    mutex.release();

    return retval;
}


#include "ClassInstaller.h"
DEFINECLASS("PCi586Disk",PCi586Disk,Disk,0,0);
DEFINECLASS("PCi586DiskBackingStore",PCi586DiskBackingStore,PersistentMemoryObject,0,0);
