/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 *
 * Copyright (c) 1987,1988,1989,1990,1991  The University of Illinois
 *                                         Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 * 
 * Author: David B. Pointer (dpointer@csrd.uiuc.edu)
 * Contributing Author: Amit Parghi (parghi@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * PCi586EthernetDriver.cc: class implementation of the 3com 3c503 
 *   Etherlink II ethernet controller card in the PCi586 machine.
 */
//#define DEBUG            // enables debug statements

	//  Note that for testing: TEST and DEBUG should always be defined,
	//  and *one* and *only*one* of TEST1 and TEST2 should be defined.
// #define TEST		// enables card receive / transmit testing
// #define TEST1	// level 1 testing -- transmit/receive test packets
// #define TEST2	// level 2 testing -- check packet sequencing

#if defined(TEST1) && defined(TEST2)
This is a syntax error to stop compilation in case of bad definitions!
#endif

#include "Assert.h"
#include "Process.h"
#include "Util.h"
#include "OutputStream.h"
#include "Raid.h"
#include "Class.h"
#include "Semaphore.h"
#include "CPU.h"
#include "Interrupt.h"
#include "Vectors.h"

#include "Types.h"

#include "NetworkBuffer.h"
#include "NetworkDriver.h"
#include "EthernetAdministrator.h"
#include "PCi586InterruptController.h"
#include "PCi586EthernetDriver.h"


static Semaphore * ethernetTxBufferMutex;

//---class PCi586EthernetDriverInterrupt---------------------------------
//
//  For PCi586EthernetDriver interrupt support.
//------------------------------------------------------------------------
class PCi586EthernetDriverInterrupt : public Interrupt {
    PCi586EthernetDriver * ethernetDriver;
protected:
    void basicRaise( char * );
public:
    PCi586EthernetDriverInterrupt(PCi586EthernetDriver * edPtr);
};

static inline
PCi586EthernetDriverInterrupt::
PCi586EthernetDriverInterrupt(PCi586EthernetDriver * edPtr) : Interrupt()
{
   Debug << this << " PCi586EthernetDriverInterrupt()\n" << eor;

   ethernetDriver = edPtr;
}

void
PCi586EthernetDriverInterrupt::basicRaise( char * )
{
  Debug << this << " basicRaise()\n" << eor;

  // Check if NIC is windowed into IO space and that the NIC
  // page 0 registers are selected.
  Assert( (In(GA_CTRL) & (GA_CTRL_ealo | GA_CTRL_eahi)) == 0 );
  Assert( (In(NIC_CR) & (NIC_CR_ps0 | NIC_CR_ps1)) == 0 );

  unsigned char interruptStatusRegister = In( NIC_ISR );
  ethernetDriver->ethernetInterruptHandler( interruptStatusRegister );
  Out( NIC_ISR, NIC_ISR_resetISR );

  Out(PIC_CTRL_REG, PIC_EOI);

}  
//------------------------------------------------------------------------

//---PCi586EthernetDriver::PCi586EthernetDriver()-----------------------
//
//  Setup the 3c503 ethernet card:
//    1. Set the cold system reset vector.
//    2. Get the card's ethernet station address from its prom.
//    3. Assign the multicast address.
//    4. Configure the card for an 8k shared memory buffer.
//    5. Test the memory buffer.
//    6. Setup receive buffer lower and upper bound addresses.
//    7. Check and make sure that ioBase (what the driver thinks the
//       I/O base address is) and GA_BCFR (what the board thinks
//       the I/O base address is) agree.
//    8. Setup the gate array interrupt register.
//    9. Set the gate array DMA registers to recommended values.
//    10. Initialize the NIC (the DP8390 LAN controller chip)
//    11. Setup the Choices interupt handler.
//    12. If TEST is defined, and the host machine is not publius,
//        transmit a burst of 4 ethernet packets to publius.
//
//  When all the dust settles, you'll have a 3c503 ethernet card:
//    - with 1 transmit buffer, 4 receive buffers in the CPU address space
//    - using BNC (ethernet thin net cable - uses in card transceiver)
//    - using interrupt 5
//    
//---WARNING---WARNING---WARNING---WARNING---WARNING---WARNING---WARNING---
//  Any writes to GA_CTRL *must* have GA_CTRL_xsel set since only the BNC
//  ethernet connector will be used.
//------------------------------------------------------------------------
PCi586EthernetDriver::PCi586EthernetDriver(unsigned char debugMask)
    : EthernetDriver(debugMask)
{

  ethernetTxBufferMutex = new Semaphore( 1 );

  // Perform a soft reset of the gate array controller.
  Out( GA_CTRL, GA_CTRL_rst | GA_CTRL_xsel );
  Out( GA_CTRL, GA_CTRL_xsel );
  Debug << this << " Ethernet controller reset\n" << eor;

  // Setup reset vector (cold system reset): if the address represented 
  // here is detected along with MEMRD- active on the PC bus, then 
  // memory mapped mode is reset (the 8k buffer memory disapears from
  // the memory address space).  Put another way, GA_GACFR bit 3 is
  // reset when the CPU reads this address.
  Out( GA_VPTR2, GA_VPTR2_value );
  Out( GA_VPTR1, GA_VPTR1_value );
  Out( GA_VPTR0, GA_VPTR0_value );
  Debug << this << " Cold reset vector set to " 
        << asHex(GA_VPTR2_value << 12 | 
                 GA_VPTR1_value << 4 | 
                 GA_VPTR0_value >> 4) << "\n" << eor;

  // Get the ethernet station address from the card's prom and
  // store it in a local private array.  Reading the prom is done
  // by first writing to the gate array control register to
  // window the ethernet address prom bytes 15:0 into the I/O
  // base address.  Bytes 5:0 of the prom can then be read
  // and stored in a private array.  MSB = byte 5.
  Out( GA_CTRL, (GA_CTRL_ealo | GA_CTRL_xsel) );
  etherAddress[5] = In( EA_BYTE5 );
  etherAddress[4] = In( EA_BYTE4 );
  etherAddress[3] = In( EA_BYTE3 );
  etherAddress[2] = In( EA_BYTE2 );
  etherAddress[1] = In( EA_BYTE1 );
  etherAddress[0] = In( EA_BYTE0 );
  Debug << "Ethernet Address is " 
        << etherAddress[5] << ":" << etherAddress[4] << ":" 
        << etherAddress[3] << ":" << etherAddress[2] << ":" 
        << etherAddress[1] << ":" << etherAddress[0] << "\n" << eor;


  _ethernetAddress = new EthernetAddress( etherAddress );


  // Assign the multicast address here.
  multicastAddress[7] = '\0';
  multicastAddress[6] = '\0';
  multicastAddress[5] = '\0';
  multicastAddress[4] = '\0';
  multicastAddress[3] = '\0';
  multicastAddress[2] = '\0';
  multicastAddress[1] = '\0';
  multicastAddress[0] = '\0';
  Debug << "Assigned Multicast address is "
        << multicastAddress[7] << ":" << multicastAddress[6] << ":"
        << multicastAddress[5] << ":" << multicastAddress[4] << ":"
        << multicastAddress[3] << ":" << multicastAddress[2] << ":"
        << multicastAddress[1] << ":" << multicastAddress[0] << "\n" << eor;

  // Configure the card for an 8k shared memory buffer. The buffer
  // will contain the transmit and receive buffers used by the
  // NIC.  These buffers will be in the CPU memory address space
  // starting at the address defined by the GA_PCFR register.
  Out( GA_GACFR, GA_GACFR_rsel | GA_GACFR_msb0 );
  int memoryOptionCheck = 0;
  unsigned char memoryOption = In( GA_PCFR );
  switch( memoryOption ) {
  case 0x80: memoryBase = (unsigned char *) 0xdc000; break;
  case 0x40: memoryBase = (unsigned char *) 0xd8000; break;
  case 0x20: memoryBase = (unsigned char *) 0xcc000; break;
  case 0x10: memoryBase = (unsigned char *) 0xc8000; break;
  default: 
    Debug << this << " Invalid PCFR register: (" 
          << asHex(memoryOption) << ")\n" << eor;
    memoryOptionCheck = 1;
    break;
  }
  Assert( memoryOptionCheck == 0 );

  // Test the shared memory buffer's integrity.
  int result = testMemory();
  if( result != memoryError )
    Debug << this << " Memory tests pass\n" << eor;
  Assert( result == noMemoryError );

  // Configure the shared memory buffer into 5 buffers:
  //   1 transimit buffer (1536 bytes max) from 0x2000-0x25ff
  //   ( memoryBase to memoryBase + 0x5ff )
  //   4 receive buffers (1664 bytes max each) from 0x2600-0x3fff
  //   ( memoryBase + 0x600 to memoryBase + 0x1fff )
  Out( GA_PSTR, GA_PSTR_pgstrt );
  Out( GA_PSPR, GA_PSPR_pgstop );
  Debug << this << " Shared memory configured for 1 tx, 4 rx buffers\n" << eor;

  //  The IO base address had to be set before any access to the board
  //  could be made. However, once the IO base address is set in
  //  PCi586EthernetDriver.h, the actual jumpered IO base address on 
  //  the board can be read and checked against what this code thinks
  //  is the board's IO base address.  Sanity check time, in other
  //  words.
  int ioBaseRegCheck = 0;
  unsigned char ioBaseReg = In( GA_BCFR );
  switch( ioBaseReg ) {
  case 0x80: Debug << this << " IO base address = 0x300\n" << eor;
             Assert( ioBase == 0x300 );  break;
  case 0x40: Debug << this << " IO base address = 0x310\n" << eor;
             Assert( ioBase == 0x310 );  break;
  case 0x20: Debug << this << " IO base address = 0x330\n" << eor;
             Assert( ioBase == 0x330 );  break;
  case 0x10: Debug << this << " IO base address = 0x350\n" << eor;
             Assert( ioBase == 0x350 );  break;
  case 0x08: Debug << this << " IO base address = 0x250\n" << eor;
             Assert( ioBase == 0x250 );  break;
  case 0x04: Debug << this << " IO base address = 0x280\n" << eor;
             Assert( ioBase == 0x280 );  break;
  case 0x02: Debug << this << " IO base address = 0x2a0\n" << eor;
             Assert( ioBase == 0x2a0 );  break;
  case 0x01: Debug << this << " IO base address = 0x2e0\n" << eor;
             Assert( ioBase == 0x2e0 );  break;
  default: Debug << this << " Invalid GA_BCFR register: ("
                 << asHex(ioBaseReg) << ")\n" << eor;
           ioBaseRegCheck = 1;
           break;
  }
  Assert( ioBaseRegCheck == 0 );

  // Setup gate array interrupt register
  Out( GA_IDCFR, GA_IDCFR_irq5 );
  Debug << this << " Using interrupt 5\n" << eor;

  // Setup DMA registers with recommended values.  There is some
  // confusion here about DMA.  This program does not intend
  // DMA between the PCi586 PC and the 3c503 card since the
  // transmit and receive buffer contents are mapped into CPU
  // memory space.  On the other hand, there are some hints in
  // the 3c503 technical reference that there is a DMA controller
  // in the gate array that transfers data between the NIC and
  // buffer memory.  I don't know if all these values are used
  // for the latter purpose, but I figure I'd better set the registers
  // up with their recommended values to be sure.
  Out( GA_DQTR, GA_DQTR_value );
  Out( GA_DAMSB, GA_DAMSB_value );
  Out( GA_DALSB, GA_DALSB_value );

  stopNIC();
  restartNIC();

  // Setup Choices interrupt handler.
  Interrupt * interrupt = new PCi586EthernetDriverInterrupt( this );
  thisCPU()->setInterrupt(ETH_Vector, interrupt);
  Out(PIC_MASK_REG, In(PIC_MASK_REG) & ~PIC_ETH_MASK);

#ifdef TEST
  // Transmit packets to publius if this machine is not publius. 
  // Receive on publius will pick it up and dump it to the diagnostic console.
 
  // Spawn off a seperate process to do the sending.  We must return
  // so that the rest of the devices get initialized.
  if( (etherAddress[5] != 69) || (etherAddress[4] != 38) ||
      (etherAddress[3] != 60) || (etherAddress[2] != 140) ||
      (etherAddress[1] != 96) || (etherAddress[0] != 2) ) {
     extern void EthernetTestEntry( int );
     Process * proc = new 
           SystemProcess( (PFV) EthernetTestEntry, thisProcess()->domain(),
                          normal, (int) this );
     Assert( proc != 0 );
     proc->ready();
  }
#endif /* TEST */
}

#ifdef TEST
void
EthernetTestEntry( int arg ) 
{
  // transmit packet to publius.
  // testing only!!!!!
  PCi586EthernetDriver * theDriver = (PCi586EthernetDriver *) arg;

  //EthernetPacket * testPacket = new EthernetPacket;
  NetworkBuffer * testPacket = new NetworkBuffer( 64, 0 );
  const unsigned short testPacketDataLength = 64;

	// As a warning to future authors: Please DON'T put ethernet addresses
	// in decimal -- they're hard to understand.  If an address format
	// commonly uses hexadecimal constants, then use hex in your code!!
	// If an address format (eg IP) uses decimal, then use that!   // Amit
  testPacket->packet()[0] = 0x02;
  testPacket->packet()[1] = 0x60;
  testPacket->packet()[2] = 0x8c;
  // testPacket->packet()[3] = 0x3c;	// Address = 02:60:8c:3c:26:45
  // testPacket->packet()[4] = 0x26;	//	publius
  // testPacket->packet()[5] = 0x45;
  testPacket->packet()[3] = 0x0e;	// Address = 02:60:8c:0e:66:fb
  testPacket->packet()[4] = 0x66;	//	popilius
  testPacket->packet()[5] = 0xfb;
  
  //testPacket->packetDataLength = testPacketDataLength;
  testPacket->setLength(testPacketDataLength);
  testPacket->packet()[12] = (EthernetSpecialTestType >> 8);
  testPacket->packet()[13] = (EthernetSpecialTestType & 0xFF);

	//  I'm adding this sequence number to each packet so as to
	//  check what packets we're getting at the other end, because
	//  I'm suspicious (esp. when, after a few packets have been
	//  received by publius, publius seems to go into some sort
	//  of loop ... very odd).		-- Amit, 1991 01 25
  unsigned short *sequenceNumber = (unsigned short *) testPacket->packet()[14];
  unsigned short counter = 1;

  //for( int i = 0; i < testPacketDataLength; i++ )
  //  testPacket->packetData[i] = 0x55;
  for ( int i = 16 ; i < testPacketDataLength ; i++ ) {
    testPacket->packet()[i] = 0x55;
  }

  for(;;) {
    char * dumpData = testPacket->packet();
    for( int burst = 0; burst < 4; burst++ ) {
      Debug << "*********** TRANSMITTING PACKET WITH SEQUENCE NUMBER " <<
		asHex(counter) << " ***********\n" << eor;
      *sequenceNumber = counter++;
      theDriver->transmitPacket( testPacket );
      //Debug << "packet:  " <<
	//asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	//asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	//asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	//asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	//asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	//asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	//asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	//asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	//asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	//asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	//"\n" << eor;
    }
    for( unsigned int j = 0; j < 2000000; j += 1 );
  }

  // not really reached
  delete testPacket;
}
#endif


//---PCi586EthernetDriver::testMemory()----------------------------------
//
//  Performs a data pattern test on the 8k shared memory starting at
//  the memory address defined by the card. Each data pattern is written
//  out to the entire 8k and then all read back.  This action is
//  repeated for all data patterns.
//
//  Future improvements: could use an address test, too.
//------------------------------------------------------------------------
int
PCi586EthernetDriver::testMemory() {

  Debug << this << " testMemory()\n" << eor;

  const int memoryByteCount = 8192;

  unsigned char *memoryPtr;
  int errorReturn = noMemoryError;

  // data pattern array
  const int numberOfTestDataPatterns = 4;
  unsigned char testData[numberOfTestDataPatterns];
  testData[0] = 0x0;
  testData[1] = 0xff;
  testData[2] = 0x55;
  testData[3] = 0xaa;

  for( int pattern = 0; pattern < numberOfTestDataPatterns; pattern++ ) {

    // write out data pattern to entire ram
    memoryPtr = (unsigned char *) memoryBase;
    for( int memCount = 0; memCount < memoryByteCount; memCount++ )
      *memoryPtr = testData[pattern];

    // now read entire ram back and check each byte
    memoryPtr = (unsigned char *) memoryBase;
    for( memCount = 0; memCount < memoryByteCount; memCount++ ) {
      if(*memoryPtr != testData[pattern]) {
        unsigned char expected = *memoryPtr;
        Debug << this << " Memory error, Address: " 
              << asAddress(memoryPtr)
              << " returns " << asHex(expected) << " instead of "
              << asHex(testData[pattern]) << "\n" << eor;
        errorReturn = memoryError;
      }
    }
  }
  return( errorReturn );  
}

//---stopNIC()------------------------------------------------------------
//
//  Stops the NIC ethernet controller and aborts any DMA in progress.
//  Waits until the NIC indicates that it has indeed reset.
//------------------------------------------------------------------------
void
PCi586EthernetDriver::stopNIC() {

  Debug << this << " resetNIC()\n" << eor;

  Out( GA_CTRL, GA_CTRL_xsel );                    // window NIC into IO space
  Out( NIC_CR, NIC_CR_rd2 | NIC_CR_stp );    
  while( ((In(NIC_ISR) & NIC_ISR_rst) == 0) )
    ;

}

//---PCi586EthernetDriver::restartNIC()----------------------------------
//
// Initialize the ethernet controller chip (NIC aka DP8390).
// Procedure recommended (demanded) by the 3c503 technical reference:
//   1. Stop NIC operation and abort remote DMA (if any)
//   2. Select byte-wide transfers, dual 16-bit DMA mode (?),
//      all packets removed from buffer ring at program control,
//      fifo threshold and burst mode set to 8 bytes.
//   3. Set the remote byte count registers to 0.
//   4. Rx operation: buffer packets to memory, check physical
//      address on incoming, do not accept multicast, do not
//      accept broadcast, runt packets rejected, error packets dumped.
//   5. Put transmitter in loopback mode.
//   6. Initialize boundary register (boundary ring read pointer).
//   7. Set page start and page stop addresses for receive ring.
//      These values *MUST* match the page start and page stop
//      values given to the gate array.
//   8. Clear the interrupt status register.
//   9. Enable all interrupts.
//  10. Select register page 1 whilst keeping the NIC stopped.
//  11. Load up the ethernet station address.
//  12. Load up the multicast address.
//  13. Initialize the current page register (buffer ring write pointer).
//  14. Go back to register page 0 whilst still keeping the NIC stopped.
//  15. Activate NIC, but keep DMA aborted.
//  16. Tx config: append crc to transmission, normal operation,
//      normal algorithm for backoff logic.
//
// Key to line comments: (n.) refers to procedure line number, and
// (rpx) refers to current register page x when it is set or changed.
//
//------------------------------------------------------------------------
void
PCi586EthernetDriver::restartNIC() {

  Debug << this << " restartNIC()\n" << eor;
  
  Out( GA_CTRL, GA_CTRL_xsel );                    // window NIC into IO space
  Out( NIC_CR, NIC_CR_rd2 | NIC_CR_stp );               // (1.) (rp0)
  Out( NIC_DCR, NIC_DCR_ft1 | NIC_DCR_bms );            // (2.)
  Out( NIC_RBCR0, 0 );                                  // (3.)
  Out( NIC_RBCR1, 0 );                                  // (3.)
  Out( NIC_RCR, NIC_RCR_operation );                    // (4.)
  Out( NIC_TCR, NIC_TCR_lb0 );                          // (5.)
  Out( NIC_BNDY, NIC_BNDY_init );                       // (6.)
  Out( NIC_PSTART, NIC_PSTART_pgstrt );                 // (7.)
  Out( NIC_PSTOP, NIC_PSTOP_pgstop );                   // (7.)
  Out( NIC_ISR, NIC_ISR_resetISR );                     // (8.)
  Out( NIC_IMR, NIC_IMR_all );                          // (9.)
  Out( NIC_CR, NIC_CR_ps0 | NIC_CR_rd2 | NIC_CR_stp );  // (10.) (rp1)
  Out( NIC_PAR5, etherAddress[5] );                     // (11.)
  Out( NIC_PAR4, etherAddress[4] );                     // (11.)
  Out( NIC_PAR3, etherAddress[3] );                     // (11.)
  Out( NIC_PAR2, etherAddress[2] );                     // (11.)
  Out( NIC_PAR1, etherAddress[1] );                     // (11.)
  Out( NIC_PAR0, etherAddress[0] );                     // (11.)
  Out( NIC_MAR7, multicastAddress[7] );                 // (12.)
  Out( NIC_MAR6, multicastAddress[6] );                 // (12.)
  Out( NIC_MAR5, multicastAddress[5] );                 // (12.)
  Out( NIC_MAR4, multicastAddress[4] );                 // (12.)
  Out( NIC_MAR3, multicastAddress[3] );                 // (12.)
  Out( NIC_MAR2, multicastAddress[2] );                 // (12.)
  Out( NIC_MAR1, multicastAddress[1] );                 // (12.)
  Out( NIC_MAR0, multicastAddress[0] );                 // (12.)
  Out( NIC_CURR, NIC_CURR_init );                       // (13.)
  Out( NIC_CR, NIC_CR_rd2 | NIC_CR_stp );               // (14.) (rp0)
  Out( NIC_CR, NIC_CR_rd2 | NIC_CR_sta );               // (15.)
  Out( NIC_TCR, NIC_TCR_operation );                    // (16.) 

  nextPacket = NIC_CURR_init;         // rd and wr ptrs equal at start
}

//---PCi586EthernetDriver::oldGetEthernetAddress()-----------------------
//
//  Returns the card's ethernet station address in the buffer indicated
//  by the calling routine.  The buffer had best be at least
//  ethernetAddressSize bytes long.  The MSB of the address will
//  be in the ethernetAddressSize-1 byte of the buffer and the LSB
//  will be in the first byte of the buffer.
//------------------------------------------------------------------------
void
PCi586EthernetDriver::oldGetEthernetAddress( unsigned char * buffer )
{

  Debug << this << " oldGetEthernetAddress()\n" << eor;

  int index = 0;
  while( index < ethernetAddressSize )
    *buffer++ = etherAddress[index++];
}

//---PCi586EthernetDriver::ethernetInterruptHandler()--------------------
//
//  Interrupt dispatch code.  Interrupt sources:
//    1. One of the NIC network statistic counters overflowed - advisory.
//    2. A received packet was overwritten (i.e., the packets are coming
//       in faster than the code is taking them out).  The recommended
//       action is to stop and restart the NIC.
//    3. Received packet with errors.  The packet reception is aborted
//       by the NIC.
//    4. Received packet with no errors. The packet reception procedure
//       is invoked.
//    5. Packet was transmitted with errors.  Cause of error is output
//       to the diagnostic console and an assertion halts the machine.
//    6. Packet was transmitted with no errors.
//------------------------------------------------------------------------
void
PCi586EthernetDriver::ethernetInterruptHandler(unsigned char 
                                                  interruptStatusRegister)
{
  Debug << this << " ethernetInterruptHandler()\n" << eor;

  // Check if NIC is windowed into IO space and that the NIC
  // page 0 registers are selected.
  Assert( (In(GA_CTRL) & (GA_CTRL_ealo | GA_CTRL_eahi)) == 0 );
  Assert( (In(NIC_CR) & (NIC_CR_ps0 | NIC_CR_ps1)) == 0 );

  if( interruptStatusRegister & NIC_ISR_cnt )
    Debug << this << " ethernetInterruptHandler(): counter overflow\n" << eor;

  if( interruptStatusRegister & NIC_ISR_ovw ) {
    Debug << this << " ethernetInterruptHandler(): overwrite warning\n" << eor;
    stopNIC();
    restartNIC();    
  }

  unsigned char receiveStatus = In( NIC_RSR );

  // -- receive packet aborted! --
  // causes:
  //   CRC error, frame alignment error, FIFO overrun, missed packet

  if( interruptStatusRegister & NIC_ISR_rxe ) {
    Debug << this <<
	" ethernetInterruptHandler(): aborted packet reception\n" << eor;
    if( (receiveStatus & NIC_RSR_crc) == NIC_RSR_crc )
      Debug << "  Receiver error: crc incorrect\n" << eor;
    if( (receiveStatus & NIC_RSR_fae) == NIC_RSR_fae )
      Debug << "  Receiver error: frame alignment error\n" << eor;
    if( (receiveStatus & NIC_RSR_fo) == NIC_RSR_fo )
      Debug << "  Receiver error: FIFO overrun\n" << eor;
    if( (receiveStatus & NIC_RSR_mpa) == NIC_RSR_mpa )
      Debug << "  Receiver error: missed packet\n" << eor;
  }

  // no receive errors

  if( interruptStatusRegister & NIC_ISR_prx ) {
    Assert( (receiveStatus & NIC_RSR_prx) == NIC_RSR_prx );
    Debug << this << " ethernetInterruptHandler: packet received\n" << eor;
    //receivePacket(0,0);
    processRecv();
  }

  unsigned char transmitStatus = In( NIC_TSR );

  // --- Transmission Aborted! ---
  // Transmit errors that can get us here are
  // excessive collisions or fifo underrun.

  int txeStatus = 0;
  if( interruptStatusRegister & NIC_ISR_txe ) {
    Debug << this << " ethernetInterruptHandler: aborted transmission\n" <<eor;
    if( (transmitStatus & NIC_TSR_abt) == NIC_TSR_abt )
      Debug << "  Transmit error: excessive collisions (>16)\n" << eor;
    if( (transmitStatus & NIC_TSR_fu) == NIC_TSR_fu )
      Debug << "  Transmit error: NIC FIFO underrun\n" << eor;
    ethernetTxBufferMutex->V();
    txeStatus = 1;
  }
  Assert( txeStatus == 0 );

  // Transmission was not aborted, but check for advisories.

  if( interruptStatusRegister & NIC_ISR_ptx ) {
    Debug << this << " ethernetInterruptHandler: packet transmission\n" << eor;
    if( (transmitStatus & NIC_TSR_ptx) == NIC_TSR_ptx )
      Debug << " Transmit status: ok\n" << eor;
    if( (transmitStatus & NIC_TSR_col) == NIC_TSR_col )
      Debug << " Transmit advisory: transmission collided " << In( NIC_NCR )
            << " times\n" << eor;
    if( (transmitStatus & NIC_TSR_crs) == NIC_TSR_crs )
      Debug << " Transmit advisory: carrier sense lost\n" << eor;
    if( (transmitStatus & NIC_TSR_cdh) == NIC_TSR_cdh )
      Debug << " Transmit advisory: failure of CD heartbeat\n" << eor;
    if( (transmitStatus & NIC_TSR_owc) == NIC_TSR_owc )
      Debug << " Transmit advisory: out of window collision\n" << eor;

    // @@@ comment added by Stan... V the mutex semaphore which locks
    // the transmit buffer.
    ethernetTxBufferMutex->V();
  }

}

//---PCi586EthernetDriver::processRecv()-----------------------
//
//  Procedure:
//    1. Get a pointer to an ethernet packet from the EthernetAdministrator.
//    2. Get the page starting address of the newly received packet.
//    3. If test is defined, output received packet to the diagnostic
//       console.
//    4. Check the status byte of the received packet.
//    5. Get the next packet pointer, and the received data byte count
//       from the 4 byte packet header.
//    6. Get the destination address.
//    7. Get the source address.
//    8. Get the packet length.
//    9. Get the packet data.
//   10. Calculate and set the new boundary pointer and next packet pointer
//------------------------------------------------------------------------
void
PCi586EthernetDriver::processRecv()	// ignore args
{
  Boolean morePackets = TRUE;

  while( morePackets ) {

    Debug << this << " receivePacket()\n" << eor;


    // Check if NIC is windowed into IO space and that the NIC
    // page 0 registers are selected.
    Assert( (In(GA_CTRL) & (GA_CTRL_ealo | GA_CTRL_eahi)) == 0 );
    Assert( (In(NIC_CR) & (NIC_CR_ps0 | NIC_CR_ps1)) == 0 );

    // Get the 256 byte page offset for the current receive buffer
    // from the rd ptr nextPacket and add it to the base address of the card.

    // @@@ Stan's comment:  the memory in this area of Choices is mapped
    // 1-1 from physical to virtual, so we *can* in fact use the VM
    // address without resorting to a PhysCopyBlock.
    unsigned short pageBoundary = ((unsigned short) nextPacket & 0x1f) << 8; 
    unsigned char * rxPacketData = 
                              (unsigned char *)(memoryBase + pageBoundary);
#ifdef TEST1
    Debug << "receivePacket(): received packet stored at "
          << asAddress(memoryBase) << " + " << asHex(pageBoundary) 
          << "\n" << eor;

    Debug << "receivePacket(): raw dump of first 8 bytes of packet\n" << eor;
    unsigned char * dumpData = rxPacketData;
    Debug << asAddress(dumpData) << ": " << asHex(*dumpData) << eor;
    dumpData++;
    Debug << "  " << asAddress(dumpData) << ": " << asHex(*dumpData) <<eor;
    dumpData++;
    Debug << "  " << asAddress(dumpData) << ": " << asHex(*dumpData) <<eor;
    dumpData++;
    Debug << "  " << asAddress(dumpData) << ": " << asHex(*dumpData) 
          << "\n" << eor;
    dumpData++;
    Debug << asAddress(dumpData) << ": " << asHex(*dumpData) << eor;
    dumpData++;
    Debug << "  " << asAddress(dumpData) << ": " << asHex(*dumpData) << eor;
    dumpData++;
    Debug << "  " << asAddress(dumpData) << ": " << asHex(*dumpData) << eor;
    dumpData++;
    Debug << "  " << asAddress(dumpData) << ": " << asHex(*dumpData) 
          << "\n" << eor;
#endif /* TEST1 */

#ifdef TEST2
    unsigned char * dumpData = rxPacketData + 2;
    unsigned short *etherType =
		(unsigned short *) (rxPacketData + 4 + 2 *ethernetAddressSize);
    unsigned short *sequenceNumber =
	(unsigned short *) (rxPacketData + 6 + 2 * ethernetAddressSize);
#ifdef TEST2A
    Debug << "receivePacket(): TEST 2: raw dump of first 20 bytes...\n" << eor;
    Debug << "packet:  " <<
	asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	asHex( *dumpData++ ) << " " << asHex( *dumpData++ ) << " " <<
	"\n" << eor;
#endif /* TEST2A */
    Debug << "**** Sequence number is " << asHex(*sequenceNumber) <<
		", ethertype is " << asHex(*etherType) << "\n\n" << eor;
#endif  /* TEST2 */

    // Check the first byte of the received packet for no errors and
    // for the address match type.
    unsigned char receiveStatus = *rxPacketData;
    Assert( (receiveStatus & NIC_RSR_prx) == NIC_RSR_prx );
    if( (receiveStatus & NIC_RSR_phy) == 0 )
      Debug << " receivePacket(): physical address matched\n" << eor;
    else
      Debug << " receivePacket(): multicast/broadcast address matched\n" << eor;
    
    // Get the pointer to the next received packet (if any) and this
    // packet's byte count.
    unsigned char headerNextPacket = *(rxPacketData + 1);
    unsigned short packetLength = 
		    (*(rxPacketData + 3) << 8) + *(rxPacketData + 2) - 4;
    rxPacketData += 4;
	//  The -4 and +4 above are to skip over the status information
	//  that the interface card puts at the front of the packet. //Amit

    Debug << "receivePacket(): received " << packetLength
          << " bytes in packet\n" << eor;


	//  Now that we know the frame has been received OK, we need to find
	//  out its ethertype and then request a buffer from the administrator.
    //EthernetPacketType ethertype;
    //ethertype = getPacketType( rxPacketData );			// Amit

    NetworkBuffer *receiveBuffer;
    receiveBuffer = _administrator->getBuffer();
	//  NOTE: we assume this is an PCi586 with a WD3C505 card, which
	//  uses only *one* size of buffer.  For use with cards which use
	//  more buffer sizes (eg the EMC on the MultiMax), this code
	//  must be rewritten to select a free list based on the size of
	//  the received packet.				// Amit

    receiveBuffer->setLength(packetLength);			// Amit

    // Get the destination and source addresses from the packet, then
    // remove the packet data.
    //for( int index = 0; index < ethernetAddressSize; index++ )
    //  receivedPacket->destinationAddress[index] = *rxPacketData++;
    //for( index = 0; index < ethernetAddressSize; index++ )
    //  receivedPacket->sourceAddress[index] = *rxPacketData++;

    //for ( int index = 0 ; index < 2 * ethernetAddressSize ; index++ ) {
    //	*destptr++ = *rxPacketData++;
    //}
    //Assert( packetLength < ( maxEthernetDataSize + 2 ) );
    //packetLength -= ( 2 * ethernetAddressSize );
    //for ( index = 0 ; index < packetLength ; index++ ) {
    //	*destptr++ = *rxPacketData++;
    //}
    receiveBuffer->copyFrom(rxPacketData, packetLength);




#ifdef TEST
    // @@@ added by Stan, to fix the allocator running out of memory
    // when we're testing.
    // delete receivedPacket;
    receiveBuffer->release();
#else

    // Now, pass the packet up to the EthernetAdministrator ...  -- Amit
    //_administrator->receive( receiveBuffer );
    NetworkDriver::receivePacket( receiveBuffer );

#endif


    // Get the current page pointer (buffer ring write pointer) to
    // see if there are any remaining packets to read.
    Assert( (In(GA_CTRL) & (GA_CTRL_ealo | GA_CTRL_eahi)) == 0 );
    unsigned char restoreNIC_CR = In( NIC_CR );
    Out( NIC_CR, ( restoreNIC_CR | NIC_CR_ps0 ) );  // select NIC pg 1 regs
    unsigned char currentPageRegister = In( NIC_CURR );
    Out( NIC_CR, restoreNIC_CR );

    // Check to see if any new packets have arrived whilst we've been
    // receiving this packet.
    if( headerNextPacket == currentPageRegister )
      morePackets = FALSE;

    nextPacket = headerNextPacket;   // update read pointer

    // This is sick, but the NIC requires its BNDY register to be 
    // updated (and wrapped around, if needed) manually.
    unsigned char newBoundary = headerNextPacket - 1;
    if( newBoundary < NIC_PSTART_pgstrt )
      newBoundary = NIC_PSTOP_pgstop - 1;
    Out( NIC_BNDY, newBoundary );

#ifndef TEST2					/* // Amit */
    Debug << this << " new NIC_BNDY is " << asHex(newBoundary) 
                  << " and CURR is " << asHex(currentPageRegister)
                  << "\n" << eor;
#endif /* TEST2 */
  }

}

//---PCi586EthernetDriver::transmitPacket()----------------------
//
//  Transmit sequence:
//    1. Create and load up transmit packet into buffer (shared memory).
//       Buffer must be padded to a minimum of 46 data bytes.
//    2. Set the Transmit Configuration Register to normal operation (NIC).
//    3. Set the Transmit Page Start Register to beginning of transmit
//       buffer (NIC).
//    4. Load the Transmit Byte Count Registers (NIC).
//    5. Set the Interrupt Mask Register (NIC).
//    6. Set Command Register to start controller (NIC).
//    7. Set the Command Register to initiate packet transmission (NIC).
//    8. Poll Interrupt Status Register to await tranmit completion (NIC).
//
//  Input parameter note: it is the responsibility of the calling routine
//  to make txPacket->destinationAddress[], txPacket->packetDataLength, and 
//  txPacket->packetData[] valid.  transmitPacket() makes
//  txPacket->sourceAddress[] valid while the DP8390 NIC makes the CRC
//  correct.
//------------------------------------------------------------------------
int
PCi586EthernetDriver::transmitPacket(NetworkBuffer * transmitBuffer)
{
  unsigned short txpacketDataLength;   // number of data bytes to transmit

  Debug << this << " transmitPacket()\n" << eor;

  // @@@ comment added by Stan... this semaphore is not V'd till an interrupt
  // goes off saying that the buffer has been freed.
  // This occurs in ethernetInterruptHandler().
  ethernetTxBufferMutex->P();

  //getEthernetAddress( &(txPacket->sourceAddress[0]) );

  // determine how many data bytes to transmit
  //if( txPacket->packetDataLength < minEthernetDataSize )
  //  txpacketDataLength = minEthernetDataSize;
  //else
  //  txpacketDataLength = txPacket->packetDataLength;
  txpacketDataLength = ( transmitBuffer->length() < minEthernetDataSize )
		? minEthernetDataSize : transmitBuffer->length() ;  // -- Amit

  //Assert( txPacket->packetDataLength < maxEthernetDataSize );
  Assert(transmitBuffer->length() <
	 (maxEthernetDataSize + 2 * ethernetAddressSize));

  //Debug << this << " Transmitting " << txpacketDataLength << " bytes to "
  //      << txPacket->destinationAddress[5] << ":" 
  //      << txPacket->destinationAddress[4] << ":" 
  //      << txPacket->destinationAddress[3] << ":" 
  //      << txPacket->destinationAddress[2] << ":" 
  //      << txPacket->destinationAddress[1] << ":" 
  //      << txPacket->destinationAddress[0] << "\n" << eor;
  Debug << this << " Transmitting " << transmitBuffer->length() << " bytes to "
	<< asHex(transmitBuffer->packet()[0]) << ":"
	<< asHex(transmitBuffer->packet()[1]) << ":"	// -- Amit
	<< asHex(transmitBuffer->packet()[2]) << ":"
	<< asHex(transmitBuffer->packet()[3]) << ":"
	<< asHex(transmitBuffer->packet()[4]) << ":"
	<< asHex(transmitBuffer->packet()[5]) << "\n" << eor;

  // Put packet together in transmit buffer (in shared memory).
  // Each multibyte component goes into the transmit buffer least
  // significant byte first.
  unsigned char *txBuffer = (unsigned char *) memoryBase;
  //for( int index = 0; index < ethernetAddressSize; index++ )
  //  *txBuffer++ = txPacket->destinationAddress[index];
  //for( index = 0; index < ethernetAddressSize; index++ )
  //  *txBuffer++ = txPacket->sourceAddress[index];
  // *txBuffer++ = (unsigned char) (txPacket->packetDataLength & 0xff);
  // *txBuffer++ = (unsigned char) (txPacket->packetDataLength >> 8);
  //for( index = 0; index < txpacketDataLength; index++ )
  //  *txBuffer++ = txPacket->packetData[index];

  //	Dave was using the fields of the EthernetPacket.  Since I don't
  //	use the EthernetPacket (I assume the data is already laid out
  //	in standard Ethernet packet format, be it old Ethernet or 802.3),
  //	I have to copy the entire buffer into the shared memory and then
  //	overlay the source Ethernet address on top before telling the
  //	NIC to transmit ...				-- Amit
  unsigned char *sourcePacket = transmitBuffer->packet();
  for ( int index = 0 ; index < transmitBuffer->length() ; index++ ) {
    *txBuffer++ = *sourcePacket++;
  }
  //	Now, stuff the Ethernet address into the packet ...	-- Amit
  oldGetEthernetAddress( (unsigned char *) memoryBase + ethernetAddressSize );

  // Fire up NIC the transmitter.  The NIC had best be windowed
  // into IO space, NIC page 0 registers better be selected,
  // and the NIC already started.
  // Then load the page address of the start of the transmit 
  // buffer, the packet length, and turn that puppy loose.
  // The transmit completion interrupt will take care of releasing
  // the transmit buffer semaphore.
  Assert( (In(GA_CTRL) & (GA_CTRL_ealo | GA_CTRL_eahi)) == 0 );
  Assert( (In(NIC_CR) & (NIC_CR_ps0 | NIC_CR_ps1)) == 0 );
  Assert( (In(NIC_CR) & NIC_CR_sta) == NIC_CR_sta );
  Out( NIC_TCR, NIC_TCR_operation );
  Out( NIC_TPSR, NIC_TPSR_start );
  Out( NIC_TBCR1, (unsigned char) (txpacketDataLength >> 8) );
  Out( NIC_TBCR0, (unsigned char) (txpacketDataLength & 0xff) );
  Out( NIC_CR, (NIC_CR_rd2 | NIC_CR_txp | NIC_CR_sta) );

  return 0;

}

//---PCi586EthernetDriver::getNewEthernetPacketPtr()---------------------
EthernetPacket *
PCi586EthernetDriver::getNewEthernetPacketPtr()
{
  return( new EthernetPacket );
}



//---Support for ProxiableObject------------------------------------------
//------------------------------------------------------------------------

#include "ClassInstaller.h"
DEFINECLASS("PCi586EthernetDriver",PCi586EthernetDriver, EthernetDriver,0,0);

//------------------------------------------------------------------------


