/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * PCi586i586CPU.cc: PCi586 i586 Processor
 */

#include "Assert.h"
#include "OutputStream.h"
#include "PCi586i586CPU.h"
#include "Vectors.h"
#include "Util.h"
#include "PCi586InterruptController.h"
#include "i586IntrExcp.h"
#include "TimerManager.h"
#include "TIMESLICE_TIMER.h"

PCi586i586CPU::PCi586i586CPU( unsigned long identifier,
	ProcessContainer * initialIdleContainer )
        : i586CPU( identifier, initialIdleContainer )
{
        Debug.printf( "PCi586i586CPU::ctor()\n" );

        _timesliceTimer = new TIMESLICETIMER;
	_timesliceTimer->setName("#PCi586i586CPUTimesliceTimer");
}

PCi586i586CPU::~PCi586i586CPU()
{
	Assert( NOTREACHED );
}

void
PCi586i586CPU::chipInitialize()
{
	Debug << "PCi586i586CPU::chipInitialize()\n" << eor;

	extern int Com0Busy;

	/* 
	 * Program the programmable interrupt controllers.
	 */
	Out(PIC_CTRL_REG, 0x11);  /* edge-triggered, cascaded, need ICW4 */
	Out(PIC_MASK_REG, HARDWARE_Vectors);    /* start of int vectors */
	Out(PIC_MASK_REG, PIC_PIC2_MASK);	/* identify slave controller */
	Out(PIC_MASK_REG, 1);			/* mcs 86 mode */

	/* disable interrupts from interrupt controller until ready for them */
	if (Com0Busy)
		/* leave Com 0 enabled for cdb */
		Out(PIC_MASK_REG, 0xff & ~PIC_COM0_MASK & ~PIC_PIC2_MASK);
	else
		/* leave Com 1 enabled instead */
		Out(PIC_MASK_REG, 0xff & ~PIC_COM1_MASK & ~PIC_PIC2_MASK);
	
	/* switch to read irr mode; this allows reading pending interrutps */
	Out(PIC_CTRL_REG, 0xa);		
	Out(PIC2_CTRL_REG, 0x11);
	Out(PIC2_MASK_REG, HARDWARE_Vectors + 8);
	Out(PIC2_MASK_REG, 2);			/* id for this controller */
	Out(PIC2_MASK_REG, 1);
	Out(PIC2_MASK_REG, 0xff);		/* start out disabling all */
	Out(PIC2_CTRL_REG, 0xa);

	/*
	 * Do the superclass stuff last.
	 */
	i586CPU::chipInitialize();

	/*
	 * Once all the interrupt handlers are setup and the IDT is valid, set IDTSetup flag to true
	 */
	i586CPU::IDTSetup = true;

	enableInterrupts();
}

TimerManagerInterrupt TimerManagerInterruptInterrupt;

void
PCi586i586CPU::installInterrupts()
{
	extern int Com0Busy;

	Debug << "PCi586i586CPU::installInterrupts()\n" << eor;
	Assert( this != 0 );

	/*
	 * First do the machine-dependent interrupt handlers
	 */
	setIdtGate( 0, InterruptGate, SYSTEM_CS, &Interrupt0, LvlThree);
	setIdtGate( 1, InterruptGate, SYSTEM_CS, &Interrupt1, LvlThree);
	setIdtGate( 2, InterruptGate, SYSTEM_CS, &Interrupt2, LvlThree);
	setIdtGate( 3, TrapGate, SYSTEM_CS, &Interrupt3, LvlThree);
	setIdtGate( 4, InterruptGate, SYSTEM_CS, &Interrupt4, LvlThree);
	setIdtGate( 5, InterruptGate, SYSTEM_CS, &Interrupt5, LvlThree);
	setIdtGate( 6, InterruptGate, SYSTEM_CS, &Interrupt6, LvlThree);
	setIdtGate( 7, InterruptGate, SYSTEM_CS, &Interrupt7, LvlThree);
	setIdtGate( 8, InterruptGate, SYSTEM_CS, &Interrupt8, LvlThree);
	setIdtGate( 9, InterruptGate, SYSTEM_CS, &Interrupt9, LvlThree);
	setIdtGate( 10, InterruptGate, SYSTEM_CS, &Interrupt10, LvlThree);
	setIdtGate( 11, InterruptGate, SYSTEM_CS, &Interrupt11, LvlThree);
	setIdtGate( 12, InterruptGate, SYSTEM_CS, &Interrupt12, LvlThree);
	setIdtGate( 13, InterruptGate, SYSTEM_CS, &Interrupt13, LvlThree);
	setIdtGate( 14, InterruptGate, SYSTEM_CS, &Interrupt14, LvlThree);
	setIdtGate( 15, InterruptGate, SYSTEM_CS, &Interrupt15, LvlThree);
	setIdtGate( 16, InterruptGate, SYSTEM_CS, &Interrupt16, LvlThree);
	setIdtGate( 17, InterruptGate, SYSTEM_CS, &Interrupt17, LvlThree);
	setIdtGate( 18, InterruptGate, SYSTEM_CS, &Interrupt18, LvlThree);
	setIdtGate( 19, InterruptGate, SYSTEM_CS, &Interrupt19, LvlThree);
	setIdtGate( 20, InterruptGate, SYSTEM_CS, &Interrupt20, LvlThree);
	setIdtGate( 21, InterruptGate, SYSTEM_CS, &Interrupt21, LvlThree);
	setIdtGate( 22, InterruptGate, SYSTEM_CS, &Interrupt22, LvlThree);
	setIdtGate( 23, InterruptGate, SYSTEM_CS, &Interrupt23, LvlThree);
	/* timer */
	setIdtGate( 24, InterruptGate, SYSTEM_CS, &Interrupt24, LvlThree);
	/* keyboard */
	setIdtGate( 25, InterruptGate, SYSTEM_CS, &Interrupt25, LvlThree);
	setIdtGate( 26, InterruptGate, SYSTEM_CS, &Interrupt26, LvlThree);


	setIdtGate( 27, InterruptGate, SYSTEM_CS, &Interrupt27, LvlThree);
	setIdtGate( 28, InterruptGate, SYSTEM_CS, &Interrupt28, LvlThree);

#ifdef STUPID
	if (Com0Busy)
		/* com port 1 */
		setIdtGate( 27, InterruptGate, SYSTEM_CS, &Interrupt27, LvlThree);
	else
		/* com port 0 */
		setIdtGate( 28, InterruptGate, SYSTEM_CS, &Interrupt28, LvlThree);
#endif

	setIdtGate( 29, InterruptGate, SYSTEM_CS, &Interrupt29, LvlThree);
	setIdtGate( 30, InterruptGate, SYSTEM_CS, &Interrupt30, LvlThree);
	setIdtGate( 31, InterruptGate, SYSTEM_CS, &Interrupt31, LvlThree);

	setIdtGate( 32, InterruptGate, SYSTEM_CS, &Interrupt32, LvlThree);
	setIdtGate( 33, InterruptGate, SYSTEM_CS, &Interrupt33, LvlThree);
	setIdtGate( 34, InterruptGate, SYSTEM_CS, &Interrupt34, LvlThree);
	setIdtGate( 35, InterruptGate, SYSTEM_CS, &Interrupt35, LvlThree);
	setIdtGate( 36, InterruptGate, SYSTEM_CS, &Interrupt36, LvlThree);
	setIdtGate( 37, InterruptGate, SYSTEM_CS, &Interrupt37, LvlThree);
	setIdtGate( 38, InterruptGate, SYSTEM_CS, &Interrupt38, LvlThree);
	setIdtGate( 39, InterruptGate, SYSTEM_CS, &Interrupt39, LvlThree);
	/*
	SVC trap
	setIdtGate( 40, InterruptGate, SYSTEM_CS, &Interrupt40, LvlThree);
	*/
	setIdtGate( 41, InterruptGate, SYSTEM_CS, &Interrupt41, LvlThree);
	setIdtGate( 42, InterruptGate, SYSTEM_CS, &Interrupt42, LvlThree);
	setIdtGate( 43, InterruptGate, SYSTEM_CS, &Interrupt43, LvlThree);
	setIdtGate( 44, InterruptGate, SYSTEM_CS, &Interrupt44, LvlThree);
	setIdtGate( 45, InterruptGate, SYSTEM_CS, &Interrupt45, LvlThree);
	setIdtGate( 46, InterruptGate, SYSTEM_CS, &Interrupt46, LvlThree);
	setIdtGate( 47, InterruptGate, SYSTEM_CS, &Interrupt47, LvlThree);
	setIdtGate( 48, InterruptGate, SYSTEM_CS, &Interrupt48, LvlThree);
	setIdtGate( 49, InterruptGate, SYSTEM_CS, &Interrupt49, LvlThree);
	setIdtGate( 50, InterruptGate, SYSTEM_CS, &Interrupt50, LvlThree);
	setIdtGate( 51, InterruptGate, SYSTEM_CS, &Interrupt51, LvlThree);
	setIdtGate( 52, InterruptGate, SYSTEM_CS, &Interrupt52, LvlThree);
	setIdtGate( 53, InterruptGate, SYSTEM_CS, &Interrupt53, LvlThree);
	setIdtGate( 54, InterruptGate, SYSTEM_CS, &Interrupt54, LvlThree);
	setIdtGate( 55, InterruptGate, SYSTEM_CS, &Interrupt55, LvlThree);
	setIdtGate( 56, InterruptGate, SYSTEM_CS, &Interrupt56, LvlThree);
	setIdtGate( 57, InterruptGate, SYSTEM_CS, &Interrupt57, LvlThree);
	setIdtGate( 58, InterruptGate, SYSTEM_CS, &Interrupt58, LvlThree);
	setIdtGate( 59, InterruptGate, SYSTEM_CS, &Interrupt59, LvlThree);
	setIdtGate( 50, InterruptGate, SYSTEM_CS, &Interrupt60, LvlThree);
	setIdtGate( 61, InterruptGate, SYSTEM_CS, &Interrupt61, LvlThree);
	setIdtGate( 62, InterruptGate, SYSTEM_CS, &Interrupt62, LvlThree);
	setIdtGate( 63, InterruptGate, SYSTEM_CS, &Interrupt63, LvlThree);
	setIdtGate( 64, InterruptGate, SYSTEM_CS, &Interrupt64, LvlThree);
	setIdtGate( 65, InterruptGate, SYSTEM_CS, &Interrupt65, LvlThree);
	setIdtGate( 66, InterruptGate, SYSTEM_CS, &Interrupt66, LvlThree);
	setIdtGate( 67, InterruptGate, SYSTEM_CS, &Interrupt67, LvlThree);
	setIdtGate( 68, InterruptGate, SYSTEM_CS, &Interrupt68, LvlThree);
	setIdtGate( 69, InterruptGate, SYSTEM_CS, &Interrupt69, LvlThree);
	setIdtGate( 70, InterruptGate, SYSTEM_CS, &Interrupt70, LvlThree);
	setIdtGate( 71, InterruptGate, SYSTEM_CS, &Interrupt71, LvlThree);
	setIdtGate( 72, InterruptGate, SYSTEM_CS, &Interrupt72, LvlThree);
	setIdtGate( 73, InterruptGate, SYSTEM_CS, &Interrupt73, LvlThree);
	setIdtGate( 74, InterruptGate, SYSTEM_CS, &Interrupt74, LvlThree);
	setIdtGate( 75, InterruptGate, SYSTEM_CS, &Interrupt75, LvlThree);
	setIdtGate( 76, InterruptGate, SYSTEM_CS, &Interrupt76, LvlThree);
	setIdtGate( 77, InterruptGate, SYSTEM_CS, &Interrupt77, LvlThree);
	setIdtGate( 78, InterruptGate, SYSTEM_CS, &Interrupt78, LvlThree);
	setIdtGate( 79, InterruptGate, SYSTEM_CS, &Interrupt79, LvlThree);
	setIdtGate( 80, InterruptGate, SYSTEM_CS, &Interrupt80, LvlThree);
	setIdtGate( 81, InterruptGate, SYSTEM_CS, &Interrupt81, LvlThree);
	setIdtGate( 82, InterruptGate, SYSTEM_CS, &Interrupt82, LvlThree);
	setIdtGate( 83, InterruptGate, SYSTEM_CS, &Interrupt83, LvlThree);
	setIdtGate( 84, InterruptGate, SYSTEM_CS, &Interrupt84, LvlThree);
	setIdtGate( 85, InterruptGate, SYSTEM_CS, &Interrupt85, LvlThree);
	setIdtGate( 86, InterruptGate, SYSTEM_CS, &Interrupt86, LvlThree);
	setIdtGate( 87, InterruptGate, SYSTEM_CS, &Interrupt87, LvlThree);
	setIdtGate( 88, InterruptGate, SYSTEM_CS, &Interrupt88, LvlThree);
	setIdtGate( 89, InterruptGate, SYSTEM_CS, &Interrupt89, LvlThree);
	setIdtGate( 90, InterruptGate, SYSTEM_CS, &Interrupt90, LvlThree);
	setIdtGate( 91, InterruptGate, SYSTEM_CS, &Interrupt91, LvlThree);
	setIdtGate( 92, InterruptGate, SYSTEM_CS, &Interrupt92, LvlThree);
	setIdtGate( 93, InterruptGate, SYSTEM_CS, &Interrupt93, LvlThree);
	setIdtGate( 94, InterruptGate, SYSTEM_CS, &Interrupt94, LvlThree);
	setIdtGate( 95, InterruptGate, SYSTEM_CS, &Interrupt95, LvlThree);
	setIdtGate( 96, InterruptGate, SYSTEM_CS, &Interrupt96, LvlThree);
	setIdtGate( 97, InterruptGate, SYSTEM_CS, &Interrupt97, LvlThree);
	setIdtGate( 98, InterruptGate, SYSTEM_CS, &Interrupt98, LvlThree);
	setIdtGate( 99, InterruptGate, SYSTEM_CS, &Interrupt99, LvlThree);
	setIdtGate( 100, InterruptGate, SYSTEM_CS, &Interrupt100, LvlThree);
	setIdtGate( 101, InterruptGate, SYSTEM_CS, &Interrupt101, LvlThree);
	setIdtGate( 102, InterruptGate, SYSTEM_CS, &Interrupt102, LvlThree);
	setIdtGate( 103, InterruptGate, SYSTEM_CS, &Interrupt103, LvlThree);
	setIdtGate( 104, InterruptGate, SYSTEM_CS, &Interrupt104, LvlThree);
	setIdtGate( 105, InterruptGate, SYSTEM_CS, &Interrupt105, LvlThree);
	setIdtGate( 106, InterruptGate, SYSTEM_CS, &Interrupt106, LvlThree);
	setIdtGate( 107, InterruptGate, SYSTEM_CS, &Interrupt107, LvlThree);
	setIdtGate( 108, InterruptGate, SYSTEM_CS, &Interrupt108, LvlThree);
	setIdtGate( 109, InterruptGate, SYSTEM_CS, &Interrupt109, LvlThree);
	setIdtGate( 110, InterruptGate, SYSTEM_CS, &Interrupt110, LvlThree);
	setIdtGate( 111, InterruptGate, SYSTEM_CS, &Interrupt111, LvlThree);
	setIdtGate( 112, InterruptGate, SYSTEM_CS, &Interrupt112, LvlThree);
	setIdtGate( 113, InterruptGate, SYSTEM_CS, &Interrupt113, LvlThree);
	setIdtGate( 114, InterruptGate, SYSTEM_CS, &Interrupt114, LvlThree);
	setIdtGate( 115, InterruptGate, SYSTEM_CS, &Interrupt115, LvlThree);
	setIdtGate( 116, InterruptGate, SYSTEM_CS, &Interrupt116, LvlThree);
	setIdtGate( 117, InterruptGate, SYSTEM_CS, &Interrupt117, LvlThree);
	setIdtGate( 118, InterruptGate, SYSTEM_CS, &Interrupt118, LvlThree);
	setIdtGate( 119, InterruptGate, SYSTEM_CS, &Interrupt119, LvlThree);
	setIdtGate( 120, InterruptGate, SYSTEM_CS, &Interrupt120, LvlThree);
	setIdtGate( 121, InterruptGate, SYSTEM_CS, &Interrupt121, LvlThree);
	setIdtGate( 122, InterruptGate, SYSTEM_CS, &Interrupt122, LvlThree);
	setIdtGate( 123, InterruptGate, SYSTEM_CS, &Interrupt123, LvlThree);
	setIdtGate( 124, InterruptGate, SYSTEM_CS, &Interrupt124, LvlThree);
	setIdtGate( 125, InterruptGate, SYSTEM_CS, &Interrupt125, LvlThree);
	setIdtGate( 126, InterruptGate, SYSTEM_CS, &Interrupt126, LvlThree);
	setIdtGate( 127, InterruptGate, SYSTEM_CS, &Interrupt127, LvlThree);
	setIdtGate( 128, InterruptGate, SYSTEM_CS, &Interrupt128, LvlThree);
	setIdtGate( 129, InterruptGate, SYSTEM_CS, &Interrupt129, LvlThree);
	setIdtGate( 130, InterruptGate, SYSTEM_CS, &Interrupt130, LvlThree);
	setIdtGate( 131, InterruptGate, SYSTEM_CS, &Interrupt131, LvlThree);
	setIdtGate( 132, InterruptGate, SYSTEM_CS, &Interrupt132, LvlThree);
	setIdtGate( 133, InterruptGate, SYSTEM_CS, &Interrupt133, LvlThree);
	setIdtGate( 134, InterruptGate, SYSTEM_CS, &Interrupt134, LvlThree);
	setIdtGate( 135, InterruptGate, SYSTEM_CS, &Interrupt135, LvlThree);
	setIdtGate( 136, InterruptGate, SYSTEM_CS, &Interrupt136, LvlThree);
	setIdtGate( 137, InterruptGate, SYSTEM_CS, &Interrupt137, LvlThree);
	setIdtGate( 138, InterruptGate, SYSTEM_CS, &Interrupt138, LvlThree);
	setIdtGate( 139, InterruptGate, SYSTEM_CS, &Interrupt139, LvlThree);
	setIdtGate( 140, InterruptGate, SYSTEM_CS, &Interrupt140, LvlThree);
	setIdtGate( 141, InterruptGate, SYSTEM_CS, &Interrupt141, LvlThree);
	setIdtGate( 142, InterruptGate, SYSTEM_CS, &Interrupt142, LvlThree);
	setIdtGate( 143, InterruptGate, SYSTEM_CS, &Interrupt143, LvlThree);
	setIdtGate( 144, InterruptGate, SYSTEM_CS, &Interrupt144, LvlThree);
	setIdtGate( 145, InterruptGate, SYSTEM_CS, &Interrupt145, LvlThree);
	setIdtGate( 146, InterruptGate, SYSTEM_CS, &Interrupt146, LvlThree);
	setIdtGate( 147, InterruptGate, SYSTEM_CS, &Interrupt147, LvlThree);
	setIdtGate( 148, InterruptGate, SYSTEM_CS, &Interrupt148, LvlThree);
	setIdtGate( 149, InterruptGate, SYSTEM_CS, &Interrupt149, LvlThree);
	setIdtGate( 150, InterruptGate, SYSTEM_CS, &Interrupt150, LvlThree);
	setIdtGate( 151, InterruptGate, SYSTEM_CS, &Interrupt151, LvlThree);
	setIdtGate( 152, InterruptGate, SYSTEM_CS, &Interrupt152, LvlThree);
	setIdtGate( 153, InterruptGate, SYSTEM_CS, &Interrupt153, LvlThree);
	setIdtGate( 154, InterruptGate, SYSTEM_CS, &Interrupt154, LvlThree);
	setIdtGate( 155, InterruptGate, SYSTEM_CS, &Interrupt155, LvlThree);
	setIdtGate( 156, InterruptGate, SYSTEM_CS, &Interrupt156, LvlThree);
	setIdtGate( 157, InterruptGate, SYSTEM_CS, &Interrupt157, LvlThree);
	setIdtGate( 158, InterruptGate, SYSTEM_CS, &Interrupt158, LvlThree);
	setIdtGate( 159, InterruptGate, SYSTEM_CS, &Interrupt159, LvlThree);
	setIdtGate( 150, InterruptGate, SYSTEM_CS, &Interrupt160, LvlThree);
	setIdtGate( 161, InterruptGate, SYSTEM_CS, &Interrupt161, LvlThree);
	setIdtGate( 162, InterruptGate, SYSTEM_CS, &Interrupt162, LvlThree);
	setIdtGate( 163, InterruptGate, SYSTEM_CS, &Interrupt163, LvlThree);
	setIdtGate( 164, InterruptGate, SYSTEM_CS, &Interrupt164, LvlThree);
	setIdtGate( 165, InterruptGate, SYSTEM_CS, &Interrupt165, LvlThree);
	setIdtGate( 166, InterruptGate, SYSTEM_CS, &Interrupt166, LvlThree);
	setIdtGate( 167, InterruptGate, SYSTEM_CS, &Interrupt167, LvlThree);
	setIdtGate( 168, InterruptGate, SYSTEM_CS, &Interrupt168, LvlThree);
	setIdtGate( 169, InterruptGate, SYSTEM_CS, &Interrupt169, LvlThree);
	setIdtGate( 170, InterruptGate, SYSTEM_CS, &Interrupt170, LvlThree);
	setIdtGate( 171, InterruptGate, SYSTEM_CS, &Interrupt171, LvlThree);
	setIdtGate( 172, InterruptGate, SYSTEM_CS, &Interrupt172, LvlThree);
	setIdtGate( 173, InterruptGate, SYSTEM_CS, &Interrupt173, LvlThree);
	setIdtGate( 174, InterruptGate, SYSTEM_CS, &Interrupt174, LvlThree);
	setIdtGate( 175, InterruptGate, SYSTEM_CS, &Interrupt175, LvlThree);
	setIdtGate( 176, InterruptGate, SYSTEM_CS, &Interrupt176, LvlThree);
	setIdtGate( 177, InterruptGate, SYSTEM_CS, &Interrupt177, LvlThree);
	setIdtGate( 178, InterruptGate, SYSTEM_CS, &Interrupt178, LvlThree);
	setIdtGate( 179, InterruptGate, SYSTEM_CS, &Interrupt179, LvlThree);
	setIdtGate( 180, InterruptGate, SYSTEM_CS, &Interrupt180, LvlThree);
	setIdtGate( 181, InterruptGate, SYSTEM_CS, &Interrupt181, LvlThree);
	setIdtGate( 182, InterruptGate, SYSTEM_CS, &Interrupt182, LvlThree);
	setIdtGate( 183, InterruptGate, SYSTEM_CS, &Interrupt183, LvlThree);
	setIdtGate( 184, InterruptGate, SYSTEM_CS, &Interrupt184, LvlThree);
	setIdtGate( 185, InterruptGate, SYSTEM_CS, &Interrupt185, LvlThree);
	setIdtGate( 186, InterruptGate, SYSTEM_CS, &Interrupt186, LvlThree);
	setIdtGate( 187, InterruptGate, SYSTEM_CS, &Interrupt187, LvlThree);
	setIdtGate( 188, InterruptGate, SYSTEM_CS, &Interrupt188, LvlThree);
	setIdtGate( 189, InterruptGate, SYSTEM_CS, &Interrupt189, LvlThree);
	setIdtGate( 190, InterruptGate, SYSTEM_CS, &Interrupt190, LvlThree);
	setIdtGate( 191, InterruptGate, SYSTEM_CS, &Interrupt191, LvlThree);
	setIdtGate( 192, InterruptGate, SYSTEM_CS, &Interrupt192, LvlThree);
	setIdtGate( 193, InterruptGate, SYSTEM_CS, &Interrupt193, LvlThree);
	setIdtGate( 194, InterruptGate, SYSTEM_CS, &Interrupt194, LvlThree);
	setIdtGate( 195, InterruptGate, SYSTEM_CS, &Interrupt195, LvlThree);
	setIdtGate( 196, InterruptGate, SYSTEM_CS, &Interrupt196, LvlThree);
	setIdtGate( 197, InterruptGate, SYSTEM_CS, &Interrupt197, LvlThree);
	setIdtGate( 198, InterruptGate, SYSTEM_CS, &Interrupt198, LvlThree);
	setIdtGate( 199, InterruptGate, SYSTEM_CS, &Interrupt199, LvlThree);
	setIdtGate( 200, InterruptGate, SYSTEM_CS, &Interrupt200, LvlThree);
	setIdtGate( 201, InterruptGate, SYSTEM_CS, &Interrupt201, LvlThree);
	setIdtGate( 202, InterruptGate, SYSTEM_CS, &Interrupt202, LvlThree);
	setIdtGate( 203, InterruptGate, SYSTEM_CS, &Interrupt203, LvlThree);
	setIdtGate( 204, InterruptGate, SYSTEM_CS, &Interrupt204, LvlThree);
	setIdtGate( 205, InterruptGate, SYSTEM_CS, &Interrupt205, LvlThree);
	setIdtGate( 206, InterruptGate, SYSTEM_CS, &Interrupt206, LvlThree);
	setIdtGate( 207, InterruptGate, SYSTEM_CS, &Interrupt207, LvlThree);
	setIdtGate( 208, InterruptGate, SYSTEM_CS, &Interrupt208, LvlThree);
	setIdtGate( 209, InterruptGate, SYSTEM_CS, &Interrupt209, LvlThree);
	setIdtGate( 210, InterruptGate, SYSTEM_CS, &Interrupt210, LvlThree);
	setIdtGate( 211, InterruptGate, SYSTEM_CS, &Interrupt211, LvlThree);
	setIdtGate( 212, InterruptGate, SYSTEM_CS, &Interrupt212, LvlThree);
	setIdtGate( 213, InterruptGate, SYSTEM_CS, &Interrupt213, LvlThree);
	setIdtGate( 214, InterruptGate, SYSTEM_CS, &Interrupt214, LvlThree);
	setIdtGate( 215, InterruptGate, SYSTEM_CS, &Interrupt215, LvlThree);
	setIdtGate( 216, InterruptGate, SYSTEM_CS, &Interrupt216, LvlThree);
	setIdtGate( 217, InterruptGate, SYSTEM_CS, &Interrupt217, LvlThree);
	setIdtGate( 218, InterruptGate, SYSTEM_CS, &Interrupt218, LvlThree);
	setIdtGate( 219, InterruptGate, SYSTEM_CS, &Interrupt219, LvlThree);
	setIdtGate( 220, InterruptGate, SYSTEM_CS, &Interrupt220, LvlThree);
	setIdtGate( 221, InterruptGate, SYSTEM_CS, &Interrupt221, LvlThree);
	setIdtGate( 222, InterruptGate, SYSTEM_CS, &Interrupt222, LvlThree);
	setIdtGate( 223, InterruptGate, SYSTEM_CS, &Interrupt223, LvlThree);
	setIdtGate( 224, InterruptGate, SYSTEM_CS, &Interrupt224, LvlThree);
	setIdtGate( 225, InterruptGate, SYSTEM_CS, &Interrupt225, LvlThree);
	setIdtGate( 226, InterruptGate, SYSTEM_CS, &Interrupt226, LvlThree);
	setIdtGate( 227, InterruptGate, SYSTEM_CS, &Interrupt227, LvlThree);
	setIdtGate( 228, InterruptGate, SYSTEM_CS, &Interrupt228, LvlThree);
	setIdtGate( 229, InterruptGate, SYSTEM_CS, &Interrupt229, LvlThree);
	setIdtGate( 230, InterruptGate, SYSTEM_CS, &Interrupt230, LvlThree);
	setIdtGate( 231, InterruptGate, SYSTEM_CS, &Interrupt231, LvlThree);
	setIdtGate( 232, InterruptGate, SYSTEM_CS, &Interrupt232, LvlThree);
	setIdtGate( 233, InterruptGate, SYSTEM_CS, &Interrupt233, LvlThree);
	setIdtGate( 234, InterruptGate, SYSTEM_CS, &Interrupt234, LvlThree);
	setIdtGate( 235, InterruptGate, SYSTEM_CS, &Interrupt235, LvlThree);
	setIdtGate( 236, InterruptGate, SYSTEM_CS, &Interrupt236, LvlThree);
	setIdtGate( 237, InterruptGate, SYSTEM_CS, &Interrupt237, LvlThree);
	setIdtGate( 238, InterruptGate, SYSTEM_CS, &Interrupt238, LvlThree);
	setIdtGate( 239, InterruptGate, SYSTEM_CS, &Interrupt239, LvlThree);
	setIdtGate( 240, InterruptGate, SYSTEM_CS, &Interrupt240, LvlThree);
	setIdtGate( 241, InterruptGate, SYSTEM_CS, &Interrupt241, LvlThree);
	setIdtGate( 242, InterruptGate, SYSTEM_CS, &Interrupt242, LvlThree);
	setIdtGate( 243, InterruptGate, SYSTEM_CS, &Interrupt243, LvlThree);
	setIdtGate( 244, InterruptGate, SYSTEM_CS, &Interrupt244, LvlThree);
	setIdtGate( 245, InterruptGate, SYSTEM_CS, &Interrupt245, LvlThree);
	setIdtGate( 246, InterruptGate, SYSTEM_CS, &Interrupt246, LvlThree);
	setIdtGate( 247, InterruptGate, SYSTEM_CS, &Interrupt247, LvlThree);
	setIdtGate( 248, InterruptGate, SYSTEM_CS, &Interrupt248, LvlThree);
	setIdtGate( 249, InterruptGate, SYSTEM_CS, &Interrupt249, LvlThree);
	setIdtGate( 250, InterruptGate, SYSTEM_CS, &Interrupt250, LvlThree);
	setIdtGate( 251, InterruptGate, SYSTEM_CS, &Interrupt251, LvlThree);
	setIdtGate( 252, InterruptGate, SYSTEM_CS, &Interrupt252, LvlThree);
	setIdtGate( 253, InterruptGate, SYSTEM_CS, &Interrupt253, LvlThree);
	setIdtGate( 254, InterruptGate, SYSTEM_CS, &Interrupt254, LvlThree);
	setIdtGate( 255, InterruptGate, SYSTEM_CS, &Interrupt255, LvlThree);
	/*
	 * Now the machine dependent interrupt handlers
	 */
	setInterrupt( INTERVAL_Vector, &TimerManagerInterruptInterrupt );

	/*
	 * Next do those that are processor (not machine) dependent
	 */
	i586CPU::installInterrupts();

}
