/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: John L. Coolidge (coolidge@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * TYPES.cc:  PCi586 implementation of machine-dependent types
 */

#include "MachineTypes.h"
#include "Byte.h"

//#define alloca __builtin_alloca

#define B 0x10000
#define low16 (B-1)

#define LOW 0
#define HIGH 1

#define big_end(n)	((n)-1)
#define little_end(n)	0
#define next_msd(i)	((i)+1)
#define next_lsd(i)	((i)-1)
#define is_not_msd(i,n)	((i)<(n))
#define is_not_lsd(i,n)	((i)>=0)

void badd (int16 *a, int16 *b, int16 *c, int n)
{
  unsigned long acc;
  int i;

  n /= sizeof *c;

  acc = 0;
  for (i = little_end (n); is_not_msd (i, n); i = next_msd (i))
    {
      /* Widen before adding to avoid loss of high bits.  */
      acc += (unsigned long) a[i] + b[i];
      c[i] = acc & low16;
      acc = acc >> 16;
    }
}

static int
bshift (int16 *u, int k, int16* w, int16 carry_in, int n)
{
  unsigned long acc;
  int i;

  if (k == 0)
    {
      ByteCopy (u, w, n * sizeof *u);
      return 0;
    }

  acc = carry_in;
  for (i = little_end (n); is_not_msd (i, n); i = next_msd (i))
    {
      acc |= (unsigned long) u[i] << k;
      w[i] = acc & low16;
      acc = acc >> 16;
    }
  return acc;
}

/* Divide a by b, producing quotient q and remainder r.

       sizeof a is m
       sizeof b is n
       sizeof q is m - n
       sizeof r is n

   The quotient must fit in m - n bytes, i.e., the most significant
   n digits of a must be less than b, and m must be greater than n.  */

/* The name of this used to be __div_internal,
   but that is too long for SYSV.  */

void
bdiv (int16* a, int16* b, int16* q, int16* r, int m, int n)
{
  unsigned long qhat, rhat;
  unsigned long acc;
  unsigned short *u = (unsigned short *) operator new(m);
  unsigned short *v = (unsigned short *) operator new(n);
  unsigned short *u0, *u1, *u2;
  unsigned short *v0;
  int d, qn;
  int i, j;

  m /= sizeof *a;
  n /= sizeof *b;
  qn = m - n;

  /* Remove leading zero digits from divisor, and the same number of
     digits (which must be zero) from dividend.  */

  while (b[big_end (n)] == 0)
    {
      r[big_end (n)] = 0;

      a += little_end (2);
      b += little_end (2);
      r += little_end (2);
      m--;
      n--;

      /* Check for zero divisor.
      if (n == 0)
        abort ();
	*/
    }

  /* If divisor is a single digit, do short division.  */

  if (n == 1)
    {
      acc = a[big_end (m)];
      a += little_end (2);
      for (j = big_end (qn); is_not_lsd (j, qn); j = next_lsd (j))
        {
          acc = (acc << 16) | a[j];
          q[j] = acc / *b;
          acc = acc % *b;
        }
      *r = acc;
      return;
    }

  /* No such luck, must do long division. Shift divisor and dividend
     left until the high bit of the divisor is 1.  */

  for (d = 0; d < 16; d++)
    if (b[big_end (n)] & (1 << (16 - 1 - d)))
      break;

  bshift (a, d, u, 0, m);
  bshift (b, d, v, 0, n);

  /* Get pointers to the high dividend and divisor digits.  */
  u0 = u + big_end (m) - big_end (qn);
  u1 = next_lsd (u0);
  u2 = next_lsd (u1);
  u += little_end (2);

  v0 = v + big_end (n);

  /* Main loop: find a quotient digit, multiply it by the divisor,
     and subtract that from the dividend, shifted over the right amount. */

  for (j = big_end (qn); is_not_lsd (j, qn); j = next_lsd (j))
    {
      /* Quotient digit initial guess: high 2 dividend digits over high
         divisor digit.  */

      if (u0[j] == *v0)
        {
          qhat = B - 1;
          rhat = (unsigned long) *v0 + u1[j];
        }
      else
        {
          unsigned long numerator = ((unsigned long) u0[j] << 16) | u1[j];
          qhat = numerator / *v0;
          rhat = numerator % *v0;
        }

      /* Now get the quotient right for high 3 dividend digits over
         high 2 divisor digits.  */

      while (rhat < B && qhat * *next_lsd (v0) > ((rhat << 16) | u2[j]))
        {
          qhat -= 1;
          rhat += *v0;
        }

      /* Multiply quotient by divisor, subtract from dividend.  */

      acc = 0;
      for (i = little_end (n); is_not_msd (i, n); i = next_msd (i))
        {
          acc += (unsigned long) (u + j)[i] - v[i] * qhat;
          (u + j)[i] = acc & low16;
          if (acc < B)
            acc = 0;
          else
            acc = (acc >> 16) | -B;
        }

      q[j] = qhat;

      /* Quotient may have been too high by 1.  If dividend went negative,
         decrement the quotient by 1 and add the divisor back.  */

      if ((signed long) (acc + u0[j]) < 0)
        {
          q[j] -= 1;
          acc = 0;
          for (i = little_end (n); is_not_msd (i, n); i = next_msd (i))
            {
              acc += (unsigned long) (u + j)[i] + v[i];
              (u + j)[i] = acc & low16;
              acc = acc >> 16;
            }
        }
    }

  /* Now the remainder is what's left of the dividend, shifted right
     by the amount of the normalizing left shift at the top.  */

  r[big_end (n)] = bshift (u + 1 + little_end (j - 1),
                           16 - d,
                           r + little_end (2),
                           u[little_end (m - 1)] >> d,
                           n - 1);
  delete u;
  delete v;
}

void bsub (int16* a, int16* b, int16* c, int n)
{
  signed long acc;
  int i;

  n /= sizeof *c;

  acc = 0;
  for (i = little_end (n); is_not_msd (i, n); i = next_msd (i))
    {
      /* Widen before subtracting to avoid loss of high bits.  */
      acc += (long) a[i] - b[i];
      c[i] = acc & low16;
      acc = acc >> 16;
    }
}

void bmul (int16* a, int16* b, int16* c, int m, int n)
{
  int i, j;
  unsigned long acc;

  ByteZero (c, m + n);

  m /= sizeof *a;
  n /= sizeof *b;

  for (j = little_end (n); is_not_msd (j, n); j = next_msd (j))
    {
      unsigned short *c1 = c + j + little_end (2);
      acc = 0;
      for (i = little_end (m); is_not_msd (i, m); i = next_msd (i))
        {
          /* Widen before arithmetic to avoid loss of high bits.  */
          acc += (unsigned long) a[i] * b[j] + c1[i];
          c1[i] = acc & low16;
          acc = acc >> 16;
        }
      c1[i] = acc;
    }
}

#if 0
int64 operator+(const int64& a,const int64& b)
{
	int64 c;
	badd((int16*)&(a._value), (int16*)&(b._value),
		(int16*)&(c._value), sizeof c._value);
	return c;
}

int64 operator-(const int64& a, const int64& b)
{
	int64 c;
	bsub( (int16*)&(a._value), (int16*)&(b._value),
		(int16*)&(c._value), sizeof (c._value) );
	return c;
}

int64 operator/(const int64& a, const int64& b)
{
	unsigned long s[4];
	s[0] = a._value.s.low;
	s[1] = a._value.s.high;
	s[2] = 0;
	s[3] = 0;
	int64 q;
	unsigned long r[2];
	bdiv( (int16*)&s, (int16*)&(b._value), (int16*)&(q._value),
		(int16*)&r, sizeof s, sizeof (b._value));
	return q;
}

int64 operator*(const int64& a, const int64& b)
{
	unsigned int s[4];
	bmul( (int16*)&(a._value), (int16*)&(b._value),
		(int16*)&s, sizeof (a._value),
		sizeof (b._value) );
	int64 c(s[0],s[1]);
	return c;
}

int64 operator%(const int64& a, const int64& b)
{
	unsigned long s[4];
	s[0] = a._value.s.low;
	s[1] = a._value.s.high;
	s[2] = 0;
	s[3] = 0;
	int64 r;
	unsigned long q[2];
	bdiv( (int16*)&s, (int16*)&(b._value), (int16*)&q,
		(int16*)&(r._value), sizeof s, sizeof (b._value));
	return r;
}

#endif
