#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */


#include "Am7990.h"
#include "OutputStream.h"
#include "Semaphore.h"
#include "FreeRunningTimer.h"
#include "FreeRunningTimerStar.h"
#include "NetworkBuffer.h"
#include "NetworkAdministrator.h"

#ifdef __GNUC__
#pragma implementation
#endif


#define NO_MUT_UNI 1
#define ADD_CHK 1

//#define PERF 1
#define DEBUG Console()
//#define DEBUG_RECEIVE 1
//#define DEBUG_INTR 1


void
Am7990BufferDescriptor::setBuffer(NetworkBuffer *buf)
{
	unsigned value = (unsigned)buf->packet();
#ifdef DEBUG_RECEIVE
	Console() << "setAm7990BufferDescriptor::setBuffer" << eor;
	Console() << asAddress((char*) buf->packet()) << eor;
#endif DEBUG_RECEIVE
		
	unsigned length = buf->length();
	Assert (! ((unsigned int)value & 0x7));
	Assert(length > 0);
	// Assert(length <= MaxPacket);
	rmd[0] = (unsigned int)value & 0xffff;
	rmd[1] &= ~0x00ff;
	rmd[1] |= ((unsigned int)value >> 16) & 0xff ;
	rmd[2] = 0 - length;

	//Console() << "setBuffer done " << eor;
}


void
Am7990::initBuffers()
{
		/*
		 * Allocate receive bufs
		 */
	for (int i = 0; i < ReceiveBuffers; i++){
		_recvBufs[i] = getBuffer(MaxPacket);
		// setReceiveBuffer(i);
		_receiveRing[i].setBuffer(_recvBufs[i]);
		_receiveRing[i].disown();
		}

}

void
Am7990::disableChip( volatile Am7990Word *rap, volatile Am7990Word *rdp)

{
	Am7990Reg _rap((Am7990Word*)rap);
	Am7990Reg _rdp((Am7990Word*)rdp);
	_rap.write(CSR0);
	_rdp.write(LE_STOP) ;
}

void
Am7990::initChip() 
{
	/*
	 * 1. sublcass has set up the init block.
	 * 2. Reset and stop the chip
	 * 3. Allocate receive buffers and stuff them into the rings.
	 * 4. Stuff the init block into the chip
	 * 5. Allocate the transmit semaphores
	 * 6. Init the chip, set _initialized with result
	 */
#ifdef DEBUG_INIT
	Console() << "Am7990::init(): Init Block at " << asAddress(_initBlock) 
		<< "\n" << eor;
	Console() << "Am7990::init(): _receiveRing at " 
		<< asAddress(_receiveRing) << "\n" << eor;
	Console() << "Am7990::init(): _transmitRing at " 
		<< asAddress(_transmitRing) << "\n" << eor;
#endif

		/*
		 * reset and stop chip
		 */
	_rap.write(CSR0);
	_rdp.write(LE_STOP) ;
#ifdef DEBUG_INIT
	Console() << "Chip stopped, CSR0: " << asHex(_rdp.read()) 
		<< "\n" << eor;
#endif

	/*
	 * The rings and initBlock should be 8 byte aligned 
	 */
	Assert( ! ((unsigned int)_initBlock & 7) );
	Assert( ! ((unsigned int)_receiveRing & 7) );
	Assert( ! ((unsigned int)_transmitRing & 7) );

		/*
		 * Sanity checks on transmit ring
		 */
	for (int i = 0; i < TransmitBuffers; i++)
		Assert( _transmitRing[i].ownedByCPU() );


	/*
	 * Tell the chip where the init block is
	 */
	_rap.write(CSR1);
	_rdp.write(((unsigned int)&_initBlock[0]) & 0xffff);
	_rap.write(CSR2);
	_rdp.write( (((unsigned int)_initBlock) >> 16) & 0xff);
	_rap.write(CSR3);
	_rdp.write(CSR3Constant);

	_currentReceiveBuffer = 0;
	_currentTransmitBuffer = 0;
	_previousTransmitBuffer = 0;

	/*
	 * Have the chip read the init block
	 */
	_rap.write(CSR0);
#ifdef DEBUG_INIT
	Console() << "LE Init Chip, CSR: " << asHex(_rdp.read()) << "\n" << eor;
	Console() << "Init Block: \n" ;
	for (i = 0; i < 12; i++)
		Console() << asAddress(&_initBlock[i]) << " " 
			<< asHex(_initBlock[i]) << "\n" ;
	Console() << eor;
#endif
	_rdp.write (LE_INIT);
	unsigned int initcount = 0;
	_initialized = 0;
	while (! _initialized) {
		unsigned int initstatus = _rdp.read();
		if ( (initstatus & LE_ERR) || (initcount++ > 1000) ) {
			Console() << "le0: Initialization error, csr: "
				<< asHex(initstatus) << "\n" << eor;
			break;
			}
		else if (initstatus & LE_IDON) {
			_rdp.write( LE_IDON); // Clears the IDON bit.
			_initialized = 1;
#ifdef DEBUG_INIT
			Console() << "le: initialized, CSR: " << asHex(_rdp.read()) << "\n" << eor;
#endif
			break;
			}
		}
}

void
Am7990::startChip()
{
#ifdef DEBUG_INIT
	Console() << "Am7990::startChip()\n" << eor;
#endif
	_rdp.write(LE_STRT | LE_INEA);
#ifdef DEBUG_INIT
	Console() << "Am7990:" << this << ": le UP\n" << eor;
#endif DEBUG_INIT
}

Am7990::Am7990(volatile Am7990Word *rap,
	       volatile Am7990Word *rdp,
	       unsigned char debugMask, NetworkBufferFreeList *list) :
	_rap((Am7990Word*)rap), _rdp((Am7990Word*)rdp),
	EthernetDriver(debugMask,list)
{
#ifdef DEBUG_INIT
	Console() << "Am7990::Am7990(" << asAddress((char*)rap) << ", " 
		<< asAddress((char*)rdp) << ")\n" << eor;
#endif

	startHandler = 0;
	dmaFlushTime = 0;
	fTimes = 0;
	sendMapDMA = 0;
	sTimes = 0;
	sendUnMapDMA = 0;
	sUTimes = 0;
	sendITime = 0;
	sITimes = 0;
	rcvITime = 0;
	rITimes = 0;
	rcvUnMapDMA = 0;
	rTimes = 0;
	rcvBuf = 0;
	rBufTimes = 0;

	Assert(ReceiveBuffers <= 128);
	Assert(TransmitBuffers <= 128);
	_rap.write(CSR0);
	_rdp.write(LE_STOP) ;
	for (int i = 0; i < TransmitBuffers; i++){
		_transmitSemaphores[i] = new Semaphore(1);
		_transmitSemaphores[i]->setName("#Am7990TransmitSemaphore");
		_completionSemaphores[i] = new Semaphore(0);
		_completionSemaphores[i]->setName("#Am7990CompletionSemaphore");
		}
	_recvBufs = new NetworkBuffer *[ReceiveBuffers];
	Assert (_recvBufs != 0);
	_transmitMutex = new Lock;
	_interruptMutex = new Lock;
	_currentReceiveBuffer = 0;
	_currentTransmitBuffer = 0;
	_previousTransmitBuffer = 0;
	_interrupts = 0;
	_missedPackets = 0;
	_inputPackets = 0;
	_inputErrors = 0;
	_outputPackets = 0;
	_outputErrors = 0;

	_receiveNetworkBuffer = new NetworkBuffer* [ReceiveBuffers];

	_multicastMutex = new MutexSemaphore;
	for (i = 0; i < MulticastAddresses; i++)
	_multicastCount[i] = 0;
}

void 
Am7990::printInfo()
{
	Console() << "Am7990(" << asAddress(this) << "):\n";
	Console() << "\tInterrupts: " << _interrupts << "\n";
	Console() << "\tInput Packets: " << _inputPackets << "\n";
	Console() << "\tInput Errors: " << _inputErrors << "\n";
	Console() << "\tOutput Packets: " << _outputPackets << "\n";
	Console() << "\tOutput Errors: " << _outputErrors << "\n";
	Console() << "\tMissed Packets: " << _missedPackets << "\n";
	Console() << "\tStatus Register: " << asHex(_rdp.read()) << "\n";
	Console() << eor;
}

Am7990::~Am7990()
{
	for (int i = 0; i < TransmitBuffers; i++){
		_transmitSemaphores[i] = 0;
		_completionSemaphores[i] = 0;
		}
}

void
Am7990::interruptHandler()
{
#ifdef DEBUG_INTR
	Console() << "\n" << this << ":Am7990::interrupt()" << eor;
#endif DEBUG_INTR
#ifdef NO_MUT_UNI
	_interruptMutex->acquire();
	_interrupts++;
#endif
	Assert(_rap.read() == CSR0);
	Am7990Word status = _rdp.read();
#ifdef DEBUG_INTR
	Console() << ": status: " << asHex(status) << "\n" ;
#endif DEBUG_INTR

#ifdef NO_DEAD
	unsigned int previousCurrentReceiveBuffer = _currentReceiveBuffer;
	unsigned int previousCurrentTransmitBuffer = _currentTransmitBuffer;
#endif
	if (status & LE_RINT) {
		receiveInterrupt();
#ifdef PERF
		unsigned end = SystemTimer->time();
        	rcvITime = end - startHandler;
		rITimes++;
#endif PERF
		}
#ifdef ADD_CHK
	if (status & LE_ERR) {
		if (status & LE_MISS) {
			Console() << "le: Missed Packet\n" ;
			_missedPackets++;
			_rdp.write(LE_MISS);
			}
		if (status & LE_BABL) {
			Console() << this << ": Babbling!\n" ;
			_rdp.write(LE_BABL);
			}
		if (status & LE_CERR) {
			Console() << this << ": Lost Heartbeat\n" ;
			_rdp.write(LE_CERR);
			}
		if (status & LE_MERR) {
			Console() << this 
				<< ": Memory Error, le going down!\n" ;
			_rdp.write(LE_STOP);
			}
		}
#endif
	if (status & LE_TINT) {
		transmitInterrupt();
#ifdef PERF
		unsigned end = SystemTimer->time();
        	sendITime = end - startHandler;
		sITimes++;
#endif PERF
		}

	_rdp.write(LE_INEA);
#ifdef DEBUG_INTR
	Console() << "Am7990::interrupt() returns, status: " << asHex(_rdp.read()) << "\n" << eor;
#endif DEBUG_INTR

#ifdef NO_MUT_UNI
	_interruptMutex->release();
#endif
}

void
Am7990::receiveInterrupt()
{
#ifdef DEBUG_RECEIVE
	Console() << "Am7990::receiveInterrupt()\n" << eor;
	Console() << "_currentReceiveBuffer = " << asHex(_currentReceiveBuffer) 
		<< "\n" ;
	Console() << "Recv decriptor statuses: \n";
	for (unsigned int i = 0; i < ReceiveBuffers; i++)
		Console() << asHex(i) << " " << asHex(_receiveRing[i].status())
		<< "\n" ;
	Console() << eor;
#endif
	/*
	 * 1. See how many buffers the chip filled
	 *	- Look for end of packet
	 *	- Check for errors in all pieces
	 */
	Am7990BufferDescriptor *bd = &_receiveRing[_currentReceiveBuffer];

	while ( ! bd->ownedByCPU()) 
	Assert(bd->ownedByCPU());

	while (bd->ownedByCPU()) {
		_rdp.write(LE_RINT);
		unsigned int bdstat = bd->status();
#ifdef DEBUG_RECEIVE
		Console() << "Am7990::receiveInterrupt(): current descriptor is "
			<< asHex(_currentReceiveBuffer) << " :status is " 
			<< asHex(bdstat) << "\n" << eor;
#endif
		if (bd->error()) {
			/*
			 * Should have FRAM, OFLO, CRC, or BUFF lit
			 */
#ifdef DEBUG_RECEIVE
			Console() << "Am7990::receiveInterrupt(): Recv packet ";
			Console() << "buffer error\n" << eor;
#endif
			_inputErrors++;
			if (bdstat & RMD_FRAM) {
				// CRC should be set too
				Assert(bdstat & RMD_CRC);
				}
			if (bdstat & RMD_OFLO) {
				Console() << "le: buffer recv overflow\n" 
					<< eor;
				}
			if (bdstat & RMD_CRC) {
				}
			if (bdstat & RMD_BUFF) {
				/*
				 * We should not get a chaining error
				 * in this implementation
				 */
				Assert( NOTREACHED );
				}
			}
		else	{
			/* This version does NOT do chaining, so we
			 * should have all the data right in this packet.
			 */
			Assert (bdstat & LMD_STP);
			Assert (bdstat & LMD_ENP);
			unsigned int length = bd->length();
			//char *data = bd->buffer();
#ifdef DEBUG_RECEIVE
			Console() << 
			"Am7990::receiveInterrupt() got packet, data at "
				<< asAddress(data) << " length is " 
				<< asHex(length) << "\n" << eor;
#endif
			_recvBufs[_currentReceiveBuffer]->setLength(length);
			receivePacket(_recvBufs[_currentReceiveBuffer]);
			_recvBufs[_currentReceiveBuffer] = getBuffer(MaxPacket);
			_inputPackets++;

			}
		_receiveRing[_currentReceiveBuffer].setBuffer(_recvBufs[_currentReceiveBuffer]);
		_receiveRing[_currentReceiveBuffer].clearError();
		_receiveRing[_currentReceiveBuffer].disown();
		Assert (! bd->ownedByCPU());
		_currentReceiveBuffer = 
			nextReceiveDescriptor(_currentReceiveBuffer);
		bd = &_receiveRing[_currentReceiveBuffer];
		}
}

void
Am7990::transmitInterrupt()
{
#ifdef DEBUG_XMIT
	Console() << "Am7990::transmitInterrupt()\n" ;
	Console() << "_currentTransmitBuffer = " << asHex(_currentTransmitBuffer) 
		<< "\n" ;
	Console() << "Xmit decriptor statuses: \n" ;
	for (unsigned int i = 0; i < TransmitBuffers; i++)
		Console() << asHex(i) << " " << asHex(_transmitRing[i].status())
		<< "\n" ;
	Console() << eor;
#endif
	Am7990BufferDescriptor *bd = &_transmitRing[_previousTransmitBuffer];

	while ( ! bd->ownedByCPU()) 
	Assert(bd->ownedByCPU());

	while (bd->ownedByCPU()) {
		_rdp.write(LE_TINT);
		unsigned int bdstat = bd->status();
#ifdef DEBUG_XMIT
		Console() << 
			"Am7990::transmitInterrupt(): current descriptor is "
			<< asHex(_previousTransmitBuffer) << " its status is " 
			<< asHex(bdstat) << "\n" << eor;
#endif

#ifdef ADD_CHK
		if (bd->error()) {
#ifdef DEBUG_XMIT
			Console() << "Am7990::transmitInterrupt(): packet error\n" 
				<< eor;
#endif
			_outputErrors++;
			unsigned int xmiterr = bd->transmitError();
			_transmitErrors[_previousTransmitBuffer] = 1;
			if ((xmiterr & TMD_BUFF) || (xmiterr & TMD_UFLO)) {
				Console() << "le: Transmit buffer underflow,"
					<< "le going down\n" << eor;
				}
			if (xmiterr & TMD_LCOL) {
				Console() << "le: late collision\n" 
					<< eor;
				}
			if (xmiterr & TMD_LCAR) {
				Console() << "le: lost carrier\n" 
					<< eor;
				}
			if (xmiterr & TMD_RTRY) {
				Console() << "le: too many retries\n" 
					<< eor;
				}
			bd->clearTransmitError();
			}
		else if (bdstat & LMD_STP) {
#else ADD_CHK
		if (bdstat & LMD_STP) {
#endif ADD_CHK

			Assert (bdstat & LMD_ENP);	// Since we don't chain
			_outputPackets++;
			_transmitErrors[_previousTransmitBuffer] = 0;
			}
		else	{
			break;	// We own it, but we have not sent anything
			}
		_transmitRing[_previousTransmitBuffer].clearError();
		if (_transmitAsync[_previousTransmitBuffer])
		{
		  releaseBuffer(_transmitBufs[_previousTransmitBuffer]);
		  _transmitSemaphores[_previousTransmitBuffer]->V();
		}
		else
		  _completionSemaphores[_previousTransmitBuffer]->V();

		_previousTransmitBuffer = 
			nextTransmitDescriptor(_previousTransmitBuffer);
		bd = &_transmitRing[_previousTransmitBuffer];
		} // end while
}


int
Am7990::transmitPacket(NetworkBuffer *buf)
{
  	int	slot;
	
  	slot = doTransmit(buf, FALSE);
	if (slot < 0)
	  return(slot);

	_completionSemaphores[slot]->P();
	_transmitSemaphores[slot]->V();

	return (_transmitErrors[slot]);
}



int
Am7990::transmitPacketAsync(NetworkBuffer *buf)
{
	int	slot;

	slot = doTransmit(buf, TRUE);
	return((slot < 0) ? slot : 0);
}


int
Am7990::doTransmit(NetworkBuffer *buf, int async)
{
	unsigned length = buf->length();

#ifdef DEBUG_XMIT
	Console() << "Am7990::send(" << asAddress(buf->packet()) << ", "
		  << asHex(length) << '\n' << eor;
#endif

	if (length < 64) {
		Console() << "Am7990::send(): packet of length < 64 (length = "
		<< length << ") ignored\n" << eor;
		return -1;
	}


	/*
	 * Grab a lock while we figure out what transmit slot
	 * we can use, from then on out we use the slot semaphore
	 * to do mutex on a slot basis.
	 */
	_transmitMutex->acquire();
	unsigned int slot = _currentTransmitBuffer;
	_currentTransmitBuffer = nextTransmitDescriptor(_currentTransmitBuffer);
	_transmitMutex->release();

	_transmitSemaphores[slot]->P();

	_transmitAsync[slot] = async;
	_transmitBufs[slot] = buf;

	length = (length <= MaxPacket) ? length : MaxPacket;
	_transmitRing[slot].setBuffer(buf);
	_transmitRing[slot].setTransmitBits();
	_transmitRing[slot].disown();

	/*
	 * If there are multiple queuers, and we are the first in the ring
	 * then we want to set transmit demand bit to get the packet out
	 * pronto.
	 */

	/* if (we are the lowest) (tru on single processor) */
		_rdp.write(LE_TDMD | LE_INEA);

	return(slot);
}

NetworkBuffer *
Am7990::getBuffer(unsigned len)
{
	NetworkBuffer *ret =  NetworkDriver::getBuffer();
	ret->setLength(len);
	return(ret);
}

void
Am7990::releaseBuffer(NetworkBuffer *buf)
{
  buf->release();
}


int
Am7990::addMulticast(EthernetAddress *address)
{
  int		err;

  if (err = registerMulticast(address))
    return(err);

  int	bit = hashAddressTo64(address);

  _multicastMutex->P();
  if (!_multicastCount[bit]++)
  {
    _initBlock[4 + bit/16] |= 1 << (bit % 16);

    initChip();
    startChip();
  }
  _multicastMutex->V();

  return(0);
}


int
Am7990::removeMulticast(EthernetAddress *address)
{
  int		err;

  if (err = unregisterMulticast(address))
    return(err);

  /* See if we need to turn off the multicast bit */
  int bit = hashAddressTo64(address);

  _multicastMutex->P();
  _multicastCount[bit]--;

  Assert(_multicastCount[bit] >= 0);

  if (_multicastCount[bit] == 0)
  {
    _initBlock[4 + bit/16] &= ~(1 << (bit % 16));

    initChip();
    startChip();
    flushInitBlock();
  }
  _multicastMutex->V();

  return(0);
}

void
Am7990::clearStats() 
{
startHandler=0;
dmaFlushTime = 0;
fTimes = 0;
sendMapDMA = 0;sTimes = 0;
sendUnMapDMA = 0;sUTimes = 0;
sendITime = 0;sITimes = 0;
rcvITime = 0;rITimes = 0;
rcvUnMapDMA = 0;rTimes = 0;
rcvBuf = 0;rBufTimes = 0;
}

void
Am7990::printStats() 
{
	Console() << "Timings breakdown for basic ethernet test " << "\n" << eor;

	Console() << "dmaFlushTime " << dmaFlushTime << " number of time init block flushed flushed " << fTimes << "\n" << eor;

	Console() << "send IntTime" << sendITime << " number of send interrupt times " << sITimes << "\n" << eor;
	Console() << "sendMapDMA " << sendMapDMA << " number of packet mapped on send " << sTimes << "\n" << eor;
	Console() << "sendUnMapDMA " << sendUnMapDMA << " number of Times packet unmapped " << sUTimes << "\n" << eor;
	Console() << "rcvIntTime" << rcvITime << " number of receive interrupt Times" << rITimes << "\n" << eor;
	Console() << "rcvUnMapDMA " << rcvUnMapDMA << " number of receive interrupt times" << rTimes << "\n" << eor;

	Console() << "rcvBuf " << rcvBuf << " number of times packet put in queue" << rBufTimes << "\n" << eor;
}

#include "ClassInstaller.h"
DEFINECLASS("Am7990", Am7990, EthernetDriver, 0, 0);

