#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#include "Assert.h"
#include "AwaitedInterruptInterrupt.h"
#include "CPU.h"
#include "Gdb.h"
#include "ESP.h"
#include "ESPStatus.h"
#include "Kernel.h"
#include "NonReentrant.h"
#include "OutputStream.h"
#include "Process.h"
#include "PROCESSOR.h"
#include "SCSICommand.h"
#include "SS1devices.h"
#include "SS1DMA.h"
#include "ThisCPU.h"
#include "ThisProcess.h"
#include "TimeoutTimer.h"
#include "Vectors.h"

#ifdef __GNUC__
#pragma implementation
#endif

#define MASKOUT_LSB24(x) ((unsigned) ((unsigned)(x) & 0xff000000))

int ESPRegisterDebug = 0;
ESP * esp0;

Process * diskControllerInterruptProcess = 0;

// #define ESPINTERRUPTPROCESS

#ifdef ESPINTERRUPTPROCESS
void
ESPInterruptProcess()
{
	AwaitedInterruptInterrupt * interrupt =
		new AwaitedInterruptInterrupt();

        thisCPU()->setInterrupt(VEC_SCSI, interrupt);
	while (1) {
		Debug << thisProcess() << ": awaiting interrupt\n" << eor;
		interrupt->await();
		Debug << thisProcess() << ": interrupt\n" << eor;
		Assert (esp0 != 0);
		esp0->interrupt();
		}
}
void
ESPInitialize()
{
	static int initialized = 0;
	Assert( initialized == 0 );
	Process * process = new InterruptProcess( (PFV) ESPInterruptProcess,
		TheKernel->domain(), small, 0);
	Debug << "ESP interrupt process " << process << "\n" << eor;
	Assert( process != 0 );
	process->setName( "#ESPInterruptProcess" );
	process->ready();
	diskControllerInterruptProcess = process;
	thisProcess()->relinquishProcessor();
	*INTREG |= 0x10;
	initialized = 1;
}
#else // not ESPINTERRUPTPROCESS
class ESPInterrupt : public Interrupt {
protected:
	virtual void basicRaise( char * info );
public:
	ESPInterrupt() : Interrupt() {}
};

void ESPInterrupt::basicRaise( char * info ) 
{
//	static NonReentrant nonreentrant("ESP");
	static int inUse = 0;

	IntStorage intState ; PROCESSOR::disableInterrupts(intState);

//	int my_nesting = nonreentrant.enter();
	Assert( inUse == 0 );
	inUse++;

	int complete = esp0->interrupt();

	inUse--;
	Assert( inUse == 0 );
//	nonreentrant.exit(my_nesting);

	if ( complete ) 
		esp0->completed();
	PROCESSOR::restoreInterrupts( intState );
}

void
ESPInitialize()
{ 
	static int initialized = 0;
	Assert( initialized == 0 );
	ESPInterrupt * interrupt = new ESPInterrupt();
        thisCPU()->setInterrupt(VEC_SCSI, interrupt);
	*INTREG |= 0x10;
	initialized = 1;
}
#endif // ESPINTERRUPTPROCESS

ESP::ESP(unsigned char *base, int step, DMAController *dma) 
{
  Debug << "ESP::ESP()\n" << eor;
  ESPInitialize();
  inmsgCount = 0;
  outmsgCount = 0;
  outcmdLength = 0;

  currentState = S_DISCON;

  Assert(dma);
  myDMA = dma;
  myStatus = new ESPStatus(this);

  /* step is a factor that controls the alignment of register addresses. 
   A value of n denotes an n-byte boundary */

  
  transferCounter = new ESPDoubleRegister(base + LO_XFER_REG_OFFSET*step,
					  base + HI_XFER_REG_OFFSET*step);
  fifo            = new ESPRegister(base + FIFO_REG_OFFS*step);
  commandRegister = new ESPRegister(base + COMMAND_REG_OFFS*step);
  
  statusRegister  = new ESPReadRegister (base + STATUS_REG_OFFS*step);
  busIdRegister   = new ESPWriteRegister(base + BUS_ID_REG_OFFS*step);
  
  intrStatusRegister= new ESPReadRegister (base + INTR_STATUS_REG_OFFS*step);
  timeoutRegister   = new ESPWriteRegister(base + TIMEOUT_REG_OFFS*step);
  
  seqStepRegister   = new ESPReadRegister (base + SEQ_STEP_REG_OFFS*step);
  periodRegister    = new ESPWriteRegister(base + PERIOD_REG_OFFS*step);
  
  fifoFlagsRegister = new ESPReadRegister (base + FIFO_FLAGS_REG_OFFS*step);
  offsetRegister    = new ESPWriteRegister(base + OFFSET_REG_OFFS*step);
  
  configRegister    = new ESPRegister (base + CONFIG_REG_OFFS*step);
  
  clockFactorRegister= new ESPWriteRegister(base + CLOCK_FCT_REG_OFFSET*step);
  
  testRegister= new ESPWriteRegister(base + TEST_REG_OFFSET*step);

	/*
	 * Initialize DMA, then enable interrupts
	 */
	myDMA->reset();
	myDMA->enableIntr();
	myDMA->enable();

	commandRegister->write(CMD_RESET_CHIP);
	commandRegister->write(CMD_NOP);
	clockFactorRegister -> write( 5 );
	periodRegister -> write( 0 );
	offsetRegister -> write( 0 );
	timeoutRegister->write( 0xc8 );
	transferCounter->write(0); 
	commandRegister->write(DMA_NOP);
	myStatus->update();

	configRegister->write( configRegister->read() | CF_SCSI_RES_INT_DIS);
	commandRegister->write(CMD_RESET_SCSI);
	TimeoutTimer *timer = new TimeoutTimer();
	timer->start(4000000);	// 4 sec.
	timer->await();
	delete timer;
	configRegister->write( 0x17);
	Assert (configRegister->read() == 0x17);

#ifdef DEBUG
	if (ESPRegisterDebug)
		dumpRegisters();
#endif
	Debug << "ESP::ESP() done\n\n" << eor;
}

ESP::~ESP()
{
  Debug << "ESP::~ESP()\n" << eor;
  delete myStatus;
  delete transferCounter;
  delete fifo;
  delete commandRegister;
  delete statusRegister;
  delete busIdRegister;
  delete intrStatusRegister;
  delete timeoutRegister;
  delete seqStepRegister;
  delete periodRegister;
  delete fifoFlagsRegister;
  delete offsetRegister;
  delete configRegister;
  delete clockFactorRegister;
  delete testRegister;
}

void 
ESP::configure(unsigned char id, int parity,
		int resetIntrDisable, int slow) 
{
  Debug << "Configured ESP\n" << eor;
}

void
ESP::checkSCSIPhase()
{
  Debug << "ESP::checkSCSIPhase()\n" << eor;

  unsigned int phase;
  phase = myStatus->phase();
  switch (phase) {
    case DATA_OUT_PHASE:
      Debug << "SCSI Data out phase\n" << eor;
      currentState = S_SEND_DATA;
      sendData();
      break;

    case DATA_IN_PHASE:
      Debug << "SCSI Data in phase\n" << eor;
      currentState = S_RECEIVE_DATA;
      receiveData();
      break;

    case COMMAND_PHASE:
      Debug << "SCSI Command phase\n" << eor;
      /* 
       * A command has been requested by the target.
       * Get the command bytes and put them in the fifo
       */
	commandRegister->write(CMD_FLUSH_FIFO);
	currentState = S_CMD_OUT;
	Debug << "Sending command: " ;
	for (int i=0; i < outcmdLength; i++) {
	  Debug << asHex(outcmd[i]);
	  fifo->write(outcmd[i]);
	  }
	Debug << "\n" << eor;
	commandRegister->write(CMD_TRANSFER_INFO);
	break;

    case STATUS_PHASE:
      Debug << "SCSI Status phase\n" << eor;
      /* use non-DMA mode */
      unsigned char s0 = fifo->read();
      unsigned char s1 = fifo->read();
      Debug << "Got status bytes: " << asHex(s0) << ", " << asHex(s1)
	<< "\n" << eor;
      currentState = S_FINISH_CMD;
      commandRegister->write(CMD_FLUSH_FIFO);
      commandRegister->write(CMD_COMPL_SEQ);
      break;

    case MSG_OUT_PHASE:
      Debug << "SCSI Message Out phase\n" << eor;
	commandRegister->write(CMD_NOP);
	commandRegister->write(CMD_FLUSH_FIFO);
      if (currentState == S_NEGOTIATE) {
	/*
	 * Load fifo with sync transfer parameters
	 */
	Debug << "Sending synchronous parameters\n" << eor;
	fifo->write(0x1);
	fifo->write(0x3);
	fifo->write(0x1);
	fifo->write(0x5);
	fifo->write(0x15);
        currentState = SYNC_OUT;
	}
      else {
	if (outmsgCount == 0)
		fifo->write(MSG_NOP);
	else {
		for (int i = 0; i < outmsgCount; i++ )
			fifo->write(outmsg[i]);
		}
	outmsgCount = 0;
        currentState = S_MSG_OUT;
	}
	commandRegister->write(CMD_TRANSFER_INFO);
      break;

    case MSG_IN_PHASE:
	Debug << "SCSI Message In phase\n" << eor;
	if (currentState != S_MSG_IN) 
		inmsgCount = 0;		// Beginning to receive message
	currentState = S_MSG_IN;
	commandRegister->write(CMD_FLUSH_FIFO);
	commandRegister->write(CMD_TRANSFER_INFO);
      break;
    default:
      Debug << "SCSI Invalid Bus Phase\n" << eor;
        /*XXXXX
          XXXX FIX ME: Specify action for this case
          XXXXX*/
      Assert (NOTREACHED);
      break;
    }
  Debug << "ESP::checkSCSIPhase() returns\n" << eor;
}

void
ESP::sendData()
{
  commandRegister->write(CMD_FLUSH_FIFO);
  Assert(!myDMA->requestPending());
  myDMA->flush();
  myDMA->setAddress((unsigned char *)_dataBuf, _dataLen);
  myDMA->memoryToDeviceDirection();
  transferCounter->write(_dataLen); 
  commandRegister->write(DMA_NOP);
  commandRegister->write(DMA_TRANSFER_INFO);
}

void
ESP::receiveData()
{
  Assert(!myDMA->requestPending());
  myDMA->flush();
  myDMA->setAddress((unsigned char *)_dataBuf, _dataLen);
  myDMA->deviceToMemoryDirection();
  transferCounter->write(_dataLen);
  commandRegister->write(DMA_NOP);
  commandRegister->write(DMA_TRANSFER_INFO);
}


void
ESP::initControllerAction(SCSICommand &command, unsigned char deviceId,
	char *dataBuf, unsigned bufLen)
{
  Debug << "ESP::initControllerAction(" << deviceId << ", " 
	<< asAddress(dataBuf) << ", " << asHex(bufLen) << ")\n" << eor;

	Assert (&command != 0);

#ifdef DEBUG
        if (ESPRegisterDebug)
                dumpRegisters();
#endif
  currentDevice = this; 
  currentState = S_INIT;


  busIdRegister->write(deviceId);
  timeoutRegister->write(currentDevice->timeout());
  //unsigned char offset = currentDevice->syncOffset();


  commandRegister->write(CMD_FLUSH_FIFO);
	for (int i = 0; i < command.length(); i++){
		fifo->write(command.byte(i));
		outcmd[i] = command.byte(i);
		}
	outcmdLength = command.length();
	Assert (outcmdLength <= MAX_CMD);

  Debug << "Initiating Command SEL_WO_ATN\n" <<eor;
  commandRegister->write(CMD_SEL_WO_ATN);
}


void
ESP::initTarget(unsigned char unit)
{
	Debug << "ESP::initTarget(" << asHex(unit) << ")\n" << eor;

	/*
	 * Clear out previous device info
	 */
	if (targets[unit] != 0)
		delete targets[unit];
	targets[unit] = 0;

	currentState = S_NEGOTIATE;
	busIdRegister->write(unit);

	/* 
	 * DMA is necessary for interrupts
	 */
        Assert(!myDMA->requestPending());
	myDMA->flush();

	commandRegister->write(CMD_FLUSH_FIFO);
	fifo->write(MSG_IDENTIFY | unit);
	commandRegister->write(DMA_NOP);

	Debug << "Initiating Command CMD_SEL_W_ATN_STOP\n" <<eor;
	commandRegister->write(CMD_SEL_W_ATN_STOP);

	waitForCompletion();
	if (currentError == ERR_NO_ERROR) {
		/*
		 * Negotiate params
		 */
		Debug << "ESP::initTarget(" << asHex(unit) << 
		"): negotiating params\n" << eor;
		}
}

int ESPWeird = 1;

int
ESP::interrupt()
{
	int completed = 0;
	Debug << "\nESP::interrupt\n" <<eor;

	/*
	 * Check for blatant errors BEFORE updating status
	 */

	myDMA->disable();
	myDMA->disableIntr();

	if (myDMA->errorPending()) {
		Console() << "ERROR SCSI interrupt: DMA error pending\n"
		  	  << eor;
		dumpRegisters();
		if ( ESPWeird ) 
			TrapToDebugger();
		Assert( NOTREACHED );
		currentError = ERR_DMA;
	}
	if (!myDMA->intrPending() && !(statusRegister->read() & STAT_INT)) {
		/*
		Console() << "ERROR SCSI interrupt: spurious interrupt????\n" 
			  << eor;
		dumpRegisters();
		 */
		myDMA->enable();
		myDMA->enableIntr();
		return 0;
	}
	if (myDMA->requestPending()) {
		Console() << "ERROR SCSI interrupt: DMA request pending\n" << eor;
		dumpRegisters();
		if ( ESPWeird ) 
			TrapToDebugger();
		Assert( NOTREACHED );
		currentError = ERR_DMA_REQ_PEND;
	}

	/* 
	 * Update Status 
	 * This resets some status registers
	 */
	myStatus->update();
#ifdef DEBUG
	if (ESPRegisterDebug)
		dumpRegisters();
#endif

	/*
	 * Check for blatant errors AFTER updating status
	 */
	if (myStatus->resetDetected()) {
		Console() << "SCSI Bus Reset!\n" << eor;
		currentError = ERR_RESET;
		dumpRegisters();
		return 0;
		}
	if (myStatus->grossError()){
		Console() << "SCSI interrupt: gross error\n" << eor;
		currentError = ERR_GROSS;
		dumpRegisters();
		}
	if (myStatus->parityError()) {
		Console() << "SCSI interrupt: parity error\n" << eor;
		currentError = ERR_PARITY;
		dumpRegisters();
		}

	currentError=ERR_NO_ERROR;

  switch (currentState) {

  case S_INIT_DMA:
    /* This has been done in DMA mode */
    myDMA->release();
  case S_INIT:
  case S_NEGOTIATE:
    Debug << "ESP State: S_INIT or S_NEGOTIATE...\n" << eor;
     if ( (currentState != S_NEGOTIATE) && (transferCounter->read() != 0)) {
	Console() << "SCSI transfer counter has unexpected value: " 
		<< asHex(transferCounter->read()) << "\n" << eor;
	abortCommand();
	completed = 1;
        currentState = S_DISCON;
      }
    if (currentError == ERR_NO_ERROR) {
      if (myStatus->reselected()) {
	Console() << "ESP: Aborting command, reselected during initiation\n" << eor;
	abortCommand();
	completed = 1;
	currentState = S_RESEL;
	/* 
	 * enable reselection 
	 * use non-DMA mode --just 2 bytes will be transfered 
	 */
	commandRegister->write(CMD_ENABLE_SR);
      }
      else if (myStatus->disconnected()) {
	/* In this case this should be a timeout error */
	/* May be we should mark the device as absent */
	Debug << "ESP: Selection Timeout for device\n" << eor;
	currentError = ERR_NO_RESPONSE;
	abortCommand();
	completed = 1;
	currentState = S_DISCON;
      }
      else if (! myStatus->busService()) {
	Console() << "Unexpected non-bus-service interrupt\n" << eor;
	abortCommand();
	completed = 1;
        currentState = S_DISCON;
      }
      else if (myStatus->functionComplete()) {
	/*
	 * Now, according to the spec, the sequence step register
	 * should have a value 0x04
	 */
	if ( (currentState != S_NEGOTIATE) && (myStatus->sequence() != 0x04)) {
	  checkSCSIPhase();
	}
	else 
	  /* This will happen if everything goes fine
	   * Information transfer will follow
	   */
	  checkSCSIPhase();
      }
      else {
	Console() << "SCSI Unidentified error during initiation\n" << eor;
	abortCommand();
	completed = 1;
	currentState = S_DISCON;
      }
    }
    else {
      /* a fatal error has already been detected */
      /* cancel current initiation */
      abortCommand();
      completed = 1;
      currentState = S_DISCON;
    }
    break;

  case S_RECEIVE_DATA:
	myDMA->drain();
  case S_SEND_DATA:
	Debug << "ESP State: S_SEND_DATA or S_RECEIVE_DATA\n" << eor;
	myDMA->release();
	transferCounter->write(0);
	commandRegister->write(DMA_NOP);
    if (currentError == ERR_NO_ERROR) {
      if (myStatus->disconnected()) {
	  Console() << "SCSI Target released BSY unexpectedly\n" <<eor;
	abortCommand();
      	completed = 1;
	currentState = S_DISCON;
      }
      else 
	checkSCSIPhase();
    }
    else {
      /* a fatal error has already been detected */
      /* cancel current initiation */
      abortCommand();
      completed = 1;
      currentState = S_DISCON;
    }
    break;

  case S_FINISH_CMD:
    Debug << "ESP State: S_FINISH_CMD\n" << eor;
    commandRegister->write(CMD_NOP);
    if (currentError == ERR_NO_ERROR) {
      if (myStatus->disconnected()) {
	Console() << "SCSI Target released BSY unexpectedly\n" <<eor;
	abortCommand();
	completed = 1;
	currentState = S_DISCON;
      }
      else if (myStatus->busService()) {
	Console() << "SCSI: no MSG_IN phase after status\n" <<  eor;
	checkSCSIPhase();
      }
      else if (myStatus->functionComplete()) {
	inmsg[inmsgCount++] = fifo->read();
	Assert (inmsgCount < MAX_MSG);
	/* 
	 * decide whether to accept message or not 
	 * Since we do not use linked commands, only
	 * MSG_CMD_COMPLETE is accepted
	 */
	currentState = S_GOT_MSG;
	if (inmsg[0] == MSG_CMD_COMPLETE) {
	  commandRegister->write(CMD_ACCEPT_MSG);
	}
	else { 
	  /* reject message */
	  /* this message will be transmitted during the next
	   * message out phase
	   */
	  messageToSend = MSG_REJECT;
	  /* the following command causes no extra interrupt */
	  commandRegister->write(CMD_SET_ATN);
	  commandRegister->write(CMD_ACCEPT_MSG);
	}
      } 
    }
    else {
      /* a fatal error has already been detected */
      /* cancel current initiation */
      abortCommand();
      completed = 1;
      currentState = S_DISCON;
    }
    break;

  case S_DISCON:
    Debug << "ESP State: S_DISCON\n" << eor;
    /* only reselection is allowed in this case */
    if (currentError == ERR_NO_ERROR &&
	myStatus->reselected()) {
      Debug << "Being reselected while in disconnect state\n" << eor;
      currentState = S_RESEL;
      commandRegister->write(CMD_ENABLE_SR);	
    }
    else {
      Console() << "SCSI invalid interrupt while in disconnect state\n" << eor;
      dumpRegisters();
    }
    break;

  case S_RESEL:
    Debug << "ESP State: S_RESEL\n" << eor;
    if (currentError == ERR_NO_ERROR &&
	myStatus->reselected()) {
      if (myStatus->busService()) {
	Console() << "SCSI Phase mismatch\n" << eor;
	currentState = S_DISCON;
      }
      else if (myStatus->functionComplete()) {
	/* read reselection ID and Identify message from fifo */
	unsigned char reselectId = fifo->read();
	inmsg[inmsgCount++] = fifo->read();
	Assert (inmsgCount < MAX_MSG);
	currentState = S_GOT_MSG;
	/* check if this is a valid identify message */
	if (inmsg[0] & 0x80) {
	  Debug <<"SCSI reselection --accepting message\n" << eor;
	  /* accept it */
	  /* encode target id */
	  unsigned char id=0;
	  while(reselectId) {
	    reselectId >>= 1;
	    id++;
	  }
	  id--; /* id should be in the range 0..7 */
	  //unsigned char lun=inmsg[0] & 0x07;
	  /*XXXXXXX
	    XXXXXX  FIX ME: Add code to finish reselection
	    XXXXXXX*/
	  commandRegister->write(CMD_ACCEPT_MSG);
	}
	else {
	  Debug <<"SCSI reselection failed --rejecting message\n" << eor;
          messageToSend = MSG_REJECT;
	  /* the following command causes no extra interrupt */
	  commandRegister->write(CMD_SET_ATN);
	  commandRegister->write(CMD_ACCEPT_MSG);
	}
      }
    }
    else {
      Debug << "SCSI reselection failed\n" << eor;
	currentState = S_DISCON;
    }
    break;

  case S_GOT_MSG:
    Debug << "ESP State: S_GOT_MSG\n" << eor;
    inmsgCount = 0;
    if (currentError == ERR_NO_ERROR) {
      if (myStatus->disconnected()) {
	/* I/O Complete */
	// complete();			 //currentDevice->complete();
	completed = 1;
	currentState = S_DISCON;
	myDMA->flush();
      }
      else {
	/* Bus service interrupt */
	checkSCSIPhase();
      }
    }
    else {
      abortCommand();
      completed = 1;
      currentState = S_DISCON;
    }
    break;

  case S_CMD_OUT:
    Debug << "ESP State: S_CMD_OUT\n" <<eor;
    if (currentError == ERR_NO_ERROR &&
	myStatus->busService()) {
      checkSCSIPhase();
    }
    else {
      abortCommand();
      completed = 1;
      currentState = S_DISCON;
    }
    break;

  case S_MSG_IN:
    Debug << "ESP State: S_MSG_IN\n" << eor;
    if (currentError == ERR_NO_ERROR &&
        myStatus->functionComplete()) {
      if (myStatus->phase() == MSG_IN_PHASE) {
	inmsg[inmsgCount++] = fifo->read();
	Debug << " Got msg byte: " << asHex(inmsg[inmsgCount-1]) << "\n" << eor;
	currentState = S_GOT_MSG;
	if (inmsg[0] == 0x1 && inmsgCount < 4) {
	  Debug << "More msg to come\n" << eor;
	  commandRegister->write(CMD_TRANSFER_INFO);
	}
	else if (inmsg[0] == 0x1 || acceptMessage(inmsg[0])) {
		Debug << "Accepting message\n" << eor;
		commandRegister->write(CMD_ACCEPT_MSG);
		}
	else {
	  Debug << "Rejecting message\n" << eor;
	  messageToSend = MSG_REJECT;
	  /* the following command causes no extra interrupt */
	  commandRegister->write(CMD_SET_ATN);
	  commandRegister->write(CMD_ACCEPT_MSG);
	}
      }
      else {
	/* phase mismatch */
	Console() << "SCSI Phase mismatch (should be message in)\n" << eor;
      }
    }
    else {
      abortCommand();
      completed = 1;
      currentState = S_DISCON;
    }
    break;

  case SYNC_IN:
    Debug << "ESP State: SYNC_IN\n" << eor;
    if (currentError == ERR_NO_ERROR && myStatus->functionComplete()) {
      if (myStatus->phase() == MSG_IN_PHASE) {
		currentState = S_MSG_OUT;
	}
	else {
	  messageToSend = MSG_REJECT;
	  /* the following command causes no extra interrupt */
	  commandRegister->write(CMD_SET_ATN);
	  commandRegister->write(CMD_ACCEPT_MSG);
		}
	}
    else if (currentError == ERR_NO_ERROR && myStatus->busService()) {
		unsigned char byte = fifo->read();
		Debug << "Got msg byte: " << asHex(byte) << "\n" << eor;
		commandRegister->write(CMD_TRANSFER_INFO);
	}
      else {
	/* phase mismatch */
	Console() << "SCSI Phase mismatch (should be message in)\n" << eor;
      abortCommand();
      completed = 1;
      currentState = S_DISCON;
	}
    break;

  case SYNC_IN3:
    Debug << "ESP State: SYNC_IN3\n" << eor;
	currentState = SYNC_IN;
	commandRegister->write(CMD_TRANSFER_INFO);
    break;

  case SYNC_IN2:
    Debug << "ESP State: SYNC_IN2\n" << eor;
	currentState = SYNC_IN;
	commandRegister->write(CMD_FLUSH_FIFO);
	commandRegister->write(CMD_NOP);
	commandRegister->write(CMD_TRANSFER_INFO);
    break;

  case S_MSG_OUT:
    Debug << "ESP State: S_MSG_OUT\n" << eor;
	commandRegister->write(CMD_NOP);
    if (currentError == ERR_NO_ERROR && myStatus->busService()) {
      if (myStatus->phase() != MSG_OUT_PHASE)
	Console() << "SCSI phase mismatch (message out expected\n" << eor;
      checkSCSIPhase();
    }
    else if (currentError == ERR_NO_ERROR && myStatus->disconnected()) {
	// complete();
	completed = 1;
    }
    else {
      abortCommand();
      completed = 1;
      currentState = S_DISCON;
    }
    break;

  case SYNC_OUT:
	Debug << "ESP State: SYNC_OUT\n" << eor;
	if (currentError == ERR_NO_ERROR && myStatus->busService()) {
	if (myStatus->phase() != MSG_IN_PHASE)
		Console() << 
		"SCSI synch negotiate phase mismatch (message in expected)\n" 
		<< eor;
	currentState = SYNC_IN;
	commandRegister->write(CMD_NOP);
	commandRegister->write(CMD_FLUSH_FIFO);
	commandRegister->write(CMD_TRANSFER_INFO);
	}
    else {
      abortCommand();
      completed = 1;
      currentState = S_DISCON;
    }
    break;

  default:
    Console() << "ESP: Unknown state\n" << eor;
    Assert (NOTREACHED);
    break;
  }


  myDMA->enable();
  myDMA->enableIntr();
#ifdef DEBUG
  if (ESPRegisterDebug)
    dumpRegisters();
#endif 
/*
  if ( completed ) 
    complete();
 */
  Debug << "ESP::interrupt() complete " << completed << "\n" <<eor;
  return completed;
}

int
ESP::acceptMessage( unsigned char msg)
{
	unsigned retval;

	switch (msg) {
		/*
		 * Only accept these
		 */
		case 0x0:
		case 0x7:
		case 0x8:
			retval = 1;
			break;
		default:
			retval = 0;
			break;

		};
	return retval;
}

void
ESP::dumpRegisters()
{
	Console() << "ESP XferCount: " << asHex(transferCounter->read()) << "\n";
	Console() << "ESP Command: " << asHex(commandRegister->read()) << "\n" ;
	Console() << "Current CSR: " << asHex(statusRegister->read()) << "\n" ;
	Console() << "Saved CSR: " << asHex(myStatus->status()) << "\n" ;
	Console() << "Saved INT_CSR: " << asHex(myStatus->intStatus()) << "\n" ;
	Console() << "Current SEQ: " << asHex(seqStepRegister->read()) << "\n" ;
	Console() << "Saved SEQ: " << asHex(myStatus->sequenceStep()) << "\n" ;
	Console() << "ESP FIFO flags: " << asHex(fifoFlagsRegister->read()) << "\n" ;
	Console() << "ESP Config: " << asHex(configRegister->read()) << "\n" ;

	Console() << "DMA CSR: " << asHex(myDMA->status()) << "\n" ;
	Console() << "DMA ADDR: " << asHex(myDMA->address()) << "\n" ;
	Console() << "DMA Byte Count: " << asHex(myDMA->byteCount()) << "\n" ;

	Console() << "Command buffer: " << asAddress(_commandBuf) << "\n" ;
	Console() << "Command Len: " << asHex(_commandLen) << "\n" ;
	Console() << "Data buffer: " << asAddress(_dataBuf) << "\n" ;
	Console() << "Data Len: " << asHex(_dataLen) << "\n" ;
	Console() << "\n" << eor;
}




void
ESP::abortCommand()
{
  Console() << "ESP::abortCommand()\n" << eor;
  commandRegister->write(CMD_FLUSH_FIFO);
  transferCounter->write(0); 
  commandRegister->write(DMA_NOP);
}

unsigned
ESP::waitForCompletion()
{
  Debug << "ESP::waitForCompletion()\n" << eor;
#ifdef DEBUG
	if (ESPRegisterDebug)
		dumpRegisters();
#endif 
	return SCSIController::waitForCompletion() ;
}
