#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *               University of Illinois at Urbana-Champaign
 *               Department of Computer Science
 *               1304 W. Springfield Ave.
 *               Urbana, IL     61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *      All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#include "Assert.h"
#include "Byte.h"
#include "Domain.h"
#include "OutputStream.h"
#include "SCSICommand.h"
#include "SCSIData.h"
#include "SCSIDevice.h"
#include "SPARCTranslation.h"
#include "Semaphore.h"
#include "PhysicalMemoryChain.h"
#include "Process.h"
#include "PROCESSOR.h"
#include "ThisProcess.h"
#include "SS1Configuration.h"

#ifdef __GNUC__
#pragma implementation
#endif

/* Default values */
static const unsigned char DEFAULT_TIMEOUT       = 0x93; // 250 ms at 24 MHz
static const unsigned char DEFAULT_SYNC_OFFSET   = 0;
static const unsigned char DEFAUT_SYNC_PERIOD    = 0; 
static const int DEFAULT_RETRIES = 3;

 SCSIDevice::SCSIDevice(unsigned char busId)
{
  myBusId = busId;

  myTimeout = DEFAULT_TIMEOUT;
  mySyncOffset = DEFAULT_SYNC_OFFSET;
  mySyncPeriod = DEFAUT_SYNC_PERIOD;

  myDataPointer = 0;
  myCommand = 0;
}


unsigned 
SCSIController::waitForCompletion()
{
	Debug << "SCSIController::waitForCompletion()\n" << eor;
	_signalSemaphore->P(); 
	Debug << "SCSIController::waitForCompletion() done\n" << eor;
	return currentError;
}

void
SCSIController::complete()
{
	Debug << "SCSIController::complete()\n" << eor;
	_signalSemaphore->V();
}

SCSIController::SCSIController() : SCSIDevice(DEFAULT_CONTROLLER_BUS_ID)
{
  Debug <<"SCSI Controller()" << eor;
  currentError = ERR_NO_ERROR;
  _mutex = new MutexSemaphore();
  _mutex->setName("SCSIDevice::_mutex");
  _signalSemaphore = new GraciousSemaphore(0, 1);
//  _signalSemaphore = new Semaphore(0, 1);
  _signalSemaphore->setName("SCSIDevice::_signalSemaphore");
  _commandBuf = 0;
  _commandLen = 0;
  currentError = ERR_NO_ERROR;
  for (int i = 0; i < MAX_TARGETS; i++)
	targets[i] = 0;
}

int
SCSIController::diskInfo(int unit, unsigned *blks, unsigned *blksize, char *vendor, unsigned vendorlen)
{
	TestUnitReadyCommand readyCommand(0,0,0);
	unsigned status = doCommand(unit, readyCommand, 0, 0, 0);

	unsigned char modeData[127];
	ByteZero(modeData, sizeof(modeData));
	
	ModeSenseCommand modeCommand(0, sizeof(modeData));
	*blks = 0;
	*blksize = 0;
	thisProcess()->relinquishProcessor();
	status = doCommand(unit, modeCommand, (char *)modeData, sizeof(modeData), 1);
	if (!status) {
		*blks = (modeData[5] << 16) | (modeData[6] << 8) | modeData[7] ;
		*blksize = (modeData[9] << 16) | (modeData[10] << 8) | modeData[11] ;
		unsigned char *vdata = modeData + 12;
		unsigned vendorleng = modeData[3] - 4 - 8;
		for (int i = 0; ((i < vendorleng) && (i < vendorlen)); i++) 
			*vendor++ = *vdata++;
		}
	else
		Console() << "SCSIController::diskInfo() failed, status: "
			<< status << "\n" << eor;
	return status;
}

unsigned
SCSIController::read(unsigned unit, unsigned startblock, 
		     unsigned blocks, char* buf, unsigned buflen)
{
	Debug 	<< "SCSIController::read(" << asHex(unit) << ", " 
		<< asHex(startblock) << ", " << asHex(blocks) << ", " 
		<< asAddress(buf) << ", " << asHex(buflen) << "\n" << eor; 

	static ReadCommandGroup0 command(0,0,0);
	unsigned remaining = blocks;
	while (1) {
		/* 64k max transfers (128 blocks) */
		unsigned nblocks = (remaining > 128 ? 128 : remaining);
		unsigned nbytes = nblocks << 9; /* 512 bytes per block */
		char * end = buf + nbytes - 1;
		if (((unsigned) buf & 0xff000000) != 
		     ((unsigned) end & 0xff000000)) {
			nbytes = 0x01000000 - ((unsigned) buf & 0x00ffffff);
			nblocks = nbytes >> 9;
			Assert(nblocks <= remaining);
		}
		command.init(0, startblock, nblocks);
                unsigned stat = doCommand(unit, command, buf, nbytes, 1);
		if (stat)
			return(blocks - remaining);
		remaining -= nblocks;
		if ( remaining == 0 )
			break;
		startblock += nblocks;
		buf += nbytes;
	}
	Debug << "SCSIController::read() done\n" << eor;
	return (blocks - remaining);
}

unsigned
SCSIController::write(unsigned unit, unsigned startblock, unsigned blocks, char* buf, unsigned buflen)
{
	Debug 	<< "SCSIController::write(" << asHex(unit) << ", " 
		<< asHex(startblock) << ", " << asAddress(buf) << ", " 
		<< asHex(buflen) << "\n" << eor;
 
	static WriteCommandGroup0 command(0,0,0);
	unsigned remaining = blocks;
	while (1) {
		/* 64k max transfers (128 blocks) */
		unsigned nblocks = (remaining > 128 ? 128 : remaining);
		unsigned nbytes = nblocks << 9; /* 512 bytes per block */
		char * end = buf + nbytes - 1;
		if (((unsigned) buf & 0xff000000) != 
		     ((unsigned) end & 0xff000000)) {
			nbytes = 0x01000000 - ((unsigned) buf & 0x00ffffff);
			nblocks = nbytes >> 9;
			Assert(nblocks <= remaining);
		}
		command.init(0, startblock, nblocks);
                unsigned stat = doCommand(unit, command, buf, nbytes, 1);
		if (stat)
			return(blocks - remaining);
		remaining -= nblocks;
		if ( remaining == 0 )
			break;
		startblock += nblocks;
		buf += nbytes;
	}
	Debug << "SCSIController::write() done\n" << eor;
	return (blocks - remaining);
}


unsigned 
SCSIController::doCommand(unsigned char unit, SCSICommand &command, char *buf, unsigned buflen, int read)
{
	currentError = ERR_NO_ERROR;
	_mutex->P();

	/*
	 * Map the command and data to DMA before the I/O.
	 */
	_commandLen = command.length();
	_commandBuf = (char *) command.dataLoc();
	if (buflen > 0) {
		if (buf >= KERNEL_START_VIRTUAL &&
		    buf < KERNEL_END_VIRTUAL) {
			_dataBuf = 
			    SPARCTranslation::prepareForDMA(buf, buflen, read);
		} else {
			_dataBuf = SPARCTranslation::mapToDMA(buf, buflen);
		}
	}
	Assert (buflen <= 0x10000);
	_dataLen = buflen; 

	initControllerAction(command, unit, buf, buflen);
	unsigned stat = waitForCompletion();
	if (buflen > 0) {
		if (_dataBuf >= (char *) DVMA_BASE) {
			SPARCTranslation::unMapDMA(_dataBuf);
		} else {
			SPARCTranslation::postDMA(_dataBuf, buflen, read);
  		}
	}

	_mutex->V();
	return stat;
}


/*
 * Go through all the targets on the bus:
 * 	- See if anybody's there.
 *	- If so, negotiate synchronous parameters, etc.
 *	- Update the table of known targets.
 */
void
SCSIController::initializeTargetTable()
{
  for (unsigned char id = 0; id < 4; id++) {
	if (id == this->busId() ) continue;	// I know myself!
	initTarget(id);
	}
}
