#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * SS1TimerManager.c - machine dependent portion of timer managing
 */

#include "Assert.h"
#include "CPU.h"
#include "ThisCPU.h"
#include "MACHINE_TIMER_MANAGER.h"
#include "OutputStream.h"
#include "SPARCasmutil.h"
#include "SS1FreeRunningTimer.h"
#include "SS1Timer.h"
#include "Lock.h"

#ifdef __GNUC__
#pragma implementation
#endif

void (*SS1TickInterruptFunc)() = 0;

#ifdef SAMPLEPC
//  Code to capture 1000 samples of the PC.
#include "SS1Gdb.h"
#include "sun4c/asm_linkage.h"
#include "sun4c/reg.h"
const pcSampleSize = 1000;
char *pcSample [pcSampleSize];
int pcSampleNext = 0, pcSampleMax = 0;


char *getPC (char *trapSP)
{
	flushWindows();			// Flush everything out to the stack
	//unsigned * lastRegisters = registers;	// Save for later restore
	/*
         * 1.  Allocate and fill in the registers structure.
	 * 2.  Point the global registers at it.
	 * 3.  Call the gdb handler
	 * 4.  Restore the registers
	 */
	unsigned myregisters[NUM_REGISTERS];
	char *userSP = ((char**)trapSP)[14];
	unsigned *source;
	unsigned *dest;

/******************************************************************************
 * Store away the registers (from the stack) into myregisters
 ******************************************************************************/
	/*
	 * Get the saved globals
	 */
	source = (unsigned *)(trapSP + MINFRAME + 3*4);
	dest = myregisters;
	for (int i = 0; i < 8; i++)
		*dest++ = *source++;
	/*
	 * Get the saved outs, the windows ins.
	 */
	source = (unsigned *) (trapSP + 8*4);
	dest = &myregisters[8];
	for (i = 0; i < 8; i++)
		*dest++ = *source++;
	/*
	 * Get the saved locals
	 */
	source = (unsigned *) userSP;
	dest = &myregisters[16];
	for (i = 0; i < 8; i++)
		*dest++ = *source++;
	/*
	 * Get the saved ins
	 */
	source = (unsigned *) (userSP + 8*4);
	dest = &myregisters[24];
	for (i = 0; i < 8; i++)
		*dest++ = *source++;
	/*
	 * Get the special registers, zero out currently unused regs
	 */
	source = (unsigned *)( trapSP + MINFRAME);
	myregisters[64] = 0;	// %y
	myregisters[65] = source[PSR];	// psr
	myregisters[66] = 0;	// %wim
	myregisters[67] = 0;	// %tbr
	myregisters[68] = source[PC];	// %pc
	myregisters[69] = source[nPC];	// %npc
	myregisters[70] = 0;	// %fpsr
	myregisters[71] = 0;	// %cpsr

	return (char *) myregisters[68];
}
#endif

extern TimerManager *TimeoutManager;
extern TimerManager *PeriodicManager;

static Lock *TimerManagerLock;
static unsigned int limitval; // for dummy read to clear interrupt.

void
TimerManagerInterrupt::basicRaise( char *stackPointer )
{
	Debug << "TimerManagerInterrupt::basicRaise()\n" << eor;
	Assert( this != 0 );

	Assert ( TimerManagerLock != 0 );
	TimerManagerLock->acquire();

	/*
	 * reset the and limit reached bits and clear the interrupt
	 */
	limitval = *LIMIT0;	// dummy read to clear interrupt.
	FORCE_USE(limitval);
	Assert (! (*COUNTER0 & LIMIT_MASK));
	TimerManagerLock->release();


    /*
     * We get some timer interrupts before the Managers are completely
     *   initialized so need to make sure the pointers are non-null
     *   before calling the handlers
     */
    if (TimeoutManager != 0) TimeoutManager->interruptHandler();

    if (PeriodicManager != 0) PeriodicManager->interruptHandler();

    if (SS1TickInterruptFunc != 0) (*SS1TickInterruptFunc)();

    CPU * currentCPU = thisCPU();
    Assert (currentCPU != 0);
    TIMESLICETIMER *ts = currentCPU->timesliceTimer();
    Assert (ts != 0);
    ts->tick();

#ifdef SAMPLEPC
    char *address = getPC (stackPointer);
    if (pcSampleNext > pcSampleMax) pcSampleMax = pcSampleNext;
    pcSample [pcSampleNext++] = address;
    if (pcSampleNext >= pcSampleSize) pcSampleNext = 0;
#endif

    Debug << "TimerManagerInterrupt::basicRaise() returning \n" << eor;
}



SS1TimerManager::SS1TimerManager(PFI ptb, PFV ep)
	: TimerManager(ptb, ep)
{
    Debug << "SS1TimerManager::ctor()\n" << eor;
	static int intervalTimerStarted = 0;

	microsecondsPerTick = TimerManageruSecPerInterrupt;
	TimerManagerLock = new Lock;
	Assert( TimerManagerLock != 0 );


    unsigned int microTicks = TimerManageruSecPerInterrupt;
    Assert(microTicks > 0);

    /*
     * Initialize the hardware clock when the first SS1TimerManager 
     *   is constructed.
     */
    TimerManagerLock->acquire();

    if (!intervalTimerStarted)
    {
	Debug << this << ": initializing timer\n" << eor;
	intervalTimerStarted = 1;

	Debug << "Timer addr = " << asHex((int)COUNTER0) << "\n";
	Debug << "limit addr = " << asHex((int)LIMIT0) << "\n";
	Debug << "Timer= " << asHex(*COUNTER0) << "\n" << eor;
	Debug << "limit= " << asHex(*LIMIT0) << "\n" << eor;
	Debug << "intreg = " << asHex(*INTREG) << "\n" << eor;
	Debug << "intreg = " << asHex(*INTREG) << "\n" << eor;

	Debug << "Setting limit to " << 
		asHex((microTicks << COUNTER_SHIFT) & COUNTER_MASK) <<
		"\n" << eor;
	*LIMIT0 = (microTicks << COUNTER_SHIFT) & COUNTER_MASK;

	Assert(((*LIMIT0 & COUNTER_MASK) >> COUNTER_SHIFT) == microTicks);

	Debug << "Timer= " << asHex(*COUNTER0) << "\n" << eor;
	Debug << "Timer= " << asHex(*COUNTER0) << "\n" << eor;
	Debug << "limit= " << asHex(*LIMIT0) << "\n" << eor;

    }

    setFreeRunningTime(0);
    TimerManagerLock->release();

}
