#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1990,1991 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 *
 * Author: Deepu Thomas (dthomas1@cs.uiuc.edu) for Solaris 5.8
 *	   changed from NIT to DLPI, modified packet filter routines
 *	   and a whole bunch of C++ generic stuff - DCT 04/15/2004
 * Author: Theron Tock (theron@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 *	VCEthernet.h
 *
 *	Interface for the Virtual Choices Ethernet driver.
 *
 * Bugs:
 *    We receive our own broadcasts.  It would be fairly simple to
 *  filter out all broadcasts originating from the machine we are
 *  running on, but then we could only telnet to VChoices from 
 *  another machine.
 *
 */

#include	"VCEthernet.h"
#include	"VCInterrupts.h"
#include	"ThisCPU.h"
#include	"ThisProcess.h"
#include	"OutputStream.h"
#include	"Process.h"
#include	"AwaitedInterruptInterrupt.h"
#include	"NetworkBuffer.h"
#include	"NetworkAdministrator.h"
#include	"FreeRunningTimer.h"
#include	"FreeRunningTimerStar.h"
#include	"NameServer.h"
extern "C" {
#include 	<sys/filio.h>
#include	<sys/dlpi.h>
#include 	<sys/pfmod.h>
#include	<memory.h>
#include	<signal.h>
#include	<stdio.h>
#include	<fcntl.h>
#include	<malloc.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<sys/param.h>
#include	<sys/file.h>
#include	<sys/ioctl.h>
#include	<sys/time.h>
#include	<stropts.h>
#include	<sys/socket.h>
#include	<net/if.h>
#include	<netinet/in.h>
#include	<netinet/if_ether.h>
#include	<netdb.h>
#include	"dlwrap.h"

#if OLDNIT
int	putmsg(int fd, struct strbuf *ctl, struct strbuf *data, int flags);
int	setpgid(pid_t, pid_t);
int	ioctl(int, int, void *);
int	vfork();
int	socket(int, int, int);
int	gethostname(char *, int);
int	connect(int, struct sockaddr *, int);
#endif
void	bcopy(char *, char *, int);
int 	dlattachreq(int, u_long);
int 	dlokack(int , char*);
int 	dlbindreq(int , u_long , u_long , u_long , u_long , u_long);
int 	dlbindack(int , char*);
int 	dlpromisconreq(int, u_long);
int	dstrioctl(int, int, int, int, char*);
int 	dlphysaddrreq(int, u_long);
int 	dlphysaddrack(int, char*);
int 	dlenabmultireq(int , char*, int);
int 	dldisabmultireq(int , char*, int);

}

#ifdef __GNUC__
#pragma implementation
#endif


/* Bleach, we need a global variable to keep track of the driver */
static VCEthernet	*ethernet;


void
ethernetAddress(octet *address)
{
  ethernet->getEthernetAddress()->copyTo(address);
}



void
VCEthernet::printInfo()
{
  Console() << "VCEthernet(" << asAddress(this) << "):\n";
  Console() << "\tInterrupts: " << _interrupts << "\n";
  Console() << "\tInput Packets: " << _inputPackets << "\n";
  Console() << "\tOutput Packets: " << _outputPackets << "\n";
}



void
VCEthernet::interruptHandler()
{
  _interrupts++;

  for ( ; ; ) {
    NetworkBuffer *buf = NetworkDriver::getBuffer();

    Assert(buf->packet() != 0);
    Assert(buf->maxLength() > 60);

	/* Read the packet from the interface */
     _data.buf = (char *) buf->packet();
     _data.maxlen = buf->maxLength();
     _data.len  = 0;
     getmsg(_fd, NULL , &_data, &_flags);	
     
     int size = _data.len;
//   fprintf(stderr, "ETH driv interrupt pkt size: %d\n", size);
     EthernetAddress	to(buf->packet());
     EthernetAddress	from(buf->packet() + EthernetAddressLength);

    	/* If nothing was read, start reading from the pseudo-interface. */
    if (size <= 0 || (to.isBroadcast() && from == *_ethernetAddress))
    {
      int	bufferSize;

      ioctl(_serverSocket, FIONREAD, &size);
      if (size <= 0)
      {
	buf->release();
	return;
      }

      	/* First we read the size of the next buffer */
      read(_serverSocket, &bufferSize, sizeof(int));

      	/* Then we actually read the buffer */
      bufferSize = ntohs(bufferSize);
      if (bufferSize > buf->maxLength())
	bufferSize = buf->maxLength();

      size = read(_serverSocket, buf->packet(), bufferSize);
      Assert(size == bufferSize);
    }

	/* Filter out multicast packets we do not want */
    //EthernetAddress	to(buf->packet());
    if (!isValidAddress(&to))
      buf->release();
    else
    {
      buf->setTime(SystemTimer->time());
      buf->setLength(size);
      NetworkDriver::receivePacket(buf);
    }
  }
}


int
VCEthernet::transmitPacket(NetworkBuffer *buf)
{
  int		retVal;

  retVal = doTransmit(buf);
  thisProcess()->relinquishProcessor();	// simulate a synchronous send

  return(retVal);
}


int
VCEthernet::transmitPacketAsync(NetworkBuffer *buf)
{
  int		retVal;

  retVal = doTransmit(buf);
  buf->release();

  return(retVal);
}


int
VCEthernet::doTransmit(NetworkBuffer *buf)
{
  EthernetAddress	dest(buf->packet());
  struct sockaddr	address;
  struct strbuf		header, data;
  octet			*packet;
  int			packetSize;

  packet = buf->packet();

  	/* If we're sending to someone not on this machine, use Ethernet */
  if (dest != *_ethernetAddress)
  {

#ifdef OLDNIT
    	/* Set up the address */
    address.sa_family = AF_UNSPEC;
    memcpy(&address.sa_data, packet, 14);
    
    	/* Set up the header and body of the message */
    header.len = sizeof(address);
    header.buf = (char *) &address;
    data.len = buf->length() - 14;
    data.buf = (char *) &packet[14];
    
#endif
    	/* Transmit the packet on the interface */
    if (write(_fd, buf->packet() , buf->length() < 0))
	perror ("Ethernet write:") ;
  }

  	/* If it is the same machine or a multicast address, use the 
	   pseudo-interface.  Note that multicast packets are sent
	   on both interfaces */
  if (dest == *_ethernetAddress || dest.isMulticast() || dest.isBroadcast())
  {
    	/* Now transmit the packet on the pseudo-interface */
    packetSize = htons(buf->length());
    write(_serverSocket, &packetSize, sizeof(int));
    write(_serverSocket, packet, buf->length());
  }

  return(0);
}



void
VCEthernetInterruptProcessEntry()
{
  int		fd, server;
  VCEthernet	*ethernet;

  ethernet = (VCEthernet *) (ProxiableObject *)StandardNameServer->lookup("ethernetDriver", "VCEthernet");

	/* Get the file descriptors associated with the interfaces */
  ethernet->getDescriptors(&fd, &server);

  AwaitedInterruptInterrupt *interrupt = new AwaitedInterruptInterrupt();

	/* Add this descriptor to list that is checked for pending
	 * input when a SIGIO signal is received.
	 */
  VCIOInterrupt::setIOInterrupt(fd, interrupt);

  	/* Add the descriptor for the pseudo-interface. */
  VCIOInterrupt::setIOInterrupt(server, interrupt);

  for ( ; ; ) {
    interrupt->await();
//    fprintf(stderr, "An ethernet terrupt occured\n");
    ethernet->interruptHandler();
  }
}


void
startLocalServer(void)
{
  int			pid;
  int			ret;

  if ((pid = vfork()) < 0){
	perror("vfork");
	Assert (NOTREACHED) ("VCVirtualEthernetDaemon vfork failed!");
	}

  if (pid == 0)  {
	ret = execlp("VCVirtualEthernetDaemon", "VCVirtualEthernetDaemon",NULL);
	Assert (NOTREACHED) ("VCVirtualEthernetDaemon execve failed!");
	}

}



bool
VCEthernet::connectToLocalServer(void)
{
  struct sockaddr_in	address;
  struct hostent	*host;
  char			hostname[MAXHOSTNAMELEN];

  	/* Create a socket to connect to the server */
  _serverSocket = socket(AF_INET, SOCK_STREAM, 0);

  	/* Ick.  To connect to ourself, we have to look up
	   our address in the host table */
  gethostname(hostname, sizeof(hostname));
  host = gethostbyname(hostname);
  bcopy(host->h_addr_list[0], (char *) &address.sin_addr.s_addr, 
	sizeof(struct in_addr));
  address.sin_family = AF_INET;
  address.sin_port = htons(networkPort);

  	/* Try to connect */
  if ((connect(_serverSocket, (sockaddr *) &address, sizeof(address)) < 0))
  {

   fprintf(stderr, "We failed in the first attempt to connect\n");
    	/* If we don\'t succeed, start the server */
    //startLocalServer();

    int connected = 0;
    while (connected == 0) {
    	/* You can not connect twice to a socket, even if
	   the first attempt failed.  So create a new socket */
	    close(_serverSocket);
	    _serverSocket = socket(AF_INET, SOCK_STREAM, 0);
	    bcopy(host->h_addr_list[0], (char *) &address.sin_addr.s_addr, 
		sizeof(struct in_addr));
	    address.sin_family = AF_INET;
	    address.sin_port = htons(networkPort);

    	/* If we can't connect this time, something is wrong */
	    connected = (connect(_serverSocket, (sockaddr *) &address, sizeof(address)) == 0) ;
		if(!connected)
		{
		    fprintf(stderr, "That was our second attempt at a connection, which also failed\n");
			return false;
		}
	}
  }

  	/* Set socket so we receive SIGIO signals */
  fcntl(_serverSocket, F_SETFL, FASYNC);
  fcntl(_serverSocket, F_SETOWN, getpid());
  fprintf(stderr, "Connected to the local daemon\n");
  return true;
}



VCEthernet::VCEthernet(unsigned char debugMask, NetworkBufferFreeList* nf) : 
	EthernetDriver(debugMask, nf)
{
  struct strioctl	strioctl;
  struct ifreq		ifreq;
  struct packetfilt	filter;
  struct sockaddr	*sockaddr;
  u_short		*fwp;
  char			hostname[64];
  struct hostent	*hostaddr;
  short			*localAddr;
  long	                buf[MAXDLBUF];
  union   DL_primitives   *dlp;

  _interrupts = 0;
  _inputPackets = 0;
  _outputPackets = 0;

#ifdef DOESNT_WORK
  _multicastMutex = new MutexSemaphore;
  for (int i = 0; i < MulticastAddresses; i++)
    _multicastCount[i] = 0;
#endif

  _ready=false;

  	/* Start up a local process to act as a pseudo-broadcast
	   network.  This is necessary since the NIT interface does
	   not send packets addressed to the local machine */
  if(!connectToLocalServer())
  {
	  return;
  }

       /* Open the DLPI enabled Fast Ethernet driver directly */
  if ((_fd = open("/dev/eri", O_RDWR)) < 0)
    perror("/dev/hme");

	/*
	 * Attach.
	 */
   if (dlattachreq(_fd, 0) < 0)
	 	perror("dlattach:");
   if (dlokack(_fd, (char *)buf) < 0)
		perror("dlokack:");
	
	/*
	 * Bind for receiving only IP packets.
	 */
   if (dlbindreq(_fd, 0x0800, 0, DL_CLDLS, 0, 0) < 0)
		perror("dlbindreq:");
   if (dlbindack(_fd, (char*)buf) < 0)
		perror("dlbindack:"); 


	/* Enable promiscuous mode @DCT verify whether it's /really/ needed? */
    if (dlpromisconreq(_fd, DL_PROMISC_SAP) < 0)
		perror("dlpromisconreq:");
    if (dlokack(_fd, (char *)buf) < 0)
		perror("dlokack");

	/*
	 * Issue DLIOCRAW ioctl.
	 */
    if (dstrioctl(_fd, DLIOCRAW, -1, 0, NULL) < 0)
	perror("DLIOCRAW:");

	/* Set mode to message no-discard mode */
  if (ioctl(_fd, I_SRDOPT, (char *) RMSGN) < 0)
    perror("ioctl: discard");


  gethostname(hostname, sizeof(hostname));
  hostaddr = gethostbyname(hostname);
  localAddr = (short *) hostaddr->h_addr_list[0];

	/* Set up the packet filter.  We discard all IP packets
	 * that are destined for the UNIX machine we're running on.
	 */
  fwp = filter.Pf_Filter;

  *fwp++ = ENF_PUSHWORD + 6;		/* Push packet type */
  *fwp++ = ENF_PUSHLIT | ENF_CNAND;	/* Return true if not IP */
  *fwp++ = ETHERTYPE_IP;

  *fwp++ = ENF_PUSHWORD + 16;		/* Push second word of IP address */
  *fwp++ = ENF_PUSHLIT | ENF_CNOR;	/* Return false if for local host */
  *fwp++ = localAddr[1];

  filter.Pf_FilterLen = 6;

  strioctl.ic_cmd = PFIOCSETF;
  strioctl.ic_timout = 3;
  strioctl.ic_len = sizeof(filter);
  strioctl.ic_dp = (char *) &filter;

  if (ioctl(_fd, I_PUSH, "pfmod") < 0) 
    perror("push protocol filter");

  if (ioctl(_fd, I_STR, (char *) &strioctl) < 0)
    perror("set protocol filter");

#ifdef OLDNIT

	/* Now bind the stream to a particular interface.  In this
	 * case we are binding to the Lance ethernet controller.
	 */
  strcpy(ifreq.ifr_name, "le0");

  strioctl.ic_cmd = NIOCBIND;
  strioctl.ic_timout = 3;
  strioctl.ic_len = sizeof(struct ifreq);
  strioctl.ic_dp = (char *) &ifreq;
  if (ioctl(_fd, I_STR, (char *) &strioctl) < 0)
    perror("bind to interface");

#endif


	/* Get the ethernet address of the interface */

#ifdef OLDNIT
  if (ioctl(_fd, SIOCGIFADDR, &ifreq) < 0)
    perror("SIOCGIFADDR");

  sockaddr = (struct sockaddr *) ifreq.ifr_data;
  _ethernetAddress = new EthernetAddress((octet *) &sockaddr->sa_data);
#endif

  /*
   * Get our physical address.
   */   
  dlphysaddrreq(_fd, DL_CURR_PHYS_ADDR);
  dlphysaddrack(_fd, (char *)buf);
  dlp = (union DL_primitives*) buf;
  _ethernetAddress = new EthernetAddress((octet *)
	OFFADDR(dlp, dlp->physaddr_ack.dl_addr_offset));

  	/* We really shouldn\'t do this */
  ethernet = this;

  StandardNameServer->bind("ethernetDriver", this);

  Process * p = new InterruptProcess ( VCEthernetInterruptProcessEntry,
				      thisProcess()->domain(),
				      normal, 0);
  Assert (p != 0);
  p->setName("#VCEthernetInterruptProcess");
  p->ready();
  //thisProcess()->relinquishProcessor(); // get it running

	/* Flush the stream, ask to receive SIGIO signals,
	 * and set descriptor to non-blocking mode.
	 */
  ioctl(_fd, I_FLUSH, (char *) FLUSHR);
  fcntl(_fd, F_SETFL, O_NDELAY);
  ioctl(_fd, I_SETSIG, (char *) S_INPUT);
  fprintf(stderr, "Finished constructor of VCEthernet\n");
  _ready=true;
}



VCEthernet::~VCEthernet()
{
  close(_fd);
  close(_serverSocket);

#ifdef DOESNT_WORK
  delete _multicastMutex;
#endif
}



int
VCEthernet::addMulticast(EthernetAddress *address)
{
  struct ifreq		ifreq;
  struct strioctl	strioctl;
  int			err;

  if (err = registerMulticast(address))
    return(err);

#ifdef DOESNT_WORK
  /* Because SUNOS accepts at most 64 multicasts, we must make sure
     that we do not use two addresses that map to the same bit in
     the filter so that we never exceed the limit */
  int index = hashAddressTo64(address);

  _multicastMutex->P();
  if (!_multicastCount[index]++)
  {
    /* Save the address for later removal, since SIOCDELMULTI expects the
       same address that was used with SIOCADDMULTI */
    _multicastAddress[index] = *address;
#endif

#ifdef OLDNIT

    strcpy(ifreq.ifr_name, "le0");
    address->copyTo((octet *) &ifreq.ifr_addr.sa_data);

    strioctl.ic_cmd = SIOCADDMULTI;
    strioctl.ic_timout = 3;
    strioctl.ic_len = sizeof(ifreq);
    strioctl.ic_dp = (char *) &ifreq;

    if ((err = ioctl(_fd, I_STR, (char *) &strioctl)) < 0)
      perror("add multicast");

#endif
		dlenabmultireq(_fd, (char *) address->bytes(), 6);




#ifdef DOESNT_WORK
  }

  _multicastMutex->V();
#endif

  return(err);
}



int
VCEthernet::removeMulticast(EthernetAddress *address)
{
  struct ifreq	ifreq;
  int		err;

  /* Notify the caller if it was not properly registered */
  if (err = unregisterMulticast(address))
    return(err);

#ifdef DOESNT_WORK
  int index = hashAddressTo64(address);
  Assert(_multicastCount[index] > 0);

  _multicastMutex->P();

  _multicastCount[index]--;
  if (_multicastCount[index] == 0)
  {
    _multicastAddress[index].copyTo((octet *) &ifreq.ifr_addr.sa_data);
#else
    address->copyTo((octet *) &ifreq.ifr_addr.sa_data);
#endif

#ifdef OLDNIT

    strcpy(ifreq.ifr_name, "le0");
    if ((err = ioctl(_fd, SIOCDELMULTI, (char *) &ifreq)) < 0)
      perror("remove multicast");

#endif


		dldisabmultireq(_fd, (char *) address->bytes(), 6);

#ifdef DOESNT_WORK
  }

  _multicastMutex->V();
#endif

  return(err);
}


#include "ClassInstaller.h"
DEFINECLASS("VCEthernet",VCEthernet,EthernetDriver,0,0);
