#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * Cache.cc
 */

#include "Assert.h"
#include "Barrier.h"
#include "Byte.h"
#include "Cache.h"
#include "CacheAsync.h"
#include "Domain.h"
#include "Kernel.h"
#include "Log2.h"
#include "MemoryObject.h"
#include "MemoryObjectCache.h"
#include "MemorySimpleInstruments.h"
#include "PhysicallyAddressableUnit.h"
#include "Process.h"
#include "Raid.h"

#ifdef __GNUC__
#pragma implementation
#endif

Lock Cache::_lock;
CacheStatistics Cache::_stats;

CacheTimer CacheCtor("Cache::Cache() timer");

Cache::Cache( MemoryObjectCache * moc )
    : _moc( moc ), _parent( 0 ), _child( 0 ), _numberOfUnits( 0 ), 
      _dirty( 0 ), _cachedUnits( 0 )
{
    SInstr( int istart = CacheCtor.start() );
    initDebug();

    Raids( RaidConstructor ) << this << "::Cache( " 
                             << moc << " )\n" << eor;

    _log2UnitSize = moc->log2UnitSize();

    _mo = moc->memoryObject();
    _log2SegmentSize = log2Low( _mo->segmentSize() );
    Assert( _mo != 0 );
    _log2MoUnitSize = _mo->log2UnitSize();
    int log2MoUnitSizeRatio = _log2UnitSize - _log2MoUnitSize;
    Assert( log2MoUnitSizeRatio >= 0 );
    _unitInfoTable = 0;
    SInstr( CacheCtor.stop( istart ) );
}

CacheTimer CacheDtor("Cache::~Cache() timer");

Cache::~Cache()
{
    SInstr( int istart = CacheDtor.start() );
    if ( _child ) 
        delete _child;
    Raids(64) << this << "::~Cache()" << "\n" << eor;
    SInstr( CacheDtor.stop( istart ) );
}

int Cache::insertAbove( Cache * cache )
{
    Raids( RaidMethod ) 
    << this << "::insertAbove( " << cache << " )\n" << eor;

    Assert( _parent == 0 );
    Assert( cache->child() == 0 );
    Assert( cache->parent() == 0 );
    _parent = cache;
    _parent->setChild( this );
    cache->setNumberOfUnits( _numberOfUnits );
    return 0;
}

void Cache::basicAllocate(unsigned  int count )
{
    /*
     * do nothing, should be re-defined by concrete class
     */
}

void Cache::basicFreeUnits( unsigned int startUnit,unsigned  int count )
{
    /*
     * do nothing, should be re-defined by concrete class to
     * free space occupied by unit
     */
}

int Cache::exists( unsigned unit )
{
    Raids( RaidMethod )
    << this << "::exists( " << unit << " )\n" << eor;

    acquire();
    int valid = unitInfo( unit ).valid();
    release();  
    if ( valid )
        return 1;
    else 
        return _child->exists( unit );
}

CacheTimer CacheSetNumberOfUnits("Cache::setNumberOfUnits() timer");

unsigned int Cache::setNumberOfUnits( unsigned int noOfUnits )
{
    SInstr( int istart = CacheSetNumberOfUnits.start() );
    Raids( RaidMethod ) 
    << this << "::setNumberOfUnits( " << noOfUnits << " ) " 
    << _numberOfUnits << "\n" << eor;

    if ( noOfUnits != _numberOfUnits ) {
        if ( noOfUnits < _numberOfUnits ) {
            invalidate( noOfUnits, _numberOfUnits - noOfUnits, 0 );
        }
        _numberOfUnits = basicSetNumberOfUnits( noOfUnits );
        Assert( _numberOfUnits == noOfUnits );  
        noOfUnits = _child->setNumberOfUnits( noOfUnits );
        Assert( _numberOfUnits == noOfUnits );
    }
    SInstr( CacheSetNumberOfUnits.stop( istart ) );
    return noOfUnits;
}

unsigned int Cache::basicSetNumberOfUnits( unsigned int noOfUnits )
{
    Assert( NOTREACHED );
    return 0;
}

CacheTimer CacheRead("Cache::read() timer");

int Cache::read( unsigned int startUnit,unsigned  int count, char * addr, IOAsyncClosure * c )
{
    SInstr( int istart = CacheRead.start() );
    Raids( RaidMethod ) << this << "::read( " << startUnit << ", " 
                        << count << ", " << asAddress( addr ) << ", " 
                        << asAddress( c ) << " )\n" << eor;

    Assert( count > 0 );
    Assert( addr != 0 );
    Debug<<this<<"::read("<<startUnit<<", "
         << count << ", " << asAddress( addr ) << ", "
         << asAddress( c ) << " )\n" << eor;

    int unit = startUnit;
    int endUnit = startUnit + count - 1;

    unsigned total = 0;
    int firstUnit;
    int noOfUnits;

    CacheIOState * state = 0;
    unsigned firstStartUnit = 0;
    unsigned firstCount = 0;
    char * firstBuf = 0;
    int firstType = 0;

    acquire();

    _stats._read.add( count );

    while ( unit <= endUnit ) {

        /* 
         * try to read from my cache if possible
         */

        firstUnit = unit;
        while( unit <= endUnit && unitInfo( unit ).valid() )
            unit++;
        noOfUnits = unit - firstUnit;
        if ( noOfUnits > 0 ) {

            Raids( RaidDetail ) << this << "::read "
                                << firstUnit << " to " << unit - 1
                                << " already cached\n" << eor;

            if ( firstBuf == 0 ) {
                firstStartUnit = firstUnit;
                firstCount = noOfUnits;
                firstBuf = addr;
                firstType = 0;
            } else {
                if ( state == 0 ) {
                    state = new CacheIOState( c );
                    state->add( firstStartUnit,
                                firstCount, firstBuf,
                                firstType );
                }
                state->add( firstUnit, noOfUnits, addr, 0 );
            }

            addr += ( noOfUnits << _log2UnitSize ) ;
            if ( unit > endUnit ) 
                break;
        }

        /*
         * try to read from child if there is child else read
         * from mo
         */

        firstUnit = unit;
        while ( unit <= endUnit && ! unitInfo( unit ).valid() ) 
            unit++;
        noOfUnits = unit - firstUnit;
        if ( noOfUnits > 0 ) {
            Raids( RaidDetail ) << this << "::read "
                                << firstUnit << " to " << unit - 1
                                << " being read from child/mo\n" << eor;

            if ( firstBuf == 0 ) {
                firstStartUnit = firstUnit;
                firstCount = noOfUnits;
                firstBuf = addr;
                firstType = 1;
            } else {
                if ( state == 0 ) {
                    state = new CacheIOState( c );
                    state->add( firstStartUnit,
                                firstCount, firstBuf,
                                firstType );
                }
                state->add( firstUnit, noOfUnits, addr, 1 );
            }
            addr += ( noOfUnits << _log2UnitSize );
            if ( unit > endUnit )
                break;
        }
    }

    release();
    
    if ( state ) {
        state->_result = 0;
        state->_expectedResult = count;
        state->_spare = 33; // ?? //(unsigned) &(Cache::read);
        CacheIOState::_Operation * op = state->_ops;
        int opCount = state->_count;
        for( int i = 0; i < opCount; i++ ) {
            if ( op->_spare == 0 ) 
                basicRead( op->_startUnit, op->_count, 
                           op->_buf, &(op->_closure) );
            else 
                _child->read( op->_startUnit, op->_count,
                              op->_buf, &(op->_closure) );
            op++;
        }
        if ( c == 0 ) {
            state->_barrier.wait();
            total = state->_result;
            delete state;
        }
    } else if ( firstBuf ) {
        if ( c != 0 && total > 0 ) {
            c = new SumResultIOAsyncClosure( c, total );
        }
        if ( firstType == 0 ) 
            total += basicRead( firstStartUnit, firstCount, 
                                firstBuf, c );
        else 
            total += _child->read( firstStartUnit, firstCount, 
                                   firstBuf, c );
    } else {
        if ( c != 0 ) {
            c->activate( total );
        }
    }

    Assert( c != 0 || total == count );

    SInstr( CacheRead.stop( istart ) );
    return total;
}

CacheTimer CacheWrite("Cache::write() timer");

int Cache::write( unsigned int startUnit, unsigned int count, CacheUnitInfo * info, char * addr,
                  IOAsyncClosure * c )
{
    SInstr( int istart = CacheWrite.start() );
    Raids( RaidMethod ) << this << "::write( " << startUnit << ", " 
                        << count << ", " << asAddress( info ) << ", " 
                        << asAddress( addr ) << ", " << asAddress( c )
                        << " )\n" << eor;

    Assert( count > 0 );
    Assert( addr != 0 );

    int unit = startUnit;
    int endUnit = startUnit + count - 1;

    unsigned int total = 0;
    unsigned int firstUnit;
    unsigned int noOfUnits;
    unsigned int newUnits = 0;

    CacheIOState * state = 0;
    unsigned firstStartUnit = 0;
    unsigned firstCount = 0;
    char * firstBuf = 0;

    acquire();

    _stats._write.add( count );

    while ( unit <= endUnit ) {

        /* 
         * skip those units that are already cached if they have
         * not been modified 
         */

        firstUnit = unit;
        while( unit <= endUnit && unitInfo( unit ).valid() &&
               !info->modified() ) {
            unit++;
            info++;
        }
        noOfUnits = unit - firstUnit;
        if ( noOfUnits > 0 ) {

            Raids( RaidDetail ) << this << "::write "
                                << firstUnit << " to " << unit - 1
                                << " already cached\n" << eor;

            total += noOfUnits;
            addr += ( noOfUnits << _log2UnitSize );
            if ( unit > endUnit )
                break;
        }

        /*
         * cache units that were not previously cached
         */

        firstUnit = unit;
        CacheUnitInfo cui = unitInfo( unit );
        while( unit <= endUnit &&
               ( !cui.valid() || info->modified() ) ) {
            if ( !cui.valid() ) 
                newUnits++;
            if ( !cui.modified() && info->modified() )
                _dirty++;
            setUnitInfo( unit, *info );
            // Assert( unitInfo( unit ).info() == info->info() );
            unit++;
            info++;
            cui = unitInfo( unit );
        }
        noOfUnits = unit - firstUnit;
        if ( noOfUnits > 0 ) {

            Assert( _dirty <= _numberOfUnits );

            Raids( RaidDetail ) << this << "::write "
                                << firstUnit << " to " << unit - 1
                                << " being cached\n" << eor;

            if ( firstBuf == 0 ) {
                firstStartUnit = firstUnit;
                firstCount = noOfUnits;
                firstBuf = addr;
            } else {
                if ( state == 0 ) {
                    state = new CacheIOState( c );
                    state->add( firstStartUnit,
                                firstCount, firstBuf );
                }
                state->add( firstUnit, noOfUnits, addr );
            }
            addr += ( noOfUnits << _log2UnitSize );
            if ( unit > endUnit ) 
                break;
        }
    }

    if ( newUnits ) {   
        _cachedUnits += newUnits; 
        Assert( _cachedUnits <= _numberOfUnits );
        Assert( _dirty <= _cachedUnits ); 
        basicAllocate( newUnits );
    }

    release();

    if ( state ) {
        state->_result = total;
        state->_spare = 43; // ??? (unsigned) &(Cache::write);
        state->_expectedResult = count;
        CacheIOState::_Operation * op = state->_ops;
        int opCount = state->_count;
        for( int i = 0; i < opCount; i++ ) {
            _child->invalidate( op->_startUnit, op->_count, 1 );
            basicWrite( op->_startUnit, op->_count, 
                        op->_buf, &(op->_closure) );
            op++;
        }
        if ( c == 0 ) {
            state->_barrier.wait();
            total = state->_result;
            delete state;
        }
    } else if ( firstBuf ) {
        if ( c != 0 && total > 0 ) {
            c = new SumResultIOAsyncClosure( c, total );
        }
        _child->invalidate( firstStartUnit, firstCount, 1 );
        total += basicWrite( firstStartUnit, firstCount, firstBuf, c );
    } else {
        if ( c != 0 ) {
            c->activate( total );
        }
    }


    Assert( c != 0 || total == count );

    SInstr( CacheWrite.stop( istart ) );
    return total;
}

CacheTimer CacheRawWrite("Cache::rawWrite() timer");

int 
Cache::rawWrite( unsigned int startUnit, unsigned int count, CacheUnitInfo * info, char * addr,
                 IOAsyncClosure * c )
{
    SInstr( int istart = CacheRawWrite.start() );
    Raids( RaidMethod ) << this << "::rawWrite( " << startUnit << ", " 
                        << count << ", " << asAddress( info ) << ", " 
                        << asAddress( addr ) << ", " << asAddress( c )
                        << " )\n" << eor;
           
    Assert( count > 0 );
    Assert( addr != 0 );


    int unit = startUnit;
    int endUnit = startUnit + count - 1;

    int firstUnit;
    int noOfUnits;
    int newUnits = 0;

    CacheUnitInfo * origInfo = info;
    CacheIOState * state = 0;
    unsigned firstStartUnit = 0;
    unsigned firstCount = 0;
    char * firstBuf = 0;
    unsigned firstSpare = 0;
    int moWriteCount = 0;
    int basicWriteCount = 0;
    CacheUnitInfo valid( CUIValid );

    acquire();

    _stats._rawWrite.add( count );

    /*
     * first pass:
     *  find the units that I need to write out to memory object
     *  using moWrite, spare is set to 0
     *  these are units are:
     *  1. dirty from memory (i.e. memory object cache ) OR
     *  2. dirty from my cache
     *     (avoid having to read it into temp buffer later during sync)
     */

    while ( unit <= endUnit ) {

        /* 
         * skip those units that are not dirty in memory and
         * not dirty in cache
         */

        firstUnit = unit;
        while( unit <= endUnit &&
               !info->modified() && !unitInfo( unit ).modified() ) {
            unit++;
            info++;
        }
        noOfUnits = unit - firstUnit;
        if ( noOfUnits > 0 ) {
            Raids( RaidDetail ) << this << "::rawWrite "
                                << firstUnit << " to " << unit - 1
                                << " need not write to mo\n" << eor;
            addr += ( noOfUnits << _log2UnitSize );
            if ( unit > endUnit )
                break;
        }

        /*
         * find dirty units in memory or dirty units in cache
         */

        firstUnit = unit;
        while( unit <= endUnit &&
               ( info->modified() || unitInfo( unit ).modified() ) ) {
            unit++;
            info++;
        }
        noOfUnits = unit - firstUnit;
        if ( noOfUnits > 0 ) {
            Raids( RaidDetail ) << this << "::rawWrite "
                                << firstUnit << " to " << unit - 1
                                << " need to write to mo\n" << eor;
            moWriteCount += noOfUnits;
            if ( firstBuf == 0 ) {
                firstStartUnit = firstUnit;
                firstCount = noOfUnits;
                firstBuf = addr;
                firstSpare = 0;
            } else {
                if ( state == 0 ) {
                    state = new CacheIOState( c );
                    state->add( firstStartUnit,
                                firstCount, firstBuf,
                                firstSpare );
                }
                state->add( firstUnit, noOfUnits, addr, 0 );
            }
            addr += ( noOfUnits << _log2UnitSize );
            if ( unit > endUnit ) 
                break;
        }
    }

    /*
     * second pass:
     *  find the units that I need to write out to my cache
     *  using basicWrite, spare is set to 1
     *  these are units are:
     *  1. dirty from memory (i.e. memory object cache ) OR
     *  2. not already cached
     *  in this pass, also update cache unit info, cached units
     *  which used to be modified are no longer modified because
     *  modified cache units will be written to mo and hence become
     *  clean
     */

    unit = startUnit;
    info = origInfo;

    while ( unit <= endUnit ) {

        /* 
         * skip those units that are already cached if they have
         * not been modified 
         */

        firstUnit = unit;
        CacheUnitInfo cui = unitInfo( unit );
        while( unit <= endUnit && cui.valid() && !info->modified() ) {
            if ( cui.modified() ) {
                setUnitInfo( unit, valid );
                _dirty--;
            }
            unit++;
            info++;
            cui = unitInfo( unit );
        }
        noOfUnits = unit - firstUnit;
        if ( noOfUnits > 0 ) {
            Raids( RaidDetail ) << this << "::rawWrite "
                                << firstUnit << " to " << unit - 1
                                << " already cached\n" << eor;
            addr += ( noOfUnits << _log2UnitSize );
            if ( unit > endUnit )
                break;
        }

        /*
         * cache units that were not previously cached
         */

        firstUnit = unit;
        while( unit <= endUnit &&
               ( !cui.valid() || info->modified() ) ) {
            if ( !cui.valid() ) 
                newUnits++;
            if ( cui.modified() ) 
                _dirty--;
            setUnitInfo( unit, valid );
            // Assert( unitInfo( unit ).info() == info->info() );
            unit++;
            info++;
            cui = unitInfo( unit );
        }
        noOfUnits = unit - firstUnit;
        if ( noOfUnits > 0 ) {
            Raids( RaidDetail ) << this << "::rawWrite "
                                << firstUnit << " to " << unit - 1
                                << " being cached\n" << eor;
            basicWriteCount += noOfUnits;
            if ( firstBuf == 0 ) {
                firstStartUnit = firstUnit;
                firstCount = noOfUnits;
                firstBuf = addr;
                firstSpare = 1;
            } else {
                if ( state == 0 ) {
                    state = new CacheIOState( c );
                    state->add( firstStartUnit,
                                firstCount, firstBuf, 
                                firstSpare );
                }
                state->add( firstUnit, noOfUnits, addr, 1 );
            }
            addr += ( noOfUnits << _log2UnitSize );
            if ( unit > endUnit ) 
                break;
        }

    }

    if ( newUnits ) {   
        _cachedUnits += newUnits; 
        Assert( _cachedUnits <= _numberOfUnits );
        Assert( _dirty <= _cachedUnits ); 
        basicAllocate( newUnits );
    }

    release();
    
    unsigned  total = count - moWriteCount - basicWriteCount;
    if ( state ) {
        state->_result = total;
        state->_spare = 67; // ? (unsigned) &(Cache::rawWrite);
        state->_expectedResult = count;
        CacheIOState::_Operation * op = state->_ops;
        int opCount = state->_count;
        for( int i = 0; i < opCount; i++ ) {
            if ( op->_spare == 0 ) {
                moWrite( op->_startUnit, op->_count,
                         op->_buf, 1, &(op->_closure) );
            } else {
                _child->invalidate( op->_startUnit, 
                                    op->_count, 1 );
                basicWrite( op->_startUnit, op->_count, 
                            op->_buf, &(op->_closure) );
            }
            op++;
        }
        if ( c == 0 ) {
            state->_barrier.wait();
            total = state->_result;
            delete state;
        }
    } else if ( firstBuf ) {
        Assert( total >= 0 );
        if ( c != 0 && total > 0 ) {
            c = new SumResultIOAsyncClosure( c, total );
        }
        if ( firstSpare == 0 ) {
            total += moWrite( firstStartUnit, firstCount, 
                              firstBuf, 1, c );
        } else {
            _child->invalidate( firstStartUnit, firstCount, 1 );
            total += basicWrite( firstStartUnit, firstCount, 
                                 firstBuf, c );
        }
    } else {
        if ( c != 0 ) {
            c->activate( total );
        }
    }

    Assert( c != 0 || total == count );

    SInstr( CacheRawWrite.stop( istart ) );
    return total;
}

CacheTimer CacheWriteMOModified("Cache::writeMOModified() timer");

int Cache::writeMOModified(unsigned  int startUnit, unsigned int count, CacheUnitInfo * info, 
                           char * addr, int writeThrough, IOAsyncClosure * c )
{
    SInstr( int istart = CacheWriteMOModified.start() );
    Raids( RaidMethod ) << this << "::writeMOModified( " 
                        << startUnit << ", " 
                        << count << ", " << asAddress( info ) << ", " 
                        << asAddress( addr ) << ", " << writeThrough << ", "
                        << asAddress( c ) << " )\n" << eor;

    Assert( count > 0 );
    Assert( addr != 0 );

    int unit = startUnit;
    int endUnit = startUnit + count - 1;

    unsigned total = 0;
    int firstUnit;
    int noOfUnits;

    CacheIOState * state = 0;
    unsigned firstStartUnit = 0;
    unsigned firstCount = 0;
    char * firstBuf = 0;

    while ( unit <= endUnit ) {

        /* 
         * skip those units that are not modified
         */

        firstUnit = unit;
        while( unit <= endUnit && !info->modified() ) {
            unit++;
            info++;
        }
        noOfUnits = unit - firstUnit;
        if ( noOfUnits > 0 ) {

            Raids( RaidDetail ) << this << "::writeMOModified "
                                << firstUnit << " to " << unit - 1
                                << " not modified\n" << eor;

            addr += ( noOfUnits << _log2UnitSize );
            total += noOfUnits;
            if ( unit > endUnit )
                break;
        }

        /*
         * write out units that has been modified
         */

        firstUnit = unit;
        while( unit <= endUnit && info->modified() ) {
            Assert( info->valid() );
            unit++;
            info++;
        }
        noOfUnits = unit - firstUnit;
        if ( noOfUnits > 0 ) {

            Raids( RaidDetail ) << this << "::writeMOModified "
                                << firstUnit << " to " << unit - 1
                                << " write to mo\n" << eor;

            if ( firstBuf == 0 ) {
                firstStartUnit = firstUnit;
                firstCount = noOfUnits;
                firstBuf = addr;
            } else {
                if ( state == 0 ) {
                    state = new CacheIOState( c );
                    state->add( firstStartUnit,
                                firstCount, firstBuf );
                }
                state->add( firstUnit, noOfUnits, addr );
            }
            addr += ( noOfUnits << _log2UnitSize );
            if ( unit > endUnit ) 
                break;
        }
    }

    if ( state ) {
        state->_result = total;
        state->_expectedResult = count;
        state->_spare = 56; // ?? (unsigned) &(Cache::writeMOModified);
        CacheIOState::_Operation * op = state->_ops;
        int opCount = state->_count;
        for( int i = 0; i < opCount; i++ ) {
            moWrite( op->_startUnit, op->_count, 
                     op->_buf, writeThrough, &(op->_closure) );
            op++;
        }
        if ( c == 0 ) {
            state->_barrier.wait();
            total = state->_result;
            delete state;
        }
    } else if ( firstBuf ) {
        if ( c != 0 && total > 0 ) {
            c = new SumResultIOAsyncClosure( c, total );
        }
        total += moWrite( firstStartUnit, firstCount, firstBuf, 
                          writeThrough, c );
    } else {
        if ( c != 0 ) {
            c->activate( total );
        }
    }

    Assert( c != 0 || total == count );

    SInstr( CacheWriteMOModified.stop( istart ) );
    return count;
}

class CacheMOReadAsyncClosure : public IOAsyncClosure {
    unsigned _expected;
    unsigned _maxExpected;
    unsigned _log2MoUnitSize;
    unsigned _log2MoUnitSizeRatio;
    char * _buf;
    IOAsyncClosure * _parent;
public:
    static void func( CacheMOReadAsyncClosure * c );
    CacheMOReadAsyncClosure
    ( IOAsyncClosure * parent, unsigned log2MoUnitSizeRatio,
      unsigned log2MoUnitSize,
      unsigned expected, unsigned maxExpected, char * buf )
        : _expected( expected ), _maxExpected( maxExpected ),
          _log2MoUnitSize( log2MoUnitSize ),
          _log2MoUnitSizeRatio( log2MoUnitSizeRatio ),
          _buf( buf ),
          _parent( parent ) 
    { _func  =(AsyncFunc)& func ;}
};

void
CacheMOReadAsyncClosure::func( CacheMOReadAsyncClosure * c )
{
    Debug<<"CacheMOReadAsyncClosure::func()\n"<<eor;
    unsigned noOfUnits = c->_result;

    if ( noOfUnits < c->_maxExpected )
        ByteZero( c->_buf + ( noOfUnits << c->_log2MoUnitSize ),
                  ( c->_maxExpected - noOfUnits ) << 
                  c->_log2MoUnitSize );
    if ( noOfUnits != c->_expected ) {
        Console()<< "CacheMOReadAsyncClosure::func() " 
                 << noOfUnits << " != " << c->_expected << "\n" << eor;
    }
    if ( noOfUnits > 0 )
        noOfUnits = 
        ( ( noOfUnits - 1 ) >> c->_log2MoUnitSizeRatio ) + 1; 
    if ( c->_parent )
    {
        Debug<<"activating parent closure (nu="<<noOfUnits<<")\n"<<eor;
        c->_parent->activate( noOfUnits );
    }
    delete c;
}


CacheTimer CacheMORead("Cache::moRead() timer");

int Cache::moRead( unsigned int startUnit, unsigned int count, char * addr, IOAsyncClosure * c )
{
    SInstr( int istart = CacheMORead.start() );
    Raids( RaidProtected ) << this << "::moRead( " << startUnit << ", " 
                           << count << ", " << asAddress( addr ) << ", "
                           << asAddress( c ) << " )\n" << eor;

    Debug<< this << "::moRead( " << startUnit << ", " 
         << count << ", " << asAddress( addr ) << ", "
         << asAddress( c ) << " )\n" << eor;

    
    Assert( addr != 0 );
    Assert( count > 0 );

    unsigned log2MoUnitSizeRatio = _log2UnitSize - _log2MoUnitSize;
    unsigned maxCount = count << log2MoUnitSizeRatio;
    unsigned moStartUnit = startUnit << log2MoUnitSizeRatio;
    unsigned moNumberOfUnits = _mo->numberOfUnits();
    unsigned moCount = maxCount;
    if ( moStartUnit + maxCount > moNumberOfUnits )
        moCount = moNumberOfUnits - moStartUnit;
    Assert( moCount >= 0 );

    unsigned  noOfUnits;

    _stats._moRead.add( count );

    if ( c == 0 ) {
        noOfUnits = _mo->rawRead( moStartUnit, moCount, addr, 
                                  moCount << _log2MoUnitSize, 0 );
        if ( noOfUnits < maxCount ) 
            ByteZero( addr + ( noOfUnits << _log2MoUnitSize ),
                      ( maxCount - noOfUnits ) << _log2MoUnitSize );
        if ( noOfUnits > 0 )
            noOfUnits = 
            ( (noOfUnits - 1) >> log2MoUnitSizeRatio ) + 1;
        if ( noOfUnits != count ) {
            Console() << this << "::moRead( "
                      << startUnit << ", "
                      << count << ", " << asAddress( addr )
                      << " ) unexpected read error\n" << eor;
        }
    } else {
        noOfUnits = moCount;
        CacheMOReadAsyncClosure * myClosure
        = new CacheMOReadAsyncClosure
        ( c, log2MoUnitSizeRatio, _log2MoUnitSize,
          moCount, maxCount, addr );
        Debug<<"_mo="<<_mo<<"\n"<<eor;
        _mo->rawRead( moStartUnit, moCount, addr, 
                      moCount << _log2MoUnitSize, myClosure );
    }

    SInstr( CacheMORead.stop( istart ) );
    return noOfUnits;
}

CacheTimer CacheMOWrite("Cache::moWrite() timer");

int Cache::moWrite( unsigned int startUnit, unsigned int count, char * addr, 
                    int writeThrough, IOAsyncClosure * c )
{
    SInstr( int istart = CacheMOWrite.start() );

    Raids( RaidProtected ) << this << "::moWrite( " 
                           << startUnit << ", " 
                           << count << ", " << asAddress( addr ) << ", "
                           << asAddress( c ) 
                           << " )\n" << eor;

    Assert( addr != 0 );
    Assert( count > 0 );

    int log2MoUnitSizeRatio = _log2UnitSize - _log2MoUnitSize;
    int moCount = count << log2MoUnitSizeRatio;
    int moStartUnit = startUnit << log2MoUnitSizeRatio;
    int moNumberOfUnits = _mo->numberOfUnits();
    if ( moStartUnit + moCount > moNumberOfUnits )
        moCount = moNumberOfUnits - moStartUnit;
    Assert( moCount >= 0 );

    unsigned  noOfUnits;

    _stats._moWrite.add( count );

    if ( c == 0 ) {
        noOfUnits = _mo->rawWrite( moStartUnit, moCount, addr, 
                                   moCount << _log2MoUnitSize, 
                                   writeThrough, 0 );
        if ( noOfUnits > 0 )
            noOfUnits = 
            ( (noOfUnits - 1) >> log2MoUnitSizeRatio ) + 1;
        if ( noOfUnits != count ) {
            Console() << this << "::moWrite( " 
                      << startUnit << ", "
                      << count << ", " << asAddress( addr )
                      << " ) unexpected write error\n" << eor;
        }
    } else {
        noOfUnits = moCount;
        CacheMOIOClosure * myClosure = 
        new CacheMOIOClosure( c, log2MoUnitSizeRatio, moCount );
        _mo->rawWrite( moStartUnit, moCount, addr, 
                       moCount << _log2MoUnitSize, 
                       writeThrough, myClosure );
    }

    SInstr( CacheMOWrite.stop( istart ) );
    return noOfUnits;
}

CacheTimer CacheInvalidate("Cache::Invalidate() timer");

void Cache::invalidate(unsigned  int startUnit, unsigned int count, int propagate )
{
    SInstr( int istart = CacheInvalidate.start() );
    Raids( RaidMethod ) << this << "::invalidate( " << startUnit << ", "
                        << count << " )\n" << eor;

    int endUnit = startUnit + count - 1;
    CacheUnitInfo invalid( CUIInvalid );
    
    int unit = startUnit;
    int noOfUnits;
    acquire();
    while ( unit <= endUnit ) {
        int firstUnit = unit;
        CacheUnitInfo cui = unitInfo( unit );
        while ( unit <= endUnit && cui.valid() ) {
            if ( cui.modified() )
                _dirty--;
            setUnitInfo( unit, invalid );
            // Assert( unitInfo( unit ).info() == invalid.info() );
            unit++;
            cui = unitInfo( unit );
        }
        noOfUnits = unit - firstUnit;
        if ( noOfUnits > 0 ) {
            _cachedUnits -= noOfUnits;
            Assert( _cachedUnits >= 0 && 
                    _cachedUnits <= _numberOfUnits );
            Assert( _dirty <= _cachedUnits);
            basicFreeUnits( firstUnit, noOfUnits );
        }
        do {
            unit++;
        } while ( unit <= endUnit && !unitInfo( unit ).valid() );
    }
    release();
    if ( propagate )
        _child->invalidate( startUnit, count, propagate ); 
    SInstr( CacheInvalidate.stop( istart ) );
}

CacheTimer CacheSynchronize("Cache::Synchronize() timer");

void Cache::synchronize(unsigned  int startUnit, unsigned int count, unsigned int & log2BufSize, 
                        char * & buffer )
{
    SInstr( int istart = CacheSynchronize.start() );
    Raids( RaidMethod ) << this << "::synchronize( " << startUnit << ", "
                        << count << ", " << log2BufSize << ", " 
                        << asHex( ( unsigned ) buffer ) << " )\n" << eor;
            
    if ( dirty() == 0 ) {
        SInstr( CacheSynchronize.stop( istart ) );
        return;
    }

    if ( log2BufSize == 0 ) {
        log2BufSize = _log2SegmentSize;
    }
    int log2BufSizeRatio = log2BufSize - _log2UnitSize;
    Assert( log2BufSizeRatio >= 0 );
    int maxUnits = 1 << log2BufSizeRatio;

    char * orgBuffer = buffer;

    int unit = startUnit;
    int endUnit = startUnit + count - 1;

    int firstUnit;
    int noOfUnits;
    int unitCount;

    acquire();
    while ( unit <= endUnit ) {

        /* 
         * skip those units which are not valid or not modified
         */

        firstUnit = unit;
        
        while( unit <= endUnit && !unitInfo( unit ).modified() ) {
            unit++;
        }
        noOfUnits = unit - firstUnit;
        if ( noOfUnits > 0 ) {

            Raids( RaidDetail ) << this << "::synchronize "
                                << firstUnit << " to " << unit - 1
                                << " not valid or not modified\n" << eor;

            release();
            _child->synchronize( firstUnit, noOfUnits,
                                 log2BufSize, buffer );
            acquire();

            if ( unit > endUnit )
                break;

        }

        /*
         * write out units which have been modified 
         */

        firstUnit = unit;
        CacheUnitInfo valid( CUIValid );
        for ( unitCount = 0; 
              unitCount < maxUnits &&
              unit <= endUnit && unitInfo( unit ).modified();
              unitCount++ ) {
            setUnitInfo( unit, valid );
            // Assert( unitInfo( unit ).info() == valid.info() );
            unit++;
        }
        noOfUnits = unit - firstUnit;
        if ( noOfUnits > 0 ) {
        
            _dirty -= noOfUnits;

            int actualNoOfUnits;

            Raids( RaidDetail ) << this << "::synchronize "
                                << firstUnit << " to " << unit - 1
                                << " being sync'ed\n" << eor;

            if ( buffer == 0 ) {
                buffer = new char[ 1 << log2BufSize ];
                Assert( buffer != 0 );
            }
            release();
            actualNoOfUnits = basicRead( firstUnit, noOfUnits, 
                                         buffer, 0 );
            Assert( actualNoOfUnits == noOfUnits );
            actualNoOfUnits = moWrite( firstUnit, noOfUnits, 
                                       buffer, 1, 0 );
            Assert( actualNoOfUnits == noOfUnits );
            acquire();
            if ( unit > endUnit ) 
                break;

        }
    }
    Assert( !( startUnit == 0 && count == _numberOfUnits ) || _dirty == 0 );

    release();

    if ( orgBuffer == 0 && buffer != 0 && _parent == 0 ) {
        delete buffer;
        buffer = 0;
    }
    SInstr( CacheSynchronize.stop( istart ) );
}

CacheSynchronizeAsyncClosure::CacheSynchronizeAsyncClosure( 
    int startUnit, int endUnit, int maxUnits, int log2BufSize,
    Cache * cache, IOAsyncClosure * parent )
    :  _buffer( _realBuffer ) 
{
    _func = (AsyncFunc)&func;
    _startUnit = startUnit;
    _endUnit = endUnit;
    _unit = startUnit;
    _maxUnits = maxUnits;
    _log2BufSize = log2BufSize;
    _cache = cache;
    _parent = parent;
    _state = 0;
}

void CacheSynchronizeAsyncClosure::func( CacheSynchronizeAsyncClosure * c )
{
    c->_cache->synchronizeClosure( c );
}

void Cache::synchronize( unsigned int startUnit, unsigned int count, unsigned int log2BufSize, 
                         char * & buffer, IOAsyncClosure * c )
{
    Raids( RaidMethod ) << this << "::synchronize( " << startUnit << ", "
                        << count << ", " << log2BufSize << ", " 
                        << asHex( ( unsigned ) buffer ) 
                        << asHex( ( unsigned ) c ) << " )\n" << eor;
    Assert( c != 0 );

    Assert( dirty() );

    if ( log2BufSize == 0 ||
         log2BufSize > (unsigned)(_log2SegmentSize + 2) ) {
        log2BufSize = _log2SegmentSize + 2;
    }
    int log2BufSizeRatio = log2BufSize - _log2UnitSize;
    Assert( log2BufSizeRatio >= 0 );
    int maxUnits = 1 << log2BufSizeRatio;

    int endUnit = startUnit + count - 1;

    CacheSynchronizeAsyncClosure * closure =
    new CacheSynchronizeAsyncClosure
    ( startUnit, endUnit, maxUnits, log2BufSize,
      this, c );
    if ( _parent == 0 ) {
        Assert( buffer == 0 );
        closure->_realBuffer = 0;
    } else {
        closure->_buffer = buffer;
    }

    synchronizeClosure( closure );
}

void Cache::synchronizeClosure( CacheSynchronizeAsyncClosure * c )
{
    Raids( RaidMethod ) << this << "::synchronizeClosure( " 
                        << asAddress( c ) << " )\n" << eor;

    if ( c->_state == 1 ) {
        c->_state = 2;
        Assert( c->_unit == c->_firstUnit + c->_noOfUnits );
        Assert( c->_result == c->_noOfUnits );
        moWrite( c->_firstUnit, c->_noOfUnits, c->_buffer, 1, c );
        return;
    }
    if ( c->_state == 2 ) {
        Assert( c->_unit == c->_firstUnit + c->_noOfUnits );
        Assert( c->_result == c->_noOfUnits );
        c->_state = 0;
    }

    unsigned unit = c->_unit;
    unsigned endUnit = c->_endUnit;

    if ( unit > endUnit ) {
        Assert ( !( c->_startUnit == 0 && 
                    endUnit == ( _numberOfUnits - 1 ) ) );
        if ( c->_realBuffer != 0 && _parent == 0 ) {
            delete c->_realBuffer;
            c->_realBuffer = 0;
        }
        c->_parent->activate( endUnit - c->_startUnit + 1 );
        delete c;
        return;
    }

    int firstUnit;
    int noOfUnits;
    int unitCount;

    acquire();

    /* 
     * skip those units which are not valid or not modified
     */

    firstUnit = unit;
    
    while( unit <= endUnit && !unitInfo( unit ).modified() ) {
        unit++;
    }
    noOfUnits = unit - firstUnit;
    if ( noOfUnits > 0 ) {
        Raids( RaidDetail ) << this << "::synchronizeClosure "
                            << firstUnit << " to " << unit - 1
                            << " not valid or not modified\n" << eor;
        if ( _child->dirty() != 0 ) {
            release();
            c->_unit = unit;
            _child->synchronize
            ( firstUnit, noOfUnits, 
              c->_log2BufSize, c->_buffer, c );
            return;
        }
    }

    /*
     * write out units which have been modified 
     */

    firstUnit = unit;
    int maxUnits = c->_maxUnits;
    CacheUnitInfo valid( CUIValid );
    for ( unitCount = 0; 
          unitCount < maxUnits &&
          unit <= endUnit && unitInfo( unit ).modified();
          unitCount++ ) {
        setUnitInfo( unit, valid );
        // Assert( unitInfo( unit ).info() == valid.info() );
        unit++;
    }
    noOfUnits = unit - firstUnit;
    if ( noOfUnits > 0 ) {
        _dirty -= noOfUnits;
        Raids( RaidDetail ) << this << "::synchronizeClosure "
                            << firstUnit << " to " << unit - 1
                            << " being sync'ed\n" << eor;

        if ( c->_buffer == 0 ) {
            c->_buffer = new char[ 1 << c->_log2BufSize ];
        }
        release();
        c->_state = 1;
        c->_firstUnit = firstUnit;
        c->_noOfUnits = noOfUnits;
        c->_unit = unit;
        basicRead( firstUnit, noOfUnits, c->_buffer, c );
        return;
    }

    /* 
     * got here when they are not more units
     */
    
    release();
    Assert( unit > endUnit );
    Assert( !( c->_startUnit == 0 && endUnit == (unsigned)( _numberOfUnits - 1 ) ) ||
            _dirty == 0 );
    if ( c->_realBuffer != 0 && _parent == 0 ) {
        delete c->_realBuffer;
        c->_realBuffer = 0;
    }
    c->_parent->activate( endUnit - c->_startUnit + 1 );
    delete c;
}

int Cache::dirty()
{
    if ( _dirty ) 
        return _dirty;
    return _child->dirty(); 
}

#include "ClassInstaller.h"
DEFINECLASS("Cache",Cache,Object,0,0);
