#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#include "COW.h"
#include "COWCache.h"
#include "MemoryObject.h"
#include "PhysicallyAddressableUnit.h"
#include "PageTableConstants.h"


#ifdef __GNUC__
#pragma implementation
#endif

COWCache::COWCache( MemoryObject * memoryObject,
	unsigned int nunits, MemoryObject * source )
	: ( memoryObject, nunits, 1 )
{
    State * state = 0;
    if (source) {
	_sourceMO = source;
	_source = source->cache();
	state = STATE_COW_NPDC;
    }
    else {
	_source = 0;
	state = STATE_VM_NPDC;
    }
    
    for (unsigned i = 0; i < _numberOfUnits; i++) {
 	_pageState[i] = state;
    }
    _shadow = 0;
}

COWCache::~COWCache()
{
    if (_source)
	_sourceMO = 0;
    else
	delete _shadow;
}

PhysicallyAddressableUnit *
COWCache::pageFromShadow(unsigned unit)
{
    Assert(_shadow != 0);
    _shadow->preFault(unit);
    PhysicallyAddressableUnit * pau = _shadow->unitToPau(unit);
    _shadow->postFault(unit);
    if (pau != 0) {
	pau->setNextLink(0);
	putPlist(pau,unit);
    }
    return pau;
}

PhysicallyAddressableUnit *
COWCache::pageFromSource(unsigned unit)
{
    PhysicallyAddressableUnit * pau = 0;
    if (isACopy()) {
	pau = ((COWCache *)_source)->pageFromSource(unit);
	Assert(pau != 0);
	Assert(0 == pau->nextLink());
	putPlist(pau,unit);
    }
    else {
	preFault(unit);
	if (0 == (pau = unitToPau(unit)))
	    if (0 == (pau = pageFromShadow(unit))) {
		Assert(0 == unitToPau(unit));
		pagesFromStore(unit,unit);
		fromDisk( unit );
		pageToShadow(unit);
		pau = unitToPau(unit);
		Assert(STATE_COW_NPDC == _pageState[unit]);
		_pageState[unit] = STATE_COW_P;
	    }
	postFault(unit);
	Assert(pau != 0);
	Assert(0 == pau->nextLink());
	return pau;
    }
}
    
PhysicallyAddressableUnit *
COWCache::pageToShadow(unsigned unit)
{
    Assert(_shadow != 0);
    PhysicallyAddressableUnit * pau = unitToPau(unit);
    Assert(pau != 0);
    pau->setNextLink(0);
    _shadow->preFault(unit);
    _shadow->putPlist(pau,unit);
    _shadow->postFault(unit);
}

PhysicallyAddressableUnit *
COWCache::dupPageSelf(unsigned unit)
{
    PhysicallyAddressableUnit * src = unitToPau(unit);
    Assert(src != 0);
    PhysicallyAddressableUnit * dst = storeAlloc(1);
    Assert(dst != 0);

    char * srcp = (char *)src->address();
    Assert(srcp != 0);
    char * dstp = (char *)dst->address();
    Assert(dstp != 0);

    for (int i = PhysicalPageSize; i > 0; i--)
	dstp[i] = srcp[i];

    untranslatable(unit,unit,(ProtectionLevel)0);
    getPlist(unit);
    Assert(dst->nextLink() == 0);
    putPlist(dst,unit);
}

MemoryObjectCache *
COWCache::createShadow()
{
    Assert(!isACopy());
    Assert(_shadow == 0);
    // Since the cache is only a shadow, use any memory object
    _shadow = new PagedMemoryObjectCache(_memoryObject,_numberOfUnits);
    Assert(_shadow != 0);
    preFault(0,_numberOfUnits-1);
    untranslatable(0,_numberOfUnits-1,(ProtectionLevel)0);
    for (unsigned i = 0; i < _numberOfUnits; i++) {
	PhysicallyAddressableUnit * pau = unitToPau(i);
	if (pau == 0) {
	    _pageState[i] = STATE_COW_NPDC;
	}
	else {
	    PhysicallyAddressableUnit * link = pau->nextLink();
	    pau->setNextLink(0);
	    _shadow->putPlist(pau,i);
	    pau->setNextLink(link);
	    _pageState[i] = STATE_COW_P;
	}
    }
    postFault(0,_numberOfUnits-1);
    return _shadow;
}

int
COWCache::pagesToStore(unsigned from, unsigned to)
{
    if (to == 0) to = from;
    Assert( from <= to && to < _numberOfUnits );
    PhysicallyAddressableUnit * pau = 0;
{
    if (from > 0)
	if (map[from-1] != 0)
	    map[from-1]->setNextLink(0);
    PhysicallyAddressableUnit *p, *q;
    p = q = 0;
    for (; from <= to; from++)
	if (map[from] != 0 && _pageState[from] == STATE_VM_P) {
	    p = q = map[from];
	    map[from] = 0;
	    from++;
	    break;
	}
    for (; from <= to ; from++)
	if (map[from] != 0 && STATE_VM_P == _pageState[from]) {
	    q->setNextLink(map[from]);
	    q = map[from];
	    map[from] = 0;
	}
    if (q != 0)
	q->setNextLink(0);
    pau = p;
}
{
    PhysicallyAddressableUnit *p = pau;
    for (PhysicallyAddressableUnit * q = p; q != 0; q = q->nextLink()) {
	q->setModified(0);
	q->setReferenced(0);
    }
    if (p != 0) {
	storeDealloc(p);
	return 1;
    } else {
	return 0;
    }
}
}

void
COWCache::flush()
{
	preFault(0,_numberOfUnits-1);

	untranslatable(0,_numberOfUnits-1,(ProtectionLevel)0);
	saveModified(0,_numberOfUnits-1);

	postFault(0,_numberOfUnits-1);
}

void
COWCache::saveModified(unsigned from, unsigned to)
{
    if (to == 0) to = from;
    Assert( from <= to && to < _numberOfUnits );
    unsigned i, j;
    
    i = j = from;
    while (j <= to) {
	PhysicallyAddressableUnit *pau;
	// skip empty slots and unmodified pau's
	for (; j <= to && (((pau = unitToPau(j)) == 0) || !pau->modified()); j++);
	i = j;
	// i points to modified pau or i > to
	// skip modified pau's, resetting the modified bit as you go.
	for (; j <= to && (((pau = unitToPau(j)) != 0) && pau->modified()); j++)
	{
	    if (STATE_VM_P == _pageState[j])
		pau->setModified(0);
	}
	// j points to empty slot or unmodified pau or j > to
	if (i <= to) {
	    Assert(unitToPau(i) != 0);
	    toDisk(i,j-1);
	}
    }
}

#include "ClassInstaller.h"
DEFINECLASS("COWCache",COWCache,SimpleVMCache,0,0);
