#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#include "Assert.h"
#include "OutputStream.h"
#include "DVMManager.h"
#include "Kernel.h"
#include "Process.h"
#include "ThisProcess.h"	// For debugging
#include "OneCopyDVMCache.h"


#ifdef __GNUC__
#pragma implementation
#endif

OutputStream&
operator<<(OutputStream& os, DVMManager * mgr)
{
    return os << "DVMManager[" << asHex((int)mgr) << "]";
}

OutputStream&
operator<<(OutputStream& os, DVMManagerAction * act)
{
    return os << "DVMManagerAction[" << asHex((int)act) << "]("
	<< act->unit() << "," << act->pau() << "," << act->target()
	<< "," << act->sender() << ")" ;
}
    
static void
DVMCommLoop(int arg)
{
    DVMManager * mgr = (DVMManager *)arg;
    Console() << thisProcess() << "->DVMCommLoop(" << mgr << ")\n" << eor;
    mgr->commLoop();
}

DVMManager::DVMManager(int numMachines, int thisMachine,
		       OneCopyDVMCache * cache)
: _awaitAction(0)
{
    Assert(0 < numMachines && numMachines <= MaxNumberOfDVMMachines);

    _numMachines = numMachines;
    _thisMachine = thisMachine;
    
    _machines = new int [ numMachines ];
    for (int i=0; i < numMachines; i++ )
    {
	_machines[i]=0;
    }
    
    _cache = cache;
    add(this);

    _commProcess = new SystemProcess((PFV)::DVMCommLoop, TheKernel->domain(),
				     normal, (int)this);

    Assert(_commProcess != 0);
    _commProcess->setName("#DVMManagerCommDaemon");
    _commProcess->ready();
    
}

DVMManager::~DVMManager()
{
    delete _machines;
    _commProcess->die();
}

void
DVMManager::add(DVMManager *machine)
{
    int j = machine->id();
    for (int i=1; (j >>= 1) > 0; i++);
    _machines[i] = (int)machine;
}

void
DVMManager::send(int machine, unsigned unit, PhysicallyAddressableUnit * pau)
{
    DVMManagerAction * action = new DVMManagerAction(machine,unit,pau);
    enqueueA(action);
}

PhysicallyAddressableUnit *
DVMManager::rcv(int machine, unsigned unit)
{
    DVMManagerAction action(machine,unit);
    enqueueA(&action);
    action.sleep();
    // When we awaken, we have the pau;
    PhysicallyAddressableUnit * pau = action._pau;
    Assert(pau != 0);
    return pau;
}

void
DVMManager::commLoop()
{
    // Loop over the action queue and take appropriate action for each packet
    
    DVMManagerAction * action = 0;
    
    for(;;) {
	action = dequeueA();
	
	if (action != 0) {
	    if (action->_target == _thisMachine) { // incoming
		if (action->_pau == 0) { // incoming request for pau
		    Assert( _cache != 0 );
		    Debug << this  << "::commLoop(" << thisProcess() << ","
			<< _cache << "," << action << ") irqfp\n" << eor;
		    // trigger remote cache event
		    _cache->remoteCache(action->_unit, RemoteRead,
					action->_sender);
		}
		else {		         // incoming response with pau
		    // search outstanding queue, set pau and awake process;
		    DVMManagerAction * outstandingAction = dequeueO(action);
		    outstandingAction->_pau = action->_pau;
		    outstandingAction->awake();
		    Debug << this  << "::commLoop(" << thisProcess() << ","
			<< action << "," << outstandingAction << ") irswp\n"
			<< eor;
		}
		delete action;
	    }
	    else {			           // outgoing
		if (action->_pau == 0) { // outgoing request for pau
		    // put request into outstanding queue
		    // place copy into remote mgr;
		    DVMManagerAction * copy = new
			DVMManagerAction(action->_target, action->_unit);
		    Debug << this  << "::commLoop(" << thisProcess() << ","
			<< action << "," << copy << ") orqfp\n" << eor;
		    enqueueO(action);
		    sendToRemoteMgr(copy);
		}
		else {		         // outgoing response with pau
		    Debug << this  << "::commLoop(" << thisProcess() << ","
			<< action << ") orswp\n" << eor;
		    // place packet into remote mgr;
		    sendToRemoteMgr(action);
		}
	    }
	    action = 0;
	}
	else {
	    Debug << this  << "::commLoop(" << thisProcess()
		<< ") waiting\n" << eor;
	    _awaitAction.P();
	}
    }
}

void
DVMManager::sendToRemoteMgr(DVMManagerAction * action)
{
    int j = action->_target;
    for (int i=1; (j >>= 1) > 0; i++);
    DVMManager * remote = ((DVMManager *)_machines[i]);
    action->_sender = _thisMachine;
    remote->enqueueA(action);
}
