#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#include "OneCopyDVM.h"
#include "OneCopyDVMCache.h"
#include "OutputStream.h"


#ifdef __GNUC__
#pragma implementation
#endif

static inline unsigned *
fragmentAddress(unsigned fragnum, unsigned * address)
{
    switch (fragnum) {
      case 0x1:
	return address;
	break;
      case 0x2:
	return address+350;
	break;
      case 0x4:
	return address+700;
	break;
      default:
	Assert(NOTREACHED);
	break;
    }
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// VM States
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//----------------------------------------------- NCDVMVMState - abstract

State *
NCDVMVMState::localRead(MemoryObjectCache * cache, unsigned unit,
			   AddressTranslation * adt)
{
    Assert(_delegatee != 0);
    return returnState(((SimpleVMState *)_delegatee)->
		       localRead(cache, unit, adt));
}

State *
NCDVMVMState::localWrite(MemoryObjectCache * cache, unsigned unit,
			    AddressTranslation * adt)
{
    Assert(_delegatee != 0);
    return returnState(((SimpleVMState *)_delegatee)->
		       localWrite(cache, unit, adt));
}

State *
NCDVMVMState::pageOut(MemoryObjectCache * cache, unsigned unit)
{
    Assert(_delegatee != 0);
    return returnState(((SimpleVMState *)_delegatee)->
		       pageOut(cache, unit));
}

FaultErrorCode
NCDVMVMState::code()
{
    Assert(_delegatee != 0);
    return ((SimpleVMState *)_delegatee)->code();
}

//----------------------------------------------- NCDVM_VM_NP - paged to disk

State *
NCDVM_VM_NP::remoteWrite(MemoryObjectCache * cache, unsigned unit)
{
    // get page and fill it. Next state will be distributed, so do not
    // make it translatable.
    
    cache->pagesFromStore(unit);
    cache->fromDisk(unit);
    return returnState(STATE_NCDVM_VM_PD);
}

//----------------------------------------------- NCDVM_VM_P - page present

State *
NCDVM_VM_P::remoteWrite(MemoryObjectCache * cache, unsigned unit)
{
    // Make page inaccessible, ready to send it off to the net
    
    cache->untranslatable(unit,unit,NoAccess);
    return returnState(STATE_NCDVM_VM_PD);
}

//----------------------------------------------- NCDVM_VM_PD - page distributed

State *
NCDVM_VM_PD::localRead(MemoryObjectCache * cache, unsigned unit,
		   AddressTranslation * adt)
{
    // get a page ready for recption
    
    cache->pagesFromStore(unit);
    return returnState(STATE_NCDVM_VM_P);
}

State *
NCDVM_VM_PD::localWrite(MemoryObjectCache * cache, unsigned unit,
		    AddressTranslation * adt)
{
    // get a page ready for recption
    
    cache->pagesFromStore(unit);
    return returnState(STATE_NCDVM_VM_P);
}

State *
NCDVM_VM_PD::pageOut(MemoryObjectCache *, unsigned)
{
    Assert(NOTREACHED);
    return 0;
}

FaultErrorCode
NCDVM_VM_PD::code()
{
    return MissingPhysicalMemory;
}

State *
NCDVM_VM_PD::remoteWrite(MemoryObjectCache *, unsigned)
{
    // Ignore remote requests
    return returnState(STATE_NCDVM_VM_PD);
}

//----------------------------------------------- The state machine

NCDVMVMMachine::NCDVMVMMachine()
: SimpleVMMachine(4)
{
    REPLACESTATE( NCDVM_VM_NP, VM_NPDC );
    REPLACESTATE( NCDVM_VM_P, VM_P );
    NEWSTATE( NCDVM_VM_PD );
    NEWSTATE( NCDVM_VM_TRANS );
}

NCDVMVMState * STATE_NCDVM_VM_TRANS = 0;
NCDVMVMState * STATE_NCDVM_VM_NP = 0;
NCDVMVMState * STATE_NCDVM_VM_P = 0;
NCDVMVMState * STATE_NCDVM_VM_PD = 0;

NCDVMVMMachine * TheNCDVMVMMachine = 0;

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Networking states
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

State *
NCDVMNetState::remoteWriteDone(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg, NCDVMState * parent )
{
    // Only old messages are legal.
    Assert(!parent->currentMsg(cache,unit,msg));
    return returnState(this);
}

State *
NCDVMNetState::hereRemoteWrite(NCDVMCache * cache, unsigned unit,
			       DVMCacheMsg * msg, NCDVMState * parent)
{
    // Only old messages are legal.
    Assert(!parent->currentMsg(cache,unit,msg));
    DVMAckPageMsg * m =
	new DVMAckPageMsg(msg->source(),
			  msg->destination(),
			  msg->timestamp(),
			  msg->unit(),
			  msg->fragment());
    cache->sendMsg(m);
    return returnState(this);
}

State *
NCDVMNetState::timeout(NCDVMCache *, unsigned)
{
    Assert(NOTREACHED);
    return 0;
}

//----------------------------------------------- NCDVM_NET_SP - send page

State *
NCDVM_NET_SP::remoteWriteDone(NCDVMCache * cache, unsigned unit,
			      DVMCacheMsg * msg, NCDVMState * parent)
{
    if (parent->currentMsg(cache,unit,msg)) {
	if ((cache->setFragment(unit,msg->fragment()) == 0x7)) {
	    // All fragments acked
	    cache->pagesToStore(unit);
	    cache->clearFragment(unit);
	    return returnState(STATE_NCDVM_NET_Q);
	}
    }
    else
	return NCDVMNetState::remoteWriteDone(cache,unit,msg,parent);
}

State *
NCDVM_NET_SP::timeout(NCDVMCache * cache, unsigned unit)
{
    unsigned received = cache->fragment(unit);
    for (unsigned fragcount=0x1; fragcount < 0x8; fragcount <<= 0x1) {
	if ((received & fragcount) == 0) {
	    cache->sendMsg(new DVMHerePageMsg(cache->owner(unit),
					      cache->thisMachine(),
					      cache->timestamp(unit),
					      unit,
					      fragmentAddress(fragcount,
							      cache->
							      unitAddress(unit)),
					      fragcount));
	}
    }
    return returnState(this);
}

//----------------------------------------------- NCDVM_NET_WP - waitfor page

State *
NCDVM_NET_WP::hereRemoteWrite(NCDVMCache * cache, unsigned unit,
			      DVMCacheMsg * msg, NCDVMState * parent)
{
    Assert(msg->unit() == unit);
    
    if (parent->currentMsg(cache,unit,msg)) {
	
	State * ret = 0;
	
	if ((cache->fragment(unit) & msg->fragment()) == 0) {
	    cache->setFragment(unit,msg->fragment());
	    unsigned * incoming = msg->address();
	    unsigned * page = cache->unitAddress(unit);
	    unsigned start = 0;
	    unsigned size = 0;
	    switch (msg->fragment()) {
	      case 0x1:
		start = 0;
		// end = 349;
		size = 350;
		break;
	      case 0x2:
		start = 350;
		// end = 699;
		size = 350;
		break;
	      case 0x4:
		start = 700;
		// end = 1023;
		size = 324;
		break;
	      default:
		Assert(NOTREACHED);
		break;
	    }
	    for(unsigned i=0; i < size; i++)
		page[start+i] = incoming[i];
	}
	
	DVMAckPageMsg * m =
	    new DVMAckPageMsg(msg->source(),
			      msg->destination(),
			      msg->timestamp(),
			      msg->unit(),
			      cache->fragment(unit));
	cache->sendMsg(m);
	
	if (cache->fragment(unit) == 0x7) {
	    cache->clearFragment(unit);
	    return returnState(STATE_NCDVM_NET_Q);
	}
	else
	    return returnState(this);
    }
    else
	return NCDVMNetState::hereRemoteWrite(cache,unit,msg,parent);
}

State *
NCDVM_NET_WP::timeout(NCDVMCache * cache, unsigned unit)
{
    unsigned received = cache->fragment(unit);
    if (received < 0x7)
	cache->sendMsg(new DVMGetPageMsg(cache->owner(unit),
					 cache->thisMachine(),
					 cache->timestamp(unit),
					 unit,
					 0,
					 (~received&0x7)));
    return returnState(this);
}

//----------------------------------------------- The state machine

NCDVMNetMachine::NCDVMNetMachine()
: StateMachine(3)
{
    NEWSTATE( NCDVM_NET_Q );
    NEWSTATE( NCDVM_NET_SP );
    NEWSTATE( NCDVM_NET_WP );
}

NCDVMNetState * STATE_NCDVM_NET_Q = 0;
NCDVMNetState * STATE_NCDVM_NET_SP = 0;
NCDVMNetState * STATE_NCDVM_NET_WP = 0;

NCDVMNetMachine * TheNCDVMNetMachine = 0;

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Ownership states
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//----------------------------------------------- NCDVMOwnerState


// Message must have smaller timestamp, since I am the timekeeper
State *
NCDVMOwnerState::ownerHint(NCDVMCache * cache, unsigned unit,
			   DVMCacheMsg * msg)
{
    Assert(msg->timestamp() < cache->timestamp(unit));
    return returnState(this);
}

// Message must have smaller timestamp, since I am the timekeeper
State *
NCDVMOwnerState::remoteWrite(NCDVMCache * cache, unsigned unit,
			     DVMCacheMsg * msg)
{
    Assert(msg->timestamp() < cache->timestamp(unit));
    return returnState(this);
}

//----------------------------------------------- NCDVM_OWN - own the page

int
NCDVM_O_OWN::currentMsg(NCDVMCache * , unsigned , DVMCacheMsg * )
{
    return 0;
}

// Message must have smaller timestamp, since I am the timekeeper
State *
NCDVM_O_OWN::remoteWrite(NCDVMCache * cache, unsigned unit,
			 DVMCacheMsg * msg)
{
    Assert(msg->timestamp() < cache->timestamp(unit));
    cache->setOwner(unit, msg->source());
    return returnState(STATE_NCDVM_O_TONOT);
}

//----------------------------------------------- NCDVM_NOT - do not own

int
NCDVM_O_NOT::currentMsg(NCDVMCache * , unsigned , DVMCacheMsg * )
{
    return 0;
}

// Message must have smaller timestamp - since we can receive remoteWrite
// only from some node that thinks we have the page. i.e. we must have
// received the page from that node before. At startup, remote node
// timestamp must be zero
State *
NCDVM_O_NOT::remoteWrite(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg)
{
    Assert(msg->unit() == unit);
    Assert(msg->timestamp() < cache->timestamp(unit));
    
    DVMOwnerHintMsg * m = 
	new DVMOwnerHintMsg(msg->source(),
			    msg->destination(),
			    cache->timestamp(unit),
			    msg->unit(),
			    cache->owner(unit));
    cache->sendMsg(m);
    return returnState(this);
}

//----------------------------------------------- NCDVM_TOOWN - to be owner

int
NCDVM_O_TOOWN::currentMsg(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg)
{
    return (msg->source() == cache->owner(unit) &&
	    msg->timestamp() > cache->timestamp(unit));
}

// Message must have greater or equal timestamp, since I am requesting
// from either the originator, or a node I gave the page to

State *
NCDVM_O_TOOWN::ownerHint(NCDVMCache * cache, unsigned unit,
			 DVMCacheMsg * msg)
{
    // Message must have greater or equal timestamp
    if (currentMsg(cache,unit,msg)) {
	cache->setOwner(unit, msg->fragment());
    }
    return returnState(this);
}

//----------------------------------------------- NCDVM_TONOT - to not be owner

int
NCDVM_O_TONOT::currentMsg(NCDVMCache * cache, unsigned unit, DVMCacheMsg * msg)
{
    return (msg->source() == cache->owner(unit) &&
	    msg->timestamp() == cache->timestamp(unit));
}

//----------------------------------------------- The state machine

NCDVMOwnerMachine::NCDVMOwnerMachine()
: StateMachine(4)
{
    NEWSTATE( NCDVM_O_OWN );
    NEWSTATE( NCDVM_O_NOT );
    NEWSTATE( NCDVM_O_TOOWN );
    NEWSTATE( NCDVM_O_TONOT );
}

NCDVMOwnerState * STATE_NCDVM_O_OWN = 0;
NCDVMOwnerState * STATE_NCDVM_O_NOT = 0;
NCDVMOwnerState * STATE_NCDVM_O_TOOWN = 0;
NCDVMOwnerState * STATE_NCDVM_O_TONOT = 0;

NCDVMOwnerMachine * TheNCDVMOwnerMachine = 0;

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Final states
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

NCDVMState::NCDVMState(StateMachine * m, State * d, NCDVMVMState * vm,
		       NCDVMNetState * net, NCDVMOwnerState * owner)
: SimpleVMState(m,d),_vm(vm),_net(net),_own(owner)
{
    Assert(d != 0);
    Assert(vm != 0);
    Assert(net != 0);
    Assert(owner != 0);
}
    
NCDVMState::NCDVMState(StateMachine * m, NCDVMVMState * vm,
		       NCDVMNetState * net, NCDVMOwnerState * owner)
: SimpleVMState(m),_vm(vm),_net(net),_own(owner)
{
    Assert(vm != 0);
    Assert(net != 0);
    Assert(owner != 0);
}

//----------------------------------------------- NCDVM_P - node owns page

State *
NCDVM_P::pageOut(MemoryObjectCache * cache, unsigned unit)
{
    _vm->pageOut(cache,unit);
    return returnState(STATE_NCDVM_NP);
}

State *
NCDVM_P::remoteWrite(NCDVMCache * cache, unsigned unit,
		     DVMCacheMsg * msg)
{
    _own->remoteWrite(cache,unit,msg);
    _vm->remoteWrite(cache,unit);
    Assert(cache->fragment(unit) == 0);
    cache->enableTimeout(unit);
    STATE_NCDVM_NET_SP->timeout(cache,unit);
    cache->disableLocalEvents(unit);
    return returnState(STATE_NCDVM_SP);
}

//----------------------------------------------- NCDVM_NP - paged out locally

State *
NCDVM_NP::localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt)
{
    _vm->localRead(cache,unit,adt);
    return returnState(STATE_NCDVM_P);
}

State *
NCDVM_NP::localWrite(MemoryObjectCache * cache, unsigned unit,
		     AddressTranslation * adt)
{
    _vm->localWrite(cache,unit,adt);
    return returnState(STATE_NCDVM_P);
}

State *
NCDVM_NP::remoteWrite(NCDVMCache * cache, unsigned unit,
		      DVMCacheMsg * msg)
{
    _vm->remoteWrite(cache,unit);
    _own->remoteWrite(cache,unit,msg);
    Assert(cache->fragment(unit) == 0);
    cache->enableTimeout(unit);
    STATE_NCDVM_NET_SP->timeout(cache,unit);
    cache->disableLocalEvents(unit);
    return returnState(STATE_NCDVM_SP);
}

//----------------------------------------------- NCDVM_PD - page distributed

State *
NCDVM_PD::localRead(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt)
{
    NCDVMCache * ncache = (NCDVMCache *)cache;
    _vm->localRead(ncache,unit,adt);
    Assert(ncache->fragment(unit) == 0);
    ncache->enableTimeout(unit);
    STATE_NCDVM_NET_WP->timeout(ncache,unit);
    ncache->disableLocalEvents(unit);
    return returnState(STATE_NCDVM_WP);
}

State *
NCDVM_PD::localWrite(MemoryObjectCache * cache, unsigned unit,
			      AddressTranslation * adt)
{
    NCDVMCache * ncache = (NCDVMCache *)cache;
    _vm->localWrite(ncache,unit,adt);
    Assert(ncache->fragment(unit) == 0);
    ncache->enableTimeout(unit);
    STATE_NCDVM_NET_WP->timeout(ncache,unit);
    ncache->disableLocalEvents(unit);
    return returnState(STATE_NCDVM_WP);
}

//----------------------------------------------- NCDVM_SP - send out page

State *
NCDVM_SP::remoteWriteDone(NCDVMCache * cache, unsigned unit,
			  DVMCacheMsg * msg)
{
    State * s = _net->remoteWriteDone(cache,unit,msg,this);
    if (s == STATE_NCDVM_NET_Q) {
	// All fragments acked
	cache->disableTimeout(unit);
	cache->enableLocalEvents(unit);
	return returnState(STATE_NCDVM_PD);
    }
    else
	return returnState(this);
}

//----------------------------------------------- NCDVM_WP - wait for page

State *
NCDVM_WP::ownerHint(NCDVMCache * cache, unsigned unit,
		    DVMCacheMsg * msg)
{
    _own->ownerHint(cache,unit,msg);
    _net->timeout(cache,unit);
    return returnState(this);
}

State *
NCDVM_WP::hereRemoteWrite(NCDVMCache * cache, unsigned unit,
			  DVMCacheMsg * msg)
{
    State * s = _net->hereRemoteWrite(cache,unit,msg,this);
    if (s == STATE_NCDVM_NET_Q) {
	cache->nextStamp(unit,msg->timestamp());
	cache->disableTimeout(unit);
	cache->enableLocalEvents(unit);
	cache->delayActionProcess();
	return returnState(STATE_NCDVM_P);
    }
    else
	return returnState(this);
}

NCDVMMachine::NCDVMMachine()
: SimpleVMMachine(5)
{
    Console() << "NCDVMVMMachine\n" << eor;
    TheNCDVMVMMachine = new NCDVMVMMachine();
    Console() << "NCDVMNetMachine\n" << eor;
    TheNCDVMNetMachine = new NCDVMNetMachine();
    Console() << "NCDVMOwnerMachine\n" << eor;
    TheNCDVMOwnerMachine = new NCDVMOwnerMachine();
    
    Console() << "NCDVM_NP\n" << eor;
    REPLACESTATE( NCDVM_NP, VM_NPDC );
    Console() << "NCDVM_P\n" << eor;
    REPLACESTATE( NCDVM_P, VM_P );
    Console() << "NCDVM_PD\n" << eor;
    NEWSTATE( NCDVM_PD );
    Console() << "NCDVM_SP\n" << eor;
    NEWSTATE( NCDVM_SP );
    Console() << "NCDVM_WP\n" << eor;
    NEWSTATE( NCDVM_WP );
}

NCDVMState * STATE_NCDVM_NP = 0;
NCDVMState * STATE_NCDVM_P = 0;
NCDVMState * STATE_NCDVM_PD = 0;
NCDVMState * STATE_NCDVM_SP = 0;
NCDVMState * STATE_NCDVM_WP = 0;

NCDVMMachine * TheNCDVMMachine = 0;
