#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#include "Assert.h"
#include "OutputStream.h"
#include "OneCopyDVMCache.h"
#include "OneCopyDVM.h"
#include "DVMAccessType.h"
#include "DVMManager.h"
#include "ThisProcess.h"
#include "Kernel.h"
#include "Process.h"
#include "BitRanges.h"
#include "LinkIterator.h"
#include "AddressTranslation.h"
#include "PhysicalMemoryChain.h"
#include "PhysicallyAddressableUnit.h"
#include "MemoryObject.h"


#ifdef __GNUC__
#pragma implementation
#endif

const long ONE_SECOND = 1000000;
const long ONE_MINUTE = ONE_SECOND*60;

void
NCDVMCache::DVMActionProcess(int arg)
{
    NCDVMCache * cache = (NCDVMCache *)arg;
    Console() << thisProcess() << "(" << cache << ")\n" << eor;
    cache->actionProcessEntry();
}

void
NCDVMCache::DVMCacheTimeout(void * cache, int unit)
{
    ((NCDVMCache *)cache)->cacheTimeout((unsigned)unit);
}


NCDVMCache::NCDVMCache( MemoryObject * memoryObject,
	unsigned int nunits, unsigned id, int copy )
: ( memoryObject, nunits, 1 ),_stateMutex(1)
{
    _manager = 0;
    
    _machines = new unsigned [ _numberOfUnits ];
    _fragments = new char [ _numberOfUnits ];
    _timestamps = new unsigned [ _numberOfUnits ];
    
    State * state = copy ? STATE_NCDVM_PD : STATE_NCDVM_NP;
    Assert(state != 0);
    unsigned ts = copy ? 0 : 1;
    
    for (int i=0; i < _numberOfUnits; i++) {
	_machines[i]=id;
	_fragments[i]=0;
	_timestamps[i]=ts;
	
 	_pageState[i] = state;
    }
    
    _localEventGate = new BitRanges(_numberOfUnits);
    _blockedEvents = new List;
    
    _awaitAction = new Semaphore(0);
    _delayFlag = 0;
    _actionProcess =
	new PreemptableSystemProcess((PFV)NCDVMCache::DVMActionProcess,
				     TheKernel->domain(),
				     normal, (int)this);
    Assert(_actionProcess != 0);
    _actionProcess->setName("#DVMCacheActionProcess");
    _actionProcess->ready();

    _cacheTimer = new GroupTimer(nunits,ONE_SECOND*30,(void *)this,
				 NCDVMCache::DVMCacheTimeout);
}

/*
 * Free data structures and paus. Assume that the MemoryObject has
 * called appropriate methods and there are no references anymore to
 * any pages that were in here. Also assume that dirty pages do not
 * matter.
 */
NCDVMCache::~NCDVMCache()
{
    delete _machines;
    delete _fragments;
    delete _timestamps;

    delete _localEventGate;
    delete _blockedEvents;		// faulted Processes
}

void
NCDVMCache::disableLocalEvents(unsigned from, unsigned to)
{
    if (to == 0) to = from;
    Assert( from <= to && to < _numberOfUnits );
    _localEventMutex.acquire();
    int ret = _localEventGate->alloc(from,to);
    _localEventMutex.release();
    Assert( ret != 0 );
}

int
NCDVMCache::areLocalEventsEnabled(unsigned from, unsigned to)
{
    if (to == 0) to = from;
    Assert( from <= to && to < _numberOfUnits );
    _localEventMutex.acquire();
    int ret = _localEventGate->isFree(from,to);
    _localEventMutex.release();
    return ret;
}


void
NCDVMCache::enableLocalEvents(unsigned from, unsigned to)
{
    if (to == 0) to = from;
    Assert( from <= to && to < _numberOfUnits );
    
    _localEventMutex.acquire();
    int ret = _localEventGate->free(from,to);
    _localEventMutex.release();
    Assert( ret != 0 );
    
    _faultInfo *f = 0;
    _blockMutex.acquire();
    LinkIterator i(_blockedEvents->theLink());
    for (int count = 0; (f = (_faultInfo *)i.next()) && count < 100; count++)
	if (f->intersect(from,to))
	    f->awake();
    _blockMutex.release();
}

void
NCDVMCache::nextStamp(unsigned unit, unsigned timestamp)
{
    Assert(timestamp > _timestamps[unit]);
    _timestamps[unit] = timestamp+1;
}

void
NCDVMCache::actionProcessEntry()
{
    DVMCacheMsg * message = 0;
    
    for(;;) {
	message = deqAction();
	if (message != 0) {
	    unsigned * packet = message->top();
	    if (_delayFlag) {
		_delayTimer.start(ONE_SECOND);
		_delayTimer.await();
		_delayFlag = 0;
	    }
#ifdef ASSERT
	    switch (message->type()) {
	      case GetPage:
	      case HerePage:
	      case AckPage:
	      case OwnerHint:
	      case GetRead:
	      case HereRead:
	      case GetUp:
	      case HereUp:
	      case AckUp:
	      case Inv:
	      case AckInv:
		Console() << "CacheAction(" << message << ")\n" << eor;
		message->action(this);
		break;
	      default:
		Assert(NOTREACHED);
		break;
	    }
#else
	    message->action(this);
#endif /* ASSERT */    
	    delete packet;
	}
	else {
	    Debug << "NCDVMCache::action wait\n" << eor;
	    _awaitAction->P();
	}
    }
}

PhysicalMemoryChain
NCDVMCache::cache(unsigned int unit,
		     ProtectionLevel requiredProtection,
		     FaultErrorCode & code,
		     AddressTranslation * adt,
		     const char * vaddr, int addToAdt)
{
    Debug << this << "::cache( unit: " << unit << ", requiredProtection: "
	<< requiredProtection << " )\n" << eor;
    Assert( isIn( unit ) );
    
    addAdt(adt,(const char *)(floorUnit((unsigned)vaddr) - unitToOffset(unit)));

    int enabled = 0;
    int done = 0;
    
    preFault(unit);
    
    Assert( requiredProtection <= _memoryObject->logicalProtectionLevel() );

    unsigned from = unit;
    unsigned to = from;
    
    _faultInfo *f = new _faultInfo(from,to);
    
    const int LIMIT = 10000;
    
    for(int tried = 0; tried < LIMIT; tried++) {

	_stateMutex.P();
	
	if ((enabled = areLocalEventsEnabled(from,to)) == 0) {
	    _blockMutex.acquire();
	    _blockedEvents->addFirst(f);
	    _blockMutex.release();
	    for (int awakened = 0;
		 awakened < LIMIT &&
		 ((enabled = areLocalEventsEnabled(from,to)) == 0);
		 awakened++)
	    {
		_stateMutex.V();
		f->sleep();
		_stateMutex.P();
	    }
	    Assert(awakened < LIMIT);
	    _blockMutex.acquire();
	    _blockedEvents->remove(f);
	    _blockMutex.release();
	}
	
	code = ((NCDVMState *)_pageState[unit])->code();
    
	if (requiredProtection == ReadOnly)
	    _pageState[unit] =
		((NCDVMState *)_pageState[unit])->localRead(this, unit, adt);
	else
	    _pageState[unit] =
		((NCDVMState *)_pageState[unit])->localWrite(this, unit, adt);

	done = areLocalEventsEnabled(from,to);

	_stateMutex.V();

	if (done)
	    break;
    }

    delete f;
    
    Assert(tried < LIMIT);
    
    PhysicallyAddressableUnit *pau = unitToPau(unit);
    
    pau->setProtection( requiredProtection );
    
    postFault(unit);
    
    return( PhysicalMemoryChain( pau, 0, pau->size()));
}

PhysicalMemoryChain
NCDVMCache::mcache(unsigned int unit, int nunits,
		   ProtectionLevel requiredProtection,
		   FaultErrorCode & code,
		   AddressTranslation * adt,
		   const char * vaddr, int addToAdt)
{
    Assert(NOTREACHED);
    return (PhysicalMemoryChain(0,0,0));
}

void
NCDVMCache::stateAction(PMFCUMS f, DVMCacheMsg * msg)
{
    unsigned unit = msg->unit();
    Assert(isIn(unit));
    _stateMutex.P();
    _pageState[unit] = (((NCDVMState *)_pageState[unit])->*f)(this, unit, msg);
    _stateMutex.V();
}

unsigned *
NCDVMCache::unitAddress(unsigned unit)
{
    Assert(isIn(unit));
    Assert(map[unit] != 0);
    return (unsigned *)map[unit]->address();
}

void
NCDVMCache::cacheTimeout(unsigned unit)
{
    Assert(isIn(unit));
    _stateMutex.P();
    _pageState[unit] = ((NCDVMState *)_pageState[unit])->timeout(this, unit);
    _stateMutex.V();
}

RCDVMCache::RCDVMCache( MemoryObject * memoryObject,
	unsigned int nunits, unsigned id, int copy )
: ( memoryObject, nunits, id, copy )
{
    _copyholders = new unsigned [ _numberOfUnits ];
    
    State * state = copy ? STATE_RCDVM_PD : STATE_RCDVM_NP;
    Assert(state != 0);
    
    for (int i=0; i < _numberOfUnits; i++) {
	_copyholders[i]=0;
 	_pageState[i] = state;
    }
}

RCDVMCache::~RCDVMCache()
{
    delete _copyholders;
}

void
RCDVMCache::stateAction(PMFCUMR f, DVMCacheMsg * msg)
{
    unsigned unit = msg->unit();
    Assert(isIn(unit));
    _stateMutex.P();
    _pageState[unit] = (((RCDVMState *)_pageState[unit])->*f)(this, unit, msg);
    _stateMutex.V();
}

void
RCDVMCache::cacheTimeout(unsigned unit)
{
    Assert(isIn(unit));
    _stateMutex.P();
    _pageState[unit] = ((RCDVMState *)_pageState[unit])->timeout(this, unit);
    _stateMutex.V();
}

#include "ClassInstaller.h"
DEFINECLASS("NCDVMCache",NCDVMCache,SimpleVMCache,0,0);
DEFINECLASS("RCDVMCache",RCDVMCache,NCDVMCache,0,0);
