#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * SimpleVMCache.c: Paged MemoryObjectCache class implementations.
 */

#include "Assert.h"
#include "AddressTranslation.h"
#include "OutputStream.h"
#include "SimpleVMCache.h"
#include "PhysicalMemoryChain.h"
#include "PhysicallyAddressableUnit.h"
#include "MemoryObject.h"
#include "SimpleVM.h"


#ifdef __GNUC__
#pragma implementation
#endif


SimpleVMCache::SimpleVMCache( MemoryObject * memoryObject,
	unsigned int nunits )
	: ( memoryObject, nunits )
{
	_pageState = new State * [ _numberOfUnits ];
	for( int i = 0; i < _numberOfUnits; i++ ) {
		_pageState[i] = STATE_VM_NPDC;
	}
}

SimpleVMCache::SimpleVMCache( MemoryObject * memoryObject,
	unsigned int nunits, int derived )
	: ( memoryObject, nunits )
{
	_pageState = new State * [ _numberOfUnits ];
}


/*
 * Free data structures and paus. Assume that the MemoryObject has
 * called appropriate methods and there are no references anymore to
 * any pages that were in here. Also assume that dirty pages do not
 * matter.
 */
SimpleVMCache::~SimpleVMCache()
{
	delete _pageState;
}

PhysicalMemoryChain
SimpleVMCache::cache(unsigned int unit,
		     ProtectionLevel requiredProtection,
		     FaultErrorCode & code,
		     AddressTranslation * adt,
		     const char * vaddr, int addToAdt)
{
    Debug << this << "::cache( unit: " << unit << ", requiredProtection: "
	<< requiredProtection << " )\n" << eor;
    Assert( isIn( unit ) );
    
    addAdt(adt,(const char *)(floorUnit((unsigned)vaddr) - unitToOffset(unit)));
    
    preFault(unit);
    
    Assert( requiredProtection <= _memoryObject->logicalProtectionLevel() );

    code = pageState(unit)->code();
    
    if (requiredProtection == ReadOnly)
	_pageState[unit] = pageState(unit)->localRead(this, unit, adt);
    else
	_pageState[unit] = pageState(unit)->localWrite(this, unit, adt);
	
    PhysicallyAddressableUnit *pau = unitToPau(unit);
    
    pau->setProtection( requiredProtection );
    
    postFault(unit);
    
    return( PhysicalMemoryChain( pau, 0, pau->size()));
}

PhysicalMemoryChain
SimpleVMCache::mcache(unsigned int unit, int nunits,
		      ProtectionLevel requiredProtection,
		      FaultErrorCode & code,
		      AddressTranslation * adt,
		      const char * vaddr, int addToAdt)
{
    unsigned from = unit; unsigned to = from+nunits-1;
    Assert( from <= to && to < _numberOfUnits );
    
    PhysicallyAddressableUnit * pau = 0;
    
    code = MissingAddressTranslationMapping;
    
    addAdt(adt,(const char *)(floorUnit((unsigned)vaddr) - unitToOffset(unit)));
    
    preFault(unit);
    
    Assert( requiredProtection <= _memoryObject->logicalProtectionLevel() );
    
    for (unsigned i = from; i < to; i++)
    {
	code = pageState(i)->code();
	
	if (requiredProtection == ReadOnly)
	    _pageState[i] = pageState(i)->localRead(this, unit, adt);
	else
	    _pageState[i] = pageState(i)->localWrite(this, unit, adt);
	
	pau = unitToPau(unit);
	
	pau->setProtection( requiredProtection );
    }
    
    postFault(unit);
    
    return( PhysicalMemoryChain( pau, 0, pau->size()));
}

PhysicallyAddressableUnit *
SimpleVMCache::pageOut()
{
    unsigned unit = victim();
    Debug << this << "::pageOut() on cpu " << thisCPU()
	<< " unit " << unit << "\n" << eor;
    if ( unit == ~0x0 )
	return 0;
    else {
	_pageState[unit] = pageState(unit)->pageOut(this,unit);
	PhysicallyAddressableUnit * p = getPlist( unit );
	postPageOut(unit);
	unuse(unit,unit);
	return p;
    }
}

#include "ClassInstaller.h"
DEFINECLASS("SimpleVMCache",SimpleVMCache,PagedMemoryObjectCache,0,0);
