#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Domain.c: domain class implementation.
 */

//#define DEBUG_COPY
#define OLDDEFAULTPROXY
#ifdef OLDDEFAULTPROXY
#define HACKFIX
#endif

/*
 * PUT THE FOLLOWING in the machine configuration file or MachineKernel.cc:
 */
const int MAXPROXIESPERDOMAIN = 512;
const int KernelDomainMaxMemoryObjectsInMap = 1024;
const int UserDomainMaxMemoryObjectsInMap = 256;


#include "Assert.h"
#include "Byte.h"
#include "CPU.h"
#include "Domain.h"
#include "DummyMemoryObject.h"
#include "FileSystemInterface.h"
#include "Kernel.h"
#include "LockHolder.h"
#include "MemoryObject.h"
#include "MemoryObjectCache.h"
#include "MemoryObjectCacheStar.h"
#include "MemoryObjectView.h"
#include "NameServer.h"
#include "ObjectProxyManager.h"
#include "OutputStream.h"
#include "PageTableConstants.h"
#include "PersistentMemoryObject.h"
#include "PersistentMemoryObjectStar.h"
#include "PhysicalMemoryChain.h"
#include "PhysicallyAddressableUnit.h"
#include "ProtectionLevel.h"
#include "Raid.h"
#include "SourceOfZerosMemoryObject.h"
#include "String.h"
#include "SystemInterface.h"
#include "SystemInterfaceStar.h"
#include "ThisProcess.h"
#include "ADDRESS_TRANSLATION.h"
#if ( MULTIMAX )
#include "PROCESSOR_CONTEXT.h"
#endif
#include "MemorySimpleInstruments.h"
#ifdef HACKFIX
/* 
 * The problem here is that during the restoreFromInitialContext,
 * proxies are allocated, this can cause a page-in to take place 
 * which causes a context switch in the middle of restoreFromInitialContext.
 * Since restoreFromInitialContext is supposed to be atomic, this
 * is a big problem.  The fix is make the proxyMo resident during
 * creation.  Ideal fix is to move the creation of the default proxy
 * outside of restoreFromInitialContext.
 */
#include "RootMemoryObjectCache.h"
#endif


#ifdef __GNUC__
#pragma implementation
#endif

#ifdef SPARC
unsigned VirtualAddressCacheSize = 65536;
#else
unsigned VirtualAddressCacheSize = 0;
#endif

#ifdef INSTRUMENT_DOMAIN
#define Instr(exp) exp
#include "Instrument.h"
#include "InstrumentEvent.h"
extern class SystemInstrumentManager : public ;
extern SystemInstrumentManager *TheInstrumentManager;

TimerInstrument *iDomainRepairFault = NULL;
TimerInstrument *iDomainRepairFaultLookup = NULL;
TimerInstrument *iDomainRepairFaultMakeResident = NULL;
TimerInstrument *iDomainRepairFaultAddMapping = NULL;

void 
DomainConstructInstruments ()
{
    iDomainRepairFault = new TimerInstrument
    (TheInstrumentManager, "DomainRepairFault", AWAKE, 1, ALL);
    iDomainRepairFaultLookup = new TimerInstrument
    (TheInstrumentManager, "DomainRepairFaultLookup", AWAKE, 1, ALL);
    iDomainRepairFaultMakeResident = new TimerInstrument
    (TheInstrumentManager, "DomainRepairFaultMakeResident", AWAKE, 1, ALL);
    iDomainRepairFaultAddMapping = new TimerInstrument
    (TheInstrumentManager, "DomainRepairFaultAddMapping", AWAKE, 1, ALL);
}

void 
DomainDeleteInstruments ()
{
    delete iDomainRepairFault;
    iDomainRepairFault = NULL;
    delete iDomainRepairFaultLookup;
    iDomainRepairFaultLookup = NULL;
    delete iDomainRepairFaultMakeResident;
    iDomainRepairFaultMakeResident = NULL;
    delete iDomainRepairFaultAddMapping;
    iDomainRepairFaultAddMapping = NULL;
}

#else
#define Instr(exp)
void 
DomainConstructInstruments () {}
void 
DomainDeleteInstruments () {}
#endif

DomainCleaner::DomainCleaner( Domain * domain )
{
    _domain = domain;
}

DomainCleaner::~DomainCleaner()
{
    _domain->cleanup();
}

DomainTimer DomainCtor("Domain::Domain() timer");

Domain::Domain( Domain * parentDomain, char *start, unsigned len ) :
    _startAddress(start), _size(len)
{
    SInstr(unsigned start = DomainCtor.start());
    initDebug();
    _translation = new ADDRESS_TRANSLATION( (ADDRESS_TRANSLATION *)(parentDomain == 0 ? 0 : parentDomain->_translation) );
#ifdef DEBUG
    _debugMask = MaxObjectDebugMask;
    _debugMessages = MaxObjectDebugMessages;
#endif
    if (parentDomain == 0)
        maxMemoryObjectsInMap = KernelDomainMaxMemoryObjectsInMap;
    else
        maxMemoryObjectsInMap = UserDomainMaxMemoryObjectsInMap;
    list = new List[maxMemoryObjectsInMap];
    Assert(list != 0);
    numberOfMemoryObjectsInMap = 0;
    _last = numberOfMemoryObjectsInMap;
    unsigned  i;
    for( i = 0; i < maxMemoryObjectsInMap; i++ ) {
        list[i].baseAddr = NotMapped;
        list[i].highAddr = NotMapped;
        list[i].length = 0;
        list[i].memoryObjectView = 0;
    }

    /*
     * Need to reference direct parent so it won't go away while this
     *   domain is alive.  Don't need to reference the other ancestors
     *   becauseeach Domain references its own parent.  That is why
     *   the _parents array is not made up of Stars and this extra
     *   _parent is here.
     */
    _parent = parentDomain;
    
    if ( parentDomain == 0 ) {
        _nameServer = new NameServer( 0 );
        _fileSystemInterface = 0;
    }
    else {
        _nameServer = new NameServer( parentDomain->nameServer() );
        _fileSystemInterface = thisProcess()->nameServer()->lookup
        ( "fileSystemInterface", "FileSystemInterface" );
        if (_fileSystemInterface != 0) {
            _fileSystemInterface = new FileSystemInterface 
            ( _fileSystemInterface, 1 );
            _nameServer->bind ( "fileSystemInterface", 
                                _fileSystemInterface );
        }
    }

// Commenting out Proxy code
#ifdef FRANCIS
    ObjectProxy * proxyMemory;
    MemoryObjectStar proxymo;
    if ( parentDomain == 0 ) {
        /*
         * This memory never gets deleted but who cares because
         *   the kernel domain never gets deleted anyway.
         */
        proxyMemory = (ObjectProxy *) new char
        [ MAXPROXIESPERDOMAIN * sizeof( ObjectProxy ) ];
        proxymo = 0;
    }
    else
    {
        /*
         * Put the proxies in a readOnly section of this Domain.
         */
#ifdef HACKFIX
        proxymo = new SourceOfZerosMemoryObject;
        Assert( proxymo != 0 );
        unsigned size = MAXPROXIESPERDOMAIN * sizeof( ObjectProxy );
        proxymo->setNumberOfUnits( size );
        proxymo->setName("ProxyMo");
        proxymo->setCacheClass( RootMemoryObjectCacheClass );
        MemoryObjectCache * cache = proxymo->buildCache();
        Assert( cache != 0 );
        const char * base = thisProcess()->domain()->add
        ( proxymo, ReadWrite );
        thisProcess()->domain()->makeAddressable( base, size );
        thisProcess()->domain()->remove( proxymo, base );
        proxyMemory = (ObjectProxy *) add( proxymo, ReadOnly );
#else
        proxymo = TheKernel->allocateMemoryObjectForBackingStore();
        Assert( proxymo != 0 );
        proxymo->setName( "ProxyMo" );
        int proxyUnitSize = proxymo->unitSize();
        proxymo->setNumberOfUnits(
            ( MAXPROXIESPERDOMAIN * sizeof( ObjectProxy ) 
              + proxyUnitSize - 1 ) / proxyUnitSize );
        proxyMemory = (ObjectProxy *) add( proxymo, ReadOnly );
#endif // HACKFIX
    }

    _proxyManager = new ObjectProxyManager( MAXPROXIESPERDOMAIN,
                                            proxyMemory, proxymo, _nameServer );
    Assert( _proxyManager != 0 );
    _returnFunct = 0;
    _defaultProxy = 0;
#ifndef OLDDEFAULTPROXY
    if ( _parent != 0 ) {
        _defaultProxy = _proxyManager->allocate
        ( TheSystemInterface, SystemInterfaceClass, _parent, 0);
        _defaultProxy = _proxyManager->kernelToDomain( _defaultProxy );
    }
#endif

#endif // FRANCIS

    _cleaner = new DomainCleaner( this );

#ifdef MULTIMAX
    _moduleTableEntry = 0;
#endif
    SInstr(DomainCtor.stop(start));
}

/*
 * The destructor for the address translation will implicitly be called and
 * result in any active mappings dumping their ref/mod type information
 * into the corresponding PhysicallyAddressableUnits.
 * The superclass destructor does the rest of the work.
 */

DomainTimer DomainDtor("Domain::~Domain() timer");

Domain::~Domain()
{
    SInstr(int start = DomainDtor.start());
    Raid( RaidConstructor )( "%N::~Domain()\n", this );

    Assert( this != 0 );

    Assert( ! this->lock.held() );

    cleanup();

#ifdef MULTIMAX
    extern void FreeModuleTableEntry( Module * );

    if ( _moduleTableEntry != 0 )
        FreeModuleTableEntry( _moduleTableEntry );
#endif
    delete _translation;
    delete list;
    SInstr(DomainDtor.stop(start));
}

DomainTimer DomainCleanup("Domain::cleanup() timer");

extern ClassStar SpecialPurposeCacheClass;

void
Domain::cleanup( )
{
    SInstr(int start = DomainCleanup.start());
    _fileSystemInterface = 0;

    _nameServer = 0;
    if ( _proxyManager != 0 ) {
        /*
         * Free all the proxies before deleting the manager because the
         *   process of freeing them all could cause proxy calls, in
         *   particular when a proxy to the nameServer is freed.
         */
        _proxyManager->freeAll();
    }
    _proxyManager = 0;

    /*
     * Loop through and unreference all the memory objects 
     *   in the list before deleting it.  Remove from the end back
     *   to avoid having to shift all the objects down in remove().
     */
    if ( _parent != 0 ) {
        while ( numberOfMemoryObjectsInMap > 0 ) {
            remove( list[numberOfMemoryObjectsInMap - 1].
                    memoryObjectView,
                    list[numberOfMemoryObjectsInMap - 1].
                    baseAddr );
        }
    } 

    SInstr(DomainCleanup.stop(start));
}

ObjectProxy *
Domain::defaultProxy( )
{
    if ( _defaultProxy == 0 ) {
#ifndef OLDDEFAULTPROXY
        Assert( _parent == 0 );
#endif
        _defaultProxy = proxyManager()->allocate( TheSystemInterface,
                                                  SystemInterfaceClass, _parent, 0);
        _defaultProxy = proxyManager()->kernelToDomain( _defaultProxy );
    }
    return( _defaultProxy );
}

#ifdef MULTIMAX
void
Domain::setModuleTableEntry( void * starttext, void * startdata )
{
    extern Module * AllocateModuleTableEntry( void * programBase,
                                              void * staticBase, void * interrupt );
    extern void FreeModuleTableEntry( Module * );

    if ( _moduleTableEntry != 0 )
        FreeModuleTableEntry( _moduleTableEntry );

    _moduleTableEntry = AllocateModuleTableEntry( starttext, startdata, 0 );
    Assert( _moduleTableEntry != 0 );
}
#endif

void
Domain::setDataEqual( Domain * oldDomain )
{
#ifdef MULTIMAX
    if ( oldDomain->moduleTableEntry() != 0 ) {
        setModuleTableEntry(
            oldDomain->moduleTableEntry()->ProgramBasePointer,
            oldDomain->moduleTableEntry()->StaticBasePointer );
    }
#endif
#ifndef OLDDEFAULTPROXY
    Assert( _parent != 0 );
    Assert( _defaultProxy != 0 );
    _defaultProxy = _proxyManager->domainToKernel( _defaultProxy );
    _proxyManager->free( _defaultProxy );
#endif
    _returnFunct = oldDomain->_returnFunct;
    _nameServer->clone( oldDomain->_nameServer );
    _nameServer->bind( "fileSystemInterface", _fileSystemInterface );
    _proxyManager->clone( oldDomain->_proxyManager );
    _defaultProxy = oldDomain->_defaultProxy;
}

/*
 * These next one-line functions can't be inline because it uses 'getParent'
 *    which requires Kernel.h.  Unfortunately, Kernel.h includes Domain.h and
 *    making these inline in Domain.h would cause a circularity.
 */

DomainTimer DomainLookup("Domain::lookup() timer");

MemoryObjectViewRef
Domain::lookup( const char * vaddr, const char * & baseAddr, unsigned & length,
                ProtectionLevel & permissionDesired )
{
    SInstr(int start = DomainLookup.start());
    MemoryObjectViewRef view;
    lock.acquire();
    if ( validAddress( vaddr ) ) {
        view = basicLookup( vaddr, baseAddr, length,
                            permissionDesired );
    } else if ( _parent != 0 ) {
        view = _parent->basicLookup( vaddr, baseAddr, length, 
                                     permissionDesired );
    } else {
        /*
         * special case for premapped addresses in root domain 
         */
        view = basicLookup( vaddr, baseAddr, length,
                            permissionDesired );
    }
    lock.release();
    SInstr(DomainLookup.stop(start));
    return view;
}

DomainTimer DomainRepairInst("Domain::repairFault() timer");

FaultErrorCode
Domain::repairFault( const char * address, AccessType attemptedAccess )
{
    SInstr(unsigned start = DomainRepairInst.start());
    FaultErrorCode ret = AccessOfUnmanagedMemory;
    if ( validAddress(address) )
        ret = basicRepairFault( address, attemptedAccess );
    else if ( _parent != 0 )
        ret = _parent->basicRepairFault( address, attemptedAccess );
    else {

        /* 
         * special case for root domain, it would only occur once in
         * in a blue moon if it even occurs, but it should be checked
         * just to be safe
         */
        const char * baseAddr;
        unsigned length;
        ProtectionLevel protection;
        MemoryObjectViewStar view = basicLookup
        ( address, baseAddr, length, protection );
        if ( view != 0 )
            ret = basicRepairFault( address, attemptedAccess );
    
    }
    
    SInstr(DomainRepairInst.stop(start));
    return ret;
}

/*
 * Add a memory object at a specified address.
 */

DomainTimer DomainAddA("Domain::add(MO *, char *) timer");
  
const char *
Domain::add( MemoryObject * memobj, const char * address,
             ProtectionLevel perms, int shared )
{
    SInstr(int start = DomainAddA.start());
    Raids( RaidMethod ) << this << "::add( memobj:" << memobj 
                        << " address:"
                        << asAddress( address ) 
                        << " perms:" << perms << " )\n" << eor; 

    Assert( this != 0 );
    Assert( memobj != 0 );
    Assert( memobj->numberOfUnits() > 0 );
    Assert( numberOfMemoryObjectsInMap < maxMemoryObjectsInMap );
    Assert( ( (unsigned) address % PhysicalPageSize ) == 0 );

#ifdef NEVER
#ifdef VCHOICES
    /*
     * In VChoices the 1-1 mapped store may not be a
     * valid address.
     * So don't assert that, but if it's not valid make sure
     * it's pre-mapped.
     */
    if ( !validAddress(address) ) {
        Assert ( memobj->isKindOf( DummyMemoryObjectClass ));
        return( address );
    }
#endif
#endif

    /*
     * Allow at most what access the memory object itself is willing
     * to allow.
     */
    if( perms > memobj->logicalProtectionLevel() ) {

        // eventually pass an indication this happenened back out!!

        Console() << this << "::add() Weakened Protection\n" << eor;
        perms = memobj->logicalProtectionLevel();
    }

    /*
     * If not view create a view for it
     */
    MemoryObjectView * view = 0;
    if ( !memobj->isKindOf( MemoryObjectViewClass ) ) {
        if (memobj->isKindOf( DummyMemoryObjectClass )) {
            view = ((DummyMemoryObject *)memobj)->view();
        }
        else {
            view = new NormalMemoryObjectView( memobj );
        }
    } else {
        view = ( MemoryObjectView * ) memobj;
    }
    Assert( view != 0 );

    lock.acquire();

    /*
     * Decide where to insert it. Keep the list sorted by address.
     */
    unsigned newSlot = numberOfMemoryObjectsInMap;
    for( unsigned i = 0; i < numberOfMemoryObjectsInMap; i++ ) {
        /*
         * If the address is less than the current thing in this slot,
         * move everything forward and put it here.
         */
        if( address < list[i].baseAddr ) {
            newSlot = i;
            for( unsigned j = numberOfMemoryObjectsInMap; j > i; j-- ) {
                list[j] = list[j-1];
            }
            break;
        }
    }

    /*
     * Insert at newSlot.
     */

    Raids( RaidMethod ) << this << "::add: putting it in slot " << newSlot
                        << "\n" << eor;
    list[newSlot].baseAddr = address;
    list[newSlot].memoryObjectView = view; 
    view->reference();
    list[newSlot].length = memobj->numberOfUnits() * memobj->unitSize();
    list[newSlot].highAddr = address + list[newSlot].length - 1;
    list[newSlot].access = perms;
    list[newSlot].shared = shared;
    _last = newSlot;

    Assert( ( newSlot == 0 ) || ( address > list[newSlot-1].highAddr ) );
    Assert( ( newSlot == numberOfMemoryObjectsInMap ) ||
            ( list[newSlot].highAddr < list[newSlot+1].baseAddr ) );

    numberOfMemoryObjectsInMap++;
    lock.release();
    view->buildCache();
    SInstr(DomainAddA.stop(start));
    return( address );
}

/*
 * Add a memory object at a random address. Makes sure the protection
 * doesnt exceed the max the MemoryObject is willing to allow.
 */

DomainTimer DomainAdd("Domain::add(MO *) timer");

const char *
Domain::add( MemoryObject * memobj, ProtectionLevel perms, int shared )
{
    SInstr(int start = DomainAdd.start());
    Raid( RaidMethod )( "%N::add( %N, protection %d )\n",
                        this, memobj, perms );
    Assert( this != 0 );
    Assert( memobj != 0 );
    Assert( memobj->numberOfUnits() > 0 );
    Assert( numberOfMemoryObjectsInMap < maxMemoryObjectsInMap );

    unsigned alignment = PhysicalPageSize;
    unsigned alignmentMask = alignment - 1;

    /*
     * Allow at most what access the MemoryObject itself is willing
     * to allow.
     */
    if( perms > memobj->logicalProtectionLevel() ) {

        // eventually pass an indication this happenened back out!!
        Console() << this
                  << "::add() Weakened Protection\n" << eor;

        perms = memobj->logicalProtectionLevel();
    }

    /*
     * If not view create a view for it
     */
    MemoryObjectView * view = 0;
    if ( memobj->isKindOf( MemoryObjectViewClass ) ) {
        view = ( MemoryObjectView * ) memobj;
    } else {
        if (memobj->isKindOf( DummyMemoryObjectClass )) {
            view = ((DummyMemoryObject *)memobj)->view();
        } else {
            view = new NormalMemoryObjectView( memobj );
        }
    }
    Assert( view != 0 );

    unsigned int newMOSize = view->unitSize() * view->numberOfUnits();

    const char * startAddress = _startAddress;
    const char * endAddress = _startAddress + _size - 1;
    const char * address = startAddress;
    const char * lowAddress = startAddress;

    unsigned vaCacheAlignment = 0;
    unsigned vaCacheMask = 0;
    if ( VirtualAddressCacheSize != 0 ) {
        vaCacheMask = VirtualAddressCacheSize - 1;
        vaCacheAlignment = view->viewOffset() & vaCacheMask;
        Assert( ( vaCacheAlignment & alignmentMask ) == 0 );
        address = ( const char * )
        ( ( (unsigned) lowAddress & ~vaCacheMask ) |
          vaCacheAlignment );
        if ( address < lowAddress ) 
            address += VirtualAddressCacheSize;
    }
    const char * highAddress = address + ( newMOSize - 1 );

    // Assert( ! lock.heldByMe() );
    lock.acquire();

    /*
     * Decide where to insert it. Keep the list sorted by address.
     */
    int newSlot = numberOfMemoryObjectsInMap;
    for( unsigned i = 0; i < numberOfMemoryObjectsInMap; i++ ) {

        Raids( RaidDetail ) << this << "::add: slot " << i
                            << " something at "
                            << asAddress( list[i].baseAddr ) << " -- "
                            << asAddress( list[i].highAddr ) << "\n" << eor;

        /*
         * We find the slot based on the addrs of the existing MO's
         * and the 1-1 MO's may not really be at the right addrs.
         */
        if ( !validAddress(list[i].baseAddr) )
            continue;

        /*
         * See if there is a hole big enough here. 
         */

        lowAddress = (const char *) ( (unsigned int)
                                      ( list[i].highAddr + 1 + alignmentMask )
                                      & ~alignmentMask );
           
        const char * nextBase;
        if ( i + 1 < numberOfMemoryObjectsInMap )
            nextBase = list[i+1].baseAddr;
        else
            nextBase = endAddress;

        /*
         * Try to perform virtual address cache alignment
         */

        if ( VirtualAddressCacheSize == 0 ) {
            address = lowAddress;
        } else {
            address = ( const char * )
            ( ( (unsigned) lowAddress & ~vaCacheMask ) |
              vaCacheAlignment );
            if ( address < lowAddress ) 
                address += VirtualAddressCacheSize;
        }
        
        /*
         * check if hole is big enough
         */

        highAddress = address + ( newMOSize - 1 );
        if ( ( address < lowAddress ) || /* wrap */
             ( address >= nextBase ) ||
             ( highAddress >= nextBase ) ) {
            /* hole not big enough */
            continue;
        }

        /*
         * Found a good spot for it.
         */

        Raids( RaidDetail ) << this << "::add: address at "
                            << asAddress( address ) << " -- " 
                            << asAddress( address + newMOSize - 1 ) << "\n" << eor;

        /*
         * FIX:
         *  Maybe do a memoryMap->lookup() to see
         *  whether it's there.
         */

        newSlot = ++i;
        for( unsigned  j = numberOfMemoryObjectsInMap; j > i; j-- ) {
            list[j] = list[j-1];

            Raids( RaidDetail ) << this << "::add: slot " << j
                                << " now has "
                                << asAddress( list[j].baseAddr ) << " -- "
                                << asAddress( list[j].highAddr ) << "\n" << eor;

        }
        break;

    }

    /*
     * Now, insert it at newSlot.
     */

    /* worry about wrap-around */

    Assert( address < endAddress );
    Assert( newSlot == 0 || address > list[newSlot-1].highAddr );
    Assert( highAddress >= address );
    Assert( highAddress <= endAddress );

    Raids( RaidMethod ) << this << "::add: putting it at "
                        << asAddress( address )
                        << " slot " << newSlot << "\n" << eor;

    list[newSlot].baseAddr = address;
    list[newSlot].memoryObjectView = view;
    view->reference();
    list[newSlot].length = newMOSize;
    list[newSlot].highAddr = highAddress;
    list[newSlot].access = perms;
    list[newSlot].shared = shared;
    _last = newSlot;

    numberOfMemoryObjectsInMap++;
    lock.release();
    view->buildCache();
    SInstr(DomainAdd.stop(start));
    return( address );
}

/*
 * Remove a MemoryObject, destroy any translation mappings corresponding to
 * it and return the base address of the block of memory it occupied. This
 * just removes a "randomly selected" instance of the MemoryObject if the
 * MemoryObject has been added multiple times. It might be necessary to call
 * this until it returns NoMapped in order to guarantee the MemoryObject is
 * not mapped into the Domain anywhere.
 */

DomainTimer DomainRemove("Domain::remove() timer");

const char *
Domain::remove( MemoryObject * memobj, const char * baseAddr )
{
    SInstr(int start = DomainRemove.start());
    Raids( RaidMethod ) << this << "::remove( memobj:" << memobj 
                        << ", " << asHex( (unsigned) baseAddr )
                        << " )\n" << eor;
    Assert( memobj != 0 );

    // Assert( ! lock.heldByMe() );
    lock.acquire();

    /*
     * Search through the list for the MemoryObject to remove
     * and compact the list when it is found.
     */
    for( unsigned i = 0; i < numberOfMemoryObjectsInMap; i++ ) {
        if( list[i].baseAddr == baseAddr &&  
            ( list[i].memoryObjectView == memobj ||
              list[i].memoryObjectView->source() == memobj ) ) {
            /*
             * Found it so compact the list.
             */
            if ( _last >= i ) 
                _last--;
            const char * where = list[i].baseAddr;
            // UNUSED ?? int sizeOfBlock = list[i].length;
            Raids( RaidDetail ) << this << "::remove: at slot "
                                << i << "\n" << eor;

            MemoryObjectView * view = list[i].memoryObjectView;

            unsigned j;
            for( j = i;
                 j < numberOfMemoryObjectsInMap - 1; j++ ) {
                list[j] = list[j+1];
            }
            list[j].memoryObjectView = 0;
            
            numberOfMemoryObjectsInMap--;

            /*
             * Remove any address translation mapping corresponding
             *   to the memory previously occupied by the memory
             *   object.
             */

            lock.release();

            if ( view->cache() != 0 )
                view->cache()->untranslatable
                ( _translation, view, where );
            
            Raids( RaidDetail ) << this << "::remove: return( "
                                << asAddress( where ) << " )\n" << eor;
            view->unreference();

            SInstr(DomainRemove.stop(start));
            return( where );
        }
    }

    /*
     * If we get here it wasn't even in the list.
     */
    lock.release();

    if ( _parent != 0 ) {
        /* 
         * Try the parent if it's not in our list.
         */
        const char * code = _parent->remove( memobj, baseAddr );
        SInstr(DomainRemove.stop(start));
        return( code );
    }

    Console() << this << "::remove( memobj:"
              << memobj << " ): it wasnt there\n" << eor;
    SInstr(DomainRemove.stop(start));
    return( NotMapped );
}

/*
 * Look through the list and find which if any memory object "contains" an
 * address at which offset. Return both the memory object and the offset.
 * MAKE THIS AS FAST AS POSSIBLE!!! 
 */

DomainTimer DomainbasicLU("Domain::basicLookup() timer");

int domainLookupHit = 0;
int domainLookupCount = 0;

MemoryObjectViewRef
Domain::basicLookup( const char * address, const char * & baseAddr,
                     unsigned & length, ProtectionLevel & prot )
{
    SInstr(int start = DomainbasicLU.start());
    Raids( RaidMethod ) << this << "::lookup( address:" <<
    asAddress( address ) << " )\n" << eor;
    Assert( this != 0 );

    domainLookupCount++;
    if ( _last < numberOfMemoryObjectsInMap ) {
        if( address >= list[_last].baseAddr &&
            address <= list[_last].highAddr ) {
            baseAddr = list[_last].baseAddr;
            length = list[_last].length;
            prot = list[_last].access;
            MemoryObjectViewRef memobj = 
            list[_last].memoryObjectView;
            domainLookupHit++;
            Raids( RaidDetail ) << this << "::lookup: return( "
                                << memobj << " )\n" << eor;
            SInstr(DomainbasicLU.stop(start));
            return( memobj );
        }
    }

    int mid;
    int low = 0;
    int high = numberOfMemoryObjectsInMap - 1;

    while( low <= high ) {
        mid = (low + high) >> 1;
        if( address < list[mid].baseAddr ) {
            high = mid - 1;
        }
        else if( address > list[mid].highAddr ) {
            low = mid + 1;
        }
        else {
            baseAddr = list[mid].baseAddr;
            length = list[mid].length;
            prot = list[mid].access;
            MemoryObjectViewRef memobj = 
            list[mid].memoryObjectView;
            _last = mid;
            Raids( RaidDetail ) << this << "::lookup: return( "
                                << memobj << " )\n" << eor;
            Assert( memobj != (MemoryObject*)0 );
            SInstr(DomainbasicLU.stop(start));
            return( memobj );
        }
    }

    Raids( RaidMethod ) << this << "::lookup: return( 0 )\n" << eor;
    SInstr(DomainbasicLU.stop(start));
    return( 0 );
}

/*
 * Handle a memory fault at a specified address.
 */

DomainTimer DomainBasicRF("Domain::basicRepairFault() timer");

FaultErrorCode
Domain::basicRepairFault( const char * address, AccessType attemptedAccess )
{
    SInstr(int start = DomainBasicRF.start());
    Instr (iDomainRepairFault->start ());

    Assert( this != 0 );

    /*
     * Find the MemoryObject managing the faulting address, 
     */

    const char * baseAddr;
    unsigned size;
    ProtectionLevel staticProtection;

    Instr (iDomainRepairFaultLookup->start ());
    MemoryObjectViewStar memobj = basicLookup( address, baseAddr, size, 
                                               staticProtection );
    int offset = address - baseAddr;
    Instr (iDomainRepairFaultLookup->stop (Domain_RepairFaultLookupEvent));

    Raids( RaidDetail ) << this << "::repairFault: memobj:" << memobj 
                        << " offset:"
                        << offset << " prot:" << staticProtection << "\n" << eor;

#ifdef NEVER
    switch( attemptedAccess ) {
        case ApplicationRead:
        case ApplicationWrite:
        {
            /*
             * check for user fault handler
             */
            Process * thisProc = thisProcess();
            FaultHandlerStar handler = thisProc->faultHandler();
            if ( handler != 0 ) {
                /*
                 * don't call the user fault handler again if
                 *   it causes another fault.
                 */
                thisProc->setFaultHandler( 0 );
                handler->fault( address, attemptedAccess );
                thisProc->setFaultHandler( handler );
            }
        }
        break;
        default:
            break;
    }
#endif

    /*
     * If there is no managing object, give up.
     */
    if( memobj == 0 ) {
        Instr (iDomainRepairFault->stop (Domain_RepairFaultEvent,
                                         "Access of Unmanaged Memory"));
        SInstr(DomainBasicRF.stop(start));
        return( AccessOfUnmanagedMemory );
    }

    /*
     * Based on the access attempted, determine the minimum
     * ProtectionLevel necessary to fullfill the request. At the same
     * time, check for violations of the static protection level.
     * I.e., the required ProtectionLevel to complete the request
     * exceeds the maximum ProtectionLevel the *Domain* is
     * willing to allow. Remember, in ::add we made sure the Domain
     * would never allow more than the MemoryObject itself would otherwise
     * allow.
     */

    ProtectionLevel requiredProtection;
    switch( attemptedAccess ) {
        case SystemRead:
        case SystemWrite:
            /*
             * 'NoAccess' actually means kernel read/write, no access
             *    by user.
             */
            requiredProtection = NoAccess;
            break;

        case ApplicationRead:
            if( staticProtection == NoAccess ) {
                Instr(iDomainRepairFault->stop (
                          Domain_RepairFaultEvent,
                          "Read of Protected System Memory"));
                SInstr(DomainBasicRF.stop(start));
                return( ReadOfProtectedSystemMemory );
            }
            requiredProtection = ReadOnly;
            break;

        case ApplicationWrite:
            if( staticProtection == NoAccess ) {
                Instr (iDomainRepairFault->stop (
                           Domain_RepairFaultEvent,
                           "Write of Protected System Memory"));
                SInstr(DomainBasicRF.stop(start));
                return( WriteOfProtectedSystemMemory );
            }
            if( staticProtection == ReadOnly ) {
                Instr (iDomainRepairFault->stop (
                           Domain_RepairFaultEvent,
                           "Write of Read-only Memory"));
                SInstr(DomainBasicRF.stop(start));
                return( WriteOfReadOnlyMemory );
            }
            requiredProtection = ReadWrite;
            break;

        default:
            requiredProtection = NoAccess;  // shut up warning
            Assert( NOTREACHED );
            break;
    }

#ifdef NEVER 
#ifdef SPARC 
    /* 
     * BUG:
     *
     * This is to solve the ldstub (i.e. test and set) problem.
     * On the SPARC, a ldstub only generates a read fault, actually
     * a write fault is required.  If the minimal read protection
     * is granted then it will fault again when the "set" takes place.
     * This fix solves the problem only when ldstub is used in a
     * read/write view, it will cause infinite loops when used in a
     * read-only view.
     *
     * However, there is an additional problem here.  If I do this
     * copy on write will not work efficiently since read accesses
     * are also considered to be write accesses and copied.
     */
    requiredProtection = staticProtection;
#endif
#endif

    /*
     * Have the managing memory object bring the data into
     * physical memory. (I'd rather it took a unit number and not
     * an offset as the argument... --Vince)
     */
    Instr (iDomainRepairFaultMakeResident->start());
    Raids( RaidDetail ) << this << "::repairFault: make offset "
                        << offset << " resident in " << memobj << "\n" << eor;
    FaultErrorCode code;
    ProtectionLevel grantedProtection;
    LockHolder * lockHolder = makeResidentInPhysicalMemory
    ( memobj, baseAddr, size, offset, 1, requiredProtection,
      grantedProtection, code, _translation, address, 
      DomainRepairFault );
    Assert( lockHolder == 0 );
    Instr (iDomainRepairFaultMakeResident->stop(Domain_RepairFaultEvent));

    // returning the protection in the PAU's is gross.


    /*
     * If the MemoryObject said no, there is nothing more we can do.
     */
    if( grantedProtection < requiredProtection ) {
        if( grantedProtection == NoAccess ) {
            Instr(iDomainRepairFault->stop (
                      Domain_RepairFaultEvent,
                      "Dynamic Access Violation"));
            SInstr(DomainBasicRF.stop(start));
            return( DynamicAccessViolation );
        }
        Instr(iDomainRepairFault->stop (
                  Domain_RepairFaultEvent,
                  "Dynamic Write of Read-only Memory"));
        SInstr(DomainBasicRF.stop(start));
        return( DynamicWriteOfReadOnlyMemory );
    }

    /*
     * We'll give the minimum of the ProtectionLevel granted by the
     * MemoryObject (via the PAU) and that allowed by the Domain.
     * I'm not sure a MemoryObject would ever give us more than we asked for
     * but its important enough to check.
     */
    if( grantedProtection > staticProtection ) {
        grantedProtection = staticProtection;
    }

    /*
     * Keep track of code 
     */
    if ( ProcessResourceUsageOn ) {
        Process * proc = thisProcess();
        if ( code == MissingPhysicalMemory ) {
            proc->usage()->majorFault++;            
        } else {
            proc->usage()->minorFault++;
        }
    }

    /*
     * Return the FaultErrorCode.
     */
    Instr(iDomainRepairFault->stop (Domain_RepairFaultCode));
    SInstr(DomainBasicRF.stop(start));
    return( code );
}

/*
 * Force the address range to be cache.
 * This make a lot of assumptions. It's only used by allocators that
 * know they are ok. (I hope!)
 * Come back and fix it later...
 *
 * I seems there might be locking problems and races with this and
 *    constructChain..
 */

DomainTimer DomainMakeAddr("Domain::makeAddressable() timer");

void
Domain::makeAddressable( const char * address, unsigned int length )
{
    SInstr(int start = DomainMakeAddr.start());
    Raids( RaidMethod ) << this << "::makeAddressable( address:" 
                        << asAddress( address ) << " length:" << length << " )\n" 
                        << eor;
    
    Assert( this != 0 );
    
    const char * baseAddr;
    unsigned size;
    ProtectionLevel protection;

    lock.acquire();
    MemoryObjectViewStar memobj = 0;
    if ( validAddress( address ) ) {
        memobj = basicLookup( address, baseAddr, size, protection );
    } else {
        if ( _parent == 0 ) {
            /*
             * special case for premapped addresses in root domain 
             */
#ifdef VCHOICES
            if ( !validAddress(address) ) {
                // should be a 1-1 MO
                lock.release();
                SInstr(DomainMakeAddr.stop(start));
                return;
            }
#endif
            memobj = basicLookup
            ( address, baseAddr, size, protection );
        } else {
            Assert (_parent != 0);
            lock.release();
            SInstr(DomainMakeAddr.stop(start));
            _parent->makeAddressable( address, length );
            return;
        }
    }
    
    Assert( memobj != 0 ); /* actually it means AccessOfUnmanagedMemory */
    
    int offsetIntoObject = address - baseAddr;
    
    /*
     * Make sure the range is managed by a single MemoryObject for now,
     * since chains currently cannot span MemoryObjects. It might be
     * possibly to remove this restriction later.
     */
    Assert(memobj->isIn
           (memobj->offsetToUnit(offsetIntoObject + length - 1)));
                    
    FaultErrorCode code;
    ProtectionLevel grantedProtection;
    lock.release();
    LockHolder * lockHolder =
    makeResidentInPhysicalMemory(memobj, baseAddr, size,
                                 offsetIntoObject, length,
                                 protection, 
                                 grantedProtection, code,
                                 _translation, address, 
                                 DomainMakeAddressable);
    Assert( lockHolder == 0 );
    SInstr(DomainMakeAddr.stop(start));
}

/*
 * Bring a whole range of virtual addresses into physical memory.
 * (Used to fill in the fields of a PhysicalMemoryChain.)
 */

DomainTimer DomainCChain("Domain::constructChain() timer");

PhysicalMemoryChain
Domain::constructChain( const char * address, unsigned int length, 
                        int makeAddr )
{
    SInstr(int start = DomainCChain.start());

    PhysicalMemoryChain chain;
    Raids( RaidMethod ) << this << "::constructChain( address:" 
                        << asAddress( address )
                        << " length:" << length << " )\n" << eor;
    Assert( this != 0 );

    const char * baseAddr;
    unsigned size;
    ProtectionLevel protection;

    MemoryObjectViewStar memobj = 0;
    if ( validAddress( address ) || _parent == 0 ) { 
        /* parent == 0,
           special case for premapped address in root domain */
        memobj = basicLookup( address, baseAddr, size, protection );
    } else {
        /* 
         * WARNING:
         *
         * cannot do:
         *
         *  return _parent->constructChain( address, length );
         *
         * due to compiler bug, the destructor will not be 
         * called resulting in diggling reference.
         */
        SInstr(DomainCChain.stop(start));
        PhysicalMemoryChain pChain = 
        _parent->constructChain ( address, length, makeAddr );
        return pChain;
    }

    Assert( memobj != 0 ); /* actually it means AccessOfUnmanagedMemory */

    Raids( RaidDetail ) << this << "::constructChain from " 
                        << memobj << "\n" << eor;

#ifdef VIRTUALADDRESSIO

    if ( _parent == 0 ) {
        extern ClassStar LockedMemoryObjectViewClass;
        extern ClassStar PagedMemoryObjectCacheClass;
        Assert( memobj->classOf() == LockedMemoryObjectViewClass ||
                memobj->cache()->classOf() != 
                PagedMemoryObjectCacheClass );
        /* this is not necessary, but a performance problem,
         * so fail assertion so we know when it occurs 
         */
        Assert( NOTREACHED ); 
        SInstr(DomainCChain.stop(start));
        return( chain );
    }

#else // not VIRTUALADDRESSIO

    int offsetIntoObject = address - baseAddr;

    FaultErrorCode code;
    ProtectionLevel grantedProtection;
    CacheReason reason;
    if ( makeAddr ) {
        reason = (CacheReason)(DomainConstructChain | DomainMakeAddressable);
    } else {
        reason = DomainConstructChain;
    }

    SInstr(start = DomainCChain.suspend(start));
    LockHolder * lockHolder =
    makeResidentInPhysicalMemory(memobj, baseAddr, size,
                                 offsetIntoObject, length,
                                 protection, grantedProtection, code,
                                 _translation, address, reason );
    SInstr(start = DomainCChain.resume(start)); 
    if ( lockHolder == 0 ) {
        SInstr(DomainCChain.stop(start));
        return( chain );
    }

    /*
     * Make sure the range is managed by a single MemoryObject for now,
     * since chains currently cannot span MemoryObjects. It might be
     * possibly to remove this restriction later.
     */
    Assert( memobj->isIn(
                memobj->offsetToUnit( offsetIntoObject + length - 1 ) ) );

    /*
     * Convert the byte offset into the view into the byte offset
     * into the unit being cached.
     */
    MemoryObjectCache * cache = memobj->cache();
    int delta = offsetIntoObject & ( ( 1 << cache->log2UnitSize() ) - 1 );
    Assert( delta < cache->unitSize() );

    chain = PhysicalMemoryChain( lockHolder, 0 );
    chain = PhysicalMemoryChain( chain, delta, length );

#endif // VIRTUALADDRESSIO

    SInstr(DomainCChain.stop(start));
    return chain;
}

/*
 * Set one Domain equal to another. Copy all writable sections.
 */

DomainTimer DomainCopy("Domain::copy() timer");

void
Domain::copy( Domain * other )
{
#ifdef DEBUG_COPY
    _debugMask = 63;
    _debugMessages = 32767;
#endif // DEBUG_COPY

    SInstr(unsigned start = DomainCopy.start());
    Raids( RaidMethod ) << this << "::copy( other:" << other << " )\n" 
                        << eor;
    Assert( this != 0 );
    Assert( other != 0 );
    Assert( other != this );

    /* 
     * Assume Proxy memory object is always the first in domain
     */

    Assert( strcmp(list[0].memoryObjectView->source()->objectName(),
                   "ProxyMo") == 0 );

    numberOfMemoryObjectsInMap =
    other->numberOfMemoryObjectsInMap;
    for( unsigned i = 0; i < numberOfMemoryObjectsInMap; i++ ) {
        
        MemoryObjectView * memObjView;
        MemoryObjectView * oldMemObjView = 
        other->list[i].memoryObjectView;
        Raids( RaidDetail ) << this << "::copy() "
                            << " examining " << oldMemObjView << "\n" << eor; 
        
        if( other->list[i].access == ReadWrite  &&
            !(other->list[i].shared) ) {

            MemoryObject * oldMemObj =
            oldMemObjView->source();
            
            Raids( RaidDetail ) << this 
                                << "::copy() oldMO "
                                << oldMemObj << " unitsz "
                                << oldMemObj->unitSize() << ", " 
                                << oldMemObj->numberOfUnits() 
                                << " units\n" << eor;

            PersistentMemoryObjectStar memObj;

            if( oldMemObj->isKindOf(PersistentMemoryObjectClass) ){

                memObj = TheKernel->
                allocateMemoryObjectForBackingStore();
                Assert( memObj != 0 );
                Assert( memObj->classOf() == 
                        oldMemObj->classOf() );

                Raids( RaidDetail ) << this << "::copy() "
                                    << "copying to "
                                    << memObj << "\n" << eor;

                oldMemObj->copy( memObj );

                Raids( RaidDetail ) << this << "::copy() " 
                                    << "new unitsz: " 
                                    << memObj->unitSize()
                                    << " new units: " 
                                    << memObj->numberOfUnits()
                                    << " new bytes: " 
                                    << memObj->numberOfBytes() 
                                    << "\n" << eor;
                Assert( memObj->unitSize() ==
                        oldMemObj->unitSize() );

            } else {

                Assert( oldMemObj->unitSize() == 1 );
                memObj = new SourceOfZerosMemoryObject;
                Assert( oldMemObj->classOf() ==
                        memObj->classOf() );
                Raids( RaidDetail ) << this << "::copy() "
                                    << "copying to " << memObj 
                                    << "\n" << eor;
                int size = oldMemObj->numberOfUnits();
                memObj->setNumberOfUnits( size );
                MemoryObjectCache * oldMemCache = 
                oldMemObj->cache();
                Assert( oldMemCache != 0 );
                memObj->setCacheClass
                ( oldMemCache->classOf() );
                MemoryObjectCache * cache = 
                memObj->buildCache();
                Assert( cache != 0 );
                Assert( oldMemCache->classOf() ==
                        cache->classOf() );

                const char * base = TheKernel->domain()->add
                ( memObj, NoAccess );
                Assert( oldMemObjView->startUnit() == 0 );
                Raids( RaidDetail ) << this << "::copy() "
                                    << "byte copying\n" << eor;
                ByteCopy( ( char * ) other->list[i].baseAddr,
                          ( char * ) base,
                          oldMemObj->numberOfUnits() );
                TheKernel->domain()->remove( memObj, base );
            }

            Raids( RaidDetail ) << this 
                                << "::copy() making view\n" << eor;
            Assert( memObj != 0 );
            memObj->setName( oldMemObj->objectName() );
            memObjView = new NormalMemoryObjectView
            ( memObj, 
              oldMemObjView->numberOfUnits(),
              oldMemObjView->log2UnitSize(),
              oldMemObjView->startUnit() );
        }
        else {
            // skip the proxy mo
            if ( other->list[i].baseAddr == _startAddress )
                continue;

            Raids( RaidDetail ) << this
                                << "::copy() read-only " 
                                << oldMemObjView << "\n" << eor;    

            memObjView = oldMemObjView;
        }

        list[i].baseAddr = other->list[i].baseAddr;
        list[i].highAddr = other->list[i].highAddr;
        list[i].length = other->list[i].length;
        list[i].access = other->list[i].access;
        list[i].shared = other->list[i].shared;
        Assert( list[i].memoryObjectView == 0 );
        list[i].memoryObjectView = memObjView;
        memObjView->reference();
    }
    SInstr(DomainCopy.stop(start));

    Raids( RaidMethod ) << this << "::copy() done\n" << eor;

#ifdef DEBUG_COPY
    initDebug();
#endif // DEBUG_COPY
}

/*
 * Prepare domain for exec, clear all memory objects except for 
 * proxy memory object
 */

DomainTimer DomainPrepareForExec("Domain::prepareForExec() timer");

void 
Domain::prepareForExec()
{
    SInstr(unsigned start = DomainPrepareForExec.start());
    Raid( RaidConstructor )( "%N::prepareForExec()\n", this );

    Assert( this != 0 );

    Assert( ! this->lock.held() );

    /* 
     * Kernel domains cannot be exec'ed
     */

    Assert( _parent != 0 );

    /* 
     * Assume Proxy memory object is always the first in domain
     */

    Assert( strcmp(list[0].memoryObjectView->source()->objectName(),
                   "ProxyMo") == 0 );
    /*
     * Loop through and unreference all the memory objects 
     *   in the list before deleting it.  Remove from the end back
     *   to avoid having to shift all the objects down in remove().
     */
    while ( numberOfMemoryObjectsInMap > 1 )
        remove( list[numberOfMemoryObjectsInMap - 1].memoryObjectView,
                list[numberOfMemoryObjectsInMap - 1].baseAddr );

    SInstr(DomainPrepareForExec.stop(start));
}

DomainTimer DomainMakeRes("Domain::makeResidentInPhysicalMemory() timer");

LockHolder *
Domain::makeResidentInPhysicalMemory
( MemoryObjectView * memobj, const char * baseAddr, unsigned size, 
  unsigned int offset, unsigned int length, 
  ProtectionLevel requestedProtection,
  ProtectionLevel & grantedProtection,
  FaultErrorCode & code, AddressTranslation * adt, 
  const char * vaddr, CacheReason reason )
{
    SInstr(unsigned start = DomainMakeRes.start());
    LockHolder * lockHolder;

    /*
     * Figure out our MemoryObjectCache.
     */

    MemoryObjectCache * cache = memobj->cache();
    if ( cache == 0 ) {
        code = AccessOfUnmanagedMemory;
        lockHolder = 0;
    } else {
        unsigned cacheUnit = memobj->offsetToCacheUnit( offset );
        Assert( cache->isIn( cacheUnit ) );
        unsigned int endUnit = 
        memobj->offsetToCacheUnit( offset + length - 1 );
        Assert( cache->isIn( endUnit ) );
        int nunits = endUnit - cacheUnit + 1;
        Assert( nunits > 0 );

#ifdef UNIX
        Assert( offset == offset );
        Assert( length == length );
        Assert( requestedProtection == requestedProtection );
        Assert( code == code );
        lockHolder = 0;
#else

        if ( nunits == 1 )
            lockHolder = 
            cache->cache( cacheUnit, requestedProtection, 
                          grantedProtection, code, 
                          adt, vaddr, 
                          reason, memobj, baseAddr );
        else
            lockHolder = 
            cache->mcache( cacheUnit, nunits, 
                           requestedProtection,
                           grantedProtection, code, 
                           adt, vaddr, 
                           reason, memobj, baseAddr );
                   
        Raids( RaidDetail ) << this 
                            << "::makeResidentInPhysicalMemory: cached offset "
                            << offset << "\n" << eor;

    }
#endif /*UNIX*/
    SInstr(DomainMakeRes.stop(start));
    return lockHolder;
}

/*
 * Force the address range to be uncached
 * The region will be narrowed to fit page boundaries.
 * Address will be rounded upward to fit a page, and length will be
 * rounded down. Any part of the area exceeding page boundaries will
 * be ignored. Thus, if length < pageSize, nothing will happen
 */

DomainTimer DomainMakeUnAddr("Domain::makeUnAddressable() timer");

void
Domain::makeUnAddressable( const char * address, unsigned int length )
{
    SInstr(unsigned start = DomainMakeUnAddr.start());
    Raids( RaidMethod ) << this << "::makeUnAddressable( address:" 
                        << asAddress( address ) << " length:" << length << " )\n" 
                        << eor;
    Assert( this != 0 );

    // Narrow the region to fit pages

    address = CeilingPage(address);
    length = FloorPage(length);

    if (length != 0) {

        const char * baseAddr;
        unsigned size;
        ProtectionLevel protection;
        
        MemoryObjectViewStar memobj = lookup( address, baseAddr, size,
                                              protection );
        Assert( memobj != 0 );
        int offsetIntoObject = address - baseAddr;
        
        /*
         * Make sure the range is managed by a single MemoryObject 
         * for now, since chains currently cannot span MemoryObjects. 
         * It might be possibly to remove this restriction later.
         */
        Assert( memobj->isIn
                (memobj->offsetToUnit( offsetIntoObject+length-1 )) );
        /*
         * Figure out our MemoryObjectCache.
         */
        MemoryObjectCacheStar cache = memobj->cache();
        Assert( cache != 0 );
        unsigned int cacheUnit = 
        memobj->offsetToCacheUnit( offsetIntoObject );
        Assert( cache->isIn( cacheUnit ) );
        
        unsigned int endUnit =
        memobj->offsetToCacheUnit( offsetIntoObject + 
                                   length - 1 );
        Assert( cache->isIn( endUnit ) );
        
        // cache->flush(cacheUnit,endUnit);
        cache->untranslatable( cacheUnit, endUnit, _translation,
                               baseAddr );
    }
    SInstr(DomainMakeUnAddr.stop(start));
}

#include "ClassInstaller.h"
DEFINECLASS("Domain",Domain,ProxiableObject,0,0);
