#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * DummyMemoryObject.c - A memory object that thinks all read succeede and
 *			 all writes are an error.
 */

#include "Assert.h"
#include "DummyMemoryObject.h"
#include "MemoryObjectAsync.h"
#include "OutputStream.h"
#include "PremappedMemoryObjectCache.h"
#include "Raid.h"
#include "RootMemoryObjectCache.h"
#include "NewAndDelete.h"
#include "MemoryObjectView.h"


#ifdef __GNUC__
#pragma implementation
#endif

DummyMemoryObject::DummyMemoryObject( unsigned int numberOfUnits,
				      int log2UnitSize )
     : MemoryObject( numberOfUnits, log2UnitSize ), _specialLock( this ),
	_view (this)
{
	Assert( this != 0 );
	_protection = ReadOnly;
	_moLock = & _specialLock;
}

DummyMemoryObject::~DummyMemoryObject()
{
}


int
DummyMemoryObject::rawRead( unsigned int, unsigned  count, char *, unsigned , 
			    IOAsyncClosure * c )
{
	if ( c ) {
		c->activate( count );
	}
	return( count );
}

int
DummyMemoryObject::rawWrite( unsigned int, unsigned , char *, unsigned , int, 
			     IOAsyncClosure * )
{
	Console().panic( "DummyMemoryObject::write() called\n" );
	Assert( NOTREACHED );
	return( 0 );
}

/*
 * The parent method would suffice but I want the extra assertion.
 * (All DummyMOs and subclasses build their caches in their constructors.
 */
MemoryObjectCache *
DummyMemoryObject::buildCache()
{
	Raids( RaidMethod ) << this << "::buildCache()\n" << eor;
	Assert( this != 0 );

	/*
	 * We prebuilt a cache for this MemoryObject in the constructor.
	 */

	return( _cache );
}

PhysicallyResidentMemoryObject::PhysicallyResidentMemoryObject
	( unsigned int numberOfUnits, unsigned log2UnitSize, const char * address,
	  int preAllocated )
	: DummyMemoryObject( numberOfUnits, log2UnitSize ),
	  _premappedCache( this, address, preAllocated )
{
	initDebug();
	Raids( RaidConstructor ) << this 
		<< "::PhysicallyResidentMemoryObject( " << numberOfUnits
		<< " units, " << log2UnitSize << " log2UnitSize, "
		<< address << " address, "
		<< preAllocated << " preAllocated )\n" << eor;
	Assert( this != 0 );

	residentAddress = address;
	_cache = & _premappedCache;
	Assert( log2UnitSize == _cache->log2UnitSize() );
}

/*
 * This constructor is used when allocation of memory is very sensitive.
 * It is assumed that the caller will have already allocated space of
 *    size 'stateSize()' and passed that in as 'stateSpace'.
 */

PhysicallyResidentMemoryObject::PhysicallyResidentMemoryObject
	( unsigned log2UnitSize, const char * base, const char * top,
	  int preAllocated )
	: DummyMemoryObject((top-base+(1 << log2UnitSize) - 1) >> log2UnitSize,
		log2UnitSize ), _premappedCache( this, base, preAllocated )
{
	initDebug();
#ifdef DEBUG
	_debugMask = MaxObjectDebugMask;
	_debugMessages = MaxObjectDebugMessages;
#endif
	Raids( RaidConstructor ) << this 
		<< "::PhysicallyResidentMemoryObject( " << log2UnitSize
		<< " log2UnitSize, " << asAddress( base ) << " base, "
		<< asAddress( top ) << " top, "
		<< preAllocated << " preAllocated )\n" << eor;
	Assert( this != 0 );

	_numberOfUnits = ( top - base + ( 1 << log2UnitSize ) ) 
		         >> log2UnitSize;

	Raids( RaidConstructor ) << this 
		<< ": to map " << asAddress( base ) << " to "
		<< asAddress( top ) << ", need " << _numberOfUnits
	        << " " << ( 1 << log2UnitSize ) << "-byte units\n" << eor;

	_cache = & _premappedCache;
}

PhysicallyResidentMemoryObject::~PhysicallyResidentMemoryObject()
{
	Assert( NOTREACHED );
}

RootMemoryObject::RootMemoryObject
	( unsigned int numberOfUnits, unsigned log2UnitSize )
	: DummyMemoryObject( numberOfUnits, log2UnitSize )
{
	initDebug();
	Raids( RaidConstructor ) << this 
		<< "::RootMemoryObject( " << numberOfUnits
		<< "units, " << log2UnitSize << " log2UnitSize )\n" << eor;
	Assert( this != 0 );
	_cache = new RootMemoryObjectCache( this );
	Assert( _cache != 0 );
	Assert( log2UnitSize == _cache->log2UnitSize() );
}

RootMemoryObject::~RootMemoryObject()
{
	Assert( NOTREACHED );
}


#include "ClassInstaller.h"
DEFINECLASS("DummyMemoryObject",DummyMemoryObject,MemoryObject,0,0);
DEFINECLASS("RootMemoryObject",RootMemoryObject,DummyMemoryObject,0,0);
DEFINECLASS("PhysicallyResidentMemoryObject",PhysicallyResidentMemoryObject,DummyMemoryObject,0,0);
