#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-1993
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * GlobalLRUReplacementPolicy.cc
 */

#include "Assert.h"
#include "Async.h"
#include "Byte.h"
#include "Class.h"
#include "ClassStar.h"
#include "FileSystemHints.h"
#include "FreeRunningTimer.h"
#include "GlobalLRUReplacementPolicy.h"
#include "Kernel.h"
#include "Log2.h"
#include "Map.h"
#include "MemoryObject.h"
#include "PageOutAsync.h"
#include "PersistentMemoryObject.h"
#include "Raid.h"
#include "SimpleInstruments.h"
#include "Store.h"
#include "ThisProcess.h"
#include "TimeoutTimer.h"

#ifdef __GNUC__
#pragma implementation
#endif

#define READAHEAD_CLUSTERS 1
#define LARGE_READAHEAD_CLUSTERS 4
#define LARGE_FREEBEHIND_CLUSTERS 2

//#define SAFE
//#define DEBUGLAG
//#define DEBUGGLOBALLRU
//#define TIMEGLOBALLRU
//#define INSTGLOBALLRU

Lock GlobalLRUInfoTable::_lock;

int Default_doSyncBehind = 1;
int Default_doPrefetchFirst = 1;

inline 
GlobalLRUInfoTable::GlobalLRUInfoTable() 
{
    _used = 0;
    _size = 0;
}

inline void
GlobalLRUInfoTable::init()
{
    Assert( _size == 0 );
    _size = 1024;
    _array = new GlobalLRUInfo[ _size ];
    Debug 
    << "GlobalLRUInfoTable::init() size " << _size << "\n" << eor;
}

void
GlobalLRUInfoTable::grow()
{
    // lock should be held by caller

    Assert( _size == _used );
    int newSize = _size << 1;
    Debug 
    << "GlobalLRUInfoTable::grow() size " << _size
    << " to " << newSize << "\n" << eor;
    GlobalLRUInfo * newArray = new GlobalLRUInfo[ newSize ];
    ByteCopy( (char *) _array, (char *) newArray, 
              _size * sizeof( GlobalLRUInfo ) );
    delete _array;
    _array = newArray;
    _size = newSize;
}

void
GlobalLRUInfoTable::garbageCollect()
{
    // lock should be held by caller

    int newSize = _size >> 2;
    Assert( newSize >= _used );
    Debug 
    << "GlobalLRUInfoTable::garbageCollect() size " << _size
    << " to " << newSize << "\n" << eor;
    GlobalLRUInfo * newArray = new GlobalLRUInfo[ newSize ];
    ByteCopy( (char *) _array, (char *) newArray, 
              _used * sizeof( GlobalLRUInfo ) );
    delete _array;
    _array = newArray;
    _size = newSize;
}

void
GlobalLRUInfoTable::check( GlobalLRUReplacementPolicy * rp, unsigned from,
                           unsigned to, unsigned locked )
{
    if ( !locked )
        acquire();
    GlobalLRUInfo * check = _array;
    for ( int count = 0; count < _used; count++ ) {
        if ( check->_rp == rp &&
             check->_from <= to && check->_to >= from ) {
            Console() << "conflict checking " << from << " to "
                      << to << " against slot " << count 
                      << " from " << check->_from
                      << " to " << check->_to << "\n" << eor;
            Assert( NOTREACHED );
        }
        check++;
    }
    if ( !locked )
        release();
}

void
GlobalLRUInfoTable::add( GlobalLRUReplacementPolicy * rp, 
                         unsigned from, unsigned to )
{
    acquire();
    Debug
    << "GlobalLRUInfoTable::add( "
    << asAddress(rp) << ", " << from << ", " << to << " ) slot "
    << _used << "\n"
    << eor;
    Assert( from <= to );
    if ( _size == _used ) {
        grow();
    }
#ifdef SAFE
    check( rp, from, to, 1 );
#endif // SAFE
    GlobalLRUInfo * info = _array + _used;
    info->_rp = rp;
    info->_from = from;
    info->_to = to;
#ifdef NEW_GLRU
    if ( rp->_head != -1 ) {
        _array[rp->_head]._prev = _used;
    }
    info->_next = rp->_head;
    info->_prev = -1;
    rp->_head = _used;
#endif // NEW_GLRU
    _used++;
    release();
}

void
GlobalLRUInfoTable::remove( GlobalLRUReplacementPolicy * rp,
                            unsigned from, unsigned to, int hint )
{
    acquire();
    int count = 0;
    int total = to - from + 1;
    Debug
    << "GlobalLRUInfoTable::remove( "
    << asAddress(rp) << ", " << from << ", " << to << " ) hint "
    << hint << " used " << _used << "\n"
    << eor;
    Assert( from <= to );
#ifdef NEW_GLRU
    int slot;
    if ( hint >= 0 && hint < _used ) {
        GlobalLRUInfo * info = _array + hint;
        if ( info->_rp == rp && 
             info->_from == to && info->_to == from ) {
            _used--;
            count += ( to - from + 1 );
            if ( info->_next != -1 ) {
                _array[info->_next]._prev = info->_prev;
            }
            if ( info->_prev != -1 ) {
                _array[info->_prev]._next = info->_next;
            } else {
                rp->_head = info->_next;
            }
            if ( hint < _used ) {
                *info = _array[ _used ];
                if ( info->_next != -1 ) {
                    _array[info->_next]._prev = hint;
                } 
                if ( info->_prev != -1 ) {
                    _array[info->_prev]._next = hint;
                } else {
                    info->_rp->_head = hint;
                }
            }
            goto done;
        }
    }
    slot = rp->_head;
    while ( count < total && slot != -1 ) {

        GlobalLRUInfo * info = _array + slot;
        Assert( info->_rp == rp );
        if ( !( info->_from <= to && info->_to >= from ) ) {
            slot = info->_next; 
            continue;
        }
        if ( info->_from >= from && info->_to <= to ) {
            _used--;
            count += ( info->_to - info->_from + 1 );
            int next = info->_next;
            if ( next != -1 ) {
                _array[next]._prev = info->_prev;
            }
            if ( info->_prev != -1 ) {
                _array[info->_prev]._next = next;
            } else {
                rp->_head = next;
            }
            if ( slot < _used ) {
                *info = _array[ _used ];
                if ( info->_next != -1 ) {
                    _array[info->_next]._prev = slot;
                } 
                if ( info->_prev != -1 ) {
                    _array[info->_prev]._next = slot;
                } else {
                    info->_rp->_head = slot;
                }
                if ( next == _used )
                    next = slot;
            }
            slot = next;
        } else if ( info->_from >= from && info->_to >= to ) {
            count += ( to - info->_from + 1 );
            info->_from = to + 1;
            slot = info->_next;
        } else if ( info->_from <= from && info->_to <= to ) {
            count += ( info->_to - from + 1 );
            info->_to = from - 1;
            slot = info->_next;
        } else {
//          Assert( info->_from < from && info->_to > to );
            count += ( to - from + 1 );
            unsigned tmp = info->_to;
            info->_to = from - 1;
            if ( _size == _used ) {
                grow();
            }
            GlobalLRUInfo * newInfo = _array + _used;
            Assert( rp->_head != -1 );
            _array[rp->_head]._prev = _used;
            newInfo->_next = rp->_head;
            newInfo->_prev = -1;
            rp->_head = _used;
            newInfo->_rp = rp;
            newInfo->_from = to + 1;
            newInfo->_to = tmp;
            _used++;
            break;
        } 
    } 
 done:
#else // not NEW_GLRU
    int slot;
    if ( hint < 0 && hint >= _used ) {
        slot = 0;
    } else {
        slot = hint;
    }
    int icount = 0;
    do {
        GlobalLRUInfo * info = _array + slot;
        if ( info->_rp == rp && 
             info->_from <= to && info->_to >= from ) {

            Debug
            << "intercept slot " << slot 
            << " rp " << asAddress( rp )
            << " from " << info->_from << " to "
            << " to " << info->_to << "\n" << eor;

            if ( info->_from >= from && 
                 info->_to <= to ) {
                _used--;
                count += ( info->_to - info->_from + 1 );
                Debug
                << "inside, new used " << _used
                << " count " << count << "\n" << eor;
                if ( slot < _used ) {
                    *info = _array[ _used ];
                    Debug
                    << "slot " << _used
                    << " moved to slot " << slot 
                    << " rp " << asAddress(info->_rp)
                    << " from " << info->_from
                    << " to " << info->_to
                    << "\n" << eor;
                }
            } else if ( info->_from >= from &&
                        info->_to >= to ) {
                count += ( to - info->_from + 1 );
                info->_from = to + 1;
                Debug
                << "right out new from " 
                << info->_from 
                << " count " << count << "\n" << eor;
                slot++;
            } else if ( info->_from <= from &&
                        info->_to <= to ) {
                count += ( info->_to - from + 1 );
                info->_to = from - 1;
                Debug
                << "left out new to " 
                << info->_to 
                << " count " << count << "\n" << eor;
                slot++;
            } else {
//              Assert( info->_from < from && info->_to > to );
                count += ( to - from + 1 );
                unsigned tmp = info->_to;
                info->_to = from - 1;
                Debug 
                << "enclosed split slot "
                << slot << " from " << info->_from
                << " to " << info->_to << "\n" 
                << eor;
                if ( _size == _used ) {
                    grow();
                }
                GlobalLRUInfo * newInfo = _array + _used;
                _used++;
                newInfo->_rp = rp;
                newInfo->_from = to + 1;
                newInfo->_to = tmp;
                Debug
                << "add split slot " << (_used - 1)
                << " from " << info->_from 
                << " to " << info->_to 
                << " count " << count << "\n" << eor;
                slot++;
            } 
        } else {
            slot++;
        }
        if ( slot >= _used ) {
            slot = 0;
        }
        icount++;
        Assert( icount <= _size );
    } while ( count < total ); 
#endif // NEW_GLRU
    Assert( count == total );
#ifdef SAFE
    check( rp, from, to, 1 );
#endif
    if ( (_used << 3) < _size ) {
        garbageCollect();
    }
    release();
}

int
GlobalLRUInfoTable::next( int current, GlobalLRUReplacementPolicy * & rp,
                          unsigned & from, unsigned & to )
{
    acquire();
    current++;
    if ( current >= _used )
        current = 0;
    GlobalLRUInfo * info = _array + current;
    rp = info->_rp;
    from = info->_from;
    to = info->_to;
    release();
    return current;
}

GlobalLRUInfoTable GlobalLRUReplacementPolicy::_table;
int GlobalLRUReplacementPolicy::_paging = 0;
Lock GlobalLRUReplacementPolicy::_pagingLock;
Semaphore GlobalLRUReplacementPolicy::_pagerWait(0);
Process * GlobalLRUReplacementPolicy::_pager = 0;
int GlobalLRUReplacementPolicy::_distance = 0;
int GlobalLRUReplacementPolicy::_firstHand = 0xffffff;

void 
GlobalLRUReplacementPolicy::initialize()
{
    _table.init();
    _pager = new PreemptableSystemProcess
    ( &pageOutDaemon, TheKernel->domain(), normal, 0 );
    _pager->setName("#GlobalLRUPageOutDaemon");
    _pager->ready();
}

GlobalLRUReplacementPolicy::GlobalLRUReplacementPolicy
( PagedMemoryObjectCache * cache ) 
    : ReplacementPolicy( cache )
{
    initDebug();
    Raids( RaidMethod ) 
    << this << "::GlobalLRUReplacementPolicy( "
    << cache << ") \n" << eor;
    _pagingLock.acquire();
    if ( _pager == 0 ) initialize();
    _pagingLock.release();
#ifdef NEW_GLRU
    _head = -1;
#endif // NEW_GLRU
}

GlobalLRUReplacementPolicy::~GlobalLRUReplacementPolicy()
{
#ifdef SAFE
    _table.check( this, 0, 9999999, 0 );
#endif
}

void
GlobalLRUReplacementPolicy::unitsToStore( unsigned from, unsigned to )
{
    Debug 
    << "GlobalLRUReplacementPolicy[" << asAddress(this) 
    << "]::unitToStore\n" << eor;
    _table.remove( this, from, to, -1 );
}

void
GlobalLRUReplacementPolicy::postPageIn( unsigned from, unsigned to )
{
    _table.add( this, from, to );
}

unsigned
GlobalLRUReplacementPolicy::pageOut( unsigned noOfUnits ) 
{
    static TimeoutTimer timer;

    if ( _table.size() == 0 )
        return 0;
    int release;
    _pagingLock.acquire();
    if ( ! _paging ) {
        _paging = 1;
        release = 1;
    } else {
        release = 0;
    }
    _pagingLock.release();
    if ( release ) {
        _pagerWait.V();
    } 
    timer.start( 200000 );
    timer.await();
    thisProcess()->relinquishProcessor();
    return 1;
}

#ifdef DEBUGGLOBALLRU
int debugGlobalLRU = 0;
int debugGlobalLRUTime = 0;
int debugGlobalLRUInst = 0;
int debugGlobalLRUPageOut = 0;
#endif
#ifdef INSTGLOBALLRU
static char *instOn[] = { "inst", "k", "SimpleTimer", "on" };
static char *instReset[] = { "inst", "k", "SimpleTimer", "r" };
static char *instOff[] = { "inst", "k", "SimpleTimer", "off" };
static char *instDisplay[] = { "inst", "k", "SimpleTimer", "d" };
#endif // INSTGLOBALLRU

void
GlobalLRUReplacementPolicy::pageOutDaemon()
{
    GlobalLRUReplacementPolicy * rp;
    unsigned from;
    unsigned to;
    unsigned mainStoreSize = MainStore->numberOfUnits();
    unsigned lotsfree = mainStoreSize / 8;
    unsigned desfree = mainStoreSize / 16;
    unsigned minfree = mainStoreSize / 32;
    Assert( desfree > 0 && lotsfree > 0 && minfree > 0 );
    int fastscan = 1000;
    int slowscan = 100;
    int fastsleep = 1000000 / fastscan;
    int slowsleep = 1000000 / slowscan;
    int veryslowsleep = 1000000 / 1;
    TimeoutTimer timer; 
#ifdef TIMEGLOBALLRU
    unsigned startClear;
    unsigned endClear;
    unsigned startCheck;
    unsigned endCheck;
    unsigned startPage;
    unsigned endPage;
#endif /*TIMEGLOBALLRU*/

    Debug << "GlobalLRUReplacementPolicy::pageOutDaemon " 
              << thisProcess() << "\n" << eor;

    while (1) {

#ifdef INSTGLOBALLRU
        if ( debugGlobalLRUInst )
            simpleInstrumentControl( 4, instOff );
#endif // INSTGLOBALLRU

        _pagerWait.P();

#ifdef INSTGLOBALLRU
        if ( debugGlobalLRUInst )
            simpleInstrumentControl( 4, instOn );
#endif // INSTGLOBALLRU

        while ( ( mainStoreSize - MainStore->totalExpectedAllocPaus() )
                <= lotsfree ) {

            _distance = _table.size() / 4;
            if ( _distance == 0 )
                break;

            unsigned startTime = SystemTimer->time();

            _firstHand = _table.next( _firstHand, rp, from, to );
            Assert( rp != 0 && rp->_cache != 0 );
#ifdef DEBUGGLOBALLRU
            if ( debugGlobalLRU ) Debug
            << "GlobalLRUReplacementPolicy::pageOutDaemon()"
            << " firstHand " << _firstHand 
            << " cache " << rp->_cache
            << " from " << from 
            << " to " << to 
            << "\n" << eor;
#endif // DEBUGGLOBALLRU

#ifdef TIMEGLOBALLRU
            if ( debugGlobalLRUTime )
                startClear = thisProcess()->processTime();
#endif // TIMEGLOBALLRU
            rp->cacheClearReferenced( from, to );
#ifdef TIMEGLOBALLRU
            if ( debugGlobalLRUTime ) 
                endClear = thisProcess()->processTime();
#endif // TIMEGLOBALLRU

            int secondHand = _firstHand - _distance;
            if ( secondHand < 0 ) 
                secondHand += _table.size();
            Assert( secondHand >= 0 );
            _table.next( secondHand, rp, from, to );
#ifdef DEBUGGLOBALLRU
            if ( debugGlobalLRU ) Debug
            << "GlobalLRUReplacementPolicy::pageOutDaemon()"
            << " secondHand " << secondHand 
            << " cache " << rp->_cache
            << " from " << from 
            << " to " << to 
            << "\n" << eor;
#endif // DEBUGGLOBALLRU

            PhysicallyAddressableUnit * notReferencedPaus;

#ifdef TIMEGLOBALLRU
            if ( debugGlobalLRUTime )
                startCheck = thisProcess()->processTime();
#endif // TIMEGLOBALLRU

#ifdef DEBUGGLOBALLRU
            if ( debugGlobalLRUPageOut ) {
                notReferencedPaus = 0;
                if (rp->acquireMo( from, to, 0, 0 ) == 0){
                    notReferencedPaus = 
                    rp->_map->pauList( from, to );
                    if ( notReferencedPaus == 0 ) {
                        rp->cachePostFault( from, to );
                    }
                }
            } else
#endif // DEBUGGLOBALLRU
                notReferencedPaus = 
                rp->cacheCheckReferenced( from, to, 0 );
#ifdef TIMEGLOBALLRU
            if ( debugGlobalLRUTime )
                endCheck = thisProcess()->processTime();
#endif // TIMEGLOBALLRU

#ifdef TIMEGLOBALLRU
            unsigned startPage = 0;
            unsigned endPage = 0;
#endif // TIMEGLOBALLRU

            if ( notReferencedPaus ) { 

                Debug
                << "GlobalLRUReplacementPolicy::pageOutDaemon "
                << "pagout remove\n" << eor; 

                _table.remove( rp, from, to, secondHand );

#ifdef DEBUGGLOBALLRU
                if ( debugGlobalLRU ) Debug
                << "GlobalLRUReplacementPolicy::pageOutDaemon()"
                << " submitting async, closure = " 
                << asAddress( closure ) 
                << " ( " << *closure << " )\n" << eor;
#endif // DEBUGGLOBALLRU

                rp->doPageOutAsync( from, to, notReferencedPaus,
                                    PostFaultGroup, 
                                    UNITSTOSTORE_SKIPRP );
            }

            unsigned remainingPaus = 
            mainStoreSize - 
            MainStore->totalExpectedAllocPaus();
            if ( remainingPaus > lotsfree )
                break;
            unsigned sleep;
            if ( remainingPaus <= lotsfree &&
                 remainingPaus > desfree ) {
                sleep = slowsleep + 
                ( remainingPaus - desfree ) *
                ( veryslowsleep - slowsleep ) /
                ( lotsfree - desfree );
            } else if ( remainingPaus <= desfree &&
                        remainingPaus > minfree ) {
                sleep = fastsleep + 
                ( remainingPaus - minfree ) *
                ( slowsleep - fastsleep ) /
                ( desfree - minfree );
            } else {
                Assert( remainingPaus <= minfree );
                sleep = 0;
            } 

            Assert( sleep >= 0 );
            unsigned endTime = SystemTimer->time();
            unsigned timeTaken =
            ( startTime > endTime ? 
              ((unsigned) -1 ) - startTime + endTime :
              endTime - startTime );

#if defined( DEBUGGLOBALLRU ) || defined( TIMEGLOBALLRU ) || defined( INSTGLOBALLRU )
            if ( debugGlobalLRU || debugGlobalLRUTime ||
                 debugGlobalLRUInst ) {
                Debug
                << "GlobalLRUReplacementPolicy::pageOutDaemon()"
                << " remainingPaus " << remainingPaus
                << " sleep time " << sleep
                << " time taken " << timeTaken
                << "\n" << eor;
#ifdef TIMEGLOBALLRU
                if ( debugGlobalLRUTime ) Debug
                << " clearTime " << endClear - startClear
                << " checkTime " << endCheck - startCheck
                << " pageTime " << endPage - startPage
                << "\n" << eor;
#endif // TIMEGLOBALLRU
#ifdef INSTGLOBALLRU
                if ( debugGlobalLRUInst )
                    simpleInstrumentControl(4,instDisplay);
#endif // INSTGLOBALLRU
            }
#endif // DEBUGGLOBALLRU || TIMEGLOBALLRU

            if ( timeTaken < sleep ) {
                timer.start( sleep - timeTaken );
                timer.await();
            }
        }

        _paging = 0;
    }
}

unsigned 
GlobalLRUReplacementPolicy::victim( unsigned count, 
                                    unsigned & start, unsigned & end,
                                    PhysicallyAddressableUnit * & pau )
{
    Assert( NOTREACHED );
    return 0;
}

Lock SmarterGlobalLRUReplacementPolicy::_sglock;

#ifdef DEBUGLAG
int debugLag = 1;
#endif // DEBUGLAG

class FreeBehindIOAsyncClosure : public IOAsyncClosure {
protected:
    SmarterGlobalLRUReplacementPolicy * _rp;
    unsigned _from;
    unsigned _to;
    static void func( FreeBehindIOAsyncClosure * c );
public:
    FreeBehindIOAsyncClosure( SmarterGlobalLRUReplacementPolicy * rp,
                              unsigned from, unsigned to ) 
        : _rp( rp ), _from( from ), _to( to ) {_func = (AsyncFunc) &func;}
};

inline
void
SmarterGlobalLRUReplacementPolicy::freeBehindDone( unsigned from, unsigned to )
{
    Debug
    << "SmarterGlobalLRUReplacementPolicy["
    << asAddress(this) << "]::freeBehindDone from "
    << from << " to " << to << "\n" << eor;
    cacheUnitsToStore( from, to, UNITSTOSTORE_QUEUED );
    releaseMo( from, to );
    _sglock.acquire();
    _queued -= ( to - from + 1 );
    Assert( _queued >= 0 );
    _sglock.release();
}

void FreeBehindIOAsyncClosure::func( FreeBehindIOAsyncClosure * c )
{
    c->_rp->freeBehindDone( c->_from, c->_to );
    delete c;
}

SmarterGlobalLRUReplacementPolicy::SmarterGlobalLRUReplacementPolicy
( PagedMemoryObjectCache * cache ) 
    : GlobalLRUReplacementPolicy( cache )
{
    unsigned log2UnitSize = _cache->log2UnitSize();
#ifdef ORIGINAL
    if ( _log2ClusterSize <= log2UnitSize )
        _clusterSize = 1;
    else
        _clusterSize = 
        1 << ( _log2ClusterSize - log2UnitSize );
#else // CRISPY
    _clusterSize = _cache->memoryObject()->segmentSize() >> log2UnitSize;
    if (_clusterSize == 0)
        _clusterSize = 1;
#endif
    _queued = 0;
    _readSequential = -1;
    _writeSequential = -1;
    _maxFreeBehindLag = 
    1 << 
    ( log2High(MainStore->numberOfUnits() * MainStore->unitSize()) 
      - log2UnitSize - 1 );
#ifdef ORIGINAL
    if ( _cache->log2UnitSize() >= _minLog2FreeBehindLag ) 
        _minFreeBehindLag = 1;
    else 
        _minFreeBehindLag = 
        1 << ( _minLog2FreeBehindLag - _cache->log2UnitSize() );
#else // CRISPY
    // defualt _minFreeBehindLag is two segment
    _minFreeBehindLag = _clusterSize * 2;
#endif
    _slowFreeBehindLag = 1 << ( _slowLog2FreeBehindLag - log2UnitSize );
    if ( _slowFreeBehindLag == 0 )
        _slowFreeBehindLag = 1;
    _freeBehindLag = _maxFreeBehindLag;
    _lastTime = SystemTimer->time();
    _nextSynchronize = 0;
#ifdef READAHEAD_CLUSTERS
    _readAheadLag = _clusterSize * READAHEAD_CLUSTERS;
#else
    _readAheadLag = _clusterSize;
#endif
    _doSyncBehind = Default_doSyncBehind;
    _doPrefetchFirst = Default_doPrefetchFirst;
}

inline
void
SmarterGlobalLRUReplacementPolicy::updateFreeBehindLag()
{
    // _sglock already held

    if ( _freeBehindLag != _maxFreeBehindLag ) {
        unsigned newTime = SystemTimer->time();
        unsigned diff = newTime - _lastTime;
        if ( diff > 700000 ) { /* 0.7 sec */
            if ( _freeBehindLag < _slowFreeBehindLag && 
                 diff < 2000000 )
                _freeBehindLag += _minFreeBehindLag;
            else
                _freeBehindLag <<= 1;
            if ( _freeBehindLag > _maxFreeBehindLag )
                _freeBehindLag = _maxFreeBehindLag;
#ifdef DEBUGLAG
            if ( debugLag ) Debug 
            << this 
            << "::updateFreeBehind() freeBehindLag " 
            << _freeBehindLag << " diff " << diff
            << "\n" << eor;
#endif // DEBUGLAG
        }
    }
}

inline
void
SmarterGlobalLRUReplacementPolicy::doFreeBehind
( unsigned & from, unsigned current )
{
    int excess = _map->residentSize() - _queued - _freeBehindLag;
    if ( excess <= 0 ) {
        _sglock.release();
        return;
    }
    unsigned size = _cache->numberOfUnits();
    if ( from >= size ) {
        from = 0;
    }
    unsigned newFrom = from + ( excess << 1 );
    if ( newFrom > size )
        newFrom = size;
    Assert( newFrom > from );
    if ( from <= current ) {
        newFrom = current - _freeBehindLag;
        if ( newFrom <= from ) {
            _sglock.release();
            return;
        }
        if ( newFrom - from < _clusterSize ) {
            _sglock.release();
            return;
        }
    }
    unsigned end = newFrom - 1;
    Assert( from >= 0 && end >= from && end < size );
    _sglock.release();
    Process * p = acquireMo( from, end, 0, 0 );
    if ( p != 0 ) {
        return;
    }
    _sglock.acquire();
    _queued += ( end - from + 1 );
    int orgFrom = from;
    from = newFrom; 
    _sglock.release();
    Debug
    << "SmarterGlobalLRUReplacementPolicy["
    << asAddress(this) << "]::doFreeBehind from "
    << from << " to " << end << "\n" << eor;
    IOAsyncClosure * c = new FreeBehindIOAsyncClosure( this, orgFrom, end );
    cacheUntranslatable( orgFrom, end );
    cacheSaveModified( ToCache, orgFrom, end, 
                       ( SAVEMODIFIED_FLUSH | SAVEMODIFIED_QUEUED ), c );
}

class SynchronizeBehindAsyncClosure : public IOAsyncClosure {
protected:
    SmarterGlobalLRUReplacementPolicy * _rp;
    unsigned _from;
    unsigned _to;
    static void func( SynchronizeBehindAsyncClosure * c );
public:
    SynchronizeBehindAsyncClosure( SmarterGlobalLRUReplacementPolicy * rp,
                                   unsigned from, unsigned to ) 
        :  _rp( rp ), _from( from ), _to( to ) {_func = (AsyncFunc)&func;}
};

inline
void
SmarterGlobalLRUReplacementPolicy::synchronizeDone( unsigned from, unsigned to )
{
    releaseMo( from, to );
}

void SynchronizeBehindAsyncClosure::func( SynchronizeBehindAsyncClosure * c )
{
    c->_rp->synchronizeDone( c->_from, c->_to );
    delete c;
}

void 
SmarterGlobalLRUReplacementPolicy::doSynchronizeBehind()
{
    _sglock.acquire();
    unsigned syncCount;
    unsigned syncEnd = 0;
    unsigned syncStart = 0;
    unsigned nextSync = 0;
    if ( _writePointer < _nextSynchronize ) {
        syncCount = _numberOfUnits - _nextSynchronize;
        if ( syncCount <= 0 ) {
            syncCount = 0;
            _nextSynchronize = 0;
        } else {
            syncEnd = _numberOfUnits - 1;
            syncStart = _nextSynchronize;
            nextSync = 0;
        }
    } else if ( _writePointer - _nextSynchronize >= _clusterSize ) {
        syncCount = _writePointer - _nextSynchronize;
        syncCount = ( syncCount / _clusterSize ) *
        _clusterSize;
        syncStart = _nextSynchronize;
        syncEnd = _nextSynchronize + syncCount - 1;
        if ( syncEnd >= _numberOfUnits )
            syncEnd = _numberOfUnits - 1;
        nextSync = syncEnd + 1;
    } else {
        syncCount = 0;
    }
    if ( syncCount > 0 ) {
        _sglock.release();
        Process * p = 
        acquireMo( syncStart, syncEnd, 0, 0 );
        if ( p == 0 ) {
            _sglock.acquire();
            // ugly hack until scheduler has different
            // priority levels. problem here is the
            // acquireMO may relinquish processor
            // if acquired exclusive
            // the solution is to bump up process priority
            // if process acquired exclusive lock
            Assert( _nextSynchronize == syncStart );
            _nextSynchronize = nextSync;
            IOAsyncClosure * c = 
            new SynchronizeBehindAsyncClosure
            ( this, syncStart, syncEnd);
            _sglock.release();
            Assert( syncStart <= syncEnd &&
                    syncEnd < _numberOfUnits );
            cacheUntranslatable( syncStart, syncEnd );
            cacheSaveModified( ToCache, syncStart, syncEnd, 
                               0, c );
        }
    } else {
        _sglock.release();
    }
}

class ReadAheadIOAsyncClosure : public IOAsyncClosure {
protected:
    SmarterGlobalLRUReplacementPolicy * _rp;
    unsigned _from;
    unsigned _to;
    static void func( ReadAheadIOAsyncClosure * c );
public:
    ReadAheadIOAsyncClosure( SmarterGlobalLRUReplacementPolicy * rp,
                             unsigned from, unsigned to ) 
        :  _rp( rp ), _from( from ), _to( to ) {_func= (AsyncFunc)&func;}
};

inline
void
SmarterGlobalLRUReplacementPolicy::readAheadDone( unsigned from, unsigned to )
{
    releaseMo( from, to );
}

void ReadAheadIOAsyncClosure::func( ReadAheadIOAsyncClosure * c )
{
    c->_rp->readAheadDone( c->_from, c->_to );
    delete c;
}

void
SmarterGlobalLRUReplacementPolicy::preCacheRead( unsigned from, unsigned to )
{
    if ( _readSequential != -1 )
        return;
    if ( _doPrefetchFirst ) {
        unsigned end = from + _clusterSize - 1;
        // int end = from + _readAheadLag - 1;
        if ( end >= _numberOfUnits )
            end = _numberOfUnits - 1;
        if ( end <= to ) {
            _readAhead = to + 1;
            return;
        }
        Process * result1;
        Process * result2;
        result1 = acquireMo( from, to, 0, 0 );
        result2 = acquireMo( to + 1, end, 0, 0 ); 
        if ( result2 == 0 )
            cacheLoad( from, end, 0 );
        if ( result1 == 0 )
            releaseMo( from, to );
        if ( result2 == 0 )
            releaseMo( to + 1, end );
        _readAhead = end + 1;
    } else {
        _readAhead = to + 1;
    }
}

void 
SmarterGlobalLRUReplacementPolicy::postCacheRead( unsigned from, unsigned to )
{
    _sglock.acquire();
    unsigned count = to - from + 1;
    unsigned next = to + 1;
    if ( _readSequential == -1 ) {
        _readSequential = 1;
        _readFreeBehind = from;
    } else {
        if ( _readPointer == from ) {
            _readSequential = 1;
            if ( _readAhead < next ) {
                _readAhead = next;
            }
        } else if ( from == 0 ) {
            _readAhead = next;
        } else {
            _readSequential = 0;
        }
    }
    _readPointer = next;
    if ( count > _readAheadLag ) {
        _readAheadLag = count;
#ifdef DEBUGLAG
        if ( debugLag ) Debug
        << this << "::postCacheRead() readAheadLag " << _readAheadLag
        << "\n" << eor;
#endif // DEBUGLAG
    }

    if ( _readSequential ) {
        
        unsigned lag = _readAhead - _readPointer;
        if ( lag >= _readAheadLag ) {
            _sglock.release();
            return;
        }

        if ( _readAhead >= _numberOfUnits ) {
            _sglock.release();
            return;
        }

        Assert ( _readAhead >= next );
        unsigned newReadAhead = _readAhead + _readAheadLag;
        if ( newReadAhead > _numberOfUnits ) {
            newReadAhead = _numberOfUnits;
        }
        unsigned start = _readAhead;
        unsigned end = newReadAhead - 1;
        _readAhead = newReadAhead;
        Assert( start <= end );
        do {
            unsigned last = start + _clusterSize;
            if ( last > end ) 
                last = end;
            Process * p = acquireMo( start, last, 0, 0 );
            if ( p == 0 ) {
                _sglock.release();
                Assert( start <= last );
                IOAsyncClosure * c = new 
                ReadAheadIOAsyncClosure
                ( this, start, last );
                Debug
                << "SmarterGlobalLRUReplacementPolicy["
                << asAddress(this) << "]::doReadAhead from "
                << start << " to " << last << "\n" << eor;
                cacheLoad( start, last, c );
                _sglock.acquire();
            }
            start = last + 1;
        } while ( start <= end );

        updateFreeBehindLag();
        /* activate free behind, _sglock released by doFreeBehind */
        doFreeBehind( _readFreeBehind, _readPointer );
    } else {
        _sglock.release();
    }
}

void 
SmarterGlobalLRUReplacementPolicy::postCacheWrite( unsigned from, unsigned to )
{
    _sglock.acquire();
    if ( _writeSequential == -1 ) {
        _writeFreeBehind = from;
        _writeSequential = 1;
    } else {
        _writeSequential = ( _writePointer == from );
    }
    _writePointer = to + 1;

    if ( _writeSequential ) {
        updateFreeBehindLag();
        /* activate free behind, _sglock released by doFreeBehind */
        doFreeBehind( _writeFreeBehind, _writePointer );
        if ( _doSyncBehind ) 
            doSynchronizeBehind();
    } else {
        _sglock.release();
    }
}

unsigned
SmarterGlobalLRUReplacementPolicy::pageOut( unsigned noOfUnits ) 
{
    // return GlobalLRUReplacementPolicy::pageOut( noOfUnits ); 

    unsigned newTime = SystemTimer->time();
#ifdef DEBUG_LAG
    unsigned diff = newTime - _lastTime;
#endif // DEBUG_LAG
    _lastTime = newTime;

    if ( _writeSequential == -1 &&
         _readSequential == -1 )
        return 1;

    if ( _freeBehindLag <= _minFreeBehindLag ) { 
        return GlobalLRUReplacementPolicy::pageOut( noOfUnits ); 
    }

    _sglock.acquire();
    unsigned oldFreeBehindLag = _freeBehindLag;
    if ( _freeBehindLag > _slowFreeBehindLag )
        _freeBehindLag -= _slowFreeBehindLag;
    else
        _freeBehindLag -= _minFreeBehindLag;
    if ( _freeBehindLag < _minFreeBehindLag )
        _freeBehindLag = _minFreeBehindLag;
#ifdef DEBUGLAG
    if ( debugLag ) Debug 
    << this << "::pageOut() freeBehindLag " 
    << _freeBehindLag
    << " diff " << diff << "\n" << eor;
#endif // DEBUGLAG
    if ( _freeBehindLag >= oldFreeBehindLag ) {
        // new freeBehindLag is less than previous value,
        // no need to check to see if can free behind
        _sglock.release();
        return GlobalLRUReplacementPolicy::pageOut( noOfUnits );
    }

    // new freeBehindLag is smaller then previous value
    // see if can doFreeBehind, i.e. avoid invoking GLRU pager daemon

    int writeLag;
    if ( _writeSequential == -1 )
        writeLag = -9999999;
    else
        writeLag = _writePointer - _freeBehindLag - _writeFreeBehind ;
    int readLag;
    if ( _readSequential == -1 )
        readLag = -99999999;
    else
        readLag = _readPointer - _freeBehindLag - _readFreeBehind ;

    if ( readLag > writeLag ) {
        doFreeBehind( _readFreeBehind, _readPointer );
    } else {
        doFreeBehind( _writeFreeBehind, _writePointer );
    }
    // _sglock released by doFreeBehind
    return 1;
}

HintedGlobalLRUReplacementPolicy::HintedGlobalLRUReplacementPolicy
( PagedMemoryObjectCache * cache )
    : SmarterGlobalLRUReplacementPolicy( cache ) 
{
    _hints = _cache->memoryObject()->hints();
    _origMaxFreeBehindLag = _maxFreeBehindLag;
    Assert( _cache->memoryObject()->isKindOf(PersistentMemoryObjectClass) );
    updateParams( _hints );
}

void 
HintedGlobalLRUReplacementPolicy::updateParams( int newHints )
{
    // parse hints and convert into cache params
    if ( h_temp(newHints) ) {
        newHints &= ~H_SINGLE;
    }
    if ( h_single(newHints) || h_large(newHints) ) {
        _doSyncBehind = 1;
    } else {
        _doSyncBehind = 0;
    }
    if ( h_single(newHints) ) {
        _maxFreeBehindLag = _clusterSize;
        _minFreeBehindLag = _clusterSize;
    } else {
        if ( h_large(newHints) ) {
            _maxFreeBehindLag = 
            _origMaxFreeBehindLag;
            _minFreeBehindLag = 
            _clusterSize * LARGE_FREEBEHIND_CLUSTERS;
        } else {
            _maxFreeBehindLag = _origMaxFreeBehindLag;
            _minFreeBehindLag = _origMaxFreeBehindLag;
        }
    }
    if ( h_random(newHints) ) {
        // disable prefetching
        _readAheadLag = 0;
        _doPrefetchFirst = 0;
    } else {
        // allow readahead
        _doPrefetchFirst = 1;
        if ( h_large(newHints) ) {
            // for large files readahead
            _readAheadLag = _clusterSize * LARGE_READAHEAD_CLUSTERS;
        } else {
            // for small files prefetch all
            _readAheadLag = _numberOfUnits;
        }
    }
    extern unsigned MemoryObjectCachingContainerPeriod;
    PersistentMemoryObject * pmo = (PersistentMemoryObject *)
    _cache->memoryObject();
    if ( h_single(newHints) ) {
        pmo->setZombieDelay( MemoryObjectCachingContainerPeriod / 2 );
    } else {
        if ( h_temp(newHints) )
            pmo->setZombieDelay( MemoryObjectCachingContainerPeriod * 2);
        else
            pmo->setZombieDelay( MemoryObjectCachingContainerPeriod );
    }
    _hints = newHints;
}

void
HintedGlobalLRUReplacementPolicy::notify( int event )
{
    int moHints = _cache->memoryObject()->hints();
    if ( _hints != moHints )
        updateParams( moHints );
}

void 
HintedGlobalLRUReplacementPolicy::preCacheRead( unsigned from, unsigned to )
{
    int moHints = _cache->memoryObject()->hints();
    if ( _hints != moHints )
        updateParams( moHints );
    SmarterGlobalLRUReplacementPolicy::preCacheRead( from, to );
}

void 
HintedGlobalLRUReplacementPolicy::preCacheWrite( unsigned from, unsigned to )
{
    int moHints = _cache->memoryObject()->hints();
    if ( _hints != moHints )
        updateParams( moHints );
    SmarterGlobalLRUReplacementPolicy::preCacheWrite( from, to );
}

extern ClassStar PagedMemoryObjectCacheClass;

ProxiableObject *
GlobalLRUReplacementPolicyConstructor( ProxiableObject * source )
{
    if( !source->isKindOf( PagedMemoryObjectCacheClass ) )
        return 0;
    return new
    GlobalLRUReplacementPolicy((PagedMemoryObjectCache *) source);
}

ProxiableObject *
SmarterGlobalLRUReplacementPolicyConstructor( ProxiableObject * source )
{
    if( !source->isKindOf( PagedMemoryObjectCacheClass ) )
        return 0;
    return new SmarterGlobalLRUReplacementPolicy
    ((PagedMemoryObjectCache *) source);
}
 
ProxiableObject *
HintedGlobalLRUReplacementPolicyConstructor( ProxiableObject * source )
{
    if( !source->isKindOf( PagedMemoryObjectCacheClass ) )
        return 0;
    return new HintedGlobalLRUReplacementPolicy
    ((PagedMemoryObjectCache *) source);
}

#include "ClassInstaller.h"
DEFINECLASS("GlobalLRUReplacementPolicy",GlobalLRUReplacementPolicy,ReplacementPolicy,&GlobalLRUReplacementPolicyConstructor,0);
DEFINECLASS("SmarterGlobalLRUReplacementPolicy",SmarterGlobalLRUReplacementPolicy,GlobalLRUReplacementPolicy,&SmarterGlobalLRUReplacementPolicyConstructor,0);
DEFINECLASS("HintedGlobalLRUReplacementPolicy",HintedGlobalLRUReplacementPolicy,SmarterGlobalLRUReplacementPolicy,&HintedGlobalLRUReplacementPolicyConstructor,0);
