#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * IOWindow.cc - 
 */

#include "Assert.h"
#include "Byte.h"
#include "Debug.h"
#include "Domain.h"
#include "Kernel.h"
#include "MemoryObjectView.h"
#include "ThisProcess.h"
#include "TimeoutTimer.h"
#include "IOWindow.h"

static const unsigned TooManyWindows = 4;
static const unsigned DefaultIOWindowCacheSize = 48;

static const unsigned primeTableSize = 20;
static unsigned primeTable[primeTableSize] = { 
	19, 41, 61, 83, 
	101, 127, 149, 179, 
	211, 229, 251, 307, 
	353, 401, 467, 523, 
	601, 673, 733, 853 
};

Lock IOWindowCache::_lock;
unsigned IOWindowCache::_age = 0;
IOWindowCache * TheIOWindowCache = 0;
	
IOWindowCache::IOWindowCache( unsigned size )
{
	if ( size == 0 ) {
		_size = DefaultIOWindowCacheSize;
	} else {
		_size = size;
	}

	Debug   << "IOWindowCache::IOWindowCache( size " << _size << " )\n";

	_infos = new IOWindowInfo[ _size ];
	Assert( _infos != 0 );

	// initialize free list

	IOWindowInfo * info = _infos;
	unsigned i;
	for( i = 0; i < _size; i++ ) {
		info->_prev = info - 1;
		info->_next = info + 1;
//		info->_list = &_freeList;
		info++;
	}
	_infos[0]._prev = 0;
	_infos[_size - 1]._next = 0;
	_freeList = _infos;

	// construct hash table

	for( i = 0; i < primeTableSize; i++ ) {
		if ( primeTable[i] >= _size ) 
			break;
	}
	i += 2;
	if ( i >= primeTableSize ) 
		i = primeTableSize - 1; 
	_hashSize = primeTable[i];
	_hashTable = new IOWindowInfo * [_hashSize];
	Assert( _hashTable != 0 );
	for( i = 0; i < _hashSize; i++ ) {
		_hashTable[i] = 0;
	}

	// find _adt
	_adt = TheKernel->domain()->translation();
}

inline
void 
IOWindowCache::free( IOWindowInfo * info, int untranslate )
{
	MemoryObjectCache * cache = info->_cache;
	Assert( info->_lock == 0 );
	cache->setNoOfWindows( cache->noOfWindows() - 1 );
       	removeFromList( _hashTable + hash( info->_cache, info->_from ), info );
	if ( untranslate ) {
		cache->untranslatable( info->_from,
 				       info->_from + 
				       cache->windowSizeRatioMask(),
				       _adt, info->_baseAddr );
	}
	info->_view->setSource( 0 );
       	info->_cache = 0;
	info->_from = 0;
       	info->_age = 0;
       	addToList( &_freeList, info );
}

IOWindowInfo *
IOWindowCache::victimize( MemoryObjectCache * cache )
{
	Debug   << "IOWindowCache::victimize( " << cache << ")\n" << eor;

	Assert( cache != 0 );

	unsigned tries = 0;
	IOWindowInfo * oldest = 0;
	IOWindowInfo * oldestMatch = 0;
	
	unsigned oldestAge;
	unsigned oldestMatchAge;
	unsigned desiredLog2WindowSize = cache->log2WindowSize();

	IOWindowInfo * unused = _freeList;
	while ( unused && unused->_log2WindowSize != desiredLog2WindowSize ) {
		unused = unused->_next;
	}
	oldest = ( unused == 0 ? _freeList : unused );
	if ( oldest != 0 ) {
		Debug 	<< "IOWindowCache::victimize() remove from free list\n"
			<< eor;
		removeFromList( &_freeList, oldest );
		Assert( oldest->_cache == 0 );
	}

	while ( oldest == 0 ) {
       		IOWindowInfo * info = _infos;
		oldestAge = 0xffffffff;
	        oldestMatchAge = 0xffffffff;
	        for (unsigned i = 0; i < _size; i++, info++ ) {
			if ( info->_lock != 0 )
				continue;
	               	if ( info->_age < oldestAge ) {
	               		oldestAge = info->_age;
	                       	oldest = info;
	               	}
			if ( info->_log2WindowSize == desiredLog2WindowSize &&
			     info->_age < oldestMatchAge ) {
				oldestMatchAge = info->_age;
				oldestMatch = info;
			}
		}
		if ( oldest != 0 ) {
			if ( oldestMatch != 0 ) {
				unsigned diff = 
					( ( _age - oldestMatchAge ) * 100 ) /
					( _age - oldestAge );
				if ( diff > 80 ) 
					oldest = oldestMatch;
			}
			break;
		}
		tries++;
		if ( tries > 1000 ) {
			Console() 
			<< "IOWindowCache::victimize() possible deadlock "
			<< eor;
			Assert( NOTREACHED );
		}
		release();
		Console() << "IOWindowCache::victimize() cannot find victim "
			  << thisProcess() << "\n" << eor;
		TimeoutTimer timer;
		timer.start( 1000 );
		timer.await();
		acquire();
	}

	Debug   << "IOWindowCache::victimize() victim "
		<< asAddress( oldest ) << " _cache " << oldest->_cache
		<< "\n" << eor;
	if ( desiredLog2WindowSize != oldest->_log2WindowSize ) {
		Debug   << "IOWindowCache::victimize() windowSize different\n"
			<< eor;
		if ( oldest->_baseAddr != 0 ) {
			char * result = (char *)
				TheKernel->domain()->remove
					( oldest->_view, oldest->_baseAddr );
			Assert( result == oldest->_baseAddr );
			oldest->_baseAddr = 0;
		} 
		oldest->_log2WindowSize = desiredLog2WindowSize;
	}
	if ( oldest->_cache != 0 && oldest->_baseAddr != 0 ) {
		Debug   << "IOWindowCache::victimize() windowSize same\n"
			<< eor;
		unsigned oldestFrom = oldest->_view->startUnit();
		unsigned oldestTo = oldestFrom + 
				    oldest->_cache->windowSizeRatioMask();
		oldest->_cache->untranslatable
			( oldestFrom, oldestTo, _adt, oldest->_baseAddr );
	}

	if ( oldest->_cache != 0 ) {
		Debug   << "IOWindowCache::victimize() not from free list\n"
			<< eor;
       		removeFromList
			( _hashTable + hash( oldest->_cache, oldest->_from ), 
		  	  oldest );
		oldest->_cache->setNoOfWindows
			( oldest->_cache->noOfWindows() - 1 );
	}
	cache->setNoOfWindows( cache->noOfWindows() + 1 );

	return oldest;
}

IOWindowInfo * 
IOWindowCache::basicAcquireIOWindow( MemoryObjectCache * cache, unsigned from )
{
	Debug   << "IOWindowCache::basicAcquireIOWindow( "
		<< cache << ", " << from << " )\n" << eor;

	unsigned slot = hash( cache, from );
	IOWindowInfo * info = _hashTable[slot];
	while( info ) {
		if ( cache == info->_cache && from == info->_from ) {
			info->_lock ++;
			info->_age = age();
			Debug   << "IOWindowCache::basicAcquireIOWindow() "
				<< "cached as " << asAddress( info ) 
				<< "\n" << eor;
			return info;
		}
		info = info->_next;
	}
	IOWindowInfo * found = victimize( cache );
	found->_view->setLog2UnitSize( cache->log2UnitSize() );
	found->_view->setNumberOfUnits( cache->windowSizeRatio() );
	found->_view->setStartUnit( from );
	found->_view->setSource( cache->memoryObject() );
	if ( found->_baseAddr == 0 ) {
		found->_baseAddr = (char *)
			TheKernel->domain()->add( found->_view, NoAccess );
	}
	found->_cache = cache;
	found->_from = from;
	found->_lock = 1;
	found->_age = age();
	addToList( _hashTable + slot, found );

	if ( cache->noOfWindows() > TooManyWindows )
		freeExcess( cache );

	return found;
}

void
IOWindowCache::freeIOWindow( MemoryObjectCache * cache, int untranslate )
{
	Debug   << "IOWindowCache::freeIOWindow( " << cache << ", "
		<< untranslate << " )\n" << eor;

	acquire();
        IOWindowInfo * info = _infos;
        for (unsigned i = 0; i < _size; i++, info++ ) {
                if ( info->_cache == cache ) {
			Debug 	<< "IOWindowCache::freeIOWindow() found "
				<< asAddress( info ) << "\n" << eor;
			free( info, untranslate );
                }
	}
	release();
}

void
IOWindowCache::freeExcess( MemoryObjectCache * cache )
{
	Debug   << "IOWindowCache::freeExcess( " << cache << " )\n" << eor;

	while ( cache->noOfWindows() > TooManyWindows ) {
        	IOWindowInfo * info = _infos;
		unsigned oldestAge = 0xffffffff;
		IOWindowInfo * oldest = 0;
        	for (unsigned i = 0; i < _size; i++, info++ ) {
               		if ( info->_cache == cache &&
			     info->_age < oldestAge &&
			     info->_lock == 0 ) {
				oldest = info;
				oldestAge = info->_age;
			}
		}
		if ( oldest == 0 )
			break;
		free( oldest, 1 );
	}
}

