#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Map.cc
 */

#include "Assert.h"
#include "Map.h"
#include "MultiLevelTable.h"
#include "PhysicallyAddressableUnit.h"
#include "PhysicalMemoryManager.h"


#ifdef __GNUC__
#pragma implementation
#endif

Map::Map( int size, int log2UnitSize )
	: _size( size ), _log2UnitSize( log2UnitSize )
{
	_residentSize = 0;
	_log2PauUnitSizeRatio = log2UnitSize - Log2PAUUnitSize;
	Assert( _log2PauUnitSizeRatio >= 0 );
	_pauUnitSizeRatio = 1 << _log2PauUnitSizeRatio;
	_pauUnitSizeRatioMask = _pauUnitSizeRatio - 1;
}

Map::~Map()
{
	Assert( _residentSize >= 0 && _residentSize <= _size );
	Assert( _residentSize == 0 );
}

PhysicallyAddressableUnit * 
Map::pauList( unsigned from, unsigned to )
{
	PhysicallyAddressableUnit * lastPau;
	acquire();
	PhysicallyAddressableUnit * pau = basicPauList( from++, lastPau );
	if ( pau ) {
		while ( from <= to ) {
			PhysicallyAddressableUnit * nextLastPau;
			PhysicallyAddressableUnit * currentPau = 
				basicPauList( from++, nextLastPau );
			Assert( currentPau != 0 );
			Assert( lastPau->nextLink() == currentPau );
			lastPau = nextLastPau;
		}
		Assert( lastPau );
	}
	release();	
	return pau;
}

PhysicallyAddressableUnit * 
Map::findAndRemove( unsigned from, unsigned to, unsigned &start, unsigned &end )
{
	PhysicallyAddressableUnit * ret;
	PhysicallyAddressableUnit * pau = 0;
	PhysicallyAddressableUnit * lastPau;
	acquire();

	// skip non-existent units
	start = from;
	while ( start <= to && (pau = basicRemoveUnit( start, lastPau )) == 0 )
		start++; 
	if ( start > to ) {
		release();
		return 0;
	}
	Assert( pau );
	ret = pau;

	// zero out link in previous unit if it exist
	if ( start > 0 ) {
		PhysicallyAddressableUnit * tmpLastPau;
		PhysicallyAddressableUnit * tmpPau = 
			basicPauList( start - 1, tmpLastPau );
		if ( tmpPau )
			tmpLastPau->setNextLink( 0 );
	}

	PhysicallyAddressableUnit * nextLastPau = 0;
	end = start + 1;
 	while ( end <= to && 
		 ( pau = basicRemoveUnit( end, nextLastPau ) ) != 0 ) {
		lastPau->setNextLink( pau );
		lastPau = nextLastPau;
		_residentSize--;
		end++;
	}
	end--;
	lastPau->setNextLink( 0 );
	_residentSize--;
	Assert( _residentSize >= 0 );
	release();	
	return ret;
}

int
Map::findNextExist( unsigned from, unsigned to )
{
        PhysicallyAddressableUnit * lastPau;
	unsigned start = from;
        acquire();
        while ( start <= to && basicPauList( start, lastPau ) == 0 )
                start++;
        release();
        return start;
}

int
Map::findNextExist( unsigned from, unsigned to, unsigned &start, unsigned &end,
		    PhysicallyAddressableUnit * & startPau )
{
	PhysicallyAddressableUnit * lastPau;
	startPau = 0;
        acquire();
	Assert( from <= to );
	start = from;
	while ( start <= to && 
	        ( startPau = basicPauList( start, lastPau ) ) == 0 )
		start++; 
	end = start + 1;
	while ( end <= to && basicPauList( end, lastPau ) != 0 )
		end++;
	end--;
        release();
        return ( start <= to ); 
}

int
Map::findNextNonExist( unsigned from, unsigned to, 
		       unsigned &start, unsigned &end,
		       PhysicallyAddressableUnit * & fromPau )

{
        PhysicallyAddressableUnit * lastPau;
        acquire();
	Assert( from <= to );
	fromPau = basicPauList( from, lastPau );
	if ( fromPau != 0 ) {
		start = from + 1;
        	while ( start <= to && basicPauList( start, lastPau ) != 0 )
               		start++; 
	} else {
		start = from;
	}
        end = start + 1; 
        while ( end <= to && basicPauList( end, lastPau ) == 0 ) 
                end++; 
        end--; 
        release();
        return ( start <= to ); 
}

inline
int 
Map::unitModified( PhysicallyAddressableUnit * pau )
{
	for ( unsigned i = 0; i < _pauUnitSizeRatio; i++, pau = pau->nextLink() ) {
//		Assert( pau && pau->owner() == this );
		if ( pau->modified() )
			return 1;
	}
	return 0;
}

int
Map::findNextModified( unsigned from, unsigned to, unsigned &start, 
		       unsigned &end, PhysicallyAddressableUnit * & startPau )
{
	PhysicallyAddressableUnit * lastPau;
	PhysicallyAddressableUnit * pau;

	startPau = 0;
        acquire();
	Assert( from <= to );
	start = from;
	while ( start <= to && 
	        ( ( startPau = basicPauList( start, lastPau ) ) == 0 ||
		  unitModified( startPau ) == 0 ) )
		start++; 
	end = start + 1;
	while ( end <= to && 
	        ( pau = basicPauList( end, lastPau ) ) != 0 &&
	        unitModified( pau ) != 0 ) {
		end++;
	}
	end--;
        release();
        return ( start <= to ); 
}

int 
Map::exists( unsigned unit )
{
	PhysicallyAddressableUnit * pau = pauList( unit, unit );
	if ( pau )
		return 1;
	return 0;
}

int
Map::modified( unsigned unit )
{
	PhysicallyAddressableUnit * pau = pauList( unit, unit );
	if ( pau ) {
		for ( unsigned i = 0; i < _pauUnitSizeRatio; i++, 
		      pau = pau->nextLink() ) {
			Assert( pau && pau->owner() == this );
			if ( pau->modified() )
				return 1;
		}
	}
	return 0;
}

int
Map::referenced( unsigned unit )
{
	PhysicallyAddressableUnit * pau = pauList( unit, unit );
	if ( pau ) {
		for ( unsigned i = 0; i < _pauUnitSizeRatio; i++, 
		      pau = pau->nextLink() ) {
			Assert( pau && pau->owner() == this );
			if ( pau->referenced() )
				return 1;
		}
	}
	return 0;
}

ProtectionLevel
Map::protection( unsigned unit )
{
	PhysicallyAddressableUnit * pau = pauList( unit, unit );
	ProtectionLevel prot = ( ProtectionLevel ) NoAccess;
	if ( pau ) {
		prot = pau->protection();
		for ( unsigned i = 0; i < _pauUnitSizeRatio; i++, 
		      pau = pau->nextLink() ) {
			Assert( pau && pau->owner() == this );
			prot = Min( prot, pau->protection() );
		}
	}
	return prot;
}

void
Map::setModified( unsigned unit, int v )
{
	PhysicallyAddressableUnit * pau = pauList( unit, unit );
	Assert( pau && pau->owner() == this );
	for ( unsigned i = 0; i < _pauUnitSizeRatio; i++, pau = pau->nextLink() ) {
		Assert( pau && pau->owner() == this );
		pau->setModified( v );
	}
}

void
Map::setModified( unsigned from, unsigned to, int v )
{
	PhysicallyAddressableUnit * pau = pauList( from, to );
	Assert( pau && pau->owner() == this );
	int count = ( to - from + 1 ) << _log2PauUnitSizeRatio;
	for ( int i = 0; i < count; i++, pau = pau->nextLink() ) {
		Assert( pau && pau->owner() == this );
		pau->setModified( v );
	}
}

void
Map::setReferenced( unsigned unit, int v )
{
	PhysicallyAddressableUnit * pau = pauList( unit, unit );
	Assert( pau && pau->owner() == this );
	for ( unsigned i = 0; i < _pauUnitSizeRatio; i++, pau = pau->nextLink() ) {
		Assert( pau && pau->owner() == this );
		pau->setReferenced( v );
	}
}

void
Map::setReferenced( unsigned from, unsigned to, int v )
{
	PhysicallyAddressableUnit * pau = pauList( from, to );
	Assert( pau && pau->owner() == this );
	int count = ( to - from + 1 ) << _log2PauUnitSizeRatio;
	for ( int i = 0; i < count; i++, pau = pau->nextLink() ) {
		Assert( pau && pau->owner() == this );
		pau->setReferenced( v );
	}
}

void 
Map::setProtection( unsigned unit, ProtectionLevel prot )
{
	PhysicallyAddressableUnit * pau = pauList( unit, unit );
	Assert( pau && pau->owner() == this );
	for ( unsigned i = 0; i < _pauUnitSizeRatio; i++, pau = pau->nextLink() ) {
		Assert( pau && pau->owner() == this );
		pau->setProtection( prot );
	}
}

void
Map::setProtection( unsigned from, unsigned to, ProtectionLevel prot )
{
	PhysicallyAddressableUnit * pau = pauList( from, to );
	Assert( pau && pau->owner() == this );
	int count = ( to - from + 1 ) << _log2PauUnitSizeRatio;
	for ( int i = 0; i < count; i++, pau = pau->nextLink() ) {
		Assert( pau && pau->owner() == this );
		pau->setProtection( prot );
	}
}

void 
Map::setReferencedModified( unsigned unit, int referenced, int modified )
{
	PhysicallyAddressableUnit * pau = pauList( unit, unit );
	Assert( pau && pau->owner() == this );
	for ( unsigned i = 0; i < _pauUnitSizeRatio; i++, pau = pau->nextLink() ) {
		Assert( pau && pau->owner() == this );
		pau->setReferenced( referenced );
		pau->setModified( modified );
	}
}    

void 
Map::setReferencedModified( unsigned from, unsigned to, 
			    int referenced, int modified )
{
	PhysicallyAddressableUnit * pau = pauList( from, to );
	Assert( pau && pau->owner() == this );
	int count = ( to - from + 1 ) << _log2PauUnitSizeRatio;
	for ( int i = 0; i < count; i++, pau = pau->nextLink() ) {
		Assert( pau && pau->owner() == this );
		pau->setReferenced( referenced );
		pau->setModified( modified );
	}
}


/*
 * TableMap
 */

PhysicallyAddressableUnit * 
TableMap::basicPauList( unsigned unit,
		        PhysicallyAddressableUnit * & lastPau )
{
	PhysicallyAddressableUnit * pau;
	PhysicallyAddressableUnit ** paup = where( unit, 0 );
	if ( paup == 0 || *paup == 0 ) {
		lastPau = 0;
		return 0;
	}
	pau = * paup;
	Assert( pau && pau->owner() == this );
	for ( unsigned i = 0; i < _pauUnitSizeRatioMask; i++ ) {
		PhysicallyAddressableUnit * prevPau = * paup;
		paup++;
		Assert( * paup && ( * paup )->owner() == this );
		Assert( prevPau->nextLink() == * paup );
	}
	lastPau = * paup;
	return pau;
}

PhysicallyAddressableUnit * 
TableMap::basicRemoveUnit( unsigned unit,
			   PhysicallyAddressableUnit * & lastPau )
{
	PhysicallyAddressableUnit * pau;
	PhysicallyAddressableUnit ** paup = where( unit, 0 );
	if ( paup == 0 || * paup == 0 ) {
		lastPau = 0;
		return 0;
	}
	pau = * paup;
	Assert( pau && pau->owner() == this );
	for ( unsigned i = 0; i < _pauUnitSizeRatioMask; i++ ) {
		PhysicallyAddressableUnit ** prevPaup = paup;
		paup++;
		Assert( * paup && ( * paup )->owner() == this );
		Assert( ( * prevPaup )->nextLink() == * paup );
	        ( * prevPaup )->setOwner( 0 );
		* prevPaup = 0;
	} 
	lastPau = *paup;
	( * paup )->setOwner( 0 );
	* paup = 0;

	return pau;
}

void 
TableMap::addPauList( PhysicallyAddressableUnit * plist, 
		      unsigned from, unsigned to )
{
	acquire();
	PhysicallyAddressableUnit * lastPau;	
	_residentSize += to - from + 1;
	Assert( _residentSize <= _size );
	if ( from > 0 ) {
		PhysicallyAddressableUnit * pau = basicPauList( from - 1,
							        lastPau );
		if ( pau )
			lastPau->setNextLink( plist );
	}
	PhysicallyAddressableUnit ** paup = 0;
	while ( from <= to ) {
	        paup = where( from++, 1 );
		Assert( paup );
		for ( unsigned i = 0; i < _pauUnitSizeRatio; i++ ) {
			Assert( plist && plist->owner() == 0 );
			Assert( * paup == 0 );
			plist->setOwner( this );
			* paup ++ = plist;
			plist = plist->nextLink();
		}
	} 
	Assert( plist == 0 );
	if ( from < _size ) {
		PhysicallyAddressableUnit * prevPau = *( paup - 1 );
		Assert( prevPau && prevPau->owner() == this );
		prevPau->setNextLink( basicPauList( from, lastPau ) );
	}
	release();
}

/*
 * MultiLevelTableMap
 */

Lock MultiLevelTableMap::_staticLock;

MultiLevelTableMap::MultiLevelTableMap( unsigned size, unsigned log2UnitSize )
	: TableMap( size, log2UnitSize )
{
	_lock = &_staticLock;

	unsigned log2ChunkSize = 3;
	unsigned log2IndexSize = 3;
	if ( _log2PauUnitSizeRatio > log2ChunkSize ) {
		log2ChunkSize = _log2PauUnitSizeRatio;
	}
	_table = new MultiLevelTable( log2ChunkSize, log2IndexSize );
}

MultiLevelTableMap::~MultiLevelTableMap()
{
	delete _table;
}

PhysicallyAddressableUnit ** 
MultiLevelTableMap::where( unsigned unit, int alloc )
{
	return ( PhysicallyAddressableUnit ** ) 
		_table->where( unit << _log2PauUnitSizeRatio, alloc );
}

void
MultiLevelTableMap::setSize( unsigned size )
{
	acquire();
	_size = size;
	Assert( _residentSize <= _size );
	release();
}

/*
 * ConstantSizeTableMap
 */

ConstantSizeTableMap::ConstantSizeTableMap( int size, int log2PauUnitSizeRatio )
	: TableMap( size, log2PauUnitSizeRatio )
{
	_lock = &_instanceLock;
	Assert( size > 0 );
	int alloc = size << _log2PauUnitSizeRatio;
	_table = new PhysicallyAddressableUnit * [ alloc ];
	PhysicallyAddressableUnit ** p = _table;
	while ( alloc-- > 0 )
		*p++ = 0;
}

ConstantSizeTableMap::~ConstantSizeTableMap()
{
	delete _table;
}

PhysicallyAddressableUnit ** 
ConstantSizeTableMap::where( unsigned unit, int alloc )
{
	return _table + ( unit << _log2PauUnitSizeRatio );
}

void
ConstantSizeTableMap::setSize( unsigned size )
{
	Assert( size == _size );
}

/*
 * PremappedMap
 */

PremappedMap::PremappedMap( int size, int log2UnitSize,
			    const char * baseAddr )
	: Map( size, log2UnitSize )
{
	_lock = &_instanceLock;
	Assert( _log2PauUnitSizeRatio == 0 );
	_baseAddr = baseAddr;
	_residentSize = _size;
}

void
PremappedMap::setSize( unsigned size )
{
	Assert( size == _size );
}

PhysicallyAddressableUnit * 
PremappedMap::basicPauList( unsigned unit, 
			    PhysicallyAddressableUnit * & lastPau )
{
	lastPau = MainMemory->lookup( _baseAddr + (unit << Log2PAUUnitSize) ); 
	// Francis - Owner is not necessarily this. If this map belongs to a preAllocated=0 cache
	// the owner might be someone else
//	Assert( lastPau && lastPau->owner() == this );
	return lastPau;
}

PhysicallyAddressableUnit * 
PremappedMap::basicRemoveUnit( unsigned unit, 
			      PhysicallyAddressableUnit * & lastPau )
{
	Assert( NOTREACHED );
	return 0;
}

void 
PremappedMap::addPauList( PhysicallyAddressableUnit * plist, 
			          unsigned from, unsigned to )
{
	Assert( NOTREACHED );
}

