#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MemoryObject.c - memory object classes.
 */

#include "Assert.h"
#include "Class.h"
#include "Domain.h"
#include "FileSystem.h"
#include "FileSystemHints.h"
#include "Kernel.h"
#include "MemoryObject.h"
#include "MemoryObjectAsync.h"
#include "MemoryObjectCache.h"
#include "PhysicallyAddressableUnit.h"
#include "Raid.h"
#include "RangeLocks.h" 
#include "ReplacementPolicy.h" 
#include "ThisProcess.h"

#ifdef __GNUC__
#pragma implementation
#endif

/* #define JEFF */

Lock MemoryObject::_lock;
Lock MemoryObject::_hintLock;

MemoryObject::MemoryObject( unsigned int numberOfUnits, unsigned log2UnitSize )
    : MemoryRange( numberOfUnits, log2UnitSize ),
      _cacheLog2UnitSize( 0 ), _cache( 0 ), _hints( 0 ), 
      _segmentSize( 16 << 20 ) // 16 Mb
{
    initDebug();
    Assert( this != 0 );
    Raids( RaidConstructor ) 
    << this << "::MemoryObject( " << numberOfUnits << ", "
    << log2UnitSize << " )\n" << eor;
    _moLock = 0;
    _cacheClass = 0;
    _replacementPolicyClass = 0;
    _cacheLog2UnitSize = 
    log2UnitSize > Log2PAUUnitSize ? 
    log2UnitSize : Log2PAUUnitSize;
    _cacheLog2UnitSizeRatio =
    _cacheLog2UnitSize - log2UnitSize;

    //
    // HACK: _protection used to be 'NoAccess'.  The subclass
    //   PersistentMemoryObject always set it to ReadWrite, and since
    //   that class was taken out of the kernel I didn't want to expose
    //   the _protection value in MemoryObject.  I don't think that
    //   the MemoryObject should have protection in at all since the
    //   it is given a protection when it is added to a Domain anyway,
    //   so I don't feel so bad about making the default to be ReadWrite.
    //   Someday the _protection should be removed from MemoryObject
    //   (and probably the PhysicallyAddressableUnit too).
    //  - Dave Dykstra, 8/27/91
    _protection = ReadWrite;
}

MemoryObject::~MemoryObject()
{
    Raid( RaidConstructor )( "%N::~MemoryObject()\n", this );

    Assert( _referenceCount.value() == 0 );
}

void
MemoryObject::noRemainingReferences()
{
    if( ! transient() ) {
        flush();
    }
    delete this;
}

#ifndef INLINE_MOLOCKS

Process * 
MemoryObject::acquire( unsigned from, unsigned to, int block, int locked )
{
    Raids( RaidMethod ) 
    << this << "::acquire( " << from << ", " << to << ", "
    << block << ", " << locked << " )\n" << eor;

    Process * process = ( _moLock ?
                          _moLock->acquire( from, to, block, locked ) : 0 );

    Raids( RaidMethod ) 
    << this << "::acquire( " << from << ", " << to << ", "
    << block << ", " << locked << " ) returns " << (int) process
    << "\n" << eor;

    return process;
}

Process *
MemoryObject::release( unsigned from, unsigned to )
{
    Raids( RaidMethod ) 
    << this << "::release( " << from << ", " << to 
    << " )\n" << eor;

    Process * result = ( _moLock ? _moLock->release( from, to ) : (Process *) 0x4 );

    Raids( RaidMethod ) 
    << this << "::release( " << from << ", " << to 
    << " ) returns " << (int) result << "\n" << eor;

    return result;
}

void 
MemoryObject::acquireAll()
{
    Raids( RaidMethod ) 
    << this << "::acquireAll()\n" << eor;

    if ( _moLock )
        _moLock->acquireAll();

    Raids( RaidMethod ) 
    << this << "::acquireAll() returns\n" << eor;
}

void
MemoryObject::releaseAll()
{
    Raids( RaidMethod ) 
    << this << "::releaseAll()\n" << eor;

    if ( _moLock )
        _moLock->releaseAll();

    Raids( RaidMethod ) 
    << this << "::releaseAll()\n" << eor;
}

void 
MemoryObject::acquireSize()
{
    Raids( RaidMethod ) 
    << this << "::acquireSize()\n" << eor;

    if ( _moLock )
        _moLock->acquire( MaxUse, MaxUse );

    Raids( RaidMethod ) 
    << this << "::acquireSize() returns\n" << eor;
}

void
MemoryObject::releaseSize()
{
    Raids( RaidMethod ) 
    << this << "::releaseSize()\n" << eor;

    if ( _moLock )
        _moLock->release( MaxUse, MaxUse );

    Raids( RaidMethod ) 
    << this << "::releaseSize()\n" << eor;
}

void 
MemoryObject::downgradeFromAll( unsigned from, unsigned to )
{
    Raids( RaidMethod ) 
    << this << "::downgradeFromAll( " << from << ", " << to
    << " )\n" << eor;

    if ( _moLock )
        _moLock->downgradeFromAll( from, to );

    Raids( RaidMethod ) 
    << this << "::downgradeFromAll( " << from << ", " << to
    << " ) returns\n" << eor;
}

#endif // not INLINE_MOLOCKS

/*
 * By default, all MemoryObjects are transient.
 * I.e. they dont cause the cache to be flushed when they are deleted.
 * Subclasses (i.e. UNIXInode) should overload this to say "no" if there
 * is a persistent reference to the object.
 */
int
MemoryObject::transient()
{
    return( 1 );
}

/*
 * By default, all MemoryObjects has a fixed size, i.e. it cannot grow
 * strink.
 */
int
MemoryObject::fixedSize()
{
    return( 1 );
}

/*
 * reads and writes
 */


int
MemoryObject::read( unsigned startUnit, int count, char * addr, int nbytes )
{
    Raids( RaidMethod ) 
    << this << "::read( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes << " )\n" << eor;

    if ( count <= 0 || nbytes <= 0 )
        return 0;

    unsigned lastUnit = startUnit + count - 1;

    acquire( startUnit, lastUnit );

    int result = basicRead( startUnit, count, addr, nbytes, 0 );

    release( startUnit, lastUnit );

    Raids( RaidMethod ) 
    << this << "::read( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes << " ) returns " << result << "\n" << eor;

    return result;
}

void
MemoryObject::asyncRead( unsigned startUnit, int count, char * addr, int nbytes,
                         IOAsyncClosure * c )
{
    Raids( RaidMethod )
    << this << "::asyncRead( " << startUnit
    << ", " << count
    << ", " << asAddress( addr )
    << ", " << nbytes 
    << ", " << asAddress( c ) << " )\n" << eor;

    if ( count <= 0 || nbytes <= 0 ) {
        c->activate( 0 );
        return;
    }

    unsigned lastUnit = startUnit + count - 1;
    int result;

    if ( _moLock ) {
        ReleaseIOAsyncClosure * myClosure = 
        new ReleaseIOAsyncClosure( c, this, startUnit, 
                                   lastUnit );
        acquire( startUnit, lastUnit );
        result = basicRead( startUnit, count, addr, nbytes, myClosure );
    } else {
        result = basicRead( startUnit, count, addr, nbytes, c );
    }

    Raids( RaidMethod )
    << this << "::asyncRead( " << startUnit
    << ", " << count
    << ", " << asAddress( addr )
    << ", " << nbytes << " ) returns " << result << "\n" << eor;

    return;
}

int
MemoryObject::basicRead( unsigned startUnit, unsigned count, char * addr, 
                         unsigned nbytes, IOAsyncClosure * c )
{
    Raids( RaidProtected ) 
    << this << "::basicRead( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes 
    << ", " << asAddress( c ) << " )\n" << eor;
#ifdef JEFF
    Console() << this << "::" << __FUNCTION__ << "( " << startUnit
              << ", " << count 
              << ", " << asAddress( addr ) 
              << ", " << nbytes 
              << ", " << asAddress( c ) << " )\n" << eor;
#endif

    if ( startUnit + count > numberOfUnits() ) {
        count = numberOfUnits() - startUnit;
        if ( count > 0 ) {
            unsigned countBytes = count << log2UnitSize();
            if ( countBytes < nbytes ) 
                nbytes = countBytes;
        }
    }

    int result = 0;
    if ( count > 0 && nbytes > 0 ) {
        if ( _cache == 0 ) {
            result = 
            rawRead( startUnit, count, addr, nbytes, c );
        } else {
            result = 
            cacheRead ( startUnit, count, addr, nbytes, c );
        }
    } else {
        if ( c ) 
            c->activate( 0 );
    }
    
    Raids( RaidProtected ) 
    << this << "::basicRead( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes << " ) returns " << result << "\n" << eor;

    return result;
}

int
MemoryObject::cacheRead( unsigned startUnit, unsigned count, char * addr, 
                         unsigned nbytes, IOAsyncClosure * c )
{
    Raids( RaidProtected ) 
    << this << "::cacheRead( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes 
    << ", " << asAddress( c ) << " )\n" << eor;

    unsigned unitsRead;
    if ( c == 0 ) {
        unitsRead = _cache->cacheRead
        ( startUnit << log2UnitSize(), nbytes, addr, 0 );
        unitsRead = ( unitsRead + ( unitSize() - 1 ) )  
        >> log2UnitSize();
#ifdef JEFF
        Console() << this << "::" << __FUNCTION__ << "( unitsRead "
                  << unitsRead << ", count " << count << " )\n" << eor;
#endif /* JEFF */
        Assert( unitsRead >= count );
    } else {
        unitsRead = count;
        TranslateBytesToUnitsAsyncClosure * myClosure
        = new TranslateBytesToUnitsAsyncClosure
        ( c, log2UnitSize(), count );
        _cache->cacheRead( startUnit << log2UnitSize(), nbytes, 
                           addr, myClosure );
    }
    return unitsRead;
}

int
MemoryObject::write( unsigned startUnit, int count, char * addr, int nbytes )
{
    Raids( RaidMethod ) 
    << this << "::write( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes << " )\n" << eor;

    if ( count <= 0 || nbytes <= 0 )
        return 0;

    unsigned lastUnit = startUnit + count - 1;

    acquire( startUnit, lastUnit );

    if ( startUnit + count > numberOfUnits() && ! fixedSize() ) {
        release( startUnit, lastUnit );
        acquireAll();
        unsigned nounits = basicSetNumberOfUnits( startUnit + count );
        downgradeFromAll( startUnit, lastUnit );
        Assert( nounits == numberOfUnits() );
    }
    int result = basicWrite( startUnit, count, addr, nbytes, 0 );

    release( startUnit, lastUnit );

    Raids( RaidMethod ) 
    << this << "::write( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes << " ) returns " << result << "\n" << eor;

    return( result );
}

void
MemoryObject::asyncWrite( unsigned startUnit, int count, char * addr, 
                          int nbytes, IOAsyncClosure * c )
{
    Raids( RaidMethod ) 
    << this << "::asyncWrite( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes 
    << ", " << asAddress( c ) << " )\n" << eor;

    if ( count <= 0 || nbytes <= 0 ) {
        c->activate( 0 );
        return;
    }

    unsigned lastUnit = startUnit + count - 1;

    acquire( startUnit, lastUnit );

    if ( startUnit + count > numberOfUnits() && ! fixedSize() ) {
        release( startUnit, lastUnit );
        acquireAll();
        unsigned nounits = basicSetNumberOfUnits( startUnit + count );
        downgradeFromAll( startUnit, lastUnit );
        Assert( nounits == numberOfUnits() );
    }

    ReleaseIOAsyncClosure * myClosure =
    new ReleaseIOAsyncClosure( c, this, startUnit, lastUnit );
 
    int result = basicWrite( startUnit, count, addr, nbytes, myClosure );

    Raids( RaidMethod ) 
    << this << "::asyncWrite( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes << " ) returns " << result << "\n" << eor;

    return;
}

int
MemoryObject::basicWrite( unsigned startUnit, unsigned count, char * addr, 
                          unsigned nbytes, IOAsyncClosure * c )
{
    Raids( RaidProtected ) 
    << this << "::basicWrite( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes 
    << ", " << asAddress( c ) << " )\n" << eor;

    if ( startUnit + count > numberOfUnits() ) {
        count = numberOfUnits() - startUnit;
        if ( count > 0 ) {
            unsigned countBytes = count << log2UnitSize();
            if ( countBytes < nbytes ) 
                nbytes = countBytes;
        }
    }

    int result = 0;
    if ( count > 0 && nbytes > 0 ) {
        if ( _cache == 0 ) {
            result = rawWrite( startUnit, count, addr, nbytes, 
                               0, c );
        } else {
            result = cacheWrite( startUnit, count, addr, nbytes,
                                 c );
        }
    } else {
        if ( c )
            c->activate( 0 );
    }

    Raids( RaidProtected ) 
    << this << "::basicWrite( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes << " ) returns " << result << "\n" << eor;

    return( result );
}

int 
MemoryObject::cacheWrite( unsigned startUnit, unsigned count, char * addr, 
                          unsigned nbytes, IOAsyncClosure * c )
{
    Raids( RaidProtected ) 
    << this << "::cacheWrite( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes 
    << ", " << asAddress( c ) << " )\n" << eor;
    
    unsigned unitsWritten;
    if ( c == 0 ) {
        unitsWritten = _cache->cacheWrite
        ( startUnit << log2UnitSize(), nbytes, addr, 0 );
        unitsWritten = ( unitsWritten + ( unitSize() - 1 ) )  
        >> log2UnitSize();
        Assert( unitsWritten >= count );
    } else {
        unitsWritten = count;
        TranslateBytesToUnitsAsyncClosure * myClosure
        = new TranslateBytesToUnitsAsyncClosure
        ( c, log2UnitSize(), count );
        _cache->cacheWrite( startUnit << log2UnitSize(), nbytes, 
                            addr, myClosure );
    }

    return unitsWritten;    
}

int
MemoryObject::writeThrough( unsigned startUnit, int count, char * addr, 
                            int nbytes )
{
    Raids( RaidMethod ) 
    << this << "::writeThrough( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes << " )\n" << eor;

    if ( count <= 0 || nbytes <= 0 )
        return 0;

    unsigned lastUnit = startUnit + count - 1;

    acquire( startUnit, lastUnit );

    if ( startUnit + count > numberOfUnits() && ! fixedSize() ) {
        release( startUnit, lastUnit );
        acquireAll();
        unsigned  nounits = basicSetNumberOfUnits( startUnit + count );
        downgradeFromAll( startUnit, lastUnit );
        Assert( nounits == numberOfUnits() );
    }
    int result = basicWriteThrough( startUnit, count, addr, nbytes, 0 );

    release( startUnit, lastUnit );

    Raids( RaidMethod ) 
    << this << "::writeThrough( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes << " ) returns " << result << "\n" << eor;

    return( result );
}

void
MemoryObject::asyncWriteThrough( unsigned startUnit, int count, char * addr, 
                                 int nbytes, IOAsyncClosure * c )
{
    Raids( RaidMethod ) 
    << this << "::asyncWriteThrough( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes 
    << ", " << asAddress( c ) << " )\n" << eor;

    if ( count <= 0 || nbytes <= 0 ) {
        c->activate( 0 );
        return;
    }

    unsigned lastUnit = startUnit + count - 1;

    acquire( startUnit, lastUnit );

    if ( startUnit + count > numberOfUnits() && ! fixedSize() ) {
        release( startUnit, lastUnit );
        acquireAll();
        unsigned  nounits = basicSetNumberOfUnits( startUnit + count );
        downgradeFromAll( startUnit, lastUnit );
        Assert( nounits == numberOfUnits() );
    }

    ReleaseIOAsyncClosure * myClosure = 
    new ReleaseIOAsyncClosure( c, this, startUnit, lastUnit );

    int result = basicWriteThrough( startUnit, count, addr, nbytes, 
                                    myClosure );

    Raids( RaidMethod ) 
    << this << "::asyncWriteThrough( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes << " ) returns " << result << "\n" << eor;

    return;
}


int
MemoryObject::basicWriteThrough( unsigned startUnit, unsigned  count, 
                                 char * addr, unsigned  nbytes, 
                                 IOAsyncClosure * c )
{
    Raids( RaidProtected ) 
    << this << "::basicWriteThrough( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes 
    << ", " << asAddress( c ) << " )\n" << eor;

    if ( startUnit + count > numberOfUnits() ) {
        count = numberOfUnits() - startUnit;
        if ( count > 0 ) {
            unsigned countBytes = count << log2UnitSize();
            if ( countBytes < nbytes ) 
                nbytes = countBytes;
        }
    }

    int result = 0;
    if ( count > 0 && nbytes > 0 ) {
        if ( _cache == 0 ) {
            result = rawWrite( startUnit, count, addr, nbytes, 
                               1, c );
        } else {
            if ( c != 0 ) {
                SynchronizeIOAsyncClosure * myClosure =
                new SynchronizeIOAsyncClosure
                ( c, this, 
                  startUnit, 
                  startUnit + count - 1 );
                result = cacheWrite( startUnit, count, addr, 
                                     nbytes, myClosure );
            } else {
                result = cacheWrite( startUnit, count, addr, 
                                     nbytes, 0 );
                basicSynchronize( startUnit,
                                  startUnit + count - 1, 0 );
            }
        }
    } else {
        if ( c )
            c->activate( 0 );
    }

    Raids( RaidProtected ) 
    << this << "::basicWriteThrough( " << startUnit 
    << ", " << count 
    << ", " << asAddress( addr ) 
    << ", " << nbytes << " ) returns " << result << "\n" << eor;

    return( result );
}

void
MemoryObject::synchronize()
{
    Raids( RaidMethod ) << this << "::synchronize()\n" << eor;
    acquireAll();
    basicSynchronize( 0 );
    releaseAll();
    Raids( RaidMethod ) << this << "::synchronize() done\n" << eor;
}

void
MemoryObject::basicSynchronize( IOAsyncClosure * c )
{
    if ( _cache != 0 ) {
        _cache->synchronize( c );
    } else {
        if ( c != 0 ) 
            c->activate();
    }
}

void
MemoryObject::basicSynchronize( unsigned from, unsigned to, IOAsyncClosure * c )
{
    Assert( _cache != 0 );
    from >>= _cacheLog2UnitSizeRatio;
    to >>= _cacheLog2UnitSizeRatio;
    _cache->synchronize( from, to, c );
}

void
MemoryObject::flush()
{
    Raids( RaidMethod ) << this << "::flush()\n" << eor;
    acquireAll();
    basicFlush();
    releaseAll();
    Raids( RaidMethod ) << this << "::flush() done\n" << eor;
}

void
MemoryObject::basicFlush()
{
    if ( _cache != 0 ) {
        _cache->flush();
    }
}

int
MemoryObject::setNumberOfUnits( unsigned units )
{
    if ( units == numberOfUnits() ) {
        return units;
    }
    if ( fixedSize() ) {
        return units;
    }
    acquireAll();
    units = basicSetNumberOfUnits( units );
    releaseAll();
    return units;
}

unsigned int
MemoryObject::basicSetNumberOfUnits( unsigned units )
{
    static unsigned large = 0;

    if ( _cache != 0 ) {
        unsigned cacheUnits = 
        ( units + ( 1 << _cacheLog2UnitSizeRatio ) - 1 ) >>
        _cacheLog2UnitSizeRatio;
        _cache->setNumberOfUnits( cacheUnits );
    }
    if ( large == 0 ) {
        large = TheKernel->physicalMemorySize() >> 2; 
    }
    unsigned largeUnits = large >> _log2UnitSize;
    unsigned largeHint = _hints & H_LARGE;
    unsigned largeMO = units >= largeUnits;
    if ( !largeHint && largeMO ) {
        setHints( H_LARGE );
    } else if ( largeHint && !largeMO ) {
        clearHints( H_LARGE );
    }
    return( _numberOfUnits = units );
}

int
MemoryObject::copy( MemoryObject * dest )
{
    Raids( RaidMethod ) 
    << this << "::copy( " << dest << " )\n" << eor;

    acquireSize();
    int result = basicCopy( dest );
    releaseSize();
    return result;
}

int
MemoryObject::basicCopy( MemoryObject * dest )
{
    Raids( RaidProtected ) 
    << this << "::basicCopy( " << dest << " )\n" << eor;

    dest->setNumberOfUnits( 0 );

    int log2UnitDifference = _log2UnitSize - dest->log2UnitSize();
    int readUnits = 1;
    int writeUnits = 1;
    int bufferSize = unitSize();

    if( log2UnitDifference > 0 ) {
        writeUnits <<= log2UnitDifference;
        dest->setNumberOfUnits( _numberOfUnits << log2UnitDifference );
    }
    else if( log2UnitDifference < 0 ) {
        log2UnitDifference = -log2UnitDifference;
        readUnits <<= log2UnitDifference;
        bufferSize <<= log2UnitDifference;
        int unitDifference = 1 << log2UnitDifference;
        dest->setNumberOfUnits
        ( ( _numberOfUnits + unitDifference - 1 ) 
          >> log2UnitDifference );
    } else {
        dest->setNumberOfUnits( _numberOfUnits );
    }

    Raids( RaidDetail )
    << this << "::baiscCopy()"
    << " ru " << readUnits << " wu " << writeUnits 
    << " bs " << bufferSize << "\n" << eor;

    char * buffer = new char[bufferSize];

    int returnValue = 0;
    int w = 0;
    for( unsigned r = 0; r < _numberOfUnits; r += readUnits, w += writeUnits ) {
        int status = read( r, readUnits, buffer, bufferSize );
        if( ( status < readUnits ) && ( r+status < _numberOfUnits ) ) {
            returnValue = ReadError;
            break;
        }
        if( dest->write( w, writeUnits, buffer, bufferSize) !=
            writeUnits ) {
            returnValue = WriteError;
            break;
        }
    }
    delete buffer;

    Raids( RaidProtected ) 
    << "::basicCopy( " << dest << " ) returns " << returnValue
    << "\n" << eor;

    return( returnValue );
}

ProtectionLevel
MemoryObject::logicalProtectionLevel()
{
    return( (ProtectionLevel) _protection );
}

RangeLock * 
MemoryObject::moLock()
{
    Assert( _moLock != 0 );
    return _moLock;
}

void
MemoryObject::setMemoryObjectCache( MemoryObjectCache * cache )
{
    _lock.acquire();
    Assert( _cache == 0 );
    if ( cache != 0 ) {
        Assert( _cacheLog2UnitSize == cache->log2UnitSize() );
        _cacheClass = cache->classOf();
        _cache = cache;
    }
    _lock.release();
}

MemoryObjectCache *
MemoryObject::cache()
{
    return _cache;
}

unsigned 
MemoryObject::cacheLog2UnitSize()
{
    return _cacheLog2UnitSize;
}

unsigned
MemoryObject::setCacheLog2UnitSize( unsigned log2Size )
{
    _lock.acquire();
    if ( _cache == 0 ) {
        if ( log2Size == 0 )
            log2Size = _log2UnitSize > Log2PAUUnitSize ?
            _log2UnitSize : Log2PAUUnitSize;
        Assert( log2Size >= Log2PAUUnitSize );
        _cacheLog2UnitSize = log2Size;
        _cacheLog2UnitSizeRatio = _cacheLog2UnitSize - _log2UnitSize;
        Assert( _cacheLog2UnitSize - _log2UnitSize >=  0 );
    }
    _lock.release();
    return _cacheLog2UnitSize;
}

extern ClassStar PagedMemoryObjectCacheClass;

Class *
MemoryObject::cacheClass()
{
    _lock.acquire();
    if ( _cacheClass == 0 ) {
        _cacheClass = PagedMemoryObjectCacheClass;
    }
    _lock.release();
    return _cacheClass;
}

void
MemoryObject::setCacheClass( Class * cacheClass )
{
    _lock.acquire();
    if ( _cache == 0 ) {
        Assert( cacheClass->isASubclassOf( MemoryObjectCacheClass ) );
        _cacheClass = cacheClass;
    }
    _lock.release();
}

Class *
MemoryObject::replacementPolicyClass()
{
    if ( _cache == 0 ) {
        return _replacementPolicyClass;
    } else {
        return _cache->replacementPolicy()->classOf();
    }
}

extern ClassStar ReplacementPolicyClass;

void
MemoryObject::setReplacementPolicyClass( Class * policyClass )
{
    Assert( policyClass->isKindOf( ReplacementPolicyClass ) );
    if ( _cache == 0 ) {
        _replacementPolicyClass = policyClass;
    } else {
        acquireAll();
        _cache->setReplacementPolicyClass( policyClass );
        releaseAll();
    }
}

unsigned
MemoryObject::maxResidentCacheUnits()
{
    if ( _cache  == 0) {
        return 0;
    } else {
        ReplacementPolicy * rp = _cache->replacementPolicy();
        Assert( rp );
        return rp->maxResidentUnits();
    }
}

unsigned
MemoryObject::setMaxResidentCacheUnits( unsigned units )
{
    if ( _cache != 0) {
        return 0;
    } else {
        ReplacementPolicy * rp = _cache->replacementPolicy();
        Assert( rp );
        return rp->setMaxResidentUnits( units );
    }
}

#ifdef UNIX

MemoryObjectCache *
MemoryObject::buildCache()
{
    Assert( NOTREACHED );
    return( 0 );
}

#else

MemoryObjectCache *
MemoryObject::buildCache()
{
    Raids( RaidMethod ) 
    << this << "::buildCache()\n" << eor;
    Assert( this != 0 );

    Class * cacheCl = cacheClass();
    Assert( cacheCl != 0 );


    if( _cache != 0 ) {
        Raids( RaidDetail ) 
        << this << "::buildCache: " << _cache 
        << " already exists\n" << eor;
        return( _cache );
    }

    Raids( RaidDetail ) 
    << this << "::buildCache() cache class " 
    << cacheCl << "\n" << eor;

    ProxiableObject * po = cacheCl->constructor( this );
    Assert( po != 0 );
    Assert( po->isKindOf( MemoryObjectCacheClass ) );
    _lock.acquire();
    _cache = ( MemoryObjectCache * ) po;

    _lock.release();
    return( _cache );
}

#endif /*UNIX*/

extern ClassStar CacheClass;

void MemoryObject::addCache( Class * cacheClass )
{
    Raids( RaidMethod ) 
    << this << "::addCache( " << cacheClass << " )\n" << eor;
        
    Assert( cacheClass->isASubclassOf( CacheClass ) );
    Assert( _cache != 0 );

    acquireAll();
    _cache->addCache( cacheClass );
    releaseAll();
}

void MemoryObject::setHints( unsigned mask )
{
    _hintLock.acquire();
    _hints |= mask;
    _hintLock.release();
}

void MemoryObject::clearHints( unsigned mask )
{
    _hintLock.acquire();
    _hints &= ~mask;
    _hintLock.release();
}

#include "ClassInstaller.h"
DEFINECLASS("MemoryObject",MemoryObject,MemoryRange,0,0);
