#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Author: Aamod Sane (sane@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * MemoryObjectCache.c: Abstract class representing a physical memory cache
 *          of a memory object.
 */

#include "AddressTranslation.h"
#include "Assert.h"
#include "Async.h"
#include "Barrier.h"
#include "BitRanges.h"
#include "Byte.h"
#include "Cache.h"
#include "CacheAsync.h"
#include "CheckSum.h"
#include "ClassStar.h"
#include "IOWindow.h"
#include "Kernel.h"
#include "LockHolder.h"
#include "Map.h"
#include "MemoryObject.h"
#include "MemoryObjectAsync.h"
#include "MemoryObjectCache.h"
#include "MemoryObjectView.h"
#include "MemorySimpleInstruments.h"
#include "OutputStream.h"
#include "PhysicalMemoryChain.h"
#include "PhysicallyAddressableUnit.h"
#include "Process.h"
#include "PROCESSOR.h"
#include "Raid.h"
#include "RangeLocks.h"
#include "ReplacementPolicy.h"
#include "Store.h"
#include "ThisCPU.h"
#include "ThisProcess.h"            /* Used in unitsFomStore */
#include "TranslationMap.h"
#include "Use.h"

#ifdef __GNUC__
#pragma implementation
#endif

/* #define JEFF */

// #define COUNTERLOCK

#define USESTACK
// #define DEBUG_GLRU

extern Store * MainStore;

#ifdef CHANGESTATUS
static AddressTranslationStatus clearMod( AddressTranslationStatus::Modified );
static AddressTranslationStatus setNone( 0 );
#endif // CHANGESTATUS

int MemoryObjectCacheDefaultLog2WindowSize = 19; /* 512K */
AddressTranslation * MemoryObjectCache::_translation = 0;


/*
 * Constructor and Destructor
 */

MemoryObjectCacheTimer 
MemoryObjectCacheCtor("MemoryObjectCache::MemoryObjectCache() timer");

MemoryObjectCache::MemoryObjectCache( MemoryObject * memObj )
    : MemoryRange( 0, memObj->cacheLog2UnitSize() ), 
      _storeInfo( this )
{
    SInstr( int istart = MemoryObjectCacheCtor.start() );
    Assert( this != 0 );

    initDebug();
    Raids( RaidConstructor ) << this 
                             << "::MemoryObjectCache( " << memObj << " )\n" << eor;

    /* initializations */
    _bytesAccessed = 0;
    _fromCacheUnits = 0;
    
    /* relationship between mo and cache unit size */

    _memoryObject = memObj;
    _moLock = _memoryObject->moLock();
    Assert( _moLock != 0 );
    Assert( log2UnitSize() >= Log2PAUUnitSize );
    _log2MoUnitSizeRatio = log2UnitSize() - _memoryObject->log2UnitSize();
    _moUnitSizeRatioMask = ( 1 << _log2MoUnitSizeRatio ) - 1;
    Assert( _log2MoUnitSizeRatio >= 0 );

    /* now I know how big I should be */

    int moUnitSizeRatio = 1 << _log2MoUnitSizeRatio;
    _numberOfUnits = ( memObj->numberOfUnits() + moUnitSizeRatio - 1 )
    >> _log2MoUnitSizeRatio;

    Debug<<"memObj->numberOfUnits() = "<<memObj->numberOfUnits()<<"\n";
    Debug<<"moUnitSizeRatio = "<<moUnitSizeRatio<<"\n";
    Debug<<"_log2MoUnitSizeRatio = "<<_log2MoUnitSizeRatio<<"\n";
    Debug<<"_noOfUnits "<<_numberOfUnits<<"\n"<<eor;


    /* data structure used unknown yet */

    _map = 0;
    _cache = 0;
    _adtMap = 0;
    _replacementPolicy = 0;

    /* window used when not mapped */

    if ( _translation == 0 ) {
        _translation = TheKernel->domain()->translation();
        Assert( _translation != 0 );
    }

    /* support for larger unit size */

    _log2PauUnitSizeRatio = log2UnitSize() - Log2PAUUnitSize;
    Assert( _log2PauUnitSizeRatio >= 0 );

    /* window unit size info */

    _log2WindowSize = computeLog2WindowSize();
    int log2WindowSizeRatio = _log2WindowSize - log2UnitSize();
    _windowSizeRatio = 1 << log2WindowSizeRatio;
    _windowSizeRatioMask = _windowSizeRatio - 1;
    _noOfWindows = 0;
    _ioWindow = 0;
    _cacheWindow = 0;

    SInstr( MemoryObjectCacheCtor.stop(istart) );
}

const unsigned MaxLog2WindowSize = 24;

int
MemoryObjectCache::computeLog2WindowSize()
{
    unsigned log2WindowSize;
    unsigned bytes = _numberOfUnits << log2UnitSize();

    if ( bytes < ( 1 << 20 ) )          // <  1 Mb
        log2WindowSize = log2UnitSize() + 4; 
    else if ( bytes < ( 1 << 22 ) )     // >= 1 Mb
        log2WindowSize = 18;        //      - 256k
    else if ( bytes < ( 1 << 24 ) )     // >= 4  Mb   
        log2WindowSize = 20;        //      - 1 Mb
    else if ( bytes < ( 1 << 26 ) )     // >= 16 Mb   
        log2WindowSize = 22;        //      - 4 Mb
    else                                // >= 64 Mb   
        log2WindowSize = 24;        //      - 16 Mb
    if ( log2WindowSize > MaxLog2WindowSize )
        return MaxLog2WindowSize;
    else
        return log2WindowSize;
}

MemoryObjectCache::~MemoryObjectCache()
{
    Raids( RaidConstructor ) << this << "::~MemoryObjectCache()\n" << eor;
    Assert( _memoryObject != 0 );
    Assert( _storeInfo.queued() == 0 );
}

void 
MemoryObjectCache::setDebugMask( unsigned short mask )
{
    _debugMask = mask;
    if ( _map ) _map->setDebugMask( mask );
    if ( _cache ) _cache->setDebugMask( mask );
    if ( _adtMap ) _adtMap->setDebugMask( mask );
}

void 
MemoryObjectCache::setDebugMessages( short messages )
{
    _debugMessages = messages;
    if ( _map ) _map->setDebugMessages( messages );
    if ( _cache ) _cache->setDebugMessages( messages );
    if ( _adtMap ) _adtMap->setDebugMessages( messages );
}

#ifdef COUNTERLOCK
static LOCK CounterLock;
#endif

inline 
void 
MemoryObjectCache::bytesAccessedAdd( unsigned bytes )
{
#ifdef COUNTERLOCK
    CounterLock.acquire();
    _bytesAccessed += bytes;
    CounterLock.release();
#else
    _bytesAccessed += bytes;
#endif
}

void
MemoryObjectCache::bytesAccessedReset()
{
#ifdef COUNTERLOCK
    CounterLock.acquire();
    _bytesAccessed = 0;
    CounterLock.release();
#else
    _bytesAccessed = 0;
#endif
}

inline 
void 
MemoryObjectCache::fromCacheUnitsAdd( unsigned bytes )
{
#ifdef COUNTERLOCK
    CounterLock.acquire();
    _fromCacheUnits += bytes;
    CounterLock.release();
#else
    _fromCacheUnits += bytes;
#endif
}

void
MemoryObjectCache::fromCacheUnitsReset()
{
#ifdef COUNTERLOCK
    CounterLock.acquire();
    _fromCacheUnits = 0;
    CounterLock.release();
#else
    _fromCacheUnits = 0;
#endif
}

PhysicallyAddressableUnit * 
MemoryObjectCache::forward( PhysicallyAddressableUnit * pau, 
                            unsigned noOfUnits )
{
    unsigned count = noOfUnits << _log2PauUnitSizeRatio;
    while ( count ) {
        pau = pau->nextLink();
        count--;
    }
    return pau;
}

ProtectionLevel 
MemoryObjectCache::protection( PhysicallyAddressableUnit * pau,
                               unsigned noOfUnits )
{
    ProtectionLevel prot = ( ProtectionLevel ) ReadWrite;
    unsigned count = noOfUnits << _log2PauUnitSizeRatio;
    while ( count ) {
        prot = Min( prot, pau->protection() );
        pau = pau->nextLink();
        count--;
    }
    return prot;
}

void
MemoryObjectCache::setProtection( PhysicallyAddressableUnit * pau,
                                  unsigned noOfUnits, ProtectionLevel prot )
{
    unsigned count = noOfUnits << _log2PauUnitSizeRatio;
    while ( count ) {
        pau->setProtection( prot );
        pau = pau->nextLink();
        count--;
    }
}

void
MemoryObjectCache::setReferenced( PhysicallyAddressableUnit * pau,
                                  unsigned noOfUnits,
                                  int referenced )
{
    unsigned count = noOfUnits << _log2PauUnitSizeRatio;
    while ( count ) {
        pau->setReferenced( referenced );
        pau = pau->nextLink();
        count--;
    }
}

void
MemoryObjectCache::setModified( PhysicallyAddressableUnit * pau,
                                unsigned noOfUnits,
                                int modified )
{
    unsigned count = noOfUnits << _log2PauUnitSizeRatio;
    while ( count ) {
        pau->setModified( modified );
        pau = pau->nextLink();
        count--;
    }
}

void 
MemoryObjectCache::setReferencedModified( PhysicallyAddressableUnit * pau,
                                          unsigned noOfUnits, 
                                          int referenced, int modified )
{
    unsigned count = noOfUnits << _log2PauUnitSizeRatio;
    while ( count ) {
        pau->setReferenced( referenced );
        pau->setModified( modified );
        pau = pau->nextLink();
        count--;
    }
}

void 
MemoryObjectCache::setCacheUnitInfo( PhysicallyAddressableUnit * pau, 
                                     unsigned noOfUnits, CacheUnitInfo * cu )
{
    int pauUnitSizeRatio = 1 << _log2PauUnitSizeRatio;
    for ( unsigned unit = 0; unit < noOfUnits; unit++, cu++ ) {
        int modified = 0;
        for (int i = 0; i < pauUnitSizeRatio; i++ ) {
            Assert( pau );
            if ( pau->modified() ) 
                modified = 1;
            pau = pau->nextLink();
        }   
        cu->is( modified ? CUIModified : CUIValid );
    }
}

MemoryObjectCacheTimer 
MemoryObjectCacheCleanUp("MemoryObjectCache::cleanUp() timer");

void
MemoryObjectCache::cleanUp()
{
    SInstr( int istart = MemoryObjectCacheCleanUp.start() );

    Raids( RaidProtected ) << this << "::cleanUp()\n" << eor;

    TheIOWindowCache->freeIOWindow( this, 1 );
    Assert( _memoryObject != 0 );
    unsigned residentSize = _map->residentSize();
    Assert( residentSize >= 0 && residentSize <= _numberOfUnits );
    if( residentSize > 0 && _numberOfUnits > 0 ) {
        unitsToStore( 0, _numberOfUnits-1, 0 );
    }

    SInstr( MemoryObjectCacheCleanUp.stop( istart ) );
}

/*
 * Support to lock memory object
 */

MemoryObjectCacheTimer
MemoryObjectCachePreFault("MemoryObjectCache::preFault() timer");

void
MemoryObjectCache::preFault( unsigned from, unsigned to )
{
    SInstr( int istart = MemoryObjectCachePreFault.start() );

    Raids( RaidProtected ) 
    << this << "::preFault( " 
    << from << ", " << to << " ) " 
    << thisProcess() << "\n" << eor;        

    acquireMo( from, to );

    Raids( RaidProtected ) 
    << this << "::preFault( " 
    << from << ", " << to << " ) " 
    << thisProcess() << " done\n" << eor;       

    SInstr( MemoryObjectCachePreFault.stop(istart) );
}

MemoryObjectCacheTimer
MemoryObjectCachePostFault("MemoryObjectCache::postFault() timer");

void
MemoryObjectCache::postFault( unsigned from, unsigned to )
{
    SInstr( int istart = MemoryObjectCachePostFault.start() );

    Raids( RaidProtected ) 
    << this << "::postFault( " 
    << from << ", " << to << " ) " 
    << thisProcess() << "\n" << eor;

    releaseMo( from, to );

    Raids( RaidProtected )
    << this << "::postFault( " << from << ", " << to 
    << " ) done\n" << eor;

    SInstr( MemoryObjectCachePostFault.stop(istart) );
}

/*
 * Moving units to and from store
 */

MemoryObjectCacheTimer
MemoryObjectCacheUnitsFromStore("MemoryObjectCache::unitsFromStore() timer");

PhysicallyAddressableUnit *
MemoryObjectCache::unitsFromStore( unsigned from, unsigned to )
{
    SInstr( int istart = MemoryObjectCacheUnitsFromStore.start() );

    Raids( RaidProtected ) 
    << this << "::unitsFromStore( "
    << from << ", " << to << " )\n" << eor;

    Assert( from <= to && to < _numberOfUnits );

    if ( _replacementPolicy )
        _replacementPolicy->unitsFromStore( from, to );

    PhysicallyAddressableUnit * pau;
    for ( ; ; ) {
        pau = storeAlloc( to - from + 1 );
        if ( pau == 0 ) {
            Assert( NOTREACHED );
            Raids( RaidDetail ) 
            << this << "::unitsFromStore("
            << from << " ," << to 
            << " ) alloc failed on cpu "
            << thisCPU() << "\n" << eor;
            thisProcess()->relinquishProcessor();
            Raids( RaidDetail ) << this << "::unitsFromStore cpu " 
                                << thisCPU() << "\n" << eor;
            while( MainStore->isPaging() ) {
                Raids( RaidDetail ) << this 
                                    << "::unitsFromStore cpu " 
                                    << thisCPU() << "\n" << eor;
                thisProcess()->relinquishProcessor();
            }
        } else {
            break;
        }
    }
    _map->addPauList( pau, from, to );
    SInstr( MemoryObjectCacheUnitsFromStore.stop(istart) );
    return pau;
}

MemoryObjectCacheTimer
MemoryObjectCacheUnitsToStore("MemoryObjectCache::unitsToStore() timer");

int
MemoryObjectCache::unitsToStore( unsigned from, unsigned to, int code )
{
    SInstr( int istart = MemoryObjectCacheUnitsToStore.start() );

    Raids( RaidProtected ) 
    << this << "::unitsToStore( "
    << from << ", " << to << ", " << asHex( code ) 
    << " )\n" << eor;

    Assert( from <= to && to < _numberOfUnits );
    unsigned first;
    unsigned start;
    unsigned end;
    unsigned result = 0;
    unsigned queued = ( UNITSTOSTORE_QUEUED & code );
    unsigned notSkipReplacementPolicy =
    !( UNITSTOSTORE_SKIPRP & code );

    first = from;
    while ( first <= to ) {
        PhysicallyAddressableUnit * p =
        _map->findAndRemove( first, to, start, end );
        if ( p == 0 )
            break;
        storeDealloc( p, queued );
        if ( notSkipReplacementPolicy && _replacementPolicy ) {
            _replacementPolicy->unitsToStore( start, end );
        }
        first = end + 1;
        result = 1;
    }

    SInstr( MemoryObjectCacheUnitsToStore.stop(istart) );
    return result;
}

unsigned
MemoryObjectCache::load( unsigned from, unsigned to, IOAsyncClosure * c )
{
    // SInstr( int istart = MemoryObjectCacheLoad.start() );

    Raids( RaidProtected ) 
    << this << "::load( " 
    << from << ", " << to << ", " << asAddress( c )
    << " )\n" << eor;

    Assert( from <= to && to < _numberOfUnits );
    unsigned first = from;
    unsigned start;
    unsigned end;
    unsigned noOfUnits;

    if ( from == to ) {

        // optimize for single unit case

        PhysicallyAddressableUnit * pau;
        if ( _map->findNextNonExist( first, to, start, end, pau ) ) {
            PhysicallyAddressableUnit * fromStorePauList =
            unitsFromStore( start, end );
            int result =
            ioOperation( FromCache, start, end, 0, 
                         fromStorePauList, c );
            if ( _replacementPolicy ) {
                /* cheat here for async case, not really 
                   paged in yet */
#ifdef DEBUG_GLRU
                RaidStream() 
                << this << "::load rp " 
                << asAddress( _replacementPolicy )
                << " single unit cheat\n" << eor;
#endif // DEBUG_GLRU
                _replacementPolicy->postPageIn( start, end );
            }
            if ( c == 0 ) {
                noOfUnits = result;
                Assert( noOfUnits == 1 );
            } else {
                noOfUnits = 1;
            }
        } else {
            noOfUnits = 0;
            if ( c != 0 ) {
                c->activate( 0 );
            }
        }

    } else {

        CacheIOState * state = 0;
        unsigned expected = 0;
        while ( first <= to ) {
            PhysicallyAddressableUnit * pau;
            if ( !_map->findNextNonExist( first, to, start, end, 
                                          pau ) )
                break;
            PhysicallyAddressableUnit * fromStorePauList =
            unitsFromStore( start, end );
            if ( state == 0 ) {
                state = new CacheIOState( c );
                Raids( RaidDetail ) 
                << this << "::load() state "
                << asAddress( state ) << "\n" << eor;
                state->_spare = 14; // ??
            }
            state->add( start, end, (char *) fromStorePauList );
            expected += end - start + 1;
            first = end + 1;
        }
    
        if ( state ) {
            state->_result = 0;
            state->_expectedResult = expected;
            CacheIOState::_Operation * op = state->_ops;
            int opCount = state->_count;
            for( int i = 0; i < opCount; i++ ) {
                // need to save value since
                // state and op may be destroyed
                // after call
                unsigned from = op->_startUnit;
                unsigned to = op->_count;
                ioOperation( FromCache,
                             from, to, 
                             0 ,
                             (PhysicallyAddressableUnit *) 
                             op->_buf,
                             &(op->_closure) );
                if ( _replacementPolicy ) {
                    /* cheat here for async case, 
                       not really paged in yet */
#ifdef DEBUG_GLRU
                    RaidStream() 
                    << this << "::load rp " 
                    << asAddress( _replacementPolicy )
                    << " multi unit cheat\n" << eor;
#endif // DEBUG_GLRU
                    _replacementPolicy->postPageIn
                    ( from, to );
                }
                op++;
            }
            if ( c == 0 ) {
                state->_barrier.wait();
                noOfUnits = state->_result;
                delete state;
            } else {
                noOfUnits = expected;
            } 
        } else {
            noOfUnits = 0;
            if ( c != 0 ) {
                c->activate( noOfUnits );
            }
        }
    }

    // SInstr( MemoryObjectCacheLoad.stop(istart) );
    
    return noOfUnits;
}

class SaveModifiedSynchronizeClosure : public IOAsyncClosure {
protected:
    IOAsyncClosure * _parent;
    IOOperationCode _operation;
    Cache * _cache;
    unsigned _startUnit;
    unsigned _count;
    unsigned _log2WindowSize;
    unsigned _state;
    static void func( SaveModifiedSynchronizeClosure * c );
public:
    Barrier _barrier;
    SaveModifiedSynchronizeClosure() : _barrier( 1 ) {}
    void init( IOAsyncClosure * parent, Cache * cache, 
               IOOperationCode operation,
               unsigned startUnit, unsigned count, 
               unsigned log2WindowSize );
};

void SaveModifiedSynchronizeClosure::init(
    IOAsyncClosure * parent, Cache * cache, 
    IOOperationCode operation,
    unsigned startUnit, unsigned count, 
    unsigned log2WindowSize )
{
    _parent = parent; _cache = cache; _operation = operation;
    _startUnit = startUnit; _count = count; 
    _log2WindowSize = log2WindowSize; _func = (AsyncFunc)&func; 
    _state = 0;
}

void SaveModifiedSynchronizeClosure::func( SaveModifiedSynchronizeClosure * c )
{
    Assert( c->_operation == ToMemoryObject );
    if ( c->_state == 0 ) {
        c->_state = 1;
        if ( c->_cache->dirty() ) {
            char * buffer = 0;
            c->_cache->synchronize
            ( c->_startUnit, c->_count, 
              c->_log2WindowSize, buffer, c );
            return;
        }
    }
    if ( c->_state == 1 ) {
        c->_state = 2;
        if ( c->_parent != 0 ) {
            c->_parent->activate();
            delete c;
        } else {
            c->_barrier.done();
        }
        return;
    }
    Assert( NOTREACHED );
}

MemoryObjectCacheTimer
MemoryObjectCacheSaveModified("MemoryObjectCache::saveModified() timer");

void
MemoryObjectCache::saveModified( IOOperationCode operation,
                                 unsigned from, unsigned to, int code,
                                 IOAsyncClosure * c )
{
    SInstr( int istart = MemoryObjectCacheSaveModified.start() );

    Raids( RaidProtected ) 
    << this << "::saveModified( " << operation << ", "
    << from << ", " << to << ", " << asHex( code ) << " )\n" << eor;

    Assert( from <= to && to < _numberOfUnits );
    unsigned first = from;
    unsigned start;
    unsigned end;
    unsigned noOfUnits = to - from + 1;
    int flush = code & SAVEMODIFIED_FLUSH;

    Assert( operation == ToCache || operation == ToMemoryObject );
    Assert( _cache );

#ifdef USESTACK
    SaveModifiedSynchronizeClosure syncStack;
#endif 
    SaveModifiedSynchronizeClosure * sync = 0;

    if ( from == to ) {

        // optimize for single unit case

        PhysicallyAddressableUnit * pau;
        int found = _map->findNextExist( first, to, start, end, pau );
        if ( found ) {
            if ( code & SAVEMODIFIED_QUEUED ) {
                Assert( operation == ToCache );
                unitsToStoreQueued( 1 );
            }
            // no need to _cache->sync since only one unit
            // and will be written through cache
            // if operation == ToMemoryObject
            ioOperation( operation, start, end, flush, pau, c ); 
        } 
        else if (operation == ToMemoryObject && _cache->dirty() != 0) {
#ifdef USESTACK
            if ( c == 0 ) {
                sync = &syncStack;
            } else {
                sync = new SaveModifiedSynchronizeClosure();
            }   
#else
            sync = new SaveModifiedSynchronizeClosure();
#endif
            sync->init( c, _cache, operation, from, noOfUnits, 
                        _log2WindowSize );
            
            sync->activate();
            if ( c == 0 ) {
                sync->_barrier.wait();
            }
        } 
        else {
            if ( c != 0 ) {
                c->activate();
            }
        }

    } else {

        CacheIOState * state = 0;
        unsigned expected = 0;
        while ( first <= to ) {
            PhysicallyAddressableUnit * pau;
            int found = _map->findNextExist
            ( first, to, start, end, pau );
            if ( !found )
                break;
            if ( state == 0 ) {
                if ( operation == ToMemoryObject ) {
#ifdef USESTACK
                    if ( c == 0 ) {
                        sync = &syncStack;
                    } else {
                        sync = new 
                        SaveModifiedSynchronizeClosure();
                    }   
#else
                    sync = new 
                    SaveModifiedSynchronizeClosure();
#endif
                    Raids( RaidDetail ) 
                    << this << "::saveModified() sync "
                    << asAddress( sync ) << "\n" << eor;
                    sync->init( c, _cache, operation,
                                from, noOfUnits, 
                                _log2WindowSize );
                    state = new CacheIOState( sync );
                } else {
                    state = new CacheIOState( c );
                }
                Raids( RaidDetail ) 
                << this << "::saveModified() state "
                << asAddress( state ) << "\n" << eor;
                //state->_spare = (unsigned) &MemoryObjectCache::saveModified;
                state->_spare = 15; // ??
            }
            state->add( start, end, (char *) pau );
            expected += end - start + 1;
            first = end + 1;
        }
    
        if ( state ) {
            state->_result = 0;
            state->_expectedResult = expected;
            if ( code & SAVEMODIFIED_QUEUED ) {
                Assert( operation == ToCache );
                unitsToStoreQueued( expected );
            }
            CacheIOState::_Operation * op = state->_ops;
            int opCount = state->_count;
            for( int i = 0; i < opCount; i++ ) {
                ioOperation( operation,
                             op->_startUnit, op->_count, 
                             flush,
                             (PhysicallyAddressableUnit *) 
                             op->_buf,
                             &(op->_closure) );
                op++;
            }
            if ( c == 0 ) {
                if ( operation == ToMemoryObject ) {
                    Assert( sync != 0 );
                    sync->_barrier.wait();
                } else {
                    Assert( sync == 0 );
                    state->_barrier.wait();
                }
            }
        }
        else if (operation == ToMemoryObject && _cache->dirty() != 0) {
#ifdef USESTACK
            if ( c == 0 ) {
                sync = &syncStack;
            } else {
                sync = new SaveModifiedSynchronizeClosure();
            }   
#else
            sync = new SaveModifiedSynchronizeClosure();
#endif
            sync->init( c, _cache, operation, from, noOfUnits, 
                        _log2WindowSize );
            
            sync->activate();
            if ( c == 0 ) {
                sync->_barrier.wait();
            }
        } 
        else {
            if ( c != 0 ) {
                c->activate();
            }
        }
    }
#ifdef USESTACK
#else
    if ( c == 0 && sync != 0 ) {
        delete sync;
    }
#endif

    Raids( RaidProtected ) 
    << this << "::saveModified() done\n" << eor;

    SInstr( MemoryObjectCacheSaveModified.stop(istart) );
}       

/*
 * Cache I/O and MO I/O
 */

class IOOperationAggregateState;

class IOOperationClosure : public IOAsyncClosure {
protected:
    IOAsyncClosure * _parent;
    IOOperationAggregateState * _state;
    MemoryObjectCache * _cache;
    int _flush;
    int _static;
    PhysicallyAddressableUnit * _pauList;

public:
    IOWindowInfo * _winInfo;
    unsigned _firstUnitOfWindow;
    unsigned _lastUnitOfWindow;
    char * _addr;
    CacheUnitInfo * _cu;
    
    static void func( IOOperationClosure * c );
    IOOperationClosure() { }
    void init( IOAsyncClosure * parent, IOOperationAggregateState * state,
               IOWindowInfo * winInfo, unsigned first, unsigned last, 
               char * addr, PhysicallyAddressableUnit * pau,
               CacheUnitInfo * cu, MemoryObjectCache * cache,
               int flush, int stat )
    { _parent = parent; _state = state;
    _winInfo = winInfo; _firstUnitOfWindow = first;
    _lastUnitOfWindow = last; _addr = addr; _pauList = pau; _cu = cu; 
    _cache = cache; _flush = flush; _static = stat;
 
//IMPORTANT CODE CHANGE BY SCOTT KIRCHER
    //For some idiotic reason, _func was not being assinged to func,
    //this was causing the default doNothing func to be called when
    //this closure was activated, so any IO operation that was supposed
    //to be finished up by this closure would FAIL. -SK
    _func=(AsyncFunc)&func; //the fix is so simple, but so HARD to find!! -SK
    }
};

class IOOperationAggregateState {
public:
    static Lock _lock;
    static const unsigned _prealloc = 8;
    unsigned _count;
    unsigned _completed;
    unsigned _result;
    IOAsyncClosure * _parent;
    IOOperationClosure * _closures;
    IOOperationClosure _preallocatedClosures[_prealloc];
    Barrier _barrier;
    unsigned _max;
    
    IOOperationAggregateState() : _barrier( 1 ), _max( _prealloc ) {}
    ~IOOperationAggregateState() 
    { if ( _max > _prealloc ) delete _closures; }
    void init( IOAsyncClosure * parent )
    { _parent = parent; _result = 0; _count = 0; _completed = 0;
    _closures = _preallocatedClosures; }
    void add( IOAsyncClosure * parent, IOOperationAggregateState * state,
              IOWindowInfo * winInfo, unsigned first, unsigned last,
              char * addr, PhysicallyAddressableUnit * pau,
              CacheUnitInfo * cu, MemoryObjectCache * cache,
              int flush, int stat );
};

Lock IOOperationAggregateState::_lock;

inline 
void
IOOperationAggregateState::add( 
    IOAsyncClosure * parent, IOOperationAggregateState * state,
    IOWindowInfo * winInfo, unsigned first, unsigned last,
    char * addr, PhysicallyAddressableUnit * pau,
    CacheUnitInfo * cu, MemoryObjectCache * cache,
    int flush, int stat ) 
{

    if ( _count == _max ) {
        int newMax = _max << 1;
        IOOperationClosure * newClosures = 
        new IOOperationClosure[newMax];
        for( unsigned  i = 0; i < _max; i++ ) {
            newClosures[i] = _closures[i];
        }
        if ( _closures != _preallocatedClosures )
            delete _closures;
        _closures = newClosures;
        _max = newMax;
    }
    Assert( _count < _max );
    _closures[_count].init( parent, state, winInfo, first, last, addr,
                            pau, cu, cache, flush, stat );
    _count++;
}

void
IOOperationClosure::func( IOOperationClosure * c )
{
    MemoryObjectCache * cache = c->_cache;


    if ( c->_cu )
        delete c->_cu;

    if ( c->_flush ) {
        cache->untranslatable
        ( c->_firstUnitOfWindow, c->_lastUnitOfWindow, 
          MemoryObjectCache::_translation, 
          c->_winInfo->baseAddr() );
    } else {
        unsigned noOfUnits =  
        c->_lastUnitOfWindow - c->_firstUnitOfWindow + 1;
#ifndef CHANGESTATUS
        cache->untranslatable
        ( c->_firstUnitOfWindow, c->_lastUnitOfWindow, 
          MemoryObjectCache::_translation, 
          c->_winInfo->baseAddr() );
#else
        cache->_translation->changeStatus
        ( c->_addr, noOfUnits << cache->log2UnitSize(),
          clearMod, setNone );
#endif // CHANGESTATUS
        cache->setModified( c->_pauList, noOfUnits, 0 );
    }

    TheIOWindowCache->releaseIOWindow( c->_winInfo );
    
    if ( c->_parent )
        c->_parent->activate( c->_result );
    IOOperationAggregateState * s = c->_state;
    if ( s ) {
        IOOperationAggregateState::_lock.acquire();
        s->_result += c->_result;
        s->_completed++;
        unsigned  completed = s->_completed;
        IOOperationAggregateState::_lock.release();
        if ( completed == s->_count ) {
            if ( s->_parent != 0 ) {
                s->_parent->activate( s->_result );
                delete s;
            } else {
                s->_barrier.done();
            }
        }
    }
    if ( !c->_static )
        delete c;
}

unsigned
MemoryObjectCache::ioOperation( IOOperationCode operation,
                                unsigned from, unsigned to, int flush,
                                PhysicallyAddressableUnit * pauList,
                                IOAsyncClosure * c )
{ 
    Raids( RaidMethod ) << this << "::ioOperation( "
                        << operation << ", " << from << ", " << to 
                        << ", " << flush << ", " << asAddress( pauList )
                        << ", " << asAddress( c ) << " )\n" << eor;

    Assert( isIn( to ) && isIn( from ) );

    unsigned totalCount = 0;

    /*
     * see if the existing view can be used
     */

    unsigned baseUnitOfFirstWindow = from & ~_windowSizeRatioMask; 

    /*
     * compute first and last unit (offset) of first and last Windows
     */

    unsigned firstUnitOffsetOfFirstWindow = 
    from - baseUnitOfFirstWindow;

    unsigned lastUnitOffsetOfLastWindow =
    ( firstUnitOffsetOfFirstWindow + to - from ) &
    _windowSizeRatioMask;

    unsigned baseUnitOfLastWindow = to - lastUnitOffsetOfLastWindow;

    unsigned firstUnitOfWindow;
    unsigned lastUnitOfWindow;
    unsigned noOfUnitsInWindow;

    if ( baseUnitOfFirstWindow == baseUnitOfLastWindow ) {

        Raids( RaidDetail ) 
        << this << "::ioOperation() shortcut\n" << eor;

        noOfUnitsInWindow = 
        lastUnitOffsetOfLastWindow -
        firstUnitOffsetOfFirstWindow + 1;
        firstUnitOfWindow = 
        firstUnitOffsetOfFirstWindow +
        baseUnitOfFirstWindow;  
        lastUnitOfWindow = 
        firstUnitOfWindow + noOfUnitsInWindow - 1;

        IOWindowInfo * winInfo = _ioWindow = 
        TheIOWindowCache->acquireIOWindow
        ( this, baseUnitOfFirstWindow, _ioWindow );

        char * addr = ( char * ) winInfo->baseAddr() + 
        ( ( firstUnitOfWindow - baseUnitOfFirstWindow ) <<
          log2UnitSize() );
        unsigned noOfUnits = lastUnitOfWindow - firstUnitOfWindow + 1;

#ifdef VIRTUALADDRESSIO
        translatable( firstUnitOfWindow, lastUnitOfWindow, NoAccess,
                      _translation, addr, winInfo->view(), 
                      winInfo->baseAddr(), pauList );
#endif // VIRTUALADDRESSIO

        CacheUnitInfo * cu;
        if ( operation != FromCache ) {
            cu = new CacheUnitInfo[ noOfUnits ];
            setCacheUnitInfo( pauList, noOfUnits, cu );
        } else {
            cu = 0;
        }

        IOOperationClosure stackClosure;

        IOOperationClosure * myClosure =
        ( c == 0 ? &stackClosure : new IOOperationClosure() );
        myClosure->init( c, 0, winInfo, firstUnitOfWindow,
                         lastUnitOfWindow, addr, pauList, cu, this,
                         flush, ( c == 0 ) );
        IOOperationClosure * paramClosure = ( c == 0 ? 0 : myClosure );

        unsigned actual;
        switch ( operation ) {
            case FromCache:   
                fromCacheUnitsAdd( noOfUnits );
                actual = _cache->read( firstUnitOfWindow, noOfUnits, 
                                       addr, paramClosure );
                break;
            case ToCache:
                actual = _cache->write( firstUnitOfWindow, 
                                        noOfUnits, cu, addr, 
                                        paramClosure );
                break;
            case ToMemoryObject:
                actual = _cache->rawWrite( firstUnitOfWindow, 
                                           noOfUnits, cu, addr, 
                                           paramClosure );
                break;
            default:
                actual = 0;
        }

        if ( c == 0 ) {
            myClosure->activate( actual );
            totalCount = actual;
        } else {
            totalCount = noOfUnits;
        }
    } else {

#ifdef USESTACK
        IOOperationAggregateState stackAggState;
#endif
        IOOperationAggregateState * aggState;
#ifdef USESTACK
        if ( c == 0 ) {
            aggState = &stackAggState;
        } else {
            aggState = new IOOperationAggregateState();
        }
#else
        aggState = new IOOperationAggregateState();
#endif
        aggState->init( c );

        for ( unsigned baseUnitOfWindow = baseUnitOfFirstWindow;
              baseUnitOfWindow <= baseUnitOfLastWindow;
              baseUnitOfWindow += _windowSizeRatio ) {

            /*
             * compute the number of units needed from this Window
             */

            if ( baseUnitOfFirstWindow == baseUnitOfWindow ) {
                noOfUnitsInWindow = 
                _windowSizeRatio - 
                firstUnitOffsetOfFirstWindow;
                firstUnitOfWindow = 
                firstUnitOffsetOfFirstWindow +
                baseUnitOfFirstWindow;  
            } else if ( baseUnitOfLastWindow == baseUnitOfWindow ){
                noOfUnitsInWindow =
                lastUnitOffsetOfLastWindow + 1;
                firstUnitOfWindow = baseUnitOfWindow;
            } else {
                noOfUnitsInWindow = _windowSizeRatio;
                firstUnitOfWindow = baseUnitOfWindow;
            }
            lastUnitOfWindow = 
            firstUnitOfWindow + noOfUnitsInWindow - 1;
            
            IOWindowInfo * winInfo = _ioWindow = 
            TheIOWindowCache->acquireIOWindow
            ( this, baseUnitOfWindow, _ioWindow );

            char * addr = ( char * ) winInfo->baseAddr() + 
            ( ( firstUnitOfWindow - baseUnitOfWindow ) 
              << log2UnitSize() );
            unsigned noOfUnits = 
            lastUnitOfWindow - firstUnitOfWindow + 1;

#ifdef VIRTUALADDRESSIO
            translatable( firstUnitOfWindow, lastUnitOfWindow, 
                          NoAccess,
                          _translation, addr, winInfo->view(), 
                          winInfo->baseAddr(), pauList );
#endif // VIRTUALADDRESSIO

            CacheUnitInfo * cu;
            if ( operation != FromCache ) {
                cu = new CacheUnitInfo[ noOfUnits ];
                setCacheUnitInfo( pauList, noOfUnits, cu );
            } else {
                cu = 0;
            }

            aggState->add( 0, aggState, 
                           winInfo, firstUnitOfWindow,
                           lastUnitOfWindow, addr, pauList, cu,
                           this, flush, 1 );

            if ( lastUnitOfWindow != to ) {
                pauList = forward
                ( pauList, 
                  lastUnitOfWindow - firstUnitOfWindow + 1 ); 
            }
            
        }

        IOOperationClosure * op = aggState->_closures;
        int opCount = aggState->_count;
        for (int i = 0; i < opCount; i++, op++ ) {
            
            unsigned noOfUnits = 
            op->_lastUnitOfWindow - op->_firstUnitOfWindow 
            + 1;
            switch ( operation ) {
                case FromCache:   
                    fromCacheUnitsAdd( noOfUnits );
                    _cache->read( op->_firstUnitOfWindow, 
                                  noOfUnits, op->_addr, op );
                    break;
                case ToCache:
                    _cache->write( op->_firstUnitOfWindow, 
                                   noOfUnits, op->_cu, 
                                   op->_addr, op );
                    break;
                case ToMemoryObject:
                    _cache->rawWrite( op->_firstUnitOfWindow, 
                                      noOfUnits, op->_cu,
                                      op->_addr, op );
                    break;
            }

        }
        if ( c == 0 ) {
            aggState->_barrier.wait();
            totalCount = aggState->_result;
        } else {
            totalCount = to - from + 1;
        }
#ifdef USESTACK
#else
        if ( c == 0 ) 
            delete aggState;
#endif
    }

    Raids( RaidMethod ) << this << "::ioOperation() done\n" << eor;
    return totalCount;
}

MemoryObjectCacheTimer
MemoryObjectCacheToMemoryObject("MemoryObjectCache::toMemoryObject() timer");

unsigned
MemoryObjectCache::toMemoryObject( unsigned from, unsigned to, int flush,
                                   PhysicallyAddressableUnit * pau,
                                   IOAsyncClosure * c )
{
    SInstr( int istart = MemoryObjectCacheToMemoryObject.start() );

    Raids( RaidProtected ) << this << "::toMemoryObject( "
                           << from << ", " << to << " )\n" << eor;

    unsigned result;
    if ( _cache != 0 ) {
        result = ioOperation( ToMemoryObject, from, to, flush, pau, c );
    } else {
        result = 0;
        if ( c )
            c->activate( 0 );
    }

    SInstr( MemoryObjectCacheToMemoryObject.stop( istart ) );
    return result;
}

MemoryObjectCacheTimer
MemoryObjectCacheToCache("MemoryObjectCache::toCache() timer");

unsigned
MemoryObjectCache::toCache( unsigned from, unsigned to, int flush,
                            PhysicallyAddressableUnit * pau,
                            IOAsyncClosure * c )
{   
    SInstr( int istart = MemoryObjectCacheToCache.start() );
    Raids( RaidProtected ) << this << "::toCache( "
                           << from << ", " << to << " )\n" << eor;

    unsigned result;
    if ( _cache != 0 ) {
        result = ioOperation( ToCache, from, to, flush, pau, c );
    } else {
        result = 0;
        if ( c )
            c->activate( 0 );
    }

    SInstr( MemoryObjectCacheToCache.stop( istart ) );
    return result;
}

MemoryObjectCacheTimer
MemoryObjectCacheFromCache("MemoryObjectCache::fromCache() timer");

unsigned
MemoryObjectCache::fromCache( unsigned from, unsigned to, 
                              PhysicallyAddressableUnit * pau,
                              IOAsyncClosure * c )
{
    SInstr( int istart = MemoryObjectCacheFromCache.start() );
    Raids( RaidProtected ) << this << "::fromCache( "
                           << from << ", " << to << " )\n" << eor;

    unsigned result;
    if ( _cache != 0 ) {
        result = ioOperation( FromCache, from, to, 0, pau, c );
    } else {
        result = 0;
        if ( c )
            c->activate( 0 );
    }

    SInstr( MemoryObjectCacheFromCache.stop( istart ) );
    return result;
}

/*
 * Net I/O, currently undefined
 */

void
MemoryObjectCache::toNet( unsigned from, unsigned to )
{
    Assert( from <= to && to < _numberOfUnits );
    Assert( NOTREACHED );
}

void
MemoryObjectCache::fromNet( unsigned from, unsigned to )
{
    Assert( from <= to && to < _numberOfUnits );
    Assert( NOTREACHED );
}

/*
 * Reference manipulations
 */

int
MemoryObjectCache::clearReferenced( unsigned from, unsigned to )
{
    if ( acquireMo( from, to, 0, 0 ) != 0 ) 
        return 1;
    PhysicallyAddressableUnit * pau = _map->pauList( from, to );
    unsigned count = ( to - from + 1 ) << _log2PauUnitSizeRatio;
    int ret = 0;
    while ( count > 0 && pau != 0 ) {
        // Assert( pau != 0 );
        if ( pau->referenced() )
            ret = 1;
        pau->setReferenced( 0 );
        pau = pau->nextLink();
        count--;
    }
    if ( _adtMap->referenced( from, to, 1 ) )
        ret = 1;
    postFault( from, to );
    return ret;
}

PhysicallyAddressableUnit *
MemoryObjectCache::checkReferenced( unsigned from, unsigned to, int mode )
{
    if ( acquireMo( from, to, 0, 0 ) != 0 )
        return 0;
    if ( _adtMap->referenced( from, to, 0 ) ) {
        if ( mode ) {
            PhysicallyAddressableUnit * pau = 
            _map->pauList( from, to );
            Assert( pau != 0 );
            return pau;
        } else {
            postFault( from, to );
            return 0;
        }
    } else {
        PhysicallyAddressableUnit * firstPau; 
        PhysicallyAddressableUnit * pau = 
        firstPau = _map->pauList( from, to );
        unsigned count = ( to - from + 1 ) << _log2PauUnitSizeRatio; 
//      Assert( pau != 0 );
        if ( pau == 0 ) {
            if ( mode == 0 ) {
                postFault( from, to );
                return 0;
            }
            Assert( NOTREACHED );
        }
        while ( count ) { 
            if ( pau->referenced() ) {
                if ( mode ) {
                    return firstPau;
                } else {
                    postFault( from, to );
                    return 0; 
                }
            }
            pau = pau->nextLink(); 
            count--; 
        } 
        if ( mode ) {
            postFault( from, to );
            return 0;   
        } else {
            return firstPau;
        }
    }
}

/*
 * VM manipulations
 */

MemoryObjectCacheTimer
MemoryObjectCacheNProtect("MemoryObjectCache::nprotect() timer");

int
MemoryObjectCache::nprotect( AddressTranslation * adt, const char * vaddr,
                             ProtectionLevel prot, unsigned from, unsigned to )
{
    SInstr( int istart = MemoryObjectCacheNProtect.start() );
    Raids( RaidMethod ) << this << "::nprotect( "
                        << adt << ", " << asHex( ( unsigned ) vaddr ) << ", "
                        << from << ", " << to << " )\n" << eor;

    Assert( adt != 0 );
    Assert( from <= to && to < _numberOfUnits );

    int result = adt->changeProtection
    ( floorUnit( vaddr ),
      ( to - from + 1 ) << log2UnitSize(),
      prot );
    SInstr( MemoryObjectCacheNProtect.stop( istart) );
    return result;
}

/* Currently undefined */

ProtectionLevel
MemoryObjectCache::whatProtection( unsigned from, unsigned to )
{
    Assert( from <= to && to < _numberOfUnits );
    Assert( NOTREACHED );
    return ( ProtectionLevel ) 0;
}

MemoryObjectCacheTimer
MemoryObjectCacheTranslatable("MemoryObjectCache::translatable() timer");

void
MemoryObjectCache::translatable( unsigned from, unsigned to,
                                 ProtectionLevel prot,
                                 AddressTranslation * adt, 
                                 const char * vaddr,
                                 MemoryObjectView * view,
                                 const char * base,
                                 PhysicallyAddressableUnit * pau )
{
    SInstr( int istart = MemoryObjectCacheTranslatable.start() );
    Raids( RaidProtected ) << this << "::translatable( "
                           << from << ", " << to << ", " << prot << ", "
                           << adt << ", " << asHex( ( unsigned ) vaddr ) 
                           << " )\n" << eor;

    Assert( _adtMap != 0 );
    _adtMap->translatable( from, to, prot, adt, vaddr, view, base, pau );
    SInstr( MemoryObjectCacheTranslatable.stop( istart ) );
}

/*
 * hackTranslat is called after ::cache gets called. It analyses the
 * cause of the fault, and updates the translations appropriately.
 * It can obviously be condensed into just nprotect and translatable calls.
 * Proper design should remove the need for so many parameters.
 * But currently, there is a bug connected to this function. If called
 * PremappedMemoryObjectCache, it causes the system to crash. Until
 * that bug is stomped, I am not going to touch this mess
 */

MemoryObjectCacheTimer
MemoryObjectCacheHackTranslat("MemoryObjectCache::hackTranslat() timer");

void
MemoryObjectCache::hackTranslat( CacheReason reason, 
                                 unsigned from, unsigned to,
                                 ProtectionLevel requiredProtection,
                                 AddressTranslation * adt, 
                                 const char * vaddr, 
                                 MemoryObjectView * view,
                                 const char * baseAddr,
                                 FaultErrorCode code,
                                 PhysicallyAddressableUnit * pau )
{
    SInstr( int istart = MemoryObjectCacheHackTranslat.start() );
    Raids( RaidProtected ) << this << "::hackTranslat( "
                           << reason << ", " << from << ", " << to << ", "
                           << requiredProtection << ", " << adt << ", "
                           << adt << ", " << asHex( (unsigned) vaddr ) << ", "
                           << code << " )\n" << eor;
            
    if ( reason & DomainRepairFault ) {
        switch ( code ) {
            case MissingPhysicalMemory:
            case MissingAddressTranslationMapping:
                translatable( from, to, requiredProtection, adt, 
                              vaddr, view, baseAddr, pau ); 
                break;
            
            case InsufficientProtectionLevel:
                if( ! nprotect( adt, vaddr, requiredProtection, from, 
                                to ) )
                    translatable( from, to, requiredProtection, 
                                  adt, vaddr, view, baseAddr, pau );
                break;
            default:
                Assert( NOTREACHED );
                break;
        }
    }

    if ( reason & DomainMakeAddressable ) {
        //nprotect( adt, vaddr, requiredProtection, from, to);
        translatable( from, to, requiredProtection, adt, vaddr,
                      view, baseAddr, pau );
    }
    SInstr( MemoryObjectCacheHackTranslat.stop( istart ) );
}

/*
 * Miscellaneous public functions
 */

void
MemoryObjectCache::writeName( OutputStream * os )
{
    ProxiableObject::writeName( os );
}

/*
 * Force a unit to be held in the cache. The initial value of 
 * requiredProtection is what is requested. It might be strengthened 
 * or weakened.
 */

MemoryObjectCacheTimer
MemoryObjectCacheCache("MemoryObjectCache::cache() timer");

LockHolder *
MemoryObjectCache::cache( unsigned int unit,
                          ProtectionLevel requiredProtection,
                          ProtectionLevel & grantedProtection, 
                          FaultErrorCode & code,
                          AddressTranslation * adt,
                          const char * vaddr, CacheReason reason,
                          MemoryObjectView * view, const char * baseAddr )
{
    SInstr( int istart = MemoryObjectCacheCache.start() );
    Raids( RaidMethod ) << this << "::cache( unit " << unit 
                        << ", requiredProtection " << requiredProtection 
                        << ", vaddr " << asHex( (unsigned) vaddr )
                        << ", reason " << reason
                        << ", view " << view
                        << ", baseAddr " << asHex( (unsigned) baseAddr )
                        << " )\n" << eor; 
    
    LockHolder * lockHolder = 0;

    int preLocked = (view->classOf() == LockedMemoryObjectViewClass);

    if ( ! preLocked )
        preFault( unit, unit );

    Assert( isIn( unit ) );
    Assert( requiredProtection <=
            _memoryObject->logicalProtectionLevel() );
    
    bytesAccessedAdd( 1 << _log2UnitSize );

#ifdef VIRTUALADDRESSIO
    if  (!preLocked) {
#endif // VIRTUALADDRESSIO
        PhysicallyAddressableUnit * pauList = _map->pauList( unit, unit );
        if ( pauList ) {
        
            // unit exists - its protection <= the MemoryObject protection
            ProtectionLevel currentProt = protection( pauList, 1 );
            Assert( currentProt <= 
                    _memoryObject->logicalProtectionLevel() );
        
            if ( requiredProtection > currentProt ) {
                code = InsufficientProtectionLevel;
                setProtection( pauList, 1, requiredProtection );
            } else {
                code = MissingAddressTranslationMapping;
            }

        } else {

            Assert( ! preLocked );
            Assert( MainStore != 0 );
            pauList = unitsFromStore( unit, unit );
            fromCache( unit, unit, pauList, 0 );
            if ( _replacementPolicy ) {
#ifdef DEBUG_GLRU
                RaidStream() 
                << this << "::cache rp " 
                << asAddress( _replacementPolicy )
                << "\n" << eor;
#endif // DEBUG_GLRU
                _replacementPolicy->postPageIn( unit, unit );
            }
            code = MissingPhysicalMemory;
            setProtection( pauList, 1, requiredProtection );
        
        }

        hackTranslat( reason, unit, unit, requiredProtection, adt, vaddr, 
                      view, baseAddr, code, pauList );
 
#ifdef VIRTUALADDRESSIO
    }
#endif // VIRTUALADDRESSIO

    grantedProtection = requiredProtection;
    
    if ( reason & DomainConstructChain ) {
        lockHolder = new LockHolder( unit, unit, this,
                                     _map->pauList( unit, unit ),
                                     !preLocked );
    }

    if ( !preLocked && lockHolder == 0 )
        postFault( unit, unit );
    
    SInstr( MemoryObjectCacheCache.stop( istart ) );
    return lockHolder;
}

/*
 * Control flow is similar to saveModified
 */

MemoryObjectCacheTimer
MemoryObjectCacheMCache("MemoryObjectCache::mcache() timer");

LockHolder *
MemoryObjectCache::mcache( unsigned int unit, int nunits,
                           ProtectionLevel requiredProtection,
                           ProtectionLevel & grantedProtection,
                           FaultErrorCode & code, AddressTranslation * adt,
                           const char * vaddr, CacheReason reason,
                           MemoryObjectView * view, const char * baseAddr )
{
    SInstr( int istart = MemoryObjectCacheMCache.start() );
    unsigned from = unit; 
    unsigned to = unit + nunits - 1;

    Raids( RaidMethod ) << this << "::mcache( from " << from
                        << ", to " << to
                        << ", requiredProtection " << requiredProtection 
                        << ", vaddr " << asHex( (unsigned) vaddr )
                        << ", reason " << reason
                        << ", view " << view
                        << ", baseAddr " << asHex( (unsigned) baseAddr )
                        << " )\n" << eor; 

    int preLocked = ( view->classOf() == LockedMemoryObjectViewClass );

    if ( !preLocked )
        preFault( from, to );

    Assert( from <= to && to < numberOfUnits() );

    bytesAccessedAdd( nunits << _log2UnitSize );

#ifdef VIRTUALADDRESSIO
    if ( !preLocked ) {
#endif // VIRTUALADDRESSIO

        code = MissingAddressTranslationMapping;
    
        unsigned first = from;
        unsigned start;
        unsigned end;
    
        Assert( requiredProtection <=
                _memoryObject->logicalProtectionLevel() );

        PhysicallyAddressableUnit * pauList = 0;
        PhysicallyAddressableUnit * pau;

        while ( first <= to ) {
            int found = _map->findNextNonExist(first, to, start, end, pau);
            if ( first == from ) {
                pauList = pau;
            }
            if ( !found )
                break;
            Assert( !preLocked );
            PhysicallyAddressableUnit * fromStorePauList =
            unitsFromStore( start, end );
            if ( start == from )
                pauList = fromStorePauList;
            fromCache( start, end, fromStorePauList, 0 );
            if ( _replacementPolicy ) {
#ifdef DEBUG_GLRU
                RaidStream() 
                << this << "::mcache rp " 
                << asAddress( _replacementPolicy )
                << "\n" << eor;
#endif // DEBUG_GLRU
                _replacementPolicy->postPageIn( start, end );
            }
            code = MissingPhysicalMemory;
            first = end + 1;
        }

        setProtection( pauList, to - from + 1, requiredProtection );
        hackTranslat( reason, from, to, requiredProtection, adt, vaddr, 
                      view, baseAddr, code, pauList );

#ifdef VIRTUALADDRESSIO
    }
#endif // VIRTUALADDRESSIO

    grantedProtection = requiredProtection;
    
    LockHolder * lockHolder = 0;

    if ( reason & DomainConstructChain ) {
        lockHolder = new LockHolder( from, to, this,
                                     _map->pauList( from, to ),
                                     !preLocked );
    }

    if ( !preLocked && lockHolder == 0 )
        postFault( from, to );
    
    SInstr( MemoryObjectCacheMCache.stop( istart ) );
    return lockHolder;
}

MemoryObjectCacheTimer
MemoryObjectCacheUntranslatable("MemoryObjectCache::untranslatable() timer");

void
MemoryObjectCache::untranslatable( unsigned from, unsigned to,
                                   AddressTranslation * adt,
                                   const char * where)
{
    SInstr( int istart = MemoryObjectCacheUntranslatable.start() );
    Raids( RaidMethod ) << this << "::untranslatable( "
                        << from << ", " << to << ", " 
                        << adt << ", " << asHex( ( unsigned ) where ) 
                        << " )\n" << eor;

    Assert( _adtMap != 0 );
    _adtMap->untranslatable( from, to, adt, where );
    SInstr( MemoryObjectCacheUntranslatable.stop( istart ) );
}

MemoryObjectCacheTimer
MemoryObjectCacheUntranslatableView("MemoryObjectCache::untranslatable(view) timer");

void
MemoryObjectCache::untranslatable(AddressTranslation * adt,
                                  MemoryObjectView * view,
                                  const char * where)
{
    SInstr( int istart = MemoryObjectCacheUntranslatableView.start() );
    Raids( RaidMethod ) << this << "::untranslatable( "
                        << adt << ", " << view << ", " 
                        << asHex( ( unsigned ) where ) << " )\n" << eor;

    unsigned from = offsetToUnit( view->viewOffset() );
    unsigned to = offsetToUnit( view->viewOffset() + view->size() - 1 );
    untranslatable( from, to, adt, where );         
    SInstr( MemoryObjectCacheUntranslatableView.stop( istart ) );
}

/*
 * Unlock 
 */

void 
MemoryObjectCache::unlock( unsigned from, unsigned to )
{
    Assert( from <= to && to < _numberOfUnits );
    postFault( from, to );
}

/*
 * Free
 */

void
MemoryObjectCache::free()
{
    Raids( RaidMethod ) 
    << this << "::free()\n" << eor;

    if ( _numberOfUnits != 0 ) {
        preFault( 0, _numberOfUnits - 1 );
        untranslatable( 0, _numberOfUnits - 1 );
        unitsToStore( 0, _numberOfUnits - 1, 0 );
        postFault( 0, _numberOfUnits - 1 );
    }
}

void
MemoryObjectCache::free( unsigned from, unsigned to )
{
    Raids( RaidMethod ) 
    << this << "::free( "
    << from << ", " << to << " )\n" << eor;

    preFault( from, to );
    Assert( from <= to && to < _numberOfUnits );

    untranslatable( from, to );
    unitsToStore( from, to, 0 );

    postFault( from, to );
}

/*
 * File system caching
 */

class MOCacheReadAggregateState;

class MOCacheReadClosure : public IOAsyncClosure {
protected:
    MOCacheReadAggregateState * _state;
    MemoryObjectCache * _cache;
    PhysicallyAddressableUnit * _pauList;

public:
    IOWindowInfo * _winInfo;
    unsigned _startUnit;
    unsigned _count;
    char * _addr;
    
    static void func( MOCacheReadClosure * c );
    MOCacheReadClosure() {}
    void init( MOCacheReadAggregateState * state,
               IOWindowInfo * winInfo, unsigned startUnit, unsigned count,
               char * addr, PhysicallyAddressableUnit * pau,
               MemoryObjectCache * cache )
    { _state = state; 
    _winInfo = winInfo; _startUnit = startUnit, _count = count;
    _addr = addr; _pauList = pau; _cache = cache;
//IMPORTANT CODE CHANGE BY SCOTT KIRCHER
    //For some idiotic reason, _func was not being assinged to func,
    //this was causing the default doNothing func to be called when
    //this closure was activated, so any IO operation that was supposed
    //to be finished up by this closure would FAIL. -SK
    _func=(AsyncFunc)&func; //the fix is so simple, but so HARD to find!! -SK
    }
};

class MOCacheReadAggregateState {
public:
    struct CopyInfo {
        IOWindowInfo * _winInfo;
        unsigned _firstUnitOfWindow;
        unsigned _lastUnitOfWindow;
        PhysicallyAddressableUnit * _pauList;
    };

    static Lock _lock;
    static const unsigned _prealloc = 8;
    unsigned _count;
    unsigned _completed;
    unsigned _log2UnitSize;
    unsigned _firstUnitOffset;
    unsigned _lastUnitNoOfBytes;
    unsigned _totalCount;
    MemoryObjectCache * _cache;
    IOAsyncClosure * _parent;
    MOCacheReadClosure * _closures;
    MOCacheReadClosure _preallocatedClosures[_prealloc];
    Barrier _barrier;
    unsigned _max;

    static const unsigned _copyPrealloc = 4;
    unsigned _copyCount;
    CopyInfo _copyInfo[_copyPrealloc];
    char * _buf;

    
    MOCacheReadAggregateState() : _barrier( 1 ), _max( _prealloc ) {}
    ~MOCacheReadAggregateState() 
    { if ( _max > _prealloc ) delete _closures; }
    void init( IOAsyncClosure * parent, MemoryObjectCache * cache,
               unsigned log2UnitSize, char * buf,
               unsigned firstUnitOffset, unsigned lastUnitNoOfBytes )
    { _parent = parent; _count = 0; _completed = 0; 
    _closures = _preallocatedClosures; _copyCount = 0; _cache = cache; 
    _log2UnitSize = log2UnitSize; _buf = buf; 
    _firstUnitOffset = firstUnitOffset; 
    _lastUnitNoOfBytes = lastUnitNoOfBytes; }
    void add( MOCacheReadAggregateState * state,
              IOWindowInfo * winInfo, unsigned startUnit, unsigned count,
              char * addr, PhysicallyAddressableUnit * pau,
              MemoryObjectCache * cache );
    void addCopyInfo( IOWindowInfo * winInfo, unsigned firstUnitOfWindow,
                      unsigned lastUnitOfWindow, 
                      PhysicallyAddressableUnit * pau );
    void doCopy();
};

Lock MOCacheReadAggregateState::_lock;

inline 
void
MOCacheReadAggregateState::add( 
    MOCacheReadAggregateState * state,
    IOWindowInfo * winInfo, unsigned startUnit, unsigned count,
    char * addr, PhysicallyAddressableUnit * pau,
    MemoryObjectCache * cache )
{
    if ( _count == _max ) {
        int newMax = _max << 1;
        MOCacheReadClosure * newClosures = 
        new MOCacheReadClosure[newMax];
        for( unsigned i = 0; i < _max; i++ ) {
            newClosures[i] = _closures[i];
        }
        if ( _closures != _preallocatedClosures )
            delete _closures;
        _closures = newClosures;
        _max = newMax;
    }
    Assert( _count < _max );
    _closures[_count].init( state, winInfo, startUnit, count, 
                            addr, pau, cache );
    _count++;
}

inline
void 
MOCacheReadAggregateState::addCopyInfo( 
    IOWindowInfo * winInfo, unsigned firstUnitOfWindow,
    unsigned lastUnitOfWindow, PhysicallyAddressableUnit * pau )
{
    Assert( _copyCount < _copyPrealloc );
    CopyInfo * ci = _copyInfo + _copyCount;
    _copyCount++;
    ci->_winInfo = winInfo;
    ci->_firstUnitOfWindow = firstUnitOfWindow;
    ci->_lastUnitOfWindow = lastUnitOfWindow;
    ci->_pauList = pau;
}

void
MOCacheReadAggregateState::doCopy()
{
    CopyInfo * copyInfo = _copyInfo;
    char * buf = _buf;
    _totalCount = 0;

    for ( unsigned i = 0; i < _copyCount; i++, copyInfo++ ) {

        /* 
         * figure out how much to copy 
         */

        IOWindowInfo * winInfo = copyInfo->_winInfo;
        unsigned baseUnitOfWindow = winInfo->from();
        unsigned firstUnitOfWindow = copyInfo->_firstUnitOfWindow;
        unsigned lastUnitOfWindow = copyInfo->_lastUnitOfWindow;

        unsigned startOffset = 
        ( ( firstUnitOfWindow - baseUnitOfWindow ) 
          << _log2UnitSize );
        unsigned count =
        ( ( lastUnitOfWindow - firstUnitOfWindow + 1 )
          << _log2UnitSize );
        if ( i == 0 ) {
            count -= _firstUnitOffset;
            startOffset += _firstUnitOffset;
        }
        if ( i+1 == _copyCount )
            count -= ( 1 << _log2UnitSize ) - _lastUnitNoOfBytes;

        char * startAddr = ( char * ) winInfo->baseAddr() + 
        ( ( firstUnitOfWindow - baseUnitOfWindow )
          << _log2UnitSize );
        _cache->translatable( firstUnitOfWindow, lastUnitOfWindow, 
                              NoAccess, 
                              MemoryObjectCache::_translation, 
                              startAddr,
                              winInfo->view(), winInfo->baseAddr(),
                              copyInfo->_pauList );
        ByteCopy( winInfo->baseAddr() + startOffset, buf, count );
        buf += count;   
        _totalCount += count;
        TheIOWindowCache->releaseIOWindow( winInfo );
    }
    if ( _cache->_replacementPolicy )
        _cache->_replacementPolicy->postCacheRead
        ( _copyInfo[0]._firstUnitOfWindow,
          _copyInfo[_copyCount - 1]._lastUnitOfWindow );

}

void
MOCacheReadClosure::func( MOCacheReadClosure * c )
{
    Assert( c->_result == c->_count );

    Debug<<"MOCacheReadClosure::func()\n"<<eor;

    MemoryObjectCache * cache = c->_cache;
    unsigned firstUnit = c->_startUnit;
    unsigned lastUnit = firstUnit + c->_count - 1;

#ifndef CHANGESTATUS
    cache->untranslatable( firstUnit, lastUnit,
                           MemoryObjectCache::_translation,
                           c->_winInfo->baseAddr() );
#else
    cache->_translation->changeStatus( c->_addr, 
                                       c->_count << cache->log2UnitSize(),
                                       clearMod, setNone );
#endif // CHANGESTATUS
    cache->setModified( c->_pauList, c->_count, 0 );

    if ( cache->_replacementPolicy ) {
#ifdef DEBUG_GLRU
        RaidStream() 
        << cache << "::MOCacheReadClosure::func() rp "
        << asAddress( cache->_replacementPolicy )
        << "\n" << eor;
#endif // DEBUG_GLRU
        cache->_replacementPolicy->postPageIn( firstUnit, lastUnit );
    }

    MOCacheReadAggregateState * s = c->_state;
    Assert( s );
    if ( s ) {
        MOCacheReadAggregateState::_lock.acquire();
        s->_completed++;
        unsigned completed = s->_completed;
        MOCacheReadAggregateState::_lock.release();
        Assert( completed <= s->_count );
        if ( completed == s->_count ) {
            if ( s->_parent ) {
                s->doCopy();
                s->_parent->activate( s->_totalCount );
                delete s;
            } else {
                s->_barrier.done();
            }
        }
    }
}

MemoryObjectCacheTimer
MemoryObjectCacheCacheRead("MemoryObjectCache::cacheRead() timer");

unsigned
MemoryObjectCache::cacheRead( unsigned offset, unsigned noOfBytes, char * buf,
                              IOAsyncClosure * c )
{
    SInstr( int istart = MemoryObjectCacheCacheRead.start() );
    Raids( RaidMethod ) << this << "::cacheRead"
                        << "( offset " << offset 
                        << ", noOfBytes " << noOfBytes
                        << ", buffer " << asAddress( buf )
                        << ", closure " << asAddress( c )
                        << " )\n" << eor;
#ifdef JEFF
    Console() << this << "::" << __FUNCTION__
              << "( offset " << offset 
              << ", noOfBytes " << noOfBytes
              << ", buffer " << asAddress( buf )
              << ", closure " << asAddress( c )
              << " )\n" << eor;
#endif

    unsigned totalCount;
    bytesAccessedAdd( noOfBytes );

    /* 
     * compute unit numbers
     */
    unsigned from = offsetToUnit( offset );
    unsigned to   = offsetToUnit( offset + noOfBytes - 1 );
    Assert( isIn( to ) && isIn( from ) );

    if ( _replacementPolicy )
        _replacementPolicy->preCacheRead( from, to );

    /*
     * compute number of bytes to copy from first and last
     * unit
     */

    unsigned unitSizeMask = unitSize() - 1;
    unsigned firstUnitOffset = offset & unitSizeMask;
    unsigned lastUnitNoOfBytes = ( offset + noOfBytes ) & unitSizeMask;
    if ( lastUnitNoOfBytes == 0 ) lastUnitNoOfBytes = unitSize();
    
    /*
     * see if the existing view can be used
     */

    unsigned baseUnitOfFirstWindow = from & ~_windowSizeRatioMask; 

    /*
     * compute first and last unit (offset) of first and last Windows
     */

    unsigned firstUnitOffsetOfFirstWindow = 
    from - baseUnitOfFirstWindow;
                         
    unsigned lastUnitOffsetOfLastWindow =
    ( firstUnitOffsetOfFirstWindow + to - from ) &
    _windowSizeRatioMask;

    unsigned baseUnitOfLastWindow = to - lastUnitOffsetOfLastWindow;

    unsigned firstUnitOfWindow;
    unsigned lastUnitOfWindow;
    unsigned noOfUnitsInWindow;

#ifdef USESTACK
    MOCacheReadAggregateState stackState;
#endif
    MOCacheReadAggregateState * state;
#ifdef USESTACK
    if ( c == 0 ) {
        state = &stackState;
    } else {
        state = new MOCacheReadAggregateState();
    }
#else
    state = new MOCacheReadAggregateState();
#endif
    state->init( c, this, _log2UnitSize, buf, firstUnitOffset,
                 lastUnitNoOfBytes );

    for ( unsigned baseUnitOfWindow = baseUnitOfFirstWindow;
          baseUnitOfWindow <= baseUnitOfLastWindow;
          baseUnitOfWindow += _windowSizeRatio ) {

        /*
         * compute the number of units needed from this Window
         */

        if ( baseUnitOfFirstWindow == baseUnitOfLastWindow ) {
            noOfUnitsInWindow = 
            lastUnitOffsetOfLastWindow -
            firstUnitOffsetOfFirstWindow + 1;
            firstUnitOfWindow = 
            firstUnitOffsetOfFirstWindow +
            baseUnitOfFirstWindow;  
        } else if ( baseUnitOfFirstWindow == baseUnitOfWindow ) {
            noOfUnitsInWindow = 
            _windowSizeRatio - 
            firstUnitOffsetOfFirstWindow;
            firstUnitOfWindow = 
            firstUnitOffsetOfFirstWindow +
            baseUnitOfFirstWindow;  
        } else if ( baseUnitOfLastWindow == baseUnitOfWindow ) {
            noOfUnitsInWindow =
            lastUnitOffsetOfLastWindow + 1;
            firstUnitOfWindow = baseUnitOfWindow;
        } else {
            noOfUnitsInWindow = _windowSizeRatio;
            firstUnitOfWindow = baseUnitOfWindow;
        }
        lastUnitOfWindow = 
        firstUnitOfWindow + noOfUnitsInWindow - 1;
        
        IOWindowInfo * winInfo = _cacheWindow =
        TheIOWindowCache->acquireIOWindow
        ( this, baseUnitOfWindow, _cacheWindow );

        unsigned unit = firstUnitOfWindow;
        PhysicallyAddressableUnit * pauList;
        PhysicallyAddressableUnit * firstPauList = 0;
        PhysicallyAddressableUnit * fromStorePauList;

        while ( unit <= lastUnitOfWindow ) {
            unsigned firstUnit;
            unsigned lastUnit;
            unsigned found = _map->findNextNonExist
            ( unit, lastUnitOfWindow, firstUnit, lastUnit, 
              pauList );
            if ( unit == firstUnitOfWindow )
                firstPauList = pauList;
            if ( !found )
                break;      
            unit = lastUnit + 1;
    
            fromStorePauList = 
            unitsFromStore( firstUnit, lastUnit );
            if ( firstUnit == firstUnitOfWindow ) 
                firstPauList = fromStorePauList;
    
            char * addr = ( char * ) winInfo->baseAddr() + 
            ( (firstUnit - baseUnitOfWindow ) <<
              log2UnitSize() );
            unsigned noOfUnits = unit - firstUnit;
#ifdef VIRTUALADDRESSIO
            translatable( firstUnit, lastUnit, NoAccess,
                          _translation, addr,
                          winInfo->view(), winInfo->baseAddr(),
                          fromStorePauList );
#endif // VIRTUALADDRESSIO
    
            state->add( state, winInfo,
                        firstUnit, noOfUnits, addr, 
                        fromStorePauList, this );

        }
        state->addCopyInfo( winInfo, firstUnitOfWindow, 
                            lastUnitOfWindow, firstPauList );
    }

    MOCacheReadClosure * op = state->_closures;
    int opCount = state->_count;
    if ( opCount > 0 ) {
        for ( int i = 0; i < opCount; i++, op++ ) {
            Raids( RaidDetail ) << this << "::cacheRead() "
                                << "issuing async _cache->read( " 
                                << op->_startUnit
                                << ", " << op->_count 
                                << ", " << asAddress( op->_addr )
                                << ", " << asAddress( op ) << " )\n" << eor;
 
            _cache->read( op->_startUnit, op->_count, op->_addr, 
                          op );
        }
        if ( c == 0 ) {
            state->_barrier.wait();
            state->doCopy();
            totalCount = state->_totalCount;
        } else {
            totalCount = noOfBytes;
        }
    } else {
        state->doCopy();
        totalCount = state->_totalCount;
        if ( c != 0 ) {
            c->activate( totalCount );
            delete state;
        }
    }
#ifdef USESTACK
#else
    if ( c == 0 ) 
        delete state;
#endif

    Raids( RaidMethod ) << this << "::cacheRead returns " 
                        << totalCount << "\n" << eor;
    SInstr( MemoryObjectCacheCacheRead.stop( istart ) );
    return totalCount;
}

class MOCacheWriteAggregateState;

class MOCacheWriteClosure : public IOAsyncClosure {
protected:
    MOCacheWriteAggregateState * _state;
    MemoryObjectCache * _cache;

public:
    unsigned _startUnit;
    unsigned _count;
    char * _addr;
    
    static void func( MOCacheWriteClosure * c );
    MOCacheWriteClosure() { }
    void init( MOCacheWriteAggregateState * state,
               unsigned startUnit, unsigned count,
               char * addr, MemoryObjectCache * cache )
    { _state = state; _startUnit = startUnit, 
                      _count = count; _addr = addr; _cache = cache;
//IMPORTANT CODE CHANGE BY SCOTT KIRCHER
                      //For some idiotic reason, _func was not being assinged to func,
                      //this was causing the default doNothing func to be called when
                      //this closure was activated, so any IO operation that was supposed
                      //to be finished up by this closure would FAIL. -SK
                      _func=(AsyncFunc)&func; //the fix is so simple, but so HARD to find!! -SK
    }
};

class MOCacheWriteAggregateState {
public:
    struct CopyInfo {
        IOWindowInfo * _winInfo;
        unsigned _firstUnitOfWindow;
        unsigned _lastUnitOfWindow;
        PhysicallyAddressableUnit * _pauList;
    };

    static Lock _lock;
    static const unsigned _prealloc = 8;
    unsigned _count;
    unsigned _completed;
    unsigned _log2UnitSize;
    unsigned _firstUnitOffset;
    unsigned _lastUnitNoOfBytes;
    unsigned _totalCount;
    MemoryObjectCache * _cache;
    IOAsyncClosure * _parent;
    MOCacheWriteClosure * _closures;
    MOCacheWriteClosure _preallocatedClosures[_prealloc];
    Barrier _barrier;
    unsigned _max;

    static const unsigned _copyPrealloc = 4;
    unsigned _copyCount;
    CopyInfo _copyInfo[_copyPrealloc];
    char * _buf;

    
    MOCacheWriteAggregateState() : _barrier( 1 ), _max( _prealloc ) {}
    ~MOCacheWriteAggregateState() 
    { if ( _max > _prealloc ) delete _closures; }
    void init( IOAsyncClosure * parent, MemoryObjectCache * cache,
               unsigned log2UnitSize, char * buf,
               unsigned firstUnitOffset, unsigned lastUnitNoOfBytes )
    { _parent = parent; _count = 0; _completed = 0; 
    _closures = _preallocatedClosures; _copyCount = 0; _cache = cache; 
    _log2UnitSize = log2UnitSize; _buf = buf; 
    _firstUnitOffset = firstUnitOffset; 
    _lastUnitNoOfBytes = lastUnitNoOfBytes; }
    void add( MOCacheWriteAggregateState * state,
              unsigned startUnit, unsigned count,
              char * addr, MemoryObjectCache * cache );
    void addCopyInfo( IOWindowInfo * winInfo, unsigned firstUnitOfWindow,
                      unsigned lastUnitOfWindow, 
                      PhysicallyAddressableUnit * pau );
    void doCopy();
};

Lock MOCacheWriteAggregateState::_lock;

inline 
void
MOCacheWriteAggregateState::add( 
    MOCacheWriteAggregateState * state,
    unsigned startUnit, unsigned count,
    char * addr, MemoryObjectCache * cache )
{
    if ( _count == _max ) {
        int newMax = _max << 1;
        MOCacheWriteClosure * newClosures = 
        new MOCacheWriteClosure[newMax];
        for( unsigned i = 0; i < _max; i++ ) {
            newClosures[i] = _closures[i];
        }
        if ( _closures != _preallocatedClosures )
            delete _closures;
        _closures = newClosures;
        _max = newMax;
    }
    Assert( _count < _max );

    Console() 
    << "MOCacheWriteAggregateState[" << asAddress(this) 
    << "]::add( " << startUnit << ", " << count << ", "
    << asAddress(addr) << ", " << cache << " ) count = " 
    << _count << "\n" << eor;

    _closures[_count].init( state, startUnit, count, addr, cache );
    _count++;
}

inline
void 
MOCacheWriteAggregateState::addCopyInfo( 
    IOWindowInfo * winInfo, unsigned firstUnitOfWindow,
    unsigned lastUnitOfWindow, PhysicallyAddressableUnit * pau )
{
    Assert( _copyCount < _copyPrealloc );
    CopyInfo * ci = _copyInfo + _copyCount;
    _copyCount++;
    ci->_winInfo = winInfo;
    ci->_firstUnitOfWindow = firstUnitOfWindow;
    ci->_lastUnitOfWindow = lastUnitOfWindow;
    ci->_pauList = pau;
}

void
MOCacheWriteAggregateState::doCopy()
{
    
    CopyInfo * copyInfo = _copyInfo;
    char * buf = _buf;
    _totalCount = 0;

    for ( unsigned i = 0; i < _copyCount; i++, copyInfo++ ) {

        /* 
         * figure out how much to copy 
         */

        IOWindowInfo * winInfo = copyInfo->_winInfo;
        unsigned baseUnitOfWindow = winInfo->from();
        unsigned firstUnitOfWindow = copyInfo->_firstUnitOfWindow;
        unsigned lastUnitOfWindow = copyInfo->_lastUnitOfWindow;

        unsigned startOffset = 
        ( ( firstUnitOfWindow - baseUnitOfWindow ) 
          << _log2UnitSize );
        unsigned count =
        ( ( lastUnitOfWindow - firstUnitOfWindow + 1 )
          << _log2UnitSize );
        if ( i == 0 ) {
            count -= _firstUnitOffset;
            startOffset += _firstUnitOffset;
        }
        if ( i+1 == _copyCount )
            count -= ( 1 << _log2UnitSize ) - _lastUnitNoOfBytes;

        char * startAddr = ( char * ) winInfo->baseAddr() + 
        ( ( firstUnitOfWindow - baseUnitOfWindow )
          << _log2UnitSize );
        _cache->translatable( firstUnitOfWindow, lastUnitOfWindow, 
                              NoAccess, 
                              MemoryObjectCache::_translation, 
                              startAddr,
                              winInfo->view(), winInfo->baseAddr(),
                              copyInfo->_pauList );
        ByteCopy( buf, winInfo->baseAddr() + startOffset, count );
        buf += count;   
        _totalCount += count;
        TheIOWindowCache->releaseIOWindow( winInfo );
    }
    if ( _cache->_replacementPolicy )
        _cache->_replacementPolicy->postCacheWrite
        ( _copyInfo[0]._firstUnitOfWindow,
          _copyInfo[_copyCount - 1]._lastUnitOfWindow );

}

void
MOCacheWriteClosure::func( MOCacheWriteClosure * c )
{
    Assert( c->_result == c->_count );


    Console() 
    << "MOCacheWrite::func( " << asAddress( c ) 
    << " ) aggregateState " << asAddress( c->_state ) 
    //<< " firstUnit " << firstUnit << " count " << c->_count 
    << "\n" << eor;

        
/*
  MemoryObjectCache * cache = c->_cache;
  if ( cache->_replacementPolicy ) {
  unsigned firstUnit = c->_startUnit;
  unsigned lastUnit = firstUnit + c->_count - 1;
  cache->_replacementPolicy->postPageIn( firstUnit, lastUnit );
  }
*/

    MOCacheWriteAggregateState * s = c->_state;
    Assert( s != 0 );
    if ( s ) {
        MOCacheWriteAggregateState::_lock.acquire();
        s->_completed++;
        unsigned completed = s->_completed;
        MOCacheWriteAggregateState::_lock.release();
/*
  Console() 
  << "MOCacheWrite::func( " << asAddress( c ) 
  << " ) aggregateState " << asAddress( c->_state ) 
  << "  completed " << completed << " count " 
  << s->_count << "\n" << eor;
*/
        Assert( completed <= s->_count );
        if ( completed == s->_count ) {
            if ( s->_parent ) {
                s->doCopy();
                s->_parent->activate( s->_totalCount );
                delete s;
            } else {
                s->_barrier.done();
            }
        }
    }
}

MemoryObjectCacheTimer
MemoryObjectCacheCacheWrite("MemoryObjectCache::cacheWrite() timer");

unsigned
MemoryObjectCache::cacheWrite( unsigned offset, unsigned noOfBytes, char * buf,
                               IOAsyncClosure * c )
{
    SInstr( int istart = MemoryObjectCacheCacheWrite.start() );
    Raids( RaidMethod ) << this << "::cacheWrite"
                        << "( offset " << offset 
                        << ", noOfBytes " << noOfBytes
                        << ", buffer " << asAddress( buf )
                        << ", closure " << asAddress( c )
                        << " )\n" << eor;

    unsigned totalCount = 0;
    bytesAccessedAdd( noOfBytes );

    /* 
     * compute unit numbers
     */
    unsigned from = offsetToUnit( offset );
    unsigned to   = offsetToUnit( offset + noOfBytes - 1 );

    Debug<<"argg!\n";
    Debug << this << "::cacheWrite"
          << "( offset " << offset 
          << ", noOfBytes " << noOfBytes
          << ", buffer " << asAddress( buf )
          << ", closure " << asAddress( c )
          << " )\n" << eor;
    Debug<<"_noOfUnits "<<_numberOfUnits<<"\n"<<eor;

    Assert( isIn( to ) && isIn( from ) );

    Debug<<"@@@@1@@@@\n"<<eor;

    if ( _replacementPolicy )
        _replacementPolicy->preCacheWrite( from, to );

    /*
     * compute number of bytes to copy from first and last
     * unit
     */

    Debug<<"@@@@2@@@@\n"<<eor;

    unsigned unitSizeMask = unitSize() - 1;
    unsigned firstUnitOffset = offset & unitSizeMask;
    unsigned firstUnitNoOfBytes = unitSize() - firstUnitOffset;
    unsigned lastUnitNoOfBytes = ( offset + noOfBytes ) & unitSizeMask;
    if ( lastUnitNoOfBytes == 0 ) lastUnitNoOfBytes = unitSize();
    
    /*
     * see if the existing view can be used
     */
    Debug<<"@@@@3@@@@\n"<<eor;

    unsigned baseUnitOfFirstWindow = from & ~_windowSizeRatioMask; 
        
    /*
     * compute first and last unit (offset) of first and last Windows
     */

    unsigned firstUnitOffsetOfFirstWindow = 
    from - baseUnitOfFirstWindow;
                         
    unsigned lastUnitOffsetOfLastWindow =
    ( firstUnitOffsetOfFirstWindow + to - from ) &
    _windowSizeRatioMask;
    unsigned baseUnitOfLastWindow = to - lastUnitOffsetOfLastWindow;

    unsigned firstUnitOfWindow;
    unsigned lastUnitOfWindow;
    unsigned noOfUnitsInWindow;

#ifdef USESTACK
    MOCacheWriteAggregateState stackState;
#endif
    MOCacheWriteAggregateState * state;
#ifdef USESTACK
    if ( c == 0 ) {
        state = &stackState;
    } else {
        state = new MOCacheWriteAggregateState();
    }
#else
    state = new MOCacheWriteAggregateState();
#endif
    Debug<<"@@@@4@@@@\n"<<eor;

    state->init( c, this, _log2UnitSize, buf, firstUnitOffset,
                 lastUnitNoOfBytes );
    Debug<<"@@@@5@@@@\n"<<eor;

    for ( unsigned baseUnitOfWindow = baseUnitOfFirstWindow;
          baseUnitOfWindow <= baseUnitOfLastWindow;
          baseUnitOfWindow += _windowSizeRatio ) {

        /*
         * compute the number of units needed from this Window
         */

        if ( baseUnitOfFirstWindow == baseUnitOfLastWindow ) {
            noOfUnitsInWindow = 
            lastUnitOffsetOfLastWindow -
            firstUnitOffsetOfFirstWindow + 1;
            firstUnitOfWindow = 
            firstUnitOffsetOfFirstWindow +
            baseUnitOfFirstWindow;  
        } else if ( baseUnitOfFirstWindow == baseUnitOfWindow ) {
            noOfUnitsInWindow = 
            _windowSizeRatio - 
            firstUnitOffsetOfFirstWindow;
            firstUnitOfWindow = 
            firstUnitOffsetOfFirstWindow +
            baseUnitOfFirstWindow;  
        } else if ( baseUnitOfLastWindow == baseUnitOfWindow ) {
            noOfUnitsInWindow =
            lastUnitOffsetOfLastWindow + 1;
            firstUnitOfWindow = baseUnitOfWindow;
        } else {
            noOfUnitsInWindow = _windowSizeRatio;
            firstUnitOfWindow = baseUnitOfWindow;
        }
        lastUnitOfWindow = 
        firstUnitOfWindow + noOfUnitsInWindow - 1;
        
        IOWindowInfo * winInfo = _cacheWindow =
        TheIOWindowCache->acquireIOWindow
        ( this, baseUnitOfWindow, _cacheWindow );

        /*
         * go through the units of this Window and 
         * make sure the required units are in core
         */

        unsigned unit = firstUnitOfWindow;
        PhysicallyAddressableUnit * pauList;
        PhysicallyAddressableUnit * fromStorePauList = 0;
        PhysicallyAddressableUnit * firstPauList = 0;
        while ( unit <= lastUnitOfWindow ) {
            unsigned firstUnit;
            unsigned lastUnit;
            unsigned found = _map->findNextNonExist
            ( unit, lastUnitOfWindow, firstUnit, lastUnit, 
              pauList );
            if ( unit == firstUnitOfWindow )
                firstPauList = pauList;
            if ( !found )
                break;      
            unit = lastUnit + 1;

            /*
             * cache these units
             *
             *   release window because page daemon
             *   may be activated, and the pager may
             *   need view for IO
             */

            fromStorePauList = 
            unitsFromStore( firstUnit, lastUnit );
            if ( firstUnit == firstUnitOfWindow )
                firstPauList = fromStorePauList;

            setReferencedModified( fromStorePauList, 
                                   lastUnit - firstUnit + 1, 0, 0 );

            /*
             * figure out if any reads are required,
             * reads are required only if the entire 
             * unit is not overwritten, reads can
             * be required only for the first and
             * last block 
             */

            int startRead = -1;
            int endRead = -1;

            if ( firstUnit == from && 
                 firstUnitNoOfBytes < unitSize() ) {
                /* 
                 * first unit needs to be read,
                 * first unit will not be completely 
                 * overwritten
                 */
                startRead = from;
            }
            if ( lastUnit == to &&
                 lastUnitNoOfBytes < unitSize() ) {
                /* 
                 * last unit needs to be read,
                 * last unit will not be completely
                 * overwritten
                 */
                endRead = to;
            }

            /* 
             * now figure out how to do read, adjacent
             * reads should be optimized and handle
             * case where firstUnit == lastUnit
             */

            if ( startRead == -1 && endRead == -1 ) {

                /*
                 * do nothing:
                 * simple optimization for case
                 * where no reads are required 
                 * skip the following checks
                 */

            } else if ( startRead != -1 && endRead != -1 &&
                        endRead - startRead <= 1 ) {

                /* 
                 * both first and last units must be read
                 * and they are either the same unit or
                 * or adjacent, this check allows 
                 * the adjacent units to handled in
                 * a single IO
                 */

                char * addr = ( char * ) winInfo->baseAddr() + 
                ( (startRead - baseUnitOfWindow ) <<
                  log2UnitSize() );
                Raids( RaidDetail ) << this 
                                    << "::cacheWrite paging in "
                                    << startRead << " to " << endRead
                                    << ", addr " 
                                    << asHex( ( unsigned ) addr )
                                    << "\n" << eor;
                unsigned noOfUnits = endRead - startRead + 1;
#ifdef VIRTUALADDRESSIO
                translatable( startRead, endRead, NoAccess, 
                              _translation, addr,
                              winInfo->view(), 
                              winInfo->baseAddr(),
                              fromStorePauList );
#endif // VIRTUALADDRESSIO
                state->add( state, startRead, noOfUnits, addr, 
                            this );

            } else {

                /* 
                 * some reads are required and the
                 * the reads will not be to adjacent units
                 */

                if ( startRead != -1 ) {
                    /* 
                     * need to read first unit 
                     */
                    Assert( startRead >= 0 );
                    char * addr = 
                    ( char * ) winInfo->baseAddr() + 
                    ( (startRead - baseUnitOfWindow ) <<
                      log2UnitSize() );
                    Raids( RaidDetail ) << this 
                                        << "::cacheWrite paging in "
                                        << startRead 
                                        << ", addr " 
                                        << asHex( ( unsigned ) addr )
                                        << "\n" << eor;
#ifdef VIRTUALADDRESSIO
                    translatable( startRead, startRead, 
                                  NoAccess, 
                                  _translation, addr,
                                  winInfo->view(), 
                                  winInfo->baseAddr(),
                                  fromStorePauList );
#endif // VIRTUALADDRESSIO
                    state->add( state, startRead, 1, addr,
                                this );

                }
                if ( endRead != -1 ) {
                    Assert( endRead >= 0 );
                    char * addr = 
                    ( char * ) winInfo->baseAddr() + 
                    ( (endRead - baseUnitOfWindow ) <<
                      log2UnitSize() );
                    Raids( RaidDetail ) << this 
                                        << "::cacheWrite paging in "
                                        << endRead
                                        << ", addr " 
                                        << asHex( ( unsigned ) addr )
                                        << "\n" << eor;
#ifdef VIRTUALADDRESSIO
                    PhysicallyAddressableUnit * endReadPau
                    = forward( fromStorePauList,
                               lastUnit-firstUnit );
                    translatable( endRead, endRead, 
                                  NoAccess, 
                                  _translation, addr,
                                  winInfo->view(), 
                                  winInfo->baseAddr(),
                                  endReadPau );
#endif // VIRTUALADDRESSIO
                    state->add( state, endRead, 1, addr,
                                this );
                }
            }
            if ( _replacementPolicy ) { 
#ifdef DEBUG_GLRU
                RaidStream() 
                << this << "::cacheWrite rp " 
                << asAddress( _replacementPolicy )
                << "\n" << eor;
#endif // DEBUG_GLRU
                _replacementPolicy->postPageIn( firstUnit, 
                                                lastUnit );
            }
        }
        state->addCopyInfo( winInfo, firstUnitOfWindow, 
                            lastUnitOfWindow, firstPauList );
    }
    Debug<<"@@@@6@@@@\n"<<eor;

    MOCacheWriteClosure * op = state->_closures;
    int opCount = state->_count;
    if ( opCount > 0 ) {
        Debug<<"@@@@6a@@@@\n"<<eor;
        Debug<<"_cache = "<<_cache<<"\n"<<eor;
        Debug<<"op = "<<asHex(op)<<"\n"<<eor;
        for ( int i = 0; i < opCount; i++, op++ ) {
            _cache->read( op->_startUnit, op->_count, op->_addr, 
                          op );
        }
        Debug<<"@@@@6a-I@@@@\n"<<eor;
        if ( c == 0 ) {
            Debug<<"@@@@6a-II@@@@\n"<<eor;
            state->_barrier.wait();
            Debug<<"@@@@6a-III@@@@\n"<<eor;
            state->doCopy();
            Debug<<"@@@@6a-IV@@@@\n"<<eor;
            totalCount = state->_totalCount;
        } else {
            Debug<<"@@@@6a-V@@@@\n"<<eor;
            totalCount = noOfBytes;
        }
    } else {
        Debug<<"@@@@6b@@@@\n"<<eor;

        state->doCopy();
        totalCount = state->_totalCount;
        if ( c != 0 ) {
            c->activate( totalCount );
            delete state;
        }
    }
    Debug<<"@@@@7@@@@\n"<<eor;

#ifdef USESTACK
#else
    delete state;
#endif
    Raids( RaidMethod ) << this << "::cacheWrite returns " 
                        << totalCount << "\n" << eor;
    SInstr( MemoryObjectCacheCacheWrite.stop( istart ) );
    Debug<<"@@@@8@@@@\n"<<eor;
    return totalCount;
}

MemoryObjectCacheTimer
MemoryObjectCacheFlush("MemoryObjectCache::flush() timer");

void
MemoryObjectCache::flush()
{
    SInstr( int istart = MemoryObjectCacheFlush.start() );
    Raids( RaidMethod ) << this << "::flush()\n" << eor;

    if ( _numberOfUnits != 0 ) {
        untranslatable( 0, _numberOfUnits - 1 );
        saveModified( ToMemoryObject, 0, _numberOfUnits - 1, 
                      SAVEMODIFIED_FLUSH, 0 );
        Assert( _cache );
        // _cache->invalidate( 0, _numberOfUnits, 1 );
        unitsToStore( 0, _numberOfUnits - 1, 0 );
    }
    SInstr( MemoryObjectCacheFlush.stop( istart ) );
}

MemoryObjectCacheTimer
MemoryObjectCacheFlushFromTo("MemoryObjectCache::flush(from,to) timer");

void
MemoryObjectCache::flush( unsigned from, unsigned to )
{
    SInstr( int istart = MemoryObjectCacheFlushFromTo.start() );
    Raids( RaidMethod ) << this << "::flush( " 
                        << from << ", " << to << " )\n" << eor;

    Assert( from <= to && to < _numberOfUnits );
    
    untranslatable( from, to );
    saveModified( ToMemoryObject, from, to, SAVEMODIFIED_FLUSH, 0 );
    Assert( _cache );
    // _cache->invalidate( from, to - from + 1, 1 );
    unitsToStore( from, to, 0 );
    SInstr( MemoryObjectCacheFlushFromTo.stop( istart ) );
}

int MemoryObjectCacheSynchronizeDoFlush = 0;

void
MemoryObjectCache::synchronize( IOAsyncClosure * c )
{
    Raids( RaidMethod ) << this << "::synchronize( "
                        << asAddress( c ) << "\n" << eor;
    
    if ( _numberOfUnits == 0 ) {
        if ( c != 0 )
            c->activate();
    } else {
        untranslatable( 0, _numberOfUnits - 1 );
        saveModified( ToMemoryObject, 0, _numberOfUnits - 1, 0, c );
        if ( c == 0 && MemoryObjectCacheSynchronizeDoFlush ) {
            unitsToStore( 0, _numberOfUnits - 1, 0 );
        }
    }
}

void
MemoryObjectCache::synchronize( unsigned from, unsigned to, IOAsyncClosure * c )
{
    Raids( RaidMethod ) << this << "::synchronize( " 
                        << from << ", " << to << " )\n" << eor;

    Assert( from <= to && to < _numberOfUnits );
    
    untranslatable( from, to );
    saveModified( ToMemoryObject, from, to, 0, c );
}

MemoryObjectCacheTimer
MemoryObjectCachePseudoSynchronize("MemoryObjectCache::pseudoSynchronize() timer");

void
MemoryObjectCache::pseudoSynchronize()
{
    SInstr( int istart = MemoryObjectCachePseudoSynchronize.start() );
    Raids( RaidMethod ) << this << "::pseudoSynchronize()\n" << eor;
    
    unsigned start;
    unsigned end;

    start = end = 0;
    while ( end < _numberOfUnits ) {
        while ( end < _numberOfUnits && 
                !_map->exists( end ) && !_cache->exists( end ) )
            end++;
        start = end;
        while ( end < _numberOfUnits && 
                ( _map->exists( end ) || _cache->exists( end ) ) )
            end++;
        if ( start < _numberOfUnits )
            memoryObjectAlloc( start, end - 1 );
    }
    SInstr( MemoryObjectCachePseudoSynchronize.stop( istart ) );
}

int MemoryObjectCache::hasCache( Class * cacheClass )
{
    Cache * cache = _cache;
    while ( cache != 0 ) {
        if ( cache->classOf() == cacheClass ) {
            return 1;
        }
        cache = cache->child();
    }
    return 0;
}

#include "ClassInstaller.h"
DEFINECLASS("MemoryObjectCache",MemoryObjectCache,MemoryRange,0,0);
