#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * PagedMemoryObjectCache.c
 */

#include "AddressTranslation.h"
#include "Assert.h"
#include "Class.h"
#include "DummyCache.h"
#include "IOWindow.h"
#include "Kernel.h"
#include "List.h"
#include "LinkIterator.h"
#include "Map.h"
#include "MemoryObjectView.h"
#include "MemorySimpleInstruments.h"
#include "PagedMemoryObjectCache.h"
#include "PersistentMemoryObject.h"
#include "Raid.h"
#include "RangeLocks.h"
#include "ReplacementPolicy.h"
#include "Store.h"
#include "ThisProcess.h"
#include "TranslationMap.h"
#include "UNIXInodeCache.h"

//#define SBLTEST

#ifdef __GNUC__
#pragma implementation
#endif

// #define MAXRESIDENTUNITS 4
// #define ADDCACHECLASS UNIXInodeCacheClass

// #define DEFAULTREPLACEMENTPOLICYCLASS SimpleReplacementPolicyClass

#ifndef DEFAULTREPLACEMENTPOLICYCLASS
// #define DEFAULTREPLACEMENTPOLICYCLASS SmarterGlobalLRUReplacementPolicyClass
#define DEFAULTREPLACEMENTPOLICYCLASS HintedGlobalLRUReplacementPolicyClass
extern ClassStar DEFAULTREPLACEMENTPOLICYCLASS;
#endif

extern Store * MainStore;

/*
 * Constructor and Destructor
 */

MemoryObjectCacheTimer
PagedMemoryObjectCacheCtor("PagedMemoryObjectCache::PagedMemoryObjectCache() timer");

PagedMemoryObjectCache::PagedMemoryObjectCache( MemoryObject * memoryObject )
     : MemoryObjectCache( memoryObject )
{
	SInstr( int istart = PagedMemoryObjectCacheCtor.start() );
	Assert( this != 0 );
	initDebug();

	Raids( RaidConstructor ) << this << "::PagedMemoryObjectCache( " 
		<< memoryObject << " )\n" << eor;

	Assert( memoryObject->isKindOf( PersistentMemoryObjectClass ) );

	_map = new MultiLevelTableMap( _numberOfUnits, _log2UnitSize );
	_adtMap = new LinkedListTranslationMap( _map );
	Assert( _map && _adtMap );

	_cache = new DummyCache( this );
	Assert( _cache != 0 );

	Class * policyClass = _memoryObject->replacementPolicyClass();
	if ( policyClass == 0 ) {
		policyClass = DEFAULTREPLACEMENTPOLICYCLASS;
	}
#ifdef ADDCACHECLASS
	addCache( ADDCACHECLASS );
#endif
	_replacementPolicy = 
		( ReplacementPolicy * ) policyClass->constructor( this );
#ifdef MAXRESIDENTUNITS
	_replacementPolicy->setMaxResidentUnits( MAXRESIDENTUNITS );
#endif
	Assert( _replacementPolicy != 0 );
	SInstr( PagedMemoryObjectCacheCtor.stop( istart ) );
}

/*
 * Free data structures and paus. Assume that the MemoryObject has
 * called appropriate methods and there are no references anymore to
 * any pages that were in here. Also assume that dirty pages do not
 * matter.
 */

MemoryObjectCacheTimer
PagedMemoryObjectCacheDtor("PagedMemoryObjectCache::~PagedMemoryObjectCache() timer");

PagedMemoryObjectCache::~PagedMemoryObjectCache()
{
	SInstr( int istart = PagedMemoryObjectCacheDtor.start() );

	Raids( RaidConstructor ) << this << "::~PagedMemoryObjectCache() "
		<< _map->residentSize() << " units resident from " 
		<< _memoryObject << "\n" << eor;
		
	Assert( this != 0 );

	cleanUp();

	delete _replacementPolicy;
	delete _map;
	delete _adtMap;
	delete _cache;
	SInstr( PagedMemoryObjectCacheDtor.stop( istart ) );
}

/* 
 * Change of cache size
 */

MemoryObjectCacheTimer
PagedMemoryObjectCacheSetNumberOfUnits("PagedMemoryObjectCache::setNumberOfUnits() timer");

int
PagedMemoryObjectCache::setNumberOfUnits( unsigned size )
{
	SInstr( int istart = PagedMemoryObjectCacheSetNumberOfUnits.start() );
	Raids( RaidMethod ) << this << "::setNumberOfUnits( " 
		<< size << " )\n" << eor;

	if ( size == _numberOfUnits )
		return size;

	if ( size < _numberOfUnits) {
		untranslatable( size, _numberOfUnits - 1 );
		unitsToStore( size, _numberOfUnits - 1, 0 );
	}
	if ( size == 0 ) {
		fromCacheUnitsReset();
		bytesAccessedReset();
	}

	_replacementPolicy->setNumberOfUnits( size );
	_cache->setNumberOfUnits( size );
	_map->setSize( size );
	_numberOfUnits = size;

	unsigned newLog2WindowSize = computeLog2WindowSize();
	if ( newLog2WindowSize > _log2WindowSize ) {

		TheIOWindowCache->freeIOWindow( this, 1 );
		_log2WindowSize = newLog2WindowSize;
	       	int log2WindowSizeRatio = _log2WindowSize - log2UnitSize();
	       	_windowSizeRatio = 1 << log2WindowSizeRatio;
	       	_windowSizeRatioMask = _windowSizeRatio - 1;

	}
	 
	SInstr( PagedMemoryObjectCacheSetNumberOfUnits.stop( istart ) );
	return size;
}

void 
PagedMemoryObjectCache::setReplacementPolicyClass( Class * policyClass )
{
	Raids( RaidMethod ) 
		<< this << "::setReplacementPolicyClass( " 
		<< policyClass << " )\n" << eor;
	Assert( policyClass->isKindOf( ReplacementPolicyClass ) );
	if ( _replacementPolicy->classOf() != policyClass ) {
		ReplacementPolicy * newPolicy = 
			(ReplacementPolicy *) policyClass->constructor( this );
		Assert( newPolicy != 0 );
		delete _replacementPolicy;
		_replacementPolicy = newPolicy;
	}
}

extern ClassStar DummyCacheClass;

void 
PagedMemoryObjectCache::addCache( Class * cacheClass )
{
	Raids( RaidMethod ) 
		<< this << "::addCache( " << cacheClass << " )\n"
		<< eor;
	Assert( cacheClass == UNIXInodeCacheClass );
	if ( _cache->classOf() == UNIXInodeCacheClass )
		return;
	Assert( _cache->classOf() == DummyCacheClass );
	int  status = 0;
	Cache * newCache = new UNIXInodeCache( this, status );
	if ( status != 0 ) {
		delete newCache;
		return;
	}
	_cache->insertAbove( newCache );
	_cache = newCache;
	return;
}

/*
 * pageOut
 */

MemoryObjectCacheTimer
PagedMemoryObjectCachePageOut("PagedMemoryObjectCache::pageOut() timer");

unsigned
PagedMemoryObjectCache::pageOut( unsigned noOfPaus )
{
	SInstr( int istart = PagedMemoryObjectCachePageOut.start() );

/*
	Raids( RaidMethod ) 
		<< this << "::pageOut( " << noOfPaus << " paus ) mo "
		<< _memoryObject << "\n"
		<< eor;
 */
	int pauUnitSizeRatioMask = ( 1 << _log2PauUnitSizeRatio ) - 1;
	unsigned noOfUnits = ( noOfPaus + pauUnitSizeRatioMask ) >>
		             _log2PauUnitSizeRatio;
	Assert( noOfUnits > 0 );
	unsigned nnoOfPaus = _replacementPolicy->pageOut( noOfUnits ) <<
		            _log2PauUnitSizeRatio;

/*
	Raids( RaidMethod )
		<< this << "::pageOut() returns " << nnoOfPaus << " paus\n" 
		<< eor;
 */

	SInstr( PagedMemoryObjectCachePageOut.stop( istart ) );	
	return nnoOfPaus;
}



PhysicallyAddressableUnit *
PagedMemoryObjectCache::storeAlloc( unsigned units )
{	
	return MainStore->userAlloc( units << _log2PauUnitSizeRatio, this );
}

void
PagedMemoryObjectCache::storeDealloc( PhysicallyAddressableUnit * p, 
				      unsigned queued )
{
	MainStore->userDealloc( p, this, queued );
}

void
PagedMemoryObjectCache::unitsToStoreQueued( unsigned nunits )
{
	MainStore->userDeallocQueued( nunits << _log2PauUnitSizeRatio, this );
}

unsigned
PagedMemoryObjectCache::memoryObjectAlloc( unsigned from, unsigned to )
{
	unsigned startUnit = from << _log2MoUnitSizeRatio;
	unsigned count = ( to - from + 1 ) << _log2MoUnitSizeRatio;
	unsigned moNumberOfUnits = _memoryObject->numberOfUnits();
	if ( startUnit + count > moNumberOfUnits ) {
		count = moNumberOfUnits - startUnit;
		if ( count <= 0 )
			return 0;
	}
	int result = ( (PersistentMemoryObject *) 
		      _memoryObject )->alloc( startUnit, count );
	return result >> _log2MoUnitSizeRatio;
}

ProxiableObject *
PagedMemoryObjectCacheConstructor( ProxiableObject * source )
{
        if( !source->isKindOf( MemoryObjectClass ) ) 
		return 0;
	
	MemoryObject * mo = ( MemoryObject * ) source;
	return new PagedMemoryObjectCache( mo );
}


#include "ClassInstaller.h"
DEFINECLASS("PagedMemoryObjectCache",PagedMemoryObjectCache,MemoryObjectCache,&PagedMemoryObjectCacheConstructor,0);

