#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Contributing Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Contributing Author: Peter W. Madany (madany@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * PhysicalMemoryChain.c - class to create physical memory chains.
 */

#include "Assert.h"
#include "LockHolder.h"
#include "MemoryObjectView.h"
#include "OutputStream.h"
#include "PhysicalMemoryChain.h"
#include "PhysicallyAddressableUnit.h"


#ifdef __GNUC__
#pragma implementation
#endif

PhysicalMemoryChain::PhysicalMemoryChain()
     : _offset( 0 ), _length( 0 ), _lockHolder( 0 ), _chain( 0 )
{
}

PhysicalMemoryChain::PhysicalMemoryChain( LockHolder * lockHolder,
					  unsigned offset)
{
	if ( lockHolder ) {
		Assert( lockHolder->chain() != 0 );
		_lockHolder = lockHolder;
		_lockHolder->reference();
		_chain = _lockHolder->chain();
		_offset = offset;
		_length = _lockHolder->size() - offset;
	} else {
		_lockHolder = 0;
		_chain = 0;
		_offset = 0;
		_length = 0;
	}
}

PhysicalMemoryChain::PhysicalMemoryChain( PhysicallyAddressableUnit * start,
					  unsigned offset, unsigned length,
					  LockHolder * lockHolder )
{
	_chain = start;
	_lockHolder = lockHolder;
	if ( lockHolder )
		lockHolder->reference();
	_offset = offset;
	_length = length;
}

PhysicalMemoryChain::~PhysicalMemoryChain()
{
	if ( _lockHolder )
		_lockHolder->unreference();
}

PhysicalMemoryChain::PhysicalMemoryChain( const PhysicalMemoryChain & chain )
{
	_lockHolder = chain._lockHolder;
	if ( _lockHolder )
		_lockHolder->reference();
	_offset = chain._offset;
	_length = chain._length;
	_chain = chain._chain;
}

void
PhysicalMemoryChain::operator = ( const PhysicalMemoryChain & chain )
{
	if ( _lockHolder != chain._lockHolder ) {
		if ( _lockHolder )
			_lockHolder->unreference();
		_lockHolder = chain._lockHolder;
		if ( _lockHolder )
			_lockHolder->reference();
	}
	_offset = chain._offset;
	_length = chain._length;
	_chain = chain._chain;
}

PhysicalMemoryChain::PhysicalMemoryChain( PhysicalMemoryChain & chain,
	unsigned subRangeOffset, unsigned int subRangeLength )
{
	Assert( subRangeLength <= ( chain._length - subRangeOffset ) );

	/*
	 * We know the length already.
	 */
	_length = subRangeLength;

	/*
	 * Loop through all the units until we find the proper starting unit.
	 */
	_chain = chain._chain;
	while( subRangeOffset >= _chain->size() ) {
		subRangeOffset -= _chain->size();
		_chain = _chain->nextLink();
	}

	/*
	 * The offset left over is the proper offset into the first unit
	 * in the new chain.
	 */
	_offset = subRangeOffset;

	_lockHolder = chain._lockHolder;
	if ( _lockHolder )
		_lockHolder->reference();
}


const char *
PhysicalMemoryChainContiguousBlockIterator::next( unsigned int & length )
{
	if( amountLeft == 0 ) {
		length = 0;
		return( 0 );
	}

	// bias is only interesting on the first call to next()
	const char * retval = currentPau->address() + bias;

	length = currentPau->size() - bias;

	bias = 0;
 
	if( amountLeft <= length ) {
		length = amountLeft;
		amountLeft = 0;
		return( retval );
	}

	amountLeft -= length;

	PhysicallyAddressableUnit * lastPau = currentPau;
	currentPau = currentPau->nextLink();
	Assert( currentPau != 0 );

	while( amountLeft >= currentPau->size() ) {
		if( currentPau->address() !=
		    ( lastPau->address() + lastPau->size() ) ) {
			return( retval );
		}
		length += currentPau->size();
		amountLeft -= currentPau->size();
		lastPau = currentPau;
		currentPau = currentPau->nextLink();
		if( ( currentPau == 0 ) && ( amountLeft == 0 ) ) {
			return( retval );
		}
		Assert( currentPau != 0 );
	}

	if( currentPau->address() !=
	     ( lastPau->address() + lastPau->size() ) ) {
		return( retval );
	}

	length += amountLeft;
	amountLeft = 0;
	return( retval );
}

/*
 * Return a sub-chain of the chain being iterated (starting at the
 * pervious stopping point) of the given length.
 */
PhysicalMemoryChain *
PhysicalMemoryChainIterator::next( unsigned int increment )
{
	PhysicallyAddressableUnit * pau = currentChain->chain();
	unsigned offset = currentChain->offset();
	unsigned int length = currentChain->lengthInBytes();

	if ( increment >= length ) {
		PhysicalMemoryChain emptyChain;
		current = emptyChain;
		return( 0 );
	}

	offset += increment;
	length -= increment;
	Assert( pau != 0 );
	while( offset >= pau->size() ) {
		offset -= pau->size();
		pau = pau->nextLink();
		Assert( pau != 0 );
	}

       	PhysicalMemoryChain newChain(  pau, offset, length, 
				       initialChain->lockHolder() );
	current = newChain;
	currentChain = &current;

	return( currentChain );
}

