#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * PremappedMemoryObjectCache.c: PremappedMemoryObjectCache class
 *				 implementation.
 */

#include "Assert.h"
#include "ClassStar.h"
#include "LockHolder.h"
#include "Map.h"
#include "MemoryObjectView.h"
#include "MemoryObjectViewStar.h"
#include "OutputStream.h"
#include "PremappedMemoryObjectCache.h"
#include "PhysicallyAddressableUnit.h"
#include "PhysicalMemoryChain.h"
#include "PhysicalMemoryManager.h"
#include "Raid.h"
#include "Store.h"
#include "MemoryObject.h"


#ifdef __GNUC__
#pragma implementation
#endif

/*
 * You have to link all of the PAU's together so that PhysicalMemoryChains
 * can be built.
 */

PremappedMemoryObjectCache::PremappedMemoryObjectCache
	( MemoryObject * memobj, const char * baseAddress, int preAllocated )
        : SpecialPurposeCache( memobj ),
	  _premappedMap( _numberOfUnits, _log2UnitSize, baseAddress ),
	  _constantSizeTranslationMap( & _premappedMap )
{
	initDebug();
	Raids( RaidConstructor ) << this << "::PremappedMemoryObjectCache()\n" 
		<< eor;

	Assert( _log2PauUnitSizeRatio == 0 );
	Assert( this != 0 );
	if ( preAllocated ) {
		PhysicallyAddressableUnit * plist =
			MainStore->alloc( baseAddress, _numberOfUnits );
		for( unsigned i = 0; i < _numberOfUnits; i++ ) {
			Assert( plist && plist->owner() == 0 );
			plist->setOwner( & _premappedMap );
			plist = plist->nextLink();
		}
	}
	_preAllocated = preAllocated;
	_map = & _premappedMap;
	_adtMap = & _constantSizeTranslationMap;
}

void
PremappedMemoryObjectCache::noRemainingReferences()
{
	Raids( RaidConstructor ) << this << "::noRemainingReferences()\n" 
		<< eor;

	Assert( NOTREACHED );
}

/*
 * A subclass of MemoryObjectCache that pre-maps all its VM to one chunk of
 * physical memory. This is useful for mapping to memory mapped I/O
 * and the like.
 */

PremappedMemoryObjectCache::~PremappedMemoryObjectCache()
{
	Assert( NOTREACHED );
}

LockHolder *
PremappedMemoryObjectCache::cache( unsigned int unit,
				   ProtectionLevel requiredProtection,
				   ProtectionLevel & grantedProtection,
				   FaultErrorCode & code,
				   AddressTranslation * adt,
				   const char * vaddr, 
				   CacheReason reason,
				   MemoryObjectView * view,
				   const char * baseAddr )
{
	Raids( RaidMethod ) 
		<< this << "::cache( unit " << unit 
		<< ", requiredProtection " << requiredProtection 
		<< ", vaddr " << asHex( (unsigned) vaddr )
		<< ", reason " << reason
		<< ", view " << view
		<< ", baseAddr " << asHex( (unsigned) baseAddr )
		<< " )\n" << eor; 

	Assert( this != 0 );
	Assert( reason != DomainRepairFault );

	LockHolder * lockHolder = 0;
	int preLocked = ( view->classOf() == LockedMemoryObjectViewClass );

	if ( ! preLocked )
		preFault( unit, unit );

	Assert( isIn( unit ) );

	PhysicallyAddressableUnit * pau = _map->pauList( unit, unit );

	Assert( pau != 0 );
	Assert( _memoryObject != 0 );
	if ( _preAllocated )
		setProtection( pau, 1, 
			       _memoryObject->logicalProtectionLevel() );
	code = MissingAddressTranslationMapping;
	
#ifndef VCHOICES
	if ( reason != DomainConstructChain ) 
		translatable( unit, unit, requiredProtection, adt, vaddr,
			      view, baseAddr, pau );
#endif

	if ( reason & DomainConstructChain ) {

		lockHolder = new LockHolder( unit, unit, this, 
					     _map->pauList( unit, unit ),
					     !preLocked );

	}
	
	if ( !preLocked && lockHolder == 0 )
		postFault( unit, unit );

	Raids( RaidMethod ) << this << "::cache(" << unit << "," 
	      << requiredProtection << "," << code << "," << adt << ","
	      << asAddress(vaddr) << "," << reason << ")\n" << eor;

	return lockHolder;
}

LockHolder *
PremappedMemoryObjectCache::mcache(unsigned int unit, int nunits,
				   ProtectionLevel requiredProtection,
				   ProtectionLevel & grantedProtection,
				   FaultErrorCode & code,
				   AddressTranslation * adt,
				   const char * vaddr, CacheReason reason,
				   MemoryObjectView * view,
				   const char * baseAddr)
{
	unsigned from = unit; 
	unsigned to = from + nunits - 1;

	Raids( RaidMethod ) 
		<< this << "::mcache( from " << from
		<< ", to " << to
		<< ", requiredProtection " << requiredProtection 
		<< ", vaddr " << asHex( (unsigned) vaddr )
		<< ", reason " << reason
		<< ", view " << view
		<< ", baseAddr " << asHex( (unsigned) baseAddr )
		<< " )\n" << eor; 

	Assert( _memoryObject != 0 );
	LockHolder * lockHolder = 0;

	int preLocked = ( view->classOf() == LockedMemoryObjectViewClass );
	if ( ! preLocked )
		preFault( from, to );

	Assert( from <= to && to < _numberOfUnits );

	PhysicallyAddressableUnit * pauList = _map->pauList( from, to );
	if ( _preAllocated ) 
		setProtection( pauList, ( to - from + 1 ),
			       _memoryObject->logicalProtectionLevel() );

	code = MissingAddressTranslationMapping;

	Assert( requiredProtection <= 
	        _memoryObject->logicalProtectionLevel() );
	
#ifndef VCHOICES
	if ( reason != DomainConstructChain )
	 	translatable( from, to, requiredProtection, adt, vaddr,
			      view, baseAddr, pauList );
#endif

	if ( reason & DomainConstructChain ) {
		lockHolder = new LockHolder( from, to, this,
					     _map->pauList( from, to ),
					     !preLocked );
	}

	if ( ! preLocked && lockHolder == 0 )
		postFault( from, to );

	Raids( RaidMethod ) << this << "::mcache(" << unit << "," 
	      << nunits << "," << requiredProtection << "," << code 
	      << "," << adt << "," << asAddress(vaddr) << "," << reason 
	      << ")\n" << eor;
    
	return lockHolder;
}

#include "ClassInstaller.h"
DEFINECLASS("PremappedMemoryObjectCache",PremappedMemoryObjectCache,SpecialPurposeCache,0,0);
