#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * Store.c - Physical memory allocation.
 */

//#define SBLTEST

#include "Allocator.h"
#include "Assert.h"
#include "Kernel.h"
#include "MemoryObjectCachingContainer.h"
#include "OutputStream.h"
#include "PagedMemoryObjectCache.h"
#include "PhysicallyAddressableUnit.h"
#include "PhysicalMemoryManager.h"
#include "PhysicalMemoryRange.h"
#include "Process.h"
#include "PROCESSOR.h"
#include "Raid.h"
#include "Store.h"
#include "StoreInfo.h"
#include "ThisProcess.h"

#ifdef __GNUC__
#pragma implementation
#endif

#ifdef SBLTEST
#define NUMBEROFUNITS 2048
#endif

#define WARNING 32

#define DEBUGFREEPAGES

Lock Store::_lock;

/*
 * Should check that base is page aligned and round it up if not.
 */
class PhysicallyAddressableUnit;    // forward reference...

Store::Store( PhysicalMemoryRange * systemMemoryMap, char * )
    : _pagerSem( 0 )
{
    Raids( RaidConstructor )
    << this << "::Store( " << systemMemoryMap << " )\n" << eor;
    Assert( this != 0 );
    Assert( systemMemoryMap != 0 );

    _log2UnitSize = systemMemoryMap->log2UnitSize();
    _numberOfUnits = 0;
    _kernelUnits = 0;
    _userUnits = 0;
    _wantUnits = 0;
    _wantWaiting = 0;
    _queued = 0;

    _paging = 0;

    _storeInfoCount = 0;
    _storeInfoHead = new StoreInfo( 0 );
    _storeInfoHead->_type = STOREINFO_LISTHEAD;
    _storeInfoHead->_next = _storeInfoHead->_prev = _storeInfoHead;
    _current = _storeInfoHead;
    
    /*
     * Build the free list. This is rather inefficient as we know the
     * pages are contiguous, but it is only done at boot time.
     */
    PhysicallyAddressableUnit * prevPAU, * PAU;
    prevPAU = 0;
    PAU = 0;
    freeList = 0;
    PhysicalMemoryRange * region = systemMemoryMap;
    for( region = systemMemoryMap; region != 0; region = region->next() ) {
        Assert( _log2UnitSize == region->log2UnitSize() );
        Raids( RaidDetail )
        << this << "::Store(): adding " << region << "\n"
        << eor;
        _numberOfUnits += region->numberOfUnits();
        const char * base = region->physicalAddress();
        for( unsigned i = 0; i < region->numberOfUnits(); i++ ) {
            PAU = MainMemory->lookup( base );
            PAU->setOwner( 0 );
            if( freeList == 0 ) {
                freeList = PAU;
            } else {
                prevPAU->setNextLink( PAU );
            }
            prevPAU = PAU;
            base += region->unitSize();
        }
    }

#ifdef NUMBEROFUNITS
    _numberOfUnits = NUMBEROFUNITS;
#endif

    Assert( _numberOfUnits > 1024 );

/*
  _high = _numberOfUnits - 64;
  _low =  _numberOfUnits - 256;
*/
    _high = _numberOfUnits - _numberOfUnits / 32;
    _low =  _numberOfUnits - _numberOfUnits / 16;

    Console() << this << "::Store(): " << _numberOfUnits 
              << " total units\n" << eor;
}

/*
 * This should probably never get called.
 */
Store::~Store()
{
    Assert( NOTREACHED );
}

/*
 * Start daemon
 */

void 
Store::startPager()
{
    _pageOutDaemon = new PreemptableSystemProcess (  (OPFV)&Store::freePages, TheKernel->domain(), normal, (int) this );
    Assert( _pageOutDaemon != 0 );
    _pageOutDaemon->setName("#PageOutDaemon");
    _pageOutDaemon->ready();
    Debug << this << " Starting pager: " << _pageOutDaemon << "\n" << eor;
}

/*
 * Info
 */

void
Store::info( OutputStream & stream )
{
    stream << this << "::info()\n" 
           << "Total                  : " << _numberOfUnits << " units\n"
           << "Non-pageable           : " << _kernelUnits << " units\n"
           << "Pageable               : " << _userUnits << " units\n"
           << "Waiting to be allocated: " << _wantUnits << " units\n"
           << "Low limit              : " << _low << " units\n" 
           << "High limit             : " << _high << " units\n"
           << "Pager active now       : " << ( _paging ? "yes" : "no" ) << "\n"
           << "No of blocked processes: " << _wantWaiting << "\n" << eor;
}

/*
 * Allocate "nunits" free units and return them in a linked list.
 */
PhysicallyAddressableUnit *
Store::alloc( unsigned int nunits )
{
    Raids( RaidMethod )
    << this << "::alloc( nunits:" << nunits << " )\n" << eor;
    Assert( this != 0 );
    if( nunits == 0 ) {
        return( 0 );
    }

    /*
     * Make sure there are "nunits" on the free list.
     * For now, Panic if not.
     */
    Assert( ! _lock.heldByMe() );
    acquire();

    PhysicallyAddressableUnit * last = freeList;
    PhysicallyAddressableUnit * prev = 0;

    for( unsigned i = 0; i < nunits; i++ ) {
        if( last == 0 ) {
            Console().panic( "Store::alloc out of memory\n" );
        }
        Assert( last->owner() == 0 );
        prev = last;
        last = last->nextLink();
    }

    PhysicallyAddressableUnit * freeUnit = freeList;
    freeList = last;

    _kernelUnits += nunits;

#ifdef WARNING
    if ( totalAllocPaus() + WARNING > _numberOfUnits ) {
        Console() << this << "::alloc( " << nunits << " ) " 
                  << "less than " << WARNING << " units left\n" << eor;
        info( Console() );
    }
#endif

    release();

    Assert( prev != 0 );
    prev->setNextLink( 0 );
    return( freeUnit );
}

/*
 * Allocate "n" particular units starting at "startAddr".
 * This is REAL expensive, but only done at boot time.
 */
PhysicallyAddressableUnit *
Store::alloc( const char * startAddr, unsigned int nunits )
{
    Raids( RaidMethod ) << this << "::alloc( " << nunits << " units starting at "
                        << asAddress( startAddr ) << " )\n" << eor;
    Assert( this != 0 );
    if( nunits == 0 ) {
        return( 0 );
    }

    Assert( ! _lock.heldByMe() );
    acquire();

    PhysicallyAddressableUnit * prevPAU = 0;
    PhysicallyAddressableUnit * chain = 0;
    PhysicallyAddressableUnit * PAU = freeList;

    /*
     * Search the free list for the PAU at startAddr.
     */
    while( PAU != 0 ) {
        if( PAU->address() == startAddr ) {
            chain = PAU;
            break;
        } else {
            prevPAU = PAU;
            PAU = PAU->nextLink();
        }
    }

    if( chain == 0 ) {
        /*
         * Didn't find the units we're looking for, probably because
         * they've already been allocated.
         */
        Console() << this << "::alloc( " << nunits
                  << " units starting at " << asAddress( startAddr )
                  << " ): UNITS NOT FOUND\n" << eor;
        Console().panic( "Store::alloc(): UNITS NOT FOUND\n" );
    }

    /*
     * We found the first unit, so follow the list and make sure
     * that all the rest are there.  For now, we assume that all the
     * desired units are linked consecutively.  In general, 
     * this will not be the case, but at boot time nothing should
     * have been deallocated, so it's a reasonable assumption.
     */
    PhysicallyAddressableUnit * nextPAU = 0;
    PAU = chain;
    for( unsigned int u = 0; u < nunits; u++ ) {
        Assert( PAU );
        Assert( PAU->owner() == 0 );
        nextPAU = PAU->nextLink();
        PAU = nextPAU;
    }

    /*
     * Now clean up the freeList.
     */
    if( prevPAU == 0 ) {
        freeList = nextPAU;
    } else {
        prevPAU->setNextLink( nextPAU );
    }

    _kernelUnits += nunits;

    release();

    return( chain );
}

/*
 * Return a unit to the Store.
 */
void
Store::dealloc( PhysicallyAddressableUnit * u )
{
    Assert( u != 0 );
    
    PhysicallyAddressableUnit * v = u;
    int i = 1;
    for (; v->nextLink() != 0; v = v->nextLink(), i++) {
        Assert( v && v->owner() == 0 );
    }
    Assert( v && v->owner() == 0 );
    
    Assert( ! _lock.heldByMe() );
    acquire();
    
    v->setNextLink( freeList );
    freeList = u;
    
    _kernelUnits -= i;
    
    release();
}

PhysicallyAddressableUnit *
Store::alloc()
{
    Assert( this != 0 );
    Assert( ! _lock.heldByMe() );
    acquire();
    if( freeList == 0 ) {
        Console().panic( "Store::alloc() out of memory\n" );
        Assert( NOTREACHED );
    }
    PhysicallyAddressableUnit * u = freeList;
    Assert( u->owner() == 0 );
    freeList = freeList->nextLink();
    _kernelUnits++;
    
#ifdef WARNING
    if ( totalAllocPaus() + WARNING > _numberOfUnits ) {
        Console() << this << "::alloc() " 
                  << "less than " << WARNING << " units left\n" << eor;
        info( Console() );
    }
#endif
    
    release();
    u->setNextLink( 0 );
    return( u );
}

void
Store::userDealloc( PhysicallyAddressableUnit * u, MemoryObjectCache * cache,
                    int queued )
{
    Assert( u != 0 );
    
    PhysicallyAddressableUnit * v = u;
    int i = 1;
    for (; v->nextLink() != 0; v = v->nextLink(), i++) {
        Assert( v && v->owner() == 0 );
    }
    Assert( v && v->owner() == 0 );
    
    Assert( ! _lock.heldByMe() );
    acquire();
    
    v->setNextLink( freeList );
    freeList = u;
    _userUnits -= i;
    StoreInfo * storeInfo = cache->storeInfo();
    storeInfo->_paus -= i;
    Assert( storeInfo->_paus >= 0 );
    if ( storeInfo->_paus == 0 ) {
        if ( storeInfo == _current )
            _current = storeInfo->_next;
        storeInfo->unlink();
        storeInfo->_type = STOREINFO_NOT;
        _storeInfoCount--;
    }
    if ( queued ) {
        storeInfo->_queued -= i;
        Assert( storeInfo->_queued >= 0 );
        _queued -= i;
        Assert( _queued >= 0 );
    }
    
    while ( _wantWaiting > 0 ) {
        Process * process = _blocked.head();
        Assert( process != 0 );
        _AwaitInfo * info = (_AwaitInfo *) process->blockedInfo();
        if ( totalAllocPaus() + info->units() >= _high ) 
            break;
        _blocked.removeHead();
        _userUnits += info->units();
        _wantUnits -= info->units();
        _wantWaiting--;
        Assert( _wantWaiting >= 0 );
        process->ready();
    }
    release();
}

PhysicallyAddressableUnit *
Store::userAlloc( MemoryObjectCache * cache)
{
    Assert( this != 0 );
    return userAlloc( 1, cache );
}

/*
 * Allocate "nunits" free units and return them in a linked list.
 */
PhysicallyAddressableUnit *
Store::userAlloc( unsigned int nunits, MemoryObjectCache * cache)
{
    Raids( RaidMethod )
    << this << "::alloc( nunits:" << nunits << " )\n" << eor;
    Assert( this != 0 );
    
    Assert( ! _lock.heldByMe() );
    IntStorage intState ; PROCESSOR::disableInterrupts(intState);
    acquire();

    if ( ( totalAllocPaus() + _wantUnits + nunits ) > _high || 
         _wantUnits > 0 ) {
        _wantUnits += nunits;
        _wantWaiting++;

        // determine whether to start pager 
        if ( _paging == 0 && totalExpectedAllocPaus() > _low ) {
            _paging = 1;
            _pagerSem.V();
        }

#ifdef DEBUGFREEPAGES
        Console()
        << this << "::userAlloc( " << nunits << ", " 
        << cache << " ) await pager  " 
        << totalAllocPaus() << ",  "
        << _wantUnits << "\n" << eor;
#endif // DEBUGFREEPAGES

        Process * process = thisProcess();
        Assert( process != _pageOutDaemon );
        _AwaitInfo * info = (_AwaitInfo *) process->blockedInfo();
        info->setUnits( nunits );
        _blocked.addTail( process );
        release();

        process->block( this );

#ifdef DEBUGFREEPAGES
        Console()
        << this << "::userAlloc( " << nunits << ", "
        << cache << " ) done waiting "
        << totalAllocPaus() << ", "
        << _wantUnits << "\n" << eor;
#endif // DEBUGFREEPAGES
        acquire();
    } else {
        _userUnits += nunits;
    }
 
    PhysicallyAddressableUnit * last = freeList;
    PhysicallyAddressableUnit * prev = 0;

    for( unsigned i = 0; i < nunits; i++ ) {
        Assert( last != 0 );
        // Console().panic( "Store::userAlloc out of memory\n" );
        Assert( last->owner() == 0 );
        prev = last;
        last = last->nextLink();
    }
    
    PhysicallyAddressableUnit * freeUnit = freeList;
    freeList = last;
    
    StoreInfo * storeInfo = cache->storeInfo();
    storeInfo->_paus += nunits;
    if ( storeInfo->_paus == nunits ) {
        Assert( storeInfo->_type == STOREINFO_NOT );
        storeInfo->_type = STOREINFO_ITEM;
        storeInfo->insertBefore( _storeInfoHead );
        _storeInfoCount++;
    }
    
    release();
    PROCESSOR::restoreInterrupts( intState );
    
    Assert( prev != 0 );
    prev->setNextLink( 0 );
    return( freeUnit );
}

OutputStream&
operator<<(OutputStream& os, PhysicallyAddressableUnit * pau)
{
    return os << "PAU[" << asHex((int)pau) << "]("
              << pau->referenced() << "," << pau->modified() << ","
              << asHex((int)pau->address()) << "," 
              << asHex((int)pau->owner()) 
              << ")" ;
}

void
Store::freePages()
{
    Debug << this << "::freePages() alive\n" << eor;

    unsigned nothingAccomplished = 0;

    for(;;) {
        _pagerSem.P();

        acquire();
        Assert( _paging );

#ifdef DEBUGFREEPAGES
        Debug 
        << "freePages() paging"
        << " kernel units: " << _kernelUnits
        << " user units: " << _userUnits
        << " want: " << _wantUnits
        << " wantWaiting: " << _wantWaiting
        << "\n" << eor;
#endif // DEBUGFREEPAGES

        while ( totalExpectedAllocPaus() > _low ) {

            release();
            int killed = MemoryObjectCachingContainer::killZombies();
            acquire();
#ifdef DEBUGFREEPAGES
            Debug 
            << "freePages() paging"
            << " zombies killed: " << killed
            << "\n" << eor;
#endif // DEBUGFREEPAGES
            if ( killed == 0 )
                break;
        }

#ifdef NEVER
        // cannot enable yet
        extern Allocator * GetNewAllocator();
        GetNewAllocator()->reclaimSpace();
#endif

        while ( totalExpectedAllocPaus() > _low ) {

            Assert( _wantUnits >= 0 );
            Assert( _wantWaiting >= 0 );

#ifdef DEBUGFREEPAGES
            Debug 
            << "freePages() paging"
            << " alloc: " << totalAllocPaus() 
            << " want: " << _wantUnits
            << " waiting: " << _wantWaiting
            << "\n" << eor;
#endif // DEBUGFREEPAGES

            if ( _wantUnits + _kernelUnits - _queued >= _high ) {
                Console()
                << "No more user memory available\n"
                << eor;
                Console()
                << " kernel units: " << _kernelUnits
                << " user units: " << _userUnits
                << " want: " << _wantUnits
                << " waiting: " << _wantWaiting
                << " queued: " << _queued
                << "\n" << eor;
                Console().panic( "Halting\n" );
                Assert( NOTREACHED );
            }

            StoreInfo * storeInfo;
            PagedMemoryObjectCache * c;
            int npaus;
            int wantToFree;
            int freePaus;
            int p = 0;

            if ( _current->_type == STOREINFO_LISTHEAD ) {
                _current = _current->_next;
                if ( _current->_type == STOREINFO_LISTHEAD ) {

                    Assert( userAllocPaus() == 0 );
#ifdef DEBUGFREEPAGES
                    Debug 
                    << "freePages adjusting _low "
                    << "new value " 
                    << ( totalAllocPaus() + _wantUnits )
                    << " old value "
                    << _low << "\n" << eor;
#endif // DEBUGFREEPAGES
                    _low = totalAllocPaus() + _wantUnits;
                    if ( _low >= _high ) {
                        Console()
                        << "No more user memory available 1\n"
                        << eor;
                        Console()
                        << " kernel units: " << _kernelUnits
                        << " user units: " << _userUnits
                        << " want: " << _wantUnits
                        << " waiting: " << _wantWaiting
                        << " queued: " << _queued
                        << "\n" << eor;
                        Console().panic
                        ( "Halting\n" );
                        Assert( NOTREACHED );
                    }
                    release();
                    goto loopEndCheck;
                }
            }

            storeInfo = _current;
            c = (PagedMemoryObjectCache *) storeInfo->_cache;
            npaus = storeInfo->_paus;
            _current = _current->_next;
        
            if ( npaus == 0 ) {
                release();
                goto loopEndCheck;
            }

            wantToFree = totalExpectedAllocPaus() - _low;
            Assert( wantToFree > 0 );
            freePaus = ( npaus * wantToFree ) / userAllocPaus();
            Assert( freePaus >= 0 );
            if ( freePaus == 0 ) {
                nothingAccomplished++;
                if ( nothingAccomplished > _storeInfoCount ) {
                    freePaus = 1;
                } else {
                    continue;
                }
            }
    
            if ( freePaus > 16 )
                freePaus++; /* round up */
            if ( freePaus > npaus )
                freePaus = npaus;
            Assert( c->alive() );
            c->reference();
            release();

#ifdef DEBUGFREEPAGES
            Debug 
            << "freePages() paging"
            << " freePaus: " << freePaus
            << " cache: " << c
            << "\n" << eor;
#endif // DEBUGFREEPAGES

            if ( (p = c->pageOut( freePaus )) == 0 ) {
#ifdef DEBUGFREEPAGES
                Debug 
                << "Store::freePages() " 
                << c << "no pages\n" << eor;
#endif // DEBUGFREEPAGES
                nothingAccomplished++;
            } else {
#ifdef DEBUGFREEPAGES
                Debug 
                << "Store::freePages() " 
                << c << " frees " << p 
                << " pages\n"  << eor;
#endif // DEBUGFREEPAGES
                nothingAccomplished = 0;
            }
            c->unreference();
        loopEndCheck:

#ifdef DEBUGFREEPAGES
            Debug 
            << "freePages() paging"
            << " totalAllocPaus: " << totalAllocPaus() 
            << " queued: " << _queued
            << " nothingAccomplished : " 
            << nothingAccomplished
            << "\n" << eor;
#endif // DEBUGFREEPAGES
     
            acquire();
            if ( nothingAccomplished > ( _storeInfoCount << 4 ) ) {
                Console() 
                << "freePages unable to reclaim "
                << "space in " << nothingAccomplished
                << " attempts\n" << eor;
                Console()
                << " kernel units: " << _kernelUnits
                << " user units: " << _userUnits
                << " want: " << _wantUnits
                << " waiting: " << _wantWaiting
                << " queued: " << _queued
                << "\n" << eor;
                Console().panic( "Halting\n" );
                Assert( NOTREACHED );
            }
        }

#ifdef DEBUGFREEPAGES
        Debug
        << "freePages() done paging"  
        << " alloc: " << totalAllocPaus() 
        << " queued: " << _queued
        << " want: " << _wantUnits
        << "\n" << eor;
#endif // DEBUGFREEPAGES

        _paging = 0;
        
        release();
    }
}

#include "ClassInstaller.h"
DEFINECLASS("Store",Store,ProxiableObject,0,0);
