#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Swee Boon Lim (sblim@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * TranslationMap.cc
 */

#include "AddressTranslation.h"
#include "Map.h"
#include "MemoryObjectView.h"
#include "PhysicallyAddressableUnit.h"
#include "PhysicalMemoryChain.h"
#include "Raid.h"
#include "TranslationMap.h"


#ifdef __GNUC__
#pragma implementation
#endif

//#define DEBUG_NEW_STUFF

TranslationMap::TranslationMap( Map * map )
     : _map( map )
{
	initDebug();
	Assert( map != 0 );
	_head = 0;
	_prev = 0;
	_current = 0;
}

TranslationMap::~TranslationMap()
{
	Assert( _head == 0 );
}

_adt *
TranslationMap::first()
{
	_prev = 0;
	_current = _head;
	return _head;
}

_adt *
TranslationMap::next()
{
	Assert( _current );

	_prev = _current;
	_current = _current->nextLink();
	return _current;
}

_adt *
TranslationMap::remove()
{
	Assert( _current );

	_adt * dead = _current;
	_current = _current->nextLink();
	if ( _prev )
		_prev->setNextLink( _current );
	else {
		Assert( _head == dead );
		_head = _current;
	}
	deleteEntry( dead );
	return _current;
}

void 
TranslationMap::add( AddressTranslation * adt, const char * baseAddr, 
		     MemoryObjectView * view )
{
	_adt a( adt, baseAddr, 0, 0 );
	_adt * t;
	for ( t = _head; t != 0; t = t->nextLink() )
	        if ( *t == a ) 
			break;
	if ( t == 0 ) {
		t = _head;
		_head = newEntry( adt, baseAddr, view );
		_head->setNextLink( t );
		Raids( RaidDetail ) 
			<< this << "::addAdt( " << _head->adt() << ", " 
			<< asHex( ( unsigned ) _head->addr() ) << ", "
			<< asHex( ( unsigned ) _head->offset() ) << ", " 
			<< asHex( ( unsigned ) _head->size() ) 
			<< " )\n" << eor;
	}
}

void
TranslationMap::translatable( unsigned from, unsigned to, 
			      ProtectionLevel prot, 
			      AddressTranslation * adt,
			      const char * vaddr,
			      MemoryObjectView * view,
			      const char * baseAddr,
			      PhysicallyAddressableUnit * pauList )
{	
	Raids( RaidMethod ) << this << "::translatable( "
		<< from << ", " << to << ", " << prot << ", "
		<< adt << ", " << asHex( ( unsigned ) vaddr ) 
		<< " )\n" << eor;

	Assert( adt != 0 );
	acquire();
	add( adt, baseAddr, view );
	release();
	unsigned log2UnitSize = _map->log2UnitSize();
	unsigned pageMask = ( 1 << Log2PAUUnitSize ) - 1;
	unsigned mask = ( 1 << log2UnitSize ) - 1;
	unsigned viewOffset = view->viewOffset();
	unsigned viewLength = 
		( ( view->numberOfUnits() << view->log2UnitSize() ) +
		  pageMask ) & ~pageMask;
	unsigned noOfBytes = ( to - from + 1 ) << log2UnitSize;
	char * startPageAddr = 
		( ( char * ) baseAddr + (from << log2UnitSize) ) - viewOffset;
#ifdef DEBUG_NEW_STUFF
	if ( pauList ) {
		PhysicallyAddressableUnit * checkList = 
			_map->pauList( from, to );
		PhysicallyAddressableUnit * tmp = pauList;
		unsigned count = ( to - from + 1 ) << 
				 ( log2UnitSize - Log2PAUUnitSize );
		while ( count -- ) {
			Assert( tmp != 0 );
			Assert( tmp == checkList );
			tmp = tmp->nextLink();
			checkList = checkList->nextLink();
		}
	}
#endif // DEBUG_NEW_STUFF

	if ( pauList == 0 )
		pauList = _map->pauList( from, to );
	
	if ( log2UnitSize == Log2PAUUnitSize || 
	     ( ( viewOffset & mask ) == 0 && ( viewLength & mask ) == 0 ) ) {
		
		/* simple case */

		PhysicalMemoryChain p1( pauList, 0, noOfBytes, 0 );
		adt->addMapping( startPageAddr, &p1, prot );
				 
		return;
	}

	/* complicated case */

	if ( baseAddr > startPageAddr ) {
		unsigned startAdjust = baseAddr - startPageAddr;
		Assert( ( startAdjust & pageMask ) == 0 );
		noOfBytes -= startAdjust;
		startPageAddr += startAdjust;
		for( ; startAdjust > 0; startAdjust -= PAUUnitSize )
			pauList = pauList->nextLink();
	}

	unsigned viewLast = viewOffset + viewLength;
	if ( ( ( viewLast - 1 ) >> log2UnitSize ) == to && 
	     ( viewLast & mask ) != 0 ) {
		unsigned endAdjust = ( 1 << log2UnitSize ) - 
			             ( viewLast & mask );
		Assert( ( endAdjust & pageMask ) == 0 );
		noOfBytes -= endAdjust;
	}
	
	PhysicalMemoryChain p2( pauList, 0, noOfBytes, 0 );
	adt->addMapping( startPageAddr, &p2, prot );
}

void 
TranslationMap::untranslatable( unsigned from, unsigned to,
			        AddressTranslation * adt,
			        const char * baseAddr )
{
	Raids( RaidMethod ) << this << "::untranslatable( " 
		<< from << ", " << to << ", " << adt << ", " 
		<< asHex( ( unsigned ) baseAddr ) << " )\n" << eor;

	Assert( adt == 0 || baseAddr != 0 );
	int log2UnitSize = _map->log2UnitSize();
	int size = ( to - from + 1 ) << log2UnitSize;
	int offset = from << log2UnitSize;
	int endOffset = offset + size;
	acquire();
	int removeEntry = 0;
	for ( _adt * a = first(); 
	      a != 0; 
	      a = ( removeEntry ? remove() : next() ) ) {
		removeEntry = 0;
		if ( adt == 0 || 
		     ( a->adt() == adt && a->addr() == baseAddr ) ) {

			Raids( RaidDetail )
				<< this << "::untranslatable() " 
				<< " adt = " << a->adt() 
				<< " where = " 
				<< asHex( ( unsigned ) a->addr() )
				<< " offset = " << asHex( a->offset() )
				<< " size = " << asHex( a->size() )
				<< "\n" << eor;

			unsigned newOffset = offset - a->offset();
			if ( newOffset < 0 ) 
				newOffset = 0; 
			else if ( newOffset > a->size() ) 
				newOffset = a->size();	

			unsigned newEndOffset = endOffset - a->offset();
			if ( newEndOffset < 0 ) 
				newEndOffset = 0;
			else if ( newEndOffset > a->size() ) 
				newEndOffset = a->size();

			unsigned newSize = newEndOffset - newOffset;

			if ( newSize > 0 ) {
				
				Raids( RaidDetail ) << this
					<< "::untranslatable() "
					<< " remove mapping( "
					<< asHex( (unsigned) a->addr()
						  + newOffset )
					<< ", " << newSize 
					<< " )\n" << eor;

				a->adt()->removeMapping( a->addr() + newOffset,
							 newSize );
				
			} 
			if ( newSize == a->size() ) {
				removeEntry = 1;
			}
		} 
	}
        release();
}

int
TranslationMap::referenced( unsigned from, unsigned to, int clear )
{
	Raids( RaidMethod ) << this << "::referenced( " 
		<< from << ", " << to << ", " << clear << " )\n" << eor;

	int log2UnitSize = _map->log2UnitSize();
	int size = ( to - from + 1 ) << log2UnitSize;
	int offset = from << log2UnitSize;
	int endOffset = offset + size;
	int ret = 0;
	acquire();
	for ( _adt * a = first(); a != 0; a = next() ) {
		Raids( RaidDetail )
			<< this << "::referenced() " 
			<< " adt = " << a->adt() 
			<< " where = " 
			<< asHex( ( unsigned ) a->addr() )
			<< " offset = " << asHex( a->offset() )
			<< " size = " << asHex( a->size() )
			<< "\n" << eor;

		unsigned newOffset = offset - a->offset();
		if ( newOffset < 0 ) 
			newOffset = 0; 
		else if ( newOffset > a->size() ) 
			newOffset = a->size();	

		unsigned newEndOffset = endOffset - a->offset();
		if ( newEndOffset < 0 ) 
			newEndOffset = 0;
		else if ( newEndOffset > a->size() ) 
			newEndOffset = a->size();

		int newSize = newEndOffset - newOffset;

		if ( newSize > 0 ) {
			
			Raids( RaidDetail ) << this
				<< "::referenced() "
				<< " update mapping( "
				<< asHex( (unsigned) a->addr()
					  + newOffset )
				<< ", " << newSize 
				<< " )\n" << eor;

			int ref = a->adt()->referenced( a->addr() + newOffset, 
							newSize, clear );
			if ( ref ) ret = 1;
			
		} 
	}

       	release();
	return ret;
}

Lock LinkedListTranslationMap::_staticLock;
_adt * LinkedListTranslationMap::_free = 0;
int LinkedListTranslationMap::_freeCount = 0;

const int MaxFreeCount = 64;

LinkedListTranslationMap::LinkedListTranslationMap( Map * map )
     : TranslationMap( map ) 
{
	_lock = &_staticLock;
	initDebug();
#ifdef DEBUG
	_debugMask = MaxObjectDebugMask;
	_debugMessages = MaxObjectDebugMessages;
#endif
}

_adt *
LinkedListTranslationMap::newEntry( AddressTranslation * adt, 
				    const char * baseAddr,
				    MemoryObjectView * view )
{
	_adt * a;
	if ( _free ) {
		a = _free;
		_free = a->nextLink();
		_freeCount--;
		a->init( adt, baseAddr, view->viewOffset(), view->size() );
	} else {
		a = new _adt( adt, baseAddr, view->viewOffset(), view->size() );
	}
	return a;
}

void
LinkedListTranslationMap::deleteEntry( _adt * a )
{
	if ( _freeCount < MaxFreeCount ) {
		a->setNextLink( _free );
		_free = a;
		_freeCount++;
	} else {
		delete a;
	}
}

ConstantSizeTranslationMap::ConstantSizeTranslationMap( Map * map )
     : LinkedListTranslationMap( map )
{
 	_lock = &_instanceLock;
	initDebug();
#ifdef DEBUG
	_debugMask = MaxObjectDebugMask;
	_debugMessages = MaxObjectDebugMessages;
#endif
	_freeList = & _array[ 0 ];
	for( int i = 0; i < ( MaxConstantSizeTranslationMapAdt - 1 ); i++ ) {
		_array[ i ].setNextLink( & _array[ i + 1 ] );
	}
	_array[ MaxConstantSizeTranslationMapAdt - 1 ].setNextLink( 0 );
}

_adt * 
ConstantSizeTranslationMap::newEntry( AddressTranslation * adt, 
				      const char * baseAddr,
				      MemoryObjectView * view )
{
	Assert( _freeList != 0 ); // out of space
	_adt * a = _freeList;
	_freeList = a->nextLink();
	a->init( adt, baseAddr, view->viewOffset(), view->size() );
	return a;
}

void
ConstantSizeTranslationMap::deleteEntry( _adt * a )
{
	Assert( a != 0 );
	a->setNextLink( _freeList );
	_freeList = a;
}

