#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * TwoLevelPageTable.cc: Generic two-level page tables.
 */

#include "Assert.h"
#include "MemoryBlockAllocator.h"
#include "OutputStream.h"
#include "PAGE_TABLE_ENTRY.h"
#include "PageTableConstants.h"
#include "PhysicalMemoryChain.h"
#include "PhysicalMemoryManager.h"
#include "PhysicallyAddressableUnit.h"
#include "TwoLevelPageTable.h"
#include "VirtualMemoryRange.h"
#include "ThisProcess.h"
#include "Kernel.h"


#ifdef __GNUC__
#pragma implementation
#endif


/*
 * Static data
 */
VirtualMemoryRange * TwoLevelPageTable::kernelVirtualMemoryMap = 0;
MemoryBlockAllocator * TwoLevelPageTable::tableMemoryAllocator = 0;

/*
 * The REAL translation information is stored by the Domains and Caches;
 * the hardware page tables are used as another level of cacheing, and
 * the information they contain can be discarded (after transferring
 * dirty and referenced information back to the Domain) at any time.
 * Therefore, at system startup time, we allocate a fixed amount of
 * space for hardware page tables.
 *
 * Set up table allocation stuff and initialize the necessary information
 * to keep the mappings common to all instances of this class.
 */

int TwoLevelPageTableClassInitialized = 0;

void
TwoLevelPageTable::init( unsigned int tableAreaSize,
                         VirtualMemoryRange * memoryMap )
{
    Debug << "TwoLevelPageTable::init( tableAreaSize: " << tableAreaSize
          << " )\n" << eor;

    /*
     * Make sure this is the first time we're called.
     */
    Assert( ! TwoLevelPageTableClassInitialized );

    kernelVirtualMemoryMap = memoryMap;

    tableMemoryAllocator = new MemoryBlockAllocator( tableAreaSize,
                                                     PhysicalPageSize );

    TwoLevelPageTableClassInitialized = 1;
}


int
TwoLevelPageTable::pointerTableIndex( const char * address )
{
    int index = (unsigned int) address >> (Log2PageTableEntries + Log2PhysicalPageSize);
    Assert( index < PointerTableEntries );
    return( index );
}

int
TwoLevelPageTable::pageTableIndex( const char * address )
{
    int index = ((unsigned int) address >> Log2PhysicalPageSize) & (PageTableEntries - 1);
    Assert( index < PageTableEntries );
    return( index );
}


/*
 * Allocate and initialize pointer and page tables.
 */

POINTER_TABLE_ENTRY *
TwoLevelPageTable::allocatePointerTable()
{
    Assert( tableMemoryAllocator != 0 );
    POINTER_TABLE_ENTRY * table = (POINTER_TABLE_ENTRY *)
    tableMemoryAllocator->allocate( PointerTableSize );

    for( int i = 0; i < PointerTableEntries; i++ ) {
        ((int *)table)[i] = 0;
    }

    return( table );
}

PAGE_TABLE_ENTRY *
TwoLevelPageTable::allocatePageTable()
{
    Assert( tableMemoryAllocator != 0 );
    PAGE_TABLE_ENTRY * table = (PAGE_TABLE_ENTRY *)
    tableMemoryAllocator->allocate( PageTableSize );

    for( int i = 0; i < PageTableEntries; i++ ) {
        ((int *)table)[i] = 0;
    }

    return( table );
}

TwoLevelPageTable::TwoLevelPageTable(  TwoLevelPageTable * parent ) : lock( "TwoLevelPageTable" )
{
    /*
     * Don't try "<< this" here; it's too early!
     */
    Debug << "TwoLevelPageTable::TwoLevelPageTable()\n" << eor;
    Assert( this != 0 );

    _parent = parent;
    /*
     * Defer allocating any tables until the first addMapping().
     */

    thisPointerTable = 0;
}

TwoLevelPageTable::~TwoLevelPageTable()
{
    Debug << this << "::~TwoLevelPageTable()\n" << eor;
    Assert( this != 0 );

    /*
     * It could be that we never called addMapping() or
     * firstLevelTable() and therefore never allocated any
     * tables for this translation.  If that's the case then
     * nothing extra need be done.  However, if we have allocated a
     * table then we need to release space we've allocated.
     */

    if( thisPointerTable != 0 ) {
        for( int i = 0; i < PointerTableEntries; i++ ) {

            POINTER_TABLE_ENTRY * pointerTableEntry = &thisPointerTable[ i ];
            if( ! pointerTableEntry->valid() ) {
                continue;
            }

            /*
             * If there is a valid entry in the parent pointer
             * table, the page tables are shared and must not
             * be deleted.
             */
            if ( ( _parent != 0 ) && _parent->pointerTable()[i].valid() 
                 && (_parent->pointerTable()[i] == *pointerTableEntry))
                continue;

            PAGE_TABLE_ENTRY * pageTable = pointerTableEntry->pageTable();
            Debug << this << "::dtor page table at "
                  << asAddress( pageTable ) << "\n" << eor;

            for( int j = 0; j < PageTableEntries; j++ ) {
                PAGE_TABLE_ENTRY * pageTableEntry = &pageTable[ j ];

                if( ! pageTableEntry->valid() ) {
                    continue;
                }

                /*
                 * If this particular PTE is valid, dump its
                 * ref/mod bits back into the machine-
                 * independent frame table info.
                 */
                Debug << this << "::dtor: " << i
                      << "[" << j << "] used\n" << eor;

                Assert( MainMemory != 0 );
                PhysicallyAddressableUnit * pau = MainMemory->lookup( pageTableEntry->addressMappedTo() );
                Assert( pau != 0 );
                pau->setReferenced( pageTableEntry->referenced() );
                pau->setModified( pageTableEntry->modified() );

                Debug << this << "::dtor dirty: "
                      << pau->modified()
                      << " ref: " << pau->referenced()
                      << "\n" << eor;
            }
            tableMemoryAllocator->free( pageTable );
        }
        tableMemoryAllocator->free( thisPointerTable );
    }
}

/*
 * setupPointerTable is called by either pointerTable()
 * or addMapping() if they discover that the first level page table
 * for this translation has not yet been allocated.
 */
void
TwoLevelPageTable::setupPointerTable()
{
    Assert( TwoLevelPageTableClassInitialized );
    Assert( thisPointerTable == 0);

    thisPointerTable = allocatePointerTable();
    Assert( thisPointerTable != 0 );
    Assert( PageAligned( (const char *) thisPointerTable ) );
    Debug << this << "::setupPointerTable: thisPointerTable "
          << asAddress( thisPointerTable ) << "\n" << eor;


    /*
     * Should assert that shared boundaries can only be
     * on 2^(Log2PointerTableSize + Log2PageTableSize)
     * boundaries.  (I.e., 4 Meg boundaries for Multimax and
     * Macintosh.)
     */

    /*
     * Copy pointer mappings for system regions from parent tables
     *  to thisPointerTable, and allocate second-level page tables
     *  for shared regions.
     */
    int pointerTableEntryScope = PhysicalPageSize * PageTableEntries;
    for( VirtualMemoryRange * region = kernelVirtualMemoryMap; region != 0;
         region = region->next() ) {
        char * base = (char *) region->virtualAddress();
        int length = region->numberOfUnits() * region->unitSize();

        Debug << this << "::setupPointerTable: adding mappings for\n\t"
              << region << " ; size = " << length << " bytes \n" << eor;

        while( length > 0 ) {
            int i = pointerTableIndex( base );
            if(_parent) {
                ((unsigned *)thisPointerTable)[i] = ((unsigned *)_parent->pointerTable())[i];
            } else {
                /*
                 * Allocate a second-level page table.
                 */
                thisPointerTable[i].mapToPageTable( allocatePageTable() );
            }
            base += pointerTableEntryScope;
            length -= pointerTableEntryScope;
        }
    }
}

POINTER_TABLE_ENTRY *
TwoLevelPageTable::pointerTable()
{
    if( thisPointerTable == 0 ) {
        lock.acquire();
        setupPointerTable();
        lock.release();
        Assert( thisPointerTable != 0 );
    }

    return( thisPointerTable );
}

void
TwoLevelPageTable::addMapping( const char * virtualAddress,
                               PhysicalMemoryChain * chain, ProtectionLevel maxProtection )
{
    // Debug << this << "::addMapping( vaddr: " << asAddress( virtualAddress )
    //          << " chain: " << asAddress( chain ) << " maxProt: "
    //          << maxProtection << "\n" << eor;
    Assert( this != 0 );
    Assert( TwoLevelPageTableClassInitialized );
    Assert( chain != 0 );
    unsigned int nbytes = chain->lengthInBytes();
    if( nbytes == 0 ) {
        return;
    }

    /*
     * Round the virtual address down to the base of the page it falls
     * on. Then calculate how many pages to map in order to map nbytes.
     */
    const char * vpagebase = FloorPage( virtualAddress );
    PhysicallyAddressableUnit * pau = chain->chain();

    Assert( (int)chain->offset() == (virtualAddress - vpagebase) );

    int npages = ( nbytes + ( virtualAddress - vpagebase )
                   + PhysicalPageSize - 1 ) / PhysicalPageSize;
    
    Debug << this << "::addMapping: " << asAddress( vpagebase )
          << " -> " << asAddress( pau->address() ) << ", "
          << npages << " pgs, prot: " << maxProtection << "\n" << eor;

    /*
     * If this translation doesn't have a pointer table yet,
     * allocate and initialize one.
     */
    Assert( ! lock.heldByMe() );
    lock.acquire();
    if( thisPointerTable == 0 ) {
        setupPointerTable();
    }
    Assert( thisPointerTable != 0 );

    /*
     * There is a race here because the table just allocated could
     * go away before we use it below.  I have to release the lock because
     * I have to get the PAU locks *first*.  I think we can fix this
     * later. --Vince
     */

    for( int page = 0; page < npages; page++ ) {
        // Debug << "Progress: " << (page+1) << "/" << npages << "\n" << eor;
        Assert( pau != 0 );

        ProtectionLevel prot = maxProtection;

        /*
         * If there isn't a valid page table allocate one.
         */
        int i = pointerTableIndex( vpagebase );
        POINTER_TABLE_ENTRY * pointerTableEntry = &thisPointerTable[ i ];
        if( ! pointerTableEntry->valid() ) {
            PAGE_TABLE_ENTRY * pageTable = allocatePageTable();

            Assert( ! pointerTableEntry->valid() );
            pointerTableEntry->mapToPageTable( pageTable );
        } else {
            // Perform COW of page table entries if parent exists
            if (_parent && (_parent->pointerTable()[i] == *pointerTableEntry)) {
                Debug << "Adding mapping to page table in parent. Performing COW\n" << eor;
                pointerTableEntry->mapToPageTable(allocatePageTable());
                PAGE_TABLE_ENTRY * pageTable = pointerTableEntry->pageTable();
                PAGE_TABLE_ENTRY * parentPageTable = _parent->pointerTable()[i].pageTable();
                for(int j = 0; j < PageTableEntries; j++)
                    ((unsigned *)pageTable)[j] = ((unsigned *)parentPageTable)[j];
            }
        }

        /*
         * Get a pointer to the proper page table.
         */
        PAGE_TABLE_ENTRY * pageTable = pointerTableEntry->pageTable();

        // Debug << this << "::addMapping: L1 table index "
        //          << pointerTableIndex( vpagebase ) << " ("
        //          << asAddress( vpagebase ) << ") at "
        //          << asAddress( pageTable ) << "\n" << eor;

        //Debug << this << "::addMapping: page table at "
        //          << asAddress( pageTable ) << "\n" << eor;
        PAGE_TABLE_ENTRY * pageTableEntry = &pageTable[ pageTableIndex( vpagebase ) ];
        
        //Debug << this << "::addMapping: 2nd index: "
        //          << pageTableIndex( vpagebase ) << " *pageTableEntry: "
        //          << asAddress( *(const void ** const)pageTableEntry )
        //          << "\n" << eor;

        /*
         * If a mapping already exists there Assert that its the
         * same one as we would add.
         */
        if( pageTableEntry->valid() ) {
            //Debug << this << "::addMapping: dup "
            //          << asAddress( vpagebase ) << " -> "
            //          << asAddress( pau->address() ) << " ("
            //          << asAddress( pageTableEntry->addressMappedTo() )
            //          << "\n" << eor;
            Assert( pageTableEntry->addressMappedTo() == pau->address() );
        }
        else {
            Assert( ! pageTableEntry->valid() );
            if( !PageAligned( pau->address() ) ) {
                Console() << "pau not aligned: " <<
                asHex( (unsigned int) pau->address() )
                          << "\n" << eor;
            }
            //Debug << "Performing AddMapping() " << asHex((void*)pau->address()) << "->" << asHex((void*)pageTableEntry->addressMappedTo()) << "\n" << eor;
            pageTableEntry->mapToAddress( pau->address(), prot );


            /*
             * There is a small (possibly innocuous) bug here.
             * For kernel memory, the ADT actually adding the
             * mapping to the shared tables will be marked in
             * the PAU as the holder of the reference.  In reality
             * some SystemADT ought to be credited with it.  --Vince
             */
        }

        vpagebase += PhysicalPageSize;

        pau = pau->nextLink();
    }
    lock.release();
}

/*
 * Remove a mapping and update the machine independent ref/mod information
 * for a range of virtual addresses.
 */
void
TwoLevelPageTable::basicRemoveMapping( const char * virtualAddress,
                                       unsigned int nbytes )
{
    Debug << this << "::removeMapping( va: "
          << asAddress( virtualAddress ) << ", nbytes: "
          << nbytes << " )\n" << eor;
    Assert( this != 0 );
    Assert( TwoLevelPageTableClassInitialized );
    if( nbytes == 0 ) {
        return;
    }

    //Assert (mapping(virtualAddress) != virtualAddress);
    Assert( ! lock.heldByMe() );
    lock.acquire(); // this lock is much much to coarse... fix it later 

    if( thisPointerTable == 0 ) {
        Debug << this << "::removeMapping: no maps\n" << eor;
        lock.release();
        return;
    }

    Assert( PageAligned( (const char *) thisPointerTable ) );

    /*
     * Round the virtual address down to the base of the page it
     * falls on. Then calculate how many pages to remove in order
     * to satisfy "nbytes".  For now, require the the virtual
     * address be page-aligned.
     */
    const char * vpagebase = FloorPage( virtualAddress );
    // for now...
    Assert( PageAligned( virtualAddress ) );

    int npages = ( nbytes + ( virtualAddress - vpagebase )
                   + PhysicalPageSize - 1 ) / PhysicalPageSize;

    Debug << this << "::removeMapping: " << asAddress( vpagebase ) << " (" << npages << ")\n" << eor;

    for( int page = 0; page < npages; page++ ) {

        /*
         * If there isn't a valid pointer table there is
         * no mapping to remove.
         */
        POINTER_TABLE_ENTRY * pointerTableEntry = &thisPointerTable[ pointerTableIndex( vpagebase ) ];
        if( ! pointerTableEntry->valid() ) {
            Debug << this 
                  << "::removeMapping: no pointer table\n"
                  << eor;
            vpagebase += PhysicalPageSize;
            continue;
        }

        // Perform COW of page table entries if parent exists
        if (_parent && (_parent->pointerTable()[pointerTableIndex( vpagebase )] == *pointerTableEntry)) {
            Debug << "Removing mapping in parent pagetable. Performing COW\n" << eor;
            pointerTableEntry->mapToPageTable(allocatePageTable());
            PAGE_TABLE_ENTRY * pageTable = pointerTableEntry->pageTable();
            PAGE_TABLE_ENTRY * parentPageTable = _parent->pointerTable()[pointerTableIndex( vpagebase )].pageTable();
            for(int i = 0; i < PageTableEntries; i++)
                ((unsigned *)pageTable)[i] = ((unsigned *)parentPageTable)[i];
        }

        /*
         * Get a pointer to the proper page table.
         */
        // Debug << this << "::removeMapping: pointer table at " << asAddress( thisPointerTable ) << "\n" << eor;
        PAGE_TABLE_ENTRY * pageTable = pointerTableEntry->pageTable();
        PAGE_TABLE_ENTRY * pageTableEntry = &pageTable[ pageTableIndex( vpagebase ) ];

        /*
         * If a mapping already exists invalidate it, otherwise 
         * continue on.
         */
        if( pageTableEntry->valid() ) {
            // Debug << this << "::removeMapping: do it \n" << eor;
            /*
             * Unmap it first to avoid race of hardware changing
             * it right after you look. (??)
             */
            pageTableEntry->unmap();

            /*
             * Copy the ref/mod info into the machine independent
             * pau's tables.
             */
            Assert( MainMemory != 0 );
            PhysicallyAddressableUnit * pau = MainMemory->lookup( pageTableEntry->addressMappedTo() );
            Assert( pau != 0 );
            pau->setReferenced( pageTableEntry->referenced() );
            pau->setModified( pageTableEntry->modified() );

            // Debug << this << "::removeMapping: dirty: " << pau->modified() << " ref: " << pau->referenced() << "\n" << eor;
        }

        vpagebase += PhysicalPageSize;
    }
    
    lock.release();
    // Debug << this << "::removeMapping: done\n" << eor;
}

int
TwoLevelPageTable::basicChangeProtection(const char * virtualAddress, unsigned
                                         nbytes, ProtectionLevel newProtection)
{
    Debug << this << "::basicChangeProtection( va: "
          << asAddress( virtualAddress ) << ", nbytes: "
          << nbytes << " )\n" << eor;

    Assert( this != 0 );
    Assert( TwoLevelPageTableClassInitialized );

    int rval = 0;
    
    if( nbytes == 0 ) {
        Console() << "::basicChangeProtection called with nbytes 0\n" << eor;
        nbytes = PhysicalPageSize;
    }

    Assert( ! lock.heldByMe() );
    lock.acquire(); // this lock is much much to coarse... fix it later 

    if( thisPointerTable == 0 ) {
        Debug << this << "::basicChangeProtection: no maps\n" << eor;
        setupPointerTable();
    } 
    
    Debug << this << "::basicChangeProtection: pointer table at "
          << asAddress( thisPointerTable ) << "\n" << eor;

    Assert( PageAligned( (const char *) thisPointerTable ) );

    /*
     * Round the virtual address down to the base of the page it
     * falls on. Then calculate how many pages to remove in order
     * to satisfy "nbytes".  For now, require the the virtual
     * address be page-aligned.
     */
    const char * vpagebase = FloorPage( virtualAddress );
    // for now...
    Assert( PageAligned( vpagebase ) );

    int npages = ( nbytes + ( virtualAddress - vpagebase )
                   + PhysicalPageSize - 1 ) / PhysicalPageSize;

    Debug << this << "::basicChangeProtection: va: "
          << asAddress( vpagebase ) << " num pages: " << npages << "\n" << eor;

    for( int page = 0; page < npages; page++ ) {

        /*
         * If there isn't a valid pointer table there is
         * no mapping to change.
         */
        POINTER_TABLE_ENTRY * pointerTableEntry = &thisPointerTable[ pointerTableIndex( vpagebase ) ];
        if( ! pointerTableEntry->valid() ) {
            Debug << this 
                  << "::basicChangeProtection: no pointer table\n"
                  << eor;
            vpagebase += PhysicalPageSize;
            continue;
        }


        // Perform COW of page table entries if parent exists
        if (_parent && (_parent->pointerTable()[pointerTableIndex( vpagebase )] == *pointerTableEntry)) {
            Debug << "Changing protection in parent pagetable. Performing COW\n" << eor;
            pointerTableEntry->mapToPageTable(allocatePageTable());
            PAGE_TABLE_ENTRY * pageTable = pointerTableEntry->pageTable();
            PAGE_TABLE_ENTRY * parentPageTable = _parent->pointerTable()[pointerTableIndex( vpagebase )].pageTable();
            for(int i = 0; i < PageTableEntries; i++)
                ((unsigned *)pageTable)[i] = ((unsigned *)parentPageTable)[i];
        }

        /*
         * Get a pointer to the proper page table.
         */
        PAGE_TABLE_ENTRY * pageTable = pointerTableEntry->pageTable();
        PAGE_TABLE_ENTRY * pageTableEntry = &pageTable[ pageTableIndex( vpagebase ) ];

        // change the protection
        
        if( pageTableEntry->valid() ) {
            rval = 1;
            
            const char * oldAddress = 
            pageTableEntry->addressMappedTo();
            Debug << this <<  "::Changing protection for physical page: "
                  << asAddress( oldAddress ) << "\n" << eor;

            pageTableEntry->setProtection( newProtection );
            
            Assert( pageTableEntry->addressMappedTo() == oldAddress );
        }

        vpagebase += PhysicalPageSize;
    }
    
    lock.release();
    Debug << this << "::basicChangeProtection: done\n" << eor;
    return rval;
}

/*
 * Return the physical address to which the given virtual address is mapped.
 */
const char *
TwoLevelPageTable::mapping( const char * vaddr )
{
    POINTER_TABLE_ENTRY * pointerTableEntry =
    &thisPointerTable[ pointerTableIndex( vaddr ) ];
    if( ! pointerTableEntry->valid() ) {
        return( (const char *) -1 );
    } else {
        PAGE_TABLE_ENTRY * pageTable = pointerTableEntry->pageTable();
        PAGE_TABLE_ENTRY * pageTableEntry =
        &pageTable[ pageTableIndex( vaddr ) ];
        if( ! pageTableEntry->valid() ) {
            return( (const char *) -1 );
        } else {
            return( pageTableEntry->addressMappedTo() );
        }
    }
}

AddressTranslationFailureCode
TwoLevelPageTable::determineError( const char * vaddr,
                                   AccessType attemptedAccess )
{
    /*
     * Get the pointer table entry.  If it isn't valid, the error is
     * NonResidentMemory.
     */
    Assert( thisPointerTable != 0 );
    POINTER_TABLE_ENTRY * pointerTableEntry;
    pointerTableEntry = &thisPointerTable[ pointerTableIndex( vaddr ) ];

    if( ! pointerTableEntry->valid() ) {
        return( NonResidentMemory );
    }

    /*
     * Same holds true for the page table entry.
     */
    PAGE_TABLE_ENTRY * pageTable = pointerTableEntry->pageTable();
    Assert( pageTable != 0 );
    PAGE_TABLE_ENTRY * pageTableEntry = &pageTable[ pageTableIndex( vaddr ) ];
    if( ! pageTableEntry->valid() ) {
        return( NonResidentMemory );
    }

    /*
     * If the PTE protection indicates that the access should have
     * been allowed, it means the MMU had cached a PTE whose protection
     * was later upgraded.
     */
    if( pageTableEntry->okToPerform( attemptedAccess ) ) {
        return( MMUCacheMiss );
    }
    return( ProtectionViolation );
}

/**
 * Mark a page as being referenced.
 * @param vaddr
 * @param nbytes
 * @param clear
 * @return Unknown.
 */
int TwoLevelPageTable::basicReferenced(const char * vaddr, unsigned int nbytes,
                                       int clear)
{
    Assert(NOTREACHED);
    return 0;
}

/**
 * Change the status of a page.
 * @param vaddr
 * @param nbytes
 * @param clear
 * @param set
 * @return Unknown.
 */
int TwoLevelPageTable::basicChangeStatus(const char * vaddr,
                                         unsigned int nbytes, AddressTranslationStatus clear,
                                         AddressTranslationStatus set)
{
    Assert(NOTREACHED);
    return 0;
}

/*
 * inhibit Data Caching
 * for a range of virtual addresses.
 */
void
TwoLevelPageTable::inhibitDataCaching( const char * virtualAddress,
                                       unsigned int nbytes )
{
    Debug << this << "::inhibitDataCaching( va: "
          << asAddress( virtualAddress ) << ", nbytes: "
          << nbytes << " )\n" << eor;
    Assert( this != 0 );
    Assert( TwoLevelPageTableClassInitialized );
    if( nbytes == 0 ) {
        return;
    }

    Assert( ! lock.heldByMe() );
    lock.acquire(); // this lock is much much to coarse... fix it later 

    if( thisPointerTable == 0 ) {
        Debug << this << "::inhibitDataCaching: no maps\n" << eor;
        lock.release();
        return;
    }

    Assert( PageAligned( (const char *) thisPointerTable ) );

    /*
     * Round the virtual address down to the base of the page it
     * falls on. Then calculate how many pages to remove in order
     * to satisfy "nbytes".  For now, require the the virtual
     * address be page-aligned.
     */
    const char * vpagebase = FloorPage( virtualAddress );
    // for now...
    Assert( PageAligned( virtualAddress ) );

    int npages = ( nbytes + ( virtualAddress - vpagebase )
                   + PhysicalPageSize - 1 ) / PhysicalPageSize;

    Debug << this << "::inhibitDataCaching: " << asAddress( vpagebase ) << " (" << npages << ")\n" << eor;

    for( int page = 0; page < npages; page++ ) {
        POINTER_TABLE_ENTRY * pointerTableEntry = &thisPointerTable[ pointerTableIndex( vpagebase ) ];
        if( ! pointerTableEntry->valid() ) {
            Debug << this 
                  << "::inhibitDataCaching: no pointer table\n"
                  << eor;
            vpagebase += PhysicalPageSize;
            continue;
        }

        // Perform COW of page table entries if parent exists
        if (_parent && (_parent->pointerTable()[pointerTableIndex( vpagebase )] == *pointerTableEntry)) {
            Debug << "Changing inhibit caching in parent pagetable. Performing COW\n" << eor;
            pointerTableEntry->mapToPageTable(allocatePageTable());
            PAGE_TABLE_ENTRY * pageTable = pointerTableEntry->pageTable();
            PAGE_TABLE_ENTRY * parentPageTable = _parent->pointerTable()[pointerTableIndex( vpagebase )].pageTable();
            for(int i = 0; i < PageTableEntries; i++)
                ((unsigned *)pageTable)[i] = ((unsigned *)parentPageTable)[i];
        }

        /*
         * Get a pointer to the proper page table.
         */
        PAGE_TABLE_ENTRY * pageTable = pointerTableEntry->pageTable();
        PAGE_TABLE_ENTRY * pageTableEntry = &pageTable[ pageTableIndex( vpagebase ) ];

        /*
         * If a mapping already exists call the inhibit function on it
         */
        if( pageTableEntry->valid() ) {
            pageTableEntry->inhibitDataCaching();
        }

        vpagebase += PhysicalPageSize;
    }
    
    lock.release();
}

#include "ClassInstaller.h"
DEFINECLASS("TwoLevelPageTable",TwoLevelPageTable,AddressTranslation,0,0);
