#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * MessageFIFO.cc
 *
 */


#include "Assert.h"
#include "OutputStream.h"
#include "Semaphore.h"
#include "ConduitTypes.h"
#include "Word.h"
#include "ZootAddress.h"
#include "ConduitMessage.h"
#include "MessageFIFO.h"


#ifdef __GNUC__
#pragma implementation
#endif

MessageFIFO::~MessageFIFO()
{
	delete msgs;	// deletes pointers to messages, not the messages
			//   themselves; ought to be empty before deleting
}

/*
 * Add an element to the FIFO queue. (Add to tail, get from head....)
 */

Return_Code
MessageFIFO::add( ConduitMessage * msg )
{
	Return_Code ret;

	//Console().OldStyleDebug( "MessageFIFO::add( msg:%x )\n", msg );
	Assert( msg != 0 );
	Assert( msgs != 0 );
	Assert( this != 0 );
	lock.acquire();

	if( head == ( tail + 1 ) % sz ) { // No room at the inn!
		ret = BUFFER_FULL;
	}
	else {
		msgs[tail] = msg;
		tail = (tail + 1) % sz;
		ret = OK;
	}

	lock.release();
	return( ret );
}

/*
 * Remove the first entry from the FIFO queue.
 */
ConduitMessage *
MessageFIFO::remove()
{
	//Console().OldStyleDebug( "MessageFIFO%T::remove()\n", this );
	Assert( this != 0 );

	/*
	 * To save bus contention by repeatedly acquiring and releasing the
	 * lock, make a good guess if there is something to deque before
	 * actually trying to acquire the lock. If it looks empty return 0.
	 * Its up to the caller to keep trying.
	 */
	if( head == tail ) {
		//Console().OldStyleDebug( "MessageFIFO::remove: looks empty\n" );
		return( 0 );
	}

	/*
	 * Lock the queue if we think something is there. (there might not
	 * actually be anything there by the time we get the lock.
	 */
	lock.acquire();

	/*
	 * Double check to make sure someone else didn't get what we thought
	 * was there while we were acquiring the lock.
	 */
	if( head == tail ) {
		lock.release();
		//Console().OldStyleDebug( "MessageFIFO::remove: lost race\n" );
		return( 0 );
	}

	/*
	 * Remove the next element.
	 */
	ConduitMessage * val = msgs[head];
	head = (head + 1) % sz;

	lock.release();

	//Console().OldStyleDebug( "FIFOScheduler::remove: val=%x\n", val );
	return( val );
}

Boolean
MessageFIFO::isEmpty()
{
	if( head == tail ) {
		return( TRUE );
	}
	else {
		return( FALSE );
	}
}

SemaphoreRegulatedMessageFIFO::~SemaphoreRegulatedMessageFIFO()
{
}


Return_Code
SemaphoreRegulatedMessageFIFO::add( ConduitMessage * msg )
{
	//Console().OldStyleDebug( "SemaphoreRegulatedMessageFIFO%T::add( msg:%A )\n", this, msg );
	MessageFIFO::add( msg );
	semaphore.V();
	//Console().OldStyleDebug( "SemaphoreRegulatedMessageFIFO%T::add: return\n" );
	return( Return_Code(0) );
}

ConduitMessage *
SemaphoreRegulatedMessageFIFO::remove()
{
	//Console().OldStyleDebug( "SemaphoreRegulatedMessageFIFO%T::remove()\n", this );
	semaphore.P();
	ConduitMessage * msg = MessageFIFO::remove();
	//Console().OldStyleDebug( "SemaphoreRegulatedMessageFIFO%T::remove: return %A\n", this, msg );
	return( msg );
}

