#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801-2910
 *
 * Copyright (c) 1991 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 *
 * Author: Amit Parghi (parghi@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 *	EthernetDriver.cc
 *
 *	Ethernet-specific, machine-independent stuff for a Driver.
 *
 * $Id: EthernetDriver.cc 2 2004-06-03 01:58:28Z jcarlyle $
 *
 * $Log: EthernetDriver.cc,v $
 * Revision 1.1  2004/06/03 01:58:57  jcarlyle
 * Initial revision
 *
 * Revision 1.2  2004/04/21 22:47:49  dthomas1
 * Compiler fixes
 *   Changed the file to add DLPI support
 * Changed the makefile to reflect changes in vcethernetu
 *
 * Revision 1.1.1.1  2004/03/10 23:02:19  kircher
 *
 *
 * Revision 1.2  1991/12/01  06:04:10  parghi
 * Minor changes to make up-to-date with Stable.9.23.1991.
 * Improved class hierarchy a bit.  Last revision prior to network redesign.
 *
 * Revision 1.1  91/09/25  16:09:35  parghi
 * Initial revision
 * 
 */

#include "Ethernet.h"
#include "EthernetDriver.h"
#include "HashTable.h"
#include "Lock.h"
#include "NetworkBuffer.h"

#ifdef __GNUC__
#pragma implementation
#endif


const int	HashTableSize = 100;


class MulticastEntry : public Link
{
private:
  EthernetAddress	_address;

public:
  MulticastEntry(EthernetAddress *address) { _address = *address; }

  virtual int		compare(Link *entry);
  virtual int		hash(int);
  EthernetAddress	address(void) { return _address; }
};


int
MulticastEntry::compare(Link *entry)
{
  return(_address == ((MulticastEntry *) entry)->address());
}


int
MulticastEntry::hash(int max)
{
  octet		*bytes = _address.bytes();
  int		value = 0;

  for (int i = 0; i < EthernetAddressLength; i++)
    value += bytes[i];

  return(value % max);
}



EthernetDriver::EthernetDriver(unsigned char debugMask,NetworkBufferFreeList* list) 
		: NetworkDriver(debugMask,list)
{	
  _multicastAddresses = new HashTable(HashTableSize);
  _lock = new Lock;
}



EthernetDriver::~EthernetDriver()
{
  delete _multicastAddresses;
  delete _lock;
}


EthernetAddress *
EthernetDriver::getEthernetAddress()	// answer our EthernetAddress
{
    return _ethernetAddress;
}


int
EthernetDriver::hashAddressTo64(EthernetAddress *address)
{
  unsigned char		*addr = address->bytes();
  unsigned int		bit, crc = 0xFFFFFFFF;

  for (bit = 0; bit < 48; bit++) {
    if (!(crc & 0x80000000) != !(addr[bit/8] & (1 << (bit % 8))))
      crc = (crc << 1) ^ 0x04C11DB7;
    else
      crc <<= 1;
  }

  return(((crc & 0x20) >> 5) + ((crc & 0x10) >> 3) + ((crc & 0x08) >> 1) +
         ((crc & 0x04) << 1) + ((crc & 0x02) << 3) + ((crc & 0x01) << 5));
}



/* This routine is called at interrupt level */
int
EthernetDriver::isValidAddress(EthernetAddress *address)
{
  if (address->isBroadcast() || *address == *_ethernetAddress)
    return(TRUE);
  else
  {
    MulticastEntry	multicast(address), *found;
#ifdef NO_USE_UNI
    _lock->acquire();
#endif
    found = (MulticastEntry *) _multicastAddresses->find(&multicast);
#ifdef NO_USE_UNI
    _lock->release();
#endif
    return(found != NULL);
  }
}



int
EthernetDriver::registerMulticast(EthernetAddress *address)
{
  MulticastEntry	*multicast;

  Assert(address->isMulticast());

  multicast = new MulticastEntry(address);

  _lock->acquire();
  _multicastAddresses->insert(multicast);
  _lock->release();

  return(0);
}


int
EthernetDriver::unregisterMulticast(EthernetAddress *address)
{
  MulticastEntry	multicast(address), *found;

  Assert(address->isMulticast());

  _lock->acquire();
  found = (MulticastEntry *) _multicastAddresses->remove(&multicast);
  _lock->release();

  if (found == NULL)
    return(1);
  else
  {
    delete found;
    return(0);
  }
}


#include "Byte.h"
#include "CharacterInputStream.h"
#include "EthernetAdministrator.h"
#include "FreeRunningTimer.h"
#include "FreeRunningTimerStar.h"
#include "HexDump.h"
#include "InputStream.h"
#include "NameServer.h"
#include "NetworkWalker.h"
#include "Semaphore.h"

typedef void	(*PFNB)(NetworkBuffer *);
extern EthernetAdministrator	*_THE_EthernetAdministrator;
extern CharacterInputStreamStar ConsoleInputStream;
static EthernetDriver		*driver;
static Semaphore		*waiter;
static const int		sendTimingType = 0xF00D;
static const int		rcvTimingType = 0xFEED;
static const int		multicastType = 0xCAFE;

class TimingWalker : public NetworkWalker
{
private:
  PFNB		_whatToDo;

public:
  TimingWalker(PFNB func) { _whatToDo = func; }
  virtual void walk(NetworkBuffer *buff) { _whatToDo(buff); }
};


class MulticastWalker : public NetworkWalker
{
public:
  virtual void walk(NetworkBuffer *);
};


void
MulticastWalker::walk(NetworkBuffer *buff)
{
  HexDump *hd = new HexDump(buff->packet(),6,':');
  Debug << "Received multicast packet addressed to:  "
	    << *hd << "\n" << eor;
  buff->release();
}


static void
reflectPacket(NetworkBuffer *buff)
{
  ByteCopy(buff->packet() + 6, buff->packet(), 6);
  buff->packet()[12] = rcvTimingType >> 8;
  buff->packet()[13] = rcvTimingType & 0xFF;
  Debug << "resending packet" << "\n" << eor ;
  driver->transmitPacket(buff);
  buff->release();
}


static void
releaseWaiter(NetworkBuffer *buff)
{
  waiter->V();
  buff->release();
}


static void
startTimingServer()
{
  TimingWalker		*reflect = new TimingWalker(reflectPacket);

  _THE_EthernetAdministrator->registerType(sendTimingType, reflect);
  Console() << "Timing server successfully started\n" << eor;
}


static void
startTimingClient()
{
  TimingWalker		release(releaseWaiter);
  int			i, packets;
  int			measuredTime;
/*
  octet                 packet[100];
  NetworkBuffer         buff(packet, sizeof(packet));
*/

  octet			*packet;
  NetworkBuffer		*buff = driver->getBuffer();
   buff->setLength(64);
  packet = (octet*) buff->packet();

  waiter = new GraciousSemaphore(0);

  Console() << "Please enter the number of packets to send.\n"
	    << "Beware: There is no recovery if a collision occurs!\n" << eor;
  ConsoleInputStream->formattedRead("%d", &packets);

  if (packets < 1 || packets > 1000000)
  {
    Console() << "??\n" << eor;
    return;
  }

  for (i = 0; i < 6; i++)
    packet[i] = 0xFF;

  driver->getEthernetAddress()->copyTo(packet + 6);

  packet[12] = sendTimingType >> 8;
  packet[13] = sendTimingType & 0xFF;

  _THE_EthernetAdministrator->registerType(rcvTimingType, &release);
  Console() << "Timing client successfully started\n" << eor;

   driver->clearStats();
   _THE_EthernetAdministrator->clearStats();

  measuredTime = SystemTimer->time();
  for (i = 0; i < packets; i++)
  {
    driver->transmitPacket(buff);
    // Console() << "encore " << "\n" << eor ;
    waiter->P();
  }
  measuredTime = SystemTimer->time() - measuredTime;

  Console() << measuredTime << " microseconds to send and receive " << packets
	    << " packets\n" << eor;

   driver->printStats();
   _THE_EthernetAdministrator->printStats();

  _THE_EthernetAdministrator->unregisterType(rcvTimingType);
  delete waiter;
}


static void
startMulticastServer()
{
  _THE_EthernetAdministrator->registerType(multicastType, new MulticastWalker);
  Console() << "Multicast server successfully started\n" << eor;
}


static void
readEthernetAddress(EthernetAddress *addr)
{
  int	a, b, c, d, e, f;

  Console() << "Enter address (xx:xx:xx:xx:xx:xx): " << eor;
  ConsoleInputStream->formattedRead("%d:%d:%d:%d", &a, &b, &c, &d);
  ConsoleInputStream->formattedRead(":%d:%d", &e, &f);

  EthernetAddress	temp(a, b, c, d, e, f);
  *addr = temp;
}


void
registerTestMulticast()
{
  EthernetAddress	addr;

  readEthernetAddress(&addr);
  driver->addMulticast(&addr);
}


static void
unregisterTestMulticast()
{
  EthernetAddress	addr;

  readEthernetAddress(&addr);
  driver->removeMulticast(&addr);
}


static void
sendTestPacket()
{
  EthernetAddress	dest;
  octet			packet[64];
  NetworkBuffer		buff(packet, sizeof(packet));

  readEthernetAddress(&dest);

  dest.copyTo(packet);
  driver->getEthernetAddress()->copyTo(packet + 6);
  packet[12] = multicastType >> 8;
  packet[13] = multicastType & 0xFF;

  driver->transmitPacket(&buff);
}

#ifdef SPARC
//#include "SS1Am7990.h"
#endif

/* Commented because of proxiable compile error DCT */ 
void
doEthernetTests(void)
{
  int	command;

  if (driver == NULL)
    driver = (EthernetDriver *) (ProxiableObject *)StandardNameServer->lookup("ethernetDriver", "EthernetDriver");

  if (driver == NULL)
  {
#ifdef SPARC
  //  extern SS1Am7990 *le0;
   // driver = le0;
#endif
    Console() << "Lookup failed!\n" << eor;
  }

  Console() << "Please choose one of:\n"
	    << "  1) Start timing server\n"
	    << "  2) Start timing client (do (1) on another machine first)\n"
	    << "  3) Start multicast server\n"
	    << "  4) Register a multicast address\n"
	    << "  5) Unregister a multicast address\n"
	    << "  6) Send a multicast packet\n\n" << eor;

  ConsoleInputStream->formattedRead("%d", &command);

  switch (command)
  {
  case 1:
    startTimingServer();
    break;

  case 2:
    startTimingClient();
    break;

  case 3:
    startMulticastServer();
    break;

  case 4:
    registerTestMulticast();
    break;

  case 5:
    unregisterTestMulticast();
    break;

  case 6:
    sendTestPacket();
    break;

  default:
    Console() << "??\n" << eor;
  }
}
#include "ClassInstaller.h"
DEFINECLASS("EthernetDriver", EthernetDriver, NetworkDriver, 0, 0);
