#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * ICMPConduit.cc 
 *
 */
/*
	See Conduit.h for much better comments about what all this stuff
	does....
 */

#include "OutputStream.h"
#include "ConduitTypes.h"
#include "Word.h"
#include "CPU.h"
#include "Process.h"
#include "ThisProcess.h"
#include "Semaphore.h"
#include "defZOOT.h"
#include "ZootConfig.h"
#include "ZootReturnCodes.h"
#include "Assert.h"
#include "IPChecksum.h"
#include "ZootAddress.h"
#include "ConduitMessage.h"
#include "ZootClock.h"
#include "ConduitMessageBuffer.h"
#include "MessageFIFO.h"
#include "ZootStateMachine.h"
#include "Conduit.h"


#ifdef __GNUC__
#pragma implementation
#endif

// -------------------- Class ICMPConduit stuff: -------------------
ICMPConduit::ICMPConduit()
{

	my_proto = ICMP;
	_IPConduit = 0;
	Debug << "ICMPConduit() -- my_proto = " << my_proto << "\n" << eor;
	ICMPIncomingMessageFIFO = new SemaphoreRegulatedMessageFIFO( 10 );
		// 10 should be enough. In reality there will not be more than one.
}


ICMPConduit::~ICMPConduit()
{
	delete   ICMPIncomingMessageFIFO;
}


Return_Code
ICMPConduit::connectTo( Conduit * c )
{
	Return_Code ret;

	// Do a double-dispatch with the other Conduit:
	ret = c->connectFromAbove( this, my_proto );
	if( ret != OK )
	{
	    // Undo the connectFromBelow() that c did
	    _IPConduit = 0;
	}
	return( ret );
}

Return_Code
ICMPConduit::connectFromAbove( Conduit * c, ProtocolID proto )
{
	int i = 0;

	// Just need to tell c what kind of Conduit I am, 
	// do not care who is connecting from above. 
	return( c->connectFromBelow( this, my_proto, 0, i ) );
}

Return_Code
ICMPConduit::connectFromBelow( Conduit * c, ProtocolID proto, int his_handy,
				int & my_handy )
{
	Debug << "ICMPConduit::cFB() protocol:" << proto << "\n" << eor;
	if( proto != IP ) return( INVALID_PROTOCOL );

	if( _IPConduit != 0 ) return( CANNOT_CONNECT );
		// Already connected to an IP Conduit....

	_IPConduit = (IPConduit *)c;
	my_handy = ICMP;	 
	return( OK );
}

Return_Code
ICMPConduit::disconnectFrom( Conduit * c )
{
	Return_Code ret;

	if( _IPConduit != c ) return( NO_SUCH_CONDUIT );

	ret = c->disconnectFromAbove( this );
	if( ret == OK ) {
	    _IPConduit = 0;
	}
	return( ret );
}

Return_Code
ICMPConduit::disconnectFromAbove( Conduit * c )
{
	// Nothing to be done here.
	return( OK );
}

extern volatile Conduit * TheIPConduit; // see ZootController.cc

Return_Code
ICMPConduit::insertFromAbove( ConduitMessage * msg, NetworkAddress * addr )
{

	Return_Code ret;

	Assert( msg != 0 );
	Assert( addr != 0 );

	Debug << "ICMPConduit::insertFromAbove( msg:" << msg << " addr:" << addr
		<< "\n" << eor;
	if( ICMP != msg->protocol() ) return( INVALID_MESSAGE_TYPE );

	ICMPMessage * icmpmsg = (ICMPMessage*)msg;
	IPMessage * ipmsg = new IPMessage( 20, 0, icmpmsg );

	// Add IP Headers here to msg
	icmpmsg->setHeader( 20, new octet[ 20 ] );
	icmpmsg->setIPmsg(ipmsg);

	ret = ((Conduit *)TheIPConduit)->insertFromAbove(ipmsg, addr );
	
	return( ret );
}

Return_Code
ICMPConduit::insertFromBelow( ConduitMessage * msg )
{
//   msg should be an IPMessage, with header/data fields
//   set up appropriately.

	Return_Code ret;
	ICMPMessage * icmpmsg;
	int hlen = 4; // ICMP headers are 4 bytes
	int dlen;
	IPMessage * ipmsg = (IPMessage *) msg;

	// Check the Checksum first
	Word16 checksum = IPChecksum(ipmsg->data(), ipmsg->dataLength() );
	if( 0 != checksum.val() ) {
	    ipmsg->del();
	    Debug << "IP Checksum Error!!! (" << asHex( checksum.val() )
			<< ")\n" << eor;
	    return( CHECKSUM_ERROR );
	}

	// Prepare the ICMPMessage
	icmpmsg = new ICMPMessage();
	icmpmsg->setHeader( hlen, ipmsg->data() ); // Get the length right
	dlen = ipmsg->dataLength() - hlen;
	icmpmsg->pleaseDontDeleteHeader();
	icmpmsg->setData( dlen, ( ipmsg->data() + hlen ) );
	icmpmsg->pleaseDontDeleteData();
	
	// Don't foget to tell the IP message who it's carrying:
	ipmsg->setLink( icmpmsg );
	icmpmsg->setIPmsg( ipmsg );	
	ipmsg->setData( 0, 0 );
	ipmsg->pleaseDontDeleteData();

	if (ICMP != msg->protocol()) {
	    icmpmsg->del();	// boo-hoo!
	    Debug << "Unable to deliver packet.\n" << eor;
	    ret = NO_SUCH_CONDUIT;
	}
	else {
		if( ICMPIncomingMessageFIFO != 0 ) {
			((SemaphoreRegulatedMessageFIFO *)ICMPIncomingMessageFIFO)->add( icmpmsg );
			ret = OK;
		}
		else {
		    icmpmsg->del(); 
		    Debug << "IP Cannot Deliver!!!\n" << eor;
		    ret = CANNOT_DELIVER;
		}
	}

	return( ret );

}

