#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * TCPEstablishedState.cc
 *
 */


#include "Semaphore.h"
#include "Assert.h"
#include "ThisProcess.h"
#include "OutputStream.h"
#include "ConduitTypes.h"
#include "Word.h"
#include "ZootConfig.h"
#include "ZootReturnCodes.h"
#include "Assert.h"
#include "ZootAddress.h"
#include "ConduitMessage.h"
#include "ZootClock.h"
#include "Conduit.h"
#include "ConduitMessageBuffer.h"
#include "ZootStateMachine.h"


#ifdef __GNUC__
#pragma implementation
#endif

#ifndef QUIET
#undef Debug
#define Debug thisProcess()->console()
#endif

//-----------------------------------------------------------------------------
//				 EstablishedState
//-----------------------------------------------------------------------------
EstablishedState::EstablishedState()
{
}
EstablishedState::~EstablishedState()
{
}
TCPState *
EstablishedState::closeConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier )
{
	ret = sendCloseRequest( cd, carrier );
	cd.gotoState( FIN_WAIT_1 );
	Debug << "ES::cC entering FIN_WAIT1 state.\n" << eor;
	return( FIN_WAIT_1 );
}
TCPState *
EstablishedState::abortConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier )
{
	MessageID id1, id2;

	// I assume the source-addr and dest-addr have already been
	//   set up, since we are in the ESTABLISHED state.

	// Need to delete any waiting messages:
	if( cd.waiting() != -1 ) {
	    id1 = cd.waiting();
	    id2 = cd.waitingBuffer()->previous( id1 );
	    while( id1 != -1 ) {
		cd.waitingBuffer()->del( id1 );
		id1 = id2;
		id2 = cd.waitingBuffer()->previous( id1 );
	    }
	    cd.linkWaiting( MessageID( -1 ) );
	}

	ret = sendAbort( cd, carrier );

	short unsigned int tmp1 = 0;

	cd.setSourcePort( tmp1 );
	cd.setDestinationPort( tmp1 );
	cd.setDestinationIPAddress( IPAddress() );	// i.e. 0.0.0.0
	cd.gotoState( CLOSED );
	Debug << "ES::aC entering CLOSED state\n" << eor;
	return( CLOSED );
}
TCPState *
EstablishedState::processOutgoingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg )
{
	// Here it is. A really meaty one, so to speak. Actually sending
	//   real live, honest-to-goodness data out.
	MessageID waitingid = -1;
	TCPMessage * waiting = 0;
	Boolean WindowHasEnufRoom = TRUE;

	Assert( carrier != 0 );
	Assert( msg != 0 );

	//Debug << "ES::pOM called (cd:%x, msg:%x)\n", &cd, msg << eor;

	// First, need to check the count in the wait-buffer and send anything
	//   that was queued before we entered the ESTABLISHED state (if the
	//   send-window is big enough):
	if( cd.waitingBuffer()->count() > 0 && cd.waiting() != -1 ) {
	    waitingid = cd.waitingBuffer()->previous( cd.waiting() );
	    waiting = (TCPMessage*)cd.waitingBuffer()->lookAt( waitingid );
	}
	while( waiting != 0 && WindowHasEnufRoom ) {
	    // There are messages waiting to be processed, so we need to chase
	    //   down the linked-list and process them:

	    if( isBetween( cd.SendUnacknowledged().val(),
		    cd.SendNext().val()           + waiting->dataLength(),
		    cd.SendUnacknowledged().val() + cd.SendWindow().val() ) ) {
		// Send the waiting message off:
		cd.pseudoheader()->setLength( waiting->length() );
		ret = sendMessage( cd, carrier, waiting );
		// Snip off from linked list and delete:
		cd.linkWaiting( cd.waitingBuffer()->previous( waitingid ) );
		cd.waitingBuffer()->del( waitingid );
	        waitingid = cd.waitingBuffer()->previous( cd.waiting() );
	        waiting = (TCPMessage*)cd.waitingBuffer()->lookAt( waitingid );
			// lookAt() returns 0 if you try to look at garbage
	    }
	    else {
		// Boo Hoo! Message is too big to be sent right now....
		WindowHasEnufRoom = FALSE;
	    }
	}

	// Now just set up the message and send it. I won't do any re-segment-
	//   izing or anything. Whatever messages get dropped into TCP go out
	//   with their data "as is" -- if segmentizing of any clever sort is
	//   required, it should be done in the application conduit (for now).
	if( isBetween( cd.SendUnacknowledged().val(),
		cd.SendNext().val() + msg->dataLength(),
		cd.SendUnacknowledged().val() + cd.SendWindow().val() ) ) {
	    cd.pseudoheader()->setLength( msg->length() );//Hochwichtig! --JZ
	    ret = sendMessage( cd, carrier, msg );
	}
	else ret = CANNOT_SEND;

	return( ESTABLISHED );	// No change
}
TCPState *
EstablishedState::processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg )
{
	MessageID id1, id2;
	TCPMessage * reply;
	Word32 sequence_number, ack_number;
	int data_length;
	Word16 window;
	Boolean fin_bit_set;

	Assert( carrier != 0 );
	Assert( msg != 0 );

	// Check whether the segment is in the receive window:
	ret = checkIncomingMessage( cd, carrier, msg );
	if( ret == OK ) {
	    if( msg->RST() ) {
		// Connection has been aborted. So blast outstanding msgs, and
		//   enter CLOSED state:
		// Delete outstanding messages:
		id1 = cd.outstanding();
		while( id1 != -1 ) {
		    id2 = cd.outgoingBuffer()->previous( id1 );
		    cd.outgoingBuffer()->del( id1 );
		    id1 = id2;
		}
		msg->del();
		cd.gotoState( CLOSED );
		Debug << "ES::pIM entering CLOSED state.\n" << eor;
		ret = CONNECTION_RESET;

		return( CLOSED );
	    }
	    // Punt security/precedence....
	    if( msg->SYN() ) {
		// Error! Connection reset! (A SYN in the window is bad news.)
		sendAbort( cd, carrier );
		// Delete outstanding messages:
		id1 = cd.outstanding();
		while( id1 != -1 ) {
		    id2 = cd.outgoingBuffer()->previous( id1 );
		    cd.outgoingBuffer()->del( id1 );
		    id1 = id2;
		}
		// Note that we have just deleted the abort-request from
		//   the ARB, so it only will get sent once. In the (hopefully
		//   unlikely) event that it gets lost, the other end will
		//   time out and get bad news about trying to talk on a non-
		//   existent connection....

		msg->del();

		short unsigned tmp1 = 0;

		cd.setSourcePort( tmp1 );
		cd.setDestinationPort( tmp1 );
		cd.setDestinationIPAddress( IPAddress() );	// i.e. 0.0.0.0
		cd.gotoState( CLOSED );
		Debug << "ES::pIM entering CLOSED state...\n" << eor;
		ret = CONNECTION_RESET;

		return( CLOSED );
	    }
	    // Copy important stuff out of message before attempting to
	    //   deliver it to the user (since delivering it may destroy
	    //   it):
	    sequence_number = msg->SequenceNumber();
	    ack_number = msg->AckNumber();
	    data_length = msg->dataLength();
	    window = msg->Window();
	    fin_bit_set = msg->FIN();

	    if( msg->ACK() ) {
	      ret = cd.user()->insertFromBelow( msg );

	      if( ret == OK ) {

		// Want to acknowledge regardless of whether it's the same
		//   sequence number or not (there might be empty packets that
		//   need to be removed from the rexmit-buffer):
		acknowledge( cd, carrier, ack_number );

		// This whole SND.WL1/SND.WL2 business gives me the heebie-
		//   jeebies -- the idea is to update the send window based on
		//   the message (we already checked if it's a retransmission),
		//   but we need to avoid failing to update the window when
		//   the second of two otherwise identical ACKs has a bigger
		//   window (which UNIX does when the process on the other end
		//   eats stuff out of its receive buffer, which ZOOT really
		//   ought to do too, I think, when an expandWindow() call is
		//   made, rather than waiting for a probe....). -JZ

		// I am trying to use the same logic as the Berkeley Reno code
		//   does.  Since isBetween(X,Y,Z) is TRUE for X==Y and Y==Z,
		//   I think the check below will make sure that the segment
		//   just received is not older than the one we used last, but
		//   will accept identical ACKs with bigger windows (a special
		//   case in the Berkeley code that assumes no God-fearing TCP
		//   would gratuitously shrink their window). -JZ

		// Do Send Window stuff:
		if( isBetween( cd.SendWL2().val(),
			       ack_number.val(),
			       cd.SendNext().val() )     &&
		    isBetween( cd.SendWL1().val(),
	    		       sequence_number.val(),
			       cd.ReceiveNext().val() +
		       		cd.ReceiveWindow().val() ) ) {
		    // Update send window:
		    cd.setSendWL1( sequence_number );
		    cd.setSendWL2( ack_number );
		    // If it's a duplicate of the last segment used to update
		    //   the window, only allow increases in window size,
		    //   since identical-looking ACK segments may have been
		    //   reordered:
		    if( cd.SendWL1().val() == sequence_number.val()   &&
			cd.SendWL2().val() == ack_number.val()        &&
			cd.SendWindow().val() < window.val() )  {
	    		cd.setSendWindow( window );
		    }
		    // Otherwise, update the window:
		    if( cd.SendWL1().val() != sequence_number.val()   ||
			cd.SendWL2().val() != ack_number.val() )  {
	    		cd.setSendWindow( window );
		    }
		}

		if( cd.SendUnacknowledged().val() != ack_number.val() ) {
		    // Update acknowledgements received:
		    cd.setSendUnacknowledged( ack_number );
		}

	    	// Punt URG flag and urgent/push nonsense....

		// Process the contents of the message:
		// (1) Update Receive-Next:
		cd.setReceiveNext( cd.ReceiveNext().val() +
					   data_length );

		short unsigned int tmp2 = cd.ReceiveWindow().val() - data_length; 
		cd.setReceiveWindow( tmp2 );


		// Send empty ACK-segment to acknowledge the segment:
		reply = new TCPMessage( 20, 0, 20, 0, 0 );
		cd.pseudoheader()->setLength( 20 );
		sendMessage( cd, carrier, reply );

		// Check FIN bit:
		if( fin_bit_set ) {
			// Send empty segment to acknowledge the FIN:
			reply = new TCPMessage( 20, 0, 20, 0, 0 );
			cd.pseudoheader()->setLength( 20 );
			sendMessage( cd, carrier, reply );
		        cd.setReceiveNext( cd.ReceiveNext().val() + 1 );
		        // (Since the FIN takes up one sequence number)

			cd.gotoState( CLOSE_WAIT );
			Debug << "ES::pIM entering CLOSE_WAIT state.\n" << eor;
			return( CLOSE_WAIT );
		}

		return( ESTABLISHED );

	      } // end of ret==OK check
	    } // end of ack-check
	}
	// Nuke this puppy!
	msg->del();
	return( ESTABLISHED );
}
