#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * TCPFin_Wait_2State.cc
 *
 */


#include "Semaphore.h"
#include "Assert.h"
#include "ThisProcess.h"
#include "OutputStream.h"
#include "ConduitTypes.h"
#include "Word.h"
#include "ZootConfig.h"
#include "ZootReturnCodes.h"
#include "Assert.h"
#include "ZootAddress.h"
#include "ConduitMessage.h"
#include "ZootClock.h"
#include "Conduit.h"
#include "ConduitMessageBuffer.h"
#include "ZootStateMachine.h"


#ifdef __GNUC__
#pragma implementation
#endif

#ifndef QUIET
#undef Debug
#define Debug thisProcess()->console()
#endif

//-----------------------------------------------------------------------------
//				 Fin_Wait_2State
//-----------------------------------------------------------------------------
Fin_Wait_2State::Fin_Wait_2State()
{
}
Fin_Wait_2State::~Fin_Wait_2State()
{
}
TCPState *
Fin_Wait_2State::abortConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier )
{
	// I assume the source-addr and dest-addr have already been
	//   set up, since we are in the FIN-WAIT-2 state.

	ret = sendAbort( cd, carrier );

	short unsigned int tmp1 = 0;

	cd.setSourcePort( tmp1 );
	cd.setDestinationPort( tmp1 );
	cd.setDestinationIPAddress( IPAddress() ); // i.e. 0.0.0.0
	cd.gotoState( CLOSED );
	return( CLOSED );
}
TCPState *
Fin_Wait_2State::processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg )
{
	MessageID id1, id2;
	TCPMessage * reply;
	Word32 sequence_number, ack_number;
	int data_length;
	Word16 window;
	Boolean fin_bit_set;

	Assert( carrier != 0 );
	Assert( msg != 0 );

	// Check whether the segment is in the receive window:
	ret = checkIncomingMessage( cd, carrier, msg );
	if( ret == OK ) {
	    if( msg->RST() ) {
		// Connection has been aborted. So blast outstanding msgs, and
		//   enter CLOSED state:
		// Delete outstanding messages:
		id1 = cd.outstanding();
		while( id1 != -1 ) {
		    id2 = cd.outgoingBuffer()->previous( id1 );
		    cd.outgoingBuffer()->del( id1 );
		    id1 = id2;
		}
		msg->del();
		cd.gotoState( CLOSED );
		ret = CONNECTION_RESET;
		return( CLOSED );
	    }
	    // Punt security/precedence....
	    if( msg->SYN() ) {
		// Error! Connection reset! (A SYN in the window is bad news.)
		sendAbort( cd, carrier );
		// Delete outstanding messages:
		id1 = cd.outstanding();
		while( id1 != -1 ) {
		    id2 = cd.outgoingBuffer()->previous( id1 );
		    cd.outgoingBuffer()->del( id1 );
		    id1 = id2;
		}
		// Note that we have just deleted the abort-request from
		//   the ARB, so it only will get sent once. In the (hopefully
		//   unlikely) event that it gets lost, the other end will
		//   time out and get bad news about trying to talk on a non-
		//   existent connection....

		short unsigned int tmp1 = 0;
		
		msg->del();
		cd.setSourcePort( tmp1 );
		cd.setDestinationPort( tmp1 );
		cd.setDestinationIPAddress( IPAddress() );	// i.e. 0.0.0.0
		cd.gotoState( CLOSED );
		ret = CONNECTION_RESET;
		return( CLOSED );
	    }
	    // Copy important stuff out of message before attempting to
	    //   deliver it to the user (since delivering it may destroy
	    //   it):
	    sequence_number = msg->SequenceNumber();
	    ack_number = msg->AckNumber();
	    data_length = msg->dataLength();
	    window = msg->Window();
	    fin_bit_set = msg->FIN();

	    if( msg->ACK() ) {
	      ret = cd.user()->insertFromBelow( msg );

	      if( ret == OK ) {

		acknowledge( cd, carrier, ack_number );
		if( cd.SendUnacknowledged().val() != ack_number.val() ) {
		    // Update acknowledgements received:
		    cd.setSendUnacknowledged( ack_number );
		    // Update window:
		    // **** RFC 793 is confusing on how this works. I think
		    //   my interpretation is adequate -- this code
		    //   prevents updating the window more than once due
		    //   to a duplicate acknowledgement
		    if(   cd.SendWL1().val() !=
					sequence_number.val()   ||
		          cd.SendWL2().val() !=
					ack_number.val() )  {
		        cd.setSendWL1( sequence_number );
		        cd.setSendWL2( ack_number );
		        cd.setSendWindow( window );
		    }
		}
		if( msg->AckNumber().val() == cd.SendNext().val() ) {
		    // This segment is acknowledging our FIN:
		    cd.gotoState( FIN_WAIT_2 );
		}
	    	// Punt URG flag and urgent/push nonsense....

		// Process the contents of the message:
		// (1) Update Receive-Next:
		cd.setReceiveNext( cd.ReceiveNext().val() +
					   data_length );

		short unsigned int tmp2 = cd.ReceiveWindow().val() - data_length;  
		cd.setReceiveWindow( tmp2 );

		// Check FIN bit:
		if( fin_bit_set ) {
		    // Send empty segment to acknowledge the FIN:
		    reply = new TCPMessage( 20, 0, 20, 0, 0 );
		    cd.pseudoheader()->setLength( 20 );
		    sendMessage( cd, carrier, reply );
		    cd.setReceiveNext( cd.ReceiveNext().val() + 1 );
		    // (Since the FIN takes up one sequence number)

		    cd.gotoState( TIME_WAIT );
		    cd.setTimestamp( now32() );
		    // Delete everything from the rexmit buffer:
		    cd.outgoingBuffer()->reset();
		    cd.linkOutstanding( -1 );
		    return( TIME_WAIT );
		}
		return( FIN_WAIT_2 ); // No change.

	      } // end of ret==OK check
	    } // end of ack-check
	}
	// Nuke this puppy!
	msg->del();
	return( FIN_WAIT_2 );
}

