#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * TCPSyn_SentState.cc
 *
 */


#include "Semaphore.h"
#include "Assert.h"
#include "ThisProcess.h"
#include "OutputStream.h"
#include "ConduitTypes.h"
#include "Word.h"
#include "ZootConfig.h"
#include "ZootReturnCodes.h"
#include "Assert.h"
#include "ZootAddress.h"
#include "ConduitMessage.h"
#include "IPChecksum.h"
#include "ZootClock.h"
#include "Conduit.h"
#include "ConduitMessageBuffer.h"
#include "ZootStateMachine.h"


#ifdef __GNUC__
#pragma interface
#endif

#ifndef QUIET
#undef Debug
#define Debug thisProcess()->console()
#endif
//-----------------------------------------------------------------------------
//				  Syn_SentState
//-----------------------------------------------------------------------------
Syn_SentState::Syn_SentState()
{
}
Syn_SentState::~Syn_SentState()
{
}
TCPState *
Syn_SentState::closeConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier )
{
	MessageID id1, id2;

	// Need to delete any waiting messages:
	if( cd.waiting() != -1 ) {
	    id1 = cd.waiting();
	    id2 = cd.waitingBuffer()->previous( id1 );
	    while( id1 != -1 ) {
		cd.waitingBuffer()->del( id1 );
		id1 = id2;
		id2 = cd.waitingBuffer()->previous( id1 );
	    }
	}
	short unsigned int tmp1 = 0;

	cd.setSourcePort( tmp1 );
	cd.gotoState( CLOSED );
	Debug << "S_SS:cC entering CLOSED state.\n" << eor;
	ret = OK;
	return( CLOSED );
}
TCPState *
Syn_SentState::abortConnection( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier )
{
	// I assume the source-addr and dest-addr have already been
	//   set up, since we are in the SYN-SENT state.

	ret = sendAbort( cd, carrier );
	
	short unsigned int tmp1 = 0;

	cd.setSourcePort( tmp1 );
	cd.setDestinationPort( tmp1 );
	cd.setDestinationIPAddress( IPAddress() ); // i.e. 0.0.0.0
	cd.gotoState( CLOSED );
	Debug << "S_SS:aC entering CLOSED state.\n" << eor;
	return( CLOSED );
}
TCPState *
Syn_SentState::processOutgoingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg )
{
	MessageID tmp;

	Assert( carrier != 0 );
	Assert( msg != 0 );

	// Queue up msg to be sent when we get to ESTABLISHED state:
	//   (Need to link it into the list of waiting messages.)
	Debug << "Syn_State::processOutgoingMessage(..) called. About to queue msg in wait-buffer:\n" << eor;
	// Look up old link:
	tmp = cd.waiting();
	cd.linkWaiting( cd.waitingBuffer()->insert( ret, msg, 0 ) );
	Debug << "S_SS::pOM: ret=" << Return_Code_Table[-ret] << ", link=" <<
		tmp << ", waitingmsg=" << cd.waiting() << "\n" << eor;
	if( ret != OK ) cd.linkWaiting( tmp );
	return( SYN_SENT );	// No Change
}
TCPState *
Syn_SentState::processIncomingMessage( Return_Code & ret,
				TCPConnectionDescriptor & cd,
				TCPConduit * carrier, TCPMessage * msg )
{
	IPAddress * replyAddr;
	TCPMessage * reply;
	MessageID tmp;

	Assert( carrier != 0 );
	Assert( msg != 0 );

	if( msg->ACK() ) {
	    // Check if the Ack is in an acceptable range, else reset:
	    if( isBetween( cd.SendInitialSequenceNumber().val(),
				msg->AckNumber().val(),
				cd.SendNext().val() ) ) {
		// Is groovy, dude (so fall through to rest of processing)
	    }
	    else {
		// Send a reset for this uncool ACK (unless it is itself a
		//   reset, in which case it can't be for this connection so
		//   just ignore the fucker):
		if( msg->RST() == FALSE ) {
	    	    ret = sendAbort( cd, carrier, msg->AckNumber() );
		}
		else {
		    // No reset, but ungroovy ACK -- so just ignore
		    ret = OK;
		}
    	    msg->del();
	    return( SYN_SENT );	// No change
	    }
	}	// End of ack-check
	if( msg->RST() ) {
	    // Just checked ACK, so this must be an acceptable request to abort
	    //   this connection (it was refused, actually):
	    ret = CONNECTION_RESET;

	    short unsigned int tmp1 = 0;

	    cd.setSourcePort( tmp1 );
	    cd.setDestinationPort( tmp1 );
	    cd.gotoState( CLOSED );
	    Debug << "S_SS:pIM entering CLOSED state.\n" << eor;
	    msg->del();
	    return( CLOSED );
	}
	else {
	    // Not a reset-request, and ack is groovy or nonexistent.
	    // I'll ignore security/precedence, and proceed with rest of
	    //   processing....
	}
	if( msg->SYN() ) {
	    // Alles in Ordnung (see RFC 793 p.67)
	    cd.setReceiveNext(
		msg->SequenceNumber().val() + msg->dataLength() + 1 );
		// since SYN-segment could carry data, conceivably
	    cd.setReceiveInitialSequenceNumber( msg->SequenceNumber() );
	    if( msg->ACK() ) {
		cd.setSendUnacknowledged( msg->AckNumber() );
		// Initialize the window:
		cd.setSendWindow( msg->Window() );
		cd.setSendWL1( msg->SequenceNumber() );
		cd.setSendWL2( msg->AckNumber() );

		// Need to check retransmit buffer and kill off anything
		//   that has just been acknowledged:
		acknowledge( cd, carrier, msg->AckNumber() );
	    }
	    if( cd.SendUnacknowledged().val() !=
			cd.SendInitialSequenceNumber().val() ) {
		// Our SYN segment has been acked, so complete the handshake
		//   and go to ESTABLISHED state:
		//   (Note how sending a message with no data ack's the right
		//   stuff and everything....)
	    	reply = new TCPMessage(  20,	// IP-header size
					 0,		// IP data size
					 20,	// TCP header size
					 0,		// TCP data size
					 0 );	// No payload
		cd.pseudoheader()->setLength( 20 );
		sendMessage( cd, carrier, reply );

		cd.gotoState( ESTABLISHED );
		Debug << "S_SS::pIM entering ESTABLISHED state\n" << eor;
		msg->del();
		return( ESTABLISHED );
	    }
	    // Connect-requests must have crossed. So enter SYN_RECEIVED and
	    //   acknowledge:
	    replyAddr = new IPAddress;
	    replyAddr->equals( cd.destinationIPAddress().val() );
	    reply = new TCPMessage(     20,	// IP-header size
					0,		// IP data size
					20,	// TCP header size
					0,		// TCP data size
					0 );	// No payload
	    cd.pseudoheader()->setLength( 20 );
	    sendMessage( cd, carrier, reply );

	    // This is a stupid place to need to know the Local IP Address:
	    reply->setSourceAddress(
		TheLocalIPAddress.val() );
	    reply->setDestinationAddress( cd.destinationIPAddress().val() );
	    reply->setSourcePort( cd.sourcePort() );
	    reply->setDestinationPort( cd.destinationPort() );
	    reply->setSequenceNumber( cd.SendInitialSequenceNumber() );
	    reply->setAckNumber( cd.ReceiveNext() );
	    reply->setACK( TRUE );
	    reply->setSYN( TRUE );

	    // Calculate Checksum:
	    reply->setChecksum( combineIPChecksums(
			IPChecksum( reply->header(), 20 ),
			cd.pseudoheader()->checksum() ) );
	    Debug << "S_SS::pIM: Connect-Request Ack Ready:\n" << eor;
	    //reply->print();

	    tmp = cd.outstanding();
	    cd.linkOutstanding( cd.outgoingBuffer()->
	    		insert( ret, reply->IPmsg(), replyAddr ) );
	    if( ret == BUFFER_FULL ) {
		// restore old link (since the buffer just ignored 
		//   this insert-attempt):
		cd.linkOutstanding( tmp );
	    }
	    cd.gotoState( SYN_RECEIVED );
	    Debug << "S_SS:pIM entering SYN_RECEIVED state.\n" << eor;
	    msg->del();
	    return( SYN_RECEIVED );
	} // end of syn-check
	// Garbage. Ignore.
	msg->del();
	return( SYN_SENT );	// No change.
}
