#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Johnny Zweig (zweig@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * UDPConduit.cc 
 *
 */
/*
	See (UDP)Conduit.h for much better comments about what all this stuff
	does....
 */

#include "Assert.h"
#include "Semaphore.h"
#include "ConduitTypes.h"
#include "Word.h"
#include "UDPConduit.h"
#include "ZootReturnCodes.h"
#include "ZootAddress.h"
#include "ConduitMessage.h"
#include "IPChecksum.h"


#ifdef __GNUC__
#pragma implementation
#endif

#define put(x)  Debug << x << eor;

//---------------------------- Inner Guts stuff -----------------------------
//  The class UDPHashTable is just a useful little hack for implementing
//  UDP's demultiplexor, since I couldn't figure out Aamod's silly general-
//  purpose hash tables, and I think that adding the two bytes together
//  and using a 256-entry hash table is the Right Thing for UDP.  -JZ

class UDPHashTable {

	struct ConduitLinkedListEntry {
		int16 port_number;
		Conduit * c;
		ConduitLinkedListEntry * next;
	};

	ConduitLinkedListEntry * _HashTable[256];

protected:
	int hash( int16 port );
public:
	UDPHashTable();
	~UDPHashTable();
	Return_Code	add( Conduit * conduit, int16 port );
	Conduit *	lookup( int16 port );
	Return_Code	remove( int16 port );
		// I assume only one entry/port, so remove() doesn't need to
		//   know which conduit to delete. add() fails iff someone is
		//   already listening to that port.
};

inline
UDPHashTable::UDPHashTable()
{
	int i;

	for( i = 0; i < 256; i++ ) _HashTable[i] = 0;
}

inline
UDPHashTable::~UDPHashTable()
{
	int i;

	for( i = 0; i < 256; i++ ) if( _HashTable[i] != 0 )
		delete _HashTable[i];
}
inline int
UDPHashTable::hash( int16 port )
{
	// folding seems as good a way as any to map 16-bit port numbers to
	//   8-bit table offsets.
	return( ( (port >> 8) & 0x00ff ) + (port & 0x00ff) );
}

Return_Code
UDPHashTable::add( Conduit * conduit, int16 port )
{
	// This code assumes that you already made sure the port is not
	//   already in use.  Also, mutual exclusion must be provided
	//   by the UDP ConduitMonster that uses this hash table!

	// Also -- notice the utter lack of mutual exclusion! This is
	//   somebody else's concern!

	ConduitLinkedListEntry * bucket;
	ConduitLinkedListEntry * entry = new ConduitLinkedListEntry;
	entry->c = conduit;
	entry->port_number = port;

	bucket = _HashTable[ hash( port ) ];
	entry->next = bucket;
	_HashTable[ hash( port ) ] = entry;

	return( OK );
}

Conduit *
UDPHashTable::lookup( int16 port )
{
	ConduitLinkedListEntry * bucket;
	Conduit * conduit = 0;

	bucket = _HashTable[ hash( port ) ];
	while( 0 != bucket && bucket->port_number != port )
		bucket = bucket->next;

	if( 0 != bucket && bucket->port_number == port )
		conduit = bucket->c;
	
	return( conduit );
}

Return_Code
UDPHashTable::remove( int16 port )
{
	ConduitLinkedListEntry * bucket;
	ConduitLinkedListEntry * previous = 0;
	Return_Code ret;

	bucket = _HashTable[ hash( port ) ];
	while( 0 != bucket && bucket->port_number != port ) {
		previous = bucket;
		bucket = bucket->next;
	}

	// loop exit invariant: bucket==0 or bucket->port_number==port

	ret = ERROR; // should never be returned, if I have my logic right

	if( 0 == bucket ) ret = ILLEGAL_PORT; // NOT_FOUND might be better

	if( 0 == previous && 0 != bucket && bucket->port_number == port ) {
		// the desired link is at the head of the linked list
		_HashTable[ hash( port ) ] = bucket->next;
		delete bucket;
		ret = OK;
	}

	if( 0 != previous && 0 != bucket && bucket->port_number == port ) {
		// the desired link is not at the head of the list
		previous->next = bucket->next;
		delete bucket;
		ret = OK;
	}

	return( ret );
}

//----------------------------------------------------------------------------
UDPPortConduit::UDPPortConduit( int16 port, UDPPortMonster * monster,
				Conduit * udp )
{
put("_UPC0_");
	_port = port;
	_monster = monster;
	_UDP = udp;
	_user = 0;
}
UDPPortConduit::~UDPPortConduit()
{
	_monster->releasePort( _port );
}
Return_Code
UDPPortConduit::connectFromAbove( Conduit * c, ProtocolID proto )
{
put("_UPC1_");
	// Ignore protocol number -- just set pointer and keep
	// our fingers crossed that this fellow wants to get UDP messages
	// Just for grins, pass the port number as a handy number....
	int handy;
	Return_Code ret = c->connectFromBelow( this, UDP, _port, handy );
	if( OK == ret ) {
	    _user = c;
put("_UPC2_");
	}
	return( ret );	// this way the person connecting might be able to
			// figure out what's wrong if we fail....
}
Return_Code
UDPPortConduit::insertFromAbove( ConduitMessage * msg, NetworkAddress * addr)
{
	IPMessage * ipm;
	UDPAddress * a = (UDPAddress*)addr;
	Word16 temp; // for holding temporary checksum as it is calculated
	ConduitMessage * next_piece;	// if msg->link() is nonzero during
					//   the checksum computation

put("_UPC3_");
	// Make sure it really is a UDP message:
	if( UDP != msg->protocol() ) return( INVALID_MESSAGE_TYPE );

	UDPMessage * m = (UDPMessage *)msg;

	// Check whether (a) m has an IP-message, or (b) addr is nonzero:
	if( 0 == m->holder() ) {
	    if( 0 == addr ) {
		// no way to figure out who the hell the message is for!
		return( INCOMPLETE_ADDRESS );
	    }
	    else {
		// cook up an IP address for m to live in, and copy addr:
		ipm = new IPMessage( 20, 0, m );
		ipm->setPROTO( 17 ); // IP protocol field value
		ipm->setDestADDR( a->destinationIPAddress() );
		ipm->setSourceADDR( TheLocalIPAddress );
		m->setSourcePort( a->sourcePort() );
		m->setDestinationPort( a->destinationPort() );
	    }
	}
	else {
	    if( IP != m->holder()->protocol() ) {
		return( INVALID_ENCAPSULATION );
	    }
	    else {
		// make sure IP message is set up correctly:
		ipm = (IPMessage*)m->holder();
		ipm->setPROTO( 17 ); // IP protocol field value
		ipm->setSourceADDR( TheLocalIPAddress );
	    }
	    if( 0 != addr ) {
		ipm->setDestADDR( a->destinationIPAddress() );
		m->setSourcePort( a->sourcePort() );
		m->setDestinationPort( a->destinationPort() );
	    }
	}
	ipm = (IPMessage*)m->holder();	// should already be true -- this
					//   just shuts the compiler up

	// do checksum if m->Checksum().val() is nonzero
	if( 0 != m->Checksum().val() ) {
	    // Create a UDP pseudoheader, using m's dest-IP-address, ports:
	    octet pseudo[12];
	    ipm->SourceADDR().copyTo( &pseudo[0] );
	    ipm->DestADDR().copyTo( &pseudo[4] );
	    pseudo[9] = 17;
	    m->Length().copyTo( &pseudo[10] );

	    // calculate checksum for pseudoheader:
	    temp = IPChecksum( &pseudo[0], 12 );

	    // combine with checksums for the rest of m's pieces:
	    temp = CombineIPChecksums( temp,
			IPChecksum( m->header(), m->headerLength() ) );
	    temp = CombineIPChecksums( temp,
			IPChecksum( m->data(), m->dataLength() ) );
	    next_piece = m->link();
	    while( 0 != next_piece ) {
		temp = CombineIPChecksums(
			temp,
			IPChecksum( next_piece->header(),
				    next_piece->headerLength() ) );
		temp = CombineIPChecksums(
			temp,
			IPChecksum( next_piece->data(),
				    next_piece->dataLength() ) );
	    }

	    // fill in m's UDP checksum field:
	    m->setChecksum( temp );
	}

	// copy address out of addr if it's nonzero, since UDP will ignore it

	return( _UDP->insertFromAbove( msg, 0 ) );
}
Return_Code
UDPPortConduit::insertFromBelow( ConduitMessage * msg )
{
put("_UPC4_");
	if( 0 == _user ) {
	    // no way to process message until user connects, so toss it
	    delete msg;
	    return( NO_SUCH_CONDUIT );	// this should almost never happen
	}
	else {
	    return( _user->insertFromBelow( msg ) );
	}
}
Return_Code
UDPPortConduit::connectTo( Conduit * c )
{
	return( ILLEGAL_OPERATION );
}
Return_Code
UDPPortConduit::connectFromBelow( Conduit *c, ProtocolID proto,
			int his_handy, int & my_handy )
{
	return( ILLEGAL_OPERATION );
}
Return_Code
UDPPortConduit::disconnectFrom( Conduit * c )
{
	return( ILLEGAL_OPERATION );
}
Return_Code
UDPPortConduit::disconnectFromAbove( Conduit * c )
{
	return( ILLEGAL_OPERATION );
}

//----------------------------------------------------------------------------
UDPPortMonster::UDPPortMonster()
{
put("_UA_")
	_mutex = new MutexSemaphore;	// defaults to count 1, right?
	_PortTable = new UDPHashTable;
	_UDP = new UDPConduit( this );
	extern volatile Conduit * TheIPConduit; // see ZootController.cc
	Assert( 0 != TheIPConduit );
	if( OK != _UDP->connectTo( (Conduit*)TheIPConduit ) )
		    Console() << "UDP sad\n" << eor;
put("_UB_")
}
UDPPortMonster::~UDPPortMonster()
{
	delete _PortTable;
	delete _mutex;
	extern Conduit * TheIPConduit; // see ZootController.cc
	Assert( OK == _UDP->disconnectFrom( TheIPConduit ) );
	delete _UDP;
}
Return_Code
UDPPortMonster::getPort( Conduit *& port_conduit, int16 port )
{
	// This function returns a UDPPortConduit which is connected to the
	//   requested port.  When the user disconnects from the port conduit,
	//   the entry in the hash table is zotzed and the port conduit is
	//   deleted.

	Return_Code ret;
	Conduit * c = 0;

	_mutex->P();
	if( 0 == _PortTable->lookup( port ) ) {
		c = new UDPPortConduit( port, this, _UDP );
		_PortTable->add( c, port );
		port_conduit = c;
		ret = OK;
	}
	else ret = ILLEGAL_PORT;
	_mutex->V();
	return( ret );
}
void
UDPPortMonster::releasePort( int16 port )
{
	Conduit * c;
	Return_Code UDP_port_table_return_code;

	_mutex->P();
	c = _PortTable->lookup( port );
	UDP_port_table_return_code = _PortTable->remove( port );
	_mutex->V();

	Assert( OK == UDP_port_table_return_code ); // for testing.

}
Conduit *
UDPPortMonster::map( int16 port )
{
	return( _PortTable->lookup( port ) );
}
//----------------------------------------------------------------------------
//                         Class UDPConduit stuff:
//----------------------------------------------------------------------------
UDPConduit::UDPConduit( UDPPortMonster * monster )
{
put("_U-1_")
	_monster = monster;
		// Monster has the demux table
put("_U0_")
	_IPConduit = 0;
		// hope to get connected to IP soon
}
UDPConduit::~UDPConduit()
{
	delete _monster;
}
Return_Code
UDPConduit::connectTo( Conduit * c )
{
	// c had better be an IPConduit, no?  So let's get cracking:
	_IPConduit = (IPConduit *)c;	// This is optimistic, I suppose.
	return ( c->connectFromAbove( this, UDP ) );
}
Return_Code
UDPConduit::connectFromBelow( Conduit * c, ProtocolID proto, int his_handy,
				int & my_handy )
{
put("_U1_")
	// c had better be the IPConduit to which I am currently being told
	//   to connect. Double-check that it is (notice that I already
	//   set _IPConduit during connectTo):
	if( _IPConduit != c ) return ( CANNOT_CONNECT );
put("_U2_")

	if( IP != proto ) return ( INVALID_PROTOCOL );
put("_U3_")

	// ignore his_handy
	my_handy = 17;  // UDP's IP-protocol number, see RFC1060
	return( OK );
}
Return_Code
UDPConduit::disconnectFrom( Conduit * c )
{
	Assert( _IPConduit == c );
	return c->disconnectFromAbove( this );
}
Return_Code
UDPConduit::insertFromAbove( ConduitMessage * msg, NetworkAddress * addr )
{
	IPMessage * ipmsg;
	UDPMessage * udpmsg = (UDPMessage*)msg; // casts are evil!
	UDPAddress * address = (UDPAddress *)addr; // casts are evil!

Assert( NOTREACHED ); // for testing
/*******************
	// Cook up a UDPMessage whose IPMessage has a header-length of 28
	//   octets, and set the UDP header to point to octet 20 of that
	//   array.

	udpmsg = new UDPMessage(	28,	// IP header array size
					0,	// IP data array size,
					0,	// UDP header array size
					0,	// UDP data array size
					msg );	// msg is the payload
	udpmsg->setheader( 8, ipmsg->header() + 20 );
	udpmsg->pleaseDontDeleteHeader();

	address->sourcePort().copyTo( ipmsg->data() );
	address->destinationPort().copyTo( ipmsg->data() + 2 );
	Word16 length = msg->length() + 8;
	length.copyTo( ipmsg->data() + 4 );
*******************/
	ipmsg = (IPMessage *) udpmsg->holder(); // PortConduit already checked

	// Checksum should have been done by the PortConduit, if necessary...

	return( _IPConduit->insertFromAbove( ipmsg,
				& address->destinationIPAddress() ) );
}
Return_Code
UDPConduit::insertFromBelow( ConduitMessage * msg )
{
	Conduit * c;
	Word16 destPort;
	UDPMessage * udpmsg;
	Return_Code ret = ERROR;

	// Assume that msg is an IP message
	udpmsg = new UDPMessage;

	// Check port number:
	destPort.copyFrom( msg->data() + 2 );
	c = _monster->map( destPort.val() );

	if( 0 != c ) ret = c->insertFromBelow( udpmsg );
	else ret = ILLEGAL_PORT; // is this a sensible thing to conclude?
	return( ret );
}
Return_Code
UDPConduit::connectFromAbove( Conduit * c, ProtocolID proto )
{
	return( ERROR ); // attach via Monster -- don't connect
}
Return_Code
UDPConduit::disconnectFromAbove( Conduit * c )
{
	return( ERROR ); // detach via Monster -- don't disconnect
}

