#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *     University of Illinois at Urbana-Champaign
 *     Department of Computer Science
 *     1304 W. Springfield Ave.
 *     Urbana, IL 61801
 *
 * Copyright (c) 1987-1992
 *    The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * ARMCPU.c: ARM CPU class implementation.
 */

#include "Assert.h"
#include "MACHINE_GDB.h"
#include "OutputStream.h"
#include "PROCESSOR_CONTEXT.h"
#include "PROM.h"
#include "ARMCPU.h"
#include "ARMInterrupts.h"
#include "ThisProcess.h"
#include "Process.h"
#include "TIMER_MANAGER.h"
#include "TIMESLICE_TIMER.h"
#include "TimerManager.h"
#include "Vectors.h"
#include "ThisCPU.h"
#include "MACHINE_CONFIGURATION.h"
#include "Kernel.h"
#include "AppInterface.h"

#ifdef __GNUC__
#pragma implementation
#endif

class ProcessContainer;
class Interrupt;

#ifdef INSTRUMENT
#include "SystemInstrumentManager.h"
#endif

/*
 * Halt the CPU. This code probably doesn't belong here but I can't think of
 * any better place to put it.
 */
void
Halt()
{
    /*
     * Halt CPU via infinite loop.
     */
    Console().criticalPrintf( "Halting.\n" );
    Console().flush(); 

    while(1);
}

void
ARMCPU::debugBreakpoint()
{
    TrapToDebugger();
}

static int ARMCPUClass_Initialized = 0;

ARMCPU::ARMCPU( unsigned long identifier, ProcessContainer * iic )
    : CPU( identifier, iic )
{
    Debug << "ARMCPU::ARMCPU()\n" << eor;

    /*
     * We know this constructor is called sequentially (not in parallel)
     * by all the processors, therefore this doesnt need locked.
     */
    Debug << "ARMCPU::ARMCPU() Clearing interrupt vectors: " << __LINE__ << "\n" << eor;

    if( ! ARMCPUClass_Initialized ) {
        ARMCPUClass_Initialized = 1;
    }

    /*
     * Set the MMU pointer and other.
     */
    _MMU = &myMMU;
    _timesliceTimer = new TIMESLICETIMER();
}


#ifdef GDB
extern "C"
void
ARMTrapSetup()
{
    /*
     * This is what the trap setup code puts as the return address 
     * in hardwareInterruptAssist so that the debugger can "see" 
     *  beyond the trap frame.
     */
}
#endif

extern void printStackFrame(char *);

/*
 * This is called directly from the assembler interruptAssist.
 * It is declare as 'static' in the class definition so it does not get
 *       a 'this' pointer.
 */
// TODO Francis : Inline for efficiency
void
ARMCPU::hardwareInterruptAssist(int vector, ARMInterruptContext *context)
{
    Assert( (vector >= 0) && (vector < nInterrupts) );
    Assert(_interruptManager != 0);

    switch(vector)
    {
        case 1:
            // UNDEFINED INSTRUCTION
            _interruptManager->interruptReceived(1, context);
            break;
      
        case 2:
        {
            // SWI
            // We'll probably want an SWI dispatcher to send the syscall 
            // to the proper object/domain soon
            volatile unsigned *regs = context->getRegisters();        
            volatile unsigned addr = regs[ARMCPU::PC];
            volatile unsigned swi = *((unsigned*)(addr-4))&0xfff;
            //Console() << "SWI at instruction at: " << asHex(addr) << " is " << asHex(swi) << "\n" << eor;
	    Console() << eor;

            switch(swi)
            {
                case 0xfff:
                {
                    // Privilege escalation - TODO Need to verify credentials
                    context->setMode(ARM_SYS);
                    break;
                }
                case 0xffe:
                    // print
                    Console() << (char*)(regs[ARMCPU::R0]) << eor;
                    break;
                case 0xffd:
                {
                    // call function
                    PFV foo = (PFV) regs[ARMCPU::R0];
                    (foo)();

                    break;
                }
                case 0xffc:
                    // die requires ARM_SYS mode
                    context->setMode(ARM_SYS);
                    // update the context's PC to point to the handle for die
                    regs[ARMCPU::PC] = (unsigned)&ARMGenericContext::doExit;

                    break;
                default:
                    if((swi & 0xf00) == 0xf00){
                        register unsigned classID = swi & 0x0ff;
                        register unsigned methodID = regs[ARMCPU::R0];
                        register void * kernelObj = (void *)(regs[ARMCPU::R1]);
                        register void * inputParam = (void *)(regs[ARMCPU::R2]);
                        register void * outputParam = (void *)(regs[ARMCPU::R3]);
                        //Console() << "ARM system call. ClassID: " << classID << ", MethodID: "
                        //      << methodID << '\n'<<  eor;
                        regs[ARMCPU::R0] = TheAppInterface->dispatch(classID, methodID, kernelObj,
                                                  inputParam, outputParam);
                    }
                    else
                        regs[ARMCPU::R0] = 0x123456;
            }

            break;
        }
        case 3:
            // PREFETCH ABORT
            _interruptManager->interruptReceived(3, context);
            // This is also used for GDB bkpt instructions
            if (Gdb::causedInterrupt(vector, context)) {
                GdbHandler(vector, context);
            }
            break;
      
        case 4:
            // DATA ABORT
            _interruptManager->interruptReceived(4, context);
            break;

        case 6:
            // IRQ
            _interruptManager->interruptReceived(context);
            break;

        case 7:
            // FIQ
            _interruptManager->interruptReceived(7,context);
            break;

        default:
            Debug << "Spurious interrupt...\n" << eor;
    }

    return;


    char *info;
#ifdef DEBUG
    static unsigned traps;
    const unsigned skipDEBUG = 2;
    if ((traps++ % skipDEBUG) == 0)
        Debug << "HWInterruptAssist(" << vector << ", " <<
        asAddress(info) << "\n" << eor;
#endif

#ifdef FRANCIS
#ifdef GDB
    unsigned regi7;         // Fake for debuggers
    asm volatile("mov %%i7,%0" : "=r"(regi7));
    asm volatile("mov %0,%%i7" : : "r"(&ARMTrapSetup) );
#endif
#endif

    switch (vector) {
#ifdef FRANCIS
#ifdef GDB
        case VEC_BPT1:
        case VEC_BPT2:
            IntStorage state ; PROCESSOR::disableInterrupts(state);
            ARMGdbHandler(vector, info);
            PROCESSOR::restoreInterrupts(state);
            break;
#endif
#endif


        default:
#ifdef INFCHECK
            /* 
             * Check for infinite traps
             */
            const int infLimit = 200;
            static int lastvector;
            static char *lastinfo;
            static unsigned lastPC;
            static unsigned infCount;
            if (vector == lastvector && info == lastinfo 
                && ((unsigned*)info)[1] == lastPC) {
                if (infCount++ > infLimit){
                    Console().printf(
                        "Infinte trap #%d at 0x%x ???\n",
                        vector, ((unsigned*)info)[1]);
                    Console().printf("TrapToDebugger()\n");
                    Console().flush();
                    TrapToDebugger();
                    lastvector = vector;
                    lastinfo = info;
                    lastPC = ((unsigned *)info)[1];
                    infCount = 0;
                }
            }
            else {
                lastvector = vector;
                lastinfo = info;
                lastPC = ((unsigned *)info)[1];
                infCount = 0;
            }
#endif

            Interrupt * interrupt = 0;
//        ((ARMCPU *) thisCPU())->interruptTable[ vector ];

            if (interrupt) {
                // Debug << "hardwareInterruptAssist( " << vector << ", " << asAddress(info) << "), interrupt( " << asAddress( interrupt ) << " )\n" << eor;
                interrupt->raise( info );
            }
            else
            {
                Console() << "SPURIOUS TRAP # " << vector << "\n" << eor;
                Assert (NOTREACHED);
            }

#ifdef DEBUG
            if ((traps % skipDEBUG) == 0)
                Debug << "HWInterruptAssist( " << vector << ", " << asAddress(info)
                //  << " ) returned, level:" << asHex(interruptLevel()) 
                      << "\n" << eor;
#endif

            break;
    };

#ifdef FRANCIS
#ifdef GDB
    asm volatile("mov %0,%%i7" : : "r"(regi7));
#endif
#endif
}


/*
 * Install an interrupt object to manage a vectored interrupt (trap or
 * interrupt) from the hardware.
 */
Interrupt *
ARMCPU::setInterrupt( int vector, Interrupt * interrupt, PFV assist  )
{
    Debug << "ARMCPU::setInterrupt(): Setting interrupt #" << vector << "\n" << eor;

    return (_interruptManager->setInterrupt(vector, interrupt));
}

/*
 * ARMCPU Class destructor.
 * Obvious side effect is to halt the processor. I doubt this could ever
 * actually be called!
 */
ARMCPU::~ARMCPU()
{
    Debug << "CPU::~CPU()\n" << eor;
    Halt();
}

ARMUndefinedInstructionInterrupt  *undefinedInstructionInterrupt;
ARMDataAccessInterrupt *dataAccessInterrupt;
ARMInstructionAccessInterrupt *instructionAccessInterrupt;
ARMProxyInterrupt     *proxyInterrupt;
MACHINEFREERUNNINGTIMERINTERRUPT *freeRunningTimerInterrupt;
TimerManagerInterrupt *timerManagerInterrupt;

void
ARMCPU::installInterrupts()
{
    Debug << "ARMCPU::installInterrupts()\n" << eor;

    undefinedInstructionInterrupt = new ARMUndefinedInstructionInterrupt();
    dataAccessInterrupt = new ARMDataAccessInterrupt();
    instructionAccessInterrupt = new ARMInstructionAccessInterrupt();
    proxyInterrupt = new ARMProxyInterrupt();
    freeRunningTimerInterrupt = new MACHINEFREERUNNINGTIMERINTERRUPT();
    timerManagerInterrupt = new TimerManagerInterrupt();

    //See Includes/MachineDependent/$MACHINE/Vectors.h 
    setInterrupt(VEC_PF_ABORT, instructionAccessInterrupt);
    setInterrupt(VEC_DATA_ABORT, dataAccessInterrupt);
    setInterrupt(VEC_UNDEF_INSN, undefinedInstructionInterrupt);
    setInterrupt(VEC_TIMER1, freeRunningTimerInterrupt);
    setInterrupt(VEC_TIMER2, timerManagerInterrupt);

    /*
     * The Proxy callgate is a special case; it doesn't go
     *  through the default interrupt handler at all.
     */
    setInterrupt(VEC_SWI, proxyInterrupt);
}

void
ARMCPU::displayMode() {
    unsigned mode = getMode();

    if (mode == ARM_ABT) Console() << "ARM CPU in ABORT mode\n" << eor;
    if (mode == ARM_FIQ) Console() << "ARM CPU in FIQ mode\n" << eor;
    if (mode == ARM_IRQ) Console() << "ARM CPU in IRQ mode\n" << eor;
    if (mode == ARM_SVC) Console() << "ARM CPU in SUPERVISOR mode\n" << eor;
    if (mode == ARM_SYS) Console() << "ARM CPU in SYSTEM mode\n" << eor;
    if (mode == ARM_UND) Console() << "ARM CPU in UNDEFINED INSTR mode\n" << eor;
    if (mode == ARM_USR) Console() << "ARM CPU in USER mode\n" << eor;

    return;
}
