#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *     University of Illinois at Urbana-Champaign
 *     Department of Computer Science
 *     1304 W. Springfield Ave.
 *     Urbana, IL 61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * ARMContext.c: Processor context for the ARM.
 */

#include "Byte.h"
#include "Domain.h"
#include "Kernel.h"
#include "MemoryObject.h"
#include "MemoryObjectView.h"
#include "OutputStream.h"
#include "RootMemoryObjectCache.h"
#include "ARMContext.h"
#include "ARMInterrupts.h"
#include "ARMPageTableConstants.h"
#include "TwoLevelPageTable.h"
#include "ARMPageTableEntry.h"
#include "SourceOfZerosMemoryObject.h"
#include "ThisProcess.h"
#include "Process.h"

#ifdef __GNUC__
#pragma implementation
#endif

class Process;

const unsigned TinyStackPages = 2;
const unsigned SmallStackPages = 4;
const unsigned NormalStackPages = 8;
const unsigned LargeStackPages = 16;
const unsigned HugeStackPages = 32;
const unsigned GiganticStackPages = 64;
const unsigned UserStackPages = 16;

inline 
const unsigned
stackPages( StackSize size)
{
    unsigned npages;
    switch( size ) {
        case none:
            npages =  UserStackPages;
            Assert (NOTREACHED);
            break;
        case heapBootStack:
        case tiny:
            npages =  TinyStackPages;
            break;
        case small:
            npages =  SmallStackPages;
            break;
        case large:
            npages =  LargeStackPages;
            break;
        case huge:
            npages =  HugeStackPages;
            break;
        case gigantic:
            npages =  GiganticStackPages;
            break;
        case normal:
            npages =  NormalStackPages;
        case application:
        default:
            npages =  UserStackPages;
            break;
    }
    return npages;
}

/*
 * Allocate a (non-paged and pre-allocated) stack of size into domain.
 * return the value to use as the stackpointer (stack grows down on ARM)
 */
void
ARMGenericContext::allocateStack(StackSize size, Domain *domain,
                                 char * &stackpointer, unsigned & stackbytes, MemoryObjectStar &mo, 
                                 char * & base)
{
    Debug << "ARMGenericContext: allocateStack(): " << "Allocating " << eor;
    Assert (domain != 0);
    int npages = stackPages(size);
    Debug << npages << " pages into domain: " << domain << "\n" << eor;
    Assert (TheKernel->domain() != 0);
    base = 0;
    stackbytes = npages*PhysicalPageSize;

    if (size == heapBootStack) {
        mo = (MemoryObject *)0;
        base = new char[stackbytes];
        Assert (base != 0);
        Debug << "ARMGenericContext::allocateStack(): " << "heapStack base:" << asAddress(base) << "\n" << eor;
    }
    else  {
        ProtectionLevel prot = (domain == TheKernel->domain()) ? NoAccess : ReadWrite;
        mo = new SourceOfZerosMemoryObject;
        mo->setName("StackMo");
        mo->setNumberOfUnits(npages*PhysicalPageSize);
        mo->setCacheClass( RootMemoryObjectCacheClass );
        MemoryObjectCache * cache = mo->buildCache();
        Assert (cache != 0);
        base = (char *)domain->add(mo, prot);
        Assert (base != 0);
        domain->makeAddressable(base, npages*PhysicalPageSize);
        Assert( base != 0 );
        Debug << "ARMGenericContext::allocateStack(): " << "MO stack base:" << asAddress(base) << "\n" << eor;
    }

    // The stack pointer should be word aligned
    stackpointer = (char*) ((unsigned)base + stackbytes);

    if ((unsigned)stackpointer & 0x3) {
        unsigned aligned;
        aligned = (unsigned)stackpointer & ~0x3;
        stackbytes -= ((unsigned)stackpointer - aligned);
        stackpointer = (char *)aligned;
    }
}

void catchAll(int arg0, int arg1, int arg2)
{
    try {
        void(* entry)(int,int) = (void(*)(int,int))arg0;
        
        entry(arg1, arg2);
    } catch(Exception *e) {
        Console() << asHex(arg0) << ": Caught Unhandled Top-Level Exception "
                  << *e << ": " << e->getMessage() << "\n" << eor;
        delete e;
    } catch(...) {
        Console() << asHex(arg0) << ": Caught Unhandled Top-Level Exception\n"
                  << eor;
    }
    
    thisProcess()->die();
}

ARMGenericContext::ARMGenericContext( Process * owner, PFV entry, 
                                      StackSize stackSize, int arg0, int arg1)
    : ProcessorContext( owner )
{
    Debug << "ARMContext.cc: Setting up generic context\n" << eor;
    Assert (entry != 0);
    Assert (TheKernel->domain() != 0);
    Assert (stackSize != none);

    allocateStack(stackSize, TheKernel->domain(), _kernelStackTop,
                  _kernelStackSize, _kernelStackMO, _kernelStack);
    _stackPointer = _kernelStackTop;

    volatileRegs[ARMCPU::R0] = (unsigned)entry;
    volatileRegs[ARMCPU::R1] = (unsigned)arg0;
    volatileRegs[ARMCPU::R2] = (unsigned)arg1;
    volatileRegs[ARMCPU::FP] = 0;
    volatileRegs[ARMCPU::IP] = 0;
    volatileRegs[ARMCPU::LR] = (unsigned)&defaultReturn;
    volatileRegs[ARMCPU::SP] = (unsigned)_stackPointer;
    volatileRegs[ARMCPU::PC] = (unsigned)catchAll;
    cpsr = 0;
    returnAddress = 0;

    /*
     * Sanity checks.  
     */
    Assert ( _kernelStackTop != 0 );
    Assert ( ! ((unsigned)_kernelStackTop & 0x3) );
    Assert ( _stackPointer != 0 );
    Assert ( ! ((unsigned)_stackPointer & 0x3) );
    Debug << "ARMContext.cc: Done setting up generic context\n" << eor;
}

ARMGenericContext::ARMGenericContext( Process * owner, PFV entry, 
                                      StackSize stackSize, OPFV arg0, int arg1)
    : ProcessorContext( owner )
{
    Assert(NOTREACHED);
}

ARMGenericContext::~ARMGenericContext()
{
    if (_kernelStackMO != 0){
        TheKernel->domain()->remove(_kernelStackMO,_kernelStack);
        _kernelStackMO = (MemoryObject*)0;
    }
}

/**
 * Handler for the 0xFFC SWI. The PC for the context is set to point to this
 * function. We override the context's stack with the kernel stack.
 */
void ARMGenericContext::doExit(void)
{ 
    Assert(ARMCPU::getMode() == ARM_SYS);
  
    ARMGenericContext *context = (ARMGenericContext *)thisProcess()->context();
    char *sp = context->_kernelStackTop;
  
    /* update the stack pointer */
    asm( "mov sp, %0" :: "r"(sp) );

    /* call die */
    thisProcess()->die();

    Assert(NOTREACHED);
}


void
ARMGenericContext::defaultReturn()
{
    Assert (thisProcess() != 0);

    /* 0xFFC is the exit system call */
    asm ("swi #0xFFC");
    Assert(NOTREACHED);
}

volatile Process *
ARMGenericContext::checkpoint()
{
    setResuming(0);

    // Save cpsr so that we can restore it later
    asm( "mrs %0, cpsr" : "=r"(cpsr) );
    // Save all the other registers
    asm( "stmia %0, {r0-r15}" :: "g"(volatileRegs) );

    // Francis - need 12 bytes of nops because store fast forwards PC by either 8 or 12 bytes
    asm ("nop\n"
         "nop\n"
         "nop\n"
        );

    unsigned *fp;
    asm ("mov %0, fp": "=r"(fp));
    returnAddress = fp[-1];

    volatile Process *giver = getResuming();
    if (giver != 0) {
//    Debug << "ARMGenericContext{"<<asHex(this)<<"::checkpoint() : resumed\n";
        // Put correct return address in frame pointed by fp
        unsigned *fp = (unsigned *)volatileRegs[ARMCPU::FP];
        fp[-1] = returnAddress;
        returnAddress = 0;
    } else {
//    Debug << "ARMGenericContext{"<<asHex(this)<<"::checkpoint() : checkpointed\n";
    }
    return((Process*)giver);
}

void
ARMGenericContext::loadFromInterruptContext(ARMInterruptContext *context) {
    volatile unsigned *sourceRegs = context->getRegisters();
    for (int i = 0; i < 16; i++) volatileRegs[i] = sourceRegs[i];
    cpsr = context->getCPSR();
}

void
ARMGenericContext::restoreToInterruptContext(ARMInterruptContext *context) {
    volatile unsigned *destRegs = context->getRegisters();
    for (int i = 0; i < 16; i++) destRegs[i] = volatileRegs[i];
    context->setCPSR(cpsr);
}

void
ARMGenericContext::restoreFromInitialContext(Process *previousProcess)
{
    Assert(!PROCESSOR::interruptsEnabled());

    /*
     * We are on the kernel stack of the giving process.
     */

    // Set the currentDomain variable inside the Process object
    process()->setCurrentDomain( process()->domain() );
    // Restore assist handles switching of domains
    restoreAssist( previousProcess );
    
    /*
     * Register names
     * Reg #    APCS    Meaning
     * R0     a1  Working registers
     * R1     a2  "
     * R2     a3  "
     * R3     a4  "
     * R4     v1  Must be preserved
     * R5     v2  "
     * R6     v3  "
     * R7     v4  "
     * R8     v5  "
     * R9     v6  "
     * R10    sl  Stack Limit
     * R11    fp  Frame Pointer
     * R12    ip   
     * R13    sp  Stack Pointer
     * R14    lr  Link Register
     * R15    pc  Program Counter
     */

    // Make this look like we are returning from an interrupt
    // so that we can switch pc and cpsr at the same time and thus renabling interrupts
    Assert(ARMCPU::getMode() == ARM_SYS);
    ARMCPU::setMode(ARM_SVC);
    asm( "mov lr, %0" :: "r"(volatileRegs[ARMCPU::PC]) );
    asm( "msr spsr_cxsf, %0" :: "r"(cpsr) );
    asm( "ldmia %0, {r0-r14}^" :: "g"(volatileRegs) );
    asm( "movs pc, lr" );

    Assert(NOTREACHED);
}

Process *
ARMGenericContext::basicRestore(Process *processGivingUpProcessor)
{
    setResuming(processGivingUpProcessor);

    // Make this look like we are returning from an interrupt
    // so that we can switch pc and cpsr at the same time and thus renabling intertupts
    Assert(ARMCPU::getMode() == ARM_SYS);
    ARMCPU::setMode(ARM_SVC);
    asm( "mov lr, %0" :: "r"(volatileRegs[ARMCPU::PC]) );
    asm( "msr spsr_cxsf, %0" :: "r"(cpsr) );
    asm( "ldmia %0, {r0-r14}^" :: "g"(volatileRegs) );
    asm( "movs pc, lr" );

    Assert(NOTREACHED);

    return processGivingUpProcessor;
}


void printStackFrame(char* s) {
    Console() << "Frame: " << asHex(s) << "\n" << eor;

    char* next = (char*)*((unsigned*)(s-12));

    unsigned* frame = (unsigned*)s;
    unsigned elements = ( (next-s)/4 > 0 ) ? (next-s)/4 : 0;
    Console() << "pc: " <<asHex(frame[0]) 
              << " lr: " << asHex(frame[-1]) 
              << " ip: " << asHex(frame[-2]) 
              << " fp: " << asHex(frame[-3]) 
              << " elements:" << elements << "\n" << eor;
  
    if(next && (elements > 0))
        printStackFrame(next);
}

void saveStackTrace(char *s, ARMStackTrace *trace) {
    char* next = (char*)*((unsigned*)(s-12));
    unsigned *frame = (unsigned*)s;
    unsigned elements = ( (next-s)/4 > 0 ) ? (next-s)/4 : 0;
    trace->addFrame( (unsigned)frame, frame[0], frame[-1], frame[-3] );
    if(next && (elements > 0))
        saveStackTrace(next, trace);
}

void
ARMInterruptContext::dumpStackTrace()
{
    Console() << "Stack BackTrace:\n" << eor;
    printStackFrame(reinterpret_cast<char *>(volatileRegs[ARMCPU::FP]));
    Console() << "End\n" << eor;
}

ARMStackTrace
ARMInterruptContext::getStackTrace()
{
    ARMStackTrace trace;
    saveStackTrace(reinterpret_cast<char *>(volatileRegs[ARMCPU::FP]), &trace);
    return trace;
}

void traceStack()
{
    Console() << "Stack BackTrace:\n" << eor;
    char *s;
    asm volatile("mov %0, fp" : "=r" (s));
    printStackFrame(s);
    Console() << "End\n" << eor;
}


void ARMInterruptContext::dumpRegisters()
{
    unsigned long flags;
    const char *processor_modes[] = {
        "USER_26",      "FIQ_26",       "IRQ_26",       "SVC_26",
        "UK4_26",       "UK5_26",       "UK6_26",       "UK7_26",
        "UK8_26",       "UK9_26",       "UK10_26",      "UK11_26",
        "UK12_26",      "UK13_26",      "UK14_26",      "UK15_26",
        "USER_32",      "FIQ_32",       "IRQ_32",       "SVC_32",
        "UK4_32",       "UK5_32",       "UK6_32",       "ABT_32",
        "UK8_32",       "UK9_32",       "UK10_32",      "UND_32",
        "UK12_32",      "UK13_32",      "UK14_32",      "SYS_32",
    };

    flags = cpsr;

    Console() << "pc : " << asAddress((void *)volatileRegs[ARMCPU::PC]) << " lr : " << asAddress((void *)volatileRegs[ARMCPU::LR]) << "\n" 
              << "sp : " <<   asAddress((void *)volatileRegs[ARMCPU::SP]) << " ip : " << asAddress((void *)volatileRegs[ARMCPU::IP]) << " fp : " 
              << asAddress((void *)volatileRegs[ARMCPU::FP]) << "\n" << eor;
    Console() << "r10: " << asAddress((void *)volatileRegs[ARMCPU::R10]) <<  " r9 : " <<  asAddress((void *)volatileRegs[ARMCPU::R9]) 
              << " r8 : " <<  asAddress((void *)volatileRegs[ARMCPU::R8]) << "\n" << eor;
    Console() << "r7 : " << asAddress((void *)volatileRegs[ARMCPU::R7]) << " r6 : " << asAddress((void *)volatileRegs[ARMCPU::R6])
              << " r5 : " << asAddress((void *)volatileRegs[ARMCPU::R5]) << " r4 : " << asAddress((void *)volatileRegs[ARMCPU::R4]) << "\n" << eor;
    Console() << "r3 : " << asAddress((void *)volatileRegs[ARMCPU::R3]) << " r2 : " << asAddress((void *)volatileRegs[ARMCPU::R2])
              << " r1 : " << asAddress((void *)volatileRegs[ARMCPU::R1]) << " r0 : "  << asAddress((void *)volatileRegs[ARMCPU::R0]) << "\n" << eor;

    Console() << "Flags: " << (flags & (1 << 31) ? 'N' : 'n') << (flags & (1 << 30) ? 'Z' : 'z') 
              << (flags & (1 << 29) ? 'C' : 'c') << (flags & (1 << 28) ? 'V' : 'v')
              << "  IRQs " << (!(cpsr & 0x80) ? "on" : "off") << " FIQs " << (!(cpsr & 0x40) ? "on" : "off") << " Mode "
              <<  processor_modes[cpsr & 0x1f] << ((cpsr & 0x20) ? " (T)" : "") << "\n" << eor;
}

// ------------------------------- ARMSystemContext --------------------------

ARMSystemContext::ARMSystemContext( Process * owner, OPFV entry, StackSize stackSize, int arg0, int arg1 ) 
    : ARMGenericContext( owner, (PFV)*((unsigned *)(&entry)), stackSize, arg0, arg1)
{
    // Set CPSR with IFT as 000 and mode ARM_SYS
    cpsr &= ~0x1f;
    cpsr |= ARM_SYS;
}

ARMSystemContext::ARMSystemContext( Process * owner, PFV entry, StackSize stackSize, int arg0 , int arg1) :
    ARMGenericContext( owner, entry, stackSize, arg0, arg1)
{
    // Set CPSR with IFT as 000 and mode ARM_SYS
    cpsr &= ~0x1f;
    cpsr |= ARM_SYS;
}

ARMSystemContext::ARMSystemContext( Process * owner, PFV entry, StackSize stackSize, OPFV arg0 , int arg1) :
    ARMGenericContext( owner, entry, stackSize, arg0, arg1)
{
    Assert(NOTREACHED);
}

// ------------------------------- ARMApplicationContext ----------------------

ARMApplicationContext::ARMApplicationContext( Process * owner,
                                              PFV entry, StackSize stackSize, Domain *initialDomain,
                                              int arg0, int arg1, StackSize userStackSize, char *userStackPointer)
    : ARMGenericContext( owner, entry, stackSize, arg0, arg1), userStackMO(0)
{
    Assert (initialDomain != 0);
    Assert (entry != 0);

    if ( userStackSize != none ) {
        allocateStack(userStackSize, initialDomain,
                      _userStackTop, _userStackSize, userStackMO, 
                      userStackBase);
        _stackPointer = _userStackTop;
    } else {
        Assert( initialDomain != 0 );
    
        const char * base;
        unsigned length;
        ProtectionLevel prot;
        MemoryObjectViewStar mov = initialDomain->lookup
        ( ( const char * ) userStackPointer,
          base, length, prot ); 
        Assert( mov != 0 );
        userStackBase = ( char * ) base;
        _userStackSize = length;
        _userStackTop = (char*) (base + length - 1 );
        if ( (unsigned) _userStackTop & 7)
            _userStackTop = 
            (char*) ((unsigned)(_userStackTop - 7) & ~7);
        _userStackTop -= 168;
        userStackMO = mov->source();
        initialDomain->makeAddressable( base, length ); 
    
        _stackPointer = userStackPointer;
    }

    /*
     * Set up additional inputs that initialRestoreAssist uses.
     */
    Debug << "User stack top is " << asHex(_stackPointer) << "\n" << eor;

    // Set PC to user PC - No catchall for user programs
    volatileRegs[ARMCPU::PC] = volatileRegs[ARMCPU::R0];
    volatileRegs[ARMCPU::R0] = volatileRegs[ARMCPU::R1];
    volatileRegs[ARMCPU::R1] = volatileRegs[ARMCPU::R2];
    volatileRegs[ARMCPU::R2] = 0;

    // Set the SP to user stack
    volatileRegs[ARMCPU::SP] = reinterpret_cast<unsigned>(_stackPointer);

    // Set CPSR with IFT as 000 and mode ARM_USR
    cpsr &= ~0x1f;
    cpsr |= ARM_USR;

    /*
     * Sanity checks.  
     */
    Assert ( _kernelStackTop != 0 );
    Assert ( ! ((unsigned)_kernelStackTop & 0x3) );
    Assert ( _stackPointer != 0 );
    Assert ( ! ((unsigned)_stackPointer & 0x3) );
}

// -------------------------- ARMUninterruptableSystemContext ---------------------

ARMUninterruptableSystemContext::ARMUninterruptableSystemContext(
    Process * owner, PFV entry, StackSize stackSize, int arg0, int arg1)
    : ARMSystemContext( owner, entry, stackSize, arg0, arg1)
{
    // Set CPSR IF to 11
    cpsr |= 0x000000C0;
}

#include "ClassInstaller.h"
DEFINECLASS("ARMGenericContext",ARMGenericContext,ProcessorContext,0,0);
DEFINECLASS("ARMSystemContext",ARMSystemContext,ProcessorContext,0,0);
DEFINECLASS("ARMApplicationContext",ARMApplicationContext,ProcessorContext,0,0);
DEFINECLASS("ARMUninterruptableSystemContext",ARMUninterruptableSystemContext,ProcessorContext,0,0);

