#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#include "AccessType.h"
#include "AddressTranslator.h"
#include "Assert.h"
#include "Class.h"
#include "ThisCPU.h"
#include "CPU.h"
#include "Domain.h"
#include "ObjectProxy.h"
#include "OutputStream.h"
#include "ProxyTable.h"
#include "ARMContext.h"
#include "ARMException.h"
#include "ARMInterrupts.h"
#include "ADDRESS_TRANSLATION.h"
#include "ThisProcess.h"
#include "Process.h"

#ifdef __GNUC__
#pragma implementation
#endif

extern "C" void
restartReturnedAssertion()
{
    Console() << "\n_restart returned!\n";
    Assert( NOTREACHED ) ;
}

extern "C" void
stackAlignment()
{
    Console() << thisProcess() << ": Stack alignment problem\n";
    Assert( NOTREACHED ) ;
}

extern "C" void
stackOverflow(unsigned psr, unsigned pc, unsigned sp, unsigned wim,
              unsigned address)
{
    Console() << thisProcess() << ": Stack overflow?:\n\t(psr)" << asHex(psr) 
              << "\n\t(pc)" << asHex(pc) << "\n\t(sp)" << asHex(sp)
              << "\n\t(wim)" << asHex(wim) << "\n\t(address)" << asHex(address)
              << "\n" ;
    Assert( NOTREACHED ) ;
}

/*
 * Used by Interrupts to kill the current process
 * Use side effect of Assert killing the Process
 */
void
killProcessDueToInterrupt( Process *myprocess, char *sp)
{
    Assert(myprocess != 0);
    Console() << " Killing process: " << myprocess << "\n" << eor;
#ifdef RAIDOUTPUTSTREAM
    extern void RaidDump();
    RaidDump();
#endif  
    myprocess->die();
    Assert (NOTREACHED);
}

#define INTERRUPTS_PC_CHANGE

// #define INTERRUPTS_INSTR_OVERWRITE - This suffers from race condition when concurrent threads
// step on the instruction we overwrite
extern void traceStack();

void throwException() {
    // Debug << thisProcess() << "::Throwing exception object after crash\n" << eor;
    throw thisProcess()->getException();
}

void setupForExceptionThrow(ARMInterruptContext *context) {
#ifdef DEBUG
    context->dumpRegisters();
    context->dumpStackTrace();
#endif

#ifdef INTERRUPTS_INSTR_OVERWRITE
    // Replace faulting instruction with bl throwException
    // e    b    ffe7a9    - addresses 0x100f5c38 from pc 0x100fbd8c (+8)
    // 1110 1011 signed 2's comp immediate
    // COND ---L
    // 1857 << 2 = 0x615c

    unsigned target = (unsigned)&throwException;
    
    int value = target-pc-8;
    value >>= 2;
    unsigned instr = ((unsigned)value & 0x00FFFFFF) | 0xEB000000;

    unsigned pc = context->getRegisters()[ARMCPU::PC];
    *((unsigned *)pc) = instr;
#endif

#ifdef INTERRUPTS_PC_CHANGE
    context->getRegisters()[ARMCPU::LR] = context->getRegisters()[ARMCPU::PC] + 4;
    context->getRegisters()[ARMCPU::PC] = reinterpret_cast<unsigned>(&throwException);
#endif

    // Francis TODO TOFIX This is the current solution for exception handling to work
    // for user-privileged system processes in the DDriver code
    // Escalate Privileges if the process is a SystemProcess
    if (thisProcess()->classOf()->isASubclassOf(SystemProcessClass))
        context->setMode(ARM_SYS);
}

extern unsigned reloadKernelFromFlash(unsigned address);

void
ARMUndefinedInstructionInterrupt::basicRaise(char *interruptContext)
{
#ifdef RELIABILITY
    // If error is in OS code, recover by reloading code
    Console() << "Undefined Instruction: fixing by reload!\n" << eor;
    if (reloadKernelFromFlash(((ARMInterruptContext *)interruptContext)->getRegisters()[ARMCPU::PC]))
        return;
#endif

    Debug << "ARMUndefinedInstructionInterrupt::basicRaise() for " << thisProcess() << "\n" << eor;
    thisProcess()->setException(new ARMUndefinedInstructionException((ARMInterruptContext *)interruptContext));
    setupForExceptionThrow((ARMInterruptContext *)interruptContext);
    return;
/*
  Process *p = thisProcess();
  killProcessDueToInterrupt(p, sp);
  Assert(NOTREACHED);
*/
}

void
ARMMemoryInterrupt(ARMInterruptContext *context, int type) {
    static char *oldAddress = (char *)0xfef0babe;

    char *address;
    unsigned writeAttempted = 0;
    volatile unsigned fsr;
    //context->dumpRegisters();
    if (type) { /* Instruction fault */
        address = (char *)context->getRegisters()[ARMCPU::PC];
    }
    else { /* Data Fault */
        address = 0;
        asm("mrc     p15, 0, %0, c6, c0, 0":"=r"(address)); // get FAR
        asm("mrc     p15, 0, %0, c5, c0, 0":"=r"(fsr));     // get FSR 

        Debug << "Data Fault status reports: ";
        switch(fsr & 0xf) {
            case FSR_ALIGNMENT1:
            case FSR_ALIGNMENT2:
                Debug << "AlignmentError\n";
                break;
            case FSR_EXT_ABORT_TRANSLATION_L1:
            case FSR_EXT_ABORT_TRANSLATION_L2:
                Debug << "ExternalAbortOnTranslation\n";
                break;
            case FSR_TRANSLATION_SECTION:
            case FSR_TRANSLATION_PAGE:
                Debug << "TranslationError\n";
                break;
            case FSR_DOMAIN_SECTION:
            case FSR_DOMAIN_PAGE:
                Debug << "DomainError\n";
                break;
            case FSR_PERMISSION_SECTION:
            case FSR_PERMISSION_PAGE:
                Debug << "PermissionError\n";
                break;
            default:
                Debug << (fsr & 0xf) << "\n" << eor;
                break;
        }
        //Debug << "Faulting instruction is " << asHex(*((volatile unsigned *)context->getRegisters()[ARMCPU::PC])) << "\n" << eor;
        writeAttempted = !((*((volatile unsigned *)context->getRegisters()[ARMCPU::PC]) & 0x00100000) >> 20);
/*
  Code from Linux Kernel that does the same as above:
        // cpsr in r3
        // address in r2
        tst     r3, #1<<24                      @ Check for Jbit (NE -> found)
        movne   r3, #-1                         @ Mark as writing
        bne     2f

        tst     r3, #1<<5                       @ Check for Thumb-bit (NE -> found)
        ldrneh  r1, [r2]                        @ Read aborted Thumb instruction
        tstne   r1, r1, lsr #12                 @ C = bit 11

        ldreq   r1, [r2]                        @ Read aborted ARM instruction
        tsteq   r1, r1, lsr #21                 @ C = bit 20
        sbc     r1, r1, r1                      @ r1 = C - 1
*/
    }

    Debug << "Faulting address is " << asAddress(address) << ":";
    Debug << (writeAttempted ? "Write\n":"Read\n") << eor;

    if (oldAddress == address) { 
        Debug << "Repeated fault. Halting.\n" << eor;
        while(1);
    }
    oldAddress = address;

    /*
     * figure out the access type
     */
    AccessType attemptedAccess;
    if (context->getMode() == ARM_SYS) {
        if (writeAttempted) {
            attemptedAccess = SystemWrite;
        }
        else {
            attemptedAccess = SystemRead;
        }
    }
    else {
        Assert(context->getMode() == ARM_USR);
        if (writeAttempted){
            attemptedAccess = ApplicationWrite;
        }
        else {
            attemptedAccess = ApplicationRead;
        }
    }
    
    Process *process = thisProcess();
    Assert ( process != 0 );
    Domain *domain = process->domain();
    Assert(domain != 0);
    FaultErrorCode reason = domain->repairFault(address, attemptedAccess);

    switch(reason) {
        /*
         * These are "normal" and the instruction can resume.
         */
        case MissingAddressTranslationMapping:
            Debug << "MissingAddressTranslationMapping\n" << eor;
            break;
        case MissingPhysicalMemory:
            Debug << "MissingPhysicalMemory\n" << eor;
            break;
        case InsufficientProtectionLevel:
            Debug << "InsufficientProtectionLevel\n" << eor;
            break;

            /*
             * These are fatal errors
             */
        case WriteOfProtectedSystemMemory:
            Console() << "WriteOfProtectedSystemMemory:";
            Console() << " Address: " << asAddress(address) << "\n" << eor;
            thisProcess()->setException(new ARMDataAccessException(context,"WriteOfProtectedSystemMemory"));
            setupForExceptionThrow(context);
            break;
        case AccessOfUnmanagedMemory:
            Console() << "AccessOfUnmanagedMemory:";
            Console() << " Address: " << asAddress(address) << "\n" << eor;
            if (type)
                thisProcess()->setException(new ARMInstructionAccessException(context,"AccessOfUnmanagedMemory"));
            else
                thisProcess()->setException(new ARMDataAccessException(context,"AccessOfUnmanagedMemory"));
            setupForExceptionThrow(context);
            break;
        case ReadOfProtectedSystemMemory:
            Console() << "ReadOfProtectedSystemMemory:";
            Console() << " Address: " << asAddress(address) << "\n" << eor;
            if (type)
                thisProcess()->setException(new ARMInstructionAccessException(context,"ReadOfProtectedSystemMemory"));
            else
                thisProcess()->setException(new ARMDataAccessException(context,"ReadOfProtectedSystemMemory"));
            setupForExceptionThrow(context);
            break;
        case WriteOfReadOnlyMemory:
            Console() << "WriteOfReadOnlyMemory:";
            Console() << " Address: " << asAddress(address) << "\n" << eor;
            thisProcess()->setException(new ARMDataAccessException(context,"WriteOfReadOnlyMemory"));
            setupForExceptionThrow(context);
            break;
        case DynamicWriteOfReadOnlyMemory:
            Console() << "DynamicWriteOfReadOnlyMemory:";
            Console() << " Address: " << asAddress(address) << "\n" << eor;
            thisProcess()->setException(new ARMDataAccessException(context,"DynamicWriteOfReadOnlyMemory"));
            setupForExceptionThrow(context);
            break;
        case DynamicAccessViolation:
            Console() << "DynamicAccessViolation:";
            Console() << " Address: " << asAddress(address) << "\n" << eor;
            thisProcess()->setException(new ARMDataAccessException(context,"DynamicAccessViolation"));
            setupForExceptionThrow(context);
            break;
        default:
            Console() << "Unknown (default) access violation:";
            Console() << " Address: " << asAddress(address) << "\n" << eor;
            Assert(NOTREACHED);
            break;
    };
    // We shouldn't crash after this. So, we allow same data abort from user/sys next time
    oldAddress++;
}

void
ARMDataAccessInterrupt::basicRaise( char *interruptContext)
{
    Debug << "ARMDataAccessInterrupt::basicRaise() for " << thisProcess() << "\n" << eor;
    // This is used for Restart Old Performance Test
    //unsigned pc = ((ARMInterruptContext *)interruptContext)->getRegisters()[ARMCPU::PC];   
    ARMMemoryInterrupt((ARMInterruptContext *)interruptContext, 0);
    //((ARMInterruptContext *)interruptContext)->getRegisters()[ARMCPU::PC] = pc + 4;
}

void
ARMInstructionAccessInterrupt::basicRaise( char *interruptContext)
{
    Debug << "ARMInstructionAccessInterrupt::basicRaise()\n" << eor;

    ARMMemoryInterrupt((ARMInterruptContext *)interruptContext, 1);
}

void
ARMDivideByZeroInterrupt::basicRaise( char *interruptContext)
{
    Process *p = thisProcess();
    Console() << p << ": Divide by zero, killing process\n" << eor;
//    killProcessDueToInterrupt(p, sp);
    Assert(NOTREACHED);
}

extern "C" {
    int
    ProxyCallAssist( unsigned int methodNumber, int fromKernel,
                     ProxyReturnFlags *returnFlags, void *refPointer,
                     ProxiableObject **from, ProxiableObject **to );

    ProxiableObject *
    ProxyReturnAssist(ProxiableObject *returnValue, ProxyReturnFlags *returnFlags);

    typedef ProxiableObject * (* proxyPFV) (ProxiableObject *, ProxiableObject *, 
                                            ProxiableObject *, ProxiableObject *, ProxiableObject *, ProxiableObject *,
                                            ProxiableObject *, ProxiableObject *, ProxiableObject *, ProxiableObject *,
                                            ProxiableObject *, ProxiableObject *);
    typedef ProxiableObjectRef (* proxyRefPFV)
        (ProxiableObject *, ProxiableObject *, 
         ProxiableObject *, ProxiableObject *, ProxiableObject *, ProxiableObject *,
         ProxiableObject *, ProxiableObject *, ProxiableObject *, ProxiableObject *,
         ProxiableObject *, ProxiableObject *);

    int ProxyCallLog = 0;
};

void
ARMProxyInterrupt::basicRaise(char *interruptContext)
{
    Debug << "ARMProxyInterruptHandler()" << eor;
    Assert(NOTREACHED);
}

void
ARMWatchdogTimeoutInterrupt(ARMInterruptContext *context) {
    // Pretend that we were a normal processor exception
    // Throw exception like we do for other processor exceptions
    ARMWatchdogTimeoutException *ex = new ARMWatchdogTimeoutException(context);
    thisProcess()->setException(ex);
    setupForExceptionThrow(context);

    // Do a return to the context set up by setupForExceptionThrow()
    ((ARMGenericContext *)thisProcess()->context())->loadFromInterruptContext(context);
    thisProcess()->context()->restoreFromInitialContext(NULL);

    Assert(NOTREACHED);
    // This function never returns
}
