#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *       University of Illinois at Urbana-Champaign
 *       Department of Computer Science
 *       1304 W. Springfield Ave.
 *       Urbana, IL 61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *  All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

#include "ADDRESS_TRANSLATION.h"
#include "Assert.h"
#include "OutputStream.h"
#include "ARMMMU.h"
#include "AsmBridge.h"

#ifdef __GNUC__
#pragma implementation
#endif


ARMMMU::ARMMMU()
{
    Debug << "ARMMMU::ctor() " << this << "\n" << eor;
    _enabled = 0;
}

void showPTChoices(ARMPointerTableEntry* pointerTable) {
    ARMPointerTableEntry* pointer;
    unsigned val;

    for(int i=0; i < 4096; i++) {
        pointer = (ARMPointerTableEntry*) &pointerTable[i];
        if(pointer->valid()) {
            Console() << "Pointer table entry=" << asHex(pointer->value()) << "\n" << eor;
            if(pointer->value() & 0x1) {
                //page table
                ARMPageTableEntry* ptbase = (ARMPageTableEntry*) (((unsigned)pointer->value()) & ~((1 << 10) -1));
                Console() << "  Page table base entry at: PTbase=" << asHex((unsigned)ptbase)
                          << " maps 1 mb of virtual addresses starting at: VA=" << asHex(i * 0x100000) << "\n" << eor;
                for(int j=0; j < 256; j++) {
                    ARMPageTableEntry* pt = &ptbase[j];
                    if(pt->valid())
                    {
                        Console() << "    Page table entry[" << asHex(i) << "][" << asHex(j) << "] entry="
                                  << asHex(pt->value()) << " physical=" << asHex(pt->value() & ~0xfff)
                                  << " maps 4kb of virtual addresses starting at: VA="
                                  << asHex(i * 0x100000 + j * 4096) << "\n" << eor;
                    }
                }
            } else if(pointer->value() & 0x2) {
                //pointer table
                val = pointer->value() & ~((1 << 20) -1);
                Console() << "  Section entry at: " << asHex(val)
                          << " - " << asHex(val + 0x100000) << " VA=" << asHex(i * 0x100000) << "\n" << eor;
            }
        }
    }
}


void
ARMMMU::basicActivate( AddressTranslation * at)
{
    ARMTranslation * tr = (ARMTranslation *)at;
    unsigned ttb = (unsigned)(tr->pointerTable());

    Debug << "Writing TTB to MMU: ttb=" << asHex(ttb) << "\n" << eor;

    if (_enabled) 
        flushTLBandCaches();

    asm ( "mcr p15, 0, %0, c2, c0, 0"
          :
          : "r" (ttb&0xffffc000)
          : );

#ifdef INTEGRATOR
    flushITLB();
#endif    
}

void
ARMMMU::enable()
{
    _enabled = 1;

    // Page Tables should have been created correctly
    // and ttb should have been written to cp15:c2

    // set page table permissions (Sloss 512)
    int permsIn = 0xCAFEBABE, permsOut = 0x55555555;

    // check original value
    asm ( "mrc p15, 0, %0, c3, c0, 0"
          : "=r" (permsIn)
          :
          : );
    Debug << "Original Perms=" << asHex(permsIn) << "\n" << eor;
    Debug << "New Perms     =" << asHex(permsOut) << "\n" << eor;

    // set new value
    asm ( "mcr p15, 0, %0, c3, c0, 0"
          :
          : "r" (permsOut)
          : );
  
    // adjust MMU control word (Sloss 514)
    // read current MMU control word
    int control;
    asm ( "mrc p15, 0, %0, c1, c0, 0"
          : "=r" (control)
          :
          : );

    // enable MMU
    control |= 0x00000001;
    
    // enable ICACHE
    control |= 0x00001000;

#ifndef RELIABILITY
    // enable DCACHE
    control |= 0x00000004;
#endif

    // enable high interrupt vectors
    control |= 1 << 13; 

    Debug << "Original Control word=" << asHex(control) << "\n" << eor;
    Debug << "New control word     =" << asHex(control) << "\n" << eor;

    flushITLBandCache();
    flushDTLBandCache();

    // set the new MMU control word
    asm ( "mcr p15, 0, %0, c1, c0, 0"
          :
          : "r" (control)
          : );
}

int
ARMMMU::enabled()
{
    return _enabled;
}
