    @
    @ boot.s  -  main entry point in Choices
    @         -  interrupt vector and assembly bridges
    @               - prepare for transition into c code in AsmBridge.cc
    @                   - AsmBridge.cc is /MachineDependent/GenericArm/
   
    @ Assembly notes:
    @   Can't use star character

    @ Are KernelReadOnlyEnd and KernelReadOnlyStart actually used?
    
    @ Relevant Assembly Files:
        @ ProcessorDependent/ARM/KernelReadOnlyEnd.s
        @ ProcessorDependent/ARM/KernelReadOnlyStart.s
        @ ProcessorDependent/ARM/boot.s - this file
        @ Configure/System/Integrator/rootfs/hello.s - a basic user mode test program
        @ Multiple files in UNIX compatability library (may not be relevant)
    
    @ Inline Assembly
        @ MachineDependent/GeneralARM/ARMCPU.h
        @ MachineDependent/GeneralARM/ARMContext.h
        @ MachineDependent/GeneralARM/ARMMMU.h
        @ Configure/System/Integrator/rootfs/data.c
        @ MachineDependent/GeneralARM/GenericARMGdb.h
        
    @ Refactoring notes:
        @ Question mark replacements
        @ Do we want to combine state saving/restoring code?
        @ What is the difference between global and globl?

    .text                           @ Code Segment
        
    .globl _choicesstart            @ Designates where to start execution and where the interrupt vector is
    .globl reset_cpu                @ Inherieted from original boot loader code and is currently unused
                                    @ Would be how to reset the CPU from software
    .globl regsave                  @ Register save area for watchdog aware bootloaders

@ This section defines some constants for latter use.  This is equivalent to a #define in C
@ These masks are for the Program Status Registers (CPSR & SPSR)
    .equ    Usr_Mode, 0x10          @ User Mode
    .equ    FIQ_Mode, 0x11          @ FIQ Mode (Entered on FIQ)
    .equ    IRQ_Mode, 0x12          @ IRQ Mode (Entered on IRQ)
    .equ    SVC_Mode, 0x13          @ Supervisor Mode
    .equ    Abt_Mode, 0x17          @ Abort Mode (Entered on Prefetch/Data Aborts)
    .equ    Und_Mode, 0x1b          @ Undefined Mode (Entered on undefined instructions)
    .equ    Sys_Mode, 0x1f          @ System Mode

    .equ    I_bit, 0x80             @ IRQ flag
    .equ    F_bit, 0x40             @ FIQ flag
    .equ    T_bit, 0x20             @ Thumb flag
    .equ    NoINT, I_bit | F_bit    @ Disable all interrupt (Combines above two)


@Note: This corresponds to address 0x00000000.    
@ This is the interrupt vector. Also, this is where the execution starts for initial
@   boot.
_choicesstart:
    ldr pc, =reset                  @ Reset Handler (and boot code!)
    ldr pc, =undefined_instruction  @ Undefinded Instruction Exception Handler
    ldr pc, =software_interrupt     @ SWI Handler
    ldr pc, =prefetch_abort         @ Prefetch Abort Handler
    ldr pc, =data_abort             @ Data Abort Handler
    ldr pc, =reserved               @ Currently unused, but reserved by ARM Ltd for future use
    ldr pc, =irq                    @ IRQ (Interrupt) Handler
    ldr pc, =fiq                    @ FIQ (Fast Interrupt) Handler
    @ FIQ Handler can continue right here but currently doesn't

regsave:
    .word 0xdeadbeef    @ r2
    .word 0x0           @ r3
    .word 0x0           @ r4
    .word 0x0           @ r5
    .word 0x0           @ r6
    .word 0x0           @ r7
    .word 0x0           @ r8
    .word 0x0           @ r9
    .word 0x0           @ r10
    .word 0x0           @ r11
    .word 0x0           @ r12
    .word 0x0           @ r13
    .word 0xdeadbeef    @ r14
        
@FIQ (Fast Interrupt) Handler
@ Francis <fdavid@uiuc.edu>: This needs to be rewritten when the time is right
fiq:
    @ Save current state to the stack so that the C handler doesn't overwrite it
    sub     lr, lr, #4              @ The stack has to look like an ARMInterruptContext object
    sub     sp, sp, #4              @ Save space for CPSR
    str     lr, [sp, #-4]!          @ Write PC
    mrs     lr, spsr                @ Save SPSR into LR
    str     lr, [sp, #4]            @ Save SPSR into stack
    stmfd   sp!, {r0-r14}^          @ Save user registers. Note: We only expect FIQs from user or system mode
    mov     r0, sp                  @ This is setup for the c call conventions
    
    @ Jump to C handler in AsmBridge.cc
    bl      fiq_handler             @ in AsmBridge.cc
    
    @ Restore the saved state
    ldmfd   sp!, {r0-r14}^          @ Restore R0-R14
    ldr     lr, [sp, #4]            @ Load CPSR into LR
    msr     spsr_cxsf, lr           @ Restore SPSR
    ldr     lr, [sp]                @ Load PC into LR
    add     sp, sp, #8              @ Restore SP to correct value
    movs    pc, lr                  @ Load PC and return
    @MOVS PC, LR has a side effect of writing CPSR = SPSR, thus restoring the original mode

    @ FIQs need to be fast, we should probably use something like the code below        
    @ stmfd sp!, {r0-r7, lr}
    @ mov r0, sp
    @ mrs   lr, spsr
    @ str   lr, [sp, #-4]!
    @ bl    fiq_handler             @ in AsmBridge.cc
    @ ldr   lr, [sp], #4
    @ msr   spsr_cxsf, lr
    @ ldmfd sp!, {r0-r7, pc}^

@IRQ (Slow Interrupt) Handler
irq:
    @ Save current state to the stack so that the C handler doesn't overwrite it
    @ Because we could get an IRQ from any mode we need to save state for all modes
    sub     lr, lr, #4
    stmfd   sp!, {lr}                   @ Save user mode pc. Note: We only expect IRQs from user or system mode
    stmfd   sp!, {sp,lr}^               @ Save user mode sp and lr
    stmfd   sp!, {r0-r3,r12}            @ Save non-callee saved registers
    bl      get_stack                   @ Get kernel stack of running process
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    mov     sp, r0                      @ Switch to kernel stack of running process
    msr     cpsr_c, #NoINT | IRQ_Mode   @ Change mode to irq
    ldmfd   sp!, {r0-r3,r12}            @ Restore registers
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    sub     sp, sp, #16                 @ Save space for cpsr,pc,lr,sp
    stmfd   sp!, {r0-r12}               @ Save all registers
    msr     cpsr_c, #NoINT | IRQ_Mode   @ Change mode to irq
    ldmfd   sp!, {r0-r2}                @ Load sp,lr,pc into r0-r2
    mrs     r3, spsr                    @ Load spsr into r3
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    add     r4, sp, #68
    stmfd   r4, {r0-r3}                 @ Save sp,lr,pc,cpsr
    mov     r0, sp                      @ Re-enable interrupts? Cannot take nested IRQs though
    
    @ Jump to C handler in AsmBridge.cc
    bl      irq_handler                 @ in AsmBridge.cc
    
    @ Restore the saved state
    add     r4, sp, #52
    ldmfd   r4, {r0-r3}                 @ Restore sp,lr,pc,cpsr from sys stack
    msr     cpsr_c, #NoINT | IRQ_Mode   @ Change mode to irq
    stmfd   sp!, {r0-r3}                @ Save sp,lr,pc,cpsr onto irq stack
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    ldmfd   sp!, {r0-r12}               @ Load registers from sys stack
    msr     cpsr_c, #NoINT | IRQ_Mode   @ Change mode to irq
    ldmfd   sp!, {sp,lr}^               @ Restore sp,lr from irq stack
    ldr     lr, [sp, #4]!               @ Restore cpsr from irq stack
    add     sp, sp, #4
    msr     spsr_cxsf, lr
    ldr     lr, [sp, #-8]               @ Restore pc from abt stack
    movs    pc,lr
    @MOVS PC, LR has a side effect of writing CPSR = SPSR, thus restoring the original mode

@Called for instructions that the hardware doesn't recognize
undefined_instruction:
    sub     lr, lr, #4
    @ Save current state to the stack so that the C handler doesn't overwrite it
    sub     sp, sp, #4                  @ Save space for CPSR
    str     lr, [sp, #-4]!              @ Write PC
    mrs     lr, spsr                    @ Save SPSR into LR
    str     lr, [sp, #4]                @ Save SPSR into stack
    stmfd   sp!, {r0-r14}^              @ Save user registers. Note: We only expect UDEFs from user or system mode
                                        @ This is only a safe assumption since we know which modes are used.  
    mov     r0, sp
    
    @ Jump to C handler in AsmBridge.cc
    bl      undef_handler               @ in AsmBridge.cc
    
    @ Restore the saved state
    ldmfd   sp!, {r0-r14}^              @ Restore R0-R14
    ldr     lr, [sp, #4]                @ Load CPSR into LR
    msr     spsr_cxsf, lr               @ Restore SPSR
    ldr     lr, [sp]                    @ Load PC into LR
    add     sp, sp, #8                  @ Restore SP to correct value
    movs    pc, lr                      @ Load PC and return
    @MOVS PC, LR has a side effect of writing CPSR = SPSR, thus restoring the original mode

@SWI, called for kernel traps
software_interrupt:
    @ Save current state to the stack so that the C handler doesn't overwrite it
    @ Because we could get an IRQ from any mode we need to save state for all modes
    stmfd   sp!, {lr}                   @ Save user mode pc. Note: We only expect SWIs from user or system mode
    stmfd   sp!, {sp,lr}^               @ Save user mode sp and lr
    stmfd   sp!, {r0-r3,r12}            @ Save non-callee saved registers
    bl      get_stack                   @ Get kernel stack of running process
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    mov     sp, r0                      @ Switch to kernel stack of running process
    msr     cpsr_c, #NoINT | SVC_Mode   @ Change mode to svc
    ldmfd   sp!, {r0-r3,r12}            @ Restore registers
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    sub     sp, sp, #16                 @ Save space for cpsr,pc,lr,sp
    stmfd   sp!, {r0-r12}               @ Save all registers
    msr     cpsr_c, #NoINT | SVC_Mode   @ Change mode to svc
    ldmfd   sp!, {r0-r2}                @ Load sp,lr,pc into r0-r2
    mrs     r3, spsr                    @ Load spsr into r3
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    add     r4, sp, #68
    stmfd   r4, {r0-r3}                 @ Save sp,lr,pc,cpsr
    mov     r0, sp                      @ Re-enable interrupts? Cannot take nested SWIs though
    
    @ Jump to C handler in AsmBridge.cc
    bl      swi_handler                 @ in AsmBridge.cc
        
    @ Restore the saved state
    add     r4, sp, #52
    ldmfd   r4, {r0-r3}                 @ Restore sp,lr,pc,cpsr from sys stack
    msr     cpsr_c, #NoINT | SVC_Mode   @ Change mode to svc
    stmfd   sp!, {r0-r3}                @ Save sp,lr,pc,cpsr onto svc stack
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    ldmfd   sp!, {r0-r12}               @ Load registers from sys stack
    msr     cpsr_c, #NoINT | SVC_Mode   @ Change mode to svc
    ldmfd   sp!, {sp,lr}^               @ Restore sp,lr from svc stack
    ldr     lr, [sp, #4]!               @ Restore cpsr from svc stack
    add     sp, sp, #4
    msr     spsr_cxsf, lr
    ldr     lr, [sp, #-8]               @ Restore pc from svc stack
    movs    pc,lr
    @MOVS PC, LR has a side effect of writing CPSR = SPSR, thus restoring the original mode

@ Handles exception caused by memory read error on instruction fetching
prefetch_abort:
    sub     lr, lr, #4
    stmfd   sp!, {lr}                   @ Save user mode pc. Note: We only expect PFAborts from user or system mode
    stmfd   sp!, {sp,lr}^               @ Save user mode sp and lr
    stmfd   sp!, {r0-r3,r12}            @ Save non-callee saved registers
    bl      get_stack                   @ Get kernel stack of running process
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    mov     sp, r0                      @ Switch to kernel stack of running process
    msr     cpsr_c, #NoINT | Abt_Mode   @ Change mode to abt
    ldmfd   sp!, {r0-r3,r12}            @ Restore registers
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    sub     sp, sp, #16                 @ Save space for cpsr,pc,lr,sp
    stmfd   sp!, {r0-r12}               @ Save all registers
    msr     cpsr_c, #NoINT | Abt_Mode   @ Change mode to abt
    ldmfd   sp!, {r0-r2}                @ Load sp,lr,pc into r0-r2
    mrs     r3, spsr                    @ Load spsr into r3
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    add     r4, sp, #68
    stmfd   r4, {r0-r3}                 @ Save sp,lr,pc,cpsr
    mov     r0, sp                      @ Re-enable interrupts? Cannot take nested SWIs though
    
    @ Jump to C handler in AsmBridge.cc
    bl      prefetch_abort_handler      @ in AsmBridge.cc
        
    @ Restore the saved state
    add     r4, sp, #52
    ldmfd   r4, {r0-r3}                 @ Restore sp,lr,pc,cpsr from sys stack
    msr     cpsr_c, #NoINT | Abt_Mode   @ Change mode to abt
    stmfd   sp!, {r0-r3}                @ Save sp,lr,pc,cpsr onto abt stack
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    ldmfd   sp!, {r0-r12}               @ Load registers from sys stack
    msr     cpsr_c, #NoINT | Abt_Mode   @ Change mode to abt
    ldmfd   sp!, {sp,lr}^               @ Restore sp,lr from abt stack
    ldr     lr, [sp, #4]!               @ Restore cpsr from abt stack
    add     sp, sp, #4
    msr     spsr_cxsf, lr
    ldr     lr, [sp, #-8]               @ Restore pc from abt stack
    movs    pc,lr
    @MOVS PC, LR has a side effect of writing CPSR = SPSR, thus restoring the original mode

@ Handle memory abort on data read
data_abort:
    sub     lr, lr, #8
    stmfd   sp!, {lr}                   @ Save user mode pc. Note: We only expect Data Aborts from user or system mode
    stmfd   sp!, {sp,lr}^               @ Save user mode sp and lr
    stmfd   sp!, {r0-r3,r12}            @ Save non-callee saved registers
    bl      get_stack                   @ Get kernel stack of running process
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    mov     sp, r0                      @ Switch to kernel stack of running process
    msr     cpsr_c, #NoINT | Abt_Mode   @ Change mode to abt
    ldmfd   sp!, {r0-r3,r12}            @ Restore registers
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    sub     sp, sp, #16                 @ Save space for cpsr,pc,lr,sp
    stmfd   sp!, {r0-r12}               @ Save all registers
    msr     cpsr_c, #NoINT | Abt_Mode   @ Change mode to abt
    ldmfd   sp!, {r0-r2}                @ Load sp,lr,pc into r0-r2
    mrs     r3, spsr                    @ Load spsr into r3
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    add     r4, sp, #68
    stmfd   r4, {r0-r3}                 @ Save sp,lr,pc,cpsr
    mov     r0, sp                      @ Should we re-enable interrupts? Cannot take nested SWIs though
    
    @ Jump to C handler in AsmBridge.cc
    bl      data_abort_handler          @ in AsmBridge.cc
        
    @ Restore the saved state
    add     r4, sp, #52
    ldmfd   r4, {r0-r3}                 @ Restore sp,lr,pc,cpsr from sys stack
    msr     cpsr_c, #NoINT | Abt_Mode   @ Change mode to abt
    stmfd   sp!, {r0-r3}                @ Save sp,lr,pc,cpsr onto abt stack
    msr     cpsr_c, #NoINT | Sys_Mode   @ Change mode to sys
    ldmfd   sp!, {r0-r12}               @ Load registers from sys stack
    msr     cpsr_c, #NoINT | Abt_Mode   @ Change mode to abt
    ldmfd   sp!, {sp,lr}^               @ Restore sp,lr from abt stack
    ldr     lr, [sp, #4]!               @ Restore cpsr from abt stack
    add     sp, sp, #4
    msr     spsr_cxsf, lr
    ldr     lr, [sp, #-8]               @ Restore pc from abt stack
    movs    pc,lr
    @MOVS PC, LR has a side effect of writing CPSR = SPSR, thus restoring the original mode
        
@ This is the handler for the interrupt address reserved by ARM Ltd for future use.  Here, we jump to uboot which ???
reserved:
    ldr pc, uboot

@ This is the reset exception handler and the initial boot code
reset:
    @ set up interrupt mode stack
    @ _*_stack are defined below
    msr cpsr_c, #NoINT | FIQ_Mode
    ldr sp, _FIQ_stack

    mov r0, #NoINT | IRQ_Mode
    msr cpsr_c, r0
    ldr sp, _IRQ_stack

    mov r0, #NoINT | Abt_Mode
    msr cpsr_c, r0
    ldr sp, _Abt_stack

    mov r0, #NoINT | Und_Mode
    msr cpsr_c, r0
    ldr sp, _Und_stack

    mov r0, #NoINT | SVC_Mode
    msr cpsr_c, r0
    ldr sp, _SVC_stack

    mov r0, #NoINT | Sys_Mode
    msr cpsr_c, r0
    ldr sp, _Usr_stack

    @ flush I/D caches
    mov r0, #0
    mcr p15, 0, r0, c7, c7, 0
    mcr p15, 0, r0, c8, c7, 0

    @ disable MMU stuff and caches
    mrc p15, 0, r0, c1, c0, 0
    bic r0, r0, #0x2300
    bic r0, r0, #0x87
    orr r0, r0, #0x2
    orr r0, r0, #0x1000
    mcr p15, 0, r0, c1, c0, 0

    @ perform platform initialization
    bl  platformsetup

    @ clear the .bss section
    @ BSS = used for uninitialized values
    @ Must be zero for C code
    ldr r0, _bss_start
    ldr r1, _bss_end
    mov r2, #0
    
@Loop through clearing memory until reach end of BSS section
clbss:
    str r2, [r0], #4
    cmp r0, r1
    blt clbss

    @ clear frame pointer
    mov fp, #0
                
    @ jump to KernelStart in Start.cc
    ldr pc, _KernelStart




@Everything below here are address constants used above

@Holds the address to start c code execution
_KernelStart:
    .word   KernelStart
    
@???
uboot:
    .word   0x100000

@These hold the address of the interrupt vector functions and are used by the LDRs in the interrupt vector
_reset_handler:
    .word   reset
_fiq_handler:
    .word   fiq
_irq_handler:
    .word   irq
_undef_handler:
    .word   undefined_instruction
_swi_handler:
    .word   software_interrupt
_prefetch_abort_handler:
    .word   prefetch_abort
_data_abort_handler:
    .word   data_abort
    
@??? Why are these needed?
__reset_handler:
    ldr pc, _reset_handler
__fiq_handler:
    ldr pc, _fiq_handler
__irq_handler:
    ldr pc, _irq_handler
__undef_handler:
    ldr pc, _undef_handler
__swi_handler:
    ldr pc, _swi_handler
__prefetch_abort_handler:
    ldr pc, _prefetch_abort_handler
__data_abort_handler:
    ldr pc, _data_abort_handler

@ Used by reset code above
_bss_start:
    .word   __bss_start__

_bss_end:
    .word   __bss_end__
    
@Defines a word of memory which holds the stack address for each mode
_SVC_stack:
    .word   SVC_stack

_IRQ_stack:
    .word   IRQ_stack

_FIQ_stack:
    .word   FIQ_stack

_Und_stack:
    .word   Und_stack

_Abt_stack:
    .word   Abt_stack

_Usr_stack:
    .word   Usr_stack

    .data
    
    .align @ Layout the memory
    @ This just runs through allocating various memory regions to different functionality

    @ reserve some stack space for supervisor state (4 KB)
    .skip   0x1000
SVC_stack:
    
    @ reserve some stack space for IRQs (4 KB)
    .skip   0x1000
IRQ_stack:

    @ reserve some stack space for FIQs (4 KB)
    .skip   0x1000
FIQ_stack:

    @ reserve some stack space for undefined instruction state (4 KB)
    .skip   0x1000
Und_stack:

    @ reserve some stack space for prefetch/data abort state (4 KB)
    .skip   0x1000
Abt_stack:

    @ reserve stack for user mode use (4 KB)
    .skip   0x1000
    
Usr_stack:
    .word   0xdeadbeef          @ This is special value used to indicate there is no user stack (yet)
                                @ Made distinctive (as per standard practice) to ease debugging

@ This is code inherited from the original boot loader 
@ It is called from ARMCPU.h
reset_cpu:
       ldr     r1, rstctl1     /* get clkm1 reset ctl */
       mov     r3, #0x0
       strh    r3, [r1]        /* clear it */
       mov     r3, #0x8
       strh    r3, [r1]        /* force dsp+arm reset */
_loop_forever:
       b       _loop_forever
       @ The above loop is just there to kill time while the processor is reseting
       @   The reset is asynchronous so we need to do something until it starts

@ this is a magic number which defines the code needed to reset the CPU
rstctl1:
       .word   0xfffece10
