#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Contributing Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MC68030CPU.c: MC68030 CPU class implementation.
 */

#include "Assert.h"
#include "CPUCollection.h"
#include "Domain.h"
#include "LOCK.h"
#include "MC68030CPU.h"
#include "MC68030Context.h"
#include "MC68030Interrupts.h"
#include "MC68030MMU.h"
#include "MC68030Vectors.h"
#include "OutputStream.h"
#include "Process.h"
#include "Timer.h"
#include "Util.h"



#ifdef __GNUC__
#pragma implementation
#endif

MC68030CPU::MC68030CPU( unsigned long identifier,
	ProcessContainer * initialIdleContainer )
	: CPU( identifier, initialIdleContainer )
{
	/*
	 * This constructor is called on behalf of this processor by the
	 * boot processor from $(MACHINE)Kernel::basicInitialize().
	 * At this stage, only one processor is running, so no locks
	 * are needed.
	 */
	Debug << this << "::MC68030CPU()\n" << eor;

	interruptVectorTable = 0;
	interruptTable = 0;

	/*
	 * Set the MMU pointer and other.
	 */
	_MMU = &myMMU;
}

void
MC68030CPU::chipInitialize()
{
	Debug << this << "::chipInitialize()\n" << eor;

	/*
	 * This code is called from CPU::bootResponsibility() as the
	 * first process is being dispatched on this processor.
	 * Since (this == thisCPU()), we can do per-processor
	 * initialization here, like loading the vector base
	 * register.
	 *
	 * Since VM is not yet enabled, DO NOT allocate any new
	 * objects here!
	 */
	Assert( this != 0 );
	Assert( this == thisCPU() );
	Assert( interruptsDisabled() );

	/*
	 * Set the vector base register to point to the interrupt
	 * vector table.
	 */
	Assert( interruptVectorTable != 0 );
	Assert( interruptTable != 0 );
	Assert( (char *) interruptTable ==
		(char *) interruptVectorTable + (nInterrupts * sizeof( PFV )));
	asm( "movec %0,%%vbr" : : "r" (interruptVectorTable) );

	/*
	 * For now, enable the processor instruction and data caches only
	 * if specified by a boot-time flag.
	 */
	extern int enableCaches;
	if( enableCaches ) {
		Console() << this << "::chipInitialize(): Enabling caches...\n"
			<< eor;
		enableProcessorCaches();
	}

	Debug << this << "::chipInitialize() done\n" << eor;
}

/*
 *
 */
void
MC68030CPU::preempt()
{
	Console() << this << "::preempt(): NOT IMPLEMENTED\n" << eor;
}

/*
 * The following three functions make sense ONLY when (this == thisCPU()).
 */
int
MC68030CPU::interruptsEnabled()
{
	unsigned short sr;

	Assert( this == thisCPU() );
	asm( "movew %%sr,%0" : "=g" (sr) );
	return( (sr & 0x0700) != 0x0700 );
}

int
MC68030CPU::interruptsDisabled()
{
	unsigned short sr;

	Assert( this == thisCPU() );
	asm( "movew %%sr,%0" : "=g" (sr) );
	return( (sr & 0x0700) == 0x0700 );
}

void
MC68030CPU::debugBreakpoint()
{
	asm( "trap #0xf" );
}

/*
 * Install an interrupt object to manage a vectored interrupt (trap or
 * interrupt) from the hardware.  For this processor, all interrupts
 * except the debug interrupts are handled by the same interruptAssist.
 * Immediately following the interruptVectorTable, we keep a parallel
 * table which contains the interrupt associated with each vector.
 * This way, interruptAssist() can easily look up the interrupt.
 */
Interrupt *
MC68030CPU::setInterrupt( int vector, Interrupt * interrupt )
{
	Debug << this << "::setInterrupt( vector " << vector << ", "
		<< interrupt << " )\n" << eor;

	Assert( interruptVectorTable != 0 );
	Assert( interruptTable != 0 );
	Assert( (char *) interruptTable ==
		(char *) interruptVectorTable + (nInterrupts * sizeof( PFV )));

	Assert( vector >= 0 );
	Assert( vector < nInterrupts );
	Assert( interrupt != 0 );

	/*
	 * Set the interrupt vector to point to the appropriate hook.
	 */
	if( (vector == 9) || (vector == 47) ) {
		interruptVectorTable[ vector ] =
					&MC68030CPU::debugInterruptAssist;
	} else {
		interruptVectorTable[ vector ] = &MC68030CPU::interruptAssist;
	}

	/*
	 * Save the pointer to the interrupt object itself.
	 */
	Interrupt * oldInterrupt = interruptTable[ vector ];
	interruptTable[ vector ] = interrupt;

	return( oldInterrupt );
}

/*
 * Return the interrupt object managing a vector.
 */
Interrupt *
MC68030CPU::interrupt( int vector )
{
	Assert( vector >= 0 );
	Assert( vector < nInterrupts );

	Interrupt * interrupt = interruptTable[ vector ];
	Assert( interrupt != 0 );

	Debug << this << "::interrupt( " << vector << " ): return "
		<< interrupt << "\n" << eor;
	return( interrupt );
}

/*
 * Halt the CPU. This code probably doesn't belong here but I can't think of
 * any better place to put it.  Maybe there should be a kernel method that
 * would halt all CPUs in the system.
 */
void
Halt()
{
	extern int useDebugger;
	if( useDebugger != 0 ) {
		thisCPU()->debugBreakpoint();
	}

	/*
	 * This should probably be done somewhere else!
	 */
	OnlineCPUs.remove( PerMemoryProcessorID() );

	/*
	 * Set interrupt level 0x7 to disable interrupts.
	 */
	asm( "oriw #0x0700,%sr" );

	/*
	 * Halt CPU via infinite loop.
	 */
	Console().criticalPrintf( "Halting.\n" );
	for ( ;; );
	Assert( NOTREACHED );
}

/*
 * MC68030CPU Class destructor.
 * Obvious side effect is to halt the processor. I doubt this could ever
 * actually be called!
 */
MC68030CPU::~MC68030CPU()
{
	Debug.printf( "CPU::~CPU()\n" );
	Halt();
}

/*
 * Stick in the interrupt object to handle the traps. Subclasses are
 * left to overload this method and install any machine dependent
 * interrupts.
 */
BusErrorInterrupt BusErrorInterrupt;
StrayInterrupt AddressErrorInterrupt( "Address Error Interrupt" );
StrayInterrupt IllegalInstructionInterrupt( "Illegal Instruction Interrupt" );
StrayInterrupt DivideByZeroInterrupt( "Divide By Zero Interrupt" );
StrayInterrupt CHKInterrupt( "CHK Instruction Interrupt" );
StrayInterrupt TRAPInterrupt( "TRAP Instruction Interrupt" );
StrayInterrupt PrivilegeViolationInterrupt( "Privilege Violation Interrupt" );
StrayInterrupt SpuriousInterrupt( "Spurious Interrupt Interrupt" );
StrayInterrupt MMUConfigurationInterrupt( "MMU Configuration Interrupt" );
ProxyGateTrap UNDInterrupt;

void
MC68030CPU::installInterrupts()
{
	Debug.printf( "MC68030CPU::installInterrupts()\n" );
	Assert( this != 0 );

	setInterrupt( BUS_ERROR_Vector, &BusErrorInterrupt );
	setInterrupt( ADDRESS_ERROR_Vector, &AddressErrorInterrupt );
	setInterrupt( ILLEGAL_INSTRUCTION_Vector,
					&IllegalInstructionInterrupt );
	setInterrupt( DIVIDE_BY_ZERO_Vector, &DivideByZeroInterrupt );
	setInterrupt( CHK_Vector, &CHKInterrupt );
	setInterrupt( TRAP_Vector, &TRAPInterrupt );
	setInterrupt( PRIVILEGE_VIOLATION_Vector,
					&PrivilegeViolationInterrupt );
	setInterrupt( SPURIOUS_Vector, &SpuriousInterrupt );
	setInterrupt( MMU_CONFIGURATION_Vector, &MMUConfigurationInterrupt );
}

int
MC68030CPU::getContextSaveHint( Interrupt *, char * )
{
/*
 * For now, everybody gets a small context.

	if( interrupt->needsAFullContext() ) {
		return( 1 );
	}
 */

	return( 0 );
}


#include "ClassInstaller.h"
DEFINECLASS MC68030CPU
