#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MC68030Context.c: Processor context for the MC68030.
 */

#include "CPU.h"
#include "Kernel.h"
#include "MC68030Context.h"
#include "MadAllocator.h"	/* TOTALLY BOGUS BUT HOPEFULLY TEMPORARY */
#include "OutputStream.h"
#include "PersistentMemoryObject.h"
#include "ThisProcess.h"


#ifdef __GNUC__
#pragma implementation
#endif


ProcessorContext *
thisProcessorContext()
{
	int oldSR = 0;

	/*
	 * Ask the CPU what process it is running only with interrupts
	 * disabled. This is to avoid the race of being rescheduled
	 * on another processor between getting the CPU pointer and
	 * getting the current process via that pointer.
	 */
	asm( "movew %%sr,%0;\
	      oriw  #0x0700,%%sr" : "=g" (oldSR) );

	CPU * cpu = thisCPU();
	if( cpu == 0 ) {
		asm( "movew %0,%%sr" : : "g" (oldSR) );
		return( 0 );
	}

	Process * process = cpu->currentProcess();
	asm( "movew %0,%%sr" : : "g" (oldSR) );

	/*
	 * Before the first process is dispatched, we return 0.
	 */
	if( process == 0 ) {
		return( 0 );
	}

	MC68030GenericContext * context =
		(MC68030GenericContext *) process->context();
	Assert( context != 0 );

	/*
	 * Make sure that the supervisor stack pointer is within bounds.
	 */
	const char * sp;
	asm( "movel %%sp,%0" : "=g" (sp) );

	Assert( sp <= context->supervisorStackBound );
	Assert( sp > context->supervisorStackBase );

	return( context );
}

MC68030GenericContext::MC68030GenericContext( Process * owner,
	StackSize stackSize )
	: ProcessorContext( owner )
{
	Debug << this << "::MC68030GenericContext()\n" << eor;

	int nbytes = 0;

	switch( stackSize ) {
		case small:
			nbytes = 2048;
			break;

		case normal:
		case application:
			nbytes = 4096;
			break;

		case large:
			nbytes = 4096;
			break;
		
		case huge:
			nbytes = 4096*2;
			break;
		
		default:
			Assert( NOTREACHED );
	}

	/*
	 * Allocate space for the process's supervisor stack and
	 * align the initial stack pointer.
	 */
	supervisorStackBase = new char[ nbytes ];
	Assert( supervisorStackBase != 0 );
	supervisorStackPointer = &supervisorStackBase[ nbytes ];
	supervisorStackPointer =
		(char *) ((int) supervisorStackPointer & 0xfffffff0);
	supervisorStackBound = supervisorStackPointer;

	Debug << this << "::MC68030GenericContext:\n\tsupervisorStackPointer "
		<< asAddress( supervisorStackPointer ) << "\n" << eor;
}

MC68030GenericContext::~MC68030GenericContext()
{
	delete supervisorStackBase;
}

/*
 * For now you can only do this to yourself.
 */
void
MC68030GenericContext::becomeInterruptable()
{
	Assert( this == thisProcessorContext() );

	asm( "andiw #0xf0ff,%sr" );
}

int
MC68030GenericContext::becomeUninterruptable()
{
	Assert( this == thisProcessorContext() );

	unsigned short oldSR;
	asm( "movew %%sr,%0;\
	      oriw  #0x0700,%%sr" : "=g" (oldSR) );
	return( (oldSR & 0x0700) != 0x0700 );
}

int
MC68030GenericContext::isInterruptable()
{
	Assert( this == thisProcessorContext() );

	unsigned short SR;
	asm( "movew %%sr,%0" : "=g" (SR) );
	return( (SR & 0x0700) != 0x0700 );
}

/*
 * Allocate a decent-size stack since memory won't actually be paged
 *   in until needed.  Backing store space will be used, however.
 * Will probably want to have a 'StackMemoryObject' per domain from
 *   which to allocate stacks instead of creating a whole new memory
 *   object for every stack.
 */
static const int UserStackSize = 4096*4;

MC68030ApplicationContext::MC68030ApplicationContext( Process * owner,
	PFV entry, StackSize stackSize, Domain *initialDomain,
	ObjectProxy *defaultProxy, int a0, int a1, StackSize userStackSize,
	char * userStackPointer = 0, int usesProxies = 1 )
		: MC68030GenericContext( owner, stackSize )
{
	Debug << "MC68030ApplicationContext::ctor()\n" << eor;

	int stackSize;
	switch( userStackSize ) {
		case none:
			stackSize = 0;
			break;
		case tiny:
			stackSize = 4096;
			break;
		case small:
			stackSize = 4096*2;
			break;
		case normal:
		case application:
			stackSize = 4096*4;
			break;
		case huge:
			stackSize = 4096*16;
			break;
		case gigantic:
			stackSize = 4096*1024;
			break;
		default:
			stackSize = (int)userStackSize;
	}

	if( ! userStackPointer ) {
		PersistentMemoryObject * memObj = 
			TheKernel->allocateMemoryObjectForBackingStore();
		Assert( memObj != 0 );

		memObj->setNumberOfUnits( stackSize /
			 memObj->unitSize() );

		userStackPointer = (char *)
			 initialDomain->add( memObj, ReadWrite );

		Assert( userStackPointer != 0 );
	}

	/*
	 * Leave some padding at the top of the stack because
	 *   proxy calls copy a fixed number of arguments.
	 */
	if( usesProxies )
		userStackPointer += ( stackSize - ( 5 * sizeof( int ) ) );
	else	userStackPointer += stackSize;
        InitialApplicationContextStackDescription * isd =
		(InitialApplicationContextStackDescription *)
		(supervisorStackPointer -
			sizeof( InitialApplicationContextStackDescription ));

	supervisorStackPointer = (char *) isd;

	/*
	 * Set the SR for user mode, interrupts enabled.  These initial
	 * stack contents are used in the assembler routine
	 * MC68030ApplicationContext::restoreFromInitialContext().
	 */

	isd->arg0 = a0;
	isd->arg1 = a1;
	isd->userStackPointer = userStackPointer;
	isd->sr = 0x0000;	// always use Interrupt stack
	isd->entryPoint = entry;
	isd->format = 0x0000;
}

MC68030SystemContext::MC68030SystemContext( Process * owner, PFV entry,
	StackSize stackSize, int a0, int a1 )
	: MC68030GenericContext( owner, stackSize )
{
	/*
	 * Put a copy of the initial entry point and its arguments in
	 * the register save area.  Also, set a flag to indicate that
	 * this process should be interruptable.  These are used by
	 * MC68030SystemContext::initialRestoreAssist(), not by
	 * any assembler routines.
	 */
	pc = (int) entry;
	d2 = a0;
	d3 = a1;
	d4 = 1;
}

void
MC68030SystemContext::initialRestoreAssist( Process * previousProcess )
{
	typedef void (* PFVII) (int, int);
	PFVII entryPoint = (PFVII) pc;

	restoreAssist( previousProcess );
	if( d4 ) becomeInterruptable();
	(* entryPoint)( d2, d3 );

	Assert( process() != 0 );
	process()->die();
	Assert( NOTREACHED );
}

MC68030UninterruptableSystemContext::MC68030UninterruptableSystemContext(
	Process * owner, PFV entry, StackSize stackSize, int a0, int a1 )
	: MC68030SystemContext( owner, entry, stackSize, a0, a1 )
{
	/*
	 * Indicate that this process should not be interruptable (see
	 * MC68030SystemContext::initialRestoreAssist()).
	 */
	d4 = 0;
}


#include "ClassInstaller.h"
DEFINECLASS MC68030SystemContext
DEFINECLASS MC68030UninterruptableSystemContext
DEFINECLASS MC68030GenericContext
DEFINECLASS MC68030ApplicationContext
