#ifdef __GNUC__
#pragma implementation
#endif

/*************
 * 
 * All commands and responses are sent with a packet which includes a 
 * checksum.  A packet consists of 
 * 
 * $<packet info>#<checksum>.
 * 
 * where
 * <packet info> :: <characters representing the command or response>
 * <checksum>    :: < two hex digits computed as modulo 256 sum of <packetinfo>>
 * 
 * When a packet is received, it is first acknowledged with either '+' or '-'.
 * '+' indicates a successful transfer.  '-' indicates a failed transfer.
 * 
 * Example:
 * 
 * Host:                  Reply:
 * $m0,10#2a               +$00010203040506070809101112131415#42
 * 
 ****************************************************************************/

#include "InputStream.h"
#include "MC68030Interrupts.h"
#include "MC68030Vectors.h"
#include "MacintoshCPU.h"
#include "MacintoshInterrupts.h"
#include "OutputStream.h"
#include "String.h"
#include "Util.h"


#ifdef __GNUC__
#pragma implementation
#endif

/************************************************************************
 *
 * external low-level support routines 
 *
 */
typedef void (*InterruptHook)(int);   /* pointer to function with int parm */
typedef void (*Function)();           /* pointer to a function */

extern char debugGetChar();		/* read and return a single char */

extern Function setInterrupt( int, Function );
					/* assign an interrupt handler */

extern OutputStream * debugOutputStream;
extern InputStream * debugInputStream;

//extern void MC68901BootWrite( const char * );
//#define DEBUG( s )	MC68901BootWrite( s )
//extern OutputStream * ConsoleOutputStream;
//#define DEBUG( s ) ConsoleOutputStream->write( s, strlen( s ) );
extern OutputStream * ConsoleOutputStream;
#define DEBUG1( s ) ConsoleOutputStream->write( s, strlen( s ) );
#define DEBUG( s )

char
debugGetChar()
{
	char c;

	debugInputStream->read( &c, 1 );
	return( c );
}

/*
 * NUMREGBYTES determines the size of the area needed to save all interesting
 * registers.  This must correspond to the size of the register area that
 * the remote debugger expects.  188 bytes is taken from gdb/sun3.h.
 */
#define NUMREGBYTES 188
int registers[NUMREGBYTES/4];

enum regnames { D0, D1, D2, D3, D4, D5, D6, D7, 
		A0, A1, A2, A3, A4, A5, A6, A7, 
		PS, PC };

int _d0;	/* Data Registers */
int _d1;
int _d2;
int _d3;
int _d4;
int _d5;
int _d6;
int _d7;

int _a0;	/* Address Registers */
int _a1;
int _a2;
int _a3;
int _a4;
int _a5;
int _a6;

int _pc;	/* Program Counter */

int _isp;	/* Interrupt Stack Pointer */
int _msp;	/* Master Stack Pointer */
int _usp;	/* User Stack Pointer */
short _sr;	/* Status Register */
int _vbr;	/* Vector Base Register */
int _cacr;	/* Cache Control Register */
int _caar;	/* Cache Address Register */
int _crp[2];	/* CPU Root Pointer */
int _srp[2];	/* Supervisor Root Pointer */
int _tc;	/* Translation Control Register */
int _tt0;	/* Transparent Translation Register 1 */
int _tt1;	/* Transparent Translation Register 2 */
int _mmusr;	/* MMU Status Register */

/*
 * BUFFER_SIZE defines the maximum number of characters in any packet
 * sent to or received from the remote debugger.  It must be at least
 * NUMREGBYTES * 2 + 5.
 */
#define BUFFER_SIZE 400

static char remoteDebug = 0;

static char hexchars[]="0123456789abcdef";

#define DEBUG_STACK_SIZE	1024

int debugStack[ DEBUG_STACK_SIZE ];
int * debugStackPointer = &debugStack[ DEBUG_STACK_SIZE - 1 ];
int * debugSupervisorSP;

/*
 * The size of the interrupt stack on the 68020 varies with the type of
 * interrupt.  The following table is the number of 8-bit BYTES used
 * for each interrupt format.
 */
int interruptStackFrameSize[] = { 8,8,12,8,8,8,8,8,8,20,32,92,8,8,8,8 };


/*
 * This function is called immediately when a level 7 interrupt occurs
 * if the previous interrupt level was 7 then we're already servicing
 * this interrupt and an rte is in order to return to the debugger.
 */
extern void
debugLevel7Interrupt();

extern void
debugReturnFromInterrupt();

int
hex( char c )
{
	if( (c >= 'a') && (c <= 'f') ) {
		return( c - 'a' + 10 );
	} else if( (c >= '0') && (c <= '9') ) {
		return (c - '0');
	} else {
		return( 0 );
	}
}


/*
 * Scan for the sequence $<data>#<checksum>
 */
static void
receivePacket( char * buffer )
{
	unsigned char checksum = 0;
	unsigned char xmitcsum = 0;
	int i;
	int count;
	char c;
  
	do {
		/*
		 * Wait around for the start character,
		 * ignore all other characters
		 */
		while( (c = debugGetChar()) != '$' )
			; 

		checksum = 0;
		count = 0;

		/*
		 * Now, read until a # or end of buffer is found
		 */
		while( count < BUFFER_SIZE ) {
			c = debugGetChar();
			if( c == '#' ) {
				break;
			}
			checksum = checksum + c;
			buffer[ count++ ] = c;
		}
		buffer[ count ] = 0;

		if( c == '#' ) {
			/*
			 * Compute and check checksum
			 */
			xmitcsum = hex( debugGetChar() ) << 4;
			xmitcsum += hex( debugGetChar() );
      			if( checksum != xmitcsum) {
				DEBUG1( "Bad checksum\n\r" );
				DEBUG1( "Buffer: " );
				DEBUG1( buffer );
				DEBUG1( "\n\r" );
				debugOutputStream->write( "-", 1 );
			} else {
				debugOutputStream->write( "+", 1 );

				/*
				 * If a sequence char is present,
				 * reply the sequence ID
				 */
				if( buffer[2] == ':' ) {
					debugOutputStream->write( buffer, 2 );

					/*
					 * Rmove sequence chars from buffer
					 */
					count = strlen( buffer );
					for( i = 3; i <= count; i++ ) {
						buffer[ i - 3 ] = buffer[ i ];
					}
				} 
			} 
		} 
	} while( checksum != xmitcsum );
}


static unsigned char
checksum( char * buffer )
{
	for( unsigned char c = 0; *buffer; c += *buffer++ )
		;
	return( c );
}

/*
 * Send the packet in buffer.  The host gets one chance to read it.  
 * This routine does not wait for a positive acknowledge.
 */
static void
sendPacket( char * buffer )
{
	/*
	 * $<packet info>#<checksum>
	 */
	debugOutputStream->write( "$", 1 );
	debugOutputStream->write( buffer, strlen( buffer ) );
	debugOutputStream->write( "#", 1 );

	unsigned char c = checksum( buffer );
	debugOutputStream->write( &hexchars[ c >> 4 ], 1 );
	debugOutputStream->write( &hexchars[ c % 16 ], 1 );
}


/*
 * Convert the memory pointed to by mem into hex, placing result in buf
 * return a pointer to the last char put in buf (null)
 */
static void
sendMemoryPacket( char * mem, int count)
{
	int i;
	unsigned char c;
	static char buffer[ BUFFER_SIZE ];
	char * ptr;

	ptr = &buffer[0];
	for( i = 0; i < count; i++ ) {
		c = *mem++;
		*ptr++ = hexchars[ c >> 4 ];
		*ptr++ = hexchars[ c % 16 ];
	}
	*ptr = 0;
	sendPacket( buffer );
}

/*
 * Convert the hex array pointed to by buf into binary to be placed in mem
 * return a pointer to the character AFTER the last byte written
 */
static char *
receiveMemory( char * buf, char * mem, int count )
{
	int i;
	unsigned char c;

	for( i = 0; i < count; i++ ) {
		c = hex( *buf++ ) << 4;
		c = c + hex( *buf++ );
		*mem++ = c;
	}
	return( mem );
}

/*
 * Convert count characters representing a hexadecimal value into its
 * value as an unsigned int.
 */
unsigned int
hex2int( char * buf, int count )
{
	unsigned int val = 0;
	while( count-- > 0 ) {
		val <<= 4;
		val += hex( *buf++ );
	}
	return( val );
}

/*
 * Take a 68000 interrupt number, map in into a "corresponding"
 * UNIX signal number, and send a packet to the remote debugger.
 */
static void
sendSignalPacket( int interruptVector )
{
	static char buffer[] = "S00";
	int sigval;

	switch( interruptVector ) {
		case 2 : sigval = 10; break; /* bus error           */
		case 3 : sigval = 10; break; /* address error       */
		case 4 : sigval = 4;  break; /* illegal instruction */
		case 5 : sigval = 8;  break; /* zero divide         */
		case 6 : sigval = 16; break; /* chk instruction     */
		case 7 : sigval = 16; break; /* trapv instruction   */
		case 8 : sigval = 11; break; /* privilege violation */
		case 9 : sigval = 5;  break; /* trace trap          */
		case 10: sigval = 4;  break; /* line 1010 emulator  */
		case 11: sigval = 4;  break; /* line 1111 emulator  */
		case 31: sigval = 2;  break; /* interrupt           */
		case 33: sigval = 5;  break; /* breakpoint          */
		case 47: sigval = 5;  break; /* breakpoint	    */
		case 48: sigval = 8;  break; /* floating point err  */
		case 49: sigval = 8;  break; /* floating point err  */
		case 50: sigval = 8;  break; /* zero divide         */
		case 51: sigval = 8;  break; /* underflow           */
		case 52: sigval = 8;  break; /* operand error       */
		case 53: sigval = 8;  break; /* overflow            */
		case 54: sigval = 8;  break; /* NAN                 */
		default: sigval = 7;         /* "software generated"*/
	}

	buffer[1] = hexchars[ sigval >> 4 ];
	buffer[2] = hexchars[ sigval % 16 ];
	sendPacket( buffer );
}

extern int bootCPUInitialized;

static setjmpBuffer environment;
static setjmpBuffer topLevelEnvironment;

/*
 * We have caused a bus error while executing a debugger command.
 * This actually gets called with a bunch of junk on the stack, but
 * since we immediately do a longjmp, we don't really care.
 * Perform a longjmp to return and handle the error.
 */
static void
busErrorHandler()
{
	longjmp( environment, 1 );
}

void
unexpectedInterruptHandler()
{
	longjmp( topLevelEnvironment, 1 );
}

static void
readMemory( char * inbuffer )
{
	char * buffer, * cursor;
	int numberLength;
	int addr, length;
	Function oldBusErrorHandler;
	Interrupt * oldInterrupt;

	DEBUG( "In readMemory()\n\r" );

	/*
	 * Note that this will only work after CPU::chipInitialize()
	 * has been called to set the VBR.
	 */
//	if( bootCPUInitialized )
//		oldInterrupt = thisCPU()->setInterrupt( BUS_ERROR_Vector,
//			&busErrorInterrupt );
//	else

	oldBusErrorHandler = setInterrupt( BUS_ERROR_Vector, busErrorHandler );

	DEBUG( "In readMemory: busError interrupt replaced\n\r" );

	if( setjmp( environment ) == 0 ) {

		DEBUG( "In readMemory: setjmp returned 0\n\r" );

		buffer = cursor = inbuffer;
		numberLength = 0;
		while( *cursor++ != ',' ) numberLength++;
		addr = hex2int( buffer, numberLength );
	
		DEBUG( "In readMemory: got address\n\r" );

		buffer = cursor;
		numberLength = 0;
		while( *cursor++ ) numberLength++;
		length = hex2int( buffer, numberLength );
	
		DEBUG( "In readMemory: got length\n\r" );

		sendMemoryPacket( (char *) addr, length );

		DEBUG( "In readMemory: memoryPacket sent\n\r" );
	} else {
		DEBUG( "In readMemory: setjmp didn't return 0\n\r" );

		sendPacket( "E03" );
	}

	/*
	 * Reinstall the usual bus error interrupt handler.
	 */
//	if( bootCPUInitialized )
//		thisCPU()->setInterrupt( BUS_ERROR_Vector, oldInterrupt );
//	else

	setInterrupt( BUS_ERROR_Vector, oldBusErrorHandler );

	DEBUG( "In readMemory: returning\n\r" );
}

static void
writeMemory( char * inbuffer )
{
	char * buffer, * cursor, * pointer;
	int numberLength;
	int addr, length;
	Function oldBusErrorHandler;
	Interrupt * oldInterrupt;

	/*
	 * Note that this will only work after CPU::chipInitialize()
	 * has been called to set the VBR.
	 */
//	if( bootCPUInitialized )
//		oldInterrupt = thisCPU()->setInterrupt( BUS_ERROR_Vector,
//			&busErrorInterrupt );
//	else

	/*
	 * Now the debugger always uses its own vector table.
	 */
	oldBusErrorHandler = setInterrupt( BUS_ERROR_Vector, busErrorHandler );

	if( setjmp( environment ) == 0 ) {
		buffer = cursor = inbuffer;
		numberLength = 0;
		while( *cursor++ != ',' ) numberLength++;
		addr = hex2int( buffer, numberLength );
	
		buffer = cursor;
		numberLength = 0;
		while( *cursor++ != ':' ) numberLength++;
		length = hex2int( buffer, numberLength );
	
		pointer = cursor;
	
		receiveMemory( pointer, (char *) addr, length );
		sendPacket( "OK" );
	} else {
		sendPacket( "E03" );
	}

	/*
	 * Reinstall the usual bus error interrupt handler.
	 */
//	if( bootCPUInitialized )
//		thisCPU()->setInterrupt( BUS_ERROR_Vector, oldInterrupt );
//	else

	setInterrupt( BUS_ERROR_Vector, oldBusErrorHandler );
}

static void
stepOrContinue( char * inbuffer )
{
	char * cursor;
	int numberLength;

	/*
	 * If the optional address is supplied, set the PC to it.
	 */
	cursor = &inbuffer[1];
	numberLength = 0;
	while( *cursor++ ) numberLength++;
	if( numberLength > 0 ) {
		registers[ PC ] = hex2int( &inbuffer[1], numberLength );
	}

	/*
	 * Set the trace bit if we're stepping; clear it otherwise.
	 */
	if( inbuffer[0] == 's' ) {
		registers[ PS ] |= 0x8000;
	} else {
		registers[ PS ] &= 0x7fff;
	}

	debugReturnFromInterrupt();
//	resumeExecution();
}

Function debugVectorTable[256];
Function oldVBR = 0;

Function
setInterrupt( int vector, Function handler )
{
	Function oldHandler;

	oldHandler = debugVectorTable[vector];
	debugVectorTable[vector] = handler;
	return( oldHandler );
}

/*
 * This function is used to set up interrupt handlers for tracing and 
 * breakpoints.
 */
void
debugInitialize()
{
	for( int interrupt = 2; interrupt < 256; interrupt++ ) {
		setInterrupt( interrupt,
			&MC68030CPU::debugInterruptAssist );   
	}

	int vectorTable = (int) &debugVectorTable[0];
	asm( "movec %%vbr,%0" : : "r" (oldVBR) );
	asm( "movec %0,%%vbr" : : "r" (vectorTable) );

	debugOutputStream->write( "Hello, debugger!\n\r", 18 );
}

/*
 * This function does all command procesing for interfacing to gdb.
 * The commands we understand are the following:
 *
 *	? *	d *	g *	G *	m *	M *	c *	s *	k
 */

int debugSyncFlag = 0;
SynchronizeInterrupt DebugSyncInterrupt( &debugSyncFlag );

extern CPUCollection RunningCPUs;

/*
 * If any other processors are running, stop them before
 * continuing.  Note that we may have already stopped them
 * at a previous breakpoint.
 */
void
stopOtherProcessors()
{
	if( debugSyncFlag == 0 ) {
		debugSyncFlag = 1;

		for( int i = 0; i < MAXPROCESSORS; i++ ) {
			MacintoshCPU * cpu = (MacintoshCPU *)
						RunningCPUs.cpu( i );
			if( (cpu != 0) && (cpu != thisCPU()) ) {
				cpu->asynchronousRaiseInterrupt(
						&DebugSyncInterrupt );
			}
		}
	}
}

void
resetInterruptVectors()
{
	for( int vector = 2; vector < 256; vector++ ) {
		setInterrupt( vector, oldVBR );
	}
}

void
processCommands( int interruptVector )
{
	DEBUG( "In processCommands!\n\r" );

	char inbuffer[ BUFFER_SIZE ];

	DEBUG( "Stopping other processors\n\r" );
	stopOtherProcessors();

//	DEBUG( "Resetting interrupt vectors\n\r" );
//	resetInterruptVectors();

	/*
	 * Indicate to remote debugger that interrupt has occured.
	 */
	DEBUG( "Sending signal to remote\n\r" );
	sendSignalPacket( interruptVector );

	while( 1 ) { 
		receivePacket( inbuffer );
		switch( inbuffer[0] ) {

			case '?' : sendSignalPacket( interruptVector );
				   break; 

			case 'd' : remoteDebug = !(remoteDebug);
				   break; 

			case 'g' : sendMemoryPacket( (char *) registers,
							NUMREGBYTES );
				   break;

			case 'G' : receiveMemory( &inbuffer[1],
					(char *) registers, NUMREGBYTES );
				   sendPacket( "OK" );
				   break;

			case 'm' : readMemory( &inbuffer[1] );
				   break;
      
			case 'M' : writeMemory( &inbuffer[1] );
				   break;

			case 'c' : debugSyncFlag = 0;

			case 's' : stepOrContinue( inbuffer );
				   break;

			case 'k' :
				   break;
		}
	}
}

