#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Contributing Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Contributing Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * MC68030Interrupts.cc: MC68030 hardware interrupt classes implementations.
 */

#include "AddressTranslation.h"
#include "AddressTranslator.h"
#include "Assert.h"
#include "CPU.h"
#include "Domain.h"
#include "FaultErrorCode.h"
#include "MC68030Context.h"		/* interruptStackFrame */
#include "MC68030Interrupts.h"
#include "MC68030Context.h"
#include "OutputStream.h"
#include "ThisProcess.h"
#include "ProcessContainer.h"
#include "Util.h"


#ifdef __GNUC__
#pragma implementation
#endif


#ifdef INSTRUMENT_INTERRUPT
void InterruptConstructInstruments () {}
#endif

#ifdef INSTRUMENT_SVCTRAP
void SVCTrapConstructInstruments () {}
#endif

void
BusErrorInterrupt::basicRaise( char * info )
{
	interruptStackFrame * frame = (interruptStackFrame *) info;
	void * pc = (void *) frame->pc;
	unsigned short sr = frame->sr;
	unsigned int specialStatusRegister = frame->specialStatusRegister;
	char * address = (char *) frame->faultAddress;

	extern void MC68901BootWrite( const char * );
	MC68901BootWrite( "BusError!\n\r" );

	extern OutputStream * ConsoleOutputStream;
	*ConsoleOutputStream << "BusError: address: " << asAddress( address )
		<< ", SSR: " << asHex( specialStatusRegister ) << ", PC: "
		<< asHex( (int) pc ) << "\n" << eor;

	Process * process = thisProcess();
	Assert( process != 0 );
	Domain * domain = process->domain();
	Assert( domain != 0 );

	Debug << this << "::basicRaise: fault at " << asAddress( address )
		<< " in\n\t" << process
		<< "\n\tpc: " << asHex( (unsigned int) pc )
		<< ", sr: " << asHex( sr ) << "\n" << eor;

	AccessType attemptedAccess;
	AddressTranslationFailureCode code = thisCPU()->MMU()->
		faultType( address, specialStatusRegister,
			domain->translation(), attemptedAccess );

	switch( code ) {
		case MMUCacheMiss:
			Debug << this << "::basicRaise: MMUCacheMiss\n" << eor;
			return;
		case SomethingElse:
			Console().panic( "Cant deal with breakpoints (yet) addr=%A\n",
				address );
			Assert( NOTREACHED );
		default:
			break;
	}

	/*
	 * If it wasn't a simple MMU cache miss, only the Domain
	 * knows what to do next.
	 */
	FaultErrorCode reason = domain->repairFault( address, attemptedAccess );
	switch( reason ) {
		case AccessOfUnmanagedMemory:
			process->console() << "No manager for address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->pc )
			   << " at time of access)\n" << eor;
			process->die();
			Assert( NOTREACHED );

		case ReadOfProtectedSystemMemory:
			process->console()
			   << "Attempted read of system memory at address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->pc )
			   << " at time of access)\n" << eor;
			process->die();
			Assert( NOTREACHED );

		case WriteOfProtectedSystemMemory:
			process->console()
			   << "Attempted write of system memory at address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->pc )
			   << " at time of access)\n" << eor;
			process->die();
			Assert( NOTREACHED );

		case WriteOfReadOnlyMemory:
			process->console()
			   << "Attempted write of read-only memory at address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->pc )
			   << " at time of access)\n" << eor;
			process->die();
			Assert( NOTREACHED );

		case DynamicWriteOfReadOnlyMemory:
		case DynamicAccessViolation:
			process->console()
			   << "Access violation referencing memory at address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->pc )
			   << " at time of access)\n" << eor;
			process->die();
			Assert( NOTREACHED );
		
		/*
		 * These are all recoverable and the instruction can
		 * be restarted/resumed.
		 */
		case MissingAddressTranslationMapping:
		case MissingPhysicalMemory:
		case InsufficientProtectionLevel:
			break;
	}

	/*
	 * All done so return.  This results in retrying/restarting the
	 * instruction which caused the fault.
	 */
	Debug << this << "::basicRaise: fault repaired - retrying\n" << eor;
}


void
LongjmpInterrupt::basicRaise( char * info )
{
	longjmp( _buffer, 1 );
}


void
ProxyGateTrap::basicRaise( char * info )
{
	interruptStackFrame * frame = (interruptStackFrame *) info;
	Debug << "ProxyGateTrap::basicRaise(): frame->d0 = "
		<< frame->d0 << " frame->d1 = " << frame->d1 << "\n" << eor;
}

void
BadObject()
{
	thisProcess()->console() << "Call to invalid persistent object\n"
		<< eor;
	thisProcess()->die();
	Assert( NOTREACHED );
}



void
StrayInterrupt::basicRaise( char * info )
{
	interruptStackFrame * frame = (interruptStackFrame *) info;

	Console() << "StrayInterrupt (" << message << ")\n\t"
		<< thisProcess() << "\n\tpc = "
		<< asAddress( (char *) frame->pc )
		<< ", sr = " << asHex( frame->sr ) << ", vector = "
		<< (frame->vector & 0x0fff) / 4 << "\n" << eor;
	Console().panic( "Stray Interrupt (%s)\n", message );
	Assert( NOTREACHED );
}


#include "ClassInstaller.h"
DEFINECLASS BusErrorInterrupt
DEFINECLASS LongjmpInterrupt
DEFINECLASS ProxyGateTrap
DEFINECLASS("StrayInterrupt",StrayInterrupt,Interrupt,0,0);
