#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MC68030MMU.c: An MC68030 as an AddressTranslator.
 */

#include "ADDRESS_TRANSLATION.h"
#include "Assert.h"
#include "LOCK.h"
#include "MC68030MMU.h"
#include "OutputStream.h"
#include "PAGE_TABLE_ENTRY.h"
#include "PageTableConstants.h"
#include "Util.h"


#ifdef __GNUC__
#pragma implementation
#endif


/*
 * TC (translation control) register definitions.
 */
static const unsigned int TC_E   = 0x80000000;	// Enable
static const unsigned int TC_SRE = 0x02000000;	// Supervisor root pointer 
static const unsigned int TC_FCL = 0x01000000;	// Function code lookup
static const unsigned int TC_PS  = 0x00c00000;	// Page size (4K => 22 bits)
static const unsigned int TC_IS  = 0x00000000;	// Initial shift
static const unsigned int TC_TIA = 0x0000a000;	// Table index A (10 bits)
static const unsigned int TC_TIB = 0x00000a00;	// Table index B (10 bits)
static const unsigned int TC_TIC = 0x00000000;	// Table index C
static const unsigned int TC_TID = 0x00000000;	// Table index D

/*
 * Set up the Translation Control word to indicate mapping enabled,
 * 4K page size, 10-bit first-level index, and 10-bit second-level
 * index.  The 68030 doesn't have a 'move immediate to %tc'
 * instruction, so we have to allocate a word statically here.
 */
//static unsigned int TC = TC_E | TC_SRE | TC_PS | TC_TIA | TC_TIB;
static unsigned int TC = TC_E | TC_PS | TC_TIA | TC_TIB;


MC68030MMU::MC68030MMU()
{
	Debug << this << "::MC68030MMU()\n" << eor;
}

void
MC68030MMU::enable()
{
	Debug << this << "::enable()\n" << eor;

	/*
	 * Disable the processor data and instruction caches temporarily
	 * to avoid problems here.
	 */
	int cachesWereEnabled = disableProcessorCaches();
	asm( "pflusha" );
	asm( "pmove %0,%%tc" : : "m" (TC) );

	if( cachesWereEnabled ) enableProcessorCaches();

	Debug << this << "::enable: translation enabled\n" << eor;
}

int
MC68030MMU::enabled()
{
	Debug << this << "::enabled()\n" << eor;

	static unsigned int TC;
	asm( "pmove %%tc,%0" : "=m" (TC) );
	return( (TC & TC_E) != 0 );
}

void
MC68030MMU::basicActivate( AddressTranslation * at )
{
	Debug << this << "::basicActivate\n" << eor;

	ADDRESS_TRANSLATION * translation = (ADDRESS_TRANSLATION *) at;
	Assert( this != 0 );
	Assert( translation != 0 );

	const POINTER_TABLE_ENTRY * systemPointerTable =
		translation->systemPointerTable();
	Assert( systemPointerTable != 0 );
	Assert( PageAligned( (const char *) systemPointerTable ) );

	const POINTER_TABLE_ENTRY * userPointerTable =
		translation->userPointerTable();
	Assert( userPointerTable != 0 );
	Assert( PageAligned( (const char *) userPointerTable ) );

	/*
	 * Since we use static data to initialize the root pointers,
	 * we must acquire a lock while we're using it.
	 */
	static LOCK activateLock;
	activateLock.acquire();

	static unsigned int systemRoot[2];
	systemRoot[0] = 0x80000002; // no limits, 4-byte descrs
	systemRoot[1] = (unsigned int) systemPointerTable;

	static unsigned int userRoot[2];
	userRoot[0] = 0x80000002; // no limits, 4-byte descrs
	userRoot[1] = (unsigned int) userPointerTable;

	Debug << this << "::activate( " << at << " )\n" << eor;
	Debug << this << "::activate: SRP "
		<< asHex( systemRoot[0] ) << " "
		<< asHex( systemRoot[1] ) << "\n" << eor;
	Debug << this << "::activate: CRP "
		<< asHex( userRoot[0] ) << " "
		<< asHex( userRoot[1] ) << "\n" << eor;

	if( processorCachesEnabled() ) {
		disableProcessorCaches();
		enableProcessorCaches();
	}
	asm( "pmove %0,%%srp" : : "m" (systemRoot[0]) );
	asm( "pmove %0,%%crp" : : "m" (userRoot[0]) );

	activateLock.release();

	Debug << this << "::activate: root pointers loaded\n" << eor;
}

void
MC68030MMU::flushCache( const char * addr, unsigned int length )
{
	Debug << this << "::flushCache( " << asAddress( addr )
		<< ", " << length << " )\n" << eor;

	/*
	 * This flushes all entries; could be optimized to flush only
	 * the appropriate entry.  This takes advantage of the fact
	 * that we flush the processor caches when we enable them.
	 */
	if( processorCachesEnabled() ) {
		disableProcessorCaches();
		enableProcessorCaches();
	}
	asm( "pflusha" );
}

AddressTranslationFailureCode
MC68030MMU::faultType( const char * address, unsigned int statusWord, 
	AddressTranslation * at, AccessType & attemptedAccess )
{
	/*
	 * Return one of
	 *
	 *	MMUCacheMiss
	 *	NonResidentMemory
	 *	ProtectionViolation
	 *	SomethingElse
	 *
	 * and set attemptedAccess appropriately.
	 */

	Debug << this << "::faultType( " << asAddress( address )
		<< ", statusWord " << asHex( statusWord )
		<< ", ... )\n" << eor;

	ADDRESS_TRANSLATION * translation = (ADDRESS_TRANSLATION *) at;
	
	int addressSpace = statusWord & 0x7;
	int dataFault = statusWord & 0x100;
	int instructionFault = statusWord & 0xc000;
	int readWrite = statusWord & 0x40;

	int user = 0;
	switch( addressSpace ) {
		case 0x1:
		case 0x2:
			user = 1;
			break;
		case 0x5:
		case 0x6:
			user = 0;
			break;
		default:
			Assert( NOTREACHED );
	}

	int write = 0;
	if( dataFault && (readWrite == 0) ) {
		write = 1;
	}

	if( user ) {
		attemptedAccess = write ? ApplicationWrite : ApplicationRead;
	} else {
		attemptedAccess = write ? SystemWrite      : SystemRead;
	}
	Debug << this << "::faultType: attemptedAccess was " << attemptedAccess
		<< "\n" << eor;

	if( dataFault && instructionFault ) {
		/*
		 * This can really happen, but I'm not sure how to handle
		 * it yet.
		 */
		Console() << this << "::faultType(): Simultaneous data "
			<< "and instruction faults!\n" << eor;
		Assert( NOTREACHED );
	}

	/*
	 * Now that we've figured out the attemptedAccess, try to determine
	 * why the fault occurred.
	 */
	AddressTranslationFailureCode error = translation->
		determineError( address, attemptedAccess );

	Debug << this << "::faultType: translation says ";
	if( error == NonResidentMemory ) {
		Debug << "NonResidentMemory\n" << eor;
		return( error );
	} else if( error == ProtectionViolation ) {
		Debug << "ProtectionViolation\n" << eor;
		return( error );
	} else if( error == MMUCacheMiss ) {
		Debug << "MMUCacheMiss\n" << eor;
		Console() << this << "::faultType: A protection level "
			<< "MMU cache miss occurred!\n" << eor;
		flushCache( address, 1 );
		return( error );
	} else {
		Debug << "Unknown!\n" << eor;
		Assert( NOTREACHED );
	}
	return( error );
}


#include "ClassInstaller.h"
DEFINECLASS MC68030MMU
