#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989 The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Bjorn A. Helgaas (helgaas@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * MC68030Translation.c: Page tables for a MC68030 MMU.
 */

#include "Assert.h"
#include "Domain.h"
#include "MC68030Translation.h"
#include "OutputStream.h"
#include "PhysicalMemoryChain.h"
#include "PhysicalMemoryManager.h"
#include "PhysicallyAddressableUnit.h"
#include "TwoLevelPageTable.h"


#ifdef __GNUC__
#pragma implementation
#endif


/*
 * Set up table allocation stuff and initialize the necessary information
 * to keep the mappings common to all instances of this class.
 */
void
MC68030Translation::init( unsigned int tableAreaSize,
	VirtualMemoryRange * systemMemoryMap )
{
	Debug << "MC68030Translation::init( tableAreaSize: " << tableAreaSize
		<< " )\n" << eor;

	TwoLevelPageTable::init( tableAreaSize, systemMemoryMap );
}

MC68030Translation::MC68030Translation()
{
	/*
	 * Don't try "<< this" here; it's too early!
	 */
	Assert( this != 0 );
	Debug << "MC68030Translation::MC68030Translation()\n" << eor;

	systemTranslation = new TwoLevelPageTable;
	userTranslation = new TwoLevelPageTable;
}

MC68030Translation::~MC68030Translation()
{
	Assert( this != 0 );
	Debug << this << "::~MC68030Translation()\n";

	delete systemTranslation;
	delete userTranslation;
}

const MC68030PointerTableEntry *
MC68030Translation::systemPointerTable()
{
	return( systemTranslation->pointerTable() );
}

const MC68030PointerTableEntry *
MC68030Translation::userPointerTable()
{
	return( userTranslation->pointerTable() );
}

void
MC68030Translation::addMapping( const char * vaddr,
	PhysicalMemoryChain * chain, ProtectionLevel maxProtection )
{
	Assert( this != 0 );
	Debug << this << "::addMapping( " << asAddress( vaddr ) << ", ..., "
		<< maxProtection << " )";

	systemTranslation->addMapping( vaddr, chain, ReadWrite );
	if( maxProtection == ReadWrite || maxProtection == ReadOnly ) {
		Debug << " (user also)";
		userTranslation->addMapping( vaddr, chain, maxProtection );
	}
	Debug << "\n" << eor;
}

/*
 * Remove a mapping and update the machine independent ref/mod information
 * for a range of virtual addresses.
 */
void
MC68030Translation::basicRemoveMapping( const char * vaddr,
	unsigned int nbytes )
{
	Assert( this != 0 );
	Debug << this << "::basicRemoveMapping( " << asAddress( vaddr ) << ", "
		<< nbytes << " )\n" << eor;

	systemTranslation->basicRemoveMapping( vaddr, nbytes );
	if( userTranslation->mapping( vaddr ) != (const char *) -1 ) {
		userTranslation->basicRemoveMapping( vaddr, nbytes );
	}
}

int
MC68030Translation::basicChangeProtection( const char * vaddr,
	unsigned int nbytes, ProtectionLevel newProt )
{
	Assert( this != 0 );
	Console() << this << "::basicChangeProtection( "
		<< asAddress( vaddr ) << ", " << nbytes << ", " << newProt
		<< " )\n" << eor;

	Assert( systemTranslation->mapping( vaddr ) != (const char *) -1 );

	if( userTranslation->mapping( vaddr ) == (const char *) -1 ) {

		/*
		 * No user mapping exists; add one if appropriate.
		 */
		if( newProt == ReadWrite || newProt == ReadOnly ) {
			PhysicalMemoryChain chain =
				thisProcess()->domain()->constructChain(
					vaddr, nbytes );
			Console() << this << "::basicChangeProtection: "
				<< "adding user mapping\n" << eor;
			userTranslation->addMapping( vaddr, &chain, newProt );
		}
	} else {

		/*
		 * A user mapping exists; remove it or change it as
		 * needed.
		 */
		if( newProt == NoAccess ) {
			Console() << this << "::basicChangeProtection: "
				<< "removing user mapping\n" << eor;
			userTranslation->basicRemoveMapping( vaddr, nbytes );
		} else {
			Console() << this << "::basicChangeProtection: "
				<< "changing user mapping\n" << eor;
			userTranslation->basicChangeProtection( vaddr, nbytes,
				newProt );
		}
	}
}

const char *
MC68030Translation::mapping( const char * vaddr )
{
	return( systemTranslation->mapping( vaddr ) );
}

AddressTranslationFailureCode
MC68030Translation::determineError( const char * vaddr,
	AccessType attemptedAccess )
{
	Assert( this != 0 );
	Debug << this << "::determineError( " << asAddress( vaddr )
		<< ", " << attemptedAccess << " )\n" << eor;

	if( (attemptedAccess == SystemRead) ||
	    (attemptedAccess == SystemWrite) ) {
		Debug << this << "::determineError: asking system translation\n"
			<< eor;
		return( systemTranslation->determineError( vaddr,
				attemptedAccess ) );
	} else {
		Debug << this << "::determineError: asking user translation\n"
			<< eor;
		return( userTranslation->determineError( vaddr,
				attemptedAccess ) );
	}

	return( ProtectionViolation );
}


#include "ClassInstaller.h"
DEFINECLASS MC68030Translation
