#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Contributing Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * NS32332CPU.c: NS32332 CPU class implementation.
 */


#include "Assert.h"
#include "OutputStream.h"
#include "Domain.h"
#include "Vectors.h"
#include "Lock.h"
#include "Timer.h"
#include "ThisCPU.h"
#include "NS32332CPU.h"
#include "NS32332Context.h"
#include "NS32332Interrupts.h"
#include "NS32382MMU.h"


#ifdef __GNUC__
#pragma implementation
#endif

#if (__GNUG__ == 2)
/* G++ 2.0 requires explicit creation of static members, which is probably
   a good thing */
Interrupt *NS32332CPU::interruptTable[ nInterrupts ];
InterruptDescriptor NS32332CPU::interruptVectorTable[ nInterrupts ];
#endif

static int NS32332CPUClass_Initialized = 0;

static const int ModuleTableSize = 1024;
static Module * ModuleTable = (Module *) 0x1000; /* skip first page */

/*
 * Initialize the module table.
 */
static Lock ModuleTableLock;

/*
 * Manage the allocation of module table entries. Currently they are only
 * allocated and never de-allocated, so the table is filled in increasing
 * order. Finding an empty slot implies that no match will be found
 * further up in the table.
 */
Module *
AllocateModuleTableEntry( void * programBase, void * staticBase,
							void * interrupt )
{
	ModuleTableLock.acquire();
	Module * mp;
	for( mp = &ModuleTable[1]; mp < &ModuleTable[ModuleTableSize]; mp++ ) {
		/*
		 * Found an empty slot so use it.
		 */
		if ( mp->ProgramBasePointer == 0 ) {
			mp->ProgramBasePointer = programBase;
			mp->StaticBasePointer = staticBase;
			mp->InterruptPointer = interrupt;
			ModuleTableLock.release();
			Debug.printf( "AllocateModuleTableEntry: return( %A )\n", mp );
			return( mp );
		}

#ifdef NEVER
		/*
		 * See if the slot matches what were installing (to save space).
		 */
		if ( ( mp->ProgramBasePointer == programBase ) &&
		     ( mp->StaticBasePointer == staticBase ) &&
		     ( mp->InterruptPointer == interrupt ) ) {
			ModuleTableLock.release();
			Debug.printf( "AllocateModuleTableEntry: return( %A )\n", mp );
			return( mp );
		}
#endif
	}

	// Couldn't find an entry to use.
	Assert (NOTREACHED) ( "Out of Module Table Entries\n" );
	return( 0 );
}

void
FreeModuleTableEntry( Module *mp )
{
	Assert (&ModuleTable[1]  <= mp);
	Assert ( mp < &ModuleTable[ModuleTableSize]);
	mp->ProgramBasePointer = 0;
	mp->StaticBasePointer = 0;
	mp->InterruptPointer = 0;
}

static StrayInterrupt uncaught( "Uncaught interrupt" );

NS32332CPU::NS32332CPU( unsigned long identifier, ProcessContainer * iic )
	: CPU( identifier, iic )
{
	/*
	 * We know this constructor is called sequentially (not in parallel)
	 * by all the processors, therefore this doesnt need locked.
	 */
	if( ! NS32332CPUClass_Initialized ) {
		Debug.printf( "Initializing CPU class\n" );
		for( int i = 0; i < nInterrupts; i++ ) {
		//	interruptTable[i] = &uncaught; //Commented out JLC
			interruptTable[i] = ( Interrupt * ) 0xfffffe80;
		}

		NS32332CPUClass_Initialized = 1;
	}

	/*
	 * Set the MMU pointer and other.
	 */
	_MMU = &myMMU;
}


void
NS32332CPU::chipInitialize()
{
	Debug << this << "::chipInitialize()\n" << eor;

	Assert( bootSanityCheck() );

	/*
	 * Set the interrupt base register to point to the classes interrupt
	 * vector table.
	 */
	asm volatile( "lprd intbase,%0" : : "g" (&interruptVectorTable[0]) );
}

/*
 * Preempt this CPU.  We want the running process to be put back on the
 * scheduler ready queue so we can get a new one.
 * This function is machine and not processor dependent and needs
 * rethought altogether.
 */
void
NS32332CPU::preempt()
{
	/* create and transmit the directed vector */
	if( this != 0 ) {
		TransmitDirectedVector( PREEMPT_Vector, (unsigned int) id() );
	}
}

/*
 * Return the interrupt object managing a vector. Nice hardware might
 * allow you to stash something like this in the vector table along
 * with the entry point of the interrupt handler. Although, keeping the
 * parallel table isn't really that expensive.
 */
Interrupt *
NS32332CPU::interrupt( int vector )
{
	Assert( vector >= 0 );
	Assert( vector < nInterrupts );

	Interrupt * interrupt = interruptTable[ vector ];

	Debug.printf( "NS32332CPU::interrupt( %d ): return( %N )\n",
		vector, interrupt );
	Assert( interrupt != 0 );
	return( interrupt );
}

/*
 * Install an interrupt object to manage a vectored interrupt (trap or
 * interrupt) from the hardware.
 */
Interrupt *
NS32332CPU::setInterrupt( int vector, Interrupt * interrupt, PFV assist )
{
	Debug.printf( "NS32332CPU::setInterrupt( vec:%d interrupt:%N assist:%x)\n",
		vector, interrupt, assist );
	if( ( vector < 0 ) || ( vector >= nInterrupts ) ) {
		Console() << "Bad vector number = " << vector << "\n" << eor;
	}
	Assert( vector >= 0 );
	Assert( vector < nInterrupts );
	Assert( interrupt != 0 );

	/*
	 * Save the pointer to the interrupt object itself.
	 */
	Interrupt * oldInterrupt = interruptTable[ vector ];
	interruptTable[ vector ] = interrupt;

	/*
	 * Set the corresponding hardware vector table entry to point to
	 * the proper assemble assist.
	 *
	 * Vector numbers less than 15 are traps on the 32k and not interrupts.
	 * Interrupts need to use reti and traps need rett, therefore,
	 * different assembler assists are needed.
	 *
         * Changed; now the setInterrupt specifies the assist, so that
         * the assist semantics can be changed for specific cases. If no
	 * assist is specified, we default to the standard behavior.
	 */

	PFV raise = assist;
	if( raise == 0 )
	  {
	      if( vector <= 15 )
		raise = trapAssist;
	      else
		raise = interruptAssist;
	  }

	Debug.printf( "Assist for vector %d is %A\n", vector, raise );

	/*
	 * Allocate a module table entry for this interrupt.
	 */
	extern void * p_glbl;
	void * mp = AllocateModuleTableEntry( (void *) raise, p_glbl, interrupt );
	Debug.printf( "Using module table entry %A\n", mp );

	/*
	 * Set the interrupt vector to point to the allocated module
	 * table entry.
	 */
	interruptVectorTable[vector].module = (unsigned short) mp;
	interruptVectorTable[vector].offset = 0;

	return( oldInterrupt );
}

/*
 * Halt the CPU. This code probably doesn't belong here but I can't think of
 * any better place to put it.
 */
void
Halt()
{
	/*
	 *  Disable interrupts.
	 */
	asm volatile( "bicpsrw $0x800" );

	/*
	 * Since no more interrupts, make sure no more vectors come to us over
	 * the Vector Bus by putting ourselves in the class which (by
	 * convention) is never sent to.
	 * Well, we could still get an NMI, but there's nothing we can do
	 * about that.
	 */
	SetCPUVectorClass( CPUCLASS_HALTED );

	/*
	 * We really need to broadcast a HALT interrupt to everybody else.
	 * Do this sooner or later.
	 * Also the CPU dtor should be called to remove it from the
	 * active cpu's collection.
	 */

	/*
	 * Do anything that needs to be done only when the first
	 * CPU Halts.
	 */
	static firstHalt = 1;
	if ( firstHalt ) {

		/*
		 * Make sure we get here only once.
		 * Yes, there is a race here.
		 */
		firstHalt = 0;

		/*
		 * Dump Locks.
		 */
		//Lock::dump();
	}

	/*
	 * Halt CPU via infinite loop.
	 */
	Console().criticalPrintf( "Halting.\n" );
	for ( ;; );
	Assert( NOTREACHED );
}

/*
 * NS32332CPU Class destructor.
 * Obvious side effect is to halt the processor. I doubt this could ever
 * actually be called!
 */
NS32332CPU::~NS32332CPU()
{
	Debug.printf( "CPU::~CPU()\n" );
	Halt();
}

/*
 * Stick in the interrupt object to handle the traps. Subclasses are
 * left to overload this method and install any machine dependent
 * interrupts.
 */
AbortTrap VMInterrupt;
SVCTrap SVCInterrupt;
StrayInterrupt DVZInterrupt( "Divide By Zero Trap" );
StrayInterrupt FLGInterrupt( "Flag Trap" );
StrayInterrupt BPTInterrupt( "Breakpoint Trap" );
StrayInterrupt TRCInterrupt( "Trace Trap" );
StrayInterrupt FPUInterrupt( "Floating Point Trap" );
StrayInterrupt BERInterrupt( "Bus Error Trap" );
StrayInterrupt UNDInterrupt( "Undefined Instruction Trap" );
StrayInterrupt ILLInterrupt( "Illegal Instruction Trap" );

void
NS32332CPU::installInterrupts()
{
	Debug.printf( "NS32332CPU::installInterrupts()\n" );
	Assert( this != 0 );
	Assert( this == thisCPU() );

	setInterrupt( ABT_Vector, &VMInterrupt, trapAssist );
	setInterrupt( SVC_Vector, &SVCInterrupt, svcAssist );
	setInterrupt( ILL_Vector, &ILLInterrupt, proxyAssist );
	setInterrupt( UND_Vector, &UNDInterrupt, trapAssist );
	setInterrupt( DVZ_Vector, &DVZInterrupt, trapAssist );
	setInterrupt( FLG_Vector, &FLGInterrupt, trapAssist );
	setInterrupt( BPT_Vector, &BPTInterrupt, trapAssist );
	setInterrupt( TRC_Vector, &TRCInterrupt, trapAssist );
	setInterrupt( FPU_Vector, &FPUInterrupt, trapAssist );
	setInterrupt( BER_Vector, &BERInterrupt, trapAssist );
}

int
NS32332CPU::bootSanityCheck()
{
	Assert( this != 0 );

	extern void * p_glbl;

	int error = 0;
	void * sb = 0;
	asm volatile( "sprd sb,%0" : "=g" (sb) );
	if ( sb != p_glbl ) {
		error++;
		Console().cpuPrintf( "BootSanityCheck: SB wrong: %x != %x\n",
			sb,p_glbl );
	}
	Module * mod = 0;
	asm volatile( "sprd mod,%0" : "=g" (mod) );
	if ( (unsigned int) mod >= 0x10000 ) {
		error++;
		Console().cpuPrintf( "BootSanityCheck: MOD out of range (?): %x\n", mod );
	}
	if ( (unsigned int) mod & 0xf ) {
		error++;
		Console().cpuPrintf( "BootSanityCheck: MOD out of alignment: %x\n", mod );
	}
	void * modsb = (void *) mod->StaticBasePointer;
	if ( modsb != sb ) {
		error++;
		Console().cpuPrintf( "BootSanityCheck: MOD wrong: %x != %x\n",
			modsb,sb );
	}
	if ( error ) {
		Console().cpuPrintf( "BootSanityCheck: ERROR: p_glbl/sb/mod/modsb: %x/%x/%x/%x\n",
			   p_glbl, sb, mod, modsb );
		unsigned int * up = (unsigned int *) mod;
		for ( int i = 0; i < 16; i++ ) {
			Console().cpuPrintf( "%A: ", up );
			for ( int j = 0; j < 8; j++, up++ )
				Console().printf( "%x ", *up );
			Console().printf( "\n" );
		}
	}
	return (error == 0);
}

void
InitModuleTable()
{
	Assert( ModuleTable == (Module *) 0x1000 );
	Assert( sizeof( Module ) == 16 );	// Hardware necessity.

	/*
 	 * Zero out the whole module table, except for the 0th entry
	 * which is used by the kernel and set in the startup code.
	 */
	Debug.printf( "Filling in %d entries of module table at %x\n",
		ModuleTableSize, ModuleTable );
	for( int i = 1; i < ModuleTableSize; i++ ) {
		ModuleTable[i].StaticBasePointer = 0;
		ModuleTable[i].LinkBasePointer = 0;
		ModuleTable[i].ProgramBasePointer = 0;
		ModuleTable[i].InterruptPointer = 0;
	}
}

int
NS32332CPU::getContextSaveHint( Interrupt * interrupt, char * )
{
/*
 * For now, everybody gets a small context.

	if( interrupt->needsAFullContext() ) {
		return( 1 );
	}
 */

	return( 0 );
}


int
NS32332CPU::disableInterrupts()
{
	unsigned short psr = 0;
	asm volatile( "sprw psr,%0" : "=g" (psr) );	// save old state
	asm volatile( "bicpsrw $0x800" );		// disable
	return( (psr & 0x800) == 0x800 );
}


void
NS32332CPU::restoreInterrupts(int val)
{
	Assert (val == 0 || val == 1);
	if (val) {
		asm volatile("bispsrw $0x800");
		}
}


int
NS32332CPU::interruptsEnabled()
{
	unsigned short psr = 0;
	asm volatile( "sprw psr,%0" : "=g" (psr) );
	return( (psr & 0x800) == 0x800 );
}

 
void
NS32332CPU::spinMutex(MutexStateAddr addr)
{
	Assert (addr != 0);
	asm(   "sbitib  $0,0(%0); \
                bfc     .+10; \
                tbitb   $0,0(%0); \
                bfc     .-11; \
                br      .-6" : : "r" (addr) );
}


unsigned
NS32332CPU::cpuID()
{
	unsigned result = *(volatile unsigned char *) 0xffffff54;
	return result;
}

void 
NS32332CPU::initMutexState(MutexStateAddr addr)
{
	Assert (addr != 0);
	*addr = 0;
}

void 
NS32332CPU::releaseMutex(MutexStateAddr addr)
{
	Assert (addr != 0);
	*addr = 0;
}
