#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * NS32332Interrupts.cc: NS32332 hardware interrupt classes implementations.
 */
#include "Assert.h"
#include "OutputStream.h"
#include "Process.h"
#include "ThisProcess.h"
#include "ProcessContainer.h"
#include "CPU.h"
#include "ThisCPU.h"
#include "NS32332Interrupts.h"
#include "PROCESSOR_CONTEXT.h"
#include "Vectors.h"
#include "FaultErrorCode.h"
#include "Domain.h"
#include "NS32382MMU.h"
#include "AddressTranslation.h"


#ifdef __GNUC__
#pragma implementation
#endif

#ifdef INSTRUMENT_PROCESSORINTERRUPT
#define Instr(exp) exp
#include "Instrument.h"
extern SystemInstrumentManager *TheInstrumentManager;

TimerInstrument *iInterruptAbortTrap = NULL;
TimerInstrument *iInterruptAbortTrapFaultType = NULL;
TimerInstrument *iInterruptAbortTrapRepairFault = NULL;

void 
ProcessorInterruptConstructInstruments ()
{
	iInterruptAbortTrap = new TimerInstrument
		(TheInstrumentManager, "InterruptAbortTrap", TALKING, 1, ALL);
	iInterruptAbortTrapFaultType = new TimerInstrument
		(TheInstrumentManager, "InterruptAbortTrapFaultType", TALKING, 1, ALL);
	iInterruptAbortTrapRepairFault = new TimerInstrument
		(TheInstrumentManager, "InterruptAbortTrapRepairFault", TALKING, 1, ALL);
}

void 
ProcessorInterruptDeleteInstruments ()
{
	delete iInterruptAbortTrap;
	iInterruptAbortTrap = NULL;
	delete iInterruptAbortTrapFaultType;
	iInterruptAbortTrapFaultType = NULL;
	delete iInterruptAbortTrapRepairFault;
	iInterruptAbortTrapRepairFault = NULL;
}

#else
#define Instr(exp)
void 
ProcessorInterruptConstructInstruments () {}
void 
ProcessorInterruptDeleteInstruments () {}
#endif


#ifdef HACK_DEBUG
int hackStrayHit;
int hackStrayPC;
int hackStrayPSR;
#endif
void
StrayInterrupt::basicRaise( char * info )
{
	stackFrame * frame = (stackFrame *) info;

#ifdef HACK_DEBUG
	hackStrayHit=1; hackStrayPC=frame->pc; hackStrayPSR=frame->psr;
	Console().write("Argh!\n",6);
#endif
	Console().panic( "Stray Interrupt (%s): process was %I, pc=%A, psr=%x\n",
		message, thisProcess(), frame->pc, frame->psr );
	Assert( NOTREACHED );
}

void
AbortTrap::basicRaise( char * isp )
{
	const char * address = 0;

	asm volatile( "smr tear,%0" : "=g" (address) );

	Instr (iInterruptAbortTrap->start ());

	/*
	 * Find the needed info back on the interrupt stack.
	 */
	stackFrame * frame = (stackFrame *) isp;
	unsigned int mode = frame->psr;

	Debug.printf( "PageFault: process:%N addr:%A psr:%A pc:%A\n",
		thisProcess(), address, mode, frame->pc );
	Process * process = thisProcess();
	Assert( process != 0 );
	Domain * domain = process->domain();
	Assert( domain != 0 );
	//PROCESSOR_CONTEXT * context = (PROCESSOR_CONTEXT *) process->context();

	Instr (iInterruptAbortTrapFaultType->start ());

	AccessType attemptedAccess;
	AddressTranslationFailureCode code =
	    ( (NS32382MMU *) thisCPU()->MMU() )->faultType(
	     address, mode, domain->translation(), attemptedAccess );

	Instr (iInterruptAbortTrapFaultType->stop ());

	switch( code ) {
		case MMUCacheMiss:
			Instr (iInterruptAbortTrap->stop ("MMUCacheMiss"));
			return;
		case SomethingElse:
			Console().panic( "Can't deal with breakpoints (yet) addr=%A\n",
				address );
			Assert( NOTREACHED );
		default:
			break;
	}

	/*
	 * If it wasnt a simple onboard MMU cache miss, only the Domain
	 * knows what to do next.
	 */
	Instr (iInterruptAbortTrapRepairFault->start ());
	FaultErrorCode reason = domain->repairFault( address, attemptedAccess );
	Instr (iInterruptAbortTrapRepairFault->stop ());

	switch( reason ) {
		case AccessOfUnmanagedMemory:
			Console() << thisProcess() << "No manager for address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->pc )
			   << " at time of access)\n" << eor;
			Assert( NOTREACHED );

		case ReadOfProtectedSystemMemory:
			Console() << thisProcess()
			   << "Attempted read of system memory at address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->pc )
			   << " at time of access)\n" << eor;
			Assert( NOTREACHED );

		case WriteOfProtectedSystemMemory:
			Console() << thisProcess()
			   << "Attempted write of system memory at address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->pc )
			   << " at time of access)\n" << eor;
			Assert( NOTREACHED );

		case WriteOfReadOnlyMemory:
			Console() << thisProcess()
			   << "Attempted write of read-only memory at address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->pc )
			   << " at time of access)\n" << eor;
			Assert( NOTREACHED );

		case DynamicWriteOfReadOnlyMemory:
		case DynamicAccessViolation:
			Console() << thisProcess()
			   << "Access violation referencing memory at address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->pc )
			   << " at time of access)\n" << eor;
			Assert( NOTREACHED );

		/*
		 * These are all recoverable and the instruction can
		 * be restarted/resumed.
		 */
		case MissingAddressTranslationMapping:
			Instr (iInterruptAbortTrap->stop ("MissingAddressTranslationMapping"));
			break;
		case MissingPhysicalMemory:
			Instr (iInterruptAbortTrap->stop ("MissingPhysicalMemory"));
			break;
		case InsufficientProtectionLevel:
			Instr (iInterruptAbortTrap->stop ("InsufficientProtectionLevel"));
			break;
	}

	/*
	 * All done so return. This results in retrying/restarting the
	 * instruction which caused the fault.
	 */
	Debug << this << "::basicRaise: fault repaired - retrying\n" << eor;
}
