#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * NS32382MMU.c: An NS32382 as an AddressTranslator.
 */

#include "NS32382PageTableEntry.h" // Create Page/Pointer TabEnt
#include "NS32382MMU.h"
#include "ADDRESS_TRANSLATION.h"
#include "Assert.h"
#include "NS32382MMU.h"
#include "OutputStream.h"
#include "PAGE_TABLE_ENTRY.h"
#include "PageTableConstants.h"
#include "PROCESSOR_CONTEXT.h"


#ifdef __GNUC__
#pragma implementation
#endif


NS32382MMU::NS32382MMU()
{
	Debug << this << "::NS32382MMU()\n" << eor;
}

void
NS32382MMU::enable()
{
	asm volatile ( "lmr mcr,$3" );
}

int
NS32382MMU::enabled()
{
	int FEW;
	asm volatile ( "smr mcr,%0" : "=r" (FEW) );
	return((FEW & 0x3) == 0x3);
}

void
NS32382MMU::basicActivate( AddressTranslation * at)
{
	load( at);
}

void
NS32382MMU::flushCache( const char * address, unsigned int length )
{
	/*
	 * I don't know if the ivar1 one is necessary, since we don't
	 * use the "dual space" mode.
	 */
//	asm volatile( "lmr ivar0,%0; lmr ivar1,%0" : : "g" (address) );

	/*
	 * For now, blow the whole cache...
	 */
	int temp;
	asm volatile( "smr ptb0,%0" : "=r" (temp) );
	asm volatile( "lmr ptb0,%0" : : "r" (temp) );
}

/*
 * TXError() values.
 */
static const int TX_NONE = 0;
static const int TX_FirstLevelPTE = 1;
static const int TX_SecondLevelPTE = 2;
static const int TX_ProtectionInterrupt = 3;

/*
 * TXStatus() and ErrStatus() values.
 * NOTE: We only define the ones we currently need.
 */
static const int SeqInstrFetch = 0x8;	 // Sequential instruction fetch.
static const int NonSeqInstrFetch = 0x9; // Non-sequential instruction fetch.
static const int DataTransfer = 0xa;	 // Data transfer.
static const int ReadRMWOperand = 0xb;	 // Read of read-modify-write operand.
static const int ReadEffectiveAddr = 0xc;    	// Read for effective addr. 
						// calculation.

class ASR {
	union {
		unsigned int all;
		struct {
			unsigned int TX	  : 2;	// Translation interrupt.
			unsigned int DDIT : 1;	// Data direction (xlate error)
			unsigned int UNST : 1;	// User/Supervisor.
			unsigned int STT  : 4;	// Status bits.
			unsigned int MCT  : 1;	// Multiple cycle.
			unsigned int BP   : 1;	// Breakpoint indicator.
			unsigned int CE	  : 1;	// CPU error.
			unsigned int ME	  : 1;	// MMU error.
			unsigned int DDIE : 1;	// Data direction (Bus error)
			unsigned int UNSE : 1;	// User/Supervisor
			unsigned int STE  : 4;	// Status bits.
			unsigned int MCE  : 1;	// Mutiple cycle.
			unsigned int	  : 13;
		} fields;
	} data;
public:
	ASR( unsigned int asr );
	operator int();

	int TXError();
	int TXWrite();
	int TXUser();
	int TXStatus();
	int Breakpoint();
	int CPUBusError();
	int MMUBusError();
};

inline
ASR::ASR( unsigned int asr )
{
	data.all = asr;
}

inline
ASR::operator int()
{
	return( (int) data.all );
}

inline int
ASR::TXError()
{
	return( data.fields.TX );
}

inline int
ASR::TXWrite()
{
	return( data.fields.DDIT != 0 );
}

inline int
ASR::TXUser()
{
	return( data.fields.UNST != 0 );
}

inline int
ASR::TXStatus()
{
	return( data.fields.STT );
}

inline int
ASR::Breakpoint()
{
	return( data.fields.BP != 0 );
}

inline int
ASR::CPUBusError()
{
	return( data.fields.CE != 0 );
}

inline int
ASR::MMUBusError()
{
	return( data.fields.ME != 0 );
}

AddressTranslationFailureCode
NS32382MMU::faultType( const char * address, unsigned int /* mode */, 
    AddressTranslation * at, AccessType & attemptedAccess )
{
	/*
	 * See note at beginning of NS32382MMU::activate(), above.
	 */
	ADDRESS_TRANSLATION * translation = (ADDRESS_TRANSLATION *) at;

	/*
	 * Read ASR.
	 */
	unsigned int uasr = 0;
	asm volatile( "smr msr,%0" : "=g" (uasr) );
	ASR asr(uasr);

	/*
	 * Check for conditions we can't deal with:
	 *	Breakpoint.
	 *	Hard bus error (CPU or MMU).
	 */
	if ( asr.Breakpoint() ) {
		Console().criticalPrintf( "MMU: Breakpoint.\n" );
		return( SomethingElse );
	}
	if ( asr.CPUBusError() ) {
		Console().criticalPrintf( "MMU: Hard CPU bus error.\n" );
		return( SomethingElse );
	}
	if ( asr.MMUBusError() ) {
		Console().criticalPrintf( "MMU: Hard MMU bus error.\n" );
		return( SomethingElse );
	}

	/*
	 * It's some sort of translation interrupt.
	 */
	int interrupt = asr.TXError();
	Assert( interrupt != TX_NONE );

	/*
	 * Determine attempted access.
	 * This is passed by reference back to the caller.
	 * NOTE: If the bus status is ReadRMWOperand, it counts as a write,
	 *	 because the page must be writable for the instruction to
	 *	 succeed.
	 * NOTE: We also check for unexpected bus status here.  If this
	 *	 fails, it may very well be that things are OK and I just
	 *	 misunderstood the hardware documentation.
	 */
	int user = asr.TXUser();
	int write = asr.TXWrite();
	int status = asr.TXStatus();
	switch ( status ) {
		case SeqInstrFetch:
		case NonSeqInstrFetch:
		case ReadEffectiveAddr:
			/*
			 * Should be a read.
			 */
			Assert( ! write );
			break;
		case DataTransfer:
			/*
			 * Normal read or write.
			 */
			break;
		case ReadRMWOperand:
			/*
			 * Read part of read/modify/write operand.
			 * This must be treated as a write.
			 */
			write = 1;
			break;
		default:
			/*
			 * Unexpected.
			 */
			Assert( NOTREACHED );
			break;
	}
	if ( user ) {
		attemptedAccess = write ? ApplicationWrite : ApplicationRead;
	}
	else {
		attemptedAccess = write ? SystemWrite      : SystemRead;
	}

	/*
	 * If it's TX_FirstLevelPTE or TX_SecondLevelPTE,
	 * then it's NonResidentMemory.
	 */
	if ( ( interrupt == TX_FirstLevelPTE ) ||
	     ( interrupt == TX_SecondLevelPTE ) ) {
		return( NonResidentMemory );
	}

	/*
	 * If we're not sure what happened, ask the AddressTranslation what
	 * it thinks.  Why can't we figure it out ourselves?
	 */
	AddressTranslationFailureCode error = translation->determineError( address, attemptedAccess);

	if( error == MMUCacheMiss ) {
		Console().cpuPrintf( "A protection level MMU cache miss happened!!\n" );
		asm volatile( "lmr ivar0,%0; lmr ivar1,%0"
			: : "g" (address) );
	}

	if( error == ProtectionViolation ) {
		Assert( interrupt == TX_ProtectionInterrupt );
	}
	return( error );
}


#include "ClassInstaller.h"
DEFINECLASS("NS32382MMU",NS32382MMU,AddressTranslator,0,0);
