#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 *    The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * SPARCCPU.c: SPARC CPU class implementation.
 */

#include "Assert.h"
#include "Gdb.h"
#include "OutputStream.h"
#include "PROCESSOR_CONTEXT.h"
#include "PROM.h"
#include "SPARCCPU.h"
#include "SPARCInterrupts.h"
#include "SPARCasmutil.h"
#include "SS1FreeRunningTimer.h"
#include "SS1FreeRunningTimer.h"
#include "ThisProcess.h"
#include "Process.h"
#include "TimerManager.h"
#include "Vectors.h"
#include "ThisCPU.h"


#ifdef __GNUC__
#pragma implementation
#endif

/*
 * static defns
 */
Interrupt * SPARCCPU::interruptTable[ nInterrupts ];

class ProcessContainer;
class Interrupt;

#ifdef INSTRUMENT
#include "SystemInstrumentManager.h"
#endif

/*
 * Halt the CPU. This code probably doesn't belong here but I can't think of
 * any better place to put it.
 */
void
Halt()
{
	/*
	 * Halt CPU via infinite loop.
	 */
	Console().criticalPrintf( "Halting.\n" );
	Console().flush(); 
	SS1PROM::exit_to_mon();
	// (*romp->v_exit_to_mon)();
}


void
SPARCCPU::debugBreakpoint()
{
	TrapToDebugger();
}

static int SPARCCPUClass_Initialized = 0;

SPARCCPU::SPARCCPU( unsigned long identifier, ProcessContainer * iic )
	: CPU( identifier, iic )
{
	Debug << "SPARCCPU::SPARCCPU()\n" << eor;
	extern unsigned nwindows;
	Console() << "SPARCCPU::SPARCCPU(): " << nwindows 
		<< " register windows\n" << eor;

	/*
	 * We know this constructor is called sequentially (not in parallel)
	 * by all the processors, therefore this doesnt need locked.
	 */
	if( ! SPARCCPUClass_Initialized ) {
		Debug <<  "Initializing CPU class\n" << eor;
		for( int i = 0; i < nInterrupts; i++ ) {
			interruptTable[i] = (Interrupt *)0;
		}

		SPARCCPUClass_Initialized = 1;
	}

	/*
	 * Set the MMU pointer and other.
	 */
	_MMU = &myMMU;
	_timesliceTimer = new TIMESLICETIMER();
}


#ifdef GDB
extern "C"
void
SPARCTrapSetup()
{
	/*
	 * This is what the trap setup code puts as the return address 
	 * in hardwareInterruptAssist so that the debugger can "see" 
	 *  beyond the trap frame.
	 */
}
#endif

/*
 * This is called directly from the assembler interruptAssist.
 * It is declare as 'static' in the class definition so it does not get
 *       a 'this' pointer.
 */
void
SPARCCPU::hardwareInterruptAssist(int vector, char *info)
{
	Assert( (vector >= 0) && (vector < nInterrupts) );
#ifdef DEBUG
	static unsigned traps;
	const unsigned skipDEBUG = 2;
	if ((traps++ % skipDEBUG) == 0)
        Debug << "HWInterruptAssist(" << vector << ", " <<
                asAddress(info) << "\n" << eor;
#endif
#ifdef GDB
	unsigned regi7;					// Fake for debuggers
	asm volatile("mov %%i7,%0" : "=r"(regi7));
	asm volatile("mov %0,%%i7" : : "r"(&SPARCTrapSetup) );
#endif


	switch (vector) {
		case VEC_FLUSH:
			flushWindows();
			break;
#ifdef GDB
		case VEC_BPT1:
		case VEC_BPT2:
			IntStorage state ; PROCESSOR::disableInterrupts(state);
			SPARCGdbHandler(vector, info);
			PROCESSOR::restoreInterrupts(state);
			break;
#endif
		default:
#ifdef INFCHECK
			/* 
			 * Check for infinite traps
			 */
			const int infLimit = 200;
			static int lastvector;
			static char *lastinfo;
			static unsigned lastPC;
			static unsigned infCount;
			if (vector == lastvector && info == lastinfo 
				&& ((unsigned*)info)[1] == lastPC) {
					if (infCount++ > infLimit){
						Console().printf(
							"Infinte trap #%d at 0x%x ???\n",
							vector, ((unsigned*)info)[1]);
						Console().printf("TrapToDebugger()\n");
						Console().flush();
						TrapToDebugger();
						lastvector = vector;
						lastinfo = info;
						lastPC = ((unsigned *)info)[1];
						infCount = 0;
						}
					}
			else {
				lastvector = vector;
				lastinfo = info;
				lastPC = ((unsigned *)info)[1];
				infCount = 0;
				}
#endif

			Interrupt * interrupt = 
				((SPARCCPU *) thisCPU())->interruptTable[ vector ];

			if (interrupt) {
				// Debug << "hardwareInterruptAssist( " << vector << ", " << asAddress(info) << "), interrupt( " << asAddress( interrupt ) << " )\n" << eor;
				interrupt->raise( info );
				}
			else
				{
				Console() << "SPURIOUS TRAP # " << vector << "\n" << eor;
				Assert (NOTREACHED);
			}

#ifdef DEBUG
			if ((traps % skipDEBUG) == 0)
			Debug << "HWInterruptAssist( " << vector << ", " << asAddress(info)
				<< " ) returned, level:" << asHex(interruptLevel()) 
				<< "\n" << eor;
#endif

			break;
	};
#ifdef GDB
	asm volatile("mov %0,%%i7" : : "r"(regi7));
#endif
}


/*
 * Install an interrupt object to manage a vectored interrupt (trap or
 * interrupt) from the hardware.
 */
Interrupt *
SPARCCPU::setInterrupt( int vector, Interrupt * interrupt, PFV assist  )
{
	//Console().debug( "SPARCCPU::setInterrupt( vec:%d interrupt:%N )\n", vector, interrupt );
	Assert( vector >= 0 );
	Assert( vector < nInterrupts );
	Assert( interrupt != 0 );

	/*
	 * Save the pointer to the interrupt object itself.
	 */
	Interrupt * oldInterrupt = interruptTable[ vector ];
	interruptTable[ vector ] = interrupt;

	return( oldInterrupt );
}

/*
 * SPARCCPU Class destructor.
 * Obvious side effect is to halt the processor. I doubt this could ever
 * actually be called!
 */
SPARCCPU::~SPARCCPU()
{
	Debug << "CPU::~CPU()\n" << eor;
	Halt();
}

TimerManagerInterrupt 			TimerManagerInterruptInterrupt;
SS1FreeRunningTimerInterrupt 		theFreeRunningTimerInterrupt;
SPARCMemoryAlignmentInterrupt		MemoryAlignmentInterrupt;
SPARCDataAccessInterrupt		DataAccessInterrupt;
SPARCInstructionAccessInterrupt		InstructionAccessInterrupt;
SPARCAsynchronousMemoryInterrupt	AsynchronousMemoryInterrupt;
SPARCFPInterrupt			fpInterrupt;
SPARCIllegalInstructionInterrupt	illegalInstructionInterrupt;
SPARCFPDisabledInterrupt		fpDisabledInterrupt;
SPARCDivideByZeroInterrupt		divideByZeroInterrupt;
SPARCProxyInterrupt			proxyInterrupt;

#ifdef INSTRUMENT
InstrumentBufferInterrupt TheInstrumentBufferInterrupt;
#endif

void
SPARCCPU::installInterrupts()
{
	Debug << "SPARCCPU::installInterrupts()\n" << eor;

	/*
	 * The Proxy callgate is a special case; it doesn't go
	 *  through the default interrupt handler at all.
	 */

	setInterrupt(VEC_INSN_ACCESS, &InstructionAccessInterrupt);
	setInterrupt(VEC_ILL_INSN, &illegalInstructionInterrupt);
	setInterrupt(VEC_PRIVL_INSN, &illegalInstructionInterrupt);
	setInterrupt(VEC_FP_DISABLED, &fpDisabledInterrupt);
	setInterrupt(VEC_MEM_ALIGN, &MemoryAlignmentInterrupt);
	setInterrupt(VEC_FP_EXCEP, &fpInterrupt);
	setInterrupt(VEC_DATA_ACCESS, &DataAccessInterrupt);
	setInterrupt(VEC_TAG_OVFL, &illegalInstructionInterrupt);
	setInterrupt(VEC_TIMER /* level 10 */, &TimerManagerInterruptInterrupt);
	setInterrupt(VEC_FRTIMER /* level 14 */, &theFreeRunningTimerInterrupt);
	setInterrupt(VEC_ASYNCH_MEM, &AsynchronousMemoryInterrupt);
	setInterrupt(VEC_PROXY, &proxyInterrupt);
	/*
	 * THis is necessary since we use the math emulation lib from
	 * libc.a and it may generate t 2
	 */
	setInterrupt(VEC_DIV_ZERO, &divideByZeroInterrupt);
#ifdef INSTRUMENT
#ifdef notdef
// figure out what the right vector should be, if any!
        setInterrupt (???, &TheInstrumentBufferInterrupt);
#endif
#endif
}
