#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * SPARCContext.c: Processor context for the SPARC.
 */

#include "Byte.h"
#include "Domain.h"
#include "Kernel.h"
#include "MemoryObject.h"
#include "MemoryObjectView.h"
#include "OutputStream.h"
#include "RootMemoryObjectCache.h"
#include "SPARCasmutil.h"
#include "SPARCContext.h"
#include "SPARCInterrupts.h"
#include "SPARCPageTableConstants.h"
#include "SourceOfZerosMemoryObject.h"
#include "ThisProcess.h"
#include "Process.h"
#include <sys/psw.h>
#include <sys/reg.h>


#ifdef __GNUC__
#pragma implementation
#endif

class Process;

const unsigned TinyStackPages = 2;
const unsigned SmallStackPages = 4;
const unsigned NormalStackPages = 8;
const unsigned LargeStackPages = 16;
const unsigned HugeStackPages = 32;
const unsigned GiganticStackPages = 64;
const unsigned UserStackPages = 16;


inline 
const unsigned
stackPages( StackSize size)
{
	unsigned npages;
	switch( size ) {
	   case none:
		npages =  UserStackPages;
		Assert (NOTREACHED);
		break;
	   case heapBootStack:
	   case tiny:
	   case small:
		npages =  TinyStackPages;
		break;
#ifdef NEVER
	   case small:
		npages =  SmallStackPages;
		break;
#endif
	   case large:
		npages =  LargeStackPages;
		break;
	   case huge:
		npages =  HugeStackPages;
		break;
	   case gigantic:
		npages =  GiganticStackPages;
		break;
	   case normal:
		npages =  NormalStackPages;
	   case application:
	   default:
		npages =  UserStackPages;
		break;
	}
	return npages;
}
/*
 * Allocate a (non-paged and pre-allocated) stack of size into domain.
 * return the value to use as the stackpointer (stack grows down on sparc)
 */

void
SPARCGenericContext::allocateStack(StackSize size, Domain *domain,
char * &stackpointer, unsigned & stackbytes, MemoryObjectStar &mo, 
char * & base)
{
	Assert (domain != 0);
	int npages = stackPages(size);
	Debug << " )" << npages << " pages into domain: " << domain 
		<< "\n" << eor;
	Assert (TheKernel->domain() != 0);
	base = 0;
	stackbytes = npages*PhysicalPageSize;

	if (size == heapBootStack) {
		mo = (MemoryObject *)0;
		base = new char[npages*PhysicalPageSize];
		Assert (base != 0);
		Debug << "SPARCGenericContext::allocateStack(): " <<
			"heapStack base:" << asAddress(base) << "\n" << eor;
	}
	else	{
		ProtectionLevel prot = (domain == TheKernel->domain()) ? NoAccess : ReadWrite;
		mo = new SourceOfZerosMemoryObject;
		mo->setName("StackMo");
		mo->setNumberOfUnits(npages*PhysicalPageSize);
		mo->setCacheClass( RootMemoryObjectCacheClass );
                MemoryObjectCache * cache = mo->buildCache();
                Assert (cache != 0);
		base = (char *)domain->add(mo, prot);
		Assert (base != 0);
		domain->makeAddressable(base, npages*PhysicalPageSize);
		Assert( base != 0 );
		Debug << "SPARCGenericContext::allocateStack(): " <<
			"MO stack base:" << asAddress(base) << "\n" << eor;
	}
	stackpointer = (char*) (base + stackbytes - 1 );
	if ((unsigned)stackpointer & 7)
		stackpointer =  (char*) ((unsigned)(stackpointer - 7) & ~7);
	stackpointer -= 168;
}

unsigned debugApplications = 0;

SPARCGenericContext::SPARCGenericContext( Process * owner, PFV entry, 
	StackSize stackSize, int arg0, int arg1)
	: ProcessorContext( owner ) , 
	     _kernelStackMO(0)
{
	Assert (entry != 0);
	Assert (TheKernel->domain() != 0);
	Assert (stackSize != none);

	allocateStack(stackSize, TheKernel->domain(),
		      _kernelStackTop, _kernelStackBytes, _kernelStackMO,
		      _kernelStackBase);
	_stackPointer = _kernelStackTop;

	/*
	 * When the context is created the inputs save area is
	 * used to communicate with the initialRestore.
	 *
	 * inputs[0]	-	owner
	 * inputs[1]	-	entry
	 * inputs[2]	-	arg0
	 * inputs[3]	-	arg1
	 * inputs[4]	-	default proxy for application (or 0 otherwise)
	 * inputs[5]	-	debugApplications Mask
	 * inputs[7]	-	default return
	 */
	unsigned * const inputs = (unsigned *) &volatileRegs[0];
	inputs[0] = INTERRUPT_LOW; // enable unless
					// overwritten in sublclass ctor
	inputs[1] = (unsigned)entry;
	inputs[2] = (unsigned)arg0;
	inputs[3] = (unsigned)arg1;
	/* 
	 * inputs[4] used for application default proxy and is reset
	 * in application class, likewise for inputs[5] and inputs[6].
	 */
	inputs[4] = 0;
	inputs[5] = debugApplications;
	inputs[7] = (unsigned) &defaultReturn;

	Debug << this << "::SPARCGenericContext(): " << "\n"
		<< "\t_stackPointer: " << asAddress (_stackPointer) << "\n"
		<< "\t_kernelStack:Top " << asAddress(_kernelStackTop) << "\n"
		<< "\tentry: " << asHex((unsigned)entry) << "\n"
		<< "\targ0: " << asHex(((unsigned *)inputs)[2]) << "\n"
		<< "\targ1: " << asHex(((unsigned *)inputs)[3]) << "\n"
		<< eor;

	/*
	 * Set up initial floating point context (only used if fp unit
	 * is used).
	 */
	fsr = 0;

	/*
	 * Sanity checks.  
	 */
	Assert ( _kernelStackTop != 0 );
	Assert ( ! ((unsigned)_kernelStackTop & 0x7) );
	Assert ( _stackPointer != 0 );
	Assert ( ! ((unsigned)_stackPointer & 0x7) );
	Assert ( inputs[0] < 17 );
}

SPARCGenericContext::~SPARCGenericContext()
{
	if (_kernelStackMO){
		TheKernel->domain()->remove(_kernelStackMO,_kernelStackBase);
		_kernelStackMO = (MemoryObject*)0;
		}
}


SPARCApplicationContext::SPARCApplicationContext( Process * owner,
	PFV entry, StackSize stackSize, Domain *initialDomain,
		int arg0, int arg1,
		StackSize userStackSize, char *userStackPointer)
	: SPARCGenericContext( owner, entry, stackSize, arg0, arg1),
	userStackMO(0)
{
	unsigned * const inputs = (unsigned *) &volatileRegs[0];
	Debug << "SPARCApplicationContext::ctor()\n" << eor;

	Assert (initialDomain != 0);
	Assert (entry != 0);

	if ( userStackSize != none ) {
		allocateStack(userStackSize, initialDomain,
			      _userStackTop, _stackBytes, userStackMO, 
			      userStackBase);
		_stackPointer = _userStackTop;
	} else {
		Assert( initialDomain != 0 );
		
		const char * base;
		unsigned length;
		ProtectionLevel prot;
		MemoryObjectViewStar mov = initialDomain->lookup
			( ( const char * ) userStackPointer,
                          base, length, prot ); 
		Assert( mov != 0 );
		userStackBase = ( char * ) base;
		_stackBytes = length;
		_userStackTop = (char*) (base + length - 1 );
		if ( (unsigned) _userStackTop & 7)
			_userStackTop = 
				(char*) ((unsigned)(_userStackTop - 7) & ~7);
		_userStackTop -= 168;
		userStackMO = mov->source();
		initialDomain->makeAddressable( base, length ); 
		
		_stackPointer = userStackPointer;
	}

	/*
	 * Set up additional inputs that initialRestoreAssist uses.
	 */
	inputs[4] = (unsigned) initialDomain->defaultProxy();
	inputs[7] = (unsigned) initialDomain->returnFunct();

	/*
	 * Sanity checks.  
	 */
	Assert ( _kernelStackTop != 0 );
	Assert ( ! ((unsigned)_kernelStackTop & 0x7) );
	Assert ( _stackPointer != 0 );
	Assert ( ! ((unsigned)_stackPointer & 0x7) );
	Assert ( inputs[0] < 17 );
}

SPARCUninterruptableSystemContext::SPARCUninterruptableSystemContext(
	Process * owner, PFV entry, StackSize stackSize, int arg0,
	int arg1)
	: SPARCSystemContext( owner, entry, stackSize, arg0, arg1)
{
	unsigned * const inputs = (unsigned *) &volatileRegs[0];
	((unsigned *)inputs)[0] = INTERRUPT_HIGH;// interrupts will NOT be enabled.
}

/*
 *  Stack backtrace for sparc
 */
static
char *
printFrame(char *s)
{
	char ** sp = (char **) s;
	Console() << "LOCALS: " << "\t" ;
	Console() << asAddress(sp[0]) << "\t" ;
	Console() << asAddress(sp[1]) << "\t" ;
	Console() << asAddress(sp[2]) << "\t" ;
	Console() << asAddress(sp[3]) << "\t" ;
	Console() << "\n\t\t" ;
	Console() << asAddress(sp[4]) << "\t" ;
	Console() << asAddress(sp[5]) << "\t" ;
	Console() << asAddress(sp[6]) << "\t" ;
	Console() << asAddress(sp[7]) << "\t" ;
	Console() << "\nINS: " << "\t\t" ;
	Console() << asAddress(sp[8]) << "\t" ;
	Console() << asAddress(sp[9]) << "\t" ;
	Console() << asAddress(sp[10]) << "\t" ;
	Console() << asAddress(sp[11]) << "\t" ;
	Console() << "\n\t\t" ;
	Console() << asAddress(sp[12]) << "\t" ;
	Console() << asAddress(sp[13]) << "\t" ;
	Console() << asAddress(sp[14]) << "\t" ;
	Console() << asAddress(sp[15]) << "\t" ;
	Console() << "\n" << eor;
	char * nextframe = (char *) ((sp)[14]);
	return nextframe;
}

void
SPARCGenericContext::traceStack()
{
	Console() << "Stack BackTrace:\n" ;
	unsigned frame = 0;
        char *s;
	if (thisProcess()->context() == this)
		asm volatile("mov %%fp, %0" : "=r" (s));
	else
		s = this->_stackPointer;


	flushWindows();		// Just make sure.
	while (
		((s > _userStackTop-_stackBytes) && ( s < _userStackTop))
				||
		((s > _kernelStackTop-_kernelStackBytes) && 
			(s < _kernelStackTop))
		) {
		Console() << "Frame " << frame++ << ":\n" ;
		s = printFrame(s);
		Console() << "\n" ;
		}
	Console() << "End of BackTrace:\n" ;

}

#include "ClassInstaller.h"
DEFINECLASS("SPARCGenericContext",SPARCGenericContext,ProcessorContext,0,0);
DEFINECLASS("SPARCSystemContext",SPARCSystemContext,ProcessorContext,0,0);
DEFINECLASS("SPARCApplicationContext",SPARCApplicationContext,ProcessorContext,0,0);
DEFINECLASS("SPARCUninterruptableSystemContext",SPARCUninterruptableSystemContext,ProcessorContext,0,0);
