#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * SPARCPageFaultInterrupt.cc: SPARC page fault interrupt class.
 */

#include "AccessType.h"
#include "AddressTranslator.h"
#include "Assert.h"
#include "Class.h"
#include "ThisCPU.h"
#include "CPU.h"
#include "Domain.h"
#include "ObjectProxy.h"
#include "OutputStream.h"
#include "ProxyTable.h"
#include "SPARCContext.h"
#include "SPARCInterrupts.h"
#include "SPARCTranslation.h"
#include "SPARCasmutil.h"
#include "SS1devices.h"
#include "ThisProcess.h"
#include "Process.h"
#include <sys/asm_linkage.h>
#include <sys/buserr.h>
#include <sys/psw.h>
#include <sys/reg.h>
#include <sys/mmu.h>

#ifdef __GNUC__
#pragma implementation
#endif

extern "C" void
restartReturnedAssertion()
{
	Console() << "\n_restart returned!\n";
	Assert( NOTREACHED ) ;
}

extern "C" void
stackAlignment()
{
	Console() << thisProcess() << ": Stack alignment problem\n";
	Assert( NOTREACHED ) ;
}

extern "C" void
stackOverflow(unsigned psr, unsigned pc, unsigned sp, unsigned wim,
	unsigned address)
{
	Console() << thisProcess() << ": Stack overflow?:\n\t(psr)" << asHex(psr) 
		<< "\n\t(pc)" << asHex(pc) << "\n\t(sp)" << asHex(sp)
		<< "\n\t(wim)" << asHex(wim) << "\n\t(address)" << asHex(address)
		<< "\n" ;
	Assert( NOTREACHED ) ;
}

/*
 * Used by Interrupts to kill the current process
 * Use side effect of Assert killing the Process
 *
 */
void
killProcessDueToInterrupt( Process *myprocess, char *sp)
{
	Assert(myprocess != 0);
	Console() << " Killing process: " << myprocess << "\n" << eor;
	myprocess->context()->traceStack();
#ifdef RAIDOUTPUTSTREAM
	extern void RaidDump();
	RaidDump();
#endif	
	myprocess->die();
	Assert (NOTREACHED);
}

void
SPARCFPDisabledInterrupt::basicRaise(char * sp)
{
	Process *p = thisProcess();
	if ( ! p->classOf()->isASubclassOf (ApplicationProcessClass) ){
		Console() << p << ": Attempt to enable floating point on non-application process\n" << eor;
		killProcessDueToInterrupt(p, sp);
		};

	Debug << p << ": Enabling FP\n" << eor;
	//SPARCApplicationContext *c = (SPARCApplicationContext*)thisProcess()->context();
	sp += MINFRAME;

	/*
	 * get pointer to saved psr and set EF bit
	 */
	unsigned *psrPointer  = &((unsigned *)sp)[PSR];
	Assert ( ! (*psrPointer & PSR_EF) );
	*psrPointer |= PSR_EF ;

	/*
	 * Set the Float bit in current %psr so we can load %fsr, etc.
	 * without re-taking the FP disabled trap.
	 */
	register unsigned psr;
	asm volatile ("mov %%psr, %0" : "=r" (psr) );
	psr |= PSR_EF;
	asm volatile ("mov %0, %%psr" : : "r" (psr) );
	asm volatile("nop; nop; nop");

	/*
	 * flush the queue and initialize %fsr
	 */
	volatile unsigned fsr;
	unsigned long long junk;
	asm volatile("st %%fsr, [%0]" : : "r"(&fsr) );
	while (fsr & FSR_QNE) {
		asm volatile("std %%fq, [%0]"  : : "r"(&junk));
		asm volatile("st %%fsr, [%0]" : : "r"(&fsr) );
		}
	unsigned zero = 0;
	asm volatile("ld [%0], %%fsr" : :"r"(&zero));

	Debug << p << ": FP enabled\n" << eor;
}


void
SPARCFPInterrupt::basicRaise(char *sp)
{
	Process *p = thisProcess();
	Console() << p << ": Floating Point Interrupt:" ;
	if ( ! ((Object *)p)->classOf()->isASubclassOf (ApplicationProcessClass) ){
		Console() << p << ": Floating point interrupt on non-application process\n" << eor;
		killProcessDueToInterrupt(p, sp);
		};
	//SPARCApplicationContext *c = (SPARCApplicationContext*)thisProcess()->context();
	unsigned psr;
	asm volatile ("mov %%psr, %0" : "=r" (psr) );
	Assert (psr & PSR_EF);

	/*
	 * flush the queue and initialize %fsr
	 */
	volatile unsigned fsr;
	unsigned long long junk;
	asm volatile("st %%fsr, [%0]" : : "r"(&fsr) );
	while (fsr & FSR_QNE) {
		asm volatile("std %%fq, [%0]"  : : "r"(&junk));
		asm volatile("st %%fsr, [%0]" : : "r"(&fsr) );
		}
	unsigned zero = 0;
	asm volatile("ld [%0], %%fsr" : :"r"(&zero));

	killProcessDueToInterrupt(p, sp);
}

void
SPARCIllegalInstructionInterrupt::basicRaise(char *sp)
{
	Process *p = thisProcess();
	Console() << p << ": Illegal Instruction:" ;
	p->context()->traceStack();
	killProcessDueToInterrupt(p, sp);
}

void
SPARCAsynchronousMemoryInterrupt::basicRaise( char *sp)
{
	Debug << "SPARCAsyncronousMemoryInterrupt::basicRaise()" ;
	Process *p = thisProcess();
	Console() << p << ": Asynchronous Memory Interrupt:" ;
	killProcessDueToInterrupt(p, sp);
}

/*
 * Memory error registers on the SS1
 * These are in ASI=2
 */
unsigned 	* const SER		= (unsigned *)0x60000000;
unsigned 	* const SEVAR		= (unsigned *)0x60000004;
unsigned 	* const ASER		= (unsigned *)0x60000008;
unsigned 	* const ASEVAR		= (unsigned *)0x6000000c;

void
SPARCMemoryAlignmentInterrupt::basicRaise( char *sp)
{
	Debug << "SPARCMemoryAlignmentInterrupt::basicRaise()" ;

	readWordASI(SER, ASI_CTL);
	unsigned sevar = readWordASI(SEVAR, ASI_CTL);
	readWordASI(ASER, ASI_CTL);
	readWordASI(ASEVAR, ASI_CTL);
	Process *p = thisProcess();
	Console() << p << " Memory alignment interrupt at " << asAddress((char*)sevar);
	killProcessDueToInterrupt(p, sp);
	Assert( NOTREACHED );
}




void
SS1MemoryInterrupt(char *stackPointer, int type)
{
	IntStorage state ; PROCESSOR::disableInterrupts(state);
	flushWindows();
	stackPointer += MINFRAME;
	unsigned ser = readWordASI(SER, ASI_CTL);
	unsigned sevar = readWordASI(SEVAR, ASI_CTL);
	unsigned aser = readWordASI(ASER, ASI_CTL);
	unsigned asevar = readWordASI(ASEVAR, ASI_CTL);
	PROCESSOR::restoreInterrupts(state);
	int writeAttempted = (ser & SE_RW) == SE_RW;
	int mode = (((unsigned *)stackPointer)[PSR] & PSR_PS) == PSR_PS;
	char *address;
	if (type) { /* Instruction fault */
		address = ((char **)stackPointer)[PC];
		}
	else { /* Data Fault */
		address = (char *)sevar;
		}

	CPU * cpu = thisCPU();
	Assert (cpu != 0);
	AddressTranslator *mmu = cpu->MMU();
	Assert (mmu != 0);
	if (! mmu->enabled() ) {
		SPARCTranslation::fixKernelInitFault(address);
		return;
		}
	Assert( mmu->enabled() );



	/* Make sure we didn't take a DMA trap */
	Assert (! ((address > DMA_BASE) && (address < DMA_END)) );

	Debug << "SS1MemoryInterrupt():"
		<< (mode ? " Kernel" : " User")
		<< (type ? " insn" : " data")
		<< (writeAttempted ? " write" : " read" )
		<< " addr " << asAddress(address)
		<< " pc: " << asAddress( ((char **)stackPointer)[PC] )
		<< "\n\t"
		<< " psr: " << asHex( ((unsigned *)stackPointer)[PSR] )
		<< " ser = " << asHex(ser)
		<< " aser = " << asHex(aser)
		<< " sevar = " << asHex(sevar)
		<< " asevar = " << asHex(asevar)
		<< "\n" << eor;

	ser &= ~SE_RW;

	/*
	 * figure out the access type
	 */
	AccessType attemptedAccess;
	if (mode) {
		// Assert ( !type )("Kernel instruction fault!");
		if (writeAttempted) {
			attemptedAccess = SystemWrite;
			}
		else	{
			attemptedAccess = SystemRead;
			}
		}
	else	{
		if (writeAttempted){
			attemptedAccess = ApplicationWrite;
			}
		else	{
			attemptedAccess = ApplicationRead;
			}
		}

	/* 
	 * Figure out the fault type
	 */
	if (ser & (ser -1)){
		/*
		 * Multiple bits, must probe the page table
		 */
		PAGE_TABLE_ENTRY pte = SPARCTranslation::pageTableEntry(address);
		if (pte.valid())
				ser = SE_PROTERR;
		else
			ser = SE_INVALID;
		}

	AddressTranslationFailureCode code;
	switch (ser) {
		case SE_PROTERR:
			code = ProtectionViolation;
			break;
		default:
		case SE_INVALID:
			code = NonResidentMemory;
			break;
		};

	Process *process = thisProcess();
	Assert ( process != 0 );
	Domain *domain = process->domain();
	Assert(domain != 0);
	FaultErrorCode reason = domain->repairFault(address, attemptedAccess);
	Debug << "Domain::repairFault returned: ";

	switch(reason) {
		/*
		 * These are "normal" and the instruction can resume.
		 */
		case MissingAddressTranslationMapping:
			Debug << "MissingAddressTranslationMapping\n" << eor;
			/*
			 * A ReadOnly write  by kernel???/
			 */
			if (mode)
			domain->translation()->changeProtection(address, 1, NoAccess);
			break;
		case MissingPhysicalMemory:
			Debug << "MissingPhysicalMemory\n" << eor;
			break;
		case InsufficientProtectionLevel:
			Debug << "InsufficientProtectionLevel\n" << eor;
			break;


		/*
		 * These are fatal errors
		 */
		case WriteOfProtectedSystemMemory:
			Console() << "WriteOfProtectedSystemMemory:";
			Console() << " Address: " << asAddress(address) << ",";
			killProcessDueToInterrupt(process, stackPointer);
			Assert (NOTREACHED);
			break;
		case AccessOfUnmanagedMemory:
			Console() << "AccessOfUnmanagedMemory:";
			Console() << " Address: " << asAddress(address) << ",";
			killProcessDueToInterrupt(process, stackPointer);
			Assert (NOTREACHED);
			break;
		case ReadOfProtectedSystemMemory:
			Console() << "ReadOfProtectedSystemMemory:";
			Console() << " Address: " << asAddress(address) << ",";
			killProcessDueToInterrupt(process, stackPointer);
			Assert (NOTREACHED);
			break;
		case WriteOfReadOnlyMemory:
			Console() << "WriteOfReadOnlyMemory:";
			Console() << " Address: " << asAddress(address) << ",";
			killProcessDueToInterrupt(process, stackPointer);
			Assert (NOTREACHED);
			break;
		case DynamicWriteOfReadOnlyMemory:
			Console() << "DynamicWriteOfReadOnlyMemory:";
			Console() << " Address: " << asAddress(address) << ",";
			killProcessDueToInterrupt(process, stackPointer);
			Assert (NOTREACHED);
			break;
		case DynamicAccessViolation:
			Console() << "DynamicAccessViolation:";
			Console() << " Address: " << asAddress(address) << ",";
			killProcessDueToInterrupt(process, stackPointer);
			Assert (NOTREACHED);
			break;
		default:
			Console() << "Unknown (default) access violation:";
			Console() << " Address: " << asAddress(address) << ",";
			Assert(NOTREACHED);
			break;
		};

	/*
	 * Return to restart the offending instruction
	 */
}


void
SPARCDataAccessInterrupt::basicRaise( char *stackPointer)
{
	Debug << "SPARCDataAccessInterrupt::basicRaise()\n" << eor;
	SS1MemoryInterrupt(stackPointer, 0);

}

void
SPARCInstructionAccessInterrupt::basicRaise( char *stackPointer)
{
	Debug << "SPARCInstructionAccessInterrupt::basicRaise()\n" << eor;
	SS1MemoryInterrupt(stackPointer, 1);
}

void
SPARCDivideByZeroInterrupt::basicRaise( char *sp)
{
	Process *p = thisProcess();
	Console() << p << ": Divide by zero, killing process\n" << eor;
	killProcessDueToInterrupt(p, sp);
}


extern "C" {
int
ProxyCallAssist( unsigned int methodNumber, int fromKernel,
	ProxyReturnFlags *returnFlags, void *refPointer,
	ProxiableObject **from, ProxiableObject **to );

ProxiableObject *
ProxyReturnAssist(ProxiableObject *returnValue, ProxyReturnFlags *returnFlags);


typedef ProxiableObject * (* proxyPFV) (ProxiableObject *, ProxiableObject *, 
ProxiableObject *, ProxiableObject *, ProxiableObject *, ProxiableObject *,
ProxiableObject *, ProxiableObject *, ProxiableObject *, ProxiableObject *,
ProxiableObject *, ProxiableObject *);
typedef ProxiableObjectRef (* proxyRefPFV)
(ProxiableObject *, ProxiableObject *, 
ProxiableObject *, ProxiableObject *, ProxiableObject *, ProxiableObject *,
ProxiableObject *, ProxiableObject *, ProxiableObject *, ProxiableObject *,
ProxiableObject *, ProxiableObject *);

int ProxyCallLog = 0;
};

void
SPARCProxyInterrupt::basicRaise(char *sp)
{
	Debug << "SPARCProxyInterruptHandler(" << asAddress((char*)sp) 
		<< ", " << ")\n" << eor;

	unsigned *msp = (unsigned *)(sp + MINFRAME);
	unsigned *usp = (unsigned *)sp;
	IntStorage state ; PROCESSOR::disableInterrupts(state);

	flushWindows();	// just be sure it's all on the stack.
	unsigned mode  = ((msp[PSR] & PSR_PS) == PSR_PS);

#ifdef NEVER
	Process * process = 0;
	if (ProcessResourceUsageOn && !mode ) {
        	process = CPU::_runningProcesses[0];
                Assert( process != 0 );
                process->stopUserTime();
		process->usage()->numSwap++;
	}
#endif

	/*
	 * Find the callers (pre-trap stub frame) stack pointer,
	 * frame pointer, and structure return slot value to
	 * pass to callee.
	 */
	unsigned *stubSP = (unsigned*)usp[14];
	unsigned *userSP = (unsigned*)stubSP[14];
	void * structReturnSlot = (void *)(userSP[16]);
	unsigned *argSP = stubSP;

	int methodNumber = usp[8];	// Trap %i0
	Assert (methodNumber < 100);

	/*
	 * Set up the table of input args to the CallAssist
	 * It expects the args contiguously, with the return
	 * address as the first arg, so fake it up.
	 *
	 * The proxyCallAssist code expects the first param to be
	 * the return address, which is in a register on the sparc,
	 * so it gets ignored.  If the methodnumber is -1 (constructor)
	 * then it expects the arguments to be shifted and the first
	 * (second really) to be the name.
	 *
	 * If it is a constructor the user frame is one level higher
	 * due to the constructor stub above the ObjectProxy stub
	 * and the string argument is the first arg in the next layer in.
	 */
	ProxiableObject* inArgs[MAXPARAMS + 1];
	int offset = 1;
	if (methodNumber == -1) {
		offset = 3;
		unsigned *constructorSP = stubSP;
		argSP = (unsigned*)stubSP[14];
		inArgs[1] = (ProxiableObject*)constructorSP[8];
		}
	for (int i = 0; i < MAXPARAMS; i++)
		inArgs[i+offset] = (i < 6) ? 
			(ProxiableObject*)argSP[8+i] :
			(ProxiableObject *)userSP[23+i-6];

	struct ProxyReturnFlags returnFlags;
	ProxiableObject *outArgs[MAXPARAMS];

	ProxiableObject * retval = 0;
	int type = ProxyCallAssist(methodNumber, mode, &returnFlags,
		structReturnSlot, inArgs, outArgs);
	PROCESSOR::restoreInterrupts(state);
	if ( ProxyCallLog ) {
		Console() << "Proxy call " 
			  << asHex((unsigned)returnFlags.calledMethod)
			  << "\n" << eor;
	}
	switch ( type ) {
		case 0:
			break;
		case 1:
			retval = (*(proxyPFV)returnFlags.calledMethod)(outArgs[0],
				outArgs[1], outArgs[2], outArgs[3], outArgs[4],
				outArgs[5]
				, outArgs[6], outArgs[7], outArgs[8],
				outArgs[9], outArgs[10], outArgs[11]);
			break;
		case 2:
			if ( returnFlags.returnType == 2 ){
				ProxiableObjectRef poRef = 
				(*(proxyRefPFV)returnFlags.calledMethod)(
				outArgs[0], outArgs[1], outArgs[2], outArgs[3],
				outArgs[4], outArgs[5], outArgs[6], outArgs[7], 
				outArgs[8], outArgs[9], outArgs[10], outArgs[11]
				);
				retval = (ProxiableObject *) structReturnSlot;
				Assert (retval != 0);
				argSP[15] +=4;
				*(ProxiableObject **)retval = 
					ProxyReturnAssist((ProxiableObject *)
							  poRef, &returnFlags);
			} else {
				retval = (*(proxyPFV)returnFlags.calledMethod)(
				outArgs[0], outArgs[1], outArgs[2], outArgs[3],
				outArgs[4], outArgs[5], outArgs[6], outArgs[7], 
				outArgs[8], outArgs[9], outArgs[10], outArgs[11]
				);
				retval = ProxyReturnAssist(retval,&returnFlags);
			}
			break;
		case 3:	
			/*
			 * This is an upcall, boo hoo hoo
			 */
			Assert (NOTREACHED);
			break;
		}

#ifdef NEVER
        if (ProcessResourceUsageOn && !mode ) {
                process->startUserTime(); 
        }
#endif

	stubSP[8] = (unsigned)retval;
	//asm volatile ("mov %0, %%i0" : : "r"(retval));
}
