#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * VCPageFaultInterrupt.cc: VC page fault interrupt class.
 *
 */

#include "AccessType.h"
#include "AddressTranslator.h"
#include "Assert.h"
#include "Class.h"
#include "ThisCPU.h"
#include "CPU.h"
#include "Domain.h"
#include "ObjectProxy.h"
#include "OutputStream.h"
#include "ProxyTable.h"
#include "VCContext.h"
#include "VCInterrupts.h"
#include "Process.h"
#include "ThisProcess.h"

extern "C" {

#if defined(linux) 
# include <asm/ioctls.h>
#elif defined(__APPLE__)
#include <sys/filio.h>
#else
# include <vm/faultcode.h>
# include <sys/filio.h>
#endif /* linux */

#include <signal.h>
  int ioctl(int, int, char *);
};

#ifdef __GNUC__
#pragma implementation
#endif

/*
 * Used by Interrupts to kill the current process
 * Use side effect of Assert killing the Process
 *
 */
void
killProcessDueToInterrupt( Process *myprocess)
{
	Assert(myprocess != 0);
	Console() << " Killing process: " << myprocess << "\n" << eor;
	if (myprocess->isKindOf(SystemProcessClass)) {
		thisCPU()->debugBreakpoint();
	}
	myprocess->context()->traceStack();
	myprocess->die();
	Assert (NOTREACHED);
}

void
VCFPInterrupt::basicRaise(char * info)
{
	Process *p = thisProcess();
	Console() << p << ": Floating Point error\n" << eor;
	killProcessDueToInterrupt(p);
}

void
VCIllegalInstructionInterrupt::basicRaise(char *info)
{
	Process *p = thisProcess();
	Console() << p << ": Illegal Instruction:" ;
	killProcessDueToInterrupt(p);
}


int VCIOInterrupt::ioInterrupts;
int VCIOInterrupt::fdTable[ nIOInterrupts ];;
Interrupt * VCIOInterrupt::interruptTable[ nIOInterrupts ];

void
VCIOInterrupt::basicRaise(char *info)
{
  long	inputAvail;

  for (int i = 0; i < ioInterrupts; i++) {
    int fd = fdTable[i];
    ioctl(fd, FIONREAD, (char *) &inputAvail);
    if (inputAvail)
      interruptTable[i]->raise(info);
  }
}


Interrupt *
VCIOInterrupt::setIOInterrupt( int fd, Interrupt * interrupt, PFV assist )
{
  Assert(fd >= 0);
  Assert(interrupt != 0);

  int i;
  for (i = 0; i < ioInterrupts; i++)
    if (fdTable[i] == fd)
      break;

  if (i == ioInterrupts) {
    Assert(ioInterrupts < nIOInterrupts);
    interruptTable[ ioInterrupts++ ] = 0;
    fdTable[ i ] = fd;
  }

  Interrupt * oldInterrupt = interruptTable[i];
  interruptTable[ i ] = interrupt;

  return(oldInterrupt);
}

void
VCMemoryInterrupt::basicRaise(char * info)
{
	VCSignalInfo *sigInfo = (VCSignalInfo *)info;
	int unixCode = sigInfo->code();
	char *address = (char*)sigInfo->address();

	int writeAttempted = 1;
	unsigned mode  = 1;	// kernel (for now)

	/*
	 * figure out the access type
	 */
	AccessType attemptedAccess;
	if (mode) {
		if (writeAttempted)
			attemptedAccess = SystemWrite;
		else
			attemptedAccess = SystemRead;
		}
	else	{
		if (writeAttempted)
			attemptedAccess = ApplicationWrite;
		else
			attemptedAccess = ApplicationRead;
		}

	AddressTranslationFailureCode code;
	switch (unixCode) {
		case SEGV_MAPERR:
			code = NonResidentMemory;
			break;
		case SEGV_ACCERR:
			code = ProtectionViolation;
			break;
		default:
			code = SomethingElse;
			//Assert( NOTREACHED );
		};

	Process *process = thisProcess();
	CPU * cpu = thisCPU();
	Assert (cpu != 0);
	AddressTranslator *mmu = cpu->MMU();
	Assert (mmu != 0);
	Assert ( process != 0 );
	Domain *domain = process->domain();
	Assert(domain != 0);
	FaultErrorCode reason = domain->repairFault(address, attemptedAccess);
	Debug << "Domain::repairFault returned: ";

	switch(reason) {
		/*
		 * These are "normal" and the instruction can resume.
		 */
		case MissingAddressTranslationMapping:
			Debug << "MissingAddressTranslationMapping\n" << eor;
			// if (mode) domain->translation()->changeProtection(address, 1, NoAccess);
			break;
		case MissingPhysicalMemory:
			Debug << "MissingPhysicalMemory\n" << eor;
			break;
		case InsufficientProtectionLevel:
			Debug << "InsufficientProtectionLevel\n" << eor;
			break;


		/*
		 * These are fatal errors
		 */
		case AccessOfUnmanagedMemory:
			Console() << "AccessOfUnmanagedMemory:";
			Console() << " Address: " << asAddress(address) << ",";
			killProcessDueToInterrupt(process);
			Assert (NOTREACHED);
			break;
		case ReadOfProtectedSystemMemory:
			Console() << "ReadOfProtectedSystemMemory:";
			Console() << " Address: " << asAddress(address) << ",";
			killProcessDueToInterrupt(process);
			Assert (NOTREACHED);
			break;
		case WriteOfProtectedSystemMemory:
			Console() << "WriteOfProtectedSystemMemory:";
			Console() << " Address: " << asAddress(address) << ",";
			killProcessDueToInterrupt(process);
			Assert (NOTREACHED);
			break;
		case WriteOfReadOnlyMemory:
			Console() << "WriteOfReadOnlyMemory:";
			Console() << " Address: " << asAddress(address) << ",";
			killProcessDueToInterrupt(process);
			Assert (NOTREACHED);
			break;
		case DynamicWriteOfReadOnlyMemory:
			Console() << "DynamicWriteOfReadOnlyMemory:";
			Console() << " Address: " << asAddress(address) << ",";
			killProcessDueToInterrupt(process);
			Assert (NOTREACHED);
			break;
		case DynamicAccessViolation:
			Console() << "DynamicAccessViolation:";
			Console() << " Address: " << asAddress(address) << ",";
			killProcessDueToInterrupt(process);
			Assert (NOTREACHED);
			break;
		default:
			Console() << "Unknown (default) access violation:";
			Console() << " Address: " << asAddress(address) << ",";
			Assert(NOTREACHED);
			break;
		};

	/*
	 * Return to restart the offending instruction
	 */

}

int ProxyCallLog = 0;

extern "C" {
int
ProxyCallAssist( unsigned int methodNumber, int fromKernel,
	ProxyReturnFlags *returnFlags, void *refPointer,
	ProxiableObject **from, ProxiableObject **to );

ProxiableObject *
ProxyReturnAssist(ProxiableObject *returnValue, ProxyReturnFlags *returnFlags);

typedef ProxiableObject * (* proxyPFV) (ProxiableObject *, ProxiableObject *, 
ProxiableObject *, ProxiableObject *, ProxiableObject *, ProxiableObject *,
ProxiableObject *, ProxiableObject *, ProxiableObject *, ProxiableObject *,
ProxiableObject *, ProxiableObject *);

typedef ProxiableObjectRef (* proxyPFR) (ProxiableObject *, ProxiableObject *, 
ProxiableObject *, ProxiableObject *, ProxiableObject *, ProxiableObject *,
ProxiableObject *, ProxiableObject *, ProxiableObject *, ProxiableObject *,
ProxiableObject *, ProxiableObject *);
};

// void VCProxyInterrupt::basicRaise(char * info)
ProxiableObject *
VCProxyInterruptbasicRaise(int methodNumber)
{

#if defined(linux)
    /* 'pushal' decrements the stack where 'ta 3' does not */
    asm volatile( "pushal; nop" );
    //asm volatile( "add $32, %sp" ); // reincrement sp
#elif defined(__APPLE__)
	Assert(NOTREACHED);
#else
    asm volatile("ta 3; nop");	// just be sure it's all on the stack.
#endif

	unsigned mode  = 0;	// user fake

	unsigned *sp;

#if defined(linux)
	asm volatile("mov %%sp, %w0" : "=r"(sp));
#elif defined(__APPLE__)
	Assert(NOTREACHED);
#else
	asm volatile("mov %%sp, %0" : "=r"(sp));
#endif

	/*
	 * Find the callers (pre-trap stub frame) stack pointer,
	 * frame pointer, and structure return slot value to
	 * pass to callee.
	 */
	unsigned *stubSP = (unsigned*)((unsigned*)sp)[14];
	unsigned *userSP = (unsigned*)((unsigned*)stubSP)[14];
	void * structReturnSlot = (void *)(userSP[16]);

	//unsigned *caller_pc = (unsigned*)(stubSP[15]);	// %i7
	//int returnsStruct = ((((*(caller_pc+8) >> 22 ) | (0x1f << 3)) ^ (0x1f << 3)) == 0);
	/*
	 * Set up the table of input args to the CallAssist
	 * It expects the args contiguously, with the return
	 * address as the first arg, so fake it up.
	 *
	 * The proxyCallAssist code expects the first param to be
	 * the return address, which is in a register on the sparc,
	 * so it gets ignored.  If the methodnumber is -1 (constructor)
	 * then it expects the arguments to be shifted and the first
	 * (second really) to be the name.
	 *
	 * If it is a constructor the user frame is one level higher
	 * due to the constructor stub above the ObjectProxy stub
	 * and the string argument is the first arg in the next layer in.
	 */
	ProxiableObject* inArgs[MAXPARAMS + 1];
	int offset = 1;

	if (methodNumber == -1) {
		offset = 3;
		unsigned *constructorSP = stubSP;
		stubSP = (unsigned*)stubSP[14];
		inArgs[1] = (ProxiableObject*)constructorSP[8];
		}

	for (int i = 0; i < MAXPARAMS; i++)
		inArgs[i+offset] = (i < 6) ? 
			(ProxiableObject*)stubSP[8+i] :
			(ProxiableObject *)userSP[23+i-6];

	struct ProxyReturnFlags returnFlags;
	ProxiableObject *outArgs[MAXPARAMS];

	int type = ProxyCallAssist(methodNumber, mode, &returnFlags,
		structReturnSlot, inArgs, outArgs);
	ProxiableObject * retval ;
	if ( ProxyCallLog ) {
		Console() << "Proxy Call " 
			<< asHex( (unsigned) returnFlags.calledMethod )
                        << " ( " << methodNumber << " ) "
			<< " type " << type << " returnType " 
                        << (int) returnFlags.returnType << "\n" << eor;
	}
	switch ( type ) {
		case 0:
			retval = 0;
			break;
		case 1:
			retval = (*(proxyPFV)returnFlags.calledMethod)(outArgs[0],
				outArgs[1], outArgs[2], outArgs[3], outArgs[4],
				outArgs[5]
				, outArgs[6], outArgs[7], outArgs[8],
				outArgs[9], outArgs[10], outArgs[11]);
			break;
		case 2:
			if ( returnFlags.returnType == 2 ){
				ProxiableObjectRef poRef  = 
				(*(proxyPFR)returnFlags.calledMethod)(outArgs[0],
				outArgs[1], outArgs[2], outArgs[3], outArgs[4],
				outArgs[5] , outArgs[6], outArgs[7], outArgs[8],
				outArgs[9], outArgs[10], outArgs[11]);
				retval = (ProxiableObject *) structReturnSlot;
				Assert (retval != 0);
				stubSP[15] +=4;
				*(ProxiableObject **)retval =
					ProxyReturnAssist(
						(ProxiableObject *)poRef,
						&returnFlags);
				}
			else {
				retval = (*(proxyPFV)returnFlags.calledMethod)(outArgs[0],
					outArgs[1], outArgs[2], outArgs[3], outArgs[4],
					outArgs[5] , outArgs[6], outArgs[7], outArgs[8],
					outArgs[9], outArgs[10], outArgs[11]);
				retval = ProxyReturnAssist(retval, &returnFlags);
				}
			break;
		default:
			/*
			 * This is an upcall, boo hoo hoo
			 */
			retval = 0;
			Assert (NOTREACHED);
			break;
		}
	return retval;
}
