#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987-1992
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */


/*
 * Turn PAGE_TABLE_SWITCHING on if you want to play with real VM 
 * domain switching.  It can be slow depending on the size of the table.
 */
// #define PAGE_TABLE_SWITCHING

#include "AddressTranslator.h"
#include "VCPageTable.h"
#include "CPU.h"
#include "ThisCPU.h"
#include "ClassStar.h"
#include "ClassInstaller.h"
#include "Domain.h"
#include "Kernel.h"
#include "ThisProcess.h"
#include "MemoryBlockAllocator.h"
#include "PhysicallyAddressableUnit.h"
#include "OutputStream.h"
extern "C" {
#include <sys/mman.h>
#include <unistd.h>
//int mprotect();
};


#ifdef __GNUC__
#pragma implementation
#endif


#ifdef PAGE_TABLE_SWITCHINGMAP_FAILED
const PAGE_MAPPINGS_PER_ADT	= 0x2000;// 8k pages
const ALLOC_SIZE = sizeof(VCPageMapping) * PAGE_MAPPINGS_PER_ADT;
#endif

Allocator * VCPageTable::_allocator = 0;
Lock * VCPageTable::_lock = 0;

extern int StoreFD;
//extern char * StoreVirtualAddress;

VCPageTable::VCPageTable( 
		VCPageTable *parent)
{
	_mappings = 0;
	_numMappings = 0;
	_baseAddr = 0;
	_endAddr = 0;
}

VCPageTable::~VCPageTable()
{
	_numMappings = 0;
	_baseAddr = 0;
	_endAddr = 0;
	if (_mappings != 0){
		_lock->acquire();
		_allocator->free(_mappings);
		_lock->release();
		}
	_mappings = 0;
}

void
VCPageTable::init( unsigned pages, VirtualMemoryRange *)
{
#ifdef PAGE_TABLE_SWITCHING
	_allocator = new MemoryBlockAllocator( pages * ALLOC_SIZE, ALLOC_SIZE);
#endif
	_lock = new Lock();
}

/* 
 * Used to add the mapping for a single page
 */
void
VCPageTable::addPageMapping( const char * vaddr,
		    PhysicallyAddressableUnit * pau, ProtectionLevel prot)
{
	pau->setReferenced(1);
	pau->setModified(1);

	/*
	 * which unit in the store file
	 */
	//unsigned unit = (pau->address() - StoreVirtualAddress);
	unsigned unit = (unsigned)pau->address();

#ifdef PAGE_TABLE_SWITCHING
	/*
	 * allocate table if it's not there
	 */
	if (_mappings == 0){
		Assert (_allocator != 0);
		_mappings = (VCPageMapping *) _allocator->allocate(ALLOC_SIZE);
		for (int i = 0 ; i < PAGE_MAPPINGS_PER_ADT; i++ )
			_mappings[i] = VCPageMapping() ;
		}
	Assert ( _mappings != 0);

	/*
	 * Find the slot for the page
	 */
	for (unsigned slot = 0; slot < _numMappings; slot++) {
		if (vaddr >= _mappings[slot].vaddress())
			break;
		}
	Assert ( slot < PAGE_MAPPINGS_PER_ADT );
	if (slot == _numMappings)
		_numMappings++;

	/*
	 * Shift if necessary to maintain ordering
	 */
	else if (slot < _numMappings && vaddr > _mappings[slot].vaddress()) {
		Assert ( _numMappings < PAGE_MAPPINGS_PER_ADT);
		for ( unsigned s = _numMappings; s > slot; s--)
			_mappings[s] = _mappings[s-1];
		_numMappings++;
		}

	/* 
	 * update the slot, and set the ref/mod bits in the pau
	 */
	_mappings[slot] = VCPageMapping (vaddr, unit, pau, prot);
#endif

	/*
	 * Now add the mmap mapping to UNIX
	 */
	int unixProt = PROT_READ|PROT_WRITE|PROT_EXEC;
	Debug << this << "::addPageMapping: mmap vaddr(" << asAddress(vaddr)
		<< ")->paddr(" << asHex(unit) << ")\n" << eor;
	void *retvaddr = mmap((char*)vaddr, PhysicalPageSize, unixProt, MAP_FIXED|MAP_SHARED, StoreFD, unit);
	if (retvaddr == MAP_FAILED) {
	  extern int errno;
	  Console() << "Mapping failed with errno " << errno << "\n" ;
	  //Assert (NOTREACHED);
	}
}

void
VCPageTable::removePageMapping( const char * vaddr, int imActive)
{
#ifdef PAGE_TABLE_SWITCHING
	/*
	 * Find the slot for the page
	 */
	for (unsigned slot = 0; slot < _numMappings; slot++) {
		if (vaddr == _mappings[slot].vaddress())
			break;
		}
	if ( slot < _numMappings ){ // Page was there
		/*
		 * Slide the higher entries down
		 */
		_numMappings--;
		for (int i = slot; i < _numMappings; i++)
			_mappings[i] = _mappings[i+1];
		}
#endif
	
	if (imActive ) {
		Debug << this << "::addPageMapping: munmap vaddr(" 
			<< asAddress(vaddr) << ")\n" << eor;
		int r = munmap((char*)vaddr, 4096);
		Assert (r == 0);
		}

}


void
VCPageTable::addMapping( const char * vaddr,
		    PhysicalMemoryChain * chain, ProtectionLevel prot )
{
	_lock->acquire();
	unsigned imActive = (this == thisCPU()->MMU()->currentlyActiveDomain()->translation()) || (this == TheKernel->domain()->translation());
	Assert (imActive);
	unsigned bytes = chain->lengthInBytes();
	Assert (bytes > 0);
	const char *base = vaddr;
	const char *end = vaddr + bytes - 1;
	PhysicallyAddressableUnit * pau = chain->chain();

	while (base < end) {
		addPageMapping(base, pau, prot);
		pau = pau->nextLink();
		base += PhysicalPageSize;
		}
	_lock->release();
}

void
VCPageTable::basicRemoveMapping( const char * vaddr,
	unsigned nbytes )
{
	Assert (nbytes > 0);
	_lock->acquire();
	const char *base = vaddr;
	const char *end = vaddr + nbytes - 1;

	unsigned imActive = (this == thisCPU()->MMU()->currentlyActiveDomain()->translation()) || (this == TheKernel->domain()->translation());
	while (base < end) {
		removePageMapping(base, imActive);
		base += PhysicalPageSize;
		}
	_lock->release();
}

void
VCPageTable::activate()
{
	_lock->acquire();
	int r =  munmap((char*)TheKernel->applicationStart(), TheKernel->applicationSize());
	Assert (r == 0);
#ifdef PAGE_TABLE_SWITCHING
	int unixProt = PROT_READ|PROT_WRITE|PROT_EXEC;
	for (int i = 0; i < _numMappings; i++ ) {
		ProtectionLevel prot = _mappings[i].prot();
		PhysicallyAddressableUnit *pau = _mappings[i].pau();
		pau->setReferenced(1);
		pau->setModified(1);
		_mappings[i].vaddress()  == mmap(_mappings[i].vaddress(), 4096, unixProt, MAP_FIXED|MAP_SHARED, StoreFD, _mappings[i].unit());
		Assert (_mappings[i].vaddress());
		}
#endif
	_lock->release();
}


int
VCPageTable::basicChangeProtection( const char * vaddr,
	unsigned nbytes, ProtectionLevel protection )
{
	_lock->acquire();
	int ret = 0;
	unsigned imActive = (this == thisCPU()->MMU()->currentlyActiveDomain()->translation()) || (this == TheKernel->domain()->translation());
	Assert (imActive);
	Assert (nbytes > 0);
	int unixProt = PROT_READ|PROT_WRITE|PROT_EXEC;
	if ( mprotect((char*)vaddr, nbytes, unixProt) == 0)
		ret = 1;
	_lock->release();
	return ret;
}

int
VCPageTable::basicChangeStatus( const char * vaddr,
	unsigned nbytes, AddressTranslationStatus clear,
				  AddressTranslationStatus set )
{
	Assert (NOTREACHED);
	return(0);
}

DEFINECLASS("VCPageTable",VCPageTable,AddressTranslation,0,0);
