/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Author: David W. Dykstra (dykstra@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * i586Context.c: Processor context for the i586.
 */

#include "i586Context.h"
#include "SegmentDescriptor.h"	/* to get PrivilegeLevel */
#include "OutputStream.h"
#include "PersistentMemoryObject.h"
#include "OutputStream.h"
#include "ThisProcess.h"
#include "Domain.h"
#include "Kernel.h"
#include "CPU.h"
#include "Util.h"
#include "Vectors.h"
#include "ProxyTable.h"
#include "TaskStateSegment.h"
#include "AddressTranslator.h"
#include "ADDRESS_TRANSLATION.h"
#include "ThisProcess.h"
#include "ThisCPU.h"

extern unsigned short DataSegments[];
extern unsigned short CodeSegments[];

i586GenericContext::i586GenericContext( Process * owner,
                                        StackSize stackSize )
    : ProcessorContext( owner )
{
    Assert( sizeof( UserProxyStack ) <= ( MAXPARAMS * sizeof( int * ) ) );

    int nbytes = 0;

    switch( stackSize ) {
        case none:
            nbytes = 0;
            break;

        case tiny:
            nbytes = 512;
            break;

        case heapBootStack:
        case small:
            nbytes = 1024;
            break;

        case normal:
            nbytes = 2048;
            break;

        case large:
        case application:
            nbytes = 4096;
            break;

        case huge:
            nbytes = 4096*4;
            break;
            /*
             * @@@ added the gigantic case since the kernel setup process
             * appears to be running out of stack space.
             *        -- Aamod & Stan
             */
        case gigantic:
            nbytes = 4096*8;
            break;

        default:
            Assert( NOTREACHED );
    }

#ifndef CFRONT
    nbytes=nbytes*4;
#endif
    stackBound[ 0 ] = new char[ nbytes ];
    Assert( stackBound[ 0 ] != 0 );
    stackBase[ 0 ] = &(stackBound[ 0 ])[ nbytes ];
    stackPointer[ 0 ] = stackBase[ 0 ];

    startKernelStack = stackPointer[ 0 ];

    sharedStackSize = nbytes;

    stackBound[ 1 ] = 0;
    stackPointer[ 1 ] = 0;
    stackBase[ 1 ] = 0;

    userStackPointer = 0;

    Debug << this << "::i586GenericContext(): stack at "
          << asAddress( stackBound ) << ", stackpointer "
          << asAddress( stackPointer ) << "\n" << eor;
}

i586GenericContext::~i586GenericContext()
{
    /* 
     * TO DO: free the user stack memory object(s)
     */
    delete stackBound[ 0 ];
}

static char *
allocateUserStack( int size, Domain * domain )
{
    char * stack;

    PersistentMemoryObjectStar memObj =
    TheKernel->allocateMemoryObjectForBackingStore();

    Assert( memObj != 0 );
    memObj->setName( "userStack" );

    memObj->setNumberOfUnits( size / memObj->unitSize() );

    stack = (char *) domain->add( memObj, ReadWrite );

    Assert( stack != 0 );

    return( stack );
}

extern "C" void _ProxyReturnCall();
extern TaskStateSegment chtss;

#define UPS ((UserProxyStack *) space)

//STATIC inline
char * 
i586GenericContext::saveUserStack( char * usp )
{
    char * oldusp = userStackPointer;
    userStackPointer = usp;
    return ( oldusp );
}

/*
 * Subclasses cannot inherit a pure virtual member.
 * Thus we must define this dummy method here.
 */
void
i586GenericContext::restoreFromInitialContext(Process * p)
{
    Assert( NOTREACHED );
}

#if (defined CFRONT)||(defined i586)
void
i586GenericContext::noOp()
{
/* hack to force vtbl generation */
}
#endif

/*
 * Allocate a  stack of the specified size. Memory won't actually be paged
 *   in until needed.  Backing store space will be used, however.
 * Will probably want to have a 'StackMemoryObject' per domain from
 *   which to allocate stacks instead of creating a whole new memory
 *   object for every stack.
 */
i586ApplicationContext::i586ApplicationContext( Process * owner,
                                                PFV entry, StackSize stackSize, Domain *initialDomain, int a0, int a1,
                                                StackSize userStackSize, char * usp )
    : i586GenericContext( owner, stackSize )
{
    Debug << "i586ApplicationContext::ctor()\n" << eor;

    userStackPointer = usp;

    int istackSize;
    switch( userStackSize ) {
        case none:
            istackSize = 0;
            break;
        case tiny:
            istackSize = 4096;
            break;
        case small:
            istackSize = 4096*2;
            break;
        case normal:
        case application:
            istackSize = 4096*4;
            break;
        case large:
            istackSize = 4096*8;
            break;
        case huge:
            istackSize = 4096*16;
            break; 
        case gigantic:
            istackSize = 4096*1024;
            break;
        default:
            istackSize = (int)userStackSize;
    }

    if( ! userStackPointer )
        userStackPointer = allocateUserStack( istackSize,
                                              initialDomain ); 

    const int userRing = 1;
    stackBound[ userRing ] = userStackPointer;
    userStackPointer += istackSize;
    stackBase[ userRing ] = userStackPointer;
    stackPointer[ userRing ] = userStackPointer;

    InitialApplicationContextStackDescription * isd =
    (InitialApplicationContextStackDescription *) (stackPointer[ 0 ]
                                                   - sizeof( InitialApplicationContextStackDescription ));

    stackPointer[ 0 ] = (char *) isd;

    isd->arg0 = a0;
    isd->arg1 = a1;
    isd->returnPoint = initialDomain->returnFunct();

    isd->iretFrame.entryPoint = entry;

    isd->iretFrame.userDataSegment = DataSegments[ userRing ];
    isd->iretFrame.userCodeSegment = CodeSegments[ userRing ];

    isd->iretFrame.returnEflags.value = 0;
    isd->iretFrame.returnEflags.flags.ief = 1;

    isd->iretFrame.userStackPointer = (unsigned int *) userStackPointer;
    isd->iretFrame.userStackSegment = isd->iretFrame.userDataSegment;

}


void
i586ApplicationContext::initialRestoreAssist( Process * previousProcess )
{
    /*
     * stuff to do in the context of the new process
     */
    InitialApplicationContextStackDescription * isd = 
    (InitialApplicationContextStackDescription *) stackPointer[ 0 ];

    /*
     * call parent method
     */
    restoreAssist( previousProcess );

    /*
     * Allocate default proxy and pass it to the new application.
     */
    isd->defaultProxy = process()->domain()->defaultProxy();

    /*
     * Will be switching to new domain soon.
     */
    process()->setCurrentDomain( process()->domain() );
}

#if (defined CFRONT)||(defined i586)
void
i586ApplicationContext::noOp()
{
/* hack to force vtbl generation */
}
#endif

i586SystemContext::i586SystemContext( Process * owner, PFV entryPoint,
                                      StackSize stackSize, int a0, int a1 )
    : i586GenericContext( owner, stackSize )
{
    /*
     * Keep a copy of the initial entry point and its arguments in
     * the save area.
     */
    pc = (int) entryPoint;
    edi = a0;
    esi = a1;

    /*
     * I stash a flag indicating whether to enable interrupts or
     * not in ebx. initialRestoreAssist() knows to find it here.
     * Subclasses can re-assign to this to create a context which
     * does not start out with interrupts enabled.
     */
    ebx = 1;
}

i586SystemContext::i586SystemContext( Process * owner, OPFV entryPoint,
                                      StackSize stackSize, int a0, int a1 )
    : i586GenericContext( owner, stackSize )
{
    /*
     * Keep a copy of the initial entry point and its arguments in
     * the save area.
     */
    pc = (int) entryPoint;
    edi = a0;
    esi = a1;
    /*
     * I stash a flag indicating whether to enable interrupts or
     * not in ebx. initialRestoreAssist() knows to find it here.
     * Subclasses can re-assign to this to create a context which
     * does not start out with interrupts enabled.
     */
    ebx = 1;
}

//TODO: FIXME
i586SystemContext::i586SystemContext( Process * owner, PFV entryPoint,
                                      StackSize stackSize, OPFV arg0, int arg1 ) 
    : i586GenericContext (owner, stackSize)
{
    Assert(NOTREACHED);
    pc = (int) entryPoint;

    ebx = 1;
}

#if (defined CFRONT)||(defined i586)
void
i586SystemContext::noOp()
{
/* hack to force vtbl generation */
}
#endif


void
i586SystemContext::initialRestoreAssist( Process * previousProcess )
{
    typedef void (* PFVII) (int, int);
    PFVII entryPoint = (PFVII) pc;

    restoreAssist( previousProcess );

    if( ebx ) {
        //becomeInterruptable();
    }

    (* entryPoint)( edi, esi );

    Assert( process() != 0 );

    process()->die();
    Assert( NOTREACHED );
}

i586UninterruptableSystemContext::i586UninterruptableSystemContext(
    Process * owner, PFV entry, StackSize stackSize, int a0, int a1 )
    : i586SystemContext( owner, entry, stackSize, a0, a1 )
{
    /*
     * See comment in SystemContext ctor.
     */
    ebx = 0;
}

/*
 * On the Multimax, start() is in assembler, but it doesn't need to be on
 *   the 80586, that's why it's here.
 */
void
start()
{
    extern void Main( int argc, char ** argv );
    extern void processorStart();

    Main( 0, 0 );

    processorStart();

    Assert( NOTREACHED );
}

void
processorStart()
{
    thisCPU()->_currentProcess = thisCPU()->_idleProcess;
    RunningProcess[ PerMachineProcessorID() ] = thisCPU()->currentProcess();
    thisCPU()->currentProcess()->context()->restoreFromInitialContext( 0 );
  
    Assert( NOTREACHED );
}

#include "ClassInstaller.h"
DEFINECLASS("i586SystemContext",i586SystemContext,i586GenericContext,0,0);
DEFINECLASS("i586UninterruptableSystemContext",i586UninterruptableSystemContext,i586SystemContext,0,0);
DEFINECLASS("i586GenericContext",i586GenericContext,ProcessorContext,0,0);
DEFINECLASS("i586ApplicationContext",i586ApplicationContext,i586GenericContext,0,0);
