/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: David W. Dykstra (dykstra@cs.uiuc.edu)
 * Contributing Author: Vincent F. Russo (russo@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * i586PageFaultInterrupt.cc: i586 page fault interrupt class.
 */
#include "Debug.h"
#include "Assert.h"
#include "OutputStream.h"
#include "ThisProcess.h"
#include "i586Interrupts.h"
#include "i586Context.h"
#include "FaultErrorCode.h"
#include "AccessType.h"
#include "Domain.h"
#include "AddressTranslator.h"
#include "AddressTranslation.h"
#include "FreeRunningTimer.h"
#include "FreeRunningTimerStar.h"

#ifdef INSTRUMENT_INTERRUPT
void InterruptConstructInstruments () {}
#endif

#ifdef INSTRUMENT_SVCTRAP
void SVCTrapConstructInstruments () {}
#endif

extern "C" const char * GetCR2();


void
i586GeneralProtectionInterrupt::basicRaise( char * isp )
{
	/* 
	 * I can't figure out why, but this happens sometimes in normal
	 *   operation.  Simply returning solves the problem.
	 */
	static int count = 0;
	static unsigned int lasttime = 0;
	static int smallelapsedcount = 0;
	stackFrame * frame = (stackFrame *) isp;
	if ((++count % 1000) == 0) {
		int elapsedtime = SystemTimer->time() - lasttime;
		Console() <<
		    "General Protection Interrupt # " << count << ", errno " <<
			asHex(frame->errno) << ", pc " << asHex(frame->eip) <<
			    ", time " << elapsedtime << "\n" << eor;
		lasttime = SystemTimer->time();
		if ( elapsedtime < 100000 ) {
			if ( ++smallelapsedcount > 2 ){
			    Console().formattedWrite(" too fast general protection interrupts\n");
			    Assert( NOTREACHED );
			    }
		}
		else
			smallelapsedcount = 0;
	}
}

void
i586PageFaultInterrupt::basicRaise( char * isp )
{
	/*
	 * Find the needed info back on the interrupt stack.
	 */
	stackFrame * frame = (stackFrame *) isp;
	const char * address = GetCR2();

	Debug << "PageFault: process:" << thisProcess() <<
	  "addr: " << asAddress(address) << " errno: " <<
	  frame->errno << " pc: " << frame->eip << "\n" << eor;

/*	Debug.printf( "PageFault: process:%N addr:%A errno:%A pc:%A\n",
 *		thisProcess(), address, frame->errno, frame->eip );
 */
	Process * process = thisProcess();
	Assert( process != 0 );
	Domain * domain = process->domain();
	Assert( domain != 0 );

	/*
	 * The low-order bit of the errno could determine an
	 *   AddressTranslationFailureCode of NonResidentMemory vs.
	 *   ProtectionViolation, but we don't need to know that here;
	 *   domain->repairFault() recalculates that based on the address.
	 */

	AccessType attemptedAccess;

	switch (frame->errno & 0x6)
	{
	case 0:
		attemptedAccess = SystemRead;
		break;
	case 2:
		attemptedAccess = SystemWrite;
		break;
	case 4:
		attemptedAccess = ApplicationRead;
		break;
	case 6:
		attemptedAccess = ApplicationWrite;
		break;
	}


	/*
	 * If it wasnt a simple onboard MMU cache miss, only the Domain
	 * knows what to do next.
	 */
	FaultErrorCode reason = domain->repairFault( address, attemptedAccess );
	switch( reason ) {
		case AccessOfUnmanagedMemory:
			Console() << "No manager for address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->eip )
			   << " at time of access)\n" << eor;
			Assert( NOTREACHED );

		case ReadOfProtectedSystemMemory:
			Console()
			   << "Attempted read of system memory at address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->eip )
			   << " at time of access)\n" << eor;
			Assert( NOTREACHED );

		case WriteOfProtectedSystemMemory:
			Console()
			   << "Attempted write of system memory at address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->eip )
			   << " at time of access)\n" << eor;
			Assert( NOTREACHED );

		case WriteOfReadOnlyMemory:
			Console()
			   << "Attempted write of read-only memory at address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->eip )
			   << " at time of access)\n" << eor;
			Assert( NOTREACHED );

		case DynamicWriteOfReadOnlyMemory:
		case DynamicAccessViolation:
			Console()
			   << "Access violation referencing memory at address "
			   << asAddress( address ) << " (PC was "
			   << asAddress( (void *) frame->eip )
			   << " at time of access)\n" << eor;
			Assert( NOTREACHED );

		/*
		 * These are all recoverable and the instruction can
		 * be restarted/resumed.
		 */
		case MissingAddressTranslationMapping:
		case MissingPhysicalMemory:
		case InsufficientProtectionLevel:
			break;
	}

	/*
	 * All done so return. This results in retrying/restarting the
	 * instruction which caused the fault.
	 */
	Debug << this << "::basicRaise: fault repaired - retrying\n" << eor;
}


void
StrayInterrupt::basicRaise( char * )
{
	Console().panic( "Stray Interrupt\n" );
	Assert( NOTREACHED );
}



#include "ClassInstaller.h"
DEFINECLASS("StrayInterrupt",StrayInterrupt,Interrupt,0,0);
