#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * DeadlineSchedulerUSS.c: A deadline scheduler
 */

#include "Assert.h"
#include "CPU.h"
#include "FreeRunningTimer.h"
#include "OutputStream.h"
#include "Process.h"
#include "SchedulerUSS.h"


#ifdef __GNUC__
#pragma implementation
#endif


/* 
 * Contains per-process scheduler info.
 */
DeadlineSchedulerInfo::DeadlineSchedulerInfo( Process * process ) : ( process )
{
	myTimeLeft = process->getServiceTime();
	mustStartTime = process->getDeadline() - myTimeLeft;
	if( mustStartTime < 0 ) mustStartTime = 0;
	Debug << this << "::DeadlineSchedulerInfo( " << process
		<< " ), deadline = " << process->getDeadline()
		<< ", service = " << myTimeLeft << "\n" << eor;
}

/*
 * Priority is the latest time the process must start.  
 */
void
DeadlineSchedulerInfo::setStartTime()
{
	mustStartTime = getProcess()->getDeadline() - myTimeLeft;
	if( mustStartTime < 0 ) mustStartTime = 0;
}


/* 
 * The deadline scheduler will have a list of ready processes ordered
 * according to which must finish soonest.  Each process may run for its
 * timeslice and then it will go through the scheduler so the deadlines
 * may be recomputed.  
 */
DeadlineSchedulerUSS::DeadlineSchedulerUSS( int thisQuantum, char * name ) 
{
	Debug << this << "( " << thisQuantum << ", " << name << " )\n" << eor;
	readyList = (SchedList *) new SchedPriorityList( name );
	quantum = thisQuantum;
}

/* 
 * Setup the process fields so the priority may be correctly calculated.
 * If the process has been running then the total service time required
 * must be updated.
 */
void
DeadlineSchedulerUSS::priority( Process * process )
{
	Debug << this << "::priority( " << process << " )\n" << eor;
	Assert( process != 0 );
	DeadlineSchedulerInfo * myInfo = 
		(DeadlineSchedulerInfo *)process->schedulerInfo();

	/* 
	 * This is a new process in the system so create a new
 	 * schedulerInfo class.
	 */
	if( myInfo == 0 ) {
	    myInfo = new DeadlineSchedulerInfo( process );
	    process->setSchedulerInfo( myInfo );
	}

	/* 
	 * Otherwise the process has been running so update the
	 * service time left so the priority can be calculated.
	 */
	else {
	    int runTime = quantum - process->getResidual();
	    myInfo->setTimeLeft( myInfo->getTimeLeft() - runTime );
	    if( myInfo->getTimeLeft() <= 0 ) {
	    	myInfo->setTimeLeft( 0 );
		Console() << this << ": bad time estimate for " << process
			<< "\n" << eor;
	    }
	    myInfo->setStartTime();
	}

	/* check if deadline has already passed */
	if( myInfo->getStartTime() < SystemTimer->time() ) 
		Console() << this << ": " << process << " missed = "
			<< process->getDeadline() << ", now = "
			<< SystemTimer->time() << ", start = "
			<< myInfo->getStartTime() << "\n" << eor;

	/* 
	 * Set the priority of the process.  The priority is the
	 * latest time at which the process must start.  It is negated
	 * because the best priority is the highest value.
	 */
	myInfo->setDoublePriority( (double)(-myInfo->getStartTime()) );
}
