#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Gary M. Johnston (johnston@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/*
 * FIFOScheduler.c: a toy dispatch queue manager.
 */

#include "Assert.h"
#include "FIFOScheduler.h"
#include "OutputStream.h"
#include "Process.h"
#include "Semaphore.h"


#ifdef __GNUC__
#pragma implementation
#endif

FIFOScheduler::~FIFOScheduler()
{
	Assert( this != 0 );
	Debug << this << "::~FIFOScheduler()\n" << eor;
}

NonLockedFIFOScheduler::~NonLockedFIFOScheduler()
{
	Assert( this != 0 );
	Debug << this <<  "%T::~NonLockedFIFOScheduler()\n" << eor;
}

/*
 * Add an element to the FIFO queue.
 */
void
FIFOScheduler::add( Process * process )
{
	Debug << this << "::add( " << process << " )\n" << eor;
	Assert( process != 0 );
	Assert( this != 0 );
	Assert( ! lock.heldByMe() );
	lock.acquire();

	process->_next = 0;
	process->_last = tail;
	if( head == 0 ) {
		Assert( tail == 0 );
		head = process;
	}
	else {
		Assert( tail != 0 );
		tail->_next = process;
	}
	tail = process;

	lock.release();
	// Debug << this << "::add: return\n" << eor;
}

void
NonLockedFIFOScheduler::add( Process * process )
{
	Debug << this << "::add( " << process << " )\n" << eor;
	Assert( process != 0 );
	Assert( this != 0 );

	process->_next = 0;
	process->_last = tail;
	if( head == 0 ) {
		Assert( tail == 0 );
		head = process;
	}
	else {
		Assert( tail != 0 );
		tail->_next = process;
	}
	tail = process;

	// Debug << this << "::add: return\n" << eor;
}
/*
 * Remove the first entry from the FIFO queue.
 */
Process *
FIFOScheduler::remove()
{
	// Debug << this << "::remove()\n" << eor;
	Assert( this != 0 );

	/*
	 * To save bus contention by repeatedly acquiring and releasing the
	 * lock, make a good guess if there is something to deque before
	 * actually trying to acquire the lock. If it looks empty return 0.
	 * Its up to the caller to keep trying.
	 */
	if( head == 0 ) {
		Debug << this << "::remove: looks empty\n" << eor;
		return( 0 );
	}

	/*
	 * Lock the queue if we think something is there. (there might not
	 * actually be anything there by the time we get the lock.
	 */
	Assert( ! lock.heldByMe() );
	lock.acquire();

	/*
	 * Double check to make sure someone else didn't get what we thought
	 * was there while we were acquiring the lock.
	 */
	if( head == 0 ) {
		Assert( tail == 0 );
		lock.release();
		Debug << this << "::remove: lost race\n" << eor;
		return( 0 );
	}

	/*
	 * Remove the next element.
	 */
	Process * val = head;
	head = head->_next;
	if( head == 0 ) {
		tail = 0;
	}
	else {
		head->_last = 0;
	}
	Assert( val != 0 );

	lock.release();

	Debug << this << "::remove: " << val << "\n" << eor;
	return( val );
}

Process *
NonLockedFIFOScheduler::remove()
{
	// Debug << this << "::remove()\n" << eor;
	Assert( this != 0 );

	/*
	 * If there is nothing there, return.
	 */
	if( head == 0 ) {
		Assert( tail == 0 );
		Debug << this << "::remove: looks empty\n" << eor;
		return( 0 );
	}

	/*
	 * Remove the next element.
	 */
	Process * val = head;
	head = head->_next;
	if( head == 0 ) {
		tail = 0;
	}
	else {
		head->_last = 0;
	}
	Assert( val != 0 );

	Debug << this << "::remove: " << val << "\n" << eor;
	return( val );
}

int
FIFOScheduler::isEmpty()
{
	return( head == 0 );
}

int
NonLockedFIFOScheduler::isEmpty()
{
	return( head == 0 );
}
