#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * MLFSchedulerUSS.c: a multilevel feedback queue scheduler.
 */

#include "Assert.h"
#include "OutputStream.h"
#include "Process.h"
#include "SchedulerUSS.h"


#ifdef __GNUC__
#pragma implementation
#endif


/*
 * The quantum will get longer for each lower level in the queue.
 */
int
MLFSchedulerUSS::calcQuantum( int thisLevel ) 
{
	return( (levels - (thisLevel - 1)) * myQuantum );
}

/*
 * Initialize the multilevel feedback queue.  
 */
MLFSchedulerUSS::MLFSchedulerUSS( int thisQuantum, int thisLevels, char * name )
{
	Debug << this << "::MLFScheduler( " << thisQuantum << " quantum, "
		<< thisLevels << " levels, " << name << " )\n" << eor;
	myQuantum = thisQuantum;
	levels = thisLevels;
	readyList = (SchedList *) new SchedFeedbackQueues( thisLevels, name );
}

/*
 * The priority of the process is its queue level.  A process starts
 * off at the top level.  It moves down one level each time it
 * completes its quantum and it moves up otherwise.
 */
void
MLFSchedulerUSS::priority( Process * process )
{
	Assert( process != 0 );
	USSSchedulerInfo * myInfo = 
		(USSSchedulerInfo *)process->schedulerInfo();

	/* queue level will be stored in the scheduler specific field */
	if( myInfo == 0 ) {

		/* put new process in top level queue */
		myInfo = new USSSchedulerInfo( process );
		process->setSchedulerInfo( myInfo ); 
		myInfo->setIntPriority( levels );
	}

	else {
		/* if quantum expired then move process down one level */
		int myPriority = myInfo->getIntPriority();
		if( process->getResidual() == 0 ) {
			if( myPriority != 1 ) 
				myInfo->setIntPriority( myPriority-1 );
		}

		/* it blocked so move it up if possible */
		else if( myPriority != levels ) 
			myInfo->setIntPriority( myPriority+1 );
	}
}

/* 
 * Get the next process.  This will be the process returned from the
 * highest level nonempty SchedQueue.
 */
Process *
MLFSchedulerUSS::remove()
{
	Process * retval = readyList->remove();
	if( retval == 0 ) return( 0 );
	USSSchedulerInfo * myInfo = 
		(USSSchedulerInfo *) retval->schedulerInfo();
	Assert( myInfo != 0 );
	Debug << this << "::remove, proc = " << retval << ", level = "
		<< myInfo->getIntPriority() << ", quantum = "
		<< calcQuantum( myInfo->getIntPriority()) << "\n" << eor;

	/* 
	 * Set the quantum according to the level of the process so
	 * the global quantum is not used.
	 */
	retval->setQuantum( calcQuantum( myInfo->getIntPriority() ) );

	/* (Preemptive Schedulers) Add the process to the run list */
	if( runList != 0 ) runList->add( retval );

	return( retval );
}
