#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * SRTSchedulerUSS.c: A shortest remaining time scheduler.  It is preemptive.
 */

#include "Assert.h"
#include "CPU.h"
#include "OutputStream.h"
#include "Process.h"
#include "SchedulerUSS.h"


#ifdef __GNUC__
#pragma implementation
#endif


/* 
 * Contains per-process scheduler info.
 */
SRTSchedulerInfo::SRTSchedulerInfo( Process * process, SchedPriorityList * myList ) 
: ( process )
{
	Debug << this << "::SRTSchedulerInfo( " << process << ", "
		<< myList << " )\n" << eor;
	timeLeft = process->getServiceTime();
	Assert( timeLeft < 0xffffffff );
	startWaiting = 0;
	runList = myList;
}

SRTSchedulerInfo::~SRTSchedulerInfo()
{
	/* remove the job from the run list */
	runList->premove( getProcess() );
}

/*
 * Return the priority of the process.  It is negated since larger values
 * are higher priorities.
 */
double
SRTSchedulerInfo::getDoublePriority( unsigned int now )
{
	/* Just need the priority for ready queue ordering */
	if( now == 0 )
		return( -doublePriority );

	/* The process has been running longer than it was supposed to. */
	else if( doublePriority < (now - startWaiting) )
		return( 0 ); 

	/* 
	 * The priority of a running process depends on how long it has
         * been running.
	 */
	else
		return( -(doublePriority - (now - startWaiting)) );
}


/* 
 * The SRT scheduler has two lists, a ready list of processes waiting
 * to run and a run list of processes that are currently running or 
 * blocked (not ready).  The lists are priority ordered according to 
 * the remaining time they need to run.
 */
SRTSchedulerUSS::SRTSchedulerUSS( char * name )
{
	Debug << this << "::SRTSchedulerUSS( " << name << " )\n" << eor;

	/* Create priority list for ready processes */
	readyList = (SchedList *) new SchedPriorityList( name );

	/* Create priority list for running processes */
	runList = (SchedList *) new SchedPriorityList( name );
}

/* 
 * The priority of the process is how long it needs to run.  If 
 * it has been running then the priority is updated by that
 * amount of time.
 */
void
SRTSchedulerUSS::priority( Process * process )
{
	Debug << this << "::priority( " << process << " )\n" << eor;
	Assert( process != 0 );

	/* this job is new to the system */
	SRTSchedulerInfo * myInfo = 
		(SRTSchedulerInfo *)process->schedulerInfo();
	if( myInfo == 0 ) {
		myInfo = new SRTSchedulerInfo( process, (SchedPriorityList *)runList );
		process->setSchedulerInfo( myInfo );
	}

	/* remove the job from the run list */
	else {

		/* remove this process from the run list */
		SchedPriorityList * temp = (SchedPriorityList *) runList;
		temp->premove( process );

		/* 
		 * Update the remaining time priority.  It may be better
		 * to set the timeslice quantum to its max value and
 	 	 * then use the residual value since the run time would
		 * be more accurate.
		 */
		unsigned int now = SystemTimer->time();
		if( (now - myInfo->getStart()) > myInfo->getTimeLeft() )

			/*
			 * If the remaining time was incorrectly estimated
			 * then give this job the worst priority.  
			 */
			myInfo->setTimeLeft( 0xffffffff );

		else if( myInfo->getTimeLeft() != 0xffffffff )
			myInfo->setTimeLeft( myInfo->getTimeLeft() -
				(now - myInfo->getStart()) );
	}

	/* The SchedPriorityList uses the double priority value */
	myInfo->setDoublePriority( (double)myInfo->getTimeLeft() );
}

/* 
 * Remove the first entry from the priority ordered ready to run list and
 * add to the appropriate position in the run list
 */
Process *
SRTSchedulerUSS::remove()
{
	Process * retval = SchedulerUSS::remove();
	if( retval == 0 ) return( 0 );

	/* set the start time */
	SRTSchedulerInfo * myInfo = 
		(SRTSchedulerInfo *) retval->schedulerInfo();
	Assert( myInfo != 0 );
	myInfo->setStart();

	Debug << this << "::remove: " << retval << ", time left "
		<< myInfo->getTimeLeft() << "\n" << eor;
	return( retval );
}
