#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * SchedList data structures used by the USS schedulers.  'SchedList' defines
 * the abstract base class used.  The derived concrete subclasses
 * are SchedQueue, SchedStack, SchedFeedbackQueues, SchedSearchList, and SchedPriorityList.
 * All subclasses use the 'schedulerInfo' class of the Process to
 * set links and determine priority.  This information is setup
 * by the USSScheduler::priority() method.
 */

#include "Assert.h"
#include "FreeRunningTimer.h"
#include "SchedList.h"
#include "OutputStream.h"
#include "SchedulerUSS.h"


#ifdef __GNUC__
#pragma implementation
#endif


/*
 * add() - Adds the process to the list data structure according to
 * the specific algorithm.  The schedulerInfo class is used to determine
 * the process priority and thus its placement in the list so the
 * schedulerInfo class must be defined at this point.  This is done in
 * the scheduler priority() method.  This method is an abstract definition
 * and must be redefined by the subclass.
 */

/*
 * remove() - Removes the best priority process from the list.  Depending
 * on the specific algorithm it may only need to remove the first process
 * or it may need to search the list.  This method is an abstract definition
 * and must be redefined by the subclass.
 */

/*
 * isEmpty() - Returns whether the list is empty.  This method is an abstract
 * definition and must be redefined by the subclass.
 */


/*
 * SchedQueue inserts processes on one end and removes them from the other.  The
 * arrival time determines the process priority so it is implicit.  
 * Constructor initializes the queue to empty.  SchedQueue is used by 
 * FIFO, RR schedulers as well as SchedStack and SchedFeedbackQueues list data 
 * structures.
 */
SchedQueue::SchedQueue( char * name ) : lock( name )
{
	Debug << this << "::SchedQueue( " << name << " )\n" << eor;
	head = 0;
	tail = 0;
}

/*
 * Add the process to the the front of the queue.  The schedulerInfo
 * class must be defined (by the scheduler priority() method) at this point.
 */
void
SchedQueue::add( Process * process )
{
	Debug << this << "::add( " << process << " )\n" << eor;
	USSSchedulerInfo * myInfo = 
		(USSSchedulerInfo *)process->schedulerInfo();
	Assert( myInfo != 0 );

	/* Easy if this is the only process in the queue */
	lock.acquire();
	if( head == 0 ) {
		Assert( tail == 0 );
		head = tail = myInfo;
		lock.release();
		return;
	}

	/* This isn't the only process in the queue so add it to the front */
	Assert( tail != 0 );
	myInfo->setNext( head );
	head->setPrev( myInfo );
	head = myInfo;
	lock.release();
}

/*
 * Remove the process from the end of the queue and return it.
 */
Process *
SchedQueue::remove()
{
	Debug << this << "::remove()\n" << eor;

	/* If the queue is empty then return 0 */
	if( head == 0 ) return( 0 );
	lock.acquire();
	if( head == 0 ) {
		Assert( tail == 0 );
		lock.release();
		return( 0 );
	}

	/* Remove the process at the end of the queue */
	USSSchedulerInfo * myInfo = tail;
	if( head == tail )
		head = tail = 0;
	else {
		tail = tail->getPrev();
		tail->setNext( 0 );
		myInfo->setPrev( 0 );
	}
	lock.release();
	Debug << this << "::remove() " << myInfo->getProcess() << "\n" << eor;
	return( myInfo->getProcess() );
}

int
SchedQueue::isEmpty()
{
	return( head == 0 );
}


/*
 * SchedStack is the same as SchedQueue except that processes are added to the
 * same end of the list from which they are removed.  Either add() or
 * remove() could have been redefined to implement this. SchedStack is used
 * by the LIFO schedulers.
 */
SchedStack::SchedStack( char * name ) : ( name )
{
	Debug << this << "::SchedStack( " << name << " )\n" << eor;
}

/*
 * Add the process to the the back of the list.  The schedulerInfo
 * class must be defined (by the scheduler priority() method) at this point.
 */
void
SchedStack::add( Process * process )
{
	Debug << this << "::add( " << process << " )\n" << eor;
	USSSchedulerInfo * myInfo = 
		(USSSchedulerInfo *)process->schedulerInfo();
	Assert( myInfo != 0 );

	/* Easy of this is the only element in the stack */
	lock.acquire();
	if( head == 0 ) {
		Assert( tail == 0 );
		head = tail = myInfo;
		lock.release();
		return;
	}

	/* Add the process to the end of the list */
	Assert( tail != 0 );
	myInfo->setPrev( tail );
	tail->setNext( myInfo );
	tail = myInfo;
	lock.release();
}


/*
 * SchedFeedbackQueues is a list of Queues, one for each indicated level.
 * It is used by the MLF and VMS schedulers.
 */
SchedFeedbackQueues::SchedFeedbackQueues( int levels, char * name )
{
	Debug << this << "::SchedFeedbackQueues( " << levels << " levels, "
		<< name << " )\n" << eor;
	numLevels = levels;
	Queues = new SchedQueue*[levels];
	for( int i = 0; i < levels; i++ )
		Queues[i] = new SchedQueue( name );
}

SchedFeedbackQueues::~SchedFeedbackQueues()
{
	Debug << this << "::~SchedFeedbackQueues()\n" << eor;
	for( int i = 0; i < numLevels; i++ )
		delete Queues[i];
	delete Queues;
}

/*
 * Determine the level (priority of the process) to put this process and
 * then add it to that queue.  The schedulerInfo class is used to read
 * the priority so it must have been defined at this point.  Priorities
 * range from 1...numLevels which corresponds to 0...numLevel-1 Queues.
 */
void
SchedFeedbackQueues::add( Process * process )
{
	Debug << this << "::add( " << process << " )\n" << eor;
	Assert( process != 0 );

	/* Determine the process priority */
	USSSchedulerInfo * myInfo = 
		(USSSchedulerInfo *) process->schedulerInfo();
	Assert( myInfo != 0 );
	int myLevel = myInfo->getIntPriority();

	/* Add the process to the appropriate queue */
	Debug << this << "::add to level " << myLevel-1  << "\n" << eor;
	Queues[myLevel-1]->add( process );
}

/*
 * Remove and return the first process in the highest queue.
 */
Process *
SchedFeedbackQueues::remove()
{
	Debug << this << "::remove()\n" << eor;

	/* Find the highest nonempty queue */
	int empty = 1;
	int thisLevel = numLevels;
	while( (empty) && (thisLevel-- > 0) ) 
	    empty = Queues[thisLevel]->isEmpty();

	/*
	 * Return the next process in this queue.
	 */
	if( empty ) return( 0 );
	Debug << this << "::remove from level " << thisLevel << "\n" << eor;
	return( Queues[thisLevel]->remove() );
}

/*
 * Empty only if all the underlying queues are empty.
 */
int 
SchedFeedbackQueues::isEmpty()
{
	int empty = 1;
	for( int i=0; i < numLevels; i++ )
		empty = empty & Queues[i]->isEmpty();
	return( empty );
}


/*
 * The SchedSearchList is used by the time-dependent schedulers, HRN and UNIX.
 * It is a singly linked list which always adds to the front and searches
 * the entire list for the highest priority at remove().
 */
SchedSearchList::SchedSearchList( char * name ) : lock( name )
{
	Debug << this << "::SchedSearchList( " << name << " )\n" << eor;
	first = 0;
}

/*
 * Always add to the front of the list.  The schedulerInfo class must
 * be defined since it contains the pointers and priority info.
 */
void
SchedSearchList::add( Process * process )
{
	Debug << this << "::add( " << process << " )" << eor;
	USSSchedulerInfo * myInfo = 
			(USSSchedulerInfo *)process->schedulerInfo();
	Assert( myInfo != 0 );

	lock.acquire();
	myInfo->setNext( first );
	first = myInfo;
	lock.release();
}

/*
 * Remove and return the process with the highest priority.  The 
 * entire list must be searched.  
 */
Process *
SchedSearchList::remove() 
{
	Debug << this << "::remove\n" << eor;

	if( first == 0 ) return( 0 );
	lock.acquire();
	if( first == 0 ) {
		lock.release();
		return( 0 );
	}

	/* find the process with the highest priority */
	USSSchedulerInfo * prev = first;
	USSSchedulerInfo * next = first->getNext();
	USSSchedulerInfo * bestprev = 0;
	USSSchedulerInfo * best = first;
	unsigned int now = SystemTimer->time();
	double bestPriority = best->getDoublePriority( now );
	while( next != 0 ) {
		double nextPriority = next->getDoublePriority( now );
		if( nextPriority >= bestPriority ) {
			bestPriority = nextPriority;
			best = next;
			bestprev = prev;
		}
		prev = next;	
		next = next->getNext();
	}

	/* remove this process from the linked list */
	if( best == first )
		first = best->getNext();
	else
		bestprev->setNext( best->getNext() );

	lock.release();
	return( best->getProcess() );
}

int
SchedSearchList::isEmpty()
{
	return( first == 0 );
}


/*
 * SchedPriorityList is a doubly linked list which places processes in 
 * their priority order at add() time.  Remove() may then get the
 * first priority off the list.  It is used by the Deadline and SRT
 * schedulers.  A priority list is also used to keep track of the
 * running processes in the preemptive schedulers, VMS and SRT.  This
 * structure is used for that as well but required two extra methods.
 * One was needed to remove a specific process from the list and
 * another was needed to search the list for one with a preemptable
 * priority which had not already been preempted.  Since these are
 * special cases they were not included in the base SchedList definition,
 * only here.
 */
SchedPriorityList::SchedPriorityList( char * name ) : lock( name )
{
	Debug << this << "::SchedPriorityList( " << name << " )\n" << eor;
	first = last = 0;
}

/*
 * Add the process to the list in order of priority.
 */
void
SchedPriorityList::add( Process * process )
{
	Debug << this << "::add( " << process << " )\n" << eor;
	Assert( process != 0 );
	USSSchedulerInfo * myInfo = 
	  			(USSSchedulerInfo *)process->schedulerInfo();
	Assert( myInfo != 0 );

	/* Easy if this is the only process in the list */
	lock.acquire();
	if( first == 0 ) {
		Assert( last == 0 );
		first = last = myInfo;
		lock.release();
		return;
	}

	/* Determine position to place this process */
	double myPriority = myInfo->getDoublePriority( 0 );
	USSSchedulerInfo * nextInfo = first;
	USSSchedulerInfo * prevInfo = 0;
	while((nextInfo != 0) 
		&& (myPriority < nextInfo->getDoublePriority( 0 )) ) {
		prevInfo = nextInfo;
		nextInfo = nextInfo->getNext();
	}

	/* Place the process at the end of the list */
	if( nextInfo == 0 ) {
		prevInfo->setNext( myInfo );
		myInfo->setPrev( prevInfo );
		last = myInfo;
	}

	/* Process goes somewhere in the middle */	
	else {
		if( prevInfo == 0 )
			first = myInfo;
		else
			prevInfo->setNext( myInfo );
		myInfo->setPrev( prevInfo );
		myInfo->setNext( nextInfo );
		nextInfo->setPrev( myInfo );
	}
	lock.release();
}

/*
 * Remove the highest priority process from the front of the list.
 */
Process *
SchedPriorityList::remove()
{
	Debug << this << "::remove\n" << eor;

	/* Return if the list is empty */	
	if( first == 0 ) return( 0 );
	lock.acquire();
	if( first == 0 ) {
		Assert( last == 0 );
		lock.release();
		return( 0 );	
	}

	/* Get the first process */
	USSSchedulerInfo * retInfo = first;
	first = first->getNext();
	if( first != 0 )
		first->setNext( 0 );
	else
		last = 0;
	retInfo->setPrev( 0 );
	retInfo->setNext( 0 );
	Process * retval = retInfo->getProcess();
	Assert( retval != 0 );
	Debug << this << "::remove() " << retval << "\n" << eor;
	lock.release();
	return( retval );
}

/*
 * Remove the specified process from the priority list.  This is used
 * by the preemptive schedulers when a process is no longer executing
 * and thus no longer eligible for preemption.
 */
void
SchedPriorityList::premove( Process * process )
{
	Debug << this << "::premove( " << process << " )\n" << eor;
	Assert( first != 0 );
	USSSchedulerInfo * myInfo =
		(USSSchedulerInfo *) process->schedulerInfo();
	Assert( myInfo != 0 );	

	/* This process is at the front of the list */
	lock.acquire();
	if( first == myInfo ) {
		first = myInfo->getNext();
		if( first != 0 )
			first->setPrev( 0 );
		else
			last = 0;
	}

	/* This process is not at the front */
	else {
		myInfo->getPrev()->setNext( myInfo->getNext() );
		if( myInfo->getNext() != 0 )
			myInfo->getNext()->setPrev( myInfo->getPrev() );
		else
			last = myInfo->getPrev();
	}
	myInfo->setPrev( 0 );
	myInfo->setNext( 0 );
	myInfo->setPreempt( 0 );
	lock.release();
	return;
}

/*
 * This process is used by the decision() method of preemptive
 * algorithms to return a process (or 0 if none) that is
 * preemptible.  It finds the process with the lowest priority
 * that is lower than the specified process and has not already
 * been preempted.
 */
Process *
SchedPriorityList::findPreempt( Process * process )
{
	Debug << this << "::findPreempt( " << process << " )\n" << eor;
	Assert( process != 0 );
	USSSchedulerInfo * myInfo = 
		(USSSchedulerInfo *)process->schedulerInfo();
	Assert( myInfo != 0 );

	/* Found process must have worse priority than this */
	unsigned int now = SystemTimer->time();
	double myPriority = myInfo->getDoublePriority( now );

	/* find a lower priority process if possible */
	lock.acquire();
	USSSchedulerInfo * next = last;
	while( next != 0 ) {

		/* 
		 * Consider priority, has it been preempted, and
		 * is it running (may be blocked somewhere) 
		 */
		if( (myPriority > next->getDoublePriority( now )) 
		    && (next->getProcess()->getMyCPU() != 0)
		    && (next->getPreempt() == 0) ) {
			next->setPreempt( 1 );
			lock.release();
			return( next->getProcess() );
		}
	
		next = next->getPrev();
	}
	lock.release();
	return( 0 );
}

int
SchedPriorityList::isEmpty()
{
	return( first == 0 );
}
