#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * Virtual function definitions for SchedulerUSS.  These functions define 
 * the Choices interface and the SchedulerUSS interface.  Since most of
 * these are the same for most schedulers they are generally inherited.  
 * The priority() function is always redefined though and some of these
 * will perform no actions for some schedulers.  
 */

#include "Assert.h"
#include "CPU.h"
#include "SchedList.h"
#include "OutputStream.h"
#include "Process.h"
#include "SchedulerUSS.h"


#ifdef __GNUC__
#pragma implementation
#endif


/*
 * Initialize the scheduler.  'readyList' is a data structure containing
 * the ready processes in the scheduler.  'runList' is a data structure
 * used by preemptive schedulers to contain the processes that have been
 * removed from the scheduler.  'quantum' is the timeslice quantum value
 * and is used by quantum-oriented schedulers with a single quantum value
 * for all processes in the system.
 */
SchedulerUSS::SchedulerUSS()
{
	Debug << this << "::SchedulerUSS\n" << eor;
	readyList = 0;
	runList = 0;		// (Preemptive Schedulers)
	quantum = 0;		// (Quantum-Oriented Schedulers)
}

/*
 * Clean up allocated space.
 */
SchedulerUSS::~SchedulerUSS()
{
	Debug << this << "::~SchedulerUSS()\n" << eor;
	if( readyList != 0) delete readyList;
	if( runList != 0) delete runList;
}

/*
 * priority() -	Used by SchedulerUSS schedulers to setup the schedulerInfo
 *		class for a process so that the priority may be determined
 *		from it using one of the getPriority functions as required
 * 		by the particular list structure used.  It is invoked by
 *		the add() method and must be redefined for all schedulers.
 */
void 
SchedulerUSS::priority( Process * process )
{
	Debug << this << "::priority( " << process << " )\n" << eor;
}

/*
 * decision() -	Used by SchedulerUSS scheduler to preempt another processor.
 *		It is invoked by the add() method and is used only
 *		by preemptive algorithms.  For all other algorithms this
 *		function performs no actions.
 */
void 
SchedulerUSS::decision( Process * process )
{
	Debug << this << "::decision( " << process << " )\n" << eor;
	Assert( runList != 0 );
	Assert( process != 0 );

	/* All preemption is done via a SchedPriorityList runList */
	SchedPriorityList * temp = (SchedPriorityList *) runList;
	Process * preemptProcess = temp->findPreempt( process );
	if( preemptProcess == 0 ) return;

	/* 
	 * We found a preemptible process so preempt it.  If the process
	 * has stopped running since we found it don't worry about it
	 * because that processor will soon be invoking the scheduler
	 * and thus we have achieved the same effect.
	 */
	if( preemptProcess->getMyCPU() != 0 )
		preemptProcess->getMyCPU()->preempt();
}

/*
 * add() - 	All schedulers must have this method.   It is used by Choices
 *	   	to put a process into the scheduler.  It calls priority() to
 *	   	add the process to the ready list and decision() to perform
 *	   	process preemption.  This function is the same for all 
 *		schedulers so it may be inherited throughout.
 */
void 
SchedulerUSS::add( Process * process )
{
	Debug << this << "::add( " << process << " )\n" << eor;
	Assert( process != 0 );

	/* Set the priority of this process. */
	priority( process );

	/* Add this process to the appropriate position in the list. */
	readyList->add( process );

	/* (Preemptive Schedulers) Preempt another processor if necessary. */
	if( runList != 0 ) decision( process );
}

/*
 * remove() - 	All schedulers must have this method.  It is used by Choices
 *	      	to remove a process from the scheduler.  It calls priority()
 *	      	for time-dependent schedulers to remove the best priority
 *	      	process.  Most schedulers can inherit this function or 
 *		invoke from within their own definition.
 */
Process * 
SchedulerUSS::remove()
{
	/* Get the next process to run */
	Process * retval = readyList->remove();
	if( retval == 0 ) return( 0 );	
	Debug << this << "::remove() " << retval << "\n" << eor;

	/* (Preemptive Schedulers) Add the process to the running list */
	if( runList != 0 ) 
		runList->add( retval );

	/* (Quantum-Oriented Schedulers) Set the timeslice quantum */
	retval->setQuantum( quantum );

	/* Clear the residual value */
	retval->setResidual( 0 );

	return( retval );
}

/*
 * isEmpty() - All schedulers must have this method.  It is used by Choices
 *	       to determine if the ready list is empty.  It may be inherited
 *	       by all schedulers since the data structure method will
 *	       provide the appropriate implementation.
 */
int 
SchedulerUSS::isEmpty()
{
	return( readyList->isEmpty() );
}


/*
 * This is the 'schedulerInfo' class base definition used by the
 * USS schedulers.  All of the list data structure operate on these
 * since they contain the list pointers as well as the process priority
 * information set up by the SchedulerUSS::priority() method.
 */
USSSchedulerInfo::USSSchedulerInfo( Process * process )
{
 	Debug << this << "::USSSchedulerInfo( " << process << " )\n" << eor;
	myProcess = process;
	next = prev = 0;
	cpuPreempted = 0;	// Flag used by Preemptive Schedulers
}

USSSchedulerInfo::~USSSchedulerInfo()
{
	Debug << this << "::~USSSchedulerInfo()\n" << eor;
}
