#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * UNIXSchedulerUSS.c: A UNIX-like scheduler.
 */

#include "Assert.h"
#include "CPU.h"
#include "OutputStream.h"
#include "Process.h"
#include "SchedulerUSS.h"


#ifdef __GNUC__
#pragma implementation
#endif


/*
 * doublePriority for this scheduler contains the cpuCount for this process.
 * This number is set to an amount of time the process executed and then
 * is slowly degraded.  It is used as an increment to the base priority.
 */
UNIXSchedulerInfo::UNIXSchedulerInfo( Process * process ) : ( process )
{
	Debug << this << "::UNIXSchedulerInfo( " << process << " )\n" << eor;
	doublePriority = 0;
}

/*
 * Recalculate and return the priority of this process.  If it is a kernel
 * mode priority then don't adjust it.  Otherwise reset the cpu count to
 * half its previous value and the new process priority will be the base
 * priority plus half of the cpu count.
 */
double 
UNIXSchedulerInfo::getDoublePriority( unsigned int )
{
	/* 
	 * Determine the new priority. Some events give a one time
	 * kernel mode priority.
	 */
	int currentPriority;
	switch( getProcess()->getProcessEvent() ) {
		case SWAPPER :
			currentPriority = 0;
			break;
		case DISK_IO :
			currentPriority = -1;
			break;
		case BUFFER :
			currentPriority = -2;
			break;
		case INODE :
			currentPriority = -3;
			break;
		case TTY_INPUT :	
			currentPriority = -4;
			break;
		case TTY_OUTPUT :
			currentPriority = -5;
			break;
		case CHILD_EXIT :
			currentPriority = -6;
			break;

		/* the default case is a user mode priority */
		default :
			doublePriority = doublePriority / 2;
			currentPriority = -(getProcess()->getPriority() 
					  + (int) (doublePriority / 2));
			break;
	}

	/* Clear the process event and return the priority */
	getProcess()->setProcessEvent( 0 );
	return( (double)currentPriority );
}


/*
 * Initialize the UNIX scheduler.  The list is empty. 
 */
UNIXSchedulerUSS::UNIXSchedulerUSS( int thisQuantum, char * name )
{
	Debug << this << "::UNIXSchedulerUSS( " << quantum << " quantum, "
		<< name << " )\n" << eor;
	readyList  = (SchedList *) new SchedSearchList( name );
	quantum = thisQuantum;
}

/*
 * If the process is new the cpuCount is 0.  Otherwise the cpuCount
 * is based on the amount of time the process ran.  This value is
 * used later for actual process priority calculation.
 */
void
UNIXSchedulerUSS::priority( Process * process )
{
	Assert( process != 0 );
	Debug << this << "::priority( " << process << " )\n" << eor;

	UNIXSchedulerInfo * myInfo = 
		(UNIXSchedulerInfo *)process->schedulerInfo();
	int count = 0;
	if( myInfo == 0 ) {
		myInfo = new UNIXSchedulerInfo( process );
		process->setSchedulerInfo( myInfo );
	}

	/*
	 * The cpu count is the number of 1/60 of a second that the process
	 * has been running.
	 */
	else
		count = (quantum - process->getResidual()) * 60 / 1000;

	/* Set the cpuCount */
	myInfo->setDoublePriority( (double) count ); 
}
