#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Author: Douglas E. Leyens (leyens@cs.uiuc.edu)
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */

/*
 * VMSSchedulerUSS.c: a VMS-like scheduler.  It is derived from MLFSchedulerUSS 
 */

#include "Assert.h"
#include "CPU.h"
#include "OutputStream.h"
#include "Process.h"
#include "SchedulerUSS.h"


#ifdef __GNUC__
#pragma implementation
#endif


/*
 * The runList must be kept so that remove() may be invoked on process
 * exit so that this process is not selected for preemption.
 */
VMSSchedulerInfo::VMSSchedulerInfo( Process * process, SchedPriorityList * myList ) 
	: ( process )
{
	runList = myList;
}

/*
 * Since the process is finished, remove it from the running list.
 */
VMSSchedulerInfo::~VMSSchedulerInfo()
{
	runList->premove( myProcess );
}


/*
 * Levels 0-15 have quantum and 16-31 have no quantum.
 */
int VMSSchedulerUSS::calcQuantum( int thisLevel )
{
	if( thisLevel < 16 )
		return( myQuantum );
	else
		return( 0 );
}

/*
 * Besides the 32 level MLF queues create queues of non-ready processes
 * to keep track of possible preemption.
 */
VMSSchedulerUSS::VMSSchedulerUSS( int thisQuantum, char * name ) 
	: ( 0, 32, name )
{
	Debug << this << "::VMSSchedulerUSS( " << thisQuantum << " quantum, "
		<< name << " )\n" << eor;
	myQuantum = thisQuantum;
	runList = new SchedPriorityList( name );
}

/*
 * Adjust the priority and save it for use by the lists.
 */
void
VMSSchedulerUSS::priority( Process * process )
{
	Assert( process != 0 );
	VMSSchedulerInfo * myInfo = 
		(VMSSchedulerInfo *)process->schedulerInfo();

	/* this is a new process */
	if( myInfo == 0 ) {
		myInfo = new VMSSchedulerInfo( process, 
			(SchedPriorityList *) runList );
		process->setSchedulerInfo( myInfo );
		myInfo->setIntPriority( process->getPriority() );
	}

	/* remove process from the possibly running queues */
	else {
		SchedPriorityList * temp = (SchedPriorityList *) runList;
		temp->premove( process );
	}

	/* decrement towards base priority */
	int myPriority = myInfo->getIntPriority();
	if( process->getPriority() < myPriority )
		myInfo->setIntPriority( myPriority - 1 );

	/* determine the priority increment */
	int priorityIncrement = 0;
	switch( process->getProcessEvent() ) {
		case DIRECT_IO :
		case BUFFERED_IO :
		case UPDATE_SECTION :
		case SET_PRIORITY :
			priorityIncrement = 2;
			break;
		case RESOURCE_AVAILABLE :
		case WAKE_PROCESS :
		case DELETE_PROCESS :
		case TIMER_EXPIRATION :
			priorityIncrement = 3;
			break;
		case TTY_OUTPUT :
			priorityIncrement = 4;
			break;
		case TTY_INPUT :
		case PROCESS_CREATION :
			priorityIncrement = 6;
			break;
	}
	process->setProcessEvent( 0 );

	/* 
	 * 1. add the priority increment to the base priority
	 * 2. use the higher of the current priority and the
	 *	result from step 1.
	 * 3. if the result from step 2 is greater than 15
	 *	then use the base priority.
	 * (realtime processes are unaffected, priority>15)
	 */
	int newPriority = process->getPriority() + priorityIncrement;
	myPriority = myInfo->getIntPriority();
	if( newPriority > myPriority )
		myInfo->setIntPriority( newPriority );
	if( myInfo->getIntPriority() > 15 )
		myInfo->setIntPriority( process->getPriority() );

	/* 
	 * Save the priority in two forms since the two data structures
 	 * used require different types.
 	 */
	myInfo->setDoublePriority( (double)myInfo->getIntPriority() );
	Debug << this << "::priority( " << process << " ): add to ready queue "
		<< myInfo << ", " << myInfo->getIntPriority() << "\n" << eor;
}
