#
# omniidl python back-end for generating Choices kernel system
# interface dispatchers.
#
# 19 Apr 2006 by Adam Boot (aboot2@uiuc.edu)
#

"""Generates Choices system interface kernel dispatchers"""

from omniidl import idlast, idltype, idlutil, idlvisitor, output
import sys, string

module_state = 0
interface_state = 0
interface_cnt = 0
operation_cnt = 0
interface_id = ""
interface_list = []
interface_dict = {}
destructor_list = []
ignore_list = ["TheInterface"]

class SwitchVisitor(idlvisitor.AstVisitor, idlvisitor.TypeVisitor):
    def __init__(self, st):
        self.st = st

    def visitAST(self, node):
        self.st.out("// This file is automatically generated")

        for n in node.declarations():
            n.accept(self)

    def visitModule(self, node):
        global interface_cnt

        self.st.out("")

        # Generate class dispatcher switch
        self.st.inc_indent()
        self.st.inc_indent()
        self.st.out("switch(classID) {")
        self.st.inc_indent()

        interface_cnt = 0
        for n in node.definitions():
            n.accept(self)

        self.st.out("default:\n  return INVALIDCLASS;")
        self.st.dec_indent()
        self.st.out("}")
        self.st.dec_indent()
        self.st.dec_indent()

    def visitInterface(self, node):
        global interface_cnt
        global interface_list
        global interface_id
        global interface_dict
        global ignore_list

        interface_list.append(node.identifier())

        # Ignore certain interfaces
        if node.identifier() in ignore_list or (node.pragmas()!=[] and str(node.pragmas()[0]) == "nonproxy"):
            return

        # Generate class dispatcher switch
        interface_cnt += 1
        if interface_cnt > 0:
            self.st.out("case @cnt@:", cnt=interface_cnt)
            self.st.out("  return @id@Dispatcher(&objTable, obj, methodID, inputParam, outputParam);", id=node.identifier())

            interface_id = node.identifier()

            for n in node.contents():
                n.accept(self)

        # Add interface to the dictionary
        interface_dict[node.identifier()] = interface_cnt

    def visitOperation(self, node):
        global interface_id
        global destructor_list

        # Strip interface name and trailing digit from method name
        method_id = node.identifier()
        if method_id.find(interface_id)==0:
            method_id = method_id[len(interface_id):]
        method_id = method_id.rstrip('0123456789')

        if method_id == "destructor":
            destructor_list.append(interface_id)


class DispatcherVisitor(idlvisitor.AstVisitor, idlvisitor.TypeVisitor):
    def __init__(self, st):
        self.st = st

    def visitAST(self, node):
        self.st.out("// This file is automatically generated")

        for n in node.declarations():
            n.accept(self)

    def visitModule(self, node):
        global module_state
        global interface_cnt

    # Generate includes
        module_state = 1
        for n in node.definitions():
            n.accept(self)

        # Generate individual dispatcher functions
        module_state = 2
        interface_cnt = 0
        for n in node.definitions():
            n.accept(self)

    def visitInterface(self, node):
        global module_state
        global operation_cnt
        global interface_state
        global interface_id
        global interface_cnt
        global ignore_list

        if module_state == 1:
            # Generate includes
            for c in node.comments():
                if c.text().strip()[-2:] == ".h":
                    for n in range(1, len(c.text())):
                        if c.text()[-n] == "/":
                            self.st.out("#include \"" + c.text()[len(c.text())-n+1:].strip() + "\"")
                            break
        elif module_state == 2:
            # Ignore certain interfaces
            if node.identifier() in ignore_list or (node.pragmas() != [] and str(node.pragmas()[0]) == "nonproxy"):
                return

            interface_cnt += 1

            if node.pragmas() == []:
                # Generate individual dispatcher functions
                interface_state = 1
                interface_id = node.identifier()
                self.st.out("\ninline AppInterface::EXCODE @id@Dispatcher(HashTable* t, ObjLink* obj, unsigned methodID, void* inputParam, void* outputParam) {", id=node.identifier())
                self.st.inc_indent()
                self.st.out("@id@* realobj = 0;", id=node.identifier())

                # Ouput local variables
                operation_cnt = 1
                for n in node.contents():
                    n.accept(self)

                # Output get and assert object
                self.st.out("\nif(methodID != 0) {")
                self.st.out("  realobj = (@id@*)(obj->getObj());", id=node.identifier())
                self.st.out("  if(!realobj->classOf()->isASubclassOf(@id@Class))", id=node.identifier())
                self.st.out("    return AppInterface::INVALIDCLASS;")
                self.st.out("}")

                # Generate method switch
                interface_state = 2
                interface_id = node.identifier()
                self.st.out("\nswitch(methodID) {")

                operation_cnt = 1
                self.st.inc_indent()
                for n in node.contents():
                    n.accept(self)
                self.st.dec_indent()

                self.st.out("}")
                self.st.out("\nreturn AppInterface::INVALIDMETHOD;")

                self.st.dec_indent()
                self.st.out("}")
            else:
                self.st.out("\n// Ignoring interface '@id@'", id=node.identifier())

    def visitOperation(self, node):
        global operation_cnt
        global interface_state
        global interface_id
        global interface_cnt
        global interface_dict
        global destructor_list

        # Strip interface name and trailing digit from method name
        method_id = node.identifier()
        if method_id.find(interface_id) == 0:
            method_id = method_id[len(interface_id):]
        method_id = method_id.rstrip('0123456789')

        # Set method number
        if method_id == "constructor":
            count = 0
        elif method_id == "destructor":
            count = 1
        else:
            operation_cnt += 1
            count = operation_cnt

        if interface_state == 1:
            # Generate local parameter variables
            param_cnt = 0
            for p in node.parameters():
                param_cnt = param_cnt+1
                p.paramType().accept(self)
                ptype = self.__result_type
                dtype = self.__declared_type
                if p.is_out():
                    typemod = "*"
                    if p.is_in():
                        dtype = "" # Ignore making a declared type a pointer
                else:
                    typemod = ""
                self.st.out("@type@@m@@d@ param@cnt@_@pcnt@;", type=ptype, m=typemod, d=dtype, cnt=count, pcnt=param_cnt)

            # Generate local return variables
            node.returnType().accept(self)
            rtype = self.__result_type
            dtype = self.__declared_type

            if rtype != "void":
                self.st.out("@type@@d@ ret@cnt@;", type=rtype, d=dtype, cnt=count)
        elif interface_state == 2:
            # Unmarshall input parameters
            param_cnt = 0
            param_offset = ""
            param_list = ""
            unmarshall = ""
            for p in node.parameters():
                param_cnt = param_cnt+1
                p.paramType().accept(self)
                ptype = self.__result_type
                dtype = self.__declared_type
                iotype = ""
                if p.is_out():
                    typemod = "*"
                    if p.is_in():
                        iotype = "*"
                        dtype = "" # Ignore making a declared type a pointer
                else:
                    typemod = ""

                if param_cnt > 1:
                    param_offset += "+sizeof(param" + str(count) + "_" + str(param_cnt-1) + ")"
                unmarshall += "  param" + str(count) + "_" + str(param_cnt) + " = *((" + ptype + typemod + dtype + "*)((char*)inputParam" + param_offset + "));\n"

                if param_cnt == 1:
                    param_list = str(iotype) + "param" + str(count) + "_" + str(param_cnt)
                else:
                    param_list += ", " + str(iotype) + "param" + str(count) + "_" + str(param_cnt)

            # Generate methods
            if method_id == "constructor":
                # Generate constructor
                self.st.out("case 0: //@iid@()", iid=interface_id)
                self.st.out(unmarshall)
                if interface_id in destructor_list:
                    self.st.out("""  realobj = new @iid@(@param@);\
\n  t->insert(new ObjLink((void*)realobj, @cnt@, thisProcess()));\
\n  *((@iid@**)outputParam) = realobj;\
\n  return AppInterface::DONE;""", iid=interface_id, param=param_list, cnt=interface_cnt)
                else:
                    self.st.out("""  realobj = new @iid@(@param@);\
\n  t->insert(new ObjLink((void*)realobj, @cnt@, 0, 0));\
\n  *((@iid@**)outputParam) = realobj;\
\n  return AppInterface::DONE;""", iid=interface_id, param=param_list, cnt=interface_cnt)
            elif method_id == "destructor":
                # Generate destructor
                self.st.out("""case 1: //~@iid@()\
\n  if(obj->refs()>0 && obj->redRef(thisProcess())==0) {\
\n    delete realobj;\
\n    t->remove(obj);\
\n    std::multimap<Process*,ObjLink*>::iterator iter;\
\n    iter = TheProcessObjMap.find(thisProcess());\
\n    while(iter != TheProcessObjMap.end()) {\
\n      Assert(iter->first == thisProcess());\
\n      if(iter->second->getObj() == realobj)\
\n        TheProcessObjMap.erase(iter);\
\n      iter++;\
\n    }\
\n  }\
\n  return AppInterface::DONE;""", iid=interface_id)
            else:
                # Generate other methods
                self.st.out("case @ocnt@: //@mid@()", ocnt=count, mid=method_id)

                # Output unmarshalling block
                self.st.out(unmarshall)

                # Make actual call
                ret_list = "  "
                node.returnType().accept(self)
                rtype = self.__result_type
                dtype = self.__declared_type
                if rtype != "void":
                    ret_list = ret_list + "ret" + str(count) + " = "
                self.st.out("@ret@realobj->@mid@(@param@);", ret=ret_list, mid=method_id, param=param_list)

                # If return type is a declared type then insert new ObjLink
                if self.__declared_type == "*":
                    if rtype in interface_dict:
                        self.st.out("  t->insert(new ObjLink((void*)ret@ocnt@, @icnt@, 0, 0));", ocnt=count, icnt=interface_dict[rtype])
                    else:
                        self.st.out("  //Error: return class '@iid@' is unknown!", iid=rtype)

                # Marshall output parameter
                if rtype != "void":
                    self.st.out("  *((@type@@d@*)outputParam) = ret@ocnt@;", type=rtype, d=dtype, ocnt=count)

                self.st.out("  return AppInterface::DONE;")

    ttsMap = {
        idltype.tk_void:       "void",
        idltype.tk_short:      "short",
        idltype.tk_long:       "long",
        idltype.tk_ushort:     "unsigned short",
        idltype.tk_ulong:      "unsigned long",
        idltype.tk_float:      "float",
        idltype.tk_double:     "double",
        idltype.tk_boolean:    "boolean",
        idltype.tk_char:       "char",
        idltype.tk_octet:      "octet",
        idltype.tk_any:        "void*",
        idltype.tk_TypeCode:   "CORBA::TypeCode",
        idltype.tk_Principal:  "CORBA::Principal",
        idltype.tk_longlong:   "long long",
        idltype.tk_ulonglong:  "unsigned long long",
        idltype.tk_longdouble: "long double",
        idltype.tk_wchar:      "wchar"
    }

    def visitBaseType(self, type):
        self.__declared_type = ""
        self.__result_type = self.ttsMap[type.kind()]

    def visitStringType(self, type):
        self.__declared_type = ""
        self.__result_type = "char*"

    def visitDeclaredType(self, type):
        global interface_list

        self.__declared_type = ""

        if type.name() == "uint":
            self.__result_type = "unsigned int"
        elif type.name() in interface_list:
            self.__declared_type = "*"
            self.__result_type = type.name()
        else:
            self.__result_type = type.name()

class CleanupVisitor(idlvisitor.AstVisitor, idlvisitor.TypeVisitor):
    def __init__(self, st):
        self.st = st

    def visitAST(self, node):
        self.st.out("// This file is automatically generated")

        for n in node.declarations():
            n.accept(self)

    def visitModule(self, node):
        global interface_cnt

        self.st.out("")

        # Generate cleanup switch
        self.st.inc_indent()
        self.st.inc_indent()
        self.st.out("switch(ol->getClassID()) {")
        self.st.inc_indent()

        interface_cnt = 0
        for n in node.definitions():
            n.accept(self)

        self.st.out("default:")
        self.st.out("""  Console() << "ERROR: try to free class " << ol->getClassID() << "\\n" << eor;""")
        self.st.dec_indent()
        self.st.out("}")
        self.st.dec_indent()
        self.st.dec_indent()

    def visitInterface(self, node):
        global interface_cnt
        global ignore_list
        global destructor_list

        if node.identifier() in ignore_list or (node.pragmas()!=[] and str(node.pragmas()[0]) == "nonproxy"):
            return

        interface_cnt += 1

        # Ignore certain interfaces
        if  node.identifier() not in destructor_list:
            return

        # Generate cleanup switch
        if interface_cnt > 0:
            self.st.out("case @cnt@:", cnt=interface_cnt)
            self.st.out("  delete (@id@*)(ol->getObj());", id=node.identifier())
            self.st.out("  break;")

def run(tree, args):
    st1 = output.Stream(open('../../Kernel/AppInterfaceSwitch.cc', 'w'), 2)
    st2 = output.Stream(open('../../Kernel/AppInterfaceDispatcher.cc', 'w'), 2)
    st3 = output.Stream(open('../../Kernel/AppInterfaceCleanup.cc', 'w'), 2)
    st2.out("""
#include "UNIXFile.h"
#include "UnixTime.h"
#include "SystemTime.h"
#include "Domain.h"
#include "Disk.h"
#include "Process.h"
#include "Process.h"
#include "MemoryRange.h"
#include "MemoryObject.h"
#include "PersistentMemoryObject.h"
#include "FileObject.h"
#include "FileObjectView.h"
#include "MemoryObjectCache.h"
#include "Semaphore.h"
#include "PersistentObject.h"
#include "PersistentArray.h"
#include "Timer.h"
#include "Semaphore.h"
#include "ConstructorDescriptor.h"
#include "TimeoutTimer.h"
#include "PeriodicTimer.h"
#include "MemoryObjectContainer.h"
#include "TimeOfDay.h"
#include "Pipe.h"
#include "ClassConstructor.h"
#include "VMInterface.h"
#include "Class.h"
#include "PersistentClass.h"
#include "SystemInterface.h"
#include "FileSystemInterface.h"
#include "MemoryObjectDictionary.h"
#include "WatchdogTimer.h"
#include "CharacterInputStream.h"
#include "SymbolicLink.h"
#include "TimingInterface.h"
#include "NameServer.h"
#include "RecordStream.h"
#include "FreeRunningTimer.h"
#include "RaidOutputStream.h"
""")
    dv1 = SwitchVisitor(st1)
    dv2 = DispatcherVisitor(st2)
    dv3 = CleanupVisitor(st3)
    tree.accept(dv1)
    tree.accept(dv2)
    tree.accept(dv3)
