#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/******************************************************************************
 *
 * declarations.c
 *
 *	This file contains the functions used to declare identifiers.
 *
 * Ed Loyot 9/14/90
 *
 *              Copyright (c) 1990, 1991
 *              University of Virginia
 *              All Rights Reserved
 *
 *		This software is the property of the University of Virginia.
 *		Permission is granted to copy and distribute this software as
 *		long as this notice is intact on all copies.  Derived works 
 *  		must retain this notice.  This software is provided `as is'
 *		and without any express or implied warrenties.
 *
 *              Send problems & suggestions to:
 *		Ed Loyot (ecl2v@virginia.edu) or
 *              Andrew Grimshaw (grimshaw@cs.virginia.edu)
 *
 *****************************************************************************/

#include <stream.h>
#include <strstream.h>
#include "String.h"

#include "common.h"
#include "tokens.h"
#include "syntax_tree.h"
#include "parser.tab.h"

extern int column;
extern int linenum;
extern char program_name[];
extern symbol_table global_table;
extern symbol_table_stack scope_stack;
extern symbol_table *parameter_table;
extern void yyerror(char *);

/******************************************************************************
 *
 * general_declaration -
 *
 * 	This function builds the canonical type for the identifier being
 * declared using the base type and the modifiers from the declarator.  The
 * canonical type is entered in the type field for the identifier then the
 * identifier is entered in the symbol table.
 *
 * type_ptr - a pointer to the base type.
 * 
 * declarator_ptr - a pointer to the declarator object for the identifier being
 *		    declared.
 * 
 * access_token - can be PUBLIC, PRIVATE, PROTECTED or NONE.
 * 
 * definition - when a function is being declared this flag indicates whether
 * 		this is a function definition or just a declaration with no
 *		body.
 *
 ******************************************************************************/

void general_declaration(type *type_ptr, declarator *declarator_ptr,
			 TOKEN access_token = NONE, p_boolean definition = FALSE)
{
    identifier *id_ptr = declarator_ptr->declarator_id;

    id_ptr->access_token = access_token;

    /*
     * Duplicate the type, because the canonical type associated with the 
     * identifier may have a different form than the type associated with the 
     * declarator object.
     */
    
    type *declarator_modifiers_ptr;
    if (declarator_ptr->declarator_type != NULL)
	declarator_modifiers_ptr=declarator_ptr->declarator_type->duplicate();
    else
	declarator_modifiers_ptr = NULL;
    
    // Expand any TYPEDEFnames in the type.
    if (type_ptr != NULL)
	type_ptr->expand_typedefs();

    /*
     * Check the declaration specifier to see if this is a typedef.  If so set 
     * the identifier's id_token to indicate this is a typedef name.
     */
    
    if (type_ptr->typedef_type())
	id_ptr->id_token = TYPEDEFname;
    
    /*
     * Set the storage class flags in the identifier.  The function 
     * set_storage_class_flags sets extern, register, auto, static, inline, 
     * virtual, overload and friend flags. 
     */
    
    type_ptr->set_storage_class_flags(id_ptr);

    /*
     * If no storage class has been specified use the default, which is extern
     * if the scope is global and auto otherwise.
     */

    if (id_ptr->storage_class_element->type_element_token == NONE)

	if (scope_stack.stack_size() == 1)
	    
	    // The scope is global(file scope) so use extern.
	    id_ptr->storage_class_element = new storage_class(EXTERN);
	else
	    
	    // The scope is not global so use auto.
	    id_ptr->storage_class_element = new storage_class(AUTO);

    /*
     * Add the base type to the modifiers that were parsed with the declarator 
     * and set the identifier type pointer to point to resulting type.
     */

    if (declarator_modifiers_ptr != NULL)
    {
	declarator_modifiers_ptr->add_type(type_ptr);
	if (id_ptr->id_token == TYPEDEFname)
	{
	    base_type *base_type_ptr = new base_type(TYPEDEFname, id_ptr);
	    base_type_ptr->typedef_type_ptr = declarator_modifiers_ptr;
	    id_ptr->type_ptr = new type(base_type_ptr);
	}
	else
	    id_ptr->type_ptr = declarator_modifiers_ptr;
    }
    else
	if (id_ptr->id_token == TYPEDEFname)
	{
	    base_type *base_type_ptr = new base_type(TYPEDEFname, id_ptr);
	    base_type_ptr->typedef_type_ptr = type_ptr;
	    id_ptr->type_ptr = new type(base_type_ptr);
	}
	else
	    id_ptr->type_ptr = type_ptr;
    
    /*
     * Check the declaration specifier to see if this is a function. If so set 
     * the identifier's id_token to indicate this is a function.
     */
    
    if ((id_ptr->id_token != TYPEDEFname)
			&& (id_ptr->type_ptr->function_declaration()))
	id_ptr->id_token = FUNCTION;
    
    // Once the type is all together set the const and volatile flags.
    id_ptr->type_ptr->set_qualifier_flags();

    /*
     * If a constructor or a destructor is being declared, the name to be
     * entered in the symbol table must be changed otherwise the class name
     * will be hidden by the constructor or destructor with the same name. If
     * this is a constructor prepend a '*' to the class name.  If this is a 
     * destructor prepend a '~'.  Duplicate the identifier object for the same
     * reasons the type was duplicated above.
     */

    identifier *final_id_ptr = id_ptr;
    if (id_ptr->ctor_or_dtor())
    {
	char *ctor_name = new char[strlen(id_ptr->id_string) + 2];
	if (id_ptr->destructor_flag)
	    strcpy(ctor_name, "~");
	else
	    strcpy(ctor_name, "*");
	strcat(ctor_name, id_ptr->id_string);
	final_id_ptr = (identifier *)id_ptr->duplicate();
	final_id_ptr->id_string = ctor_name;
    }

    /*
     * Enter the identifier being declared in the symbol table. If there
     * is already an identifier with this name in the table, then a name is 
     * being illegaly redclared.  Unless this is a foreward defined function.
     * Log the error then remove the original entry, insert to new entry and 
     * continue the parse.
     */

    identifier *temp_id_ptr = scope_stack.insert(final_id_ptr);
    if (temp_id_ptr == NULL)
    {
	/*
	 * There is already an identifier with this name in the symbol table.
	 * Check to see if this is a forward defined function.
	 */

	if (final_id_ptr->type_ptr->function_declaration())
	{
	    /*
	     * The identifier being declared here is a function name with the
	     * same type as the one already in the symbol table.
	     */

	    temp_id_ptr = scope_stack.lookup(final_id_ptr);
	    if (definition)
	    {
		// This is a function definition not just a declaration.
		if (temp_id_ptr->function_definition_ptr != NULL)
		{
		    /*
		     * There is already a body associated with this function
		     * name.  Log an error.
		     */

		    ostrstream error(message_buffer, max_message_size);
		    error << "Identifier ";
		    error << final_id_ptr->id_string;
		    error << " is being redeclared!\n";
		    error.put('\0');
		    yyerror(message_buffer);
		    temp_id_ptr = scope_stack.remove(final_id_ptr);
		    temp_id_ptr = scope_stack.insert(final_id_ptr);
		}
		else
		{
		    /*
		     * This function was forward defined, now that the body
		     * has been seen replace the old declaration with this
		     * new one.
		     */

		    temp_id_ptr = scope_stack.remove(final_id_ptr);
		    temp_id_ptr = scope_stack.insert(final_id_ptr);
		}
	    }
	}
	else
	{
	    if (final_id_ptr->id_token != TYPEDEFname)
	    {
		// The identifier was not a function or typedef,
		//      thus there is an error.
		ostrstream error(message_buffer, max_message_size);
		error << "Identifier " << final_id_ptr->id_string;
		error << " is being redeclared!\n";
		error.put('\0');
		yyerror(message_buffer);
	    }
	    temp_id_ptr = scope_stack.remove(final_id_ptr);
	    temp_id_ptr = scope_stack.insert(final_id_ptr);
	}
    }
} 

/******************************************************************************
 *
 * parameter_declaration -
 *
 * 	This function builds the canonical type for the identifier being
 * declared using the base type and the modifiers from the declarator.  The
 * canonical type is entered in the type field for the identifier then the
 * identifier is entered in the symbol table.
 *
 * type_ptr - a pointer to the base type.
 * 
 * declarator_ptr - a pointer to the declarator object for the identifier being
 *		    declared.
 * 
 ******************************************************************************/

void parameter_declaration(type *type_ptr, declarator *declarator_ptr)
{
    identifier *id_ptr = declarator_ptr->declarator_id;

    if (id_ptr == NULL)
	return;

    /*
     * Duplicate the type, because the canonical type associated with the 
     * identifier may have a different form than the type associated with the 
     * declarator object.
     */
    
    type *declarator_modifiers_ptr;
    if (declarator_ptr->declarator_type != NULL)
	declarator_modifiers_ptr=declarator_ptr->declarator_type->duplicate();
    else
	declarator_modifiers_ptr = NULL;
    
    // Expand any TYPEDEFnames in the type.
    if (type_ptr != NULL)
	type_ptr->expand_typedefs();

    /*
     * Check the declaration specifier to see if this is a typedef.  If so set 
     * the identifier's id_token to indicate this is a typedef name.
     */
    
    if (type_ptr->typedef_type())
	id_ptr->id_token = TYPEDEFname;
    
    /*
     * Set the storage class flags in the identifier.  The function 
     * set_storage_class_flags sets extern, register, auto, static, inline, 
     * virtual, overload and friend flags. 
     */
    
    type_ptr->set_storage_class_flags(id_ptr);

    /*
     * Add the base type to the modifiers that were parsed with the declarator 
     * and set the identifier type pointer to point to resulting type.
     */

    if (declarator_modifiers_ptr != NULL)
    {
	declarator_modifiers_ptr->add_type(type_ptr);
	id_ptr->type_ptr = declarator_modifiers_ptr;
    }
    else
	id_ptr->type_ptr = type_ptr;
    
    // Once the type is all together set the const and volatile flags.
    id_ptr->type_ptr->set_qualifier_flags();

    /*
     * Enter the identifier being declared in the symbol table. If there
     * is already an identifier with this name in the table, then a name is 
     * being illegaly redclared.  Log the error then remove the original entry, 
     * insert to new entry and continue the parse.
     */

    identifier *temp_id_ptr = parameter_table->insert(id_ptr);
    if (temp_id_ptr == NULL)
    {
	 //There is already an identifier with this name in the symbol table.
	ostrstream error(message_buffer, max_message_size);
	error << "Identifier " << id_ptr->id_string;
	error << " is being redeclared!\n";
	error.put('\0');
	yyerror(message_buffer);
	temp_id_ptr = parameter_table->remove(id_ptr);
	temp_id_ptr = parameter_table->insert(id_ptr);
    }
} 

/******************************************************************************
 *
 * get_parameter_type -
 *
 * 	This function builds  and returns the canonical type for the arguments.
 *
 * type_ptr - a pointer to the base type.
 * 
 * declarator_ptr - a pointer to the declarator object for the identifier being
 *		    declared.
 * 
 ******************************************************************************/

type *get_parameter_type(type *type_ptr, declarator *declarator_ptr)
{
    type *parameter_type_ptr = NULL;    
    if (declarator_ptr != NULL && declarator_ptr->declarator_type != NULL)
	parameter_type_ptr = declarator_ptr->declarator_type->duplicate();
    
    // Add the base type to the modifiers that were parsed with the declarator.
    if (parameter_type_ptr != NULL)
	parameter_type_ptr->add_type(type_ptr->duplicate());
    else
	parameter_type_ptr = type_ptr->duplicate();
    
    // Once the type is all together set the const and volatile flags.
    parameter_type_ptr->set_qualifier_flags();

    return parameter_type_ptr;
} 




