#ifdef __GNUC__
#pragma implementation
#endif

/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/******************************************************************************
 *
 * symbol_table.c
 *
 *	This file contains the member functions used by the symbol table. The
 * class definitions for these member functions are contained in 
 * symbol_table.h.
 *
 * Ed Loyot 7/12/90
 *
 *              Copyright (c) 1990, 1991
 *              University of Virginia
 *              All Rights Reserved
 *
 *		This software is the property of the University of Virginia.
 *		Permission is granted to copy and distribute this software as
 *		long as this notice is intact on all copies.  Derived works 
 *  		must retain this notice.  This software is provided `as is'
 *		and without any express or implied warrenties.
 *
 *              Send problems & suggestions to:
 *		Ed Loyot (ecl2v@virginia.edu) or
 *              Andrew Grimshaw (grimshaw@cs.virginia.edu)
 *
 *****************************************************************************/

#ifdef SABER_CC
#include <libc.h>
#endif

#include <stddef.h>
#include <stream.h>
#include "String.h"
#include <stdlib.h>

#include "syntax_tree.h"
#include "symbol_table.h"
#include "parser.tab.h"

symbol_table::symbol_table()
{
    for (int list_number = 0; list_number < HASH_CONSTANT; list_number++)
	array[list_number] = NULL;
#ifdef MPLFRONT
    prv_list = NULL;
#endif
}

int symbol_table::hash(char *hash_string)
{
    /*
     * This algorithm computes the hash number for a given string.
     * For each character in the string, the current value of the hash number 
     * is multiplied by 16 then the character is added to this value.  
     * The final value is forced to be in the desired range by taking the 
     * modulus with respect to the hashing constant.
     */

    unsigned hash_number = 0;
    for(char *temp_string = hash_string; *temp_string != '\0'; temp_string++)
	hash_number = (hash_number * 16) + (*temp_string);
    return (hash_number % HASH_CONSTANT);
}

/*
 * lookup -
 *
 * 	This function is used to lookup identifiers in the symbol table.
 */

identifier *symbol_table::lookup(char *id_string)
{
    // Hash the id_string.
    int hash_number = hash(id_string);

    // Follow the linked list of buckets until a match is found.
    for (bucket *bucket_ptr = array[hash_number]; bucket_ptr != NULL; 
	 bucket_ptr = bucket_ptr->get_next())
    {
	identifier *temp_id = bucket_ptr->get_data();
	char *temp_string = temp_id->id_string;
	if (strcmp(temp_string, id_string) == 0)
	    return temp_id;
    }
    return NULL;
}    

#ifdef MPLFRONT
void symbol_table::build_prv_list()
{
    expression_list *expr_list_ptr = new expression_list(NULL);
    for (int count = 0; count < HASH_CONSTANT; count++)
    {
	if (array[count] == NULL)
	    continue;
	for (bucket *bucket_ptr = array[count]; bucket_ptr != NULL;
	     bucket_ptr = bucket_ptr->get_next())
	    if (bucket_ptr->get_data()->prv_flag)
		expr_list_ptr->add_expression(bucket_ptr->get_data());
    }
    prv_list = expr_list_ptr;
}

expression_list *symbol_table::get_prvs()
{
    if (prv_list == NULL)
	build_prv_list();
    if (prv_list->get_size() != 0)
	return prv_list;
    else
	return NULL;
}
#endif

/*
 * lookup -
 *
 * 	This function is used to lookup identifiers in the symbol table.
 */

identifier *symbol_table::lookup(identifier *id_ptr)
{
    return lookup(id_ptr->id_string);
}    

/*
 * insert -
 *
 * 	If an identifier is not present in the local symbol table insert it.
 */

identifier *symbol_table::insert_identifier(identifier *id_ptr)
{
    /*
     * If the identifier is already in the local table the insert has failed.
     * In the case of functions the return type and number and type of arguments
     * must also be checked.  This is because of function overloading.
     */

    identifier *temp_id_ptr;
    if ((temp_id_ptr = lookup(id_ptr->id_string)) != NULL)
    {
	/*
	 * Check to see if this is an overloaded function.  If the identifier
	 * in the symbol table and the identifier being inserted are both
	 * function names and the number and  type of the arguments are 
	 * different then this is an overloaded function and should be entered 
	 * in the table.
	 */

	if (((temp_id_ptr->type_ptr != NULL) && 
	     temp_id_ptr->type_ptr->function_declaration()) 
	    && 
	    ((id_ptr->type_ptr != NULL) && 
	     id_ptr->type_ptr->function_declaration()))
	{
	    if (id_ptr->type_ptr->compare(temp_id_ptr->type_ptr))
		// The types are the same, this is not an overloaded function.
		return NULL;
	}
	else
	    /*
	     * One of the identifiers is not a function name, this is not an 
	     * overloaded function.
	     */

	    return NULL;
    }

    // Create a bucket object for the new entry.
    bucket *new_bucket = new bucket(id_ptr);

    // Hash the identifier string.
    int hash_number = hash(id_ptr->id_string);

    /*
     * Use the hash value to find the right array slot.  Then check to see if
     * this slot is empty.  If so, make it point to the new bucket.
     * If there is already an entry for this slot, chain the new bucket onto
     * the end of the list.
     */

    bucket *bucket_ptr = array[hash_number];
    if (bucket_ptr == NULL)
	array[hash_number] = new_bucket;
    else
    {
	bucket *temp_bucket_ptr;
	while ((temp_bucket_ptr = bucket_ptr->get_next()) != NULL)
	    bucket_ptr = temp_bucket_ptr;
	bucket_ptr->add_link(new_bucket);
    }
    return id_ptr;
}

identifier *symbol_table::insert(char *id_string, TOKEN id_token)
{
    // Create a new identifier object.
    identifier *id_ptr;
/*    switch(id_token)
    {
      case IDENTIFIER:
      case LABEL:
      case TYPEDEFname:
	id_ptr = new identifier(id_string, id_token);
	break;
      case CLASS:
      case STRUCT:
      case UNION:
      case ENUM:
	id_ptr = new class_identifier(id_string, id_token);
	break;
      case FUNCTION:
	id_ptr = new function_identifier(id_string, id_token);
	break;
      default:
	cerr << Me << ": Internal error!: symbol_table::insert: Unknown ";
	cerr << "token : " << id_token << ": " << Me << " will abort!\n";
	exit(-1);
    }
*/

    id_ptr = new identifier(id_string, id_token);

    // Enter it in the table.  Return NULL if it is already in the table.
    return insert_identifier(id_ptr);
}

identifier *symbol_table::insert(identifier *id_ptr)
{
    // Enter it in the table.  Return NULL if it is already in the table.
    return insert_identifier(id_ptr);
}

identifier *symbol_table::remove(char *id_string)
{
    // Check the table for the identifier.
    identifier *temp_id = lookup(id_string);

    if (temp_id != NULL)
    {
	// Remove the identifier from the table!

        // Hash the identifier string.
	int hash_number = hash(id_string);

	/*
	 * Use the hash value to find the right array slot.  Then check to see
	 * if this slot has only one entry.  If so, simply delete the bucket.
	 * If there is more than one entry for this slot, remove the 
	 * appropriate bucket from the list.
	 */

	bucket *bucket_ptr = array[hash_number];
	if (bucket_ptr->get_data() == temp_id)
	{
	    // This head of the list is the entry to be removed.
	    array[hash_number] = bucket_ptr->get_next();
	    delete bucket_ptr;
	}
	else
	{
	    // Find the right bucket in the chain for this hash number.
	    bucket* temp_bucket_ptr;
	    for (temp_bucket_ptr = bucket_ptr;
		 temp_bucket_ptr->get_next() != NULL && 
		 temp_bucket_ptr->get_next()->get_data() != temp_id;
		 temp_bucket_ptr = temp_bucket_ptr->get_next())
		{
		}	

	    bucket *delete_bucket_ptr = temp_bucket_ptr->get_next();
	    if (delete_bucket_ptr == NULL)
	    {
		// Didnt find the entry! Log an error and abort.
		cerr << "Internal error in symbol_table::remove, " << Me;
		cerr << " will abort!\n";
		exit(-1);
	    }
	    else
	    {
		// Remove the bucket to be deleted from the list and delete it.
		temp_bucket_ptr->add_link(delete_bucket_ptr->get_next());
		delete delete_bucket_ptr;
	    }
	}

	return temp_id;
    }

    // If the identifier is not found indicate the remove failed.
    return NULL;
}

identifier *symbol_table::remove(identifier *id)
{
    return remove(id->id_string);
}


extern ostream *Debug_stream;

void symbol_table::print()
{
    for (int list_number = 0; list_number < HASH_CONSTANT; list_number++)
    {
	*Debug_stream << "List " << list_number << ":\n";
	if (array[list_number] == NULL)
	    *Debug_stream << "\tempty\n";
	else
	{
	    bucket *temp_bucket_ptr = array[list_number];
	    *Debug_stream << "\t" << temp_bucket_ptr->get_data()->id_string << "\n";
	    if (temp_bucket_ptr->get_data()->type_ptr != NULL)
	    {
		*Debug_stream << "\t\tDeclared as";
		temp_bucket_ptr->get_data()->type_ptr->dump_type(Debug_stream);
		*Debug_stream << "\n\t\tStorage_class is";
		temp_bucket_ptr->get_data()->storage_class_element->dump_type(Debug_stream);
		*Debug_stream << "\n";
	    }
	    while ((temp_bucket_ptr = temp_bucket_ptr->get_next()) != NULL)
	    {
		*Debug_stream << "\t" << temp_bucket_ptr->get_data()->id_string << "\n";
		if (temp_bucket_ptr->get_data()->type_ptr != NULL)
		{
		    *Debug_stream << "\t\tDeclared as";
		    temp_bucket_ptr->get_data()->type_ptr->dump_type(Debug_stream);
		    *Debug_stream << "\n\t\tStorage_class is";
		    temp_bucket_ptr->get_data()->storage_class_element->dump_type(Debug_stream);
		    *Debug_stream << "\n";
		}
	    }
	}
    }
}

identifier *symbol_table_stack::lookup(identifier *id_ptr)
{
    if (top == NULL)
    {
	cerr << "Internal error:lookup attempted on empty symbol_table_stack! ";
	cerr << Me << " will abort!\n";
	exit(-1);
    }

    identifier *temp_id_ptr = NULL;
    
    for (symbol_table_bucket *bucket_ptr = top;
	 bucket_ptr != NULL;
	 bucket_ptr = bucket_ptr->get_link())
    {
	temp_id_ptr = bucket_ptr->get_data()->lookup(id_ptr);
	if (temp_id_ptr != NULL)
	    break;
    }
    
    return temp_id_ptr;
}

identifier *symbol_table_stack::insert(char *string, TOKEN id_token)
{
    return top->get_data()->insert(string, id_token);
}

identifier *symbol_table_stack::remove(char *string)
{
    return top->get_data()->remove(string);
}

identifier *symbol_table_stack::lookup(char *string)
{
    if (top == NULL)
    {
	cerr << "Internal error:lookup attempted on empty symbol_table_stack! ";
	cerr << Me << " will abort!\n";
	exit(-1);
    }

    identifier *temp_id_ptr = NULL;
    
    for (symbol_table_bucket *bucket_ptr = top;
	 bucket_ptr != NULL;
	 bucket_ptr = bucket_ptr->get_link())
    {
	temp_id_ptr = bucket_ptr->get_data()->lookup(string);
	if (temp_id_ptr != NULL)
	    break;
    }
    
    return temp_id_ptr;
}

void symbol_table_stack::push(symbol_table_stack *stack_ptr)
{
    if (stack_ptr == NULL)
	return;

    symbol_table_stack *duplicate_stack_ptr = stack_ptr->duplicate();

    if (top == NULL)
    {
	top = duplicate_stack_ptr->top;
	return;
    }
    
    // Get the bottom of the duplicate stack.
    if (duplicate_stack_ptr->top == NULL)
    {
	return;
    }

    symbol_table_bucket* bucket_ptr;
    for (bucket_ptr = duplicate_stack_ptr->top;
	 bucket_ptr->get_link() != NULL;
	 bucket_ptr = bucket_ptr->get_link())
	;
    
    bucket_ptr->set_link(top);
    top = duplicate_stack_ptr->top;
    size = size + duplicate_stack_ptr->size;
    return;
}

symbol_table *symbol_table_stack::pop()
{
    if (top == NULL)
	return NULL;
	
    symbol_table_bucket *bucket_ptr = top;
    top = top->get_link();
    symbol_table *table_ptr = bucket_ptr->get_data();
    delete bucket_ptr;
    size = size - 1;
    return table_ptr;
}

symbol_table_stack *symbol_table_stack::duplicate()
{
    symbol_table_stack *new_stack_ptr = new symbol_table_stack();
    new_stack_ptr->size = size;
    
    if (top == NULL)
	return new_stack_ptr;
    
    new_stack_ptr->top = new symbol_table_bucket(top->get_data());

    symbol_table_bucket *new_bucket_ptr;

    for (symbol_table_bucket *old_stack_bucket_ptr = top,
	 *new_stack_bucket_ptr = new_stack_ptr->top;
	 old_stack_bucket_ptr->get_link() != NULL;
	 old_stack_bucket_ptr = old_stack_bucket_ptr->get_link())
    {
	new_bucket_ptr = 
	    new symbol_table_bucket(old_stack_bucket_ptr->get_link()->get_data());
	new_stack_bucket_ptr->set_link(new_bucket_ptr);
	new_stack_bucket_ptr = new_bucket_ptr;
    }
    
    return new_stack_ptr;
}

void symbol_table_stack::dump()
{
    for (symbol_table_bucket *bucket_ptr = top;
	 bucket_ptr != NULL;
	 bucket_ptr = bucket_ptr->get_link())
	bucket_ptr->get_data()->print();
}


