/*
 * This file is part of the Choices Operating System
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Ave.
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 * CONFIDENTIAL INFORMATION. Distribution restricted under license agreement.
 *
 * Project Manager and Principal Investigator: Roy Campbell (roy@cs.uiuc.edu)
 *
 * Funded by: NSF TAPESTRY Grant No. 1-5-30035, NASA ICLASS Grant 
 *   No. 1-5-25469 and No. NSG1471 and AT&T Metronet Grant No. 1-5-37411.
 */
/******************************************************************************
 *
 * symbol_table.h
 * 
 * 	This file contains the definitions needed for the symbol_table.
 * 
 * Ed Loyot 7/12/90
 *
 *              Copyright (c) 1990, 1991
 *              University of Virginia
 *              All Rights Reserved
 *
 *		This software is the property of the University of Virginia.
 *		Permission is granted to copy and distribute this software as
 *		long as this notice is intact on all copies.  Derived works 
 *  		must retain this notice.  This software is provided `as is'
 *		and without any express or implied warrenties.
 *
 *              Send problems & suggestions to:
 *		Ed Loyot (ecl2v@virginia.edu) or
 *              Andrew Grimshaw (grimshaw@cs.virginia.edu)
 *
 *****************************************************************************/

#ifndef SYMBOL_TABLE
#define SYMBOL_TABLE

#include "common.h"
#include "tokens.h"

//define HASH_CONSTANT 211
#define HASH_CONSTANT 27

class identifier;

/*
 * class bucket -
 *	
 * 	This class provides a bucket object to be used as a node in the linked
 * lists used in the symbol table class.
 */

class bucket
{
    identifier *data;			    
    bucket *link;
  public:
    bucket(identifier *new_data)            {  data = new_data; link = NULL;  }
    bucket 	*get_next()      	    {  return link;  }
    void	add_link(bucket *new_link)  {  link = new_link;  }
    identifier 	*get_data()		    {  return data;  }
};

/*
 * class symbol_table -
 * 
 * 	This class provides symbol table objects.  Identifiers can be inserted
 * and looked up.  Symbol tables can be kept on a stack to model nested scope.
 * The lookup function checks the table on the top of the stack first, then if
 * a match is not found it checks the tables beneath in order.  The symbol 
 * tables are implemented using a hashed array with linked lists to resolve 
 * collisions.
 */

class expression_list;

class symbol_table
{
    bucket *array[HASH_CONSTANT];
    int hash(char *);			
    identifier *insert_identifier(identifier *);
#ifdef MPLFRONT
    expression_list *prv_list;
    void build_prv_list();
#endif
  public:
    symbol_table();
    identifier *insert(char *, TOKEN);
    identifier *insert(identifier *);
    identifier *lookup(char *);
    identifier *lookup(identifier *);
    identifier *remove(char *);
    identifier *remove(identifier *);
#ifdef MPLFRONT
    expression_list *get_prvs();
#endif
    void       print();
};

/*
 * class symbol_table_bucket -
 *	
 * 	This class provides a bucket object to be used as a node in the linked
 * lists used in the symbol_table_stack class.
 */

class symbol_table_bucket
{
    symbol_table *data;			    
    symbol_table_bucket *link;
  public:
    symbol_table_bucket(symbol_table *new_data)
    {
	data = new_data;
	link = NULL;
    }

    symbol_table_bucket(symbol_table *new_data, symbol_table_bucket *new_link)
    {
	data = new_data;
	link = new_link;
    }

    symbol_table_bucket *get_link()
    {
	return link;
    }
    void set_link(symbol_table_bucket *new_link)
    {
	link = new_link;
    }
    symbol_table *get_data()
    {
	return data;
    }
};

/*
 * Class symbol_table_stack -
 *
 * 	This class provides a stack of symbol tables.  The stack operations
 * are as follows:
 * push         - add an element to the top of the stack.
 * pop          - remove an element from the top of the stack.
 * local_lookup - lookup an identifier in the symbol table on the top of the
 *                stack.
 * lookup       - lookup an identifier, start at the symbol table on top of
 *                the stack, if the identifier is not in that table look in
 *                the symbol table below it and so forth until the identifier
 *                is found.
 * insert       - insert the identifier in the symbol_table at the top of the
 *                stack.
 *
 */

class symbol_table_stack
{
    symbol_table_bucket *top;
    int size;
  public:
    symbol_table_stack()
    {
	top  = NULL;
	size = 0;
    }
    symbol_table_stack(symbol_table *new_top)
    {
	top  = new symbol_table_bucket(new_top);
	size = 1;
    }

    void push(symbol_table *new_top)
    {
	symbol_table_bucket *bucket_ptr=new symbol_table_bucket(new_top, top);
	top = bucket_ptr;
	size = size + 1;
    }
    void push(symbol_table_stack *);
    symbol_table *pop();
    identifier *insert(identifier *id_ptr)
    {
	return top->get_data()->insert(id_ptr);
    }
    identifier *insert(char *string, TOKEN);
    identifier *remove(identifier *id_ptr)
    {
	return top->get_data()->remove(id_ptr);
    }
    identifier *remove(char *);
    identifier *local_lookup(identifier *id_ptr)
    {
	return top->get_data()->lookup(id_ptr);
    }
    identifier *local_lookup(char *string)
    {
	return top->get_data()->lookup(string);
    }
    identifier *lookup(identifier *);
    identifier *lookup(char *);
    symbol_table_stack *duplicate();
    symbol_table_stack *pop_bottom();
    int stack_size()
    {
	return size;
    }
    void dump();
    void dump_top()
    {
	top->get_data()->print();
    }
};    

#endif //SYMBOL_TABLE
