#include "request.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

/*
 *
 * DEFINES
 *
 */

#ifdef SPARC_2
#define ONE_HZ 40300
#endif

#define DEBUG
/* #define CHECK_OFFSET */

#define FLAGS_CLOSE_ON_EXEC 1
#define FLAGS_CONSOLE 2
#define FLAGS_PIPE 4
#define FLAGS_FAKE 8

#define MAXFDS 			256
#define MAXFILENAMES 		39989
#define MAXPIDS 		32
#define MAXSTRLEN 		512
#define MAXPIDMAP		3989
#define FAKESTART		4096

#define DEBUG_DETAILS 1
#define DEBUG_WARNING 2
#define DEBUG_OPEN 4 
#define DEBUG_ALL DEBUG_DETAILS | DEBUG_WARNING | DEBUG_OPEN

#define MODE_SINGLE 1
#define MODE_MULTI 0

/*
 *
 * DATA TYPES
 *
 */

struct fd_info_t {
  int actual;				/* system assigned fd number */
  int flags;				/* flags */
  unsigned long offset;			/* offset */
};

struct fd_info_table_t {
  struct fd_info_t fd[MAXFDS];
};
int fd_info_table_ctor(struct fd_info_table_t *this);
int fd_info_table_close_on_exec(struct fd_info_table_t *this);
int fd_info_table_close_on_exit(struct fd_info_table_t *this);
int fd_info_table_fork(struct fd_info_table_t *this, 
	               struct fd_info_table_t *src);

struct pid_info_table_t {
  int used;
  int log_pid[MAXPIDS];
  int replay_pid[MAXPIDS];
};

int pid_info_table_ctor(struct pid_info_table_t *this);
int pid_info_table_find(struct pid_info_table_t *this, int pid);
int pid_info_table_add(struct pid_info_table_t *this, int log_pid, int rp_pid);
int pid_info_table_update(struct pid_info_table_t *this, int pid, unsigned tm);

struct filename_map_table_t {
  int used;
  struct {
    char *src;
    char *dest;
  } map[MAXFILENAMES];
}; 

int filename_map_table_ctor(struct filename_map_table_t *this);
unsigned long filename_map_table_hash(char *name, int tries, unsigned long l);
char* filename_map_table_find(struct filename_map_table_t *this, char *name);
char* filename_map_table_add(struct filename_map_table_t *this, char *buf);

struct cache_t {
  char *base;
  char *next;
  char *last;
};

struct input_t {
  FILE *file;
  struct cache_t *cache;
};

int input_ctor(struct input_t *this);
int input_dtor(struct input_t *this);
int input_copy(struct input_t *this, struct input_t *src);
int input_open(struct input_t *this, char *filename);
int input_get(struct input_t *this, struct iod_msg *msg);

struct replay_t {
  int debug;				/* debug flag */
  int pid;				/* current pid */
  int parent;				/* parent */
  int cwd;				/* fd of current dir */
  int lasttm;				/* last time */
  struct input_t input;			/* current input file */
  char *process_name[MAXSTRLEN]; 	/* process name */
  struct fd_info_table_t fd_table;	/* fd info table */
  struct pid_info_table_t pid_table;	/* pid info table */
};

int replay_ctor(struct replay_t *this);
int replay_dtor(struct replay_t *this);
int replay_open_log(struct replay_t *this, char *filename);
int replay_actual_fd(struct replay_t *this, int fd);
int replay_set_fd(struct replay_t *this, int fd, int actual_fd, int flags);
int replay_set_fd_close_on_exec(struct replay_t *this, int fd, int truefalse);
int replay_loop(struct replay_t *this, struct iod_msg *msg);
int replay_decode(struct replay_t *this, struct iod_msg *msg);
int replay_ninit(struct replay_t *this, struct iod_msg *msg);
int replay_stat(struct replay_t *this, struct iod_msg *msg);
int replay_lstat(struct replay_t *this, struct iod_msg *msg);
int replay_fstat(struct replay_t *this, struct iod_msg *msg);
int replay_fork(struct replay_t *this, struct iod_msg *msg);
int replay_open(struct replay_t *this, struct iod_msg *msg);
int replay_creat(struct replay_t *this, struct iod_msg *msg);
int replay_dup(struct replay_t *this, struct iod_msg *msg);
int replay_close(struct replay_t *this, struct iod_msg *msg);
int replay_read(struct replay_t *this, struct iod_msg *msg);
int replay_write(struct replay_t *this, struct iod_msg *msg);
int replay_exit(struct replay_t *this, struct iod_msg *msg);
int replay_unlink(struct replay_t *this, struct iod_msg *msg);
int replay_wait(struct replay_t *this, struct iod_msg *msg);
int replay_fcntl(struct replay_t *this, struct iod_msg *msg);
int replay_pipe(struct replay_t *this, struct iod_msg *msg);
int replay_chdir(struct replay_t *this, struct iod_msg *msg);
int replay_mkdir(struct replay_t *this, struct iod_msg *msg);
int replay_rmdir(struct replay_t *this, struct iod_msg *msg);
int replay_link(struct replay_t *this, struct iod_msg *msg);
int replay_symlink(struct replay_t *this, struct iod_msg *msg);
int replay_rename(struct replay_t *this, struct iod_msg *msg);
int replay_socket(struct replay_t *this, struct iod_msg *msg);

typedef int (* pfn_t)(struct replay_t *, struct iod_msg *msg);

struct pid_map_table_t {
  struct replay_t *replay[MAXPIDMAP];
};

int pid_map_table_ctor(struct pid_map_table_t *this);
struct replay_t * pid_map_table_find(struct pid_map_table_t *this, int pid);
int pid_map_table_add(struct pid_map_table_t *this, int pid, 
		      struct replay_t * replay);

/* 
 *
 * GLOBALS
 *
 */

int  replay_mode;		/* mode */
int  replay_debug;		/* debug */
int  replay_precache;		/* precache */
int  replay_o_flag;		/* o_flag */
int  replay_instdisk;		/* inst_disk */
int  replay_inst;		/* instrument */
int  replay_syscalls;		/* no of syscalls */
int  replay_nostat;		/* no stat and fstat */
int  replay_stdout;		/* stdout fd */
int  replay_stderr;		/* stderr fd */
int  replay_noprint;		/* no console output */
int  replay_delay;		/* delay */
int  replay_hzcount;		/* replay hzcount */
int  open_count = 0;		/* count of opens */
char dummy[65536];		/* dummy buffer */
char *replay_output;		/* output file name */
char *replay_progname;		/* argv[0] name of program */
char *replay_filename; 		/* replay file name */
char *replay_logpath;		/* path where log files can be found */
struct cache_t *replay_cache;   /* replay cache */

int time_start(int real);
int time_done(int real);

extern int errno;

struct filename_map_table_t _filename_map_table;
struct filename_map_table_t * filename_map_table = &_filename_map_table;

struct pid_map_table_t _pid_map_table;
struct pid_map_table_t * pid_map_table = &_pid_map_table;

struct {
  int cmd;
  pfn_t fn;
} replay_table[] = 
{ { IOD_NINITIATE, replay_ninit },
  { IOD_FORK, replay_fork },
  { IOD_STAT, replay_stat },
  { IOD_FSTAT, replay_fstat },
  { IOD_OPEN, replay_open },
  { IOD_CREAT, replay_creat },
  { IOD_DUP, replay_dup },
  { IOD_CLOSE, replay_close },
  { IOD_IOCTL, 0 },
  { IOD_READ, replay_read },
  { IOD_WRITE, replay_write },
  { IOD_READA, replay_read },
  { IOD_WRITEA, replay_write },
  { IOD_EXIT, replay_exit },
  { IOD_SOCKET, replay_socket },
  { IOD_BIND, 0 },
  { IOD_LISTEN, 0 },
  { IOD_ACCEPT, 0 },
  { IOD_CONNECT, 0 },
  { IOD_SOCKETPAIR, 0 },
  { IOD_SENDTO, 0 },
  { IOD_SEND, 0 },
  { IOD_SENDMSG, 0 },
  { IOD_RECVFROM, 0 },
  { IOD_RECV, 0 },
  { IOD_RECVMSG, 0 },
  { IOD_SHUTDOWN, 0 },
  { DAEMON_START, 0 },
  { IOD_READLA, replay_read },
  { IOD_READL, replay_read },
  { IOD_WRITELA, replay_write },
  { IOD_WRITEL, replay_write },
  { IOD_READV, replay_read },
  { IOD_WRITEV, replay_write },
  { IOD_FCNTL, replay_fcntl },
  { IOD_PIPE, replay_pipe },
  { IOD_UNLINK, replay_unlink },
  { IOD_VFORK, replay_fork },
  { IOD_WAIT, replay_wait },
  { XIO_OPEN, 0 },
  { XIO_CLOSE, 0 },
  { XIO_READ, 0 },
  { XIO_WRITE, 0 },
  { XIO_REQUEST, 0 },
  { XIO_REPLY, 0 },
  { XIO_EVENT, 0 },
  { XIO_FLUSH, 0 },
  { XIO_FLUSH_INFO, 0 },
  { IOD_CHDIR, replay_chdir },
  { IOD_LINK, replay_link },
  { IOD_SYMLINK, replay_symlink },
  { IOD_RENAME, replay_rename },
  { IOD_MKDIR, replay_mkdir },
  { IOD_RMDIR, replay_rmdir },
  { IOD_LSTAT, replay_lstat },
  { XIO_RESERVED8, 0 },
  { XIO_RESERVED9, 0 },
  { XIO_RESERVED10, 0 },
  { XIO_RESERVED11, 0 },
  { XIO_RESERVED12, 0 },
  { CNTL_FLUSH, 0 },
  { NET_REQ, 0 }
};

/*
 * utility
 */

char age = 0;

void filldummy(int size)
{
  int i;
  for( i = 128; i < size; i += 256 ) 
    dummy[i] = age++;
}

int delay(int no_hz)
{
  while(no_hz-- > 0) {
    register int count = replay_hzcount;
    while (count-- > 0);
  }
  return 0;
}
 
/*
 * fake routines
 */

int fake_fd = FAKESTART;

int dup_fake(int fd)
{
  if (fd < FAKESTART) {
    fprintf(stderr,"ERROR: cannot dup_fake real fd\n");
    return -1;
  }
  return fake_fd++;
}

int pipe_fake(int *fd) 
{
  fd[0] = fake_fd++;
  fd[1] = fake_fd++;
  return 0;
}

int socket_fake()
{
  return fake_fd++;
}

/*
 *
 * fd_info_table
 *
 */

int
fd_info_table_ctor(
  struct fd_info_table_t *this
)
{
  int i;
  this->fd[0].actual = dup(0);
  this->fd[1].actual = dup(replay_stdout);
  this->fd[2].actual = dup(replay_stderr);
  this->fd[0].flags = FLAGS_CONSOLE;
  this->fd[1].flags = FLAGS_CONSOLE;
  this->fd[2].flags = FLAGS_CONSOLE;
  this->fd[0].offset = 0;
  this->fd[1].offset = 0;
  this->fd[2].offset = 0;
  for(i = 3; i < MAXFDS; i++) {
    this->fd[i].actual = -1; 
    this->fd[i].flags = 0; 
    this->fd[i].offset = 0;
  }
}

int
fd_info_table_close_on_exec(
  struct fd_info_table_t *this
)
{
  int i;
  for(i = 0; i < MAXFDS; i++) {
    if (this->fd[i].actual != -1) {
      if (this->fd[i].flags & FLAGS_CLOSE_ON_EXEC) {
        if (!(this->fd[i].flags & FLAGS_FAKE)) {
          close(this->fd[i].actual);
	  replay_syscalls++;
     	}
	this->fd[i].actual = -1;
	this->fd[i].flags = 0;
	this->fd[i].offset = 0;
      }
    }
  }
  return 0;
}

int
fd_info_table_close_on_exit(
  struct fd_info_table_t *this
)
{
  int i;
  int status;
  for(i = 0; i < MAXFDS; i++) {
    if (this->fd[i].actual != -1) {
      if (!(this->fd[i].flags & FLAGS_FAKE)) {
        status = close(this->fd[i].actual);
        replay_syscalls++;
        if (status < 0) {
   	  fprintf(stderr,"ERROR: fd_info_table_close_on_exit() close failed assigned fd %d errno %d",this->fd[i].actual,errno);
 	  exit(1);
        }
      }
      if (replay_debug & DEBUG_DETAILS) {
 	fprintf(stdout,"fd_info_table_close_on_exit() fd %d assigned %d\n",
	        i,this->fd[i].actual);
      }
      this->fd[i].actual = -1;
      this->fd[i].flags = 0;
      this->fd[i].offset = 0;
    }
  }
  return 0;
}

int
fd_info_table_fork(
  struct fd_info_table_t *this,
  struct fd_info_table_t *src
)
{
  int i;
  int new_fd;
  int error;
  error = 0;
  for(i = 0; i < MAXFDS; i++) {
    if (src->fd[i].actual == -1) {
      this->fd[i].actual = -1;
      this->fd[i].flags = 0;
      this->fd[i].offset = 0;
    } else {
      if (src->fd[i].flags & FLAGS_FAKE) {
        new_fd = dup_fake(src->fd[i].actual);
      } else {
        new_fd = dup(src->fd[i].actual);
        replay_syscalls++;
      }
      if (new_fd < 0) {
 	fprintf(stderr,"ERROR: fd_info_table_fork dup error, errno %d\n",
		errno);
   	error = -1;
      }
      if (replay_debug & DEBUG_DETAILS) {
 	fprintf(stdout,"fd_info_table_fork() dup fd %d assigned %d\n",
		i,new_fd);
      }
      this->fd[i].actual = new_fd;
      this->fd[i].flags = src->fd[i].flags;
      this->fd[i].offset = src->fd[i].offset;
    }
  }
  return error;
}

/* 
 * 
 * pid_info_table
 *
 */

int
pid_info_table_ctor(
  struct pid_info_table_t *this
)
{
  this->used = 0;
}

int 
pid_info_table_find(
  struct pid_info_table_t *this,
  int pid
)
{
  int i;
  for (i = 0; i < this->used; i++) {
    if (this->log_pid[i] == pid)
      return this->replay_pid[i];
  }
  return -1;
}

int 
pid_info_table_add(
  struct pid_info_table_t *this,
  int log_pid,
  int replay_pid
)
{
  if (this->used == MAXPIDS) {
    fprintf(stderr,"ERROR: pid_info_table overflow\n");
    exit(1);
  }
  this->log_pid[this->used] = log_pid;
  this->replay_pid[this->used] = replay_pid;
  this->used++;
  return replay_pid;
}

/*
 *
 * filename_map_table
 *
 */

int
filename_map_table_ctor(
  struct filename_map_table_t *this
)
{
  int i;
  this->used = 0;
  for(i = 0; i < MAXFILENAMES; i++) {
    this->map[i].src = 0;
    this->map[i].dest = 0;
  }
  return 0;
}

unsigned long
filename_map_table_hash(
  char * name,
  int tries,
  unsigned long prev
)
{
  unsigned long val;
  unsigned long count = 0;
  int shift = 0;

  if (tries > 0) {
    int len = strlen(name);
    unsigned long val = name[( len - 1 ) - ( (tries - 1) % len )] << 8;
    return ( val + prev ) % MAXFILENAMES; 
  } else {
    val = 0;
  }
  while (*name != 0) {
    count++;
    val += ( (int)(*name) ) << shift;
    shift = (shift + 4) & 31;
    name++;
  }
  return (count + val) % MAXFILENAMES;
}

char *
filename_map_table_find(
  struct filename_map_table_t *this,
  char * filename
)
{
  unsigned long hash;
  int tries = 0;
  hash = filename_map_table_hash(filename,tries,0UL);
  while (this->map[hash].src != 0) {
    if (strcmp(this->map[hash].src, filename) == 0)
      return this->map[hash].dest;
    tries++;
    hash = filename_map_table_hash(filename,tries,hash);
  }
  return filename;
}

char *
filename_map_table_add(
  struct filename_map_table_t *this,
  char *buf
)
{
  int probes;
  int buflen;
  int tries;
  int warned = 0;
  unsigned long hash;
  char *s;
  char *start;
  char *d;
  char *src;
  char *dest;

  if (this->used == MAXFILENAMES) {
    fprintf(stderr,"ERROR: filename_map_table overflow\n");
    exit(1);
  }
  buflen = strlen(buf);
  if (buflen > 0 && buf[buflen-1] == '\n') {
    buf[buflen-1] = 0;
    buflen--;
  }
  if (buflen == 0)
    return 0;
  s = buf;
  while( isspace(*s) && *s != 0 ) s++;
  start = s;
  while( !isspace(*s) && *s != 0 ) s++;
  src = d = (char *)malloc( (s - start) + 1 );
  while( start != s ) *d++ = *start++; 
  *d = 0;
  while( isspace(*s) && *s != 0 ) s++;
  start = s;
  while( !isspace(*s) && *s != 0 ) s++;
  dest = d = (char *)malloc( (s - start) + 1 );
  while( start != s ) *d++ = *start++; 
  *d = 0;
  this->used++;
  tries = 0;
  hash = filename_map_table_hash(src,tries,0UL);
  /* fprintf(stderr,"%5d %s\n", hash, src); */
  while (this->map[hash].src != 0) {
    tries++;
    hash = filename_map_table_hash(src,tries,hash);
    if (tries > 30 && !warned) {
      fprintf(stderr,"%s more than 30 probes\n",src);
      warned = 1;
    }
  }
  this->map[hash].src = src;
  this->map[hash].dest = dest;
  return 0;
}

int 
pid_map_table_ctor(
  struct pid_map_table_t *this
)
{
  int i;
  for(i = 0; i < MAXPIDMAP; i++) {
    this->replay[i] = 0;
  }
  return 0;
}

struct replay_t * 
pid_map_table_find(
  struct pid_map_table_t *this, 
  int pid
)
{
  int tries;
  int slot;
  tries = 0;
  slot = pid % MAXPIDMAP;
  while(this->replay[slot] != 0 && tries < 50) {
    if (this->replay[slot]->pid == pid)
      return this->replay[slot];
    slot = (slot + MAXPIDMAP-3) % MAXPIDMAP;
    tries++;
  }
  if (this->replay[slot] == 0)
    return 0;
  fprintf(stderr,"ERROR: pid_map_table_find cannot find\n");
  exit(1);
}

int
pid_map_table_add(
  struct pid_map_table_t *this, 
  int pid, 
  struct replay_t * replay
)
{
  int tries;
  int slot;
  tries = 0;
  slot = pid % MAXPIDMAP;
  while(this->replay[slot] != 0 && tries < 50) {
    if (this->replay[slot]->pid == pid)
      return 0;
    slot = (slot + MAXPIDMAP-3) % MAXPIDMAP;
    tries++;
  }
  if (this->replay[slot] == 0) {
    this->replay[slot] = replay;
    return 0;
  }
  fprintf(stderr,"ERROR: pid_map_table_add cannot add\n"); 
  exit(1);   
}

/*
 *
 * input
 *
 */

int 
input_ctor(
  struct input_t *this
)
{
  this->file = 0;
  this->cache = 0;
  return 0;
}

int 
input_dtor(
  struct input_t *this
)
{
  if (this->file != 0 && replay_mode != MODE_SINGLE) {
    fclose(this->file);
  }
  this->file = 0;
  this->cache = 0;
}

int 
input_open(
  struct input_t *this, 
  char *filename
)
{
  int fd;
  int status;
  int size;
  char replay_filename[MAXSTRLEN];
  struct stat stat_buf;
  
  if (replay_precache) {
    if (replay_cache) {
      this->cache = replay_cache;
      this->file = 0;
    } else {
      fd = open(filename, O_RDONLY);
      if (fd < 0) {
        fprintf(stderr,"ERROR: cannot open %s, errno %d\n",filename, errno);
        exit(1);
      }
      status = fstat(fd, &stat_buf);
      if (status < 0) {
        fprintf(stderr,"ERROR: cannot fstat %s, errno %d\n",filename, errno);
        exit(1);
      }
      size = stat_buf.st_size;
      replay_cache = (struct cache_t *) malloc(sizeof(struct replay_t));
      if (replay_cache == 0) {
        fprintf(stderr,"ERROR: cannot malloc replay_cache\n");
        exit(1);
      }
      replay_cache->base = (char *) malloc(size);
      if (replay_cache->base == 0) {
        fprintf(stderr,"ERROR: cannot malloc replay_cache buffer\n");
        exit(1);
      }
      status = read(fd, replay_cache->base, size);
      if (size != status) {
        fprintf(stderr,"ERROR: read into replay_cache failed\n");
	exit(1);
      } 
      close(fd);
      replay_cache->next = replay_cache->base;
      replay_cache->last = replay_cache->base + status;
      this->file = 0;
      this->cache = replay_cache;
      fprintf(stdout,"replay_cache constructed size = %d\n", size);
    }
  } else {
    if (this->file)
      fclose(this->file);
    this->file = fopen(filename,"r");
    if (this->file == 0 ) {
      fprintf(stderr,"ERROR: cannot open %s, errno %d\n",filename,errno);
      exit(1);
    }
  }
  return 0;
}

int 
input_get(
  struct input_t *this, 
  struct iod_msg *msg
)
{
  if (this->file) {
    return fread((char *) msg, sizeof(struct iod_msg), 1, this->file);
  } else {
    if (this->cache->next >= this->cache->last) 
      return 0;
    memcpy((char*) msg, this->cache->next, sizeof(struct iod_msg));
    this->cache->next += 64;
    return 64;
  }
}

int 
input_copy(
  struct input_t *this,
  struct input_t *src
)
{
  this->file = src->file;
  this->cache = src->cache;
}

/*
 *
 * replay
 *
 */

int
replay_ctor(
  struct replay_t *this
)
{
  this->debug = 0;
  this->pid = -1;
  this->parent = -1;
  this->lasttm = 0;
  fd_info_table_ctor(&(this->fd_table));
  pid_info_table_ctor(&(this->pid_table));
  input_ctor(&(this->input));
}

int
replay_dtor(
  struct replay_t *this
)
{
  input_dtor(&(this->input));
}

int
replay_actual_fd(
  struct replay_t *this,
  int fd
)
{
  return this->fd_table.fd[fd].actual;
}

int
replay_set_fd(
  struct replay_t *this,
  int fd, 
  int actual,
  int flags
)
{
  this->fd_table.fd[fd].actual = actual;
  this->fd_table.fd[fd].flags = flags;
  this->fd_table.fd[fd].offset = 0;
  return actual; 
}
 
int 
replay_set_fd_close_on_exec(
  struct replay_t *this, 
  int fd, 
  int truefalse
)
{
  if (truefalse) {
    this->fd_table.fd[fd].flags |= FLAGS_CLOSE_ON_EXEC;
  } else {
    this->fd_table.fd[fd].flags &= ~FLAGS_CLOSE_ON_EXEC;
  }
  return fd;
}

int
replay_open_log(
  struct replay_t *this,
  char *filename
)
{
  int result;
  char replay_filename[MAXSTRLEN];
  if (filename == 0) 
    sprintf(replay_filename,"%s/%d",replay_logpath,this->pid);
  else
    strcpy(replay_filename,filename);
  result = input_open(&(this->input), filename);
  if (result != 0 ) {
    fprintf(stderr,"ERROR: cannot open %s\n",replay_filename);
    exit(1);
  }
}

int
replay_loop(
  struct replay_t *this,
  struct iod_msg *ret_msg
)
{
  int status = 1;
  struct iod_msg msg;
  struct replay_t * replay;
  while (input_get(&(this->input), &msg)) {
    if (this->debug & DEBUG_DETAILS)
      fdump_iod_msg(stdout,&msg);
    while (msg.pid != this->pid && msg.pid != 0 && replay_mode == MODE_SINGLE) {
      if (this->parent == -1) {
        replay = pid_map_table_find(pid_map_table,msg.pid);
        if (replay == 0) {
	  fprintf(stderr,"ERROR: unknown pid %d\n",msg.pid);
	  exit(1);
   	}
        if (replay->lasttm < msg.tm) {
          if (replay_delay) delay(msg.tm - replay->lasttm);
          replay->lasttm = msg.tm;
        }
	replay_decode(replay,&msg);
	status = replay_loop(replay,&msg);
        if (status != 0) {
	  continue;
	} else {
	  break;
	}
      } else {
	memcpy((char *)ret_msg,(char *)&msg,sizeof(msg));
        return 1;
      }
    }
    if (status == 0)
      break;
    if (msg.pid == 0)
      continue;
    if (msg.pid != this->pid) {
      fdump_iod_msg(stderr,&msg);
      fprintf(stderr,"ERROR: incorrect pid\n");
      exit(1);
    }
    if (this->lasttm < msg.tm) {
      if (replay_delay) delay(msg.tm - this->lasttm);
      this->lasttm = msg.tm;
    }
    replay_decode(this,&msg);
  }
  return 0;
}


int
replay_decode(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int cmd;
  pfn_t fn;

  cmd = msgp->cmd;
  if (replay_table[cmd].cmd != cmd) {
    fprintf(stderr,"ERROR: replay_table fatal error\n");
    exit(1);
  }
  fn = replay_table[cmd].fn;
  if (fn == 0) {
    if (this->debug & DEBUG_WARNING) {
      fdump_iod_msg(stderr,msgp);
      fprintf(stderr,"WARNING: %d cannot replay %s\n",
	      msgp->pid,symtable_lookup_int(cmd));
    }
    return -1;
  }
  return fn(this,msgp);
}

int
replay_ninit(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int i;
  strncpy((char *)this->process_name, (char *)iod_msg_name(msgp),
	  sizeof(this->process_name)-1);
  if (this->debug & DEBUG_DETAILS) {
    fprintf(stderr,"%d replaying %s\n",this->pid,this->process_name);
  }
  fd_info_table_close_on_exec(&(this->fd_table));
  return 0;
}

int 
replay_fork(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int i;
  int status;
  int new_pid;
  int new_fd;
  int actual_fd;
  struct replay_t *forked_replay;
  new_pid = msgp->arg.fork_arg.new_pid;
  if (new_pid < 0)
    return 0;
  if (this->debug & DEBUG_DETAILS) {
    fdump_iod_msg(stdout,msgp);
    fprintf(stdout,"forking\n");
  }
  if (replay_mode == MODE_SINGLE) {
    forked_replay = (struct replay_t *) malloc(sizeof(struct replay_t));
    if (forked_replay == 0) {
      fprintf(stderr,"ERROR: fork malloc error\n");
      exit(1);
    }
    forked_replay->pid = new_pid;
    forked_replay->debug = this->debug;
    forked_replay->parent = this->pid;
    pid_info_table_ctor(&(forked_replay->pid_table));
    fd_info_table_fork(&(forked_replay->fd_table),&(this->fd_table));
    input_copy(&(forked_replay->input),&(this->input));
    pid_map_table_add(pid_map_table,new_pid,forked_replay);
  } else {
    replay_syscalls++;
    status = fork();
    if (status == 0) {
      this->pid = new_pid;
      replay_open_log(this,0);
      return 0;
    }
    if (status > 0) {
      pid_info_table_add(&(this->pid_table),new_pid,status);
      return 0;
    }
    fprintf(stderr,"ERROR: %d fork %d error\n",msgp->pid,new_pid);
    return -1;
  } 
}

int 
replay_stat(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int status;
  char filename[25];
  char *realname;
  struct stat buf;

  if (replay_nostat)
    return 0;
  strncpy(filename,msgp->arg.string,24);
  filename[24] = 0;
  realname = filename_map_table_find(filename_map_table,filename);
  status = stat(realname,&buf);
  replay_syscalls++;
  return 0;
}

int 
replay_lstat(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int status;
  char filename[25];
  char *realname;
  struct stat buf;

  if (replay_nostat)
    return 0;
  strncpy(filename,msgp->arg.string,24);
  filename[24] = 0;
  realname = filename_map_table_find(filename_map_table,filename);
  status = lstat(realname,&buf);
  replay_syscalls++;
  return 0;
}

int 
replay_link(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int status;
  int result;
  char filename[25];
  char destfilename[25];
  char *realname;
  char *destrealname;
  struct stat buf;
  struct iod_msg parttwo;

  status = msgp->fd;
  strncpy(filename,msgp->arg.string,24);
  filename[24] = 0;
  realname = filename_map_table_find(filename_map_table,filename);
  if (input_get(&(this->input), &parttwo) == 0) {
    fprintf(stderr, "ERROR: premature end of input after\n");
    fdump_iod_msg(stdout, msgp);
    return -1;
  }
  if (msgp->cmd != IOD_LINK) {
    fprintf(stderr, "ERROR: inconsistent input, LINK expected, after\n");
    fdump_iod_msg(stdout, msgp);
    fdump_iod_msg(stdout, &parttwo);
    return -1;
  }
  strncpy(destfilename,parttwo.arg.string,24);
  destfilename[24] = 0;
  destrealname = filename_map_table_find(filename_map_table,destfilename);
  if (status < 0) 
    return 0;
  result = link(realname, destrealname);
  if (result != 0 && status == 0) {
    fprintf(stderr, "ERROR: link error\n");
    fdump_iod_msg(stdout, msgp);
    fdump_iod_msg(stdout, &parttwo);
    return -1;
  }
  replay_syscalls++;
  return 0;
}

int 
replay_symlink(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int status;
  int result;
  char filename[25];
  char destfilename[25];
  char *realname;
  char *destrealname;
  struct stat buf;
  struct iod_msg parttwo;

  status = msgp->fd;
  strncpy(filename,msgp->arg.string,24);
  filename[24] = 0;
  realname = filename_map_table_find(filename_map_table,filename);
  if (input_get(&(this->input), &parttwo) == 0) {
    fprintf(stderr, "ERROR: premature end of input after\n");
    fdump_iod_msg(stdout, msgp);
    return -1;
  }
  if (msgp->cmd != IOD_SYMLINK) {
    fprintf(stderr, "ERROR: inconsistent input, SYMLINK expected, after\n");
    fdump_iod_msg(stdout, msgp);
    fdump_iod_msg(stdout, &parttwo);
    return -1;
  }
  strncpy(destfilename,parttwo.arg.string,24);
  destfilename[24] = 0;
  destrealname = filename_map_table_find(filename_map_table,destfilename);
  if (status < 0) 
    return 0;
  result = symlink(realname, destrealname);
  if (result != 0 && status == 0) {
    fprintf(stderr, "ERROR: symlink error\n");
    fdump_iod_msg(stdout, msgp);
    fdump_iod_msg(stdout, &parttwo);
    return -1;
  }
  replay_syscalls++;
  return 0;
}

int 
replay_rename(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int status;
  int result;
  char filename[25];
  char destfilename[25];
  char *realname;
  char *destrealname;
  struct stat buf;
  struct iod_msg parttwo;

  status = msgp->fd;
  strncpy(filename,msgp->arg.string,24);
  filename[24] = 0;
  realname = filename_map_table_find(filename_map_table,filename);
  if (input_get(&(this->input), &parttwo) == 0) {
    fprintf(stderr, "ERROR: premature end of input after\n");
    fdump_iod_msg(stdout, msgp);
    return -1;
  }
  if (msgp->cmd != IOD_RENAME) {
    fprintf(stderr, "ERROR: inconsistent input, RENAME expected, after\n");
    fdump_iod_msg(stdout, msgp);
    fdump_iod_msg(stdout, &parttwo);
    return -1;
  }
  strncpy(destfilename,parttwo.arg.string,24);
  destfilename[24] = 0;
  destrealname = filename_map_table_find(filename_map_table,destfilename);
  if (status < 0) 
    return 0;
  result = rename(realname, destrealname);
  if (result != 0 && status == 0) {
    fprintf(stderr, "ERROR: rename error\n");
    fdump_iod_msg(stdout, msgp);
    fdump_iod_msg(stdout, &parttwo);
    return -1;
  }
  replay_syscalls++;
  return 0;
}

int 
replay_chdir(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int status;
  int result;
  char filename[25];
  char *realname;

  status = msgp->fd;
  if (status < 0)
    return 0;
  strncpy(filename,msgp->arg.string,24);
  filename[24] = 0;
  realname = filename_map_table_find(filename_map_table,filename);
  result = chdir(realname);
  replay_syscalls++;
  if (result != 0 && status == 0) {
    fdump_iod_msg(stderr, msgp);
    fprintf(stderr,"chdir failed\n");
    return -1;
  }
  return 0;
}

int 
replay_mkdir(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int status;
  int result;
  char filename[25];
  char *realname;

  status = msgp->fd;
  if (status < 0)
    return 0;
  strncpy(filename,msgp->arg.string,24);
  filename[24] = 0;
  realname = filename_map_table_find(filename_map_table,filename);
  result = mkdir(realname,0777);
  replay_syscalls++;
  if (result != 0 && status == 0) {
    fdump_iod_msg(stderr, msgp);
    fprintf(stderr,"mkdir failed\n");
    return -1;
  }
  return 0;
}

int 
replay_rmdir(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int status;
  int result;
  char filename[25];
  char *realname;

  status = msgp->fd;
  if (status < 0)
    return 0;
  strncpy(filename,msgp->arg.string,24);
  filename[24] = 0;
  realname = filename_map_table_find(filename_map_table,filename);
  result = rmdir(realname);
  replay_syscalls++;
  if (result != 0 && status == 0) {
    fdump_iod_msg(stderr, msgp);
    fprintf(stderr,"rmdir failed\n");
    return -1;
  }
  return 0;
}

int
replay_fstat(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int fd;
  int actual_fd;
  struct stat buf;
  fd = msgp->fd;
  if (fd < 0)
    return 0;
  if (replay_nostat)
    return 0;
  actual_fd = replay_actual_fd(this,fd);
  if (actual_fd == -1) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: fd not inuse\n");
    return -1;
  }
  if (!(this->fd_table.fd[fd].flags & FLAGS_FAKE)) {
    fstat(actual_fd,&buf);
    replay_syscalls++;
  }
  return 0;
}

int
replay_pipe(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  static int first = 0;
  static int fd[2] = {-1,-1};
  int assigned_fd[2];
  int status;
  int i;
  int fd_flags;
  
  if (msgp->fd == -1)
    return 0;
  if (first == 0) {
    fd[0] = msgp->fd;
    first = 1;
    return 0;
  }  
  fd[1] = msgp->fd;
  first = 0;
  if (replay_mode == MODE_SINGLE) {
    status = pipe_fake(assigned_fd);
    fd_flags = FLAGS_PIPE | FLAGS_FAKE;
  } else {
    status = pipe(assigned_fd);
    replay_syscalls++;
    fd_flags = FLAGS_PIPE;
  }
  if (status != 0) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: pipe() failed, errno %d\n",errno);
    return -1;
  }
  status = 0;
  for(i = 0; i < 2; i++) {
    if (replay_actual_fd(this,fd[i]) != -1) {
      fdump_iod_msg(stderr,msgp);
      fprintf(stderr,"ERROR: fd %d inuse, forced substitution\n",fd[i]);
      status = -1;
    }
    replay_set_fd(this,fd[i],assigned_fd[i],fd_flags);
    if (this->debug & DEBUG_DETAILS) {
      fprintf(stdout,"fd %d assigned to %d\n",fd[i],assigned_fd[i]);
    }
  }   
  return status;
}

int
replay_socket(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int status;
  int i;
  int fd_flags;
  int fd;
  int assigned_fd;
  
  if (msgp->fd == -1)
    return 0;
  fd = msgp->fd;
  fd_flags = FLAGS_FAKE | FLAGS_PIPE;
  assigned_fd = socket_fake();
  status = 0;
  if (replay_actual_fd(this,fd) != -1) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: fd %d inuse, forced substitution\n",fd);
    status = -1;
  }
  replay_set_fd(this,fd,assigned_fd,fd_flags);
  if (this->debug & DEBUG_DETAILS) {
    fprintf(stdout,"fd %d assigned to %d\n",fd,assigned_fd);
  }   
  return status;
}

int
replay_fcntl(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int fd;
  int status;
  int cmd;
  int arg;
  int assigned_fd;

  fd = msgp->fd;
  status = msgp->arg.fcntl_arg.status;
  cmd = msgp->arg.fcntl_arg.cmd;
  arg = msgp->arg.fcntl_arg.arg;
  if (status == -1)
    return 0;
  assigned_fd = replay_actual_fd(this, fd);
  if (assigned_fd == -1) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: fcntl on invalid fd\n");
    return -1;
  }
  switch(cmd) {
    case 2:
      replay_set_fd_close_on_exec(this, fd, arg & 1);
      break;
    default:
      fdump_iod_msg(stderr,msgp);
      fprintf(stderr,"ERROR: unknown fcntl command\n");
      return -1;
  }
  return 0;
}

int
replay_open(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int assigned_fd;
  int fd;
  char filename[21];
  char *realname;
  fd = msgp->fd;
  if (fd < 0)
    return 0;
  strncpy(filename,msgp->arg.open_arg.path,20);
  filename[20] = 0;
  realname = filename_map_table_find(filename_map_table,filename);
  if (this->debug & (DEBUG_DETAILS | DEBUG_OPEN)) {
    fprintf(stdout,"%5d open() realname %s, logname %s\n",++open_count,realname,filename);
  }
  assigned_fd = open(realname,msgp->arg.open_arg.flags | replay_o_flag,0777);
  replay_syscalls++;
  if (fd >= 0 && assigned_fd < 0) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: open() cannot open file %s, errno %d\n",
	    realname,errno);
    return -1;
  }
  if (replay_actual_fd(this,fd) != -1) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: open() fd %d slot in use, overriding it\n",fd);
  }
  replay_set_fd(this,fd,assigned_fd,0);
  if (this->debug & DEBUG_DETAILS) {
    fprintf(stdout,"fd %d assigned to %d\n",fd,assigned_fd);
  }
  return 0;
}

int
replay_creat(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int assigned_fd;
  int fd;
  char filename[21];
  char * realname;
  fd = msgp->fd;
  if (fd < 0)
    return 0;
  strncpy(filename,msgp->arg.open_arg.path,20);
  filename[20] = 0;
  realname = filename_map_table_find(filename_map_table,filename);
  if (this->debug & (DEBUG_DETAILS | DEBUG_OPEN)) {
    fprintf(stdout,"%5d open() realname %s, logname %s\n",++open_count,realname,filename);
  }
  assigned_fd = creat(realname,msgp->arg.open_arg.flags);
  replay_syscalls++;
  if (fd >= 0 && assigned_fd < 0) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: creat() cannot creat file %s, errno %d\n",
	    realname,errno);
    return -1;
  }
  if (replay_actual_fd(this,fd) != -1) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: creat() fd %d slot in use, overriding it\n",fd);
  }
  replay_set_fd(this,fd,assigned_fd,0);
  if (this->debug & DEBUG_DETAILS) {
    fprintf(stdout,"fd %d assigned to %d\n",fd,assigned_fd);
  }
  return 0;
}

int
replay_dup(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int fd;
  int dupfd;
  int assigned_fd;
  int actual_fd;
  int actual_dup_fd;
  int dup_flags;
  fd = msgp->fd;
  dupfd = msgp->arg.io_arg.actual;
  if (fd < 0)
    return 0;
  actual_dup_fd = replay_actual_fd(this,dupfd);
  dup_flags = this->fd_table.fd[dupfd].flags;
  if (actual_dup_fd == -1) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: dup() trying to dup unused fd\n");
    return -1;
  }
  actual_fd = replay_actual_fd(this,fd);
  if (actual_fd != -1) {
    close(actual_fd);
    replay_set_fd(this,fd,-1,0);
  }
  if (dup_flags & FLAGS_FAKE) {
    assigned_fd = dup_fake(actual_dup_fd);
  } else {
    assigned_fd = dup(actual_dup_fd);
    replay_syscalls++;
  }
  if (assigned_fd < 0) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: dup() dup failed, errno %d\n",errno);
    return -1;
  }
  replay_set_fd(this,fd,assigned_fd,dup_flags);
  this->fd_table.fd[fd].offset = this->fd_table.fd[dupfd].offset;
  if (this->debug & DEBUG_DETAILS) {
    fprintf(stdout,"fd %d assigned to %d\n",fd,assigned_fd);
  }
  return 0;
}

int
replay_close(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int fd;
  int actual_fd;
  fd = msgp->fd;
  if (fd < 0)
    return 0;
  actual_fd = replay_actual_fd(this,fd);
  if (actual_fd == -1) {
    if (this->debug & DEBUG_WARNING) {
      fdump_iod_msg(stderr,msgp);
      fprintf(stderr,"WARNING: fd not inuse\n");
      return -1;
    }
  }
  if (!(this->fd_table.fd[fd].flags & FLAGS_FAKE)) {
    close(actual_fd);
    replay_syscalls++;
  }
  replay_set_fd(this,fd,-1,0);
  return 0;
}

int 
replay_read(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int fd;
  int assigned_fd;
  int length;
  int actual;
  int offset;
  int cmd;
  int status;
  long current_offset;

  fd = msgp->fd;
  cmd = msgp->cmd;
  length = msgp->arg.io_arg.length;
  offset = msgp->arg.io_arg.offset;
  if (cmd == IOD_READA || cmd == IOD_READLA) {
    actual = length;
  } else {
    actual = msgp->arg.io_arg.actual; 
  }
  if (actual < 0)
    return 0;
  assigned_fd = replay_actual_fd(this,fd);
  if (assigned_fd == -1) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: read() fd not inuse\n");
    return -1;
  }
  if (!(this->fd_table.fd[fd].flags & FLAGS_PIPE) &&
      !(this->fd_table.fd[fd].flags & FLAGS_CONSOLE)) {
#ifdef CHECK_OFFSET
    current_offset = tell(assigned_fd);
    replay_syscalls++;
#else
    current_offset = this->fd_table.fd[fd].offset;
#endif /* CHECK_OFFSET */
    if (current_offset != offset && 
        !(this->fd_table.fd[fd].flags & FLAGS_CONSOLE)) {
      status = lseek(assigned_fd,offset,SEEK_SET);
      replay_syscalls++;
      if (status != offset) {
        fdump_iod_msg(stderr,msgp);
        fprintf(stderr,"ERROR: read() lseek(%d,%d) returns %d\n",
  	      assigned_fd,offset,status);
        return -1;
      }
    }
  }
/*
  if (actual > sizeof(dummy)) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"FATAL ERROR: dummy too small\n");
    return -1;
  }
 */
  if (replay_mode == MODE_SINGLE && 
      (this->fd_table.fd[fd].flags & (FLAGS_PIPE | FLAGS_FAKE))) 
    status = actual;
  else if (actual <= sizeof(dummy)) {
    status = read(assigned_fd,dummy,actual);
    replay_syscalls++;
  } else {
    int size;
    int remaining;
    remaining = actual;
    while(remaining > 0) {
      size = (remaining > sizeof(dummy) ? sizeof(dummy) : remaining);
      status = read(assigned_fd,dummy,size);
      replay_syscalls++;
      if (status <= 0)
	break;
      remaining -= size;
    }
    status = actual - remaining;
  }
  this->fd_table.fd[fd].offset += status;
  if (status != actual) {
    if (errno != EISDIR || (this->debug & DEBUG_WARNING) != 0) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: read() read returns %d should be %d, errno %d\n",
	    status,actual,errno);
    return -1;
    }
  }
  return 0;
}

int 
replay_write(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int fd;
  int assigned_fd;
  int length;
  int actual;
  int offset;
  int cmd;
  int status;
  long current_offset;

  fd = msgp->fd;
  cmd = msgp->cmd;
  length = msgp->arg.io_arg.length;
  offset = msgp->arg.io_arg.offset;
  if (cmd == IOD_WRITEA || cmd == IOD_WRITELA) {
    actual = length;
  } else {
    actual = msgp->arg.io_arg.actual; 
  }
  if (actual < 0)
    return 0;
  assigned_fd = replay_actual_fd(this,fd);
  if (assigned_fd == -1) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: write() fd not inuse\n");
    return -1;
  }
  if (!(this->fd_table.fd[fd].flags & FLAGS_PIPE) &&
      !(this->fd_table.fd[fd].flags & FLAGS_CONSOLE)) {
#ifdef CHECK_OFFSET
    current_offset = tell(assigned_fd);
    replay_syscalls++;
#else
    current_offset = this->fd_table.fd[fd].offset;
#endif /* CHECK_OFFSET */
    if (current_offset != offset && 
        !(this->fd_table.fd[fd].flags & FLAGS_CONSOLE)) {
      status = lseek(assigned_fd,offset,SEEK_SET);
      replay_syscalls++;
      if (status != offset) {
        fdump_iod_msg(stderr,msgp);
        fprintf(stderr,"ERROR: write() lseek(%d,%d) returns %d, errno %d\n",
		assigned_fd,offset,status,errno);
        return -1;
      }
    }
  }
/*
  if (actual > sizeof(dummy)) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: dummy too small\n");
    return -1;
  }
 */
  if (this->fd_table.fd[fd].flags & FLAGS_CONSOLE) {
    if (replay_noprint) {
      return 0;
    }
    if (actual > 0) {
      int len;
      char * p;
      len = strlen((char *)this->process_name);
      strcpy(dummy,(char *)this->process_name);
      if (actual > len) {
        p = dummy + len;
	len = actual-len-1;
        while(len-- > 0)
          *p++ = ' ';
      }
      dummy[actual-1] = '\n';
    }
  }
  if (replay_mode == MODE_SINGLE && 
      (this->fd_table.fd[fd].flags & (FLAGS_PIPE | FLAGS_FAKE))) 
    status = actual;
  else if (actual <= sizeof(dummy)) {
    filldummy(actual);
    status = write(assigned_fd,dummy,actual);
    replay_syscalls++;
  } else {
    int size; 
    int remaining; 
    remaining = actual;
    while(remaining > 0) { 
      size = (remaining > sizeof(dummy) ? sizeof(dummy) : remaining); 
      filldummy(size);
      status = write(assigned_fd,dummy,size); 
      replay_syscalls++;
      if (status <= 0) 
        break;   
      remaining -= size; 
    }    
    status = actual - remaining; 
  } 
  this->fd_table.fd[fd].offset += status;
  if (status != actual) {
    fdump_iod_msg(stderr,msgp);
    fprintf(stderr,"ERROR: write() write(%d,%d) returns %d should be %d, errno %d\n",
	    assigned_fd,actual,status,actual,errno);
    return -1;
  }
  return 0;
}

int
replay_exit(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
/*
  exit(0);
*/
  return 0;
}

int 
replay_unlink(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int status;
  char filename[25];
  char *realname;

  strncpy(filename,msgp->arg.string,24);
  filename[24] = 0;
  realname = filename_map_table_find(filename_map_table,filename);
  status = unlink(realname);
  replay_syscalls++;
  return 0;
}

int 
replay_wait(
  struct replay_t *this,
  struct iod_msg *msgp
)
{
  int dummy;
  int options = 0;
  int replay_pid;
  int replay_precache = 0;
  int result;
  int wait_pid;
  struct replay_t *wait_replay;
  wait_pid = msgp->arg.wait_arg.result;
  if (wait_pid == -1)
    return 0;
  if (replay_mode == MODE_SINGLE) {
    wait_replay = pid_map_table_find(pid_map_table,wait_pid);
    if (wait_replay == 0) {
      fdump_iod_msg(stderr,msgp);
      fprintf(stderr,"ERROR: wait() unknown pid\n");
    }
    fd_info_table_close_on_exit(&(wait_replay->fd_table));
    replay_dtor(wait_replay);
    free(wait_replay);
  } else {
    replay_pid = pid_info_table_find(&(this->pid_table),wait_pid);
    if (replay_pid == -1) {
      fdump_iod_msg(stderr,msgp);
      fprintf(stderr,"ERROR: wait() unknown pid\n");
      return -1;
    }
    result = waitpid(replay_pid,&dummy,0);
    replay_syscalls++;
    if (result != replay_pid) {
      fdump_iod_msg(stderr,msgp);
      fprintf(stderr,"ERROR: wait() return %d, errno %d\n",result,errno);
      return -1;
    }
  }
  return 0;
}

/*
 * INITIALIZATION
 */

int 
usage()
{
    fprintf(stdout,"%s -logfile filename -hz hz -subfile subfile -logpath logpath -startpid pid -debug debug -precache -nostat -noprint -output output -delay -nodelay -hzcount loopcount\n", replay_progname);
}

int
init_params(argc, argv)
int argc;
char *argv[];
{
  int i;
  char *key;
  extern int hz;
  char buf[MAXSTRLEN];
  int startpid;
  FILE * subfile;
  char * subfilename;
  struct stat stat_buf;

  replay_progname = argv[0];
  if (argc <= 1) {
    usage();
    exit(0);
  }

  i = 1;
  hz = 60;
  replay_filename = 0;
  subfilename = "FilenameMap";
  subfilename = 0;
  replay_hzcount = ONE_HZ;
  replay_logpath = ".";
  replay_debug = 0;
  replay_o_flag = 0;
  replay_nostat = 0;
  replay_noprint = 0;
  replay_delay = 1;
  replay_output = 0;
  replay_stdout = 1;
  replay_stderr = 2;
#ifdef CHOICES
  replay_instdisk = 0;
  replay_inst = 0;
#endif
  startpid = 0;
  while(i < argc) {
    key = argv[i];
    if (strcmp(key,"-logfile") == 0) {
      i++;
      replay_filename = argv[i];
      i++;
      continue;
    }
    if (strcmp(key,"-hz") == 0) {
      i++;
      hz = atoi(argv[i]);  
      i++;
      continue;
    } 
    if (strcmp(key,"-startpid") == 0) {
      i++;
      startpid = atoi(argv[i]);  
      i++;
      continue;
    } 
    if (strcmp(key,"-subfile") == 0) {
      i++;
      subfilename = argv[i];  
      i++;
      continue;
    } 
    if (strcmp(key,"-logpath") == 0) {
      i++;
      replay_logpath = argv[i];
      i++;
      continue;
    }
    if (strcmp(key,"-debug") == 0) {
      i++;
      replay_debug = atoi(argv[i]);
      i++;
      continue;
    }
    if (strcmp(key,"-output") == 0) {
      i++;
      replay_output = argv[i];
      i++;
      continue;
    }
    if (strcmp(key,"-precache") == 0) {
      replay_precache = 1;
      i++;
      continue;
    }
    if (strcmp(key,"-nostat") == 0) {
      replay_nostat = 1;
      i++;
      continue;
    }
    if (strcmp(key,"-noprint") == 0) {
      replay_noprint = 1;
      i++;
      continue;
    }
    if (strcmp(key,"-delay") == 0) {
      replay_delay = 1;
      i++;
      continue;
    }
    if (strcmp(key,"-nodelay") == 0) {
      replay_delay = 0;
      i++;
      continue;
    }
    if (strcmp(key,"-hzcount") == 0) {
      i++;
      replay_hzcount = atoi(argv[i]);
      i++;
      continue;
    }

#ifdef CHOICES
    if (strcmp(key,"-O_LCACHE") == 0) {
      replay_o_flag |= 0x08000000;
      i++;
      continue;
    }
    if (strcmp(key,"-instdisk") == 0) {
      replay_instdisk = 1;
      i++;
      continue;
    }
    if (strcmp(key,"-inst") == 0) {
      replay_inst = 1;
      i++;
      continue;
    }
#endif
    usage();
    exit(1);
  }
  if (hz == 0) {
    fprintf(stderr,"ERROR: hz != 0\n",hz);
    exit(1);
  }

  if (subfilename != 0) {
    fprintf(stdout,"Processing filename substitutions (%s)\n",subfilename);
    subfile = fopen(subfilename,"r");
    if (subfile == 0) {
      fprintf(stderr,"ERROR: cannot open %s\n",subfilename);
      exit(1);
    }
    while(fgets(buf,sizeof(buf),subfile) != 0) {
      filename_map_table_add(filename_map_table,buf);
    }
    fclose(subfile);
  }

  if (stat(replay_logpath,&stat_buf) != 0) {
    fprintf(stderr,"ERROR: %s does not exist\n",replay_logpath);
    exit(1);
  }

  if (replay_filename != 0) {
    replay_mode = MODE_SINGLE;
  } else {
    replay_mode = MODE_MULTI;
  }

  if (replay_precache && replay_filename == 0) {
    fprintf(stderr,"ERROR: -precache requires -logfile\n");
    exit(1);
  }

  if (replay_output != 0) {
    int fd = open(replay_output, O_WRONLY | O_CREAT | O_TRUNC, 0777);
    if (fd < 0) {
      fprintf(stderr,"ERROR: cannot open output file %s errno %d\n",
	      replay_output, errno);
      fprintf(stderr,"       using stdin and stdout for output\n");
    } else {
      replay_stderr = fd;
      replay_stdout = dup(fd);
    }
  }

  return startpid;
}

int
main(argc, argv)
int argc;
char *argv[];
{
  int start_pid;
  int status;
  struct iod_msg msg;
  struct replay_t _replay_info;
  struct replay_t *replay_info;

  time_start(0);
  replay_cache = 0;
  replay_syscalls = 0;
  symtable_init();
  filename_map_table_ctor(filename_map_table);
  start_pid = init_params(argc, argv);
  replay_info = &_replay_info;
  replay_ctor(replay_info);
  replay_info->debug = replay_debug;
  replay_info->pid = start_pid;
  replay_open_log(replay_info,replay_filename);
  pid_map_table_add(pid_map_table,start_pid,replay_info);
  time_done(0);
  time_start(1);
  status = replay_loop(replay_info,&msg);
  if (status == 0) {
    fprintf(stdout,"successful completion\n");
  }
  time_done(1);
  exit(status);
}

#include <sys/time.h>
#include <sys/resource.h>

extern void inst( int argc, char * argv[] );

struct timezone time_zone;
struct timeval start_timeval;
struct timeval done_timeval;
struct rusage start_usage;
struct rusage done_usage;
struct rusage total_usage;

int subtract_timeval(struct timeval *out, struct timeval *in1, 
		     struct timeval *in2)
{
  out->tv_sec = in1->tv_sec - in2->tv_sec;
  if (in2->tv_usec > in1->tv_usec) {
    out->tv_sec--;
    out->tv_usec = in1->tv_usec + 1000000 - in2->tv_usec;
  } else {
    out->tv_usec = in1->tv_usec - in2->tv_usec;
  }
}

int time_start(int real)
{
  int status;

#ifdef CHOICES
  static char * reset_disk[] = { "inst", "k", "DiskTimer", "r" };
  static char * on_disk[] = { "inst", "k", "DiskTimer", "on" };
  static char * reset_all[] = { "inst", "k", "SimpleInstrument", "r" };
  static char * on_all[] = { "inst", "k", "SimpleInstrument", "on" };
  if ( replay_instdisk ) {
    inst( 4, on_disk );
    inst( 4, reset_disk );
  }
  if ( replay_inst ) {
    inst( 4, on_all );
    inst( 4, reset_all );
  }
#endif // CHOICES

  status = gettimeofday(&start_timeval, &time_zone);
  status = getrusage(RUSAGE_SELF, &start_usage);
  return 0;
}

int time_done(int real)
{
  int status;
  struct timeval wall_clock_time;

#ifdef CHOICES
  static char * dpy_disk[] = { "inst", "k", "DiskTimer", "d" };
  static char * dpy_all[] = { "inst", "k", "SimpleInstrument", "d" };
#endif 

  status = getrusage(RUSAGE_SELF, &done_usage);
  status = gettimeofday(&done_timeval, &time_zone);
#ifdef CHOICES
  if ( real ) {
    if ( replay_instdisk ) {
      inst( 4, dpy_disk );
    }
    if ( replay_inst ) {
      inst( 4, dpy_all );
    }
  }
#endif

  subtract_timeval(&wall_clock_time, &done_timeval, &start_timeval);
  subtract_timeval(&(total_usage.ru_utime), 
		   &(done_usage.ru_utime), &(start_usage.ru_utime));
  subtract_timeval(&(total_usage.ru_stime), 
		   &(done_usage.ru_stime), &(start_usage.ru_stime));
  subtract_timeval(&wall_clock_time, &done_timeval, &start_timeval);
  total_usage.ru_minflt = done_usage.ru_minflt - start_usage.ru_minflt;
  total_usage.ru_majflt = done_usage.ru_majflt - start_usage.ru_majflt;
  total_usage.ru_nswap = done_usage.ru_nswap - start_usage.ru_nswap;
  total_usage.ru_inblock = done_usage.ru_inblock - start_usage.ru_inblock;
  total_usage.ru_oublock = done_usage.ru_oublock - start_usage.ru_oublock;
  total_usage.ru_msgsnd = done_usage.ru_msgsnd - start_usage.ru_msgsnd;
  total_usage.ru_msgrcv = done_usage.ru_msgrcv - start_usage.ru_msgrcv;
  total_usage.ru_nvcsw = done_usage.ru_nvcsw - start_usage.ru_nvcsw;
  total_usage.ru_nivcsw = done_usage.ru_nivcsw - start_usage.ru_nivcsw;

  printf("Wall Clock Time: %ld sec, %ld usec\n",
	 wall_clock_time.tv_sec, wall_clock_time.tv_usec);
  if ( real ) {
    printf("Resource Usage :\n");
    printf("  user time     : %ld sec %ld usec\n", 
  	   total_usage.ru_utime.tv_sec, total_usage.ru_utime.tv_usec);
    printf("  system time   : %ld sec %ld usec\n", 
  	   total_usage.ru_stime.tv_sec, total_usage.ru_stime.tv_usec);
    printf("  minor faults  : %d\n", total_usage.ru_minflt);
    printf("  major faults  : %d\n", total_usage.ru_majflt);
    printf("  swaps         : %d\n", total_usage.ru_nswap);
    printf("  in blocks     : %d\n", total_usage.ru_inblock);
    printf("  out blocks    : %d\n", total_usage.ru_oublock);
    printf("  msg sends     : %d\n", total_usage.ru_msgsnd);
    printf("  msg receives  : %d\n", total_usage.ru_msgrcv);
    printf("  volunt. csw   : %d\n", total_usage.ru_nvcsw);
    printf("  involunt. csw : %d\n", total_usage.ru_nivcsw);
    printf("Replay syscalls : %d\n", replay_syscalls);
  }
  return 0;
}


